/************************************************************************

The PLANETARY SOLUTION VSOP87 by Bretagnon P. and Francou G. can be found at
ftp://ftp.imcce.fr/pub/ephem/planets/vsop87

I (Johannes Gajdosik) have just taken the data obtained from above
(VSOP87.mer,...,VSOP87.nep) and rearranged it into this piece of software.

I can neigther allow nor forbid the usage of VSOP87.
The copyright notice below covers not the work of Bretagnon P. and Francou G.
but just my work, that is the compilation of the VSOP87 data
into the software supplied in this file.


Copyright (c) 2006 Johannes Gajdosik

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

This is my implementation of the VSOP87 planetary solution.
I tried to optimize for speed by rearranging the terms so that
for a given touple of (a[0],a[1],...,a[11]) the values
  (cos,sin)(a[0]*lambda[0](T)+...a[11]*lambda[11](T))
have only to be calculated once.
Furthermore I used the addition formulas
  (cos,sin)(x+y) = ...
so that for given T the functions cos and sin have only to be called 12 times.

****************************************************************/

#include "vsop87.h"
#include "elliptic_to_rectangular.h"

#include <string.h>
#include <math.h>

static
const int vsop87_max_lambda_factor[12] = {
  16,
  27,
  27,
  32,
  22,
  26,
  24,
  23,
  2,
  1,
  2,
  1,
};

static
const double lambda_0[12] = {
  4.40260884240,
  3.17614669689,
  1.75347045953,
  6.20347611291,
  0.59954649739,
  0.87401675650,
  5.48129387159,
  5.31188628676,
  5.19846674103,
  1.62790523337,
  2.35555589827,
  3.81034454697,
};

static
const double lambda_1[12] = {
  26087.9031415742,
  10213.2855462110,
   6283.0758499914,
   3340.6124266998,
    529.6909650946,
    213.2990954380,
     74.7815985673,
     38.1330356378,
  77713.7714681205,
  84334.6615813083,
  83286.9142695536,
  83997.0911355954,
};


  /* taken out from vsop87_constants: */
static
const double vsop87_l[8] = {
   26087.90314157420,
   10213.28554621100,
    6283.07584999140,
    3340.61242669981,
     529.69096509460,
     213.29909543800,
      74.78159856730,
      38.13303563780,
};

static
const double vsop87_constants[253] = {
       0.38709830982,
       0.00000000000,
       0.00000000000,
       4.40260884240,
       0.0,
      -0.00000934669,
       0.00000003100,
       0.04466059760,
      -0.00552114624,
      -0.00001860568,
       0.00000079123,
       0.00000000590,
      -0.00000000016,
       0.20072331368,
       0.00143750118,
      -0.00007974120,
      -0.00000030459,
       0.00000000812,
      -0.00000000008,
       0.04061563384,
       0.00065433117,
      -0.00001071215,
       0.00000022459,
      -0.00000000377,
      -0.00000000003,
       0.04563550461,
      -0.00127633657,
      -0.00000913350,
       0.00000018993,
      -0.00000000640,
      -0.00000000002,
       0.72332981996,
       0.00000000000,
       0.00000000000,
       3.17614669689,
       0.0,
       0.00000287889,
      -0.00000003038,
      -0.00449282133,
       0.00031259019,
       0.00000604065,
      -0.00000068351,
       0.00000000494,
       0.00000000060,
       0.00506684726,
      -0.00036121239,
       0.00001846764,
       0.00000003281,
      -0.00000000614,
      -0.00000000017,
       0.00682410142,
       0.00138133826,
      -0.00001090942,
      -0.00000186418,
       0.00000000602,
       0.00000000075,
       0.02882285775,
      -0.00040384791,
      -0.00006232891,
       0.00000024731,
       0.00000004229,
      -0.00000000006,
       1.00000101778,
       0.00000000000,
       0.00000000000,
       1.75347045953,
       0.0,
      -0.00000991013,
      -0.00000002537,
       0.00000000000,
       0.00000000000,
      -0.00374081650,
      -0.00082267418,
       0.00002762465,
       0.00000116955,
      -0.00000002696,
      -0.00000000071,
       0.01628447663,
      -0.00062029655,
      -0.00003382631,
       0.00000085101,
       0.00000002770,
      -0.00000000047,
       0.00000000000,
      -0.00113468869,
       0.00001237314,
       0.00000126542,
      -0.00000001372,
      -0.00000000032,
       0.00000000000,
       0.00010180375,
       0.00004701998,
      -0.00000054174,
      -0.00000002508,
       0.00000000046,
       1.52367934191,
       0.00000000000,
       0.00000000000,
       6.20347611291,
       0.0,
       0.00000457003,
      -0.00000005057,
       0.00000000000,
       0.00000000000,
       0.08536560252,
       0.00376330152,
      -0.00024657776,
      -0.00000367305,
       0.00000011112,
       0.00000000026,
      -0.03789973236,
       0.00624657465,
       0.00015529482,
      -0.00000634879,
      -0.00000006593,
       0.00000000073,
       0.01047042574,
       0.00017138526,
      -0.00004077491,
      -0.00000138835,
       0.00000000916,
       0.00000000176,
       0.01228449307,
      -0.00108020083,
      -0.00001922224,
       0.00000087186,
       0.00000003090,
       5.20260319132,
       0.00000191323,
       0.00000000000,
       0.00000000000,
       0.00000000000,
       0.00000000000,
       0.59954649739,
       0.0,
      -0.00014837133,
       0.00000007482,
       0.00000000000,
       0.00000000000,
       0.04698572124,
       0.00113010377,
      -0.00010930126,
      -0.00000428748,
       0.00000020539,
       0.01200385748,
       0.00217149360,
       0.00009858539,
      -0.00000513109,
      -0.00000009007,
      -0.00206561098,
      -0.00031340156,
      -0.00001667392,
       0.00000076926,
       0.01118377157,
      -0.00023427562,
       0.00002086760,
       0.00000050721,
       9.55490959574,
      -0.00002138957,
       0.00000000000,
       0.00000000000,
       0.00000000000,
       0.00000000000,
       0.87401675650,
       0.0,
       0.00036659741,
      -0.00000033330,
       0.00000000217,
       0.00000000000,
      -0.00296003595,
      -0.00529602626,
       0.00030928405,
       0.00001296215,
      -0.00000059959,
       0.00000000000,
       0.05542964254,
      -0.00375593887,
      -0.00031990236,
       0.00001598633,
       0.00000032451,
       0.00000000000,
      -0.00871747436,
       0.00080171499,
       0.00004142282,
      -0.00000196049,
      -0.00000009439,
       0.01989147301,
       0.00059439766,
      -0.00005235117,
      -0.00000127219,
       0.00000008295,
      19.21844606178,
      -0.00000037163,
       0.00000009791,
       0.00000000000,
       0.00000000000,
       0.00000000000,
       5.48129387159,
       0.0,
      -0.00000848828,
       0.00000010450,
       0.00000000000,
       0.00000000000,
      -0.04595132376,
       0.00018344115,
      -0.00000080849,
      -0.00000045396,
       0.00000002185,
       0.00563791307,
      -0.00074964350,
       0.00001210200,
      -0.00000042088,
      -0.00000001714,
       0.00185915075,
      -0.00012449382,
      -0.00000207373,
       0.00000007621,
       0.00648617008,
      -0.00011744733,
       0.00000317799,
       0.00000007317,
      30.11038686942,
      -0.00000166346,
       0.00000006857,
       0.00000000000,
       0.00000000000,
       0.00000000000,
       5.31188628676,
       0.0,
       0.00000102311,
      -0.00000004340,
       0.00000000000,
       0.00000000000,
       0.00599977571,
       0.00000871304,
      -0.00000119902,
      -0.00000004034,
       0.00000000000,
       0.00000000000,
       0.00669242413,
       0.00007824336,
       0.00000080801,
      -0.00000003955,
       0.00000000000,
       0.00000000000,
      -0.01029147819,
      -0.00000072727,
      -0.00000006568,
       0.00000001668,
       0.01151683985,
       0.00002575536,
       0.00000019377,
       0.00000001331,
};

static
const double vsop87_coefficients[61126*2] = {
      -0.00000002893,    -0.00000011799,
       0.00000000332,    -0.00000000078,
       0.00000000001,     0.00000000003,
      -0.00000091933,     0.00000023240,
      -0.00000000656,    -0.00000002636,
       0.00000000033,    -0.00000000009,
      -0.00000063987,     0.00000001677,
       0.00000000006,    -0.00000000055,
       0.00000001754,    -0.00000064219,
      -0.00000000058,    -0.00000000056,
      -0.00000001018,     0.00000001921,
       0.00000000087,     0.00000000151,
      -0.00000004656,    -0.00000001902,
      -0.00000000125,     0.00000000062,
       0.00000002356,    -0.00000000340,
       0.00000034105,    -0.00000000214,
      -0.00000000097,     0.00000033741,
       0.00000000025,    -0.00000000050,
       0.00000000125,     0.00000000028,
       0.00000001423,     0.00000000081,
      -0.00000000035,     0.00000000059,
       0.00000039352,    -0.00000000242,
      -0.00000000244,     0.00000038921,
      -0.00000002629,    -0.00000006147,
       0.00000000453,    -0.00000000270,
       0.00000048318,    -0.00000000210,
      -0.00000000203,     0.00000047817,
      -0.00000000124,    -0.00000000139,
       0.00000000059,     0.00000000064,
       0.00000001642,    -0.00000006257,
       0.00000000289,    -0.00000000150,
       0.00000089329,    -0.00000000565,
      -0.00000000372,     0.00000088559,
      -0.00000000180,    -0.00000000143,
       0.00000010075,     0.00000000464,
      -0.00000000678,     0.00000000954,
       0.00000120906,    -0.00000000875,
      -0.00000000297,     0.00000120136,
      -0.00000000332,    -0.00000000171,
       0.00000001393,     0.00000011733,
      -0.00000000173,    -0.00000000044,
       0.00000171894,    -0.00000000976,
      -0.00000000859,     0.00000170186,
       0.00000000307,     0.00000000307,
       0.00000002168,    -0.00000001934,
       0.00000214912,    -0.00000001640,
      -0.00000000844,     0.00000213398,
      -0.00000000066,    -0.00000000043,
      -0.00000000073,    -0.00000000049,
      -0.00000000005,    -0.00000000003,
      -0.00000000003,     0.00000000005,
      -0.00000000007,     0.00000000009,
      -0.00000000007,     0.00000000027,
       0.00000000005,     0.00000000001,
       0.00000000028,     0.00000000006,
       0.00000000001,    -0.00000000005,
       0.00000000002,     0.00000000004,
       0.00000000005,     0.00000000003,
      -0.00000000001,     0.00000000004,
       0.00000000054,     0.00000000015,
       0.00000000002,    -0.00000000011,
       0.00000000142,    -0.00000000013,
      -0.00000000002,    -0.00000000022,
      -0.00000000014,    -0.00000000143,
      -0.00000000023,     0.00000000002,
       0.00000000011,     0.00000000010,
       0.00000000013,    -0.00000000012,
       0.00000000020,     0.00000000001,
       0.00000000023,    -0.00000000002,
       0.00000000000,    -0.00000000004,
      -0.00000000020,    -0.00000000285,
      -0.00000000045,     0.00000000005,
      -0.00000000297,    -0.00000000677,
      -0.00000000091,     0.00000000040,
       0.00000000002,     0.00000000004,
      -0.00000000686,     0.00000000303,
       0.00000000040,     0.00000000092,
       0.00000000004,    -0.00000000002,
       0.00000000026,    -0.00000000060,
      -0.00000000005,    -0.00000000001,
      -0.00000000070,    -0.00000000042,
      -0.00000000003,     0.00000000007,
       0.00000000024,     0.00000000056,
       0.00000000057,    -0.00000000024,
       0.00000000002,    -0.00000000008,
      -0.00000000008,    -0.00000000002,
      -0.00000000005,     0.00000000000,
       0.00000000000,     0.00000000005,
       0.00000000006,    -0.00000000032,
       0.00000000000,     0.00000000005,
       0.00000000005,     0.00000000000,
       0.00000000014,     0.00000000000,
      -0.00000000063,    -0.00000000146,
      -0.00000000018,     0.00000000008,
      -0.00000001283,     0.00000000552,
       0.00000000078,     0.00000000171,
       0.00000000007,    -0.00000000003,
      -0.00000003012,     0.00000002660,
       0.00000000258,     0.00000000304,
       0.00000000012,    -0.00000000011,
       0.00000002703,     0.00000003054,
       0.00000000309,    -0.00000000262,
      -0.00000000011,    -0.00000000013,
      -0.00000000283,    -0.00000000033,
       0.00000000005,     0.00000000013,
      -0.00000000099,     0.00000000359,
       0.00000000023,     0.00000000001,
       0.00000000029,     0.00000000068,
       0.00000000072,    -0.00000000031,
       0.00000000250,    -0.00000000219,
      -0.00000000221,    -0.00000000259,
      -0.00000000003,     0.00000000011,
       0.00000000011,     0.00000000003,
       0.00000000001,    -0.00000000005,
       0.00000000048,     0.00000000012,
      -0.00000000033,    -0.00000000001,
      -0.00000000001,     0.00000000034,
       0.00000000026,     0.00000000003,
       0.00000002208,    -0.00000017037,
       0.00000000500,     0.00000000347,
       0.00000000001,    -0.00000000093,
      -0.00000000091,     0.00000000002,
       0.00000000001,    -0.00000000009,
      -0.00000000224,     0.00000001642,
      -0.00000000048,    -0.00000000042,
       0.00000000084,    -0.00000000576,
      -0.00000000969,     0.00000000846,
       0.00000000083,     0.00000000098,
       0.00000000004,    -0.00000000004,
       0.00000003239,     0.00000003521,
       0.00000000345,    -0.00000000319,
      -0.00000000012,    -0.00000000012,
       0.00000020786,     0.00000012709,
       0.00000000881,    -0.00000001318,
      -0.00000000038,    -0.00000000023,
       0.00000012949,    -0.00000021142,
      -0.00000001346,    -0.00000000909,
      -0.00000000025,     0.00000000039,
       0.00000000221,     0.00000001231,
       0.00000000006,    -0.00000000030,
       0.00000001794,     0.00000000034,
      -0.00000000018,    -0.00000000061,
       0.00000000359,    -0.00000000314,
      -0.00000000029,    -0.00000000034,
      -0.00000000185,    -0.00000000203,
      -0.00000001743,    -0.00000001067,
      -0.00000000069,     0.00000000104,
      -0.00000001115,     0.00000001770,
       0.00000000107,     0.00000000074,
      -0.00000000003,    -0.00000000068,
      -0.00000000083,    -0.00000000006,
      -0.00000000004,     0.00000000017,
      -0.00000000696,    -0.00000000163,
       0.00000000000,     0.00000000018,
       0.00000000109,     0.00000000000,
       0.00000000000,    -0.00000000112,
       0.00000000005,     0.00000000000,
       0.00000000015,    -0.00000000159,
       0.00000000003,     0.00000000002,
       0.00000000001,    -0.00000000015,
      -0.00000000011,     0.00000000002,
       0.00000000007,    -0.00000000004,
       0.00000016706,     0.00000010365,
       0.00000000718,    -0.00000001078,
      -0.00000000031,    -0.00000000019,
      -0.00000146974,     0.00000231921,
       0.00000015141,     0.00000010270,
       0.00000000280,    -0.00000000452,
       0.00000064161,    -0.00000238769,
      -0.00000007418,    -0.00000002531,
      -0.00000000018,     0.00000000119,
      -0.00000243967,    -0.00000066559,
      -0.00000002771,     0.00000007730,
       0.00000000132,     0.00000000035,
       0.00000006138,    -0.00000003078,
      -0.00000000053,     0.00000000274,
      -0.00000000003,    -0.00000000007,
      -0.00000002102,    -0.00000012748,
      -0.00000000093,     0.00000000113,
      -0.00000000014,     0.00000000009,
      -0.00000004647,    -0.00000002879,
      -0.00000000200,     0.00000000299,
       0.00000000009,     0.00000000005,
       0.00000007314,    -0.00000011549,
      -0.00000000753,    -0.00000000511,
      -0.00000004963,     0.00000020844,
       0.00000000598,     0.00000000171,
       0.00000021294,     0.00000005391,
       0.00000000212,    -0.00000000638,
      -0.00000000412,     0.00000000085,
       0.00000000044,     0.00000000587,
      -0.00000000002,     0.00000000000,
       0.00000000000,     0.00000000044,
      -0.00000000010,    -0.00000000002,
      -0.00000000009,     0.00000000004,
      -0.00000016521,     0.00000071340,
       0.00000002014,     0.00000000574,
       0.00000000005,    -0.00000000027,
       0.00001686982,     0.00000394035,
       0.00000013916,    -0.00000047867,
      -0.00000000613,    -0.00000000132,
       0.00000445156,    -0.00000003021,
       0.00000000012,     0.00000000363,
       0.00000000011,     0.00000000015,
       0.00000001223,    -0.00000463407,
      -0.00000000322,    -0.00000000275,
       0.00000000003,     0.00000000013,
      -0.00000001112,    -0.00000001817,
      -0.00000000122,     0.00000000203,
      -0.00000000006,    -0.00000000010,
       0.00000013525,    -0.00000000665,
      -0.00000000081,    -0.00000000152,
      -0.00000000012,    -0.00000000007,
       0.00000002806,    -0.00000013517,
      -0.00000000356,    -0.00000000086,
      -0.00000081288,    -0.00000019014,
      -0.00000000674,     0.00000002308,
       0.00000012697,     0.00000000229,
      -0.00000000427,    -0.00000011417,
       0.00000000043,     0.00000000029,
       0.00000000123,    -0.00000000225,
      -0.00000000624,    -0.00000000122,
       0.00000013008,     0.00000000057,
       0.00000000017,    -0.00000000006,
       0.00000000449,    -0.00000201760,
      -0.00000000058,    -0.00000000289,
       0.00000000004,     0.00000000005,
      -0.00000000260,    -0.00000018158,
       0.00000000039,    -0.00000000170,
      -0.00000000001,     0.00000000007,
      -0.00000002130,     0.00000005242,
       0.00000000042,    -0.00000000584,
      -0.00000000654,     0.00000001479,
      -0.00000000021,    -0.00000000047,
       0.00000000001,     0.00000000007,
       0.00000000357,    -0.00000005883,
      -0.00000000031,     0.00000000044,
       0.00000043444,    -0.00000000082,
       0.00000000015,     0.00000044908,
       0.00000000376,     0.00000000274,
      -0.00000000039,     0.00000000018,
       0.00000001221,    -0.00000000294,
       0.00000000528,     0.00000000186,
       0.00000000226,    -0.00000000238,
       0.00000000016,     0.00000000010,
      -0.00000000010,    -0.00000000003,
      -0.00000000003,     0.00000000010,
      -0.00000000018,     0.00000000032,
      -0.00000000003,     0.00000000012,
       0.00000000011,     0.00000000003,
       0.00000000020,     0.00000000005,
       0.00000000005,     0.00000000000,
       0.00000000000,    -0.00000000005,
       0.00000000004,     0.00000000007,
       0.00000000070,     0.00000000041,
      -0.00000000005,     0.00000000006,
       0.00000000421,    -0.00000000815,
       0.00000000081,     0.00000000065,
       0.00000001241,    -0.00000000892,
       0.00000000057,     0.00000000089,
       0.00000000838,     0.00000001262,
      -0.00000000087,     0.00000000049,
       0.00000000069,    -0.00000000415,
       0.00000000005,     0.00000000010,
       0.00000000468,     0.00000000042,
      -0.00000000009,     0.00000000010,
      -0.00000000021,     0.00000000043,
      -0.00000000046,     0.00000000041,
      -0.00000000040,    -0.00000000046,
       0.00000000001,    -0.00000000011,
       0.00000000002,     0.00000000000,
      -0.00000000156,    -0.00000000006,
       0.00000000006,    -0.00000000021,
      -0.00000000245,    -0.00000000153,
       0.00000000018,    -0.00000000023,
       0.00000000160,    -0.00000000239,
       0.00000000022,     0.00000000019,
      -0.00000000082,     0.00000000014,
       0.00000000001,    -0.00000000004,
      -0.00000000023,    -0.00000000089,
       0.00000000006,     0.00000000000,
      -0.00000000032,    -0.00000000098,
      -0.00000000002,     0.00000000000,
       0.00000000011,     0.00000000026,
      -0.00000000005,     0.00000000001,
      -0.00000000009,     0.00000000055,
      -0.00000000007,    -0.00000000002,
      -0.00000000054,    -0.00000000011,
       0.00000000003,    -0.00000000007,
       0.00000000008,     0.00000000015,
      -0.00000000016,     0.00000000010,
       0.00000000004,    -0.00000000004,
       0.00000000011,    -0.00000000002,
       0.00000000002,     0.00000000011,
       0.00000000003,     0.00000000002,
       0.00000000002,     0.00000000005,
      -0.00000000003,    -0.00000000004,
      -0.00000000004,     0.00000000003,
       0.00000000016,    -0.00000000016,
       0.00000000003,    -0.00000000005,
      -0.00000000005,    -0.00000000003,
       0.00000000005,     0.00000000000,
       0.00000000001,     0.00000000001,
       0.00000037671,    -0.00000029029,
       0.00000007389,     0.00000010232,
       0.00000000021,    -0.00000000019,
      -0.00000000020,    -0.00000000021,
      -0.00000003026,     0.00000002275,
      -0.00000000612,    -0.00000000742,
       0.00000013396,    -0.00000010090,
       0.00000002593,     0.00000003165,
       0.00000000008,     0.00000000001,
       0.00000000003,     0.00000000002,
       0.00000000004,     0.00000000003,
       0.00000000016,     0.00000000004,
       0.00000000004,    -0.00000000016,
       0.00000000005,     0.00000000001,
       0.00000000004,    -0.00000000030,
      -0.00000000023,    -0.00000000153,
      -0.00000000012,    -0.00000000001,
      -0.00000000154,     0.00000000023,
      -0.00000000001,     0.00000000012,
       0.00000000003,    -0.00000000015,
      -0.00000000017,    -0.00000000006,
      -0.00000000010,    -0.00000000073,
      -0.00000000005,     0.00000000000,
       0.00000000457,    -0.00000000010,
       0.00000000012,    -0.00000000039,
      -0.00000001699,     0.00000000989,
       0.00000000042,     0.00000000124,
       0.00000001001,     0.00000001720,
       0.00000000128,    -0.00000000042,
      -0.00000000166,     0.00000000028,
       0.00000000005,     0.00000000207,
       0.00000000001,     0.00000000005,
       0.00000000136,    -0.00000000078,
      -0.00000000078,    -0.00000000138,
       0.00000002569,    -0.00000001468,
      -0.00000000060,    -0.00000000180,
      -0.00000000008,    -0.00000000003,
      -0.00000171761,    -0.00000297696,
      -0.00000020815,     0.00000007247,
      -0.00000000284,     0.00000000933,
      -0.00000047216,    -0.00000037766,
      -0.00000002269,     0.00000001739,
      -0.00000000007,     0.00000000093,
      -0.00000038482,     0.00000047909,
       0.00000001759,     0.00000002397,
       0.00000000098,     0.00000000008,
      -0.00000002101,    -0.00000003442,
       0.00000000055,    -0.00000000091,
       0.00000000001,     0.00000000004,
      -0.00000004920,     0.00000001886,
      -0.00000000122,    -0.00000000015,
      -0.00000000902,     0.00000000516,
       0.00000000021,     0.00000000063,
       0.00000005863,     0.00000010166,
       0.00000000733,    -0.00000000238,
       0.00000003610,     0.00000003135,
       0.00000000185,    -0.00000000104,
       0.00000003208,    -0.00000003629,
      -0.00000000105,    -0.00000000191,
       0.00000000051,     0.00000000128,
       0.00000000142,    -0.00000000047,
      -0.00000002090,    -0.00000001846,
      -0.00000000118,     0.00000000069,
      -0.00000000001,     0.00000000005,
      -0.00000039197,     0.00000046999,
       0.00000001701,     0.00000002406,
       0.00000000110,     0.00000000007,
      -0.00000012312,     0.00000037424,
       0.00000000836,     0.00000000583,
       0.00000000018,     0.00000000004,
       0.00000038180,     0.00000012845,
       0.00000000677,    -0.00000000874,
       0.00000000001,    -0.00000000020,
      -0.00000000936,     0.00000001288,
      -0.00000000068,    -0.00000000053,
       0.00000001496,     0.00000001958,
      -0.00000000044,     0.00000000083,
       0.00000000525,     0.00000000498,
       0.00000000032,    -0.00000000013,
       0.00000001330,    -0.00000001591,
      -0.00000000049,    -0.00000000076,
       0.00000000728,    -0.00000004191,
      -0.00000000118,    -0.00000000003,
      -0.00000004223,    -0.00000000771,
      -0.00000000006,     0.00000000119,
       0.00000000045,    -0.00000000033,
      -0.00000000034,    -0.00000000054,
       0.00000000002,    -0.00000000016,
      -0.00000002929,     0.00000012451,
       0.00000000338,     0.00000000092,
       0.00000000002,    -0.00000000004,
       0.00000155661,     0.00000037045,
       0.00000001172,    -0.00000004206,
      -0.00000000050,    -0.00000000023,
       0.00000077060,    -0.00000000470,
      -0.00000000113,     0.00000000109,
       0.00000000006,    -0.00000000002,
       0.00000000788,    -0.00000081360,
      -0.00000000146,    -0.00000000051,
      -0.00000000538,    -0.00000000701,
      -0.00000000002,     0.00000000037,
       0.00000001609,    -0.00000000928,
       0.00000000077,     0.00000000023,
       0.00000000226,    -0.00000003441,
      -0.00000000117,     0.00000000033,
      -0.00000004831,    -0.00000001230,
      -0.00000000052,     0.00000000133,
       0.00000018697,     0.00000000024,
      -0.00000000051,    -0.00000018586,
       0.00000000034,    -0.00000000030,
      -0.00000000050,    -0.00000000025,
       0.00000000036,    -0.00000000063,
      -0.00000000013,    -0.00000000009,
       0.00000003048,     0.00000000041,
      -0.00000000003,     0.00000000000,
       0.00000000379,    -0.00000040772,
       0.00000000001,     0.00000000035,
      -0.00000000098,    -0.00000003461,
       0.00000000037,     0.00000000009,
      -0.00000001177,     0.00000001064,
       0.00000000041,    -0.00000000077,
      -0.00000000121,     0.00000000965,
      -0.00000000016,    -0.00000000034,
       0.00000000054,    -0.00000001177,
       0.00000000003,    -0.00000000074,
       0.00000064299,    -0.00000000002,
      -0.00000000013,     0.00000047031,
       0.00000000137,    -0.00000000529,
       0.00000000050,    -0.00000000132,
       0.00000000657,     0.00000000701,
       0.00000000030,    -0.00000000021,
       0.00000000725,    -0.00000000619,
      -0.00000000022,    -0.00000000027,
      -0.00000000047,     0.00000000244,
       0.00000000027,     0.00000000001,
      -0.00000000001,    -0.00000000009,
      -0.00000000006,     0.00000000001,
       0.00000000004,    -0.00000000002,
       0.00000000033,     0.00000000015,
       0.00000000110,    -0.00000000287,
       0.00000000002,     0.00000000008,
       0.00000000424,    -0.00000000352,
       0.00000000004,     0.00000000018,
       0.00000000334,     0.00000000427,
      -0.00000000017,     0.00000000004,
       0.00000000096,    -0.00000000155,
       0.00000000168,     0.00000000094,
       0.00000000028,    -0.00000000029,
       0.00000000028,     0.00000000028,
       0.00000000000,    -0.00000000003,
      -0.00000000037,     0.00000000004,
      -0.00000000067,    -0.00000000027,
       0.00000000028,    -0.00000000065,
      -0.00000000020,     0.00000000000,
      -0.00000000001,    -0.00000000022,
       0.00000000002,     0.00000000004,
       0.00000000000,     0.00000000010,
      -0.00000000010,     0.00000000000,
       0.00000000005,     0.00000000000,
      -0.00000000005,     0.00000000001,
       0.00000000001,     0.00000000005,
      -0.00000000007,    -0.00000000022,
      -0.00000000003,    -0.00000000005,
      -0.00000000005,     0.00000000003,
       0.00000000039,    -0.00000000123,
      -0.00000000061,    -0.00000000007,
       0.00000000000,     0.00000000013,
      -0.00000000014,     0.00000000009,
       0.00000000005,     0.00000000005,
       0.00000000010,     0.00000000014,
       0.00000000005,    -0.00000000005,
      -0.00000000025,    -0.00000000045,
      -0.00000000016,     0.00000000013,
       0.00000000000,    -0.00000000018,
      -0.00000000006,     0.00000000001,
      -0.00000000019,     0.00000000000,
       0.00000000001,     0.00000000006,
       0.00000000000,    -0.00000000003,
      -0.00000000065,    -0.00000000015,
      -0.00000000002,     0.00000000021,
      -0.00000000030,    -0.00000000021,
      -0.00000000004,     0.00000000009,
      -0.00000000022,     0.00000000031,
       0.00000000009,     0.00000000004,
       0.00000000025,     0.00000000036,
       0.00000000037,    -0.00000000025,
      -0.00000000007,    -0.00000000007,
      -0.00000000002,     0.00000000002,
      -0.00000000133,     0.00000000094,
       0.00000000027,     0.00000000029,
       0.00000000002,    -0.00000000003,
      -0.00000000095,     0.00000000035,
       0.00000000010,     0.00000000017,
       0.00000000037,     0.00000000098,
       0.00000000018,    -0.00000000011,
      -0.00000000003,     0.00000000004,
       0.00000000031,     0.00000000057,
       0.00000000038,    -0.00000000026,
       0.00000000114,     0.00000000059,
       0.00000000058,    -0.00000000115,
      -0.00000000035,    -0.00000000002,
       0.00000000000,     0.00000000006,
      -0.00000000063,     0.00000000459,
       0.00000000079,     0.00000000002,
      -0.00000000076,     0.00000000279,
       0.00000000037,     0.00000000002,
       0.00000000292,     0.00000000080,
       0.00000000003,    -0.00000000040,
      -0.00000000006,    -0.00000000001,
       0.00000000002,     0.00000000011,
       0.00000000191,     0.00000000160,
       0.00000000036,    -0.00000000032,
       0.00000000023,    -0.00000000126,
       0.00000000452,     0.00000000164,
       0.00000000038,    -0.00000000053,
       0.00000000160,    -0.00000000451,
      -0.00000000053,    -0.00000000037,
      -0.00000000085,     0.00000000071,
       0.00000000008,     0.00000000008,
       0.00000000774,     0.00000000994,
       0.00000000090,    -0.00000000085,
      -0.00000000003,    -0.00000000003,
       0.00000000481,     0.00000000593,
       0.00000000031,    -0.00000000038,
       0.00000000635,    -0.00000000498,
      -0.00000000040,    -0.00000000036,
      -0.00000000006,     0.00000000003,
       0.00000000019,     0.00000000008,
       0.00000001147,     0.00000000762,
       0.00000000145,    -0.00000000142,
      -0.00000000010,    -0.00000000013,
      -0.00000000219,    -0.00000000239,
       0.00000002841,     0.00000000892,
       0.00000000113,    -0.00000000205,
       0.00000000879,    -0.00000002827,
      -0.00000000203,    -0.00000000110,
      -0.00000000037,     0.00000000252,
       0.00000000007,     0.00000000000,
       0.00000002540,     0.00000000312,
       0.00000000001,    -0.00000000075,
       0.00000001485,    -0.00000000068,
      -0.00000000009,    -0.00000001576,
      -0.00000000003,    -0.00000000003,
      -0.00000000008,    -0.00000000007,
       0.00000000019,    -0.00000000018,
       0.00000010520,     0.00000004122,
       0.00000000443,    -0.00000000762,
      -0.00000000031,    -0.00000000019,
      -0.00000001293,     0.00000001462,
       0.00000000112,     0.00000000084,
       0.00000019758,     0.00000000056,
       0.00000000006,    -0.00000019665,
       0.00000000016,     0.00000000037,
       0.00000000036,    -0.00000000014,
       0.00000000050,    -0.00000000017,
      -0.00000000155,    -0.00000000626,
      -0.00000000097,    -0.00000000045,
       0.00000000036,     0.00000000026,
      -0.00000000007,    -0.00000000001,
      -0.00000000002,     0.00000000012,
      -0.00000000001,    -0.00000000015,
       0.00000110067,     0.00000000014,
       0.00000000041,     0.00000060244,
      -0.00000001530,     0.00000001368,
      -0.00000000069,    -0.00000000225,
       0.00000000745,     0.00000003097,
      -0.00000000112,     0.00000000135,
       0.00000000549,     0.00000000600,
       0.00000000601,    -0.00000000545,
      -0.00000000005,    -0.00000000004,
      -0.00000000031,     0.00000000050,
      -0.00000000004,    -0.00000000004,
      -0.00000000027,     0.00000000043,
      -0.00000000041,    -0.00000000027,
       0.00000000005,     0.00000000006,
      -0.00000000008,     0.00000000005,
       0.00000000371,    -0.00000000304,
      -0.00000000276,    -0.00000000339,
      -0.00000000610,     0.00000000241,
      -0.00000000024,    -0.00000000043,
      -0.00000000254,    -0.00000000616,
       0.00000000043,    -0.00000000026,
       0.00000000424,     0.00000000464,
      -0.00000000463,     0.00000000422,
       0.00000000006,     0.00000000004,
       0.00000000014,     0.00000000002,
      -0.00000000002,     0.00000000014,
       0.00000000004,    -0.00000000002,
       0.00000000002,     0.00000000004,
       0.00000000038,    -0.00000000035,
       0.00000000035,     0.00000000038,
       0.00000000021,    -0.00000000020,
       0.00000000002,    -0.00000000005,
      -0.00000000005,    -0.00000000002,
      -0.00000000005,    -0.00000000001,
       0.00000000003,    -0.00000000004,
       0.00000000004,     0.00000000003,
       0.00000000000,     0.00000000002,
       0.00000000016,     0.00000000001,
       0.00000000009,     0.00000000006,
       0.00000000006,    -0.00000000010,
       0.00000000005,     0.00000000003,
       0.00000000001,    -0.00000000004,
       0.00000000001,     0.00000000023,
       0.00000000003,     0.00000000002,
       0.00000000196,     0.00000000020,
       0.00000000019,    -0.00000000026,
       0.00000000119,     0.00000000049,
       0.00000000015,    -0.00000000011,
       0.00000000051,    -0.00000000129,
      -0.00000000013,    -0.00000000016,
      -0.00000000004,     0.00000000002,
       0.00000000005,     0.00000000003,
      -0.00000000014,     0.00000000252,
       0.00000000028,     0.00000000019,
       0.00000000002,    -0.00000000001,
       0.00000002130,     0.00000000295,
       0.00000000181,    -0.00000000238,
      -0.00000000010,    -0.00000000017,
       0.00000001360,     0.00000000334,
       0.00000000114,    -0.00000000117,
      -0.00000000004,    -0.00000000005,
       0.00000000359,    -0.00000001465,
      -0.00000000130,    -0.00000000125,
      -0.00000000009,     0.00000000005,
      -0.00000000032,     0.00000000010,
       0.00000000042,     0.00000000017,
       0.00000000132,    -0.00000000073,
      -0.00000000073,    -0.00000000132,
      -0.00000000303,     0.00000002265,
       0.00000000154,     0.00000000099,
       0.00000000006,    -0.00000000005,
       0.00000019335,     0.00000003356,
       0.00000000899,    -0.00000001323,
      -0.00000000045,    -0.00000000057,
       0.00000012914,     0.00000001306,
       0.00000000447,    -0.00000000596,
      -0.00000000019,    -0.00000000023,
       0.00000001591,    -0.00000013842,
      -0.00000000678,    -0.00000000515,
      -0.00000000025,     0.00000000020,
      -0.00000000152,    -0.00000000007,
      -0.00000000010,    -0.00000000003,
       0.00000000284,     0.00000000005,
       0.00000000000,    -0.00000000006,
       0.00000001517,    -0.00000000836,
       0.00000002208,    -0.00000000570,
       0.00000000040,    -0.00000000084,
      -0.00000000569,    -0.00000002207,
      -0.00000000084,    -0.00000000040,
       0.00000000021,    -0.00000000004,
       0.00000000003,    -0.00000000002,
      -0.00000000002,    -0.00000000003,
      -0.00000000003,     0.00000000004,
      -0.00000002952,     0.00000013854,
       0.00000000386,     0.00000000084,
       0.00000000002,    -0.00000000005,
       0.00000119357,     0.00000024982,
       0.00000000707,    -0.00000003341,
      -0.00000000037,    -0.00000000017,
       0.00000083538,    -0.00000001560,
      -0.00000000131,     0.00000000143,
       0.00000000007,    -0.00000000006,
       0.00000000607,    -0.00000089213,
      -0.00000000165,    -0.00000000048,
      -0.00000000010,     0.00000000003,
      -0.00000000189,    -0.00000000677,
      -0.00000000010,     0.00000000014,
       0.00000001143,    -0.00000000787,
       0.00000000016,     0.00000000007,
       0.00000030149,    -0.00000007751,
       0.00000000710,    -0.00000001386,
       0.00000000509,     0.00000002005,
       0.00000000076,     0.00000000037,
       0.00000030723,    -0.00000000042,
      -0.00000000024,    -0.00000030691,
       0.00000002845,    -0.00000000732,
       0.00000000038,     0.00000000048,
       0.00000000005,    -0.00000000008,
       0.00000000000,     0.00000000003,
       0.00000000022,     0.00000000000,
      -0.00000000001,    -0.00000000007,
       0.00000000014,     0.00000000001,
       0.00000000001,    -0.00000000016,
      -0.00000000005,     0.00000000000,
       0.00000000001,    -0.00000000002,
      -0.00000000017,    -0.00000000007,
      -0.00000000011,    -0.00000000004,
      -0.00000000004,     0.00000000011,
       0.00000000011,     0.00000000008,
       0.00000000008,     0.00000000004,
       0.00000000005,    -0.00000000008,
       0.00000000960,    -0.00000000510,
       0.00000000003,    -0.00000000014,
      -0.00000004360,    -0.00000011071,
      -0.00000000122,    -0.00000000020,
      -0.00000003339,    -0.00000000905,
      -0.00000000070,    -0.00000000029,
       0.00000000005,    -0.00000000001,
      -0.00000000964,     0.00000000941,
      -0.00000000152,    -0.00000000006,
      -0.00000000005,    -0.00000000001,
      -0.00000000057,     0.00000000247,
      -0.00000000115,    -0.00000000333,
      -0.00000000009,    -0.00000000007,
       0.00000629494,    -0.00000000680,
       0.00000000094,     0.00000117441,
      -0.00000001423,    -0.00000000720,
      -0.00000000026,    -0.00000000108,
      -0.00000000356,     0.00000002845,
      -0.00000000054,     0.00000000053,
       0.00000001300,     0.00000001037,
       0.00000001036,    -0.00000001298,
       0.00000007193,     0.00000000167,
       0.00000026441,    -0.00000000015,
       0.00000062162,    -0.00000000092,
      -0.00000000017,     0.00000000001,
      -0.00000000005,    -0.00000000001,
       0.00000000033,     0.00000000036,
       0.00000000001,     0.00000000020,
       0.00000000006,    -0.00000000001,
       0.00000000167,     0.00000000003,
      -0.00000000003,    -0.00000000057,
       0.00000000109,    -0.00000000006,
      -0.00000000004,    -0.00000000036,
      -0.00000000004,    -0.00000000118,
      -0.00000000039,     0.00000000003,
       0.00000000002,    -0.00000000007,
      -0.00000000062,    -0.00000000022,
      -0.00000000005,     0.00000000004,
      -0.00000000046,    -0.00000000006,
      -0.00000000003,     0.00000000003,
      -0.00000000009,     0.00000000049,
       0.00000000003,     0.00000000003,
      -0.00000000001,     0.00000000002,
       0.00000000019,     0.00000000006,
       0.00000000013,     0.00000000002,
       0.00000000002,    -0.00000000014,
       0.00000000120,     0.00000000000,
       0.00000000000,    -0.00000000120,
       0.00000000137,    -0.00000000001,
      -0.00000000001,     0.00000000134,
       0.00000000179,    -0.00000000001,
      -0.00000000001,    -0.00000000179,
       0.00000000192,    -0.00000000002,
      -0.00000000002,     0.00000000189,
      -0.00000000006,     0.00000000000,
       0.00000000001,     0.00000000004,
      -0.00000000007,     0.00000000080,
       0.00000000025,    -0.00000000001,
       0.00000000692,     0.00000000054,
      -0.00000000010,    -0.00000000220,
       0.00000000470,    -0.00000000072,
      -0.00000000040,    -0.00000000134,
      -0.00000000064,    -0.00000000504,
      -0.00000000144,     0.00000000039,
       0.00000000005,    -0.00000000005,
       0.00000000156,    -0.00000000025,
      -0.00000000025,    -0.00000000156,
       0.00000000000,    -0.00000000015,
      -0.00000000003,     0.00000000001,
      -0.00000000127,     0.00000000002,
       0.00000000012,     0.00000000026,
      -0.00000000072,     0.00000000020,
       0.00000000011,     0.00000000011,
       0.00000000019,     0.00000000078,
       0.00000000013,    -0.00000000011,
      -0.00000000253,     0.00000000028,
       0.00000000028,     0.00000000253,
       0.00000000121,     0.00000000016,
      -0.00000000005,    -0.00000000017,
       0.00000000125,    -0.00000001015,
      -0.00000000141,     0.00000000046,
       0.00000000003,     0.00000000005,
      -0.00000000088,    -0.00000000650,
      -0.00000000067,     0.00000000054,
       0.00000000003,     0.00000000003,
      -0.00000000700,     0.00000000072,
       0.00000000054,     0.00000000076,
       0.00000000003,    -0.00000000003,
      -0.00000000010,     0.00000000003,
      -0.00000000003,    -0.00000000006,
      -0.00000005431,     0.00000000611,
       0.00000000785,     0.00000000986,
      -0.00000000190,    -0.00000003401,
      -0.00000000321,     0.00000000225,
      -0.00000003404,     0.00000000190,
       0.00000000226,     0.00000000321,
      -0.00000000143,     0.00000000649,
       0.00000000016,     0.00000000005,
       0.00000005517,     0.00000001223,
       0.00000000044,    -0.00000000149,
       0.00000003928,    -0.00000000070,
      -0.00000000001,     0.00000000006,
       0.00000000036,    -0.00000004183,
      -0.00000000005,    -0.00000000007,
      -0.00000000001,    -0.00000000035,
       0.00000000060,    -0.00000000024,
      -0.00000005015,    -0.00000086743,
      -0.00000008186,     0.00000005805,
       0.00000003288,    -0.00000000181,
      -0.00000000220,    -0.00000000310,
       0.00000040548,    -0.00000000094,
      -0.00000000101,    -0.00000040611,
       0.00000000030,     0.00000000008,
       0.00000000071,    -0.00000000326,
       0.00000000010,     0.00000000003,
       0.00000000008,    -0.00000000055,
       0.00000000014,     0.00000000002,
      -0.00000000195,     0.00000000001,
       0.00000000010,    -0.00000000003,
       0.00000000001,     0.00000000005,
      -0.00000000002,    -0.00000000006,
       0.00001549886,    -0.00000003717,
       0.00000000287,     0.00000160007,
       0.00000000120,    -0.00000004488,
       0.00000000212,     0.00000000299,
      -0.00000002240,    -0.00000000240,
       0.00000000148,    -0.00000000424,
       0.00000002012,     0.00000001034,
       0.00000001035,    -0.00000002012,
       0.00000006506,     0.00000000152,
       0.00000014646,    -0.00000000026,
       0.00000000239,     0.00000000059,
       0.00000000059,    -0.00000000239,
       0.00000000001,    -0.00000000002,
      -0.00000000017,    -0.00000000005,
      -0.00000000012,     0.00000000000,
       0.00000000000,     0.00000000013,
       0.00000010429,     0.00000002565,
      -0.00000003949,    -0.00000000484,
      -0.00000000483,     0.00000003947,
       0.00000000005,     0.00000000002,
       0.00000000001,    -0.00000000004,
      -0.00000006311,     0.00000000182,
      -0.00000000003,    -0.00000000131,
      -0.00000000834,     0.00000000057,
       0.00000000056,     0.00000000838,
      -0.00000000003,     0.00000000012,
       0.00000000103,     0.00000000023,
       0.00000000073,    -0.00000000001,
       0.00000000001,    -0.00000000078,
      -0.00000203626,    -0.00000024696,
       0.00000000815,     0.00000003294,
       0.00000000479,    -0.00000003910,
       0.00000057124,    -0.00000000012,
      -0.00000000027,    -0.00000057065,
      -0.00000000045,    -0.00000000058,
      -0.00000000058,     0.00000000045,
      -0.00000033926,     0.00000001958,
      -0.00000000099,    -0.00000001725,
       0.00000000004,    -0.00000000003,
       0.00004401244,    -0.00000001497,
       0.00000000058,     0.00000227581,
       0.00000002633,    -0.00000000652,
      -0.00000000329,    -0.00000005261,
       0.00000002229,     0.00000002229,
       0.00000000046,    -0.00000000066,
       0.00000002227,    -0.00000002228,
      -0.00000000066,    -0.00000000046,
       0.00000007344,    -0.00000000610,
       0.00000000701,    -0.00000000048,
      -0.00000000048,    -0.00000000701,
      -0.00000002953,    -0.00000000538,
       0.00000000804,     0.00000000053,
       0.00000000052,    -0.00000000803,
       0.00000045604,    -0.00000002698,
       0.00000000086,     0.00000001467,
       0.00000020752,    -0.00000023191,
      -0.00000001256,    -0.00000001132,
       0.00000001323,    -0.00000000171,
       0.00000000189,     0.00000001337,
       0.00000002216,     0.00000002216,
       0.00000000045,    -0.00000000066,
      -0.00000002215,     0.00000002215,
       0.00000000066,     0.00000000046,
      -0.00000000691,    -0.00000000045,
      -0.00000000045,     0.00000000691,
      -0.00000042001,    -0.00000002325,
       0.00000000080,    -0.00000001433,
       0.00000000001,     0.00000000545,
      -0.00000000001,    -0.00000000185,
       0.00000000252,    -0.00000000296,
       0.00000000295,     0.00000000252,
      -0.00000000057,    -0.00000000045,
       0.00000000044,    -0.00000000057,
       0.00000000643,    -0.00000000721,
      -0.00000000721,    -0.00000000643,
      -0.00000000001,     0.00000000004,
       0.00000000032,     0.00000000007,
       0.00000000022,     0.00000000000,
       0.00000000000,    -0.00000000024,
       0.00000032094,     0.00000001807,
      -0.00000000096,     0.00000001693,
       0.00000000203,    -0.00000000004,
       0.00000000000,    -0.00000000221,
       0.00000051623,    -0.00000057823,
       0.00000000715,     0.00000000639,
       0.00000071582,    -0.00000000134,
      -0.00000000132,    -0.00000071582,
       0.00008622130,    -0.00000016016,
       0.00000000396,     0.00000285067,
      -0.00000000429,    -0.00000000954,
      -0.00000000475,     0.00000000858,
       0.00000003660,     0.00000002421,
       0.00000000048,    -0.00000000092,
       0.00000002421,    -0.00000003659,
      -0.00000000092,    -0.00000000048,
       0.00000030410,    -0.00000049951,
      -0.00000000234,     0.00000002902,
      -0.00000001693,    -0.00000001141,
       0.00000000093,     0.00000000009,
      -0.00000000210,    -0.00000000231,
       0.00000000231,    -0.00000000215,
       0.00000003648,     0.00000002414,
       0.00000000048,    -0.00000000092,
      -0.00000002413,     0.00000003649,
       0.00000000092,     0.00000000048,
       0.00000000252,    -0.00000000414,
       0.00000000414,     0.00000000252,
       0.00000001056,    -0.00000000122,
       0.00000000037,     0.00000000045,
       0.00000001114,    -0.00000001622,
       0.00000000014,    -0.00000000061,
      -0.00000000850,    -0.00000004045,
       0.00000000101,    -0.00000000010,
       0.00000004223,    -0.00000001012,
       0.00000000030,     0.00000000116,
       0.00000001901,     0.00000001328,
       0.00000000010,    -0.00000000050,
      -0.00000001310,     0.00000001949,
       0.00000000047,     0.00000000009,
      -0.00000000075,     0.00000000063,
       0.00000001482,     0.00000006566,
      -0.00000000168,     0.00000000046,
      -0.00000006540,     0.00000001444,
      -0.00000000043,    -0.00000000165,
      -0.00000000046,    -0.00000000020,
       0.00000000021,    -0.00000000050,
      -0.00000000014,     0.00000000203,
       0.00000001769,     0.00000007764,
      -0.00000000198,     0.00000000051,
      -0.00000007724,     0.00000001674,
      -0.00000000051,    -0.00000000194,
       0.00000001100,    -0.00000000813,
       0.00000000096,     0.00000000105,
       0.00000002165,     0.00000009568,
      -0.00000000245,     0.00000000064,
      -0.00000009516,     0.00000002082,
      -0.00000000061,    -0.00000000242,
       0.00000000020,    -0.00000000033,
       0.00000001317,     0.00000000048,
       0.00000000034,     0.00000000076,
       0.00000004040,     0.00000017690,
      -0.00000000453,     0.00000000118,
      -0.00000017630,     0.00000003859,
      -0.00000000114,    -0.00000000445,
       0.00000000351,     0.00000002019,
      -0.00000000271,    -0.00000000081,
       0.00000005490,     0.00000023938,
      -0.00000000616,     0.00000000171,
      -0.00000023907,     0.00000005277,
      -0.00000000162,    -0.00000000599,
      -0.00000002272,     0.00000000797,
      -0.00000000022,    -0.00000000094,
       0.00000007752,     0.00000034045,
      -0.00000000872,     0.00000000231,
      -0.00000033886,     0.00000007388,
      -0.00000000223,    -0.00000000854,
      -0.00000000048,     0.00000000074,
       0.00000000480,     0.00000000344,
       0.00000009775,     0.00000042547,
      -0.00000001183,     0.00000000311,
      -0.00000042480,     0.00000009310,
      -0.00000000299,    -0.00000001157,
       0.00000000005,     0.00000000000,
      -0.00000000003,    -0.00000000004,
      -0.00000000012,    -0.00000000006,
      -0.00000000007,     0.00000000012,
       0.00000000003,     0.00000000000,
       0.00000000134,    -0.00000000840,
      -0.00000000062,     0.00000000006,
       0.00000000001,    -0.00000000004,
      -0.00000000007,    -0.00000000001,
      -0.00000000007,     0.00000000069,
      -0.00000000003,    -0.00000000001,
      -0.00000000013,     0.00000000022,
       0.00000000005,     0.00000000003,
      -0.00000000012,     0.00000000066,
       0.00000000012,     0.00000000002,
       0.00000000067,     0.00000000012,
       0.00000000002,    -0.00000000013,
      -0.00000000005,     0.00000000003,
       0.00000000003,     0.00000000007,
      -0.00000000002,     0.00000000003,
       0.00000000012,    -0.00000000001,
      -0.00000000005,     0.00000000000,
       0.00000000000,     0.00000000006,
      -0.00000000003,     0.00000000018,
       0.00000000003,     0.00000000001,
       0.00000000112,     0.00000000023,
       0.00000000004,    -0.00000000021,
       0.00000000339,    -0.00000000046,
      -0.00000000007,    -0.00000000053,
      -0.00000000046,    -0.00000000347,
      -0.00000000055,     0.00000000006,
       0.00000000020,     0.00000000017,
       0.00000000025,    -0.00000000022,
      -0.00000000003,    -0.00000000003,
       0.00000000006,     0.00000000003,
       0.00000000004,    -0.00000000006,
      -0.00000000006,     0.00000000009,
       0.00000000001,    -0.00000000005,
      -0.00000000006,    -0.00000000002,
      -0.00000000001,     0.00000000003,
       0.00000000166,     0.00000000076,
      -0.00000000005,     0.00000000001,
       0.00000000012,     0.00000000002,
       0.00000000003,    -0.00000000013,
       0.00000000003,     0.00000000003,
       0.00000000011,     0.00000000003,
       0.00000000120,    -0.00000000015,
      -0.00000000002,    -0.00000000019,
      -0.00000000054,    -0.00000000412,
      -0.00000000069,     0.00000000010,
      -0.00000000814,    -0.00000001711,
      -0.00000000228,     0.00000000103,
       0.00000000006,     0.00000000013,
      -0.00000001767,     0.00000000822,
       0.00000000104,     0.00000000238,
       0.00000000014,    -0.00000000006,
       0.00000000046,    -0.00000000111,
      -0.00000000009,    -0.00000000001,
      -0.00000000138,    -0.00000000088,
      -0.00000000006,     0.00000000014,
      -0.00000000039,     0.00000000005,
       0.00000000068,     0.00000000146,
       0.00000000151,    -0.00000000068,
       0.00000000008,     0.00000000009,
      -0.00000000009,    -0.00000000005,
      -0.00000000006,     0.00000000009,
       0.00000000037,     0.00000000024,
       0.00000000001,    -0.00000000002,
      -0.00000005038,     0.00000007774,
       0.00000000439,     0.00000000283,
       0.00000000079,    -0.00000000268,
      -0.00000000007,     0.00000000000,
      -0.00000000283,    -0.00000000092,
      -0.00000000001,     0.00000000009,
       0.00000000004,    -0.00000000002,
       0.00000000000,    -0.00000000016,
       0.00000000396,    -0.00000000609,
      -0.00000000038,    -0.00000000024,
      -0.00000000102,     0.00000000157,
       0.00000000030,     0.00000000009,
       0.00000000007,     0.00000000000,
       0.00000000000,    -0.00000000008,
       0.00000000001,    -0.00000000010,
       0.00000000014,    -0.00000000003,
      -0.00000000458,    -0.00000000991,
      -0.00000000133,     0.00000000059,
       0.00000000003,     0.00000000007,
       0.00000003528,    -0.00000001557,
      -0.00000000196,    -0.00000000479,
      -0.00000000029,     0.00000000011,
      -0.00000009998,     0.00000009204,
       0.00000000870,     0.00000001003,
       0.00000000041,    -0.00000000037,
       0.00000009397,     0.00000010458,
       0.00000001075,    -0.00000000893,
      -0.00000000038,    -0.00000000046,
      -0.00000000671,    -0.00000000061,
       0.00000000013,     0.00000000028,
      -0.00000000300,     0.00000000945,
       0.00000000067,     0.00000000012,
       0.00000000128,     0.00000000276,
       0.00000000035,    -0.00000000015,
      -0.00000000146,     0.00000000064,
       0.00000000863,    -0.00000000789,
      -0.00000000070,    -0.00000000081,
      -0.00000000798,    -0.00000000906,
      -0.00000000088,     0.00000000071,
       0.00000000016,    -0.00000000043,
       0.00000000007,     0.00000000004,
       0.00000000139,    -0.00000000230,
      -0.00000000010,    -0.00000000003,
       0.00000000011,    -0.00000000045,
      -0.00000000048,    -0.00000000013,
      -0.00000000002,     0.00000000005,
       0.00000000078,     0.00000000023,
       0.00000000021,     0.00000000000,
       0.00000000001,    -0.00000000021,
       0.00000000000,    -0.00000000014,
      -0.00000000002,     0.00000000004,
      -0.00000000002,     0.00000000000,
       0.00000000056,     0.00000000202,
      -0.00000000006,     0.00000000002,
       0.00000000007,     0.00000000009,
       0.00000000009,    -0.00000000008,
       0.00000027093,    -0.00000024393,
      -0.00000002331,    -0.00000002755,
      -0.00000000116,     0.00000000100,
      -0.00002382689,    -0.00002642046,
      -0.00000268486,     0.00000227661,
       0.00000009746,     0.00000011260,
      -0.00000290473,    -0.00000171580,
      -0.00000011708,     0.00000018101,
       0.00000000524,     0.00000000296,
      -0.00000185004,     0.00000300434,
       0.00000019125,     0.00000013414,
       0.00000000395,    -0.00000000576,
      -0.00000002399,    -0.00000011183,
       0.00000000027,     0.00000000287,
      -0.00000000010,    -0.00000000006,
      -0.00000019945,    -0.00000001680,
       0.00000000070,     0.00000000794,
       0.00000000021,     0.00000000019,
      -0.00000006279,     0.00000005653,
       0.00000000540,     0.00000000638,
       0.00000000027,    -0.00000000023,
       0.00000118533,     0.00000131440,
       0.00000013357,    -0.00000011325,
      -0.00000000485,    -0.00000000560,
       0.00000025416,     0.00000015119,
       0.00000001030,    -0.00000001592,
      -0.00000000046,    -0.00000000026,
       0.00000016221,    -0.00000026022,
      -0.00000001666,    -0.00000001172,
      -0.00000000034,     0.00000000050,
      -0.00000000028,     0.00000000683,
       0.00000000896,     0.00000000153,
       0.00000000001,    -0.00000000005,
      -0.00000000056,    -0.00000000015,
      -0.00000000013,     0.00000000001,
       0.00000000000,     0.00000000015,
       0.00000000005,     0.00000000000,
       0.00000000002,    -0.00000000044,
       0.00000002285,     0.00000000789,
       0.00000000207,    -0.00000000371,
       0.00000000015,     0.00000000002,
       0.00000000002,    -0.00000000016,
      -0.00000000117,    -0.00000000046,
      -0.00000000001,     0.00000000001,
       0.00000000258,     0.00000000125,
       0.00000000020,    -0.00000000022,
       0.00000000013,     0.00000000001,
       0.00000000002,    -0.00000000014,
       0.00000000019,     0.00000000009,
       0.00000000010,    -0.00000000010,
      -0.00000011573,    -0.00000007038,
      -0.00000000495,     0.00000000733,
       0.00000000022,     0.00000000014,
      -0.00000099815,     0.00000165967,
       0.00000010447,     0.00000006980,
       0.00000000188,    -0.00000000304,
      -0.00000020001,     0.00000079557,
       0.00000002368,     0.00000000692,
       0.00000000002,    -0.00000000035,
       0.00000084438,     0.00000024070,
       0.00000001092,    -0.00000002765,
      -0.00000000051,    -0.00000000017,
      -0.00000001045,     0.00000000669,
       0.00000000001,    -0.00000000084,
       0.00000000204,     0.00000003749,
       0.00000000068,    -0.00000000016,
       0.00000000004,    -0.00000000002,
       0.00000002146,     0.00000001305,
       0.00000000092,    -0.00000000136,
       0.00000004934,    -0.00000008205,
      -0.00000000516,    -0.00000000345,
       0.00000001838,    -0.00000007146,
      -0.00000000201,    -0.00000000064,
      -0.00000007424,    -0.00000002114,
      -0.00000000089,     0.00000000225,
       0.00000000103,    -0.00000000003,
       0.00000000016,    -0.00000000167,
       0.00000000040,    -0.00000000024,
      -0.00000000002,    -0.00000000005,
      -0.00000000006,     0.00000000005,
       0.00000000005,     0.00000000006,
      -0.00000000008,    -0.00000000009,
      -0.00000000003,    -0.00000000002,
      -0.00000000002,     0.00000000003,
      -0.00000000004,     0.00000000000,
       0.00000000000,     0.00000000033,
       0.00000000008,    -0.00000000037,
      -0.00000000005,    -0.00000000001,
      -0.00000006169,     0.00000022782,
       0.00000000716,     0.00000000249,
       0.00000000002,    -0.00000000012,
       0.00000204097,     0.00000054296,
       0.00000002183,    -0.00000006329,
      -0.00000000104,    -0.00000000021,
       0.00000066978,    -0.00000003128,
      -0.00000000175,     0.00000000324,
       0.00000000009,     0.00000000008,
       0.00000002150,    -0.00000077593,
      -0.00000000271,    -0.00000000182,
      -0.00000000003,     0.00000000011,
      -0.00000000354,    -0.00000000799,
      -0.00000000016,     0.00000000034,
       0.00000002648,     0.00000000134,
      -0.00000000014,    -0.00000000055,
       0.00000000857,    -0.00000003168,
      -0.00000000099,    -0.00000000035,
      -0.00000009706,    -0.00000002577,
      -0.00000000099,     0.00000000280,
      -0.00000006318,     0.00000000110,
      -0.00000000163,     0.00000006816,
       0.00000000036,    -0.00000000036,
      -0.00000000119,    -0.00000000044,
      -0.00000000024,     0.00000000020,
      -0.00000000004,     0.00000000007,
       0.00000000007,     0.00000000005,
      -0.00000000004,    -0.00000000004,
      -0.00000000003,     0.00000000006,
       0.00000037683,    -0.00000000075,
       0.00000000005,     0.00000000002,
      -0.00000000477,    -0.00000238782,
       0.00000000013,    -0.00000000067,
       0.00000000007,     0.00000000001,
      -0.00000001090,     0.00000018615,
       0.00000000137,     0.00000000138,
      -0.00000000001,    -0.00000000006,
      -0.00000005006,    -0.00000003844,
      -0.00000000005,     0.00000000492,
      -0.00000001092,     0.00000000693,
       0.00000000000,    -0.00000000025,
      -0.00000000070,    -0.00000002596,
      -0.00000000004,     0.00000000023,
      -0.00000003458,    -0.00000000043,
       0.00000000036,     0.00000010540,
       0.00000000827,     0.00000002882,
      -0.00000000094,     0.00000000018,
       0.00000003654,    -0.00000000642,
       0.00000000024,     0.00000000073,
      -0.00000000028,    -0.00000000034,
      -0.00000000037,     0.00000000126,
      -0.00000000002,    -0.00000000010,
      -0.00000000002,     0.00000000001,
       0.00000000273,     0.00000000527,
      -0.00000000003,    -0.00000000038,
       0.00000000012,     0.00000000012,
       0.00000000013,    -0.00000000013,
       0.00000000006,     0.00000000004,
       0.00000000005,     0.00000000001,
      -0.00000000331,    -0.00000001257,
       0.00000000038,    -0.00000000006,
      -0.00000009708,     0.00000002747,
      -0.00000000069,    -0.00000000312,
      -0.00000003591,     0.00000000332,
      -0.00000000017,    -0.00000000022,
       0.00000000267,    -0.00000003985,
      -0.00000000018,     0.00000000001,
      -0.00000000366,     0.00000000627,
       0.00000000032,     0.00000000012,
      -0.00000001058,    -0.00000000435,
      -0.00000000002,     0.00000000026,
       0.00000001724,     0.00000007675,
      -0.00000000212,     0.00000000055,
      -0.00000007485,     0.00000001643,
      -0.00000000050,    -0.00000000188,
      -0.00000000249,    -0.00000000009,
       0.00000000007,    -0.00000000295,
      -0.00000000006,     0.00000000070,
       0.00000000151,     0.00000000046,
       0.00000000002,    -0.00000000006,
      -0.00000000003,    -0.00000000001,
      -0.00000000026,     0.00000000022,
      -0.00000000015,     0.00000000017,
      -0.00000000014,    -0.00000000014,
      -0.00000000022,     0.00000000008,
      -0.00000000010,    -0.00000000023,
       0.00000000006,     0.00000000003,
       0.00000000005,     0.00000000004,
      -0.00000000004,     0.00000000004,
       0.00000000003,     0.00000000004,
      -0.00000000004,     0.00000000004,
      -0.00000000067,    -0.00000000005,
       0.00000000031,    -0.00000000020,
      -0.00000000111,     0.00000000019,
       0.00000000012,    -0.00000000001,
      -0.00000000010,     0.00000000005,
       0.00000000005,     0.00000000010,
       0.00000000002,     0.00000000005,
       0.00000000003,    -0.00000000003,
       0.00000000006,    -0.00000000003,
      -0.00000000007,     0.00000000011,
       0.00000000011,     0.00000000007,
       0.00000000005,    -0.00000000007,
      -0.00000000442,    -0.00000000299,
      -0.00000000045,     0.00000000048,
      -0.00000000166,    -0.00000000034,
      -0.00000000007,     0.00000000015,
      -0.00000000035,     0.00000000169,
       0.00000000015,     0.00000000007,
      -0.00000000013,    -0.00000000008,
      -0.00000000012,     0.00000000014,
      -0.00000000013,    -0.00000000003,
      -0.00000000055,     0.00000000269,
       0.00000000024,     0.00000000010,
       0.00000000038,     0.00000000214,
       0.00000000017,     0.00000000001,
       0.00000000219,    -0.00000000037,
       0.00000000002,    -0.00000000018,
      -0.00000000003,     0.00000000016,
       0.00000000020,     0.00000000007,
       0.00000000007,     0.00000000047,
       0.00000000004,     0.00000000000,
       0.00000000679,    -0.00000000125,
       0.00000000000,    -0.00000000060,
       0.00000000545,    -0.00000000332,
      -0.00000000014,    -0.00000000042,
      -0.00000000334,    -0.00000000566,
      -0.00000000045,     0.00000000013,
       0.00000000039,    -0.00000000008,
       0.00000000000,    -0.00000000054,
      -0.00000000045,     0.00000000027,
       0.00000000027,     0.00000000046,
       0.00000000177,    -0.00000000103,
      -0.00000000004,    -0.00000000014,
      -0.00000001226,    -0.00000001973,
      -0.00000000150,     0.00000000055,
      -0.00000001658,    -0.00000001287,
      -0.00000000081,     0.00000000062,
      -0.00000001374,     0.00000001698,
       0.00000000063,     0.00000000092,
      -0.00000000055,    -0.00000000079,
      -0.00000000138,     0.00000000046,
      -0.00000000039,     0.00000000023,
       0.00000000043,     0.00000000069,
       0.00000000138,     0.00000000109,
       0.00000000113,    -0.00000000139,
      -0.00000000725,    -0.00000000583,
      -0.00000000038,     0.00000000026,
      -0.00000005085,     0.00000006646,
       0.00000000258,     0.00000000318,
       0.00000000013,     0.00000000000,
      -0.00000001913,     0.00000006084,
       0.00000000135,     0.00000000087,
       0.00000006393,     0.00000002251,
       0.00000000123,    -0.00000000151,
      -0.00000000064,     0.00000000132,
      -0.00000000008,    -0.00000000007,
       0.00000000172,     0.00000000274,
      -0.00000000005,     0.00000000011,
       0.00000000127,     0.00000000102,
       0.00000000172,    -0.00000000224,
       0.00000000168,    -0.00000000523,
      -0.00000000532,    -0.00000000180,
      -0.00000001266,     0.00000003785,
       0.00000000085,     0.00000000065,
       0.00000000002,     0.00000000000,
       0.00000026961,     0.00000008558,
       0.00000000414,    -0.00000000630,
       0.00000000002,    -0.00000000011,
       0.00000013325,    -0.00000000615,
      -0.00000000067,     0.00000000032,
       0.00000000500,    -0.00000015164,
      -0.00000000051,    -0.00000000026,
      -0.00000000172,    -0.00000000218,
       0.00000000356,    -0.00000000143,
       0.00000000016,     0.00000000004,
       0.00000000164,    -0.00000000502,
      -0.00000000867,    -0.00000000275,
      -0.00000001112,    -0.00000000029,
      -0.00000000058,     0.00000001155,
       0.00000012974,     0.00000000022,
      -0.00000000006,    -0.00000000003,
       0.00000000128,    -0.00000070309,
      -0.00000000016,     0.00000000035,
      -0.00000000421,     0.00000007648,
       0.00000000054,     0.00000000068,
      -0.00000001933,    -0.00000001675,
       0.00000000006,     0.00000000214,
      -0.00000000366,     0.00000000557,
      -0.00000000022,    -0.00000000019,
      -0.00000000313,    -0.00000000587,
       0.00000000012,    -0.00000000032,
      -0.00000000940,    -0.00000000096,
       0.00000000018,     0.00000002119,
       0.00000000738,     0.00000003170,
      -0.00000000084,     0.00000000020,
       0.00000003311,    -0.00000000700,
       0.00000000021,     0.00000000081,
      -0.00000000089,    -0.00000000364,
       0.00000000009,     0.00000000006,
      -0.00000002246,     0.00000000604,
       0.00000000030,    -0.00000000049,
      -0.00000000713,     0.00000000156,
       0.00000000010,     0.00000000006,
       0.00000000132,    -0.00000000787,
       0.00000000005,    -0.00000000013,
      -0.00000000157,     0.00000000108,
       0.00000000004,     0.00000000017,
      -0.00000000186,    -0.00000000208,
      -0.00000000019,     0.00000000005,
       0.00000002580,     0.00000011594,
      -0.00000000318,     0.00000000081,
      -0.00000009641,     0.00000002141,
      -0.00000000065,    -0.00000000245,
       0.00000000008,    -0.00000000154,
      -0.00000000100,     0.00000000144,
       0.00000000154,     0.00000000113,
      -0.00000000001,     0.00000000005,
       0.00000000006,     0.00000000009,
      -0.00000000009,     0.00000000006,
      -0.00000000002,     0.00000000003,
      -0.00000000003,    -0.00000000002,
      -0.00000000035,     0.00000000134,
       0.00000000021,    -0.00000000008,
       0.00000000000,     0.00000000011,
       0.00000000012,     0.00000000000,
       0.00000000005,     0.00000000000,
      -0.00000000001,    -0.00000000004,
      -0.00000000006,    -0.00000000001,
      -0.00000000005,    -0.00000000003,
      -0.00000000003,     0.00000000005,
      -0.00000000003,    -0.00000000002,
      -0.00000000016,     0.00000000018,
       0.00000000005,     0.00000000003,
      -0.00000000018,     0.00000000007,
       0.00000000002,     0.00000000003,
       0.00000000008,     0.00000000019,
       0.00000000003,    -0.00000000002,
      -0.00000000012,     0.00000000005,
       0.00000000001,     0.00000000002,
       0.00000000020,     0.00000000083,
       0.00000000015,    -0.00000000007,
      -0.00000000016,     0.00000000051,
       0.00000000007,     0.00000000001,
       0.00000000058,     0.00000000013,
       0.00000000000,    -0.00000000008,
       0.00000000006,    -0.00000000046,
      -0.00000000047,    -0.00000000005,
      -0.00000000011,     0.00000000045,
       0.00000000006,     0.00000000000,
       0.00000000260,     0.00000000097,
       0.00000000005,    -0.00000000035,
       0.00000000063,     0.00000000101,
       0.00000000006,    -0.00000000003,
       0.00000000111,    -0.00000000093,
      -0.00000000008,    -0.00000000007,
      -0.00000000001,    -0.00000000078,
      -0.00000000067,    -0.00000000025,
      -0.00000000118,    -0.00000000041,
      -0.00000000041,     0.00000000123,
       0.00000000091,     0.00000000112,
       0.00000000006,    -0.00000000007,
       0.00000000639,    -0.00000000457,
      -0.00000000037,    -0.00000000033,
       0.00000000142,     0.00000000048,
       0.00000000004,     0.00000000005,
      -0.00000000006,    -0.00000000234,
      -0.00000000004,     0.00000000001,
      -0.00000000005,    -0.00000000001,
       0.00000000002,    -0.00000000006,
      -0.00000000273,     0.00000000018,
       0.00000000012,     0.00000000032,
      -0.00000000165,     0.00000000102,
      -0.00000000299,     0.00000000528,
       0.00000000538,     0.00000000305,
       0.00000000317,    -0.00000000007,
      -0.00000000043,    -0.00000001574,
      -0.00000000026,     0.00000000211,
      -0.00000000032,    -0.00000000031,
      -0.00000000002,     0.00000000006,
      -0.00000000007,     0.00000000009,
      -0.00000000007,    -0.00000000009,
      -0.00000000736,     0.00000002054,
       0.00000000101,     0.00000000043,
      -0.00000000381,     0.00000000269,
       0.00000000787,     0.00000003515,
      -0.00000000090,     0.00000000024,
       0.00000003530,    -0.00000000810,
       0.00000000022,     0.00000000090,
      -0.00000000013,     0.00000000001,
      -0.00000000001,     0.00000000063,
      -0.00000000004,     0.00000000002,
      -0.00000000015,    -0.00000000010,
       0.00000000005,    -0.00000000012,
      -0.00000000003,    -0.00000000002,
       0.00000004597,     0.00000020628,
      -0.00000000569,     0.00000000145,
      -0.00000012341,     0.00000002731,
      -0.00000000081,    -0.00000000315,
      -0.00000000345,    -0.00000000232,
       0.00000000045,    -0.00000000030,
      -0.00000000550,     0.00000000234,
      -0.00000000037,    -0.00000000029,
      -0.00000000090,     0.00000000127,
       0.00000000129,     0.00000000090,
       0.00000000062,     0.00000000040,
       0.00000000032,    -0.00000000055,
      -0.00000000085,    -0.00000000117,
       0.00000000117,    -0.00000000087,
      -0.00000000080,     0.00000000113,
      -0.00000000112,    -0.00000000080,
       0.00000000002,     0.00000000001,
       0.00000000009,    -0.00000000009,
       0.00000000012,     0.00000000006,
       0.00000000004,    -0.00000000013,
       0.00000000001,    -0.00000000004,
       0.00000000030,     0.00000000011,
       0.00000000004,    -0.00000000003,
       0.00000000083,    -0.00000000124,
      -0.00000000010,    -0.00000000020,
       0.00000000113,     0.00000000045,
       0.00000000010,    -0.00000000006,
       0.00000000020,    -0.00000000120,
      -0.00000000009,    -0.00000000007,
       0.00000000335,     0.00000000074,
       0.00000000028,    -0.00000000030,
      -0.00000000001,    -0.00000000002,
       0.00000000593,    -0.00000001490,
      -0.00000000125,    -0.00000000146,
      -0.00000000006,     0.00000000004,
       0.00000000821,     0.00000000310,
       0.00000000051,    -0.00000000022,
       0.00000000032,    -0.00000000892,
      -0.00000000049,    -0.00000000022,
      -0.00000000017,     0.00000000007,
       0.00000000005,    -0.00000000009,
       0.00000003183,     0.00000000309,
       0.00000000113,    -0.00000000151,
      -0.00000000005,    -0.00000000006,
       0.00000002587,    -0.00000014608,
      -0.00000000699,    -0.00000000559,
      -0.00000000026,     0.00000000021,
       0.00000003694,     0.00000002150,
       0.00000000106,     0.00000000088,
       0.00000000004,    -0.00000000004,
      -0.00000000451,    -0.00000004304,
      -0.00000000040,     0.00000000063,
      -0.00000000101,     0.00000000052,
      -0.00000000002,     0.00000000004,
      -0.00000000038,    -0.00000000116,
       0.00000000202,     0.00000000401,
       0.00000000401,    -0.00000000202,
       0.00000020727,    -0.00000000111,
      -0.00000000012,    -0.00000000003,
      -0.00000000558,    -0.00000096272,
      -0.00000000013,     0.00000000051,
      -0.00000000008,    -0.00000000001,
      -0.00000000758,     0.00000013893,
       0.00000000071,     0.00000000087,
       0.00000000002,    -0.00000000004,
      -0.00000002716,    -0.00000002953,
      -0.00000000023,     0.00000000387,
      -0.00000000588,     0.00000000498,
      -0.00000000013,    -0.00000000014,
      -0.00000000474,    -0.00000000593,
       0.00000000013,    -0.00000000012,
       0.00000002799,     0.00000005534,
      -0.00000000384,     0.00000000199,
       0.00000001339,     0.00000005981,
      -0.00000000152,     0.00000000038,
       0.00000005975,    -0.00000001335,
       0.00000000038,     0.00000000153,
       0.00000000004,     0.00000000000,
       0.00000000004,    -0.00000000016,
      -0.00000000005,    -0.00000000001,
       0.00000000009,     0.00000000002,
      -0.00000000001,    -0.00000000009,
      -0.00000000003,    -0.00000000001,
      -0.00000000005,     0.00000000012,
      -0.00000000004,    -0.00000000003,
       0.00000000000,     0.00000000005,
       0.00000000002,     0.00000000001,
       0.00000000005,    -0.00000000008,
      -0.00000000501,    -0.00000000198,
      -0.00000000008,    -0.00000000022,
      -0.00000000990,     0.00000002337,
      -0.00000000101,     0.00000000037,
      -0.00000000003,    -0.00000000003,
      -0.00000000133,    -0.00000000358,
       0.00000000012,    -0.00000000009,
       0.00000000133,    -0.00000000184,
       0.00000000003,    -0.00000000008,
       0.00000000026,     0.00000000056,
       0.00000000005,     0.00000000002,
      -0.00000000070,     0.00000000017,
      -0.00000000002,     0.00000000005,
       0.00000027675,     0.00000123744,
      -0.00000003405,     0.00000000872,
      -0.00000023439,     0.00000005235,
      -0.00000000155,    -0.00000000596,
       0.00000000079,    -0.00000000311,
      -0.00000000575,     0.00000000054,
      -0.00000000148,     0.00000000301,
       0.00000000300,     0.00000000148,
       0.00000000288,     0.00000001446,
       0.00000001180,     0.00000005294,
       0.00000002760,     0.00000012316,
       0.00000000027,    -0.00000000002,
      -0.00000000002,    -0.00000000009,
       0.00000000002,    -0.00000000126,
      -0.00000000042,     0.00000000002,
       0.00000000031,     0.00000000012,
       0.00000000003,    -0.00000000010,
      -0.00000000008,    -0.00000000032,
      -0.00000000010,     0.00000000004,
      -0.00000000011,    -0.00000000001,
      -0.00000000009,     0.00000000053,
       0.00000000003,     0.00000000003,
       0.00000000000,    -0.00000000008,
       0.00000000001,     0.00000000004,
       0.00000000003,     0.00000000000,
       0.00000000001,    -0.00000000015,
       0.00000000001,     0.00000000005,
       0.00000000117,    -0.00000000016,
      -0.00000000009,    -0.00000000033,
      -0.00000000077,    -0.00000000542,
      -0.00000000154,     0.00000000044,
       0.00000000002,     0.00000000081,
       0.00000000022,    -0.00000000005,
      -0.00000000014,    -0.00000000009,
      -0.00000000004,     0.00000000006,
      -0.00000000003,     0.00000000003,
      -0.00000000003,    -0.00000000003,
      -0.00000000018,     0.00000000005,
       0.00000000003,     0.00000000003,
       0.00000000024,     0.00000000078,
       0.00000000013,    -0.00000000012,
      -0.00000000005,    -0.00000000036,
      -0.00000000004,     0.00000000003,
      -0.00000000020,     0.00000000005,
      -0.00000000019,    -0.00000000164,
      -0.00000000017,     0.00000000013,
      -0.00000000729,     0.00000000096,
       0.00000000061,     0.00000000079,
       0.00000000003,    -0.00000000003,
       0.00000000212,    -0.00000000007,
      -0.00000000006,    -0.00000000011,
      -0.00000000024,    -0.00000000074,
       0.00000000006,     0.00000000001,
       0.00000000003,     0.00000000005,
      -0.00000000004,     0.00000000003,
      -0.00000000356,    -0.00000001038,
       0.00000000659,    -0.00000000186,
      -0.00000000186,    -0.00000000659,
       0.00000000986,    -0.00000000004,
      -0.00000000017,    -0.00000004552,
      -0.00000000025,     0.00000000660,
       0.00000000004,     0.00000000005,
      -0.00000000134,    -0.00000000148,
       0.00000000000,     0.00000000016,
      -0.00000000032,     0.00000000017,
      -0.00000000017,    -0.00000000032,
       0.00000016864,    -0.00000004790,
      -0.00000001365,    -0.00000000932,
       0.00000000183,     0.00000000651,
       0.00000001796,     0.00000007984,
      -0.00000000202,     0.00000000061,
       0.00000007996,    -0.00000001800,
       0.00000000060,     0.00000000201,
       0.00000000001,    -0.00000000030,
       0.00000000002,     0.00000000001,
      -0.00000000140,    -0.00000000003,
       0.00000000007,    -0.00000000010,
       0.00000000010,     0.00000000003,
      -0.00000000005,    -0.00000000007,
       0.00000068945,     0.00000306393,
      -0.00000008415,     0.00000002288,
      -0.00000031860,     0.00000007149,
      -0.00000000219,    -0.00000000809,
       0.00000000893,    -0.00000000174,
      -0.00000000034,     0.00000000068,
      -0.00000000051,    -0.00000000454,
      -0.00000000117,     0.00000000443,
       0.00000000443,     0.00000000117,
       0.00000000652,     0.00000002907,
      -0.00000000003,     0.00000000000,
       0.00000000000,     0.00000000014,
      -0.00000000054,     0.00000002157,
      -0.00000000078,    -0.00000000802,
      -0.00000000802,     0.00000000078,
       0.00000000001,    -0.00000000005,
      -0.00000000048,    -0.00000000162,
      -0.00000000163,     0.00000000048,
       0.00000000018,     0.00000000000,
       0.00000000000,    -0.00000000085,
       0.00000000000,     0.00000000012,
      -0.00000000002,    -0.00000000003,
      -0.00000004074,    -0.00000041392,
       0.00000000800,    -0.00000000078,
       0.00000002519,     0.00000011312,
      -0.00000000287,     0.00000000076,
       0.00000011300,    -0.00000002520,
       0.00000000076,     0.00000000287,
      -0.00000001881,    -0.00000006623,
       0.00000000323,    -0.00000000095,
       0.00000194516,     0.00000872801,
      -0.00000023961,     0.00000006227,
      -0.00000045227,     0.00000010075,
      -0.00000000323,    -0.00000001245,
       0.00000000245,     0.00000000493,
       0.00000001028,    -0.00000000297,
      -0.00000000345,     0.00000000540,
       0.00000000540,     0.00000000345,
       0.00000000040,     0.00000000137,
       0.00000000137,    -0.00000000040,
       0.00000000025,     0.00000000162,
       0.00000000162,    -0.00000000025,
       0.00000002550,     0.00000008931,
      -0.00000000275,     0.00000000081,
       0.00000004545,     0.00000001819,
       0.00000000103,    -0.00000000248,
       0.00000000092,     0.00000000255,
      -0.00000000256,     0.00000000096,
      -0.00000000344,     0.00000000539,
      -0.00000000538,    -0.00000000344,
      -0.00000000022,    -0.00000000139,
      -0.00000000139,     0.00000000022,
      -0.00000001393,    -0.00000008438,
       0.00000000276,    -0.00000000043,
      -0.00000000108,     0.00000000024,
       0.00000000171,     0.00000000096,
       0.00000000096,    -0.00000000171,
       0.00000000006,     0.00000000000,
       0.00000000000,    -0.00000000026,
       0.00000001056,     0.00000006428,
      -0.00000000327,     0.00000000050,
       0.00000013759,     0.00000007697,
      -0.00000000096,     0.00000000171,
       0.00000003185,     0.00000014190,
      -0.00000000361,     0.00000000095,
       0.00000014190,    -0.00000003185,
       0.00000000095,     0.00000000361,
       0.00000383909,     0.00001710374,
      -0.00000047061,     0.00000012267,
      -0.00000056617,     0.00000012680,
      -0.00000000406,    -0.00000001559,
       0.00000000170,    -0.00000000127,
      -0.00000000191,    -0.00000000057,
      -0.00000000320,     0.00000000833,
       0.00000000833,     0.00000000320,
       0.00000009341,     0.00000001335,
       0.00000000069,    -0.00000000323,
      -0.00000000320,     0.00000000831,
      -0.00000000831,    -0.00000000320,
       0.00000000116,     0.00000000220,
      -0.00000000016,     0.00000000015,
       0.00000000203,    -0.00000000086,
       0.00000000018,     0.00000000015,
       0.00000000958,    -0.00000000363,
       0.00000000029,     0.00000000050,
       0.00000000395,     0.00000001002,
      -0.00000000053,     0.00000000033,
      -0.00000000092,     0.00000000165,
      -0.00000000161,    -0.00000000096,
      -0.00000001401,     0.00000000666,
      -0.00000000037,    -0.00000000071,
      -0.00000000657,    -0.00000001399,
       0.00000000071,    -0.00000000036,
      -0.00000000045,     0.00000000009,
      -0.00000001647,     0.00000000782,
      -0.00000000043,    -0.00000000083,
      -0.00000000761,    -0.00000001645,
       0.00000000082,    -0.00000000042,
       0.00000000237,     0.00000000206,
      -0.00000000024,     0.00000000033,
      -0.00000002029,     0.00000000958,
      -0.00000000053,    -0.00000000103,
      -0.00000000940,    -0.00000002024,
       0.00000000102,    -0.00000000052,
       0.00000000055,     0.00000000297,
      -0.00000003749,     0.00000001779,
      -0.00000000098,    -0.00000000190,
      -0.00000001742,    -0.00000003750,
       0.00000000188,    -0.00000000096,
      -0.00000000433,     0.00000000178,
      -0.00000005072,     0.00000002413,
      -0.00000000135,    -0.00000000257,
      -0.00000002371,    -0.00000005082,
       0.00000000254,    -0.00000000132,
      -0.00000000291,    -0.00000000468,
      -0.00000007217,     0.00000003419,
      -0.00000000189,    -0.00000000365,
      -0.00000003341,    -0.00000007208,
       0.00000000362,    -0.00000000185,
      -0.00000000053,     0.00000000124,
      -0.00000009014,     0.00000004292,
      -0.00000000238,    -0.00000000456,
      -0.00000004199,    -0.00000009034,
       0.00000000453,    -0.00000000233,
      -0.00000000003,    -0.00000000015,
      -0.00000000002,     0.00000000004,
       0.00000000004,    -0.00000000004,
      -0.00000000004,    -0.00000000004,
      -0.00000000063,     0.00000000183,
       0.00000000017,     0.00000000016,
       0.00000000004,     0.00000000001,
       0.00000000001,    -0.00000000001,
      -0.00000000007,    -0.00000000007,
      -0.00000000027,    -0.00000000013,
      -0.00000000003,     0.00000000006,
      -0.00000000014,     0.00000000027,
       0.00000000006,     0.00000000003,
      -0.00000000008,    -0.00000000013,
      -0.00000000010,    -0.00000000005,
      -0.00000000001,     0.00000000002,
      -0.00000000018,     0.00000000036,
       0.00000000008,     0.00000000004,
      -0.00000000022,     0.00000000153,
       0.00000000029,     0.00000000004,
       0.00000000158,     0.00000000025,
       0.00000000005,    -0.00000000030,
      -0.00000000009,     0.00000000006,
       0.00000000006,     0.00000000013,
       0.00000000008,    -0.00000000004,
      -0.00000000005,     0.00000000004,
       0.00000000004,     0.00000000005,
       0.00000000001,    -0.00000000001,
      -0.00000000077,    -0.00000000091,
      -0.00000000005,     0.00000000005,
      -0.00000000006,    -0.00000000004,
      -0.00000000004,     0.00000000007,
      -0.00000000002,     0.00000000003,
      -0.00000000012,     0.00000000075,
       0.00000000014,     0.00000000002,
      -0.00000000009,    -0.00000000003,
       0.00000000928,    -0.00000000154,
      -0.00000000023,    -0.00000000153,
      -0.00000000009,     0.00000000001,
      -0.00000000145,    -0.00000000967,
      -0.00000000162,     0.00000000020,
       0.00000000001,     0.00000000009,
       0.00000000044,     0.00000000035,
       0.00000000003,    -0.00000000005,
       0.00000000059,    -0.00000000051,
      -0.00000000007,    -0.00000000007,
      -0.00000000080,     0.00000000013,
       0.00000000011,     0.00000000083,
       0.00000000002,     0.00000000004,
      -0.00000000320,     0.00000000153,
       0.00000000019,     0.00000000037,
       0.00000000024,    -0.00000000023,
      -0.00000000024,    -0.00000000026,
       0.00000000001,    -0.00000000001,
      -0.00000000032,    -0.00000000038,
      -0.00000000012,    -0.00000000007,
      -0.00000000007,     0.00000000011,
      -0.00000000003,     0.00000000005,
      -0.00000000016,     0.00000000008,
       0.00000001173,    -0.00000000177,
      -0.00000000026,    -0.00000000196,
      -0.00000000014,     0.00000000002,
       0.00000003981,     0.00000026747,
       0.00000004464,    -0.00000000580,
      -0.00000000040,    -0.00000000318,
      -0.00000005460,    -0.00000010873,
      -0.00000001437,     0.00000000678,
       0.00000000038,     0.00000000080,
      -0.00000011505,     0.00000005485,
       0.00000000675,     0.00000001556,
       0.00000000090,    -0.00000000038,
       0.00000000211,    -0.00000000554,
      -0.00000000043,    -0.00000000002,
      -0.00000000735,    -0.00000000496,
      -0.00000000044,     0.00000000079,
      -0.00000000290,     0.00000000044,
       0.00000000006,     0.00000000045,
      -0.00000000193,    -0.00000001294,
      -0.00000000202,     0.00000000025,
       0.00000000470,     0.00000000952,
       0.00000000118,    -0.00000000054,
       0.00000001000,    -0.00000000466,
      -0.00000000053,    -0.00000000127,
       0.00000000032,     0.00000000025,
       0.00000000015,     0.00000000016,
       0.00000000007,     0.00000000004,
       0.00000000005,    -0.00000000007,
      -0.00000000002,    -0.00000000001,
      -0.00000000012,     0.00000000019,
      -0.00000000002,     0.00000000008,
       0.00000000008,     0.00000000003,
       0.00000000985,     0.00000002042,
       0.00000000274,    -0.00000000123,
      -0.00000000007,    -0.00000000016,
       0.00000042419,    -0.00000020631,
      -0.00000002571,    -0.00000005671,
      -0.00000000320,     0.00000000145,
       0.00000013314,    -0.00000012702,
      -0.00000001183,    -0.00000001322,
      -0.00000000054,     0.00000000051,
      -0.00000013074,    -0.00000014498,
      -0.00000001510,     0.00000001229,
       0.00000000053,     0.00000000067,
       0.00000000678,     0.00000000039,
      -0.00000000015,    -0.00000000024,
       0.00000000399,    -0.00000001085,
      -0.00000000080,    -0.00000000022,
      -0.00000000213,    -0.00000000442,
      -0.00000000056,     0.00000000024,
      -0.00000002143,     0.00000001042,
       0.00000000120,     0.00000000269,
      -0.00000001190,     0.00000001119,
       0.00000000097,     0.00000000112,
       0.00000001136,     0.00000001271,
       0.00000000124,    -0.00000000100,
      -0.00000000038,    -0.00000000012,
      -0.00000000022,     0.00000000048,
       0.00000000025,     0.00000000145,
       0.00000000021,    -0.00000000003,
       0.00000000004,     0.00000000007,
       0.00000000007,    -0.00000000004,
       0.00000000001,     0.00000000001,
       0.00000000008,    -0.00000000014,
       0.00000000001,    -0.00000000006,
      -0.00000000007,    -0.00000000002,
      -0.00000000001,     0.00000000002,
       0.00000000019,     0.00000000005,
       0.00000000001,    -0.00000000007,
      -0.00000000006,     0.00000000018,
       0.00000000004,    -0.00000000003,
       0.00000002335,    -0.00000002146,
      -0.00000000202,    -0.00000000238,
      -0.00000000010,     0.00000000009,
      -0.00000023991,    -0.00000025816,
      -0.00000002617,     0.00000002249,
       0.00000000096,     0.00000000110,
      -0.00000016455,    -0.00000009330,
      -0.00000000613,     0.00000001005,
       0.00000000029,     0.00000000014,
      -0.00000010864,     0.00000017430,
       0.00000001113,     0.00000000811,
       0.00000000025,    -0.00000000035,
      -0.00000000120,    -0.00000000427,
       0.00000000007,     0.00000000012,
      -0.00000000992,    -0.00000000154,
      -0.00000000004,     0.00000000045,
      -0.00000000433,     0.00000000398,
       0.00000000035,     0.00000000041,
       0.00000001207,     0.00000001298,
       0.00000000124,    -0.00000000105,
       0.00000001486,     0.00000000862,
       0.00000000055,    -0.00000000085,
       0.00000000955,    -0.00000001536,
      -0.00000000091,    -0.00000000066,
       0.00000000043,     0.00000000012,
       0.00000000010,    -0.00000000001,
       0.00000000001,    -0.00000000002,
      -0.00000000016,    -0.00000000004,
       0.00000000000,     0.00000000005,
       0.00000000002,     0.00000000000,
       0.00000000000,    -0.00000000015,
      -0.00000005334,    -0.00000003202,
      -0.00000000226,     0.00000000335,
       0.00000000010,     0.00000000006,
      -0.00000025877,     0.00000043691,
       0.00000002722,     0.00000001814,
       0.00000000049,    -0.00000000080,
      -0.00000005150,     0.00000023045,
       0.00000000638,     0.00000000127,
      -0.00000000001,    -0.00000000008,
       0.00000025881,     0.00000007778,
       0.00000000387,    -0.00000000883,
      -0.00000000018,    -0.00000000007,
      -0.00000000101,     0.00000000141,
      -0.00000000002,    -0.00000000024,
      -0.00000000055,     0.00000001075,
       0.00000000027,     0.00000000000,
       0.00000000824,     0.00000000495,
       0.00000000033,    -0.00000000048,
       0.00000001273,    -0.00000002151,
      -0.00000000124,    -0.00000000084,
       0.00000000530,    -0.00000002164,
      -0.00000000059,    -0.00000000017,
      -0.00000002298,    -0.00000000665,
      -0.00000000029,     0.00000000071,
       0.00000000013,    -0.00000000047,
       0.00000000000,     0.00000000007,
      -0.00000003003,     0.00000011251,
       0.00000000351,     0.00000000120,
       0.00000000001,    -0.00000000006,
       0.00000072487,     0.00000019058,
       0.00000000760,    -0.00000002237,
      -0.00000000037,    -0.00000000006,
       0.00000018236,    -0.00000002347,
      -0.00000000149,     0.00000000242,
       0.00000000009,     0.00000000003,
       0.00000001668,    -0.00000025709,
      -0.00000000198,    -0.00000000123,
      -0.00000000002,     0.00000000008,
      -0.00000000161,    -0.00000000435,
      -0.00000000005,     0.00000000010,
       0.00000001028,     0.00000000140,
      -0.00000000004,    -0.00000000026,
       0.00000000371,    -0.00000001391,
      -0.00000000040,    -0.00000000014,
      -0.00000003438,    -0.00000000901,
      -0.00000000034,     0.00000000099,
      -0.00000001935,     0.00000000103,
      -0.00000000063,     0.00000002231,
      -0.00000000046,    -0.00000000023,
       0.00000000001,    -0.00000000013,
       0.00000014650,    -0.00000000176,
      -0.00000000006,     0.00000000012,
      -0.00000000750,    -0.00000076027,
       0.00000000050,     0.00000000012,
      -0.00000000614,     0.00000013946,
       0.00000000067,     0.00000000125,
      -0.00000000001,    -0.00000000005,
      -0.00000001716,    -0.00000003260,
      -0.00000000011,     0.00000000398,
      -0.00000000483,     0.00000000193,
      -0.00000000005,    -0.00000000005,
       0.00000000055,    -0.00000001017,
      -0.00000000007,     0.00000000005,
      -0.00000001358,     0.00000000016,
       0.00000000038,     0.00000003383,
      -0.00000000034,    -0.00000000689,
      -0.00000000206,     0.00000000154,
      -0.00000000018,     0.00000000049,
       0.00000001135,     0.00000005140,
      -0.00000000140,     0.00000000042,
      -0.00000000001,    -0.00000000002,
       0.00000016159,    -0.00000003394,
       0.00000000135,     0.00000000427,
      -0.00000000006,     0.00000000003,
       0.00000008864,     0.00000000469,
      -0.00000000030,    -0.00000000044,
       0.00000000470,     0.00000009955,
      -0.00000000043,     0.00000000016,
      -0.00000000305,     0.00000000267,
       0.00000000023,     0.00000000004,
      -0.00000000370,    -0.00000000318,
      -0.00000000002,     0.00000000022,
      -0.00000000058,    -0.00000000313,
      -0.00000000738,     0.00000000156,
      -0.00000000801,     0.00000000272,
       0.00000000271,     0.00000000400,
       0.00000000137,     0.00000000037,
       0.00000000264,    -0.00000000496,
      -0.00000000015,     0.00000000010,
      -0.00000000052,    -0.00000000308,
       0.00000000007,     0.00000000000,
       0.00000000330,    -0.00000000061,
       0.00000000000,     0.00000000008,
       0.00000000408,     0.00000000197,
      -0.00000000008,    -0.00000000010,
      -0.00000000191,     0.00000000416,
       0.00000000009,    -0.00000000008,
      -0.00000001533,     0.00000000718,
      -0.00000000043,    -0.00000000083,
      -0.00000000834,    -0.00000001723,
       0.00000000089,    -0.00000000046,
      -0.00000000016,    -0.00000000048,
       0.00000000059,    -0.00000000011,
      -0.00000000042,    -0.00000000010,
      -0.00000000003,    -0.00000000002,
      -0.00000000010,    -0.00000000006,
      -0.00000000013,    -0.00000000002,
      -0.00000000003,     0.00000000014,
      -0.00000000005,    -0.00000000001,
      -0.00000000008,     0.00000000047,
       0.00000000004,     0.00000000002,
       0.00000000010,     0.00000000051,
       0.00000000053,    -0.00000000009,
       0.00000000001,    -0.00000000005,
       0.00000000004,     0.00000000002,
       0.00000000004,     0.00000000021,
       0.00000000002,     0.00000000000,
       0.00000000190,    -0.00000000040,
       0.00000000001,    -0.00000000016,
       0.00000000165,    -0.00000000105,
      -0.00000000003,    -0.00000000012,
      -0.00000000105,    -0.00000000177,
      -0.00000000014,     0.00000000003,
       0.00000000009,    -0.00000000002,
       0.00000000000,    -0.00000000014,
       0.00000000085,    -0.00000000051,
      -0.00000000001,    -0.00000000006,
      -0.00000000412,    -0.00000000644,
      -0.00000000052,     0.00000000018,
      -0.00000000538,    -0.00000000402,
      -0.00000000024,     0.00000000018,
      -0.00000000456,     0.00000000557,
       0.00000000021,     0.00000000032,
      -0.00000000014,    -0.00000000017,
      -0.00000000039,     0.00000000011,
       0.00000000046,     0.00000000036,
       0.00000000037,    -0.00000000047,
      -0.00000000334,    -0.00000000265,
      -0.00000000018,     0.00000000012,
      -0.00000001685,     0.00000002227,
       0.00000000087,     0.00000000110,
      -0.00000000518,     0.00000001805,
       0.00000000036,     0.00000000021,
       0.00000001972,     0.00000000725,
       0.00000000041,    -0.00000000049,
      -0.00000000001,     0.00000000026,
       0.00000000039,     0.00000000077,
       0.00000000049,     0.00000000039,
       0.00000000057,    -0.00000000075,
       0.00000000051,    -0.00000000161,
      -0.00000000165,    -0.00000000057,
      -0.00000000490,     0.00000001478,
       0.00000000034,     0.00000000025,
       0.00000008160,     0.00000002591,
       0.00000000126,    -0.00000000194,
       0.00000003010,    -0.00000000370,
      -0.00000000034,     0.00000000014,
       0.00000000255,    -0.00000003967,
      -0.00000000023,    -0.00000000010,
      -0.00000000067,    -0.00000000082,
       0.00000000110,    -0.00000000032,
       0.00000000004,     0.00000000001,
       0.00000000057,    -0.00000000173,
      -0.00000000263,    -0.00000000083,
      -0.00000000301,     0.00000000010,
      -0.00000000003,     0.00000000320,
       0.00000003724,    -0.00000000053,
      -0.00000000009,    -0.00000000004,
      -0.00000000162,    -0.00000016838,
      -0.00000000015,     0.00000000036,
      -0.00000000181,     0.00000003572,
       0.00000000002,     0.00000000040,
      -0.00000000288,    -0.00000000919,
      -0.00000000006,     0.00000000094,
      -0.00000000105,     0.00000000146,
      -0.00000000007,    -0.00000000005,
      -0.00000000076,    -0.00000000162,
       0.00000000003,    -0.00000000009,
      -0.00000000326,     0.00000000004,
       0.00000000006,     0.00000000505,
      -0.00000000007,    -0.00000000131,
      -0.00000000073,     0.00000000023,
       0.00000000471,     0.00000002126,
      -0.00000000059,     0.00000000019,
       0.00000006460,    -0.00000001410,
       0.00000000067,     0.00000000180,
       0.00000003737,     0.00000000200,
      -0.00000000008,    -0.00000000013,
       0.00000000179,     0.00000004271,
      -0.00000000011,     0.00000000000,
      -0.00000000130,     0.00000000013,
       0.00000000003,     0.00000000002,
      -0.00000000021,    -0.00000000134,
      -0.00000000002,     0.00000000004,
       0.00000000001,    -0.00000000087,
      -0.00000000210,     0.00000000051,
      -0.00000000683,     0.00000000296,
      -0.00000000017,    -0.00000000032,
       0.00000000296,     0.00000000595,
      -0.00000000032,     0.00000000017,
       0.00000000040,     0.00000000030,
       0.00000000003,     0.00000000003,
       0.00000000139,    -0.00000000130,
       0.00000000012,    -0.00000000015,
      -0.00000000049,    -0.00000000147,
       0.00000000003,     0.00000000003,
       0.00000000158,    -0.00000000025,
      -0.00000000001,     0.00000000004,
       0.00000000104,     0.00000000117,
       0.00000000004,    -0.00000000004,
      -0.00000000115,     0.00000000101,
       0.00000000004,     0.00000000004,
      -0.00000002372,     0.00000001109,
      -0.00000000066,    -0.00000000129,
      -0.00000000991,    -0.00000002102,
       0.00000000106,    -0.00000000054,
      -0.00000000040,    -0.00000000020,
      -0.00000000001,    -0.00000000001,
      -0.00000000005,     0.00000000006,
      -0.00000000005,     0.00000000002,
       0.00000000003,     0.00000000006,
      -0.00000000005,     0.00000000002,
       0.00000000007,     0.00000000025,
       0.00000000005,    -0.00000000002,
      -0.00000000006,     0.00000000012,
       0.00000000015,     0.00000000003,
      -0.00000000005,     0.00000000015,
       0.00000000002,     0.00000000000,
       0.00000000069,     0.00000000027,
       0.00000000002,    -0.00000000009,
       0.00000000007,     0.00000000024,
       0.00000000023,    -0.00000000020,
       0.00000000020,     0.00000000032,
       0.00000000002,    -0.00000000001,
       0.00000000144,    -0.00000000087,
      -0.00000000006,    -0.00000000008,
       0.00000000015,     0.00000000030,
      -0.00000000007,    -0.00000000037,
      -0.00000000031,    -0.00000000048,
      -0.00000000035,     0.00000000020,
      -0.00000000041,    -0.00000000017,
      -0.00000000013,     0.00000000044,
       0.00000000051,     0.00000000018,
       0.00000000051,    -0.00000000221,
       0.00000000026,     0.00000000080,
      -0.00000000020,     0.00000000028,
       0.00000000002,     0.00000000003,
      -0.00000000098,    -0.00000000038,
      -0.00000000011,     0.00000000046,
      -0.00000000120,    -0.00000000027,
      -0.00000000019,     0.00000000116,
       0.00000000009,     0.00000000061,
       0.00000000185,    -0.00000000020,
       0.00000000000,     0.00000000005,
       0.00000000098,    -0.00000000002,
       0.00000000006,     0.00000000117,
      -0.00000000466,    -0.00000000033,
      -0.00000000045,    -0.00000000091,
      -0.00000000727,     0.00000000337,
      -0.00000000019,    -0.00000000036,
       0.00000000341,     0.00000000712,
      -0.00000000037,     0.00000000018,
      -0.00000000001,    -0.00000000002,
      -0.00000000006,     0.00000000003,
      -0.00000000005,     0.00000000000,
       0.00000000000,    -0.00000000005,
      -0.00000004292,     0.00000002011,
      -0.00000000119,    -0.00000000235,
      -0.00000001239,    -0.00000002650,
       0.00000000134,    -0.00000000068,
       0.00000000035,    -0.00000000085,
      -0.00000000080,    -0.00000000108,
       0.00000000004,     0.00000000002,
       0.00000000010,    -0.00000000015,
       0.00000000011,     0.00000000005,
       0.00000000003,    -0.00000000010,
       0.00000000040,     0.00000000013,
       0.00000000003,    -0.00000000002,
       0.00000000064,    -0.00000000147,
      -0.00000000009,    -0.00000000013,
       0.00000000069,     0.00000000035,
       0.00000000005,    -0.00000000002,
       0.00000000012,    -0.00000000054,
       0.00000000293,     0.00000000092,
       0.00000000015,    -0.00000000010,
       0.00000000352,    -0.00000001108,
      -0.00000000041,    -0.00000000055,
       0.00000000327,     0.00000000217,
       0.00000000013,    -0.00000000002,
       0.00000000048,    -0.00000000122,
       0.00000000010,     0.00000000010,
      -0.00000000010,     0.00000000004,
      -0.00000000003,    -0.00000000010,
       0.00000001412,     0.00000000637,
       0.00000000016,     0.00000000018,
       0.00000001817,    -0.00000005318,
       0.00000000053,    -0.00000000045,
      -0.00000000005,    -0.00000000003,
       0.00000001122,     0.00000001002,
       0.00000000028,    -0.00000000050,
       0.00000000247,     0.00000001061,
       0.00000000057,     0.00000000065,
      -0.00000000047,     0.00000000020,
      -0.00000000018,    -0.00000000047,
      -0.00000000079,     0.00000000065,
       0.00000000065,     0.00000000079,
       0.00000000881,     0.00000004130,
      -0.00000000115,     0.00000000026,
       0.00000011824,    -0.00000002465,
       0.00000000072,     0.00000000331,
       0.00000006950,     0.00000000189,
      -0.00000000019,    -0.00000000031,
       0.00000000299,     0.00000008093,
      -0.00000000028,     0.00000000007,
      -0.00000000078,    -0.00000000059,
       0.00000000004,    -0.00000000004,
       0.00000000059,    -0.00000000077,
       0.00000000004,     0.00000000004,
      -0.00000001090,     0.00000000896,
      -0.00000000064,    -0.00000000076,
      -0.00000001261,     0.00000000593,
       0.00000000591,     0.00000001260,
       0.00000000003,     0.00000000000,
       0.00000000002,    -0.00000000012,
       0.00000000004,     0.00000000002,
      -0.00000000001,    -0.00000000001,
      -0.00000000002,     0.00000000004,
       0.00000000016,    -0.00000000096,
       0.00000000007,    -0.00000000002,
      -0.00000000272,    -0.00000000055,
      -0.00000000005,    -0.00000000020,
      -0.00000000174,    -0.00000000092,
      -0.00000000001,    -0.00000000005,
       0.00000000092,    -0.00000000197,
       0.00000000008,    -0.00000000002,
       0.00000000031,     0.00000000035,
      -0.00000000035,     0.00000000029,
      -0.00000026185,     0.00000012294,
      -0.00000000729,    -0.00000001429,
      -0.00000002339,    -0.00000004985,
       0.00000000251,    -0.00000000129,
      -0.00000000040,    -0.00000000125,
      -0.00000002630,     0.00000001237,
       0.00000000013,     0.00000000004,
       0.00000000001,    -0.00000000004,
       0.00000000009,    -0.00000000046,
      -0.00000000014,    -0.00000000001,
       0.00000000010,     0.00000000006,
       0.00000000002,    -0.00000000003,
       0.00000000003,     0.00000000008,
       0.00000000000,    -0.00000000002,
      -0.00000000007,     0.00000000002,
       0.00000000001,    -0.00000000004,
       0.00000000007,     0.00000000023,
       0.00000000006,    -0.00000000003,
       0.00000000067,    -0.00000000018,
      -0.00000000009,    -0.00000000018,
       0.00000000035,    -0.00000000006,
      -0.00000000003,    -0.00000000011,
       0.00000000006,     0.00000000042,
       0.00000000013,    -0.00000000004,
      -0.00000000002,    -0.00000000012,
      -0.00000000001,     0.00000000001,
      -0.00000000040,     0.00000000008,
       0.00000000004,     0.00000000004,
      -0.00000000003,    -0.00000000012,
       0.00000000066,    -0.00000000011,
      -0.00000000002,    -0.00000000002,
      -0.00000000032,    -0.00000000221,
      -0.00000000005,     0.00000000007,
      -0.00000000002,    -0.00000000032,
      -0.00000000005,     0.00000000004,
       0.00000000071,    -0.00000000014,
      -0.00000000004,    -0.00000000005,
       0.00000000074,     0.00000000137,
       0.00000000138,    -0.00000000074,
       0.00000000045,     0.00000000196,
      -0.00000000005,     0.00000000002,
       0.00000000559,    -0.00000000129,
       0.00000000005,     0.00000000014,
       0.00000000333,     0.00000000011,
       0.00000000014,     0.00000000386,
      -0.00000000003,    -0.00000000004,
       0.00000000004,    -0.00000000003,
       0.00000001905,     0.00000003522,
      -0.00000000137,     0.00000000073,
      -0.00000001690,     0.00000000796,
      -0.00000000046,    -0.00000000085,
       0.00000000798,     0.00000001692,
      -0.00000000084,     0.00000000045,
       0.00000000006,    -0.00000000001,
      -0.00000000002,    -0.00000000018,
       0.00000000000,    -0.00000000011,
       0.00000000013,     0.00000000000,
      -0.00000064942,     0.00000030596,
      -0.00000001836,    -0.00000003536,
      -0.00000003183,    -0.00000006768,
       0.00000000341,    -0.00000000177,
       0.00000000083,     0.00000000191,
       0.00000000099,    -0.00000000034,
      -0.00000000105,    -0.00000000004,
      -0.00000000004,     0.00000000105,
       0.00000000175,    -0.00000000057,
      -0.00000000057,    -0.00000000175,
       0.00000000001,     0.00000000004,
       0.00000000011,    -0.00000000002,
       0.00000000006,     0.00000000000,
       0.00000000000,     0.00000000007,
       0.00000008994,    -0.00000002942,
       0.00000000057,     0.00000000175,
      -0.00000002400,     0.00000001124,
      -0.00000000062,    -0.00000000121,
       0.00000001124,     0.00000002397,
      -0.00000000121,     0.00000000062,
      -0.00000185249,     0.00000086809,
      -0.00000005164,    -0.00000010086,
      -0.00000004505,    -0.00000009616,
       0.00000000484,    -0.00000000250,
      -0.00000000098,     0.00000000079,
       0.00000000117,     0.00000000215,
      -0.00000000138,    -0.00000000050,
      -0.00000000050,     0.00000000138,
      -0.00000000138,    -0.00000000050,
       0.00000000050,    -0.00000000137,
      -0.00000001033,     0.00000003449,
      -0.00000003010,     0.00000001416,
      -0.00000000078,    -0.00000000152,
       0.00000001416,     0.00000003011,
      -0.00000000152,     0.00000000078,
      -0.00000362969,     0.00000170763,
      -0.00000010165,    -0.00000019798,
      -0.00000005654,    -0.00000012033,
       0.00000000606,    -0.00000000313,
      -0.00000000202,    -0.00000000030,
      -0.00000000030,     0.00000000202,
      -0.00000000202,    -0.00000000030,
       0.00000000030,    -0.00000000202,
      -0.00000000049,     0.00000000043,
      -0.00000000005,    -0.00000000004,
       0.00000000017,     0.00000000037,
       0.00000000139,     0.00000000231,
      -0.00000000016,     0.00000000013,
      -0.00000000237,     0.00000000149,
      -0.00000000014,    -0.00000000016,
      -0.00000000033,    -0.00000000005,
       0.00000000005,    -0.00000000033,
      -0.00000000229,    -0.00000000295,
       0.00000000295,    -0.00000000228,
      -0.00000000270,    -0.00000000347,
       0.00000000347,    -0.00000000265,
      -0.00000000036,     0.00000000066,
      -0.00000000331,    -0.00000000427,
       0.00000000032,    -0.00000000027,
       0.00000000427,    -0.00000000327,
       0.00000000026,     0.00000000032,
      -0.00000000615,    -0.00000000789,
       0.00000000059,    -0.00000000049,
       0.00000000792,    -0.00000000607,
       0.00000000049,     0.00000000059,
      -0.00000000065,    -0.00000000093,
      -0.00000000833,    -0.00000001067,
       0.00000000080,    -0.00000000068,
       0.00000001072,    -0.00000000825,
       0.00000000067,     0.00000000080,
       0.00000000095,    -0.00000000093,
      -0.00000001182,    -0.00000001519,
       0.00000000114,    -0.00000000095,
       0.00000001522,    -0.00000001165,
       0.00000000094,     0.00000000113,
      -0.00000001481,    -0.00000001896,
       0.00000000142,    -0.00000000120,
       0.00000001907,    -0.00000001463,
       0.00000000118,     0.00000000142,
      -0.00000000042,     0.00000000088,
       0.00000000007,     0.00000000006,
      -0.00000000003,    -0.00000000003,
       0.00000000004,    -0.00000000004,
      -0.00000000003,    -0.00000000003,
       0.00000000003,    -0.00000000002,
       0.00000000009,    -0.00000000010,
      -0.00000000010,    -0.00000000009,
       0.00000000004,    -0.00000000002,
       0.00000000004,    -0.00000000004,
      -0.00000000001,    -0.00000000001,
      -0.00000000007,    -0.00000000006,
      -0.00000000066,    -0.00000000030,
      -0.00000000007,     0.00000000014,
      -0.00000000032,     0.00000000068,
       0.00000000015,     0.00000000007,
      -0.00000000001,    -0.00000000004,
      -0.00000000006,     0.00000000001,
      -0.00000000047,    -0.00000000022,
      -0.00000000005,     0.00000000010,
       0.00000000133,    -0.00000000281,
      -0.00000000065,    -0.00000000032,
      -0.00000000003,     0.00000000006,
      -0.00000000064,     0.00000000550,
       0.00000000109,     0.00000000014,
       0.00000000001,    -0.00000000008,
       0.00000000575,     0.00000000080,
       0.00000000019,    -0.00000000115,
      -0.00000000008,    -0.00000000002,
      -0.00000000026,     0.00000000019,
       0.00000000003,     0.00000000003,
       0.00000000018,     0.00000000041,
       0.00000000006,    -0.00000000003,
       0.00000000006,    -0.00000000048,
      -0.00000000050,    -0.00000000007,
       0.00000000092,    -0.00000000688,
      -0.00000000137,    -0.00000000021,
      -0.00000000002,     0.00000000012,
      -0.00000035550,    -0.00000004729,
      -0.00000001065,     0.00000007080,
       0.00000000605,     0.00000000095,
      -0.00000006242,     0.00000001174,
       0.00000000170,     0.00000001025,
       0.00000000072,    -0.00000000012,
       0.00000001064,     0.00000006632,
       0.00000001111,    -0.00000000141,
      -0.00000000009,    -0.00000000080,
      -0.00000000246,    -0.00000000183,
      -0.00000000013,     0.00000000029,
      -0.00000000357,     0.00000000295,
       0.00000000042,     0.00000000046,
      -0.00000000021,     0.00000000160,
       0.00000000030,     0.00000000005,
       0.00000001786,     0.00000000237,
       0.00000000052,    -0.00000000332,
       0.00000000551,    -0.00000000098,
      -0.00000000013,    -0.00000000085,
      -0.00000000088,    -0.00000000577,
      -0.00000000090,     0.00000000011,
      -0.00000000437,     0.00000000074,
       0.00000000010,     0.00000000073,
       0.00000000005,    -0.00000000001,
       0.00000001031,     0.00000005979,
       0.00000000992,    -0.00000000146,
      -0.00000000010,    -0.00000000070,
       0.00000001555,     0.00000002956,
       0.00000000387,    -0.00000000190,
      -0.00000000011,    -0.00000000021,
       0.00000003224,    -0.00000001555,
      -0.00000000188,    -0.00000000438,
      -0.00000000026,     0.00000000010,
      -0.00000000042,     0.00000000122,
       0.00000000009,     0.00000000000,
       0.00000000176,     0.00000000127,
       0.00000000011,    -0.00000000018,
       0.00000000091,    -0.00000000015,
      -0.00000000053,    -0.00000000306,
      -0.00000000047,     0.00000000007,
      -0.00000000136,    -0.00000000266,
      -0.00000000282,     0.00000000135,
       0.00000000015,     0.00000000036,
       0.00000000130,    -0.00000000364,
      -0.00000000067,    -0.00000000030,
      -0.00000000001,     0.00000000007,
       0.00000000007,     0.00000000001,
       0.00000000363,     0.00000000730,
       0.00000000098,    -0.00000000044,
      -0.00000000003,    -0.00000000006,
       0.00000006921,    -0.00000003495,
      -0.00000000427,    -0.00000000923,
      -0.00000000052,     0.00000000024,
       0.00000003428,    -0.00000003394,
      -0.00000000312,    -0.00000000334,
      -0.00000000013,     0.00000000013,
      -0.00000003533,    -0.00000003927,
      -0.00000000414,     0.00000000330,
       0.00000000014,     0.00000000019,
       0.00000000136,     0.00000000001,
      -0.00000000004,    -0.00000000004,
       0.00000000108,    -0.00000000254,
      -0.00000000018,    -0.00000000007,
      -0.00000000067,    -0.00000000136,
      -0.00000000017,     0.00000000008,
      -0.00000000353,     0.00000000178,
       0.00000000020,     0.00000000044,
      -0.00000000317,     0.00000000305,
       0.00000000026,     0.00000000030,
       0.00000000312,     0.00000000345,
       0.00000000034,    -0.00000000027,
       0.00000001257,    -0.00000001174,
      -0.00000000109,    -0.00000000128,
      -0.00000000005,     0.00000000005,
      -0.00000008984,    -0.00000009489,
      -0.00000000962,     0.00000000833,
       0.00000000036,     0.00000000041,
      -0.00000006477,    -0.00000003480,
      -0.00000000214,     0.00000000387,
       0.00000000011,     0.00000000004,
      -0.00000004489,     0.00000007081,
       0.00000000456,     0.00000000346,
       0.00000000011,    -0.00000000015,
      -0.00000000046,    -0.00000000109,
       0.00000000005,     0.00000000004,
      -0.00000000364,    -0.00000000084,
      -0.00000000004,     0.00000000017,
      -0.00000000204,     0.00000000191,
       0.00000000016,     0.00000000020,
       0.00000000451,     0.00000000477,
       0.00000000045,    -0.00000000039,
       0.00000000602,     0.00000000340,
       0.00000000021,    -0.00000000034,
       0.00000000391,    -0.00000000630,
      -0.00000000037,    -0.00000000027,
      -0.00000002999,    -0.00000001776,
      -0.00000000125,     0.00000000187,
       0.00000000006,     0.00000000003,
      -0.00000011163,     0.00000019094,
       0.00000001182,     0.00000000781,
       0.00000000021,    -0.00000000035,
      -0.00000001611,     0.00000008917,
       0.00000000218,     0.00000000004,
       0.00000010855,     0.00000003474,
       0.00000000189,    -0.00000000389,
      -0.00000000009,    -0.00000000004,
       0.00000000031,     0.00000000044,
      -0.00000000001,    -0.00000000010,
      -0.00000000068,     0.00000000444,
       0.00000000015,     0.00000000002,
       0.00000000417,     0.00000000247,
       0.00000000016,    -0.00000000024,
       0.00000000547,    -0.00000000936,
      -0.00000000054,    -0.00000000036,
       0.00000000200,    -0.00000000879,
      -0.00000000961,    -0.00000000282,
      -0.00000001465,     0.00000005624,
       0.00000000174,     0.00000000056,
       0.00000000000,    -0.00000000003,
       0.00000030319,     0.00000007812,
       0.00000000303,    -0.00000000928,
      -0.00000000015,    -0.00000000002,
       0.00000004723,    -0.00000001577,
      -0.00000000105,     0.00000000163,
       0.00000000006,     0.00000000001,
       0.00000001128,    -0.00000009635,
      -0.00000000131,    -0.00000000079,
      -0.00000000002,     0.00000000006,
      -0.00000000069,    -0.00000000232,
      -0.00000000002,     0.00000000003,
       0.00000000450,     0.00000000100,
       0.00000000000,    -0.00000000014,
       0.00000000170,    -0.00000000652,
      -0.00000000019,    -0.00000000006,
      -0.00000001434,    -0.00000000367,
      -0.00000000014,     0.00000000041,
      -0.00000000616,     0.00000000066,
      -0.00000000035,     0.00000000794,
       0.00000005462,    -0.00000000139,
      -0.00000000007,     0.00000000010,
      -0.00000000540,    -0.00000025445,
       0.00000000039,     0.00000000026,
      -0.00000000252,     0.00000008453,
       0.00000000027,     0.00000000076,
      -0.00000000585,    -0.00000002001,
      -0.00000000003,     0.00000000244,
      -0.00000000213,     0.00000000031,
      -0.00000000004,     0.00000000001,
       0.00000000066,    -0.00000000413,
      -0.00000000006,     0.00000000000,
      -0.00000000506,     0.00000000013,
       0.00000000027,     0.00000001130,
      -0.00000000040,    -0.00000000455,
      -0.00000000187,     0.00000000107,
       0.00000000791,     0.00000003464,
      -0.00000000097,     0.00000000029,
      -0.00000000001,    -0.00000000002,
       0.00000011179,    -0.00000002450,
       0.00000000096,     0.00000000303,
      -0.00000000003,     0.00000000002,
       0.00000003429,     0.00000000396,
      -0.00000000032,    -0.00000000041,
       0.00000000527,     0.00000004892,
      -0.00000000054,     0.00000000020,
      -0.00000000187,     0.00000000124,
       0.00000000014,     0.00000000001,
      -0.00000000134,    -0.00000000184,
      -0.00000000001,     0.00000000013,
      -0.00000000055,    -0.00000000241,
      -0.00000000516,     0.00000000112,
      -0.00000000042,    -0.00000000029,
      -0.00000000031,    -0.00000000146,
      -0.00000000853,     0.00000000468,
      -0.00000000028,    -0.00000000052,
       0.00000001109,     0.00000001766,
      -0.00000000119,     0.00000000064,
       0.00000000552,     0.00000002688,
      -0.00000000070,     0.00000000015,
      -0.00000002808,     0.00000000690,
      -0.00000000027,    -0.00000000083,
       0.00000000174,     0.00000000070,
      -0.00000000003,    -0.00000000003,
      -0.00000000067,     0.00000000173,
       0.00000000003,    -0.00000000003,
       0.00000000039,    -0.00000000019,
      -0.00000000055,    -0.00000000086,
      -0.00000000103,    -0.00000000167,
      -0.00000000071,     0.00000000079,
       0.00000000000,     0.00000000022,
       0.00000000047,    -0.00000000009,
       0.00000000068,    -0.00000000012,
       0.00000000012,     0.00000000075,
      -0.00000000015,     0.00000000047,
      -0.00000000046,    -0.00000000017,
      -0.00000000242,    -0.00000000313,
       0.00000000023,    -0.00000000020,
       0.00000000372,    -0.00000000289,
       0.00000000023,     0.00000000028,
      -0.00000000004,    -0.00000000002,
      -0.00000000005,    -0.00000000001,
      -0.00000000001,     0.00000000006,
      -0.00000000003,     0.00000000000,
      -0.00000000003,     0.00000000020,
       0.00000000005,     0.00000000021,
       0.00000000023,    -0.00000000004,
       0.00000000002,     0.00000000012,
       0.00000000081,    -0.00000000018,
       0.00000000000,    -0.00000000007,
       0.00000000066,    -0.00000000043,
      -0.00000000001,    -0.00000000005,
      -0.00000000043,    -0.00000000073,
      -0.00000000006,     0.00000000001,
       0.00000000000,    -0.00000000005,
       0.00000000043,    -0.00000000027,
      -0.00000000001,    -0.00000000003,
      -0.00000000172,    -0.00000000263,
      -0.00000000020,     0.00000000006,
      -0.00000000202,    -0.00000000144,
      -0.00000000009,     0.00000000007,
      -0.00000000176,     0.00000000213,
       0.00000000007,     0.00000000012,
      -0.00000000005,    -0.00000000004,
      -0.00000000013,     0.00000000003,
      -0.00000000152,    -0.00000000119,
      -0.00000000008,     0.00000000005,
      -0.00000000632,     0.00000000845,
       0.00000000034,     0.00000000042,
      -0.00000000142,     0.00000000581,
       0.00000000012,     0.00000000004,
       0.00000000669,     0.00000000258,
       0.00000000014,    -0.00000000016,
       0.00000000005,     0.00000000005,
       0.00000000010,     0.00000000025,
       0.00000000016,    -0.00000000054,
      -0.00000000056,    -0.00000000019,
      -0.00000000181,     0.00000000561,
       0.00000000012,     0.00000000009,
       0.00000002680,     0.00000000837,
       0.00000000040,    -0.00000000066,
       0.00000000665,    -0.00000000189,
      -0.00000000016,     0.00000000006,
       0.00000000119,    -0.00000001119,
      -0.00000000011,    -0.00000000004,
      -0.00000000025,    -0.00000000030,
       0.00000000036,    -0.00000000007,
       0.00000000020,    -0.00000000062,
      -0.00000000086,    -0.00000000027,
      -0.00000000078,     0.00000000005,
       0.00000000000,     0.00000000086,
       0.00000001037,    -0.00000000036,
      -0.00000000006,    -0.00000000002,
      -0.00000000111,    -0.00000004255,
      -0.00000000008,     0.00000000020,
      -0.00000000055,     0.00000001518,
      -0.00000000001,     0.00000000016,
      -0.00000000034,    -0.00000000398,
      -0.00000000003,     0.00000000037,
      -0.00000000029,     0.00000000036,
      -0.00000000017,    -0.00000000045,
      -0.00000000091,     0.00000000003,
       0.00000000004,     0.00000000128,
      -0.00000000010,    -0.00000000069,
      -0.00000000047,     0.00000000017,
       0.00000000231,     0.00000000992,
      -0.00000000027,     0.00000000008,
       0.00000002988,    -0.00000000677,
       0.00000000028,     0.00000000083,
       0.00000001036,     0.00000000103,
      -0.00000000010,    -0.00000000012,
       0.00000000161,     0.00000001485,
      -0.00000000009,     0.00000000000,
      -0.00000000060,     0.00000000001,
       0.00000000001,    -0.00000000058,
      -0.00000000015,    -0.00000000065,
      -0.00000000094,     0.00000000021,
      -0.00000000377,     0.00000000199,
      -0.00000000010,    -0.00000000020,
       0.00000000447,     0.00000000777,
      -0.00000000040,     0.00000000019,
       0.00000000224,     0.00000001147,
      -0.00000000028,     0.00000000008,
      -0.00000001205,     0.00000000295,
      -0.00000000011,    -0.00000000032,
       0.00000000032,     0.00000000025,
      -0.00000000024,     0.00000000031,
      -0.00000000102,    -0.00000000140,
      -0.00000000120,     0.00000000097,
      -0.00000000007,     0.00000000008,
       0.00000000018,     0.00000000019,
       0.00000000026,     0.00000000002,
      -0.00000000006,     0.00000000028,
      -0.00000000018,     0.00000000023,
      -0.00000000022,    -0.00000000018,
      -0.00000000378,    -0.00000000491,
       0.00000000037,    -0.00000000030,
       0.00000000448,    -0.00000000347,
       0.00000000028,     0.00000000033,
      -0.00000000002,     0.00000000000,
       0.00000000055,     0.00000000396,
       0.00000000060,    -0.00000000005,
       0.00000000008,     0.00000000016,
       0.00000000017,    -0.00000000008,
       0.00000000004,    -0.00000000002,
      -0.00000000003,     0.00000000005,
       0.00000000007,     0.00000000002,
      -0.00000000002,     0.00000000001,
       0.00000000002,     0.00000000008,
      -0.00000000002,     0.00000000004,
       0.00000000018,     0.00000000008,
      -0.00000000001,     0.00000000006,
       0.00000000005,    -0.00000000005,
       0.00000000003,     0.00000000008,
       0.00000000033,    -0.00000000014,
      -0.00000000001,     0.00000000010,
      -0.00000000002,    -0.00000000007,
       0.00000000006,     0.00000000007,
       0.00000000021,    -0.00000000028,
       0.00000000004,     0.00000000023,
      -0.00000000007,     0.00000000007,
      -0.00000000002,     0.00000000017,
       0.00000000052,     0.00000000002,
       0.00000000012,     0.00000000011,
      -0.00000000009,     0.00000000022,
      -0.00000000012,     0.00000000004,
       0.00000000007,     0.00000000025,
       0.00000000005,     0.00000000033,
      -0.00000000036,     0.00000000006,
      -0.00000000020,    -0.00000000108,
      -0.00000000115,    -0.00000000151,
      -0.00000000147,     0.00000000115,
      -0.00000000691,    -0.00000000896,
       0.00000000067,    -0.00000000056,
       0.00000000562,    -0.00000000433,
       0.00000000035,     0.00000000042,
       0.00000000004,     0.00000000002,
       0.00000000007,    -0.00000000013,
       0.00000000006,     0.00000000005,
       0.00000000023,     0.00000000014,
       0.00000000044,    -0.00000000081,
      -0.00000000001,    -0.00000000004,
       0.00000000027,     0.00000000027,
       0.00000000005,     0.00000000002,
       0.00000000083,     0.00000000090,
       0.00000000002,     0.00000000003,
       0.00000000243,    -0.00000000294,
       0.00000000007,    -0.00000000003,
       0.00000000107,     0.00000000125,
       0.00000000004,    -0.00000000002,
      -0.00000000006,     0.00000000126,
       0.00000000003,     0.00000000008,
      -0.00000000040,     0.00000000437,
      -0.00000000014,     0.00000000003,
       0.00000001158,     0.00000000022,
       0.00000000006,     0.00000000035,
       0.00000000496,     0.00000000396,
       0.00000000005,     0.00000000004,
      -0.00000000341,     0.00000000690,
      -0.00000000015,     0.00000000012,
      -0.00000000007,    -0.00000000008,
       0.00000000008,    -0.00000000007,
      -0.00000000782,     0.00000000353,
      -0.00000000020,    -0.00000000043,
       0.00000000687,     0.00000001554,
      -0.00000000086,     0.00000000040,
       0.00000000470,     0.00000002321,
      -0.00000000061,     0.00000000011,
      -0.00000002463,     0.00000000571,
      -0.00000000019,    -0.00000000072,
       0.00000000008,    -0.00000000011,
       0.00000000011,     0.00000000008,
      -0.00000000205,    -0.00000000265,
      -0.00000000265,     0.00000000204,
       0.00000000018,    -0.00000000001,
       0.00000000001,     0.00000000002,
       0.00000000000,    -0.00000000036,
       0.00000000004,    -0.00000000001,
       0.00000000010,    -0.00000000057,
       0.00000000060,     0.00000000010,
       0.00000000001,     0.00000000004,
      -0.00000000006,     0.00000000009,
      -0.00000000009,    -0.00000000006,
      -0.00000004258,    -0.00000005510,
       0.00000000412,    -0.00000000343,
       0.00000001052,    -0.00000000812,
       0.00000000065,     0.00000000079,
      -0.00000000004,     0.00000000002,
       0.00000000001,     0.00000000001,
       0.00000000003,     0.00000000008,
       0.00000000004,     0.00000000011,
       0.00000000003,    -0.00000000002,
      -0.00000000012,     0.00000000004,
       0.00000000002,     0.00000000004,
       0.00000000002,    -0.00000000003,
       0.00000000000,     0.00000000004,
       0.00000000009,    -0.00000000001,
       0.00000000004,     0.00000000002,
      -0.00000000002,     0.00000000006,
       0.00000000004,    -0.00000000001,
      -0.00000000003,    -0.00000000012,
       0.00000000006,    -0.00000000002,
       0.00000000006,     0.00000000016,
       0.00000000039,    -0.00000000013,
       0.00000000030,    -0.00000000005,
       0.00000000008,     0.00000000035,
      -0.00000000037,     0.00000000018,
      -0.00000000001,    -0.00000000002,
       0.00000000035,     0.00000000073,
      -0.00000000004,     0.00000000002,
       0.00000000024,     0.00000000111,
      -0.00000000117,     0.00000000029,
      -0.00000000728,     0.00000000632,
      -0.00000000275,    -0.00000000356,
      -0.00000000356,     0.00000000276,
      -0.00000010596,    -0.00000013674,
       0.00000001103,    -0.00000000923,
       0.00000001427,    -0.00000001104,
       0.00000000089,     0.00000000106,
       0.00000001163,     0.00000001962,
      -0.00000000390,    -0.00000000506,
      -0.00000000505,     0.00000000390,
      -0.00000030127,    -0.00000039048,
       0.00000003155,    -0.00000002615,
       0.00000002028,    -0.00000001564,
       0.00000000126,     0.00000000151,
      -0.00000000491,    -0.00000000634,
      -0.00000000634,     0.00000000491,
      -0.00000059197,    -0.00000076483,
       0.00000006182,    -0.00000005138,
       0.00000002536,    -0.00000001961,
       0.00000000158,     0.00000000189,
      -0.00000000014,    -0.00000000010,
       0.00000000001,    -0.00000000002,
      -0.00000000003,     0.00000000006,
      -0.00000000053,     0.00000000052,
      -0.00000000006,    -0.00000000005,
      -0.00000000055,    -0.00000000054,
       0.00000000005,    -0.00000000007,
       0.00000000000,    -0.00000000005,
       0.00000000005,     0.00000000000,
       0.00000000059,    -0.00000000072,
       0.00000000071,     0.00000000059,
       0.00000000069,    -0.00000000084,
       0.00000000083,     0.00000000070,
       0.00000000086,    -0.00000000103,
       0.00000000102,     0.00000000086,
       0.00000000158,    -0.00000000191,
       0.00000000190,     0.00000000159,
       0.00000000214,    -0.00000000259,
       0.00000000258,     0.00000000216,
       0.00000000304,    -0.00000000368,
       0.00000000365,     0.00000000306,
       0.00000000380,    -0.00000000461,
       0.00000000458,     0.00000000384,
      -0.00000003095,     0.00000006745,
       0.00000000253,    -0.00000000529,
      -0.00000000093,     0.00000000199,
       0.00000000003,     0.00000000005,
       0.00000000005,    -0.00000000003,
       0.00000000002,     0.00000000003,
      -0.00000000004,     0.00000000002,
       0.00000000024,    -0.00000000028,
      -0.00000000007,    -0.00000000006,
      -0.00000000029,    -0.00000000025,
      -0.00000000007,     0.00000000007,
       0.00000000044,    -0.00000000050,
      -0.00000000012,    -0.00000000011,
      -0.00000000001,     0.00000000002,
       0.00000001849,     0.00000001623,
       0.00000000440,    -0.00000000489,
      -0.00000000056,    -0.00000000051,
      -0.00000000598,    -0.00000000250,
      -0.00000000060,     0.00000000137,
       0.00000000012,     0.00000000006,
      -0.00000000281,     0.00000000624,
       0.00000000144,     0.00000000069,
       0.00000000007,    -0.00000000012,
      -0.00000000012,    -0.00000000034,
      -0.00000000005,     0.00000000003,
      -0.00000000051,     0.00000000007,
       0.00000000002,     0.00000000010,
      -0.00000000090,    -0.00000000079,
       0.00000000052,     0.00000000022,
       0.00000000025,    -0.00000000054,
       0.00000000065,     0.00000000029,
       0.00000000007,    -0.00000000014,
      -0.00000000001,    -0.00000000001,
       0.00000000531,    -0.00000001209,
      -0.00000000279,    -0.00000000128,
      -0.00000000013,     0.00000000028,
       0.00000000056,    -0.00000000598,
      -0.00000000117,    -0.00000000013,
      -0.00000000001,     0.00000000008,
      -0.00000000635,    -0.00000000080,
      -0.00000000020,     0.00000000126,
       0.00000000009,     0.00000000002,
       0.00000000023,    -0.00000000018,
      -0.00000000003,    -0.00000000003,
      -0.00000000017,    -0.00000000040,
      -0.00000000006,     0.00000000003,
      -0.00000000027,     0.00000000062,
      -0.00000000006,     0.00000000053,
       0.00000000055,     0.00000000007,
       0.00000000012,    -0.00000000106,
      -0.00000000020,    -0.00000000003,
       0.00000000000,     0.00000000002,
      -0.00000001161,    -0.00000000130,
      -0.00000000031,     0.00000000229,
       0.00000000016,     0.00000000003,
      -0.00000000749,     0.00000000156,
       0.00000000022,     0.00000000122,
       0.00000000007,    -0.00000000001,
       0.00000000136,     0.00000000814,
       0.00000000136,    -0.00000000017,
       0.00000000000,    -0.00000000008,
      -0.00000000025,    -0.00000000017,
      -0.00000000040,     0.00000000031,
       0.00000000004,     0.00000000005,
       0.00000000060,     0.00000000007,
       0.00000000067,    -0.00000000013,
      -0.00000000012,    -0.00000000071,
       0.00000000003,    -0.00000000003,
      -0.00000000218,     0.00000000040,
       0.00000000005,     0.00000000036,
       0.00000000003,     0.00000000000,
       0.00000000350,     0.00000001861,
       0.00000000307,    -0.00000000048,
      -0.00000000003,    -0.00000000022,
       0.00000000628,     0.00000001141,
       0.00000000148,    -0.00000000076,
      -0.00000000003,    -0.00000000006,
       0.00000001291,    -0.00000000626,
      -0.00000000074,    -0.00000000176,
      -0.00000000008,     0.00000000003,
      -0.00000000012,     0.00000000039,
       0.00000000062,     0.00000000048,
       0.00000000005,    -0.00000000006,
       0.00000000040,    -0.00000000007,
      -0.00000000018,    -0.00000000096,
      -0.00000000056,    -0.00000000105,
      -0.00000000113,     0.00000000055,
       0.00000000223,     0.00000000437,
       0.00000000058,    -0.00000000027,
      -0.00000000002,    -0.00000000003,
       0.00000003126,    -0.00000001621,
      -0.00000000196,    -0.00000000416,
      -0.00000000024,     0.00000000011,
       0.00000001556,    -0.00000001607,
      -0.00000000146,    -0.00000000148,
      -0.00000000004,     0.00000000004,
      -0.00000001697,    -0.00000001899,
      -0.00000000203,     0.00000000158,
       0.00000000004,     0.00000000007,
       0.00000000048,    -0.00000000003,
       0.00000000054,    -0.00000000110,
      -0.00000000008,    -0.00000000004,
      -0.00000000037,    -0.00000000073,
      -0.00000000159,     0.00000000083,
      -0.00000000149,     0.00000000147,
       0.00000000151,     0.00000000166,
       0.00000000758,    -0.00000000720,
      -0.00000000067,    -0.00000000077,
      -0.00000000003,     0.00000000003,
      -0.00000004508,    -0.00000004688,
      -0.00000000474,     0.00000000415,
       0.00000000018,     0.00000000020,
      -0.00000003010,    -0.00000001503,
      -0.00000000083,     0.00000000175,
      -0.00000002221,     0.00000003430,
       0.00000000223,     0.00000000176,
       0.00000000005,    -0.00000000004,
      -0.00000000023,    -0.00000000028,
      -0.00000000165,    -0.00000000051,
      -0.00000000003,     0.00000000009,
      -0.00000000113,     0.00000000107,
       0.00000000009,     0.00000000011,
       0.00000000226,     0.00000000234,
       0.00000000288,     0.00000000157,
       0.00000000190,    -0.00000000306,
      -0.00000001698,    -0.00000000990,
      -0.00000000069,     0.00000000105,
       0.00000000003,     0.00000000002,
      -0.00000005355,     0.00000009291,
       0.00000000572,     0.00000000372,
       0.00000000010,    -0.00000000017,
      -0.00000000406,     0.00000003567,
       0.00000000069,    -0.00000000026,
       0.00000004873,     0.00000001679,
       0.00000000100,    -0.00000000186,
      -0.00000000003,    -0.00000000002,
       0.00000000043,     0.00000000017,
       0.00000000000,    -0.00000000005,
      -0.00000000051,     0.00000000202,
       0.00000000008,     0.00000000002,
       0.00000000220,     0.00000000128,
       0.00000000008,    -0.00000000013,
       0.00000000261,    -0.00000000454,
       0.00000000075,    -0.00000000374,
      -0.00000000425,    -0.00000000127,
      -0.00000000684,     0.00000002724,
       0.00000000083,     0.00000000025,
       0.00000013133,     0.00000003280,
       0.00000000120,    -0.00000000397,
       0.00000000681,    -0.00000000974,
      -0.00000000067,     0.00000000101,
       0.00000000700,    -0.00000003703,
      -0.00000000081,    -0.00000000047,
      -0.00000000025,    -0.00000000119,
       0.00000000203,     0.00000000064,
       0.00000000001,    -0.00000000008,
       0.00000000076,    -0.00000000303,
      -0.00000000618,    -0.00000000153,
      -0.00000000174,     0.00000000039,
      -0.00000000019,     0.00000000278,
       0.00000001847,    -0.00000000094,
      -0.00000000006,     0.00000000007,
      -0.00000000347,    -0.00000008146,
       0.00000000026,     0.00000000023,
      -0.00000000080,     0.00000004576,
       0.00000000009,     0.00000000037,
      -0.00000000172,    -0.00000001085,
       0.00000000001,     0.00000000132,
      -0.00000000092,    -0.00000000015,
      -0.00000000003,     0.00000000002,
       0.00000000050,    -0.00000000167,
      -0.00000000004,    -0.00000000001,
      -0.00000000171,     0.00000000009,
       0.00000000017,     0.00000000356,
      -0.00000000027,    -0.00000000259,
      -0.00000000121,     0.00000000061,
       0.00000000469,     0.00000002003,
      -0.00000000057,     0.00000000017,
       0.00000006458,    -0.00000001452,
       0.00000000056,     0.00000000178,
       0.00000000876,     0.00000000316,
      -0.00000000027,    -0.00000000033,
       0.00000000429,     0.00000002181,
      -0.00000000042,     0.00000000016,
      -0.00000000103,     0.00000000056,
       0.00000000007,     0.00000000000,
      -0.00000000046,    -0.00000000098,
      -0.00000000001,     0.00000000007,
      -0.00000000035,    -0.00000000149,
      -0.00000000299,     0.00000000067,
      -0.00000000029,    -0.00000000086,
      -0.00000000722,     0.00000000385,
      -0.00000000023,    -0.00000000043,
       0.00000000942,     0.00000001663,
      -0.00000000103,     0.00000000057,
       0.00000000307,     0.00000001649,
      -0.00000000037,     0.00000000004,
      -0.00000001843,     0.00000000508,
      -0.00000000023,    -0.00000000059,
       0.00000000069,     0.00000000018,
      -0.00000000017,     0.00000000069,
       0.00000000040,    -0.00000000021,
      -0.00000000046,    -0.00000000081,
       0.00000000040,    -0.00000000014,
      -0.00000000131,    -0.00000000145,
       0.00000000012,    -0.00000000011,
      -0.00000000214,     0.00000000211,
      -0.00000000017,    -0.00000000019,
      -0.00000000609,     0.00000000296,
      -0.00000000016,    -0.00000000032,
      -0.00000000322,    -0.00000000621,
       0.00000000033,    -0.00000000019,
      -0.00000000009,     0.00000000037,
      -0.00000000037,    -0.00000000009,
       0.00000000032,    -0.00000000032,
      -0.00000000005,     0.00000000000,
       0.00000000001,     0.00000000009,
       0.00000000002,     0.00000000015,
      -0.00000000016,     0.00000000002,
      -0.00000000010,     0.00000000001,
       0.00000000000,    -0.00000000010,
       0.00000000062,    -0.00000000074,
       0.00000000091,     0.00000000075,
      -0.00000000001,     0.00000000010,
       0.00000000002,     0.00000000010,
       0.00000000010,    -0.00000000002,
       0.00000000001,     0.00000000006,
       0.00000000037,    -0.00000000009,
       0.00000000027,    -0.00000000019,
      -0.00000000019,    -0.00000000032,
       0.00000000021,    -0.00000000013,
       0.00000000000,    -0.00000000002,
      -0.00000000075,    -0.00000000112,
      -0.00000000009,     0.00000000003,
      -0.00000000078,    -0.00000000052,
      -0.00000000003,     0.00000000003,
      -0.00000000070,     0.00000000084,
       0.00000000003,     0.00000000005,
      -0.00000000005,     0.00000000001,
      -0.00000000066,    -0.00000000051,
      -0.00000000003,     0.00000000002,
      -0.00000000242,     0.00000000328,
       0.00000000012,     0.00000000015,
      -0.00000000034,     0.00000000187,
       0.00000000231,     0.00000000093,
       0.00000000005,    -0.00000000006,
       0.00000000002,     0.00000000009,
      -0.00000000063,     0.00000000205,
       0.00000000005,     0.00000000003,
       0.00000000891,     0.00000000269,
       0.00000000012,    -0.00000000021,
       0.00000000117,    -0.00000000087,
      -0.00000000007,     0.00000000003,
       0.00000000052,    -0.00000000320,
      -0.00000000005,    -0.00000000002,
      -0.00000000009,    -0.00000000011,
       0.00000000012,    -0.00000000001,
       0.00000000271,    -0.00000000020,
      -0.00000000003,    -0.00000000001,
      -0.00000000059,    -0.00000001053,
      -0.00000000003,     0.00000000010,
      -0.00000000012,     0.00000000593,
      -0.00000000001,     0.00000000005,
       0.00000000007,    -0.00000000156,
      -0.00000000001,     0.00000000014,
      -0.00000000007,     0.00000000007,
      -0.00000000003,    -0.00000000012,
       0.00000000099,     0.00000000408,
      -0.00000000010,     0.00000000003,
       0.00000001216,    -0.00000000285,
       0.00000000009,     0.00000000031,
       0.00000000208,     0.00000000056,
      -0.00000000006,    -0.00000000008,
       0.00000000095,     0.00000000481,
      -0.00000000005,     0.00000000000,
      -0.00000000025,     0.00000000000,
       0.00000000003,    -0.00000000023,
      -0.00000000221,     0.00000000116,
      -0.00000000006,    -0.00000000011,
       0.00000000266,     0.00000000490,
      -0.00000000024,     0.00000000012,
       0.00000000092,     0.00000000499,
      -0.00000000011,     0.00000000002,
      -0.00000000562,     0.00000000157,
      -0.00000000005,    -0.00000000016,
       0.00000000009,     0.00000000002,
      -0.00000000002,     0.00000000010,
      -0.00000000060,    -0.00000000066,
       0.00000000005,    -0.00000000004,
      -0.00000000097,     0.00000000097,
      -0.00000000006,    -0.00000000007,
      -0.00000000269,     0.00000000129,
      -0.00000000007,    -0.00000000013,
      -0.00000000143,    -0.00000000276,
       0.00000000014,    -0.00000000007,
      -0.00000000007,     0.00000000011,
      -0.00000000010,    -0.00000000007,
       0.00000000027,    -0.00000000031,
      -0.00000000030,    -0.00000000024,
      -0.00000000002,    -0.00000000002,
      -0.00000000003,     0.00000000004,
      -0.00000000003,     0.00000000006,
      -0.00000000006,    -0.00000000003,
      -0.00000000005,    -0.00000000002,
       0.00000000002,    -0.00000000005,
       0.00000000097,    -0.00000000117,
       0.00000000109,     0.00000000091,
       0.00000000004,     0.00000000003,
       0.00000000000,     0.00000000002,
       0.00000000007,    -0.00000000001,
       0.00000000000,     0.00000000002,
       0.00000000006,    -0.00000000002,
       0.00000000000,     0.00000000006,
      -0.00000000001,     0.00000000004,
       0.00000000012,     0.00000000003,
      -0.00000000004,     0.00000000003,
      -0.00000000005,     0.00000000001,
       0.00000000001,     0.00000000010,
      -0.00000000002,     0.00000000009,
      -0.00000000010,    -0.00000000001,
      -0.00000000001,    -0.00000000002,
      -0.00000000004,     0.00000000002,
      -0.00000000008,     0.00000000003,
      -0.00000000003,    -0.00000000009,
       0.00000000030,    -0.00000000036,
      -0.00000000035,    -0.00000000029,
       0.00000000179,    -0.00000000214,
       0.00000000023,     0.00000000018,
       0.00000000136,     0.00000000113,
       0.00000000002,     0.00000000002,
       0.00000000005,    -0.00000000006,
       0.00000000002,     0.00000000003,
       0.00000000004,     0.00000000010,
       0.00000000027,    -0.00000000015,
       0.00000000007,     0.00000000014,
      -0.00000000004,     0.00000000012,
      -0.00000000015,     0.00000000038,
       0.00000000100,     0.00000000026,
       0.00000000020,     0.00000000054,
      -0.00000000051,     0.00000000045,
      -0.00000000107,     0.00000000018,
      -0.00000000002,    -0.00000000006,
       0.00000000050,     0.00000000212,
      -0.00000000012,     0.00000000004,
      -0.00000000050,     0.00000000245,
      -0.00000000007,     0.00000000000,
      -0.00000000276,    -0.00000000034,
      -0.00000000002,    -0.00000000010,
      -0.00000000108,    -0.00000000146,
       0.00000000011,    -0.00000000009,
      -0.00000000214,     0.00000000156,
      -0.00000000012,    -0.00000000016,
      -0.00000000568,     0.00000000252,
      -0.00000000013,    -0.00000000028,
      -0.00000000273,    -0.00000000587,
       0.00000000030,    -0.00000000015,
       0.00000000053,    -0.00000000064,
      -0.00000000064,    -0.00000000053,
       0.00000000001,     0.00000000004,
       0.00000000005,    -0.00000000001,
       0.00000000014,    -0.00000000001,
       0.00000000001,     0.00000000014,
       0.00000001105,    -0.00000001328,
       0.00000000254,     0.00000000211,
      -0.00000000003,     0.00000000002,
       0.00000000004,     0.00000000006,
       0.00000000004,     0.00000000010,
      -0.00000000011,     0.00000000004,
      -0.00000000005,    -0.00000000007,
      -0.00000000010,     0.00000000008,
      -0.00000000027,     0.00000000013,
      -0.00000000014,    -0.00000000028,
       0.00000000071,    -0.00000000086,
      -0.00000000086,    -0.00000000071,
       0.00000002742,    -0.00000003305,
       0.00000000345,     0.00000000286,
       0.00000000102,    -0.00000000122,
      -0.00000000122,    -0.00000000101,
       0.00000007830,    -0.00000009396,
       0.00000000489,     0.00000000407,
       0.00000000127,    -0.00000000153,
      -0.00000000153,    -0.00000000127,
       0.00000015348,    -0.00000018476,
       0.00000001952,     0.00000001501,
       0.00000000612,     0.00000000509,
      -0.00000000050,     0.00000000065,
       0.00000000001,    -0.00000000004,
      -0.00000000018,    -0.00000000009,
       0.00000000009,    -0.00000000019,
       0.00000000076,     0.00000000039,
      -0.00000000039,     0.00000000076,
       0.00000000108,     0.00000000055,
      -0.00000000056,     0.00000000107,
       0.00000000135,     0.00000000068,
      -0.00000000070,     0.00000000134,
      -0.00000000002,     0.00000000004,
       0.00000000003,    -0.00000000015,
      -0.00000000002,     0.00000000001,
      -0.00000000005,    -0.00000000018,
      -0.00000000006,     0.00000000001,
       0.00000000013,     0.00000000019,
       0.00000000005,    -0.00000000003,
       0.00000000020,    -0.00000000013,
      -0.00000000004,    -0.00000000006,
      -0.00000000009,    -0.00000000014,
      -0.00000000004,     0.00000000002,
      -0.00000000518,     0.00000000331,
       0.00000000098,     0.00000000156,
       0.00000000018,    -0.00000000011,
      -0.00000000097,     0.00000000120,
       0.00000000029,     0.00000000024,
       0.00000000003,    -0.00000000003,
       0.00000000124,     0.00000000106,
       0.00000000027,    -0.00000000030,
      -0.00000000003,    -0.00000000003,
      -0.00000000008,     0.00000000001,
      -0.00000000002,     0.00000000012,
      -0.00000000009,     0.00000000011,
       0.00000000003,     0.00000000002,
       0.00000000146,     0.00000000122,
       0.00000000031,    -0.00000000036,
      -0.00000000004,    -0.00000000003,
       0.00000000109,     0.00000000043,
       0.00000000010,    -0.00000000023,
       0.00000000050,    -0.00000000115,
      -0.00000000025,    -0.00000000012,
       0.00000000002,     0.00000000005,
       0.00000000008,    -0.00000000001,
       0.00000000026,     0.00000000011,
       0.00000000003,    -0.00000000006,
       0.00000000102,    -0.00000000246,
      -0.00000000056,    -0.00000000025,
      -0.00000000002,     0.00000000005,
       0.00000000014,    -0.00000000194,
      -0.00000000035,    -0.00000000004,
      -0.00000000209,    -0.00000000025,
      -0.00000000006,     0.00000000039,
       0.00000000006,    -0.00000000005,
      -0.00000000005,    -0.00000000012,
       0.00000000006,    -0.00000000062,
      -0.00000000011,    -0.00000000002,
      -0.00000000491,    -0.00000000047,
      -0.00000000012,     0.00000000096,
       0.00000000007,     0.00000000001,
      -0.00000000356,     0.00000000081,
       0.00000000012,     0.00000000057,
       0.00000000068,     0.00000000398,
       0.00000000067,    -0.00000000008,
      -0.00000000010,    -0.00000000007,
      -0.00000000018,     0.00000000014,
      -0.00000000140,     0.00000000027,
       0.00000000003,     0.00000000022,
       0.00000000193,     0.00000000963,
       0.00000000158,    -0.00000000026,
      -0.00000000001,    -0.00000000009,
       0.00000000340,     0.00000000588,
       0.00000000075,    -0.00000000041,
       0.00000000695,    -0.00000000337,
      -0.00000000040,    -0.00000000096,
      -0.00000000005,     0.00000000001,
      -0.00000000004,     0.00000000017,
       0.00000000030,     0.00000000025,
       0.00000000003,    -0.00000000003,
      -0.00000000010,    -0.00000000049,
      -0.00000000031,    -0.00000000056,
      -0.00000000061,     0.00000000030,
       0.00000000146,     0.00000000280,
       0.00000000037,    -0.00000000017,
      -0.00000000001,    -0.00000000002,
       0.00000001704,    -0.00000000904,
      -0.00000000108,    -0.00000000226,
      -0.00000000010,     0.00000000004,
       0.00000000777,    -0.00000000844,
      -0.00000000076,    -0.00000000071,
      -0.00000000908,    -0.00000001025,
      -0.00000000111,     0.00000000084,
       0.00000000002,     0.00000000004,
       0.00000000019,    -0.00000000004,
       0.00000000030,    -0.00000000054,
      -0.00000000004,    -0.00000000002,
      -0.00000000023,    -0.00000000043,
      -0.00000000086,     0.00000000046,
      -0.00000000077,     0.00000000079,
       0.00000000081,     0.00000000089,
       0.00000000459,    -0.00000000442,
      -0.00000000041,    -0.00000000046,
      -0.00000000002,     0.00000000002,
      -0.00000002433,    -0.00000002493,
      -0.00000000251,     0.00000000223,
       0.00000000010,     0.00000000011,
      -0.00000001438,    -0.00000000646,
      -0.00000000029,     0.00000000080,
      -0.00000001149,     0.00000001730,
       0.00000000114,     0.00000000094,
       0.00000000003,    -0.00000000002,
      -0.00000000012,    -0.00000000004,
      -0.00000000080,    -0.00000000032,
      -0.00000000002,     0.00000000005,
      -0.00000000064,     0.00000000061,
       0.00000000121,     0.00000000124,
       0.00000000142,     0.00000000074,
       0.00000000095,    -0.00000000154,
      -0.00000000936,    -0.00000000535,
      -0.00000000037,     0.00000000058,
       0.00000000002,     0.00000000001,
      -0.00000002620,     0.00000004625,
       0.00000000283,     0.00000000180,
       0.00000000004,    -0.00000000005,
      -0.00000000002,     0.00000001363,
       0.00000000014,    -0.00000000027,
       0.00000002209,     0.00000000828,
       0.00000000054,    -0.00000000091,
       0.00000000032,     0.00000000008,
      -0.00000000033,     0.00000000095,
       0.00000000005,     0.00000000002,
       0.00000000116,     0.00000000066,
       0.00000000127,    -0.00000000225,
       0.00000000026,    -0.00000000157,
      -0.00000000188,    -0.00000000057,
      -0.00000000299,     0.00000001259,
       0.00000000037,     0.00000000010,
       0.00000005635,     0.00000001342,
       0.00000000044,    -0.00000000167,
      -0.00000000354,    -0.00000000561,
      -0.00000000039,     0.00000000059,
       0.00000000406,    -0.00000001401,
      -0.00000000047,    -0.00000000026,
      -0.00000000006,    -0.00000000059,
       0.00000000091,     0.00000000038,
       0.00000000001,    -0.00000000004,
       0.00000000032,    -0.00000000136,
      -0.00000000264,    -0.00000000062,
      -0.00000000031,     0.00000000022,
      -0.00000000010,     0.00000000089,
       0.00000000503,    -0.00000000057,
      -0.00000000004,     0.00000000005,
      -0.00000000204,    -0.00000002227,
       0.00000000016,     0.00000000016,
      -0.00000000009,     0.00000002277,
       0.00000000002,     0.00000000017,
      -0.00000000034,    -0.00000000538,
       0.00000000002,     0.00000000066,
      -0.00000000038,    -0.00000000021,
       0.00000000031,    -0.00000000065,
      -0.00000000047,     0.00000000005,
       0.00000000010,     0.00000000093,
      -0.00000000016,    -0.00000000133,
      -0.00000000067,     0.00000000032,
       0.00000000253,     0.00000001048,
      -0.00000000031,     0.00000000009,
       0.00000003366,    -0.00000000778,
       0.00000000029,     0.00000000095,
      -0.00000000051,     0.00000000223,
      -0.00000000019,    -0.00000000024,
       0.00000000302,     0.00000000909,
      -0.00000000030,     0.00000000011,
      -0.00000000053,     0.00000000024,
      -0.00000000014,    -0.00000000049,
      -0.00000000020,    -0.00000000081,
      -0.00000000155,     0.00000000036,
      -0.00000000022,    -0.00000000049,
      -0.00000000518,     0.00000000272,
      -0.00000000017,    -0.00000000030,
       0.00000000686,     0.00000001265,
      -0.00000000076,     0.00000000042,
       0.00000000130,     0.00000000861,
      -0.00000000017,    -0.00000000002,
      -0.00000001070,     0.00000000333,
      -0.00000000015,    -0.00000000035,
       0.00000000025,     0.00000000000,
       0.00000000001,     0.00000000027,
       0.00000000032,    -0.00000000017,
      -0.00000000033,    -0.00000000061,
      -0.00000000126,    -0.00000000140,
       0.00000000012,    -0.00000000010,
      -0.00000000244,     0.00000000232,
      -0.00000000019,    -0.00000000021,
      -0.00000000466,     0.00000000221,
      -0.00000000011,    -0.00000000024,
      -0.00000000264,    -0.00000000490,
       0.00000000027,    -0.00000000016,
      -0.00000000004,     0.00000000023,
      -0.00000000023,    -0.00000000004,
       0.00000000023,    -0.00000000034,
       0.00000000004,     0.00000000003,
      -0.00000000041,    -0.00000000024,
       0.00000000003,    -0.00000000005,
      -0.00000000104,    -0.00000000124,
       0.00000000010,    -0.00000000008,
       0.00000000125,    -0.00000000109,
       0.00000000009,     0.00000000010,
      -0.00000000009,     0.00000000001,
       0.00000000000,    -0.00000000009,
      -0.00000000001,     0.00000000005,
       0.00000000001,     0.00000000004,
       0.00000000005,    -0.00000000001,
       0.00000000001,     0.00000000003,
       0.00000000017,    -0.00000000004,
       0.00000000011,    -0.00000000008,
      -0.00000000008,    -0.00000000014,
       0.00000000010,    -0.00000000006,
      -0.00000000032,    -0.00000000048,
      -0.00000000004,     0.00000000001,
      -0.00000000030,    -0.00000000018,
      -0.00000000028,     0.00000000033,
      -0.00000000028,    -0.00000000021,
      -0.00000000001,     0.00000000001,
      -0.00000000091,     0.00000000126,
       0.00000000005,     0.00000000006,
      -0.00000000005,     0.00000000058,
       0.00000000079,     0.00000000034,
      -0.00000000021,     0.00000000072,
       0.00000000002,     0.00000000001,
       0.00000000293,     0.00000000084,
       0.00000000003,    -0.00000000007,
       0.00000000004,    -0.00000000038,
       0.00000000021,    -0.00000000090,
      -0.00000000003,    -0.00000000004,
       0.00000000062,    -0.00000000009,
      -0.00000000028,    -0.00000000237,
      -0.00000000001,     0.00000000004,
       0.00000000000,     0.00000000217,
       0.00000000008,    -0.00000000057,
       0.00000000000,     0.00000000005,
       0.00000000040,     0.00000000155,
      -0.00000000004,     0.00000000001,
       0.00000000458,    -0.00000000112,
       0.00000000003,     0.00000000012,
       0.00000000009,     0.00000000028,
      -0.00000000003,    -0.00000000004,
       0.00000000048,     0.00000000150,
      -0.00000000009,     0.00000000000,
       0.00000000001,    -0.00000000008,
      -0.00000000111,     0.00000000058,
      -0.00000000003,    -0.00000000005,
       0.00000000138,     0.00000000259,
      -0.00000000013,     0.00000000007,
       0.00000000030,     0.00000000187,
      -0.00000000233,     0.00000000074,
      -0.00000000003,    -0.00000000007,
      -0.00000000040,    -0.00000000045,
       0.00000000003,    -0.00000000003,
      -0.00000000076,     0.00000000072,
      -0.00000000005,    -0.00000000005,
      -0.00000000145,     0.00000000069,
      -0.00000000003,    -0.00000000007,
      -0.00000000083,    -0.00000000153,
       0.00000000008,    -0.00000000004,
      -0.00000000002,     0.00000000004,
      -0.00000000004,    -0.00000000002,
       0.00000000011,    -0.00000000016,
       0.00000000001,     0.00000000001,
      -0.00000000020,    -0.00000000011,
      -0.00000000047,    -0.00000000056,
       0.00000000004,    -0.00000000003,
       0.00000000057,    -0.00000000050,
       0.00000000004,     0.00000000004,
       0.00000000034,     0.00000000018,
       0.00000000063,     0.00000000032,
      -0.00000000021,     0.00000000040,
      -0.00000000002,     0.00000000003,
       0.00000000008,     0.00000000005,
       0.00000000000,     0.00000000006,
      -0.00000000006,     0.00000000002,
      -0.00000000012,    -0.00000000001,
       0.00000000001,     0.00000000024,
      -0.00000000013,     0.00000000020,
      -0.00000000025,    -0.00000000011,
      -0.00000000010,    -0.00000000024,
       0.00000000002,    -0.00000000001,
      -0.00000000037,     0.00000000018,
      -0.00000000076,     0.00000000007,
      -0.00000000001,    -0.00000000004,
      -0.00000000012,    -0.00000000081,
       0.00000000005,    -0.00000000001,
       0.00000000026,    -0.00000000030,
       0.00000000003,     0.00000000003,
      -0.00000000033,    -0.00000000029,
       0.00000000003,    -0.00000000003,
      -0.00000000091,    -0.00000000124,
       0.00000000009,    -0.00000000007,
       0.00000000126,    -0.00000000096,
       0.00000000008,     0.00000000010,
       0.00000000389,     0.00000000200,
      -0.00000000038,     0.00000000074,
       0.00000000001,    -0.00000000001,
      -0.00000000005,    -0.00000000006,
       0.00000000006,    -0.00000000005,
       0.00000000966,     0.00000000495,
      -0.00000000052,     0.00000000101,
       0.00000002752,     0.00000001418,
      -0.00000000074,     0.00000000143,
       0.00000005401,     0.00000002773,
      -0.00000000092,     0.00000000179,
       0.00000000002,    -0.00000000006,
       0.00000000006,     0.00000000001,
       0.00000000000,    -0.00000000005,
       0.00000000355,     0.00000000007,
       0.00000000004,    -0.00000000130,
      -0.00000000020,    -0.00000000001,
      -0.00000000042,     0.00000000013,
       0.00000000004,     0.00000000013,
       0.00000000013,     0.00000000045,
       0.00000000014,    -0.00000000004,
      -0.00000000002,     0.00000000003,
       0.00000000002,    -0.00000000001,
      -0.00000000011,    -0.00000000036,
      -0.00000000011,     0.00000000003,
      -0.00000000011,    -0.00000000016,
      -0.00000000004,     0.00000000003,
      -0.00000000017,     0.00000000011,
       0.00000000003,     0.00000000005,
      -0.00000000002,    -0.00000000003,
      -0.00000000032,     0.00000000021,
       0.00000000006,     0.00000000009,
      -0.00000000017,     0.00000000022,
       0.00000000005,     0.00000000004,
       0.00000000023,     0.00000000019,
       0.00000000005,    -0.00000000006,
      -0.00000000005,     0.00000000006,
       0.00000000001,     0.00000000001,
       0.00000000052,     0.00000000042,
       0.00000000011,    -0.00000000013,
       0.00000000049,     0.00000000018,
       0.00000000004,    -0.00000000010,
       0.00000000022,    -0.00000000052,
      -0.00000000011,    -0.00000000005,
       0.00000000017,     0.00000000007,
       0.00000000002,    -0.00000000004,
       0.00000000049,    -0.00000000124,
      -0.00000000026,    -0.00000000012,
       0.00000000006,    -0.00000000105,
      -0.00000000019,    -0.00000000002,
      -0.00000000116,    -0.00000000013,
      -0.00000000004,     0.00000000021,
       0.00000000003,    -0.00000000002,
      -0.00000000002,    -0.00000000006,
       0.00000000004,    -0.00000000042,
      -0.00000000008,    -0.00000000001,
      -0.00000000280,    -0.00000000023,
      -0.00000000006,     0.00000000051,
      -0.00000000201,     0.00000000050,
       0.00000000007,     0.00000000030,
       0.00000000040,     0.00000000232,
       0.00000000036,    -0.00000000004,
      -0.00000000005,    -0.00000000003,
      -0.00000000010,     0.00000000007,
      -0.00000000094,     0.00000000019,
       0.00000000002,     0.00000000015,
       0.00000000119,     0.00000000564,
       0.00000000093,    -0.00000000016,
      -0.00000000001,    -0.00000000005,
       0.00000000196,     0.00000000323,
       0.00000000038,    -0.00000000022,
       0.00000000402,    -0.00000000194,
      -0.00000000023,    -0.00000000056,
      -0.00000000001,     0.00000000008,
       0.00000000016,     0.00000000014,
      -0.00000000035,     0.00000000018,
       0.00000000096,     0.00000000179,
       0.00000000024,    -0.00000000011,
      -0.00000000001,    -0.00000000001,
       0.00000000976,    -0.00000000529,
      -0.00000000063,    -0.00000000129,
      -0.00000000006,     0.00000000002,
       0.00000000391,    -0.00000000452,
      -0.00000000040,    -0.00000000034,
      -0.00000000497,    -0.00000000569,
      -0.00000000063,     0.00000000046,
       0.00000000007,    -0.00000000003,
       0.00000000018,    -0.00000000028,
      -0.00000000014,    -0.00000000026,
      -0.00000000049,     0.00000000027,
      -0.00000000041,     0.00000000043,
       0.00000000045,     0.00000000049,
       0.00000000270,    -0.00000000265,
      -0.00000000024,    -0.00000000027,
      -0.00000000001,     0.00000000001,
      -0.00000001331,    -0.00000001342,
      -0.00000000135,     0.00000000121,
       0.00000000003,     0.00000000004,
      -0.00000000677,    -0.00000000259,
      -0.00000000006,     0.00000000033,
      -0.00000000598,     0.00000000874,
       0.00000000059,     0.00000000051,
      -0.00000000007,     0.00000000003,
      -0.00000000039,    -0.00000000019,
      -0.00000000036,     0.00000000035,
       0.00000000066,     0.00000000066,
       0.00000000069,     0.00000000034,
       0.00000000048,    -0.00000000077,
      -0.00000000498,    -0.00000000278,
      -0.00000000019,     0.00000000030,
      -0.00000001267,     0.00000002285,
       0.00000000139,     0.00000000085,
       0.00000000103,     0.00000000462,
      -0.00000000003,    -0.00000000021,
       0.00000000988,     0.00000000408,
       0.00000000029,    -0.00000000044,
       0.00000000020,     0.00000000004,
      -0.00000000020,     0.00000000045,
       0.00000000059,     0.00000000033,
       0.00000000061,    -0.00000000110,
       0.00000000007,    -0.00000000062,
      -0.00000000081,    -0.00000000025,
      -0.00000000120,     0.00000000548,
       0.00000000015,     0.00000000003,
       0.00000002332,     0.00000000516,
       0.00000000014,    -0.00000000067,
      -0.00000000461,    -0.00000000303,
      -0.00000000020,     0.00000000030,
       0.00000000222,    -0.00000000505,
      -0.00000000024,    -0.00000000013,
       0.00000000001,    -0.00000000028,
       0.00000000040,     0.00000000022,
       0.00000000013,    -0.00000000058,
      -0.00000000108,    -0.00000000023,
       0.00000000058,    -0.00000000031,
      -0.00000000003,     0.00000000003,
      -0.00000000111,    -0.00000000348,
       0.00000000009,     0.00000000010,
       0.00000000014,     0.00000001046,
      -0.00000000001,     0.00000000007,
       0.00000000005,    -0.00000000245,
       0.00000000002,     0.00000000028,
      -0.00000000015,    -0.00000000017,
       0.00000000018,    -0.00000000024,
      -0.00000000009,    -0.00000000063,
       0.00000000127,     0.00000000503,
      -0.00000000014,     0.00000000004,
       0.00000001615,    -0.00000000388,
       0.00000000013,     0.00000000043,
      -0.00000000272,     0.00000000145,
      -0.00000000012,    -0.00000000015,
       0.00000000192,     0.00000000353,
      -0.00000000019,     0.00000000008,
      -0.00000000025,     0.00000000010,
      -0.00000000003,    -0.00000000023,
      -0.00000000010,    -0.00000000040,
      -0.00000000074,     0.00000000018,
      -0.00000000333,     0.00000000175,
      -0.00000000011,    -0.00000000020,
       0.00000000449,     0.00000000845,
      -0.00000000050,     0.00000000028,
       0.00000000034,     0.00000000389,
      -0.00000000006,    -0.00000000004,
      -0.00000000572,     0.00000000206,
      -0.00000000011,    -0.00000000021,
       0.00000000008,    -0.00000000005,
       0.00000000005,     0.00000000010,
      -0.00000000022,    -0.00000000041,
      -0.00000000103,    -0.00000000115,
       0.00000000009,    -0.00000000009,
      -0.00000000222,     0.00000000206,
      -0.00000000017,    -0.00000000019,
      -0.00000000307,     0.00000000139,
      -0.00000000007,    -0.00000000015,
      -0.00000000189,    -0.00000000337,
       0.00000000019,    -0.00000000012,
      -0.00000000001,     0.00000000013,
      -0.00000000013,    -0.00000000001,
       0.00000000025,    -0.00000000036,
       0.00000000004,     0.00000000003,
      -0.00000000052,    -0.00000000033,
       0.00000000004,    -0.00000000006,
      -0.00000000090,    -0.00000000109,
       0.00000000009,    -0.00000000007,
       0.00000000111,    -0.00000000099,
       0.00000000008,     0.00000000009,
      -0.00000000007,     0.00000000001,
       0.00000000000,    -0.00000000007,
       0.00000000009,     0.00000000003,
       0.00000000002,    -0.00000000008,
       0.00000000023,    -0.00000000031,
       0.00000000003,     0.00000000002,
       0.00000000032,     0.00000000023,
      -0.00000000002,     0.00000000004,
       0.00000000000,     0.00000000002,
       0.00000000008,    -0.00000000002,
       0.00000000004,    -0.00000000004,
      -0.00000000003,    -0.00000000006,
       0.00000000004,    -0.00000000003,
      -0.00000000014,    -0.00000000020,
      -0.00000000011,    -0.00000000006,
      -0.00000000011,     0.00000000013,
      -0.00000000011,    -0.00000000008,
      -0.00000000034,     0.00000000048,
       0.00000000001,     0.00000000017,
       0.00000000027,     0.00000000012,
      -0.00000000006,     0.00000000024,
       0.00000000093,     0.00000000024,
      -0.00000000011,    -0.00000000015,
       0.00000000009,    -0.00000000024,
       0.00000000010,    -0.00000000004,
      -0.00000000012,    -0.00000000042,
       0.00000000002,     0.00000000074,
       0.00000000004,    -0.00000000020,
       0.00000000015,     0.00000000055,
       0.00000000162,    -0.00000000042,
      -0.00000000021,     0.00000000013,
       0.00000000022,     0.00000000045,
      -0.00000000051,     0.00000000027,
      -0.00000000001,    -0.00000000003,
       0.00000000065,     0.00000000122,
      -0.00000000006,     0.00000000003,
       0.00000000007,     0.00000000062,
      -0.00000000091,     0.00000000034,
      -0.00000000023,    -0.00000000026,
       0.00000000002,    -0.00000000002,
      -0.00000000048,     0.00000000044,
      -0.00000000003,    -0.00000000003,
      -0.00000000068,     0.00000000031,
      -0.00000000043,    -0.00000000075,
       0.00000000004,    -0.00000000002,
       0.00000000008,    -0.00000000012,
       0.00000000001,     0.00000000001,
      -0.00000000017,    -0.00000000011,
      -0.00000000029,    -0.00000000035,
       0.00000000036,    -0.00000000032,
       0.00000000002,     0.00000000003,
       0.00000000004,     0.00000000002,
       0.00000000011,    -0.00000000014,
       0.00000000015,     0.00000000011,
      -0.00000000001,    -0.00000000003,
      -0.00000000005,     0.00000000002,
      -0.00000000008,    -0.00000000001,
       0.00000000001,    -0.00000000009,
       0.00000000005,    -0.00000000004,
      -0.00000000005,    -0.00000000006,
      -0.00000000007,    -0.00000000020,
       0.00000000021,    -0.00000000008,
       0.00000000008,     0.00000000004,
       0.00000000004,    -0.00000000007,
       0.00000000025,    -0.00000000028,
       0.00000000003,     0.00000000002,
       0.00000000029,     0.00000000025,
      -0.00000000002,     0.00000000003,
      -0.00000000004,     0.00000000013,
       0.00000000005,     0.00000000002,
       0.00000000006,     0.00000000000,
       0.00000000004,    -0.00000000001,
      -0.00000000001,    -0.00000000004,
      -0.00000000003,    -0.00000000009,
      -0.00000000004,    -0.00000000006,
      -0.00000000006,     0.00000000004,
      -0.00000000001,    -0.00000000002,
      -0.00000000014,     0.00000000010,
       0.00000000003,     0.00000000004,
      -0.00000000009,     0.00000000012,
       0.00000000003,     0.00000000002,
       0.00000000012,     0.00000000010,
       0.00000000003,    -0.00000000003,
      -0.00000000003,     0.00000000004,
       0.00000000030,     0.00000000023,
       0.00000000006,    -0.00000000007,
       0.00000000029,     0.00000000010,
       0.00000000002,    -0.00000000006,
       0.00000000013,    -0.00000000032,
      -0.00000000007,    -0.00000000003,
       0.00000000012,     0.00000000005,
       0.00000000001,    -0.00000000003,
       0.00000000029,    -0.00000000076,
      -0.00000000016,    -0.00000000007,
       0.00000000002,    -0.00000000063,
      -0.00000000011,    -0.00000000001,
      -0.00000000072,    -0.00000000008,
      -0.00000000002,     0.00000000013,
       0.00000000002,    -0.00000000029,
      -0.00000000005,    -0.00000000001,
      -0.00000000174,    -0.00000000012,
      -0.00000000004,     0.00000000032,
      -0.00000000118,     0.00000000032,
       0.00000000004,     0.00000000017,
       0.00000000025,     0.00000000142,
       0.00000000022,    -0.00000000002,
      -0.00000000006,     0.00000000004,
      -0.00000000063,     0.00000000014,
       0.00000000002,     0.00000000010,
       0.00000000076,     0.00000000342,
       0.00000000052,    -0.00000000009,
       0.00000000115,     0.00000000178,
       0.00000000020,    -0.00000000013,
       0.00000000236,    -0.00000000113,
      -0.00000000012,    -0.00000000031,
       0.00000000008,     0.00000000008,
       0.00000000061,     0.00000000112,
       0.00000000014,    -0.00000000007,
       0.00000000563,    -0.00000000311,
      -0.00000000037,    -0.00000000074,
       0.00000000191,    -0.00000000240,
      -0.00000000020,    -0.00000000014,
      -0.00000000272,    -0.00000000315,
      -0.00000000033,     0.00000000023,
       0.00000000010,    -0.00000000014,
       0.00000000024,     0.00000000026,
       0.00000000154,    -0.00000000154,
      -0.00000000014,    -0.00000000015,
      -0.00000000722,    -0.00000000715,
      -0.00000000071,     0.00000000066,
      -0.00000000306,    -0.00000000088,
       0.00000000001,     0.00000000014,
      -0.00000000308,     0.00000000435,
       0.00000000028,     0.00000000025,
      -0.00000000004,     0.00000000004,
      -0.00000000019,    -0.00000000011,
      -0.00000000020,     0.00000000020,
       0.00000000036,     0.00000000035,
       0.00000000023,    -0.00000000038,
      -0.00000000254,    -0.00000000137,
      -0.00000000008,     0.00000000014,
      -0.00000000595,     0.00000001104,
       0.00000000067,     0.00000000039,
       0.00000000103,     0.00000000117,
      -0.00000000006,    -0.00000000014,
       0.00000000429,     0.00000000199,
       0.00000000014,    -0.00000000020,
       0.00000000011,     0.00000000003,
      -0.00000000012,     0.00000000021,
       0.00000000029,     0.00000000016,
       0.00000000028,    -0.00000000053,
      -0.00000000042,     0.00000000220,
       0.00000000005,     0.00000000000,
       0.00000000907,     0.00000000177,
       0.00000000002,    -0.00000000023,
      -0.00000000338,    -0.00000000154,
      -0.00000000010,     0.00000000016,
       0.00000000115,    -0.00000000167,
      -0.00000000012,    -0.00000000006,
       0.00000000002,    -0.00000000013,
       0.00000000017,     0.00000000012,
      -0.00000000056,    -0.00000000016,
      -0.00000000001,     0.00000000001,
      -0.00000000055,     0.00000000138,
       0.00000000005,     0.00000000005,
       0.00000000017,     0.00000000438,
       0.00000000011,    -0.00000000101,
       0.00000000001,     0.00000000012,
      -0.00000000006,    -0.00000000011,
       0.00000000010,    -0.00000000008,
       0.00000000059,     0.00000000221,
      -0.00000000007,     0.00000000002,
       0.00000000712,    -0.00000000180,
       0.00000000005,     0.00000000020,
      -0.00000000247,     0.00000000087,
      -0.00000000007,    -0.00000000009,
       0.00000000112,     0.00000000125,
      -0.00000000011,     0.00000000005,
      -0.00000000011,     0.00000000004,
       0.00000000000,    -0.00000000010,
      -0.00000000198,     0.00000000104,
      -0.00000000006,    -0.00000000011,
       0.00000000272,     0.00000000514,
      -0.00000000028,     0.00000000016,
      -0.00000000008,     0.00000000146,
      -0.00000000286,     0.00000000122,
      -0.00000000007,    -0.00000000012,
       0.00000000002,    -0.00000000005,
       0.00000000005,     0.00000000003,
      -0.00000000075,    -0.00000000085,
       0.00000000007,    -0.00000000006,
      -0.00000000175,     0.00000000160,
      -0.00000000013,    -0.00000000014,
      -0.00000000181,     0.00000000077,
      -0.00000000003,    -0.00000000009,
      -0.00000000125,    -0.00000000212,
       0.00000000012,    -0.00000000008,
       0.00000000000,     0.00000000006,
      -0.00000000007,     0.00000000000,
       0.00000000022,    -0.00000000033,
       0.00000000004,     0.00000000002,
      -0.00000000053,    -0.00000000034,
       0.00000000004,    -0.00000000006,
      -0.00000000067,    -0.00000000083,
       0.00000000006,    -0.00000000005,
       0.00000000086,    -0.00000000079,
       0.00000000007,     0.00000000007,
       0.00000000010,     0.00000000004,
       0.00000000004,    -0.00000000011,
       0.00000000022,    -0.00000000030,
       0.00000000003,     0.00000000002,
       0.00000000032,     0.00000000022,
      -0.00000000002,     0.00000000004,
       0.00000000000,     0.00000000002,
       0.00000000009,     0.00000000004,
      -0.00000000004,     0.00000000009,
       0.00000000002,    -0.00000000001,
      -0.00000000006,    -0.00000000008,
      -0.00000000004,    -0.00000000002,
      -0.00000000004,     0.00000000005,
      -0.00000000004,    -0.00000000003,
      -0.00000000012,     0.00000000018,
       0.00000000001,     0.00000000004,
       0.00000000009,     0.00000000004,
      -0.00000000002,     0.00000000008,
       0.00000000028,     0.00000000006,
      -0.00000000008,    -0.00000000006,
       0.00000000003,    -0.00000000006,
      -0.00000000005,    -0.00000000001,
       0.00000000001,     0.00000000024,
       0.00000000002,    -0.00000000006,
       0.00000000006,     0.00000000018,
       0.00000000054,    -0.00000000015,
      -0.00000000015,     0.00000000006,
       0.00000000010,     0.00000000013,
      -0.00000000022,     0.00000000012,
       0.00000000029,     0.00000000054,
       0.00000000000,     0.00000000018,
      -0.00000000033,     0.00000000015,
      -0.00000000012,    -0.00000000013,
      -0.00000000027,     0.00000000024,
      -0.00000000029,     0.00000000013,
      -0.00000000020,    -0.00000000034,
       0.00000000005,    -0.00000000007,
      -0.00000000012,    -0.00000000008,
      -0.00000000015,    -0.00000000019,
       0.00000000020,    -0.00000000018,
       0.00000000003,     0.00000000001,
       0.00000000007,    -0.00000000010,
       0.00000000010,     0.00000000007,
       0.00000000004,     0.00000000002,
      -0.00000000002,     0.00000000004,
       0.00000000001,     0.00000000001,
       0.00000000005,    -0.00000000003,
       0.00000000003,     0.00000000005,
      -0.00000000001,     0.00000000002,
       0.00000000008,     0.00000000004,
      -0.00000000004,     0.00000000008,
      -0.00000000001,    -0.00000000001,
       0.00000001232,    -0.00000000993,
      -0.00000000454,    -0.00000000573,
      -0.00000000167,     0.00000000166,
      -0.00000000009,     0.00000000014,
       0.00000000005,     0.00000000003,
       0.00000000015,     0.00000000010,
       0.00000000003,    -0.00000000005,
      -0.00000000061,     0.00000000049,
      -0.00000000002,    -0.00000000005,
      -0.00000000003,    -0.00000000003,
      -0.00000000004,     0.00000000003,
      -0.00000000001,    -0.00000000001,
      -0.00000000009,     0.00000000006,
      -0.00000000005,     0.00000000007,
       0.00000000008,     0.00000000006,
      -0.00000000002,     0.00000000003,
       0.00000000019,     0.00000000015,
       0.00000000004,    -0.00000000005,
       0.00000000019,     0.00000000006,
       0.00000000001,    -0.00000000004,
       0.00000000008,    -0.00000000021,
      -0.00000000004,    -0.00000000002,
       0.00000000009,     0.00000000003,
       0.00000000001,    -0.00000000002,
       0.00000000018,    -0.00000000050,
      -0.00000000010,    -0.00000000004,
       0.00000000001,    -0.00000000039,
      -0.00000000007,     0.00000000000,
      -0.00000000046,    -0.00000000005,
      -0.00000000001,     0.00000000008,
       0.00000000001,    -0.00000000020,
      -0.00000000004,     0.00000000000,
      -0.00000000110,    -0.00000000007,
      -0.00000000002,     0.00000000020,
      -0.00000000070,     0.00000000021,
       0.00000000003,     0.00000000010,
       0.00000000015,     0.00000000087,
       0.00000000014,    -0.00000000001,
      -0.00000000004,     0.00000000002,
      -0.00000000041,     0.00000000009,
       0.00000000001,     0.00000000006,
       0.00000000048,     0.00000000208,
       0.00000000032,    -0.00000000006,
       0.00000000067,     0.00000000097,
       0.00000000011,    -0.00000000007,
       0.00000000138,    -0.00000000066,
      -0.00000000007,    -0.00000000018,
       0.00000000004,     0.00000000005,
       0.00000000038,     0.00000000068,
       0.00000000008,    -0.00000000004,
       0.00000000321,    -0.00000000182,
      -0.00000000020,    -0.00000000039,
       0.00000000089,    -0.00000000124,
      -0.00000000010,    -0.00000000006,
      -0.00000000146,    -0.00000000173,
      -0.00000000018,     0.00000000013,
       0.00000000006,    -0.00000000007,
       0.00000000085,    -0.00000000087,
      -0.00000000007,    -0.00000000008,
      -0.00000000385,    -0.00000000373,
      -0.00000000037,     0.00000000035,
      -0.00000000130,    -0.00000000019,
       0.00000000003,     0.00000000005,
      -0.00000000156,     0.00000000212,
       0.00000000014,     0.00000000013,
      -0.00000000002,     0.00000000003,
      -0.00000000009,    -0.00000000007,
      -0.00000000123,    -0.00000000064,
      -0.00000000004,     0.00000000007,
      -0.00000000268,     0.00000000516,
       0.00000000029,     0.00000000016,
       0.00000000075,     0.00000000000,
      -0.00000000006,    -0.00000000008,
       0.00000000179,     0.00000000095,
       0.00000000007,    -0.00000000009,
       0.00000000006,     0.00000000002,
      -0.00000000007,     0.00000000009,
      -0.00000000011,     0.00000000078,
       0.00000000002,     0.00000000000,
       0.00000000318,     0.00000000047,
      -0.00000000001,    -0.00000000007,
      -0.00000000204,    -0.00000000073,
      -0.00000000005,     0.00000000008,
       0.00000000056,    -0.00000000046,
      -0.00000000006,    -0.00000000003,
       0.00000000002,    -0.00000000005,
       0.00000000007,     0.00000000006,
      -0.00000000062,    -0.00000000007,
      -0.00000000025,     0.00000000189,
       0.00000000013,     0.00000000161,
       0.00000000008,    -0.00000000036,
       0.00000000001,     0.00000000004,
      -0.00000000002,    -0.00000000006,
       0.00000000005,    -0.00000000002,
       0.00000000025,     0.00000000086,
      -0.00000000003,     0.00000000001,
       0.00000000283,    -0.00000000077,
       0.00000000002,     0.00000000009,
      -0.00000000167,     0.00000000049,
      -0.00000000004,    -0.00000000005,
       0.00000000061,     0.00000000038,
      -0.00000000006,     0.00000000003,
      -0.00000000004,     0.00000000002,
      -0.00000000109,     0.00000000058,
      -0.00000000003,    -0.00000000006,
       0.00000000153,     0.00000000289,
      -0.00000000016,     0.00000000009,
      -0.00000000020,     0.00000000037,
      -0.00000000134,     0.00000000069,
      -0.00000000004,    -0.00000000007,
      -0.00000000051,    -0.00000000057,
       0.00000000005,    -0.00000000004,
      -0.00000000124,     0.00000000113,
      -0.00000000010,    -0.00000000010,
      -0.00000000098,     0.00000000038,
      -0.00000000001,    -0.00000000005,
      -0.00000000078,    -0.00000000124,
       0.00000000008,    -0.00000000006,
       0.00000000018,    -0.00000000026,
       0.00000000003,     0.00000000002,
      -0.00000000047,    -0.00000000031,
       0.00000000003,    -0.00000000005,
      -0.00000000045,    -0.00000000057,
       0.00000000004,    -0.00000000003,
       0.00000000061,    -0.00000000057,
       0.00000000005,     0.00000000005,
       0.00000000009,     0.00000000004,
       0.00000000004,    -0.00000000013,
       0.00000000018,    -0.00000000025,
       0.00000000027,     0.00000000019,
      -0.00000000002,     0.00000000003,
       0.00000000000,     0.00000000002,
       0.00000000009,     0.00000000004,
      -0.00000000004,     0.00000000009,
      -0.00000000002,    -0.00000000003,
      -0.00000000002,    -0.00000000001,
      -0.00000000004,     0.00000000006,
       0.00000000000,     0.00000000002,
       0.00000000008,     0.00000000001,
      -0.00000000004,    -0.00000000002,
      -0.00000000002,     0.00000000004,
       0.00000000001,     0.00000000007,
       0.00000000002,     0.00000000006,
       0.00000000017,    -0.00000000005,
      -0.00000000008,     0.00000000003,
       0.00000000004,     0.00000000004,
      -0.00000000009,     0.00000000005,
       0.00000000012,     0.00000000022,
      -0.00000000012,     0.00000000006,
      -0.00000000006,    -0.00000000006,
      -0.00000000013,     0.00000000012,
      -0.00000000011,     0.00000000005,
      -0.00000000009,    -0.00000000014,
       0.00000000003,    -0.00000000004,
      -0.00000000007,    -0.00000000005,
      -0.00000000007,    -0.00000000009,
       0.00000000010,    -0.00000000009,
       0.00000000002,     0.00000000001,
       0.00000000004,    -0.00000000006,
       0.00000000006,     0.00000000004,
      -0.00000000006,     0.00000000004,
      -0.00000000003,     0.00000000005,
       0.00000000005,     0.00000000004,
      -0.00000000002,     0.00000000002,
       0.00000000013,     0.00000000010,
       0.00000000002,    -0.00000000003,
       0.00000000012,     0.00000000004,
       0.00000000005,    -0.00000000014,
       0.00000000006,     0.00000000002,
       0.00000000012,    -0.00000000033,
      -0.00000000007,    -0.00000000003,
       0.00000000000,    -0.00000000024,
      -0.00000000029,    -0.00000000003,
      -0.00000000001,     0.00000000005,
       0.00000000001,    -0.00000000014,
      -0.00000000003,     0.00000000000,
      -0.00000000070,    -0.00000000004,
      -0.00000000001,     0.00000000013,
      -0.00000000040,     0.00000000013,
       0.00000000002,     0.00000000006,
       0.00000000009,     0.00000000053,
       0.00000000008,    -0.00000000001,
      -0.00000000026,     0.00000000006,
       0.00000000001,     0.00000000004,
       0.00000000030,     0.00000000125,
       0.00000000019,    -0.00000000004,
       0.00000000038,     0.00000000051,
       0.00000000005,    -0.00000000004,
       0.00000000080,    -0.00000000037,
      -0.00000000004,    -0.00000000011,
       0.00000000002,     0.00000000003,
       0.00000000023,     0.00000000040,
       0.00000000005,    -0.00000000003,
       0.00000000180,    -0.00000000104,
      -0.00000000011,    -0.00000000022,
       0.00000000038,    -0.00000000062,
      -0.00000000005,    -0.00000000002,
      -0.00000000077,    -0.00000000093,
      -0.00000000010,     0.00000000007,
       0.00000000003,    -0.00000000004,
       0.00000000045,    -0.00000000047,
      -0.00000000004,    -0.00000000004,
      -0.00000000200,    -0.00000000189,
      -0.00000000017,     0.00000000017,
      -0.00000000050,     0.00000000005,
      -0.00000000077,     0.00000000100,
       0.00000000007,     0.00000000007,
      -0.00000000004,    -0.00000000004,
      -0.00000000057,    -0.00000000028,
      -0.00000000001,     0.00000000003,
      -0.00000000114,     0.00000000231,
       0.00000000013,     0.00000000006,
       0.00000000047,    -0.00000000027,
      -0.00000000004,    -0.00000000005,
       0.00000000071,     0.00000000044,
       0.00000000003,    -0.00000000004,
      -0.00000000004,     0.00000000004,
       0.00000000000,     0.00000000021,
       0.00000000091,     0.00000000004,
      -0.00000000110,    -0.00000000032,
      -0.00000000002,     0.00000000003,
       0.00000000026,    -0.00000000007,
       0.00000000003,     0.00000000003,
      -0.00000000043,    -0.00000000003,
      -0.00000000010,     0.00000000135,
       0.00000000009,     0.00000000046,
       0.00000000005,    -0.00000000010,
       0.00000000010,     0.00000000028,
       0.00000000095,    -0.00000000030,
      -0.00000000097,     0.00000000027,
       0.00000000031,     0.00000000008,
      -0.00000000056,     0.00000000031,
      -0.00000000002,    -0.00000000003,
       0.00000000081,     0.00000000152,
      -0.00000000008,     0.00000000005,
      -0.00000000019,    -0.00000000003,
      -0.00000000059,     0.00000000038,
      -0.00000000003,    -0.00000000004,
      -0.00000000032,    -0.00000000036,
       0.00000000003,    -0.00000000003,
      -0.00000000082,     0.00000000074,
      -0.00000000006,    -0.00000000007,
      -0.00000000048,     0.00000000015,
      -0.00000000047,    -0.00000000068,
       0.00000000004,    -0.00000000003,
       0.00000000013,    -0.00000000020,
       0.00000000002,     0.00000000001,
      -0.00000000037,    -0.00000000025,
       0.00000000003,    -0.00000000004,
      -0.00000000027,    -0.00000000036,
       0.00000000040,    -0.00000000039,
       0.00000000004,     0.00000000003,
       0.00000000008,     0.00000000003,
       0.00000000004,    -0.00000000012,
       0.00000000014,    -0.00000000018,
       0.00000000021,     0.00000000014,
       0.00000000000,     0.00000000003,
       0.00000000008,     0.00000000003,
      -0.00000000003,     0.00000000008,
       0.00000000001,     0.00000000002,
       0.00000000005,    -0.00000000002,
      -0.00000000003,     0.00000000002,
       0.00000000005,     0.00000000008,
      -0.00000000004,     0.00000000003,
      -0.00000000003,    -0.00000000003,
      -0.00000000006,     0.00000000006,
      -0.00000000004,     0.00000000001,
      -0.00000000004,    -0.00000000006,
       0.00000000002,    -0.00000000002,
      -0.00000000004,    -0.00000000003,
      -0.00000000003,    -0.00000000004,
       0.00000000005,    -0.00000000005,
       0.00000000002,    -0.00000000003,
       0.00000000004,     0.00000000002,
      -0.00000000004,     0.00000000003,
      -0.00000000002,     0.00000000003,
       0.00000000004,     0.00000000003,
      -0.00000000001,     0.00000000002,
       0.00000000009,     0.00000000006,
       0.00000000008,     0.00000000002,
       0.00000000004,    -0.00000000009,
       0.00000000004,     0.00000000001,
       0.00000000007,    -0.00000000022,
      -0.00000000005,    -0.00000000002,
       0.00000000000,    -0.00000000015,
      -0.00000000019,    -0.00000000002,
       0.00000000000,    -0.00000000009,
      -0.00000000044,    -0.00000000002,
      -0.00000000001,     0.00000000008,
      -0.00000000023,     0.00000000008,
       0.00000000006,     0.00000000032,
       0.00000000005,    -0.00000000001,
      -0.00000000016,     0.00000000004,
       0.00000000000,     0.00000000002,
       0.00000000019,     0.00000000074,
       0.00000000011,    -0.00000000002,
       0.00000000021,     0.00000000026,
       0.00000000003,    -0.00000000002,
       0.00000000045,    -0.00000000021,
      -0.00000000002,    -0.00000000006,
       0.00000000013,     0.00000000023,
       0.00000000003,    -0.00000000001,
       0.00000000099,    -0.00000000059,
      -0.00000000006,    -0.00000000012,
       0.00000000014,    -0.00000000030,
      -0.00000000039,    -0.00000000049,
      -0.00000000005,     0.00000000003,
       0.00000000022,    -0.00000000024,
      -0.00000000002,    -0.00000000002,
      -0.00000000101,    -0.00000000093,
      -0.00000000008,     0.00000000008,
      -0.00000000016,     0.00000000010,
      -0.00000000037,     0.00000000046,
       0.00000000003,     0.00000000003,
      -0.00000000024,    -0.00000000011,
      -0.00000000045,     0.00000000098,
       0.00000000005,     0.00000000002,
       0.00000000027,    -0.00000000026,
      -0.00000000002,    -0.00000000003,
       0.00000000026,     0.00000000020,
       0.00000000002,     0.00000000002,
       0.00000000012,    -0.00000000007,
      -0.00000000053,    -0.00000000012,
       0.00000000011,     0.00000000003,
      -0.00000000024,    -0.00000000001,
      -0.00000000003,     0.00000000078,
       0.00000000005,     0.00000000005,
       0.00000000003,     0.00000000005,
       0.00000000022,    -0.00000000010,
      -0.00000000051,     0.00000000013,
       0.00000000015,     0.00000000000,
      -0.00000000027,     0.00000000015,
      -0.00000000001,    -0.00000000002,
       0.00000000041,     0.00000000074,
      -0.00000000004,     0.00000000002,
      -0.00000000014,    -0.00000000014,
      -0.00000000024,     0.00000000020,
      -0.00000000019,    -0.00000000022,
       0.00000000002,    -0.00000000002,
      -0.00000000050,     0.00000000045,
      -0.00000000004,    -0.00000000004,
      -0.00000000021,     0.00000000004,
      -0.00000000027,    -0.00000000036,
       0.00000000009,    -0.00000000014,
       0.00000000002,     0.00000000001,
      -0.00000000027,    -0.00000000018,
       0.00000000002,    -0.00000000003,
      -0.00000000015,    -0.00000000021,
       0.00000000024,    -0.00000000025,
       0.00000000007,     0.00000000002,
       0.00000000004,    -0.00000000011,
       0.00000000010,    -0.00000000013,
       0.00000000016,     0.00000000010,
       0.00000000000,     0.00000000002,
       0.00000000006,     0.00000000003,
      -0.00000000003,     0.00000000007,
      -0.00000000001,     0.00000000001,
      -0.00000000001,    -0.00000000001,
      -0.00000000003,     0.00000000003,
      -0.00000000001,     0.00000000001,
       0.00000000006,     0.00000000004,
       0.00000000005,     0.00000000001,
       0.00000000002,    -0.00000000006,
       0.00000000003,     0.00000000001,
       0.00000000005,    -0.00000000014,
      -0.00000000001,    -0.00000000009,
      -0.00000000012,    -0.00000000001,
       0.00000000000,    -0.00000000006,
      -0.00000000027,    -0.00000000001,
       0.00000000000,     0.00000000005,
      -0.00000000013,     0.00000000005,
       0.00000000003,     0.00000000019,
      -0.00000000010,     0.00000000003,
       0.00000000011,     0.00000000043,
       0.00000000006,    -0.00000000001,
       0.00000000012,     0.00000000012,
       0.00000000025,    -0.00000000011,
       0.00000000008,     0.00000000013,
       0.00000000002,    -0.00000000001,
       0.00000000052,    -0.00000000032,
      -0.00000000003,    -0.00000000006,
       0.00000000004,    -0.00000000014,
      -0.00000000020,    -0.00000000025,
       0.00000000011,    -0.00000000012,
      -0.00000000001,    -0.00000000001,
      -0.00000000049,    -0.00000000043,
      -0.00000000004,     0.00000000004,
      -0.00000000003,     0.00000000009,
      -0.00000000017,     0.00000000020,
      -0.00000000010,    -0.00000000004,
      -0.00000000015,     0.00000000038,
       0.00000000014,    -0.00000000018,
       0.00000000008,     0.00000000008,
       0.00000000002,    -0.00000000003,
      -0.00000000009,    -0.00000000007,
      -0.00000000023,    -0.00000000004,
       0.00000000004,     0.00000000004,
      -0.00000000012,     0.00000000000,
       0.00000000000,     0.00000000040,
       0.00000000002,    -0.00000000006,
       0.00000000001,    -0.00000000002,
      -0.00000000024,     0.00000000006,
       0.00000000006,    -0.00000000002,
      -0.00000000012,     0.00000000007,
       0.00000000019,     0.00000000034,
      -0.00000000009,    -0.00000000013,
      -0.00000000009,     0.00000000010,
      -0.00000000011,    -0.00000000012,
       0.00000000001,    -0.00000000001,
      -0.00000000029,     0.00000000026,
      -0.00000000008,     0.00000000000,
      -0.00000000015,    -0.00000000017,
       0.00000000006,    -0.00000000009,
      -0.00000000019,    -0.00000000012,
      -0.00000000008,    -0.00000000012,
       0.00000000014,    -0.00000000015,
       0.00000000005,     0.00000000002,
       0.00000000003,    -0.00000000009,
       0.00000000007,    -0.00000000008,
       0.00000000011,     0.00000000007,
       0.00000000000,     0.00000000002,
       0.00000000005,     0.00000000002,
      -0.00000000002,     0.00000000006,
       0.00000000004,     0.00000000003,
       0.00000000001,    -0.00000000004,
       0.00000000002,     0.00000000001,
       0.00000000003,    -0.00000000009,
       0.00000000000,    -0.00000000005,
      -0.00000000007,    -0.00000000001,
       0.00000000000,    -0.00000000004,
      -0.00000000017,     0.00000000000,
      -0.00000000007,     0.00000000003,
       0.00000000002,     0.00000000011,
      -0.00000000006,     0.00000000002,
       0.00000000007,     0.00000000024,
       0.00000000006,     0.00000000005,
       0.00000000014,    -0.00000000006,
       0.00000000004,     0.00000000007,
       0.00000000027,    -0.00000000017,
      -0.00000000002,    -0.00000000003,
       0.00000000000,    -0.00000000006,
      -0.00000000009,    -0.00000000012,
       0.00000000005,    -0.00000000006,
      -0.00000000023,    -0.00000000019,
       0.00000000001,     0.00000000006,
      -0.00000000008,     0.00000000008,
      -0.00000000003,    -0.00000000001,
      -0.00000000004,     0.00000000013,
       0.00000000007,    -0.00000000011,
       0.00000000002,     0.00000000003,
       0.00000000001,    -0.00000000003,
      -0.00000000010,    -0.00000000005,
      -0.00000000009,    -0.00000000001,
      -0.00000000005,     0.00000000000,
       0.00000000000,     0.00000000018,
       0.00000000001,    -0.00000000007,
       0.00000000000,    -0.00000000003,
      -0.00000000007,     0.00000000001,
      -0.00000000010,     0.00000000003,
      -0.00000000005,     0.00000000003,
       0.00000000008,     0.00000000014,
      -0.00000000006,    -0.00000000009,
      -0.00000000003,     0.00000000005,
      -0.00000000006,    -0.00000000006,
      -0.00000000016,     0.00000000014,
      -0.00000000008,    -0.00000000008,
       0.00000000004,    -0.00000000006,
      -0.00000000012,    -0.00000000008,
      -0.00000000004,    -0.00000000006,
       0.00000000008,    -0.00000000009,
       0.00000000003,     0.00000000001,
       0.00000000002,    -0.00000000006,
       0.00000000004,    -0.00000000005,
       0.00000000007,     0.00000000004,
      -0.00000000001,     0.00000000004,
       0.00000000002,    -0.00000000006,
       0.00000000000,    -0.00000000002,
      -0.00000000010,     0.00000000000,
      -0.00000000003,     0.00000000002,
       0.00000000001,     0.00000000006,
      -0.00000000003,     0.00000000001,
       0.00000000004,     0.00000000013,
       0.00000000003,     0.00000000002,
       0.00000000007,    -0.00000000003,
       0.00000000002,     0.00000000003,
       0.00000000013,    -0.00000000009,
      -0.00000000004,    -0.00000000006,
       0.00000000002,    -0.00000000003,
      -0.00000000010,    -0.00000000008,
       0.00000000002,     0.00000000004,
      -0.00000000003,     0.00000000003,
       0.00000000003,    -0.00000000006,
       0.00000000001,    -0.00000000002,
      -0.00000000008,    -0.00000000003,
      -0.00000000002,     0.00000000000,
       0.00000000000,     0.00000000007,
       0.00000000000,    -0.00000000005,
       0.00000000000,    -0.00000000002,
      -0.00000000006,     0.00000000001,
      -0.00000000001,     0.00000000001,
       0.00000000003,     0.00000000005,
      -0.00000000003,    -0.00000000005,
      -0.00000000003,    -0.00000000003,
      -0.00000000008,     0.00000000008,
      -0.00000000004,    -0.00000000003,
       0.00000000002,    -0.00000000003,
      -0.00000000007,    -0.00000000005,
       0.00000000004,    -0.00000000005,
       0.00000000002,     0.00000000001,
       0.00000000002,    -0.00000000004,
       0.00000000002,    -0.00000000003,
       0.00000000005,     0.00000000003,
      -0.00000000006,     0.00000000000,
      -0.00000000002,     0.00000000001,
       0.00000000002,     0.00000000007,
       0.00000000004,    -0.00000000002,
       0.00000000001,     0.00000000002,
       0.00000000006,    -0.00000000004,
      -0.00000000004,    -0.00000000003,
      -0.00000000004,    -0.00000000001,
      -0.00000000001,    -0.00000000002,
      -0.00000000004,     0.00000000004,
       0.00000000001,    -0.00000000002,
      -0.00000000004,    -0.00000000003,
       0.00000000002,     0.00000000001,
       0.00000000003,    -0.00000000002,
      -0.00000090259,     0.00000024787,
      -0.00000001030,    -0.00000002831,
       0.00000000040,    -0.00000000012,
      -0.00000071575,     0.00000002818,
      -0.00000000252,    -0.00000000265,
       0.00000000009,     0.00000000001,
       0.00000002641,    -0.00000066591,
      -0.00000000237,     0.00000000059,
       0.00000000146,     0.00000001219,
       0.00000000064,     0.00000000088,
       0.00000000002,    -0.00000000007,
      -0.00000006714,    -0.00000000406,
       0.00000000009,    -0.00000000043,
       0.00000000005,    -0.00000000004,
       0.00000001200,    -0.00000004702,
      -0.00000000037,     0.00000000081,
       0.00000013481,    -0.00000015004,
       0.00000000555,     0.00000001500,
      -0.00000000083,     0.00000000000,
      -0.00000252407,     0.00000001767,
       0.00000000306,     0.00000000547,
       0.00000000050,    -0.00000000020,
       0.00000001787,    -0.00000246150,
       0.00000000546,    -0.00000000241,
      -0.00000000022,    -0.00000000051,
      -0.00000000309,     0.00000000241,
       0.00000000026,    -0.00000000119,
       0.00000000304,     0.00000000105,
       0.00000000097,     0.00000000015,
       0.00000003900,     0.00000003996,
      -0.00000000256,    -0.00000000174,
       0.00000483421,    -0.00000000660,
      -0.00000000123,    -0.00000000251,
      -0.00000000645,     0.00000480730,
      -0.00000000188,     0.00000000092,
       0.00000000227,    -0.00000000081,
      -0.00000000008,     0.00000000084,
      -0.00000000063,    -0.00000000332,
      -0.00000000081,     0.00000000006,
      -0.00000028112,    -0.00000063967,
       0.00000004679,    -0.00000002861,
       0.00000000119,     0.00000000187,
       0.00000535145,     0.00000000430,
      -0.00000000051,    -0.00000000046,
      -0.00000000027,     0.00000535053,
      -0.00000000025,    -0.00000000048,
       0.00000000164,    -0.00000000855,
      -0.00000000049,    -0.00000000074,
      -0.00000000051,     0.00000000352,
       0.00000000044,    -0.00000000042,
       0.00000017738,    -0.00000069321,
       0.00000003192,    -0.00000001656,
       0.00000000097,     0.00000000099,
       0.00000982940,    -0.00000000570,
      -0.00000000118,    -0.00000000100,
      -0.00000000035,     0.00000982483,
      -0.00000000412,    -0.00000000817,
      -0.00000000105,    -0.00000000209,
      -0.00000001322,     0.00000001297,
       0.00000110891,     0.00000005831,
      -0.00000007438,     0.00000010486,
      -0.00000000444,    -0.00000000371,
       0.00001331586,    -0.00000001394,
       0.00000000215,     0.00000000014,
       0.00000000669,     0.00001332016,
       0.00000000040,     0.00000000310,
       0.00000000092,     0.00000000048,
      -0.00000001151,    -0.00000000661,
       0.00000000019,     0.00000000110,
      -0.00000007255,    -0.00000000931,
       0.00000015958,     0.00000130188,
      -0.00000002125,    -0.00000000515,
       0.00001890430,    -0.00000000373,
      -0.00000000028,    -0.00000000171,
      -0.00000000444,     0.00001888635,
      -0.00000000076,     0.00000000078,
       0.00000000432,     0.00000001944,
       0.00000000129,    -0.00000000033,
      -0.00000000053,    -0.00000000239,
       0.00000023768,    -0.00000021313,
       0.00000000271,    -0.00000000032,
       0.00002367415,    -0.00000003166,
      -0.00000000218,    -0.00000000137,
       0.00000000388,     0.00002367379,
      -0.00000000024,     0.00000000077,
      -0.00000000243,    -0.00000000246,
      -0.00000000271,    -0.00000000274,
      -0.00000000306,     0.00000000260,
      -0.00000000016,    -0.00000000025,
       0.00000003290,     0.00000003965,
      -0.00000000324,     0.00000000207,
      -0.00000000002,    -0.00000000003,
       0.00000001563,     0.00000004960,
      -0.00000000167,     0.00000000071,
      -0.00000004978,     0.00000001258,
      -0.00000000044,    -0.00000000148,
       0.00000001111,     0.00000000476,
      -0.00000000016,    -0.00000000028,
      -0.00000000432,     0.00000001407,
       0.00000000012,    -0.00000000020,
       0.00000000052,    -0.00000000043,
      -0.00000000361,    -0.00000000417,
       0.00000000033,    -0.00000000017,
      -0.00000000092,    -0.00000000394,
       0.00000000392,    -0.00000000083,
       0.00000000174,     0.00000000090,
      -0.00000000092,     0.00000000161,
       0.00000000004,    -0.00000000004,
      -0.00000000011,    -0.00000000007,
       0.00000000016,    -0.00000000008,
      -0.00000000001,     0.00000000001,
      -0.00000000034,    -0.00000000173,
       0.00000000017,     0.00000000010,
       0.00000000029,    -0.00000000095,
       0.00000000005,     0.00000000007,
       0.00000000085,     0.00000000042,
      -0.00000000007,     0.00000000003,
       0.00000000017,    -0.00000000146,
       0.00000000003,     0.00000000004,
       0.00000000157,     0.00000000009,
      -0.00000000003,     0.00000000004,
      -0.00000000017,     0.00000000074,
      -0.00000000074,    -0.00000000017,
      -0.00000000005,     0.00000000007,
       0.00000000001,    -0.00000000005,
      -0.00000000002,     0.00000000003,
      -0.00000000004,    -0.00000000003,
       0.00000000334,    -0.00000000478,
       0.00000000184,     0.00000000105,
      -0.00000001103,     0.00000001607,
      -0.00000000658,    -0.00000000391,
       0.00000000107,     0.00000000074,
      -0.00000000026,     0.00000000045,
       0.00000006835,    -0.00000010030,
       0.00000004093,     0.00000002369,
       0.00000000072,     0.00000000112,
      -0.00000000041,     0.00000000032,
      -0.00000000110,     0.00000000034,
      -0.00000000038,     0.00000000019,
       0.00000000198,    -0.00000000086,
       0.00000000041,     0.00000000061,
       0.00000000082,     0.00000000004,
      -0.00000000027,     0.00000000032,
       0.00000000055,    -0.00000000133,
      -0.00000000131,    -0.00000000053,
      -0.00000000025,     0.00000000048,
       0.00000000049,     0.00000000026,
      -0.00000000054,     0.00000000130,
       0.00000000129,     0.00000000054,
      -0.00000000033,    -0.00000000018,
       0.00000000034,     0.00000000022,
       0.00000000020,    -0.00000000041,
      -0.00000000368,    -0.00000000169,
       0.00000000023,    -0.00000000026,
       0.00000001953,     0.00000000737,
      -0.00000000148,     0.00000000177,
       0.00000000715,    -0.00000001928,
       0.00000000173,     0.00000000142,
      -0.00000000502,    -0.00000000086,
      -0.00000000095,     0.00000000514,
      -0.00000000069,     0.00000000166,
       0.00000000263,     0.00000000119,
      -0.00000000072,    -0.00000000016,
      -0.00000001817,    -0.00000000763,
       0.00000000136,    -0.00000000196,
      -0.00000000758,     0.00000001810,
      -0.00000000195,    -0.00000000134,
       0.00000000343,     0.00000000059,
       0.00000000065,    -0.00000000351,
       0.00000000052,     0.00000000095,
       0.00000000790,     0.00000000498,
      -0.00000000052,     0.00000000113,
      -0.00000000015,    -0.00000000005,
       0.00000002503,    -0.00000005489,
       0.00000000804,     0.00000000433,
      -0.00000000032,     0.00000000088,
       0.00000000224,     0.00000023510,
       0.00000002013,    -0.00000000638,
      -0.00000000040,    -0.00000000088,
       0.00000023134,    -0.00000000381,
      -0.00000000616,    -0.00000002042,
      -0.00000000090,     0.00000000042,
       0.00000002221,    -0.00000009611,
       0.00000000838,     0.00000000153,
      -0.00000009809,    -0.00000002372,
       0.00000000186,    -0.00000000874,
       0.00000000039,     0.00000000011,
      -0.00000002705,    -0.00000001190,
       0.00000000196,    -0.00000000280,
       0.00000000043,     0.00000000011,
      -0.00000001674,     0.00000003822,
      -0.00000000549,    -0.00000000304,
       0.00000000021,    -0.00000000060,
       0.00000000857,    -0.00000015189,
      -0.00000001477,     0.00000000357,
       0.00000000027,     0.00000000065,
      -0.00000015081,    -0.00000000821,
       0.00000000350,     0.00000001485,
       0.00000000064,    -0.00000000027,
      -0.00000001524,     0.00000006575,
      -0.00000000575,    -0.00000000106,
       0.00000006699,     0.00000001613,
      -0.00000000126,     0.00000000596,
       0.00000000000,     0.00000000061,
       0.00000000102,     0.00000000001,
      -0.00000000003,     0.00000000207,
       0.00000000079,     0.00000000171,
      -0.00000000050,    -0.00000000108,
       0.00000000028,     0.00000000062,
       0.00000000011,     0.00000000000,
       0.00000000009,     0.00000000000,
       0.00000000000,    -0.00000000008,
      -0.00000007038,     0.00000004731,
       0.00000001625,     0.00000000237,
      -0.00000000015,    -0.00000000081,
       0.00000018950,    -0.00000043606,
      -0.00000002754,     0.00000005927,
       0.00000000290,    -0.00000000012,
      -0.00001705724,    -0.00000003294,
       0.00000000765,    -0.00000001009,
       0.00000000119,     0.00000000080,
       0.00000003965,     0.00001690123,
       0.00000001346,     0.00000000482,
      -0.00000000010,     0.00000000119,
      -0.00000003934,    -0.00000000848,
       0.00000000221,     0.00000000492,
       0.00000000626,     0.00000000166,
       0.00000000374,    -0.00000000129,
       0.00000008520,    -0.00000029880,
      -0.00000004136,     0.00000000355,
       0.00000000061,     0.00000000194,
      -0.00000012498,     0.00000031722,
       0.00000002246,    -0.00000004963,
      -0.00000000220,     0.00000000022,
       0.00000823038,     0.00000000268,
      -0.00000000267,     0.00000000168,
      -0.00000000770,    -0.00000819544,
      -0.00000000180,    -0.00000000025,
       0.00000002823,     0.00000000570,
      -0.00000000144,    -0.00000000340,
      -0.00000000217,    -0.00000000081,
      -0.00000000302,     0.00000000078,
       0.00000000029,    -0.00000000057,
      -0.00000000002,    -0.00000000101,
       0.00000000045,     0.00000000023,
      -0.00000000023,     0.00000000045,
       0.00000177364,    -0.00000000058,
      -0.00000000033,    -0.00000000001,
      -0.00000000519,    -0.00001974219,
      -0.00000000034,     0.00000000477,
       0.00000003128,     0.00000010015,
       0.00000000641,    -0.00000000193,
      -0.00000000024,    -0.00000000075,
      -0.00000012222,    -0.00000002241,
       0.00000000457,     0.00000001086,
      -0.00000000403,     0.00000044543,
      -0.00000000157,    -0.00000000797,
       0.00000000002,    -0.00000000167,
      -0.00000001760,    -0.00000010540,
       0.00000000031,    -0.00000003895,
       0.00000760929,     0.00000000055,
       0.00000000036,    -0.00000000061,
       0.00000000357,     0.00001664099,
       0.00000000042,    -0.00000000328,
      -0.00000001591,    -0.00000018965,
       0.00000000286,     0.00000000621,
      -0.00000000001,     0.00000000065,
      -0.00000000604,    -0.00000002016,
      -0.00000000058,    -0.00000001257,
       0.00000001288,    -0.00000023794,
       0.00000000481,     0.00000000428,
      -0.00000000005,     0.00000000089,
       0.00000005480,     0.00000005629,
      -0.00000000096,     0.00000002080,
       0.00000004787,    -0.00000000004,
      -0.00000000008,     0.00000000068,
       0.00000000049,     0.00000000030,
      -0.00000000238,     0.00000000107,
       0.00000000813,    -0.00000000382,
      -0.00000000067,    -0.00000000125,
       0.00000000364,     0.00000000794,
       0.00000000120,    -0.00000000061,
      -0.00000000246,     0.00000000044,
      -0.00000000049,    -0.00000000252,
      -0.00000000017,    -0.00000000040,
       0.00000000174,    -0.00000000077,
      -0.00000000067,     0.00000000020,
      -0.00000000795,     0.00000000385,
       0.00000000064,     0.00000000127,
      -0.00000000380,    -0.00000000788,
      -0.00000000125,     0.00000000062,
       0.00000000177,    -0.00000000030,
       0.00000000033,     0.00000000181,
       0.00000000208,     0.00000000095,
       0.00000000023,    -0.00000000057,
      -0.00000000006,    -0.00000000080,
      -0.00000000080,     0.00000000006,
       0.00000000051,    -0.00000000004,
       0.00000000024,     0.00000000056,
      -0.00000000054,     0.00000000023,
      -0.00000000021,    -0.00000000033,
       0.00000000034,    -0.00000000022,
      -0.00000000024,    -0.00000000058,
       0.00000000057,    -0.00000000024,
       0.00000000035,     0.00000000048,
       0.00000000039,     0.00000000018,
       0.00000000079,     0.00000000027,
       0.00000000021,    -0.00000000061,
      -0.00000000656,    -0.00000000221,
      -0.00000000190,     0.00000000499,
      -0.00000000056,     0.00000000068,
       0.00000000026,     0.00000000021,
      -0.00000000054,     0.00000000240,
       0.00000000072,     0.00000000018,
       0.00000000240,     0.00000000054,
       0.00000000018,    -0.00000000072,
      -0.00000000064,    -0.00000000245,
      -0.00000000057,     0.00000000012,
      -0.00000000246,     0.00000000064,
       0.00000000011,     0.00000000057,
      -0.00000000151,     0.00000000641,
       0.00000000193,     0.00000000047,
       0.00000000009,    -0.00000000029,
       0.00000000102,     0.00000000023,
       0.00000000400,     0.00000001706,
       0.00000000388,    -0.00000000081,
      -0.00000000007,    -0.00000000045,
       0.00000001709,    -0.00000000399,
      -0.00000000081,    -0.00000000389,
      -0.00000000045,     0.00000000007,
      -0.00000000039,     0.00000000046,
       0.00000000046,     0.00000000036,
      -0.00000000007,    -0.00000000063,
       0.00000000139,    -0.00000000003,
      -0.00000001607,    -0.00000001903,
      -0.00000000304,     0.00000000225,
      -0.00000001914,     0.00000001602,
       0.00000000224,     0.00000000307,
       0.00000000020,    -0.00000000069,
      -0.00000000066,    -0.00000000014,
       0.00000001128,     0.00000005326,
       0.00000001193,    -0.00000000251,
      -0.00000000023,    -0.00000000138,
      -0.00000000365,    -0.00000000075,
       0.00000000015,     0.00000000059,
       0.00000008336,     0.00000010501,
       0.00000001582,    -0.00000001220,
      -0.00000000089,    -0.00000000125,
       0.00000010532,    -0.00000008342,
      -0.00000001218,    -0.00000001591,
      -0.00000000126,     0.00000000088,
      -0.00000000106,     0.00000000359,
       0.00000000348,     0.00000000082,
      -0.00000000032,     0.00000000029,
       0.00000000012,     0.00000000011,
       0.00000000008,     0.00000000004,
       0.00000000003,    -0.00000000007,
       0.00000002793,     0.00000004248,
       0.00000000537,    -0.00000000469,
      -0.00000000035,    -0.00000000039,
       0.00000488028,    -0.00000321911,
      -0.00000054001,    -0.00000062030,
      -0.00000004496,     0.00000003989,
       0.00000151234,     0.00000069449,
       0.00000006029,    -0.00000011089,
      -0.00000000438,    -0.00000000230,
       0.00000070632,    -0.00000151403,
      -0.00000011157,    -0.00000006297,
      -0.00000000265,     0.00000000441,
       0.00000000143,     0.00000002341,
      -0.00000000128,     0.00000000139,
       0.00000001988,    -0.00000000443,
       0.00000000109,     0.00000000145,
      -0.00000078253,    -0.00000124810,
      -0.00000015107,     0.00000013638,
       0.00000001012,     0.00000001064,
      -0.00002549814,     0.00001681596,
       0.00000282143,     0.00000324010,
       0.00000023495,    -0.00000020825,
      -0.00000001020,    -0.00000001140,
      -0.00000248561,    -0.00000135751,
      -0.00000005980,     0.00000018540,
       0.00000000669,     0.00000000151,
      -0.00000139206,     0.00000250339,
       0.00000018697,     0.00000006899,
       0.00000000265,    -0.00000000672,
      -0.00000000729,    -0.00000012194,
       0.00000000669,    -0.00000000763,
      -0.00000000014,     0.00000000034,
      -0.00000010621,     0.00000001996,
      -0.00000000616,    -0.00000000740,
       0.00000000990,     0.00000000552,
       0.00000000025,    -0.00000000074,
       0.00000005623,    -0.00000010407,
      -0.00000000771,    -0.00000000289,
       0.00000025350,    -0.00000000153,
      -0.00000000033,     0.00000000017,
       0.00000000162,    -0.00000025594,
      -0.00000000039,     0.00000000078,
       0.00000130003,     0.00000055313,
       0.00000006130,    -0.00000009486,
      -0.00000000387,    -0.00000000259,
      -0.00000031943,     0.00000060623,
       0.00000004493,     0.00000001830,
       0.00000000066,    -0.00000000166,
       0.00000253235,     0.00000001090,
       0.00000000026,    -0.00000000095,
      -0.00000000581,    -0.00000251558,
       0.00000000116,     0.00000000296,
       0.00000000083,    -0.00000000215,
       0.00000000212,     0.00000000021,
       0.00000001579,    -0.00000000003,
      -0.00000000019,    -0.00000012234,
      -0.00000000916,    -0.00000000109,
      -0.00000000041,     0.00000000015,
       0.00000000402,    -0.00000000194,
      -0.00000000066,    -0.00000000005,
       0.00000000000,     0.00000000134,
       0.00001412662,     0.00000000661,
      -0.00000000139,    -0.00000000020,
      -0.00000000053,     0.00000651923,
      -0.00000019996,     0.00000019102,
      -0.00000000846,    -0.00000003134,
       0.00000000057,    -0.00000000079,
       0.00000009043,     0.00000042967,
      -0.00000001460,     0.00000001878,
      -0.00000000035,    -0.00000000125,
      -0.00000000845,     0.00000003197,
       0.00000000281,     0.00000000129,
       0.00000003853,     0.00000000714,
       0.00000000159,    -0.00000000235,
      -0.00000000088,    -0.00000000056,
      -0.00000000285,     0.00000000130,
      -0.00000000134,    -0.00000000283,
      -0.00000000260,     0.00000000539,
      -0.00000000062,    -0.00000000080,
       0.00000001011,     0.00000000450,
      -0.00000000150,     0.00000000119,
      -0.00000004118,     0.00000001770,
      -0.00000000191,    -0.00000000299,
      -0.00000001765,    -0.00000004130,
       0.00000000297,    -0.00000000190,
      -0.00000000400,     0.00000001783,
       0.00000000131,     0.00000000072,
      -0.00000001787,    -0.00000000396,
      -0.00000000072,     0.00000000131,
      -0.00000000026,     0.00000000034,
      -0.00000000034,    -0.00000000025,
       0.00000000093,    -0.00000000005,
       0.00000000034,    -0.00000000055,
       0.00000000059,     0.00000000032,
       0.00000000098,    -0.00000000131,
      -0.00000000131,    -0.00000000098,
      -0.00000000015,    -0.00000000085,
       0.00000001257,    -0.00000001047,
      -0.00000000004,    -0.00000000248,
      -0.00000001048,    -0.00000001257,
      -0.00000000248,     0.00000000004,
       0.00000000088,    -0.00000000083,
      -0.00000000083,    -0.00000000088,
       0.00000000154,     0.00000000019,
       0.00000000003,    -0.00000000018,
       0.00000000050,    -0.00000000822,
      -0.00000000104,    -0.00000000016,
       0.00000013448,    -0.00000006702,
       0.00000000323,    -0.00000001475,
      -0.00000000068,    -0.00000000049,
      -0.00000006714,    -0.00000013446,
      -0.00000001478,    -0.00000000322,
      -0.00000000049,     0.00000000068,
      -0.00000000047,     0.00000000208,
       0.00000000203,     0.00000000045,
       0.00000000909,    -0.00000000853,
       0.00000001620,    -0.00000000890,
       0.00000000030,    -0.00000000148,
      -0.00000000890,    -0.00000001620,
      -0.00000000148,    -0.00000000030,
       0.00000000033,    -0.00000000097,
       0.00000001239,     0.00000000443,
       0.00000000002,    -0.00000000104,
       0.00000001851,    -0.00000006633,
      -0.00000000582,    -0.00000000028,
       0.00000114984,    -0.00000027045,
       0.00000002529,    -0.00000005335,
      -0.00000000170,    -0.00000000168,
      -0.00000027177,    -0.00000115027,
      -0.00000005363,    -0.00000002529,
      -0.00000000170,     0.00000000167,
      -0.00000000358,     0.00000000749,
      -0.00000000067,    -0.00000000007,
       0.00000000702,     0.00000000331,
      -0.00000000009,     0.00000000066,
       0.00000000043,     0.00000000060,
       0.00000000064,    -0.00000000044,
       0.00000018238,    -0.00000009998,
       0.00000000467,    -0.00000002032,
       0.00000000190,     0.00000000357,
       0.00000026546,    -0.00000006808,
       0.00000000621,    -0.00000001220,
      -0.00000006797,    -0.00000026534,
      -0.00000001218,    -0.00000000622,
       0.00000000017,    -0.00000000051,
      -0.00000000051,    -0.00000000016,
      -0.00000000058,    -0.00000000211,
       0.00000000043,    -0.00000000046,
      -0.00000000046,    -0.00000000043,
       0.00000000305,     0.00000001102,
      -0.00000000078,     0.00000000073,
       0.00000000074,     0.00000000079,
       0.00000000020,    -0.00000000062,
      -0.00000000061,    -0.00000000020,
       0.00000000056,    -0.00000000172,
       0.00000000079,     0.00000000026,
       0.00000000021,     0.00000000032,
       0.00000000033,    -0.00000000023,
       0.00000007295,     0.00000003944,
      -0.00000000224,    -0.00000000480,
      -0.00000000009,     0.00000000012,
       0.00000017410,    -0.00000040149,
      -0.00000002537,     0.00000000940,
       0.00000000050,     0.00000000050,
       0.00000646057,     0.00000000052,
      -0.00000000309,     0.00000000048,
       0.00000000002,    -0.00000000064,
      -0.00000001014,    -0.00000647289,
      -0.00000000156,     0.00000000241,
      -0.00000000067,     0.00000000000,
      -0.00000000008,     0.00000000244,
      -0.00000000065,    -0.00000000223,
       0.00000353237,    -0.00000090393,
       0.00000008354,    -0.00000016245,
      -0.00000000518,    -0.00000000530,
       0.00000005279,     0.00000020746,
       0.00000000955,     0.00000000491,
       0.00000359990,    -0.00000000199,
      -0.00000000117,    -0.00000000033,
       0.00000000005,    -0.00000359617,
       0.00000000001,     0.00000000124,
       0.00000000114,     0.00000000130,
       0.00000000131,    -0.00000000114,
       0.00000003654,    -0.00000000518,
       0.00000013191,    -0.00000003359,
       0.00000031599,    -0.00000008104,
      -0.00000000080,     0.00000000008,
       0.00000000063,     0.00000000098,
       0.00000000096,    -0.00000000064,
       0.00000000107,     0.00000000163,
       0.00000000021,    -0.00000000018,
       0.00000000209,    -0.00000000137,
       0.00000000120,     0.00000000001,
       0.00000000139,    -0.00000000098,
      -0.00000000033,    -0.00000000050,
      -0.00000000098,    -0.00000000139,
      -0.00000000050,     0.00000000033,
      -0.00000000102,    -0.00000000016,
      -0.00000000016,     0.00000000102,
       0.00000000058,     0.00000000012,
       0.00000000012,    -0.00000000059,
       0.00000000097,     0.00000000025,
       0.00000000025,    -0.00000000097,
       0.00000003520,     0.00000000924,
      -0.00000000049,     0.00000000191,
       0.00000000099,    -0.00000000025,
      -0.00000000025,    -0.00000000099,
       0.00000002995,     0.00000000113,
       0.00000000101,    -0.00000000026,
       0.00000000026,     0.00000000101,
       0.00000020642,    -0.00000000221,
      -0.00000001032,    -0.00000131647,
      -0.00000024054,    -0.00000001331,
      -0.00000000570,    -0.00000000168,
      -0.00000006121,     0.00000004030,
      -0.00000001096,     0.00000000116,
      -0.00000000057,     0.00000001428,
      -0.00000000136,    -0.00000000712,
       0.00000000001,    -0.00000000136,
       0.00007170005,    -0.00000001946,
      -0.00000000715,    -0.00000000317,
       0.00000000254,     0.00001277074,
      -0.00000016184,    -0.00000008266,
      -0.00000000383,    -0.00000001487,
      -0.00000004124,     0.00000032382,
      -0.00000000742,     0.00000000768,
       0.00000003069,     0.00000006087,
       0.00000000586,    -0.00000000058,
       0.00000006085,    -0.00000003065,
      -0.00000000058,    -0.00000000586,
       0.00000080102,     0.00000002057,
      -0.00000000063,     0.00000000397,
      -0.00000000842,     0.00000000160,
      -0.00000000051,    -0.00000000056,
       0.00000000035,     0.00000000848,
      -0.00000000056,     0.00000000051,
       0.00000293427,    -0.00000000038,
       0.00000000000,     0.00000000224,
       0.00000000182,    -0.00000000001,
       0.00000688153,    -0.00000000304,
       0.00000000001,     0.00000000182,
      -0.00000000022,    -0.00000000043,
       0.00000000006,    -0.00000000053,
       0.00000000929,    -0.00000000371,
      -0.00000000131,    -0.00000000296,
      -0.00000000372,    -0.00000000929,
      -0.00000000296,     0.00000000132,
       0.00000000195,    -0.00000000083,
      -0.00000000083,    -0.00000000195,
      -0.00000000362,    -0.00000000052,
      -0.00000000043,     0.00000000354,
      -0.00000000232,    -0.00000000038,
      -0.00000000036,     0.00000000237,
       0.00000000077,     0.00000000003,
       0.00000000003,    -0.00000000077,
       0.00000000125,     0.00000000002,
       0.00000000002,    -0.00000000125,
       0.00000000145,    -0.00000002977,
       0.00000000762,     0.00000000121,
       0.00000000121,    -0.00000000765,
      -0.00000009296,    -0.00000001194,
       0.00000001390,     0.00000000000,
      -0.00000000001,    -0.00000001392,
      -0.00000010240,     0.00000001360,
       0.00000001541,     0.00000000000,
       0.00000000000,     0.00000001545,
       0.00000003426,    -0.00000003301,
       0.00000001954,    -0.00000000002,
      -0.00000000002,    -0.00000001954,
       0.00000003079,     0.00000003366,
       0.00000002078,    -0.00000000003,
      -0.00000000001,     0.00000002076,
       0.00000000044,     0.00000000016,
       0.00000000063,    -0.00000000238,
      -0.00000000081,    -0.00000000001,
       0.00000003646,    -0.00000000437,
      -0.00000000264,    -0.00000001061,
      -0.00000000456,    -0.00000003654,
      -0.00000001062,     0.00000000267,
       0.00000001885,    -0.00000000802,
      -0.00000000279,    -0.00000000600,
       0.00000000044,     0.00000000102,
       0.00000002000,    -0.00000000296,
      -0.00000000160,    -0.00000000574,
      -0.00000000291,    -0.00000001997,
      -0.00000000574,     0.00000000160,
       0.00000000035,     0.00000000214,
       0.00000000214,    -0.00000000035,
      -0.00000000793,     0.00000000115,
       0.00000000117,     0.00000000146,
       0.00000000115,     0.00000000793,
       0.00000000146,    -0.00000000117,
       0.00000000638,     0.00000004155,
       0.00000001087,    -0.00000000958,
      -0.00000002929,     0.00000000316,
       0.00000000421,     0.00000000534,
       0.00000000317,     0.00000002930,
       0.00000000534,    -0.00000000421,
       0.00000000029,    -0.00000000039,
      -0.00000000179,    -0.00000000119,
      -0.00000000321,    -0.00000005776,
      -0.00000000554,     0.00000000385,
      -0.00000005770,     0.00000000323,
       0.00000000383,     0.00000000552,
      -0.00000000063,    -0.00000000095,
      -0.00000000095,     0.00000000063,
      -0.00000062190,     0.00000006871,
       0.00000009043,     0.00000011372,
       0.00000000682,    -0.00000000960,
      -0.00000000096,    -0.00000000895,
      -0.00000000163,     0.00000000129,
      -0.00000002133,    -0.00000039011,
      -0.00000003681,     0.00000002575,
       0.00000000130,     0.00000000153,
      -0.00000039041,     0.00000002138,
       0.00000002582,     0.00000003688,
       0.00000000154,    -0.00000000130,
       0.00000000700,     0.00000000464,
       0.00000000183,     0.00000000124,
      -0.00000000020,    -0.00000000010,
       0.00000000410,    -0.00000000808,
      -0.00000000038,     0.00000000103,
       0.00000029010,    -0.00000000045,
      -0.00000000050,    -0.00000028927,
      -0.00000002512,    -0.00000000348,
       0.00000000068,    -0.00000000438,
      -0.00000000209,    -0.00000000002,
       0.00000000002,     0.00000000200,
      -0.00000055161,    -0.00000982542,
      -0.00000092854,     0.00000065667,
       0.00000003254,     0.00000003928,
       0.00000035363,    -0.00000001923,
      -0.00000002369,    -0.00000003341,
      -0.00000000142,     0.00000000118,
       0.00000459198,    -0.00000000318,
       0.00000000492,     0.00000000060,
      -0.00000000394,    -0.00000459906,
      -0.00000000069,    -0.00000000389,
       0.00000000104,    -0.00000000205,
      -0.00000000205,    -0.00000000104,
      -0.00000000153,    -0.00000003807,
      -0.00000000682,    -0.00000008536,
       0.00000000072,    -0.00000000031,
       0.00000000192,    -0.00000001188,
       0.00000000110,     0.00000000015,
       0.00000000525,    -0.00000000008,
      -0.00000000039,    -0.00000003295,
       0.00000000072,    -0.00000000262,
       0.00000000124,     0.00000000013,
      -0.00000001321,    -0.00000000020,
       0.00000000088,    -0.00000000019,
       0.00000001014,     0.00000000019,
      -0.00000000004,     0.00000000279,
       0.00017342289,    -0.00000013377,
       0.00000003360,    -0.00000000170,
       0.00000001017,     0.00001756850,
       0.00000000012,     0.00000000141,
       0.00000001302,    -0.00000050282,
       0.00000002368,     0.00000003357,
      -0.00000000085,     0.00000000201,
      -0.00000025148,    -0.00000002689,
       0.00000001668,    -0.00000004753,
       0.00000000100,     0.00000000169,
       0.00000007053,     0.00000004051,
       0.00000000259,    -0.00000000455,
       0.00000004055,    -0.00000007055,
      -0.00000000457,    -0.00000000259,
       0.00000071936,     0.00000001767,
      -0.00000000018,     0.00000000125,
      -0.00000000027,    -0.00000000369,
      -0.00000000366,     0.00000000075,
       0.00000162402,    -0.00000000095,
       0.00000000302,     0.00000000241,
       0.00000000039,    -0.00000000079,
      -0.00000001564,    -0.00000001114,
      -0.00000000167,     0.00000000400,
       0.00000000483,    -0.00000000001,
      -0.00000000001,    -0.00000000483,
       0.00000000548,    -0.00000000001,
       0.00000000000,     0.00000000548,
      -0.00000000157,    -0.00000000060,
      -0.00000000060,     0.00000000157,
       0.00000000173,     0.00000000017,
       0.00000000017,    -0.00000000173,
      -0.00000000130,    -0.00000000002,
      -0.00000000002,     0.00000000129,
      -0.00000006232,    -0.00000002357,
       0.00000002686,     0.00000000669,
       0.00000000669,    -0.00000002685,
       0.00000008326,     0.00000000868,
      -0.00000001084,     0.00000000050,
       0.00000000051,     0.00000001081,
       0.00000003050,     0.00000000042,
      -0.00000005042,    -0.00000000413,
       0.00000000006,    -0.00000000151,
      -0.00000000341,    -0.00000000063,
      -0.00000000063,     0.00000000373,
      -0.00000000085,    -0.00000000010,
      -0.00000000010,     0.00000000086,
      -0.00000000148,    -0.00000001019,
       0.00000000111,    -0.00000000015,
       0.00000116896,     0.00000028798,
      -0.00000000475,    -0.00000003937,
      -0.00000000234,     0.00000000941,
      -0.00000044201,    -0.00000005429,
       0.00000000169,     0.00000000719,
      -0.00000005420,     0.00000044179,
       0.00000000718,    -0.00000000169,
      -0.00000070466,     0.00000001989,
      -0.00000000031,    -0.00000001449,
      -0.00000009696,     0.00000000567,
       0.00000000034,    -0.00000000134,
       0.00000000561,     0.00000009747,
      -0.00000000134,    -0.00000000033,
      -0.00000014567,    -0.00000002693,
       0.00000000135,    -0.00000000722,
       0.00000000529,     0.00000000000,
       0.00000000000,    -0.00000000530,
      -0.00002269434,    -0.00000276043,
       0.00000009059,     0.00000036817,
       0.00000005242,    -0.00000042712,
      -0.00000000696,     0.00000000166,
       0.00000636503,     0.00000000095,
      -0.00000000071,    -0.00000635838,
      -0.00000000033,    -0.00000000337,
      -0.00000000337,     0.00000000033,
       0.00000002524,     0.00000000789,
      -0.00000000933,    -0.00000000178,
      -0.00000000178,     0.00000000932,
      -0.00000003781,    -0.00000000016,
       0.00000000000,    -0.00000000168,
      -0.00000004549,    -0.00000000273,
       0.00000000230,     0.00000000011,
       0.00000000012,    -0.00000000238,
      -0.00000376442,     0.00000021592,
       0.00000001311,    -0.00000005236,
      -0.00000001087,    -0.00000018990,
       0.00000000264,     0.00000000066,
      -0.00000000835,     0.00000000064,
       0.00000000054,    -0.00000000205,
      -0.00000000029,    -0.00000000100,
      -0.00000000100,     0.00000000029,
       0.00000017220,     0.00000000997,
      -0.00000000015,     0.00000000259,
       0.00000000303,     0.00000000045,
       0.00000000045,    -0.00000000397,
       0.00000022780,     0.00000003235,
      -0.00000000091,     0.00000000704,
       0.00048838076,     0.00000000985,
      -0.00000000033,     0.00002508604,
       0.00000029147,    -0.00000007158,
      -0.00000000113,     0.00000000945,
      -0.00000003578,    -0.00000058316,
       0.00000000433,     0.00000000226,
       0.00000003154,     0.00000014196,
       0.00000000957,    -0.00000000182,
       0.00000014188,    -0.00000003154,
      -0.00000000182,    -0.00000000956,
       0.00000081488,    -0.00000006729,
       0.00000000079,     0.00000000792,
       0.00000008121,    -0.00000000479,
      -0.00000000028,     0.00000000113,
      -0.00000000476,    -0.00000008122,
       0.00000000113,     0.00000000028,
      -0.00000002383,    -0.00000000424,
       0.00000000349,     0.00000000042,
       0.00000000042,    -0.00000000347,
      -0.00000033147,    -0.00000006017,
       0.00000000117,    -0.00000000608,
       0.00000009352,     0.00000000524,
      -0.00000000034,    -0.00000000134,
       0.00000000521,    -0.00000009342,
      -0.00000000134,     0.00000000034,
       0.00000000829,    -0.00000000122,
      -0.00000000122,    -0.00000000829,
       0.00000505266,    -0.00000029366,
      -0.00000001785,     0.00000007080,
       0.00000000935,     0.00000016197,
      -0.00000000224,    -0.00000000056,
       0.00000000252,    -0.00000000049,
      -0.00000000023,     0.00000000331,
       0.00000000070,     0.00000000062,
       0.00000000062,    -0.00000000070,
      -0.00000000182,     0.00000000017,
      -0.00000000017,    -0.00000000182,
       0.00000000163,    -0.00000000014,
       0.00000000014,     0.00000000163,
      -0.00000000070,    -0.00000000062,
       0.00000000062,    -0.00000000070,
       0.00000010781,    -0.00000007196,
      -0.00000001288,    -0.00000003527,
      -0.00000000556,    -0.00000001040,
      -0.00000000183,     0.00000000077,
       0.00000014384,    -0.00000001763,
      -0.00000000057,     0.00000000211,
       0.00000001782,     0.00000014422,
      -0.00000000210,    -0.00000000058,
       0.00000003110,     0.00000013989,
       0.00000000943,    -0.00000000180,
      -0.00000013982,     0.00000003106,
       0.00000000180,     0.00000000943,
      -0.00000003134,     0.00000000113,
      -0.00000006302,    -0.00000015402,
       0.00000001282,     0.00000000806,
       0.00000000428,    -0.00000000097,
       0.00000000349,     0.00000000365,
       0.00000000352,     0.00000000468,
      -0.00000000110,     0.00000000111,
       0.00000000111,     0.00000000110,
       0.00000015392,    -0.00000002180,
       0.00000000104,     0.00000000798,
       0.00000000179,    -0.00000000016,
      -0.00000000016,    -0.00000000189,
      -0.00000005867,     0.00000006117,
      -0.00000000072,    -0.00000000072,
      -0.00000008027,    -0.00000000444,
       0.00000000028,     0.00000000113,
      -0.00000000444,     0.00000008027,
       0.00000000113,    -0.00000000028,
       0.00000002243,    -0.00000000170,
       0.00000000009,     0.00000000133,
      -0.00000465345,    -0.00000026259,
       0.00000001651,     0.00000006527,
       0.00000000892,    -0.00000015790,
      -0.00000000224,     0.00000000056,
      -0.00000000646,     0.00000000056,
       0.00000000031,     0.00000000623,
      -0.00000000061,    -0.00000000071,
      -0.00000000071,     0.00000000061,
       0.00000005326,     0.00000000002,
       0.00000000013,     0.00000006159,
      -0.00000000182,    -0.00000000017,
       0.00000000017,    -0.00000000182,
       0.00000011418,     0.00000000015,
      -0.00000000014,    -0.00000002158,
       0.00000000158,     0.00000000014,
      -0.00000000014,     0.00000000158,
      -0.00000000109,    -0.00000000013,
       0.00000000012,    -0.00000000110,
       0.00000000061,     0.00000000071,
      -0.00000000070,     0.00000000061,
      -0.00000000110,    -0.00000000310,
       0.00000000310,    -0.00000000110,
      -0.00000002494,     0.00000016723,
      -0.00000001166,    -0.00000001670,
      -0.00000000758,    -0.00000000108,
       0.00000000077,    -0.00000000054,
      -0.00000000341,     0.00000000045,
       0.00000000045,     0.00000000341,
      -0.00000017273,     0.00000003197,
      -0.00000000087,    -0.00000000666,
      -0.00000000227,     0.00000000016,
       0.00000000016,     0.00000000227,
       0.00000000037,     0.00000000498,
      -0.00000003682,     0.00000000272,
      -0.00000000048,    -0.00000000652,
       0.00000002570,    -0.00000000002,
      -0.00000006854,    -0.00000000017,
       0.00000000559,     0.00000000005,
      -0.00000000004,     0.00000000562,
      -0.00000000016,    -0.00000000102,
       0.00000000102,    -0.00000000016,
       0.00000000464,    -0.00000000044,
      -0.00000000044,    -0.00000000461,
      -0.00000000014,    -0.00000000128,
      -0.00000000128,     0.00000000014,
      -0.00000003301,     0.00000000010,
      -0.00000000930,    -0.00000008660,
       0.00000007170,    -0.00000008013,
      -0.00000008013,    -0.00000007170,
       0.00000000163,     0.00000000000,
       0.00000000000,    -0.00000000163,
       0.00000356138,     0.00000020172,
      -0.00000001259,    -0.00000005026,
      -0.00000001057,     0.00000018638,
       0.00000000264,    -0.00000000066,
       0.00000002316,    -0.00000000046,
      -0.00000000006,    -0.00000002531,
       0.00000000016,     0.00000000103,
       0.00000000103,    -0.00000000016,
       0.00000574236,    -0.00000641495,
       0.00000000890,    -0.00000007344,
       0.00000007865,     0.00000007041,
       0.00000794801,    -0.00000000613,
      -0.00000000581,    -0.00000794805,
       0.00000003729,    -0.00000000016,
       0.00000000000,     0.00000000237,
       0.00095530543,    -0.00000071699,
      -0.00000002834,    -0.00000002281,
       0.00000001769,     0.00003147352,
       0.00000000056,    -0.00000000063,
      -0.00000004752,    -0.00000010594,
      -0.00000000006,    -0.00000000094,
      -0.00000005295,     0.00000009501,
       0.00000013563,     0.00000013707,
       0.00000001095,    -0.00000000341,
       0.00000013710,    -0.00000013559,
      -0.00000000341,    -0.00000001095,
       0.00000000828,    -0.00000056622,
      -0.00000005890,    -0.00000003227,
      -0.00000001902,    -0.00000000616,
      -0.00000000140,     0.00000000182,
      -0.00000002353,    -0.00000002622,
       0.00000002623,    -0.00000002359,
       0.00000013459,     0.00000013608,
       0.00000001087,    -0.00000000338,
      -0.00000013605,     0.00000013462,
       0.00000000338,     0.00000001087,
       0.00000000103,    -0.00000000020,
       0.00000000020,     0.00000000103,
       0.00000000029,     0.00000000099,
      -0.00000000099,     0.00000000029,
       0.00000000008,    -0.00000000475,
       0.00000000475,     0.00000000008,
       0.00000011808,     0.00000019927,
      -0.00000001268,     0.00000000761,
      -0.00000000015,    -0.00000000030,
       0.00000011790,     0.00000048067,
      -0.00000001398,     0.00000000441,
      -0.00000000003,    -0.00000000019,
      -0.00000048045,     0.00000011053,
      -0.00000000376,    -0.00000001352,
       0.00000000019,    -0.00000000004,
       0.00000002476,     0.00000001145,
      -0.00000000034,    -0.00000000075,
      -0.00000001012,     0.00000003603,
       0.00000000016,    -0.00000000047,
      -0.00000000005,    -0.00000000001,
      -0.00000001515,     0.00000000961,
       0.00000000062,     0.00000000134,
       0.00000000017,    -0.00000000007,
       0.00000001733,     0.00000003216,
       0.00000000809,    -0.00000000409,
      -0.00000000028,    -0.00000000068,
      -0.00000000968,    -0.00000004819,
       0.00000000140,     0.00000000003,
       0.00000004712,    -0.00000000927,
       0.00000000009,     0.00000000155,
       0.00000001967,     0.00000006724,
       0.00000000603,    -0.00000000099,
      -0.00000006795,     0.00000001982,
       0.00000000113,     0.00000000619,
      -0.00000001401,    -0.00000002675,
      -0.00000000462,     0.00000000214,
       0.00000000019,     0.00000000043,
      -0.00000001881,     0.00000002751,
       0.00000000098,    -0.00000000172,
      -0.00000002668,    -0.00000001912,
       0.00000000168,     0.00000000090,
      -0.00000000782,    -0.00000003372,
      -0.00000000292,     0.00000000055,
       0.00000003425,    -0.00000000824,
      -0.00000000066,    -0.00000000304,
       0.00000000599,     0.00000000183,
      -0.00000000058,     0.00000000036,
      -0.00000002226,     0.00000002613,
       0.00000000170,    -0.00000000179,
      -0.00000002603,    -0.00000002226,
       0.00000000179,     0.00000000172,
       0.00000000043,    -0.00000000186,
       0.00000000187,     0.00000000040,
       0.00000000271,     0.00000000401,
      -0.00000004413,     0.00000004983,
       0.00000000307,    -0.00000000353,
      -0.00000004978,    -0.00000004414,
       0.00000000353,     0.00000000306,
      -0.00000000528,     0.00000000536,
       0.00000000019,    -0.00000000107,
      -0.00000005975,     0.00000006754,
       0.00000000414,    -0.00000000478,
      -0.00000006752,    -0.00000005981,
       0.00000000477,     0.00000000414,
      -0.00000000732,    -0.00000000505,
      -0.00000008491,     0.00000009581,
       0.00000000590,    -0.00000000679,
      -0.00000009568,    -0.00000008487,
       0.00000000678,     0.00000000588,
       0.00000000001,     0.00000000216,
      -0.00000010620,     0.00000012010,
       0.00000000739,    -0.00000000852,
      -0.00000011997,    -0.00000010634,
       0.00000000850,     0.00000000737,
       0.00000000093,    -0.00000000034,
       0.00000000027,     0.00000000071,
      -0.00000000021,     0.00000000030,
      -0.00000000034,     0.00000000017,
      -0.00000000153,     0.00000000076,
      -0.00000000024,    -0.00000000039,
      -0.00000000031,    -0.00000000019,
      -0.00000000397,    -0.00000000045,
       0.00000000006,    -0.00000000097,
       0.00000000613,     0.00000000067,
      -0.00000000008,     0.00000000144,
       0.00000000020,    -0.00000000169,
       0.00000000041,     0.00000000003,
      -0.00000002315,    -0.00000000292,
       0.00000000039,    -0.00000000566,
       0.00000000091,    -0.00000000152,
       0.00000000145,     0.00000000095,
       0.00000000034,     0.00000000042,
      -0.00000000041,    -0.00000000074,
      -0.00000001013,    -0.00000000717,
       0.00000000103,    -0.00000000155,
       0.00000001647,     0.00000001169,
      -0.00000000170,     0.00000000256,
       0.00000000042,    -0.00000000025,
       0.00000000022,     0.00000000042,
      -0.00000000263,     0.00000000354,
      -0.00000000058,    -0.00000000038,
      -0.00000007068,    -0.00000005197,
       0.00000000758,    -0.00000001157,
      -0.00000000369,     0.00000000172,
      -0.00000000148,    -0.00000000369,
      -0.00000000030,     0.00000000037,
       0.00000000075,    -0.00000000191,
      -0.00000000188,    -0.00000000073,
      -0.00000000028,     0.00000000057,
       0.00000000058,     0.00000000029,
      -0.00000000072,     0.00000000184,
       0.00000000183,     0.00000000072,
       0.00000000019,    -0.00000000039,
      -0.00000000040,    -0.00000000020,
      -0.00000000092,    -0.00000000167,
       0.00000000152,     0.00000000276,
      -0.00000000194,     0.00000000086,
      -0.00000000585,    -0.00000001278,
       0.00000000097,    -0.00000000055,
      -0.00000000157,    -0.00000000015,
       0.00000000027,    -0.00000000150,
       0.00000000047,    -0.00000000122,
      -0.00000000392,    -0.00000000189,
       0.00000000024,    -0.00000000028,
       0.00000002937,     0.00000000959,
      -0.00000000217,     0.00000000180,
      -0.00000000044,    -0.00000000006,
       0.00000000935,    -0.00000002906,
       0.00000000175,     0.00000000210,
      -0.00000000656,    -0.00000000114,
      -0.00000000127,     0.00000000673,
      -0.00000000110,     0.00000000288,
      -0.00000000013,    -0.00000000018,
       0.00000000282,     0.00000000134,
      -0.00000000083,    -0.00000000020,
      -0.00000002769,    -0.00000000951,
       0.00000000205,    -0.00000000199,
      -0.00000000947,     0.00000002762,
      -0.00000000198,    -0.00000000203,
       0.00000000449,     0.00000000078,
       0.00000000086,    -0.00000000459,
       0.00000000039,    -0.00000000120,
      -0.00000000001,     0.00000000181,
      -0.00000000152,    -0.00000000015,
       0.00000000051,    -0.00000000691,
       0.00000000023,     0.00000000042,
       0.00000000052,    -0.00000000029,
       0.00000003444,     0.00000001105,
      -0.00000000251,     0.00000000202,
      -0.00000000052,    -0.00000000006,
       0.00000001018,     0.00000001414,
       0.00000000642,    -0.00000000082,
       0.00000000990,     0.00000051997,
       0.00000004444,    -0.00000001432,
      -0.00000000090,    -0.00000000199,
       0.00000051409,    -0.00000001214,
      -0.00000001395,    -0.00000004488,
      -0.00000000196,     0.00000000092,
       0.00000003252,    -0.00000014095,
       0.00000001236,     0.00000000218,
      -0.00000000010,     0.00000000048,
      -0.00000014482,    -0.00000003517,
       0.00000000277,    -0.00000001294,
       0.00000000057,     0.00000000017,
      -0.00000006521,    -0.00000002092,
       0.00000000475,    -0.00000000383,
       0.00000000098,     0.00000000011,
      -0.00000000690,    -0.00000000985,
      -0.00000000443,     0.00000000060,
      -0.00000000805,    -0.00000044831,
      -0.00000003881,     0.00000001215,
       0.00000000079,     0.00000000172,
      -0.00000044717,     0.00000000832,
       0.00000001207,     0.00000003890,
       0.00000000172,    -0.00000000079,
      -0.00000002238,     0.00000009642,
      -0.00000000851,    -0.00000000152,
       0.00000009863,     0.00000002384,
      -0.00000000185,     0.00000000883,
      -0.00000000038,    -0.00000000011,
       0.00000000015,     0.00000000003,
       0.00000000011,     0.00000000000,
       0.00000000000,    -0.00000000009,
      -0.00000002452,    -0.00000103973,
      -0.00000008937,     0.00000002857,
       0.00000000183,     0.00000000394,
      -0.00001609597,     0.00000023696,
       0.00000043521,     0.00000140641,
       0.00000006242,    -0.00000002843,
       0.00002247399,     0.00000002556,
      -0.00000000869,     0.00000000790,
      -0.00000000119,    -0.00000000112,
      -0.00000003667,    -0.00002230650,
      -0.00000001508,    -0.00000000547,
      -0.00000000033,    -0.00000000115,
       0.00000007500,     0.00000001121,
      -0.00000000371,    -0.00000000656,
      -0.00000000041,     0.00000000016,
      -0.00000001215,     0.00000000532,
      -0.00000000531,     0.00000000182,
       0.00000000022,     0.00000000042,
       0.00000003482,     0.00000147617,
       0.00000012689,    -0.00000004055,
      -0.00000000259,    -0.00000000558,
       0.00001089494,    -0.00000015977,
      -0.00000029444,    -0.00000095091,
      -0.00000004216,     0.00000001913,
      -0.00001988852,     0.00000000540,
       0.00000000192,     0.00000000053,
       0.00000000009,     0.00000000068,
       0.00000000751,     0.00001986929,
       0.00000000242,     0.00000000003,
       0.00000000066,     0.00000000009,
      -0.00000005173,    -0.00000000484,
       0.00000000272,     0.00000000548,
       0.00000000039,    -0.00000000016,
       0.00000000666,     0.00000000831,
       0.00000000500,    -0.00000000169,
      -0.00000000019,    -0.00000000039,
       0.00000000000,     0.00000000028,
       0.00000432219,    -0.00000000076,
      -0.00000000015,    -0.00000000044,
       0.00000000000,    -0.00000000010,
      -0.00000000391,    -0.00002003673,
      -0.00000000189,     0.00000000162,
      -0.00000000116,     0.00000000003,
       0.00000002551,     0.00000002965,
       0.00000000398,    -0.00000000164,
      -0.00000008445,     0.00000001134,
       0.00000000329,    -0.00000000039,
      -0.00000001009,     0.00000021472,
      -0.00000000379,    -0.00000000383,
       0.00000000004,    -0.00000000077,
      -0.00000004297,    -0.00000005082,
       0.00000000075,    -0.00000001879,
      -0.00000408978,     0.00000000122,
       0.00000000008,     0.00000000032,
       0.00000000263,     0.00001268612,
       0.00000000135,    -0.00000000116,
      -0.00000002937,     0.00000003362,
       0.00000000066,    -0.00000000168,
       0.00000004426,     0.00000001859,
      -0.00000000256,     0.00000000038,
      -0.00000000706,    -0.00000014697,
      -0.00000000256,     0.00000000262,
       0.00000000002,     0.00000000053,
      -0.00000002957,     0.00000003479,
       0.00000000053,     0.00000001286,
       0.00000000029,     0.00000000025,
       0.00000000025,    -0.00000000029,
       0.00000000568,    -0.00000000878,
      -0.00000000002,     0.00000000049,
       0.00000000112,    -0.00000002043,
       0.00000000073,     0.00000000234,
      -0.00000038651,     0.00000001063,
       0.00000000186,     0.00000000370,
       0.00000001066,    -0.00000034379,
       0.00000000364,    -0.00000000159,
       0.00000000019,     0.00000000077,
       0.00000000001,    -0.00000000079,
       0.00000000024,     0.00000000236,
       0.00000000072,    -0.00000000006,
      -0.00000003647,     0.00000004226,
       0.00000000235,    -0.00000000293,
       0.00000000005,     0.00000000015,
      -0.00000002647,    -0.00000000855,
       0.00000000154,     0.00000000045,
      -0.00000004548,    -0.00000000455,
      -0.00000000092,    -0.00000000167,
      -0.00000000459,    -0.00000006555,
      -0.00000000166,     0.00000000073,
      -0.00000000047,    -0.00000000069,
       0.00000000000,     0.00000000059,
      -0.00000000039,    -0.00000000119,
      -0.00000000052,     0.00000000027,
      -0.00000000088,     0.00000000119,
       0.00000000044,     0.00000000035,
       0.00000000119,     0.00000000088,
       0.00000000035,    -0.00000000044,
       0.00000000071,     0.00000000050,
       0.00000000029,    -0.00000000158,
      -0.00000000048,    -0.00000000011,
      -0.00000000158,    -0.00000000029,
      -0.00000000012,     0.00000000048,
      -0.00000000379,     0.00000000516,
       0.00000000190,     0.00000000153,
       0.00000000031,    -0.00000000035,
      -0.00000000334,    -0.00000000235,
      -0.00000000097,     0.00000000122,
      -0.00000000286,     0.00000001512,
       0.00000000455,     0.00000000110,
       0.00000000021,    -0.00000000069,
       0.00000001511,     0.00000000282,
       0.00000000109,    -0.00000000455,
      -0.00000000069,    -0.00000000021,
      -0.00000000132,     0.00000000708,
       0.00000000213,     0.00000000051,
       0.00000000010,    -0.00000000032,
       0.00000040420,     0.00000007556,
       0.00000002944,    -0.00000012197,
      -0.00000001848,    -0.00000000564,
       0.00000001428,     0.00000005320,
       0.00000001231,    -0.00000000250,
      -0.00000000021,    -0.00000000144,
       0.00000005326,    -0.00000001422,
      -0.00000000248,    -0.00000001234,
      -0.00000000145,     0.00000000020,
      -0.00000000109,     0.00000000122,
       0.00000000122,     0.00000000089,
       0.00000004334,    -0.00000023368,
      -0.00000007053,    -0.00000001698,
      -0.00000000326,     0.00000001068,
      -0.00000212481,    -0.00000039635,
      -0.00000015456,     0.00000064184,
       0.00000009717,     0.00000002967,
       0.00000000360,    -0.00000000970,
      -0.00000013630,    -0.00000051510,
      -0.00000011966,     0.00000002411,
       0.00000000197,     0.00000001401,
      -0.00000051366,     0.00000013767,
       0.00000002459,     0.00000011910,
       0.00000001390,    -0.00000000207,
       0.00000000571,    -0.00000000632,
      -0.00000000020,    -0.00000000083,
      -0.00000000632,    -0.00000000512,
      -0.00000000070,     0.00000000023,
       0.00000000012,     0.00000000056,
       0.00000000101,     0.00000000379,
       0.00000000088,    -0.00000000018,
      -0.00000000001,    -0.00000000010,
       0.00000002834,    -0.00000000743,
      -0.00000000133,    -0.00000000654,
      -0.00000000077,     0.00000000011,
       0.00000001427,     0.00000001687,
       0.00000000270,    -0.00000000199,
       0.00000001691,    -0.00000001426,
      -0.00000000199,    -0.00000000272,
      -0.00000000015,     0.00000000041,
      -0.00000002790,    -0.00000010429,
      -0.00000002421,     0.00000000493,
       0.00000000040,     0.00000000283,
      -0.00000015136,     0.00000003972,
       0.00000000716,     0.00000003502,
       0.00000000409,    -0.00000000059,
      -0.00000013994,    -0.00000016646,
      -0.00000002684,     0.00000001961,
       0.00000000142,     0.00000000216,
      -0.00000016532,     0.00000014026,
       0.00000001970,     0.00000002650,
       0.00000000211,    -0.00000000143,
       0.00000000081,    -0.00000000208,
      -0.00000000196,    -0.00000000057,
       0.00000000646,     0.00000000764,
       0.00000000123,    -0.00000000090,
      -0.00000000007,    -0.00000000010,
       0.00000003602,    -0.00000003021,
      -0.00000000423,    -0.00000000575,
      -0.00000000046,     0.00000000031,
       0.00000003326,     0.00000001535,
       0.00000000133,    -0.00000000244,
       0.00000001547,    -0.00000003332,
      -0.00000000245,    -0.00000000137,
      -0.00000014240,    -0.00000016843,
      -0.00000002715,     0.00000001992,
       0.00000000144,     0.00000000218,
      -0.00000018789,     0.00000015756,
       0.00000002210,     0.00000002998,
       0.00000000241,    -0.00000000160,
      -0.00000033461,    -0.00000015596,
      -0.00000001384,     0.00000002466,
       0.00000000098,     0.00000000058,
      -0.00000015265,     0.00000033392,
       0.00000002451,     0.00000001311,
       0.00000000049,    -0.00000000097,
       0.00000000034,    -0.00000000201,
      -0.00000000160,     0.00000000000,
       0.00000002544,     0.00000001175,
       0.00000000104,    -0.00000000186,
      -0.00000000007,    -0.00000000004,
       0.00000004321,    -0.00000009304,
      -0.00000000682,    -0.00000000376,
       0.00000004913,    -0.00000000009,
       0.00000000018,    -0.00000004951,
      -0.00000042110,    -0.00000019408,
      -0.00000001712,     0.00000003104,
       0.00000000123,     0.00000000068,
      -0.00000021479,     0.00000046238,
       0.00000003391,     0.00000001866,
       0.00000000075,    -0.00000000134,
      -0.00000051569,    -0.00000000323,
      -0.00000000004,     0.00000000069,
       0.00000000548,     0.00000050879,
      -0.00000000032,    -0.00000000127,
       0.00000000108,    -0.00000000163,
      -0.00000000052,    -0.00000000077,
       0.00000006063,     0.00000000002,
       0.00000000002,    -0.00000018760,
      -0.00000000379,     0.00000000116,
       0.00000000159,     0.00000000207,
       0.00000000006,     0.00000000069,
      -0.00000067567,     0.00000000356,
       0.00000000102,     0.00000000047,
      -0.00000000058,     0.00000086772,
      -0.00000000207,    -0.00000000572,
       0.00000000122,     0.00000000198,
       0.00000000882,    -0.00000002835,
      -0.00000000025,    -0.00000000238,
       0.00000000050,    -0.00000000365,
      -0.00000000178,    -0.00000000068,
      -0.00000000157,     0.00000000066,
      -0.00000000007,    -0.00000000011,
       0.00000000139,     0.00000000344,
      -0.00000000301,    -0.00000000008,
      -0.00000000007,    -0.00000000302,
      -0.00000004770,     0.00000005987,
       0.00000000437,    -0.00000000389,
       0.00000000004,     0.00000000025,
      -0.00000003961,    -0.00000004578,
       0.00000000352,     0.00000000119,
      -0.00000003364,    -0.00000000119,
      -0.00000000166,    -0.00000003450,
       0.00000000010,    -0.00000000065,
       0.00000000022,     0.00000000040,
       0.00000000038,    -0.00000000051,
      -0.00000000135,    -0.00000000103,
       0.00000000220,    -0.00000000183,
      -0.00000000183,    -0.00000000219,
       0.00000000014,    -0.00000000042,
       0.00000000512,    -0.00000000428,
      -0.00000000002,    -0.00000000101,
      -0.00000000007,    -0.00000000003,
      -0.00000001104,    -0.00000001347,
      -0.00000000264,     0.00000000003,
       0.00000001975,    -0.00000000982,
       0.00000000047,    -0.00000000216,
      -0.00000000985,    -0.00000001966,
      -0.00000000216,    -0.00000000047,
       0.00000005698,    -0.00000002855,
       0.00000000137,    -0.00000000626,
      -0.00000000029,    -0.00000000021,
      -0.00000007226,    -0.00000014685,
      -0.00000001609,    -0.00000000349,
      -0.00000000053,     0.00000000074,
       0.00000013378,    -0.00000003113,
       0.00000000292,    -0.00000000622,
      -0.00000003155,    -0.00000013302,
      -0.00000000618,    -0.00000000290,
      -0.00000000026,     0.00000000101,
      -0.00000000086,     0.00000000031,
       0.00000000031,     0.00000000086,
       0.00000050560,    -0.00000011974,
       0.00000001116,    -0.00000002350,
      -0.00000000075,    -0.00000000074,
      -0.00000029778,    -0.00000127874,
      -0.00000005956,    -0.00000002789,
      -0.00000000186,     0.00000000189,
       0.00000054257,     0.00000000390,
      -0.00000000039,    -0.00000000026,
      -0.00000000091,    -0.00000053763,
      -0.00000000076,     0.00000000368,
      -0.00000000351,    -0.00000000100,
      -0.00000000551,     0.00000000252,
       0.00000000023,     0.00000000113,
      -0.00000000428,     0.00000000155,
       0.00000000154,     0.00000000426,
      -0.00000000959,     0.00000000636,
       0.00000000396,    -0.00000000213,
       0.00000000615,    -0.00000000368,
      -0.00000000148,     0.00000000144,
       0.00000000022,    -0.00000000072,
      -0.00000000033,     0.00000000101,
       0.00000000100,     0.00000000033,
      -0.00000000015,     0.00000000049,
       0.00000000029,    -0.00000000089,
      -0.00000000089,    -0.00000000029,
       0.00000000060,     0.00000000020,
      -0.00000000038,    -0.00000000012,
       0.00000293981,    -0.00000000212,
      -0.00000000123,    -0.00000000025,
       0.00000000000,    -0.00000000031,
      -0.00000000528,    -0.00000731372,
      -0.00000000057,     0.00000000303,
      -0.00000000077,    -0.00000000001,
      -0.00000002542,     0.00000002056,
      -0.00000000101,    -0.00000000269,
      -0.00000001266,     0.00000004216,
      -0.00000000099,    -0.00000000113,
      -0.00000000893,     0.00000001907,
      -0.00000000170,    -0.00000000019,
      -0.00000001801,    -0.00000000945,
       0.00000000018,    -0.00000000180,
      -0.00000003095,     0.00000002095,
      -0.00000000136,     0.00000000531,
      -0.00000001536,     0.00000001745,
       0.00000000108,    -0.00000000123,
       0.00000001738,     0.00000001517,
      -0.00000000124,    -0.00000000108,
      -0.00000000129,    -0.00000000014,
       0.00000000059,    -0.00000000042,
      -0.00000000014,    -0.00000000020,
      -0.00000000106,    -0.00000000152,
      -0.00000000053,     0.00000000036,
       0.00000000144,    -0.00000000057,
      -0.00000000020,    -0.00000000045,
      -0.00000000057,    -0.00000000143,
      -0.00000000045,     0.00000000020,
      -0.00000000045,    -0.00000000006,
      -0.00000000017,     0.00000000114,
      -0.00000000050,    -0.00000000008,
      -0.00000000008,     0.00000000049,
       0.00000000013,    -0.00000000066,
      -0.00000006633,    -0.00000001728,
      -0.00000000160,    -0.00000000300,
       0.00000000010,    -0.00000000009,
      -0.00000004239,     0.00000016148,
      -0.00000000730,     0.00000000393,
      -0.00000003574,    -0.00000000016,
       0.00000000096,    -0.00000003581,
       0.00000000127,     0.00000000093,
      -0.00000000089,     0.00000000124,
      -0.00000031389,     0.00000035445,
       0.00000002205,    -0.00000002514,
      -0.00000006554,    -0.00000005826,
       0.00000000466,     0.00000000403,
       0.00000000065,    -0.00000000044,
      -0.00000000142,    -0.00000000209,
      -0.00000001342,     0.00000001513,
      -0.00000003080,     0.00000003478,
       0.00000000053,    -0.00000000014,
      -0.00000000047,     0.00000000012,
       0.00000000040,    -0.00000000104,
       0.00000000408,    -0.00000000164,
      -0.00000000058,    -0.00000000130,
      -0.00000000410,    -0.00000001033,
      -0.00000000329,     0.00000000146,
       0.00000000454,    -0.00000000066,
      -0.00000000034,    -0.00000000132,
      -0.00000000070,    -0.00000000449,
      -0.00000000130,     0.00000000035,
      -0.00000000151,    -0.00000000015,
      -0.00000000053,     0.00000000389,
       0.00000000047,     0.00000000001,
       0.00000000003,    -0.00000000116,
       0.00000001659,    -0.00000000203,
      -0.00000000121,    -0.00000000482,
      -0.00000000506,    -0.00000004128,
      -0.00000001199,     0.00000000298,
      -0.00000000083,     0.00000000030,
       0.00000000013,     0.00000000093,
      -0.00000000365,     0.00000000053,
       0.00000000054,     0.00000000067,
       0.00000000004,    -0.00000000006,
       0.00000000129,     0.00000000900,
       0.00000000166,    -0.00000000133,
      -0.00000000019,    -0.00000000378,
      -0.00000000030,     0.00000000020,
      -0.00000000374,     0.00000000016,
       0.00000000019,     0.00000000030,
      -0.00000000142,    -0.00000002679,
      -0.00000000258,     0.00000000179,
       0.00000000009,     0.00000000011,
      -0.00000006597,     0.00000000342,
       0.00000000439,     0.00000000634,
       0.00000001366,    -0.00000000009,
       0.00000000018,    -0.00000001335,
       0.00000000162,     0.00000000160,
       0.00000000160,    -0.00000000162,
       0.00000013663,    -0.00000000023,
      -0.00000000051,    -0.00000033399,
       0.00000000026,     0.00000000060,
      -0.00000000113,     0.00000000103,
      -0.00000000074,     0.00000000043,
      -0.00000000042,    -0.00000000075,
       0.00000004846,     0.00000004053,
      -0.00000000279,    -0.00000001027,
      -0.00000000153,     0.00000000260,
      -0.00000002020,     0.00000002285,
       0.00000000138,    -0.00000000160,
       0.00000002286,     0.00000002023,
      -0.00000000160,    -0.00000000139,
       0.00000000021,    -0.00000000377,
       0.00000000035,     0.00000000025,
      -0.00000000916,    -0.00000000052,
       0.00000000061,    -0.00000000086,
      -0.00000000087,     0.00000000000,
       0.00000000004,    -0.00000000090,
      -0.00000077053,     0.00000087024,
       0.00000005379,    -0.00000006157,
      -0.00000008961,    -0.00000007937,
       0.00000000635,     0.00000000550,
       0.00000000241,     0.00000000230,
       0.00000000125,    -0.00000000120,
      -0.00000000041,    -0.00000000005,
      -0.00000000012,     0.00000000099,
       0.00000000220,    -0.00000000197,
      -0.00000000197,    -0.00000000220,
       0.00000000253,     0.00000000000,
       0.00000000000,    -0.00000000615,
       0.00000012050,    -0.00000010178,
       0.00000000195,     0.00000000226,
      -0.00000002840,     0.00000003202,
       0.00000000197,    -0.00000000227,
       0.00000003200,     0.00000002837,
      -0.00000000227,    -0.00000000197,
       0.00000001571,    -0.00000001992,
       0.00000000097,     0.00000000075,
      -0.00000218643,     0.00000246532,
       0.00000015194,    -0.00000017541,
      -0.00000012748,    -0.00000011303,
       0.00000000904,     0.00000000784,
      -0.00000000095,     0.00000000179,
       0.00000000310,     0.00000000243,
      -0.00000000086,    -0.00000000048,
      -0.00000000048,     0.00000000086,
      -0.00000002115,     0.00000002681,
      -0.00000000083,    -0.00000000065,
      -0.00000000056,     0.00000000081,
      -0.00000000081,    -0.00000000056,
      -0.00000000085,    -0.00000000047,
       0.00000000047,    -0.00000000085,
       0.00000002213,    -0.00000002229,
       0.00000000072,     0.00000000073,
       0.00000000078,     0.00000000000,
       0.00000000000,    -0.00000000189,
      -0.00000001694,     0.00000001707,
      -0.00000000086,    -0.00000000087,
       0.00000002154,     0.00000005770,
      -0.00000003555,     0.00000004016,
       0.00000000247,    -0.00000000285,
       0.00000004015,     0.00000003555,
      -0.00000000285,    -0.00000000247,
      -0.00000428023,     0.00000483344,
       0.00000029806,    -0.00000034468,
      -0.00000015988,    -0.00000014159,
       0.00000001134,     0.00000000983,
      -0.00000000130,     0.00000000007,
       0.00000000007,     0.00000000130,
      -0.00000000130,     0.00000000007,
      -0.00000000007,    -0.00000000130,
       0.00000000001,     0.00000000001,
       0.00000002254,    -0.00000002813,
       0.00000000279,     0.00000000241,
      -0.00000000010,     0.00000000010,
       0.00000006792,    -0.00000004341,
       0.00000000299,     0.00000000445,
      -0.00000000012,     0.00000000008,
       0.00000004247,     0.00000006845,
      -0.00000000443,     0.00000000285,
      -0.00000000007,    -0.00000000012,
       0.00000000115,    -0.00000000717,
       0.00000000007,     0.00000000017,
       0.00000000897,     0.00000000033,
      -0.00000000014,     0.00000000024,
       0.00000000078,     0.00000000044,
      -0.00000000010,     0.00000000011,
      -0.00000000253,     0.00000000207,
      -0.00000000349,     0.00000000278,
      -0.00000000035,    -0.00000000060,
      -0.00000000300,    -0.00000000367,
       0.00000000057,    -0.00000000032,
      -0.00000000126,     0.00000000059,
      -0.00000000067,    -0.00000000126,
       0.00000000094,    -0.00000000038,
      -0.00000000318,     0.00000000134,
       0.00000000027,     0.00000000057,
      -0.00000000128,    -0.00000000312,
      -0.00000000055,     0.00000000025,
       0.00000000080,    -0.00000000008,
       0.00000000009,     0.00000000082,
      -0.00000000015,    -0.00000000121,
       0.00000000121,    -0.00000000015,
      -0.00000000039,    -0.00000000015,
      -0.00000000029,    -0.00000000133,
       0.00000000055,     0.00000000101,
      -0.00000000017,    -0.00000000462,
       0.00000000050,    -0.00000000004,
       0.00000000107,    -0.00000000047,
      -0.00000000032,     0.00000000040,
       0.00000000101,    -0.00000000272,
      -0.00000000269,    -0.00000000098,
      -0.00000000032,     0.00000000069,
       0.00000000071,     0.00000000033,
       0.00000000023,    -0.00000000030,
      -0.00000000097,     0.00000000260,
       0.00000000259,     0.00000000097,
       0.00000000022,    -0.00000000047,
      -0.00000000048,    -0.00000000023,
       0.00000000106,     0.00000000067,
       0.00000068483,    -0.00000108450,
      -0.00000000650,    -0.00000000281,
      -0.00000000961,     0.00000000618,
      -0.00000000055,    -0.00000000072,
       0.00000000617,     0.00000000952,
      -0.00000000071,     0.00000000055,
      -0.00000000235,    -0.00000000149,
      -0.00000067541,     0.00000106616,
       0.00000000619,     0.00000000293,
       0.00000000829,    -0.00000000616,
       0.00000000044,     0.00000000067,
      -0.00000000479,    -0.00000000866,
       0.00000000063,    -0.00000000057,
       0.00000001571,     0.00000001003,
       0.00000157712,    -0.00000247151,
      -0.00000001361,    -0.00000000731,
       0.00000000173,     0.00000000869,
       0.00000000077,     0.00000000004,
      -0.00000000915,     0.00000000048,
       0.00000000002,     0.00000000071,
       0.00000000099,    -0.00000000275,
       0.00000000008,     0.00000000018,
      -0.00000000188,    -0.00000000139,
       0.00000004778,     0.00000001352,
      -0.00000000345,     0.00000000174,
      -0.00000000071,    -0.00000000004,
       0.00000001322,    -0.00000004737,
       0.00000000168,     0.00000000337,
      -0.00000000003,     0.00000000070,
      -0.00000000937,    -0.00000000164,
      -0.00000000185,     0.00000000966,
      -0.00000000188,     0.00000000521,
      -0.00000000015,    -0.00000000033,
       0.00000000149,     0.00000000102,
      -0.00000000092,    -0.00000000032,
      -0.00000004472,    -0.00000001332,
       0.00000000324,    -0.00000000205,
       0.00000000067,     0.00000000005,
      -0.00000001330,     0.00000004466,
      -0.00000000204,    -0.00000000322,
       0.00000000005,    -0.00000000066,
       0.00000000641,     0.00000000113,
       0.00000000125,    -0.00000000657,
       0.00000000000,    -0.00000000008,
       0.00000027909,     0.00000007664,
      -0.00000002006,     0.00000000905,
      -0.00000000412,    -0.00000000017,
      -0.00000177229,     0.00000678821,
      -0.00000017676,    -0.00000048691,
       0.00000000185,    -0.00000009924,
       0.00000008132,     0.00000333344,
       0.00000028441,    -0.00000009219,
      -0.00000000581,    -0.00000001258,
       0.00000330679,    -0.00000009136,
      -0.00000009050,    -0.00000028660,
      -0.00000001247,     0.00000000591,
       0.00000015427,    -0.00000066914,
       0.00000005921,     0.00000000998,
      -0.00000000048,     0.00000000218,
      -0.00000069282,    -0.00000016928,
       0.00000001328,    -0.00000006213,
       0.00000000274,     0.00000000086,
      -0.00000044032,    -0.00000012093,
       0.00000003164,    -0.00000001427,
       0.00000000649,     0.00000000026,
       0.00000121408,    -0.00000464908,
       0.00000012079,     0.00000033366,
      -0.00000000130,     0.00000006791,
      -0.00000007178,    -0.00000294355,
      -0.00000025352,     0.00000008071,
       0.00000000516,     0.00000001111,
      -0.00000294157,     0.00000007138,
       0.00000008051,     0.00000025385,
       0.00000001110,    -0.00000000515,
      -0.00000010651,     0.00000045767,
      -0.00000004085,    -0.00000000703,
       0.00000000035,    -0.00000000152,
       0.00000047038,     0.00000011434,
      -0.00000000878,     0.00000004238,
      -0.00000000180,    -0.00000000057,
      -0.00000001094,    -0.00000038153,
      -0.00000003255,     0.00000001058,
       0.00000000067,     0.00000000142,
      -0.00000215710,     0.00000002793,
       0.00000005820,     0.00000018936,
       0.00000000837,    -0.00000000380,
       0.00000535428,     0.00000000405,
      -0.00000000146,     0.00000000111,
      -0.00000000019,    -0.00000000031,
      -0.00000000681,    -0.00000532790,
      -0.00000000359,    -0.00000000091,
       0.00000002242,     0.00000000273,
      -0.00000000106,    -0.00000000166,
      -0.00000000383,     0.00000000211,
      -0.00000000136,     0.00000000049,
       0.00000001381,     0.00000048155,
       0.00000004108,    -0.00000001333,
      -0.00000000084,    -0.00000000180,
       0.00000144597,    -0.00000001836,
      -0.00000003889,    -0.00000012682,
      -0.00000000562,     0.00000000254,
      -0.00000483174,     0.00000000220,
       0.00000000020,     0.00000000067,
       0.00000000041,     0.00000483355,
       0.00000000084,    -0.00000000016,
      -0.00000001543,    -0.00000000077,
       0.00000000081,     0.00000000152,
       0.00000000230,     0.00000000323,
       0.00000000141,    -0.00000000052,
       0.00000000000,    -0.00000000012,
       0.00000264045,    -0.00000000018,
      -0.00000000010,    -0.00000000051,
       0.00000000000,    -0.00000000009,
      -0.00000000150,    -0.00000943109,
      -0.00000000162,     0.00000000080,
       0.00000000762,    -0.00000003977,
       0.00000000080,     0.00000000060,
      -0.00000001528,     0.00000000409,
       0.00000000057,    -0.00000000507,
      -0.00000000616,     0.00000011626,
      -0.00000000232,    -0.00000000207,
      -0.00000002619,    -0.00000002754,
       0.00000000045,    -0.00000001020,
      -0.00000249981,     0.00000000017,
       0.00000000010,     0.00000000048,
       0.00000000103,     0.00000603627,
       0.00000000115,    -0.00000000057,
      -0.00000000498,     0.00000003335,
      -0.00000000060,    -0.00000000071,
       0.00000001186,    -0.00000000064,
      -0.00000000046,     0.00000000354,
      -0.00000000431,    -0.00000007958,
      -0.00000000155,     0.00000000142,
      -0.00000001810,     0.00000001885,
       0.00000000033,     0.00000000698,
      -0.00000000370,    -0.00000000704,
       0.00000000150,    -0.00000000006,
      -0.00000001392,     0.00000000199,
      -0.00000000004,    -0.00000000278,
       0.00000037206,     0.00000000809,
       0.00000000141,     0.00000000280,
       0.00000000829,     0.00000040486,
       0.00000000285,    -0.00000000124,
       0.00000000140,     0.00000000018,
      -0.00000000007,    -0.00000000058,
       0.00000000032,     0.00000000244,
       0.00000000055,    -0.00000000010,
       0.00000000233,     0.00000000444,
      -0.00000000094,     0.00000000004,
       0.00000000892,    -0.00000000106,
       0.00000000018,     0.00000000213,
      -0.00000014782,    -0.00000000454,
      -0.00000000075,    -0.00000000146,
      -0.00000000450,    -0.00000016544,
      -0.00000000151,     0.00000000066,
      -0.00000000096,    -0.00000000009,
      -0.00000000020,    -0.00000000153,
      -0.00000000830,     0.00000000420,
       0.00000000067,     0.00000000143,
       0.00000000013,    -0.00000000006,
       0.00000001595,     0.00000003197,
       0.00000000555,    -0.00000000261,
      -0.00000000022,    -0.00000000050,
       0.00000000088,    -0.00000000678,
       0.00000000023,     0.00000000028,
       0.00000000577,     0.00000000127,
      -0.00000000022,     0.00000000033,
       0.00000000796,     0.00000003426,
       0.00000000296,    -0.00000000057,
      -0.00000003475,     0.00000000835,
       0.00000000066,     0.00000000308,
       0.00000000257,    -0.00000000168,
      -0.00000000020,    -0.00000000040,
      -0.00000001065,    -0.00000002190,
      -0.00000000384,     0.00000000179,
       0.00000000015,     0.00000000034,
       0.00000000065,     0.00000000253,
      -0.00000000192,     0.00000000041,
      -0.00000000542,    -0.00000002346,
      -0.00000000202,     0.00000000038,
       0.00000002388,    -0.00000000577,
      -0.00000000046,    -0.00000000213,
      -0.00000000088,     0.00000000048,
       0.00000000020,     0.00000000041,
       0.00000000048,     0.00000000088,
       0.00000000041,    -0.00000000020,
      -0.00000000026,     0.00000000014,
       0.00000000006,     0.00000000012,
       0.00000000556,     0.00000001022,
       0.00000000478,    -0.00000000233,
      -0.00000000047,    -0.00000000113,
      -0.00000000134,     0.00000000187,
       0.00000000068,     0.00000000055,
       0.00000000186,     0.00000000133,
       0.00000000054,    -0.00000000068,
       0.00000000873,    -0.00000000477,
      -0.00000000199,    -0.00000000408,
      -0.00000000097,     0.00000000040,
      -0.00000002935,    -0.00000005396,
      -0.00000002526,     0.00000001229,
       0.00000000250,     0.00000000598,
       0.00000001306,    -0.00000001807,
      -0.00000000661,    -0.00000000533,
      -0.00000000110,     0.00000000121,
      -0.00000001808,    -0.00000001294,
      -0.00000000526,     0.00000000662,
       0.00000000121,     0.00000000108,
       0.00000000131,     0.00000000095,
       0.00000000039,    -0.00000000048,
      -0.00000000025,     0.00000000141,
       0.00000000042,     0.00000000010,
       0.00000000140,     0.00000000025,
       0.00000000010,    -0.00000000042,
       0.00000000389,    -0.00000000544,
      -0.00000000199,    -0.00000000159,
      -0.00000000033,     0.00000000036,
      -0.00000000718,    -0.00000000522,
      -0.00000000211,     0.00000000263,
       0.00000000048,     0.00000000043,
       0.00000000252,    -0.00000001381,
      -0.00000000416,    -0.00000000102,
      -0.00000000020,     0.00000000063,
      -0.00000001377,    -0.00000000243,
      -0.00000000098,     0.00000000414,
       0.00000000063,     0.00000000019,
      -0.00000000010,     0.00000000059,
       0.00000000018,     0.00000000004,
       0.00000000269,     0.00000000050,
       0.00000000020,    -0.00000000081,
       0.00000000076,     0.00000000286,
       0.00000000066,    -0.00000000013,
       0.00000000283,    -0.00000000078,
      -0.00000000014,    -0.00000000065,
       0.00000000269,    -0.00000001509,
      -0.00000000454,    -0.00000000110,
      -0.00000000021,     0.00000000069,
      -0.00000001474,    -0.00000000274,
      -0.00000000107,     0.00000000443,
       0.00000000067,     0.00000000020,
      -0.00000000756,    -0.00000002839,
      -0.00000000660,     0.00000000129,
       0.00000000010,     0.00000000077,
      -0.00000002818,     0.00000000773,
       0.00000000136,     0.00000000652,
       0.00000000076,    -0.00000000012,
       0.00000000048,     0.00000000176,
       0.00000000041,    -0.00000000008,
       0.00000000638,    -0.00000000169,
      -0.00000000030,    -0.00000000147,
       0.00000000427,     0.00000000513,
       0.00000000083,    -0.00000000059,
       0.00000000499,    -0.00000000429,
      -0.00000000060,    -0.00000000079,
      -0.00000001050,    -0.00000003874,
      -0.00000000899,     0.00000000180,
       0.00000000015,     0.00000000105,
      -0.00000003434,     0.00000000910,
       0.00000000160,     0.00000000793,
       0.00000000093,    -0.00000000013,
      -0.00000004312,    -0.00000005153,
      -0.00000000834,     0.00000000599,
       0.00000000043,     0.00000000068,
      -0.00000005062,     0.00000004331,
       0.00000000606,     0.00000000807,
       0.00000000064,    -0.00000000045,
       0.00000000024,    -0.00000000046,
      -0.00000000042,    -0.00000000016,
       0.00000000352,     0.00000000416,
       0.00000000067,    -0.00000000049,
       0.00000001294,    -0.00000001087,
      -0.00000000152,    -0.00000000207,
       0.00000001174,     0.00000000569,
       0.00000000054,    -0.00000000087,
       0.00000000513,    -0.00000001160,
      -0.00000000084,    -0.00000000041,
      -0.00000006459,    -0.00000007630,
      -0.00000001231,     0.00000000901,
       0.00000000065,     0.00000000099,
      -0.00000006742,     0.00000005666,
       0.00000000792,     0.00000001078,
       0.00000000086,    -0.00000000058,
      -0.00000012045,    -0.00000005712,
      -0.00000000520,     0.00000000890,
       0.00000000036,     0.00000000023,
      -0.00000005404,     0.00000011970,
       0.00000000875,     0.00000000450,
       0.00000000015,    -0.00000000034,
       0.00000000024,    -0.00000000060,
      -0.00000000043,    -0.00000000009,
       0.00000001265,     0.00000000586,
       0.00000000052,    -0.00000000093,
       0.00000001626,    -0.00000003499,
      -0.00000000257,    -0.00000000141,
       0.00000001634,     0.00000000094,
      -0.00000000087,    -0.00000001486,
      -0.00000018561,    -0.00000008595,
      -0.00000000758,     0.00000001364,
       0.00000000054,     0.00000000030,
      -0.00000008116,     0.00000017458,
       0.00000001281,     0.00000000705,
       0.00000000028,    -0.00000000051,
      -0.00000016814,    -0.00000000414,
      -0.00000000043,     0.00000000045,
       0.00000000360,     0.00000016179,
      -0.00000000045,    -0.00000000078,
       0.00000000051,    -0.00000000071,
      -0.00000000026,    -0.00000000035,
       0.00000002325,     0.00000000005,
       0.00000000009,    -0.00000005893,
       0.00000000204,     0.00000000261,
      -0.00000000095,     0.00000000550,
      -0.00000025599,    -0.00000000061,
      -0.00000000047,     0.00000027839,
       0.00000000420,    -0.00000001330,
       0.00000000013,     0.00000000204,
      -0.00000000213,    -0.00000002790,
       0.00000000030,    -0.00000000107,
       0.00000000024,    -0.00000000135,
      -0.00000000066,    -0.00000000020,
      -0.00000000177,     0.00000000091,
      -0.00000000006,    -0.00000000013,
       0.00000000188,     0.00000000379,
       0.00000001056,    -0.00000000006,
      -0.00000000006,     0.00000001051,
       0.00000001300,    -0.00000000677,
       0.00000000040,     0.00000000095,
      -0.00000000926,    -0.00000001872,
       0.00000000136,    -0.00000000066,
      -0.00000002177,     0.00000000031,
       0.00000000041,    -0.00000002131,
      -0.00000000044,     0.00000000019,
      -0.00000000019,    -0.00000000044,
       0.00000000093,     0.00000000005,
       0.00000000084,    -0.00000000045,
       0.00000000049,     0.00000000081,
       0.00000000018,    -0.00000000082,
       0.00000000082,     0.00000000018,
      -0.00000000034,    -0.00000000026,
       0.00000000044,    -0.00000000037,
      -0.00000000035,    -0.00000000043,
       0.00000000132,    -0.00000000111,
       0.00000000000,    -0.00000000026,
      -0.00000000239,    -0.00000000290,
      -0.00000000047,     0.00000000002,
       0.00000000338,    -0.00000000175,
      -0.00000000153,    -0.00000000321,
       0.00000001202,    -0.00000000599,
       0.00000000029,    -0.00000000132,
      -0.00000000006,    -0.00000000004,
      -0.00000001281,    -0.00000002600,
      -0.00000000285,    -0.00000000062,
       0.00000001938,    -0.00000000514,
       0.00000000039,    -0.00000000107,
      -0.00000000333,    -0.00000001707,
      -0.00000000062,    -0.00000000040,
       0.00000008229,    -0.00000001918,
       0.00000000182,    -0.00000000382,
      -0.00000000012,    -0.00000000013,
      -0.00000004074,    -0.00000017615,
      -0.00000000819,    -0.00000000384,
      -0.00000000027,     0.00000000025,
       0.00000006966,    -0.00000000555,
       0.00000000023,    -0.00000000140,
       0.00000000658,    -0.00000004420,
       0.00000000136,     0.00000000036,
      -0.00000000012,     0.00000000059,
      -0.00000000057,    -0.00000000016,
       0.00000033621,     0.00000000261,
      -0.00000000016,    -0.00000000011,
       0.00000000378,    -0.00000071368,
      -0.00000000018,     0.00000000043,
       0.00000010705,    -0.00000002461,
       0.00000000246,    -0.00000000560,
       0.00000002671,     0.00000011873,
       0.00000000495,     0.00000000238,
      -0.00000000101,     0.00000000218,
      -0.00000000207,    -0.00000000107,
      -0.00000000047,    -0.00000000030,
       0.00000000032,     0.00000000021,
       0.00000000042,     0.00000000014,
      -0.00000000021,    -0.00000000031,
      -0.00000002851,     0.00000001060,
       0.00000000054,    -0.00000000174,
       0.00000001386,     0.00000005243,
      -0.00000000306,    -0.00000000061,
       0.00000065984,    -0.00000000056,
       0.00000000052,     0.00000065963,
       0.00000010491,     0.00000005234,
       0.00000000089,    -0.00000000035,
      -0.00000000012,    -0.00000000028,
      -0.00000000074,    -0.00000000190,
      -0.00000000060,     0.00000000026,
       0.00000000070,    -0.00000000016,
      -0.00000000039,    -0.00000000003,
      -0.00000000008,     0.00000000081,
       0.00000000030,    -0.00000000037,
      -0.00000001519,    -0.00000000404,
      -0.00000000037,    -0.00000000068,
       0.00000000404,    -0.00000001518,
       0.00000000068,    -0.00000000037,
       0.00000000104,     0.00000000208,
      -0.00000000208,     0.00000000104,
       0.00000000107,    -0.00000000036,
      -0.00000000030,    -0.00000000045,
       0.00000000281,    -0.00000000041,
      -0.00000000021,    -0.00000000081,
      -0.00000000088,    -0.00000000597,
      -0.00000000173,     0.00000000045,
       0.00000000078,    -0.00000000029,
       0.00000000037,     0.00000000089,
      -0.00000000057,     0.00000000012,
       0.00000000010,     0.00000000011,
       0.00000000034,     0.00000000120,
       0.00000000367,    -0.00000000076,
      -0.00000000030,    -0.00000000107,
       0.00000000019,     0.00000000370,
       0.00000000107,    -0.00000000030,
      -0.00000000010,    -0.00000000239,
      -0.00000000023,     0.00000000016,
      -0.00000000500,     0.00000000022,
       0.00000000033,     0.00000000049,
      -0.00000000006,    -0.00000000160,
       0.00000000866,     0.00000000009,
       0.00000000013,    -0.00000001803,
      -0.00000000032,    -0.00000000604,
      -0.00000000057,     0.00000000040,
       0.00000000601,    -0.00000000023,
      -0.00000000040,    -0.00000000059,
      -0.00000000058,     0.00000000021,
       0.00000000004,     0.00000000086,
       0.00000003104,    -0.00000000005,
       0.00000000005,     0.00000003104,
       0.00000000005,    -0.00000000087,
       0.00000000087,     0.00000000005,
      -0.00000000108,    -0.00000000874,
       0.00000000091,    -0.00000000011,
       0.00000000058,     0.00000000000,
       0.00000000000,     0.00000000058,
      -0.00000000300,    -0.00000002486,
       0.00000000129,    -0.00000000016,
      -0.00000000592,    -0.00000004874,
       0.00000000162,    -0.00000000020,
      -0.00000000583,    -0.00000000193,
       0.00000000033,    -0.00000000079,
      -0.00000001101,    -0.00000000905,
       0.00000000092,    -0.00000000111,
       0.00000000003,     0.00000000002,
       0.00000000920,    -0.00000001091,
       0.00000000108,     0.00000000093,
      -0.00000000002,     0.00000000003,
      -0.00000000165,     0.00000000023,
       0.00000000003,    -0.00000000008,
      -0.00000000048,    -0.00000000192,
       0.00000000012,     0.00000000000,
       0.00000000090,     0.00000000077,
      -0.00000000077,     0.00000000091,
      -0.00000000022,     0.00000000061,
       0.00000000065,    -0.00000000091,
      -0.00000000122,     0.00000000413,
      -0.00000000186,    -0.00000000044,
      -0.00000000024,    -0.00000000014,
      -0.00000000030,     0.00000000034,
       0.00000000141,    -0.00000000405,
      -0.00000000401,    -0.00000000139,
      -0.00000000039,     0.00000000091,
       0.00000000093,     0.00000000041,
       0.00000000046,     0.00000000026,
      -0.00000000136,     0.00000000384,
       0.00000000384,     0.00000000136,
       0.00000000027,    -0.00000000062,
      -0.00000000063,    -0.00000000028,
       0.00000000263,    -0.00000000773,
       0.00000000012,     0.00000000051,
       0.00000002931,     0.00000000827,
      -0.00000000203,    -0.00000000008,
       0.00000010763,     0.00000002653,
      -0.00000000766,     0.00000000172,
      -0.00000000158,     0.00000000001,
       0.00000002603,    -0.00000010692,
       0.00000000162,     0.00000000751,
       0.00000000003,     0.00000000156,
      -0.00000001877,    -0.00000000330,
      -0.00000000376,     0.00000001942,
      -0.00000000006,    -0.00000000046,
      -0.00000000436,     0.00000001281,
      -0.00000000020,    -0.00000000085,
      -0.00000001964,    -0.00000000552,
      -0.00000000017,    -0.00000000087,
      -0.00000010013,    -0.00000002602,
       0.00000000715,    -0.00000000240,
       0.00000000148,     0.00000000002,
      -0.00000002604,     0.00000010012,
      -0.00000000239,    -0.00000000713,
       0.00000000002,    -0.00000000148,
       0.00000001283,     0.00000000228,
       0.00000000253,    -0.00000001319,
      -0.00000007939,    -0.00000001886,
       0.00000000563,    -0.00000000093,
       0.00000000116,    -0.00000000003,
      -0.00000019403,     0.00000072899,
      -0.00000001779,    -0.00000005171,
       0.00000000021,    -0.00000001083,
      -0.00000002234,    -0.00000080176,
      -0.00000006822,     0.00000002223,
       0.00000000139,     0.00000000299,
      -0.00000079740,     0.00000002391,
       0.00000002195,     0.00000006867,
       0.00000000298,    -0.00000000141,
      -0.00000002924,     0.00000012678,
      -0.00000001134,    -0.00000000182,
       0.00000000010,    -0.00000000041,
       0.00000013248,     0.00000003259,
      -0.00000000255,     0.00000001192,
      -0.00000000052,    -0.00000000017,
       0.00000011272,     0.00000002678,
      -0.00000000798,     0.00000000137,
      -0.00000000164,     0.00000000004,
       0.00000013277,    -0.00000049891,
       0.00000001223,     0.00000003530,
      -0.00000000014,     0.00000000742,
       0.00000002027,     0.00000071804,
       0.00000006153,    -0.00000001978,
      -0.00000000126,    -0.00000000268,
       0.00000071862,    -0.00000001978,
      -0.00000001979,    -0.00000006159,
      -0.00000000269,     0.00000000125,
       0.00000002027,    -0.00000008670,
       0.00000000785,     0.00000000128,
      -0.00000008960,    -0.00000002194,
       0.00000000166,    -0.00000000814,
      -0.00000000764,    -0.00000024273,
      -0.00000002061,     0.00000000676,
       0.00000000042,     0.00000000090,
      -0.00000095536,     0.00000001419,
       0.00000002576,     0.00000008380,
       0.00000000368,    -0.00000000168,
       0.00000251699,     0.00000000100,
      -0.00000000042,     0.00000000022,
      -0.00000000210,    -0.00000251069,
      -0.00000000181,    -0.00000000029,
       0.00000001269,     0.00000000133,
      -0.00000000058,    -0.00000000084,
      -0.00000000225,     0.00000000142,
      -0.00000000070,     0.00000000025,
       0.00000000904,     0.00000028722,
       0.00000002439,    -0.00000000797,
      -0.00000000050,    -0.00000000106,
       0.00000063925,    -0.00000000931,
      -0.00000001715,    -0.00000005605,
      -0.00000000247,     0.00000000112,
      -0.00000229665,     0.00000000144,
      -0.00000000008,     0.00000000051,
      -0.00000000034,     0.00000230089,
       0.00000000058,    -0.00000000013,
      -0.00000000872,    -0.00000000018,
       0.00000000046,     0.00000000082,
       0.00000000142,     0.00000000209,
       0.00000000077,    -0.00000000030,
       0.00000168309,     0.00000000003,
      -0.00000000004,    -0.00000000047,
      -0.00000000068,    -0.00000510584,
      -0.00000000128,     0.00000000037,
       0.00000000282,    -0.00000004778,
       0.00000000013,     0.00000000086,
       0.00000000040,     0.00000000355,
      -0.00000000006,    -0.00000000555,
      -0.00000000393,     0.00000006905,
      -0.00000000149,    -0.00000000122,
      -0.00000001665,    -0.00000001637,
       0.00000000028,    -0.00000000607,
      -0.00000159326,    -0.00000000003,
       0.00000000004,     0.00000000045,
       0.00000000046,     0.00000329412,
       0.00000000090,    -0.00000000027,
      -0.00000000148,     0.00000003545,
      -0.00000000059,    -0.00000000073,
       0.00000000518,    -0.00000000149,
      -0.00000000010,     0.00000000383,
      -0.00000000275,    -0.00000004727,
      -0.00000000098,     0.00000000084,
      -0.00000001157,     0.00000001121,
       0.00000000021,     0.00000000416,
      -0.00000000410,    -0.00000000986,
       0.00000000186,    -0.00000000003,
      -0.00000002386,     0.00000000446,
       0.00000000009,    -0.00000000400,
       0.00000033836,     0.00000000629,
       0.00000000106,     0.00000000210,
       0.00000000630,     0.00000036351,
       0.00000000223,    -0.00000000097,
       0.00000000176,    -0.00000000007,
      -0.00000000009,    -0.00000000043,
       0.00000000032,     0.00000000224,
       0.00000000042,    -0.00000000011,
       0.00000000291,     0.00000000700,
      -0.00000000132,     0.00000000002,
       0.00000001569,    -0.00000000279,
       0.00000000004,     0.00000000286,
      -0.00000014388,    -0.00000000377,
      -0.00000000063,    -0.00000000124,
      -0.00000000373,    -0.00000015889,
      -0.00000000132,     0.00000000058,
      -0.00000000120,     0.00000000009,
      -0.00000000019,    -0.00000000136,
      -0.00000000765,     0.00000000385,
       0.00000000061,     0.00000000133,
       0.00000000012,    -0.00000000005,
       0.00000001254,     0.00000002511,
       0.00000000439,    -0.00000000204,
      -0.00000000017,    -0.00000000040,
      -0.00000000167,    -0.00000000316,
       0.00000000247,    -0.00000000143,
       0.00000000000,     0.00000000049,
       0.00000000549,     0.00000002368,
       0.00000000204,    -0.00000000039,
      -0.00000002404,     0.00000000579,
       0.00000000046,     0.00000000214,
       0.00000000362,    -0.00000000182,
      -0.00000000029,    -0.00000000063,
      -0.00000000858,    -0.00000001717,
      -0.00000000300,     0.00000000140,
       0.00000000071,     0.00000000161,
      -0.00000000117,     0.00000000054,
      -0.00000000374,    -0.00000001622,
      -0.00000000139,     0.00000000026,
       0.00000001655,    -0.00000000401,
      -0.00000000033,    -0.00000000148,
      -0.00000000094,     0.00000000035,
       0.00000000071,    -0.00000000034,
       0.00000000022,     0.00000000060,
      -0.00000000073,     0.00000000004,
      -0.00000000005,    -0.00000000075,
       0.00000000066,    -0.00000000024,
      -0.00000000182,     0.00000000091,
      -0.00000000085,    -0.00000000176,
       0.00000000050,    -0.00000000002,
       0.00000000004,     0.00000000052,
      -0.00000000016,     0.00000000038,
       0.00000000083,    -0.00000000200,
      -0.00000000120,    -0.00000000055,
      -0.00000000019,     0.00000000036,
       0.00000000103,    -0.00000000006,
       0.00000000000,    -0.00000000055,
      -0.00000000006,    -0.00000000102,
      -0.00000000054,     0.00000000001,
      -0.00000000003,    -0.00000000050,
       0.00000000075,    -0.00000000041,
      -0.00000000017,    -0.00000000035,
      -0.00000000042,    -0.00000000074,
      -0.00000000035,     0.00000000017,
       0.00000000085,    -0.00000000048,
      -0.00000000020,    -0.00000000040,
      -0.00000000043,    -0.00000000078,
      -0.00000000037,     0.00000000018,
       0.00000000114,    -0.00000000160,
      -0.00000000058,    -0.00000000047,
      -0.00000000160,    -0.00000000112,
      -0.00000000046,     0.00000000058,
       0.00000000033,     0.00000000023,
      -0.00000000009,     0.00000000051,
       0.00000000051,     0.00000000008,
       0.00000000160,    -0.00000000227,
      -0.00000000083,    -0.00000000066,
      -0.00000000014,     0.00000000015,
      -0.00000000182,    -0.00000000131,
      -0.00000000053,     0.00000000066,
       0.00000000092,    -0.00000000510,
      -0.00000000153,    -0.00000000038,
      -0.00000000507,    -0.00000000086,
      -0.00000000035,     0.00000000152,
      -0.00000000006,     0.00000000036,
       0.00000000111,     0.00000000020,
       0.00000000034,     0.00000000129,
       0.00000000126,    -0.00000000036,
       0.00000000138,    -0.00000000793,
      -0.00000000238,    -0.00000000058,
      -0.00000000011,     0.00000000036,
      -0.00000000610,    -0.00000000111,
      -0.00000000044,     0.00000000183,
      -0.00000000345,    -0.00000001302,
      -0.00000000303,     0.00000000058,
      -0.00000001285,     0.00000000359,
       0.00000000063,     0.00000000297,
       0.00000000030,     0.00000000110,
       0.00000000025,    -0.00000000005,
       0.00000000304,    -0.00000000081,
      -0.00000000014,    -0.00000000070,
       0.00000000194,     0.00000000238,
       0.00000000039,    -0.00000000027,
       0.00000000224,    -0.00000000197,
      -0.00000000028,    -0.00000000035,
      -0.00000000577,    -0.00000002114,
      -0.00000000491,     0.00000000098,
       0.00000000008,     0.00000000057,
      -0.00000001631,     0.00000000437,
       0.00000000076,     0.00000000377,
       0.00000000044,    -0.00000000006,
      -0.00000002001,    -0.00000002410,
      -0.00000000393,     0.00000000277,
       0.00000000020,     0.00000000032,
      -0.00000002334,     0.00000002016,
       0.00000000283,     0.00000000370,
       0.00000000029,    -0.00000000021,
       0.00000000204,     0.00000000240,
       0.00000000039,    -0.00000000028,
       0.00000000613,    -0.00000000516,
      -0.00000000072,    -0.00000000098,
       0.00000000499,     0.00000000256,
       0.00000000026,    -0.00000000037,
       0.00000000201,    -0.00000000484,
      -0.00000003361,    -0.00000003969,
      -0.00000000640,     0.00000000467,
       0.00000000034,     0.00000000051,
      -0.00000003191,     0.00000002686,
       0.00000000375,     0.00000000511,
       0.00000000041,    -0.00000000027,
      -0.00000005171,    -0.00000002505,
      -0.00000000234,     0.00000000384,
      -0.00000002262,     0.00000005109,
       0.00000000372,     0.00000000179,
       0.00000000631,     0.00000000293,
       0.00000000026,    -0.00000000046,
       0.00000000698,    -0.00000001500,
      -0.00000000110,    -0.00000000061,
       0.00000000618,     0.00000000077,
      -0.00000000072,    -0.00000000494,
      -0.00000008678,    -0.00000004028,
      -0.00000000354,     0.00000000636,
       0.00000000025,     0.00000000014,
      -0.00000003497,     0.00000007513,
       0.00000000552,     0.00000000303,
      -0.00000006194,    -0.00000000294,
      -0.00000000034,     0.00000000031,
       0.00000000253,     0.00000005739,
      -0.00000000033,    -0.00000000052,
       0.00000000025,    -0.00000000033,
       0.00000000920,     0.00000000004,
       0.00000000006,    -0.00000002073,
       0.00000000154,     0.00000000161,
      -0.00000000071,     0.00000000341,
      -0.00000010129,    -0.00000000045,
      -0.00000000033,     0.00000009910,
       0.00000000189,    -0.00000000777,
       0.00000000004,     0.00000000119,
      -0.00000000102,    -0.00000001627,
       0.00000000013,    -0.00000000062,
       0.00000000012,    -0.00000000054,
      -0.00000000117,     0.00000000060,
       0.00000000118,     0.00000000237,
       0.00000000491,    -0.00000000004,
      -0.00000000007,     0.00000000486,
       0.00000000966,    -0.00000000493,
       0.00000000031,     0.00000000070,
      -0.00000000585,    -0.00000001183,
       0.00000000086,    -0.00000000041,
      -0.00000001081,     0.00000000039,
       0.00000000048,    -0.00000001033,
      -0.00000000046,     0.00000000025,
      -0.00000000025,    -0.00000000046,
       0.00000000051,     0.00000000018,
       0.00000000073,    -0.00000000040,
       0.00000000040,     0.00000000072,
       0.00000000030,    -0.00000000025,
      -0.00000000049,    -0.00000000060,
       0.00000000058,    -0.00000000031,
      -0.00000000024,    -0.00000000052,
       0.00000000231,    -0.00000000115,
       0.00000000006,    -0.00000000025,
      -0.00000000223,    -0.00000000453,
      -0.00000000041,    -0.00000000008,
       0.00000000292,    -0.00000000086,
      -0.00000000032,    -0.00000000224,
       0.00000001284,    -0.00000000294,
       0.00000000028,    -0.00000000060,
      -0.00000000572,    -0.00000002502,
      -0.00000000116,    -0.00000000054,
       0.00000000956,    -0.00000000135,
       0.00000000160,    -0.00000000345,
       0.00000004020,     0.00000000067,
       0.00000000098,    -0.00000007799,
       0.00000001994,    -0.00000000451,
       0.00000000046,    -0.00000000103,
       0.00000000490,     0.00000002182,
       0.00000000092,     0.00000000045,
      -0.00000000510,     0.00000000179,
       0.00000000009,    -0.00000000031,
       0.00000000236,     0.00000000885,
      -0.00000000047,    -0.00000000014,
       0.00000008536,     0.00000000089,
      -0.00000000092,     0.00000008529,
      -0.00000000842,     0.00000000396,
       0.00000000023,    -0.00000000050,
      -0.00000000396,    -0.00000000842,
       0.00000000050,     0.00000000023,
       0.00000000045,    -0.00000000006,
      -0.00000000012,    -0.00000000080,
       0.00000000065,    -0.00000000010,
       0.00000000010,     0.00000000065,
       0.00000000057,     0.00000000001,
       0.00000000002,    -0.00000000109,
      -0.00000000003,    -0.00000000062,
       0.00000000061,    -0.00000000002,
       0.00000000224,     0.00000000003,
      -0.00000000003,     0.00000000224,
       0.00000000006,     0.00000000110,
      -0.00000000018,    -0.00000000001,
      -0.00000000091,     0.00000000245,
      -0.00000000031,    -0.00000000013,
      -0.00000000245,    -0.00000000095,
       0.00000000013,    -0.00000000031,
       0.00000000016,     0.00000000033,
      -0.00000000036,     0.00000000022,
      -0.00000000001,    -0.00000000004,
      -0.00000000044,    -0.00000000029,
       0.00000000052,     0.00000000034,
      -0.00000000085,    -0.00000000063,
      -0.00000000037,    -0.00000000020,
      -0.00000000020,     0.00000000037,
       0.00000000035,     0.00000000019,
       0.00000000019,    -0.00000000035,
      -0.00000000050,    -0.00000000026,
       0.00000000231,    -0.00000000703,
       0.00000000004,     0.00000000048,
      -0.00000000699,    -0.00000000227,
       0.00000000047,    -0.00000000003,
      -0.00000000057,     0.00000000141,
       0.00000000145,     0.00000000061,
       0.00000000085,     0.00000000045,
      -0.00000000222,     0.00000000665,
      -0.00000000007,    -0.00000000044,
       0.00000000665,     0.00000000222,
      -0.00000000044,     0.00000000007,
       0.00000000039,    -0.00000000096,
      -0.00000000098,    -0.00000000041,
      -0.00000001098,     0.00000003430,
      -0.00000000008,    -0.00000000232,
      -0.00000000005,    -0.00000000033,
       0.00000099460,     0.00000032529,
      -0.00000006689,     0.00000000456,
      -0.00000000948,     0.00000000112,
      -0.00000040378,    -0.00000008716,
       0.00000002836,     0.00000000084,
       0.00000000588,    -0.00000000036,
      -0.00000008582,     0.00000040182,
       0.00000000114,    -0.00000002795,
      -0.00000000042,    -0.00000000583,
       0.00000006328,     0.00000001115,
      -0.00000000124,    -0.00000000002,
       0.00000001288,    -0.00000006581,
       0.00000000018,     0.00000000157,
       0.00000001665,    -0.00000005199,
       0.00000000015,     0.00000000352,
       0.00000000007,     0.00000000050,
      -0.00000068293,    -0.00000022332,
       0.00000005059,    -0.00000000024,
       0.00000000640,    -0.00000000082,
       0.00000037598,     0.00000008528,
      -0.00000002645,     0.00000000181,
      -0.00000000548,     0.00000000024,
       0.00000008558,    -0.00000037630,
       0.00000000179,     0.00000002643,
       0.00000000024,     0.00000000549,
      -0.00000004327,    -0.00000000774,
       0.00000000087,    -0.00000000002,
      -0.00000000863,     0.00000004459,
      -0.00000000013,    -0.00000000104,
      -0.00000000047,    -0.00000000006,
      -0.00000003759,    -0.00000000779,
       0.00000000263,     0.00000000017,
       0.00000000054,    -0.00000000004,
      -0.00000004836,     0.00000019419,
      -0.00000000251,    -0.00000001358,
      -0.00000000004,    -0.00000000289,
      -0.00000001006,    -0.00000033147,
      -0.00000002820,     0.00000000917,
       0.00000000057,     0.00000000123,
      -0.00000033044,     0.00000001043,
       0.00000000910,     0.00000002837,
       0.00000000123,    -0.00000000058,
      -0.00000000994,     0.00000004302,
      -0.00000000390,    -0.00000000059,
       0.00000004543,     0.00000001126,
      -0.00000000087,     0.00000000410,
       0.00000004982,     0.00000001032,
      -0.00000000349,    -0.00000000022,
      -0.00000000072,     0.00000000006,
       0.00000003308,    -0.00000013286,
       0.00000000177,     0.00000000927,
       0.00000000003,     0.00000000198,
       0.00000000924,     0.00000029974,
       0.00000002565,    -0.00000000824,
      -0.00000000052,    -0.00000000111,
       0.00000030038,    -0.00000000891,
      -0.00000000828,    -0.00000002568,
      -0.00000000112,     0.00000000052,
       0.00000000692,    -0.00000002941,
       0.00000000271,     0.00000000042,
      -0.00000003060,    -0.00000000756,
       0.00000000056,    -0.00000000281,
      -0.00000000568,    -0.00000017073,
      -0.00000001445,     0.00000000475,
       0.00000000030,     0.00000000063,
      -0.00000054774,     0.00000000941,
       0.00000001478,     0.00000004791,
       0.00000000210,    -0.00000000096,
       0.00000140152,     0.00000000019,
      -0.00000000055,    -0.00000140191,
      -0.00000000116,    -0.00000000009,
       0.00000000828,     0.00000000077,
      -0.00000000037,    -0.00000000050,
      -0.00000000150,     0.00000000103,
      -0.00000000042,     0.00000000016,
       0.00000000645,     0.00000019394,
       0.00000001636,    -0.00000000539,
      -0.00000000034,    -0.00000000071,
       0.00000036661,    -0.00000000619,
      -0.00000000986,    -0.00000003210,
      -0.00000000141,     0.00000000064,
      -0.00000128829,     0.00000000093,
      -0.00000000016,     0.00000000038,
      -0.00000000052,     0.00000129280,
       0.00000000047,    -0.00000000010,
      -0.00000000568,    -0.00000000001,
       0.00000000030,     0.00000000052,
       0.00000000098,     0.00000000148,
       0.00000000049,    -0.00000000020,
       0.00000004116,     0.00000023950,
       0.00000003076,    -0.00000001764,
      -0.00000000021,    -0.00000000063,
      -0.00000000063,     0.00000000021,
      -0.00000002043,    -0.00000012335,
      -0.00000001571,     0.00000000833,
       0.00000000019,     0.00000000060,
       0.00000000060,    -0.00000000019,
      -0.00000005731,    -0.00000032631,
      -0.00000004119,     0.00000002623,
       0.00000000081,    -0.00000000065,
      -0.00000000065,    -0.00000000080,
       0.00000109965,     0.00000000017,
       0.00000000000,    -0.00000000042,
      -0.00000000018,    -0.00000297815,
      -0.00000000101,     0.00000000015,
       0.00000000055,    -0.00000004544,
      -0.00000000015,     0.00000000086,
       0.00000000691,     0.00000000285,
      -0.00000000033,    -0.00000000512,
      -0.00000000258,     0.00000004303,
      -0.00000000098,    -0.00000000076,
      -0.00000001084,    -0.00000001022,
       0.00000000018,    -0.00000000380,
      -0.00000104096,    -0.00000000016,
       0.00000000000,     0.00000000040,
       0.00000000012,     0.00000193313,
       0.00000000072,    -0.00000000011,
       0.00000000011,     0.00000003206,
      -0.00000000052,    -0.00000000064,
       0.00000000179,    -0.00000000165,
       0.00000000006,     0.00000000351,
      -0.00000000179,    -0.00000002946,
      -0.00000000064,     0.00000000052,
      -0.00000000759,     0.00000000700,
       0.00000000014,     0.00000000260,
      -0.00000000383,    -0.00000001030,
       0.00000000186,    -0.00000000001,
      -0.00000000004,     0.00000000005,
      -0.00000002471,     0.00000000486,
       0.00000000011,    -0.00000000403,
       0.00000026608,     0.00000000484,
       0.00000000081,     0.00000000157,
       0.00000000479,     0.00000028532,
       0.00000000172,    -0.00000000074,
       0.00000000172,    -0.00000000018,
       0.00000000029,     0.00000000192,
       0.00000000290,     0.00000000780,
      -0.00000000141,     0.00000000001,
       0.00000001637,    -0.00000000312,
       0.00000000000,     0.00000000283,
      -0.00000011661,    -0.00000000311,
      -0.00000000051,    -0.00000000100,
      -0.00000000306,    -0.00000012895,
      -0.00000000109,     0.00000000048,
      -0.00000000117,     0.00000000016,
      -0.00000000017,    -0.00000000115,
      -0.00000000661,     0.00000000334,
       0.00000000053,     0.00000000115,
       0.00000000010,    -0.00000000005,
       0.00000000973,     0.00000001941,
       0.00000000339,    -0.00000000157,
      -0.00000000162,    -0.00000000285,
       0.00000000240,    -0.00000000149,
       0.00000000001,     0.00000000050,
       0.00000000378,     0.00000001635,
       0.00000000141,    -0.00000000027,
      -0.00000001662,     0.00000000401,
       0.00000000032,     0.00000000148,
       0.00000000376,    -0.00000000190,
      -0.00000000030,    -0.00000000065,
      -0.00000000666,    -0.00000001328,
      -0.00000000232,     0.00000000108,
       0.00000000068,     0.00000000131,
      -0.00000000100,     0.00000000058,
      -0.00000000257,    -0.00000001120,
      -0.00000000096,     0.00000000018,
       0.00000001146,    -0.00000000278,
      -0.00000000023,    -0.00000000103,
      -0.00000000065,     0.00000000021,
       0.00000000045,    -0.00000000015,
      -0.00000000100,     0.00000000051,
      -0.00000000045,    -0.00000000095,
       0.00000000049,    -0.00000000028,
      -0.00000000019,    -0.00000000035,
       0.00000000059,    -0.00000000083,
      -0.00000000030,    -0.00000000025,
      -0.00000000083,    -0.00000000058,
      -0.00000000024,     0.00000000030,
       0.00000000098,    -0.00000000140,
      -0.00000000051,    -0.00000000041,
      -0.00000000097,    -0.00000000069,
      -0.00000000028,     0.00000000035,
       0.00000000051,    -0.00000000281,
      -0.00000000085,    -0.00000000021,
      -0.00000000278,    -0.00000000046,
      -0.00000000019,     0.00000000084,
       0.00000000062,     0.00000000011,
       0.00000000017,     0.00000000069,
       0.00000000066,    -0.00000000020,
       0.00000000083,    -0.00000000483,
      -0.00000000145,    -0.00000000035,
      -0.00000000007,     0.00000000022,
      -0.00000000340,    -0.00000000061,
      -0.00000000025,     0.00000000102,
      -0.00000000185,    -0.00000000703,
      -0.00000000164,     0.00000000030,
      -0.00000000689,     0.00000000196,
       0.00000000035,     0.00000000159,
       0.00000000019,     0.00000000069,
       0.00000000016,    -0.00000000003,
       0.00000000166,    -0.00000000045,
       0.00000000096,     0.00000000121,
       0.00000000108,    -0.00000000098,
      -0.00000000334,    -0.00000001222,
      -0.00000000284,     0.00000000057,
       0.00000000005,     0.00000000033,
      -0.00000000886,     0.00000000239,
       0.00000000041,     0.00000000205,
      -0.00000001002,    -0.00000001221,
      -0.00000000200,     0.00000000138,
      -0.00000001161,     0.00000001013,
       0.00000000143,     0.00000000182,
       0.00000000116,     0.00000000137,
       0.00000000022,    -0.00000000016,
       0.00000000310,    -0.00000000261,
      -0.00000000036,    -0.00000000050,
       0.00000000223,     0.00000000123,
       0.00000000079,    -0.00000000212,
      -0.00000001785,    -0.00000002110,
      -0.00000000340,     0.00000000248,
       0.00000000018,     0.00000000027,
      -0.00000001612,     0.00000001357,
       0.00000000189,     0.00000000258,
      -0.00000002326,    -0.00000001158,
      -0.00000000112,     0.00000000173,
      -0.00000000984,     0.00000002280,
       0.00000000166,     0.00000000072,
       0.00000000310,     0.00000000144,
       0.00000000013,    -0.00000000023,
       0.00000000312,    -0.00000000670,
      -0.00000000049,    -0.00000000027,
       0.00000000250,     0.00000000053,
      -0.00000000050,    -0.00000000163,
      -0.00000004095,    -0.00000001908,
      -0.00000000166,     0.00000000300,
      -0.00000001570,     0.00000003365,
       0.00000000247,     0.00000000135,
      -0.00000002389,    -0.00000000188,
       0.00000000161,     0.00000002098,
      -0.00000000021,    -0.00000000033,
       0.00000000367,     0.00000000003,
       0.00000000004,    -0.00000000765,
       0.00000000093,     0.00000000090,
      -0.00000000043,     0.00000000190,
      -0.00000004043,    -0.00000000029,
      -0.00000000021,     0.00000003685,
       0.00000000086,    -0.00000000420,
       0.00000000001,     0.00000000064,
      -0.00000000048,    -0.00000000878,
      -0.00000000068,     0.00000000034,
       0.00000000065,     0.00000000131,
       0.00000000217,    -0.00000000003,
      -0.00000000005,     0.00000000213,
       0.00000000597,    -0.00000000302,
       0.00000000019,     0.00000000043,
      -0.00000000323,    -0.00000000657,
       0.00000000048,    -0.00000000023,
      -0.00000000499,     0.00000000030,
       0.00000000037,    -0.00000000459,
       0.00000000041,     0.00000000019,
       0.00000000054,    -0.00000000030,
       0.00000000028,     0.00000000053,
       0.00000000042,    -0.00000000021,
      -0.00000000038,    -0.00000000077,
       0.00000000045,    -0.00000000015,
       0.00000000195,    -0.00000000043,
      -0.00000000080,    -0.00000000358,
       0.00000000140,    -0.00000000028,
       0.00000000492,     0.00000000014,
       0.00000000021,    -0.00000000900,
       0.00000000335,    -0.00000000074,
       0.00000000081,     0.00000000364,
      -0.00000000084,     0.00000000028,
       0.00000000038,     0.00000000140,
       0.00000001092,     0.00000000023,
      -0.00000000024,     0.00000001091,
      -0.00000000155,     0.00000000066,
      -0.00000000066,    -0.00000000155,
       0.00000000019,    -0.00000000009,
       0.00000000001,     0.00000000004,
       0.00000000049,     0.00000000001,
      -0.00000000001,     0.00000000008,
      -0.00000000001,     0.00000000049,
      -0.00000000008,    -0.00000000001,
       0.00000000006,    -0.00000000005,
       0.00000000007,     0.00000000006,
      -0.00000000143,    -0.00000000078,
       0.00000000156,     0.00000000136,
      -0.00000000325,    -0.00000000058,
      -0.00000000068,     0.00000000005,
      -0.00000000056,    -0.00000000028,
      -0.00000000028,     0.00000000056,
       0.00000000054,     0.00000000028,
       0.00000000028,    -0.00000000054,
      -0.00000000195,    -0.00000000096,
       0.00000000024,    -0.00000000003,
       0.00000000986,    -0.00000002125,
       0.00000000015,     0.00000000258,
      -0.00000000017,     0.00000000034,
       0.00000000786,    -0.00000002491,
      -0.00000000015,     0.00000000169,
      -0.00000002376,    -0.00000000712,
       0.00000000166,     0.00000000017,
      -0.00000000169,     0.00000000441,
      -0.00000000018,    -0.00000000042,
       0.00000000455,     0.00000000182,
      -0.00000000045,     0.00000000019,
       0.00000000311,     0.00000000153,
      -0.00000000039,     0.00000000005,
      -0.00000000682,     0.00000001462,
      -0.00000000027,    -0.00000000140,
      -0.00000000752,     0.00000002346,
       0.00000000004,    -0.00000000157,
       0.00000002257,     0.00000000697,
      -0.00000000157,    -0.00000000004,
       0.00000000116,    -0.00000000301,
      -0.00000000309,    -0.00000000123,
      -0.00000000187,     0.00000000623,
       0.00000000005,    -0.00000000043,
       0.00000004229,     0.00000001382,
      -0.00000000286,     0.00000000002,
      -0.00000006409,    -0.00000001218,
       0.00000000445,     0.00000000099,
       0.00000000092,    -0.00000000010,
      -0.00000001203,     0.00000006389,
       0.00000000104,    -0.00000000440,
      -0.00000000011,    -0.00000000092,
       0.00000000908,     0.00000000160,
       0.00000000188,    -0.00000000949,
       0.00000000266,    -0.00000000884,
      -0.00000000007,     0.00000000061,
      -0.00000002919,    -0.00000000953,
       0.00000000259,     0.00000000037,
       0.00000005964,     0.00000001189,
      -0.00000000415,    -0.00000000057,
      -0.00000000086,     0.00000000008,
       0.00000001197,    -0.00000005975,
      -0.00000000058,     0.00000000415,
       0.00000000008,     0.00000000086,
      -0.00000000620,    -0.00000000112,
      -0.00000000125,     0.00000000641,
      -0.00000002429,    -0.00000000442,
       0.00000000168,     0.00000000045,
       0.00000000035,    -0.00000000004,
      -0.00000002172,     0.00000009509,
      -0.00000000003,    -0.00000000657,
      -0.00000000007,    -0.00000000141,
      -0.00000000615,    -0.00000019078,
      -0.00000001615,     0.00000000527,
       0.00000000033,     0.00000000070,
      -0.00000019062,     0.00000000622,
       0.00000000526,     0.00000001626,
       0.00000000071,    -0.00000000033,
      -0.00000000484,     0.00000002090,
      -0.00000000193,    -0.00000000027,
       0.00000002234,     0.00000000558,
      -0.00000000043,     0.00000000202,
       0.00000003065,     0.00000000557,
      -0.00000000213,    -0.00000000057,
      -0.00000000044,     0.00000000005,
       0.00000001486,    -0.00000006505,
       0.00000000003,     0.00000000448,
       0.00000000005,     0.00000000096,
       0.00000000568,     0.00000017374,
       0.00000001471,    -0.00000000478,
      -0.00000000029,    -0.00000000060,
       0.00000017434,    -0.00000000543,
      -0.00000000482,    -0.00000001473,
      -0.00000000065,     0.00000000030,
       0.00000000339,    -0.00000001428,
       0.00000000134,     0.00000000019,
      -0.00000001497,    -0.00000000374,
       0.00000000027,    -0.00000000139,
      -0.00000000427,    -0.00000012407,
      -0.00000001050,     0.00000000345,
       0.00000000021,     0.00000000046,
      -0.00000034789,     0.00000000672,
       0.00000000941,     0.00000003039,
       0.00000000132,    -0.00000000061,
       0.00000084190,    -0.00000000004,
       0.00000000006,    -0.00000084482,
      -0.00000000084,     0.00000000000,
       0.00000000570,     0.00000000048,
      -0.00000000025,    -0.00000000032,
      -0.00000000105,     0.00000000077,
       0.00000000472,     0.00000013702,
       0.00000001154,    -0.00000000382,
      -0.00000000024,    -0.00000000050,
       0.00000023306,    -0.00000000444,
      -0.00000000628,    -0.00000002018,
      -0.00000000089,     0.00000000041,
      -0.00000077808,     0.00000000058,
      -0.00000000054,     0.00000078224,
      -0.00000000390,     0.00000000005,
       0.00000000020,     0.00000000035,
       0.00000000070,     0.00000000108,
       0.00000073026,     0.00000000025,
       0.00000000002,    -0.00000000036,
       0.00000000010,    -0.00000181822,
      -0.00000000081,     0.00000000004,
      -0.00000000047,    -0.00000003946,
      -0.00000000025,     0.00000000076,
       0.00000000902,     0.00000000221,
      -0.00000000041,    -0.00000000439,
      -0.00000000172,     0.00000002761,
      -0.00000000066,    -0.00000000048,
      -0.00000000716,    -0.00000000658,
       0.00000000012,    -0.00000000245,
      -0.00000069128,    -0.00000000023,
      -0.00000000002,     0.00000000034,
      -0.00000000007,     0.00000118606,
       0.00000000057,    -0.00000000003,
       0.00000000077,     0.00000002705,
      -0.00000000045,    -0.00000000053,
       0.00000000010,    -0.00000000152,
       0.00000000013,     0.00000000300,
      -0.00000000118,    -0.00000001890,
      -0.00000000042,     0.00000000033,
      -0.00000000506,     0.00000000450,
       0.00000000010,     0.00000000168,
      -0.00000000330,    -0.00000000953,
       0.00000000170,    -0.00000000001,
      -0.00000002208,     0.00000000444,
       0.00000000009,    -0.00000000359,
       0.00000019834,     0.00000000373,
       0.00000000061,     0.00000000116,
       0.00000000362,     0.00000021300,
       0.00000000132,    -0.00000000056,
       0.00000000152,    -0.00000000021,
       0.00000000025,     0.00000000159,
       0.00000000260,     0.00000000752,
      -0.00000000134,     0.00000000001,
       0.00000001469,    -0.00000000289,
      -0.00000000001,     0.00000000250,
      -0.00000008840,    -0.00000000252,
      -0.00000000041,    -0.00000000079,
      -0.00000000245,    -0.00000009832,
      -0.00000000087,     0.00000000038,
      -0.00000000104,     0.00000000018,
      -0.00000000014,    -0.00000000093,
      -0.00000000551,     0.00000000279,
       0.00000000044,     0.00000000095,
       0.00000000009,    -0.00000000004,
       0.00000000747,     0.00000001486,
       0.00000000258,    -0.00000000120,
      -0.00000000142,    -0.00000000263,
       0.00000000234,    -0.00000000137,
       0.00000000261,     0.00000001130,
       0.00000000097,    -0.00000000018,
      -0.00000001150,     0.00000000278,
       0.00000000023,     0.00000000103,
       0.00000000348,    -0.00000000176,
      -0.00000000028,    -0.00000000060,
      -0.00000000511,    -0.00000001016,
      -0.00000000176,     0.00000000082,
       0.00000000060,     0.00000000116,
      -0.00000000095,     0.00000000056,
      -0.00000000177,    -0.00000000774,
      -0.00000000066,     0.00000000012,
       0.00000000794,    -0.00000000193,
      -0.00000000016,    -0.00000000071,
      -0.00000000044,     0.00000000012,
      -0.00000000040,     0.00000000019,
      -0.00000000029,    -0.00000000048,
      -0.00000000051,     0.00000000026,
      -0.00000000021,    -0.00000000047,
       0.00000000036,    -0.00000000051,
      -0.00000000051,    -0.00000000035,
       0.00000000064,    -0.00000000092,
      -0.00000000033,    -0.00000000027,
      -0.00000000060,    -0.00000000042,
       0.00000000030,    -0.00000000167,
      -0.00000000050,    -0.00000000013,
      -0.00000000165,    -0.00000000026,
      -0.00000000011,     0.00000000049,
       0.00000000051,    -0.00000000301,
      -0.00000000090,    -0.00000000022,
      -0.00000000203,    -0.00000000036,
      -0.00000000015,     0.00000000061,
      -0.00000000102,    -0.00000000394,
      -0.00000000092,     0.00000000016,
      -0.00000000382,     0.00000000110,
       0.00000000020,     0.00000000088,
       0.00000000012,     0.00000000043,
       0.00000000093,    -0.00000000025,
       0.00000000048,     0.00000000063,
       0.00000000053,    -0.00000000050,
      -0.00000000194,    -0.00000000709,
      -0.00000000165,     0.00000000033,
       0.00000000003,     0.00000000019,
      -0.00000000497,     0.00000000134,
       0.00000000023,     0.00000000115,
      -0.00000000510,    -0.00000000630,
      -0.00000000104,     0.00000000070,
      -0.00000000586,     0.00000000518,
       0.00000000073,     0.00000000091,
       0.00000000064,     0.00000000076,
       0.00000000012,    -0.00000000009,
       0.00000000159,    -0.00000000134,
       0.00000000102,     0.00000000062,
       0.00000000030,    -0.00000000094,
      -0.00000000943,    -0.00000001117,
      -0.00000000180,     0.00000000131,
       0.00000000010,     0.00000000014,
      -0.00000000826,     0.00000000695,
       0.00000000097,     0.00000000132,
      -0.00000001060,    -0.00000000546,
      -0.00000000055,     0.00000000080,
      -0.00000000430,     0.00000001030,
       0.00000000074,     0.00000000028,
       0.00000000150,     0.00000000070,
       0.00000000142,    -0.00000000303,
       0.00000000105,     0.00000000033,
      -0.00000000032,    -0.00000000050,
      -0.00000001925,    -0.00000000901,
      -0.00000000078,     0.00000000141,
      -0.00000000714,     0.00000001525,
       0.00000000112,     0.00000000061,
      -0.00000000939,    -0.00000000113,
       0.00000000096,     0.00000000765,
       0.00000000146,     0.00000000002,
       0.00000000002,    -0.00000000288,
       0.00000000052,     0.00000000047,
      -0.00000000024,     0.00000000100,
      -0.00000001609,    -0.00000000018,
      -0.00000000012,     0.00000001393,
       0.00000000039,    -0.00000000216,
      -0.00000000022,    -0.00000000452,
      -0.00000000037,     0.00000000018,
       0.00000000033,     0.00000000068,
       0.00000000093,    -0.00000000002,
      -0.00000000003,     0.00000000090,
       0.00000000336,    -0.00000000168,
      -0.00000000168,    -0.00000000343,
      -0.00000000223,     0.00000000021,
       0.00000000025,    -0.00000000194,
       0.00000000034,    -0.00000000019,
       0.00000000017,     0.00000000034,
      -0.00000000011,    -0.00000000051,
       0.00000000061,     0.00000000003,
       0.00000000004,    -0.00000000107,
       0.00000000053,    -0.00000000011,
       0.00000000012,     0.00000000058,
       0.00000000139,     0.00000000005,
      -0.00000000005,     0.00000000139,
      -0.00000000003,    -0.00000000003,
      -0.00000000003,    -0.00000000009,
       0.00000000009,    -0.00000000003,
      -0.00000000111,    -0.00000000053,
      -0.00000000053,     0.00000000111,
       0.00000000106,     0.00000000052,
       0.00000000052,    -0.00000000106,
       0.00000000137,     0.00000000064,
       0.00000000760,    -0.00000001555,
       0.00000000012,     0.00000000189,
      -0.00000000442,     0.00000001521,
       0.00000000026,    -0.00000000107,
       0.00000001516,     0.00000000438,
      -0.00000000106,    -0.00000000026,
       0.00000000092,    -0.00000000251,
      -0.00000000260,    -0.00000000100,
      -0.00000000204,    -0.00000000095,
      -0.00000000524,     0.00000001074,
       0.00000000001,    -0.00000000150,
       0.00000000424,    -0.00000001432,
      -0.00000000018,     0.00000000100,
      -0.00000001434,    -0.00000000426,
       0.00000000100,     0.00000000018,
      -0.00000000063,     0.00000000172,
       0.00000000176,     0.00000000067,
      -0.00000000098,     0.00000000348,
       0.00000000006,    -0.00000000025,
       0.00000001593,     0.00000000506,
      -0.00000000109,    -0.00000000012,
      -0.00000003197,    -0.00000000537,
       0.00000000220,     0.00000000089,
       0.00000000046,    -0.00000000007,
      -0.00000000533,     0.00000003192,
       0.00000000091,    -0.00000000218,
      -0.00000000007,    -0.00000000046,
       0.00000000413,     0.00000000073,
       0.00000000087,    -0.00000000434,
       0.00000000133,    -0.00000000470,
      -0.00000000009,     0.00000000033,
      -0.00000001101,    -0.00000000349,
       0.00000000101,     0.00000000024,
       0.00000002975,     0.00000000524,
      -0.00000000205,    -0.00000000067,
       0.00000000529,    -0.00000002983,
      -0.00000000068,     0.00000000206,
      -0.00000000282,    -0.00000000051,
      -0.00000000058,     0.00000000293,
      -0.00000001725,    -0.00000000277,
       0.00000000119,     0.00000000052,
       0.00000000025,    -0.00000000004,
      -0.00000001180,     0.00000005667,
       0.00000000068,    -0.00000000388,
      -0.00000000008,    -0.00000000084,
      -0.00000000413,    -0.00000012240,
      -0.00000001037,     0.00000000338,
       0.00000000020,     0.00000000041,
      -0.00000012257,     0.00000000409,
       0.00000000339,     0.00000001045,
       0.00000000042,    -0.00000000020,
      -0.00000000269,     0.00000001155,
      -0.00000000109,    -0.00000000014,
       0.00000001252,     0.00000000315,
      -0.00000000024,     0.00000000114,
       0.00000002099,     0.00000000337,
      -0.00000000145,    -0.00000000064,
      -0.00000000030,     0.00000000005,
       0.00000000807,    -0.00000003877,
      -0.00000000046,     0.00000000264,
       0.00000000005,     0.00000000057,
       0.00000000383,     0.00000011207,
       0.00000000949,    -0.00000000308,
      -0.00000000019,    -0.00000000038,
       0.00000011260,    -0.00000000362,
      -0.00000000311,    -0.00000000951,
      -0.00000000039,     0.00000000019,
       0.00000000189,    -0.00000000789,
       0.00000000076,     0.00000000010,
      -0.00000000834,    -0.00000000211,
       0.00000000015,    -0.00000000078,
      -0.00000000322,    -0.00000009131,
      -0.00000000770,     0.00000000254,
       0.00000000016,     0.00000000033,
      -0.00000023251,     0.00000000492,
       0.00000000630,     0.00000002020,
       0.00000000088,    -0.00000000041,
       0.00000052756,    -0.00000000008,
       0.00000000029,    -0.00000053130,
      -0.00000000065,     0.00000000003,
       0.00000000403,     0.00000000031,
      -0.00000000075,     0.00000000058,
       0.00000000349,     0.00000009879,
       0.00000000832,    -0.00000000275,
      -0.00000000017,    -0.00000000036,
       0.00000015594,    -0.00000000327,
      -0.00000000421,    -0.00000001349,
      -0.00000000059,     0.00000000027,
      -0.00000048964,     0.00000000035,
      -0.00000000050,     0.00000049329,
      -0.00000000276,     0.00000000007,
       0.00000000051,     0.00000000080,
       0.00000049057,     0.00000000028,
       0.00000000003,    -0.00000000031,
       0.00000000025,    -0.00000114451,
      -0.00000000064,    -0.00000000002,
      -0.00000000087,    -0.00000003268,
      -0.00000000027,     0.00000000064,
       0.00000000906,     0.00000000168,
      -0.00000000041,    -0.00000000361,
      -0.00000000116,     0.00000001806,
      -0.00000000045,    -0.00000000031,
      -0.00000000479,    -0.00000000431,
       0.00000000007,    -0.00000000161,
      -0.00000046439,    -0.00000000027,
      -0.00000000003,     0.00000000029,
      -0.00000000017,     0.00000074962,
       0.00000000097,     0.00000002198,
      -0.00000000037,    -0.00000000042,
      -0.00000000068,    -0.00000000130,
       0.00000000015,     0.00000000247,
      -0.00000000079,    -0.00000001236,
      -0.00000000028,     0.00000000021,
      -0.00000000341,     0.00000000295,
       0.00000000007,     0.00000000111,
      -0.00000000272,    -0.00000000826,
       0.00000000147,    -0.00000000001,
      -0.00000001844,     0.00000000375,
       0.00000000006,    -0.00000000302,
       0.00000014421,     0.00000000286,
       0.00000000046,     0.00000000086,
       0.00000000271,     0.00000015533,
       0.00000000101,    -0.00000000042,
       0.00000000128,    -0.00000000020,
       0.00000000021,     0.00000000128,
       0.00000000221,     0.00000000670,
      -0.00000000119,     0.00000000001,
       0.00000001229,    -0.00000000245,
      -0.00000000001,     0.00000000209,
      -0.00000006495,    -0.00000000201,
      -0.00000000032,    -0.00000000061,
      -0.00000000192,    -0.00000007279,
      -0.00000000069,     0.00000000030,
      -0.00000000087,     0.00000000017,
      -0.00000000012,    -0.00000000075,
      -0.00000000447,     0.00000000227,
       0.00000000035,     0.00000000076,
       0.00000000007,    -0.00000000003,
       0.00000000569,     0.00000001129,
       0.00000000195,    -0.00000000091,
      -0.00000000119,    -0.00000000235,
       0.00000000217,    -0.00000000118,
       0.00000000180,     0.00000000783,
       0.00000000067,    -0.00000000013,
      -0.00000000798,     0.00000000193,
       0.00000000016,     0.00000000071,
       0.00000000302,    -0.00000000154,
      -0.00000000024,    -0.00000000052,
      -0.00000000389,    -0.00000000772,
      -0.00000000133,     0.00000000062,
       0.00000000051,     0.00000000102,
      -0.00000000089,     0.00000000050,
      -0.00000000122,    -0.00000000536,
      -0.00000000046,     0.00000000008,
       0.00000000552,    -0.00000000135,
      -0.00000000011,    -0.00000000050,
      -0.00000000055,     0.00000000027,
      -0.00000000035,    -0.00000000061,
       0.00000000023,    -0.00000000033,
      -0.00000000032,    -0.00000000022,
       0.00000000042,    -0.00000000060,
      -0.00000000022,    -0.00000000017,
      -0.00000000038,    -0.00000000027,
       0.00000000018,    -0.00000000100,
      -0.00000000099,    -0.00000000015,
       0.00000000032,    -0.00000000186,
      -0.00000000056,    -0.00000000014,
      -0.00000000122,    -0.00000000021,
      -0.00000000056,    -0.00000000221,
      -0.00000000052,     0.00000000009,
      -0.00000000212,     0.00000000062,
       0.00000000011,     0.00000000049,
       0.00000000052,    -0.00000000014,
       0.00000000024,     0.00000000033,
       0.00000000025,    -0.00000000025,
      -0.00000000111,    -0.00000000406,
      -0.00000000094,     0.00000000019,
      -0.00000000279,     0.00000000075,
       0.00000000013,     0.00000000065,
      -0.00000000259,    -0.00000000325,
      -0.00000000054,     0.00000000035,
      -0.00000000295,     0.00000000264,
       0.00000000038,     0.00000000045,
       0.00000000035,     0.00000000041,
       0.00000000082,    -0.00000000069,
       0.00000000047,     0.00000000031,
       0.00000000010,    -0.00000000041,
      -0.00000000492,    -0.00000000584,
      -0.00000000094,     0.00000000068,
      -0.00000000423,     0.00000000355,
       0.00000000050,     0.00000000068,
      -0.00000000484,    -0.00000000259,
      -0.00000000027,     0.00000000037,
      -0.00000000186,     0.00000000464,
       0.00000000071,     0.00000000034,
       0.00000000065,    -0.00000000138,
       0.00000000045,     0.00000000020,
      -0.00000000897,    -0.00000000422,
      -0.00000000036,     0.00000000066,
      -0.00000000325,     0.00000000692,
       0.00000000051,     0.00000000028,
      -0.00000000372,    -0.00000000064,
       0.00000000055,     0.00000000272,
       0.00000000058,     0.00000000001,
       0.00000000001,    -0.00000000109,
       0.00000000027,     0.00000000024,
      -0.00000000013,     0.00000000050,
      -0.00000000633,    -0.00000000010,
      -0.00000000007,     0.00000000527,
       0.00000000018,    -0.00000000108,
      -0.00000000010,    -0.00000000226,
       0.00000000017,     0.00000000034,
       0.00000000179,    -0.00000000089,
      -0.00000000084,    -0.00000000172,
      -0.00000000097,     0.00000000013,
       0.00000000016,    -0.00000000078,
      -0.00000000034,    -0.00000000020,
      -0.00000000055,    -0.00000000017,
       0.00000000163,    -0.00000000132,
      -0.00000000110,     0.00000000091,
       0.00000000014,     0.00000000007,
       0.00000000008,     0.00000000004,
       0.00000000004,    -0.00000000007,
       0.00000000870,    -0.00000001373,
       0.00000000008,     0.00000000215,
      -0.00000000016,     0.00000000025,
      -0.00001109612,    -0.00000709634,
       0.00000173774,    -0.00000007188,
       0.00000020095,     0.00000012827,
       0.00000015951,     0.00000007283,
      -0.00000001846,    -0.00000000160,
      -0.00000000262,    -0.00000000094,
       0.00000007223,    -0.00000015897,
      -0.00000000163,     0.00000001825,
      -0.00000000091,     0.00000000259,
      -0.00000002945,    -0.00000001434,
       0.00000000319,    -0.00000000094,
       0.00000000040,     0.00000000016,
      -0.00000001540,     0.00000003037,
      -0.00000000102,    -0.00000000347,
       0.00000000018,    -0.00000000041,
      -0.00000001338,     0.00000002112,
      -0.00000000012,    -0.00000000331,
       0.00000000024,    -0.00000000038,
       0.00000759750,     0.00000485907,
      -0.00000119205,     0.00000004765,
      -0.00000013776,    -0.00000008780,
      -0.00000015176,    -0.00000007034,
       0.00000001769,     0.00000000073,
       0.00000000241,     0.00000000098,
      -0.00000007061,     0.00000015199,
       0.00000000075,    -0.00000001774,
       0.00000000103,    -0.00000000240,
       0.00000002010,     0.00000000977,
      -0.00000000220,     0.00000000066,
       0.00000001031,    -0.00000002058,
       0.00000000069,     0.00000000234,
       0.00000000053,     0.00000000023,
       0.00000000144,    -0.00000000299,
      -0.00000000147,     0.00000000537,
       0.00000000014,    -0.00000000039,
       0.00000000536,     0.00000000146,
      -0.00000000038,    -0.00000000015,
       0.00000000028,    -0.00000000082,
      -0.00000000085,    -0.00000000031,
      -0.00000000076,    -0.00000000033,
      -0.00000000100,     0.00000000208,
       0.00000000141,    -0.00000000505,
      -0.00000000506,    -0.00000000142,
      -0.00000000020,     0.00000000056,
       0.00000000057,     0.00000000021,
      -0.00000000063,     0.00000000236,
       0.00000000007,    -0.00000000017,
       0.00000000870,     0.00000000266,
      -0.00000000060,    -0.00000000014,
      -0.00000001968,    -0.00000000293,
       0.00000000134,     0.00000000075,
      -0.00000000292,     0.00000001968,
       0.00000000076,    -0.00000000134,
       0.00000000233,     0.00000000041,
       0.00000000050,    -0.00000000247,
       0.00000000082,    -0.00000000307,
      -0.00000000009,     0.00000000022,
      -0.00000000602,    -0.00000000184,
       0.00000000055,     0.00000000018,
       0.00000001832,     0.00000000286,
      -0.00000000125,    -0.00000000062,
       0.00000000290,    -0.00000001839,
      -0.00000000062,     0.00000000126,
      -0.00000000159,    -0.00000000029,
      -0.00000000033,     0.00000000166,
      -0.00000001275,    -0.00000000181,
       0.00000000087,     0.00000000052,
       0.00000000018,    -0.00000000004,
      -0.00000000700,     0.00000003700,
       0.00000000085,    -0.00000000251,
      -0.00000000007,    -0.00000000054,
      -0.00000000288,    -0.00000008257,
      -0.00000000697,     0.00000000227,
      -0.00000008288,     0.00000000280,
       0.00000000230,     0.00000000703,
      -0.00000000159,     0.00000000681,
      -0.00000000065,    -0.00000000008,
       0.00000000750,     0.00000000191,
      -0.00000000014,     0.00000000068,
       0.00000001508,     0.00000000215,
      -0.00000000103,    -0.00000000061,
      -0.00000000021,     0.00000000004,
       0.00000000479,    -0.00000002531,
      -0.00000000058,     0.00000000171,
       0.00000000268,     0.00000007593,
       0.00000000643,    -0.00000000209,
       0.00000007639,    -0.00000000252,
      -0.00000000212,    -0.00000000645,
       0.00000000113,    -0.00000000465,
       0.00000000046,     0.00000000005,
      -0.00000000496,    -0.00000000127,
       0.00000000008,    -0.00000000047,
      -0.00000000243,    -0.00000006756,
      -0.00000000571,     0.00000000188,
       0.00000000012,     0.00000000025,
      -0.00000015993,     0.00000000364,
       0.00000000434,     0.00000001389,
       0.00000000060,    -0.00000000028,
       0.00000033948,    -0.00000000006,
       0.00000000036,    -0.00000034327,
      -0.00000000051,     0.00000000005,
       0.00000000289,     0.00000000021,
      -0.00000000055,     0.00000000043,
       0.00000000258,     0.00000007195,
       0.00000000605,    -0.00000000200,
      -0.00000000012,    -0.00000000026,
       0.00000010738,    -0.00000000242,
      -0.00000000290,    -0.00000000928,
      -0.00000000037,     0.00000000018,
      -0.00000031620,     0.00000000019,
      -0.00000000044,     0.00000031929,
      -0.00000000198,     0.00000000007,
       0.00000000038,     0.00000000059,
       0.00000033240,     0.00000000029,
       0.00000000004,    -0.00000000026,
       0.00000000031,    -0.00000073623,
      -0.00000000051,    -0.00000000005,
      -0.00000000095,    -0.00000002628,
      -0.00000000025,     0.00000000051,
       0.00000000818,     0.00000000126,
      -0.00000000037,    -0.00000000290,
      -0.00000000079,     0.00000001197,
      -0.00000000031,    -0.00000000020,
      -0.00000000322,    -0.00000000287,
       0.00000000005,    -0.00000000108,
      -0.00000031465,    -0.00000000027,
      -0.00000000004,     0.00000000024,
      -0.00000000021,     0.00000048385,
       0.00000000096,     0.00000001744,
      -0.00000000030,    -0.00000000032,
      -0.00000000098,    -0.00000000105,
       0.00000000015,     0.00000000198,
      -0.00000000053,    -0.00000000820,
      -0.00000000233,     0.00000000196,
       0.00000000005,     0.00000000074,
      -0.00000000217,    -0.00000000687,
       0.00000000122,    -0.00000000001,
      -0.00000001482,     0.00000000303,
       0.00000000004,    -0.00000000246,
       0.00000010350,     0.00000000219,
       0.00000000035,     0.00000000063,
       0.00000000203,     0.00000011191,
       0.00000000076,    -0.00000000031,
       0.00000000104,    -0.00000000018,
       0.00000000017,     0.00000000101,
       0.00000000180,     0.00000000569,
      -0.00000000101,     0.00000000001,
       0.00000000990,    -0.00000000199,
       0.00000000000,     0.00000000170,
      -0.00000004693,    -0.00000000158,
      -0.00000000025,    -0.00000000047,
      -0.00000000149,    -0.00000005306,
      -0.00000000054,     0.00000000023,
      -0.00000000071,     0.00000000015,
      -0.00000000009,    -0.00000000059,
      -0.00000000357,     0.00000000182,
       0.00000000028,     0.00000000060,
       0.00000000430,     0.00000000854,
       0.00000000146,    -0.00000000068,
      -0.00000000096,    -0.00000000204,
       0.00000000193,    -0.00000000098,
       0.00000000125,     0.00000000543,
       0.00000000047,    -0.00000000009,
      -0.00000000554,     0.00000000134,
       0.00000000011,     0.00000000050,
       0.00000000253,    -0.00000000129,
      -0.00000000020,    -0.00000000043,
      -0.00000000294,    -0.00000000584,
      -0.00000000100,     0.00000000046,
       0.00000000042,     0.00000000089,
      -0.00000000080,     0.00000000042,
      -0.00000000084,    -0.00000000372,
       0.00000000384,    -0.00000000094,
      -0.00000000057,     0.00000000028,
      -0.00000000035,    -0.00000000063,
       0.00000000011,    -0.00000000060,
      -0.00000000059,    -0.00000000009,
       0.00000000019,    -0.00000000113,
      -0.00000000034,    -0.00000000008,
      -0.00000000073,    -0.00000000013,
      -0.00000000030,    -0.00000000123,
      -0.00000000116,     0.00000000035,
      -0.00000000062,    -0.00000000229,
      -0.00000000053,     0.00000000011,
      -0.00000000155,     0.00000000042,
      -0.00000000130,    -0.00000000167,
      -0.00000000147,     0.00000000134,
       0.00000000018,     0.00000000022,
       0.00000000041,    -0.00000000035,
      -0.00000000253,    -0.00000000302,
      -0.00000000048,     0.00000000035,
      -0.00000000215,     0.00000000180,
       0.00000000025,     0.00000000034,
      -0.00000000220,    -0.00000000123,
      -0.00000000078,     0.00000000208,
       0.00000000033,     0.00000000016,
       0.00000000029,    -0.00000000062,
      -0.00000000414,    -0.00000000196,
      -0.00000000017,     0.00000000030,
      -0.00000000148,     0.00000000312,
      -0.00000000147,    -0.00000000035,
       0.00000000030,     0.00000000092,
      -0.00000000244,    -0.00000000005,
      -0.00000000004,     0.00000000197,
       0.00000000008,    -0.00000000052,
      -0.00000000004,    -0.00000000110,
       0.00000000091,    -0.00000000045,
      -0.00000000041,    -0.00000000084,
       0.00000000041,     0.00000000034,
      -0.00000000073,    -0.00000000046,
       0.00000000101,     0.00000000043,
       0.00000000043,    -0.00000000101,
       0.00000000032,     0.00000000020,
      -0.00000000096,    -0.00000000042,
      -0.00000000042,     0.00000000096,
       0.00000000033,     0.00000000014,
       0.00000000064,    -0.00000000138,
      -0.00000000078,     0.00000000304,
       0.00000000304,     0.00000000078,
       0.00000000014,    -0.00000000043,
      -0.00000000045,    -0.00000000016,
      -0.00000000045,    -0.00000000019,
      -0.00000000044,     0.00000000096,
       0.00000000075,    -0.00000000286,
      -0.00000000286,    -0.00000000076,
      -0.00000000043,     0.00000000172,
       0.00000000006,    -0.00000000013,
       0.00000000550,     0.00000000161,
      -0.00000000038,    -0.00000000013,
      -0.00000001320,    -0.00000000175,
       0.00000000089,     0.00000000062,
      -0.00000000175,     0.00000001322,
       0.00000000063,    -0.00000000089,
       0.00000000144,     0.00000000026,
       0.00000000031,    -0.00000000154,
       0.00000000054,    -0.00000000217,
      -0.00000000380,    -0.00000000111,
       0.00000000034,     0.00000000014,
       0.00000001229,     0.00000000171,
      -0.00000000083,    -0.00000000053,
       0.00000000174,    -0.00000001235,
      -0.00000000054,     0.00000000084,
      -0.00000000098,    -0.00000000018,
      -0.00000000021,     0.00000000103,
      -0.00000000960,    -0.00000000122,
       0.00000000065,     0.00000000047,
       0.00000000014,    -0.00000000003,
      -0.00000000436,     0.00000002539,
       0.00000000084,    -0.00000000171,
      -0.00000000205,    -0.00000005723,
      -0.00000000484,     0.00000000157,
      -0.00000005759,     0.00000000197,
       0.00000000160,     0.00000000489,
      -0.00000000099,     0.00000000418,
       0.00000000468,     0.00000000120,
      -0.00000000009,     0.00000000043,
       0.00000001111,     0.00000000141,
      -0.00000000075,    -0.00000000055,
      -0.00000000016,     0.00000000004,
       0.00000000298,    -0.00000001736,
      -0.00000000057,     0.00000000116,
       0.00000000191,     0.00000005282,
       0.00000000448,    -0.00000000145,
       0.00000005322,    -0.00000000178,
      -0.00000000147,    -0.00000000450,
       0.00000000070,    -0.00000000285,
      -0.00000000307,    -0.00000000080,
      -0.00000000182,    -0.00000005008,
      -0.00000000424,     0.00000000139,
       0.00000000009,     0.00000000018,
      -0.00000011197,     0.00000000270,
       0.00000000304,     0.00000000967,
       0.00000000038,    -0.00000000019,
       0.00000022244,    -0.00000000003,
       0.00000000035,    -0.00000022594,
       0.00000000209,     0.00000000014,
      -0.00000000040,     0.00000000032,
       0.00000000192,     0.00000005267,
       0.00000000443,    -0.00000000146,
      -0.00000000009,    -0.00000000019,
       0.00000007525,    -0.00000000180,
      -0.00000000204,    -0.00000000651,
      -0.00000020783,     0.00000000009,
      -0.00000000037,     0.00000021039,
      -0.00000000143,     0.00000000006,
       0.00000000028,     0.00000000044,
       0.00000022672,     0.00000000027,
       0.00000000004,    -0.00000000021,
       0.00000000033,    -0.00000048126,
      -0.00000000089,    -0.00000002073,
      -0.00000000021,     0.00000000040,
       0.00000000698,     0.00000000094,
      -0.00000000032,    -0.00000000229,
      -0.00000000055,     0.00000000802,
      -0.00000000218,    -0.00000000193,
       0.00000000003,    -0.00000000073,
      -0.00000021462,    -0.00000000026,
      -0.00000000022,     0.00000031720,
       0.00000000085,     0.00000001362,
      -0.00000000024,    -0.00000000024,
      -0.00000000103,    -0.00000000083,
       0.00000000014,     0.00000000156,
      -0.00000000036,    -0.00000000549,
      -0.00000000160,     0.00000000132,
       0.00000000003,     0.00000000050,
      -0.00000000170,    -0.00000000556,
       0.00000000100,    -0.00000000001,
      -0.00000001163,     0.00000000239,
       0.00000000002,    -0.00000000195,
       0.00000007373,     0.00000000167,
       0.00000000027,     0.00000000046,
       0.00000000151,     0.00000008007,
       0.00000000058,    -0.00000000023,
       0.00000000083,    -0.00000000015,
       0.00000000013,     0.00000000079,
       0.00000000143,     0.00000000468,
      -0.00000000084,     0.00000000001,
       0.00000000778,    -0.00000000158,
       0.00000000001,     0.00000000135,
      -0.00000003358,    -0.00000000123,
      -0.00000000019,    -0.00000000035,
      -0.00000000115,    -0.00000003832,
      -0.00000000041,     0.00000000017,
      -0.00000000056,     0.00000000012,
      -0.00000000007,    -0.00000000045,
      -0.00000000281,     0.00000000143,
       0.00000000022,     0.00000000047,
       0.00000000324,     0.00000000642,
       0.00000000109,    -0.00000000051,
      -0.00000000076,    -0.00000000171,
       0.00000000166,    -0.00000000079,
       0.00000000086,     0.00000000377,
      -0.00000000386,     0.00000000093,
       0.00000000207,    -0.00000000105,
      -0.00000000016,    -0.00000000035,
      -0.00000000222,    -0.00000000439,
      -0.00000000074,     0.00000000035,
       0.00000000033,     0.00000000075,
      -0.00000000070,     0.00000000035,
      -0.00000000058,    -0.00000000259,
       0.00000000268,    -0.00000000066,
      -0.00000000054,     0.00000000027,
      -0.00000000032,    -0.00000000058,
       0.00000000012,    -0.00000000068,
      -0.00000000043,    -0.00000000008,
      -0.00000000016,    -0.00000000068,
      -0.00000000063,     0.00000000019,
      -0.00000000035,    -0.00000000127,
      -0.00000000085,     0.00000000023,
      -0.00000000065,    -0.00000000085,
      -0.00000000072,     0.00000000067,
      -0.00000000129,    -0.00000000154,
      -0.00000000025,     0.00000000018,
      -0.00000000108,     0.00000000090,
      -0.00000000099,    -0.00000000059,
      -0.00000000032,     0.00000000092,
      -0.00000000188,    -0.00000000090,
      -0.00000000067,     0.00000000140,
      -0.00000000057,    -0.00000000019,
      -0.00000000091,    -0.00000000003,
      -0.00000000002,     0.00000000072,
      -0.00000000002,    -0.00000000052,
      -0.00000000019,    -0.00000000040,
       0.00000000022,    -0.00000000028,
       0.00000000048,     0.00000000019,
       0.00000000019,    -0.00000000047,
      -0.00000000045,    -0.00000000019,
      -0.00000000019,     0.00000000045,
       0.00000000036,    -0.00000000081,
      -0.00000000048,     0.00000000199,
       0.00000000199,     0.00000000048,
      -0.00000000025,     0.00000000057,
       0.00000000046,    -0.00000000186,
      -0.00000000187,    -0.00000000047,
      -0.00000000031,     0.00000000130,
       0.00000000374,     0.00000000104,
      -0.00000000924,    -0.00000000109,
       0.00000000062,     0.00000000051,
      -0.00000000110,     0.00000000927,
       0.00000000052,    -0.00000000062,
       0.00000000093,     0.00000000017,
       0.00000000021,    -0.00000000101,
       0.00000000038,    -0.00000000160,
      -0.00000000259,    -0.00000000072,
       0.00000000861,     0.00000000107,
      -0.00000000058,    -0.00000000045,
       0.00000000109,    -0.00000000866,
      -0.00000000045,     0.00000000059,
      -0.00000000064,    -0.00000000012,
      -0.00000000014,     0.00000000067,
      -0.00000000730,    -0.00000000083,
       0.00000000049,     0.00000000042,
       0.00000000010,    -0.00000000003,
      -0.00000000280,     0.00000001794,
       0.00000000075,    -0.00000000120,
      -0.00000000148,    -0.00000004032,
      -0.00000000341,     0.00000000110,
      -0.00000004067,     0.00000000140,
       0.00000000113,     0.00000000346,
      -0.00000000063,     0.00000000263,
       0.00000000300,     0.00000000078,
       0.00000000829,     0.00000000094,
      -0.00000000056,    -0.00000000047,
      -0.00000000012,     0.00000000003,
       0.00000000192,    -0.00000001227,
      -0.00000000052,     0.00000000082,
       0.00000000137,     0.00000003732,
       0.00000000317,    -0.00000000102,
       0.00000003765,    -0.00000000128,
      -0.00000000104,    -0.00000000319,
       0.00000000045,    -0.00000000179,
      -0.00000000195,    -0.00000000052,
      -0.00000000136,    -0.00000003713,
      -0.00000000313,     0.00000000103,
       0.00000000006,     0.00000000014,
      -0.00000007928,     0.00000000200,
       0.00000000215,     0.00000000685,
       0.00000014765,    -0.00000000001,
       0.00000000031,    -0.00000015072,
       0.00000000152,     0.00000000010,
      -0.00000000029,     0.00000000024,
       0.00000000142,     0.00000003867,
       0.00000000326,    -0.00000000107,
      -0.00000000007,    -0.00000000014,
       0.00000005334,    -0.00000000134,
      -0.00000000144,    -0.00000000461,
      -0.00000013834,     0.00000000003,
      -0.00000000031,     0.00000014044,
      -0.00000000104,     0.00000000005,
       0.00000000021,     0.00000000033,
       0.00000015546,     0.00000000024,
       0.00000000004,    -0.00000000017,
       0.00000000031,    -0.00000031849,
      -0.00000000077,    -0.00000001613,
       0.00000000576,     0.00000000069,
      -0.00000000026,    -0.00000000179,
      -0.00000000038,     0.00000000541,
      -0.00000000149,    -0.00000000131,
       0.00000000002,    -0.00000000049,
      -0.00000014716,    -0.00000000023,
      -0.00000000021,     0.00000021044,
       0.00000000071,     0.00000001051,
      -0.00000000097,    -0.00000000065,
       0.00000000012,     0.00000000122,
      -0.00000000024,    -0.00000000371,
      -0.00000000110,     0.00000000089,
      -0.00000000132,    -0.00000000441,
       0.00000000080,    -0.00000000001,
      -0.00000000898,     0.00000000184,
       0.00000000000,    -0.00000000153,
       0.00000005230,     0.00000000128,
       0.00000000020,     0.00000000033,
       0.00000000112,     0.00000005707,
       0.00000000044,    -0.00000000017,
       0.00000000065,    -0.00000000012,
       0.00000000010,     0.00000000061,
       0.00000000112,     0.00000000376,
      -0.00000000068,     0.00000000001,
       0.00000000602,    -0.00000000122,
       0.00000000001,     0.00000000105,
      -0.00000002389,    -0.00000000095,
      -0.00000000087,    -0.00000002753,
      -0.00000000044,     0.00000000010,
      -0.00000000219,     0.00000000112,
       0.00000000017,     0.00000000036,
       0.00000000243,     0.00000000482,
       0.00000000081,    -0.00000000038,
      -0.00000000059,    -0.00000000141,
       0.00000000138,    -0.00000000063,
       0.00000000060,     0.00000000263,
      -0.00000000269,     0.00000000065,
       0.00000000166,    -0.00000000085,
      -0.00000000013,    -0.00000000027,
      -0.00000000166,    -0.00000000330,
      -0.00000000055,     0.00000000026,
       0.00000000026,     0.00000000062,
      -0.00000000060,     0.00000000028,
      -0.00000000040,    -0.00000000180,
       0.00000000187,    -0.00000000046,
      -0.00000000047,     0.00000000023,
      -0.00000000028,    -0.00000000051,
       0.00000000521,    -0.00000000946,
      -0.00000001389,    -0.00000000774,
      -0.00000000019,    -0.00000000069,
      -0.00000000046,     0.00000000012,
      -0.00000000032,    -0.00000000043,
      -0.00000000035,     0.00000000033,
      -0.00000000064,    -0.00000000077,
      -0.00000000054,     0.00000000045,
      -0.00000000044,    -0.00000000028,
      -0.00000000013,     0.00000000040,
      -0.00000000085,    -0.00000000041,
      -0.00000000030,     0.00000000062,
       0.00000000023,    -0.00000000054,
      -0.00000000032,     0.00000000139,
       0.00000000139,     0.00000000032,
      -0.00000000016,     0.00000000037,
       0.00000000030,    -0.00000000130,
      -0.00000000131,    -0.00000000031,
      -0.00000000022,     0.00000000100,
       0.00000000265,     0.00000000070,
      -0.00000000663,    -0.00000000070,
       0.00000000044,     0.00000000041,
      -0.00000000071,     0.00000000667,
       0.00000000042,    -0.00000000045,
       0.00000000062,     0.00000000011,
       0.00000000014,    -0.00000000068,
       0.00000000027,    -0.00000000121,
      -0.00000000183,    -0.00000000049,
       0.00000000619,     0.00000000069,
      -0.00000000042,    -0.00000000037,
       0.00000000070,    -0.00000000623,
      -0.00000000037,     0.00000000042,
      -0.00000000043,    -0.00000000008,
      -0.00000000009,     0.00000000045,
      -0.00000000558,    -0.00000000056,
       0.00000000038,     0.00000000036,
       0.00000000008,    -0.00000000002,
      -0.00000000184,     0.00000001292,
       0.00000000065,    -0.00000000086,
      -0.00000000107,    -0.00000002869,
      -0.00000000243,     0.00000000078,
      -0.00000002902,     0.00000000100,
       0.00000000081,     0.00000000247,
      -0.00000000040,     0.00000000168,
       0.00000000196,     0.00000000051,
       0.00000000624,     0.00000000063,
      -0.00000000042,    -0.00000000040,
       0.00000000126,    -0.00000000884,
      -0.00000000044,     0.00000000058,
       0.00000000100,     0.00000002663,
       0.00000000226,    -0.00000000073,
       0.00000002691,    -0.00000000092,
      -0.00000000075,    -0.00000000228,
       0.00000000029,    -0.00000000115,
      -0.00000000127,    -0.00000000034,
      -0.00000000102,    -0.00000002752,
      -0.00000000232,     0.00000000076,
       0.00000000005,     0.00000000010,
      -0.00000005657,     0.00000000148,
       0.00000000153,     0.00000000489,
       0.00000009896,     0.00000000001,
       0.00000000027,    -0.00000010157,
       0.00000000110,     0.00000000007,
       0.00000000105,     0.00000002842,
       0.00000000240,    -0.00000000079,
      -0.00000000005,    -0.00000000010,
       0.00000003809,    -0.00000000099,
      -0.00000000103,    -0.00000000330,
      -0.00000009297,     0.00000000000,
      -0.00000000025,     0.00000009466,
      -0.00000000075,     0.00000000004,
       0.00000010705,     0.00000000021,
       0.00000000028,    -0.00000021283,
      -0.00000000064,    -0.00000001242,
       0.00000000464,     0.00000000051,
      -0.00000000021,    -0.00000000138,
      -0.00000000026,     0.00000000368,
      -0.00000000101,    -0.00000000089,
      -0.00000010134,    -0.00000000020,
      -0.00000000019,     0.00000014092,
       0.00000000058,     0.00000000804,
      -0.00000000085,    -0.00000000050,
       0.00000000010,     0.00000000094,
      -0.00000000016,    -0.00000000252,
      -0.00000000077,     0.00000000061,
      -0.00000000100,    -0.00000000345,
       0.00000000063,    -0.00000000001,
      -0.00000000685,     0.00000000140,
      -0.00000000001,    -0.00000000119,
       0.00000003700,     0.00000000097,
       0.00000000082,     0.00000004058,
       0.00000000050,    -0.00000000010,
       0.00000000008,     0.00000000046,
       0.00000000086,     0.00000000297,
      -0.00000000054,     0.00000000001,
       0.00000000460,    -0.00000000093,
       0.00000000001,     0.00000000082,
      -0.00000001693,    -0.00000000073,
      -0.00000000066,    -0.00000001971,
      -0.00000000169,     0.00000000086,
       0.00000000013,     0.00000000028,
       0.00000000182,     0.00000000360,
       0.00000000060,    -0.00000000028,
      -0.00000000045,    -0.00000000113,
       0.00000000113,    -0.00000000049,
       0.00000000042,     0.00000000183,
      -0.00000000188,     0.00000000045,
       0.00000000131,    -0.00000000067,
      -0.00000000010,    -0.00000000021,
      -0.00000000125,    -0.00000000247,
      -0.00000000041,     0.00000000019,
       0.00000000020,     0.00000000050,
      -0.00000000050,     0.00000000022,
      -0.00000000028,    -0.00000000126,
       0.00000000131,    -0.00000000032,
      -0.00000000040,     0.00000000020,
      -0.00000000023,    -0.00000000043,
       0.00000000016,    -0.00000000038,
      -0.00000000022,     0.00000000100,
       0.00000000101,     0.00000000022,
       0.00000000021,    -0.00000000094,
      -0.00000000095,    -0.00000000021,
      -0.00000000016,     0.00000000078,
       0.00000000193,     0.00000000049,
      -0.00000000484,    -0.00000000046,
       0.00000000032,     0.00000000033,
      -0.00000000047,     0.00000000487,
       0.00000000034,    -0.00000000033,
       0.00000000042,     0.00000000008,
       0.00000000010,    -0.00000000046,
       0.00000000019,    -0.00000000092,
      -0.00000000133,    -0.00000000034,
       0.00000000452,     0.00000000045,
      -0.00000000030,    -0.00000000030,
       0.00000000046,    -0.00000000455,
      -0.00000000030,     0.00000000030,
      -0.00000000427,    -0.00000000039,
       0.00000000029,     0.00000000030,
      -0.00000000122,     0.00000000943,
       0.00000000054,    -0.00000000062,
      -0.00000000078,    -0.00000002055,
      -0.00000000175,     0.00000000056,
      -0.00000002084,     0.00000000072,
       0.00000000058,     0.00000000178,
      -0.00000000027,     0.00000000109,
       0.00000000130,     0.00000000034,
       0.00000000471,     0.00000000043,
      -0.00000000032,    -0.00000000033,
       0.00000000083,    -0.00000000645,
      -0.00000000037,     0.00000000042,
       0.00000000072,     0.00000001912,
       0.00000000163,    -0.00000000052,
       0.00000001935,    -0.00000000066,
      -0.00000000054,    -0.00000000164,
       0.00000000019,    -0.00000000074,
      -0.00000000083,    -0.00000000023,
      -0.00000000076,    -0.00000002038,
      -0.00000000173,     0.00000000056,
       0.00000000003,     0.00000000007,
      -0.00000004058,     0.00000000110,
       0.00000000110,     0.00000000351,
       0.00000006683,     0.00000000002,
       0.00000000022,    -0.00000006899,
       0.00000000080,     0.00000000005,
       0.00000000077,     0.00000002090,
       0.00000000177,    -0.00000000058,
       0.00000002735,    -0.00000000074,
      -0.00000000074,    -0.00000000237,
      -0.00000006294,    -0.00000000002,
      -0.00000000020,     0.00000006430,
      -0.00000000055,     0.00000000003,
       0.00000007398,     0.00000000018,
       0.00000000024,    -0.00000014333,
      -0.00000000051,    -0.00000000949,
       0.00000000367,     0.00000000037,
      -0.00000000017,    -0.00000000106,
      -0.00000000018,     0.00000000251,
      -0.00000000069,    -0.00000000061,
      -0.00000007003,    -0.00000000017,
      -0.00000000017,     0.00000009509,
       0.00000000046,     0.00000000611,
      -0.00000000071,    -0.00000000037,
       0.00000000008,     0.00000000072,
      -0.00000000011,    -0.00000000172,
      -0.00000000053,     0.00000000042,
      -0.00000000076,    -0.00000000267,
       0.00000000049,    -0.00000000001,
      -0.00000000518,     0.00000000106,
      -0.00000000001,    -0.00000000091,
       0.00000002613,     0.00000000074,
       0.00000000061,     0.00000002881,
       0.00000000066,     0.00000000232,
      -0.00000000043,     0.00000000001,
       0.00000000348,    -0.00000000070,
       0.00000000001,     0.00000000063,
      -0.00000001197,    -0.00000000056,
      -0.00000000050,    -0.00000001408,
      -0.00000000130,     0.00000000066,
       0.00000000010,     0.00000000021,
       0.00000000136,     0.00000000269,
       0.00000000044,    -0.00000000021,
      -0.00000000034,    -0.00000000090,
       0.00000000090,    -0.00000000038,
       0.00000000029,     0.00000000128,
      -0.00000000131,     0.00000000032,
       0.00000000102,    -0.00000000052,
      -0.00000000093,    -0.00000000184,
       0.00000000015,     0.00000000040,
      -0.00000000040,     0.00000000018,
      -0.00000000019,    -0.00000000088,
       0.00000000092,    -0.00000000023,
      -0.00000000019,    -0.00000000035,
      -0.00000000015,     0.00000000074,
       0.00000000074,     0.00000000015,
       0.00000000014,    -0.00000000069,
      -0.00000000070,    -0.00000000015,
      -0.00000000012,     0.00000000061,
       0.00000000143,     0.00000000034,
      -0.00000000356,    -0.00000000030,
       0.00000000024,     0.00000000026,
      -0.00000000031,     0.00000000359,
       0.00000000027,    -0.00000000024,
       0.00000000014,    -0.00000000071,
      -0.00000000099,    -0.00000000024,
       0.00000000333,     0.00000000029,
       0.00000000031,    -0.00000000336,
      -0.00000000327,    -0.00000000027,
       0.00000000022,     0.00000000025,
      -0.00000000082,     0.00000000693,
       0.00000000044,    -0.00000000046,
      -0.00000000057,    -0.00000001479,
      -0.00000000126,     0.00000000040,
      -0.00000001504,     0.00000000052,
       0.00000000042,     0.00000000129,
      -0.00000000018,     0.00000000071,
       0.00000000087,     0.00000000023,
       0.00000000357,     0.00000000029,
      -0.00000000024,    -0.00000000027,
       0.00000000056,    -0.00000000474,
      -0.00000000030,     0.00000000031,
       0.00000000053,     0.00000001379,
       0.00000000118,    -0.00000000037,
       0.00000001398,    -0.00000000048,
      -0.00000000039,    -0.00000000119,
       0.00000000013,    -0.00000000049,
      -0.00000000055,    -0.00000000015,
      -0.00000000056,    -0.00000001508,
      -0.00000000128,     0.00000000042,
      -0.00000002922,     0.00000000081,
       0.00000000079,     0.00000000253,
       0.00000004539,     0.00000000003,
       0.00000000018,    -0.00000004716,
       0.00000000059,     0.00000000003,
       0.00000000057,     0.00000001538,
       0.00000000130,    -0.00000000042,
       0.00000001970,    -0.00000000054,
      -0.00000000053,    -0.00000000171,
      -0.00000004286,    -0.00000000003,
      -0.00000000016,     0.00000004394,
       0.00000005128,     0.00000000015,
       0.00000000021,    -0.00000009714,
      -0.00000000040,    -0.00000000720,
       0.00000000287,     0.00000000027,
      -0.00000000013,    -0.00000000081,
      -0.00000000013,     0.00000000172,
      -0.00000000048,    -0.00000000042,
      -0.00000004854,    -0.00000000014,
      -0.00000000014,     0.00000006455,
       0.00000000036,     0.00000000461,
      -0.00000000059,    -0.00000000028,
       0.00000000006,     0.00000000055,
      -0.00000000008,    -0.00000000118,
      -0.00000000037,     0.00000000029,
      -0.00000000057,    -0.00000000205,
       0.00000000038,    -0.00000000001,
      -0.00000000390,     0.00000000079,
      -0.00000000001,    -0.00000000070,
       0.00000001844,     0.00000000056,
       0.00000000045,     0.00000002044,
       0.00000000050,     0.00000000179,
      -0.00000000033,     0.00000000001,
       0.00000000262,    -0.00000000053,
       0.00000000001,     0.00000000048,
      -0.00000000845,    -0.00000000043,
      -0.00000000037,    -0.00000001005,
      -0.00000000099,     0.00000000051,
       0.00000000007,     0.00000000016,
       0.00000000101,     0.00000000200,
      -0.00000000026,    -0.00000000070,
       0.00000000072,    -0.00000000029,
       0.00000000020,     0.00000000089,
      -0.00000000092,     0.00000000022,
       0.00000000079,    -0.00000000041,
      -0.00000000069,    -0.00000000137,
      -0.00000000013,    -0.00000000061,
       0.00000000064,    -0.00000000016,
      -0.00000000011,     0.00000000055,
       0.00000000056,     0.00000000011,
       0.00000000010,    -0.00000000052,
      -0.00000000052,    -0.00000000010,
      -0.00000000009,     0.00000000048,
       0.00000000107,     0.00000000024,
      -0.00000000264,    -0.00000000020,
      -0.00000000021,     0.00000000267,
       0.00000000010,    -0.00000000055,
      -0.00000000074,    -0.00000000017,
       0.00000000247,     0.00000000020,
       0.00000000021,    -0.00000000249,
      -0.00000000250,    -0.00000000018,
       0.00000000017,     0.00000000020,
      -0.00000000055,     0.00000000513,
       0.00000000036,    -0.00000000034,
      -0.00000000041,    -0.00000001067,
      -0.00000000091,     0.00000000029,
      -0.00000001089,     0.00000000037,
       0.00000000030,     0.00000000093,
      -0.00000000012,     0.00000000047,
       0.00000000059,     0.00000000016,
       0.00000000271,     0.00000000020,
      -0.00000000018,    -0.00000000022,
       0.00000000038,    -0.00000000351,
       0.00000000038,     0.00000000997,
       0.00000000085,    -0.00000000027,
       0.00000001012,    -0.00000000035,
      -0.00000000028,    -0.00000000086,
      -0.00000000041,    -0.00000001115,
      -0.00000000095,     0.00000000031,
      -0.00000002110,     0.00000000060,
       0.00000000057,     0.00000000183,
       0.00000003098,     0.00000000003,
       0.00000000014,    -0.00000003241,
       0.00000000042,     0.00000001131,
       0.00000000096,    -0.00000000031,
       0.00000001424,    -0.00000000040,
      -0.00000000038,    -0.00000000124,
      -0.00000002932,    -0.00000000003,
      -0.00000000013,     0.00000003017,
       0.00000003563,     0.00000000013,
       0.00000000017,    -0.00000006619,
      -0.00000000031,    -0.00000000544,
       0.00000000223,     0.00000000019,
      -0.00000000010,    -0.00000000061,
      -0.00000000009,     0.00000000118,
      -0.00000000033,    -0.00000000029,
      -0.00000003373,    -0.00000000012,
      -0.00000000012,     0.00000004405,
       0.00000000027,     0.00000000347,
      -0.00000000047,    -0.00000000021,
      -0.00000000005,    -0.00000000081,
      -0.00000000043,    -0.00000000156,
       0.00000000029,    -0.00000000001,
      -0.00000000291,     0.00000000059,
      -0.00000000001,    -0.00000000053,
       0.00000001301,     0.00000000042,
       0.00000000033,     0.00000001450,
       0.00000000037,     0.00000000137,
      -0.00000000026,     0.00000000001,
       0.00000000196,    -0.00000000039,
      -0.00000000595,    -0.00000000033,
      -0.00000000028,    -0.00000000716,
      -0.00000000075,     0.00000000039,
       0.00000000075,     0.00000000149,
      -0.00000000019,    -0.00000000055,
       0.00000000056,    -0.00000000022,
       0.00000000014,     0.00000000063,
      -0.00000000064,     0.00000000016,
       0.00000000061,    -0.00000000031,
      -0.00000000051,    -0.00000000102,
      -0.00000000009,    -0.00000000043,
       0.00000000045,    -0.00000000011,
      -0.00000000007,     0.00000000037,
       0.00000000081,     0.00000000018,
      -0.00000000196,    -0.00000000013,
      -0.00000000014,     0.00000000199,
      -0.00000000056,    -0.00000000012,
       0.00000000184,     0.00000000013,
       0.00000000014,    -0.00000000186,
      -0.00000000191,    -0.00000000012,
       0.00000000013,     0.00000000016,
      -0.00000000037,     0.00000000381,
       0.00000000029,    -0.00000000025,
      -0.00000000030,    -0.00000000772,
      -0.00000000066,     0.00000000021,
      -0.00000000790,     0.00000000027,
       0.00000000022,     0.00000000068,
       0.00000000040,     0.00000000011,
       0.00000000205,     0.00000000013,
      -0.00000000014,    -0.00000000017,
       0.00000000025,    -0.00000000260,
       0.00000000028,     0.00000000723,
       0.00000000062,    -0.00000000019,
       0.00000000735,    -0.00000000025,
      -0.00000000020,    -0.00000000063,
      -0.00000000031,    -0.00000000823,
      -0.00000000070,     0.00000000023,
      -0.00000001526,     0.00000000044,
       0.00000000041,     0.00000000133,
       0.00000002123,     0.00000000003,
       0.00000000011,    -0.00000002237,
       0.00000000031,     0.00000000831,
       0.00000000071,    -0.00000000023,
       0.00000001030,    -0.00000000030,
      -0.00000000028,    -0.00000000090,
      -0.00000002014,    -0.00000000003,
      -0.00000000010,     0.00000002081,
       0.00000002481,     0.00000000010,
       0.00000000014,    -0.00000004530,
      -0.00000000024,    -0.00000000409,
       0.00000000171,     0.00000000014,
      -0.00000000008,    -0.00000000046,
      -0.00000000007,     0.00000000081,
      -0.00000002349,    -0.00000000010,
      -0.00000000010,     0.00000003019,
       0.00000000021,     0.00000000260,
      -0.00000000038,    -0.00000000015,
      -0.00000000004,    -0.00000000056,
      -0.00000000032,    -0.00000000118,
       0.00000000023,    -0.00000000001,
      -0.00000000217,     0.00000000044,
       0.00000000917,     0.00000000032,
       0.00000000024,     0.00000001028,
       0.00000000028,     0.00000000104,
       0.00000000146,    -0.00000000029,
      -0.00000000420,    -0.00000000025,
      -0.00000000021,    -0.00000000510,
      -0.00000000057,     0.00000000029,
       0.00000000056,     0.00000000111,
      -0.00000000014,    -0.00000000042,
       0.00000000043,    -0.00000000017,
       0.00000000010,     0.00000000044,
      -0.00000000045,     0.00000000011,
       0.00000000047,    -0.00000000024,
      -0.00000000038,    -0.00000000076,
       0.00000000061,     0.00000000013,
      -0.00000000146,    -0.00000000009,
      -0.00000000009,     0.00000000148,
      -0.00000000042,    -0.00000000009,
       0.00000000137,     0.00000000009,
       0.00000000009,    -0.00000000139,
      -0.00000000146,    -0.00000000009,
       0.00000000010,     0.00000000013,
      -0.00000000025,     0.00000000283,
      -0.00000000022,    -0.00000000559,
      -0.00000000048,     0.00000000015,
      -0.00000000574,     0.00000000019,
       0.00000000016,     0.00000000050,
       0.00000000156,     0.00000000009,
       0.00000000017,    -0.00000000194,
       0.00000000020,     0.00000000524,
       0.00000000045,    -0.00000000014,
       0.00000000534,    -0.00000000018,
      -0.00000000015,    -0.00000000046,
      -0.00000000022,    -0.00000000607,
      -0.00000000052,     0.00000000017,
      -0.00000001106,     0.00000000032,
       0.00000000030,     0.00000000096,
       0.00000001459,     0.00000000003,
       0.00000000009,    -0.00000001550,
       0.00000000023,     0.00000000611,
       0.00000000052,    -0.00000000017,
       0.00000000747,    -0.00000000022,
      -0.00000000020,    -0.00000000065,
      -0.00000001388,    -0.00000000003,
      -0.00000000008,     0.00000001440,
       0.00000001731,     0.00000000008,
       0.00000000011,    -0.00000003112,
      -0.00000000018,    -0.00000000306,
       0.00000000131,     0.00000000010,
      -0.00000000005,     0.00000000056,
      -0.00000001639,    -0.00000000008,
      -0.00000000008,     0.00000002076,
       0.00000000016,     0.00000000194,
      -0.00000000023,    -0.00000000089,
      -0.00000000161,     0.00000000032,
       0.00000000647,     0.00000000024,
       0.00000000017,     0.00000000729,
       0.00000000021,     0.00000000079,
       0.00000000108,    -0.00000000021,
      -0.00000000295,    -0.00000000019,
      -0.00000000015,    -0.00000000364,
      -0.00000000043,     0.00000000022,
       0.00000000041,     0.00000000082,
       0.00000000036,    -0.00000000018,
      -0.00000000028,    -0.00000000056,
       0.00000000046,     0.00000000009,
      -0.00000000109,    -0.00000000006,
      -0.00000000006,     0.00000000111,
       0.00000000102,     0.00000000006,
       0.00000000006,    -0.00000000104,
      -0.00000000111,    -0.00000000006,
      -0.00000000017,     0.00000000211,
      -0.00000000016,    -0.00000000405,
      -0.00000000418,     0.00000000014,
       0.00000000118,     0.00000000006,
       0.00000000012,    -0.00000000144,
       0.00000000015,     0.00000000381,
       0.00000000389,    -0.00000000013,
      -0.00000000017,    -0.00000000447,
      -0.00000000038,     0.00000000012,
      -0.00000000802,     0.00000000024,
       0.00000000022,     0.00000000070,
       0.00000001006,     0.00000000003,
       0.00000000007,    -0.00000001077,
       0.00000000017,     0.00000000448,
       0.00000000038,    -0.00000000012,
       0.00000000542,    -0.00000000016,
      -0.00000000015,    -0.00000000047,
      -0.00000000959,    -0.00000000003,
      -0.00000000006,     0.00000001000,
       0.00000001210,     0.00000000007,
       0.00000000009,    -0.00000002145,
      -0.00000000014,    -0.00000000229,
       0.00000000099,     0.00000000007,
      -0.00000001145,    -0.00000000006,
      -0.00000000006,     0.00000001433,
       0.00000000012,     0.00000000144,
      -0.00000000017,    -0.00000000066,
      -0.00000000119,     0.00000000024,
       0.00000000456,     0.00000000018,
       0.00000000013,     0.00000000517,
       0.00000000015,     0.00000000059,
       0.00000000080,    -0.00000000016,
      -0.00000000208,    -0.00000000014,
      -0.00000000011,    -0.00000000259,
      -0.00000000032,     0.00000000017,
       0.00000000031,     0.00000000061,
      -0.00000000021,    -0.00000000041,
      -0.00000000082,    -0.00000000004,
      -0.00000000004,     0.00000000083,
       0.00000000077,     0.00000000004,
       0.00000000004,    -0.00000000078,
      -0.00000000085,    -0.00000000004,
      -0.00000000012,     0.00000000158,
      -0.00000000012,    -0.00000000294,
      -0.00000000304,     0.00000000010,
       0.00000000089,     0.00000000004,
       0.00000000008,    -0.00000000108,
       0.00000000011,     0.00000000277,
       0.00000000283,    -0.00000000010,
      -0.00000000012,    -0.00000000329,
      -0.00000000028,     0.00000000009,
      -0.00000000582,     0.00000000017,
       0.00000000016,     0.00000000051,
       0.00000000695,     0.00000000002,
       0.00000000005,    -0.00000000750,
       0.00000000012,     0.00000000329,
       0.00000000028,    -0.00000000009,
       0.00000000393,    -0.00000000012,
      -0.00000000664,    -0.00000000002,
      -0.00000000005,     0.00000000696,
       0.00000000847,     0.00000000005,
       0.00000000007,    -0.00000001483,
      -0.00000000010,    -0.00000000170,
       0.00000000075,     0.00000000005,
      -0.00000000802,    -0.00000000005,
      -0.00000000005,     0.00000000991,
       0.00000000009,     0.00000000107,
      -0.00000000013,    -0.00000000050,
      -0.00000000088,     0.00000000017,
       0.00000000321,     0.00000000013,
       0.00000000009,     0.00000000367,
       0.00000000011,     0.00000000044,
       0.00000000059,    -0.00000000011,
      -0.00000000146,    -0.00000000011,
      -0.00000000008,    -0.00000000184,
      -0.00000000024,     0.00000000012,
       0.00000000023,     0.00000000045,
      -0.00000000061,    -0.00000000003,
      -0.00000000003,     0.00000000062,
       0.00000000057,     0.00000000003,
       0.00000000003,    -0.00000000058,
      -0.00000000065,    -0.00000000003,
      -0.00000000008,     0.00000000118,
      -0.00000000009,    -0.00000000213,
      -0.00000000222,     0.00000000007,
       0.00000000068,     0.00000000003,
       0.00000000005,    -0.00000000081,
       0.00000000008,     0.00000000201,
       0.00000000206,    -0.00000000007,
      -0.00000000009,    -0.00000000242,
      -0.00000000021,     0.00000000007,
      -0.00000000422,     0.00000000013,
       0.00000000481,     0.00000000002,
       0.00000000004,    -0.00000000524,
       0.00000000009,     0.00000000241,
       0.00000000021,    -0.00000000007,
       0.00000000286,    -0.00000000009,
      -0.00000000461,    -0.00000000002,
      -0.00000000004,     0.00000000485,
       0.00000000594,     0.00000000004,
       0.00000000006,    -0.00000001027,
      -0.00000000008,    -0.00000000127,
       0.00000000057,     0.00000000004,
      -0.00000000562,    -0.00000000004,
      -0.00000000004,     0.00000000687,
       0.00000000006,     0.00000000079,
      -0.00000000009,    -0.00000000037,
      -0.00000000065,     0.00000000013,
       0.00000000227,     0.00000000010,
       0.00000000007,     0.00000000260,
       0.00000000044,    -0.00000000008,
      -0.00000000103,    -0.00000000008,
      -0.00000000006,    -0.00000000131,
      -0.00000000049,    -0.00000000002,
      -0.00000000005,     0.00000000088,
      -0.00000000006,    -0.00000000155,
      -0.00000000162,     0.00000000005,
       0.00000000051,     0.00000000002,
       0.00000000004,    -0.00000000060,
       0.00000000006,     0.00000000146,
       0.00000000150,    -0.00000000005,
      -0.00000000006,    -0.00000000178,
      -0.00000000015,     0.00000000005,
      -0.00000000307,     0.00000000009,
       0.00000000333,     0.00000000002,
       0.00000000003,    -0.00000000367,
       0.00000000006,     0.00000000177,
       0.00000000208,    -0.00000000006,
      -0.00000000320,    -0.00000000002,
      -0.00000000003,     0.00000000339,
       0.00000000416,     0.00000000003,
       0.00000000004,    -0.00000000713,
      -0.00000000006,    -0.00000000094,
      -0.00000000394,    -0.00000000003,
      -0.00000000003,     0.00000000478,
       0.00000000005,     0.00000000059,
      -0.00000000047,     0.00000000009,
       0.00000000160,     0.00000000008,
       0.00000000005,     0.00000000185,
      -0.00000000072,    -0.00000000006,
      -0.00000000005,    -0.00000000094,
      -0.00000000037,    -0.00000000001,
      -0.00000000004,     0.00000000066,
      -0.00000000005,    -0.00000000113,
      -0.00000000118,     0.00000000004,
       0.00000000004,     0.00000000106,
       0.00000000110,    -0.00000000004,
      -0.00000000005,    -0.00000000131,
      -0.00000000223,     0.00000000007,
       0.00000000231,     0.00000000001,
       0.00000000002,    -0.00000000257,
       0.00000000005,     0.00000000130,
       0.00000000151,    -0.00000000005,
      -0.00000000223,    -0.00000000001,
      -0.00000000002,     0.00000000237,
       0.00000000292,     0.00000000003,
       0.00000000003,    -0.00000000496,
      -0.00000000004,    -0.00000000069,
      -0.00000000277,    -0.00000000002,
      -0.00000000002,     0.00000000333,
       0.00000000113,     0.00000000006,
       0.00000000004,     0.00000000131,
      -0.00000000051,    -0.00000000004,
      -0.00000000003,    -0.00000000067,
      -0.00000000003,     0.00000000049,
      -0.00000000003,    -0.00000000082,
      -0.00000000086,     0.00000000003,
       0.00000000003,     0.00000000077,
       0.00000000080,    -0.00000000003,
      -0.00000000003,    -0.00000000096,
      -0.00000000162,     0.00000000005,
       0.00000000161,     0.00000000001,
       0.00000000002,    -0.00000000181,
       0.00000000003,     0.00000000095,
       0.00000000110,    -0.00000000003,
      -0.00000000156,    -0.00000000001,
      -0.00000000002,     0.00000000166,
       0.00000000206,     0.00000000002,
       0.00000000003,    -0.00000000346,
      -0.00000000003,    -0.00000000051,
      -0.00000000195,    -0.00000000002,
      -0.00000000002,     0.00000000232,
       0.00000000080,     0.00000000004,
       0.00000000003,     0.00000000093,
      -0.00000000002,    -0.00000000059,
      -0.00000000063,     0.00000000002,
       0.00000000002,     0.00000000056,
       0.00000000058,    -0.00000000002,
      -0.00000000002,    -0.00000000070,
      -0.00000000118,     0.00000000004,
       0.00000000112,     0.00000000001,
       0.00000000001,    -0.00000000127,
       0.00000000002,     0.00000000069,
       0.00000000080,    -0.00000000002,
      -0.00000000109,    -0.00000000001,
      -0.00000000001,     0.00000000117,
       0.00000000145,     0.00000000001,
       0.00000000002,    -0.00000000241,
      -0.00000000137,    -0.00000000001,
      -0.00000000001,     0.00000000162,
       0.00000000056,     0.00000000003,
       0.00000000002,     0.00000000066,
      -0.00000000002,    -0.00000000052,
      -0.00000000086,     0.00000000003,
       0.00000000078,     0.00000000001,
       0.00000000001,    -0.00000000090,
       0.00000000002,     0.00000000051,
       0.00000000058,    -0.00000000002,
      -0.00000000076,    -0.00000000001,
      -0.00000000001,     0.00000000082,
       0.00000000102,     0.00000000001,
       0.00000000002,    -0.00000000169,
      -0.00000000096,    -0.00000000001,
      -0.00000000001,     0.00000000113,
      -0.00000000001,    -0.00000000038,
      -0.00000000062,     0.00000000002,
       0.00000000054,     0.00000000001,
       0.00000000001,    -0.00000000063,
      -0.00000000053,    -0.00000000001,
      -0.00000000001,     0.00000000058,
       0.00000000072,     0.00000000001,
       0.00000000001,    -0.00000000118,
      -0.00000000068,    -0.00000000001,
      -0.00000000001,     0.00000000079,
       0.00000000051,     0.00000000001,
       0.00000000001,    -0.00000000083,
      -0.00000000001,     0.00000000056,
       0.00000000036,     0.00000000001,
       0.00000000001,    -0.00000000058,
      -0.00000032706,     0.00000010513,
      -0.00000000336,    -0.00000000742,
       0.00000000006,    -0.00000000016,
      -0.00000031729,     0.00000001670,
       0.00000000057,    -0.00000000015,
      -0.00000000001,    -0.00000000004,
       0.00000001172,    -0.00000030212,
      -0.00000000010,     0.00000000002,
      -0.00000000003,     0.00000000002,
       0.00000000393,     0.00000000097,
       0.00000000010,     0.00000000131,
      -0.00000002424,    -0.00000000886,
      -0.00000000187,     0.00000000030,
       0.00000024930,    -0.00000001510,
      -0.00000000600,     0.00000002377,
      -0.00000000123,    -0.00000000041,
      -0.00000239035,     0.00000001703,
       0.00000000348,     0.00000000577,
       0.00000000056,    -0.00000000025,
       0.00000001800,    -0.00000231835,
       0.00000000591,    -0.00000000227,
      -0.00000000021,    -0.00000000056,
      -0.00000001111,     0.00000000287,
       0.00000000064,    -0.00000000184,
       0.00000000143,     0.00000000091,
       0.00000000155,     0.00000000033,
       0.00000007946,    -0.00000005296,
      -0.00000000115,     0.00000000468,
      -0.00000000008,    -0.00000000012,
      -0.00000047675,    -0.00000035113,
       0.00000003164,    -0.00000004241,
       0.00000000202,     0.00000000129,
       0.00000074513,    -0.00000000781,
      -0.00000000192,    -0.00000000053,
      -0.00000000891,     0.00000073024,
      -0.00000000043,    -0.00000000136,
       0.00000000840,    -0.00000000811,
      -0.00000000056,     0.00000000071,
       0.00000000322,     0.00000000312,
      -0.00000000040,    -0.00000000041,
      -0.00000004843,    -0.00000031704,
       0.00000002023,    -0.00000000849,
       0.00000000050,     0.00000000095,
       0.00000597350,     0.00000001556,
      -0.00000000401,    -0.00000000140,
       0.00000001818,     0.00000599392,
      -0.00000000188,     0.00000000407,
       0.00000001018,     0.00000001955,
       0.00000000106,    -0.00000000063,
      -0.00000001242,    -0.00000000874,
       0.00000000017,     0.00000000084,
       0.00000018763,    -0.00000073235,
       0.00000003369,    -0.00000001743,
       0.00000000103,     0.00000000105,
       0.00001038475,     0.00000000064,
       0.00000000065,     0.00001038160,
      -0.00000000101,     0.00000000547,
      -0.00000000026,     0.00000000139,
      -0.00000001397,     0.00000001394,
       0.00000117212,     0.00000006274,
      -0.00000007855,     0.00000011076,
      -0.00000000469,    -0.00000000392,
       0.00001407001,     0.00000000511,
       0.00000000373,    -0.00000000032,
       0.00000000509,     0.00001406014,
      -0.00000000154,     0.00000000231,
       0.00000000036,    -0.00000000083,
       0.00000000045,    -0.00000000124,
      -0.00000000682,     0.00000001556,
       0.00000000189,    -0.00000000044,
      -0.00000007670,    -0.00000000965,
       0.00000016896,     0.00000137675,
      -0.00000002240,    -0.00000000534,
       0.00001996913,    -0.00000000018,
      -0.00000000018,     0.00001996633,
      -0.00000000049,     0.00000000056,
      -0.00000000171,    -0.00000000596,
      -0.00000000092,     0.00000000022,
       0.00000025109,    -0.00000022499,
       0.00000000289,    -0.00000000031,
       0.00002501595,     0.00000000631,
       0.00000000125,    -0.00000000002,
       0.00000000631,     0.00002501304,
      -0.00000000123,     0.00000000033,
      -0.00000000154,     0.00000000173,
      -0.00000000172,     0.00000000193,
      -0.00000000073,     0.00000000099,
       0.00000001103,     0.00000000883,
      -0.00000000027,    -0.00000000002,
       0.00000000450,     0.00000001186,
      -0.00000000024,     0.00000000012,
      -0.00000001188,     0.00000000352,
      -0.00000000015,    -0.00000000023,
       0.00000000332,     0.00000000357,
       0.00000000019,    -0.00000000012,
      -0.00000000371,     0.00000000407,
       0.00000000012,     0.00000000019,
       0.00000000208,    -0.00000000229,
      -0.00000000389,    -0.00000000346,
      -0.00000000006,    -0.00000000172,
       0.00000000173,    -0.00000000007,
       0.00000000392,     0.00000000437,
      -0.00000000436,     0.00000000389,
       0.00000000002,    -0.00000000008,
      -0.00000000532,    -0.00000000130,
       0.00000000000,     0.00000000007,
      -0.00000000053,     0.00000000000,
       0.00000000000,     0.00000000055,
       0.00000000005,    -0.00000000002,
      -0.00000000010,    -0.00000000040,
      -0.00000000001,    -0.00000000021,
       0.00000000019,     0.00000000001,
       0.00000000011,    -0.00000000029,
       0.00000000030,     0.00000000010,
       0.00000000056,     0.00000000024,
       0.00000000029,    -0.00000000066,
      -0.00000000083,     0.00000000117,
      -0.00000000117,    -0.00000000083,
       0.00000000023,     0.00000000025,
      -0.00000000006,    -0.00000000003,
      -0.00000000004,     0.00000000006,
       0.00000000011,    -0.00000000019,
       0.00000000001,    -0.00000000005,
      -0.00000000005,    -0.00000000001,
      -0.00000000008,    -0.00000000002,
       0.00000000000,     0.00000000006,
      -0.00000000020,     0.00000000010,
       0.00000000005,     0.00000000001,
      -0.00000000452,     0.00000000226,
       0.00000000036,     0.00000000074,
       0.00000000007,    -0.00000000003,
       0.00000002201,     0.00000004500,
       0.00000000741,    -0.00000000361,
      -0.00000000029,    -0.00000000071,
      -0.00000000072,    -0.00000002601,
       0.00000000206,     0.00000000073,
       0.00000002399,     0.00000000015,
      -0.00000000061,     0.00000000223,
       0.00000001265,     0.00000005464,
       0.00000000473,    -0.00000000089,
      -0.00000005556,     0.00000001338,
       0.00000000106,     0.00000000494,
      -0.00000000247,     0.00000000131,
       0.00000000019,     0.00000000040,
      -0.00000001787,    -0.00000003688,
      -0.00000000607,     0.00000000296,
       0.00000000024,     0.00000000058,
       0.00000000306,     0.00000000900,
      -0.00000000107,    -0.00000000014,
      -0.00000000806,     0.00000000268,
       0.00000000008,    -0.00000000115,
      -0.00000000487,    -0.00000002145,
      -0.00000000185,     0.00000000033,
       0.00000002211,    -0.00000000539,
      -0.00000000045,    -0.00000000199,
      -0.00000000051,    -0.00000000006,
      -0.00000000115,    -0.00000000012,
       0.00000000009,    -0.00000000071,
       0.00000000861,     0.00000000148,
       0.00000000014,    -0.00000000192,
      -0.00000000065,    -0.00000000103,
      -0.00000000109,     0.00000000063,
       0.00000003201,    -0.00000000003,
      -0.00000000012,    -0.00000000030,
      -0.00000000183,     0.00000000079,
       0.00000000469,    -0.00000000227,
      -0.00000000040,    -0.00000000076,
       0.00000000211,     0.00000000453,
       0.00000000072,    -0.00000000035,
      -0.00000000181,     0.00000000026,
      -0.00000000029,    -0.00000000185,
       0.00000000132,    -0.00000000054,
      -0.00000000524,     0.00000000263,
       0.00000000043,     0.00000000087,
      -0.00000000257,    -0.00000000517,
      -0.00000000086,     0.00000000041,
       0.00000000106,    -0.00000000027,
       0.00000000030,     0.00000000109,
      -0.00000002631,    -0.00000007710,
      -0.00000002873,     0.00000000856,
      -0.00000000643,    -0.00000001770,
      -0.00000000656,     0.00000000198,
      -0.00000000020,     0.00000000055,
       0.00000000055,     0.00000000021,
       0.00000000243,    -0.00000000084,
      -0.00000000029,    -0.00000000094,
       0.00000018402,     0.00000053521,
       0.00000020168,    -0.00000006179,
       0.00000000303,    -0.00000000364,
      -0.00000000105,    -0.00000000101,
      -0.00000000368,    -0.00000000303,
      -0.00000000097,     0.00000000107,
       0.00000000025,     0.00000000063,
      -0.00000000090,     0.00000000064,
       0.00000000025,     0.00000000045,
       0.00000000064,     0.00000000091,
       0.00000000046,    -0.00000000025,
       0.00000000032,    -0.00000000064,
      -0.00000000065,    -0.00000000033,
      -0.00000000200,     0.00000000136,
       0.00000000052,     0.00000000098,
       0.00000000020,    -0.00000000009,
       0.00000000063,     0.00000000088,
       0.00000000052,    -0.00000000026,
      -0.00000000313,     0.00000000612,
       0.00000000205,     0.00000000130,
       0.00000000022,    -0.00000000031,
       0.00000000615,     0.00000000317,
       0.00000000133,    -0.00000000205,
      -0.00000000031,    -0.00000000022,
       0.00000000063,     0.00000000047,
       0.00000000043,    -0.00000000064,
      -0.00000000062,    -0.00000000035,
      -0.00000000017,    -0.00000000443,
      -0.00000000110,    -0.00000000004,
      -0.00000000446,     0.00000000015,
      -0.00000000006,     0.00000000111,
      -0.00000000737,     0.00000001387,
       0.00000000458,     0.00000000299,
       0.00000000050,    -0.00000000069,
       0.00000000606,     0.00000000321,
       0.00000000153,    -0.00000000219,
      -0.00000000033,    -0.00000000025,
       0.00000000093,     0.00000004175,
       0.00000001053,     0.00000000057,
       0.00000000010,    -0.00000000120,
       0.00000004201,    -0.00000000073,
       0.00000000069,    -0.00000001063,
      -0.00000000122,    -0.00000000012,
       0.00000000320,     0.00000000047,
       0.00000000023,    -0.00000000311,
       0.00000000025,    -0.00000000141,
      -0.00000000036,    -0.00000000008,
      -0.00000000435,    -0.00000000007,
      -0.00000000011,     0.00000000122,
      -0.00000001443,    -0.00000002354,
      -0.00000000391,     0.00000000208,
      -0.00000002381,     0.00000001435,
       0.00000000201,     0.00000000402,
      -0.00000000150,     0.00000000053,
       0.00000000067,     0.00000000134,
       0.00000000035,     0.00000010209,
       0.00000002562,     0.00000000177,
       0.00000000030,    -0.00000000293,
       0.00000004048,    -0.00000000055,
       0.00000000097,    -0.00000001101,
      -0.00000000121,    -0.00000000014,
       0.00000012469,     0.00000021541,
       0.00000003613,    -0.00000001816,
      -0.00000000135,    -0.00000000275,
       0.00000021767,    -0.00000012421,
      -0.00000001785,    -0.00000003691,
      -0.00000000286,     0.00000000129,
       0.00000001153,    -0.00000000413,
      -0.00000000512,    -0.00000001036,
      -0.00000000399,    -0.00000001162,
      -0.00000000191,     0.00000000065,
       0.00000000005,     0.00000000014,
      -0.00000001626,     0.00000001116,
       0.00000000157,     0.00000000309,
      -0.00000016464,    -0.00000009093,
      -0.00000000762,     0.00000001250,
       0.00000000050,     0.00000000031,
      -0.00000009381,     0.00000016505,
       0.00000001247,     0.00000000850,
       0.00000000034,    -0.00000000050,
      -0.00000000371,     0.00000000385,
       0.00000000418,     0.00000000226,
       0.00000033284,     0.00000061970,
       0.00000010384,    -0.00000004860,
      -0.00000000359,    -0.00000000798,
       0.00000014402,    -0.00000010345,
      -0.00000001462,    -0.00000002704,
      -0.00000000190,     0.00000000101,
       0.00000127464,     0.00000075805,
       0.00000006251,    -0.00000009814,
      -0.00000000392,    -0.00000000234,
       0.00000077738,    -0.00000127986,
      -0.00000009870,    -0.00000006756,
      -0.00000000290,     0.00000000395,
       0.00000002842,    -0.00000003046,
       0.00000000141,     0.00000000146,
      -0.00000000021,     0.00000000033,
      -0.00000003237,    -0.00000001820,
       0.00000000211,    -0.00000000123,
       0.00000000033,     0.00000000017,
      -0.00000033179,    -0.00000027534,
      -0.00000002037,     0.00000002734,
       0.00000000110,     0.00000000082,
       0.00000556228,    -0.00000647254,
      -0.00000053870,    -0.00000040635,
      -0.00000001636,     0.00000002166,
      -0.00000441068,     0.00000007640,
       0.00000001147,    -0.00000000623,
      -0.00000000045,    -0.00000000084,
      -0.00000003944,     0.00000448918,
       0.00000000511,     0.00000001104,
       0.00000000062,    -0.00000000032,
      -0.00000001689,     0.00000002937,
       0.00000000282,    -0.00000000086,
       0.00000001587,    -0.00000002530,
      -0.00000000015,    -0.00000000290,
       0.00000971790,     0.00000729223,
       0.00000056361,    -0.00000078322,
      -0.00000003135,    -0.00000002271,
      -0.00004297154,     0.00004999277,
       0.00000415956,     0.00000314064,
       0.00000012658,    -0.00000016754,
      -0.00000000450,    -0.00000000340,
       0.00001710249,    -0.00000049113,
      -0.00000004985,     0.00000004008,
       0.00000000290,     0.00000000511,
       0.00000015007,    -0.00001767582,
      -0.00000003661,    -0.00000006572,
      -0.00000000382,     0.00000000231,
       0.00000011405,    -0.00000024480,
      -0.00000002506,     0.00000000587,
       0.00000000012,     0.00000000357,
      -0.00000014157,     0.00000017501,
       0.00000000160,     0.00000002624,
       0.00000000302,    -0.00000000066,
       0.00000013961,    -0.00000000115,
      -0.00000000639,    -0.00000099878,
       0.00000000010,    -0.00000000069,
      -0.00000003960,    -0.00000002603,
      -0.00000000165,     0.00000000316,
       0.00000001712,    -0.00000003883,
      -0.00000000280,    -0.00000000230,
       0.00000000047,    -0.00000001000,
      -0.00000000002,     0.00000000082,
       0.00000000073,     0.00000000838,
       0.00000000000,     0.00000000087,
       0.00001580815,     0.00000001347,
      -0.00000000216,     0.00000000036,
       0.00000000001,     0.00000000020,
       0.00000004895,     0.00001262890,
      -0.00000000096,     0.00000000507,
      -0.00000023346,     0.00000045347,
      -0.00000001298,    -0.00000006853,
       0.00000000070,    -0.00000000172,
       0.00000015605,     0.00000091588,
      -0.00000001872,     0.00000004357,
      -0.00000000050,    -0.00000000261,
      -0.00000002812,     0.00000004134,
      -0.00000000290,    -0.00000000289,
       0.00000000018,    -0.00000000035,
      -0.00000003432,    -0.00000003411,
       0.00000000276,    -0.00000000295,
       0.00000000052,    -0.00000000101,
       0.00000000021,     0.00000000008,
      -0.00000000640,    -0.00000000359,
       0.00000000056,    -0.00000000137,
      -0.00000001203,     0.00000000679,
      -0.00000000054,    -0.00000000093,
      -0.00000000710,    -0.00000001210,
       0.00000000091,    -0.00000000070,
      -0.00000000143,    -0.00000000163,
       0.00000000168,    -0.00000000120,
       0.00000000248,    -0.00000000873,
       0.00000000138,     0.00000000042,
       0.00000005341,     0.00000003256,
      -0.00000000511,     0.00000001205,
      -0.00000000063,    -0.00000000032,
       0.00000003977,    -0.00000003047,
       0.00000000192,     0.00000000346,
       0.00000003504,     0.00000004010,
      -0.00000000332,     0.00000000341,
       0.00000000179,     0.00000000166,
      -0.00000000205,     0.00000000000,
      -0.00000000091,    -0.00000000007,
      -0.00000000115,     0.00000000186,
      -0.00000000035,    -0.00000000018,
      -0.00000000189,    -0.00000000115,
       0.00000000016,    -0.00000000037,
      -0.00000000061,    -0.00000000022,
       0.00000000025,    -0.00000000058,
       0.00000000013,    -0.00000000306,
       0.00000000106,    -0.00000000013,
       0.00000000745,     0.00000000068,
       0.00000000030,     0.00000000268,
       0.00000000874,    -0.00000001487,
       0.00000000296,     0.00000000141,
       0.00000001542,     0.00000000856,
      -0.00000000125,     0.00000000321,
       0.00000000476,     0.00000000175,
      -0.00000000196,     0.00000000452,
       0.00000000025,    -0.00000000017,
       0.00000000090,    -0.00000000027,
       0.00000000031,     0.00000000045,
       0.00000000027,    -0.00000000298,
       0.00000000099,    -0.00000000007,
       0.00000000304,     0.00000000023,
       0.00000000012,     0.00000000102,
       0.00000000125,    -0.00000000017,
       0.00000000011,     0.00000000123,
      -0.00000000014,    -0.00000000044,
       0.00000000044,    -0.00000000015,
       0.00000000041,    -0.00000000046,
      -0.00000000046,    -0.00000000041,
       0.00000000017,    -0.00000000063,
      -0.00000000063,    -0.00000000017,
       0.00000000039,    -0.00000000085,
       0.00000000458,    -0.00000000372,
       0.00000000000,    -0.00000000129,
      -0.00000000372,    -0.00000000458,
      -0.00000000130,     0.00000000000,
      -0.00000000027,    -0.00000000057,
      -0.00000000057,     0.00000000028,
      -0.00000000099,    -0.00000000011,
      -0.00000000011,     0.00000000100,
      -0.00000000025,     0.00000000027,
       0.00000000027,     0.00000000025,
       0.00000000105,     0.00000000035,
       0.00000000035,    -0.00000000105,
       0.00000000147,     0.00000000103,
       0.00000000030,    -0.00000000020,
       0.00000000447,    -0.00000000651,
      -0.00000000100,    -0.00000000137,
       0.00000004761,    -0.00000002696,
       0.00000000152,    -0.00000000884,
      -0.00000000055,    -0.00000000041,
      -0.00000002699,    -0.00000004763,
      -0.00000000885,    -0.00000000151,
      -0.00000000041,     0.00000000055,
      -0.00000000052,    -0.00000000084,
      -0.00000000082,     0.00000000052,
       0.00000000101,    -0.00000000105,
      -0.00000000105,    -0.00000000101,
       0.00000001098,     0.00000001116,
       0.00000000209,    -0.00000000112,
      -0.00000000004,    -0.00000000019,
       0.00000004431,    -0.00000004593,
      -0.00000000516,    -0.00000000879,
      -0.00000000079,     0.00000000020,
       0.00000044498,    -0.00000016089,
       0.00000001711,    -0.00000004853,
      -0.00000000220,    -0.00000000204,
      -0.00000016120,    -0.00000044527,
      -0.00000004863,    -0.00000001707,
      -0.00000000204,     0.00000000221,
      -0.00000000198,    -0.00000000519,
      -0.00000000500,     0.00000000206,
       0.00000001038,    -0.00000001029,
       0.00000001874,    -0.00000001076,
       0.00000000032,    -0.00000000171,
      -0.00000001075,    -0.00000001874,
      -0.00000000171,    -0.00000000032,
       0.00000000046,     0.00000000035,
       0.00000000034,    -0.00000000044,
       0.00000000088,     0.00000000018,
       0.00000000018,    -0.00000000088,
       0.00000000034,    -0.00000000102,
       0.00000006804,     0.00000009723,
       0.00000001115,    -0.00000000609,
      -0.00000000023,    -0.00000000077,
       0.00000035970,    -0.00000027906,
      -0.00000002573,    -0.00000004365,
      -0.00000000303,     0.00000000097,
       0.00000348726,    -0.00000061913,
       0.00000009345,    -0.00000016686,
      -0.00000000541,    -0.00000000576,
      -0.00000062178,    -0.00000349111,
      -0.00000016772,    -0.00000009337,
      -0.00000000577,     0.00000000546,
      -0.00000000350,    -0.00000002026,
       0.00000000134,     0.00000000057,
      -0.00000001884,     0.00000000425,
       0.00000000057,    -0.00000000141,
      -0.00000000042,     0.00000000010,
      -0.00000000004,     0.00000000078,
       0.00000000082,    -0.00000000003,
       0.00000020777,    -0.00000011649,
       0.00000000513,    -0.00000002316,
       0.00000000095,     0.00000000305,
       0.00000030337,    -0.00000007930,
       0.00000000704,    -0.00000001388,
      -0.00000007916,    -0.00000030324,
      -0.00000001385,    -0.00000000704,
      -0.00000000074,    -0.00000000071,
      -0.00000000218,     0.00000000226,
      -0.00000000069,    -0.00000000038,
      -0.00000000036,     0.00000000069,
       0.00000000078,     0.00000000139,
       0.00000000150,    -0.00000000073,
      -0.00000000038,     0.00000000042,
       0.00000000047,     0.00000000047,
       0.00000000054,     0.00000000056,
       0.00000000060,     0.00000000026,
       0.00000000004,     0.00000000053,
       0.00000000051,     0.00000000017,
       0.00000000056,    -0.00000000044,
      -0.00000000044,    -0.00000000056,
      -0.00000000039,    -0.00000000013,
      -0.00000000013,     0.00000000039,
       0.00000000037,     0.00000000032,
       0.00000000032,    -0.00000000037,
       0.00000031059,     0.00000056203,
       0.00000003464,    -0.00000003121,
      -0.00000000117,    -0.00000000146,
       0.00000196611,    -0.00000133507,
      -0.00000012227,    -0.00000012989,
      -0.00000000556,     0.00000000450,
       0.00001859231,    -0.00000000481,
      -0.00000000698,     0.00000000383,
       0.00000000014,    -0.00000000162,
      -0.00000002478,    -0.00001864028,
      -0.00000000218,     0.00000000604,
      -0.00000000172,     0.00000000008,
       0.00000000088,    -0.00000000632,
       0.00000000248,     0.00000000348,
       0.00000000412,     0.00000000350,
       0.00000000026,    -0.00000000034,
       0.00000000449,    -0.00000000538,
      -0.00000000047,    -0.00000000035,
       0.00000000084,    -0.00000000017,
       0.00000000019,    -0.00000000100,
       0.00000394891,    -0.00000101785,
       0.00000009319,    -0.00000018108,
      -0.00000000575,    -0.00000000586,
       0.00000004933,     0.00000020481,
       0.00000000949,     0.00000000500,
       0.00000402497,    -0.00000000110,
      -0.00000000115,    -0.00000000009,
       0.00000000127,    -0.00000402081,
       0.00000000029,     0.00000000122,
      -0.00000000008,    -0.00000000111,
      -0.00000000111,     0.00000000008,
       0.00000003883,    -0.00000000512,
       0.00000013983,    -0.00000003560,
       0.00000033447,    -0.00000008553,
       0.00000000026,     0.00000000059,
      -0.00000000251,    -0.00000000510,
      -0.00000000036,    -0.00000000090,
      -0.00000000224,     0.00000000077,
       0.00000000011,     0.00000000038,
       0.00000000252,     0.00000000227,
       0.00000000233,    -0.00000000244,
       0.00000000333,     0.00000000011,
       0.00000000006,    -0.00000000096,
       0.00000000050,     0.00000000043,
      -0.00000000043,     0.00000000049,
       0.00000000038,    -0.00000000054,
       0.00000000470,    -0.00000000256,
      -0.00000000083,    -0.00000000172,
      -0.00000000256,    -0.00000000470,
      -0.00000000172,     0.00000000084,
       0.00000000101,     0.00000000002,
       0.00000000002,    -0.00000000101,
      -0.00000000039,     0.00000000018,
      -0.00000000243,    -0.00000000084,
      -0.00000000021,     0.00000000029,
      -0.00000000084,     0.00000000243,
       0.00000000029,     0.00000000021,
       0.00000000186,     0.00000000095,
       0.00000000095,    -0.00000000187,
       0.00000000103,     0.00000000027,
       0.00000000026,    -0.00000000103,
       0.00000003738,     0.00000000981,
      -0.00000000052,     0.00000000202,
       0.00000000113,    -0.00000000029,
      -0.00000000029,    -0.00000000113,
       0.00000003740,     0.00000000122,
       0.00000000115,    -0.00000000029,
       0.00000000029,     0.00000000115,
       0.00000000000,     0.00000000008,
       0.00000109757,    -0.00000002286,
       0.00000000040,    -0.00000000017,
      -0.00000007819,    -0.00000514214,
      -0.00000000062,    -0.00000000213,
      -0.00000066453,    -0.00000006314,
      -0.00000001577,    -0.00000000334,
       0.00000000094,    -0.00000000004,
      -0.00000017342,     0.00000009595,
      -0.00000003042,     0.00000000061,
      -0.00000000092,    -0.00000000008,
      -0.00000000055,    -0.00000003521,
       0.00000000020,     0.00000000101,
       0.00000000314,    -0.00000000633,
      -0.00000000003,     0.00000000254,
       0.00000000000,     0.00000000056,
       0.00007808378,     0.00000000016,
      -0.00000000474,     0.00000000195,
       0.00000000011,     0.00001327642,
      -0.00000017629,    -0.00000009021,
      -0.00000000419,    -0.00000001622,
      -0.00000004470,     0.00000035289,
      -0.00000000811,     0.00000000837,
       0.00000000775,    -0.00000004198,
      -0.00000000307,     0.00000000127,
      -0.00000004200,    -0.00000000776,
       0.00000000127,     0.00000000307,
       0.00000085242,     0.00000002308,
      -0.00000000053,     0.00000000393,
      -0.00000000909,     0.00000000176,
      -0.00000000056,    -0.00000000061,
       0.00000000039,     0.00000000916,
      -0.00000000061,     0.00000000056,
       0.00000311006,    -0.00000000024,
       0.00000000000,     0.00000000213,
       0.00000000205,    -0.00000000001,
       0.00000728234,     0.00000000115,
       0.00000000000,     0.00000000204,
      -0.00000000053,     0.00000000000,
       0.00000000040,     0.00000000054,
       0.00000000098,    -0.00000000090,
       0.00000000069,     0.00000000073,
       0.00000000026,    -0.00000000029,
       0.00000000266,    -0.00000000279,
      -0.00000000114,    -0.00000000096,
       0.00000002866,    -0.00000000966,
      -0.00000000340,    -0.00000000920,
      -0.00000000968,    -0.00000002869,
      -0.00000000921,     0.00000000341,
       0.00000000223,    -0.00000000096,
      -0.00000000039,    -0.00000000068,
      -0.00000000096,    -0.00000000223,
      -0.00000000068,     0.00000000039,
      -0.00000000018,     0.00000000036,
       0.00000000036,     0.00000000018,
       0.00000000037,     0.00000000043,
       0.00000000052,    -0.00000000049,
      -0.00000000013,    -0.00000000045,
      -0.00000000162,     0.00000000059,
      -0.00000001027,    -0.00000000158,
      -0.00000000049,     0.00000000056,
      -0.00000000141,     0.00000001026,
       0.00000000056,     0.00000000049,
      -0.00000000270,    -0.00000000039,
      -0.00000000037,     0.00000000276,
       0.00000000043,    -0.00000000029,
       0.00000000281,     0.00000000043,
       0.00000000044,    -0.00000000282,
       0.00000000122,     0.00000000002,
       0.00000000002,    -0.00000000122,
       0.00000000228,    -0.00000003210,
      -0.00000000098,     0.00000000004,
       0.00000000846,     0.00000000134,
       0.00000000134,    -0.00000000849,
      -0.00000009786,    -0.00000001271,
      -0.00000000012,     0.00000000135,
       0.00000001400,     0.00000000001,
      -0.00000000001,    -0.00000001395,
      -0.00000010872,     0.00000001443,
       0.00000001576,    -0.00000000001,
      -0.00000000001,     0.00000001570,
       0.00000003882,    -0.00000003513,
       0.00000002073,     0.00000000000,
       0.00000000000,    -0.00000002073,
       0.00000003238,     0.00000003561,
       0.00000002199,     0.00000000001,
       0.00000000000,     0.00000002199,
       0.00000000212,     0.00000000289,
       0.00000000093,    -0.00000000091,
       0.00000000988,    -0.00000000881,
      -0.00000000367,    -0.00000000318,
       0.00000010514,    -0.00000001294,
      -0.00000000755,    -0.00000003060,
      -0.00000001311,    -0.00000010540,
      -0.00000003067,     0.00000000761,
       0.00000002107,    -0.00000000900,
      -0.00000000313,    -0.00000000670,
       0.00000000042,     0.00000000101,
       0.00000002237,    -0.00000000330,
      -0.00000000179,    -0.00000000642,
      -0.00000000325,    -0.00000002234,
      -0.00000000642,     0.00000000178,
       0.00000000039,     0.00000000240,
       0.00000000240,    -0.00000000039,
      -0.00000000021,    -0.00000000061,
      -0.00000000201,     0.00000000068,
       0.00000000041,     0.00000000036,
      -0.00000002115,     0.00000000264,
       0.00000000317,     0.00000000391,
       0.00000000023,    -0.00000000032,
       0.00000000264,     0.00000002114,
       0.00000000392,    -0.00000000315,
      -0.00000000032,    -0.00000000023,
       0.00000000677,     0.00000004614,
       0.00000001206,    -0.00000001062,
      -0.00000003246,     0.00000000348,
       0.00000000465,     0.00000000592,
       0.00000000349,     0.00000003248,
       0.00000000592,    -0.00000000466,
       0.00000000445,    -0.00000000139,
      -0.00000000059,    -0.00000000056,
      -0.00000000463,    -0.00000001441,
      -0.00000000178,     0.00000000189,
      -0.00000000853,    -0.00000015629,
      -0.00000001502,     0.00000001053,
       0.00000000052,     0.00000000062,
      -0.00000015610,     0.00000000807,
       0.00000001050,     0.00000001499,
       0.00000000062,    -0.00000000052,
      -0.00000000034,    -0.00000000071,
      -0.00000000065,     0.00000000029,
      -0.00000000069,    -0.00000000103,
      -0.00000000103,     0.00000000069,
      -0.00000068159,     0.00000007485,
       0.00000009899,     0.00000012467,
       0.00000000748,    -0.00000001052,
      -0.00000000091,    -0.00000000857,
      -0.00000000156,     0.00000000124,
      -0.00000002305,    -0.00000042733,
      -0.00000004031,     0.00000002817,
       0.00000000142,     0.00000000168,
      -0.00000042766,     0.00000002310,
       0.00000002824,     0.00000004040,
       0.00000000168,    -0.00000000143,
       0.00000000555,     0.00000002101,
       0.00000000083,    -0.00000000062,
       0.00000006499,    -0.00000001826,
      -0.00000000171,    -0.00000000226,
       0.00000079084,    -0.00000000018,
       0.00000000066,    -0.00000000005,
       0.00000000009,    -0.00000078833,
       0.00000000026,    -0.00000000090,
      -0.00000002722,    -0.00000000380,
       0.00000000064,    -0.00000000415,
      -0.00000000241,    -0.00000000002,
      -0.00000000001,     0.00000000242,
      -0.00000058863,    -0.00001063777,
      -0.00000100530,     0.00000070994,
       0.00000003526,     0.00000004251,
       0.00000036320,    -0.00000001949,
      -0.00000002427,    -0.00000003426,
      -0.00000000146,     0.00000000122,
       0.00000497112,     0.00000000042,
       0.00000000547,     0.00000000079,
      -0.00000000041,    -0.00000497881,
      -0.00000000036,    -0.00000000438,
      -0.00000000283,    -0.00000000106,
      -0.00000000106,     0.00000000283,
      -0.00000000161,    -0.00000004134,
      -0.00000000721,    -0.00000009034,
       0.00000002707,    -0.00000000020,
      -0.00000000068,    -0.00000012341,
       0.00000000199,    -0.00000000757,
       0.00000000338,     0.00000000049,
      -0.00000003558,    -0.00000000063,
       0.00000000240,    -0.00000000070,
      -0.00000000006,    -0.00000000150,
       0.00000000016,    -0.00000000065,
       0.00000001097,     0.00000000021,
      -0.00000000004,     0.00000000277,
       0.00018549188,     0.00000000069,
       0.00000003914,     0.00000000816,
       0.00000000008,     0.00001843741,
      -0.00000000059,     0.00000000169,
       0.00000001405,    -0.00000053802,
       0.00000002540,     0.00000003592,
      -0.00000000090,     0.00000000214,
      -0.00000026887,    -0.00000002891,
       0.00000001790,    -0.00000005084,
       0.00000000107,     0.00000000180,
       0.00000004231,    -0.00000009657,
      -0.00000000943,    -0.00000000243,
      -0.00000009661,    -0.00000004235,
      -0.00000000242,     0.00000000946,
       0.00000076240,     0.00000001893,
      -0.00000000018,     0.00000000171,
      -0.00000000029,    -0.00000000395,
      -0.00000000392,     0.00000000080,
       0.00000171862,     0.00000000037,
       0.00000000519,     0.00000000000,
       0.00000000000,    -0.00000000518,
       0.00000000586,     0.00000000000,
       0.00000000000,     0.00000000586,
      -0.00000000169,    -0.00000000065,
      -0.00000000065,     0.00000000169,
       0.00000000187,     0.00000000019,
       0.00000000018,    -0.00000000186,
       0.00000001929,     0.00000000141,
      -0.00000000139,    -0.00000000002,
      -0.00000000002,     0.00000000139,
      -0.00000006705,    -0.00000002537,
       0.00000002889,     0.00000000720,
      -0.00000000012,    -0.00000000090,
       0.00000000720,    -0.00000002888,
      -0.00000000090,     0.00000000012,
       0.00000008990,     0.00000000931,
      -0.00000001161,     0.00000000054,
       0.00000000054,     0.00000001158,
       0.00000003260,     0.00000000044,
      -0.00000005363,    -0.00000000439,
       0.00000000007,    -0.00000000163,
      -0.00000000363,    -0.00000000067,
      -0.00000000068,     0.00000000398,
      -0.00000000229,    -0.00000000024,
      -0.00000000024,     0.00000000229,
      -0.00000000157,    -0.00000001091,
       0.00000000116,    -0.00000000016,
       0.00000125181,     0.00000030837,
      -0.00000000501,    -0.00000004209,
      -0.00000000240,     0.00000000966,
      -0.00000047338,    -0.00000005814,
       0.00000000181,     0.00000000768,
      -0.00000005805,     0.00000047315,
       0.00000000767,    -0.00000000181,
      -0.00000075130,     0.00000002130,
      -0.00000000033,    -0.00000001506,
      -0.00000010345,     0.00000000605,
       0.00000000036,    -0.00000000143,
       0.00000000599,     0.00000010399,
      -0.00000000143,    -0.00000000036,
      -0.00000015469,    -0.00000002868,
       0.00000000142,    -0.00000000761,
       0.00000000007,     0.00000000046,
       0.00000000149,    -0.00000000019,
       0.00000001412,     0.00000000000,
       0.00000000000,    -0.00000001416,
      -0.00002419922,    -0.00000294324,
       0.00000009665,     0.00000039161,
       0.00000005487,    -0.00000044720,
      -0.00000000727,     0.00000000174,
       0.00000678722,     0.00000000088,
      -0.00000000088,    -0.00000678013,
       0.00000000017,     0.00000000105,
       0.00000000105,    -0.00000000017,
       0.00000002698,     0.00000000846,
      -0.00000000999,    -0.00000000191,
      -0.00000000191,     0.00000000999,
      -0.00000004018,    -0.00000000017,
       0.00000000000,    -0.00000000177,
      -0.00000004848,    -0.00000000295,
       0.00000000249,     0.00000000013,
       0.00000000013,    -0.00000000249,
      -0.00000399571,     0.00000022920,
       0.00000001392,    -0.00000005558,
      -0.00000001138,    -0.00000020020,
       0.00000000278,     0.00000000070,
      -0.00000000887,     0.00000000068,
       0.00000000058,    -0.00000000218,
       0.00000018275,     0.00000001074,
      -0.00000000015,     0.00000000268,
       0.00000000317,     0.00000000048,
       0.00000000048,    -0.00000000417,
       0.00000024133,     0.00000003454,
      -0.00000000114,     0.00000000711,
       0.00000000006,    -0.00000000111,
       0.00000000054,    -0.00000000002,
       0.00051844499,     0.00000000082,
       0.00000000004,     0.00002645235,
       0.00000030934,    -0.00000007598,
      -0.00000000119,     0.00000001005,
      -0.00000003796,    -0.00000061912,
       0.00000000462,     0.00000000240,
      -0.00000001254,    -0.00000004375,
      -0.00000000682,     0.00000000147,
      -0.00000004372,     0.00000001254,
       0.00000000147,     0.00000000681,
       0.00000086437,    -0.00000007080,
       0.00000000082,     0.00000000831,
       0.00000008621,    -0.00000000502,
      -0.00000000030,     0.00000000120,
      -0.00000000499,    -0.00000008625,
       0.00000000120,     0.00000000030,
      -0.00000002565,    -0.00000000471,
       0.00000000375,     0.00000000045,
       0.00000000045,    -0.00000000374,
      -0.00000035348,    -0.00000006416,
       0.00000000122,    -0.00000000637,
       0.00000009985,     0.00000000559,
      -0.00000000036,    -0.00000000143,
       0.00000000556,    -0.00000009974,
      -0.00000000143,     0.00000000036,
       0.00000000882,    -0.00000000129,
      -0.00000000129,    -0.00000000882,
       0.00000535112,    -0.00000030704,
      -0.00000001879,     0.00000007507,
       0.00000000975,     0.00000017090,
      -0.00000000237,    -0.00000000059,
       0.00000000267,    -0.00000000052,
      -0.00000000024,     0.00000000350,
      -0.00000000193,     0.00000000018,
      -0.00000000018,    -0.00000000193,
       0.00000000176,    -0.00000000010,
       0.00000000010,     0.00000000176,
      -0.00000009180,     0.00000002921,
       0.00000000096,     0.00000000218,
       0.00000015116,    -0.00000001853,
      -0.00000000059,     0.00000000225,
       0.00000001858,     0.00000015134,
      -0.00000000226,    -0.00000000059,
      -0.00000001225,    -0.00000004271,
      -0.00000000665,     0.00000000144,
       0.00000004269,    -0.00000001224,
      -0.00000000144,    -0.00000000665,
      -0.00000003385,     0.00000000122,
      -0.00000006716,    -0.00000016405,
       0.00000001366,     0.00000000858,
       0.00000000451,    -0.00000000101,
       0.00000000374,     0.00000000387,
       0.00000000381,     0.00000000497,
      -0.00000000117,     0.00000000118,
       0.00000000118,     0.00000000117,
       0.00000016343,    -0.00000002317,
       0.00000000110,     0.00000000842,
       0.00000000191,    -0.00000000017,
      -0.00000000017,    -0.00000000201,
      -0.00000006232,     0.00000006486,
      -0.00000000076,    -0.00000000075,
      -0.00000008527,    -0.00000000478,
       0.00000000030,     0.00000000120,
      -0.00000000479,     0.00000008527,
       0.00000000120,    -0.00000000030,
       0.00000002376,    -0.00000000180,
       0.00000000009,     0.00000000140,
      -0.00000492873,    -0.00000028176,
       0.00000001728,     0.00000006902,
       0.00000000951,    -0.00000016675,
      -0.00000000237,     0.00000000059,
      -0.00000000685,     0.00000000059,
       0.00000000033,     0.00000000660,
       0.00000005629,     0.00000000002,
       0.00000000014,     0.00000006515,
      -0.00000000193,    -0.00000000018,
       0.00000000018,    -0.00000000193,
       0.00000012063,     0.00000000022,
      -0.00000000013,    -0.00000002277,
       0.00000000167,     0.00000000010,
      -0.00000000010,     0.00000000167,
      -0.00000000115,    -0.00000000014,
       0.00000000013,    -0.00000000116,
      -0.00000000292,     0.00000000078,
      -0.00000000079,    -0.00000000293,
       0.00000000040,     0.00000000093,
      -0.00000000093,     0.00000000040,
       0.00000000016,     0.00000000036,
      -0.00000002688,     0.00000017743,
      -0.00000001237,    -0.00000001772,
      -0.00000000799,    -0.00000000114,
       0.00000000076,    -0.00000000054,
      -0.00000000367,     0.00000000048,
       0.00000000048,     0.00000000367,
      -0.00000018295,     0.00000003376,
      -0.00000000109,    -0.00000000671,
      -0.00000000240,     0.00000000017,
       0.00000000017,     0.00000000241,
       0.00000000039,     0.00000000521,
      -0.00000003881,     0.00000000284,
      -0.00000000050,    -0.00000000689,
       0.00000002709,    -0.00000000002,
      -0.00000007224,     0.00000000010,
       0.00000000588,     0.00000000005,
      -0.00000000005,     0.00000000591,
      -0.00000000088,     0.00000000018,
       0.00000000018,     0.00000000088,
       0.00000000749,     0.00000000493,
       0.00000000206,    -0.00000001270,
       0.00000000115,     0.00000000016,
      -0.00000003472,     0.00000000008,
       0.00000000495,    -0.00000000047,
      -0.00000000047,    -0.00000000492,
      -0.00000000015,    -0.00000000137,
      -0.00000000137,     0.00000000015,
      -0.00000000984,    -0.00000009217,
       0.00000007634,    -0.00000008520,
      -0.00000008520,    -0.00000007635,
       0.00000000046,    -0.00000000011,
       0.00000000432,     0.00000000000,
       0.00000000000,    -0.00000000432,
       0.00000378084,     0.00000021402,
      -0.00000001336,    -0.00000005336,
      -0.00000001123,     0.00000019657,
       0.00000000279,    -0.00000000070,
       0.00000002459,    -0.00000000049,
      -0.00000000006,    -0.00000002687,
       0.00000610076,    -0.00000680529,
       0.00000001020,    -0.00000007808,
       0.00000008268,     0.00000007410,
       0.00000843696,    -0.00000000017,
       0.00000000017,    -0.00000843700,
       0.00000003955,    -0.00000000017,
       0.00000000000,     0.00000000250,
       0.00101174992,     0.00000000018,
       0.00000003201,     0.00000002429,
       0.00000000000,     0.00003322163,
      -0.00000000060,     0.00000000086,
      -0.00000005031,    -0.00000011229,
      -0.00000000007,    -0.00000000100,
      -0.00000005613,     0.00000010063,
       0.00000008633,    -0.00000009664,
      -0.00000000951,     0.00000000064,
      -0.00000009661,    -0.00000008635,
       0.00000000064,     0.00000000951,
       0.00000005934,     0.00000024191,
       0.00000001835,    -0.00000002502,
       0.00000000434,    -0.00000000152,
      -0.00000000061,    -0.00000000043,
      -0.00000002485,    -0.00000002773,
       0.00000002772,    -0.00000002486,
       0.00000008531,    -0.00000009545,
      -0.00000000939,     0.00000000063,
       0.00000009547,     0.00000008529,
      -0.00000000063,    -0.00000000939,
       0.00000000107,    -0.00000000021,
       0.00000000021,     0.00000000107,
       0.00000000050,     0.00000000191,
      -0.00000000191,     0.00000000050,
       0.00000000020,    -0.00000000088,
       0.00000000101,     0.00000000023,
      -0.00000000862,     0.00000000000,
       0.00000000000,    -0.00000000863,
       0.00000000330,     0.00000000000,
       0.00000000000,     0.00000000330,
       0.00000000058,     0.00000000000,
       0.00000000000,     0.00000000058,
       0.00000000124,     0.00000000000,
       0.00000000000,     0.00000000124,
       0.00000004753,     0.00000000000,
       0.00000000000,    -0.00000000015,
       0.00000000000,     0.00000004753,
       0.00000000015,     0.00000000000,
      -0.00000000515,     0.00000000055,
       0.00000000081,     0.00000000000,
       0.00000000000,     0.00000000081,
      -0.00000000116,     0.00000000000,
       0.00000000000,    -0.00000000116,
       0.00000000082,     0.00000000000,
       0.00000000000,     0.00000000082,
       0.00000006464,     0.00000010152,
      -0.00000000499,     0.00000000222,
      -0.00000000013,    -0.00000000012,
       0.00000007462,     0.00000030197,
      -0.00000000814,     0.00000000223,
      -0.00000000005,    -0.00000000008,
      -0.00000030224,     0.00000007072,
      -0.00000000234,    -0.00000000813,
       0.00000000008,    -0.00000000004,
       0.00000001421,     0.00000001685,
       0.00000000082,    -0.00000000060,
      -0.00000001863,     0.00000002015,
       0.00000000055,     0.00000000085,
       0.00000002506,     0.00000005357,
       0.00000000831,    -0.00000000427,
      -0.00000000033,    -0.00000000085,
      -0.00000001567,    -0.00000003991,
       0.00000000578,     0.00000000034,
       0.00000003724,    -0.00000001451,
      -0.00000000018,     0.00000000599,
       0.00000001853,     0.00000007985,
       0.00000000693,    -0.00000000130,
      -0.00000008116,     0.00000001953,
       0.00000000153,     0.00000000721,
      -0.00000001820,     0.00000000670,
       0.00000000113,     0.00000000236,
       0.00000000024,    -0.00000000009,
      -0.00000002442,    -0.00000003715,
      -0.00000000432,     0.00000000206,
       0.00000000023,     0.00000000059,
      -0.00000000939,     0.00000003908,
      -0.00000000591,    -0.00000000203,
      -0.00000003810,    -0.00000000996,
       0.00000000198,    -0.00000000605,
      -0.00000000766,    -0.00000008972,
      -0.00000000740,     0.00000000189,
       0.00000009061,    -0.00000000829,
      -0.00000000205,    -0.00000000757,
       0.00000002866,     0.00000000817,
      -0.00000000169,     0.00000000656,
       0.00000000230,     0.00000007793,
      -0.00000000353,    -0.00000000153,
      -0.00000007636,     0.00000000258,
       0.00000000160,    -0.00000000282,
       0.00000000919,     0.00000001082,
       0.00000000104,    -0.00000000090,
      -0.00000001087,     0.00000000849,
       0.00000000095,     0.00000000101,
       0.00000001123,     0.00000000579,
      -0.00000000049,     0.00000000098,
      -0.00000003885,     0.00000016908,
      -0.00000000842,    -0.00000000642,
      -0.00000016904,    -0.00000003882,
       0.00000000642,    -0.00000000843,
      -0.00000000540,     0.00000001885,
      -0.00000000242,    -0.00000000247,
      -0.00000005271,     0.00000022909,
      -0.00000001141,    -0.00000000864,
      -0.00000022897,    -0.00000005251,
       0.00000000866,    -0.00000001144,
      -0.00000002307,    -0.00000000240,
       0.00000000088,    -0.00000000154,
      -0.00000007469,     0.00000032516,
      -0.00000001620,    -0.00000001233,
       0.00000000051,    -0.00000000064,
      -0.00000032512,    -0.00000007469,
       0.00000001233,    -0.00000001620,
       0.00000000064,     0.00000000051,
       0.00000000272,     0.00000000493,
      -0.00000009367,     0.00000040732,
      -0.00000002029,    -0.00000001544,
       0.00000000064,    -0.00000000081,
      -0.00000040733,    -0.00000009346,
       0.00000001545,    -0.00000002031,
       0.00000000081,     0.00000000064,
      -0.00000000001,    -0.00000000009,
       0.00000000002,    -0.00000000003,
      -0.00000000004,    -0.00000000001,
      -0.00000000012,    -0.00000000006,
      -0.00000000001,     0.00000000003,
      -0.00000028326,     0.00000063190,
       0.00000013905,     0.00000006615,
      -0.00000000011,     0.00000000116,
       0.00000000016,    -0.00000000003,
       0.00000000123,     0.00000000016,
      -0.00000000003,    -0.00000000017,
      -0.00000000003,     0.00000000002,
       0.00000000003,     0.00000000007,
       0.00000001537,    -0.00000003431,
      -0.00000000742,    -0.00000000350,
      -0.00000000096,     0.00000000215,
       0.00000000046,     0.00000000022,
      -0.00000000015,    -0.00000000036,
       0.00000000015,     0.00000000042,
      -0.00000000050,     0.00000000006,
       0.00000000006,     0.00000000050,
      -0.00000000117,     0.00000000013,
       0.00000000000,     0.00000000063,
      -0.00000000245,     0.00000000170,
       0.00000000064,     0.00000000114,
       0.00000000171,     0.00000000246,
       0.00000000114,    -0.00000000064,
       0.00000000077,    -0.00000000154,
      -0.00000000049,    -0.00000000030,
      -0.00000000154,    -0.00000000078,
      -0.00000000030,     0.00000000049,
      -0.00000000608,     0.00000000418,
       0.00000000155,     0.00000000277,
       0.00000000057,    -0.00000000028,
       0.00000000121,     0.00000000192,
       0.00000000100,    -0.00000000048,
      -0.00000000786,     0.00000001567,
       0.00000000503,     0.00000000308,
       0.00000000052,    -0.00000000076,
       0.00000001571,     0.00000000789,
       0.00000000310,    -0.00000000505,
      -0.00000000077,    -0.00000000052,
       0.00000000096,     0.00000000073,
       0.00000000063,    -0.00000000097,
       0.00000000061,    -0.00000000121,
      -0.00000000039,    -0.00000000024,
      -0.00000000084,    -0.00000000055,
      -0.00000000024,     0.00000000032,
      -0.00000000042,    -0.00000001004,
      -0.00000000245,    -0.00000000007,
      -0.00000001009,     0.00000000040,
      -0.00000000009,     0.00000000247,
      -0.00000000062,    -0.00000000010,
      -0.00000000003,     0.00000000059,
      -0.00000002131,     0.00000004231,
       0.00000001350,     0.00000000826,
       0.00000000140,    -0.00000000205,
       0.00000000865,     0.00000000530,
       0.00000000231,    -0.00000000306,
      -0.00000000046,    -0.00000000037,
       0.00000000418,     0.00000010369,
       0.00000002546,     0.00000000088,
       0.00000000021,    -0.00000000296,
       0.00000010395,    -0.00000000421,
       0.00000000091,    -0.00000002556,
      -0.00000000297,    -0.00000000022,
       0.00000000476,     0.00000000074,
       0.00000000024,    -0.00000000457,
      -0.00000000049,    -0.00000001189,
      -0.00000000291,    -0.00000000010,
      -0.00000000002,     0.00000000034,
       0.00000000541,    -0.00000000114,
      -0.00000000017,    -0.00000000111,
      -0.00000003386,    -0.00000005479,
      -0.00000000903,     0.00000000483,
       0.00000000036,     0.00000000071,
      -0.00000005529,     0.00000003386,
       0.00000000476,     0.00000000923,
       0.00000000074,    -0.00000000035,
      -0.00000000241,     0.00000000078,
       0.00000000114,     0.00000000199,
       0.00000001407,     0.00000034574,
       0.00000008473,     0.00000000287,
       0.00000000073,    -0.00000000985,
      -0.00000003034,     0.00000000814,
       0.00000000160,     0.00000000597,
       0.00000000079,    -0.00000000010,
       0.00000035280,     0.00000057543,
       0.00000009589,    -0.00000005027,
      -0.00000000371,    -0.00000000754,
       0.00000057638,    -0.00000035416,
      -0.00000005054,    -0.00000009623,
      -0.00000000758,     0.00000000374,
       0.00000001852,    -0.00000000647,
      -0.00000000868,    -0.00000001587,
      -0.00000014301,    -0.00000023249,
      -0.00000003872,     0.00000002030,
       0.00000000150,     0.00000000306,
       0.00000236133,    -0.00000147105,
      -0.00000020840,    -0.00000039296,
      -0.00000003116,     0.00000001534,
      -0.00000083790,    -0.00000046168,
      -0.00000003908,     0.00000006351,
       0.00000000254,     0.00000000147,
      -0.00000047227,     0.00000084300,
       0.00000006407,     0.00000004235,
       0.00000000181,    -0.00000000258,
      -0.00000001522,     0.00000001386,
      -0.00000000039,    -0.00000000040,
       0.00000001569,     0.00000000555,
      -0.00000000101,     0.00000000023,
       0.00000332593,     0.00000540750,
       0.00000090183,    -0.00000047228,
      -0.00000003477,    -0.00000007118,
      -0.00001803251,     0.00001123127,
       0.00000159028,     0.00000300061,
       0.00000023794,    -0.00000011710,
      -0.00000000570,    -0.00000001260,
       0.00000893490,     0.00000498403,
       0.00000043574,    -0.00000067881,
      -0.00000002715,    -0.00000001753,
       0.00000496790,    -0.00000896318,
      -0.00000068405,    -0.00000043282,
      -0.00000001713,     0.00000002765,
       0.00000011132,    -0.00000011679,
       0.00000000058,     0.00000000252,
      -0.00000000098,     0.00000000167,
      -0.00000012654,    -0.00000005491,
       0.00000000632,     0.00000000031,
       0.00000000164,     0.00000000073,
       0.00000028052,     0.00000015593,
       0.00000001366,    -0.00000002128,
      -0.00000000086,    -0.00000000055,
       0.00000081640,    -0.00000144852,
      -0.00000011014,    -0.00000007115,
      -0.00000000287,     0.00000000442,
       0.00000053846,    -0.00000000708,
      -0.00000000118,     0.00000000090,
       0.00000000210,    -0.00000055026,
      -0.00000000065,    -0.00000000081,
       0.00000000353,    -0.00000000454,
      -0.00000000050,     0.00000000013,
      -0.00000000151,     0.00000000486,
       0.00000000002,     0.00000000052,
      -0.00000489816,    -0.00000271942,
      -0.00000023875,     0.00000037212,
       0.00000001496,     0.00000000963,
      -0.00000607805,     0.00001078556,
       0.00000082113,     0.00000053100,
       0.00000002138,    -0.00000003289,
      -0.00000596347,     0.00000002043,
      -0.00000000083,    -0.00000000026,
       0.00000007156,     0.00000596132,
      -0.00000000134,    -0.00000000650,
      -0.00000000057,     0.00000000020,
      -0.00000001566,     0.00000004174,
       0.00000000583,    -0.00000000103,
      -0.00000000001,    -0.00000000063,
       0.00000002498,    -0.00000002910,
      -0.00000000076,    -0.00000000578,
      -0.00000000051,     0.00000000016,
       0.00000000001,    -0.00000000092,
       0.00000036238,    -0.00000000134,
      -0.00000000439,    -0.00000105932,
      -0.00000002291,     0.00000000332,
      -0.00000000089,    -0.00000000004,
       0.00000000897,     0.00000000052,
      -0.00000000158,     0.00000000005,
       0.00000000105,    -0.00000000481,
       0.00000000141,     0.00000000397,
      -0.00000423084,     0.00000003738,
       0.00000000132,    -0.00000000040,
      -0.00000000004,    -0.00000000019,
       0.00000003194,     0.00000738484,
      -0.00000000106,     0.00000000001,
       0.00000004138,    -0.00000001103,
      -0.00000000169,     0.00000000721,
       0.00000005228,    -0.00000011681,
       0.00000000095,    -0.00000000265,
       0.00000000648,     0.00000003683,
       0.00000000144,    -0.00000000290,
       0.00000000622,    -0.00000003054,
      -0.00000000112,    -0.00000000293,
      -0.00000000352,     0.00000000039,
      -0.00000000007,    -0.00000000021,
      -0.00000000550,    -0.00000000284,
       0.00000000040,    -0.00000000048,
      -0.00000002129,    -0.00000000051,
      -0.00000000039,    -0.00000002202,
      -0.00000000033,    -0.00000000114,
       0.00000000097,     0.00000000023,
      -0.00000003361,     0.00000023249,
      -0.00000001134,    -0.00000000775,
       0.00000000030,    -0.00000000042,
      -0.00000006303,     0.00000000147,
       0.00000000062,    -0.00000000161,
      -0.00000008128,     0.00000000186,
      -0.00000000074,    -0.00000000070,
      -0.00000000043,    -0.00000007072,
      -0.00000000082,     0.00000000125,
       0.00000000295,     0.00000000846,
       0.00000000069,    -0.00000000023,
      -0.00000000791,    -0.00000000228,
       0.00000000011,     0.00000000074,
      -0.00000000036,    -0.00000000035,
       0.00000000105,    -0.00000000087,
      -0.00000000087,    -0.00000000104,
      -0.00000000022,     0.00000000036,
       0.00000000069,     0.00000000043,
      -0.00000000067,    -0.00000000018,
       0.00000000197,    -0.00000000165,
      -0.00000000001,    -0.00000000056,
      -0.00000000315,    -0.00000000421,
      -0.00000000117,    -0.00000000004,
       0.00000001012,    -0.00000000582,
       0.00000000031,    -0.00000000188,
      -0.00000000586,    -0.00000001007,
      -0.00000000187,    -0.00000000030,
      -0.00000000059,     0.00000000028,
      -0.00000000055,    -0.00000000007,
      -0.00000000013,     0.00000000105,
       0.00000000029,     0.00000000027,
       0.00000000065,     0.00000000021,
       0.00000000039,    -0.00000000119,
       0.00000002288,    -0.00000001333,
       0.00000000069,    -0.00000000426,
      -0.00000000027,    -0.00000000019,
      -0.00000002451,    -0.00000004646,
      -0.00000000856,    -0.00000000168,
      -0.00000000041,     0.00000000053,
       0.00000008319,    -0.00000003019,
       0.00000000319,    -0.00000000905,
      -0.00000000041,    -0.00000000039,
      -0.00000003067,    -0.00000008275,
      -0.00000000901,    -0.00000000307,
      -0.00000000037,     0.00000000041,
      -0.00000000029,    -0.00000000069,
      -0.00000000055,     0.00000000020,
       0.00000023380,    -0.00000008666,
       0.00000000882,    -0.00000002553,
      -0.00000000116,    -0.00000000106,
      -0.00000015455,    -0.00000045583,
      -0.00000004964,    -0.00000001827,
      -0.00000000213,     0.00000000225,
       0.00000053520,    -0.00000009200,
       0.00000001450,    -0.00000002550,
      -0.00000000081,    -0.00000000093,
      -0.00000009726,    -0.00000053204,
      -0.00000002532,    -0.00000001367,
      -0.00000000087,     0.00000000083,
      -0.00000000074,    -0.00000000267,
      -0.00000000128,     0.00000000004,
      -0.00000000343,     0.00000000094,
       0.00000000094,     0.00000000342,
       0.00000197366,    -0.00000035872,
       0.00000005250,    -0.00000009441,
      -0.00000000306,    -0.00000000323,
      -0.00000062276,    -0.00000372125,
      -0.00000017945,    -0.00000010123,
      -0.00000000618,     0.00000000583,
       0.00000211630,     0.00000003694,
      -0.00000001006,    -0.00000210036,
       0.00000000079,     0.00000000431,
      -0.00000000206,    -0.00000000612,
       0.00000000553,    -0.00000000190,
       0.00000000026,     0.00000000033,
      -0.00000000035,     0.00000000029,
       0.00000000031,     0.00000000030,
      -0.00000002066,     0.00000000717,
       0.00000000073,     0.00000000486,
      -0.00000001423,     0.00000000475,
       0.00000000472,     0.00000001414,
      -0.00000000206,     0.00000000049,
      -0.00000000019,    -0.00000000050,
      -0.00000000048,     0.00000000020,
      -0.00000000034,     0.00000000034,
       0.00000000040,     0.00000000012,
      -0.00000000047,     0.00000000010,
       0.00000000012,     0.00000000045,
       0.00000000029,    -0.00000000023,
      -0.00000000041,    -0.00000000054,
       0.00000000110,    -0.00000000060,
      -0.00000000021,    -0.00000000038,
      -0.00000000060,    -0.00000000109,
      -0.00000000038,     0.00000000021,
      -0.00000000013,     0.00000000047,
      -0.00000000068,    -0.00000000009,
      -0.00000000008,     0.00000000068,
       0.00001120472,    -0.00000000788,
      -0.00000000376,     0.00000000049,
       0.00000000002,    -0.00000000091,
      -0.00000001742,    -0.00002056367,
       0.00000000084,     0.00000000686,
      -0.00000000166,    -0.00000000004,
      -0.00000010382,     0.00000020427,
      -0.00000000046,    -0.00000001190,
       0.00000000011,    -0.00000000053,
      -0.00000007471,     0.00000012430,
      -0.00000000350,     0.00000001284,
      -0.00000000006,    -0.00000000045,
      -0.00000000581,    -0.00000003512,
       0.00000000226,     0.00000000106,
       0.00000003153,    -0.00000000645,
      -0.00000000095,     0.00000000252,
      -0.00000008052,     0.00000006329,
      -0.00000000383,     0.00000002486,
      -0.00000001379,     0.00000006105,
      -0.00000000303,    -0.00000000229,
       0.00000006088,     0.00000001307,
      -0.00000000232,     0.00000000301,
      -0.00000000096,     0.00000000006,
       0.00000000102,     0.00000000167,
       0.00000000167,    -0.00000000094,
       0.00000000086,     0.00000000197,
       0.00000000031,    -0.00000000016,
       0.00000000238,    -0.00000000121,
      -0.00000000019,    -0.00000000037,
       0.00000000047,     0.00000000026,
       0.00000000034,    -0.00000000062,
       0.00000000099,     0.00000000055,
       0.00000000055,    -0.00000000097,
      -0.00000000019,     0.00000000035,
       0.00000000246,    -0.00000000137,
      -0.00000000044,    -0.00000000089,
      -0.00000000250,    -0.00000000483,
      -0.00000000176,     0.00000000081,
       0.00000000581,    -0.00000000191,
      -0.00000000068,    -0.00000000187,
      -0.00000000196,    -0.00000000577,
      -0.00000000185,     0.00000000072,
       0.00000000058,     0.00000000001,
       0.00000000002,    -0.00000000109,
      -0.00000000154,    -0.00000000044,
      -0.00000000011,     0.00000000017,
      -0.00000000088,     0.00000000287,
       0.00000000032,     0.00000000020,
      -0.00000000175,    -0.00000000011,
      -0.00000000040,     0.00000000175,
       0.00000000121,     0.00000000052,
       0.00000000098,    -0.00000000226,
       0.00000000073,     0.00000000003,
       0.00000000001,    -0.00000000073,
      -0.00000024472,    -0.00000007125,
      -0.00000000548,    -0.00000001099,
       0.00000000034,    -0.00000000032,
      -0.00000012991,     0.00000043104,
      -0.00000001935,     0.00000000931,
      -0.00000000055,    -0.00000000058,
      -0.00000013629,    -0.00000000615,
       0.00000000231,    -0.00000013792,
       0.00000000027,    -0.00000000032,
       0.00000000025,    -0.00000000190,
       0.00000000176,     0.00000000034,
      -0.00000027987,     0.00000122246,
      -0.00000006113,    -0.00000004640,
      -0.00000022065,    -0.00000005128,
       0.00000000840,    -0.00000001101,
       0.00000000023,    -0.00000000243,
      -0.00000000536,    -0.00000000379,
      -0.00000000356,     0.00000001413,
      -0.00000001202,     0.00000005234,
      -0.00000002721,     0.00000011835,
      -0.00000000010,    -0.00000000055,
       0.00000001606,    -0.00000000548,
      -0.00000000195,    -0.00000000521,
      -0.00000000990,    -0.00000003035,
      -0.00000000987,     0.00000000353,
       0.00000001797,    -0.00000000216,
      -0.00000000129,    -0.00000000514,
      -0.00000000261,    -0.00000001760,
      -0.00000000507,     0.00000000139,
       0.00000000035,    -0.00000000032,
      -0.00000000022,    -0.00000000033,
      -0.00000000032,     0.00000000023,
      -0.00000000623,    -0.00000000082,
      -0.00000000030,     0.00000000034,
      -0.00000000161,     0.00000001135,
       0.00000000063,     0.00000000054,
      -0.00000000093,    -0.00000000013,
      -0.00000000001,     0.00000000100,
       0.00000000171,     0.00000000026,
       0.00000000049,    -0.00000000312,
      -0.00000000053,     0.00000000011,
       0.00000000020,     0.00000000097,
       0.00000006323,    -0.00000000786,
      -0.00000000455,    -0.00000001841,
      -0.00000001453,    -0.00000011605,
      -0.00000003381,     0.00000000840,
      -0.00000000313,     0.00000000213,
       0.00000000086,     0.00000000053,
       0.00000000059,     0.00000000336,
       0.00000000071,    -0.00000000036,
      -0.00000001308,     0.00000000164,
       0.00000000195,     0.00000000242,
       0.00000000014,    -0.00000000020,
       0.00000000288,     0.00000002338,
       0.00000000435,    -0.00000000349,
      -0.00000000036,    -0.00000000026,
      -0.00000000078,    -0.00000001415,
      -0.00000000137,     0.00000000096,
      -0.00000001382,     0.00000000067,
       0.00000000089,     0.00000000130,
      -0.00000000532,    -0.00000009867,
      -0.00000000947,     0.00000000662,
       0.00000000032,     0.00000000040,
      -0.00000017598,     0.00000000925,
       0.00000001185,     0.00000001695,
       0.00000000069,    -0.00000000059,
       0.00000005216,    -0.00000000021,
       0.00000000048,    -0.00000004966,
      -0.00000000049,     0.00000000023,
      -0.00000000024,    -0.00000000046,
       0.00000000518,     0.00000000117,
       0.00000000117,    -0.00000000517,
       0.00000050713,    -0.00000000004,
       0.00000000051,     0.00000000005,
      -0.00000000010,    -0.00000090341,
       0.00000000009,    -0.00000000089,
       0.00000000064,     0.00000000807,
       0.00000000037,    -0.00000000026,
      -0.00000000538,     0.00000000224,
       0.00000000025,     0.00000000034,
      -0.00000000111,    -0.00000000259,
       0.00000000251,    -0.00000000115,
       0.00000015477,     0.00000002922,
      -0.00000001344,    -0.00000000971,
      -0.00000000108,     0.00000000888,
      -0.00000001801,     0.00000007850,
      -0.00000000393,    -0.00000000292,
       0.00000007854,     0.00000001803,
      -0.00000000292,     0.00000000393,
       0.00000000075,    -0.00000001395,
       0.00000000132,     0.00000000093,
      -0.00000002463,    -0.00000000136,
       0.00000000164,    -0.00000000233,
      -0.00000000318,    -0.00000000006,
       0.00000000002,    -0.00000000327,
      -0.00000068245,     0.00000297114,
      -0.00000014889,    -0.00000011213,
       0.00000000467,    -0.00000000590,
      -0.00000030303,    -0.00000006962,
       0.00000001147,    -0.00000001511,
       0.00000000060,     0.00000000048,
       0.00000000835,     0.00000000221,
      -0.00000000087,     0.00000000065,
       0.00000000145,    -0.00000000441,
       0.00000000082,     0.00000000057,
       0.00000000139,     0.00000000103,
       0.00000000103,    -0.00000000139,
      -0.00000000649,     0.00000002821,
      -0.00000000150,    -0.00000000016,
      -0.00000000026,     0.00000000265,
      -0.00000000954,     0.00000001885,
       0.00000000254,    -0.00000000736,
      -0.00000000736,    -0.00000000254,
       0.00000000053,     0.00000000011,
       0.00000000020,    -0.00000000097,
       0.00000000026,    -0.00000000159,
      -0.00000000159,    -0.00000000026,
       0.00000000931,     0.00000000000,
       0.00000000000,    -0.00000001638,
       0.00000014091,    -0.00000037837,
       0.00000001137,     0.00000001723,
       0.00000000718,     0.00000000283,
      -0.00000002511,     0.00000010926,
      -0.00000000544,    -0.00000000414,
       0.00000010915,     0.00000002507,
      -0.00000000413,     0.00000000544,
       0.00000001111,    -0.00000006532,
       0.00000000315,     0.00000000051,
      -0.00000192802,     0.00000839283,
      -0.00000042147,    -0.00000031891,
       0.00000001323,    -0.00000001658,
      -0.00000043204,    -0.00000009924,
       0.00000001640,    -0.00000002155,
       0.00000000086,     0.00000000068,
       0.00000000006,     0.00000000529,
       0.00000001016,     0.00000000167,
      -0.00000000022,     0.00000000135,
       0.00000000135,     0.00000000022,
      -0.00000000047,     0.00000000155,
       0.00000000154,     0.00000000046,
      -0.00000001498,     0.00000008804,
      -0.00000000271,    -0.00000000044,
      -0.00000000027,     0.00000000255,
      -0.00000000255,    -0.00000000026,
       0.00000000039,    -0.00000000131,
      -0.00000000131,    -0.00000000039,
       0.00000002292,    -0.00000007898,
       0.00000000256,     0.00000000077,
       0.00000000106,     0.00000000155,
       0.00000000155,    -0.00000000106,
       0.00000000286,     0.00000000000,
       0.00000000000,    -0.00000000502,
      -0.00000001748,     0.00000006028,
      -0.00000000306,    -0.00000000092,
       0.00000010346,     0.00000012411,
      -0.00000000153,     0.00000000112,
      -0.00000003137,     0.00000013658,
      -0.00000000681,    -0.00000000518,
       0.00000013658,     0.00000003138,
      -0.00000000518,     0.00000000681,
      -0.00000377292,     0.00001642508,
      -0.00000082507,    -0.00000062393,
       0.00000002581,    -0.00000003247,
      -0.00000054180,    -0.00000012446,
       0.00000002056,    -0.00000002703,
       0.00000000107,     0.00000000085,
       0.00000000201,    -0.00000000040,
      -0.00000000142,    -0.00000000129,
       0.00000000125,     0.00000000176,
       0.00000000176,    -0.00000000125,
       0.00000000124,     0.00000000175,
      -0.00000000175,     0.00000000124,
       0.00000000017,    -0.00000000074,
       0.00000000074,     0.00000000017,
       0.00000000560,    -0.00000001015,
       0.00000000057,     0.00000000033,
       0.00000002154,    -0.00000001883,
       0.00000000105,     0.00000000078,
       0.00000001853,     0.00000002173,
      -0.00000000078,     0.00000000106,
       0.00000000155,    -0.00000000280,
      -0.00000000004,    -0.00000000007,
       0.00000000342,     0.00000000145,
       0.00000000008,    -0.00000000003,
      -0.00000000308,     0.00000000138,
       0.00000001184,    -0.00000000613,
      -0.00000000090,    -0.00000000170,
       0.00000000593,     0.00000001162,
       0.00000000165,    -0.00000000084,
      -0.00000000386,     0.00000000064,
      -0.00000000071,    -0.00000000394,
      -0.00000000002,    -0.00000000082,
       0.00000000182,    -0.00000000066,
      -0.00000001474,     0.00000000702,
       0.00000000092,     0.00000000176,
      -0.00000000698,    -0.00000001468,
      -0.00000000175,     0.00000000091,
       0.00000000284,    -0.00000000051,
       0.00000000055,     0.00000000290,
       0.00000000054,     0.00000000061,
       0.00000000002,     0.00000000050,
      -0.00000000067,    -0.00000000186,
       0.00000000056,     0.00000000004,
       0.00000000190,    -0.00000000063,
      -0.00000000001,     0.00000000061,
       0.00000000154,     0.00000000066,
      -0.00000000070,     0.00000000149,
      -0.00000000293,    -0.00000000142,
       0.00000000142,    -0.00000000293,
      -0.00000000397,    -0.00000000193,
       0.00000000193,    -0.00000000397,
      -0.00000000564,    -0.00000000274,
       0.00000000059,    -0.00000000049,
       0.00000000274,    -0.00000000564,
       0.00000000049,     0.00000000059,
      -0.00000000707,    -0.00000000343,
       0.00000000074,    -0.00000000061,
       0.00000000343,    -0.00000000707,
       0.00000000061,     0.00000000074,
      -0.00000000015,     0.00000000027,
      -0.00000000118,     0.00000000012,
       0.00000000000,     0.00000000062,
       0.00000000013,     0.00000000118,
       0.00000000062,    -0.00000000001,
       0.00000000053,    -0.00000000036,
      -0.00000000037,    -0.00000000053,
      -0.00000000299,     0.00000000030,
       0.00000000001,     0.00000000155,
       0.00000000037,     0.00000000002,
       0.00000000005,     0.00000000080,
       0.00000000046,     0.00000000003,
      -0.00000000549,     0.00000000377,
       0.00000000138,     0.00000000244,
       0.00000000050,    -0.00000000025,
       0.00000000379,     0.00000000548,
       0.00000000244,    -0.00000000139,
      -0.00000000025,    -0.00000000050,
       0.00000000018,     0.00000000040,
       0.00000000038,    -0.00000000020,
       0.00000000049,    -0.00000000034,
      -0.00000000012,    -0.00000000022,
       0.00000000162,    -0.00000000323,
      -0.00000000102,    -0.00000000061,
      -0.00000000324,    -0.00000000162,
      -0.00000000061,     0.00000000102,
      -0.00000001516,     0.00000001041,
       0.00000000381,     0.00000000669,
       0.00000000137,    -0.00000000068,
       0.00000000157,     0.00000000280,
       0.00000000140,    -0.00000000061,
      -0.00000001697,     0.00000003376,
       0.00000001066,     0.00000000645,
       0.00000000109,    -0.00000000162,
       0.00000003384,     0.00000001683,
       0.00000000639,    -0.00000001069,
      -0.00000000163,    -0.00000000108,
       0.00000000145,     0.00000000111,
       0.00000000093,    -0.00000000148,
       0.00000000208,    -0.00000000416,
      -0.00000000131,    -0.00000000079,
      -0.00000000013,     0.00000000020,
       0.00000000221,     0.00000000077,
       0.00000000026,    -0.00000000061,
      -0.00000000096,    -0.00000002219,
      -0.00000000536,    -0.00000000015,
      -0.00000000004,     0.00000000063,
      -0.00000002222,     0.00000000103,
      -0.00000000014,     0.00000000538,
       0.00000000063,     0.00000000004,
      -0.00000000101,    -0.00000000019,
      -0.00000000001,     0.00000000094,
      -0.00000005644,     0.00000011291,
       0.00000003558,     0.00000002132,
       0.00000000362,    -0.00000000541,
      -0.00000001177,    -0.00000000340,
      -0.00000000098,     0.00000000321,
       0.00000000050,     0.00000000020,
       0.00000000966,     0.00000023479,
       0.00000005713,     0.00000000182,
       0.00000000049,    -0.00000000667,
       0.00000023454,    -0.00000001107,
       0.00000000128,    -0.00000005705,
      -0.00000000665,    -0.00000000039,
       0.00000000783,     0.00000000127,
       0.00000000018,    -0.00000000068,
       0.00000000019,    -0.00000000741,
      -0.00000000061,     0.00000000008,
      -0.00000000326,    -0.00000007290,
      -0.00000001768,    -0.00000000048,
      -0.00000000014,     0.00000000207,
       0.00000046649,    -0.00000002479,
       0.00000000224,    -0.00000011261,
      -0.00000001320,    -0.00000000082,
      -0.00000013397,    -0.00000021758,
      -0.00000003596,     0.00000001900,
       0.00000000140,     0.00000000286,
      -0.00000021693,     0.00000013532,
       0.00000001927,     0.00000003588,
       0.00000000284,    -0.00000000144,
      -0.00000000748,     0.00000000213,
       0.00000000374,     0.00000000556,
       0.00000007605,     0.00000169596,
       0.00000041201,     0.00000001140,
       0.00000000328,    -0.00000004808,
      -0.00000351952,     0.00000018716,
      -0.00000001656,     0.00000085094,
       0.00000009955,     0.00000000616,
       0.00000000080,    -0.00000000800,
       0.00000143244,     0.00000233868,
       0.00000038937,    -0.00000020235,
      -0.00000001479,    -0.00000003103,
       0.00000231966,    -0.00000145058,
      -0.00000020821,    -0.00000038350,
      -0.00000003010,     0.00000001572,
       0.00000005742,    -0.00000001951,
      -0.00000000067,    -0.00000000200,
      -0.00000000050,     0.00000000031,
      -0.00000002853,    -0.00000004630,
      -0.00000000041,     0.00000000214,
       0.00000000042,     0.00000000035,
       0.00000007762,     0.00000012533,
       0.00000002075,    -0.00000001100,
      -0.00000000081,    -0.00000000165,
       0.00000072857,    -0.00000044471,
      -0.00000006318,    -0.00000012084,
      -0.00000000957,     0.00000000468,
       0.00000029218,     0.00000016430,
       0.00000001446,    -0.00000002203,
      -0.00000000088,    -0.00000000061,
       0.00000015987,    -0.00000029364,
      -0.00000002230,    -0.00000001360,
      -0.00000000050,     0.00000000091,
       0.00000000485,    -0.00000000385,
      -0.00000000457,    -0.00000000071,
      -0.00000150508,    -0.00000243003,
      -0.00000040312,     0.00000021333,
       0.00000001578,     0.00000003201,
      -0.00000555601,     0.00000339233,
       0.00000048236,     0.00000092299,
       0.00000007298,    -0.00000003568,
      -0.00000318842,    -0.00000180041,
      -0.00000016107,     0.00000024137,
       0.00000000963,     0.00000000696,
      -0.00000173656,     0.00000319826,
       0.00000024408,     0.00000014587,
       0.00000000517,    -0.00000000996,
      -0.00000003415,     0.00000003499,
       0.00000000135,     0.00000000025,
       0.00000000037,    -0.00000000063,
       0.00000003854,     0.00000001183,
      -0.00000000140,    -0.00000000166,
      -0.00000000061,    -0.00000000023,
      -0.00000000384,     0.00000000297,
       0.00000000079,     0.00000000149,
       0.00000002600,    -0.00000001940,
      -0.00000000518,    -0.00000001025,
       0.00000000067,     0.00000000031,
       0.00000000030,    -0.00000000067,
       0.00000014764,     0.00000008195,
       0.00000000715,    -0.00000001119,
      -0.00000000045,    -0.00000000029,
       0.00000025087,    -0.00000044576,
      -0.00000003377,    -0.00000002176,
      -0.00000000087,     0.00000000136,
       0.00000018414,     0.00000000389,
       0.00000000054,    -0.00000000020,
      -0.00000000520,    -0.00000018055,
       0.00000000035,     0.00000000081,
       0.00000000162,    -0.00000000186,
      -0.00000000043,     0.00000000218,
      -0.00000228921,    -0.00000127057,
      -0.00000011125,     0.00000017371,
       0.00000000699,     0.00000000446,
      -0.00000185803,     0.00000330148,
       0.00000025083,     0.00000016153,
       0.00000000645,    -0.00000001007,
      -0.00000206033,    -0.00000003499,
      -0.00000000546,     0.00000000255,
       0.00000000015,     0.00000000046,
       0.00000005278,     0.00000201949,
      -0.00000000358,    -0.00000000859,
      -0.00000000066,     0.00000000033,
      -0.00000000645,     0.00000001779,
       0.00000000282,    -0.00000000036,
       0.00000001030,    -0.00000001257,
      -0.00000000048,    -0.00000000263,
       0.00000019349,     0.00000000001,
      -0.00000000105,    -0.00000044533,
       0.00000000000,     0.00000001423,
       0.00000000008,    -0.00000000172,
      -0.00000000055,     0.00000002288,
       0.00000000007,     0.00000000111,
       0.00000000061,    -0.00000000247,
       0.00000000087,     0.00000000199,
      -0.00000225048,    -0.00000000010,
       0.00000000062,    -0.00000000043,
       0.00000000806,     0.00000314113,
      -0.00000000088,    -0.00000000055,
       0.00000003205,    -0.00000011133,
       0.00000000107,     0.00000001392,
      -0.00000000012,     0.00000000043,
      -0.00000001479,    -0.00000018538,
       0.00000000202,    -0.00000000891,
       0.00000000011,     0.00000000062,
       0.00000000311,     0.00000001894,
       0.00000000099,    -0.00000000132,
       0.00000000247,    -0.00000001528,
      -0.00000000069,    -0.00000000119,
      -0.00000000906,     0.00000000675,
      -0.00000000048,    -0.00000000070,
       0.00000000863,     0.00000001268,
      -0.00000000092,     0.00000000054,
       0.00000002994,    -0.00000000055,
      -0.00000000020,     0.00000002943,
      -0.00000000019,    -0.00000000075,
       0.00000000069,     0.00000000004,
       0.00000006985,    -0.00000005255,
       0.00000000376,     0.00000000540,
      -0.00000000025,     0.00000000020,
      -0.00000006120,    -0.00000009057,
       0.00000000638,    -0.00000000397,
       0.00000000029,     0.00000000037,
      -0.00000012153,     0.00000000372,
      -0.00000000112,    -0.00000000049,
       0.00000000412,    -0.00000011470,
      -0.00000000058,     0.00000000133,
       0.00000000109,     0.00000000599,
       0.00000000076,    -0.00000000015,
      -0.00000000567,    -0.00000000001,
       0.00000000014,     0.00000000076,
      -0.00000000079,    -0.00000000061,
      -0.00000000195,     0.00000000086,
      -0.00000000086,    -0.00000000198,
      -0.00000000075,    -0.00000000088,
       0.00000000089,    -0.00000000071,
      -0.00000000449,    -0.00000000074,
      -0.00000000009,    -0.00000000048,
       0.00000000697,     0.00000000273,
      -0.00000000073,     0.00000000284,
       0.00000001040,    -0.00000000692,
       0.00000000052,     0.00000000094,
       0.00000000730,     0.00000001044,
      -0.00000000086,     0.00000000086,
       0.00000000576,     0.00000000673,
       0.00000000067,    -0.00000000058,
      -0.00000000677,     0.00000000541,
       0.00000000061,     0.00000000066,
       0.00000000032,    -0.00000000027,
      -0.00000000026,    -0.00000000031,
       0.00000000080,    -0.00000000068,
      -0.00000000106,    -0.00000000137,
       0.00000000269,    -0.00000000158,
      -0.00000000153,    -0.00000000264,
       0.00000000799,    -0.00000000468,
       0.00000000024,    -0.00000000149,
      -0.00000000009,    -0.00000000007,
      -0.00000000717,    -0.00000001320,
      -0.00000000243,    -0.00000000045,
       0.00000001909,    -0.00000000707,
       0.00000000071,    -0.00000000212,
      -0.00000000668,    -0.00000001843,
      -0.00000000195,    -0.00000000069,
       0.00000006730,    -0.00000002484,
       0.00000000253,    -0.00000000731,
      -0.00000000033,    -0.00000000031,
      -0.00000003750,    -0.00000010856,
      -0.00000001176,    -0.00000000426,
      -0.00000000051,     0.00000000054,
       0.00000010315,    -0.00000001879,
       0.00000000276,    -0.00000000535,
      -0.00000001558,    -0.00000009585,
      -0.00000000405,    -0.00000000243,
      -0.00000000016,    -0.00000000047,
       0.00000044229,    -0.00000007737,
       0.00000001169,    -0.00000002093,
      -0.00000000068,    -0.00000000075,
      -0.00000011612,    -0.00000069884,
      -0.00000003339,    -0.00000001877,
      -0.00000000119,     0.00000000109,
       0.00000034285,    -0.00000000482,
       0.00000000099,    -0.00000000389,
       0.00000001638,    -0.00000026948,
       0.00000000404,     0.00000000222,
      -0.00000000045,    -0.00000000134,
       0.00000000124,    -0.00000000045,
       0.00000000015,     0.00000000092,
      -0.00000000214,    -0.00000000003,
      -0.00000000004,     0.00000000211,
       0.00000178050,     0.00000002908,
      -0.00000000206,     0.00000000038,
       0.00000000003,    -0.00000000028,
       0.00000003135,    -0.00000277278,
       0.00000000031,     0.00000000281,
       0.00000028257,    -0.00000002534,
       0.00000000699,    -0.00000001705,
      -0.00000000041,    -0.00000000060,
       0.00000006458,     0.00000033643,
       0.00000001317,     0.00000001045,
       0.00000000045,    -0.00000000056,
      -0.00000000087,    -0.00000000559,
       0.00000000035,     0.00000000017,
       0.00000000508,    -0.00000000096,
      -0.00000000016,     0.00000000039,
       0.00000000027,     0.00000000029,
      -0.00000001258,    -0.00000000021,
      -0.00000000004,     0.00000000345,
      -0.00000000216,    -0.00000000064,
       0.00000000089,    -0.00000000049,
      -0.00000000017,    -0.00000000031,
      -0.00000000075,    -0.00000000143,
      -0.00000000050,     0.00000000026,
       0.00000000140,    -0.00000000047,
      -0.00000000017,    -0.00000000045,
      -0.00000000041,    -0.00000000130,
      -0.00000000041,     0.00000000015,
      -0.00000000056,    -0.00000000007,
      -0.00000000012,     0.00000000088,
      -0.00000000053,    -0.00000000005,
       0.00000000008,    -0.00000000044,
      -0.00000012077,     0.00000015362,
      -0.00000001063,    -0.00000001041,
       0.00000000041,    -0.00000000045,
       0.00000016235,     0.00000016805,
      -0.00000001381,     0.00000001148,
      -0.00000000053,    -0.00000000047,
       0.00000178132,    -0.00000000321,
       0.00000000056,     0.00000178133,
      -0.00000000003,    -0.00000000075,
       0.00000000039,    -0.00000000054,
       0.00000000053,     0.00000000039,
      -0.00000000143,    -0.00000000050,
      -0.00000000049,     0.00000000063,
       0.00000000058,     0.00000000030,
       0.00000000003,     0.00000000072,
       0.00000000072,    -0.00000000004,
       0.00000000046,    -0.00000000045,
       0.00000000029,     0.00000000048,
       0.00000000059,    -0.00000000036,
       0.00000000480,    -0.00000000159,
      -0.00000000057,    -0.00000000155,
      -0.00000000244,    -0.00000000759,
      -0.00000000246,     0.00000000088,
       0.00000000379,    -0.00000000057,
      -0.00000000031,    -0.00000000110,
      -0.00000000022,    -0.00000000301,
      -0.00000000088,     0.00000000018,
      -0.00000000158,    -0.00000000023,
      -0.00000000030,     0.00000000248,
       0.00000000048,     0.00000000002,
       0.00000000002,    -0.00000000075,
       0.00000000184,    -0.00000000373,
       0.00000000039,     0.00000000031,
      -0.00000000409,    -0.00000000196,
       0.00000000044,    -0.00000000039,
      -0.00000004060,    -0.00000001250,
      -0.00000000089,    -0.00000000181,
       0.00000001251,    -0.00000004062,
       0.00000000181,    -0.00000000088,
       0.00000000095,    -0.00000000502,
      -0.00000000037,     0.00000000015,
       0.00000000502,     0.00000000095,
      -0.00000000015,    -0.00000000037,
      -0.00000002092,    -0.00000001014,
       0.00000000188,    -0.00000000386,
      -0.00000000043,    -0.00000000042,
       0.00000001509,    -0.00000000182,
      -0.00000000111,    -0.00000000431,
      -0.00000000298,    -0.00000002344,
      -0.00000000673,     0.00000000173,
       0.00000000226,    -0.00000000042,
      -0.00000000019,    -0.00000000077,
       0.00000000111,     0.00000000280,
       0.00000000087,    -0.00000000036,
      -0.00000000024,    -0.00000000035,
      -0.00000000054,     0.00000000037,
      -0.00000000078,    -0.00000000009,
      -0.00000000010,     0.00000000121,
      -0.00000000093,    -0.00000000018,
       0.00000000019,    -0.00000000091,
      -0.00000000290,     0.00000000173,
       0.00000000067,     0.00000000051,
       0.00000000222,     0.00000000436,
       0.00000000080,    -0.00000000092,
       0.00000001001,    -0.00000000279,
      -0.00000000079,    -0.00000000287,
      -0.00000000013,     0.00000001014,
       0.00000000302,    -0.00000000058,
      -0.00000000061,    -0.00000001225,
      -0.00000000119,     0.00000000083,
      -0.00000001853,     0.00000000093,
       0.00000000124,     0.00000000178,
       0.00000000212,     0.00000000032,
       0.00000000029,     0.00000000037,
      -0.00000000025,    -0.00000000619,
      -0.00000000038,     0.00000000029,
       0.00000004565,     0.00000000028,
       0.00000000031,    -0.00000006877,
      -0.00000000089,    -0.00000001530,
      -0.00000000150,     0.00000000105,
       0.00000001589,    -0.00000000063,
      -0.00000000105,    -0.00000000150,
      -0.00000000174,     0.00000000570,
      -0.00000000018,    -0.00000000014,
       0.00000000524,     0.00000000186,
       0.00000008297,    -0.00000000002,
       0.00000000000,     0.00000008300,
      -0.00000000137,    -0.00000000065,
      -0.00000000065,     0.00000000132,
       0.00000000013,    -0.00000000234,
       0.00000000234,     0.00000000013,
      -0.00000005125,    -0.00000002487,
       0.00000000533,    -0.00000000445,
       0.00000000256,    -0.00000000527,
       0.00000000046,     0.00000000055,
      -0.00000000003,    -0.00000000061,
       0.00000000154,     0.00000000000,
       0.00000000000,     0.00000000154,
      -0.00000000189,    -0.00000000092,
      -0.00000000091,     0.00000000189,
      -0.00000014515,    -0.00000007040,
       0.00000001513,    -0.00000001260,
       0.00000000364,    -0.00000000751,
       0.00000000065,     0.00000000078,
      -0.00000028474,    -0.00000013818,
       0.00000002964,    -0.00000002468,
       0.00000000457,    -0.00000000942,
       0.00000000082,     0.00000000098,
      -0.00000000136,    -0.00000000017,
       0.00000000002,    -0.00000000008,
      -0.00000000297,    -0.00000000163,
       0.00000000006,    -0.00000000017,
       0.00000000166,    -0.00000000295,
       0.00000000017,     0.00000000006,
      -0.00000000044,    -0.00000000005,
       0.00000000002,    -0.00000000050,
      -0.00000000024,    -0.00000000028,
       0.00000000040,     0.00000000087,
      -0.00000000086,     0.00000000038,
      -0.00000000023,    -0.00000000040,
       0.00000000041,    -0.00000000024,
      -0.00000000015,    -0.00000000078,
       0.00000000078,    -0.00000000015,
      -0.00000000042,    -0.00000000018,
      -0.00000000018,     0.00000000042,
      -0.00000000108,    -0.00000000046,
      -0.00000000033,     0.00000000062,
       0.00000000017,     0.00000000011,
      -0.00000000246,     0.00000000024,
       0.00000000001,     0.00000000126,
       0.00000000025,     0.00000000246,
       0.00000000125,    -0.00000000002,
       0.00000000107,    -0.00000000073,
      -0.00000000027,    -0.00000000046,
      -0.00000000074,    -0.00000000106,
      -0.00000000046,     0.00000000027,
      -0.00000000687,     0.00000000068,
       0.00000000005,     0.00000000347,
       0.00000000083,     0.00000000004,
       0.00000000001,     0.00000000109,
       0.00000000062,     0.00000000006,
      -0.00000001124,     0.00000000768,
       0.00000000279,     0.00000000490,
       0.00000000100,    -0.00000000050,
       0.00000000775,     0.00000001117,
       0.00000000486,    -0.00000000283,
      -0.00000000051,    -0.00000000099,
       0.00000000027,     0.00000000062,
       0.00000000057,    -0.00000000031,
       0.00000000142,    -0.00000000098,
      -0.00000000035,    -0.00000000061,
      -0.00000000013,     0.00000000006,
       0.00000000053,     0.00000000060,
       0.00000000344,    -0.00000000686,
      -0.00000000214,    -0.00000000128,
      -0.00000000022,     0.00000000033,
      -0.00000000688,    -0.00000000339,
      -0.00000000126,     0.00000000215,
       0.00000000033,     0.00000000021,
      -0.00000000030,    -0.00000000025,
      -0.00000003710,     0.00000002554,
       0.00000000928,     0.00000001607,
       0.00000000328,    -0.00000000166,
      -0.00000000270,    -0.00000000263,
      -0.00000000090,     0.00000000092,
      -0.00000003669,     0.00000007292,
       0.00000002278,     0.00000001374,
       0.00000000234,    -0.00000000348,
       0.00000007306,     0.00000003589,
       0.00000001336,    -0.00000002286,
      -0.00000000350,    -0.00000000225,
       0.00000000240,     0.00000000186,
       0.00000000035,    -0.00000000037,
       0.00000000147,    -0.00000000245,
      -0.00000000040,    -0.00000000022,
       0.00000000946,    -0.00000001905,
      -0.00000000594,    -0.00000000352,
      -0.00000000060,     0.00000000091,
       0.00000008662,     0.00000004182,
       0.00000001552,    -0.00000002679,
      -0.00000000410,    -0.00000000265,
      -0.00000000310,    -0.00000007440,
      -0.00000001789,    -0.00000000053,
      -0.00000000016,     0.00000000211,
      -0.00000007415,     0.00000000393,
      -0.00000000024,     0.00000001782,
       0.00000000209,     0.00000000010,
      -0.00000000269,    -0.00000000057,
       0.00000000009,     0.00000000244,
      -0.00000021996,     0.00000044298,
       0.00000013861,     0.00000008227,
       0.00000001394,    -0.00000002110,
      -0.00000064376,    -0.00000031064,
      -0.00000011557,     0.00000020005,
       0.00000003049,     0.00000001971,
       0.00000003244,     0.00000079985,
       0.00000019359,     0.00000000641,
       0.00000000180,    -0.00000002272,
       0.00000079552,    -0.00000004270,
       0.00000000219,    -0.00000019207,
      -0.00000002244,    -0.00000000095,
       0.00000002112,     0.00000000357,
       0.00000000051,    -0.00000000212,
      -0.00000000017,    -0.00000001962,
      -0.00000000183,     0.00000000039,
       0.00000000330,     0.00000006937,
       0.00000001673,     0.00000000038,
       0.00000000013,    -0.00000000196,
       0.00000046719,    -0.00000001938,
       0.00000000313,    -0.00000011296,
      -0.00000001323,    -0.00000000090,
       0.00000009290,     0.00000015257,
       0.00000002523,    -0.00000001298,
      -0.00000000094,    -0.00000000205,
       0.00000014969,    -0.00000009523,
      -0.00000001368,    -0.00000002441,
      -0.00000000191,     0.00000000105,
       0.00000000443,    -0.00000000108,
      -0.00000000234,    -0.00000000288,
      -0.00000006728,    -0.00000141255,
      -0.00000034093,    -0.00000000786,
      -0.00000000255,     0.00000003993,
      -0.00000359641,     0.00000015035,
      -0.00000002477,     0.00000087020,
       0.00000010177,     0.00000000691,
       0.00000000090,    -0.00000000800,
      -0.00000101037,    -0.00000166226,
      -0.00000027680,     0.00000014157,
       0.00000001022,     0.00000002239,
      -0.00000162822,     0.00000103431,
       0.00000014978,     0.00000026619,
       0.00000002075,    -0.00000001156,
      -0.00000003399,     0.00000001122,
       0.00000000074,     0.00000000189,
       0.00000000040,    -0.00000000024,
       0.00000001793,     0.00000002551,
       0.00000000063,    -0.00000000185,
      -0.00000000033,    -0.00000000027,
       0.00000003870,     0.00000006226,
       0.00000001026,    -0.00000000549,
      -0.00000000041,    -0.00000000082,
       0.00000019638,    -0.00000011974,
      -0.00000001696,    -0.00000003236,
      -0.00000000257,     0.00000000127,
       0.00000010587,     0.00000006145,
       0.00000000567,    -0.00000000795,
      -0.00000000031,    -0.00000000027,
       0.00000005592,    -0.00000010640,
      -0.00000000811,    -0.00000000441,
       0.00000000174,    -0.00000000120,
      -0.00000000149,     0.00000000008,
      -0.00000067305,    -0.00000108489,
      -0.00000017943,     0.00000009564,
       0.00000000711,     0.00000001429,
      -0.00000149411,     0.00000090908,
       0.00000012960,     0.00000024702,
       0.00000001959,    -0.00000000964,
      -0.00000117272,    -0.00000067281,
      -0.00000006220,     0.00000008867,
       0.00000000351,     0.00000000292,
      -0.00000062017,     0.00000117572,
       0.00000008997,     0.00000004965,
       0.00000000144,    -0.00000000370,
      -0.00000001183,     0.00000001185,
       0.00000000096,     0.00000000044,
       0.00000001320,     0.00000000247,
      -0.00000000034,    -0.00000000107,
       0.00000009225,     0.00000005122,
       0.00000000445,    -0.00000000697,
      -0.00000000028,    -0.00000000018,
       0.00000012141,    -0.00000021620,
      -0.00000001636,    -0.00000001048,
      -0.00000000042,     0.00000000066,
       0.00000008428,     0.00000000571,
       0.00000000081,    -0.00000000038,
      -0.00000000582,    -0.00000007838,
       0.00000000048,     0.00000000101,
       0.00000000096,    -0.00000000102,
      -0.00000000015,     0.00000000127,
      -0.00000133883,    -0.00000073853,
      -0.00000006486,     0.00000010171,
       0.00000000409,     0.00000000257,
      -0.00000089232,     0.00000159793,
       0.00000012147,     0.00000007762,
       0.00000000307,    -0.00000000487,
      -0.00000094673,    -0.00000004187,
      -0.00000000593,     0.00000000300,
       0.00000000018,     0.00000000052,
       0.00000004590,     0.00000090063,
      -0.00000000380,    -0.00000000827,
      -0.00000000062,     0.00000000034,
      -0.00000000363,     0.00000001000,
       0.00000000172,    -0.00000000013,
       0.00000000551,    -0.00000000717,
      -0.00000000036,    -0.00000000149,
       0.00000010618,     0.00000000039,
      -0.00000000008,    -0.00000021104,
       0.00000000317,     0.00000001314,
       0.00000000020,    -0.00000000161,
      -0.00000000172,     0.00000002145,
       0.00000000028,     0.00000000104,
       0.00000000037,    -0.00000000139,
       0.00000000056,     0.00000000108,
      -0.00000123422,    -0.00000000386,
       0.00000000071,    -0.00000000041,
       0.00000000107,     0.00000149860,
      -0.00000000072,    -0.00000000075,
       0.00000001948,    -0.00000009989,
       0.00000000070,     0.00000001250,
      -0.00000001388,    -0.00000016561,
       0.00000000122,    -0.00000000804,
       0.00000000009,     0.00000000055,
       0.00000000146,     0.00000001067,
       0.00000000071,    -0.00000000062,
       0.00000000071,    -0.00000000830,
      -0.00000000045,    -0.00000000044,
      -0.00000000950,     0.00000000689,
      -0.00000000050,    -0.00000000073,
       0.00000000979,     0.00000001440,
      -0.00000000106,     0.00000000066,
       0.00000002184,    -0.00000000041,
      -0.00000000046,     0.00000002137,
      -0.00000000011,    -0.00000000051,
       0.00000000048,    -0.00000000003,
       0.00000008294,    -0.00000005998,
       0.00000000434,     0.00000000634,
      -0.00000000029,     0.00000000023,
      -0.00000007110,    -0.00000010493,
       0.00000000763,    -0.00000000492,
       0.00000000031,     0.00000000040,
      -0.00000009496,     0.00000000431,
      -0.00000000112,    -0.00000000039,
       0.00000000469,    -0.00000009029,
      -0.00000000045,     0.00000000126,
       0.00000000051,     0.00000000413,
       0.00000000065,    -0.00000000010,
      -0.00000000413,     0.00000000054,
       0.00000000012,     0.00000000065,
       0.00000000014,    -0.00000000040,
      -0.00000000062,    -0.00000000036,
      -0.00000000201,     0.00000000156,
      -0.00000000154,    -0.00000000203,
      -0.00000000046,    -0.00000000054,
       0.00000000054,    -0.00000000044,
      -0.00000000079,     0.00000000234,
      -0.00000000075,    -0.00000000020,
       0.00000000473,     0.00000000276,
      -0.00000000059,     0.00000000193,
       0.00000000786,    -0.00000000590,
       0.00000000046,     0.00000000069,
       0.00000000580,     0.00000000803,
      -0.00000000068,     0.00000000059,
       0.00000000352,     0.00000000414,
       0.00000000042,    -0.00000000036,
      -0.00000000416,     0.00000000335,
       0.00000000038,     0.00000000041,
       0.00000000032,    -0.00000000021,
       0.00000000021,     0.00000000035,
       0.00000000073,     0.00000000036,
      -0.00000000037,     0.00000000071,
       0.00000000029,    -0.00000000025,
      -0.00000000035,    -0.00000000044,
       0.00000000072,    -0.00000000042,
      -0.00000000040,    -0.00000000069,
       0.00000000250,    -0.00000000146,
       0.00000000007,    -0.00000000046,
      -0.00000000203,    -0.00000000369,
      -0.00000000068,    -0.00000000012,
       0.00000000447,    -0.00000000167,
       0.00000000013,    -0.00000000042,
      -0.00000000146,    -0.00000000417,
       0.00000001779,    -0.00000000650,
       0.00000000067,    -0.00000000192,
      -0.00000000009,    -0.00000000008,
      -0.00000000897,    -0.00000002579,
      -0.00000000279,    -0.00000000100,
       0.00000002083,    -0.00000000390,
       0.00000000056,    -0.00000000117,
      -0.00000000246,    -0.00000001790,
      -0.00000000064,    -0.00000000044,
       0.00000009533,    -0.00000001592,
       0.00000000250,    -0.00000000446,
      -0.00000000015,    -0.00000000017,
      -0.00000002210,    -0.00000013642,
      -0.00000000645,    -0.00000000362,
       0.00000006060,    -0.00000000199,
       0.00000000034,    -0.00000000129,
       0.00000000565,    -0.00000003626,
       0.00000000133,     0.00000000067,
       0.00000029554,     0.00000001012,
      -0.00000000066,     0.00000000011,
       0.00000001082,    -0.00000041926,
       0.00000000009,     0.00000000083,
       0.00000007369,    -0.00000000703,
       0.00000000180,    -0.00000000426,
       0.00000001662,     0.00000008567,
       0.00000000337,     0.00000000259,
      -0.00000000013,    -0.00000000094,
       0.00000000086,    -0.00000000015,
      -0.00000000020,    -0.00000000037,
      -0.00000003069,     0.00000003670,
      -0.00000000255,    -0.00000000260,
       0.00000000011,    -0.00000000011,
       0.00000003842,     0.00000003974,
      -0.00000000324,     0.00000000274,
       0.00000032224,     0.00000000658,
      -0.00000000757,     0.00000032228,
       0.00000000098,     0.00000000100,
       0.00000000040,    -0.00000000037,
      -0.00000000791,    -0.00000000850,
       0.00000000072,    -0.00000000139,
       0.00000000031,    -0.00000000055,
      -0.00000000055,    -0.00000000029,
       0.00000000129,    -0.00000000041,
      -0.00000000015,    -0.00000000041,
      -0.00000000058,    -0.00000000185,
      -0.00000000059,     0.00000000021,
       0.00000000076,    -0.00000000013,
      -0.00000000003,     0.00000000067,
      -0.00000000155,    -0.00000000301,
       0.00000000048,    -0.00000000009,
      -0.00000000281,     0.00000000079,
      -0.00000002471,     0.00000003894,
      -0.00000000248,    -0.00000000231,
      -0.00000003892,    -0.00000002474,
       0.00000000233,    -0.00000000248,
       0.00000000017,    -0.00000000085,
       0.00000000085,     0.00000000017,
       0.00000000067,     0.00000000097,
       0.00000000040,    -0.00000000027,
       0.00000000034,     0.00000000028,
       0.00000000028,    -0.00000000035,
       0.00000000029,    -0.00000000053,
       0.00000000033,     0.00000000067,
       0.00000000059,    -0.00000000025,
       0.00000000332,    -0.00000000032,
      -0.00000000022,    -0.00000000097,
      -0.00000000050,    -0.00000000469,
      -0.00000000136,     0.00000000032,
       0.00000000072,    -0.00000000015,
       0.00000000033,     0.00000000086,
       0.00000000048,    -0.00000000092,
       0.00000000093,     0.00000000048,
      -0.00000000029,    -0.00000000038,
       0.00000000032,    -0.00000000022,
      -0.00000000023,    -0.00000000031,
       0.00000000055,     0.00000000022,
       0.00000000022,    -0.00000000057,
       0.00000000097,     0.00000000036,
       0.00000000029,    -0.00000000108,
       0.00000000039,     0.00000000103,
       0.00000000109,    -0.00000000039,
       0.00000000081,     0.00000000201,
       0.00000000192,    -0.00000000074,
      -0.00000000028,     0.00000000049,
       0.00000000054,     0.00000000035,
       0.00000000270,    -0.00000000036,
      -0.00000000019,    -0.00000000079,
       0.00000000035,     0.00000000270,
       0.00000000079,    -0.00000000019,
      -0.00000000006,    -0.00000000145,
      -0.00000000200,     0.00000000008,
      -0.00000000025,    -0.00000000084,
       0.00000000424,     0.00000000005,
       0.00000000006,    -0.00000000580,
      -0.00000000011,    -0.00000000218,
       0.00000000227,    -0.00000000008,
      -0.00000000024,     0.00000000075,
       0.00000000069,     0.00000000025,
       0.00000000848,     0.00000000007,
      -0.00000000007,     0.00000000848,
      -0.00000000044,     0.00000000156,
      -0.00000000156,    -0.00000000044,
       0.00000000005,     0.00000000016,
      -0.00000000004,     0.00000000042,
      -0.00000000042,    -0.00000000005,
       0.00000000002,     0.00000000006,
      -0.00000000006,     0.00000000002,
      -0.00000000026,     0.00000000017,
      -0.00000000084,    -0.00000000036,
      -0.00000000026,     0.00000000047,
      -0.00000000035,     0.00000000084,
       0.00000000048,     0.00000000025,
       0.00000000046,    -0.00000000004,
      -0.00000000005,    -0.00000000046,
      -0.00000000234,    -0.00000000100,
      -0.00000000071,     0.00000000132,
       0.00000000036,     0.00000000022,
      -0.00000000018,     0.00000000034,
      -0.00000000486,     0.00000000046,
       0.00000000003,     0.00000000244,
       0.00000000058,     0.00000000003,
       0.00000000051,     0.00000000484,
       0.00000000242,    -0.00000000006,
       0.00000000002,    -0.00000000058,
       0.00000000062,    -0.00000000006,
      -0.00000000001,    -0.00000000031,
       0.00000000220,    -0.00000000150,
      -0.00000000054,    -0.00000000094,
      -0.00000000152,    -0.00000000218,
      -0.00000000093,     0.00000000055,
      -0.00000001586,     0.00000000157,
       0.00000000014,     0.00000000791,
       0.00000000189,     0.00000000008,
      -0.00000000034,    -0.00000000097,
      -0.00000000038,     0.00000000015,
      -0.00000002345,     0.00000001598,
       0.00000000577,     0.00000001009,
       0.00000000206,    -0.00000000103,
       0.00000001619,     0.00000002316,
       0.00000000993,    -0.00000000590,
      -0.00000000107,    -0.00000000202,
       0.00000000045,     0.00000000104,
       0.00000000093,    -0.00000000053,
       0.00000000536,    -0.00000000370,
      -0.00000000134,    -0.00000000229,
      -0.00000000047,     0.00000000024,
       0.00000001285,     0.00000001805,
       0.00000000766,    -0.00000000460,
      -0.00000000083,    -0.00000000156,
       0.00000001027,    -0.00000002037,
      -0.00000000631,    -0.00000000379,
      -0.00000000065,     0.00000000096,
      -0.00000002042,    -0.00000000992,
      -0.00000000363,     0.00000000634,
       0.00000000097,     0.00000000061,
      -0.00000000073,    -0.00000000062,
      -0.00000000041,     0.00000000076,
      -0.00000012461,     0.00000008608,
       0.00000003120,     0.00000005344,
       0.00000001089,    -0.00000000559,
      -0.00000009400,    -0.00000013195,
      -0.00000005596,     0.00000003385,
       0.00000000608,     0.00000001143,
      -0.00000011057,     0.00000021925,
       0.00000006820,     0.00000004108,
       0.00000000704,    -0.00000001040,
       0.00000021954,     0.00000010645,
       0.00000003917,    -0.00000006847,
      -0.00000001048,    -0.00000000657,
       0.00000000586,     0.00000000461,
       0.00000000089,    -0.00000000096,
       0.00000000341,    -0.00000000601,
      -0.00000000104,    -0.00000000051,
      -0.00000001518,     0.00000003071,
       0.00000000953,     0.00000000560,
       0.00000000095,    -0.00000000146,
       0.00000025350,     0.00000012688,
       0.00000004681,    -0.00000007882,
      -0.00000001205,    -0.00000000796,
       0.00000000355,     0.00000009422,
       0.00000002260,     0.00000000081,
       0.00000000023,    -0.00000000267,
       0.00000009328,    -0.00000000564,
       0.00000000000,    -0.00000002228,
      -0.00000000261,    -0.00000000007,
       0.00000000287,     0.00000000069,
      -0.00000000024,    -0.00000000250,
       0.00000031743,    -0.00000064239,
      -0.00000019985,    -0.00000011750,
      -0.00000001995,     0.00000003048,
      -0.00000195554,    -0.00000097812,
      -0.00000036239,     0.00000060972,
       0.00000009299,     0.00000006141,
       0.00000000680,    -0.00000000940,
      -0.00000003927,    -0.00000102446,
      -0.00000024709,    -0.00000000897,
      -0.00000000253,     0.00000002913,
      -0.00000101354,     0.00000006070,
      -0.00000000010,     0.00000024318,
       0.00000002842,     0.00000000073,
      -0.00000002270,    -0.00000000400,
      -0.00000000058,     0.00000000263,
       0.00000000098,     0.00000002065,
       0.00000000221,    -0.00000000063,
      -0.00000000034,    -0.00000000165,
       0.00000000031,    -0.00000000026,
      -0.00000000005,    -0.00000000054,
       0.00000000043,     0.00000000361,
       0.00000000165,    -0.00000000004,
      -0.00000000037,     0.00000000024,
       0.00000000025,     0.00000000036,
       0.00000000014,    -0.00000000062,
       0.00000000152,     0.00000003109,
       0.00000000744,     0.00000000015,
       0.00000000005,    -0.00000000088,
       0.00000010168,    -0.00000000398,
       0.00000000069,    -0.00000002443,
      -0.00000000287,    -0.00000000019,
       0.00000003228,     0.00000005408,
       0.00000000900,    -0.00000000444,
      -0.00000000031,    -0.00000000074,
       0.00000005186,    -0.00000003377,
      -0.00000000493,    -0.00000000833,
      -0.00000000064,     0.00000000039,
       0.00000000140,    -0.00000000028,
      -0.00000000078,    -0.00000000077,
      -0.00000002835,    -0.00000057815,
      -0.00000013896,    -0.00000000275,
      -0.00000000098,     0.00000001631,
      -0.00000078555,     0.00000003107,
      -0.00000000533,     0.00000018921,
       0.00000002219,     0.00000000148,
      -0.00000035717,    -0.00000059406,
      -0.00000009914,     0.00000004954,
       0.00000000351,     0.00000000813,
      -0.00000057288,     0.00000037047,
       0.00000005418,     0.00000009256,
       0.00000000712,    -0.00000000428,
      -0.00000001071,     0.00000000342,
       0.00000000034,     0.00000000084,
       0.00000000598,     0.00000000740,
       0.00000000032,    -0.00000000078,
       0.00000002547,     0.00000004092,
       0.00000000671,    -0.00000000360,
      -0.00000000027,    -0.00000000054,
       0.00000009854,    -0.00000006016,
      -0.00000000853,    -0.00000001618,
      -0.00000000129,     0.00000000064,
       0.00000005416,     0.00000003294,
       0.00000000323,    -0.00000000405,
       0.00000002721,    -0.00000005439,
      -0.00000000416,    -0.00000000189,
       0.00000000093,    -0.00000000055,
      -0.00000000072,     0.00000000020,
      -0.00000041454,    -0.00000066604,
      -0.00000010969,     0.00000005877,
       0.00000000438,     0.00000000874,
      -0.00000074897,     0.00000045744,
       0.00000006511,     0.00000012322,
       0.00000000979,    -0.00000000486,
      -0.00000060726,    -0.00000035981,
      -0.00000003430,     0.00000004581,
       0.00000000179,     0.00000000176,
      -0.00000031316,     0.00000060839,
       0.00000004665,     0.00000002314,
       0.00000000044,    -0.00000000193,
      -0.00000000611,     0.00000000597,
       0.00000000072,     0.00000000040,
       0.00000000670,     0.00000000051,
      -0.00000000013,    -0.00000000077,
       0.00000005946,     0.00000003312,
       0.00000000285,    -0.00000000449,
      -0.00000000018,    -0.00000000011,
       0.00000006713,    -0.00000011954,
      -0.00000000904,    -0.00000000576,
      -0.00000000023,     0.00000000036,
       0.00000004265,     0.00000000578,
       0.00000000082,    -0.00000000040,
      -0.00000000542,    -0.00000003649,
       0.00000000048,     0.00000000097,
       0.00000000061,    -0.00000000062,
      -0.00000000004,     0.00000000079,
      -0.00000082976,    -0.00000046212,
      -0.00000003992,     0.00000006287,
       0.00000000253,     0.00000000156,
      -0.00000049690,     0.00000088469,
       0.00000006711,     0.00000004263,
       0.00000000166,    -0.00000000268,
      -0.00000047647,    -0.00000003967,
      -0.00000000538,     0.00000000279,
       0.00000000017,     0.00000000048,
       0.00000003729,     0.00000043343,
      -0.00000000342,    -0.00000000712,
      -0.00000000052,     0.00000000030,
      -0.00000000228,     0.00000000612,
       0.00000000112,    -0.00000000002,
       0.00000000315,    -0.00000000445,
      -0.00000000029,    -0.00000000088,
       0.00000005860,     0.00000000056,
       0.00000000037,    -0.00000010543,
       0.00000000365,     0.00000001043,
       0.00000000020,    -0.00000000129,
      -0.00000000182,     0.00000001715,
       0.00000000030,     0.00000000083,
       0.00000000023,    -0.00000000082,
       0.00000000037,     0.00000000061,
      -0.00000068168,    -0.00000000650,
       0.00000000068,    -0.00000000038,
      -0.00000000281,     0.00000075190,
      -0.00000000058,    -0.00000000071,
       0.00000001156,    -0.00000007924,
       0.00000000043,     0.00000000978,
      -0.00000000999,    -0.00000012966,
       0.00000000074,    -0.00000000627,
       0.00000000007,     0.00000000043,
       0.00000000061,     0.00000000630,
       0.00000000052,    -0.00000000028,
      -0.00000000009,    -0.00000000468,
      -0.00000000817,     0.00000000583,
      -0.00000000042,    -0.00000000062,
       0.00000000830,     0.00000001223,
      -0.00000000091,     0.00000000057,
       0.00000001425,    -0.00000000040,
      -0.00000000056,     0.00000001380,
       0.00000007597,    -0.00000005423,
       0.00000000391,     0.00000000579,
      -0.00000000027,     0.00000000021,
      -0.00000006087,    -0.00000008979,
       0.00000000657,    -0.00000000425,
       0.00000000027,     0.00000000034,
      -0.00000006467,     0.00000000435,
      -0.00000000106,    -0.00000000038,
       0.00000000532,    -0.00000006014,
      -0.00000000041,     0.00000000117,
       0.00000000022,     0.00000000283,
       0.00000000050,    -0.00000000007,
      -0.00000000293,     0.00000000063,
       0.00000000009,     0.00000000051,
       0.00000000016,    -0.00000000052,
      -0.00000000068,    -0.00000000031,
      -0.00000000183,     0.00000000142,
      -0.00000000137,    -0.00000000185,
      -0.00000000028,    -0.00000000033,
       0.00000000033,    -0.00000000027,
      -0.00000000111,     0.00000000361,
      -0.00000000092,    -0.00000000025,
       0.00000000515,     0.00000000239,
      -0.00000000049,     0.00000000162,
       0.00000000691,    -0.00000000537,
       0.00000000044,     0.00000000058,
       0.00000000507,     0.00000000708,
      -0.00000000060,     0.00000000045,
       0.00000000213,     0.00000000252,
      -0.00000000254,     0.00000000204,
       0.00000000023,     0.00000000026,
       0.00000000072,    -0.00000000042,
      -0.00000000055,    -0.00000000099,
       0.00000000104,    -0.00000000039,
      -0.00000000031,    -0.00000000093,
       0.00000000444,    -0.00000000159,
       0.00000000017,    -0.00000000048,
      -0.00000000208,    -0.00000000602,
      -0.00000000065,    -0.00000000023,
       0.00000000431,    -0.00000000083,
      -0.00000000033,    -0.00000000332,
       0.00000002004,    -0.00000000312,
       0.00000000052,    -0.00000000092,
      -0.00000000414,    -0.00000002691,
      -0.00000000126,    -0.00000000070,
       0.00000001145,    -0.00000000061,
       0.00000000163,    -0.00000000451,
       0.00000005029,     0.00000000296,
       0.00000000313,    -0.00000006715,
       0.00000001721,    -0.00000000161,
       0.00000000042,    -0.00000000097,
       0.00000000376,     0.00000001979,
       0.00000000077,     0.00000000059,
      -0.00000000711,     0.00000000813,
      -0.00000000057,    -0.00000000060,
       0.00000000845,     0.00000000877,
      -0.00000000066,     0.00000000057,
       0.00000005746,     0.00000000232,
      -0.00000000262,     0.00000005747,
      -0.00000000039,    -0.00000000096,
      -0.00000000091,     0.00000000020,
      -0.00000000655,     0.00000000926,
      -0.00000000058,    -0.00000000056,
      -0.00000000925,    -0.00000000656,
       0.00000000056,    -0.00000000058,
      -0.00000000055,    -0.00000000083,
       0.00000000083,    -0.00000000055,
       0.00000000001,    -0.00000000053,
      -0.00000000012,    -0.00000000040,
       0.00000000071,    -0.00000000007,
      -0.00000000010,    -0.00000000086,
       0.00000000059,    -0.00000000007,
       0.00000000007,     0.00000000059,
       0.00000000085,     0.00000000001,
      -0.00000000001,     0.00000000085,
       0.00000000005,    -0.00000000002,
       0.00000000001,     0.00000000005,
      -0.00000000057,    -0.00000000073,
      -0.00000000053,     0.00000000034,
       0.00000000010,     0.00000000018,
      -0.00000000160,    -0.00000000070,
      -0.00000000048,     0.00000000090,
      -0.00000000068,     0.00000000160,
       0.00000000090,     0.00000000047,
       0.00000000092,    -0.00000000008,
      -0.00000000010,    -0.00000000092,
      -0.00000000516,    -0.00000000222,
      -0.00000000153,     0.00000000288,
       0.00000000078,     0.00000000048,
      -0.00000000985,     0.00000000091,
       0.00000000006,     0.00000000488,
       0.00000000117,     0.00000000006,
       0.00000000104,     0.00000000978,
       0.00000000484,    -0.00000000015,
       0.00000000003,    -0.00000000116,
       0.00000000205,    -0.00000000020,
      -0.00000000002,    -0.00000000101,
      -0.00000000024,    -0.00000000001,
       0.00000000064,     0.00000000544,
       0.00000000265,    -0.00000000010,
       0.00000000001,    -0.00000000063,
       0.00000000597,    -0.00000000405,
      -0.00000000146,    -0.00000000254,
      -0.00000000052,     0.00000000026,
      -0.00000000413,    -0.00000000586,
      -0.00000000248,     0.00000000150,
       0.00000000027,     0.00000000050,
      -0.00000004767,     0.00000000474,
       0.00000000055,     0.00000002351,
       0.00000000562,     0.00000000021,
      -0.00000000460,    -0.00000003901,
      -0.00000001904,     0.00000000080,
      -0.00000000008,     0.00000000454,
      -0.00000006433,     0.00000004371,
       0.00000001570,     0.00000002745,
       0.00000000562,    -0.00000000280,
       0.00000004448,     0.00000006309,
       0.00000002677,    -0.00000001619,
      -0.00000000295,    -0.00000000543,
       0.00000000101,     0.00000000238,
       0.00000000062,    -0.00000000021,
       0.00000000208,    -0.00000000123,
      -0.00000000031,    -0.00000000050,
      -0.00000001509,     0.00000001046,
       0.00000000377,     0.00000000642,
       0.00000000131,    -0.00000000068,
       0.00000011611,     0.00000016894,
       0.00000007195,    -0.00000004199,
      -0.00000000754,    -0.00000001466,
      -0.00000002397,     0.00000004715,
       0.00000001454,     0.00000000881,
       0.00000000152,    -0.00000000222,
       0.00000004723,     0.00000002257,
       0.00000000817,    -0.00000001461,
      -0.00000000224,    -0.00000000137,
       0.00000000143,     0.00000000126,
       0.00000000073,    -0.00000000150,
       0.00000032172,    -0.00000022288,
      -0.00000008064,    -0.00000013715,
      -0.00000002789,     0.00000001446,
      -0.00000089512,    -0.00000130224,
      -0.00000055609,     0.00000032425,
       0.00000005816,     0.00000011310,
       0.00000001530,    -0.00000000690,
       0.00000025945,    -0.00000051196,
      -0.00000015869,    -0.00000009601,
      -0.00000001654,     0.00000002421,
      -0.00000051231,    -0.00000024523,
      -0.00000008924,     0.00000015929,
       0.00000002443,     0.00000001489,
      -0.00000001161,    -0.00000000925,
      -0.00000000186,     0.00000000205,
       0.00000000029,     0.00000000023,
      -0.00000000636,     0.00000001195,
       0.00000000222,     0.00000000094,
       0.00000000065,    -0.00000000124,
      -0.00000000079,    -0.00000000039,
      -0.00000000417,     0.00000001048,
       0.00000000589,     0.00000000296,
       0.00000000056,    -0.00000000004,
      -0.00000000005,    -0.00000000055,
       0.00000000069,    -0.00000000089,
      -0.00000000038,    -0.00000000028,
       0.00000000142,    -0.00000000327,
      -0.00000000580,     0.00000001176,
       0.00000000363,     0.00000000212,
       0.00000000036,    -0.00000000056,
       0.00000004018,     0.00000002030,
       0.00000000743,    -0.00000001246,
      -0.00000000191,    -0.00000000126,
       0.00000000093,     0.00000002966,
       0.00000000711,     0.00000000032,
       0.00000000009,    -0.00000000084,
       0.00000002910,    -0.00000000201,
      -0.00000000011,    -0.00000000691,
      -0.00000000081,     0.00000000000,
       0.00000000079,     0.00000000021,
      -0.00000000011,    -0.00000000066,
       0.00000011235,    -0.00000022799,
      -0.00000007057,    -0.00000004114,
      -0.00000000700,     0.00000001078,
      -0.00000031338,    -0.00000015822,
      -0.00000005796,     0.00000009734,
       0.00000001487,     0.00000000982,
      -0.00000001122,    -0.00000032529,
      -0.00000007828,    -0.00000000324,
      -0.00000000090,     0.00000000926,
      -0.00000031981,     0.00000002110,
       0.00000000088,     0.00000007630,
       0.00000000891,     0.00000000007,
      -0.00000000631,    -0.00000000116,
      -0.00000000017,     0.00000000083,
       0.00000000051,     0.00000000560,
       0.00000000068,    -0.00000000024,
       0.00000000103,     0.00000002073,
       0.00000000495,     0.00000000009,
       0.00000000003,    -0.00000000058,
       0.00000005069,    -0.00000000198,
       0.00000000033,    -0.00000001214,
      -0.00000000143,    -0.00000000009,
       0.00000001734,     0.00000002989,
       0.00000000502,    -0.00000000234,
      -0.00000000016,    -0.00000000042,
       0.00000002780,    -0.00000001862,
      -0.00000000277,    -0.00000000439,
      -0.00000000033,     0.00000000023,
       0.00000000071,    -0.00000000011,
      -0.00000000042,    -0.00000000032,
      -0.00000001797,    -0.00000036143,
      -0.00000008650,    -0.00000000149,
      -0.00000000059,     0.00000001018,
      -0.00000039177,     0.00000001544,
      -0.00000000246,     0.00000009407,
       0.00000001104,     0.00000000071,
      -0.00000019471,    -0.00000032946,
      -0.00000005523,     0.00000002670,
       0.00000000184,     0.00000000459,
      -0.00000031174,     0.00000020520,
       0.00000003039,     0.00000004974,
       0.00000000375,    -0.00000000246,
      -0.00000000546,     0.00000000169,
       0.00000000023,     0.00000000055,
       0.00000000321,     0.00000000344,
       0.00000000021,    -0.00000000050,
       0.00000001769,     0.00000002846,
       0.00000000466,    -0.00000000251,
      -0.00000000019,    -0.00000000037,
       0.00000005834,    -0.00000003566,
      -0.00000000506,    -0.00000000956,
      -0.00000000076,     0.00000000038,
       0.00000003053,     0.00000001974,
       0.00000000207,    -0.00000000227,
       0.00000001430,    -0.00000003060,
      -0.00000000235,    -0.00000000079,
       0.00000000056,    -0.00000000029,
      -0.00000000040,     0.00000000019,
      -0.00000027475,    -0.00000044192,
      -0.00000007246,     0.00000003897,
       0.00000000291,     0.00000000578,
      -0.00000044357,     0.00000027111,
       0.00000003862,     0.00000007272,
       0.00000000577,    -0.00000000289,
      -0.00000034533,    -0.00000021235,
      -0.00000002101,     0.00000002591,
       0.00000000100,     0.00000000118,
      -0.00000017176,     0.00000034557,
       0.00000002650,     0.00000001134,
       0.00000000004,    -0.00000000110,
      -0.00000000361,     0.00000000344,
       0.00000000053,     0.00000000033,
       0.00000000386,    -0.00000000010,
      -0.00000000007,    -0.00000000055,
       0.00000003840,     0.00000002152,
       0.00000000184,    -0.00000000290,
      -0.00000000012,    -0.00000000007,
       0.00000003917,    -0.00000006958,
      -0.00000000526,    -0.00000000333,
       0.00000002260,     0.00000000513,
       0.00000000073,    -0.00000000037,
      -0.00000000462,    -0.00000001706,
       0.00000000043,     0.00000000084,
       0.00000000040,    -0.00000000039,
       0.00000000000,     0.00000000051,
      -0.00000052102,    -0.00000029191,
      -0.00000002500,     0.00000003947,
       0.00000000158,     0.00000000096,
      -0.00000029003,     0.00000051487,
       0.00000003904,     0.00000002466,
       0.00000000094,    -0.00000000155,
      -0.00000024859,    -0.00000003386,
      -0.00000000450,     0.00000000235,
       0.00000000014,     0.00000000041,
       0.00000002925,     0.00000021241,
      -0.00000000285,    -0.00000000576,
      -0.00000000042,     0.00000000025,
      -0.00000000152,     0.00000000388,
       0.00000000074,     0.00000000003,
       0.00000000183,    -0.00000000287,
      -0.00000000023,    -0.00000000052,
       0.00000003217,     0.00000000059,
       0.00000000053,    -0.00000005385,
       0.00000000317,     0.00000000769,
       0.00000000017,    -0.00000000096,
      -0.00000000155,     0.00000001269,
       0.00000000025,     0.00000000061,
       0.00000000014,    -0.00000000050,
       0.00000000025,     0.00000000035,
      -0.00000037425,    -0.00000000682,
       0.00000000057,    -0.00000000033,
      -0.00000000402,     0.00000038456,
      -0.00000000046,    -0.00000000058,
       0.00000000710,    -0.00000005812,
       0.00000000025,     0.00000000715,
      -0.00000000710,    -0.00000009475,
       0.00000000045,    -0.00000000455,
       0.00000000020,     0.00000000383,
       0.00000000038,    -0.00000000012,
      -0.00000000040,    -0.00000000269,
      -0.00000000638,     0.00000000450,
      -0.00000000032,    -0.00000000048,
       0.00000000630,     0.00000000930,
      -0.00000000069,     0.00000000043,
       0.00000000881,    -0.00000000036,
      -0.00000000058,     0.00000000839,
       0.00000006186,    -0.00000004361,
       0.00000000311,     0.00000000470,
      -0.00000000022,     0.00000000017,
      -0.00000004635,    -0.00000006852,
       0.00000000503,    -0.00000000322,
      -0.00000004122,     0.00000000421,
      -0.00000000095,    -0.00000000035,
       0.00000000535,    -0.00000003700,
      -0.00000000037,     0.00000000105,
       0.00000000009,     0.00000000192,
      -0.00000000204,     0.00000000056,
       0.00000000018,    -0.00000000057,
      -0.00000000072,    -0.00000000030,
      -0.00000000153,     0.00000000117,
      -0.00000000112,    -0.00000000155,
      -0.00000000137,     0.00000000442,
      -0.00000000098,    -0.00000000028,
       0.00000000545,     0.00000000226,
      -0.00000000044,     0.00000000140,
       0.00000000589,    -0.00000000464,
       0.00000000040,     0.00000000048,
       0.00000000420,     0.00000000606,
      -0.00000000051,     0.00000000032,
       0.00000000128,     0.00000000152,
      -0.00000000153,     0.00000000122,
       0.00000000107,    -0.00000000037,
      -0.00000000047,    -0.00000000138,
       0.00000000090,    -0.00000000018,
      -0.00000000003,    -0.00000000060,
       0.00000000414,    -0.00000000058,
       0.00000000011,    -0.00000000019,
      -0.00000000075,    -0.00000000532,
       0.00000000225,    -0.00000000016,
       0.00000000042,    -0.00000000044,
       0.00000000867,     0.00000000078,
       0.00000000082,    -0.00000001110,
       0.00000000380,    -0.00000000032,
       0.00000000078,     0.00000000435,
      -0.00000000157,     0.00000000172,
       0.00000000178,     0.00000000188,
       0.00000001022,     0.00000000068,
      -0.00000000076,     0.00000001022,
      -0.00000000156,     0.00000000205,
      -0.00000000204,    -0.00000000156,
      -0.00000000038,    -0.00000000050,
      -0.00000000035,     0.00000000023,
      -0.00000000049,     0.00000000039,
       0.00000000023,     0.00000000035,
      -0.00000000122,    -0.00000000156,
      -0.00000000110,     0.00000000072,
       0.00000000021,     0.00000000037,
      -0.00000000316,    -0.00000000139,
      -0.00000000095,     0.00000000176,
       0.00000000048,     0.00000000030,
      -0.00000000133,     0.00000000316,
       0.00000000176,     0.00000000090,
       0.00000000028,    -0.00000000048,
       0.00000000061,     0.00000000026,
       0.00000000018,    -0.00000000034,
      -0.00000000009,    -0.00000000006,
      -0.00000000053,     0.00000000131,
       0.00000000072,     0.00000000036,
       0.00000000233,    -0.00000000021,
      -0.00000000001,    -0.00000000114,
      -0.00000000025,    -0.00000000231,
      -0.00000000113,     0.00000000004,
      -0.00000001423,    -0.00000000611,
      -0.00000000417,     0.00000000792,
       0.00000000213,     0.00000000130,
       0.00000000373,    -0.00000000924,
      -0.00000000506,    -0.00000000250,
      -0.00000000078,     0.00000000136,
      -0.00000002507,     0.00000000228,
       0.00000000015,     0.00000001233,
       0.00000000295,     0.00000000014,
       0.00000000272,     0.00000002482,
       0.00000001219,    -0.00000000045,
       0.00000000004,    -0.00000000291,
      -0.00000000009,     0.00000000089,
       0.00000000084,    -0.00000000002,
      -0.00000001182,     0.00000000119,
       0.00000000014,     0.00000000581,
       0.00000000139,     0.00000000005,
       0.00000002120,     0.00000021575,
       0.00000010625,    -0.00000000240,
       0.00000000088,    -0.00000002534,
      -0.00000000400,    -0.00000000040,
      -0.00000002930,     0.00000001976,
       0.00000000706,     0.00000001240,
       0.00000000254,    -0.00000000125,
       0.00000002028,     0.00000002846,
       0.00000001195,    -0.00000000737,
      -0.00000000134,    -0.00000000242,
       0.00000000049,     0.00000000132,
       0.00000000104,    -0.00000000070,
       0.00000025517,    -0.00000002565,
      -0.00000000309,    -0.00000012558,
      -0.00000002994,    -0.00000000098,
      -0.00000016311,    -0.00000165816,
      -0.00000081682,     0.00000001828,
      -0.00000000678,     0.00000019490,
       0.00000003100,     0.00000000290,
       0.00000000060,    -0.00000000370,
      -0.00000000034,    -0.00000000007,
       0.00000031700,    -0.00000021458,
      -0.00000007710,    -0.00000013484,
      -0.00000002761,     0.00000001369,
      -0.00000021950,    -0.00000030848,
      -0.00000013029,     0.00000008015,
       0.00000001463,     0.00000002634,
      -0.00000000426,    -0.00000001014,
      -0.00000000271,     0.00000000092,
       0.00000000015,     0.00000000041,
      -0.00000000860,     0.00000000537,
       0.00000000141,     0.00000000213,
       0.00000000030,    -0.00000000025,
      -0.00000001693,     0.00000000834,
       0.00000000493,     0.00000000989,
       0.00000000046,     0.00000000094,
       0.00000000054,    -0.00000000027,
       0.00000012754,    -0.00000006218,
      -0.00000003754,    -0.00000007434,
       0.00000000189,     0.00000000092,
       0.00000000045,    -0.00000000084,
       0.00000000087,    -0.00000000190,
      -0.00000000085,    -0.00000000042,
      -0.00000000454,     0.00000000315,
       0.00000000113,     0.00000000191,
       0.00000000039,    -0.00000000020,
       0.00000001129,     0.00000001661,
       0.00000000703,    -0.00000000408,
      -0.00000000073,    -0.00000000143,
      -0.00000000616,     0.00000001196,
       0.00000000368,     0.00000000226,
       0.00000000039,    -0.00000000056,
       0.00000001197,     0.00000000562,
       0.00000000201,    -0.00000000370,
      -0.00000000057,    -0.00000000033,
       0.00000000031,     0.00000000029,
       0.00000009045,    -0.00000006275,
      -0.00000002265,    -0.00000003822,
      -0.00000000777,     0.00000000406,
      -0.00000008851,    -0.00000013019,
      -0.00000005521,     0.00000003203,
       0.00000000574,     0.00000001122,
       0.00000006701,    -0.00000013118,
      -0.00000004051,    -0.00000002471,
      -0.00000000427,     0.00000000617,
      -0.00000013117,    -0.00000006201,
      -0.00000002228,     0.00000004066,
       0.00000000624,     0.00000000369,
      -0.00000000260,    -0.00000000210,
      -0.00000000044,     0.00000000049,
      -0.00000000132,     0.00000000269,
       0.00000000054,     0.00000000019,
      -0.00000000384,     0.00000000780,
       0.00000000240,     0.00000000140,
       0.00000000024,    -0.00000000037,
       0.00000001939,     0.00000000982,
       0.00000000357,    -0.00000000600,
      -0.00000000092,    -0.00000000060,
       0.00000000038,     0.00000001668,
       0.00000000400,     0.00000000023,
       0.00000000006,    -0.00000000048,
       0.00000001618,    -0.00000000127,
      -0.00000000013,    -0.00000000382,
       0.00000000040,     0.00000000012,
       0.00000007016,    -0.00000014261,
      -0.00000004402,    -0.00000002557,
      -0.00000000435,     0.00000000673,
      -0.00000015167,    -0.00000007675,
      -0.00000002792,     0.00000004700,
       0.00000000718,     0.00000000473,
      -0.00000000535,    -0.00000018409,
      -0.00000004425,    -0.00000000216,
      -0.00000000058,     0.00000000525,
      -0.00000017965,     0.00000001296,
       0.00000000101,     0.00000004265,
       0.00000000496,    -0.00000000006,
      -0.00000000322,    -0.00000000062,
      -0.00000000009,     0.00000000048,
       0.00000000039,     0.00000000278,
       0.00000000039,    -0.00000000016,
       0.00000000075,     0.00000001506,
       0.00000000359,     0.00000000006,
       0.00000000002,    -0.00000000042,
       0.00000003112,    -0.00000000122,
       0.00000000019,    -0.00000000743,
      -0.00000000087,    -0.00000000005,
       0.00000001038,     0.00000001859,
       0.00000000316,    -0.00000000136,
       0.00000001663,    -0.00000001152,
      -0.00000000175,    -0.00000000256,
      -0.00000001241,    -0.00000024999,
      -0.00000005968,    -0.00000000095,
      -0.00000000039,     0.00000000702,
      -0.00000024028,     0.00000000951,
      -0.00000000140,     0.00000005753,
       0.00000000675,     0.00000000042,
      -0.00000011835,    -0.00000020480,
      -0.00000003455,     0.00000001601,
       0.00000000107,     0.00000000292,
      -0.00000018944,     0.00000012711,
       0.00000001910,     0.00000002979,
       0.00000000219,    -0.00000000158,
      -0.00000000322,     0.00000000096,
       0.00000000016,     0.00000000040,
       0.00000000199,     0.00000000183,
       0.00000000015,    -0.00000000035,
       0.00000001245,     0.00000002009,
       0.00000000328,    -0.00000000176,
      -0.00000000013,    -0.00000000026,
       0.00000003693,    -0.00000002256,
      -0.00000000320,    -0.00000000602,
      -0.00000000048,     0.00000000024,
       0.00000001784,     0.00000001245,
       0.00000000140,    -0.00000000132,
       0.00000000758,    -0.00000001784,
      -0.00000000138,    -0.00000000026,
       0.00000000036,    -0.00000000017,
      -0.00000018615,    -0.00000030030,
      -0.00000004910,     0.00000002644,
       0.00000000198,     0.00000000391,
      -0.00000027992,     0.00000017097,
       0.00000002438,     0.00000004573,
       0.00000000363,    -0.00000000183,
      -0.00000020322,    -0.00000013084,
      -0.00000001352,     0.00000001520,
       0.00000000057,     0.00000000083,
      -0.00000009662,     0.00000020305,
       0.00000001562,     0.00000000538,
      -0.00000000013,    -0.00000000065,
      -0.00000000228,     0.00000000211,
       0.00000000039,     0.00000000027,
       0.00000000236,    -0.00000000028,
       0.00000002459,     0.00000001384,
       0.00000000118,    -0.00000000186,
       0.00000002329,    -0.00000004133,
      -0.00000000312,    -0.00000000197,
       0.00000001225,     0.00000000422,
       0.00000000060,    -0.00000000031,
      -0.00000000372,    -0.00000000766,
       0.00000000035,     0.00000000068,
       0.00000000027,    -0.00000000025,
      -0.00000032723,    -0.00000018495,
      -0.00000001567,     0.00000002478,
       0.00000000098,     0.00000000059,
      -0.00000017331,     0.00000030596,
       0.00000002311,     0.00000001454,
       0.00000000055,    -0.00000000091,
      -0.00000013139,    -0.00000002722,
      -0.00000000357,     0.00000000186,
       0.00000002225,     0.00000010264,
      -0.00000000226,    -0.00000000446,
      -0.00000000032,     0.00000000020,
      -0.00000000104,     0.00000000250,
       0.00000000050,     0.00000000006,
       0.00000000105,    -0.00000000188,
       0.00000001741,     0.00000000054,
       0.00000000053,    -0.00000002760,
       0.00000000244,     0.00000000543,
       0.00000000013,    -0.00000000067,
      -0.00000000121,     0.00000000896,
       0.00000000019,     0.00000000043,
      -0.00000020259,    -0.00000000621,
       0.00000000046,    -0.00000000028,
      -0.00000000403,     0.00000019703,
      -0.00000000036,    -0.00000000044,
       0.00000000442,    -0.00000004080,
       0.00000000012,     0.00000000501,
      -0.00000000476,    -0.00000006636,
       0.00000000028,    -0.00000000316,
       0.00000000000,     0.00000000238,
      -0.00000000048,    -0.00000000156,
      -0.00000000470,     0.00000000328,
      -0.00000000023,    -0.00000000036,
       0.00000000450,     0.00000000666,
      -0.00000000049,     0.00000000031,
       0.00000000526,    -0.00000000030,
      -0.00000000054,     0.00000000489,
       0.00000004689,    -0.00000003265,
       0.00000000230,     0.00000000355,
      -0.00000000017,     0.00000000013,
      -0.00000003314,    -0.00000004921,
       0.00000000361,    -0.00000000227,
      -0.00000002527,     0.00000000379,
      -0.00000000081,    -0.00000000031,
       0.00000000493,    -0.00000002155,
      -0.00000000032,     0.00000000089,
       0.00000000003,     0.00000000128,
      -0.00000000140,     0.00000000044,
       0.00000000018,    -0.00000000056,
      -0.00000000069,    -0.00000000027,
      -0.00000000120,     0.00000000091,
      -0.00000000086,    -0.00000000122,
      -0.00000000148,     0.00000000463,
      -0.00000000094,    -0.00000000029,
       0.00000000526,     0.00000000208,
      -0.00000000039,     0.00000000118,
       0.00000000473,    -0.00000000378,
       0.00000000034,     0.00000000037,
       0.00000000324,     0.00000000490,
      -0.00000000041,     0.00000000022,
       0.00000000076,     0.00000000091,
      -0.00000000092,     0.00000000071,
       0.00000000084,    -0.00000000010,
      -0.00000000013,    -0.00000000105,
       0.00000000151,     0.00000000020,
       0.00000000020,    -0.00000000187,
       0.00000000081,    -0.00000000006,
       0.00000000015,     0.00000000093,
      -0.00000000034,     0.00000000036,
       0.00000000037,     0.00000000039,
       0.00000000182,     0.00000000018,
      -0.00000000020,     0.00000000182,
      -0.00000000035,     0.00000000043,
      -0.00000000043,    -0.00000000035,
      -0.00000000011,    -0.00000000069,
      -0.00000000052,     0.00000000004,
       0.00000000000,     0.00000000019,
      -0.00000000073,    -0.00000000096,
      -0.00000000067,     0.00000000043,
      -0.00000000094,     0.00000000074,
       0.00000000044,     0.00000000066,
      -0.00000000029,     0.00000000024,
       0.00000000070,     0.00000000031,
       0.00000000021,    -0.00000000039,
       0.00000000029,    -0.00000000070,
      -0.00000000039,    -0.00000000020,
      -0.00000000312,    -0.00000000402,
      -0.00000000281,     0.00000000184,
       0.00000000053,     0.00000000094,
       0.00000000201,    -0.00000000165,
      -0.00000000097,    -0.00000000138,
      -0.00000000046,     0.00000000028,
      -0.00000000758,    -0.00000000334,
      -0.00000000226,     0.00000000418,
       0.00000000113,     0.00000000071,
      -0.00000000316,     0.00000000758,
       0.00000000419,     0.00000000212,
       0.00000000066,    -0.00000000113,
      -0.00000001456,    -0.00000000624,
      -0.00000000420,     0.00000000804,
       0.00000000218,     0.00000000131,
      -0.00000038499,     0.00000089705,
       0.00000049590,     0.00000025963,
       0.00000008140,    -0.00000013485,
      -0.00000002440,    -0.00000001680,
      -0.00000000260,     0.00000000330,
       0.00000000040,     0.00000000030,
      -0.00000004805,     0.00000000418,
       0.00000000024,     0.00000002352,
       0.00000000561,     0.00000000027,
       0.00000000540,     0.00000004735,
       0.00000002313,    -0.00000000102,
       0.00000000003,    -0.00000000550,
      -0.00000000030,     0.00000000208,
       0.00000000067,     0.00000000016,
       0.00000000188,    -0.00000000013,
      -0.00000000004,    -0.00000000059,
       0.00000031706,     0.00000013592,
       0.00000009168,    -0.00000017535,
      -0.00000004754,    -0.00000002865,
       0.00000294204,    -0.00000685519,
      -0.00000379283,    -0.00000198525,
      -0.00000062276,     0.00000103187,
       0.00000018640,     0.00000012870,
       0.00000001970,    -0.00000002510,
      -0.00000000269,    -0.00000000239,
       0.00000051975,    -0.00000004638,
      -0.00000000291,    -0.00000025490,
      -0.00000006107,    -0.00000000304,
      -0.00000000110,     0.00000000970,
      -0.00000005813,    -0.00000051260,
      -0.00000025078,     0.00000001088,
      -0.00000000037,     0.00000005987,
       0.00000000950,     0.00000000050,
       0.00000000170,    -0.00000001622,
      -0.00000000541,    -0.00000000099,
      -0.00000000019,     0.00000000096,
      -0.00000001514,     0.00000000061,
       0.00000000018,     0.00000000496,
       0.00000000087,    -0.00000000009,
       0.00000000192,    -0.00000000004,
      -0.00000000002,    -0.00000000134,
      -0.00000000196,     0.00000000070,
       0.00000000038,     0.00000000156,
      -0.00000000232,     0.00000000023,
       0.00000000003,     0.00000000113,
       0.00000000027,     0.00000000001,
       0.00000000083,     0.00000000906,
       0.00000000443,    -0.00000000008,
       0.00000000004,    -0.00000000106,
      -0.00000000499,     0.00000000333,
       0.00000000119,     0.00000000210,
       0.00000000043,    -0.00000000021,
       0.00000000345,     0.00000000478,
       0.00000000200,    -0.00000000126,
      -0.00000000023,    -0.00000000040,
       0.00000004728,    -0.00000000477,
      -0.00000000065,    -0.00000002308,
      -0.00000000549,    -0.00000000016,
      -0.00000000653,    -0.00000007118,
      -0.00000003483,     0.00000000062,
      -0.00000000032,     0.00000000830,
       0.00000005438,    -0.00000003658,
      -0.00000001304,    -0.00000002298,
      -0.00000000471,     0.00000000231,
      -0.00000003764,    -0.00000005246,
      -0.00000002193,     0.00000001371,
       0.00000000252,     0.00000000443,
      -0.00000000063,    -0.00000000153,
      -0.00000000042,     0.00000000014,
      -0.00000000126,     0.00000000083,
       0.00000000023,     0.00000000032,
      -0.00000000295,     0.00000000205,
       0.00000000074,     0.00000000124,
       0.00000000025,    -0.00000000013,
       0.00000000516,     0.00000000762,
       0.00000000321,    -0.00000000186,
      -0.00000000033,    -0.00000000065,
      -0.00000000351,     0.00000000672,
       0.00000000206,     0.00000000129,
       0.00000000023,    -0.00000000031,
       0.00000000672,     0.00000000309,
       0.00000000109,    -0.00000000207,
      -0.00000000032,    -0.00000000018,
       0.00000005574,    -0.00000003872,
      -0.00000001397,    -0.00000002344,
      -0.00000000477,     0.00000000251,
      -0.00000004068,    -0.00000006006,
      -0.00000002530,     0.00000001466,
       0.00000000264,     0.00000000514,
       0.00000003826,    -0.00000007416,
      -0.00000002278,    -0.00000001405,
      -0.00000000245,     0.00000000348,
      -0.00000007409,    -0.00000003457,
      -0.00000001227,     0.00000002286,
       0.00000000352,     0.00000000202,
      -0.00000000132,    -0.00000000108,
      -0.00000000024,     0.00000000027,
      -0.00000000062,     0.00000000137,
      -0.00000000015,     0.00000000025,
      -0.00000000024,     0.00000000056,
      -0.00000000285,     0.00000000580,
       0.00000000178,     0.00000000103,
       0.00000000018,    -0.00000000027,
       0.00000001207,     0.00000000612,
       0.00000000221,    -0.00000000372,
      -0.00000000057,    -0.00000000037,
       0.00000000012,     0.00000001075,
       0.00000000258,     0.00000000019,
       0.00000001028,    -0.00000000092,
      -0.00000000013,    -0.00000000241,
       0.00000004978,    -0.00000010120,
      -0.00000003118,    -0.00000001805,
      -0.00000000307,     0.00000000477,
      -0.00000009445,    -0.00000004782,
      -0.00000001727,     0.00000002912,
       0.00000000446,     0.00000000292,
      -0.00000000265,    -0.00000011911,
      -0.00000002855,    -0.00000000167,
      -0.00000000043,     0.00000000341,
      -0.00000011525,     0.00000000907,
       0.00000000099,     0.00000002716,
       0.00000000316,    -0.00000000011,
      -0.00000000193,    -0.00000000039,
       0.00000000031,     0.00000000162,
       0.00000000055,     0.00000001118,
       0.00000000266,     0.00000000004,
       0.00000000002,    -0.00000000031,
       0.00000002065,    -0.00000000081,
       0.00000000012,    -0.00000000492,
      -0.00000000058,    -0.00000000003,
       0.00000000644,     0.00000001210,
       0.00000000209,    -0.00000000082,
       0.00000001032,    -0.00000000745,
      -0.00000000116,    -0.00000000154,
      -0.00000000873,    -0.00000017862,
      -0.00000004256,    -0.00000000066,
      -0.00000000027,     0.00000000501,
      -0.00000015933,     0.00000000628,
      -0.00000000088,     0.00000003793,
       0.00000000446,     0.00000000027,
      -0.00000007469,    -0.00000013297,
      -0.00000002252,     0.00000000990,
       0.00000000063,     0.00000000194,
      -0.00000011974,     0.00000008206,
       0.00000001249,     0.00000001845,
       0.00000000131,    -0.00000000107,
      -0.00000000204,     0.00000000058,
       0.00000000130,     0.00000000102,
       0.00000000870,     0.00000001411,
       0.00000000229,    -0.00000000123,
      -0.00000000009,    -0.00000000018,
       0.00000002396,    -0.00000001461,
      -0.00000000208,    -0.00000000389,
       0.00000001055,     0.00000000806,
       0.00000000098,    -0.00000000077,
       0.00000000390,    -0.00000001052,
      -0.00000000082,     0.00000000000,
      -0.00000012650,    -0.00000020506,
      -0.00000003343,     0.00000001799,
       0.00000000135,     0.00000000266,
      -0.00000018135,     0.00000011051,
       0.00000001573,     0.00000002946,
       0.00000000234,    -0.00000000118,
      -0.00000012097,    -0.00000008233,
      -0.00000000892,     0.00000000901,
       0.00000000033,     0.00000000059,
      -0.00000005432,     0.00000012064,
       0.00000000931,     0.00000000226,
      -0.00000000019,    -0.00000000039,
      -0.00000000150,     0.00000000133,
       0.00000000028,     0.00000000021,
       0.00000000148,    -0.00000000032,
       0.00000001561,     0.00000000893,
       0.00000000075,    -0.00000000118,
       0.00000001413,    -0.00000002477,
      -0.00000000187,    -0.00000000118,
       0.00000000671,     0.00000000331,
       0.00000000048,    -0.00000000024,
      -0.00000000288,    -0.00000000310,
       0.00000000028,     0.00000000053,
      -0.00000020425,    -0.00000011682,
      -0.00000000976,     0.00000001545,
       0.00000000060,     0.00000000036,
      -0.00000010455,     0.00000018302,
       0.00000001381,     0.00000000866,
       0.00000000032,    -0.00000000053,
      -0.00000006938,    -0.00000002097,
      -0.00000000274,     0.00000000142,
       0.00000001651,     0.00000004746,
      -0.00000000172,    -0.00000000334,
      -0.00000000072,     0.00000000162,
       0.00000000058,    -0.00000000125,
       0.00000000921,     0.00000000045,
       0.00000000046,    -0.00000001401,
       0.00000000176,     0.00000000371,
       0.00000000010,    -0.00000000046,
      -0.00000000090,     0.00000000611,
      -0.00000010717,    -0.00000000519,
       0.00000000034,    -0.00000000022,
      -0.00000000351,     0.00000009967,
      -0.00000000028,    -0.00000000032,
       0.00000000279,    -0.00000002774,
       0.00000000005,     0.00000000340,
      -0.00000000304,    -0.00000004505,
       0.00000000017,    -0.00000000211,
      -0.00000000008,     0.00000000149,
      -0.00000000045,    -0.00000000091,
      -0.00000000333,     0.00000000229,
      -0.00000000016,    -0.00000000025,
       0.00000000308,     0.00000000460,
      -0.00000000034,     0.00000000021,
       0.00000000304,    -0.00000000025,
      -0.00000000046,     0.00000000274,
       0.00000003387,    -0.00000002326,
       0.00000000161,     0.00000000255,
      -0.00000000012,     0.00000000009,
      -0.00000002273,    -0.00000003397,
       0.00000000249,    -0.00000000153,
      -0.00000001504,     0.00000000321,
      -0.00000000066,    -0.00000000026,
       0.00000000425,    -0.00000001191,
      -0.00000000026,     0.00000000073,
       0.00000000000,     0.00000000085,
      -0.00000000094,     0.00000000032,
       0.00000000017,    -0.00000000050,
      -0.00000000062,    -0.00000000024,
      -0.00000000089,     0.00000000068,
      -0.00000000062,    -0.00000000091,
      -0.00000000144,     0.00000000437,
      -0.00000000084,    -0.00000000027,
       0.00000000468,     0.00000000182,
      -0.00000000033,     0.00000000096,
       0.00000000360,    -0.00000000293,
       0.00000000028,     0.00000000027,
       0.00000000235,     0.00000000377,
       0.00000000045,     0.00000000053,
      -0.00000000055,     0.00000000041,
      -0.00000000009,     0.00000000044,
      -0.00000000043,    -0.00000000011,
      -0.00000000026,    -0.00000000168,
      -0.00000000126,     0.00000000009,
       0.00000000000,     0.00000000046,
       0.00000000068,    -0.00000000013,
      -0.00000000006,    -0.00000000050,
      -0.00000000167,    -0.00000000220,
      -0.00000000153,     0.00000000098,
       0.00000000028,     0.00000000051,
      -0.00000000214,     0.00000000170,
       0.00000000101,     0.00000000148,
       0.00000000049,    -0.00000000029,
       0.00000000234,     0.00000000302,
       0.00000000209,    -0.00000000138,
      -0.00000000040,    -0.00000000070,
      -0.00000011387,     0.00000008858,
       0.00000005214,     0.00000007889,
       0.00000002621,    -0.00000001503,
      -0.00000000280,    -0.00000000570,
       0.00000001079,     0.00000000482,
       0.00000000320,    -0.00000000591,
      -0.00000000160,    -0.00000000100,
       0.00000000446,    -0.00000001079,
      -0.00000000592,    -0.00000000296,
      -0.00000000092,     0.00000000161,
       0.00000000033,    -0.00000000044,
      -0.00000000045,    -0.00000000022,
      -0.00000005135,    -0.00000006612,
      -0.00000004587,     0.00000003025,
       0.00000000876,     0.00000001532,
       0.00000086298,    -0.00000067138,
      -0.00000039570,    -0.00000059940,
      -0.00000019914,     0.00000011425,
       0.00000002170,     0.00000004390,
       0.00000000720,    -0.00000000300,
      -0.00000000033,    -0.00000000095,
      -0.00000011686,    -0.00000005199,
      -0.00000003473,     0.00000006428,
       0.00000001734,     0.00000001085,
      -0.00000004853,     0.00000011671,
       0.00000006434,     0.00000003219,
       0.00000000992,    -0.00000001742,
      -0.00000000242,     0.00000000349,
       0.00000000137,     0.00000000112,
       0.00000000025,    -0.00000000028,
       0.00000000353,     0.00000000181,
       0.00000000077,    -0.00000000142,
      -0.00000000092,    -0.00000000039,
      -0.00000000026,     0.00000000050,
       0.00000000014,     0.00000000008,
      -0.00000000169,     0.00000000385,
       0.00000000212,     0.00000000113,
       0.00000000035,    -0.00000000057,
      -0.00000000265,     0.00000000022,
       0.00000000001,     0.00000000129,
       0.00000000031,     0.00000000260,
       0.00000000126,    -0.00000000007,
       0.00000001892,     0.00000000812,
       0.00000000542,    -0.00000001040,
      -0.00000000280,    -0.00000000168,
       0.00000001331,    -0.00000003031,
      -0.00000001667,    -0.00000000886,
      -0.00000000275,     0.00000000451,
       0.00000002893,    -0.00000000250,
      -0.00000000014,    -0.00000001409,
      -0.00000000336,    -0.00000000017,
      -0.00000000331,    -0.00000002842,
      -0.00000001380,     0.00000000069,
       0.00000000001,     0.00000000328,
       0.00000000009,    -0.00000000080,
      -0.00000000074,     0.00000000004,
      -0.00000000146,     0.00000000015,
       0.00000000002,     0.00000000071,
       0.00000000017,     0.00000000000,
       0.00000000034,     0.00000000386,
       0.00000000188,    -0.00000000003,
      -0.00000000278,     0.00000000184,
       0.00000000065,     0.00000000117,
       0.00000000192,     0.00000000263,
       0.00000000109,    -0.00000000070,
       0.00000002838,    -0.00000000287,
      -0.00000000040,    -0.00000001378,
      -0.00000000328,    -0.00000000009,
      -0.00000000273,    -0.00000003062,
      -0.00000001492,     0.00000000026,
      -0.00000000013,     0.00000000355,
       0.00000003040,    -0.00000002031,
      -0.00000000721,    -0.00000001281,
      -0.00000000263,     0.00000000128,
      -0.00000002103,    -0.00000002905,
      -0.00000001207,     0.00000000767,
       0.00000000141,     0.00000000243,
      -0.00000000032,    -0.00000000077,
      -0.00000000061,     0.00000000043,
      -0.00000000222,     0.00000000154,
       0.00000000055,     0.00000000093,
       0.00000000019,    -0.00000000010,
       0.00000000321,     0.00000000475,
       0.00000000199,    -0.00000000116,
      -0.00000000021,    -0.00000000040,
      -0.00000000237,     0.00000000443,
       0.00000000136,     0.00000000087,
       0.00000000443,     0.00000000199,
       0.00000000069,    -0.00000000136,
       0.00000004014,    -0.00000002789,
      -0.00000001006,    -0.00000001681,
      -0.00000000342,     0.00000000181,
      -0.00000002537,    -0.00000003750,
      -0.00000001573,     0.00000000913,
       0.00000000165,     0.00000000319,
       0.00000002572,    -0.00000004923,
      -0.00000001509,    -0.00000000944,
      -0.00000000165,     0.00000000230,
      -0.00000004913,    -0.00000002261,
      -0.00000000794,     0.00000001514,
       0.00000000233,     0.00000000129,
      -0.00000000081,    -0.00000000067,
      -0.00000000034,     0.00000000084,
      -0.00000000220,     0.00000000446,
       0.00000000137,     0.00000000079,
       0.00000000013,    -0.00000000021,
       0.00000000826,     0.00000000419,
       0.00000000151,    -0.00000000254,
      -0.00000000039,    -0.00000000025,
      -0.00000000002,     0.00000000729,
       0.00000000175,     0.00000000017,
       0.00000000685,    -0.00000000070,
      -0.00000000013,    -0.00000000160,
       0.00000003688,    -0.00000007490,
      -0.00000002303,    -0.00000001331,
      -0.00000000226,     0.00000000352,
      -0.00000006457,    -0.00000003270,
      -0.00000001175,     0.00000001987,
       0.00000000305,     0.00000000199,
      -0.00000000113,    -0.00000008093,
      -0.00000001940,    -0.00000000136,
      -0.00000000033,     0.00000000232,
      -0.00000007754,     0.00000000664,
       0.00000000092,     0.00000001819,
       0.00000000210,    -0.00000000013,
      -0.00000000124,    -0.00000000026,
       0.00000000025,     0.00000000101,
       0.00000000039,     0.00000000831,
       0.00000000197,     0.00000000003,
       0.00000000001,    -0.00000000023,
       0.00000001416,    -0.00000000055,
       0.00000000008,    -0.00000000336,
       0.00000000403,     0.00000000804,
       0.00000000141,    -0.00000000049,
       0.00000000647,    -0.00000000490,
      -0.00000000079,    -0.00000000093,
      -0.00000000612,    -0.00000012873,
      -0.00000003062,    -0.00000000047,
      -0.00000000019,     0.00000000360,
      -0.00000010904,     0.00000000424,
      -0.00000000058,     0.00000002590,
       0.00000000305,     0.00000000018,
      -0.00000004764,    -0.00000008780,
      -0.00000001500,     0.00000000618,
       0.00000000037,     0.00000000131,
      -0.00000007662,     0.00000005377,
       0.00000000834,     0.00000001156,
       0.00000000078,    -0.00000000073,
      -0.00000000133,     0.00000000036,
       0.00000000088,     0.00000000059,
       0.00000000602,     0.00000000983,
       0.00000000159,    -0.00000000085,
      -0.00000000006,    -0.00000000013,
       0.00000001570,    -0.00000000953,
      -0.00000000136,    -0.00000000254,
       0.00000000625,     0.00000000530,
       0.00000000069,    -0.00000000045,
       0.00000000187,    -0.00000000620,
      -0.00000000049,     0.00000000012,
      -0.00000008553,    -0.00000013959,
      -0.00000002268,     0.00000001219,
       0.00000000091,     0.00000000180,
      -0.00000011871,     0.00000007202,
       0.00000001026,     0.00000001920,
       0.00000000152,    -0.00000000077,
      -0.00000007202,    -0.00000005235,
      -0.00000000596,     0.00000000534,
       0.00000000018,     0.00000000043,
      -0.00000003003,     0.00000007166,
       0.00000000556,     0.00000000066,
      -0.00000000100,     0.00000000086,
       0.00000000094,    -0.00000000029,
       0.00000000976,     0.00000000567,
       0.00000000047,    -0.00000000074,
       0.00000000855,    -0.00000001481,
      -0.00000000112,    -0.00000000070,
       0.00000000369,     0.00000000251,
       0.00000000036,    -0.00000000018,
      -0.00000000216,    -0.00000000096,
       0.00000000021,     0.00000000039,
      -0.00000012617,    -0.00000007327,
      -0.00000000603,     0.00000000954,
       0.00000000036,     0.00000000022,
      -0.00000006319,     0.00000010931,
       0.00000000823,     0.00000000515,
       0.00000000018,    -0.00000000031,
      -0.00000003623,    -0.00000001563,
      -0.00000000203,     0.00000000104,
       0.00000001201,     0.00000002003,
      -0.00000000128,    -0.00000000244,
      -0.00000000051,     0.00000000105,
       0.00000000031,    -0.00000000083,
       0.00000000471,     0.00000000035,
       0.00000000037,    -0.00000000693,
       0.00000000121,     0.00000000247,
      -0.00000000064,     0.00000000406,
      -0.00000005474,    -0.00000000409,
       0.00000000025,    -0.00000000018,
      -0.00000000282,     0.00000004904,
       0.00000000178,    -0.00000001838,
       0.00000000001,     0.00000000225,
      -0.00000000184,    -0.00000002981,
       0.00000000011,    -0.00000000138,
      -0.00000000010,     0.00000000094,
      -0.00000000038,    -0.00000000054,
      -0.00000000228,     0.00000000154,
      -0.00000000010,    -0.00000000017,
       0.00000000205,     0.00000000308,
       0.00000000170,    -0.00000000019,
      -0.00000000038,     0.00000000147,
       0.00000002360,    -0.00000001596,
       0.00000000109,     0.00000000177,
      -0.00000001510,    -0.00000002278,
       0.00000000167,    -0.00000000100,
      -0.00000000871,     0.00000000259,
      -0.00000000052,    -0.00000000021,
       0.00000000347,    -0.00000000619,
      -0.00000000021,     0.00000000057,
      -0.00000000001,     0.00000000055,
      -0.00000000062,     0.00000000022,
       0.00000000014,    -0.00000000042,
      -0.00000000052,    -0.00000000020,
      -0.00000000064,     0.00000000049,
      -0.00000000043,    -0.00000000066,
      -0.00000000130,     0.00000000383,
      -0.00000000070,    -0.00000000024,
       0.00000000392,     0.00000000152,
      -0.00000000027,     0.00000000076,
       0.00000000263,    -0.00000000219,
       0.00000000162,     0.00000000278,
       0.00000000026,     0.00000000031,
      -0.00000000032,     0.00000000023,
      -0.00000000009,     0.00000000041,
      -0.00000000040,    -0.00000000012,
      -0.00000000043,    -0.00000000020,
      -0.00000000009,     0.00000000017,
      -0.00000000013,    -0.00000000092,
      -0.00000000069,     0.00000000004,
      -0.00000000091,     0.00000000015,
       0.00000000006,     0.00000000068,
       0.00000000008,     0.00000000054,
       0.00000000040,    -0.00000000003,
       0.00000000000,    -0.00000000014,
      -0.00000000724,     0.00000000115,
       0.00000000043,     0.00000000539,
       0.00000000196,    -0.00000000003,
       0.00000000102,     0.00000000136,
       0.00000000093,    -0.00000000060,
       0.00000000131,    -0.00000000104,
      -0.00000000061,    -0.00000000090,
      -0.00000000183,    -0.00000001172,
      -0.00000000873,     0.00000000067,
       0.00000000004,     0.00000000320,
       0.00000005409,    -0.00000000860,
      -0.00000000324,    -0.00000004030,
      -0.00000001474,     0.00000000022,
      -0.00000001099,    -0.00000001453,
      -0.00000001004,     0.00000000642,
       0.00000000186,     0.00000000338,
      -0.00000001408,     0.00000001121,
       0.00000000663,     0.00000000968,
       0.00000000324,    -0.00000000196,
      -0.00000000058,     0.00000000029,
       0.00000000033,     0.00000000050,
      -0.00000000026,    -0.00000000034,
      -0.00000000023,     0.00000000015,
      -0.00000000159,     0.00000000122,
       0.00000000072,     0.00000000110,
       0.00000000037,    -0.00000000021,
      -0.00000000108,    -0.00000000049,
      -0.00000000032,     0.00000000059,
      -0.00000000044,     0.00000000108,
       0.00000000059,     0.00000000029,
       0.00000000552,     0.00000000712,
       0.00000000487,    -0.00000000322,
      -0.00000000094,    -0.00000000163,
       0.00000001253,    -0.00000000957,
      -0.00000000562,    -0.00000000864,
      -0.00000000288,     0.00000000163,
       0.00000001180,     0.00000000527,
       0.00000000350,    -0.00000000645,
      -0.00000000174,    -0.00000000109,
       0.00000000484,    -0.00000001178,
      -0.00000000646,    -0.00000000319,
      -0.00000000098,     0.00000000175,
       0.00000000022,    -0.00000000031,
      -0.00000000055,    -0.00000000024,
      -0.00000000016,     0.00000000030,
       0.00000000008,     0.00000000005,
      -0.00000000066,     0.00000000149,
       0.00000000082,     0.00000000044,
      -0.00000000143,     0.00000000011,
       0.00000000000,     0.00000000069,
       0.00000000017,     0.00000000139,
       0.00000000067,    -0.00000000004,
       0.00000001091,     0.00000000468,
       0.00000000309,    -0.00000000596,
      -0.00000000161,    -0.00000000096,
       0.00000000526,    -0.00000001189,
      -0.00000000652,    -0.00000000348,
      -0.00000000107,     0.00000000176,
       0.00000001564,    -0.00000000130,
      -0.00000000005,    -0.00000000760,
      -0.00000000181,    -0.00000000010,
      -0.00000000184,    -0.00000001530,
      -0.00000000740,     0.00000000041,
       0.00000000002,     0.00000000175,
      -0.00000000111,     0.00000000011,
       0.00000000002,     0.00000000053,
       0.00000000013,     0.00000000000,
       0.00000000021,     0.00000000239,
       0.00000000116,    -0.00000000002,
      -0.00000000188,     0.00000000122,
       0.00000000043,     0.00000000079,
       0.00000000130,     0.00000000175,
       0.00000000072,    -0.00000000047,
       0.00000002056,    -0.00000000208,
      -0.00000000030,    -0.00000000995,
      -0.00000000237,    -0.00000000006,
      -0.00000000167,    -0.00000001896,
      -0.00000000921,     0.00000000016,
      -0.00000000008,     0.00000000219,
       0.00000002063,    -0.00000001366,
      -0.00000000483,    -0.00000000867,
      -0.00000000178,     0.00000000085,
      -0.00000001424,    -0.00000001950,
      -0.00000000806,     0.00000000521,
       0.00000000096,     0.00000000162,
      -0.00000000019,    -0.00000000048,
      -0.00000000037,     0.00000000027,
      -0.00000000175,     0.00000000121,
       0.00000000044,     0.00000000073,
       0.00000000015,    -0.00000000008,
       0.00000000224,     0.00000000332,
       0.00000000139,    -0.00000000081,
      -0.00000000170,     0.00000000310,
       0.00000000095,     0.00000000063,
       0.00000000309,     0.00000000136,
       0.00000000046,    -0.00000000095,
       0.00000003047,    -0.00000002116,
      -0.00000000762,    -0.00000001272,
      -0.00000000258,     0.00000000137,
      -0.00000001768,    -0.00000002615,
      -0.00000001093,     0.00000000636,
       0.00000000115,     0.00000000222,
       0.00000001835,    -0.00000003460,
      -0.00000001059,    -0.00000000673,
      -0.00000000118,     0.00000000162,
      -0.00000003449,    -0.00000001565,
      -0.00000000543,     0.00000001061,
       0.00000000164,     0.00000000087,
      -0.00000000053,    -0.00000000045,
      -0.00000000020,     0.00000000055,
      -0.00000000170,     0.00000000345,
       0.00000000106,     0.00000000061,
       0.00000000010,    -0.00000000016,
       0.00000000589,     0.00000000299,
       0.00000000107,    -0.00000000181,
      -0.00000000009,     0.00000000504,
       0.00000000122,     0.00000000014,
       0.00000000465,    -0.00000000055,
      -0.00000000012,    -0.00000000108,
       0.00000002773,    -0.00000005617,
      -0.00000001725,    -0.00000000994,
      -0.00000000169,     0.00000000264,
      -0.00000004593,    -0.00000002328,
      -0.00000000832,     0.00000001411,
       0.00000000216,     0.00000000141,
      -0.00000000023,    -0.00000005604,
      -0.00000001345,    -0.00000000112,
      -0.00000000027,     0.00000000162,
      -0.00000005309,     0.00000000495,
       0.00000000083,     0.00000001239,
       0.00000000143,    -0.00000000013,
      -0.00000000083,    -0.00000000018,
       0.00000000020,     0.00000000065,
       0.00000000028,     0.00000000613,
       0.00000000145,     0.00000000002,
       0.00000000001,    -0.00000000017,
       0.00000000983,    -0.00000000037,
       0.00000000005,    -0.00000000233,
       0.00000000251,     0.00000000539,
       0.00000000096,    -0.00000000028,
       0.00000000405,    -0.00000000325,
      -0.00000000054,    -0.00000000055,
      -0.00000000421,    -0.00000009265,
      -0.00000002200,    -0.00000000035,
      -0.00000000014,     0.00000000259,
      -0.00000007557,     0.00000000286,
      -0.00000000039,     0.00000001791,
       0.00000000211,     0.00000000012,
      -0.00000003036,    -0.00000005832,
      -0.00000001006,     0.00000000383,
       0.00000000021,     0.00000000090,
      -0.00000004906,     0.00000003537,
       0.00000000560,     0.00000000722,
       0.00000000046,    -0.00000000051,
      -0.00000000089,     0.00000000023,
       0.00000000060,     0.00000000034,
       0.00000000411,     0.00000000677,
       0.00000000109,    -0.00000000058,
      -0.00000000004,    -0.00000000009,
       0.00000001031,    -0.00000000622,
      -0.00000000089,    -0.00000000166,
       0.00000000368,     0.00000000351,
       0.00000000049,    -0.00000000026,
       0.00000000078,    -0.00000000363,
      -0.00000005732,    -0.00000009437,
      -0.00000001528,     0.00000000818,
       0.00000000061,     0.00000000120,
      -0.00000007779,     0.00000004691,
       0.00000000670,     0.00000001253,
       0.00000000099,    -0.00000000050,
      -0.00000004259,    -0.00000003345,
      -0.00000000400,     0.00000000313,
      -0.00000001607,     0.00000004228,
       0.00000000329,    -0.00000000011,
      -0.00000000068,     0.00000000056,
       0.00000000060,    -0.00000000024,
       0.00000000601,     0.00000000356,
       0.00000000029,    -0.00000000045,
       0.00000000516,    -0.00000000880,
      -0.00000000066,    -0.00000000042,
       0.00000000202,     0.00000000184,
      -0.00000000158,    -0.00000000002,
      -0.00000007691,    -0.00000004553,
      -0.00000000368,     0.00000000579,
       0.00000000021,     0.00000000013,
      -0.00000003808,     0.00000006484,
       0.00000000487,     0.00000000304,
      -0.00000001851,    -0.00000001135,
      -0.00000000148,     0.00000000075,
       0.00000000857,     0.00000000685,
      -0.00000000092,    -0.00000000174,
      -0.00000000036,     0.00000000068,
       0.00000000015,    -0.00000000055,
       0.00000000228,     0.00000000026,
       0.00000000028,    -0.00000000328,
       0.00000000080,     0.00000000160,
      -0.00000000045,     0.00000000263,
      -0.00000002652,    -0.00000000307,
      -0.00000000214,     0.00000002296,
       0.00000000115,    -0.00000001190,
      -0.00000000001,     0.00000000145,
      -0.00000000105,    -0.00000001929,
       0.00000000007,    -0.00000000087,
      -0.00000000010,     0.00000000060,
      -0.00000000030,    -0.00000000031,
      -0.00000000153,     0.00000000101,
       0.00000000132,     0.00000000202,
       0.00000000092,    -0.00000000015,
      -0.00000000029,     0.00000000074,
       0.00000001597,    -0.00000001061,
       0.00000000070,     0.00000000120,
      -0.00000000976,    -0.00000001492,
       0.00000000109,    -0.00000000063,
      -0.00000000491,     0.00000000202,
      -0.00000000040,    -0.00000000017,
       0.00000000271,    -0.00000000294,
      -0.00000000016,     0.00000000043,
      -0.00000000041,     0.00000000015,
      -0.00000000041,    -0.00000000016,
      -0.00000000044,     0.00000000034,
      -0.00000000029,    -0.00000000046,
      -0.00000000110,     0.00000000317,
      -0.00000000056,    -0.00000000020,
       0.00000000313,     0.00000000122,
      -0.00000000021,     0.00000000058,
       0.00000000184,    -0.00000000159,
       0.00000000105,     0.00000000199,
      -0.00000000093,    -0.00000000033,
      -0.00000000033,     0.00000000074,
       0.00000000095,    -0.00000000260,
      -0.00000000204,    -0.00000000093,
      -0.00000000044,     0.00000000081,
       0.00000000693,     0.00000000249,
       0.00000000244,    -0.00000000542,
      -0.00000000215,    -0.00000000115,
      -0.00000000061,    -0.00000000424,
      -0.00000000315,     0.00000000020,
       0.00000000000,     0.00000000114,
      -0.00000000417,     0.00000000070,
       0.00000000029,     0.00000000310,
       0.00000000112,    -0.00000000004,
      -0.00000000099,     0.00000000015,
       0.00000000005,     0.00000000073,
      -0.00000000032,    -0.00000000043,
      -0.00000000041,     0.00000000033,
       0.00000000062,     0.00000000396,
       0.00000000291,    -0.00000000023,
      -0.00000000002,    -0.00000000106,
       0.00000000773,    -0.00000000115,
      -0.00000000041,    -0.00000000575,
      -0.00000000209,     0.00000000001,
       0.00000000351,     0.00000000466,
       0.00000000320,    -0.00000000204,
      -0.00000000059,    -0.00000000107,
       0.00000000449,    -0.00000000359,
      -0.00000000212,    -0.00000000307,
      -0.00000000101,     0.00000000062,
      -0.00000000055,     0.00000000042,
       0.00000000024,     0.00000000038,
      -0.00000000055,    -0.00000000025,
      -0.00000000022,     0.00000000055,
       0.00000000301,     0.00000000388,
       0.00000000264,    -0.00000000175,
      -0.00000000051,    -0.00000000088,
       0.00000000439,    -0.00000000332,
      -0.00000000195,    -0.00000000301,
      -0.00000000100,     0.00000000056,
       0.00000000607,     0.00000000273,
       0.00000000180,    -0.00000000331,
      -0.00000000089,    -0.00000000056,
       0.00000000247,    -0.00000000605,
      -0.00000000331,    -0.00000000161,
      -0.00000000049,     0.00000000089,
      -0.00000000042,    -0.00000000018,
      -0.00000000012,     0.00000000023,
      -0.00000000040,     0.00000000091,
       0.00000000050,     0.00000000026,
      -0.00000000097,     0.00000000007,
       0.00000000012,     0.00000000094,
       0.00000000791,     0.00000000339,
       0.00000000223,    -0.00000000431,
      -0.00000000116,    -0.00000000069,
       0.00000000322,    -0.00000000724,
      -0.00000000396,    -0.00000000211,
      -0.00000000065,     0.00000000107,
       0.00000001067,    -0.00000000085,
      -0.00000000002,    -0.00000000517,
      -0.00000000123,    -0.00000000007,
      -0.00000000128,    -0.00000001038,
      -0.00000000500,     0.00000000031,
       0.00000000002,     0.00000000118,
      -0.00000000088,     0.00000000009,
       0.00000000001,     0.00000000043,
       0.00000000015,     0.00000000168,
       0.00000000081,    -0.00000000001,
      -0.00000000137,     0.00000000087,
       0.00000000031,     0.00000000057,
       0.00000000094,     0.00000000124,
       0.00000000051,    -0.00000000034,
       0.00000001589,    -0.00000000161,
      -0.00000000024,    -0.00000000767,
      -0.00000000182,    -0.00000000004,
      -0.00000000117,    -0.00000001337,
      -0.00000000647,     0.00000000012,
      -0.00000000005,     0.00000000154,
       0.00000001499,    -0.00000000982,
      -0.00000000346,    -0.00000000629,
      -0.00000000129,     0.00000000061,
      -0.00000001032,    -0.00000001400,
      -0.00000000576,     0.00000000378,
       0.00000000070,     0.00000000115,
      -0.00000000139,     0.00000000097,
       0.00000000035,     0.00000000058,
       0.00000000012,    -0.00000000006,
       0.00000000164,     0.00000000243,
       0.00000000101,    -0.00000000059,
      -0.00000000125,     0.00000000222,
       0.00000000068,     0.00000000046,
       0.00000000221,     0.00000000094,
       0.00000000032,    -0.00000000068,
       0.00000002358,    -0.00000001635,
      -0.00000000589,    -0.00000000982,
      -0.00000000199,     0.00000000106,
      -0.00000001293,    -0.00000001913,
      -0.00000000796,     0.00000000465,
       0.00000000084,     0.00000000162,
       0.00000001342,    -0.00000002487,
      -0.00000000759,    -0.00000000493,
      -0.00000000087,     0.00000000116,
      -0.00000002475,    -0.00000001107,
      -0.00000000380,     0.00000000761,
       0.00000000118,     0.00000000060,
      -0.00000000036,    -0.00000000031,
      -0.00000000132,     0.00000000266,
       0.00000000081,     0.00000000047,
       0.00000000008,    -0.00000000012,
       0.00000000426,     0.00000000217,
       0.00000000077,    -0.00000000131,
      -0.00000000013,     0.00000000352,
       0.00000000085,     0.00000000012,
       0.00000000317,    -0.00000000043,
      -0.00000000010,    -0.00000000073,
       0.00000002089,    -0.00000004217,
      -0.00000001291,    -0.00000000746,
      -0.00000000126,     0.00000000198,
      -0.00000003320,    -0.00000001685,
      -0.00000000600,     0.00000001018,
       0.00000000156,     0.00000000101,
       0.00000000028,    -0.00000003903,
      -0.00000000938,    -0.00000000093,
      -0.00000000021,     0.00000000113,
      -0.00000003652,     0.00000000370,
       0.00000000072,     0.00000000848,
       0.00000000097,    -0.00000000012,
      -0.00000000056,    -0.00000000013,
       0.00000000016,     0.00000000042,
       0.00000000019,     0.00000000448,
       0.00000000106,     0.00000000002,
       0.00000000001,    -0.00000000012,
       0.00000000685,    -0.00000000025,
       0.00000000004,    -0.00000000162,
       0.00000000155,     0.00000000362,
       0.00000000066,    -0.00000000016,
       0.00000000250,    -0.00000000215,
      -0.00000000037,    -0.00000000032,
      -0.00000000284,    -0.00000006628,
      -0.00000001571,    -0.00000000027,
      -0.00000000010,     0.00000000184,
      -0.00000005252,     0.00000000190,
      -0.00000000027,     0.00000001242,
       0.00000000146,     0.00000000008,
      -0.00000001919,    -0.00000003874,
      -0.00000000675,     0.00000000234,
       0.00000000011,     0.00000000061,
      -0.00000003123,     0.00000002323,
       0.00000000376,     0.00000000445,
       0.00000000026,    -0.00000000035,
      -0.00000000061,     0.00000000014,
       0.00000000042,     0.00000000019,
       0.00000000276,     0.00000000460,
       0.00000000074,    -0.00000000039,
       0.00000000674,    -0.00000000404,
      -0.00000000058,    -0.00000000108,
       0.00000000214,     0.00000000234,
       0.00000000035,    -0.00000000015,
       0.00000000021,    -0.00000000210,
      -0.00000003800,    -0.00000006326,
      -0.00000001019,     0.00000000544,
       0.00000000041,     0.00000000080,
      -0.00000005082,     0.00000003039,
       0.00000000435,     0.00000000815,
       0.00000000064,    -0.00000000033,
      -0.00000002490,    -0.00000002141,
      -0.00000000270,     0.00000000181,
      -0.00000000816,     0.00000002465,
       0.00000000193,    -0.00000000044,
      -0.00000000047,     0.00000000036,
       0.00000000038,    -0.00000000019,
       0.00000000364,     0.00000000221,
       0.00000000017,    -0.00000000027,
       0.00000000309,    -0.00000000518,
      -0.00000000039,    -0.00000000024,
       0.00000000110,     0.00000000132,
      -0.00000000113,     0.00000000033,
      -0.00000004618,    -0.00000002799,
      -0.00000000221,     0.00000000347,
       0.00000000012,     0.00000000008,
      -0.00000002280,     0.00000003805,
       0.00000000285,     0.00000000178,
      -0.00000000912,    -0.00000000805,
      -0.00000000105,     0.00000000052,
       0.00000000602,     0.00000000093,
      -0.00000000065,    -0.00000000122,
      -0.00000000025,     0.00000000044,
       0.00000000101,     0.00000000019,
       0.00000000021,    -0.00000000144,
       0.00000000052,     0.00000000102,
      -0.00000000031,     0.00000000167,
      -0.00000001176,    -0.00000000223,
      -0.00000000156,     0.00000000984,
       0.00000000075,    -0.00000000754,
      -0.00000000002,     0.00000000092,
      -0.00000000056,    -0.00000001222,
       0.00000000005,    -0.00000000054,
      -0.00000000100,     0.00000000065,
       0.00000000084,     0.00000000130,
       0.00000000048,    -0.00000000011,
      -0.00000000022,     0.00000000034,
       0.00000001054,    -0.00000000685,
       0.00000000044,     0.00000000079,
      -0.00000000616,    -0.00000000957,
       0.00000000070,    -0.00000000038,
      -0.00000000269,     0.00000000152,
       0.00000000205,    -0.00000000119,
      -0.00000000030,     0.00000000023,
      -0.00000000090,     0.00000000251,
      -0.00000000043,    -0.00000000016,
       0.00000000241,     0.00000000095,
      -0.00000000016,     0.00000000043,
       0.00000000126,    -0.00000000113,
       0.00000000065,     0.00000000138,
       0.00000000057,    -0.00000000049,
      -0.00000000040,    -0.00000000055,
      -0.00000000026,     0.00000000016,
       0.00000000089,     0.00000000101,
       0.00000000097,    -0.00000000073,
      -0.00000000029,    -0.00000000045,
       0.00000000041,    -0.00000000109,
      -0.00000000086,    -0.00000000041,
      -0.00000000109,    -0.00000000038,
      -0.00000000038,     0.00000000086,
      -0.00000000049,    -0.00000000018,
      -0.00000000018,     0.00000000039,
      -0.00000000063,     0.00000000172,
       0.00000000136,     0.00000000062,
       0.00000000029,    -0.00000000053,
       0.00000000385,     0.00000000143,
       0.00000000142,    -0.00000000305,
      -0.00000000119,    -0.00000000065,
       0.00000000038,     0.00000000267,
       0.00000000197,    -0.00000000012,
       0.00000000000,    -0.00000000072,
       0.00000000262,    -0.00000000045,
      -0.00000000019,    -0.00000000193,
      -0.00000000070,     0.00000000003,
       0.00000000031,     0.00000000201,
       0.00000000148,    -0.00000000012,
      -0.00000000001,    -0.00000000054,
       0.00000000232,    -0.00000000033,
      -0.00000000012,    -0.00000000172,
      -0.00000000062,     0.00000000000,
       0.00000000168,     0.00000000224,
       0.00000000153,    -0.00000000098,
      -0.00000000028,    -0.00000000051,
       0.00000000215,    -0.00000000173,
      -0.00000000102,    -0.00000000146,
      -0.00000000048,     0.00000000030,
      -0.00000000033,     0.00000000025,
      -0.00000000038,    -0.00000000017,
      -0.00000000015,     0.00000000037,
       0.00000000217,     0.00000000280,
       0.00000000191,    -0.00000000127,
      -0.00000000037,    -0.00000000063,
       0.00000000261,    -0.00000000197,
      -0.00000000115,    -0.00000000179,
      -0.00000000059,     0.00000000033,
       0.00000000414,     0.00000000187,
       0.00000000123,    -0.00000000225,
      -0.00000000060,    -0.00000000038,
       0.00000000167,    -0.00000000412,
      -0.00000000225,    -0.00000000108,
      -0.00000000033,     0.00000000061,
      -0.00000000009,     0.00000000018,
      -0.00000000029,     0.00000000064,
       0.00000000035,     0.00000000019,
      -0.00000000071,     0.00000000005,
       0.00000000009,     0.00000000068,
       0.00000000618,     0.00000000265,
       0.00000000173,    -0.00000000336,
      -0.00000000091,    -0.00000000053,
       0.00000000228,    -0.00000000513,
      -0.00000000280,    -0.00000000149,
      -0.00000000046,     0.00000000075,
       0.00000000788,    -0.00000000060,
       0.00000000000,    -0.00000000381,
      -0.00000000091,    -0.00000000006,
      -0.00000000097,    -0.00000000763,
      -0.00000000366,     0.00000000025,
       0.00000000002,     0.00000000087,
      -0.00000000072,     0.00000000007,
       0.00000000001,     0.00000000035,
       0.00000000011,     0.00000000126,
       0.00000000061,    -0.00000000001,
      -0.00000000102,     0.00000000064,
       0.00000000022,     0.00000000043,
       0.00000000070,     0.00000000091,
       0.00000000037,    -0.00000000026,
       0.00000001258,    -0.00000000126,
      -0.00000000019,    -0.00000000606,
      -0.00000000144,    -0.00000000003,
      -0.00000000087,    -0.00000000998,
      -0.00000000482,     0.00000000009,
      -0.00000000004,     0.00000000114,
       0.00000001120,    -0.00000000725,
      -0.00000000255,    -0.00000000470,
      -0.00000000097,     0.00000000045,
      -0.00000000768,    -0.00000001033,
      -0.00000000422,     0.00000000282,
       0.00000000053,     0.00000000084,
      -0.00000000111,     0.00000000077,
       0.00000000028,     0.00000000046,
       0.00000000009,    -0.00000000005,
       0.00000000123,     0.00000000182,
       0.00000000075,    -0.00000000044,
      -0.00000000093,     0.00000000160,
       0.00000000049,     0.00000000035,
       0.00000000159,     0.00000000065,
       0.00000000021,    -0.00000000049,
       0.00000001833,    -0.00000001268,
      -0.00000000455,    -0.00000000760,
      -0.00000000154,     0.00000000082,
      -0.00000000964,    -0.00000001429,
      -0.00000000593,     0.00000000347,
       0.00000000063,     0.00000000120,
       0.00000000991,    -0.00000001800,
      -0.00000000549,    -0.00000000364,
      -0.00000000065,     0.00000000084,
      -0.00000001789,    -0.00000000788,
      -0.00000000267,     0.00000000550,
       0.00000000085,     0.00000000042,
      -0.00000000101,     0.00000000203,
       0.00000000062,     0.00000000036,
       0.00000000006,    -0.00000000010,
       0.00000000310,     0.00000000158,
       0.00000000056,    -0.00000000095,
      -0.00000000014,     0.00000000245,
       0.00000000060,     0.00000000011,
       0.00000000215,    -0.00000000033,
      -0.00000000009,    -0.00000000049,
       0.00000001569,    -0.00000003151,
      -0.00000000963,    -0.00000000558,
      -0.00000000094,     0.00000000148,
      -0.00000002411,    -0.00000001228,
      -0.00000000435,     0.00000000739,
       0.00000000113,     0.00000000073,
       0.00000000056,    -0.00000002717,
      -0.00000000654,    -0.00000000076,
      -0.00000000017,     0.00000000079,
      -0.00000002506,     0.00000000276,
       0.00000000061,     0.00000000578,
       0.00000000065,    -0.00000000011,
       0.00000000013,     0.00000000324,
       0.00000000076,     0.00000000001,
       0.00000000476,    -0.00000000016,
       0.00000000003,    -0.00000000112,
       0.00000000093,     0.00000000243,
       0.00000000045,    -0.00000000008,
       0.00000000152,    -0.00000000142,
      -0.00000000185,    -0.00000004705,
      -0.00000001112,    -0.00000000020,
      -0.00000000007,     0.00000000130,
      -0.00000003641,     0.00000000123,
      -0.00000000019,     0.00000000859,
       0.00000000101,     0.00000000005,
      -0.00000001198,    -0.00000002566,
      -0.00000000453,     0.00000000140,
      -0.00000001968,     0.00000001518,
       0.00000000252,     0.00000000270,
      -0.00000000041,     0.00000000009,
       0.00000000183,     0.00000000309,
       0.00000000049,    -0.00000000026,
       0.00000000439,    -0.00000000260,
      -0.00000000037,    -0.00000000070,
       0.00000000122,     0.00000000156,
      -0.00000000006,    -0.00000000120,
      -0.00000002487,    -0.00000004196,
      -0.00000000672,     0.00000000355,
       0.00000000027,     0.00000000052,
      -0.00000003301,     0.00000001953,
       0.00000000280,     0.00000000527,
       0.00000000041,    -0.00000000021,
      -0.00000001432,    -0.00000001369,
      -0.00000000182,     0.00000000103,
      -0.00000000379,     0.00000001414,
       0.00000000111,    -0.00000000052,
      -0.00000000032,     0.00000000024,
       0.00000000216,     0.00000000135,
       0.00000000010,    -0.00000000016,
       0.00000000184,    -0.00000000300,
       0.00000000059,     0.00000000093,
      -0.00000000080,     0.00000000041,
      -0.00000002723,    -0.00000001700,
      -0.00000000131,     0.00000000204,
      -0.00000001352,     0.00000002199,
       0.00000000164,     0.00000000103,
      -0.00000000424,    -0.00000000559,
      -0.00000000073,     0.00000000036,
       0.00000000417,    -0.00000000139,
      -0.00000000045,    -0.00000000084,
       0.00000000038,     0.00000000013,
       0.00000000015,    -0.00000000054,
       0.00000000032,     0.00000000064,
      -0.00000000021,     0.00000000103,
      -0.00000000436,    -0.00000000156,
      -0.00000000110,     0.00000000349,
       0.00000000049,    -0.00000000468,
      -0.00000000002,     0.00000000057,
      -0.00000000026,    -0.00000000759,
      -0.00000000064,     0.00000000040,
       0.00000000052,     0.00000000082,
       0.00000000680,    -0.00000000430,
       0.00000000027,     0.00000000051,
      -0.00000000379,    -0.00000000601,
       0.00000000044,    -0.00000000023,
      -0.00000000141,     0.00000000112,
       0.00000000150,    -0.00000000031,
      -0.00000000070,     0.00000000192,
      -0.00000000033,    -0.00000000012,
       0.00000000179,     0.00000000071,
       0.00000000083,    -0.00000000078,
       0.00000000037,     0.00000000093,
      -0.00000000066,     0.00000000057,
       0.00000000047,     0.00000000063,
       0.00000000029,    -0.00000000019,
       0.00000000153,     0.00000000178,
       0.00000000171,    -0.00000000125,
      -0.00000000051,    -0.00000000080,
      -0.00000000046,     0.00000000121,
       0.00000000096,     0.00000000046,
       0.00000000021,    -0.00000000037,
       0.00000000121,     0.00000000042,
       0.00000000041,    -0.00000000095,
      -0.00000000037,    -0.00000000019,
      -0.00000000029,     0.00000000079,
       0.00000000062,     0.00000000028,
       0.00000000013,    -0.00000000024,
       0.00000000094,     0.00000000036,
       0.00000000035,    -0.00000000074,
       0.00000000016,     0.00000000117,
       0.00000000086,    -0.00000000005,
       0.00000000114,    -0.00000000020,
      -0.00000000009,    -0.00000000084,
       0.00000000022,     0.00000000143,
       0.00000000105,    -0.00000000008,
      -0.00000000001,    -0.00000000038,
       0.00000000134,    -0.00000000019,
      -0.00000000007,    -0.00000000099,
       0.00000000113,     0.00000000152,
       0.00000000104,    -0.00000000066,
      -0.00000000019,    -0.00000000035,
       0.00000000145,    -0.00000000117,
      -0.00000000069,    -0.00000000098,
      -0.00000000032,     0.00000000020,
       0.00000000171,     0.00000000221,
       0.00000000150,    -0.00000000100,
      -0.00000000029,    -0.00000000049,
       0.00000000185,    -0.00000000139,
      -0.00000000081,    -0.00000000126,
      -0.00000000042,     0.00000000024,
       0.00000000309,     0.00000000141,
       0.00000000093,    -0.00000000168,
      -0.00000000045,    -0.00000000029,
       0.00000000123,    -0.00000000307,
      -0.00000000167,    -0.00000000079,
      -0.00000000024,     0.00000000045,
      -0.00000000022,     0.00000000048,
      -0.00000000054,     0.00000000003,
       0.00000000007,     0.00000000052,
       0.00000000498,     0.00000000214,
       0.00000000139,    -0.00000000270,
      -0.00000000073,    -0.00000000043,
       0.00000000173,    -0.00000000388,
      -0.00000000211,    -0.00000000112,
      -0.00000000034,     0.00000000057,
       0.00000000601,    -0.00000000043,
       0.00000000002,    -0.00000000290,
      -0.00000000069,    -0.00000000005,
      -0.00000000076,    -0.00000000579,
      -0.00000000277,     0.00000000021,
       0.00000000002,     0.00000000065,
      -0.00000000059,     0.00000000006,
       0.00000000001,     0.00000000028,
       0.00000000008,     0.00000000096,
      -0.00000000077,     0.00000000047,
       0.00000000016,     0.00000000033,
       0.00000000052,     0.00000000067,
       0.00000001003,    -0.00000000100,
      -0.00000000015,    -0.00000000482,
      -0.00000000114,    -0.00000000002,
      -0.00000000066,    -0.00000000764,
      -0.00000000368,     0.00000000007,
      -0.00000000003,     0.00000000087,
       0.00000000846,    -0.00000000540,
      -0.00000000189,    -0.00000000355,
      -0.00000000073,     0.00000000033,
      -0.00000000577,    -0.00000000769,
      -0.00000000313,     0.00000000213,
       0.00000000040,     0.00000000062,
      -0.00000000088,     0.00000000061,
       0.00000000022,     0.00000000037,
       0.00000000092,     0.00000000137,
       0.00000000057,    -0.00000000033,
      -0.00000000069,     0.00000000116,
       0.00000000035,     0.00000000026,
       0.00000000114,     0.00000000045,
       0.00000000014,    -0.00000000035,
       0.00000001421,    -0.00000000980,
      -0.00000000351,    -0.00000000588,
      -0.00000000119,     0.00000000064,
      -0.00000000725,    -0.00000001076,
      -0.00000000445,     0.00000000260,
       0.00000000047,     0.00000000090,
       0.00000000734,    -0.00000001303,
      -0.00000000397,    -0.00000000270,
      -0.00000000048,     0.00000000060,
      -0.00000001293,    -0.00000000560,
      -0.00000000186,     0.00000000397,
       0.00000000061,     0.00000000028,
      -0.00000000077,     0.00000000154,
       0.00000000047,     0.00000000027,
       0.00000000226,     0.00000000115,
       0.00000000041,    -0.00000000069,
      -0.00000000014,     0.00000000170,
       0.00000000042,     0.00000000009,
       0.00000000145,    -0.00000000026,
       0.00000001170,    -0.00000002336,
      -0.00000000714,    -0.00000000414,
      -0.00000000070,     0.00000000109,
      -0.00000001749,    -0.00000000894,
      -0.00000000315,     0.00000000535,
       0.00000000082,     0.00000000053,
       0.00000000067,    -0.00000001883,
      -0.00000000454,    -0.00000000062,
      -0.00000000013,     0.00000000055,
      -0.00000001709,     0.00000000206,
       0.00000000050,     0.00000000392,
       0.00000000044,    -0.00000000009,
       0.00000000008,     0.00000000231,
       0.00000000054,     0.00000000001,
       0.00000000329,    -0.00000000010,
       0.00000000002,    -0.00000000077,
       0.00000000055,     0.00000000163,
       0.00000000090,    -0.00000000094,
      -0.00000000117,    -0.00000003305,
      -0.00000000778,    -0.00000000016,
      -0.00000000005,     0.00000000091,
      -0.00000002511,     0.00000000077,
      -0.00000000014,     0.00000000591,
       0.00000000069,     0.00000000004,
      -0.00000000735,    -0.00000001691,
      -0.00000000302,     0.00000000081,
      -0.00000001224,     0.00000000985,
       0.00000000168,     0.00000000159,
       0.00000000120,     0.00000000205,
       0.00000000033,    -0.00000000017,
       0.00000000283,    -0.00000000166,
      -0.00000000024,    -0.00000000045,
       0.00000000068,     0.00000000104,
      -0.00000000017,    -0.00000000066,
      -0.00000001606,    -0.00000002753,
      -0.00000000438,     0.00000000230,
       0.00000000017,     0.00000000034,
      -0.00000002127,     0.00000001242,
       0.00000000178,     0.00000000338,
      -0.00000000806,    -0.00000000873,
      -0.00000000122,     0.00000000057,
      -0.00000000146,     0.00000000793,
       0.00000000063,    -0.00000000050,
       0.00000000125,     0.00000000081,
       0.00000000108,    -0.00000000171,
       0.00000000031,     0.00000000064,
      -0.00000000055,     0.00000000037,
      -0.00000001571,    -0.00000001018,
      -0.00000000076,     0.00000000117,
      -0.00000000792,     0.00000001247,
       0.00000000092,     0.00000000058,
      -0.00000000178,    -0.00000000381,
      -0.00000000050,     0.00000000024,
       0.00000000285,    -0.00000000202,
      -0.00000000031,    -0.00000000056,
       0.00000000019,     0.00000000039,
      -0.00000000014,     0.00000000063,
      -0.00000000089,    -0.00000000106,
      -0.00000000074,     0.00000000062,
       0.00000000033,    -0.00000000284,
      -0.00000000009,    -0.00000000461,
      -0.00000000040,     0.00000000024,
       0.00000000031,     0.00000000050,
       0.00000000429,    -0.00000000263,
       0.00000000016,     0.00000000032,
      -0.00000000228,    -0.00000000371,
      -0.00000000071,     0.00000000080,
       0.00000000107,     0.00000000009,
      -0.00000000054,     0.00000000142,
       0.00000000130,     0.00000000052,
       0.00000000053,    -0.00000000054,
       0.00000000020,     0.00000000062,
       0.00000000006,     0.00000000044,
       0.00000000021,    -0.00000000001,
       0.00000000032,     0.00000000151,
       0.00000000152,    -0.00000000019,
      -0.00000000005,    -0.00000000074,
      -0.00000000051,     0.00000000043,
       0.00000000035,     0.00000000049,
       0.00000000044,     0.00000000049,
       0.00000000047,    -0.00000000036,
       0.00000000019,     0.00000000025,
       0.00000000012,    -0.00000000007,
       0.00000000028,     0.00000000033,
       0.00000000032,    -0.00000000023,
      -0.00000000018,     0.00000000046,
       0.00000000036,     0.00000000017,
       0.00000000046,     0.00000000016,
       0.00000000016,    -0.00000000036,
       0.00000000043,     0.00000000020,
       0.00000000009,    -0.00000000017,
       0.00000000052,     0.00000000020,
       0.00000000019,    -0.00000000041,
       0.00000000011,     0.00000000078,
       0.00000000057,    -0.00000000003,
       0.00000000076,    -0.00000000013,
      -0.00000000006,    -0.00000000056,
       0.00000000018,     0.00000000113,
       0.00000000083,    -0.00000000007,
      -0.00000000001,    -0.00000000030,
       0.00000000094,    -0.00000000013,
      -0.00000000005,    -0.00000000069,
       0.00000000085,     0.00000000115,
       0.00000000078,    -0.00000000049,
       0.00000000109,    -0.00000000088,
      -0.00000000052,    -0.00000000073,
       0.00000000140,     0.00000000180,
       0.00000000122,    -0.00000000081,
      -0.00000000024,    -0.00000000040,
       0.00000000142,    -0.00000000106,
      -0.00000000062,    -0.00000000096,
      -0.00000000032,     0.00000000018,
       0.00000000240,     0.00000000111,
       0.00000000072,    -0.00000000130,
      -0.00000000035,    -0.00000000023,
       0.00000000095,    -0.00000000238,
      -0.00000000129,    -0.00000000061,
      -0.00000000018,     0.00000000035,
      -0.00000000017,     0.00000000038,
       0.00000000406,     0.00000000175,
       0.00000000113,    -0.00000000220,
      -0.00000000059,    -0.00000000035,
       0.00000000135,    -0.00000000302,
      -0.00000000164,    -0.00000000087,
      -0.00000000027,     0.00000000044,
       0.00000000465,    -0.00000000031,
       0.00000000002,    -0.00000000224,
      -0.00000000053,    -0.00000000004,
      -0.00000000060,    -0.00000000445,
      -0.00000000212,     0.00000000017,
       0.00000000002,     0.00000000050,
      -0.00000000048,     0.00000000005,
       0.00000000006,     0.00000000075,
      -0.00000000059,     0.00000000035,
       0.00000000039,     0.00000000049,
       0.00000000800,    -0.00000000079,
      -0.00000000012,    -0.00000000384,
      -0.00000000091,    -0.00000000002,
      -0.00000000051,    -0.00000000591,
      -0.00000000284,     0.00000000006,
      -0.00000000002,     0.00000000067,
       0.00000000641,    -0.00000000402,
      -0.00000000140,    -0.00000000269,
      -0.00000000056,     0.00000000024,
      -0.00000000435,    -0.00000000573,
      -0.00000000232,     0.00000000161,
       0.00000000030,     0.00000000046,
      -0.00000000069,     0.00000000048,
       0.00000000017,     0.00000000029,
       0.00000000070,     0.00000000103,
       0.00000000043,    -0.00000000025,
      -0.00000000052,     0.00000000083,
       0.00000000082,     0.00000000031,
       0.00000001096,    -0.00000000752,
      -0.00000000270,    -0.00000000452,
      -0.00000000092,     0.00000000049,
      -0.00000000545,    -0.00000000811,
      -0.00000000335,     0.00000000196,
       0.00000000035,     0.00000000068,
       0.00000000542,    -0.00000000939,
      -0.00000000285,    -0.00000000200,
      -0.00000000036,     0.00000000043,
      -0.00000000930,    -0.00000000395,
      -0.00000000129,     0.00000000286,
       0.00000000044,     0.00000000019,
      -0.00000000058,     0.00000000115,
       0.00000000035,     0.00000000020,
       0.00000000163,     0.00000000084,
       0.00000000029,    -0.00000000050,
      -0.00000000013,     0.00000000118,
       0.00000000097,    -0.00000000020,
       0.00000000866,    -0.00000001716,
      -0.00000000524,    -0.00000000305,
      -0.00000000051,     0.00000000080,
      -0.00000001263,    -0.00000000649,
      -0.00000000228,     0.00000000386,
       0.00000000059,     0.00000000038,
       0.00000000068,    -0.00000001298,
      -0.00000000314,    -0.00000000050,
      -0.00000000011,     0.00000000038,
      -0.00000001156,     0.00000000152,
       0.00000000041,     0.00000000263,
       0.00000000005,     0.00000000163,
       0.00000000038,     0.00000000001,
       0.00000000225,    -0.00000000006,
       0.00000000001,    -0.00000000053,
       0.00000000031,     0.00000000109,
       0.00000000052,    -0.00000000061,
      -0.00000000070,    -0.00000002298,
      -0.00000000540,    -0.00000000013,
      -0.00000000004,     0.00000000063,
      -0.00000001719,     0.00000000047,
      -0.00000000010,     0.00000000404,
       0.00000000047,     0.00000000003,
      -0.00000000442,    -0.00000001107,
      -0.00000000201,     0.00000000045,
      -0.00000000748,     0.00000000634,
       0.00000000112,     0.00000000091,
       0.00000000077,     0.00000000134,
       0.00000000021,    -0.00000000011,
       0.00000000181,    -0.00000000104,
       0.00000000037,     0.00000000069,
      -0.00000000019,    -0.00000000036,
      -0.00000001022,    -0.00000001786,
      -0.00000000283,     0.00000000146,
       0.00000000011,     0.00000000022,
      -0.00000001358,     0.00000000780,
       0.00000000112,     0.00000000214,
      -0.00000000441,    -0.00000000555,
      -0.00000000082,     0.00000000030,
      -0.00000000029,     0.00000000433,
       0.00000000035,    -0.00000000042,
       0.00000000071,     0.00000000048,
       0.00000000063,    -0.00000000095,
       0.00000000016,     0.00000000043,
      -0.00000000037,     0.00000000030,
      -0.00000000883,    -0.00000000599,
      -0.00000000043,     0.00000000065,
      -0.00000000457,     0.00000000690,
       0.00000000050,     0.00000000032,
      -0.00000000059,    -0.00000000255,
      -0.00000000034,     0.00000000016,
       0.00000000192,    -0.00000000192,
      -0.00000000020,    -0.00000000037,
       0.00000000055,    -0.00000000070,
      -0.00000000049,    -0.00000000052,
       0.00000000022,    -0.00000000168,
      -0.00000000001,    -0.00000000273,
       0.00000000264,    -0.00000000156,
      -0.00000000133,    -0.00000000224,
      -0.00000000033,     0.00000000056,
       0.00000000075,     0.00000000024,
      -0.00000000040,     0.00000000103,
       0.00000000092,     0.00000000038,
       0.00000000033,    -0.00000000036,
       0.00000000013,     0.00000000006,
      -0.00000000039,     0.00000000128,
       0.00000000134,     0.00000000053,
       0.00000000032,    -0.00000000068,
       0.00000000034,     0.00000000016,
       0.00000000007,    -0.00000000013,
       0.00000000036,     0.00000000014,
       0.00000000008,     0.00000000059,
       0.00000000057,    -0.00000000010,
       0.00000000014,     0.00000000093,
       0.00000000068,    -0.00000000006,
      -0.00000000001,    -0.00000000025,
       0.00000000072,    -0.00000000010,
      -0.00000000004,    -0.00000000053,
       0.00000000067,     0.00000000091,
       0.00000000062,    -0.00000000039,
       0.00000000085,    -0.00000000070,
      -0.00000000041,    -0.00000000057,
       0.00000000116,     0.00000000150,
       0.00000000101,    -0.00000000067,
      -0.00000000019,    -0.00000000033,
       0.00000000112,    -0.00000000084,
      -0.00000000049,    -0.00000000076,
       0.00000000189,     0.00000000088,
       0.00000000058,    -0.00000000102,
       0.00000000074,    -0.00000000188,
      -0.00000000102,    -0.00000000047,
       0.00000000331,     0.00000000143,
       0.00000000092,    -0.00000000179,
      -0.00000000048,    -0.00000000028,
       0.00000000107,    -0.00000000239,
      -0.00000000130,    -0.00000000069,
      -0.00000000021,     0.00000000035,
       0.00000000361,    -0.00000000022,
       0.00000000003,    -0.00000000174,
      -0.00000000047,    -0.00000000343,
      -0.00000000163,     0.00000000014,
       0.00000000005,     0.00000000058,
      -0.00000000044,     0.00000000026,
       0.00000000030,     0.00000000036,
       0.00000000634,    -0.00000000061,
      -0.00000000009,    -0.00000000304,
      -0.00000000072,    -0.00000000001,
      -0.00000000039,    -0.00000000458,
      -0.00000000220,     0.00000000004,
      -0.00000000002,     0.00000000052,
       0.00000000484,    -0.00000000298,
      -0.00000000103,    -0.00000000203,
      -0.00000000042,     0.00000000018,
      -0.00000000326,    -0.00000000426,
      -0.00000000172,     0.00000000121,
       0.00000000023,     0.00000000034,
      -0.00000000054,     0.00000000037,
       0.00000000013,     0.00000000022,
       0.00000000052,     0.00000000078,
       0.00000000032,    -0.00000000019,
      -0.00000000039,     0.00000000059,
       0.00000000058,     0.00000000021,
       0.00000000838,    -0.00000000573,
      -0.00000000205,    -0.00000000345,
      -0.00000000070,     0.00000000037,
      -0.00000000408,    -0.00000000609,
      -0.00000000251,     0.00000000146,
       0.00000000027,     0.00000000051,
       0.00000000399,    -0.00000000672,
      -0.00000000204,    -0.00000000147,
      -0.00000000027,     0.00000000031,
      -0.00000000665,    -0.00000000277,
      -0.00000000089,     0.00000000204,
      -0.00000000043,     0.00000000085,
       0.00000000026,     0.00000000015,
       0.00000000117,     0.00000000061,
       0.00000000021,    -0.00000000036,
      -0.00000000011,     0.00000000081,
       0.00000000064,    -0.00000000015,
       0.00000000636,    -0.00000001249,
      -0.00000000381,    -0.00000000223,
      -0.00000000037,     0.00000000058,
      -0.00000000906,    -0.00000000468,
      -0.00000000164,     0.00000000277,
       0.00000000042,     0.00000000027,
       0.00000000064,    -0.00000000887,
      -0.00000000215,    -0.00000000040,
      -0.00000000775,     0.00000000112,
       0.00000000032,     0.00000000175,
       0.00000000003,     0.00000000113,
       0.00000000027,     0.00000000001,
       0.00000000153,    -0.00000000004,
       0.00000000017,     0.00000000072,
       0.00000000029,    -0.00000000040,
      -0.00000000040,    -0.00000001581,
      -0.00000000371,    -0.00000000010,
      -0.00000000003,     0.00000000043,
      -0.00000001167,     0.00000000026,
      -0.00000000008,     0.00000000274,
      -0.00000000258,    -0.00000000720,
      -0.00000000133,     0.00000000023,
      -0.00000000449,     0.00000000404,
       0.00000000074,     0.00000000050,
       0.00000000049,     0.00000000087,
       0.00000000115,    -0.00000000065,
       0.00000000019,     0.00000000046,
      -0.00000000640,    -0.00000001144,
      -0.00000000180,     0.00000000092,
       0.00000000007,     0.00000000014,
      -0.00000000858,     0.00000000483,
       0.00000000070,     0.00000000134,
      -0.00000000232,    -0.00000000351,
      -0.00000000054,     0.00000000015,
       0.00000000024,     0.00000000228,
       0.00000000019,    -0.00000000034,
       0.00000000039,     0.00000000028,
       0.00000000036,    -0.00000000051,
      -0.00000000479,    -0.00000000346,
      -0.00000000024,     0.00000000035,
      -0.00000000259,     0.00000000369,
      -0.00000000006,    -0.00000000168,
       0.00000000128,    -0.00000000157,
       0.00000000100,    -0.00000000045,
      -0.00000000031,    -0.00000000085,
       0.00000000014,    -0.00000000097,
       0.00000000003,    -0.00000000158,
       0.00000000159,    -0.00000000089,
      -0.00000000075,    -0.00000000132,
      -0.00000000013,     0.00000000039,
       0.00000000051,     0.00000000026,
      -0.00000000029,     0.00000000073,
       0.00000000064,     0.00000000027,
      -0.00000000024,    -0.00000000028,
       0.00000000009,     0.00000000012,
      -0.00000000178,     0.00000000172,
       0.00000000170,     0.00000000210,
       0.00000000130,    -0.00000000097,
       0.00000000028,     0.00000000013,
       0.00000000006,     0.00000000047,
       0.00000000046,    -0.00000000008,
       0.00000000012,     0.00000000079,
       0.00000000057,    -0.00000000005,
       0.00000000000,    -0.00000000021,
       0.00000000058,    -0.00000000008,
       0.00000000054,     0.00000000074,
       0.00000000050,    -0.00000000031,
       0.00000000068,    -0.00000000056,
      -0.00000000033,    -0.00000000046,
       0.00000000096,     0.00000000124,
       0.00000000084,    -0.00000000056,
      -0.00000000016,    -0.00000000028,
       0.00000000090,    -0.00000000068,
      -0.00000000039,    -0.00000000061,
       0.00000000150,     0.00000000071,
       0.00000000046,    -0.00000000081,
       0.00000000058,    -0.00000000149,
      -0.00000000081,    -0.00000000037,
       0.00000000269,     0.00000000117,
       0.00000000075,    -0.00000000145,
      -0.00000000039,    -0.00000000023,
       0.00000000085,    -0.00000000190,
      -0.00000000103,    -0.00000000055,
       0.00000000280,    -0.00000000015,
       0.00000000003,    -0.00000000135,
      -0.00000000038,    -0.00000000264,
      -0.00000000125,     0.00000000012,
      -0.00000000034,     0.00000000019,
       0.00000000500,    -0.00000000047,
      -0.00000000007,    -0.00000000239,
      -0.00000000057,    -0.00000000001,
      -0.00000000030,    -0.00000000354,
      -0.00000000170,     0.00000000003,
       0.00000000364,    -0.00000000220,
      -0.00000000076,    -0.00000000153,
      -0.00000000244,    -0.00000000315,
      -0.00000000126,     0.00000000091,
      -0.00000000042,     0.00000000028,
       0.00000000010,     0.00000000017,
       0.00000000039,     0.00000000058,
      -0.00000000029,     0.00000000042,
       0.00000000041,     0.00000000014,
       0.00000000636,    -0.00000000432,
      -0.00000000155,    -0.00000000261,
      -0.00000000053,     0.00000000028,
      -0.00000000303,    -0.00000000455,
      -0.00000000187,     0.00000000109,
       0.00000000020,     0.00000000038,
       0.00000000292,    -0.00000000478,
      -0.00000000145,    -0.00000000108,
      -0.00000000472,    -0.00000000192,
      -0.00000000060,     0.00000000145,
      -0.00000000032,     0.00000000062,
       0.00000000019,     0.00000000011,
       0.00000000084,     0.00000000044,
      -0.00000000010,     0.00000000055,
       0.00000000042,    -0.00000000012,
       0.00000000462,    -0.00000000899,
      -0.00000000274,    -0.00000000161,
      -0.00000000027,     0.00000000042,
      -0.00000000645,    -0.00000000336,
      -0.00000000117,     0.00000000197,
       0.00000000030,     0.00000000019,
       0.00000000056,    -0.00000000602,
      -0.00000000146,    -0.00000000031,
      -0.00000000513,     0.00000000082,
       0.00000000025,     0.00000000115,
       0.00000000001,     0.00000000078,
       0.00000000103,    -0.00000000002,
       0.00000000008,     0.00000000048,
      -0.00000000020,    -0.00000001076,
      -0.00000000252,    -0.00000000008,
      -0.00000000002,     0.00000000029,
      -0.00000000785,     0.00000000013,
      -0.00000000006,     0.00000000184,
      -0.00000000146,    -0.00000000465,
      -0.00000000087,     0.00000000010,
      -0.00000000262,     0.00000000255,
       0.00000000048,     0.00000000025,
       0.00000000030,     0.00000000055,
       0.00000000072,    -0.00000000040,
      -0.00000000393,    -0.00000000724,
      -0.00000000113,     0.00000000057,
      -0.00000000536,     0.00000000294,
       0.00000000042,     0.00000000083,
      -0.00000000115,    -0.00000000221,
       0.00000000042,     0.00000000113,
      -0.00000000248,    -0.00000000195,
      -0.00000000143,     0.00000000189,
       0.00000000014,    -0.00000000109,
       0.00000000084,    -0.00000000118,
       0.00000000101,    -0.00000000028,
      -0.00000000019,    -0.00000000083,
       0.00000000009,    -0.00000000054,
       0.00000000004,    -0.00000000088,
       0.00000000093,    -0.00000000049,
      -0.00000000041,    -0.00000000076,
       0.00000000034,     0.00000000023,
       0.00000000043,     0.00000000019,
       0.00000000140,    -0.00000000038,
      -0.00000000033,    -0.00000000176,
      -0.00000000101,     0.00000000014,
       0.00000000036,    -0.00000000007,
      -0.00000000003,    -0.00000000021,
      -0.00000001026,     0.00000000278,
       0.00000000237,     0.00000001278,
       0.00000000789,    -0.00000000105,
       0.00000000031,     0.00000000033,
       0.00000000037,    -0.00000000031,
       0.00000000032,    -0.00000000032,
      -0.00000000031,    -0.00000000037,
       0.00000000010,     0.00000000066,
       0.00000000048,    -0.00000000004,
       0.00000000047,    -0.00000000007,
       0.00000000043,     0.00000000060,
       0.00000000040,    -0.00000000025,
       0.00000000055,    -0.00000000045,
      -0.00000000027,    -0.00000000037,
       0.00000000080,     0.00000000103,
       0.00000000069,    -0.00000000046,
      -0.00000000013,    -0.00000000023,
       0.00000000073,    -0.00000000055,
      -0.00000000032,    -0.00000000049,
       0.00000000119,     0.00000000057,
       0.00000000037,    -0.00000000064,
       0.00000000045,    -0.00000000117,
      -0.00000000064,    -0.00000000029,
       0.00000000217,     0.00000000095,
       0.00000000061,    -0.00000000117,
      -0.00000000031,    -0.00000000018,
       0.00000000068,    -0.00000000151,
      -0.00000000081,    -0.00000000043,
       0.00000000216,    -0.00000000010,
       0.00000000003,    -0.00000000104,
      -0.00000000030,    -0.00000000202,
      -0.00000000096,     0.00000000010,
       0.00000000390,    -0.00000000036,
      -0.00000000005,    -0.00000000186,
      -0.00000000044,    -0.00000000001,
      -0.00000000022,    -0.00000000273,
      -0.00000000130,     0.00000000003,
       0.00000000272,    -0.00000000161,
      -0.00000000055,    -0.00000000114,
      -0.00000000181,    -0.00000000231,
      -0.00000000092,     0.00000000068,
      -0.00000000032,     0.00000000021,
       0.00000000029,     0.00000000043,
      -0.00000000021,     0.00000000030,
       0.00000000478,    -0.00000000322,
      -0.00000000115,    -0.00000000196,
      -0.00000000039,     0.00000000021,
      -0.00000000224,    -0.00000000338,
      -0.00000000138,     0.00000000080,
       0.00000000213,    -0.00000000337,
      -0.00000000102,    -0.00000000079,
      -0.00000000331,    -0.00000000132,
      -0.00000000040,     0.00000000102,
      -0.00000000023,     0.00000000045,
       0.00000000059,     0.00000000031,
       0.00000000333,    -0.00000000641,
      -0.00000000195,    -0.00000000116,
      -0.00000000019,     0.00000000030,
      -0.00000000455,    -0.00000000239,
      -0.00000000083,     0.00000000139,
       0.00000000047,    -0.00000000405,
      -0.00000000099,    -0.00000000024,
      -0.00000000336,     0.00000000059,
       0.00000000020,     0.00000000074,
       0.00000000001,     0.00000000053,
       0.00000000069,    -0.00000000001,
      -0.00000000008,    -0.00000000724,
      -0.00000000169,    -0.00000000006,
      -0.00000000001,     0.00000000020,
      -0.00000000523,     0.00000000006,
      -0.00000000004,     0.00000000122,
      -0.00000000079,    -0.00000000297,
      -0.00000000057,     0.00000000004,
      -0.00000000149,     0.00000000159,
       0.00000000018,     0.00000000035,
       0.00000000044,    -0.00000000024,
      -0.00000000237,    -0.00000000451,
      -0.00000000070,     0.00000000034,
      -0.00000000330,     0.00000000176,
       0.00000000025,     0.00000000051,
      -0.00000000052,    -0.00000000138,
       0.00000000044,     0.00000000052,
      -0.00000000120,    -0.00000000106,
      -0.00000000076,     0.00000000090,
       0.00000000019,    -0.00000000069,
       0.00000000055,    -0.00000000084,
       0.00000000084,    -0.00000000016,
      -0.00000000012,    -0.00000000068,
       0.00000000004,    -0.00000000047,
       0.00000000053,    -0.00000000025,
      -0.00000000021,    -0.00000000042,
       0.00000000041,    -0.00000000003,
       0.00000000035,     0.00000000049,
       0.00000000033,    -0.00000000020,
       0.00000000044,    -0.00000000037,
      -0.00000000022,    -0.00000000029,
       0.00000000066,     0.00000000086,
       0.00000000057,    -0.00000000038,
      -0.00000000011,    -0.00000000019,
       0.00000000059,    -0.00000000044,
      -0.00000000026,    -0.00000000040,
       0.00000000094,     0.00000000046,
       0.00000000030,    -0.00000000050,
       0.00000000035,    -0.00000000093,
      -0.00000000050,    -0.00000000022,
       0.00000000174,     0.00000000076,
       0.00000000049,    -0.00000000094,
      -0.00000000025,    -0.00000000015,
       0.00000000054,    -0.00000000119,
      -0.00000000064,    -0.00000000034,
       0.00000000165,    -0.00000000006,
       0.00000000003,    -0.00000000079,
      -0.00000000023,    -0.00000000153,
      -0.00000000073,     0.00000000008,
       0.00000000302,    -0.00000000027,
      -0.00000000004,    -0.00000000144,
      -0.00000000034,    -0.00000000001,
      -0.00000000017,    -0.00000000209,
      -0.00000000100,     0.00000000002,
       0.00000000202,    -0.00000000116,
      -0.00000000040,    -0.00000000085,
      -0.00000000133,    -0.00000000168,
      -0.00000000066,     0.00000000050,
       0.00000000021,     0.00000000032,
       0.00000000355,    -0.00000000238,
      -0.00000000085,    -0.00000000145,
      -0.00000000029,     0.00000000015,
      -0.00000000164,    -0.00000000249,
      -0.00000000102,     0.00000000059,
       0.00000000154,    -0.00000000235,
      -0.00000000071,    -0.00000000057,
      -0.00000000231,    -0.00000000089,
      -0.00000000026,     0.00000000071,
       0.00000000041,     0.00000000022,
       0.00000000238,    -0.00000000452,
      -0.00000000137,    -0.00000000082,
      -0.00000000014,     0.00000000021,
      -0.00000000318,    -0.00000000169,
      -0.00000000058,     0.00000000097,
       0.00000000039,    -0.00000000270,
      -0.00000000066,    -0.00000000019,
      -0.00000000217,     0.00000000043,
       0.00000000015,     0.00000000047,
      -0.00000000001,    -0.00000000482,
      -0.00000000112,    -0.00000000005,
      -0.00000000345,     0.00000000001,
      -0.00000000003,     0.00000000080,
      -0.00000000039,    -0.00000000189,
      -0.00000000081,     0.00000000098,
      -0.00000000140,    -0.00000000277,
      -0.00000000042,     0.00000000020,
      -0.00000000201,     0.00000000103,
      -0.00000000019,    -0.00000000085,
       0.00000000039,     0.00000000020,
      -0.00000000051,    -0.00000000056,
      -0.00000000039,     0.00000000038,
       0.00000000018,    -0.00000000043,
       0.00000000035,    -0.00000000057,
      -0.00000000007,    -0.00000000051,
       0.00000000028,     0.00000000040,
       0.00000000036,    -0.00000000030,
       0.00000000054,     0.00000000070,
       0.00000000047,    -0.00000000031,
      -0.00000000009,    -0.00000000015,
       0.00000000048,    -0.00000000036,
      -0.00000000021,    -0.00000000032,
       0.00000000073,     0.00000000036,
       0.00000000024,    -0.00000000039,
       0.00000000027,    -0.00000000073,
      -0.00000000039,    -0.00000000017,
       0.00000000138,     0.00000000061,
       0.00000000039,    -0.00000000074,
      -0.00000000020,    -0.00000000012,
       0.00000000042,    -0.00000000093,
      -0.00000000050,    -0.00000000027,
       0.00000000126,    -0.00000000004,
       0.00000000003,    -0.00000000060,
      -0.00000000018,    -0.00000000116,
      -0.00000000055,     0.00000000006,
       0.00000000232,    -0.00000000020,
      -0.00000000003,    -0.00000000110,
      -0.00000000026,    -0.00000000001,
      -0.00000000012,    -0.00000000159,
      -0.00000000076,     0.00000000001,
       0.00000000149,    -0.00000000083,
      -0.00000000028,    -0.00000000063,
      -0.00000000097,    -0.00000000121,
      -0.00000000047,     0.00000000037,
       0.00000000262,    -0.00000000174,
      -0.00000000062,    -0.00000000107,
      -0.00000000021,     0.00000000011,
      -0.00000000119,    -0.00000000182,
      -0.00000000074,     0.00000000043,
       0.00000000111,    -0.00000000162,
      -0.00000000049,    -0.00000000041,
      -0.00000000159,    -0.00000000060,
      -0.00000000017,     0.00000000049,
       0.00000000168,    -0.00000000315,
      -0.00000000096,    -0.00000000058,
      -0.00000000010,     0.00000000015,
      -0.00000000220,    -0.00000000119,
      -0.00000000041,     0.00000000067,
       0.00000000031,    -0.00000000178,
      -0.00000000044,    -0.00000000014,
      -0.00000000139,     0.00000000031,
       0.00000000002,    -0.00000000317,
      -0.00000000073,    -0.00000000004,
      -0.00000000225,    -0.00000000001,
      -0.00000000002,     0.00000000052,
      -0.00000000017,    -0.00000000118,
      -0.00000000041,     0.00000000060,
      -0.00000000080,    -0.00000000167,
      -0.00000000120,     0.00000000059,
      -0.00000000004,    -0.00000000052,
       0.00000000022,    -0.00000000038,
       0.00000000022,     0.00000000032,
       0.00000000028,    -0.00000000024,
       0.00000000044,     0.00000000057,
       0.00000000038,    -0.00000000025,
      -0.00000000007,    -0.00000000013,
       0.00000000039,    -0.00000000029,
       0.00000000057,     0.00000000029,
       0.00000000019,    -0.00000000031,
       0.00000000021,    -0.00000000056,
       0.00000000109,     0.00000000048,
       0.00000000031,    -0.00000000059,
      -0.00000000016,    -0.00000000009,
       0.00000000033,    -0.00000000073,
      -0.00000000039,    -0.00000000021,
       0.00000000095,    -0.00000000002,
       0.00000000003,    -0.00000000046,
      -0.00000000014,    -0.00000000087,
       0.00000000176,    -0.00000000015,
      -0.00000000002,    -0.00000000084,
      -0.00000000020,    -0.00000000001,
      -0.00000000009,    -0.00000000120,
      -0.00000000057,     0.00000000001,
       0.00000000109,    -0.00000000059,
      -0.00000000020,    -0.00000000046,
      -0.00000000070,    -0.00000000086,
      -0.00000000033,     0.00000000027,
       0.00000000191,    -0.00000000126,
      -0.00000000045,    -0.00000000078,
      -0.00000000016,     0.00000000008,
      -0.00000000085,    -0.00000000132,
      -0.00000000054,     0.00000000031,
       0.00000000079,    -0.00000000111,
      -0.00000000033,    -0.00000000030,
      -0.00000000109,    -0.00000000039,
       0.00000000118,    -0.00000000217,
      -0.00000000066,    -0.00000000040,
      -0.00000000151,    -0.00000000082,
      -0.00000000028,     0.00000000046,
       0.00000000024,    -0.00000000116,
      -0.00000000087,     0.00000000022,
       0.00000000004,    -0.00000000206,
      -0.00000000048,    -0.00000000003,
      -0.00000000145,    -0.00000000002,
      -0.00000000006,    -0.00000000074,
      -0.00000000019,     0.00000000036,
      -0.00000000044,    -0.00000000099,
      -0.00000000070,     0.00000000032,
       0.00000000035,     0.00000000046,
       0.00000000031,    -0.00000000020,
       0.00000000031,    -0.00000000023,
       0.00000000044,     0.00000000023,
       0.00000000016,    -0.00000000044,
       0.00000000085,     0.00000000038,
       0.00000000024,    -0.00000000046,
      -0.00000000012,    -0.00000000007,
       0.00000000026,    -0.00000000056,
       0.00000000071,    -0.00000000001,
      -0.00000000011,    -0.00000000064,
       0.00000000133,    -0.00000000011,
      -0.00000000001,    -0.00000000063,
      -0.00000000006,    -0.00000000089,
       0.00000000079,    -0.00000000041,
      -0.00000000050,    -0.00000000061,
       0.00000000139,    -0.00000000090,
      -0.00000000032,    -0.00000000056,
      -0.00000000061,    -0.00000000095,
      -0.00000000038,     0.00000000022,
       0.00000000056,    -0.00000000075,
      -0.00000000073,    -0.00000000025,
       0.00000000082,    -0.00000000148,
      -0.00000000045,    -0.00000000028,
      -0.00000000102,    -0.00000000057,
      -0.00000000019,     0.00000000031,
       0.00000000019,    -0.00000000075,
      -0.00000000054,     0.00000000015,
       0.00000000004,    -0.00000000132,
      -0.00000000092,    -0.00000000003,
      -0.00000000023,    -0.00000000057,
      -0.00000000040,     0.00000000017,
       0.00000000025,    -0.00000000016,
       0.00000000034,     0.00000000018,
       0.00000000066,     0.00000000030,
       0.00000000019,    -0.00000000035,
       0.00000000020,    -0.00000000043,
       0.00000000053,     0.00000000000,
      -0.00000000008,    -0.00000000047,
       0.00000000099,    -0.00000000007,
      -0.00000000001,    -0.00000000047,
      -0.00000000005,    -0.00000000066,
       0.00000000057,    -0.00000000029,
      -0.00000000036,    -0.00000000042,
       0.00000000099,    -0.00000000064,
      -0.00000000023,    -0.00000000040,
      -0.00000000043,    -0.00000000068,
       0.00000000040,    -0.00000000050,
      -0.00000000049,    -0.00000000016,
       0.00000000057,    -0.00000000100,
      -0.00000000030,    -0.00000000019,
      -0.00000000069,    -0.00000000039,
       0.00000000014,    -0.00000000048,
       0.00000000004,    -0.00000000083,
      -0.00000000058,    -0.00000000002,
       0.00000000014,    -0.00000000027,
       0.00000000073,    -0.00000000005,
      -0.00000000003,    -0.00000000049,
       0.00000000041,    -0.00000000020,
      -0.00000000025,    -0.00000000029,
       0.00000000071,    -0.00000000045,
      -0.00000000016,    -0.00000000029,
      -0.00000000030,    -0.00000000048,
       0.00000000028,    -0.00000000033,
       0.00000000039,    -0.00000000067,
      -0.00000000046,    -0.00000000026,
       0.00000000050,    -0.00000000031,
      -0.00000000021,    -0.00000000033,
      -0.00000001806,    -0.00000001040,
       0.00000000086,    -0.00000000086,
       0.00000000003,     0.00000000004,
      -0.00000001174,    -0.00000000306,
       0.00000000008,    -0.00000000012,
       0.00000000174,    -0.00000000992,
      -0.00000000003,    -0.00000000011,
       0.00000000025,     0.00000000029,
      -0.00000000002,     0.00000000003,
      -0.00000000069,    -0.00000000060,
      -0.00000002184,    -0.00000004643,
       0.00000000340,    -0.00000000184,
      -0.00000005966,    -0.00000000110,
      -0.00000000094,    -0.00000006092,
       0.00000000035,     0.00000000083,
      -0.00000008344,    -0.00000014007,
       0.00000001077,    -0.00000000736,
       0.00000000029,     0.00000000041,
      -0.00000017970,    -0.00000000604,
       0.00000000097,     0.00000000049,
      -0.00000000418,    -0.00000018578,
       0.00000000045,    -0.00000000124,
      -0.00000000165,    -0.00000000359,
       0.00000000227,     0.00000000239,
      -0.00000108962,     0.00000053133,
      -0.00000005123,    -0.00000008549,
       0.00000000370,    -0.00000000182,
       0.00000468021,     0.00000790415,
      -0.00000061937,     0.00000038886,
      -0.00000001497,    -0.00000002527,
      -0.00000590672,     0.00000004912,
      -0.00000000377,    -0.00000000970,
       0.00000000024,    -0.00000000086,
       0.00000005206,    -0.00000583999,
      -0.00000001000,     0.00000001146,
      -0.00000000093,    -0.00000000026,
      -0.00000006083,    -0.00000001919,
      -0.00000000766,     0.00000001075,
       0.00000000039,     0.00000000073,
       0.00000005078,    -0.00000004193,
      -0.00000001110,    -0.00000000609,
      -0.00000000075,     0.00000000041,
       0.00000001531,    -0.00000006389,
       0.00000000297,    -0.00000000123,
       0.00000088874,     0.00000000046,
       0.00000000094,     0.00000089030,
       0.00000010016,     0.00000000526,
      -0.00000000672,     0.00000000948,
       0.00000120316,     0.00000000058,
       0.00000000209,     0.00000120568,
      -0.00000000128,     0.00000000051,
       0.00000001455,     0.00000011807,
      -0.00000000177,    -0.00000000045,
       0.00000170832,     0.00000000107,
       0.00000000138,     0.00000171098,
       0.00000000068,     0.00000000046,
       0.00000002145,    -0.00000001928,
       0.00000213898,     0.00000000062,
       0.00000000273,     0.00000214398,
      -0.00000000146,     0.00000000090,
       0.00000000012,     0.00000000011,
       0.00000000028,    -0.00000000077,
       0.00000000079,     0.00000000027,
      -0.00000000017,    -0.00000000045,
       0.00000000045,    -0.00000000017,
       0.00000000013,    -0.00000000042,
      -0.00000000032,     0.00000000046,
      -0.00000000046,    -0.00000000033,
      -0.00000000039,    -0.00000000015,
       0.00000000017,    -0.00000000038,
       0.00000000219,     0.00000000058,
       0.00000000012,     0.00000000113,
       0.00000000263,    -0.00000000511,
       0.00000000119,     0.00000000048,
       0.00000000529,     0.00000000262,
      -0.00000000041,     0.00000000130,
       0.00000000334,     0.00000000101,
      -0.00000000110,     0.00000000323,
       0.00000000014,     0.00000000083,
      -0.00000000030,    -0.00000000021,
       0.00000000034,     0.00000000792,
       0.00000000155,     0.00000000026,
      -0.00000000031,     0.00000000022,
       0.00000000004,    -0.00000000070,
       0.00000000181,    -0.00000000004,
       0.00000000019,     0.00000000078,
      -0.00000000041,    -0.00000000076,
       0.00000000532,     0.00000000431,
      -0.00000000117,     0.00000000108,
      -0.00000000067,    -0.00000000024,
      -0.00000000023,     0.00000000067,
      -0.00000000109,     0.00000000026,
      -0.00000002277,    -0.00000009668,
       0.00000000277,     0.00000000169,
       0.00000000243,    -0.00000000293,
       0.00000000018,     0.00000000036,
      -0.00000000289,    -0.00000000234,
       0.00000000035,    -0.00000000018,
       0.00000000009,    -0.00000000176,
      -0.00000000040,     0.00000000085,
       0.00000000304,    -0.00000000023,
      -0.00000000019,     0.00000000031,
      -0.00000000036,    -0.00000000025,
      -0.00000000024,     0.00000000036,
      -0.00000000291,     0.00000000698,
       0.00000000071,     0.00000000044,
      -0.00000000114,     0.00000000058,
       0.00000000060,     0.00000000113,
      -0.00000002474,    -0.00000001604,
      -0.00000000074,    -0.00000000108,
      -0.00000005494,     0.00000003811,
       0.00000000418,     0.00000000235,
       0.00000000540,     0.00000000174,
       0.00000000182,    -0.00000000556,
       0.00000000009,    -0.00000000050,
      -0.00000000146,    -0.00000000388,
      -0.00000000062,     0.00000000102,
       0.00000000102,     0.00000000060,
       0.00000000080,    -0.00000000013,
      -0.00000000035,    -0.00000000044,
      -0.00000000432,     0.00000000342,
       0.00000000036,     0.00000000025,
       0.00000000240,     0.00000000074,
       0.00000000073,    -0.00000000263,
      -0.00000000153,    -0.00000000278,
      -0.00000000061,     0.00000000131,
       0.00000000131,     0.00000000059,
       0.00000000063,    -0.00000000022,
       0.00000000064,    -0.00000000023,
      -0.00000000038,    -0.00000000024,
      -0.00000000023,     0.00000000038,
       0.00000000106,    -0.00000000003,
      -0.00000000079,    -0.00000000056,
      -0.00000000054,     0.00000000080,
       0.00000000456,     0.00000000003,
       0.00000000092,    -0.00000000093,
      -0.00000000019,    -0.00000000085,
      -0.00000000087,     0.00000000018,
      -0.00000000046,    -0.00000000307,
      -0.00000000156,     0.00000000156,
       0.00000000157,     0.00000000153,
       0.00000000039,    -0.00000000110,
      -0.00000000066,     0.00000000053,
       0.00000000057,     0.00000000042,
      -0.00000000020,    -0.00000000048,
       0.00000000022,     0.00000000066,
       0.00000000044,    -0.00000000054,
      -0.00000000054,    -0.00000000041,
      -0.00000000058,     0.00000000086,
       0.00000000264,     0.00000000179,
       0.00000000004,    -0.00000000432,
      -0.00000000429,     0.00000000002,
       0.00000000038,    -0.00000000078,
      -0.00000001471,     0.00000000937,
       0.00000000123,     0.00000000159,
       0.00000000099,    -0.00000000190,
      -0.00000000488,     0.00000001770,
       0.00000000217,     0.00000000351,
       0.00000000044,     0.00000000002,
       0.00000036798,     0.00000009667,
       0.00000007267,    -0.00000004508,
       0.00000000045,    -0.00000000911,
      -0.00000002619,     0.00000002482,
       0.00000000145,     0.00000000004,
       0.00000002550,     0.00000002670,
       0.00000000026,    -0.00000000153,
       0.00000000203,     0.00000000191,
       0.00000000197,    -0.00000000206,
       0.00000000019,     0.00000000408,
       0.00000000214,     0.00000000062,
      -0.00000000081,    -0.00000000186,
       0.00000000036,    -0.00000000020,
      -0.00000000184,     0.00000000070,
      -0.00000000021,    -0.00000000034,
       0.00000000067,     0.00000000033,
       0.00000000116,    -0.00000000212,
      -0.00000000025,    -0.00000000027,
      -0.00000000343,    -0.00000000044,
      -0.00000000007,     0.00000000048,
      -0.00000000032,     0.00000000345,
       0.00000000047,     0.00000000007,
      -0.00000000053,    -0.00000000115,
      -0.00000000132,     0.00000000141,
       0.00000000142,     0.00000000131,
      -0.00000000088,    -0.00000000117,
      -0.00000000336,     0.00000000232,
       0.00000000408,     0.00000000152,
       0.00000000141,    -0.00000000403,
      -0.00000003656,     0.00000005234,
       0.00000000339,     0.00000001050,
       0.00000000090,     0.00000000024,
       0.00000029596,     0.00000018257,
       0.00000005512,    -0.00000001966,
       0.00000000134,    -0.00000000476,
      -0.00000014020,    -0.00000003765,
       0.00000000141,     0.00000000312,
      -0.00000000031,    -0.00000000023,
      -0.00000003599,     0.00000014274,
       0.00000000383,    -0.00000000148,
       0.00000000025,     0.00000000794,
      -0.00000000044,     0.00000000043,
       0.00000000830,    -0.00000000027,
       0.00000000047,     0.00000000044,
       0.00000005384,    -0.00000006527,
       0.00000000563,     0.00000000313,
      -0.00000000104,     0.00000000052,
      -0.00000000058,    -0.00000000018,
       0.00000000014,    -0.00000000043,
       0.00000000000,     0.00000000068,
      -0.00000000046,    -0.00000000012,
      -0.00000000043,     0.00000000073,
       0.00000000076,    -0.00000000012,
      -0.00000000013,    -0.00000000075,
       0.00000000031,     0.00000000101,
      -0.00000000092,    -0.00000000044,
      -0.00000000030,     0.00000000101,
       0.00000000046,     0.00000000046,
       0.00000000036,    -0.00000000086,
      -0.00000041836,     0.00000025212,
       0.00000000260,     0.00000006551,
       0.00000000386,     0.00000000182,
       0.00000101623,     0.00000142941,
       0.00000022221,    -0.00000002519,
       0.00000000556,    -0.00000001255,
      -0.00000038613,    -0.00000121593,
      -0.00000001139,     0.00000000978,
      -0.00000000032,    -0.00000000185,
      -0.00000121244,     0.00000040956,
       0.00000001440,     0.00000001238,
      -0.00000000163,    -0.00000000008,
      -0.00000002911,     0.00000003093,
      -0.00000000323,    -0.00000000057,
       0.00000003383,     0.00000003101,
      -0.00000000032,     0.00000000330,
       0.00000000357,    -0.00000000119,
      -0.00000000119,    -0.00000000357,
       0.00000000685,     0.00000000078,
       0.00000000083,    -0.00000000287,
      -0.00000000036,     0.00000000047,
       0.00000000048,     0.00000000037,
       0.00000000039,     0.00000000045,
      -0.00000000070,     0.00000000039,
       0.00000000054,     0.00000000026,
       0.00000000025,    -0.00000000054,
      -0.00000000082,    -0.00000000062,
      -0.00000000162,     0.00000000211,
       0.00000000393,     0.00000000085,
       0.00000000079,    -0.00000000390,
      -0.00000000040,     0.00000000071,
       0.00000000063,     0.00000000001,
       0.00000000000,    -0.00000000063,
       0.00000000012,    -0.00000000042,
      -0.00000382434,     0.00000023302,
      -0.00000010219,     0.00000038485,
       0.00000001691,     0.00000001078,
       0.00000122987,     0.00001117396,
       0.00000110637,     0.00000020868,
       0.00000002423,    -0.00000004658,
       0.00000722998,    -0.00000925166,
      -0.00000004192,    -0.00000004738,
       0.00000000323,    -0.00000000747,
      -0.00000931747,    -0.00000705137,
      -0.00000001913,     0.00000005947,
      -0.00000000550,    -0.00000000534,
      -0.00000016700,     0.00000000332,
      -0.00000000540,    -0.00000001278,
       0.00000000040,    -0.00000000048,
       0.00000001193,     0.00000019684,
      -0.00000001175,     0.00000000666,
      -0.00000000045,    -0.00000000038,
       0.00000002826,    -0.00000001198,
      -0.00000000021,    -0.00000000202,
       0.00000003598,    -0.00000000507,
       0.00000000081,    -0.00000000158,
      -0.00000000504,    -0.00000003596,
      -0.00000000157,    -0.00000000081,
      -0.00000000044,     0.00000000040,
       0.00000000040,     0.00000000043,
      -0.00000000040,     0.00000000082,
       0.00000000031,     0.00000000019,
       0.00000000463,     0.00000000196,
       0.00000000093,    -0.00000000175,
      -0.00000000208,    -0.00000000024,
       0.00000000010,     0.00000000066,
      -0.00000000021,     0.00000000211,
       0.00000000067,    -0.00000000011,
       0.00000000030,     0.00000000158,
       0.00000000067,    -0.00000000110,
      -0.00000000111,    -0.00000000065,
       0.00000000004,    -0.00000000062,
       0.00000000049,     0.00000000030,
       0.00000000030,    -0.00000000050,
      -0.00000000150,     0.00000000036,
       0.00000000077,     0.00000000093,
       0.00000000090,    -0.00000000077,
       0.00000000019,    -0.00000000078,
      -0.00000000061,     0.00000000000,
       0.00000000001,     0.00000000061,
      -0.00001962728,    -0.00000997777,
      -0.00000089784,     0.00000148864,
       0.00000005925,     0.00000003857,
      -0.00002697858,     0.00005072656,
       0.00000389175,     0.00000235831,
       0.00000010068,    -0.00000015543,
      -0.00000000410,    -0.00000000280,
       0.00008228335,     0.00000075920,
       0.00000003608,    -0.00000005207,
      -0.00000000321,    -0.00000001510,
      -0.00000016314,    -0.00008158375,
       0.00000004600,     0.00000010337,
      -0.00000000328,    -0.00000000339,
       0.00000013445,     0.00000001437,
       0.00000000698,     0.00000000535,
      -0.00000000031,     0.00000000057,
      -0.00000007893,     0.00000003686,
       0.00000000432,    -0.00000000082,
       0.00000000053,     0.00000000031,
       0.00000040592,    -0.00000008511,
       0.00000001011,    -0.00000001937,
       0.00000000828,     0.00000000561,
       0.00000038637,    -0.00000000043,
      -0.00000000013,    -0.00000038580,
       0.00000002863,    -0.00000000733,
      -0.00000000428,     0.00000000355,
       0.00000000137,     0.00000000169,
       0.00000001397,     0.00000001443,
       0.00000000557,    -0.00000000535,
      -0.00000000644,    -0.00000001309,
      -0.00000000352,     0.00000000280,
      -0.00000001302,     0.00000000670,
       0.00000000292,     0.00000000349,
      -0.00000000025,     0.00000000036,
       0.00000000040,     0.00000000027,
      -0.00000000124,    -0.00000000056,
      -0.00000000185,     0.00000000404,
       0.00000000544,    -0.00000000051,
      -0.00000000057,    -0.00000000539,
       0.00000000197,     0.00000000110,
       0.00000000030,    -0.00000000038,
       0.00000000269,    -0.00000000549,
      -0.00000000110,    -0.00000000072,
      -0.00000000426,     0.00000000129,
      -0.00000000003,     0.00000000051,
       0.00000000148,     0.00000000395,
       0.00000000050,    -0.00000000001,
      -0.00000000159,    -0.00000000168,
      -0.00000000035,     0.00000000015,
      -0.00000000447,     0.00000000388,
       0.00000000039,     0.00000000094,
       0.00000000656,     0.00000000312,
       0.00000000051,    -0.00000000034,
       0.00000000320,    -0.00000000664,
      -0.00000000040,    -0.00000000049,
       0.00001038213,     0.00000047210,
       0.00000001010,    -0.00000000358,
      -0.00000000011,    -0.00000000159,
       0.00000111717,    -0.00003319998,
      -0.00000000853,    -0.00000002166,
      -0.00000000425,     0.00000000023,
      -0.00000274393,     0.00000049397,
      -0.00000006950,    -0.00000010351,
       0.00000000416,    -0.00000000225,
      -0.00000058523,     0.00000149999,
      -0.00000012988,     0.00000006075,
      -0.00000000382,    -0.00000000402,
      -0.00000002703,     0.00000002706,
      -0.00000000038,    -0.00000001699,
       0.00000000010,    -0.00000000093,
      -0.00000000014,    -0.00000025545,
       0.00000000093,    -0.00000000848,
       0.00000000004,     0.00000000054,
       0.00000713818,    -0.00000000233,
       0.00000000001,     0.00000109752,
      -0.00000001524,    -0.00000000825,
      -0.00000000026,    -0.00000000124,
      -0.00000000450,     0.00000003247,
      -0.00000000055,     0.00000000060,
       0.00000000430,     0.00000000038,
       0.00000000038,    -0.00000000426,
       0.00000007415,     0.00000000190,
       0.00000026752,    -0.00000000005,
       0.00000062540,    -0.00000000021,
      -0.00000000025,    -0.00000000077,
      -0.00000000037,     0.00000000076,
       0.00000000076,     0.00000000036,
      -0.00000003215,     0.00000000690,
       0.00000000308,     0.00000001203,
       0.00000002492,     0.00000009351,
       0.00000003464,    -0.00000001109,
       0.00000004949,    -0.00000008826,
      -0.00000002841,    -0.00000001047,
      -0.00000008858,    -0.00000004789,
      -0.00000000976,     0.00000002853,
      -0.00000000140,     0.00000000024,
       0.00000000035,     0.00000000164,
      -0.00000000071,    -0.00000000067,
      -0.00000000034,     0.00000000076,
       0.00000000076,     0.00000000031,
       0.00000000038,    -0.00000000109,
      -0.00000000027,    -0.00000000014,
      -0.00000000314,    -0.00000000124,
      -0.00000000040,     0.00000000076,
       0.00000000213,     0.00000000464,
       0.00000000080,    -0.00000000012,
       0.00000000450,    -0.00000000221,
      -0.00000000014,    -0.00000000086,
       0.00000000993,     0.00000000536,
       0.00000000130,    -0.00000000101,
       0.00000001337,    -0.00000002498,
      -0.00000000254,    -0.00000000328,
      -0.00000003983,    -0.00000000385,
      -0.00000000199,     0.00000000222,
      -0.00000000343,     0.00000003948,
       0.00000000214,     0.00000000184,
      -0.00000000196,    -0.00000000169,
      -0.00000000383,     0.00000000443,
       0.00000000034,     0.00000000025,
       0.00000001080,     0.00000000323,
       0.00000000332,    -0.00000001068,
       0.00000000143,     0.00000000000,
       0.00000000000,    -0.00000000143,
       0.00000000154,     0.00000000000,
       0.00000000000,     0.00000000154,
       0.00000000211,     0.00000000000,
       0.00000000000,    -0.00000000211,
       0.00000000219,     0.00000000000,
       0.00000000000,     0.00000000220,
       0.00000000070,     0.00000000140,
       0.00000000050,    -0.00000000052,
       0.00000000097,    -0.00000000246,
      -0.00000000124,     0.00000000001,
      -0.00000000246,    -0.00000000094,
       0.00000000002,     0.00000000124,
       0.00000000103,     0.00000000028,
       0.00000000078,    -0.00000000266,
      -0.00000000387,     0.00000000076,
       0.00000000080,     0.00000000383,
      -0.00000012018,    -0.00000004435,
      -0.00000001258,     0.00000004562,
       0.00000000019,     0.00000000021,
      -0.00000012216,     0.00000031324,
       0.00000011946,     0.00000003485,
       0.00000000057,    -0.00000000056,
       0.00000048300,    -0.00000005344,
      -0.00000003349,    -0.00000014106,
      -0.00000002106,     0.00000000747,
      -0.00000005828,    -0.00000047830,
      -0.00000013887,     0.00000003542,
       0.00000000789,     0.00000002065,
       0.00000000105,    -0.00000000004,
      -0.00000000064,    -0.00000000019,
       0.00000000197,    -0.00000000031,
      -0.00000000031,    -0.00000000197,
       0.00000003154,     0.00000000128,
      -0.00000000254,    -0.00000000931,
      -0.00000000105,     0.00000000048,
       0.00000000004,    -0.00000008171,
      -0.00000002399,     0.00000000724,
       0.00000000131,     0.00000000270,
      -0.00000010838,     0.00000007321,
       0.00000002546,     0.00000001742,
       0.00000000099,    -0.00000000237,
       0.00000007372,     0.00000010700,
       0.00000001684,    -0.00000002554,
      -0.00000000237,    -0.00000000091,
      -0.00000000049,     0.00000000152,
       0.00000000169,     0.00000000002,
      -0.00000000303,     0.00000000018,
       0.00000000018,     0.00000000303,
       0.00000000054,     0.00000000000,
       0.00000000000,    -0.00000000053,
       0.00000000050,     0.00000000029,
       0.00000000028,    -0.00000000049,
      -0.00000005043,     0.00000019248,
       0.00000003757,     0.00000000124,
      -0.00000000014,    -0.00000000309,
       0.00000047166,     0.00000011591,
       0.00000000262,    -0.00000009233,
      -0.00000000762,     0.00000000038,
      -0.00000020915,    -0.00000071653,
      -0.00000006921,     0.00000005437,
       0.00000000251,     0.00000000299,
      -0.00000070846,     0.00000021340,
       0.00000005523,     0.00000006697,
       0.00000000273,    -0.00000000259,
      -0.00000000354,    -0.00000000325,
      -0.00000000083,     0.00000000530,
      -0.00000000032,    -0.00000000035,
      -0.00000006240,     0.00000000514,
       0.00000000877,     0.00000001146,
      -0.00000000170,    -0.00000003931,
      -0.00000000371,     0.00000000258,
      -0.00000003934,     0.00000000170,
       0.00000000259,     0.00000000372,
      -0.00000078110,    -0.00000038939,
      -0.00000003927,     0.00000005946,
       0.00000000239,     0.00000000167,
      -0.00000093096,     0.00000179793,
       0.00000013907,     0.00000009366,
       0.00000000398,    -0.00000000561,
       0.00000299233,     0.00000001963,
       0.00000000376,    -0.00000000344,
      -0.00000000409,    -0.00000295539,
       0.00000000224,    -0.00000000060,
       0.00000000543,     0.00000000458,
      -0.00000000011,     0.00000000049,
      -0.00000000572,     0.00000000261,
       0.00000000042,     0.00000000009,
      -0.00000004967,    -0.00000096006,
      -0.00000009044,     0.00000006386,
       0.00000002958,    -0.00000000232,
      -0.00000000201,    -0.00000000279,
       0.00000044953,    -0.00000000029,
      -0.00000000033,    -0.00000045022,
       0.00000023157,     0.00000003300,
      -0.00000000006,    -0.00000000208,
       0.00000007303,    -0.00000070680,
      -0.00000000461,     0.00000000027,
       0.00000000330,    -0.00000001011,
       0.00000001197,    -0.00000000042,
      -0.00000012738,     0.00000002608,
       0.00000000855,    -0.00000000093,
       0.00000000050,    -0.00000000001,
      -0.00000000036,    -0.00000000288,
       0.00000000021,    -0.00000000081,
       0.00000000002,    -0.00000000815,
       0.00000000009,     0.00000000045,
       0.00001634104,    -0.00000001044,
       0.00000000068,     0.00000155503,
       0.00000000132,    -0.00000004739,
       0.00000000186,     0.00000000317,
      -0.00000002371,    -0.00000000249,
       0.00000000126,    -0.00000000448,
       0.00000000820,    -0.00000000325,
      -0.00000000325,    -0.00000000820,
       0.00000006545,     0.00000000155,
       0.00000014726,    -0.00000000005,
      -0.00000000051,     0.00000000058,
       0.00000000128,     0.00000000045,
       0.00000000044,    -0.00000000127,
       0.00000000256,     0.00000000064,
       0.00000000064,    -0.00000000256,
       0.00000000225,     0.00000000141,
       0.00000000319,    -0.00000000506,
      -0.00000000046,    -0.00000000034,
      -0.00000000976,    -0.00000000169,
      -0.00000000158,     0.00000000970,
       0.00000011009,     0.00000002714,
      -0.00000004173,    -0.00000000511,
      -0.00000000510,     0.00000004171,
      -0.00000001356,    -0.00000000591,
      -0.00000000058,     0.00000000097,
      -0.00000001295,     0.00000002988,
       0.00000000212,     0.00000000127,
       0.00000005075,     0.00000000021,
      -0.00000000028,    -0.00000005042,
      -0.00000211341,    -0.00000025652,
       0.00000000832,     0.00000003411,
       0.00000000463,    -0.00000003762,
       0.00000059316,     0.00000000002,
      -0.00000000013,    -0.00000059254,
       0.00000000197,    -0.00000000016,
      -0.00000000035,    -0.00000000595,
       0.00000000188,     0.00000000006,
       0.00004484929,    -0.00000000392,
       0.00000000016,     0.00000225295,
       0.00000002676,    -0.00000000657,
      -0.00000000328,    -0.00000005354,
       0.00000000502,     0.00000000339,
       0.00000000339,    -0.00000000502,
       0.00000007481,    -0.00000000611,
       0.00000000709,    -0.00000000048,
      -0.00000000048,    -0.00000000709,
      -0.00000000021,     0.00000000042,
       0.00000000073,     0.00000000005,
       0.00000000004,    -0.00000000072,
      -0.00000003070,    -0.00000000561,
       0.00000000852,     0.00000000057,
       0.00000000056,    -0.00000000851,
       0.00000046051,    -0.00000002661,
       0.00000000084,     0.00000001459,
      -0.00000000028,    -0.00000000045,
       0.00000000089,    -0.00000000015,
       0.00000000016,     0.00000000089,
      -0.00000004444,     0.00000003339,
       0.00000000165,     0.00000000208,
       0.00000001278,    -0.00000000156,
       0.00000000154,     0.00000001277,
       0.00000000480,     0.00000000324,
      -0.00000000324,     0.00000000480,
      -0.00000000710,    -0.00000000047,
      -0.00000000047,     0.00000000710,
      -0.00000042431,    -0.00000002393,
       0.00000000081,    -0.00000001424,
       0.00000000001,     0.00000000542,
      -0.00000000001,    -0.00000000205,
       0.00000000006,    -0.00000000048,
      -0.00000000048,    -0.00000000006,
       0.00000000664,    -0.00000000741,
      -0.00000000741,    -0.00000000664,
      -0.00000000400,    -0.00000000178,
      -0.00000000018,     0.00000000029,
      -0.00000000385,     0.00000000863,
       0.00000000063,     0.00000000038,
       0.00000001534,     0.00000000007,
      -0.00000000009,    -0.00000001521,
       0.00000032736,     0.00000001847,
      -0.00000000097,     0.00000001695,
       0.00000000198,    -0.00000000004,
       0.00000000000,    -0.00000000217,
       0.00000052757,    -0.00000058887,
       0.00000000701,     0.00000000626,
       0.00000073000,    -0.00000000038,
      -0.00000000035,    -0.00000073001,
       0.00000000007,    -0.00000000118,
       0.00008707312,    -0.00000004306,
       0.00000000105,     0.00000283649,
      -0.00000000433,    -0.00000000967,
      -0.00000000483,     0.00000000866,
       0.00000001498,     0.00000000056,
       0.00000000057,    -0.00000001498,
      -0.00000009915,     0.00000006566,
       0.00000000191,     0.00000000304,
      -0.00000000212,    -0.00000000237,
       0.00000000237,    -0.00000000211,
       0.00000001464,     0.00000000055,
      -0.00000000055,     0.00000001464,
      -0.00000000081,     0.00000000053,
      -0.00000000053,    -0.00000000081,
       0.00000000016,    -0.00000000044,
      -0.00000000073,     0.00000000005,
       0.00000000005,     0.00000000072,
      -0.00000006529,     0.00000000188,
      -0.00000000003,    -0.00000000130,
      -0.00000000851,     0.00000000057,
       0.00000000057,     0.00000000850,
      -0.00000034526,     0.00000001987,
      -0.00000000096,    -0.00000001690,
       0.00000000031,     0.00000000548,
      -0.00000000028,    -0.00000000014,
       0.00000000186,     0.00000001636,
      -0.00000000045,    -0.00000000003,
      -0.00000001633,     0.00000000197,
      -0.00000000001,    -0.00000000046,
       0.00000000052,     0.00000000060,
      -0.00000000077,     0.00000000065,
      -0.00000000353,    -0.00000000269,
       0.00000000042,    -0.00000000052,
      -0.00000000939,     0.00000000483,
      -0.00000000027,    -0.00000000069,
      -0.00000000502,    -0.00000000938,
       0.00000000070,    -0.00000000031,
      -0.00000000017,     0.00000000112,
      -0.00000000105,    -0.00000000022,
      -0.00000001524,    -0.00000000871,
       0.00000000129,    -0.00000000289,
      -0.00000003016,     0.00000002098,
      -0.00000000153,    -0.00000000242,
      -0.00000002199,    -0.00000003028,
       0.00000000238,    -0.00000000189,
      -0.00000000244,    -0.00000000271,
       0.00000000283,    -0.00000000189,
       0.00000010267,    -0.00000018939,
       0.00000003267,     0.00000001071,
      -0.00000000121,     0.00000000212,
       0.00000020147,     0.00000012960,
      -0.00000002447,     0.00000003490,
      -0.00000000286,    -0.00000000126,
       0.00000057433,    -0.00000037861,
       0.00000002751,     0.00000004572,
      -0.00000000184,     0.00000000095,
       0.00000039249,     0.00000057581,
      -0.00000004597,     0.00000003114,
      -0.00000000137,    -0.00000000183,
       0.00000015968,     0.00000002224,
       0.00000000589,    -0.00000001129,
      -0.00000000040,    -0.00000000056,
      -0.00000002321,     0.00000015623,
       0.00000001151,     0.00000000584,
       0.00000000056,    -0.00000000040,
      -0.00000000102,    -0.00000000596,
       0.00000007567,    -0.00000003357,
       0.00000000330,     0.00000000552,
       0.00000003369,     0.00000007574,
      -0.00000000552,     0.00000000333,
       0.00000000874,    -0.00000000333,
       0.00000000018,     0.00000000152,
       0.00000010252,    -0.00000004549,
       0.00000000450,     0.00000000747,
       0.00000004573,     0.00000010260,
      -0.00000000748,     0.00000000453,
       0.00000000570,     0.00000000951,
      -0.00000000085,     0.00000000056,
       0.00000014557,    -0.00000006457,
       0.00000000636,     0.00000001062,
       0.00000006477,     0.00000014567,
      -0.00000001062,     0.00000000641,
       0.00000000110,    -0.00000000245,
       0.00000018224,    -0.00000008091,
       0.00000000797,     0.00000001330,
       0.00000008124,     0.00000018250,
      -0.00000001331,     0.00000000804,
      -0.00000000001,     0.00000000386,
       0.00000000001,    -0.00000000265,
       0.00000001072,    -0.00000000057,
      -0.00000000018,    -0.00000000087,
       0.00000000246,     0.00000000001,
       0.00000000001,    -0.00000000244,
      -0.00000000757,     0.00000000061,
       0.00000000012,     0.00000000066,
      -0.00000000218,    -0.00000000001,
      -0.00000000001,     0.00000000217,
       0.00000000001,    -0.00000000079,
       0.00000000780,     0.00000000000,
      -0.00000000018,    -0.00000000065,
       0.00000000016,    -0.00000000079,
       0.00000000081,     0.00000000016,
       0.00000000069,    -0.00000000008,
       0.00000000006,     0.00000000068,
      -0.00000000078,     0.00000000037,
       0.00000000006,     0.00000000045,
       0.00000000089,    -0.00000000087,
      -0.00000000026,    -0.00000000036,
       0.00000000132,    -0.00000000208,
      -0.00000000085,    -0.00000000130,
       0.00000000038,     0.00000000024,
       0.00000000023,    -0.00000000045,
       0.00000000056,     0.00000000008,
      -0.00000000036,    -0.00000000042,
      -0.00000000060,     0.00000000051,
      -0.00000000034,     0.00000000062,
       0.00000000009,    -0.00000000094,
      -0.00000000098,    -0.00000000007,
       0.00000000037,    -0.00000000026,
      -0.00000000168,     0.00000000111,
       0.00000000110,     0.00000000168,
      -0.00000000050,     0.00000000045,
       0.00000000071,     0.00000000059,
      -0.00000000106,    -0.00000000007,
      -0.00000000006,     0.00000000112,
       0.00000000020,     0.00000000046,
      -0.00000000027,     0.00000000034,
       0.00000000032,     0.00000000034,
      -0.00000000049,    -0.00000000039,
       0.00000000005,    -0.00000000290,
      -0.00000000403,     0.00000000007,
       0.00000000117,     0.00000000050,
       0.00000000027,    -0.00000000128,
      -0.00000000293,     0.00000000190,
       0.00000000553,     0.00000000254,
      -0.00000000009,    -0.00000000040,
      -0.00000002250,    -0.00000000906,
      -0.00000000008,     0.00000000045,
      -0.00000000928,     0.00000002259,
       0.00000000045,     0.00000000011,
       0.00000000005,     0.00000000084,
       0.00000000087,    -0.00000000006,
      -0.00000000208,    -0.00000000014,
      -0.00000000018,     0.00000000278,
      -0.00000000003,     0.00000000057,
      -0.00000000094,     0.00000000099,
       0.00000000131,     0.00000000125,
       0.00000000329,     0.00000000118,
       0.00000000150,    -0.00000000418,
      -0.00000004742,    -0.00000001938,
      -0.00000000025,     0.00000000113,
       0.00000000613,     0.00000007849,
       0.00000000267,    -0.00000000286,
      -0.00000006129,    -0.00000025477,
      -0.00000000276,     0.00000000061,
      -0.00000025916,     0.00000006125,
       0.00000000020,     0.00000000310,
      -0.00000000375,     0.00000000416,
      -0.00000000039,    -0.00000000012,
       0.00000000443,     0.00000000404,
      -0.00000000013,     0.00000000038,
      -0.00000000069,    -0.00000000032,
      -0.00000000041,     0.00000000088,
      -0.00000015563,    -0.00000067236,
      -0.00000000861,     0.00000000156,
      -0.00000000020,    -0.00000000079,
      -0.00000081043,     0.00000051101,
       0.00000002832,     0.00000000338,
      -0.00000000159,    -0.00000000052,
       0.00000146854,    -0.00000179162,
      -0.00000000427,    -0.00000001012,
       0.00000000065,    -0.00000000147,
      -0.00000184622,    -0.00000150769,
      -0.00000001644,     0.00000000721,
      -0.00000000151,    -0.00000000036,
      -0.00000002174,     0.00000000174,
      -0.00000000087,    -0.00000000156,
      -0.00000000001,     0.00000002547,
      -0.00000000156,     0.00000000074,
      -0.00000000072,     0.00000000089,
       0.00000000089,     0.00000000072,
       0.00000000078,     0.00000000000,
       0.00000000001,    -0.00000000109,
       0.00000519865,    -0.00000616688,
      -0.00000002103,    -0.00000004675,
       0.00000000189,    -0.00000000508,
      -0.00000902252,    -0.00000527209,
       0.00000006129,     0.00000010218,
      -0.00000000343,    -0.00000000740,
       0.00001300576,    -0.00000003050,
       0.00000000781,     0.00000003675,
       0.00000000111,    -0.00000000155,
      -0.00000025243,    -0.00001375873,
      -0.00000003472,    -0.00000000865,
      -0.00000000213,     0.00000000171,
       0.00000000312,     0.00000001497,
      -0.00000000076,     0.00000000033,
      -0.00000002347,    -0.00000000966,
       0.00000000052,    -0.00000000139,
      -0.00000000373,     0.00000000428,
       0.00000000157,     0.00000000090,
      -0.00000000575,    -0.00000000158,
      -0.00000000158,     0.00000000578,
      -0.00000000078,     0.00000000143,
      -0.00000000071,    -0.00000000017,
       0.00000000023,     0.00000000100,
       0.00000000033,    -0.00000000019,
      -0.00000000141,    -0.00000000345,
      -0.00000000095,     0.00000000063,
      -0.00000000351,     0.00000000141,
       0.00000000063,     0.00000000097,
      -0.00000000073,    -0.00000000025,
       0.00000000116,    -0.00000000014,
      -0.00000000016,    -0.00000000121,
       0.00000000022,    -0.00000000060,
      -0.00000000133,     0.00000000086,
       0.00000000086,     0.00000000136,
       0.00000000047,     0.00000000064,
       0.00000000076,    -0.00000000044,
       0.00000000173,     0.00000000013,
       0.00000000024,    -0.00000000174,
       0.00006601665,     0.00000020968,
      -0.00000003916,    -0.00000000215,
       0.00000000019,    -0.00000000781,
       0.00000033019,    -0.00008395476,
      -0.00000000317,     0.00000004126,
      -0.00000001007,    -0.00000000014,
      -0.00000159441,    -0.00000205443,
      -0.00000005685,     0.00000030315,
       0.00000000345,     0.00000000759,
       0.00000024603,    -0.00000398491,
      -0.00000009868,    -0.00000018183,
      -0.00000000308,     0.00000001229,
      -0.00000013815,     0.00000001540,
      -0.00000000506,    -0.00000001164,
       0.00000000027,    -0.00000000054,
      -0.00000001144,    -0.00000017699,
       0.00000000901,    -0.00000000661,
       0.00000000041,     0.00000000034,
      -0.00000001588,    -0.00000002013,
       0.00000000099,     0.00000001465,
       0.00000002943,    -0.00000001277,
       0.00000000128,     0.00000000213,
      -0.00000001299,    -0.00000002940,
       0.00000000215,    -0.00000000127,
      -0.00000000289,    -0.00000000730,
      -0.00000000203,     0.00000000130,
      -0.00000000827,     0.00000000713,
       0.00000000287,     0.00000000200,
       0.00000001354,    -0.00000002281,
      -0.00000000732,    -0.00000000297,
      -0.00000002349,    -0.00000001389,
      -0.00000000306,     0.00000000757,
       0.00000000352,    -0.00000000045,
      -0.00000000059,    -0.00000000476,
       0.00000000016,     0.00000000095,
       0.00000000106,    -0.00000000051,
      -0.00000000333,     0.00000000087,
      -0.00000000003,     0.00000000042,
       0.00000000188,     0.00000000318,
       0.00000000042,    -0.00000000011,
      -0.00000001052,    -0.00000000054,
      -0.00000000048,     0.00000000052,
      -0.00000000063,     0.00000001099,
       0.00000000065,     0.00000000048,
       0.00000000481,     0.00000000198,
       0.00000000036,    -0.00000000031,
       0.00000000233,    -0.00000000568,
      -0.00000000040,    -0.00000000045,
       0.00000000398,     0.00000000004,
       0.00000000018,    -0.00000000433,
      -0.00000175805,    -0.00000013708,
      -0.00000005146,    -0.00000009241,
       0.00000000315,    -0.00000000248,
      -0.00000008473,     0.00000221058,
      -0.00000011962,     0.00000006717,
      -0.00000000315,    -0.00000000417,
      -0.00000081785,     0.00000006845,
       0.00000000005,    -0.00000000639,
       0.00000000021,    -0.00000000045,
      -0.00000003560,    -0.00000073433,
      -0.00000000562,     0.00000000649,
      -0.00000000033,    -0.00000000029,
      -0.00000000547,    -0.00000001310,
      -0.00000000149,     0.00000000179,
       0.00000001563,    -0.00000000036,
      -0.00000000175,    -0.00000000129,
       0.00000056764,    -0.00000025202,
       0.00000002483,     0.00000004139,
       0.00000004311,     0.00000009664,
      -0.00000000706,     0.00000000424,
      -0.00000000240,    -0.00000000009,
       0.00000000081,     0.00000000187,
       0.00000002432,    -0.00000001080,
       0.00000005298,    -0.00000002354,
      -0.00000000028,     0.00000000051,
       0.00000000072,     0.00000000024,
      -0.00000000092,     0.00000000019,
       0.00000000022,     0.00000000097,
       0.00000003608,    -0.00000005915,
      -0.00000001911,    -0.00000000802,
      -0.00000008365,    -0.00000003302,
      -0.00000000506,     0.00000002682,
       0.00000011061,    -0.00000001647,
      -0.00000000858,    -0.00000003143,
      -0.00000001932,    -0.00000011671,
      -0.00000003360,     0.00000000977,
      -0.00000000013,     0.00000000066,
       0.00000000072,     0.00000000007,
      -0.00000000032,     0.00000000086,
       0.00000000092,     0.00000000033,
       0.00000000138,     0.00000000353,
       0.00000000064,    -0.00000000001,
       0.00000000493,    -0.00000000191,
      -0.00000000003,    -0.00000000091,
      -0.00000000144,    -0.00000000240,
      -0.00000000221,     0.00000000150,
      -0.00000003287,    -0.00000000298,
      -0.00000000153,     0.00000000178,
      -0.00000000330,     0.00000004086,
       0.00000000219,     0.00000000192,
      -0.00000000350,     0.00000000143,
      -0.00000000018,     0.00000000607,
       0.00000000957,     0.00000000289,
       0.00000000359,    -0.00000001185,
      -0.00000000137,    -0.00000000004,
       0.00000000046,     0.00000000028,
       0.00000000022,    -0.00000000050,
      -0.00000000051,    -0.00000000021,
       0.00000000073,    -0.00000000184,
      -0.00000000093,    -0.00000000002,
       0.00000000004,     0.00000000015,
      -0.00000000274,    -0.00000000078,
       0.00000000012,     0.00000000133,
       0.00000000263,     0.00000000231,
       0.00000000039,    -0.00000000119,
       0.00000000222,    -0.00000000275,
      -0.00000000124,    -0.00000000033,
      -0.00000000313,     0.00000000063,
       0.00000000080,     0.00000000397,
       0.00000038598,    -0.00000004483,
      -0.00000002770,    -0.00000011236,
      -0.00000000067,    -0.00000000003,
      -0.00000005785,    -0.00000049431,
      -0.00000014388,     0.00000003562,
       0.00000000002,     0.00000000094,
      -0.00000002964,     0.00000000440,
       0.00000000194,     0.00000000820,
       0.00000000018,    -0.00000000035,
       0.00000001387,    -0.00000000425,
      -0.00000000442,    -0.00000000449,
      -0.00000000040,    -0.00000000012,
      -0.00000000076,     0.00000000031,
       0.00000000002,    -0.00000000091,
      -0.00000009158,     0.00000005996,
       0.00000002124,     0.00000001460,
       0.00000000081,    -0.00000000198,
       0.00000008722,     0.00000011504,
       0.00000001796,    -0.00000002874,
      -0.00000000266,    -0.00000000098,
      -0.00000002283,    -0.00000008906,
      -0.00000000873,     0.00000000559,
       0.00000000018,     0.00000000044,
      -0.00000008804,     0.00000003205,
       0.00000000820,     0.00000000768,
       0.00000000027,    -0.00000000047,
       0.00000000000,     0.00000000080,
      -0.00000018192,    -0.00000063429,
      -0.00000006052,     0.00000004866,
       0.00000000227,     0.00000000254,
      -0.00000078593,     0.00000025642,
       0.00000006161,     0.00000007515,
       0.00000000316,    -0.00000000285,
       0.00000026440,     0.00000001428,
       0.00000000587,     0.00000000063,
      -0.00000004230,    -0.00000027446,
       0.00000000106,     0.00000000475,
       0.00000000046,    -0.00000000004,
      -0.00000000001,     0.00000000243,
      -0.00000000257,    -0.00000000003,
      -0.00000000228,    -0.00000000303,
      -0.00000000302,     0.00000000228,
       0.00000277534,     0.00000000598,
       0.00000000155,    -0.00000000053,
       0.00000000974,    -0.00000331140,
      -0.00000000078,    -0.00000000198,
      -0.00000003738,    -0.00000009732,
      -0.00000000005,     0.00000001479,
      -0.00000000086,    -0.00000019298,
      -0.00000000158,    -0.00000000973,
      -0.00000000001,     0.00000000057,
      -0.00000000874,    -0.00000000360,
       0.00000000035,    -0.00000000083,
       0.00000000332,    -0.00000000950,
       0.00000000077,     0.00000000037,
      -0.00000004659,    -0.00000007546,
       0.00000000260,    -0.00000000005,
       0.00000003613,    -0.00000001600,
       0.00000000162,     0.00000000262,
      -0.00000001608,    -0.00000003616,
       0.00000000262,    -0.00000000161,
      -0.00000000602,    -0.00000006943,
       0.00000000651,     0.00000000433,
      -0.00000000021,     0.00000000027,
      -0.00000007926,     0.00000001000,
       0.00000000481,    -0.00000000739,
       0.00000000030,     0.00000000024,
      -0.00000001993,     0.00000000180,
       0.00000000038,    -0.00000000030,
      -0.00000000631,    -0.00000001826,
      -0.00000000121,    -0.00000000049,
       0.00000000065,    -0.00000000108,
       0.00000135059,    -0.00000060078,
       0.00000005938,     0.00000009857,
       0.00000005992,     0.00000013472,
      -0.00000000982,     0.00000000592,
      -0.00000000173,    -0.00000000396,
      -0.00000000205,     0.00000000056,
       0.00000000056,    -0.00000000057,
      -0.00000000057,    -0.00000000056,
      -0.00000000015,    -0.00000000061,
       0.00000000060,     0.00000000025,
       0.00000000036,    -0.00000000101,
      -0.00000000077,     0.00000000062,
       0.00000000096,     0.00000000087,
      -0.00000000020,    -0.00000000040,
      -0.00000000042,     0.00000000022,
      -0.00000000048,    -0.00000000102,
      -0.00000000142,     0.00000000071,
       0.00000000010,    -0.00000000033,
      -0.00000000043,    -0.00000000013,
      -0.00000000017,    -0.00000000061,
       0.00000000093,     0.00000000019,
       0.00000000018,    -0.00000000098,
       0.00000000313,     0.00000000065,
       0.00000000083,    -0.00000000398,
       0.00000000120,     0.00000000041,
       0.00000000054,    -0.00000000149,
      -0.00000000066,    -0.00000000015,
      -0.00000000008,     0.00000000058,
      -0.00000000937,    -0.00000000156,
      -0.00000000196,     0.00000001129,
       0.00000000252,     0.00000000030,
      -0.00000000016,    -0.00000000181,
      -0.00000000369,     0.00000000111,
       0.00000000111,     0.00000000369,
      -0.00000000074,     0.00000000004,
       0.00000000006,     0.00000000083,
       0.00000004965,    -0.00000000003,
      -0.00000000004,    -0.00000005784,
      -0.00000000056,    -0.00000000156,
       0.00000000029,    -0.00000000358,
      -0.00000018705,     0.00000005683,
      -0.00000000105,    -0.00000000335,
       0.00000004944,    -0.00000002195,
       0.00000000217,     0.00000000360,
      -0.00000002194,    -0.00000004938,
       0.00000000360,    -0.00000000217,
      -0.00000002831,     0.00000001454,
      -0.00000000071,    -0.00000000135,
       0.00000000105,    -0.00000000011,
      -0.00000000011,    -0.00000000119,
       0.00000378020,    -0.00000167945,
       0.00000016594,     0.00000027614,
       0.00000008578,     0.00000019319,
      -0.00000001409,     0.00000000847,
       0.00000000200,    -0.00000000155,
      -0.00000000228,    -0.00000000440,
       0.00000000075,     0.00000000004,
       0.00000000005,    -0.00000000083,
       0.00000003790,    -0.00000001951,
       0.00000000060,     0.00000000114,
       0.00000000104,    -0.00000000062,
       0.00000000062,     0.00000000104,
      -0.00000003677,     0.00000001388,
      -0.00000000045,    -0.00000000121,
      -0.00000000057,    -0.00000000002,
      -0.00000000003,    -0.00000000049,
       0.00000000027,    -0.00000000087,
      -0.00000000087,    -0.00000000027,
       0.00000001517,    -0.00000000001,
      -0.00000000001,    -0.00000001758,
      -0.00000000022,    -0.00000000048,
       0.00000000009,    -0.00000000107,
       0.00000002828,    -0.00000001067,
       0.00000000052,     0.00000000143,
       0.00000002212,    -0.00000006946,
       0.00000000084,     0.00000000029,
       0.00000006147,    -0.00000002733,
       0.00000000270,     0.00000000448,
      -0.00000002733,    -0.00000006147,
       0.00000000448,    -0.00000000270,
       0.00000737700,    -0.00000328098,
       0.00000032445,     0.00000053985,
      -0.00000002158,     0.00000001339,
       0.00000010778,     0.00000024251,
      -0.00000001769,     0.00000001064,
      -0.00000000044,    -0.00000000070,
      -0.00000000073,    -0.00000000066,
       0.00000000129,    -0.00000000052,
      -0.00000000052,    -0.00000000129,
       0.00000000128,    -0.00000000051,
       0.00000000051,     0.00000000128,
       0.00000000111,     0.00000000072,
      -0.00000000011,     0.00000000012,
       0.00000000309,     0.00000000251,
      -0.00000000028,     0.00000000028,
      -0.00000000251,     0.00000000309,
      -0.00000000028,    -0.00000000027,
       0.00000000021,     0.00000000001,
      -0.00000000001,     0.00000000026,
      -0.00000000049,    -0.00000000007,
      -0.00000000098,     0.00000000139,
      -0.00000000140,    -0.00000000097,
      -0.00000000231,    -0.00000000007,
      -0.00000000004,    -0.00000000071,
      -0.00000000281,     0.00000000611,
      -0.00000000103,    -0.00000000044,
      -0.00000000621,    -0.00000000278,
       0.00000000041,    -0.00000000108,
      -0.00000000105,    -0.00000000039,
       0.00000000045,    -0.00000000098,
       0.00000000133,    -0.00000003566,
       0.00000000926,    -0.00000000076,
       0.00000000009,     0.00000000101,
       0.00000002511,     0.00000000056,
       0.00000000021,     0.00000000709,
      -0.00000000076,     0.00000000005,
       0.00000006115,    -0.00000010400,
       0.00000001752,     0.00000000885,
      -0.00000000069,     0.00000000126,
       0.00000010559,     0.00000006059,
      -0.00000000856,     0.00000001808,
      -0.00000000134,    -0.00000000064,
       0.00000001611,    -0.00000000044,
       0.00000000047,    -0.00000000022,
      -0.00000000004,     0.00000001563,
       0.00000000029,     0.00000000039,
       0.00000000582,    -0.00000000644,
       0.00000000110,     0.00000000078,
       0.00000000645,     0.00000000582,
      -0.00000000078,     0.00000000111,
       0.00000000070,    -0.00000000069,
       0.00000000789,    -0.00000000872,
       0.00000000150,     0.00000000105,
       0.00000000875,     0.00000000789,
      -0.00000000105,     0.00000000150,
       0.00000000095,     0.00000000067,
       0.00000001120,    -0.00000001238,
       0.00000000213,     0.00000000149,
       0.00000001240,     0.00000001120,
      -0.00000000149,     0.00000000213,
       0.00000001402,    -0.00000001551,
       0.00000000266,     0.00000000187,
       0.00000001555,     0.00000001403,
      -0.00000000187,     0.00000000267,
      -0.00000000059,     0.00000000001,
      -0.00000000061,     0.00000000001,
      -0.00000000020,     0.00000000042,
      -0.00000000070,     0.00000000036,
       0.00000000037,     0.00000000070,
       0.00000000042,     0.00000000015,
       0.00000000008,    -0.00000000111,
      -0.00000000125,    -0.00000000006,
       0.00000000013,    -0.00000000043,
      -0.00000000176,     0.00000000086,
       0.00000000196,     0.00000000134,
      -0.00000000790,    -0.00000000386,
      -0.00000000389,     0.00000000799,
      -0.00000000083,    -0.00000000001,
      -0.00000000005,     0.00000000119,
      -0.00000000037,     0.00000000040,
       0.00000000042,     0.00000000038,
       0.00000000144,     0.00000000035,
       0.00000000044,    -0.00000000156,
      -0.00000002272,    -0.00000001180,
      -0.00000000035,     0.00000000036,
      -0.00000000222,     0.00000002902,
       0.00000000083,    -0.00000000063,
      -0.00000001675,    -0.00000008307,
      -0.00000000099,    -0.00000000029,
      -0.00000008427,     0.00000001708,
      -0.00000000024,     0.00000000110,
      -0.00000000084,     0.00000000099,
       0.00000000105,     0.00000000097,
       0.00000000009,     0.00000000062,
      -0.00000005131,    -0.00000029088,
      -0.00000000379,    -0.00000000136,
      -0.00000000016,    -0.00000000034,
      -0.00000028849,     0.00000013767,
       0.00000000590,     0.00000000207,
      -0.00000000048,    -0.00000000006,
       0.00000040645,    -0.00000049146,
      -0.00000000020,    -0.00000000449,
       0.00000000006,    -0.00000000046,
      -0.00000050147,    -0.00000041712,
      -0.00000000511,     0.00000000128,
      -0.00000000041,    -0.00000000008,
      -0.00000000423,     0.00000000068,
      -0.00000000021,    -0.00000000031,
      -0.00000000028,     0.00000000533,
       0.00000179425,    -0.00000209856,
      -0.00000000162,    -0.00000002293,
       0.00000000032,    -0.00000000185,
      -0.00000244390,    -0.00000159345,
       0.00000000299,     0.00000001887,
      -0.00000000128,    -0.00000000142,
       0.00000286834,    -0.00000013592,
       0.00000000787,     0.00000000597,
      -0.00000000003,    -0.00000000056,
      -0.00000007471,    -0.00000301036,
      -0.00000000451,    -0.00000000638,
      -0.00000000026,     0.00000000034,
      -0.00000000099,     0.00000000523,
      -0.00000000670,    -0.00000000312,
      -0.00000000174,     0.00000000004,
       0.00000000003,     0.00000000175,
      -0.00000000049,    -0.00000000136,
      -0.00000000038,     0.00000000019,
      -0.00000000138,     0.00000000050,
       0.00000000019,     0.00000000038,
      -0.00000000058,     0.00000000050,
       0.00000000049,     0.00000000058,
       0.00000000046,    -0.00000000017,
      -0.00000000010,    -0.00000000042,
       0.00001639012,    -0.00000053314,
       0.00000003699,     0.00000000427,
      -0.00000000085,    -0.00000000161,
      -0.00000036940,    -0.00001733481,
       0.00000000290,    -0.00000002588,
      -0.00000000211,     0.00000000066,
       0.00000071000,    -0.00000110928,
       0.00000001271,     0.00000006836,
      -0.00000000088,     0.00000000134,
       0.00000032264,    -0.00000044608,
       0.00000002609,    -0.00000004724,
       0.00000000090,     0.00000000319,
      -0.00000003381,     0.00000000423,
      -0.00000000135,    -0.00000000272,
      -0.00000000357,    -0.00000004237,
       0.00000000217,    -0.00000000172,
      -0.00000000980,     0.00000000016,
       0.00000000007,     0.00000000284,
      -0.00000000136,    -0.00000000400,
      -0.00000000112,     0.00000000062,
      -0.00000000376,     0.00000000256,
       0.00000000105,     0.00000000096,
       0.00000000467,    -0.00000000774,
      -0.00000000249,    -0.00000000106,
      -0.00000000790,    -0.00000000480,
      -0.00000000108,     0.00000000257,
       0.00000000137,    -0.00000000019,
      -0.00000000023,    -0.00000000150,
      -0.00000000185,     0.00000000080,
       0.00000000117,     0.00000000146,
      -0.00000000338,    -0.00000000010,
      -0.00000000017,     0.00000000350,
       0.00000000207,     0.00000000044,
       0.00000000023,    -0.00000000204,
       0.00000000136,    -0.00000000002,
      -0.00000000004,    -0.00000000143,
       0.00000441293,    -0.00000242623,
       0.00000020433,     0.00000034289,
      -0.00000001430,     0.00000000764,
      -0.00000186684,    -0.00000312528,
       0.00000024838,    -0.00000014894,
       0.00000000552,     0.00000001046,
       0.00000636710,    -0.00000002663,
       0.00000000270,     0.00000000381,
      -0.00000009205,     0.00000631301,
       0.00000000439,    -0.00000001642,
       0.00000000155,     0.00000000045,
      -0.00000000608,    -0.00000000030,
      -0.00000000128,    -0.00000000050,
       0.00000000045,    -0.00000000580,
       0.00000000050,    -0.00000000127,
       0.00000000048,     0.00000000061,
       0.00000000186,    -0.00000000237,
      -0.00000000237,    -0.00000000244,
       0.00000001667,    -0.00000002682,
      -0.00000000863,    -0.00000000377,
      -0.00000003043,    -0.00000001381,
      -0.00000000250,     0.00000000972,
       0.00000003228,    -0.00000000620,
      -0.00000000285,    -0.00000000912,
      -0.00000000575,    -0.00000003400,
      -0.00000000976,     0.00000000284,
       0.00000000039,     0.00000000013,
       0.00000000054,     0.00000000108,
       0.00000000133,    -0.00000000055,
      -0.00000000025,    -0.00000000088,
      -0.00000000090,     0.00000000075,
      -0.00000001331,    -0.00000000047,
      -0.00000000059,     0.00000000073,
      -0.00000000058,     0.00000001368,
       0.00000000073,     0.00000000064,
      -0.00000000233,     0.00000000064,
      -0.00000000004,     0.00000000225,
       0.00000000559,     0.00000000032,
       0.00000000013,    -0.00000000550,
       0.00000000044,    -0.00000000018,
      -0.00000013500,     0.00000004659,
      -0.00000001082,    -0.00000001674,
       0.00000000081,    -0.00000000106,
       0.00000003768,     0.00000009926,
      -0.00000001279,     0.00000000832,
      -0.00000000084,    -0.00000000061,
      -0.00000020355,    -0.00000000814,
      -0.00000000636,    -0.00000001109,
       0.00000000040,    -0.00000000033,
       0.00000001113,    -0.00000020069,
       0.00000001049,    -0.00000000565,
       0.00000000024,     0.00000000036,
       0.00000003323,     0.00000000236,
       0.00000000129,    -0.00000000220,
      -0.00000000224,     0.00000003338,
       0.00000000219,     0.00000000131,
       0.00000004268,    -0.00000004726,
       0.00000000808,     0.00000000567,
       0.00000000836,     0.00000000755,
      -0.00000000101,     0.00000000144,
      -0.00000000116,     0.00000000028,
       0.00000000028,     0.00000000123,
       0.00000013864,    -0.00000002494,
      -0.00000001188,    -0.00000003950,
      -0.00000002489,    -0.00000014711,
      -0.00000004199,     0.00000001225,
       0.00000000395,    -0.00000000887,
      -0.00000000271,    -0.00000000027,
       0.00000000354,    -0.00000000353,
      -0.00000000112,    -0.00000000124,
      -0.00000000038,     0.00000000112,
       0.00000000115,     0.00000000040,
      -0.00000000203,    -0.00000000287,
      -0.00000000032,     0.00000000020,
      -0.00000000293,     0.00000000205,
      -0.00000000049,     0.00000000021,
      -0.00000000019,     0.00000000065,
      -0.00000000777,     0.00000000139,
       0.00000000108,     0.00000000722,
      -0.00000000334,     0.00000000000,
       0.00000000047,    -0.00000000262,
       0.00000000009,     0.00000000097,
       0.00000000105,     0.00000000020,
      -0.00000000015,     0.00000000111,
       0.00000000027,    -0.00000000072,
      -0.00000000071,    -0.00000000035,
       0.00000000055,     0.00000000049,
       0.00000000048,    -0.00000000062,
       0.00000000325,     0.00000000276,
       0.00000000043,    -0.00000000147,
      -0.00000000019,    -0.00000000001,
       0.00000000283,    -0.00000000390,
      -0.00000000169,    -0.00000000040,
      -0.00000000299,     0.00000000181,
       0.00000000060,     0.00000000049,
       0.00000000228,     0.00000000289,
       0.00000000043,    -0.00000000084,
      -0.00000000005,     0.00000000152,
       0.00000000141,    -0.00000000091,
       0.00000000017,    -0.00000000046,
       0.00000000834,     0.00000000817,
       0.00000000064,    -0.00000000309,
      -0.00000000051,    -0.00000000018,
       0.00000003328,    -0.00000001734,
      -0.00000000390,    -0.00000000995,
      -0.00000000855,     0.00000004218,
       0.00000001200,    -0.00000000159,
      -0.00000003729,    -0.00000010909,
      -0.00000000968,     0.00000000885,
       0.00000000041,     0.00000000041,
      -0.00000011517,     0.00000004081,
       0.00000000931,     0.00000001044,
       0.00000000045,    -0.00000000043,
       0.00000002007,     0.00000000427,
       0.00000000278,     0.00000000170,
      -0.00000001340,    -0.00000003924,
      -0.00000000114,     0.00000000250,
       0.00000033085,    -0.00000004355,
       0.00000000069,     0.00000000269,
      -0.00000003124,    -0.00000034353,
       0.00000000193,    -0.00000000068,
      -0.00000000889,    -0.00000007894,
      -0.00000000533,     0.00000000692,
       0.00000005655,    -0.00000004335,
      -0.00000000404,    -0.00000000755,
      -0.00000000020,     0.00000000029,
      -0.00000000111,    -0.00000000032,
       0.00000000028,    -0.00000000120,
       0.00000023030,    -0.00000012139,
       0.00000001203,     0.00000001804,
      -0.00000000073,     0.00000000050,
      -0.00000009370,    -0.00000016797,
       0.00000001345,    -0.00000000929,
       0.00000000038,     0.00000000055,
       0.00000028652,    -0.00000000303,
       0.00000000064,     0.00000000028,
      -0.00000000521,     0.00000028173,
       0.00000000038,    -0.00000000076,
      -0.00000000082,    -0.00000000011,
       0.00000000011,    -0.00000000082,
      -0.00000000717,    -0.00000000523,
       0.00000000271,    -0.00000000301,
      -0.00000000302,    -0.00000000274,
      -0.00000000275,    -0.00000000627,
       0.00000000113,    -0.00000000023,
      -0.00000000452,     0.00000000202,
      -0.00000000020,    -0.00000000079,
      -0.00000000115,    -0.00000000718,
       0.00000000069,     0.00000000041,
       0.00000000699,    -0.00000000098,
      -0.00000000045,     0.00000000065,
       0.00000000110,    -0.00000000047,
       0.00000000048,     0.00000000111,
       0.00000010304,    -0.00000011421,
       0.00000001958,     0.00000001369,
       0.00000001153,     0.00000001042,
      -0.00000000139,     0.00000000198,
      -0.00000000076,    -0.00000000016,
      -0.00000000016,     0.00000000081,
       0.00000000243,     0.00000000022,
       0.00000000015,    -0.00000000263,
      -0.00000000089,    -0.00000000026,
       0.00000000013,    -0.00000000121,
       0.00000000435,    -0.00000000189,
       0.00000000018,     0.00000000031,
      -0.00000000136,    -0.00000000316,
       0.00000000527,    -0.00000000007,
      -0.00000000006,     0.00000000515,
      -0.00000001527,     0.00000001319,
       0.00000000378,    -0.00000000418,
       0.00000000072,     0.00000000050,
      -0.00000000417,    -0.00000000377,
       0.00000000050,    -0.00000000072,
       0.00000029041,    -0.00000032157,
       0.00000005494,     0.00000003850,
       0.00000001647,     0.00000001489,
      -0.00000000198,     0.00000000283,
      -0.00000000003,    -0.00000000063,
       0.00000000129,    -0.00000000058,
      -0.00000000042,    -0.00000000092,
       0.00000000162,    -0.00000000002,
      -0.00000000002,     0.00000000159,
       0.00000000471,    -0.00000000521,
       0.00000000090,     0.00000000063,
      -0.00000000521,    -0.00000000471,
       0.00000000063,    -0.00000000090,
       0.00000056761,    -0.00000062896,
       0.00000010820,     0.00000007588,
       0.00000002067,     0.00000001867,
      -0.00000000249,     0.00000000355,
       0.00000000024,    -0.00000000007,
       0.00000000001,     0.00000000005,
       0.00000000075,    -0.00000000008,
       0.00000000000,     0.00000000013,
       0.00000000008,     0.00000000075,
      -0.00000000013,     0.00000000000,
       0.00000000003,    -0.00000000003,
       0.00000000004,     0.00000000004,
       0.00000000008,     0.00000000115,
      -0.00000000116,     0.00000000008,
      -0.00000000222,    -0.00000000472,
       0.00000000164,    -0.00000000102,
       0.00000000017,     0.00000000024,
       0.00000000299,    -0.00000000143,
       0.00000000077,     0.00000000115,
       0.00000000089,    -0.00000001834,
       0.00000000477,    -0.00000000018,
       0.00000000002,     0.00000000055,
       0.00000001849,     0.00000000074,
       0.00000000027,     0.00000000483,
      -0.00000000056,     0.00000000004,
       0.00000000245,    -0.00000000087,
       0.00000000075,     0.00000000241,
       0.00000000030,    -0.00000000091,
       0.00000000091,     0.00000000030,
       0.00000000041,    -0.00000000123,
       0.00000000124,     0.00000000040,
       0.00000000058,    -0.00000000175,
       0.00000000175,     0.00000000058,
       0.00000000072,    -0.00000000219,
       0.00000000220,     0.00000000072,
      -0.00000000096,     0.00000000039,
       0.00000000078,     0.00000000070,
      -0.00000000301,    -0.00000000161,
      -0.00000000163,     0.00000000305,
       0.00000000012,    -0.00000000053,
      -0.00000001033,    -0.00000000589,
      -0.00000000207,     0.00000001144,
      -0.00000000535,    -0.00000002841,
      -0.00000000035,    -0.00000000019,
      -0.00000002902,     0.00000000534,
      -0.00000000017,     0.00000000040,
       0.00000000028,     0.00000000027,
      -0.00000001852,    -0.00000011604,
      -0.00000000154,    -0.00000000093,
      -0.00000000008,    -0.00000000013,
      -0.00000010398,     0.00000004156,
       0.00000000131,     0.00000000092,
       0.00000011668,    -0.00000014523,
       0.00000000022,    -0.00000000157,
      -0.00000014791,    -0.00000012264,
      -0.00000000185,     0.00000000015,
      -0.00000000094,     0.00000000027,
      -0.00000000015,     0.00000000129,
       0.00000058383,    -0.00000069382,
       0.00000000057,    -0.00000000891,
       0.00000000003,    -0.00000000061,
      -0.00000070801,    -0.00000048214,
      -0.00000000164,     0.00000000379,
      -0.00000000042,    -0.00000000029,
       0.00000069096,    -0.00000007487,
       0.00000000377,     0.00000000186,
      -0.00000003074,    -0.00000073959,
      -0.00000000122,    -0.00000000366,
      -0.00000000063,     0.00000000174,
      -0.00000000200,    -0.00000000098,
      -0.00000000017,    -0.00000000050,
      -0.00000000051,     0.00000000018,
       0.00000421912,    -0.00000029388,
       0.00000002110,     0.00000000179,
      -0.00000000043,    -0.00000000029,
      -0.00000020488,    -0.00000401196,
       0.00000000134,    -0.00000001467,
      -0.00000000045,     0.00000000033,
       0.00000032620,    -0.00000041538,
       0.00000000787,     0.00000002137,
      -0.00000000048,     0.00000000044,
       0.00000012001,    -0.00000011222,
       0.00000001366,    -0.00000001579,
       0.00000000049,     0.00000000106,
      -0.00000000864,     0.00000000127,
      -0.00000000039,    -0.00000000067,
      -0.00000000116,    -0.00000001071,
       0.00000000054,    -0.00000000048,
      -0.00000000061,    -0.00000000191,
      -0.00000000054,     0.00000000028,
      -0.00000000165,     0.00000000097,
       0.00000000040,     0.00000000043,
       0.00000000160,    -0.00000000273,
      -0.00000000085,    -0.00000000037,
      -0.00000000279,    -0.00000000170,
      -0.00000000040,     0.00000000087,
      -0.00000000013,    -0.00000000037,
      -0.00000000072,     0.00000000051,
       0.00000000064,     0.00000000061,
      -0.00000000109,     0.00000000001,
      -0.00000000006,     0.00000000115,
       0.00000000083,    -0.00000000002,
      -0.00000000014,    -0.00000000066,
       0.00000000050,    -0.00000000003,
      -0.00000000002,    -0.00000000054,
       0.00000164888,    -0.00000097009,
       0.00000008465,     0.00000012680,
      -0.00000000534,     0.00000000337,
      -0.00000071248,    -0.00000115502,
       0.00000009085,    -0.00000005889,
       0.00000000227,     0.00000000386,
       0.00000184951,    -0.00000012048,
       0.00000001171,     0.00000000379,
      -0.00000000040,     0.00000000037,
       0.00000004066,     0.00000180850,
       0.00000000241,    -0.00000000110,
       0.00000000066,     0.00000000015,
      -0.00000000214,     0.00000000035,
      -0.00000000049,    -0.00000000029,
      -0.00000000043,    -0.00000000230,
       0.00000000029,    -0.00000000050,
       0.00000000672,    -0.00000001094,
      -0.00000000352,    -0.00000000154,
      -0.00000001095,    -0.00000000528,
      -0.00000000102,     0.00000000351,
       0.00000000953,    -0.00000000240,
      -0.00000000100,    -0.00000000268,
      -0.00000000189,    -0.00000001031,
      -0.00000000298,     0.00000000087,
      -0.00000000031,     0.00000000038,
      -0.00000000472,    -0.00000000003,
      -0.00000000022,     0.00000000025,
      -0.00000000003,     0.00000000437,
      -0.00000000101,     0.00000000044,
      -0.00000000011,     0.00000000073,
       0.00000000216,    -0.00000000006,
      -0.00000000007,    -0.00000000178,
       0.00000000035,    -0.00000000018,
       0.00000024772,    -0.00000035347,
       0.00000005842,     0.00000003501,
      -0.00000000278,     0.00000000470,
      -0.00000017718,    -0.00000010962,
       0.00000001583,    -0.00000002890,
       0.00000000230,     0.00000000132,
       0.00000071040,    -0.00000037297,
       0.00000003285,     0.00000005440,
      -0.00000000226,     0.00000000130,
       0.00000036512,     0.00000071029,
      -0.00000005448,     0.00000003100,
      -0.00000000108,    -0.00000000227,
       0.00000000841,     0.00000000081,
       0.00000000024,    -0.00000000061,
      -0.00000000079,     0.00000000844,
       0.00000000061,     0.00000000024,
       0.00000004729,    -0.00000001017,
      -0.00000000443,    -0.00000001340,
      -0.00000000888,    -0.00000004519,
      -0.00000001285,     0.00000000404,
       0.00000000276,    -0.00000000475,
      -0.00000000144,    -0.00000000040,
       0.00000000163,    -0.00000000126,
      -0.00000000036,    -0.00000000059,
       0.00000000047,     0.00000000010,
      -0.00000000071,    -0.00000000106,
      -0.00000000107,     0.00000000071,
      -0.00000000353,     0.00000000062,
       0.00000000051,     0.00000000329,
      -0.00000000159,     0.00000000007,
       0.00000000012,    -0.00000000125,
       0.00000000064,    -0.00000000003,
      -0.00000000009,     0.00000000067,
      -0.00000000900,     0.00000000829,
      -0.00000000213,    -0.00000000164,
       0.00000000011,    -0.00000000026,
       0.00000000434,     0.00000000439,
      -0.00000000084,     0.00000000106,
      -0.00000002247,     0.00000000766,
      -0.00000000185,    -0.00000000269,
      -0.00000000744,    -0.00000002239,
       0.00000000267,    -0.00000000178,
       0.00000000316,    -0.00000000095,
       0.00000000096,     0.00000000316,
       0.00000000216,    -0.00000000658,
       0.00000000119,     0.00000000039,
       0.00000001193,    -0.00000000811,
      -0.00000000219,    -0.00000000233,
      -0.00000000555,    -0.00000000803,
      -0.00000000155,     0.00000000151,
       0.00000001561,    -0.00000000320,
      -0.00000000141,    -0.00000000442,
       0.00000000277,     0.00000001533,
       0.00000000436,    -0.00000000135,
       0.00000000041,    -0.00000000017,
      -0.00000000095,    -0.00000000009,
      -0.00000000008,     0.00000000069,
      -0.00000000034,    -0.00000000030,
      -0.00000000113,     0.00000000070,
       0.00000000044,     0.00000000074,
      -0.00000000132,     0.00000000020,
      -0.00000000015,    -0.00000000129,
       0.00000000088,     0.00000000066,
       0.00000000010,    -0.00000000035,
       0.00000000063,    -0.00000000087,
      -0.00000000052,     0.00000000031,
       0.00000000049,     0.00000000056,
      -0.00000000421,     0.00000000304,
       0.00000000104,     0.00000000062,
       0.00000000311,     0.00000000406,
       0.00000000060,    -0.00000000102,
      -0.00000000034,    -0.00000000162,
      -0.00000000220,     0.00000000133,
      -0.00000000637,    -0.00000001942,
      -0.00000000172,     0.00000000142,
      -0.00000001803,     0.00000000666,
       0.00000000142,     0.00000000157,
       0.00000000434,    -0.00000000111,
      -0.00000000113,    -0.00000000232,
       0.00000003985,    -0.00000001159,
       0.00000000025,     0.00000000070,
      -0.00000000825,    -0.00000003843,
       0.00000000050,    -0.00000000019,
      -0.00000000258,    -0.00000001509,
      -0.00000000088,     0.00000000138,
       0.00000001067,    -0.00000000875,
      -0.00000000085,    -0.00000000134,
       0.00000004164,    -0.00000002688,
       0.00000000271,     0.00000000342,
      -0.00000000014,     0.00000000013,
      -0.00000001938,    -0.00000003049,
       0.00000000251,    -0.00000000195,
       0.00000003763,    -0.00000000811,
       0.00000000030,     0.00000000053,
       0.00000000538,     0.00000003608,
      -0.00000000031,    -0.00000000018,
       0.00000001457,    -0.00000001984,
       0.00000000346,     0.00000000199,
      -0.00000000015,     0.00000000029,
      -0.00000001045,    -0.00000000706,
       0.00000000097,    -0.00000000179,
       0.00000003633,    -0.00000001865,
       0.00000000185,     0.00000000278,
       0.00000001803,     0.00000003629,
      -0.00000000278,     0.00000000170,
      -0.00000000046,    -0.00000000046,
      -0.00000000046,    -0.00000000097,
       0.00000000096,    -0.00000000045,
       0.00000000524,    -0.00000001599,
       0.00000000164,     0.00000000054,
       0.00000000069,    -0.00000000031,
       0.00000000030,     0.00000000069,
       0.00000001483,    -0.00000004518,
       0.00000000233,     0.00000000077,
       0.00000002899,    -0.00000008843,
       0.00000000292,     0.00000000096,
       0.00000000002,    -0.00000000004,
       0.00000000008,    -0.00000000010,
       0.00000000010,     0.00000000008,
      -0.00000000057,    -0.00000000046,
       0.00000000019,    -0.00000000032,
       0.00000000029,    -0.00000000033,
      -0.00000000112,    -0.00000000245,
       0.00000000089,    -0.00000000052,
       0.00000000246,    -0.00000000114,
       0.00000000054,     0.00000000089,
       0.00000000037,    -0.00000000035,
       0.00000000033,     0.00000000038,
       0.00000000031,     0.00000000034,
      -0.00000000115,    -0.00000000064,
      -0.00000000066,     0.00000000117,
      -0.00000000445,    -0.00000000265,
      -0.00000000114,     0.00000000448,
      -0.00000000185,    -0.00000000973,
      -0.00000001002,     0.00000000177,
      -0.00000000707,    -0.00000004398,
      -0.00000000061,    -0.00000000046,
      -0.00000003691,     0.00000001330,
       0.00000000025,     0.00000000039,
       0.00000003360,    -0.00000004429,
       0.00000000014,    -0.00000000056,
      -0.00000004490,    -0.00000003652,
      -0.00000000067,    -0.00000000003,
       0.00000018393,    -0.00000022670,
       0.00000000060,    -0.00000000334,
      -0.00000000001,    -0.00000000019,
      -0.00000021175,    -0.00000014503,
      -0.00000000115,     0.00000000063,
       0.00000017458,    -0.00000003355,
       0.00000000158,     0.00000000053,
      -0.00000001203,    -0.00000019136,
      -0.00000000033,    -0.00000000156,
      -0.00000000025,     0.00000000057,
      -0.00000000061,    -0.00000000030,
       0.00000111844,    -0.00000013302,
       0.00000000969,     0.00000000063,
      -0.00000000019,    -0.00000000001,
      -0.00000009210,    -0.00000099062,
       0.00000000051,    -0.00000000667,
       0.00000012143,    -0.00000014518,
       0.00000000348,     0.00000000653,
       0.00000004045,    -0.00000003032,
       0.00000000536,    -0.00000000527,
       0.00000000020,     0.00000000034,
      -0.00000000227,     0.00000000041,
      -0.00000000040,    -0.00000000280,
       0.00000056367,    -0.00000035099,
       0.00000003130,     0.00000004285,
      -0.00000000182,     0.00000000130,
      -0.00000025000,    -0.00000039098,
       0.00000003043,    -0.00000002120,
       0.00000000085,     0.00000000130,
       0.00000051878,    -0.00000006488,
       0.00000000636,     0.00000000182,
       0.00000002503,     0.00000049803,
       0.00000000128,    -0.00000000008,
      -0.00000000072,     0.00000000021,
      -0.00000000026,    -0.00000000082,
      -0.00000000005,    -0.00000000057,
       0.00000010997,    -0.00000016954,
       0.00000002800,     0.00000001541,
      -0.00000000121,     0.00000000229,
      -0.00000008990,    -0.00000005330,
       0.00000000771,    -0.00000001463,
       0.00000000119,     0.00000000063,
       0.00000026417,    -0.00000015384,
       0.00000001433,     0.00000001997,
      -0.00000000084,     0.00000000064,
       0.00000014721,     0.00000026407,
      -0.00000002004,     0.00000001276,
      -0.00000000045,    -0.00000000084,
       0.00000000218,     0.00000000025,
      -0.00000000025,     0.00000000219,
      -0.00000000026,     0.00000000035,
       0.00000000713,    -0.00000003937,
       0.00000000939,     0.00000000086,
      -0.00000000005,     0.00000000113,
      -0.00000001372,    -0.00000000159,
       0.00000000009,    -0.00000000329,
       0.00000005227,    -0.00000007081,
       0.00000001186,     0.00000000735,
      -0.00000000056,     0.00000000096,
       0.00000007031,     0.00000005251,
      -0.00000000744,     0.00000001170,
      -0.00000000094,    -0.00000000058,
       0.00000000112,    -0.00000000037,
       0.00000000037,     0.00000000112,
      -0.00000000045,     0.00000000104,
      -0.00000000032,    -0.00000000009,
       0.00000000038,     0.00000000015,
      -0.00000000188,     0.00000000170,
      -0.00000000042,    -0.00000000033,
      -0.00000000168,    -0.00000000189,
       0.00000000033,    -0.00000000041,
      -0.00000000079,     0.00000000064,
       0.00000000060,     0.00000000072,
      -0.00000000035,     0.00000000028,
      -0.00000000147,    -0.00000000277,
      -0.00000000019,     0.00000000026,
      -0.00000000255,     0.00000000127,
      -0.00000000070,    -0.00000000087,
       0.00000000446,    -0.00000000277,
      -0.00000000184,    -0.00000000423,
      -0.00000000041,    -0.00000000301,
       0.00000000217,    -0.00000000141,
      -0.00000000027,    -0.00000000084,
      -0.00000000063,     0.00000000034,
       0.00000000053,    -0.00000000097,
      -0.00000000099,    -0.00000000059,
       0.00000000253,    -0.00000000427,
      -0.00000000136,    -0.00000000058,
      -0.00000000393,    -0.00000000191,
      -0.00000000038,     0.00000000124,
       0.00000000286,    -0.00000000097,
      -0.00000000035,    -0.00000000078,
      -0.00000000068,    -0.00000000325,
      -0.00000000091,     0.00000000030,
       0.00000001556,    -0.00000000410,
      -0.00000000162,    -0.00000000438,
      -0.00000000326,    -0.00000001376,
      -0.00000000392,     0.00000000137,
       0.00000000134,    -0.00000000214,
      -0.00000000063,    -0.00000000022,
       0.00000000061,    -0.00000000041,
       0.00000000542,    -0.00000000413,
      -0.00000000108,    -0.00000000107,
      -0.00000000281,    -0.00000000369,
      -0.00000000072,     0.00000000076,
       0.00000000572,    -0.00000000150,
      -0.00000000059,    -0.00000000162,
       0.00000000116,     0.00000000554,
       0.00000000159,    -0.00000000053,
       0.00000000073,    -0.00000000120,
      -0.00000000051,    -0.00000000032,
       0.00000000193,    -0.00000000137,
      -0.00000000036,    -0.00000000038,
       0.00000000134,     0.00000000192,
       0.00000000038,    -0.00000000035,
       0.00000000677,    -0.00000000521,
       0.00000000055,     0.00000000059,
      -0.00000000365,    -0.00000000500,
       0.00000000041,    -0.00000000037,
       0.00000000471,    -0.00000000212,
       0.00000000143,     0.00000000434,
       0.00000000308,    -0.00000000501,
       0.00000000087,     0.00000000042,
      -0.00000000278,    -0.00000000167,
       0.00000000023,    -0.00000000048,
       0.00000000636,    -0.00000000431,
       0.00000000044,     0.00000000050,
       0.00000000405,     0.00000000634,
      -0.00000000050,     0.00000000038,
       0.00000000046,    -0.00000000235,
       0.00000000057,     0.00000000005,
      -0.00000000090,    -0.00000000013,
       0.00000000293,    -0.00000000386,
       0.00000000067,     0.00000000040,
       0.00000000382,     0.00000000295,
      -0.00000000040,     0.00000000065,
       0.00000000037,     0.00000000028,
      -0.00000000163,     0.00000000010,
       0.00000000002,     0.00000000137,
      -0.00000000031,     0.00000000019,
      -0.00000000156,     0.00000000034,
       0.00000000027,     0.00000000129,
      -0.00000000068,     0.00000000003,
       0.00000000003,    -0.00000000054,
      -0.00000000049,     0.00000000034,
      -0.00000000058,     0.00000000013,
      -0.00000000010,    -0.00000000056,
      -0.00000000030,    -0.00000000024,
       0.00000000025,    -0.00000000030,
      -0.00000000044,    -0.00000000025,
      -0.00000000026,     0.00000000044,
      -0.00000000183,    -0.00000000111,
      -0.00000000053,     0.00000000172,
      -0.00000000067,    -0.00000000331,
      -0.00000000343,     0.00000000061,
      -0.00000000273,    -0.00000001603,
      -0.00000000023,    -0.00000000020,
      -0.00000001288,     0.00000000435,
       0.00000000961,    -0.00000001374,
      -0.00000001381,    -0.00000001091,
       0.00000005647,    -0.00000007352,
       0.00000000032,    -0.00000000125,
      -0.00000006449,    -0.00000004321,
      -0.00000000055,     0.00000000005,
       0.00000004540,    -0.00000001362,
       0.00000000062,     0.00000000014,
      -0.00000000451,    -0.00000005104,
      -0.00000000009,    -0.00000000065,
       0.00000030239,    -0.00000005464,
       0.00000000401,     0.00000000014,
      -0.00000003748,    -0.00000025425,
       0.00000000014,    -0.00000000273,
       0.00000004141,    -0.00000004932,
       0.00000000139,     0.00000000192,
       0.00000001316,    -0.00000000859,
       0.00000000190,    -0.00000000176,
      -0.00000000061,     0.00000000014,
      -0.00000000014,    -0.00000000075,
       0.00000018446,    -0.00000012271,
       0.00000001126,     0.00000001380,
      -0.00000000059,     0.00000000049,
      -0.00000008522,    -0.00000012691,
       0.00000000974,    -0.00000000744,
       0.00000000031,     0.00000000042,
       0.00000014552,    -0.00000002916,
       0.00000000287,     0.00000000072,
       0.00000001179,     0.00000013669,
       0.00000000056,     0.00000000006,
       0.00000004339,    -0.00000007127,
       0.00000001178,     0.00000000604,
      -0.00000000047,     0.00000000097,
      -0.00000003923,    -0.00000002226,
       0.00000000320,    -0.00000000639,
       0.00000000052,     0.00000000026,
       0.00000008971,    -0.00000005685,
       0.00000000551,     0.00000000671,
      -0.00000000028,     0.00000000027,
       0.00000005291,     0.00000008967,
      -0.00000000674,     0.00000000458,
       0.00000000057,     0.00000000008,
      -0.00000000008,     0.00000000057,
       0.00000000274,    -0.00000002105,
       0.00000000500,     0.00000000023,
       0.00000000000,     0.00000000061,
      -0.00000000836,    -0.00000000063,
       0.00000000001,    -0.00000000201,
       0.00000002202,    -0.00000003287,
       0.00000000555,     0.00000000305,
      -0.00000000023,     0.00000000046,
       0.00000003238,     0.00000002224,
      -0.00000000313,     0.00000000540,
      -0.00000000043,    -0.00000000024,
       0.00000000038,    -0.00000000013,
       0.00000000013,     0.00000000038,
      -0.00000000098,    -0.00000000378,
       0.00000000116,    -0.00000000042,
       0.00000000008,     0.00000000018,
      -0.00000000094,     0.00000000033,
       0.00000000186,    -0.00000000929,
       0.00000000223,     0.00000000024,
       0.00000000926,     0.00000000189,
      -0.00000000025,     0.00000000222,
      -0.00000000036,     0.00000000023,
       0.00000000043,    -0.00000000051,
      -0.00000000037,    -0.00000000043,
      -0.00000000015,    -0.00000000042,
       0.00000000026,    -0.00000000030,
       0.00000000101,    -0.00000000098,
      -0.00000000067,    -0.00000000076,
       0.00000000053,    -0.00000000047,
       0.00000000031,     0.00000000046,
       0.00000000090,    -0.00000000160,
      -0.00000000050,    -0.00000000021,
      -0.00000000139,    -0.00000000066,
      -0.00000000014,     0.00000000043,
       0.00000000078,    -0.00000000036,
      -0.00000000022,    -0.00000000090,
       0.00000000495,    -0.00000000162,
      -0.00000000060,    -0.00000000140,
      -0.00000000122,    -0.00000000420,
      -0.00000000120,     0.00000000047,
       0.00000000056,    -0.00000000088,
       0.00000000223,    -0.00000000184,
      -0.00000000047,    -0.00000000043,
      -0.00000000125,    -0.00000000152,
      -0.00000000029,     0.00000000032,
       0.00000000199,    -0.00000000067,
      -0.00000000027,    -0.00000000055,
       0.00000000047,     0.00000000189,
       0.00000000053,    -0.00000000020,
       0.00000000038,    -0.00000000071,
      -0.00000000041,    -0.00000000017,
       0.00000000091,    -0.00000000071,
       0.00000000067,     0.00000000091,
       0.00000000058,    -0.00000000111,
      -0.00000000064,    -0.00000000034,
       0.00000000102,    -0.00000000086,
       0.00000000078,     0.00000000102,
       0.00000000058,    -0.00000000095,
       0.00000000093,     0.00000000059,
       0.00000000012,    -0.00000000054,
       0.00000000053,     0.00000000012,
       0.00000000008,     0.00000000052,
      -0.00000000062,     0.00000000018,
       0.00000000012,     0.00000000038,
      -0.00000000073,    -0.00000000044,
      -0.00000000022,     0.00000000065,
      -0.00000000025,    -0.00000000111,
      -0.00000000116,     0.00000000022,
      -0.00000000106,    -0.00000000569,
      -0.00000000442,     0.00000000146,
       0.00000000271,    -0.00000000430,
      -0.00000000428,    -0.00000000325,
       0.00000001697,    -0.00000002378,
       0.00000000014,    -0.00000000045,
      -0.00000001984,    -0.00000001268,
       0.00000001199,    -0.00000000520,
      -0.00000000164,    -0.00000001392,
       0.00000008271,    -0.00000002110,
       0.00000000155,     0.00000000000,
      -0.00000001436,    -0.00000006690,
       0.00000000002,    -0.00000000105,
       0.00000001345,    -0.00000001649,
       0.00000000053,     0.00000000055,
       0.00000000418,    -0.00000000252,
       0.00000000064,    -0.00000000059,
       0.00000005863,    -0.00000004211,
       0.00000000398,     0.00000000429,
      -0.00000000018,     0.00000000019,
      -0.00000002865,    -0.00000004009,
       0.00000000301,    -0.00000000259,
       0.00000004102,    -0.00000001194,
       0.00000000118,     0.00000000026,
       0.00000000494,     0.00000003761,
       0.00000001596,    -0.00000002805,
       0.00000000464,     0.00000000219,
      -0.00000000017,     0.00000000039,
      -0.00000001580,    -0.00000000852,
       0.00000000123,    -0.00000000264,
       0.00000002924,    -0.00000002034,
       0.00000000206,     0.00000000213,
       0.00000001836,     0.00000002925,
      -0.00000000223,     0.00000000163,
       0.00000000093,    -0.00000000980,
       0.00000000233,     0.00000000003,
       0.00000000001,     0.00000000028,
      -0.00000000424,    -0.00000000021,
      -0.00000000002,    -0.00000000102,
       0.00000000835,    -0.00000001349,
       0.00000000230,     0.00000000114,
       0.00000001317,     0.00000000850,
      -0.00000000120,     0.00000000220,
      -0.00000000068,    -0.00000000215,
       0.00000000066,    -0.00000000028,
      -0.00000000065,     0.00000000026,
       0.00000000070,    -0.00000000471,
       0.00000000113,     0.00000000006,
       0.00000000468,     0.00000000073,
      -0.00000000008,     0.00000000112,
      -0.00000000024,    -0.00000000100,
       0.00000000099,    -0.00000000023,
       0.00000000030,    -0.00000000059,
      -0.00000000045,    -0.00000000020,
       0.00000000156,    -0.00000000063,
      -0.00000000021,    -0.00000000043,
      -0.00000000046,    -0.00000000128,
      -0.00000000035,     0.00000000018,
       0.00000000019,    -0.00000000032,
       0.00000000089,    -0.00000000078,
      -0.00000000052,    -0.00000000055,
       0.00000000061,    -0.00000000022,
       0.00000000022,     0.00000000056,
       0.00000000033,    -0.00000000032,
       0.00000000029,     0.00000000034,
      -0.00000000041,    -0.00000000197,
      -0.00000000149,     0.00000000050,
       0.00000000075,    -0.00000000136,
      -0.00000000133,    -0.00000000096,
       0.00000000501,    -0.00000000766,
      -0.00000000614,    -0.00000000368,
       0.00000000319,    -0.00000000191,
      -0.00000000058,    -0.00000000385,
       0.00000002279,    -0.00000000782,
       0.00000000058,    -0.00000000002,
      -0.00000000528,    -0.00000001790,
       0.00000000422,    -0.00000000546,
       0.00000000131,    -0.00000000076,
       0.00000001824,    -0.00000001429,
       0.00000000139,     0.00000000129,
      -0.00000000956,    -0.00000001242,
       0.00000000095,    -0.00000000093,
       0.00000001160,    -0.00000000461,
       0.00000000047,     0.00000000013,
       0.00000000193,     0.00000001037,
       0.00000000558,    -0.00000001058,
       0.00000000177,     0.00000000074,
      -0.00000000006,     0.00000000015,
      -0.00000000605,    -0.00000000307,
       0.00000000044,    -0.00000000101,
       0.00000000925,    -0.00000000714,
       0.00000000078,     0.00000000068,
       0.00000000624,     0.00000000927,
      -0.00000000070,     0.00000000056,
      -0.00000000016,    -0.00000000035,
       0.00000000026,    -0.00000000422,
       0.00000000102,    -0.00000000002,
      -0.00000000194,    -0.00000000005,
      -0.00000000002,    -0.00000000047,
       0.00000000297,    -0.00000000523,
       0.00000000090,     0.00000000040,
       0.00000000505,     0.00000000306,
      -0.00000000043,     0.00000000084,
      -0.00000000039,    -0.00000000107,
       0.00000000033,    -0.00000000015,
      -0.00000000037,     0.00000000016,
       0.00000000023,    -0.00000000210,
       0.00000000051,     0.00000000001,
       0.00000000208,     0.00000000025,
      -0.00000000002,     0.00000000050,
      -0.00000000016,    -0.00000000053,
       0.00000000053,    -0.00000000015,
      -0.00000000016,    -0.00000000067,
      -0.00000000050,     0.00000000017,
       0.00000000020,    -0.00000000043,
      -0.00000000042,    -0.00000000028,
       0.00000000145,    -0.00000000246,
      -0.00000000191,    -0.00000000105,
       0.00000000085,    -0.00000000068,
      -0.00000000020,    -0.00000000108,
       0.00000000631,    -0.00000000281,
      -0.00000000188,    -0.00000000484,
       0.00000000129,    -0.00000000180,
       0.00000000040,    -0.00000000024,
       0.00000000558,    -0.00000000481,
       0.00000000049,     0.00000000039,
      -0.00000000317,    -0.00000000379,
       0.00000000028,    -0.00000000032,
       0.00000000328,    -0.00000000171,
       0.00000000072,     0.00000000286,
       0.00000000187,    -0.00000000387,
       0.00000000066,     0.00000000025,
      -0.00000000224,    -0.00000000106,
       0.00000000015,    -0.00000000038,
       0.00000000286,    -0.00000000247,
       0.00000000028,     0.00000000020,
       0.00000000209,     0.00000000288,
       0.00000000005,    -0.00000000172,
       0.00000000042,    -0.00000000002,
      -0.00000000083,     0.00000000000,
       0.00000000101,    -0.00000000195,
       0.00000000187,     0.00000000105,
       0.00000000006,    -0.00000000088,
       0.00000000087,     0.00000000008,
       0.00000000041,    -0.00000000079,
      -0.00000000060,    -0.00000000029,
       0.00000000175,    -0.00000000099,
      -0.00000000066,    -0.00000000132,
       0.00000000038,    -0.00000000059,
       0.00000000168,    -0.00000000161,
      -0.00000000105,    -0.00000000114,
       0.00000000093,    -0.00000000062,
       0.00000000026,     0.00000000079,
       0.00000000061,    -0.00000000138,
      -0.00000000081,    -0.00000000035,
       0.00000000087,    -0.00000000085,
       0.00000000070,     0.00000000088,
       0.00000000033,    -0.00000000071,
       0.00000000067,     0.00000000035,
       0.00000000048,    -0.00000000034,
      -0.00000000023,    -0.00000000036,
       0.00000000050,    -0.00000000054,
      -0.00000000035,    -0.00000000034,
       0.00000000026,    -0.00000000029,
      -0.00000085202,    -0.00000258727,
       0.00000012190,    -0.00000011320,
       0.00000000474,     0.00000000400,
      -0.00000145785,    -0.00000067437,
       0.00000001093,    -0.00000002236,
       0.00000000047,     0.00000000091,
       0.00000033880,    -0.00000140246,
       0.00000000574,    -0.00000001730,
      -0.00000000006,    -0.00000000064,
      -0.00000002150,     0.00000006922,
      -0.00000000329,     0.00000000403,
       0.00000000000,    -0.00000000014,
      -0.00000006423,    -0.00000014744,
       0.00000000033,    -0.00000000679,
       0.00000000022,     0.00000000011,
       0.00000166317,    -0.00000740629,
       0.00000034623,    -0.00000016102,
       0.00000000986,     0.00000001075,
      -0.00000653094,    -0.00000003346,
       0.00000000095,     0.00000000603,
       0.00000005753,    -0.00000658243,
       0.00000000498,    -0.00000000261,
      -0.00000003388,     0.00000009079,
      -0.00000000654,     0.00000000164,
      -0.00000000002,    -0.00000000053,
       0.00000002160,    -0.00000006350,
       0.00000000400,    -0.00000001279,
       0.00000173739,    -0.00001251515,
       0.00000061248,    -0.00000036454,
       0.00000002059,     0.00000001981,
      -0.00001497222,    -0.00000025523,
       0.00000000288,     0.00000000818,
       0.00000014566,    -0.00001510907,
       0.00000001091,    -0.00000001517,
       0.00000000088,     0.00000000028,
       0.00000002161,    -0.00000009633,
       0.00000000609,    -0.00000000122,
      -0.00000000006,     0.00000000050,
      -0.00000001306,    -0.00000000904,
       0.00000000136,     0.00000000831,
      -0.00000000109,    -0.00000000083,
       0.00001661638,    -0.00000288424,
      -0.00000047152,     0.00000041009,
      -0.00000000628,    -0.00000002852,
      -0.00004629930,     0.00000132824,
      -0.00000003925,    -0.00000007681,
       0.00000000260,    -0.00000000693,
      -0.00000133848,    -0.00004514523,
      -0.00000005651,    -0.00000001916,
       0.00000000324,    -0.00000000193,
      -0.00000022876,    -0.00000007195,
      -0.00000000773,     0.00000002288,
       0.00000000124,     0.00000000266,
       0.00000028283,    -0.00000014632,
      -0.00000002366,    -0.00000001278,
      -0.00000000296,     0.00000000105,
      -0.00001980506,     0.00000803192,
      -0.00000156484,    -0.00000190713,
       0.00000013674,    -0.00000010442,
       0.00000000319,     0.00000000700,
       0.00000738864,     0.00001668416,
      -0.00000157595,     0.00000141386,
      -0.00000009761,    -0.00000010834,
       0.00000000539,    -0.00000000331,
      -0.00002955053,    -0.00000167795,
       0.00000044260,    -0.00000006986,
       0.00000000176,     0.00000005913,
      -0.00000000572,    -0.00000000063,
      -0.00000155854,    -0.00002838458,
      -0.00000007342,    -0.00000040437,
       0.00000006875,    -0.00000000208,
      -0.00000000079,     0.00000000554,
      -0.00000004362,     0.00000019800,
       0.00000001470,     0.00000004685,
       0.00000000135,    -0.00000000104,
      -0.00000006497,    -0.00000000900,
      -0.00000004325,     0.00000000606,
       0.00000000288,     0.00000000038,
       0.00003726944,    -0.00002189390,
       0.00000290759,     0.00000487997,
      -0.00000039375,     0.00000012344,
       0.00000000400,    -0.00000002122,
       0.00014340344,    -0.00002986817,
      -0.00000724800,     0.00001208562,
      -0.00000070090,    -0.00000062225,
       0.00000003188,    -0.00000003157,
       0.00197267636,     0.00000666684,
      -0.00000160045,     0.00000030172,
      -0.00000000898,     0.00000001620,
       0.00000002275,     0.00000000384,
       0.00000660261,     0.00196886512,
       0.00000018228,     0.00000204169,
      -0.00000057013,     0.00000000105,
       0.00000000429,    -0.00000002372,
       0.00000015541,    -0.00000056136,
       0.00000002656,    -0.00000020416,
      -0.00000001937,    -0.00000000023,
       0.00000000049,     0.00000000260,
      -0.00000056425,     0.00000098506,
       0.00000027088,     0.00000001022,
      -0.00000000514,    -0.00000002092,
      -0.00000000294,    -0.00000000023,
       0.00000927275,     0.00000382185,
       0.00000065308,     0.00000082103,
      -0.00000005360,     0.00000005275,
       0.00002355019,     0.00018719456,
      -0.00000301110,    -0.00000080670,
       0.00000002539,     0.00000001041,
      -0.00000000178,     0.00000000161,
       0.00274532741,    -0.00000017282,
      -0.00000020643,    -0.00000007682,
       0.00000000626,     0.00000040469,
      -0.00000040413,     0.00274676146,
      -0.00000008068,    -0.00000015382,
      -0.00000041175,    -0.00000000521,
      -0.00000050054,     0.00000060786,
      -0.00000002215,    -0.00000001699,
      -0.00000000186,     0.00000000228,
       0.00000005141,    -0.00000007253,
      -0.00000000559,     0.00000001169,
       0.00004794660,     0.00001167738,
       0.00000128875,     0.00000226145,
      -0.00000011869,     0.00000011382,
       0.00003456218,    -0.00003125715,
       0.00000041534,    -0.00000004997,
       0.00000000989,     0.00000001234,
       0.00343810366,    -0.00000119588,
      -0.00000025921,    -0.00000004221,
       0.00000001157,     0.00000050330,
       0.00000026002,     0.00344034783,
      -0.00000012926,    -0.00000024898,
      -0.00000051722,    -0.00000000399,
      -0.00000016857,     0.00000000606,
       0.00000000620,     0.00000000514,
      -0.00000018831,     0.00000000821,
       0.00000000523,     0.00000000606,
      -0.00000000038,    -0.00000000185,
       0.00000000014,    -0.00000000009,
      -0.00000001485,     0.00000000374,
      -0.00000000080,    -0.00000000115,
       0.00000000002,    -0.00000000004,
      -0.00000001172,     0.00000000295,
      -0.00000000046,    -0.00000000055,
      -0.00000000221,    -0.00000001181,
       0.00000000057,    -0.00000000046,
       0.00000000109,     0.00000000152,
      -0.00000000192,     0.00000000097,
       0.00000002460,    -0.00000003270,
      -0.00000000664,    -0.00000000530,
      -0.00000000685,    -0.00000000163,
       0.00000000157,    -0.00000000695,
       0.00000001247,     0.00000000996,
      -0.00000000995,     0.00000001247,
       0.00000000000,     0.00000000002,
      -0.00000000046,    -0.00000000004,
       0.00000000017,    -0.00000000003,
      -0.00000000003,    -0.00000000018,
       0.00000000013,     0.00000000016,
       0.00000000008,     0.00000000002,
       0.00000000051,    -0.00000000069,
      -0.00000000004,    -0.00000000001,
       0.00000000005,    -0.00000000014,
      -0.00000000014,    -0.00000000005,
      -0.00000000021,    -0.00000000005,
      -0.00000000007,     0.00000000000,
       0.00000000000,     0.00000000008,
       0.00000000011,     0.00000000014,
       0.00000000334,    -0.00000000001,
      -0.00000000001,     0.00000000332,
       0.00000000011,     0.00000000005,
       0.00000000018,    -0.00000000045,
       0.00000000004,    -0.00000000003,
      -0.00000000014,    -0.00000000068,
       0.00000000003,    -0.00000000002,
       0.00000000069,    -0.00000000011,
       0.00000000002,     0.00000000003,
       0.00000000036,     0.00000000036,
      -0.00000000036,     0.00000000035,
       0.00000000630,     0.00000000170,
       0.00000000041,    -0.00000000128,
       0.00000000003,    -0.00000000144,
       0.00000000146,     0.00000000001,
      -0.00000000145,     0.00000000295,
      -0.00000000294,    -0.00000000145,
      -0.00000000059,    -0.00000000013,
       0.00000000004,     0.00000000002,
       0.00000000110,    -0.00000000257,
      -0.00000000014,    -0.00000000002,
       0.00000000002,    -0.00000000026,
      -0.00000000028,    -0.00000000003,
      -0.00000000020,    -0.00000000004,
       0.00000000000,     0.00000000010,
      -0.00000000001,     0.00000000002,
       0.00000000004,     0.00000000003,
       0.00000000007,     0.00000000001,
      -0.00000000001,     0.00000000007,
      -0.00000000004,     0.00000000006,
      -0.00000000006,    -0.00000000004,
      -0.00000000012,     0.00000000006,
       0.00000000004,     0.00000000005,
      -0.00000000001,    -0.00000000007,
       0.00000000102,     0.00000000013,
       0.00000000004,     0.00000000016,
       0.00000000142,    -0.00000000540,
       0.00000000087,    -0.00000000022,
      -0.00000009107,    -0.00000002070,
      -0.00000000201,    -0.00000000417,
       0.00000002117,    -0.00000009059,
       0.00000000428,    -0.00000000201,
       0.00000000346,     0.00000000708,
       0.00000000067,    -0.00000000006,
      -0.00000000701,     0.00000000343,
       0.00000000007,     0.00000000067,
      -0.00000000164,    -0.00000002903,
      -0.00000000277,    -0.00000000559,
      -0.00000000556,    -0.00000000260,
      -0.00000000102,     0.00000000052,
      -0.00000007298,    -0.00000001853,
      -0.00000000172,    -0.00000000333,
       0.00000001856,    -0.00000007317,
       0.00000000334,    -0.00000000172,
       0.00000002764,     0.00000005488,
       0.00000000528,    -0.00000000052,
      -0.00000005484,     0.00000002764,
       0.00000000052,     0.00000000528,
       0.00000000144,    -0.00000000020,
       0.00000000020,     0.00000000144,
       0.00000000000,    -0.00000000080,
       0.00000000046,    -0.00000000032,
       0.00000000097,    -0.00000000220,
      -0.00000000078,     0.00000000001,
       0.00000000002,     0.00000000078,
      -0.00000000140,     0.00000000003,
      -0.00000004447,    -0.00000000290,
       0.00000005977,    -0.00000006541,
       0.00000003758,    -0.00000000003,
       0.00000000002,     0.00000003754,
       0.00000000032,     0.00000000141,
      -0.00000000140,     0.00000000032,
      -0.00000000044,    -0.00000000015,
       0.00000000071,     0.00000000001,
       0.00000000001,    -0.00000000071,
      -0.00000000062,    -0.00000000001,
      -0.00000000001,     0.00000000062,
      -0.00000000072,    -0.00000000001,
      -0.00000000014,     0.00000000842,
       0.00000000097,     0.00000000002,
       0.00000000009,    -0.00000000554,
       0.00000000065,    -0.00000000019,
       0.00000000221,     0.00000000246,
       0.00000000009,    -0.00000000070,
       0.00000005904,     0.00000000747,
      -0.00000000132,    -0.00000000483,
       0.00000000984,     0.00000000051,
       0.00000000047,    -0.00000000976,
      -0.00000000015,     0.00000000110,
      -0.00000004174,    -0.00000000528,
       0.00000000095,     0.00000000346,
      -0.00000000891,    -0.00000000044,
      -0.00000000044,     0.00000000890,
      -0.00000000003,     0.00000000179,
       0.00000000053,     0.00000000001,
       0.00000000002,    -0.00000000154,
      -0.00000000083,    -0.00000000001,
      -0.00000000001,     0.00000000083,
       0.00000000074,     0.00000000001,
       0.00000000001,    -0.00000000074,
      -0.00000000001,     0.00000000108,
       0.00000000014,    -0.00000000057,
       0.00000000545,    -0.00000000621,
       0.00000000084,     0.00000000068,
      -0.00000001895,    -0.00000002015,
       0.00000000259,    -0.00000000291,
      -0.00000024210,    -0.00000003919,
      -0.00000000677,    -0.00000001167,
       0.00000000038,    -0.00000000039,
       0.00000003884,    -0.00000024174,
       0.00000001176,    -0.00000000679,
       0.00000000039,     0.00000000038,
       0.00000000312,    -0.00000001623,
      -0.00000000117,     0.00000000045,
       0.00000001607,     0.00000000320,
      -0.00000000045,    -0.00000000118,
      -0.00000000697,    -0.00000000531,
      -0.00000000052,     0.00000000069,
      -0.00000007439,    -0.00000001910,
      -0.00000000176,    -0.00000000341,
       0.00000001917,    -0.00000007445,
       0.00000000342,    -0.00000000176,
       0.00000000656,    -0.00000003547,
      -0.00000000259,     0.00000000107,
       0.00000003545,     0.00000000654,
      -0.00000000107,    -0.00000000259,
       0.00000000149,    -0.00000000021,
       0.00000000021,     0.00000000149,
      -0.00000000097,    -0.00000000061,
       0.00000000055,    -0.00000000062,
      -0.00000000115,     0.00000000004,
       0.00000000002,     0.00000000118,
       0.00000000309,     0.00000001140,
       0.00000000244,    -0.00000000084,
      -0.00000002381,    -0.00000009055,
      -0.00000000296,    -0.00000000341,
      -0.00000000016,    -0.00000000490,
       0.00000006357,    -0.00000006943,
       0.00000003975,     0.00000000000,
       0.00000000002,     0.00000003973,
       0.00000000009,    -0.00000000101,
       0.00000000100,     0.00000000009,
       0.00000000088,    -0.00000000210,
       0.00000000210,     0.00000000088,
       0.00000000062,    -0.00000000119,
       0.00000000120,     0.00000000062,
      -0.00000000062,    -0.00000000032,
      -0.00000000032,     0.00000000062,
       0.00000000071,     0.00000000032,
      -0.00000001087,     0.00000002385,
       0.00000000180,     0.00000000096,
       0.00000000151,    -0.00000000008,
      -0.00000000004,    -0.00000000155,
      -0.00000001664,    -0.00000000768,
      -0.00000000067,     0.00000000124,
       0.00000009258,    -0.00000020053,
      -0.00000001485,    -0.00000000808,
      -0.00000001422,     0.00000000223,
       0.00000000099,     0.00000001549,
       0.00000000188,    -0.00000000009,
      -0.00000000018,    -0.00000000373,
      -0.00000000219,     0.00000000154,
       0.00000000020,     0.00000000032,
       0.00000000036,     0.00000000019,
       0.00000000019,    -0.00000000036,
      -0.00000000277,    -0.00000000368,
      -0.00000000061,     0.00000000039,
       0.00000001667,    -0.00000001263,
      -0.00000000180,    -0.00000000279,
      -0.00000000427,    -0.00000000192,
      -0.00000000186,     0.00000000429,
       0.00000000035,     0.00000000015,
      -0.00000000017,     0.00000000033,
       0.00000000174,     0.00000000088,
       0.00000000173,    -0.00000000342,
       0.00000000155,    -0.00000000002,
      -0.00000000009,    -0.00000000152,
       0.00000000105,    -0.00000000003,
      -0.00000000005,    -0.00000000146,
      -0.00000000051,     0.00000000004,
      -0.00000000019,    -0.00000000142,
      -0.00000000036,     0.00000000003,
       0.00000000356,    -0.00000000055,
      -0.00000000010,    -0.00000000094,
      -0.00000000089,    -0.00000000134,
      -0.00000000132,     0.00000000091,
       0.00000000122,    -0.00000000163,
      -0.00000000462,    -0.00000000302,
      -0.00000000025,    -0.00000000040,
       0.00000000461,     0.00000001083,
       0.00000001064,    -0.00000000470,
       0.00000000612,     0.00000001658,
       0.00000002111,    -0.00000000779,
       0.00000000089,    -0.00000000054,
       0.00000000095,    -0.00000000075,
       0.00000000203,     0.00000000131,
       0.00000000090,    -0.00000000178,
       0.00000000077,     0.00000000232,
      -0.00000000232,     0.00000000072,
       0.00000000029,     0.00000000029,
      -0.00000000029,     0.00000000029,
       0.00000000116,     0.00000000053,
      -0.00000000004,    -0.00000000060,
      -0.00000000059,     0.00000000005,
       0.00000000009,     0.00000000112,
       0.00000000253,    -0.00000000019,
      -0.00000000003,    -0.00000000054,
       0.00000000066,     0.00000000102,
       0.00000000098,    -0.00000000068,
       0.00000000039,     0.00000000058,
       0.00000000075,    -0.00000000049,
       0.00000000047,     0.00000000026,
       0.00000000022,    -0.00000000047,
       0.00000000057,     0.00000000030,
       0.00000000032,    -0.00000000062,
      -0.00000000029,     0.00000000063,
       0.00000000139,     0.00000000064,
       0.00000000022,    -0.00000000046,
       0.00000000004,     0.00000000066,
       0.00000000065,    -0.00000000006,
       0.00000000020,     0.00000000032,
       0.00000000030,    -0.00000000022,
      -0.00000000062,     0.00000000020,
       0.00000000019,    -0.00000000037,
       0.00000000079,     0.00000000106,
       0.00000000042,    -0.00000000030,
      -0.00000000019,     0.00000000040,
       0.00000000040,     0.00000000017,
       0.00000000019,     0.00000000146,
       0.00000000072,    -0.00000000010,
      -0.00000000098,     0.00000000315,
       0.00000000170,     0.00000000071,
       0.00000000104,    -0.00000000086,
      -0.00000000049,    -0.00000000058,
      -0.00000000666,     0.00000000652,
       0.00000000369,     0.00000000466,
      -0.00000015636,    -0.00000001613,
      -0.00000000208,    -0.00000001680,
       0.00000000061,    -0.00000000036,
      -0.00000006056,     0.00000029511,
      -0.00000003184,    -0.00000000184,
      -0.00000000031,    -0.00000000096,
      -0.00000023794,    -0.00000049352,
       0.00000000744,     0.00000000563,
      -0.00000000059,    -0.00000000010,
       0.00000049481,    -0.00000023898,
      -0.00000000567,     0.00000000723,
       0.00000000005,    -0.00000000056,
       0.00000009789,     0.00000000453,
       0.00000000329,    -0.00000000649,
      -0.00000000583,     0.00000009958,
       0.00000000644,     0.00000000327,
       0.00000000029,    -0.00000000021,
      -0.00000000607,     0.00000000334,
       0.00000000084,     0.00000000127,
      -0.00000000608,    -0.00000000136,
       0.00000000137,    -0.00000000607,
       0.00000000311,     0.00000000028,
      -0.00000000028,     0.00000000311,
       0.00000000296,     0.00000000000,
       0.00000000000,     0.00000000296,
      -0.00000000370,     0.00000000247,
      -0.00000000038,    -0.00000000078,
       0.00000000197,     0.00000000341,
      -0.00000000071,     0.00000000023,
      -0.00000001510,    -0.00000000192,
      -0.00000000055,    -0.00000000137,
       0.00000000192,    -0.00000001514,
       0.00000000137,    -0.00000000056,
       0.00000000639,     0.00000000428,
       0.00000000059,    -0.00000000007,
      -0.00000000429,     0.00000000641,
       0.00000000007,     0.00000000059,
      -0.00000000071,     0.00000000035,
      -0.00000000035,    -0.00000000071,
       0.00000000049,     0.00000000010,
      -0.00000000010,     0.00000000049,
       0.00000000001,     0.00000000005,
       0.00000000005,    -0.00000000001,
      -0.00000000017,    -0.00000000047,
      -0.00000000017,     0.00000000049,
       0.00000000049,     0.00000000016,
       0.00000000043,     0.00000000014,
      -0.00000000054,     0.00000000018,
       0.00000000020,     0.00000000057,
      -0.00000000067,     0.00000000023,
       0.00000000025,     0.00000000072,
       0.00000002447,     0.00000000255,
      -0.00000002901,     0.00000003553,
      -0.00000000219,    -0.00000000124,
      -0.00000000046,     0.00000000121,
       0.00000001003,     0.00000000165,
      -0.00000000034,     0.00000000141,
      -0.00000008966,    -0.00000002321,
       0.00000000163,    -0.00000000614,
      -0.00000000037,    -0.00000000071,
      -0.00000000055,     0.00000000194,
      -0.00000000096,    -0.00000000122,
       0.00000000068,    -0.00000000052,
       0.00000003306,     0.00000004578,
      -0.00000000313,     0.00000000249,
      -0.00000000405,    -0.00000000122,
       0.00000000085,    -0.00000000417,
       0.00000000793,    -0.00000001729,
       0.00000000130,     0.00000000021,
       0.00000000161,     0.00000000209,
      -0.00000000206,     0.00000000172,
      -0.00000000039,    -0.00000000064,
       0.00000000111,     0.00000000094,
       0.00000000056,    -0.00000000105,
      -0.00000000002,    -0.00000000008,
      -0.00000000007,     0.00000000002,
       0.00000004481,    -0.00000001198,
      -0.00000001431,    -0.00000003428,
      -0.00000001304,     0.00000000758,
       0.00000000249,     0.00000000325,
       0.00000000488,     0.00000001695,
       0.00000001289,    -0.00000000570,
      -0.00000000291,    -0.00000000493,
      -0.00000000121,     0.00000000094,
      -0.00000020346,     0.00000005655,
       0.00000006631,     0.00000015458,
       0.00000005837,    -0.00000003480,
      -0.00000001146,    -0.00000001450,
      -0.00000001112,    -0.00000003841,
      -0.00000002916,     0.00000001293,
       0.00000000671,     0.00000001109,
       0.00000000269,    -0.00000000221,
      -0.00000000410,     0.00000000369,
       0.00000000266,     0.00000000207,
       0.00000000060,    -0.00000000094,
       0.00000000385,     0.00000000245,
       0.00000000055,    -0.00000000278,
      -0.00000000050,    -0.00000000068,
       0.00000000045,    -0.00000000087,
      -0.00000000087,    -0.00000000044,
       0.00000000056,    -0.00000000111,
      -0.00000000111,    -0.00000000055,
       0.00000000016,    -0.00000000041,
      -0.00000000045,    -0.00000000019,
      -0.00000000046,    -0.00000000015,
       0.00000000314,    -0.00000000913,
       0.00000001384,    -0.00000002498,
      -0.00000000117,    -0.00000000058,
       0.00000000994,     0.00000004985,
       0.00000002338,    -0.00000001038,
       0.00000000358,     0.00000001803,
       0.00000000865,    -0.00000000473,
      -0.00000000790,     0.00000000128,
       0.00000000183,     0.00000000388,
       0.00000000342,    -0.00000000343,
      -0.00000000254,    -0.00000000168,
       0.00000000294,     0.00000000332,
       0.00000000174,    -0.00000000257,
      -0.00000002586,    -0.00000011303,
      -0.00000005396,     0.00000003060,
       0.00000001016,     0.00000000960,
       0.00000001879,    -0.00000000331,
      -0.00000000457,    -0.00000000931,
      -0.00000000167,     0.00000000157,
      -0.00000001357,     0.00000001471,
       0.00000001129,     0.00000000691,
       0.00000000135,    -0.00000000376,
      -0.00000001295,    -0.00000001542,
      -0.00000000783,     0.00000001091,
       0.00000000381,     0.00000000154,
       0.00000010011,    -0.00000001434,
      -0.00000001237,    -0.00000003676,
      -0.00000000039,     0.00000000096,
      -0.00000000026,     0.00000000116,
       0.00000000118,     0.00000000033,
       0.00000000518,    -0.00000001550,
       0.00000000220,     0.00000000081,
       0.00000000267,     0.00000000042,
       0.00000000049,    -0.00000000279,
      -0.00000009461,     0.00000017278,
       0.00000003537,     0.00000002147,
      -0.00000001013,    -0.00000000450,
      -0.00000000033,     0.00000000155,
      -0.00000000556,    -0.00000000076,
      -0.00000000145,     0.00000000684,
       0.00000000062,     0.00000000050,
       0.00000007506,     0.00000000437,
      -0.00000000185,     0.00000001294,
       0.00000000087,     0.00000000093,
       0.00000000010,    -0.00000000148,
      -0.00000068774,    -0.00000012844,
      -0.00000006400,     0.00000003700,
       0.00000000992,    -0.00000005414,
      -0.00000000380,    -0.00000000430,
      -0.00000000216,     0.00000000819,
       0.00000000107,     0.00000000005,
      -0.00000000390,    -0.00000000369,
      -0.00000000043,     0.00000000084,
       0.00000000426,    -0.00000001700,
       0.00000000408,    -0.00000000226,
       0.00000000387,     0.00000000070,
      -0.00000000048,     0.00000000422,
       0.00000005123,     0.00000023716,
       0.00000001670,    -0.00000002011,
      -0.00000001886,     0.00000001325,
       0.00000000181,     0.00000000103,
      -0.00000003503,    -0.00000000608,
      -0.00000000278,     0.00000000247,
       0.00000000528,    -0.00000003620,
      -0.00000000230,    -0.00000000260,
      -0.00000001228,    -0.00000000109,
      -0.00000000096,    -0.00000000091,
      -0.00000000048,    -0.00000000131,
       0.00000000138,    -0.00000000058,
       0.00000012228,    -0.00000002640,
       0.00000000514,     0.00000000260,
      -0.00000000069,    -0.00000000074,
       0.00000000514,     0.00000001177,
       0.00000000073,    -0.00000000126,
      -0.00000001197,     0.00000000557,
       0.00000000128,     0.00000000063,
      -0.00000000037,     0.00000000065,
      -0.00000004449,    -0.00000002236,
       0.00000000070,    -0.00000000227,
       0.00000000387,    -0.00000000224,
      -0.00000000060,     0.00000000048,
       0.00000000237,     0.00000000382,
      -0.00000000049,    -0.00000000060,
      -0.00000000179,    -0.00000000083,
       0.00000000079,    -0.00000000182,
       0.00000002625,    -0.00000004727,
      -0.00000002417,    -0.00000001029,
      -0.00000000144,    -0.00000000073,
      -0.00000000029,     0.00000000072,
      -0.00000004367,    -0.00000000297,
       0.00000000023,    -0.00000000188,
      -0.00000000123,    -0.00000000012,
       0.00000000014,    -0.00000000123,
      -0.00000002745,     0.00000007656,
       0.00000005472,     0.00000001038,
       0.00000000054,    -0.00000001918,
      -0.00000004057,    -0.00000001079,
      -0.00000000319,     0.00000002773,
       0.00000000946,    -0.00000000017,
      -0.00000000038,    -0.00000000216,
       0.00000003441,     0.00000001844,
       0.00000000574,    -0.00000002139,
      -0.00000000687,    -0.00000000018,
       0.00000002500,    -0.00000001253,
      -0.00000000482,    -0.00000001333,
      -0.00000000417,     0.00000000066,
       0.00000000071,     0.00000000187,
       0.00000000180,    -0.00000000083,
       0.00000010801,    -0.00000036189,
      -0.00000025790,    -0.00000003658,
       0.00000000216,     0.00000008901,
       0.00000002010,    -0.00000000401,
       0.00000009764,     0.00000002605,
       0.00000000778,    -0.00000006634,
      -0.00000002250,     0.00000000060,
       0.00000000090,     0.00000000517,
      -0.00000017379,    -0.00000013140,
      -0.00000004565,     0.00000010884,
       0.00000003497,     0.00000000550,
      -0.00000000088,    -0.00000000760,
      -0.00000016056,     0.00000007871,
       0.00000003681,     0.00000007869,
       0.00000002254,    -0.00000000776,
      -0.00000000083,    -0.00000000468,
      -0.00000001347,    -0.00000001601,
      -0.00000000346,     0.00000000455,
      -0.00000001588,     0.00000001397,
       0.00000000480,     0.00000000333,
       0.00000000366,    -0.00000000571,
       0.00000000268,     0.00000000171,
      -0.00000001567,     0.00000002606,
      -0.00000000588,    -0.00000000373,
      -0.00000000368,     0.00000000041,
       0.00000000023,     0.00000000104,
       0.00000000042,     0.00000000370,
       0.00000000104,    -0.00000000026,
      -0.00000000023,    -0.00000000092,
      -0.00000000057,     0.00000000287,
       0.00000000287,     0.00000000058,
      -0.00000033558,     0.00000004713,
       0.00000003644,     0.00000013242,
       0.00000002628,    -0.00000001146,
      -0.00000000042,    -0.00000000248,
      -0.00000000564,     0.00000000796,
       0.00000000193,     0.00000000142,
       0.00000000800,     0.00000000610,
       0.00000000141,    -0.00000000224,
       0.00000000001,    -0.00000000006,
       0.00000000003,    -0.00000000002,
       0.00000000107,    -0.00000000016,
      -0.00000000016,    -0.00000000105,
      -0.00000000641,    -0.00000000207,
       0.00000000105,    -0.00000000300,
      -0.00000000126,    -0.00000000009,
       0.00000000322,     0.00000000301,
       0.00000000123,    -0.00000000141,
      -0.00000000393,    -0.00000000174,
       0.00000000048,     0.00000000218,
       0.00000006162,     0.00000000240,
      -0.00000000379,    -0.00000002503,
      -0.00000000812,     0.00000000248,
       0.00000000213,     0.00000000254,
      -0.00000004815,    -0.00000000499,
       0.00000000104,     0.00000001360,
       0.00000000239,    -0.00000000060,
       0.00000002433,     0.00000004175,
       0.00000000565,    -0.00000001520,
      -0.00000000311,     0.00000000034,
      -0.00000010077,     0.00000035897,
       0.00000007642,     0.00000001059,
       0.00000005280,     0.00000004276,
       0.00000001193,    -0.00000001741,
      -0.00000000300,    -0.00000000153,
      -0.00000007652,     0.00000000859,
       0.00000000982,     0.00000001378,
       0.00000000135,    -0.00000000249,
       0.00000000763,     0.00000007686,
       0.00000001385,    -0.00000000967,
      -0.00000000247,    -0.00000000136,
      -0.00000000096,    -0.00000000062,
      -0.00000007891,     0.00000001239,
       0.00000001268,     0.00000001553,
       0.00000000160,    -0.00000000301,
       0.00000001242,     0.00000007895,
       0.00000001553,    -0.00000001269,
      -0.00000000301,    -0.00000000161,
       0.00000000014,     0.00000000079,
      -0.00000340052,     0.00000053264,
       0.00000054783,     0.00000067010,
       0.00000006916,    -0.00000012977,
       0.00000001588,     0.00000010135,
       0.00000001995,    -0.00000001631,
      -0.00000000386,    -0.00000000206,
       0.00000013902,     0.00000000072,
       0.00000000723,    -0.00000000830,
      -0.00000000445,     0.00000002662,
       0.00000000185,     0.00000000177,
       0.00000000255,    -0.00000000108,
       0.00000000097,    -0.00000000219,
       0.00000219594,    -0.00000106601,
      -0.00000070316,    -0.00000072642,
      -0.00000008602,     0.00000009329,
       0.00000002385,     0.00000001037,
      -0.00000001848,    -0.00000027690,
      -0.00000004480,     0.00000001863,
       0.00000000424,     0.00000000425,
       0.00000000046,    -0.00000000107,
      -0.00000001759,     0.00000003895,
       0.00000000890,     0.00000000346,
       0.00000000046,    -0.00000000111,
      -0.00000003146,    -0.00000001624,
      -0.00000000367,     0.00000000815,
       0.00000000105,     0.00000000042,
      -0.00000000075,     0.00000000403,
      -0.00000000394,    -0.00000000089,
      -0.00000003525,     0.00000037676,
       0.00000001955,     0.00000001382,
       0.00000001134,     0.00000000109,
       0.00000000042,    -0.00000000060,
       0.00000002989,    -0.00000002335,
       0.00000000758,    -0.00000000337,
      -0.00000000076,    -0.00000000028,
       0.00000000737,     0.00000000024,
       0.00000000055,    -0.00000000058,
      -0.00000000221,     0.00000000937,
       0.00000000055,     0.00000000051,
      -0.00000041566,     0.00000058126,
       0.00000011475,     0.00000003743,
       0.00000000421,    -0.00000001798,
      -0.00000003199,     0.00000001983,
       0.00000000485,     0.00000000042,
       0.00000006844,    -0.00000004431,
      -0.00000002893,    -0.00000002589,
      -0.00000000341,     0.00000000392,
       0.00000000080,     0.00000000035,
       0.00000004361,     0.00000006596,
       0.00000002608,    -0.00000002860,
      -0.00000000392,    -0.00000000341,
      -0.00000000035,     0.00000000080,
      -0.00000000060,     0.00000000245,
       0.00000000245,     0.00000000060,
      -0.00000000086,     0.00000000355,
       0.00000000355,     0.00000000086,
      -0.00000000104,     0.00000000433,
       0.00000000434,     0.00000000104,
      -0.00000011513,     0.00000036965,
       0.00000001113,     0.00000000347,
      -0.00000003905,    -0.00000001660,
      -0.00000000222,    -0.00000000091,
      -0.00000000010,    -0.00000000244,
       0.00000000256,    -0.00000000023,
       0.00000025383,    -0.00000004744,
      -0.00000003278,     0.00000000887,
       0.00000000587,     0.00000000921,
      -0.00000000112,    -0.00000000158,
      -0.00000000782,     0.00000002778,
       0.00000000449,     0.00000000038,
      -0.00000002829,    -0.00000000712,
      -0.00000000032,     0.00000000434,
      -0.00000000040,     0.00000000105,
       0.00000000106,     0.00000000040,
      -0.00000000056,     0.00000000148,
       0.00000000148,     0.00000000056,
      -0.00000000071,     0.00000000189,
       0.00000000189,     0.00000000071,
      -0.00000007862,     0.00000022505,
       0.00000000666,     0.00000000233,
       0.00000010573,    -0.00000029757,
      -0.00000001284,    -0.00000000451,
      -0.00000000112,     0.00000000325,
      -0.00000000325,    -0.00000000112,
       0.00000000681,     0.00000000715,
      -0.00000000084,     0.00000000127,
      -0.00000000139,     0.00000000016,
      -0.00000000065,    -0.00000000084,
      -0.00000007994,    -0.00000005169,
       0.00000000179,    -0.00000000411,
      -0.00000000026,     0.00000000099,
       0.00000000887,    -0.00000000415,
      -0.00000000130,     0.00000000075,
       0.00000000442,     0.00000000885,
      -0.00000000078,    -0.00000000129,
      -0.00000000067,     0.00000000103,
       0.00000000104,     0.00000000067,
      -0.00000000095,     0.00000000147,
       0.00000000147,     0.00000000095,
      -0.00000000118,     0.00000000184,
       0.00000000184,     0.00000000119,
      -0.00000005215,     0.00000008267,
       0.00000000246,     0.00000000155,
       0.00000005537,    -0.00000016748,
      -0.00000000703,    -0.00000000232,
      -0.00000000049,     0.00000000079,
      -0.00000000079,    -0.00000000049,
      -0.00000000964,     0.00000004790,
       0.00000000339,    -0.00000001013,
       0.00000001011,     0.00000000339,
       0.00000000080,    -0.00000000023,
       0.00000001200,     0.00000002905,
      -0.00000000269,     0.00000000103,
      -0.00000000357,    -0.00000000168,
       0.00000000077,     0.00000000027,
       0.00000000162,    -0.00000000364,
      -0.00000000034,     0.00000000077,
      -0.00000002712,     0.00000004229,
       0.00000000180,     0.00000000115,
       0.00000000180,    -0.00000000536,
       0.00000000535,     0.00000000180,
      -0.00000000031,     0.00000000149,
      -0.00000000149,    -0.00000000031,
       0.00000000073,     0.00000000220,
      -0.00000000222,     0.00000000073,
       0.00000000770,     0.00000000410,
      -0.00000001110,    -0.00000000640,
      -0.00000002706,     0.00000002946,
      -0.00000003216,    -0.00000002245,
       0.00000002910,     0.00000001310,
      -0.00000000330,    -0.00000000900,
       0.00000000820,    -0.00000000280,
       0.00000000131,    -0.00000000041,
      -0.00000000553,     0.00000000085,
       0.00000000044,     0.00000000152,
       0.00000000084,     0.00000000553,
       0.00000000152,    -0.00000000044,
      -0.00000000088,    -0.00000000041,
       0.00000000301,    -0.00000001185,
       0.00000000190,     0.00000000090,
      -0.00000002025,     0.00000004667,
      -0.00000000148,     0.00000000015,
       0.00000001223,    -0.00000000028,
      -0.00000000020,     0.00000000205,
      -0.00000012575,    -0.00000001013,
       0.00000000080,    -0.00000000586,
      -0.00000000003,     0.00000000102,
      -0.00000000086,     0.00000000031,
      -0.00000003921,     0.00000001200,
      -0.00000000056,    -0.00000000124,
      -0.00000000374,    -0.00000000041,
       0.00000000045,    -0.00000000374,
      -0.00000000103,     0.00000000035,
      -0.00000000035,    -0.00000000103,
      -0.00000000080,     0.00000000841,
       0.00000000276,     0.00000000013,
      -0.00000000200,     0.00000000430,
       0.00000000123,     0.00000000054,
       0.00000000332,     0.00000000198,
       0.00000000051,    -0.00000000088,
      -0.00000000004,    -0.00000000002,
       0.00000000006,     0.00000000040,
       0.00000000012,    -0.00000000002,
       0.00000000822,     0.00000002035,
       0.00000000644,    -0.00000000288,
      -0.00000001857,     0.00000000294,
       0.00000000154,     0.00000000528,
       0.00000000322,     0.00000001855,
       0.00000000531,    -0.00000000157,
       0.00000001083,     0.00000003483,
       0.00000001128,    -0.00000000380,
      -0.00000004169,     0.00000000799,
       0.00000000375,     0.00000001188,
       0.00000000776,     0.00000004214,
       0.00000001209,    -0.00000000368,
       0.00000004849,    -0.00000000066,
      -0.00000000398,    -0.00000001519,
      -0.00000013923,     0.00000002279,
       0.00000001133,     0.00000004010,
       0.00000002222,     0.00000013594,
       0.00000003885,    -0.00000001103,
      -0.00000000061,     0.00000000047,
       0.00000000096,     0.00000000031,
       0.00000003610,    -0.00000002765,
      -0.00000002725,    -0.00000000210,
       0.00000000395,     0.00000000723,
       0.00000006035,    -0.00000012304,
      -0.00000006542,    -0.00000002672,
      -0.00000000597,     0.00000001837,
       0.00000000340,     0.00000000070,
      -0.00000060252,     0.00000014774,
       0.00000009256,     0.00000023149,
       0.00000004598,    -0.00000002726,
      -0.00000000511,    -0.00000000558,
       0.00000009692,     0.00000058840,
       0.00000022771,    -0.00000005797,
      -0.00000001548,    -0.00000004647,
      -0.00000000607,     0.00000000257,
      -0.00000003346,     0.00000000138,
       0.00000000156,     0.00000000432,
       0.00000000261,     0.00000003293,
       0.00000000406,    -0.00000000195,
      -0.00000195244,    -0.00000133275,
      -0.00000050835,     0.00000098424,
       0.00000025842,     0.00000011221,
       0.00000001633,    -0.00000004641,
       0.00000028405,     0.00000020985,
       0.00000006551,    -0.00000013698,
      -0.00000003247,    -0.00000000962,
      -0.00000000042,     0.00000000505,
       0.00000766512,    -0.00000177884,
      -0.00000106818,    -0.00000290321,
      -0.00000056498,     0.00000029758,
       0.00000005348,     0.00000007053,
       0.00000000617,    -0.00000000698,
      -0.00000156717,    -0.00000761124,
      -0.00000288911,     0.00000092365,
       0.00000024836,     0.00000056738,
       0.00000007267,    -0.00000004225,
      -0.00000000520,    -0.00000000675,
       0.00000009877,    -0.00000000689,
      -0.00000000461,    -0.00000001257,
      -0.00000001244,    -0.00000010045,
      -0.00000001273,     0.00000000672,
       0.00000002720,    -0.00000001606,
      -0.00000001491,    -0.00000054266,
      -0.00000022212,     0.00000003777,
       0.00000001484,     0.00000004580,
       0.00000000611,    -0.00000000315,
      -0.00000000046,    -0.00000000059,
       0.00000779617,    -0.00000119997,
      -0.00000088144,    -0.00000302300,
      -0.00000059425,     0.00000026340,
       0.00000004798,     0.00000007537,
       0.00000000678,    -0.00000000619,
      -0.00000120240,    -0.00000779997,
      -0.00000302485,     0.00000088249,
       0.00000026365,     0.00000059473,
       0.00000007545,    -0.00000004802,
      -0.00000000619,    -0.00000000680,
      -0.00000000169,    -0.00000000147,
      -0.00000000055,     0.00000000077,
       0.00000013508,    -0.00000005694,
      -0.00000002756,    -0.00000005736,
       0.00000011217,     0.00000007408,
       0.00000002469,    -0.00000004826,
      -0.00000001057,    -0.00000000398,
      -0.00000000035,     0.00000000152,
       0.00000976393,    -0.00000149996,
      -0.00000110281,    -0.00000378641,
      -0.00000074438,     0.00000032970,
       0.00000006009,     0.00000009445,
       0.00000000850,    -0.00000000777,
      -0.00000000077,    -0.00000000043,
      -0.00000150374,    -0.00000976991,
      -0.00000378906,     0.00000110466,
       0.00000033015,     0.00000074501,
       0.00000009456,    -0.00000006016,
      -0.00000000777,    -0.00000000852,
      -0.00000000043,     0.00000000076,
       0.00000000000,    -0.00000000003,
      -0.00000000029,    -0.00000000009,
      -0.00000000006,     0.00000000007,
      -0.00000000015,     0.00000000013,
       0.00000000003,     0.00000000004,
       0.00000000012,     0.00000000010,
       0.00000001729,    -0.00000000262,
      -0.00000000138,    -0.00000000495,
       0.00000000048,     0.00000000299,
       0.00000000094,    -0.00000000026,
      -0.00000000003,     0.00000000003,
       0.00000000014,     0.00000000014,
       0.00000000057,    -0.00000000009,
      -0.00000000005,    -0.00000000016,
      -0.00000000059,    -0.00000000366,
      -0.00000000107,     0.00000000031,
      -0.00000000101,     0.00000000091,
       0.00000000029,     0.00000000027,
       0.00000000070,     0.00000000051,
       0.00000019702,    -0.00000002962,
      -0.00000001558,    -0.00000005635,
       0.00000000544,     0.00000003546,
       0.00000001020,    -0.00000000284,
      -0.00000000122,     0.00000000156,
       0.00000000173,     0.00000000162,
      -0.00000000010,    -0.00000000166,
      -0.00000000163,     0.00000000012,
      -0.00000000042,     0.00000000046,
       0.00000000011,     0.00000000009,
       0.00000000109,     0.00000000095,
       0.00000000023,    -0.00000000027,
       0.00000000299,    -0.00000000060,
      -0.00000000028,    -0.00000000089,
      -0.00000000275,    -0.00000001443,
      -0.00000000419,     0.00000000131,
      -0.00000000331,     0.00000000288,
       0.00000000072,     0.00000000073,
       0.00000000231,     0.00000000176,
       0.00000000025,    -0.00000000047,
       0.00000021479,    -0.00000003220,
      -0.00000001696,    -0.00000006139,
       0.00000000567,     0.00000003691,
       0.00000001061,    -0.00000000296,
      -0.00000000133,     0.00000000170,
       0.00000000189,     0.00000000177,
      -0.00000000012,    -0.00000000191,
      -0.00000000187,     0.00000000014,
      -0.00000000163,     0.00000000168,
       0.00000000038,     0.00000000032,
       0.00000000296,     0.00000000288,
       0.00000000058,    -0.00000000069,
      -0.00000000024,     0.00000000027,
      -0.00000000027,    -0.00000000024,
       0.00000000066,    -0.00000000054,
       0.00000003277,    -0.00000000533,
      -0.00000000264,    -0.00000000937,
      -0.00000001660,    -0.00000010224,
      -0.00000002933,     0.00000000832,
      -0.00000001313,     0.00000001157,
       0.00000000293,     0.00000000272,
       0.00000000759,     0.00000000983,
       0.00000000190,    -0.00000000169,
      -0.00000000010,    -0.00000000065,
       0.00000001992,    -0.00000000306,
      -0.00000000159,    -0.00000000570,
       0.00000000044,     0.00000000280,
       0.00000000088,    -0.00000000024,
      -0.00000001005,     0.00000000862,
       0.00000000204,     0.00000000204,
       0.00000001022,     0.00000001217,
       0.00000000247,    -0.00000000245,
      -0.00000000149,     0.00000000071,
       0.00000000020,     0.00000000053,
       0.00000000030,    -0.00000000213,
      -0.00000000058,     0.00000000005,
       0.00000000014,     0.00000000153,
       0.00000000112,    -0.00000000021,
      -0.00000000139,     0.00000000088,
       0.00000000028,     0.00000000023,
      -0.00000000090,    -0.00000000081,
      -0.00000000024,     0.00000000027,
      -0.00000000128,     0.00000000014,
       0.00000000014,     0.00000000128,
      -0.00000002992,     0.00000000461,
      -0.00000000011,    -0.00000000188,
       0.00000000020,     0.00000000268,
       0.00000000101,    -0.00000000022,
      -0.00000007796,    -0.00000000104,
       0.00000000346,     0.00000002842,
      -0.00000000003,     0.00000000251,
       0.00000000090,    -0.00000000011,
      -0.00000000102,     0.00000000007,
      -0.00000000007,    -0.00000000102,
      -0.00000000142,    -0.00000000007,
       0.00000000007,    -0.00000000142,
      -0.00000000154,     0.00000000062,
       0.00000000076,    -0.00000000116,
       0.00000000116,     0.00000000075,
      -0.00000000803,    -0.00000001804,
       0.00000000389,    -0.00000000155,
      -0.00000000310,     0.00000000503,
      -0.00000000496,    -0.00000000321,
      -0.00000008469,     0.00000002137,
       0.00000001235,     0.00000003897,
      -0.00000000156,    -0.00000000467,
      -0.00000000217,     0.00000000081,
       0.00000005133,    -0.00000000120,
      -0.00000000349,    -0.00000002102,
      -0.00000000433,     0.00000000138,
      -0.00000000117,    -0.00000005136,
      -0.00000002103,     0.00000000349,
       0.00000000138,     0.00000000433,
      -0.00000001007,    -0.00000000902,
      -0.00000000343,     0.00000000483,
       0.00000000114,     0.00000000061,
      -0.00000000902,     0.00000001006,
       0.00000000483,     0.00000000343,
       0.00000000061,    -0.00000000114,
      -0.00000205144,     0.00000046911,
       0.00000029313,     0.00000075121,
       0.00000013927,    -0.00000007749,
       0.00000001446,     0.00000006349,
       0.00000002324,    -0.00000000905,
      -0.00000000239,    -0.00000000431,
      -0.00000000310,     0.00000000020,
       0.00000000025,     0.00000000121,
      -0.00000000030,    -0.00000000298,
      -0.00000000116,     0.00000000028,
       0.00000007245,     0.00000000089,
      -0.00000000456,    -0.00000002687,
       0.00000000004,    -0.00000000326,
      -0.00000000125,     0.00000000022,
      -0.00000003460,    -0.00000002120,
      -0.00000003755,     0.00000000820,
       0.00000000525,     0.00000001379,
       0.00000000257,    -0.00000000142,
      -0.00000000820,    -0.00000003755,
      -0.00000001379,     0.00000000525,
       0.00000000143,     0.00000000257,
      -0.00000000076,    -0.00000000038,
       0.00000000038,    -0.00000000075,
       0.00000000041,     0.00000000280,
       0.00000000108,    -0.00000000031,
       0.00000000016,    -0.00000000186,
      -0.00000000186,    -0.00000000015,
       0.00022765928,    -0.00003506189,
      -0.00002574652,    -0.00008828540,
      -0.00001734471,     0.00000768642,
       0.00000140135,     0.00000220090,
       0.00000019843,    -0.00000018136,
      -0.00000001787,    -0.00000000998,
      -0.00000149242,    -0.00000969008,
      -0.00000375789,     0.00000109596,
       0.00000032760,     0.00000073877,
       0.00000009375,    -0.00000005971,
      -0.00000000772,    -0.00000000844,
      -0.00000000044,     0.00000000078,
       0.00000014624,     0.00000000818,
      -0.00000000532,    -0.00000006248,
      -0.00000001339,     0.00000000310,
       0.00000000076,     0.00000000188,
       0.00000002220,     0.00000027295,
       0.00000010941,    -0.00000002475,
      -0.00000000851,    -0.00000002208,
      -0.00000000289,     0.00000000168,
      -0.00000001296,    -0.00000001130,
      -0.00000000452,     0.00000000641,
       0.00000000151,     0.00000000081,
       0.00000001196,    -0.00000001332,
      -0.00000000638,    -0.00000000456,
      -0.00000000081,     0.00000000151,
       0.00000038058,    -0.00000002847,
      -0.00000003066,    -0.00000014940,
      -0.00000002968,     0.00000001021,
      -0.00000000002,    -0.00000000262,
      -0.00000000105,     0.00000000016,
       0.00000003828,    -0.00000000375,
      -0.00000000373,    -0.00000001553,
      -0.00000000317,     0.00000000126,
       0.00000000360,     0.00000003824,
       0.00000001554,    -0.00000000371,
      -0.00000000126,    -0.00000000317,
       0.00000223441,    -0.00000021631,
      -0.00000021809,    -0.00000090861,
      -0.00000018572,     0.00000007399,
       0.00000001465,     0.00000002439,
      -0.00000000627,    -0.00000006467,
      -0.00000002628,     0.00000000631,
       0.00000000214,     0.00000000537,
       0.00000000070,    -0.00000000042,
       0.00000000121,     0.00000000004,
      -0.00000000034,    -0.00000000150,
       0.00000000452,    -0.00000000014,
      -0.00000000027,    -0.00000000179,
       0.00000000014,     0.00000000452,
       0.00000000179,    -0.00000000027,
      -0.00000001015,     0.00000000732,
      -0.00000000752,    -0.00000000702,
      -0.00000000191,     0.00000000221,
       0.00000001168,     0.00000000422,
      -0.00000000028,    -0.00000000366,
       0.00000000633,     0.00000000383,
       0.00000000229,    -0.00000000144,
      -0.00000000224,     0.00000002315,
       0.00000000371,     0.00000000039,
      -0.00000000837,    -0.00000001332,
      -0.00000000191,     0.00000000210,
      -0.00000001331,     0.00000000754,
       0.00000000204,     0.00000000190,
      -0.00001074115,     0.00000246166,
       0.00000165112,     0.00000398349,
       0.00000073796,    -0.00000046100,
      -0.00000007828,    -0.00000008627,
       0.00000009727,     0.00000019795,
       0.00000006669,    -0.00000004169,
      -0.00000000911,    -0.00000001142,
      -0.00000000127,     0.00000000130,
       0.00000327897,    -0.00000050975,
      -0.00000037355,    -0.00000127661,
      -0.00000025121,     0.00000011139,
       0.00000002029,     0.00000003185,
       0.00000000287,    -0.00000000262,
       0.00000050613,     0.00000330999,
       0.00000127838,    -0.00000037146,
      -0.00000011131,    -0.00000025111,
      -0.00000003185,     0.00000002028,
       0.00000000261,     0.00000000287,
       0.00000009740,    -0.00000002969,
      -0.00000001827,    -0.00000003982,
      -0.00000000083,    -0.00000000270,
      -0.00000000110,     0.00000000051,
      -0.00000001870,     0.00000000300,
       0.00000000118,    -0.00000000024,
       0.00000000025,     0.00000000122,
      -0.00000001371,    -0.00000000499,
      -0.00000000086,    -0.00000002076,
      -0.00000000421,     0.00000000291,
      -0.00000000728,     0.00000001302,
       0.00000000209,     0.00000000086,
       0.00000001193,    -0.00000000042,
      -0.00000000257,    -0.00000000115,
      -0.00000301324,     0.00000060321,
       0.00000056144,     0.00000063768,
       0.00000006186,    -0.00000011197,
       0.00000003015,     0.00000030495,
       0.00000005410,    -0.00000003652,
      -0.00000000881,    -0.00000000504,
       0.00000001014,    -0.00000004862,
      -0.00000000779,     0.00000000154,
       0.00000000054,     0.00000000105,
      -0.00000003007,     0.00000000412,
       0.00000000478,     0.00000000154,
      -0.00000000035,    -0.00000000077,
      -0.00000236953,    -0.00000028237,
      -0.00000021444,     0.00000004397,
      -0.00000001533,     0.00000002599,
       0.00000001225,    -0.00000027482,
      -0.00000002360,    -0.00000000905,
       0.00000000048,     0.00000000193,
      -0.00000020842,     0.00000003441,
       0.00000002798,     0.00000007764,
       0.00000001474,    -0.00000000853,
      -0.00000000154,    -0.00000000172,
      -0.00000002589,    -0.00000020547,
      -0.00000007847,     0.00000002778,
       0.00000000854,     0.00000001475,
       0.00000000173,    -0.00000000154,
       0.00000000020,     0.00000000104,
      -0.00000000064,     0.00000000074,
       0.00000000276,    -0.00000000024,
       0.00000000107,    -0.00000000137,
       0.00000012086,    -0.00000002516,
       0.00000001518,    -0.00000000150,
      -0.00000000152,    -0.00000000041,
       0.00000001309,     0.00000000166,
       0.00000000112,    -0.00000000125,
      -0.00000000355,     0.00000002336,
       0.00000000124,     0.00000000106,
      -0.00000026227,     0.00000063714,
       0.00000002378,    -0.00000008383,
      -0.00000006202,     0.00000000929,
       0.00000000807,     0.00000000160,
       0.00000000075,    -0.00000000049,
      -0.00000015575,    -0.00000001057,
      -0.00000000883,     0.00000000961,
       0.00000000076,     0.00000000038,
       0.00000001101,    -0.00000016067,
      -0.00000000942,    -0.00000000833,
      -0.00000000039,     0.00000000072,
      -0.00000000093,    -0.00000000116,
      -0.00000000095,     0.00000000103,
      -0.00000006738,    -0.00000003667,
      -0.00000000081,    -0.00000000843,
       0.00000000010,     0.00000000172,
       0.00000000289,    -0.00000000251,
       0.00000000797,    -0.00000000194,
       0.00000055082,     0.00000005003,
      -0.00000004937,     0.00000002330,
       0.00000000453,     0.00000003241,
      -0.00000000246,    -0.00000000375,
       0.00000000212,     0.00000004015,
       0.00000000146,    -0.00000000559,
      -0.00000004155,     0.00000000331,
       0.00000000573,     0.00000000122,
      -0.00000000112,    -0.00000000030,
      -0.00000000045,    -0.00000000137,
      -0.00000000124,    -0.00000000022,
       0.00000002305,     0.00000000535,
      -0.00000000110,     0.00000000381,
       0.00000000021,    -0.00000000175,
      -0.00000000424,     0.00000000273,
       0.00000000254,    -0.00000000418,
      -0.00000020309,    -0.00000015975,
       0.00000003083,     0.00000001302,
       0.00000001267,    -0.00000001831,
      -0.00000000086,     0.00000000312,
       0.00000002936,    -0.00000000210,
      -0.00000000305,     0.00000000198,
       0.00000000259,     0.00000002940,
      -0.00000000203,    -0.00000000303,
      -0.00000002726,     0.00000007244,
       0.00000000556,     0.00000000201,
      -0.00000000565,    -0.00000001433,
       0.00000000266,     0.00000000833,
       0.00000003116,    -0.00000001505,
      -0.00000002087,     0.00000000665,
       0.00000000738,     0.00000000762,
      -0.00000000720,     0.00000000730,
       0.00000009211,     0.00000023498,
      -0.00000002306,    -0.00000004827,
      -0.00000010464,     0.00000004344,
       0.00000001904,    -0.00000000995,
      -0.00000000194,     0.00000000272,
      -0.00000004360,    -0.00000002682,
       0.00000000732,     0.00000000248,
       0.00000002669,    -0.00000004404,
      -0.00000000335,     0.00000000711,
      -0.00000001198,     0.00000001859,
       0.00000000141,     0.00000000088,
      -0.00000000077,    -0.00000000031,
      -0.00000000079,    -0.00000000185,
       0.00000000189,    -0.00000000076,
       0.00000000683,     0.00000001232,
      -0.00000000274,    -0.00000000070,
      -0.00000006391,     0.00000002265,
       0.00000001339,     0.00000002626,
       0.00000000092,     0.00000000259,
       0.00000000101,    -0.00000000052,
      -0.00000002562,     0.00000000661,
      -0.00000000156,     0.00000000063,
       0.00000000038,     0.00000000068,
      -0.00000000063,    -0.00000000173,
      -0.00000000076,     0.00000000040,
      -0.00000000126,    -0.00000000026,
      -0.00000004285,     0.00000017120,
       0.00000001343,     0.00000000333,
      -0.00000011028,     0.00000007301,
       0.00000003569,     0.00000004878,
      -0.00000000319,    -0.00000000352,
      -0.00000000156,     0.00000000145,
      -0.00000000734,     0.00000001152,
       0.00000000495,     0.00000000241,
       0.00000000036,    -0.00000000108,
       0.00000001151,     0.00000000734,
       0.00000000241,    -0.00000000495,
      -0.00000000108,    -0.00000000036,
       0.00000002543,    -0.00000000266,
      -0.00000000321,    -0.00000000892,
      -0.00000000160,     0.00000000099,
      -0.00000000265,    -0.00000002543,
      -0.00000000892,     0.00000000321,
       0.00000000099,     0.00000000160,
       0.00042044816,    -0.00006465355,
      -0.00004751361,    -0.00016305561,
      -0.00003203640,     0.00001418380,
       0.00000258471,     0.00000406633,
       0.00000036672,    -0.00000033398,
      -0.00000003294,    -0.00000001846,
      -0.00000193044,    -0.00001254911,
      -0.00000486662,     0.00000141838,
       0.00000042340,     0.00000095617,
       0.00000012144,    -0.00000007721,
      -0.00000000997,    -0.00000001094,
      -0.00000000054,     0.00000000097,
      -0.00000001239,     0.00000004862,
       0.00000002011,     0.00000000294,
       0.00000000016,    -0.00000000425,
      -0.00000003074,    -0.00000003857,
      -0.00000001386,     0.00000001402,
       0.00000000321,     0.00000000250,
       0.00000002933,    -0.00000000306,
      -0.00000000370,    -0.00000001029,
      -0.00000000184,     0.00000000114,
       0.00000000306,     0.00000002931,
       0.00000001028,    -0.00000000370,
      -0.00000000114,    -0.00000000184,
       0.00000003645,     0.00000000444,
      -0.00000000204,     0.00000000741,
       0.00000000122,    -0.00000000167,
       0.00000000173,     0.00000000032,
      -0.00000037621,    -0.00000000057,
       0.00000000279,    -0.00000001874,
      -0.00000000134,    -0.00000000119,
       0.00000004785,     0.00000002890,
       0.00000000937,    -0.00000001948,
      -0.00000000411,    -0.00000000158,
      -0.00000002844,     0.00000004873,
       0.00000001948,     0.00000000937,
       0.00000000158,    -0.00000000411,
       0.00000001560,    -0.00000000205,
       0.00000000054,     0.00000000235,
       0.00000000127,     0.00000000012,
      -0.00000000015,     0.00000000127,
       0.00000000233,    -0.00000000004,
      -0.00000000018,    -0.00000000093,
       0.00000000003,     0.00000000231,
       0.00000000093,    -0.00000000018,
      -0.00000012040,     0.00000003192,
      -0.00000000148,    -0.00000000409,
      -0.00000001150,    -0.00000000137,
      -0.00000000072,     0.00000000081,
       0.00000000148,    -0.00000001149,
      -0.00000000082,    -0.00000000070,
      -0.00000000343,     0.00000000092,
      -0.00000000092,    -0.00000000344,
       0.00000000039,    -0.00000000038,
      -0.00000000006,    -0.00000000005,
       0.00000000102,    -0.00000000127,
      -0.00000000016,    -0.00000000017,
      -0.00000000127,    -0.00000000102,
      -0.00000000015,     0.00000000015,
       0.00000000008,    -0.00000000002,
      -0.00000000001,    -0.00000000009,
      -0.00000000018,    -0.00000000033,
      -0.00000000030,    -0.00000000057,
       0.00000000031,    -0.00000000058,
      -0.00000000052,    -0.00000000037,
      -0.00000000156,    -0.00000000118,
       0.00000000407,     0.00000000536,
       0.00000000105,    -0.00000000055,
       0.00000000535,    -0.00000000399,
      -0.00000000053,    -0.00000000108,
       0.00000000044,    -0.00000000028,
      -0.00000001848,     0.00000131659,
       0.00000035079,    -0.00000022729,
      -0.00000006049,    -0.00000003613,
      -0.00000000090,     0.00000000867,
       0.00000053523,     0.00000214361,
       0.00000086999,    -0.00000031173,
      -0.00000008737,    -0.00000017524,
      -0.00000002230,     0.00000001559,
       0.00000000206,     0.00000000198,
       0.00000265660,    -0.00000647364,
      -0.00000180447,    -0.00000049883,
      -0.00000003216,     0.00000025844,
       0.00000002390,    -0.00000000174,
      -0.00000646403,    -0.00000276842,
      -0.00000055204,     0.00000180639,
       0.00000025967,     0.00000004546,
       0.00000000040,    -0.00000002449,
       0.00000011831,    -0.00000020775,
       0.00000000373,     0.00000000352,
       0.00000000020,    -0.00000000166,
      -0.00000022071,    -0.00000009247,
       0.00000001016,    -0.00000000120,
      -0.00000000142,    -0.00000000115,
       0.00037209830,    -0.00006993652,
      -0.00004538284,    -0.00014238841,
      -0.00002789200,     0.00001276757,
       0.00000224930,     0.00000354808,
       0.00000032574,    -0.00000028398,
      -0.00000002534,    -0.00000002001,
      -0.00000440350,    -0.00002558829,
      -0.00001002883,     0.00000305256,
       0.00000089147,     0.00000198404,
       0.00000025340,    -0.00000015967,
      -0.00000002015,    -0.00000002319,
      -0.00000000138,     0.00000000166,
      -0.00000320162,     0.00002580001,
       0.00000656715,     0.00000043405,
      -0.00000006463,    -0.00000094408,
      -0.00000009035,     0.00000002847,
       0.00000000445,     0.00000000667,
       0.00002401002,     0.00000397341,
       0.00000101700,    -0.00000573383,
      -0.00000075285,    -0.00000013030,
      -0.00000001211,     0.00000006259,
       0.00000000397,     0.00000000119,
      -0.00000021717,     0.00000057286,
       0.00000000330,    -0.00000003117,
      -0.00000000369,     0.00000000299,
       0.00000047388,     0.00000030878,
      -0.00000000001,     0.00000001796,
       0.00000000529,    -0.00000000166,
       0.00000109957,    -0.00000017685,
      -0.00000013153,    -0.00000042624,
      -0.00000008290,     0.00000003879,
      -0.00000000243,     0.00000000218,
      -0.00000006143,    -0.00000000352,
       0.00000001110,     0.00000001187,
       0.00000000081,    -0.00000000234,
      -0.00000001570,     0.00000006126,
       0.00000001344,    -0.00000000512,
      -0.00000000086,    -0.00000000164,
       0.00000000048,     0.00000000101,
       0.00000000097,    -0.00000000051,
       0.00000240625,    -0.00000032880,
      -0.00000026017,    -0.00000094128,
      -0.00000018564,     0.00000007895,
       0.00000001449,     0.00000002394,
      -0.00000000087,    -0.00000000191,
      -0.00000007656,    -0.00000001304,
       0.00000000995,     0.00000001603,
       0.00000000159,    -0.00000000199,
      -0.00000001310,     0.00000007657,
       0.00000001607,    -0.00000000992,
      -0.00000000198,    -0.00000000161,
       0.00000000007,     0.00000000002,
       0.00000000004,     0.00000000002,
       0.00000000002,    -0.00000000004,
       0.00000000001,    -0.00000000005,
       0.00000000035,    -0.00000000025,
      -0.00000000025,    -0.00000000034,
      -0.00000001103,    -0.00000001389,
       0.00000000158,    -0.00000000061,
       0.00000000891,     0.00000000137,
       0.00000000050,    -0.00000000055,
       0.00000000128,    -0.00000000892,
      -0.00000000055,    -0.00000000050,
       0.00000000018,    -0.00000000034,
       0.00000000086,     0.00000000055,
       0.00000000055,    -0.00000000085,
      -0.00000001188,    -0.00000001541,
       0.00000000165,    -0.00000000061,
       0.00000000959,     0.00000000148,
       0.00000000048,    -0.00000000053,
       0.00000000138,    -0.00000000960,
      -0.00000000054,    -0.00000000048,
       0.00000000058,     0.00000000039,
       0.00000000069,    -0.00000000102,
      -0.00000000071,    -0.00000000093,
      -0.00000000210,     0.00000000166,
       0.00000000023,     0.00000000034,
       0.00000000355,     0.00000000182,
       0.00000000025,    -0.00000000025,
       0.00000000186,    -0.00000000370,
      -0.00000000025,    -0.00000000024,
       0.00000000361,     0.00000000155,
       0.00000000022,    -0.00000000022,
       0.00000000129,    -0.00000000396,
      -0.00000000026,    -0.00000000026,
      -0.00000000045,    -0.00000000092,
      -0.00000000099,     0.00000000043,
      -0.00000004978,    -0.00000016074,
      -0.00000007112,     0.00000005127,
       0.00000001957,     0.00000001420,
      -0.00000000572,     0.00000000186,
       0.00000000185,     0.00000000254,
       0.00000000049,    -0.00000000068,
       0.00000000020,    -0.00000000080,
       0.00000000043,     0.00000000091,
      -0.00000000099,     0.00000000040,
       0.00000000899,     0.00000000136,
      -0.00000000062,     0.00000000294,
      -0.00000000196,    -0.00000000391,
       0.00000000421,    -0.00000000165,
      -0.00000006472,     0.00000023050,
       0.00000011546,    -0.00000000468,
      -0.00000000968,    -0.00000002784,
       0.00000000512,     0.00000000078,
      -0.00000000079,    -0.00000000281,
      -0.00000000068,     0.00000000038,
       0.00000000466,    -0.00000000479,
      -0.00000000239,    -0.00000000160,
       0.00000000239,    -0.00000000733,
      -0.00000000412,    -0.00000000075,
      -0.00000000009,     0.00000000107,
      -0.00000000296,    -0.00000000400,
      -0.00000000397,     0.00000000300,
       0.00000000761,    -0.00000002558,
      -0.00000000167,    -0.00000000019,
       0.00000000841,     0.00000000830,
       0.00000000089,    -0.00000000119,
       0.00000000842,    -0.00000000748,
      -0.00000000095,    -0.00000000123,
      -0.00000198114,    -0.00000036652,
       0.00000024792,     0.00000041863,
       0.00000004275,    -0.00000005148,
      -0.00000000193,     0.00000006866,
       0.00000001419,    -0.00000001118,
      -0.00000000219,    -0.00000000125,
      -0.00000001199,    -0.00000000473,
      -0.00000000189,    -0.00000000054,
      -0.00000000066,     0.00000000038,
      -0.00000000143,     0.00000001964,
       0.00000000275,     0.00000000059,
      -0.00000000076,    -0.00000000073,
      -0.00000000043,    -0.00000000061,
       0.00000000070,     0.00000000036,
      -0.00000000733,     0.00000001722,
       0.00000002487,     0.00000000335,
       0.00000000140,     0.00000001042,
      -0.00000004974,     0.00000000391,
       0.00000000419,     0.00000004930,
       0.00000009333,    -0.00000025975,
      -0.00000001654,     0.00000001327,
      -0.00000003192,     0.00000012278,
       0.00000001022,     0.00000000811,
       0.00000011278,    -0.00000001545,
      -0.00000000108,     0.00000000016,
      -0.00000001153,    -0.00000010939,
      -0.00000525572,    -0.00000081844,
      -0.00000035355,     0.00000037093,
       0.00000002429,     0.00000001664,
       0.00000011126,    -0.00000064505,
      -0.00000004740,    -0.00000004059,
      -0.00000000154,     0.00000000282,
      -0.00000004439,    -0.00000000534,
       0.00000000267,     0.00000000639,
       0.00000000074,    -0.00000000057,
       0.00000001790,    -0.00000002818,
      -0.00000000715,     0.00000000233,
       0.00000000055,     0.00000000073,
       0.00000000034,     0.00000000227,
      -0.00000000129,     0.00000000160,
      -0.00000002541,     0.00000000710,
       0.00000000019,     0.00000000094,
      -0.00000000071,     0.00000000222,
       0.00000008456,     0.00000004889,
      -0.00000000015,    -0.00000001230,
      -0.00000004148,     0.00000014037,
       0.00000001877,    -0.00000000555,
      -0.00000000060,    -0.00000000192,
      -0.00000021175,    -0.00000006038,
       0.00000002400,     0.00000002724,
       0.00000000218,    -0.00000000326,
      -0.00000010203,    -0.00000001910,
      -0.00000001890,     0.00000001409,
       0.00000000410,     0.00000000219,
       0.00000000108,    -0.00000000052,
      -0.00000050935,    -0.00000128106,
      -0.00000008543,     0.00000023540,
       0.00000002398,    -0.00000000113,
       0.00000014464,    -0.00000031890,
      -0.00000004938,     0.00000001174,
       0.00000000297,     0.00000000323,
       0.00000011726,     0.00000020891,
       0.00000001136,     0.00000000854,
       0.00000000212,    -0.00000000445,
       0.00000015236,    -0.00000023286,
      -0.00000002879,     0.00000003320,
       0.00000000326,     0.00000000165,
      -0.00000000271,     0.00000000084,
       0.00000000090,     0.00000000140,
      -0.00000161295,     0.00000159623,
       0.00000008569,    -0.00000018050,
      -0.00000001161,    -0.00000001859,
      -0.00000021791,    -0.00000010646,
       0.00000002211,     0.00000001035,
       0.00000000268,    -0.00000000140,
      -0.00000044336,    -0.00000007161,
      -0.00000002938,     0.00000003244,
       0.00000000204,     0.00000000113,
       0.00000007545,    -0.00000045276,
      -0.00000003235,    -0.00000002828,
      -0.00000000123,     0.00000000195,
       0.00000001337,    -0.00000000147,
       0.00000001098,     0.00000007258,
      -0.00000001018,    -0.00000001689,
       0.00000001150,     0.00000001792,
       0.00000001660,    -0.00000000851,
      -0.00000039868,    -0.00000004623,
       0.00000003574,    -0.00000000977,
      -0.00000000669,    -0.00000036822,
       0.00000001456,     0.00000003528,
      -0.00000000272,    -0.00000000299,
       0.00000000629,    -0.00000005447,
      -0.00000000105,     0.00000000412,
       0.00000001412,     0.00000002724,
      -0.00000000343,    -0.00000000157,
       0.00000330952,     0.00000036189,
      -0.00000030046,     0.00000010045,
      -0.00000006295,     0.00000089196,
      -0.00000003196,    -0.00000008444,
       0.00000000061,     0.00000000278,
      -0.00000014651,     0.00000029238,
       0.00000001664,    -0.00000003231,
      -0.00000000179,    -0.00000000329,
      -0.00000030132,    -0.00000014216,
       0.00000003349,     0.00000001574,
       0.00000000318,    -0.00000000207,
      -0.00000000428,     0.00000000322,
      -0.00000000381,    -0.00000000471,
       0.00000002691,    -0.00000000936,
      -0.00000000033,    -0.00000000093,
       0.00000000895,    -0.00000001084,
       0.00000000026,     0.00000000271,
      -0.00000000188,     0.00000001558,
      -0.00000002443,     0.00000002920,
       0.00000000158,    -0.00000000348,
       0.00000000721,    -0.00000000027,
       0.00000000033,     0.00000000675,
      -0.00000000200,     0.00000000097,
       0.00000017926,     0.00000010357,
      -0.00000002874,    -0.00000000861,
       0.00000000933,    -0.00000001660,
      -0.00000000071,     0.00000000218,
      -0.00000005584,    -0.00000000207,
       0.00000000540,    -0.00000000205,
       0.00000000114,    -0.00000005616,
       0.00000000211,     0.00000000552,
       0.00000000441,     0.00000000112,
      -0.00000000169,     0.00000000591,
      -0.00000000579,    -0.00000000121,
      -0.00000002784,     0.00000002008,
      -0.00000001321,    -0.00000000198,
      -0.00000000088,    -0.00000000124,
      -0.00000000131,     0.00000000586,
      -0.00000000583,    -0.00000000178,
       0.00000000107,    -0.00000000034,
       0.00000000034,     0.00000000106,
      -0.00000000262,     0.00000000556,
       0.00000000134,    -0.00000000005,
       0.00000000022,     0.00000001392,
      -0.00000000602,     0.00000000005,
       0.00000000070,    -0.00000000045,
       0.00000000375,     0.00000000238,
      -0.00000000244,     0.00000000373,
       0.00000000197,     0.00000000055,
      -0.00000000094,     0.00000000227,
      -0.00000000234,    -0.00000000068,
      -0.00000001307,     0.00000000684,
      -0.00000000627,    -0.00000000077,
      -0.00000000093,     0.00000000246,
      -0.00000000250,    -0.00000000110,
       0.00000000080,    -0.00000000091,
       0.00000001731,    -0.00000000611,
      -0.00000000330,    -0.00000000934,
       0.00000000107,     0.00000000083,
       0.00000000090,     0.00000000102,
       0.00000000102,    -0.00000000090,
       0.00000005010,     0.00000005669,
      -0.00000000170,     0.00000000150,
      -0.00000000113,     0.00000000228,
       0.00000000228,     0.00000000113,
      -0.00000003513,     0.00000007557,
      -0.00000000111,    -0.00000000042,
       0.00000000757,    -0.00000002264,
       0.00000000105,     0.00000000061,
      -0.00000002261,    -0.00000000756,
       0.00000000061,    -0.00000000105,
       0.00000026120,    -0.00000078183,
       0.00000003548,     0.00000002033,
       0.00000003317,     0.00000001108,
      -0.00000000086,     0.00000000151,
      -0.00000000100,    -0.00000000023,
       0.00000001113,    -0.00000002348,
       0.00000000818,    -0.00000000121,
      -0.00000000092,    -0.00000000317,
      -0.00000000084,     0.00000000038,
      -0.00000000121,    -0.00000000818,
      -0.00000000317,     0.00000000092,
       0.00000000038,     0.00000000084,
      -0.00000000052,    -0.00000000072,
      -0.00000000077,     0.00000000050,
       0.00000000202,     0.00000000175,
       0.00000000178,    -0.00000000186,
      -0.00000342243,    -0.00000058289,
       0.00000044986,     0.00000071378,
       0.00000007003,    -0.00000009007,
      -0.00000001589,     0.00000009628,
       0.00000002017,    -0.00000001255,
      -0.00000000251,    -0.00000000200,
       0.00000000438,    -0.00000000200,
      -0.00000000114,    -0.00000000274,
      -0.00000000091,     0.00000000045,
       0.00000000052,     0.00000001122,
       0.00000000321,    -0.00000000066,
      -0.00000000045,    -0.00000000093,
       0.00000005189,    -0.00000000326,
       0.00000000124,     0.00000000434,
       0.00000000501,     0.00000000199,
      -0.00000000157,     0.00000000523,
      -0.00000042426,     0.00000007733,
      -0.00000000427,    -0.00000001489,
      -0.00000003939,    -0.00000000647,
      -0.00000000255,     0.00000000253,
       0.00000000809,    -0.00000003884,
      -0.00000000268,    -0.00000000234,
       0.00000000123,     0.00000000066,
       0.00000000195,    -0.00000000027,
       0.00000000026,     0.00000000193,
      -0.00000003169,     0.00000008282,
      -0.00000001205,     0.00000000267,
      -0.00000000265,    -0.00000001208,
      -0.00000000076,     0.00000000216,
      -0.00000000216,    -0.00000000076,
      -0.00000000367,     0.00000000287,
       0.00000000032,     0.00000000046,
      -0.00000000183,     0.00000000347,
       0.00000000005,     0.00000000012,
       0.00000000332,     0.00000000136,
       0.00000000008,    -0.00000000006,
       0.00000000001,     0.00000000018,
       0.00000000011,     0.00000000020,
      -0.00000000693,     0.00000000693,
       0.00000000079,     0.00000000115,
      -0.00000000209,    -0.00000000455,
      -0.00000000043,     0.00000000030,
      -0.00000000402,     0.00000000259,
       0.00000000031,     0.00000000043,
       0.00000000008,     0.00000000047,
       0.00000000051,    -0.00000000003,
      -0.00000001234,     0.00000001169,
       0.00000000125,     0.00000000212,
      -0.00000000522,    -0.00000000785,
      -0.00000000086,     0.00000000067,
      -0.00000000783,     0.00000000640,
       0.00000000074,     0.00000000085,
       0.00000000038,    -0.00000000025,
      -0.00000000638,     0.00000002185,
       0.00000000223,     0.00000000065,
      -0.00000002866,    -0.00000003484,
      -0.00000000368,     0.00000000332,
      -0.00000003247,     0.00000003237,
       0.00000000330,     0.00000000330,
       0.00000000000,     0.00000000084,
       0.00000000099,     0.00000000014,
      -0.00002665750,    -0.00004802018,
      -0.00000814970,     0.00000536732,
       0.00000061750,     0.00000092608,
       0.00000006995,    -0.00000005993,
      -0.00000000543,    -0.00000000421,
       0.00003776099,    -0.00003320646,
      -0.00000719739,    -0.00000491632,
      -0.00000043206,     0.00000088002,
       0.00000008303,     0.00000002268,
       0.00000000072,    -0.00000000714,
       0.00008128460,     0.00010802139,
       0.00001252649,    -0.00001530856,
      -0.00000174045,    -0.00000045172,
       0.00000003660,     0.00000015731,
       0.00000001141,    -0.00000000732,
       0.00011128760,    -0.00006090816,
      -0.00000751593,    -0.00001487050,
      -0.00000112780,     0.00000021173,
      -0.00000003736,     0.00000008205,
       0.00000000721,     0.00000000654,
       0.00000112943,     0.00000189529,
      -0.00000026655,     0.00000021537,
      -0.00000000759,    -0.00000000744,
       0.00000000116,    -0.00000000164,
       0.00000169772,    -0.00000137853,
       0.00000018623,     0.00000028063,
      -0.00000000655,     0.00000000986,
      -0.00000000158,    -0.00000000118,
       0.00011679772,     0.00080983754,
       0.00017716666,    -0.00002967631,
      -0.00000530057,    -0.00002306892,
      -0.00000196617,     0.00000080045,
       0.00000008975,     0.00000012777,
      -0.00009299940,     0.00008319766,
       0.00001787085,     0.00001186112,
       0.00000102095,    -0.00000215559,
      -0.00000020135,    -0.00000005244,
      -0.00000000137,     0.00000001709,
      -0.00026024173,    -0.00017377845,
      -0.00003074529,     0.00002336010,
       0.00000121809,     0.00000200582,
       0.00000011416,    -0.00000001603,
       0.00000000464,    -0.00000000898,
      -0.00016900676,     0.00028321911,
       0.00003226683,     0.00002790797,
       0.00000123945,    -0.00000298199,
      -0.00000024435,     0.00000001770,
       0.00000000877,     0.00000001672,
      -0.00000267543,    -0.00000450142,
       0.00000066886,    -0.00000057403,
       0.00000000944,     0.00000001495,
      -0.00000000356,     0.00000000485,
      -0.00000433975,     0.00000319584,
      -0.00000050561,    -0.00000068285,
       0.00000001661,    -0.00000001390,
       0.00000000484,     0.00000000313,
      -0.00000040242,     0.00000027739,
       0.00000012083,     0.00000006234,
       0.00000000271,    -0.00000001953,
       0.00000003183,    -0.00000001737,
      -0.00000000215,    -0.00000000579,
      -0.00000000091,    -0.00000000011,
       0.00000034806,    -0.00000007256,
       0.00000000946,    -0.00000001865,
      -0.00000000103,    -0.00000000020,
      -0.00000007199,    -0.00000033905,
      -0.00000001528,    -0.00000000997,
      -0.00000092436,     0.00000066146,
       0.00000027826,     0.00000014545,
       0.00000000545,    -0.00000004394,
       0.00000000958,     0.00000000108,
       0.00000040629,    -0.00000009432,
       0.00000000981,    -0.00000002140,
      -0.00000000127,    -0.00000000015,
      -0.00000009163,    -0.00000039456,
      -0.00000001707,    -0.00000001128,
      -0.00000000006,    -0.00000000001,
       0.00000001458,    -0.00000000506,
      -0.00000000037,     0.00000000179,
      -0.00000000370,    -0.00000000611,
       0.00000000239,    -0.00000000158,
       0.00000016366,    -0.00000005653,
      -0.00000000425,     0.00000001978,
       0.00000000121,     0.00000000114,
       0.00000000108,    -0.00000000016,
      -0.00000000029,    -0.00000000093,
       0.00000000041,    -0.00000000025,
      -0.00000000107,     0.00000000000,
       0.00000000001,     0.00000000105,
      -0.00000000398,    -0.00000000661,
       0.00000000248,    -0.00000000165,
       0.00000017450,    -0.00000006052,
      -0.00000000444,     0.00000002082,
       0.00000000127,     0.00000000120,
       0.00000000107,    -0.00000000014,
      -0.00000000029,    -0.00000000091,
      -0.00000000077,     0.00000000000,
       0.00000000000,     0.00000000135,
       0.00000000134,     0.00000000157,
       0.00000000469,    -0.00000000416,
      -0.00000000049,    -0.00000000046,
      -0.00000000312,    -0.00000000021,
       0.00000000028,     0.00000000291,
       0.00000001502,    -0.00000000499,
      -0.00000000057,     0.00000000159,
      -0.00000000331,     0.00000000007,
       0.00000000005,     0.00000000382,
      -0.00000000189,    -0.00000000030,
      -0.00000000035,     0.00000000186,
       0.00000000667,    -0.00000000003,
      -0.00000000087,    -0.00000000209,
       0.00000000012,    -0.00000000683,
      -0.00000000216,     0.00000000077,
      -0.00000000877,     0.00000001563,
       0.00000000095,    -0.00000000029,
      -0.00000000088,    -0.00000000087,
      -0.00000000211,     0.00000000032,
       0.00000000046,     0.00000000130,
      -0.00000000059,    -0.00000000069,
      -0.00000000045,     0.00000000073,
       0.00000003735,    -0.00000002393,
      -0.00000001383,    -0.00000001240,
       0.00000000039,    -0.00000000076,
      -0.00000000219,     0.00000000008,
       0.00000000027,     0.00000000090,
      -0.00000000016,     0.00000000198,
       0.00000000088,    -0.00000000015,
      -0.00000000369,     0.00000000401,
       0.00000008553,     0.00000007870,
       0.00000003795,    -0.00000004219,
       0.00000000889,     0.00000001789,
      -0.00000020257,    -0.00000018862,
       0.00000008344,    -0.00000004147,
      -0.00000003171,    -0.00000000630,
       0.00000000819,    -0.00000003081,
      -0.00000000735,     0.00000000326,
       0.00000000019,     0.00000000221,
       0.00000000134,    -0.00000000045,
      -0.00000000148,     0.00000000065,
       0.00000000051,     0.00000000118,
      -0.00000004228,    -0.00000000534,
       0.00000000100,     0.00000001895,
       0.00000000014,     0.00000000691,
       0.00000000258,    -0.00000000060,
      -0.00000000216,     0.00000000613,
       0.00000000121,     0.00000000062,
       0.00000000623,     0.00000000236,
       0.00000000052,    -0.00000000111,
      -0.00000010575,     0.00000003154,
       0.00000000067,     0.00000000260,
      -0.00000000055,    -0.00000000439,
       0.00000000119,    -0.00000000147,
       0.00000002630,    -0.00000000678,
       0.00000000098,     0.00000000258,
      -0.00000000192,     0.00000000075,
      -0.00000000063,    -0.00000000190,
      -0.00000001070,    -0.00000000070,
       0.00000000688,     0.00000000086,
      -0.00000000674,     0.00000000028,
      -0.00000000043,    -0.00000000683,
      -0.00000005562,    -0.00000000299,
      -0.00000000248,     0.00000000328,
       0.00000000692,    -0.00000021516,
      -0.00000001413,    -0.00000001070,
       0.00000011056,    -0.00000000304,
      -0.00000000307,    -0.00000010974,
       0.00000083357,     0.00000017049,
       0.00000005381,    -0.00000006308,
      -0.00000003827,     0.00000075557,
       0.00000005071,     0.00000003874,
       0.00000000059,    -0.00000000108,
      -0.00000011675,     0.00000000750,
      -0.00000000108,     0.00000011797,
       0.00001129697,    -0.00000325824,
       0.00000023754,    -0.00000049895,
      -0.00000001847,    -0.00000000774,
       0.00000001261,    -0.00000126262,
      -0.00000007991,    -0.00000005836,
      -0.00000000218,     0.00000000465,
       0.00000000198,     0.00000000074,
      -0.00000002960,    -0.00000002242,
      -0.00000000405,     0.00000000097,
       0.00000003245,     0.00000005740,
       0.00000000123,    -0.00000000272,
      -0.00000000275,     0.00000000374,
       0.00000000011,     0.00000000331,
      -0.00000000033,    -0.00000000241,
       0.00000000204,    -0.00000000060,
       0.00000000051,     0.00000000194,
       0.00000011581,    -0.00000002043,
      -0.00000000029,    -0.00000000231,
      -0.00000002920,    -0.00000005510,
      -0.00000002909,     0.00000000842,
       0.00000010505,     0.00000036359,
      -0.00000000850,    -0.00000001990,
       0.00000008192,    -0.00000005266,
      -0.00000003712,    -0.00000005796,
       0.00000001491,     0.00000000122,
      -0.00000000153,     0.00000001488,
       0.00000016331,    -0.00000003516,
      -0.00000015265,    -0.00000075249,
       0.00000000687,    -0.00000000300,
       0.00000001985,     0.00000009800,
      -0.00000000777,    -0.00000000648,
       0.00000006863,    -0.00000000026,
      -0.00000000530,     0.00000000906,
      -0.00000000488,     0.00000035524,
      -0.00000001858,    -0.00000001140,
       0.00000039308,     0.00000190114,
      -0.00000001583,     0.00000000365,
      -0.00000000400,    -0.00000000006,
      -0.00000007428,    -0.00000007696,
      -0.00000000231,     0.00000001357,
      -0.00000001046,    -0.00000015042,
      -0.00000000278,    -0.00000001378,
      -0.00000000002,     0.00000000388,
       0.00000000068,     0.00000000264,
       0.00000376881,    -0.00000139244,
       0.00000002462,     0.00000010379,
       0.00000000457,     0.00000001746,
      -0.00000008719,    -0.00000039471,
       0.00000000228,    -0.00000000403,
       0.00000043361,    -0.00000005080,
       0.00000001719,    -0.00000002404,
      -0.00000000076,    -0.00000000067,
       0.00000003615,     0.00000044702,
       0.00000002490,     0.00000001513,
       0.00000000066,    -0.00000000077,
      -0.00000000086,    -0.00000000120,
      -0.00000000140,    -0.00000000130,
       0.00000000027,     0.00000000095,
      -0.00000000111,     0.00000000027,
       0.00000000027,     0.00000000110,
      -0.00000002460,    -0.00000002370,
       0.00000009260,     0.00000011149,
      -0.00000004802,    -0.00000004363,
      -0.00000001898,    -0.00000002211,
       0.00000000930,     0.00000000280,
       0.00000008160,     0.00000008272,
      -0.00000000785,    -0.00000000435,
       0.00000016680,    -0.00000016414,
      -0.00000000971,     0.00000001459,
       0.00000000242,    -0.00000005171,
       0.00000000932,     0.00000000677,
      -0.00000009598,    -0.00000002226,
       0.00000001221,    -0.00000000193,
       0.00000106274,     0.00000043077,
      -0.00000002853,    -0.00000001950,
      -0.00000041004,     0.00000041342,
       0.00000002320,    -0.00000003551,
       0.00000000225,     0.00000000252,
       0.00000009039,    -0.00000013443,
      -0.00000000202,     0.00000001459,
      -0.00000000162,     0.00000000040,
      -0.00000007139,     0.00000001802,
       0.00000001721,    -0.00000001581,
      -0.00000000080,     0.00000000107,
       0.00000000039,     0.00000000140,
       0.00000019845,     0.00000017936,
      -0.00000001017,    -0.00000001957,
       0.00000001582,    -0.00000001981,
      -0.00000000063,     0.00000000188,
       0.00000005553,    -0.00000002073,
       0.00000000014,     0.00000000224,
       0.00000002146,     0.00000005584,
      -0.00000000221,     0.00000000008,
       0.00000000092,    -0.00000000080,
       0.00000000090,     0.00000000110,
       0.00000000046,    -0.00000000072,
      -0.00000000084,     0.00000000073,
       0.00000005110,     0.00000013210,
      -0.00000034460,    -0.00000039870,
       0.00000009520,    -0.00000002249,
       0.00000053260,     0.00000215783,
      -0.00000014060,    -0.00000001370,
       0.00000010020,    -0.00000102650,
      -0.00000000288,     0.00000000120,
      -0.00000003078,    -0.00000004821,
       0.00000000465,     0.00000000594,
       0.00000000955,    -0.00000000393,
      -0.00000000151,     0.00000000119,
      -0.00000000158,    -0.00000000117,
       0.00000000062,    -0.00000000048,
      -0.00000000051,     0.00000000070,
      -0.00000009701,    -0.00000015396,
       0.00000001490,     0.00000001928,
      -0.00000000484,     0.00000000162,
      -0.00000000682,    -0.00000000067,
       0.00000000100,    -0.00000000697,
      -0.00000000055,    -0.00000000193,
       0.00000000045,     0.00000000107,
       0.00000000112,    -0.00000000028,
       0.00000000028,     0.00000000112,
       0.00000000101,     0.00000002000,
      -0.00000000253,    -0.00000000016,
       0.00000000108,    -0.00000000027,
       0.00000000074,     0.00000000123,
      -0.00000000124,     0.00000000057,
       0.00000000632,     0.00000005602,
       0.00000000164,     0.00000000021,
       0.00000000038,     0.00000000068,
       0.00000000092,     0.00000000121,
      -0.00000000117,     0.00000000104,
       0.00000000926,     0.00000000210,
       0.00000000101,     0.00000000089,
      -0.00000000131,    -0.00000000417,
       0.00000000369,    -0.00000000068,
       0.00000000518,    -0.00000000813,
       0.00000000640,    -0.00000001955,
      -0.00000000108,    -0.00000000041,
      -0.00000000053,     0.00000000048,
      -0.00000000036,    -0.00000000127,
       0.00000000177,     0.00000000020,
       0.00000013410,    -0.00000002177,
      -0.00000001588,    -0.00000004998,
       0.00000000190,     0.00000000859,
       0.00000000312,    -0.00000000106,
      -0.00000000265,     0.00000000321,
       0.00000000098,     0.00000000044,
       0.00000000324,     0.00000000264,
       0.00000000043,    -0.00000000098,
       0.00000000011,    -0.00000000127,
      -0.00000000127,    -0.00000000011,
      -0.00000000650,    -0.00000000027,
      -0.00000000040,    -0.00000001131,
      -0.00000000080,    -0.00000000055,
       0.00000001262,     0.00000000009,
       0.00000000020,    -0.00000001246,
       0.00000009210,     0.00000003260,
       0.00000000714,    -0.00000000747,
      -0.00000000076,     0.00000003951,
       0.00000000222,     0.00000000180,
      -0.00000001794,    -0.00000000030,
      -0.00000000035,     0.00000001708,
       0.00002131145,    -0.00000503704,
       0.00000055406,    -0.00000101406,
      -0.00000004480,    -0.00000001920,
      -0.00000009917,    -0.00000047259,
      -0.00000002335,    -0.00000001329,
      -0.00000000054,     0.00000000108,
      -0.00000000675,     0.00000000364,
       0.00000000066,     0.00000000074,
      -0.00000000127,    -0.00000000248,
      -0.00000000073,     0.00000000055,
       0.00000000024,    -0.00000000252,
       0.00000000269,     0.00000000022,
       0.00000000555,    -0.00000001474,
      -0.00000056352,    -0.00000021346,
       0.00000003250,    -0.00000005870,
      -0.00000007053,     0.00000019386,
      -0.00000001587,    -0.00000002123,
       0.00000170747,     0.00000067106,
      -0.00000017445,     0.00000013400,
      -0.00000006508,     0.00000001421,
      -0.00000001037,    -0.00000006648,
      -0.00000001152,     0.00000002975,
       0.00000004129,     0.00000001599,
       0.00000052683,    -0.00000169762,
       0.00000021877,     0.00000011053,
      -0.00000072566,    -0.00000024201,
       0.00000004731,    -0.00000009267,
       0.00000000591,     0.00000000300,
       0.00000060162,    -0.00000011977,
       0.00000000710,     0.00000000127,
       0.00000011946,     0.00000060310,
      -0.00000000161,     0.00000000767,
       0.00000000083,     0.00000000120,
      -0.00000000048,     0.00000000144,
      -0.00000000144,    -0.00000000049,
       0.00000000112,    -0.00000000414,
       0.00000000112,     0.00000000074,
       0.00000000413,     0.00000000115,
      -0.00000000074,     0.00000000112,
       0.00000000575,    -0.00000002316,
      -0.00000000129,    -0.00000000120,
       0.00000000391,    -0.00000000324,
      -0.00000000017,    -0.00000000013,
      -0.00000000022,    -0.00000000286,
       0.00000000004,    -0.00000000003,
      -0.00000000008,     0.00000000065,
       0.00000000019,    -0.00000000091,
      -0.00000000004,    -0.00000000003,
       0.00000000419,    -0.00000006281,
      -0.00000000337,    -0.00000000269,
       0.00000001848,     0.00000000050,
      -0.00000000014,    -0.00000001970,
      -0.00000000007,     0.00000000067,
       0.00000001020,    -0.00000010769,
      -0.00000000583,    -0.00000000484,
       0.00000004419,     0.00000000048,
      -0.00000000052,    -0.00000004712,
       0.00000000006,    -0.00000000057,
      -0.00000000039,    -0.00000000013,
      -0.00000002380,    -0.00000015513,
      -0.00000000657,    -0.00000000639,
       0.00000014248,     0.00000000935,
       0.00000000046,    -0.00000000048,
      -0.00000000339,    -0.00000013623,
       0.00000000042,     0.00000000090,
       0.00000000095,     0.00000000038,
      -0.00000000136,    -0.00000000150,
      -0.00032143485,    -0.00002877728,
      -0.00001795333,     0.00002208978,
       0.00000136434,     0.00000098000,
       0.00000003090,    -0.00000007958,
      -0.00009023546,     0.00061651695,
       0.00004439093,     0.00003795126,
       0.00000212804,    -0.00000281588,
      -0.00000016558,    -0.00000007053,
      -0.00000000035,     0.00000000766,
       0.00065285562,    -0.00000434708,
      -0.00000219775,    -0.00000082612,
      -0.00000005261,     0.00000090653,
       0.00000003320,    -0.00000000176,
      -0.00000000050,    -0.00000000230,
       0.00001479520,    -0.00064261110,
       0.00000072372,    -0.00000281199,
       0.00000016957,    -0.00000003415,
       0.00000000302,     0.00000003462,
       0.00000000248,    -0.00000000080,
      -0.00000034453,     0.00000111939,
      -0.00000007951,     0.00000020398,
       0.00000001137,     0.00000000206,
       0.00000000101,    -0.00000000150,
      -0.00000039212,     0.00000057997,
       0.00000016206,     0.00000000688,
      -0.00000001400,    -0.00000001962,
      -0.00000000206,     0.00000000159,
       0.00275615018,    -0.00222937739,
      -0.00007588378,    -0.00002995472,
       0.00000384125,     0.00000542839,
       0.00000014906,    -0.00000021108,
      -0.00000000898,    -0.00000000259,
       0.00022957813,    -0.00148093026,
      -0.00010747843,    -0.00009236368,
      -0.00000518255,     0.00000684342,
       0.00000040270,     0.00000017113,
       0.00000000050,    -0.00000001880,
       0.00065045580,    -0.00000771755,
       0.00000543625,     0.00000402878,
       0.00000040616,     0.00000007643,
      -0.00000006669,    -0.00000003164,
      -0.00000000356,     0.00000000348,
      -0.00003136834,    -0.00068657275,
      -0.00000331713,     0.00000278843,
       0.00000135570,     0.00000040184,
       0.00000002818,    -0.00000005976,
      -0.00000000321,    -0.00000000145,
       0.00000156956,    -0.00000304994,
       0.00000003851,    -0.00000046625,
      -0.00000003822,     0.00000000500,
      -0.00000000326,     0.00000000528,
      -0.00000055815,    -0.00000102480,
      -0.00000049827,     0.00000009636,
       0.00000004630,     0.00000004852,
       0.00000000574,    -0.00000000462,
      -0.00000006718,    -0.00000577886,
      -0.00000055490,     0.00000039383,
       0.00000003775,     0.00000003340,
      -0.00000002410,     0.00000012847,
       0.00000001083,     0.00000000912,
       0.00000000064,    -0.00000000096,
       0.00000028461,     0.00000004063,
       0.00000000670,    -0.00000000541,
      -0.00000000069,    -0.00000000044,
      -0.00000003735,    -0.00000024051,
       0.00000000536,     0.00000000605,
       0.00000000042,    -0.00000000075,
       0.00000000016,    -0.00000000156,
       0.00000000063,    -0.00000000099,
      -0.00000074303,    -0.00001295637,
      -0.00000123168,     0.00000088550,
       0.00000008438,     0.00000007311,
       0.00000000121,     0.00000002434,
       0.00000000153,     0.00000000108,
       0.00000036327,     0.00000004852,
       0.00000000802,    -0.00000000677,
      -0.00000000094,    -0.00000000062,
      -0.00000004434,    -0.00000030421,
       0.00000000673,     0.00000000779,
       0.00000000051,    -0.00000000098,
       0.00000005065,    -0.00000000020,
       0.00000000229,     0.00000001008,
       0.00000000000,    -0.00000000054,
       0.00000001943,     0.00000000013,
       0.00000000005,    -0.00000000720,
       0.00000057150,    -0.00000000146,
      -0.00000000016,    -0.00000000726,
       0.00000000160,    -0.00000000027,
       0.00000000053,     0.00000000152,
       0.00000000001,    -0.00000000148,
       0.00000002099,     0.00000000015,
       0.00000000005,    -0.00000000773,
       0.00000060975,    -0.00000000106,
      -0.00000000017,    -0.00000000839,
       0.00000000168,    -0.00000000028,
       0.00000000055,     0.00000000159,
      -0.00000000218,     0.00000000957,
       0.00000000382,     0.00000000031,
       0.00000000078,    -0.00000001297,
      -0.00000000157,     0.00000000020,
       0.00000000032,    -0.00000000114,
       0.00000005333,    -0.00000000015,
      -0.00000000087,     0.00000000000,
      -0.00000000005,    -0.00000000086,
       0.00000000340,    -0.00000001150,
       0.00000000920,     0.00000000270,
      -0.00000000295,     0.00000000673,
      -0.00000000830,     0.00000000293,
      -0.00000000070,    -0.00000000468,
      -0.00000000101,     0.00000000063,
       0.00000001904,    -0.00000001545,
      -0.00000000589,    -0.00000000224,
      -0.00000001671,    -0.00000001830,
      -0.00000000195,     0.00000000628,
       0.00000016271,    -0.00000004219,
      -0.00000003200,    -0.00000004884,
      -0.00000000752,     0.00000000852,
       0.00000000207,     0.00000001131,
       0.00000000064,    -0.00000000059,
      -0.00000002753,     0.00000005504,
       0.00000001464,     0.00000000085,
      -0.00000000143,    -0.00000000111,
       0.00000005274,     0.00000002883,
       0.00000000173,    -0.00000001369,
      -0.00000000084,     0.00000000064,
       0.00000000076,     0.00000000072,
       0.00000002110,    -0.00000000736,
       0.00000000044,    -0.00000000067,
      -0.00000000253,    -0.00000000255,
      -0.00000000263,     0.00000000235,
       0.00000001024,    -0.00000000039,
       0.00000000134,     0.00000003390,
       0.00000001110,    -0.00000000303,
      -0.00000001191,     0.00000000268,
       0.00000000109,     0.00000000183,
       0.00000000581,     0.00000001243,
       0.00000000140,    -0.00000000172,
      -0.00000006572,    -0.00000000971,
       0.00000000097,     0.00000001926,
      -0.00000000071,    -0.00000008271,
      -0.00000002755,     0.00000000517,
       0.00000000283,     0.00000000612,
       0.00000000941,    -0.00000003114,
      -0.00000000566,    -0.00000000033,
      -0.00000003245,    -0.00000000763,
      -0.00000000033,     0.00000000430,
      -0.00000012271,    -0.00000008892,
      -0.00000000521,    -0.00000000036,
       0.00000000157,     0.00000000158,
       0.00000001546,     0.00000004951,
       0.00000000835,    -0.00000000603,
      -0.00000000026,    -0.00000000097,
       0.00000004968,    -0.00000001526,
      -0.00000000609,    -0.00000000807,
      -0.00000000100,     0.00000000037,
       0.00000000121,     0.00000000195,
       0.00000000212,    -0.00000000118,
      -0.00000000307,    -0.00000000130,
      -0.00000000537,     0.00000000954,
       0.00000005099,     0.00000002877,
       0.00000000565,    -0.00000001359,
       0.00000000697,    -0.00000001903,
      -0.00000001843,     0.00000000035,
       0.00000004697,    -0.00000009944,
      -0.00000001488,    -0.00000000560,
      -0.00000009400,    -0.00000007510,
      -0.00000001118,     0.00000001631,
       0.00000003446,     0.00000000162,
      -0.00000000088,    -0.00000003969,
      -0.00000000185,    -0.00000000113,
       0.00000101024,     0.00000145848,
       0.00000024761,    -0.00000021661,
      -0.00000002675,    -0.00000002256,
      -0.00000019597,     0.00000000620,
       0.00000001749,     0.00000002881,
       0.00000000476,    -0.00000000334,
       0.00000000022,     0.00000000209,
      -0.00000013056,    -0.00000005305,
      -0.00000001178,     0.00000001203,
      -0.00000004711,     0.00000012990,
       0.00000001141,     0.00000000981,
      -0.00000000106,    -0.00000000150,
      -0.00000000064,     0.00000000207,
      -0.00000000977,     0.00000000003,
       0.00000000030,     0.00000002036,
      -0.00000000057,    -0.00000000123,
      -0.00000064669,     0.00000001737,
      -0.00000000229,    -0.00000006885,
      -0.00000000887,     0.00000000251,
      -0.00000000047,     0.00000000283,
      -0.00000331314,     0.00000015196,
       0.00000001713,    -0.00000001167,
       0.00000000754,    -0.00000037766,
       0.00000002679,    -0.00000002452,
      -0.00000000201,     0.00000000048,
       0.00000000059,     0.00000000043,
       0.00000003074,     0.00000001365,
      -0.00000000067,    -0.00000000136,
      -0.00000000108,     0.00000000034,
       0.00000000034,     0.00000000108,
       0.00000005404,    -0.00000000577,
       0.00000000005,    -0.00000000217,
      -0.00000000673,     0.00000001149,
       0.00000001458,     0.00000000549,
      -0.00000000161,     0.00000000185,
      -0.00000000208,    -0.00000000078,
      -0.00000000138,    -0.00000000112,
       0.00000020610,     0.00000049816,
      -0.00000004902,    -0.00000002716,
      -0.00000007002,    -0.00000000675,
       0.00000000647,    -0.00000000841,
       0.00000000098,     0.00000000089,
       0.00000003591,     0.00000000460,
      -0.00000000167,     0.00000000078,
       0.00000000476,     0.00000002725,
       0.00000000057,     0.00000000153,
       0.00000229288,     0.00000101769,
      -0.00000012752,    -0.00000004575,
       0.00000002376,    -0.00000000112,
      -0.00000000198,     0.00000000277,
      -0.00000012419,    -0.00000000505,
      -0.00000000365,    -0.00000011486,
       0.00000000047,    -0.00000000082,
      -0.00000002696,     0.00000000921,
      -0.00000000034,    -0.00000000091,
       0.00000003729,    -0.00000000635,
       0.00000000018,     0.00000000100,
      -0.00000000387,    -0.00000001109,
      -0.00000001228,     0.00000000345,
       0.00000000069,    -0.00000000148,
      -0.00000000522,    -0.00000000496,
       0.00000000090,     0.00000000063,
       0.00000000299,    -0.00000000561,
       0.00000047766,    -0.00000230428,
       0.00000017555,     0.00000031314,
      -0.00000002884,    -0.00000000324,
       0.00000016468,     0.00000001596,
      -0.00000001964,     0.00000001378,
       0.00000000008,    -0.00000000150,
       0.00000002030,     0.00000002175,
      -0.00000000139,    -0.00000000299,
      -0.00000001690,     0.00000003351,
      -0.00000000023,    -0.00000000200,
      -0.00000056066,    -0.00000069771,
       0.00000007908,     0.00000007374,
      -0.00000000658,    -0.00000000107,
       0.00000000171,    -0.00000000087,
       0.00000003326,    -0.00000000336,
      -0.00000000093,     0.00000000310,
      -0.00000002946,     0.00000003363,
       0.00000000369,    -0.00000000197,
       0.00000000835,    -0.00000054056,
       0.00000001619,     0.00000011429,
      -0.00000000996,    -0.00000000991,
       0.00000001444,     0.00000000048,
      -0.00000000366,     0.00000000058,
      -0.00000000695,    -0.00000000560,
       0.00000000127,     0.00000000056,
       0.00000000964,    -0.00000000377,
      -0.00000000189,     0.00000000053,
      -0.00000006415,     0.00000011883,
       0.00000001264,    -0.00000003637,
      -0.00000000453,    -0.00000000230,
       0.00000000112,     0.00000000035,
      -0.00000000034,    -0.00000000124,
       0.00000000140,    -0.00000000027,
      -0.00000000824,    -0.00000000465,
       0.00000000686,     0.00000000911,
      -0.00000002693,     0.00000002028,
      -0.00000000219,    -0.00000000042,
      -0.00000000021,     0.00000000108,
       0.00000000076,     0.00000000366,
       0.00000000064,    -0.00000000286,
       0.00000000363,    -0.00000000078,
      -0.00000000286,    -0.00000000064,
       0.00000000231,    -0.00000000035,
      -0.00000000059,    -0.00000000139,
      -0.00000000134,     0.00000000075,
      -0.00000000230,    -0.00000000784,
      -0.00000000324,     0.00000000055,
       0.00000000242,    -0.00000000111,
      -0.00000000117,    -0.00000000291,
       0.00000150134,     0.00000243344,
       0.00000042484,    -0.00000035313,
      -0.00000004770,    -0.00000004939,
      -0.00000005497,     0.00000003511,
       0.00000000839,     0.00000000942,
       0.00000000110,    -0.00000000113,
      -0.00000001083,    -0.00000004115,
      -0.00000000440,     0.00000000326,
      -0.00000004107,     0.00000001110,
       0.00000000339,     0.00000000431,
       0.00000000126,     0.00000000026,
       0.00000000027,    -0.00000000125,
      -0.00000000022,     0.00000000500,
      -0.00000015077,    -0.00000000062,
       0.00000000022,    -0.00000001681,
      -0.00000000059,     0.00000000080,
       0.00000746721,    -0.00000003221,
       0.00000000365,    -0.00000050660,
       0.00000000167,    -0.00000003963,
      -0.00000000347,     0.00000000249,
       0.00000003867,     0.00000000030,
      -0.00000000236,    -0.00000000364,
       0.00000000070,    -0.00000000182,
       0.00000000180,     0.00000000073,
       0.00000000111,    -0.00000000036,
       0.00000000035,     0.00000000110,
      -0.00000000088,     0.00000003222,
      -0.00000002737,    -0.00000000075,
       0.00000000248,    -0.00000000033,
       0.00000000773,    -0.00000001793,
      -0.00000003447,     0.00000000202,
       0.00000000011,     0.00000000248,
       0.00000000029,     0.00000000072,
      -0.00000000063,     0.00000001704,
       0.00000000004,     0.00000000247,
      -0.00000000011,    -0.00000000037,
       0.00000000001,    -0.00000000036,
      -0.00000000001,     0.00000000061,
      -0.00000000237,     0.00000004308,
       0.00000000053,     0.00000000007,
      -0.00000000008,     0.00000000053,
       0.00000000000,     0.00000000050,
      -0.00000000381,     0.00000007194,
       0.00000000172,     0.00000000052,
      -0.00000000060,     0.00000000225,
      -0.00000000803,     0.00000014334,
       0.00000000444,    -0.00000000515,
       0.00000000024,     0.00000000043,
       0.00000000003,    -0.00000000278,
       0.00000000029,    -0.00000000025,
      -0.00000000001,     0.00000000166,
       0.00020577646,     0.00000563936,
      -0.00000004946,    -0.00000007307,
       0.00000000005,     0.00000006877,
       0.00000478257,    -0.00023786377,
      -0.00000007144,     0.00000001308,
       0.00000009154,     0.00000000287,
      -0.00000076476,     0.00000316508,
       0.00000083609,    -0.00000075580,
      -0.00000004547,    -0.00000004052,
      -0.00000000194,     0.00000000333,
      -0.00000910427,     0.00000790299,
       0.00000052757,     0.00000061041,
       0.00000004545,    -0.00000005899,
      -0.00000000273,    -0.00000000052,
      -0.00000002429,    -0.00000138227,
       0.00000001567,    -0.00000015354,
      -0.00000000057,     0.00000000934,
       0.00000009154,    -0.00000101673,
       0.00000001502,     0.00000018780,
       0.00000000033,     0.00000000555,
       0.03363446980,    -0.00003436390,
       0.00000455602,     0.00000164427,
      -0.00000005740,     0.00001756000,
      -0.00000001058,    -0.00000001604,
      -0.00001166001,     0.00259781931,
       0.00000013279,    -0.00000012452,
      -0.00000126412,    -0.00000000824,
       0.00002518229,    -0.00012217352,
       0.00000523135,     0.00000931792,
      -0.00000029027,     0.00000075946,
      -0.00000001572,    -0.00000003667,
      -0.00004522399,    -0.00002069197,
       0.00000463524,    -0.00001223248,
       0.00000032828,     0.00000073342,
      -0.00000001169,     0.00000003159,
       0.00000574731,    -0.00000447757,
      -0.00000098997,    -0.00000037064,
      -0.00000004416,     0.00000004518,
       0.00000000282,    -0.00000000092,
      -0.00000768129,    -0.00000350556,
      -0.00000074629,     0.00000054728,
       0.00000006515,     0.00000003454,
       0.00000000166,    -0.00000000116,
       0.00011329540,     0.00000329550,
      -0.00000021719,    -0.00000029746,
       0.00000001743,     0.00000005023,
      -0.00000003075,     0.00000015317,
      -0.00000000127,    -0.00000000549,
      -0.00000077148,    -0.00000088501,
       0.00000002394,     0.00000010732,
      -0.00000000085,     0.00000000663,
      -0.00000058829,    -0.00000056413,
       0.00000000107,    -0.00000011759,
       0.00000000082,     0.00000000778,
       0.00000000234,    -0.00000000040,
      -0.00000000230,     0.00000000023,
       0.00024649447,    -0.00000001095,
       0.00000002829,     0.00000001127,
      -0.00000000029,     0.00000012673,
      -0.00000001073,     0.00000010941,
      -0.00000000009,     0.00000000116,
       0.00000000108,     0.00000000001,
      -0.00000087495,    -0.00000019012,
      -0.00000005773,     0.00000006758,
       0.00000000429,     0.00000000361,
       0.00000017238,    -0.00000087800,
      -0.00000006641,    -0.00000005900,
      -0.00000000350,     0.00000000436,
      -0.00000000218,     0.00000000194,
      -0.00000000182,     0.00000000157,
       0.00000000004,    -0.00000000002,
       0.00000000131,     0.00000000000,
      -0.00000000001,     0.00000000130,
      -0.00000000139,    -0.00000000015,
       0.00000001412,     0.00000000010,
      -0.00000000012,     0.00000001416,
      -0.00000000144,    -0.00000000020,
       0.00000001466,     0.00000000002,
      -0.00000000003,     0.00000001470,
       0.00000000022,    -0.00000000034,
       0.00000000051,    -0.00000000028,
      -0.00000000093,    -0.00000000146,
       0.00000000304,     0.00000000006,
      -0.00000000005,     0.00000000304,
       0.00000000133,     0.00000000000,
       0.00000000000,     0.00000000133,
      -0.00000000045,     0.00000000214,
       0.00000000085,     0.00000000054,
       0.00000000054,    -0.00000000085,
      -0.00000000234,    -0.00000000112,
       0.00000000111,    -0.00000000103,
      -0.00000000118,    -0.00000000019,
      -0.00000000081,    -0.00000000057,
       0.00000000127,     0.00000000056,
       0.00000000047,    -0.00000000127,
      -0.00000000645,    -0.00000001919,
      -0.00000001670,     0.00000000469,
       0.00000001153,     0.00000000913,
       0.00000000457,     0.00000002805,
       0.00000001631,    -0.00000001994,
      -0.00000000607,    -0.00000000290,
      -0.00000000480,    -0.00000000098,
      -0.00000000408,     0.00000001150,
      -0.00000007497,     0.00000004407,
       0.00000001664,    -0.00000002398,
       0.00000001965,     0.00000003176,
      -0.00000000857,    -0.00000000797,
       0.00000000085,     0.00000000316,
       0.00000004055,    -0.00000000574,
      -0.00000000821,    -0.00000000391,
      -0.00000000588,    -0.00000004046,
      -0.00000000372,     0.00000000797,
      -0.00000000012,     0.00000004752,
       0.00000000814,    -0.00000000328,
       0.00000010417,     0.00000000537,
      -0.00000000626,    -0.00000001869,
      -0.00000000140,     0.00000000126,
      -0.00000001346,    -0.00000009567,
      -0.00000000685,     0.00000001276,
      -0.00000010084,     0.00000000762,
       0.00000001092,     0.00000000905,
      -0.00000030339,    -0.00000034021,
      -0.00000000009,     0.00000007294,
       0.00000000493,    -0.00000000499,
      -0.00000025202,    -0.00000000969,
       0.00000001534,     0.00000004576,
       0.00000000566,    -0.00000000406,
       0.00000000568,    -0.00000005254,
      -0.00000000862,     0.00000000282,
      -0.00000004325,     0.00000000186,
       0.00000000531,     0.00000000374,
       0.00000008946,    -0.00000005220,
      -0.00000003288,    -0.00000000268,
      -0.00000000338,    -0.00000002310,
      -0.00000000244,     0.00000000425,
      -0.00000001286,    -0.00000000566,
       0.00000000087,     0.00000000652,
       0.00000000089,    -0.00000000035,
      -0.00000000593,     0.00000001306,
       0.00000000652,    -0.00000000090,
      -0.00000000032,    -0.00000000079,
       0.00000000187,    -0.00000000530,
      -0.00000000618,    -0.00000000178,
      -0.00000000405,     0.00000001064,
       0.00000001051,     0.00000000092,
      -0.00000000111,    -0.00000000076,
       0.00000010256,     0.00000012056,
       0.00000000780,    -0.00000002524,
       0.00000001674,    -0.00000000805,
      -0.00000000242,    -0.00000000199,
      -0.00000001331,     0.00000004225,
       0.00000000354,    -0.00000000336,
       0.00000004200,     0.00000002118,
      -0.00000000101,    -0.00000000417,
       0.00000089923,     0.00000029296,
       0.00000008474,    -0.00000000923,
      -0.00000002461,     0.00000006106,
       0.00000001310,     0.00000000164,
      -0.00000000009,    -0.00000000163,
      -0.00000035120,    -0.00000007629,
      -0.00000002272,     0.00000002697,
       0.00000000144,     0.00000000114,
      -0.00000007533,     0.00000035149,
       0.00000002687,     0.00000002231,
       0.00000000114,    -0.00000000141,
       0.00000000131,     0.00000000155,
       0.00000000168,    -0.00000000129,
      -0.00000000095,    -0.00000000007,
       0.00000000373,     0.00000000196,
       0.00000000152,    -0.00000000385,
       0.00000007131,    -0.00000011916,
      -0.00000001182,     0.00000000644,
       0.00000000356,     0.00000000174,
       0.00000003630,    -0.00000000032,
       0.00000000160,    -0.00000003618,
      -0.00002609903,    -0.00000546301,
      -0.00000144765,     0.00000178162,
       0.00000010800,     0.00000008622,
       0.00000017669,    -0.00000078575,
      -0.00000006193,    -0.00000005379,
      -0.00000000354,     0.00000000398,
       0.00000056393,     0.00000007984,
      -0.00000000614,    -0.00000000663,
       0.00000007544,    -0.00000055371,
      -0.00000000611,     0.00000000719,
       0.00000000452,    -0.00000000345,
      -0.00000000030,    -0.00000000074,
      -0.00000000350,    -0.00000000437,
      -0.00000000074,     0.00000000031,
      -0.00000003562,    -0.00000000808,
      -0.00000000371,    -0.00000000021,
      -0.00000000029,     0.00000000366,
      -0.00000003969,    -0.00000003916,
       0.00000000084,    -0.00000000626,
       0.00000000399,    -0.00000000024,
      -0.00000000018,    -0.00000000417,
      -0.00000000413,     0.00000000023,
      -0.00000000023,    -0.00000000414,
      -0.00000000275,    -0.00000000097,
       0.00000000111,     0.00000000152,
       0.00000000164,     0.00000000141,
      -0.00000003224,    -0.00000003048,
       0.00000000568,     0.00000000024,
      -0.00000000062,     0.00000000071,
      -0.00000000607,    -0.00000000108,
      -0.00000000102,    -0.00000000504,
      -0.00001066470,    -0.00002826123,
       0.00000283399,     0.00000199540,
      -0.00000016822,     0.00000014904,
       0.00000051559,    -0.00000014426,
      -0.00000003499,     0.00000005082,
      -0.00000000339,    -0.00000000330,
       0.00000052007,     0.00000067920,
      -0.00000005237,    -0.00000008979,
       0.00000000864,     0.00000000122,
       0.00000065962,     0.00000077628,
      -0.00000008816,     0.00000005150,
       0.00000000047,    -0.00000000854,
       0.00000000388,    -0.00000000162,
       0.00000000302,     0.00000000172,
      -0.00000000221,     0.00000000131,
       0.00000000082,    -0.00000000076,
       0.00000000068,    -0.00000000071,
       0.00000001118,     0.00000002842,
      -0.00000000863,     0.00000000135,
       0.00000000497,    -0.00000000306,
       0.00000000053,     0.00000000145,
      -0.00000000363,    -0.00000000282,
       0.00000000057,    -0.00000000106,
       0.00000000302,     0.00000000300,
      -0.00000000153,     0.00000000014,
      -0.00000427208,     0.00002724471,
      -0.00000217934,    -0.00000353569,
       0.00000034391,     0.00000002383,
      -0.00000001380,     0.00000001907,
      -0.00000107785,    -0.00000017529,
       0.00000013927,    -0.00000008664,
      -0.00000000083,     0.00000001353,
      -0.00000000067,    -0.00000000048,
      -0.00000005255,    -0.00000008192,
       0.00000000847,     0.00000000431,
       0.00000012567,    -0.00000004921,
      -0.00000000996,     0.00000001303,
      -0.00000000055,    -0.00000000092,
      -0.00000000032,    -0.00000000304,
      -0.00000000013,     0.00000000269,
       0.00000000153,    -0.00000000446,
      -0.00000000243,    -0.00000000072,
       0.00000000111,     0.00000000003,
      -0.00000001165,     0.00000003148,
      -0.00000000731,    -0.00000000703,
       0.00000000650,     0.00000000224,
      -0.00000000137,     0.00000000144,
      -0.00000000049,    -0.00000000686,
       0.00000000145,     0.00000000035,
       0.00000000146,     0.00000000240,
      -0.00000000098,    -0.00000000084,
       0.00000003457,    -0.00000010128,
       0.00000000316,     0.00000002100,
      -0.00000000244,    -0.00000000227,
      -0.00000001052,     0.00000003527,
      -0.00000000212,    -0.00000000530,
      -0.00000003532,    -0.00000001059,
       0.00000000530,    -0.00000000213,
      -0.00000000262,    -0.00000000107,
       0.00000000105,    -0.00000000259,
       0.00000000088,    -0.00000000018,
       0.00000000019,     0.00000000089,
       0.00000000018,    -0.00000000540,
       0.00000000062,     0.00000000072,
       0.00000000541,     0.00000000017,
      -0.00000000069,     0.00000000059,
       0.00000000079,     0.00000000098,
      -0.00000000098,     0.00000000079,
       0.00000000074,     0.00000000079,
       0.00000000082,    -0.00000000081,
       0.00000002619,     0.00000002707,
      -0.00000000104,     0.00000000093,
      -0.00000000081,    -0.00000000096,
      -0.00000000096,     0.00000000081,
      -0.00000005379,    -0.00000005745,
       0.00000000157,    -0.00000000141,
      -0.00000002033,    -0.00000001018,
       0.00000000341,     0.00000000064,
       0.00000000064,    -0.00000000341,
      -0.00000000092,    -0.00000000034,
      -0.00000000035,     0.00000000092,
      -0.00000004820,     0.00000000140,
       0.00000000285,     0.00000000461,
       0.00000000452,    -0.00000000328,
       0.00000012112,     0.00000015541,
      -0.00000001605,    -0.00000000783,
      -0.00000000603,     0.00000000434,
      -0.00000000078,    -0.00000000311,
      -0.00000000312,     0.00000000078,
       0.00000000112,    -0.00000000030,
      -0.00000000125,     0.00000000032,
      -0.00000000084,     0.00000000130,
       0.00000000125,     0.00000000084,
      -0.00000000699,     0.00000000234,
       0.00000000006,     0.00000000123,
       0.00000000231,     0.00000000704,
       0.00000000123,    -0.00000000007,
       0.00000018432,     0.00000005774,
       0.00000001364,    -0.00000001713,
      -0.00000000105,     0.00000000511,
      -0.00000000369,    -0.00000000018,
      -0.00000000017,     0.00000000370,
       0.00000003189,    -0.00000000145,
      -0.00000004778,    -0.00000019317,
       0.00000002351,     0.00000001262,
       0.00000000550,    -0.00000000137,
      -0.00000000036,     0.00000000067,
      -0.00000000361,    -0.00000002291,
       0.00000000069,    -0.00000000093,
       0.00000000652,    -0.00000000013,
       0.00000000004,    -0.00000000647,
      -0.00003806857,    -0.00000967540,
      -0.00000262362,     0.00000299690,
       0.00000019500,     0.00000016074,
       0.00000022643,    -0.00000109628,
      -0.00000008392,    -0.00000007298,
      -0.00000000443,     0.00000000540,
       0.00000077127,     0.00000000426,
       0.00000000271,    -0.00000077313,
       0.00000000053,    -0.00000000374,
      -0.00000000370,    -0.00000000055,
       0.00000017285,     0.00000001091,
      -0.00000000007,     0.00000000114,
       0.00000000428,     0.00000000022,
      -0.00000000029,     0.00000000430,
       0.00000003558,    -0.00000001622,
       0.00000000056,     0.00000000106,
       0.00000000381,    -0.00000000763,
       0.00000000155,     0.00000000120,
      -0.00000000153,    -0.00000000029,
      -0.00000000023,    -0.00000000133,
       0.00000150416,     0.00000190612,
      -0.00000003369,     0.00000000005,
      -0.00000000356,    -0.00000000307,
       0.00000083058,     0.00000000097,
      -0.00000000128,     0.00000083107,
       0.00000000045,    -0.00000000294,
       0.00000000295,     0.00000000042,
       0.00000000013,     0.00000000040,
       0.00000000004,    -0.00000000001,
       0.00000000067,     0.00000000141,
      -0.00000000140,     0.00000000069,
       0.00000000006,     0.00000000006,
      -0.00000000007,     0.00000000008,
      -0.00000000037,     0.00000000015,
      -0.00000000083,    -0.00000000025,
      -0.00000000054,     0.00000000064,
      -0.00000000071,    -0.00000000060,
      -0.00000000146,    -0.00000000059,
       0.00000000335,    -0.00000000289,
       0.00000000272,     0.00000000334,
      -0.00000288855,    -0.00000868060,
       0.00000236144,    -0.00000038114,
       0.00000000534,     0.00000033865,
      -0.00000002995,    -0.00000000532,
      -0.00000426333,     0.00000132337,
      -0.00000019303,    -0.00000112297,
       0.00000016382,    -0.00000000952,
      -0.00000000152,     0.00000001429,
       0.00000300587,    -0.00000197942,
       0.00000013191,     0.00000075107,
      -0.00000011798,    -0.00000002416,
       0.00000000643,    -0.00000001381,
       0.00000239362,     0.00000039929,
       0.00000025216,     0.00000042893,
      -0.00000005922,     0.00000008040,
      -0.00000001180,    -0.00000000783,
       0.00000021405,    -0.00000031736,
      -0.00000004416,    -0.00000003397,
      -0.00000000127,     0.00000000194,
       0.00000030706,     0.00000022984,
       0.00000003185,    -0.00000004506,
      -0.00000000192,    -0.00000000153,
       0.00002011092,     0.00004112347,
      -0.00001289900,     0.00000299312,
      -0.00000010224,    -0.00000197387,
       0.00000017118,     0.00000002953,
      -0.00000000444,     0.00000001139,
       0.00001002688,    -0.00000338571,
       0.00000053982,     0.00000274966,
      -0.00000042425,     0.00000003463,
       0.00000000340,    -0.00000003571,
       0.00000000254,     0.00000000055,
      -0.00001064184,    -0.00000871279,
       0.00000114884,    -0.00000198926,
       0.00000040465,     0.00000020878,
      -0.00000003208,     0.00000005401,
      -0.00000000511,    -0.00000000364,
       0.00000683737,     0.00000062961,
      -0.00000236653,    -0.00000016149,
       0.00000015612,    -0.00000045432,
       0.00000005684,     0.00000002994,
      -0.00000000369,     0.00000000542,
       0.00000204123,    -0.00000256001,
      -0.00000032521,    -0.00000023675,
      -0.00000001977,     0.00000002216,
       0.00000259096,     0.00000198015,
       0.00000024430,    -0.00000032142,
      -0.00000002231,    -0.00000001890,
      -0.00000002188,     0.00000000835,
       0.00000000593,     0.00000000535,
      -0.00000000067,     0.00000000070,
       0.00000014652,    -0.00000008313,
       0.00000001012,     0.00000001249,
      -0.00000000064,     0.00000000053,
       0.00000008333,     0.00000014698,
      -0.00000001269,     0.00000001013,
      -0.00000000052,    -0.00000000069,
       0.00000000258,    -0.00000000313,
       0.00000000316,     0.00000000256,
       0.00000000047,    -0.00000000152,
       0.00000010975,    -0.00000010601,
       0.00000001274,     0.00000001431,
      -0.00000000079,     0.00000000068,
       0.00000010617,     0.00000011040,
      -0.00000001455,     0.00000001277,
      -0.00000000066,    -0.00000000086,
      -0.00000000261,    -0.00000000256,
       0.00000000255,    -0.00000000260,
      -0.00000000041,    -0.00000000416,
       0.00000000032,     0.00000000273,
       0.00000000052,     0.00000000203,
      -0.00000002500,    -0.00000000940,
      -0.00000000990,     0.00000000300,
       0.00000000112,    -0.00000000032,
      -0.00000000295,     0.00000000497,
       0.00000000075,    -0.00000000036,
       0.00000000500,     0.00000000282,
      -0.00000000036,    -0.00000000075,
       0.00000000089,    -0.00000000062,
      -0.00000000013,     0.00000000111,
       0.00000000109,     0.00000000008,
      -0.00000006180,     0.00000018088,
       0.00000002919,    -0.00000001817,
      -0.00000000339,     0.00000000218,
       0.00000000068,     0.00000000058,
       0.00000000931,    -0.00000008002,
      -0.00000000986,     0.00000000844,
       0.00000000122,     0.00000000018,
      -0.00000007999,    -0.00000000932,
       0.00000000847,     0.00000000980,
       0.00000000017,    -0.00000000120,
      -0.00000000097,     0.00000000030,
       0.00000000024,     0.00000000221,
       0.00000000224,    -0.00000000023,
      -0.00000000768,     0.00000002841,
      -0.00000000039,     0.00000000170,
       0.00000000202,     0.00000000092,
       0.00000000089,    -0.00000000198,
       0.00000015604,    -0.00000236521,
      -0.00000028740,     0.00000025233,
       0.00000003473,     0.00000000965,
       0.00000003342,    -0.00000001215,
      -0.00000000541,    -0.00000000286,
       0.00000006613,     0.00000074384,
       0.00000006273,    -0.00000008843,
      -0.00000000915,     0.00000000005,
       0.00000074307,    -0.00000006641,
      -0.00000008834,    -0.00000006257,
       0.00000000007,     0.00000000914,
       0.00000003908,    -0.00000004203,
       0.00000000144,     0.00000000136,
       0.00000007193,    -0.00000004646,
       0.00000000367,    -0.00000000536,
      -0.00000000100,     0.00000000010,
      -0.00000000375,    -0.00000000119,
      -0.00000000117,     0.00000000349,
       0.00000309535,     0.00002169914,
       0.00000178742,    -0.00000262838,
      -0.00000025982,     0.00000000339,
       0.00000001009,     0.00000001491,
      -0.00000098260,     0.00000017833,
       0.00000012948,     0.00000008003,
      -0.00000000053,    -0.00000001249,
      -0.00000000080,     0.00000000058,
       0.00000019500,    -0.00000056224,
      -0.00000005647,     0.00000003235,
       0.00000000311,     0.00000000368,
      -0.00000054792,    -0.00000019281,
       0.00000003098,     0.00000005446,
       0.00000000359,    -0.00000000290,
      -0.00000000586,    -0.00000000657,
      -0.00000000657,     0.00000000586,
       0.00000000075,    -0.00000000302,
      -0.00000000301,    -0.00000000075,
       0.00000003647,    -0.00000014647,
       0.00000000483,     0.00000000120,
       0.00000001380,    -0.00000001138,
      -0.00000000885,    -0.00000001148,
       0.00000000088,     0.00000000139,
       0.00000000309,     0.00000000790,
      -0.00000000100,    -0.00000000111,
       0.00000000669,    -0.00000000196,
      -0.00000000079,     0.00000000112,
      -0.00000035670,     0.00000014406,
       0.00000002041,    -0.00000000736,
      -0.00000006331,    -0.00000009620,
       0.00000000314,     0.00000000733,
      -0.00000000876,    -0.00000002800,
       0.00000000448,     0.00000000372,
      -0.00000003461,     0.00000001163,
       0.00000000456,    -0.00000000416,
      -0.00000614219,    -0.00001003602,
      -0.00000081341,     0.00000066409,
       0.00000004534,     0.00000004979,
       0.00000134699,     0.00000030297,
      -0.00000009120,    -0.00000014130,
      -0.00000000795,     0.00000000728,
      -0.00000005332,    -0.00000002320,
      -0.00000000179,     0.00000000457,
       0.00000002428,     0.00000002401,
      -0.00000000439,    -0.00000000058,
       0.00000000111,     0.00000000142,
       0.00000000196,    -0.00000000115,
      -0.00000001984,    -0.00000001501,
       0.00000001203,     0.00000000032,
       0.00000000021,    -0.00000000863,
       0.00000086254,    -0.00000000814,
      -0.00000000185,    -0.00000008010,
      -0.00000000390,     0.00000000167,
      -0.00000000474,    -0.00000000491,
       0.00008139242,     0.00000024852,
       0.00000002524,    -0.00000002688,
      -0.00000000055,    -0.00000002040,
       0.00000001058,    -0.00000071149,
      -0.00000000026,    -0.00000000122,
       0.00000003773,    -0.00000016575,
      -0.00000000897,     0.00000001375,
       0.00000000052,     0.00000000099,
       0.00000011052,    -0.00000002946,
      -0.00000000484,    -0.00000001688,
      -0.00000000056,     0.00000000106,
       0.00000000187,    -0.00000000796,
      -0.00000000120,    -0.00000000534,
       0.00000007212,     0.00000000280,
       0.00000001209,    -0.00000000903,
       0.00000000246,     0.00000000026,
       0.00000000010,     0.00000000234,
       0.00000272541,    -0.00000191927,
       0.00000026853,     0.00000031624,
      -0.00000002439,     0.00000001619,
       0.00000008511,     0.00000003809,
      -0.00000001336,     0.00000001330,
      -0.00000000072,    -0.00000000086,
       0.00000094573,     0.00000000437,
      -0.00000000884,     0.00000094139,
      -0.00000000013,    -0.00000000103,
       0.00000000102,    -0.00000000059,
       0.00000182917,    -0.00000010665,
      -0.00000000748,     0.00000002902,
      -0.00000000068,    -0.00000001255,
       0.00000000080,    -0.00000000031,
       0.00000031659,    -0.00000005672,
       0.00000002665,    -0.00000002211,
       0.00000000042,     0.00000000074,
       0.00000000684,    -0.00000000040,
       0.00000000040,     0.00000000685,
       0.00000000010,    -0.00000000101,
      -0.00000003874,     0.00000005705,
      -0.00000000659,    -0.00000000033,
       0.00000000039,    -0.00000000652,
      -0.00000006018,     0.00000001454,
       0.00000000106,     0.00000000046,
       0.00000005594,    -0.00000000832,
       0.00000000823,    -0.00000010655,
       0.00000000372,    -0.00000000188,
      -0.00000000186,    -0.00000000373,
       0.00000000026,     0.00000000130,
       0.00000000131,    -0.00000000026,
       0.00000000627,     0.00000002442,
      -0.00000000132,     0.00000000029,
      -0.00000000829,    -0.00000001295,
       0.00000000147,     0.00000000033,
      -0.00000173997,    -0.00000009677,
       0.00000000754,     0.00000002738,
      -0.00000000069,     0.00000001231,
       0.00000000082,    -0.00000000030,
      -0.00000001145,    -0.00000000218,
      -0.00000000127,     0.00000000243,
       0.00000243746,    -0.00000367879,
      -0.00000009266,     0.00000003244,
      -0.00000001431,    -0.00000001792,
      -0.00000001551,    -0.00000000017,
      -0.00000000013,     0.00000001553,
       0.00000518509,    -0.00000388431,
       0.00000053441,     0.00000061435,
      -0.00000004737,     0.00000003182,
       0.00000010451,     0.00000017176,
      -0.00000001904,     0.00000001653,
      -0.00000000087,    -0.00000000106,
       0.00000130964,    -0.00000000061,
       0.00000000048,     0.00000131093,
       0.00000000057,    -0.00000000112,
      -0.00000000111,    -0.00000000056,
       0.00000001900,    -0.00000004617,
      -0.00000000284,     0.00000000276,
       0.00000000268,     0.00000000283,
      -0.00000008387,     0.00000010147,
      -0.00000000476,    -0.00000000444,
       0.00000000144,    -0.00000000124,
      -0.00000000157,    -0.00000000134,
      -0.00000000130,     0.00000000089,
      -0.00000004292,    -0.00000000950,
       0.00000005520,     0.00000001580,
      -0.00000000069,     0.00000000055,
       0.00000000056,     0.00000000065,
      -0.00000001954,     0.00000001986,
      -0.00000000093,    -0.00000000099,
       0.00000000065,    -0.00000000056,
      -0.00000000083,    -0.00000000044,
      -0.00000000080,    -0.00000000121,
       0.00000000010,    -0.00000000019,
      -0.00000000145,    -0.00000000422,
       0.00000000027,    -0.00000000029,
       0.00000000422,    -0.00000000144,
       0.00000000030,     0.00000000030,
      -0.00000000019,    -0.00000000014,
       0.00000000014,    -0.00000000023,
       0.00000000058,    -0.00000000039,
       0.00000000039,     0.00000000058,
       0.00000000050,    -0.00000000177,
       0.00000000177,     0.00000000050,
       0.00000000179,     0.00000000105,
      -0.00000000017,     0.00000000044,
      -0.00000001350,     0.00000000732,
      -0.00000000178,    -0.00000000136,
      -0.00000000726,    -0.00000001360,
       0.00000000135,    -0.00000000176,
      -0.00000000059,     0.00000000006,
       0.00000000014,    -0.00000000069,
       0.00000880151,     0.00000130148,
      -0.00000097946,     0.00000340387,
      -0.00000067089,    -0.00000028235,
       0.00000004919,    -0.00000008628,
       0.00000000809,     0.00000000607,
       0.00000055583,    -0.00000335451,
       0.00000129041,     0.00000039020,
      -0.00000011028,     0.00000025424,
      -0.00000003262,    -0.00000001905,
       0.00000000234,    -0.00000000304,
      -0.00000002208,     0.00000000011,
       0.00000000952,     0.00000001411,
      -0.00000000633,     0.00000000177,
      -0.00000004723,    -0.00000012943,
       0.00000004290,    -0.00000001423,
       0.00000000391,     0.00000000911,
      -0.00000000165,     0.00000000055,
       0.00000002000,     0.00000005397,
       0.00000000353,    -0.00000000121,
      -0.00000005796,     0.00000001447,
       0.00000000279,     0.00000000274,
      -0.00000000099,     0.00000000054,
      -0.00003924689,    -0.00000560014,
       0.00000433560,    -0.00001534188,
       0.00000302464,     0.00000126202,
      -0.00000022122,     0.00000038922,
      -0.00000003670,    -0.00000002718,
      -0.00000126142,     0.00000758797,
      -0.00000291547,    -0.00000088286,
       0.00000024929,    -0.00000057592,
       0.00000007434,     0.00000004325,
      -0.00000000530,     0.00000000699,
       0.00000007685,    -0.00000039775,
       0.00000000590,    -0.00000004191,
       0.00000002240,    -0.00000000353,
       0.00000000130,     0.00000000461,
       0.00000055931,     0.00000043632,
      -0.00000015195,     0.00000009031,
      -0.00000001765,    -0.00000003100,
       0.00000000529,    -0.00000000215,
      -0.00000003436,    -0.00000011134,
      -0.00000000591,    -0.00000000422,
       0.00000000027,    -0.00000000127,
       0.00000013690,    -0.00000000442,
      -0.00000000491,    -0.00000000063,
       0.00000000075,    -0.00000000115,
      -0.00000000273,    -0.00000000409,
       0.00000000044,    -0.00000000126,
      -0.00000006038,     0.00000003717,
      -0.00000001731,    -0.00000000750,
       0.00000000061,    -0.00000000241,
      -0.00000003724,    -0.00000006030,
       0.00000000747,    -0.00000001734,
       0.00000000241,     0.00000000059,
      -0.00000000042,     0.00000000111,
      -0.00000007510,     0.00000004249,
      -0.00000002217,    -0.00000000880,
       0.00000000074,    -0.00000000301,
      -0.00000004256,    -0.00000007503,
       0.00000000876,    -0.00000002221,
       0.00000000302,     0.00000000073,
      -0.00000000006,    -0.00000000077,
      -0.00000001020,     0.00000001440,
       0.00000000106,     0.00000000094,
       0.00000000018,     0.00000000501,
      -0.00000000003,    -0.00000000106,
       0.00000000500,    -0.00000000024,
      -0.00000000106,     0.00000000003,
       0.00000000633,     0.00000012334,
       0.00000000267,    -0.00000002868,
      -0.00000000378,     0.00000000240,
       0.00000000137,    -0.00000000029,
       0.00000000436,    -0.00000001650,
      -0.00000000059,     0.00000000287,
      -0.00000001642,    -0.00000000438,
       0.00000000283,     0.00000000059,
      -0.00000000672,     0.00000000727,
      -0.00000000044,     0.00000000115,
       0.00000000170,    -0.00000000024,
      -0.00000000019,    -0.00000000170,
       0.00000012201,    -0.00000047548,
      -0.00000003305,     0.00000007745,
       0.00000002682,    -0.00000000393,
      -0.00000000600,    -0.00000000011,
      -0.00000001231,     0.00000000001,
       0.00000000030,     0.00000000170,
      -0.00000000026,     0.00000001180,
       0.00000000173,    -0.00000000024,
      -0.00000000090,    -0.00000000067,
       0.00000000045,     0.00000000131,
       0.00000000131,    -0.00000000051,
       0.00000004183,    -0.00000001800,
      -0.00000000330,    -0.00000000678,
      -0.00000000316,    -0.00000000007,
      -0.00000000036,     0.00000000284,
      -0.00000089756,     0.00000094176,
       0.00000009793,    -0.00000007054,
       0.00000002198,     0.00000003149,
       0.00000000027,    -0.00000000229,
       0.00000002946,    -0.00000000499,
       0.00000000085,    -0.00000000073,
       0.00000000104,    -0.00000002937,
      -0.00000000155,    -0.00000000141,
       0.00000000069,     0.00000000064,
       0.00000000965,    -0.00000000969,
      -0.00000000170,    -0.00000000156,
       0.00000000069,     0.00000000314,
       0.00000000315,    -0.00000000085,
      -0.00000014767,     0.00000003682,
       0.00000000118,     0.00000001328,
       0.00000000060,     0.00000000048,
       0.00000000162,    -0.00000000066,
       0.00000364597,    -0.00000147889,
      -0.00000017038,     0.00000004075,
      -0.00000000692,    -0.00000005502,
      -0.00000000190,    -0.00000000111,
      -0.00000002847,     0.00000000275,
       0.00000000249,     0.00000002758,
       0.00000000835,     0.00000003369,
      -0.00000000460,    -0.00000000464,
       0.00000001516,    -0.00000000418,
      -0.00000000187,     0.00000000217,
      -0.00000001605,    -0.00000000138,
       0.00000000249,    -0.00000000278,
      -0.00000000391,     0.00000001146,
      -0.00000000268,    -0.00000000218,
       0.00000023601,    -0.00000021615,
       0.00000003008,     0.00000002898,
      -0.00000003888,     0.00000000527,
       0.00000000473,    -0.00000000548,
       0.00000005146,     0.00000001479,
      -0.00000000975,     0.00000001153,
      -0.00000000113,    -0.00000000201,
       0.00000000765,    -0.00000006749,
       0.00000001203,     0.00000001053,
      -0.00000000185,     0.00000000104,
      -0.00000000045,     0.00000000056,
      -0.00000853681,    -0.00000042659,
       0.00000001555,     0.00000003482,
       0.00000000179,     0.00000000958,
       0.00000001689,    -0.00000000431,
       0.00000000117,     0.00000000136,
       0.00000000048,    -0.00000000843,
       0.00000000113,    -0.00000000142,
      -0.00000000760,     0.00000000162,
       0.00000000105,     0.00000000336,
      -0.00000000153,    -0.00000000005,
       0.00000000024,    -0.00000000149,
       0.00000000808,     0.00000001332,
      -0.00000000320,    -0.00000000138,
       0.00000000036,    -0.00000000067,
      -0.00000007749,     0.00000420213,
      -0.00000037180,    -0.00000026385,
       0.00000002419,    -0.00000002067,
       0.00000004207,    -0.00000001816,
      -0.00000000146,     0.00000000304,
      -0.00000003519,     0.00000001230,
      -0.00000000654,    -0.00000000276,
       0.00000000023,    -0.00000000094,
       0.00000001390,     0.00000001051,
      -0.00000000284,     0.00000000653,
      -0.00000000095,    -0.00000000024,
      -0.00000167065,     0.00000094675,
      -0.00000048934,    -0.00000019560,
       0.00000001595,    -0.00000006918,
      -0.00000003962,    -0.00000007798,
       0.00000000883,    -0.00000002217,
       0.00000000306,     0.00000000074,
      -0.00000000323,     0.00000004844,
      -0.00000000487,    -0.00000000327,
      -0.00000004639,    -0.00000000048,
       0.00000000296,    -0.00000000390,
      -0.00000000367,     0.00000009669,
      -0.00000003160,     0.00000001865,
       0.00000003250,    -0.00000000026,
       0.00000001159,    -0.00000008998,
      -0.00000000050,     0.00000000090,
       0.00000000100,     0.00000000041,
       0.00000000141,     0.00000000005,
       0.00000000005,    -0.00000000142,
      -0.00000004866,    -0.00000015880,
      -0.00000002253,     0.00000001577,
       0.00000000031,    -0.00000000234,
      -0.00000000623,     0.00000000032,
       0.00000000031,     0.00000000623,
       0.00000004240,     0.00000000035,
      -0.00000206360,     0.00000014316,
       0.00000000119,    -0.00000000385,
       0.00000000254,    -0.00000000012,
       0.00000000034,    -0.00000000268,
      -0.00000316988,     0.00000180452,
      -0.00000092873,    -0.00000037177,
       0.00000003047,    -0.00000013077,
      -0.00000005697,    -0.00000009986,
       0.00000001173,    -0.00000002901,
       0.00000000394,     0.00000000097,
      -0.00000000463,     0.00000007001,
      -0.00000000673,    -0.00000000466,
      -0.00000006974,    -0.00000000542,
       0.00000000488,    -0.00000000680,
       0.00000000008,    -0.00000000003,
       0.00000000019,     0.00000000013,
      -0.00000000013,     0.00000000019,
      -0.00000000071,    -0.00000000120,
       0.00000000124,    -0.00000000063,
       0.00000042563,     0.00000001500,
      -0.00000000352,     0.00000020148,
      -0.00000004993,    -0.00000000034,
       0.00000000001,    -0.00000000844,
      -0.00000000534,    -0.00000008794,
       0.00000004338,    -0.00000000430,
       0.00000000115,     0.00000001101,
      -0.00000000191,     0.00000000023,
       0.00000029942,     0.00000003108,
      -0.00000002526,     0.00000011718,
      -0.00000002318,    -0.00000000695,
       0.00000000101,    -0.00000000304,
      -0.00000005900,     0.00000030407,
      -0.00000011755,    -0.00000004372,
       0.00000001315,    -0.00000002250,
       0.00000000270,     0.00000000233,
      -0.00000001136,    -0.00000000174,
       0.00000000058,    -0.00000000141,
       0.00000000206,    -0.00000001120,
       0.00000000134,     0.00000000068,
      -0.00000100925,    -0.00000035319,
       0.00000013267,    -0.00000046071,
       0.00000011121,     0.00000002336,
      -0.00000000248,     0.00000001844,
       0.00000001973,     0.00000019494,
      -0.00000009665,     0.00000001318,
      -0.00000000355,    -0.00000002424,
       0.00000000421,    -0.00000000049,
      -0.00000019568,     0.00000008369,
      -0.00000003290,    -0.00000008501,
       0.00000001671,    -0.00000000909,
       0.00000000208,     0.00000000241,
       0.00000009809,    -0.00000030080,
       0.00000012834,     0.00000007456,
      -0.00000002399,     0.00000002449,
      -0.00000000292,    -0.00000000484,
       0.00000002881,     0.00000000285,
      -0.00000000148,     0.00000000349,
      -0.00000000306,     0.00000002887,
      -0.00000000346,    -0.00000000143,
       0.00000000019,     0.00000000217,
       0.00000003130,     0.00000000106,
      -0.00000000253,     0.00000001127,
      -0.00000000216,    -0.00000000081,
      -0.00000000107,     0.00000003133,
      -0.00000001128,    -0.00000000255,
       0.00000000081,    -0.00000000216,
       0.00000003933,     0.00000000134,
      -0.00000000320,     0.00000001414,
      -0.00000000271,    -0.00000000101,
      -0.00000000136,     0.00000003937,
      -0.00000001415,    -0.00000000321,
       0.00000000102,    -0.00000000271,
       0.00000000210,     0.00000000140,
      -0.00000000684,     0.00000001234,
       0.00000000090,    -0.00000000173,
      -0.00000000241,     0.00000003386,
       0.00000000089,     0.00000000050,
       0.00000000271,    -0.00000000169,
      -0.00000000165,    -0.00000000272,
       0.00000000053,     0.00000000078,
       0.00000000102,    -0.00000000020,
      -0.00000000015,    -0.00000000100,
       0.00000007981,    -0.00000012751,
      -0.00000001368,     0.00000001254,
      -0.00000000383,    -0.00000000397,
      -0.00000000992,     0.00000000276,
       0.00000000022,     0.00000000098,
       0.00000000265,     0.00000000986,
       0.00000000101,    -0.00000000017,
       0.00000002485,    -0.00000001133,
      -0.00000000065,    -0.00000000307,
      -0.00000000174,     0.00000000005,
      -0.00000000019,     0.00000000147,
      -0.00000046511,     0.00000035793,
       0.00000003732,    -0.00000001365,
       0.00000000392,     0.00000001417,
       0.00000000146,    -0.00000000034,
       0.00000001976,    -0.00000000236,
       0.00000000068,    -0.00000000084,
      -0.00000000161,    -0.00000001912,
      -0.00000000080,    -0.00000000083,
      -0.00000000076,     0.00000000028,
       0.00000000349,     0.00000000917,
       0.00000000169,    -0.00000000063,
      -0.00000007704,     0.00000001818,
      -0.00000000011,     0.00000000594,
       0.00000000090,    -0.00000000051,
       0.00000169181,    -0.00000064060,
      -0.00000006140,     0.00000001237,
       0.00000000004,    -0.00000002478,
      -0.00000000132,    -0.00000000124,
      -0.00000001721,     0.00000000477,
       0.00000000293,     0.00000001403,
      -0.00000000121,    -0.00000000478,
      -0.00000000181,     0.00000000044,
       0.00000000554,     0.00000002186,
       0.00000000410,    -0.00000000106,
       0.00000012465,    -0.00000000972,
      -0.00000000073,    -0.00000000155,
      -0.00000000158,    -0.00000000004,
      -0.00000000007,    -0.00000000167,
      -0.00000342242,     0.00000011013,
      -0.00000001737,    -0.00000000711,
       0.00000000625,     0.00000000342,
       0.00000000527,    -0.00000000586,
      -0.00000000262,     0.00000000417,
      -0.00000009215,    -0.00000001509,
       0.00000001521,    -0.00000001837,
      -0.00000000587,     0.00000003474,
      -0.00000000682,    -0.00000000584,
       0.00000000113,    -0.00000000064,
      -0.00000000064,     0.00000000075,
       0.00000000041,     0.00000000157,
       0.00000041709,     0.00000008522,
      -0.00000007034,     0.00000008167,
      -0.00000000796,    -0.00000001389,
       0.00000001474,    -0.00000007833,
       0.00000001549,     0.00000001325,
      -0.00000000263,     0.00000000149,
       0.00000000410,    -0.00000000291,
       0.00000000036,     0.00000000184,
      -0.00000000165,    -0.00000000414,
       0.00000000181,    -0.00000000008,
      -0.00000001854,    -0.00000026742,
       0.00000002572,     0.00000002498,
      -0.00000000578,    -0.00000000336,
       0.00000000073,    -0.00000000081,
      -0.00000000959,     0.00000000033,
      -0.00000000077,    -0.00000000750,
      -0.00000028892,    -0.00000003857,
       0.00000004276,    -0.00000005922,
      -0.00000000013,     0.00000000222,
       0.00000001222,    -0.00000000044,
      -0.00000000086,     0.00000000433,
      -0.00000000082,    -0.00000000031,
       0.00000000125,    -0.00000001132,
       0.00000000415,     0.00000000106,
      -0.00000000031,     0.00000000082,
       0.00000085028,     0.00000003398,
      -0.00000007040,     0.00000030496,
      -0.00000005909,    -0.00000002212,
      -0.00000000136,     0.00000004056,
      -0.00000001456,    -0.00000000330,
       0.00000000103,    -0.00000000276,
      -0.00000000231,    -0.00000000043,
       0.00000000031,    -0.00000000384,
      -0.00000000385,     0.00000000794,
       0.00000000122,     0.00000000053,
      -0.00000000083,    -0.00000000133,
      -0.00000000131,     0.00000000091,
      -0.00000001324,    -0.00000004576,
      -0.00000000099,     0.00000000018,
       0.00000000017,     0.00000000099,
       0.00000001141,    -0.00000000077,
      -0.00000060746,     0.00000000620,
       0.00000000660,    -0.00000006504,
      -0.00000000060,    -0.00000000060,
       0.00000161862,     0.00000006381,
      -0.00000013366,     0.00000058054,
      -0.00000011110,    -0.00000004189,
      -0.00000000171,     0.00000005151,
      -0.00000001853,    -0.00000000418,
       0.00000000133,    -0.00000000354,
      -0.00000000425,    -0.00000000069,
       0.00000000040,    -0.00000000403,
       0.00000000006,    -0.00000000003,
       0.00000000003,     0.00000000006,
      -0.00000002252,    -0.00000008972,
       0.00000005910,    -0.00000000545,
      -0.00000000108,     0.00000001975,
      -0.00000000437,    -0.00000000089,
      -0.00000002934,     0.00000000764,
      -0.00000000200,    -0.00000001926,
       0.00000000645,     0.00000000022,
       0.00000000586,    -0.00000000388,
       0.00000000051,     0.00000000506,
      -0.00000000203,    -0.00000000048,
       0.00000000699,    -0.00000000701,
       0.00000000471,     0.00000000439,
      -0.00000000163,     0.00000000169,
       0.00000000253,    -0.00000000278,
       0.00000000056,     0.00000000079,
       0.00000000273,     0.00000000262,
      -0.00000000083,     0.00000000053,
       0.00000105159,     0.00000115891,
      -0.00000062466,     0.00000048278,
      -0.00000011088,    -0.00000017683,
       0.00000003429,    -0.00000001598,
       0.00000000853,     0.00000006090,
      -0.00000003309,     0.00000001847,
      -0.00000000908,    -0.00000001029,
       0.00000000232,    -0.00000000246,
      -0.00000004308,    -0.00000000943,
       0.00000001083,    -0.00000002784,
       0.00000000967,     0.00000000515,
      -0.00000000166,     0.00000000234,
      -0.00000000917,     0.00000001078,
      -0.00000001354,    -0.00000000904,
       0.00000000486,    -0.00000000616,
       0.00000000167,     0.00000000157,
      -0.00000000569,     0.00000000703,
      -0.00000000149,    -0.00000000184,
      -0.00000000703,    -0.00000000675,
       0.00000000227,    -0.00000000142,
      -0.00000000075,    -0.00000000102,
       0.00000000099,    -0.00000000077,
      -0.00000000095,    -0.00000000126,
       0.00000000126,    -0.00000000095,
      -0.00000000090,     0.00000000102,
       0.00000000102,     0.00000000090,
      -0.00000004982,     0.00000005640,
      -0.00000000170,    -0.00000000150,
      -0.00000000423,     0.00000001919,
       0.00000000175,    -0.00000000121,
      -0.00000000117,    -0.00000000175,
       0.00000005093,    -0.00000007180,
      -0.00000000178,    -0.00000000238,
      -0.00000000557,     0.00000000148,
       0.00000000129,     0.00000000551,
       0.00000001559,    -0.00000000680,
      -0.00000000045,    -0.00000000179,
      -0.00000000094,     0.00000000008,
      -0.00000025638,     0.00000018796,
       0.00000001888,    -0.00000000450,
       0.00000000201,     0.00000000789,
       0.00000000083,    -0.00000000020,
       0.00000001011,    -0.00000000127,
      -0.00000000063,    -0.00000000959,
      -0.00000003815,     0.00000000858,
       0.00000000003,     0.00000000295,
       0.00000081346,    -0.00000030385,
      -0.00000002557,     0.00000000424,
       0.00000000045,    -0.00000001169,
      -0.00000000066,    -0.00000000070,
      -0.00000000667,     0.00000000235,
       0.00000000103,     0.00000000439,
       0.00000005147,    -0.00000000723,
      -0.00000000060,     0.00000000064,
      -0.00000000006,    -0.00000000096,
      -0.00000138798,     0.00000007110,
       0.00000000253,    -0.00000000060,
       0.00000000168,    -0.00000000056,
       0.00000000054,     0.00000000424,
       0.00000000433,     0.00000000829,
      -0.00000008289,    -0.00000014881,
       0.00000001717,     0.00000000636,
      -0.00000000390,     0.00000000092,
       0.00000000015,    -0.00000000416,
      -0.00000000757,     0.00000000149,
       0.00000000084,     0.00000000204,
      -0.00000000287,    -0.00000000157,
       0.00000000101,    -0.00000000036,
      -0.00000000043,    -0.00000000334,
       0.00000003052,    -0.00000000823,
      -0.00000000115,     0.00000001215,
      -0.00000000199,    -0.00000000481,
       0.00000000196,    -0.00000000007,
       0.00000000172,     0.00000000447,
      -0.00000000252,    -0.00000000067,
       0.00000000353,     0.00000000466,
      -0.00000000202,     0.00000000145,
       0.00000000104,    -0.00000000030,
      -0.00000002055,    -0.00000002527,
       0.00000000965,    -0.00000001017,
       0.00000000147,    -0.00000000100,
       0.00000000046,     0.00000000060,
       0.00000000080,    -0.00000000053,
      -0.00000000742,    -0.00000001095,
      -0.00000000181,     0.00000000121,
      -0.00000016367,     0.00000000160,
       0.00000000205,    -0.00000003218,
      -0.00000003870,    -0.00000005298,
       0.00000002140,    -0.00000001860,
       0.00000000166,    -0.00000000125,
       0.00000000061,     0.00000000066,
       0.00000000047,     0.00000000250,
       0.00000000250,    -0.00000000047,
       0.00000001724,     0.00000008122,
      -0.00000000114,     0.00000000034,
      -0.00000000756,    -0.00000002264,
      -0.00000000105,     0.00000000061,
      -0.00000002261,     0.00000000755,
       0.00000000061,     0.00000000105,
      -0.00000026082,    -0.00000078063,
      -0.00000003550,     0.00000002034,
       0.00000003319,    -0.00000001108,
      -0.00000000086,    -0.00000000151,
      -0.00000000093,     0.00000000041,
       0.00000000054,     0.00000000193,
      -0.00000000072,     0.00000000018,
       0.00000000002,     0.00000000639,
      -0.00000000223,    -0.00000000006,
      -0.00000000640,    -0.00000000002,
       0.00000000007,    -0.00000000224,
       0.00000000020,     0.00000000024,
      -0.00000000008,     0.00000000006,
      -0.00000000029,     0.00000000029,
      -0.00000000007,    -0.00000000010,
      -0.00000000069,     0.00000000006,
      -0.00000000006,    -0.00000000069,
      -0.00000000188,     0.00000000168,
      -0.00000000059,    -0.00000000076,
      -0.00000000168,    -0.00000000188,
       0.00000000076,    -0.00000000059,
      -0.00000000046,    -0.00000000256,
       0.00000000115,    -0.00000000021,
       0.00000001858,     0.00000000437,
      -0.00000000198,     0.00000000694,
      -0.00000000438,     0.00000001859,
      -0.00000000694,    -0.00000000198,
       0.00000000032,     0.00000000044,
      -0.00000000044,     0.00000000030,
       0.00000004697,     0.00000001041,
      -0.00000001352,     0.00000003585,
      -0.00000001365,    -0.00000000736,
       0.00000000244,    -0.00000000341,
       0.00000000363,    -0.00000001427,
       0.00000001082,     0.00000000444,
      -0.00000000235,     0.00000000409,
      -0.00000000043,    -0.00000000290,
       0.00000000200,     0.00000000019,
       0.00000000271,    -0.00000000176,
       0.00000000102,     0.00000000190,
      -0.00000021582,    -0.00000004809,
       0.00000006234,    -0.00000016516,
       0.00000006294,     0.00000003398,
      -0.00000001134,     0.00000001575,
      -0.00000000544,     0.00000003524,
      -0.00000002640,    -0.00000000862,
       0.00000000496,    -0.00000000995,
       0.00000000246,     0.00000000169,
       0.00000003373,     0.00000003091,
      -0.00000001501,     0.00000001508,
      -0.00000000325,    -0.00000000378,
      -0.00000003015,     0.00000003903,
      -0.00000001991,    -0.00000001462,
       0.00000000373,    -0.00000000549,
      -0.00000000223,     0.00000000374,
      -0.00000000184,    -0.00000000105,
       0.00000004959,     0.00000003850,
      -0.00000001737,     0.00000002389,
      -0.00000000609,    -0.00000000403,
       0.00000000062,    -0.00000000107,
      -0.00000003851,     0.00000004960,
      -0.00000002389,    -0.00000001738,
       0.00000000403,    -0.00000000609,
       0.00000000107,     0.00000000062,
       0.00000000106,    -0.00000000008,
       0.00000006202,     0.00000004815,
      -0.00000002173,     0.00000002988,
      -0.00000000762,    -0.00000000504,
       0.00000000078,    -0.00000000134,
      -0.00000004818,     0.00000006203,
      -0.00000002989,    -0.00000002174,
       0.00000000504,    -0.00000000763,
       0.00000000135,     0.00000000078,
      -0.00000000118,     0.00000000027,
       0.00000000081,    -0.00000000021,
       0.00000024425,    -0.00000003943,
      -0.00000000034,    -0.00000000208,
      -0.00000190606,     0.00000030904,
      -0.00000000086,    -0.00000000189,
      -0.00000000093,     0.00000000017,
       0.00000000110,    -0.00000000082,
      -0.00000000078,    -0.00000000110,
       0.00000003201,    -0.00000004422,
      -0.00000000117,    -0.00000000148,
      -0.00000000321,     0.00000000099,
       0.00000000084,     0.00000000316,
       0.00000000913,    -0.00000000425,
      -0.00000000029,    -0.00000000102,
      -0.00000014297,     0.00000010590,
       0.00000000134,     0.00000000453,
       0.00000000458,    -0.00000000055,
      -0.00000000014,    -0.00000000419,
      -0.00000002016,     0.00000000555,
       0.00000000006,     0.00000000147,
       0.00000039846,    -0.00000015778,
      -0.00000000005,    -0.00000000556,
      -0.00000000218,     0.00000000211,
       0.00000000084,     0.00000000039,
       0.00000002161,    -0.00000000480,
      -0.00000057124,     0.00000006521,
       0.00000000231,    -0.00000000164,
      -0.00000000235,    -0.00000000256,
       0.00000000068,    -0.00000000105,
      -0.00000000151,     0.00000000557,
      -0.00000000106,    -0.00000000071,
      -0.00000012507,    -0.00000013542,
       0.00000002590,    -0.00000002022,
      -0.00000000262,     0.00000000467,
      -0.00000000138,    -0.00000000090,
       0.00000001085,     0.00000000471,
      -0.00000000207,     0.00000000443,
      -0.00000000080,    -0.00000000048,
       0.00000000396,    -0.00000001434,
       0.00000000437,     0.00000000200,
      -0.00000000048,     0.00000000079,
      -0.00000000509,    -0.00000000177,
      -0.00000000061,     0.00000000169,
       0.00000029927,     0.00000009714,
      -0.00000004761,     0.00000010393,
      -0.00000001848,    -0.00000001117,
       0.00000000650,    -0.00000001823,
       0.00000000633,     0.00000000295,
      -0.00000000068,     0.00000000113,
      -0.00000002826,    -0.00000000390,
       0.00000000302,    -0.00000001090,
       0.00000000214,     0.00000000092,
      -0.00000000310,     0.00000002713,
      -0.00000001059,    -0.00000000280,
       0.00000000088,    -0.00000000209,
       0.00000001620,    -0.00000000567,
       0.00000000207,     0.00000000773,
      -0.00000000186,    -0.00000000501,
       0.00000000239,    -0.00000000071,
       0.00000000084,     0.00000000102,
       0.00000000069,    -0.00000000101,
       0.00000000056,     0.00000000072,
      -0.00000001766,     0.00000003564,
      -0.00000001773,    -0.00000001367,
       0.00000000644,     0.00000000055,
      -0.00000000138,     0.00000000308,
      -0.00000000372,    -0.00000000395,
       0.00000000306,    -0.00000000193,
      -0.00000000366,     0.00000000439,
      -0.00000000220,    -0.00000000306,
      -0.00000063199,    -0.00000008395,
       0.00000006776,    -0.00000024503,
       0.00000004814,     0.00000002070,
      -0.00000000613,     0.00000004572,
      -0.00000001775,    -0.00000000494,
       0.00000000149,    -0.00000000347,
      -0.00000000459,    -0.00000000206,
       0.00000000107,    -0.00000000213,
      -0.00000000357,     0.00000000869,
      -0.00000000355,    -0.00000000174,
       0.00000000043,    -0.00000000076,
       0.00000000027,    -0.00000000080,
       0.00000000198,     0.00000000029,
      -0.00000000028,     0.00000000195,
      -0.00000003506,    -0.00000003962,
       0.00000001822,    -0.00000001686,
      -0.00000000104,     0.00000000562,
      -0.00000000231,    -0.00000000064,
       0.00000001282,     0.00000001325,
      -0.00000000574,     0.00000000665,
      -0.00000000182,    -0.00000000129,
       0.00000001461,    -0.00000002278,
       0.00000001052,     0.00000000682,
      -0.00000000162,     0.00000000258,
       0.00000132874,     0.00000103330,
      -0.00000046618,     0.00000064014,
      -0.00000016333,    -0.00000010804,
       0.00000001676,    -0.00000002881,
      -0.00000005006,     0.00000006444,
      -0.00000003105,    -0.00000002259,
       0.00000000524,    -0.00000000793,
       0.00000000140,     0.00000000082,
       0.00000000182,     0.00000000083,
      -0.00000000040,     0.00000000085,
       0.00000000074,    -0.00000000080,
      -0.00000004560,     0.00000000057,
      -0.00000000348,    -0.00000000050,
       0.00000000038,    -0.00000000135,
      -0.00000000048,     0.00000000348,
      -0.00000000135,    -0.00000000038,
      -0.00000011531,    -0.00000001918,
       0.00000001410,    -0.00000004450,
      -0.00000000080,     0.00000000552,
      -0.00000000214,    -0.00000000061,
       0.00000255364,     0.00000198392,
      -0.00000089518,     0.00000123025,
      -0.00000031390,    -0.00000020745,
       0.00000003220,    -0.00000005539,
      -0.00000006346,     0.00000008165,
      -0.00000003933,    -0.00000002863,
       0.00000000663,    -0.00000001004,
       0.00000000177,     0.00000000103,
       0.00000000213,     0.00000000050,
      -0.00000000024,     0.00000000090,
      -0.00000000121,     0.00000000217,
      -0.00000000094,    -0.00000000058,
       0.00000000485,     0.00000000036,
       0.00000000001,    -0.00000000112,
       0.00000000199,     0.00000000033,
      -0.00000000044,     0.00000000153,
      -0.00000000056,     0.00000000196,
      -0.00000000147,    -0.00000000068,
      -0.00000001656,    -0.00000000234,
       0.00000000283,    -0.00000001412,
      -0.00000000001,     0.00000000262,
      -0.00000000232,    -0.00000000012,
      -0.00000000232,    -0.00000000005,
       0.00000000018,    -0.00000000188,
       0.00000000106,    -0.00000000222,
       0.00000000164,     0.00000000123,
      -0.00000000050,     0.00000000126,
      -0.00000000052,     0.00000000121,
       0.00000000020,    -0.00000000136,
      -0.00000000112,     0.00000000003,
      -0.00000000003,    -0.00000000107,
      -0.00000000038,     0.00000000255,
      -0.00000000258,     0.00000000004,
      -0.00000000001,    -0.00000000249,
      -0.00000000007,     0.00000000215,
      -0.00000000122,    -0.00000000002,
      -0.00000001252,     0.00000001072,
      -0.00000000010,    -0.00000000261,
      -0.00000000214,     0.00000000051,
       0.00000000024,     0.00000000215,
       0.00000003019,    -0.00000000041,
      -0.00000000109,     0.00000000002,
      -0.00000000001,    -0.00000000109,
       0.00000000067,    -0.00000000053,
       0.00000001174,    -0.00000001681,
      -0.00000000100,     0.00000000039,
       0.00000000032,     0.00000000097,
      -0.00000002352,     0.00000001914,
       0.00000002328,    -0.00000001221,
       0.00000000030,    -0.00000000114,
       0.00000000080,     0.00000000126,
       0.00000005058,    -0.00000000105,
      -0.00000003651,     0.00000000068,
       0.00000002287,     0.00000000341,
       0.00000001953,    -0.00000002741,
      -0.00000000075,    -0.00000000090,
      -0.00000000181,     0.00000000064,
       0.00000000053,     0.00000000177,
      -0.00000004353,     0.00000003392,
       0.00000000049,     0.00000000137,
       0.00000000118,    -0.00000000028,
       0.00000004791,    -0.00000002284,
      -0.00000001998,     0.00000000772,
      -0.00000000066,     0.00000000119,
      -0.00000000016,     0.00000000143,
      -0.00000002347,    -0.00000000161,
       0.00000008626,    -0.00000000146,
      -0.00000000002,    -0.00000000132,
      -0.00000006120,     0.00000000084,
       0.00000003639,     0.00000000539,
      -0.00000003327,    -0.00000000450,
      -0.00000007903,     0.00000006001,
       0.00000000082,     0.00000000251,
       0.00000000233,    -0.00000000041,
      -0.00000000013,    -0.00000000208,
       0.00000009770,    -0.00000004284,
      -0.00000000019,    -0.00000000148,
      -0.00000004672,     0.00000001333,
      -0.00000003733,    -0.00000000264,
       0.00000014801,    -0.00000000196,
      -0.00000000003,    -0.00000000205,
      -0.00000010268,     0.00000000105,
       0.00000000001,     0.00000000099,
       0.00000005809,     0.00000000842,
       0.00000000127,     0.00000000002,
      -0.00000005210,    -0.00000000691,
      -0.00000000014,     0.00000000110,
       0.00000000024,    -0.00000000100,
       0.00000000126,     0.00000000017,
      -0.00000000095,    -0.00000000011,
       0.00000000011,    -0.00000000095,
       0.00000000248,     0.00000000011,
       0.00000000017,    -0.00000000252,
      -0.00000001017,     0.00000000314,
       0.00000019620,    -0.00000008205,
      -0.00000000022,    -0.00000000306,
      -0.00000000080,     0.00000000093,
      -0.00000010845,     0.00000002225,
      -0.00000000882,    -0.00000001549,
      -0.00000000092,     0.00000000032,
      -0.00000005972,    -0.00000000396,
       0.00000025704,    -0.00000000274,
      -0.00000000003,    -0.00000000343,
      -0.00000017369,     0.00000000132,
       0.00000000001,     0.00000000169,
       0.00000009136,     0.00000001234,
       0.00000000015,    -0.00000000158,
       0.00000000254,     0.00000000016,
       0.00000000016,     0.00000000146,
      -0.00000008196,    -0.00000001059,
      -0.00000000019,     0.00000000159,
      -0.00000000158,     0.00000000000,
       0.00000000000,    -0.00000000157,
       0.00000002003,     0.00000000488,
       0.00000000027,    -0.00000000122,
       0.00000000184,     0.00000000026,
      -0.00000000012,     0.00000000140,
      -0.00000000132,    -0.00000000015,
       0.00000000015,    -0.00000000131,
       0.00000000470,    -0.00000007101,
      -0.00000000880,    -0.00000000240,
      -0.00000000930,    -0.00000000130,
      -0.00000000130,     0.00000000930,
       0.00000000824,     0.00000000328,
       0.00000001761,    -0.00000004425,
      -0.00000079280,    -0.00000004775,
      -0.00000004551,     0.00000013288,
      -0.00000004156,     0.00000075565,
       0.00000012435,     0.00000004259,
       0.00000026146,     0.00000000176,
       0.00000000659,    -0.00000003031,
       0.00000000146,    -0.00000026024,
      -0.00000003009,    -0.00000000673,
       0.00000000900,    -0.00000000290,
      -0.00000024477,     0.00000003811,
       0.00000000099,    -0.00000000101,
      -0.00000000004,     0.00000000197,
      -0.00000001550,    -0.00000002923,
       0.00000000034,     0.00000000095,
      -0.00000000278,    -0.00000000159,
      -0.00000000159,     0.00000000282,
      -0.00000009469,    -0.00000000706,
       0.00000000177,    -0.00000000040,
      -0.00000000043,    -0.00000000163,
       0.00000002998,     0.00000000312,
       0.00000000011,    -0.00000000155,
       0.00000043963,    -0.00000001144,
      -0.00000000005,    -0.00000000643,
       0.00000000395,     0.00000000100,
       0.00000000101,    -0.00000000557,
      -0.00000029570,     0.00000000136,
       0.00000000000,     0.00000000297,
       0.00000014967,     0.00000001925,
       0.00000000023,    -0.00000000236,
       0.00000000483,     0.00000000034,
       0.00000000032,     0.00000000215,
      -0.00000013063,    -0.00000001628,
      -0.00000000012,     0.00000000126,
      -0.00000000273,     0.00000000000,
       0.00000000000,    -0.00000000271,
       0.00000005174,     0.00000001339,
       0.00000000030,    -0.00000000139,
      -0.00000000058,     0.00000000355,
      -0.00000003351,    -0.00000000763,
      -0.00000000187,    -0.00000000020,
       0.00000000020,    -0.00000000186,
      -0.00000000009,     0.00000000478,
       0.00000000645,     0.00000000075,
       0.00000000076,    -0.00000000650,
       0.00000000011,    -0.00000000111,
       0.00000048910,    -0.00000002219,
       0.00000001711,    -0.00000005809,
       0.00000000010,    -0.00000005806,
      -0.00000000734,    -0.00000000135,
      -0.00000014500,     0.00000001899,
      -0.00000000366,     0.00000001508,
       0.00000000106,     0.00000000051,
       0.00000001885,     0.00000014509,
       0.00000001526,     0.00000000373,
       0.00000000051,    -0.00000000106,
       0.00000000053,     0.00000000083,
       0.00000000083,    -0.00000000053,
      -0.00000003886,    -0.00000005698,
      -0.00000000035,     0.00000000096,
       0.00000002450,     0.00000001556,
       0.00000000047,     0.00000000077,
      -0.00000000683,    -0.00000000199,
      -0.00000000197,     0.00000000682,
      -0.00000030465,    -0.00000006786,
      -0.00000000067,    -0.00000000252,
       0.00000004140,     0.00000000579,
       0.00000000572,    -0.00000004136,
       0.00000003719,     0.00000000501,
       0.00000000014,    -0.00000000152,
       0.00000080837,    -0.00000001906,
      -0.00000000008,    -0.00000001104,
       0.00000000963,     0.00000000155,
       0.00000000156,    -0.00000001147,
      -0.00000050816,     0.00000000156,
       0.00000000000,     0.00000000525,
       0.00000043699,     0.00000007399,
       0.00000000023,    -0.00000000255,
      -0.00000000787,    -0.00000000252,
      -0.00000000250,     0.00000002027,
      -0.00000021132,    -0.00000002533,
      -0.00000000019,     0.00000000205,
      -0.00000000481,    -0.00000000002,
       0.00000000000,    -0.00000000473,
       0.00000010588,     0.00000002361,
       0.00000000028,    -0.00000000148,
       0.00000000126,    -0.00000000004,
      -0.00000000055,     0.00000000350,
      -0.00000005342,    -0.00000001143,
      -0.00000000020,     0.00000000111,
      -0.00000000274,    -0.00000000028,
       0.00000000027,    -0.00000000270,
      -0.00000006678,    -0.00000001699,
       0.00000000034,    -0.00000000130,
      -0.00000000104,    -0.00000000043,
      -0.00000000073,     0.00000000265,
       0.00000001037,    -0.00000000055,
      -0.00000000062,    -0.00000001455,
      -0.00000000146,    -0.00000000035,
      -0.00000002846,    -0.00000000259,
      -0.00000000037,     0.00000000199,
      -0.00000000264,     0.00000002842,
       0.00000000208,     0.00000000039,
       0.00000000044,    -0.00000000103,
      -0.00000000087,     0.00000000164,
       0.00000000185,     0.00000000097,
      -0.00000138541,     0.00000050442,
      -0.00000004313,     0.00000015347,
       0.00000001024,     0.00000001030,
       0.00000001186,     0.00000017274,
       0.00000001790,     0.00000000406,
       0.00000000065,    -0.00000000125,
       0.00000026364,    -0.00000020176,
       0.00000001185,    -0.00000003181,
      -0.00000000258,    -0.00000000240,
      -0.00000020146,    -0.00000026139,
      -0.00000003141,    -0.00000001178,
      -0.00000000240,     0.00000000255,
      -0.00000000189,    -0.00000000289,
      -0.00000000284,     0.00000000185,
      -0.00000000143,    -0.00000000096,
      -0.00000000096,     0.00000000141,
      -0.00000052344,    -0.00000008321,
      -0.00000000090,    -0.00000000201,
       0.00000007017,     0.00000000292,
       0.00000000282,    -0.00000007002,
       0.00000004131,     0.00000000843,
       0.00000000019,    -0.00000000139,
       0.00000000928,     0.00000000059,
       0.00000000050,     0.00000000067,
      -0.00000000117,    -0.00000000032,
      -0.00000000021,     0.00000000121,
       0.00000326597,     0.00000025121,
       0.00000000009,    -0.00000000922,
      -0.00000016355,    -0.00000001183,
      -0.00000000046,    -0.00000000270,
      -0.00000001157,     0.00000016097,
      -0.00000000276,     0.00000000046,
      -0.00000088723,     0.00000000175,
       0.00000000000,     0.00000000934,
       0.00000083652,     0.00000011129,
       0.00000000007,    -0.00000000269,
       0.00000000123,    -0.00000000193,
      -0.00000000189,     0.00000002139,
      -0.00000035424,    -0.00000003999,
      -0.00000000030,     0.00000000346,
      -0.00000000860,    -0.00000000004,
      -0.00000000001,    -0.00000000833,
      -0.00000075188,    -0.00000013212,
       0.00000000584,    -0.00000002320,
       0.00000000214,    -0.00000001187,
      -0.00000002474,    -0.00000000704,
       0.00000000057,    -0.00000000118,
      -0.00000000812,     0.00000003418,
      -0.00000000133,    -0.00000000058,
      -0.00000009393,    -0.00000001858,
      -0.00000000027,     0.00000000161,
      -0.00000000321,    -0.00000000033,
       0.00000000026,    -0.00000000302,
      -0.00000003904,    -0.00000000914,
       0.00000000047,    -0.00000000199,
       0.00000000138,     0.00000000024,
      -0.00000000027,     0.00000000153,
      -0.00000002531,    -0.00000000714,
       0.00000000027,    -0.00000000136,
      -0.00000007782,    -0.00000002377,
       0.00000000093,    -0.00000000300,
      -0.00000006143,     0.00000000571,
      -0.00000000130,     0.00000000496,
       0.00000000633,     0.00000006592,
       0.00000000535,     0.00000000138,
       0.00000014930,     0.00000000943,
       0.00000000104,    -0.00000000704,
       0.00000000964,    -0.00000014907,
      -0.00000000701,    -0.00000000105,
      -0.00000000066,    -0.00000000091,
      -0.00000000083,     0.00000000060,
      -0.00000001290,    -0.00000000657,
       0.00000000063,     0.00000000302,
       0.00000000593,    -0.00000000364,
      -0.00000000059,     0.00000000069,
      -0.00000000370,    -0.00000000627,
      -0.00000325666,    -0.00000617591,
       0.00000019772,    -0.00000011360,
      -0.00000003284,    -0.00000004488,
      -0.00000009873,    -0.00000074278,
      -0.00000006250,    -0.00000001457,
      -0.00000000246,     0.00000000378,
       0.00000335148,     0.00000184777,
      -0.00000007464,    -0.00000004084,
       0.00000000616,     0.00000001056,
       0.00000000055,    -0.00000000048,
       0.00000184483,    -0.00000336027,
      -0.00000004186,     0.00000007445,
       0.00000001050,    -0.00000000609,
      -0.00000000051,    -0.00000000057,
       0.00000000748,     0.00000001031,
       0.00000001006,    -0.00000000730,
      -0.00000008189,    -0.00000002897,
      -0.00000000034,    -0.00000000156,
       0.00000002591,     0.00000000421,
       0.00000000419,    -0.00000002589,
       0.00000561117,     0.00000009245,
      -0.00000000038,    -0.00000001318,
      -0.00000009835,    -0.00000000848,
       0.00000000032,    -0.00000000154,
      -0.00000000816,     0.00000009499,
      -0.00000000160,    -0.00000000032,
      -0.00000160205,     0.00000000293,
       0.00000000000,     0.00000001725,
      -0.00000000005,     0.00000000144,
      -0.00000000140,    -0.00000000032,
      -0.00000000029,     0.00000000136,
      -0.00000899697,    -0.00000074015,
      -0.00000000233,    -0.00000014892,
       0.00000001119,    -0.00000012233,
       0.00000000240,     0.00000000029,
      -0.00000057403,    -0.00000013030,
       0.00000000642,    -0.00000001765,
      -0.00000013203,     0.00000062793,
      -0.00000001767,    -0.00000000648,
      -0.00000064702,    -0.00000006712,
      -0.00000000045,     0.00000000588,
      -0.00000001625,    -0.00000000024,
      -0.00000000018,    -0.00000001489,
      -0.00000033489,    -0.00000005615,
       0.00000000236,    -0.00000001449,
       0.00000000883,     0.00000000085,
      -0.00000000100,     0.00000000999,
      -0.00000020166,    -0.00000004146,
      -0.00000000033,     0.00000000236,
      -0.00000000154,     0.00000000040,
       0.00000000133,    -0.00000000966,
      -0.00000122047,    -0.00000029551,
       0.00000002354,    -0.00000005258,
       0.00000001213,    -0.00000004969,
       0.00000000206,     0.00000000094,
      -0.00000000516,    -0.00000000090,
       0.00000000066,    -0.00000000303,
       0.00000023680,     0.00000005165,
      -0.00000000081,     0.00000000369,
       0.00000000089,    -0.00000000409,
       0.00000013575,     0.00000003856,
      -0.00000000102,     0.00000000353,
       0.00000035012,    -0.00000007234,
       0.00000000611,    -0.00000001902,
      -0.00000006915,    -0.00000031081,
      -0.00000001639,    -0.00000000513,
      -0.00000000066,     0.00000000101,
      -0.00000075256,    -0.00000002622,
      -0.00000000192,     0.00000001760,
      -0.00000002508,     0.00000075220,
       0.00000001745,     0.00000000191,
       0.00000000187,     0.00000000236,
       0.00000000198,    -0.00000000145,
      -0.00000021916,    -0.00000196369,
      -0.00000017099,     0.00000023620,
       0.00000002445,     0.00000000088,
       0.00000017986,    -0.00000004547,
      -0.00000002571,    -0.00000001343,
       0.00000000015,     0.00000000271,
      -0.00000010268,     0.00000005573,
       0.00000000975,    -0.00000000375,
       0.00000004988,     0.00000009851,
      -0.00000000340,    -0.00000000704,
       0.00019067135,     0.00006490653,
      -0.00000188050,    -0.00000518837,
       0.00000007908,     0.00000028874,
       0.00000001364,    -0.00000001256,
       0.00000047772,     0.00000385739,
       0.00000023722,     0.00000006963,
       0.00000000798,    -0.00000001237,
      -0.00000000057,    -0.00000000047,
      -0.00008161648,    -0.00001250408,
       0.00000051966,     0.00000112911,
      -0.00000001551,    -0.00000004745,
      -0.00000000166,     0.00000000162,
      -0.00001247361,     0.00008163282,
       0.00000113234,    -0.00000051859,
      -0.00000004711,     0.00000001521,
       0.00000000160,     0.00000000167,
      -0.00000002466,    -0.00000002518,
      -0.00000000096,     0.00000000037,
      -0.00000002367,     0.00000002353,
       0.00000000028,     0.00000000080,
       0.00000006047,     0.00000000090,
      -0.00000000126,    -0.00000000015,
      -0.00000000015,     0.00000000127,
      -0.00000019322,    -0.00000002567,
      -0.00000000018,     0.00000000175,
      -0.00000020545,    -0.00000002728,
      -0.00000000019,     0.00000000188,
       0.00000000000,    -0.00000000005,
       0.00000000005,     0.00000000000,
       0.00000000838,     0.00000000750,
      -0.00000000608,     0.00000000429,
       0.00000000231,    -0.00000000250,
       0.00000000126,     0.00000000187,
      -0.00000003852,    -0.00000003454,
       0.00000002808,    -0.00000001970,
       0.00000000396,     0.00000000933,
      -0.00000000536,     0.00000000576,
      -0.00000000294,    -0.00000000430,
       0.00000000142,    -0.00000000060,
      -0.00000000094,     0.00000000015,
      -0.00000001868,    -0.00000001625,
       0.00000000052,    -0.00000000060,
      -0.00000000034,     0.00000000077,
      -0.00000000064,    -0.00000000037,
       0.00000000063,    -0.00000000048,
       0.00000000406,     0.00000000088,
       0.00000000088,    -0.00000000404,
       0.00000108447,     0.00000014222,
      -0.00000000598,    -0.00000001631,
       0.00000000018,     0.00000000824,
      -0.00000019861,    -0.00000000123,
      -0.00000000112,     0.00000019841,
      -0.00000002667,     0.00000002489,
       0.00000000015,     0.00000000098,
      -0.00000000356,     0.00000000006,
       0.00000000009,     0.00000000303,
      -0.00000298337,    -0.00000022981,
       0.00000000981,    -0.00000004705,
       0.00000000870,    -0.00000012114,
       0.00000000231,     0.00000000046,
       0.00000005035,     0.00000000039,
      -0.00000000041,     0.00000005974,
      -0.00000143702,    -0.00000018363,
      -0.00000000043,     0.00000000974,
       0.00000001033,     0.00000000543,
      -0.00000000037,     0.00000000126,
       0.00000000554,    -0.00000007184,
       0.00000000126,     0.00000000038,
      -0.00002341903,    -0.00000439715,
       0.00000024613,    -0.00000069329,
       0.00000018310,    -0.00000096757,
       0.00000002799,     0.00000001002,
      -0.00000007828,    -0.00000000686,
       0.00000000037,    -0.00000000187,
       0.00000000411,    -0.00000003664,
       0.00000000134,    -0.00000000122,
      -0.00000000098,    -0.00000000102,
       0.00000254397,     0.00000035998,
      -0.00000002347,     0.00000007742,
      -0.00000000485,     0.00000003387,
      -0.00000000110,    -0.00000000034,
       0.00000002297,     0.00000000589,
      -0.00000000060,     0.00000000136,
       0.00000000782,    -0.00000004702,
       0.00000000178,     0.00000000069,
       0.00000000151,     0.00000000009,
      -0.00000000009,     0.00000000152,
       0.00000201260,     0.00000042749,
      -0.00000003736,     0.00000008809,
      -0.00000001199,     0.00000005589,
      -0.00000000237,    -0.00000000101,
       0.00000000637,     0.00000000097,
      -0.00000000132,     0.00000000661,
      -0.00000000174,    -0.00000000022,
       0.00000000023,    -0.00000000174,
       0.00000000159,     0.00000000021,
      -0.00000000021,     0.00000000159,
      -0.00000000001,    -0.00000000008,
      -0.00000000005,    -0.00000000075,
      -0.00000157388,     0.00000032700,
      -0.00000003265,     0.00000006694,
       0.00000000317,     0.00000000263,
       0.00000024066,     0.00000123656,
       0.00000005124,     0.00000002443,
       0.00000000198,    -0.00000000240,
       0.00000281947,     0.00000000030,
       0.00000000660,    -0.00000282064,
       0.00000000050,     0.00000000054,
      -0.00000000079,    -0.00000000152,
       0.00000000029,    -0.00000000119,
      -0.00000028197,     0.00000065245,
       0.00000006021,    -0.00000003431,
      -0.00000004773,     0.00000001739,
       0.00000000496,     0.00000000618,
      -0.00000003019,     0.00000000084,
      -0.00000000175,     0.00000003066,
      -0.00402884328,    -0.00053443547,
       0.00001983726,     0.00006213859,
      -0.00000028677,    -0.00000188448,
      -0.00000006080,     0.00000005337,
       0.00000299329,    -0.00005851850,
      -0.00000129089,    -0.00000008168,
      -0.00000000321,     0.00000002543,
       0.00000000077,     0.00000000035,
       0.00116379909,     0.00000012915,
      -0.00000001991,    -0.00000001801,
      -0.00000000152,     0.00000015376,
      -0.00000025193,    -0.00116288286,
       0.00000000092,     0.00000000816,
       0.00000015183,     0.00000000241,
       0.00000011760,    -0.00000009293,
       0.00000000233,     0.00000000281,
      -0.00000010106,    -0.00000011111,
       0.00000000309,    -0.00000000176,
      -0.00000148181,    -0.00000008742,
       0.00000000035,     0.00000000167,
      -0.00000001724,    -0.00000000782,
      -0.00000000778,     0.00000001739,
       0.00000018326,    -0.00000000007,
       0.00000000000,    -0.00000000217,
       0.00000000803,     0.00000003604,
       0.00000203793,     0.00000000009,
       0.00000000000,    -0.00000002386,
       0.00000216261,     0.00000000011,
       0.00000000000,    -0.00000002524,
      -0.00000000775,     0.00000003374,
       0.00000018652,    -0.00000000001,
       0.00000000000,    -0.00000000220,
       0.00000001470,    -0.00000000653,
      -0.00000000027,    -0.00000000128,
       0.00000000585,     0.00000000257,
       0.00000000256,    -0.00000000585,
       0.00000015441,     0.00000002354,
      -0.00000004063,    -0.00000000226,
      -0.00000000224,     0.00000004057,
      -0.00000000476,    -0.00000000001,
      -0.00000000001,     0.00000000477,
      -0.00000665209,    -0.00000003193,
       0.00000000110,    -0.00000022055,
       0.00000001095,     0.00000000220,
       0.00000000088,     0.00000000040,
       0.00000000086,    -0.00000000072,
      -0.00000000062,    -0.00000000110,
      -0.00000716438,    -0.00000048348,
       0.00000000116,     0.00000005382,
       0.00000041646,     0.00000002038,
      -0.00000000111,     0.00000000685,
       0.00000002160,    -0.00000042917,
       0.00000000661,     0.00000000118,
      -0.00000006776,    -0.00000000656,
       0.00000000870,    -0.00000000025,
       0.00000000023,     0.00000000875,
       0.00003507387,     0.00000504568,
      -0.00000035068,     0.00000104994,
      -0.00000014499,     0.00000099788,
      -0.00000002934,    -0.00000000985,
       0.00000008475,     0.00000000386,
      -0.00000000015,     0.00000000127,
      -0.00000000998,     0.00000008949,
      -0.00000000145,    -0.00000000038,
       0.00000000279,     0.00000000024,
       0.00000000015,    -0.00000000210,
      -0.00000002102,    -0.00000000132,
       0.00000000134,    -0.00000002103,
      -0.00000000093,     0.00000000093,
      -0.00000000091,    -0.00000000089,
       0.00000003551,     0.00000000739,
       0.00000000041,    -0.00000000170,
       0.00000001882,     0.00000000122,
      -0.00000000123,     0.00000001887,
      -0.00000000258,    -0.00000000007,
       0.00000000007,    -0.00000000257,
       0.00000000057,     0.00000000085,
      -0.00000000085,     0.00000000056,
      -0.00000000156,    -0.00000000027,
       0.00000000027,    -0.00000000156,
       0.00000000000,     0.00000000014,
       0.00000000000,     0.00000000066,
       0.00000000000,     0.00000000087,
      -0.00000000001,     0.00000000118,
       0.00000112354,     0.00000001766,
       0.00000001212,    -0.00000104959,
       0.00000000095,    -0.00000000022,
       0.00000009448,    -0.00000000725,
      -0.00000000043,    -0.00000000093,
      -0.00000001200,     0.00000000660,
       0.00000000135,     0.00000000061,
       0.00000000015,     0.00000000318,
      -0.00000000035,     0.00000000254,
       0.00000041973,     0.00000001001,
      -0.00000000061,     0.00000004604,
      -0.00000000153,    -0.00000000101,
       0.08030476237,     0.00000227945,
      -0.00000528673,    -0.00000004972,
       0.00000001768,     0.00001126661,
      -0.00000019098,     0.00341045476,
      -0.00000001337,    -0.00000024232,
      -0.00000047835,     0.00000000104,
       0.00004902137,    -0.00001179678,
      -0.00000016584,     0.00000156940,
      -0.00000000015,    -0.00000001866,
       0.00000000016,     0.00000000089,
      -0.00000561309,    -0.00009620973,
       0.00000083426,     0.00000039151,
       0.00000000329,    -0.00000000265,
       0.00000000028,    -0.00000000087,
      -0.00000409462,     0.00000486953,
      -0.00000019661,    -0.00000012136,
      -0.00000001591,     0.00000001778,
       0.00000000075,     0.00000000044,
       0.00000490329,     0.00000406461,
      -0.00000012191,     0.00000019540,
       0.00000001790,     0.00000001580,
       0.00000000044,    -0.00000000075,
       0.00013196740,    -0.00001107697,
       0.00000011569,     0.00000006286,
       0.00000009574,     0.00000104974,
      -0.00000000118,     0.00000000062,
       0.00001337101,    -0.00000076725,
      -0.00000006130,     0.00000020997,
       0.00000000076,    -0.00000000197,
      -0.00000076253,    -0.00001337864,
       0.00000020981,     0.00000006133,
      -0.00000000197,    -0.00000000075,
       0.00000000290,    -0.00000000034,
      -0.00000000030,    -0.00000000298,
       0.00000000221,    -0.00000000001,
      -0.00000000001,     0.00000000219,
      -0.00000000025,    -0.00000000226,
       0.00000002484,    -0.00000000001,
       0.00000000000,     0.00000002479,
       0.00000004690,    -0.00000000312,
      -0.00000000025,    -0.00000000239,
       0.00000002630,     0.00000000000,
      -0.00000000001,     0.00000002622,
       0.00000004973,    -0.00000000326,
       0.00000000197,     0.00000000000,
       0.00000000000,     0.00000000196,
       0.00000022765,     0.00000005161,
       0.00000000102,     0.00000000460,
      -0.00000009529,    -0.00000000433,
       0.00000000121,     0.00000000427,
      -0.00000000428,     0.00000009533,
       0.00000000424,    -0.00000000122,
       0.00000002914,     0.00000000719,
      -0.00000001174,    -0.00000000097,
      -0.00000000096,     0.00000001173,
       0.00000000141,    -0.00000000016,
      -0.00000000016,    -0.00000000141,
      -0.00000136475,    -0.00000005245,
       0.00000000555,     0.00000002172,
       0.00000000021,    -0.00000003002,
       0.00000018810,    -0.00000000029,
      -0.00000000044,    -0.00000018703,
       0.00000125278,    -0.00000189390,
       0.00000000208,    -0.00000000011,
       0.00000152384,    -0.00000022217,
       0.00000000261,     0.00000000100,
      -0.00000022205,    -0.00000152439,
       0.00000000098,    -0.00000000260,
       0.00000035144,     0.00000001364,
      -0.00000000118,     0.00000002817,
       0.00078267799,    -0.00004400186,
      -0.00000357265,     0.00001228523,
       0.00000004304,    -0.00000011470,
       0.00000128132,     0.00002278238,
      -0.00000035738,    -0.00000010431,
       0.00000000334,     0.00000000126,
       0.00000040996,    -0.00000007918,
       0.00000000133,    -0.00000000123,
      -0.00000003876,     0.00000053106,
      -0.00000000150,    -0.00000000100,
       0.00000005395,    -0.00000000580,
      -0.00000000229,    -0.00000000070,
      -0.00000000583,    -0.00000005380,
      -0.00000000070,     0.00000000229,
       0.00000021095,     0.00000008106,
       0.00000000399,    -0.00000000708,
       0.00000001780,     0.00000001746,
      -0.00000001744,     0.00000001780,
      -0.00000004184,     0.00000000351,
       0.00000000191,     0.00000000061,
      -0.00000000352,    -0.00000004184,
      -0.00000000061,     0.00000000191,
       0.00000000578,     0.00000000003,
      -0.00000000003,     0.00000000580,
      -0.00000000141,     0.00000000114,
      -0.00000000114,    -0.00000000141,
       0.00000000331,     0.00000000139,
      -0.00000000139,     0.00000000331,
       0.00000002092,     0.00000000350,
       0.00000000165,    -0.00000001427,
       0.00000000088,    -0.00000000027,
       0.00000004170,    -0.00000000085,
       0.00000000013,     0.00000000101,
       0.00000000086,     0.00000004185,
      -0.00000000102,     0.00000000014,
      -0.00000000105,     0.00000000131,
      -0.00000000128,    -0.00000000102,
      -0.00000000089,     0.00000000040,
       0.00000000259,     0.00000000031,
      -0.00003018966,    -0.00000217794,
      -0.00000051754,    -0.00000222651,
       0.00000012964,    -0.00000006653,
      -0.00000024665,     0.00000169720,
      -0.00000013844,     0.00000003970,
      -0.00000000489,    -0.00000000894,
       0.00001764172,    -0.00000230167,
      -0.00000007777,     0.00000028014,
       0.00000000129,    -0.00000000245,
       0.00000231158,     0.00001765222,
      -0.00000028002,    -0.00000007834,
       0.00000000252,     0.00000000136,
      -0.00000307052,     0.00000367741,
      -0.00000014758,    -0.00000009161,
      -0.00000001200,     0.00000001338,
      -0.00000367534,    -0.00000306985,
       0.00000009154,    -0.00000014748,
      -0.00000001337,    -0.00000001200,
      -0.00000007846,    -0.00000001079,
      -0.00000000174,    -0.00000000483,
       0.00000000483,    -0.00000000173,
       0.00000000246,     0.00000000021,
      -0.00000000021,     0.00000000246,
      -0.00000017148,     0.00000001178,
      -0.00000000038,    -0.00000000267,
       0.00000004678,    -0.00000000962,
      -0.00000000087,    -0.00000000206,
      -0.00000000963,    -0.00000004674,
      -0.00000000202,     0.00000000086,
       0.00000000034,     0.00000000550,
       0.00000000550,    -0.00000000034,
      -0.00000005068,     0.00000000025,
       0.00000001468,    -0.00000000199,
      -0.00000000199,    -0.00000001465,
       0.00000002826,     0.00000024414,
      -0.00000021724,     0.00000020709,
       0.00000000351,     0.00000000535,
       0.00000020709,     0.00000021725,
       0.00000000535,    -0.00000000351,
      -0.00000000046,    -0.00000000252,
      -0.00000000758,    -0.00000000042,
      -0.00000000042,     0.00000000758,
       0.00002541682,    -0.00000362337,
      -0.00000025697,    -0.00000076059,
      -0.00000000603,     0.00000001826,
       0.00000015646,     0.00000111527,
       0.00000003412,    -0.00000001148,
      -0.00000000083,    -0.00000000028,
       0.00000030191,    -0.00000002922,
      -0.00000000121,    -0.00000000442,
      -0.00000002937,    -0.00000031795,
      -0.00000000512,     0.00000000141,
      -0.00000000591,    -0.00000000030,
      -0.00000000028,     0.00000000588,
      -0.00001008134,     0.00001018562,
       0.00000017798,     0.00000023089,
      -0.00000009183,    -0.00000007587,
      -0.00000000105,     0.00000000215,
      -0.00001329581,    -0.00000072292,
       0.00000006105,     0.00000020815,
      -0.00000000053,    -0.00000000595,
      -0.00000072351,     0.00001329594,
       0.00000020817,    -0.00000006104,
      -0.00000000595,     0.00000000053,
      -0.00000004715,    -0.00000000310,
      -0.00000000007,     0.00000000093,
       0.00000363069,    -0.00000028826,
      -0.00000001171,    -0.00000005757,
       0.00000001324,     0.00000018854,
       0.00000000270,    -0.00000000057,
      -0.00000002647,    -0.00000000043,
      -0.00000000036,     0.00000002259,
      -0.00072442666,    -0.00004022516,
       0.00000333836,     0.00001133539,
      -0.00000002930,    -0.00000032376,
       0.00000122541,    -0.00002216464,
      -0.00000034706,     0.00000010185,
       0.00000000998,    -0.00000000089,
      -0.00000107643,     0.00000009236,
       0.00000000144,    -0.00000000012,
       0.00000005218,     0.00000103859,
      -0.00000000094,    -0.00000000074,
      -0.00000005225,     0.00000000114,
       0.00000000215,     0.00000000265,
       0.00000000110,     0.00000005229,
       0.00000000265,    -0.00000000215,
       0.00000019642,    -0.00000002093,
      -0.00000000065,    -0.00000000632,
       0.00000002259,    -0.00000000010,
       0.00000000009,     0.00000002241,
      -0.00000066713,    -0.00000019728,
      -0.00000000913,     0.00000002067,
      -0.00000014953,    -0.00000001751,
       0.00000001736,    -0.00000014926,
       0.00000004417,     0.00000000230,
      -0.00000000186,    -0.00000000212,
      -0.00000000230,     0.00000004399,
       0.00000000212,    -0.00000000186,
      -0.00000000456,    -0.00000000031,
       0.00000000029,    -0.00000000458,
      -0.00000002582,     0.00000000023,
       0.00000000176,    -0.00000001026,
       0.00000000009,     0.00000000413,
      -0.00000000165,    -0.00000000028,
      -0.00000000042,    -0.00000000065,
       0.00000086077,    -0.00000000369,
       0.00000002025,     0.00000008113,
       0.00000000353,    -0.00000004219,
       0.00000000407,    -0.00000000115,
      -0.00000044216,     0.00000002538,
      -0.00000000368,    -0.00000002640,
       0.00000000126,    -0.00000000051,
      -0.00000002553,    -0.00000044214,
       0.00000002639,    -0.00000000366,
       0.00000000049,     0.00000000124,
       0.00000004770,    -0.00000007498,
       0.00000000372,     0.00000000252,
       0.00000007498,     0.00000004773,
      -0.00000000252,     0.00000000373,
       0.00000000304,    -0.00000000087,
      -0.00000000087,    -0.00000000302,
       0.00000000184,    -0.00000000038,
      -0.00000000038,    -0.00000000185,
      -0.00000000474,     0.00000000881,
       0.00000000881,     0.00000000474,
       0.00000137933,    -0.00000029898,
      -0.00000002579,    -0.00000005980,
       0.00000001315,     0.00000006187,
       0.00000000277,    -0.00000000119,
       0.00000002649,    -0.00000000456,
      -0.00000000028,    -0.00000000080,
      -0.00000000463,    -0.00000002738,
      -0.00000000085,     0.00000000030,
      -0.00000020544,     0.00000044011,
      -0.00000000402,    -0.00000000180,
      -0.00000047468,     0.00000009999,
       0.00000000510,     0.00000001398,
       0.00000009998,     0.00000047470,
       0.00000001398,    -0.00000000510,
       0.00000046494,    -0.00000007369,
       0.00000000355,     0.00000002392,
      -0.00000000535,     0.00000000050,
       0.00000000047,     0.00000000484,
      -0.00002653091,     0.00000495816,
       0.00000027904,     0.00000078545,
      -0.00000015279,    -0.00000083474,
      -0.00000002516,     0.00000000890,
      -0.00000035219,     0.00000003140,
       0.00000000128,     0.00000000555,
       0.00000002959,     0.00000035159,
       0.00000000553,    -0.00000000122,
       0.00000000183,     0.00000000212,
       0.00000000212,    -0.00000000183,
       0.00000119231,    -0.00000013000,
      -0.00000000356,    -0.00000001908,
       0.00000006186,     0.00000065766,
       0.00000001023,    -0.00000000210,
      -0.00000001891,     0.00000000152,
      -0.00000000166,    -0.00000002334,
      -0.00000000140,     0.00000000085,
       0.00000000120,     0.00000000076,
      -0.00000564229,     0.00000042605,
       0.00000001859,     0.00000008889,
      -0.00000007455,    -0.00000094271,
      -0.00000001462,     0.00000000314,
       0.00000001700,    -0.00000000087,
       0.00000000123,     0.00000001714,
      -0.00000000244,     0.00000000022,
       0.00000000014,     0.00000000246,
       0.00000421395,    -0.00000000567,
      -0.00000000031,    -0.00000014694,
       0.00000001997,    -0.00000000097,
       0.00000000051,     0.00000001649,
      -0.00001051232,     0.00000001086,
      -0.00000000043,     0.00000011332,
      -0.00000010148,     0.00000000682,
       0.00000000033,     0.00000000160,
      -0.00000000581,    -0.00000010116,
      -0.00000000162,     0.00000000033,
       0.00000000586,     0.00000000098,
      -0.00000000123,     0.00000000452,
      -0.00000000224,     0.00000000027,
      -0.00000000027,    -0.00000000223,
      -0.00000000667,     0.00000000084,
      -0.00000000084,    -0.00000000668,
       0.00000000247,     0.00000000313,
       0.00000000136,    -0.00000000101,
      -0.00000001307,    -0.00000001876,
       0.00000000865,    -0.00000000344,
      -0.00000000320,     0.00000000245,
      -0.00000000065,    -0.00000000151,
      -0.00000000003,     0.00000000101,
       0.00000000148,     0.00000000109,
      -0.00000001784,     0.00000000496,
      -0.00000000008,     0.00000000204,
       0.00000001115,    -0.00000000161,
       0.00000000021,     0.00000000094,
       0.00000000160,     0.00000001117,
      -0.00000000094,     0.00000000021,
      -0.00000000075,     0.00000000188,
      -0.00000000188,    -0.00000000074,
       0.00000000088,    -0.00000000073,
      -0.00000000084,    -0.00000000083,
      -0.00000000034,    -0.00000000072,
      -0.00000000095,     0.00000000035,
       0.00000000122,     0.00000000086,
       0.00000000038,    -0.00000000098,
       0.00000002332,    -0.00000002134,
      -0.00000001723,    -0.00000001184,
      -0.00000000235,     0.00000000571,
       0.00000000810,     0.00000000870,
       0.00000000440,    -0.00000000652,
      -0.00000000215,    -0.00000000087,
      -0.00000010673,     0.00000009790,
       0.00000007891,     0.00000005415,
       0.00000001085,    -0.00000002631,
      -0.00000001839,    -0.00000001975,
      -0.00000001000,     0.00000001480,
       0.00000000490,     0.00000000197,
      -0.00000000067,    -0.00000000082,
      -0.00000000102,    -0.00000000080,
       0.00000008595,    -0.00000002520,
       0.00000000119,     0.00000000416,
       0.00000000240,    -0.00000000059,
      -0.00000000060,    -0.00000000247,
      -0.00000002361,     0.00000000633,
       0.00000000049,     0.00000000105,
       0.00000000633,     0.00000002361,
       0.00000000105,    -0.00000000049,
       0.00000005766,    -0.00000001324,
       0.00000000056,     0.00000000261,
      -0.00000130518,     0.00000032855,
       0.00000002573,     0.00000005663,
      -0.00000001035,    -0.00000004253,
      -0.00000000192,     0.00000000086,
      -0.00000002898,     0.00000000532,
       0.00000000030,     0.00000000088,
       0.00000000524,     0.00000002899,
       0.00000000088,    -0.00000000030,
       0.00000014834,    -0.00000002850,
       0.00000000974,     0.00000005596,
       0.00000000168,    -0.00000000055,
      -0.00000000190,     0.00000000027,
      -0.00000000025,    -0.00000000163,
      -0.00000068306,     0.00000012474,
      -0.00000001348,    -0.00000007942,
      -0.00000000236,     0.00000000077,
       0.00000000241,    -0.00000000027,
       0.00000062871,    -0.00000007790,
      -0.00000000267,    -0.00000002440,
       0.00000000609,    -0.00000000037,
       0.00000000000,    -0.00000000175,
      -0.00000181603,     0.00000020930,
       0.00000000529,     0.00000002864,
       0.00000000207,     0.00000001937,
      -0.00000001972,     0.00000000145,
      -0.00000000123,     0.00000000178,
       0.00000208094,     0.00000001910,
       0.00000000085,    -0.00000002852,
       0.00000005357,    -0.00000000469,
       0.00000000456,     0.00000005241,
      -0.00000000291,    -0.00000000281,
       0.00000000272,    -0.00000000294,
      -0.00000505841,     0.00000001463,
       0.00000000012,     0.00000002015,
       0.00000000051,     0.00000000051,
      -0.00000003956,    -0.00000000526,
      -0.00000000027,    -0.00000000143,
      -0.00000000707,     0.00000000016,
      -0.00000000020,    -0.00000000710,
      -0.00000000388,     0.00000000002,
       0.00000000000,    -0.00000000385,
       0.00000006902,    -0.00000000411,
      -0.00000000653,    -0.00000002564,
       0.00000000087,     0.00000000033,
      -0.00000000148,     0.00000000051,
       0.00000000051,     0.00000000148,
      -0.00000007787,     0.00000002566,
      -0.00000000087,    -0.00000000279,
      -0.00000000276,     0.00000000069,
       0.00000000068,     0.00000000275,
       0.00000001762,    -0.00000000475,
       0.00000000122,     0.00000000493,
      -0.00000008336,     0.00000002136,
      -0.00000000174,    -0.00000000722,
       0.00000007814,    -0.00000001701,
      -0.00000000075,    -0.00000000369,
      -0.00000024183,     0.00000005060,
       0.00000000065,     0.00000000323,
      -0.00000000214,     0.00000000037,
       0.00000035290,    -0.00000004607,
      -0.00000000050,    -0.00000000478,
       0.00000000733,    -0.00000000083,
       0.00000000076,     0.00000000207,
      -0.00000093826,     0.00000012093,
       0.00000000046,     0.00000000353,
      -0.00000000704,     0.00000000002,
       0.00000000010,     0.00000000704,
       0.00000121198,    -0.00000000318,
      -0.00000000012,    -0.00000001183,
      -0.00000000030,    -0.00000000144,
      -0.00000265960,     0.00000000773,
       0.00000000005,     0.00000000584,
      -0.00000005863,    -0.00000000227,
       0.00000006646,     0.00000000343,
      -0.00000000338,     0.00000000002,
      -0.00000000001,    -0.00000000336,
       0.00000004211,     0.00000000126,
       0.00000000439,    -0.00000001062,
       0.00000000141,     0.00000000061,
       0.00000000147,     0.00000000025,
       0.00000000032,    -0.00000000157,
      -0.00000006772,     0.00000010172,
       0.00000001985,     0.00000001628,
      -0.00000000617,    -0.00000000287,
      -0.00000000032,     0.00000000100,
      -0.00000000292,    -0.00000000037,
      -0.00000000100,     0.00000000398,
      -0.00000002923,     0.00000000860,
       0.00000004582,    -0.00000001083,
      -0.00000012860,     0.00000002984,
      -0.00000000097,     0.00000000013,
       0.00000000015,     0.00000000098,
       0.00000022295,    -0.00000003181,
      -0.00000000023,    -0.00000000176,
      -0.00000054038,     0.00000007385,
       0.00000000015,     0.00000000099,
      -0.00000000431,     0.00000000004,
       0.00000000002,     0.00000000404,
       0.00000068211,    -0.00000000099,
      -0.00000000006,    -0.00000000477,
      -0.00000147016,     0.00000000323,
       0.00000000002,     0.00000000129,
      -0.00000005824,    -0.00000000001,
       0.00000000231,    -0.00000000003,
       0.00000000002,     0.00000000231,
       0.00000007277,     0.00000000134,
      -0.00000000150,     0.00000000001,
      -0.00000000001,    -0.00000000148,
       0.00000000322,    -0.00000000689,
       0.00000000085,     0.00000000041,
       0.00000000020,    -0.00000000085,
      -0.00000004514,     0.00000006117,
       0.00000001167,     0.00000001105,
      -0.00000000374,    -0.00000000195,
      -0.00000000146,    -0.00000000015,
      -0.00000000069,     0.00000000223,
       0.00000003917,     0.00000000376,
      -0.00000000106,     0.00000000619,
      -0.00000036471,    -0.00000006972,
      -0.00000003354,     0.00000002038,
       0.00000000529,    -0.00000002669,
      -0.00000000179,    -0.00000000232,
      -0.00000000140,     0.00000000531,
      -0.00000000262,    -0.00000000278,
       0.00000003013,    -0.00000000771,
      -0.00000007754,     0.00000001921,
       0.00000013871,    -0.00000002048,
      -0.00000032652,     0.00000004638,
       0.00000000016,     0.00000000102,
      -0.00000000171,     0.00000000002,
       0.00000000000,     0.00000000155,
       0.00000039076,     0.00000000013,
      -0.00000000002,    -0.00000000211,
      -0.00000083386,     0.00000000106,
      -0.00000004561,     0.00000000023,
       0.00000006437,     0.00000000034,
       0.00000001994,     0.00000000265,
      -0.00000000060,     0.00000000297,
      -0.00000018275,    -0.00000004023,
      -0.00000001905,     0.00000000916,
       0.00000000293,    -0.00000001286,
      -0.00000000083,    -0.00000000128,
      -0.00000000089,     0.00000000327,
      -0.00000000163,    -0.00000000190,
      -0.00000000238,    -0.00000001093,
       0.00000000221,    -0.00000000120,
       0.00000000188,     0.00000000044,
      -0.00000000023,     0.00000000207,
       0.00000006954,     0.00000013869,
       0.00000000908,    -0.00000001302,
      -0.00000001018,     0.00000000677,
       0.00000000096,     0.00000000063,
      -0.00000001715,    -0.00000000353,
      -0.00000000155,     0.00000000120,
       0.00000000283,    -0.00000001772,
      -0.00000000107,    -0.00000000142,
       0.00000001978,    -0.00000000531,
      -0.00000004932,     0.00000001284,
       0.00000008748,    -0.00000001322,
      -0.00000020181,     0.00000002950,
       0.00000022698,     0.00000000051,
      -0.00000048056,    -0.00000000009,
       0.00000000000,     0.00000000103,
      -0.00000003367,     0.00000000025,
       0.00000005115,    -0.00000000003,
      -0.00000001803,     0.00000002012,
      -0.00000000124,    -0.00000000077,
      -0.00000000309,    -0.00000000659,
       0.00000000122,    -0.00000000084,
       0.00000000091,     0.00000000028,
      -0.00000000011,     0.00000000101,
       0.00000005120,     0.00000008021,
      -0.00000000565,     0.00000000423,
      -0.00000000837,    -0.00000000208,
      -0.00000000087,     0.00000000057,
       0.00000000155,    -0.00000000864,
      -0.00000000049,    -0.00000000080,
      -0.00000000055,    -0.00000000059,
      -0.00000000040,    -0.00000000073,
       0.00000000076,    -0.00000000048,
       0.00000004568,    -0.00000003174,
       0.00000000301,     0.00000000168,
       0.00000000402,     0.00000000661,
       0.00000000045,    -0.00000000068,
      -0.00000000666,     0.00000000432,
       0.00000000068,     0.00000000039,
      -0.00000003212,     0.00000000868,
       0.00000005563,    -0.00000000854,
      -0.00000012637,     0.00000001885,
       0.00000013294,     0.00000000059,
      -0.00000027995,    -0.00000000060,
      -0.00000002405,     0.00000000020,
       0.00000003832,    -0.00000000015,
      -0.00000001086,     0.00000001102,
      -0.00000004293,    -0.00000001333,
       0.00000000092,    -0.00000000291,
      -0.00000000034,     0.00000000112,
      -0.00000000055,    -0.00000000075,
       0.00000001934,    -0.00000002465,
       0.00000000203,     0.00000000061,
       0.00000000265,     0.00000000372,
      -0.00000000371,     0.00000000284,
      -0.00000002398,    -0.00000000990,
       0.00000000151,    -0.00000000179,
       0.00000000187,     0.00000000146,
       0.00000003543,    -0.00000000549,
      -0.00000007981,     0.00000001209,
       0.00000007823,     0.00000000056,
      -0.00000016428,    -0.00000000074,
      -0.00000002016,    -0.00000000757,
       0.00000000052,    -0.00000000136,
       0.00000001985,     0.00000002577,
      -0.00000000172,     0.00000000141,
      -0.00000000193,    -0.00000000072,
       0.00000000047,    -0.00000000199,
       0.00000000058,    -0.00000000127,
       0.00000000132,     0.00000000055,
       0.00000002268,    -0.00000000353,
      -0.00000005054,     0.00000000773,
       0.00000004626,     0.00000000047,
      -0.00000009691,    -0.00000000071,
       0.00000001102,     0.00000001422,
      -0.00000000094,     0.00000000077,
      -0.00000000092,    -0.00000000041,
       0.00000000093,     0.00000000117,
      -0.00000000113,     0.00000000100,
      -0.00000003209,     0.00000000494,
       0.00000002742,     0.00000000039,
      -0.00000005742,    -0.00000000062,
      -0.00000003405,    -0.00000000050,
      -0.00000000593,     0.00000000021,
       0.00000000021,     0.00000000590,
      -0.00000000155,    -0.00000000006,
      -0.00000000007,     0.00000000155,
      -0.00000000284,    -0.00000000191,
      -0.00000000191,     0.00000000284,
       0.00000000027,     0.00000000090,
       0.00000000204,    -0.00000000025,
      -0.00000000025,    -0.00000000205,
      -0.00000289818,    -0.00000002874,
       0.00000003621,     0.00000013259,
      -0.00000000299,    -0.00000004266,
      -0.00000000232,     0.00000000042,
       0.00000081378,    -0.00000010246,
      -0.00000000805,    -0.00000002451,
      -0.00000010266,    -0.00000081316,
      -0.00000002445,     0.00000000805,
      -0.00000015783,    -0.00000010980,
      -0.00000002720,    -0.00000026700,
      -0.00000000425,    -0.00000000321,
      -0.00000026700,     0.00000002720,
      -0.00000000321,     0.00000000425,
       0.00000000047,     0.00000000718,
       0.00000000507,    -0.00000000090,
      -0.00000000874,    -0.00000001193,
      -0.00000000058,     0.00000000094,
      -0.00000001192,     0.00000000873,
       0.00000000094,     0.00000000058,
      -0.00000000190,     0.00000000778,
       0.00000000078,     0.00000000054,
      -0.00000000168,    -0.00000000176,
      -0.00000000179,     0.00000000173,
      -0.00000040207,    -0.00000002021,
       0.00000000020,    -0.00000000504,
       0.00000010870,    -0.00000000695,
      -0.00000000036,    -0.00000000172,
      -0.00000000698,    -0.00000010857,
      -0.00000000171,     0.00000000036,
      -0.00000168329,    -0.00001616382,
      -0.00000021196,    -0.00000019643,
       0.00000000041,     0.00000000399,
       0.00000000094,     0.00000000171,
       0.00001351141,    -0.00001422599,
       0.00000004201,    -0.00000020860,
      -0.00000000267,    -0.00000000175,
      -0.00001422581,    -0.00001351177,
      -0.00000020859,    -0.00000004202,
      -0.00000000175,     0.00000000267,
       0.00000000410,    -0.00000000023,
      -0.00000000023,    -0.00000000410,
      -0.00000033809,     0.00000008560,
      -0.00000000850,     0.00000001611,
       0.00000005757,     0.00000022633,
       0.00000001093,     0.00000000570,
       0.00000000048,    -0.00000000054,
       0.00000071822,    -0.00000000076,
       0.00000000041,    -0.00000071510,
      -0.00000002475,     0.00000004523,
      -0.00000000643,     0.00000000729,
      -0.00000000019,     0.00000000084,
       0.00058751896,     0.00003239962,
      -0.00000274776,    -0.00000918829,
       0.00000003252,     0.00000027860,
       0.00000001367,    -0.00000000815,
      -0.00000140987,     0.00002533360,
       0.00000039596,    -0.00000011872,
      -0.00000001214,     0.00000000142,
       0.00000384682,    -0.00000007531,
      -0.00000000536,     0.00000000638,
      -0.00000000856,    -0.00000419306,
      -0.00000000488,     0.00000000756,
      -0.00000003229,     0.00000003261,
      -0.00000000073,    -0.00000000084,
       0.00000003297,     0.00000003194,
      -0.00000000084,     0.00000000073,
       0.00097919295,    -0.00106537829,
       0.00000223461,    -0.00001378576,
      -0.00000010800,    -0.00000005385,
       0.00000942573,     0.00000712491,
       0.00000003516,    -0.00000002936,
      -0.00000000345,     0.00000000362,
       0.00131042157,    -0.00000031181,
      -0.00000009910,     0.00000001649,
       0.00000000141,     0.00000019713,
      -0.00000025533,    -0.00131043133,
       0.00000001676,     0.00000009824,
       0.00000019709,    -0.00000000138,
       0.00000003171,    -0.00000003841,
      -0.00000000236,    -0.00000000033,
      -0.00000003835,    -0.00000003178,
      -0.00000000035,     0.00000000236,
       0.00000042428,    -0.00000000088,
       0.00000000000,    -0.00000000331,
       0.00000001868,     0.00000008316,
       0.00000470006,    -0.00000000450,
      -0.00000000002,    -0.00000003689,
      -0.00000002100,     0.00000002373,
       0.00000497719,    -0.00000000101,
       0.00000000000,    -0.00000003901,
      -0.00000001850,     0.00000008061,
       0.00000042827,    -0.00000000030,
       0.00000000000,    -0.00000000312,
       0.00000003618,    -0.00000001609,
      -0.00000000018,     0.00000000124,
       0.00000002790,     0.00000000112,
      -0.00000000030,     0.00000000211,
       0.00000001456,    -0.00000000738,
       0.00000000064,     0.00000000083,
       0.00000000162,     0.00000000021,
      -0.00000000028,     0.00000000169,
       0.00000000000,     0.00000000417,
       0.00000000433,    -0.00000000054,
      -0.00000000166,    -0.00000000120,
       0.00000000037,     0.00000000590,
       0.00000000591,    -0.00000000039,
       0.00000000046,     0.00000000712,
       0.00000000712,    -0.00000000046,
       0.00000005320,    -0.00000032419,
      -0.00000006790,    -0.00000000915,
      -0.00000000319,     0.00000001760,
      -0.00000001530,     0.00000000308,
      -0.00000000025,     0.00000000313,
       0.00000000072,    -0.00000000127,
       0.00000000148,     0.00000000069,
      -0.00000000167,     0.00000000494,
      -0.00000000494,    -0.00000000165,
       0.00000002433,     0.00000003443,
       0.00000000094,     0.00000000197,
       0.00000000162,    -0.00000001190,
      -0.00000000190,     0.00000000014,
       0.00000001180,     0.00000000169,
      -0.00000000015,    -0.00000000189,
      -0.00000000887,    -0.00000001393,
       0.00000000064,    -0.00000000081,
       0.00000000107,     0.00000000076,
      -0.00000000074,     0.00000000112,
      -0.00000000850,    -0.00000000230,
      -0.00000000085,     0.00000000061,
       0.00000000061,     0.00000000085,
      -0.00000176985,     0.00000017217,
       0.00000017288,     0.00000071789,
       0.00000014647,    -0.00000005861,
      -0.00000001165,    -0.00000001926,
       0.00000000718,     0.00000007362,
       0.00000002979,    -0.00000000718,
      -0.00000000243,    -0.00000000608,
      -0.00000000082,     0.00000000049,
      -0.00000000417,     0.00000000033,
       0.00000000036,     0.00000000169,
       0.00000000039,     0.00000000089,
       0.00000007948,    -0.00000001711,
      -0.00000001130,    -0.00000003181,
      -0.00000000021,    -0.00000000100,
       0.00000000151,    -0.00000000046,
       0.00000000053,     0.00000000196,
       0.00000000080,    -0.00000000034,
      -0.00000037960,     0.00000005645,
       0.00000003949,     0.00000014687,
       0.00000002889,    -0.00000001131,
      -0.00000000043,     0.00000000400,
       0.00000000166,    -0.00000000015,
      -0.00000005083,     0.00000000494,
       0.00000000495,     0.00000002060,
       0.00000000421,    -0.00000000168,
      -0.00000000495,    -0.00000005107,
      -0.00000002068,     0.00000000497,
       0.00000000169,     0.00000000421,
      -0.00000000173,     0.00000000016,
       0.00000001314,     0.00000024037,
       0.00000005094,    -0.00000000395,
       0.00000001902,    -0.00000000282,
      -0.00000000037,    -0.00000000415,
      -0.00000000182,    -0.00000000014,
      -0.00000000059,     0.00000000074,
       0.00000010284,     0.00000000272,
      -0.00000000176,    -0.00000002091,
      -0.00000000092,     0.00000000554,
      -0.00000000727,     0.00000000165,
       0.00000000080,     0.00000000273,
      -0.00000000167,    -0.00000000728,
      -0.00000000274,     0.00000000081,
      -0.00000000032,    -0.00000000168,
       0.00000006139,    -0.00000001571,
       0.00000000134,     0.00000000502,
       0.00000000572,     0.00000000059,
      -0.00000000074,     0.00000000576,
      -0.00000001358,    -0.00000001974,
       0.00000000197,    -0.00000000126,
       0.00000000390,    -0.00000000124,
       0.00000000125,     0.00000000400,
      -0.00000000100,    -0.00000000081,
      -0.00000000122,    -0.00000000168,
       0.00000000175,    -0.00000000126,
       0.00000000476,    -0.00000000752,
       0.00000002197,     0.00000000319,
      -0.00000000085,     0.00000000806,
      -0.00000000278,     0.00000000166,
      -0.00000000164,    -0.00000000282,
       0.00000000019,    -0.00000000231,
      -0.00000000178,     0.00000000001,
      -0.00000000009,     0.00000000173,
       0.00000008188,     0.00000001813,
      -0.00000000192,     0.00000000908,
       0.00000000077,     0.00000000064,
       0.00000000195,    -0.00000000384,
       0.00000000693,     0.00000000179,
      -0.00000000276,     0.00000000908,
       0.00000000169,    -0.00000000031,
      -0.00000000709,     0.00000000385,
       0.00000000000,     0.00000000213,
      -0.00000006707,     0.00000003291,
       0.00000000574,     0.00000000743,
      -0.00000001812,    -0.00000000626,
      -0.00000000012,     0.00000000317,
       0.00000000623,     0.00000000693,
       0.00000000220,    -0.00000000006,
      -0.00000000852,    -0.00000000752,
       0.00000000032,     0.00000000180,
       0.00000014738,    -0.00000001628,
       0.00000000233,     0.00000001634,
       0.00000000996,     0.00000000169,
       0.00000000046,    -0.00000000074,
      -0.00000000192,     0.00000000984,
       0.00000000066,     0.00000000058,
       0.00000000473,    -0.00000000429,
       0.00000000303,     0.00000001390,
      -0.00000001652,     0.00000000542,
       0.00000000166,    -0.00000000201,
      -0.00000000455,     0.00000000023,
      -0.00000000009,    -0.00000000376,
      -0.00000003254,    -0.00000010657,
       0.00000002982,    -0.00000001141,
      -0.00000000213,     0.00000000113,
       0.00000002267,    -0.00000000526,
       0.00000000485,     0.00000002324,
       0.00000000131,     0.00000000236,
      -0.00000000228,     0.00000000143,
       0.00000000062,    -0.00000000043,
       0.00000000115,    -0.00000000463,
      -0.00000009071,     0.00000002201,
      -0.00000000121,     0.00000000732,
       0.00000000091,     0.00000000108,
       0.00000000248,     0.00000001835,
       0.00000000910,    -0.00000000696,
      -0.00000000465,    -0.00000003872,
       0.00000000340,    -0.00000000111,
       0.00000000172,     0.00000000069,
      -0.00000000014,     0.00000000209,
      -0.00000013265,     0.00000002465,
       0.00000000131,     0.00000000602,
       0.00000000054,     0.00000000055,
      -0.00000000711,     0.00000000000,
      -0.00000000080,     0.00000000055,
      -0.00000000012,    -0.00000000682,
      -0.00000000062,    -0.00000000074,
      -0.00000000087,    -0.00000000641,
      -0.00000000004,    -0.00000000158,
      -0.00000000376,     0.00000000304,
       0.00000002996,    -0.00000002762,
       0.00000000181,     0.00000001625,
      -0.00000000382,    -0.00000000802,
       0.00000000031,    -0.00000000093,
      -0.00000000018,    -0.00000000137,
      -0.00000000282,     0.00000000062,
      -0.00000000057,    -0.00000000293,
       0.00000000035,    -0.00000000134,
      -0.00000000037,     0.00000000149,
       0.00000000320,     0.00000000999,
       0.00000002136,    -0.00000000511,
      -0.00000000307,    -0.00000001289,
      -0.00000000573,    -0.00000002255,
      -0.00000001436,     0.00000000334,
       0.00000000140,     0.00000000310,
       0.00000000320,     0.00000000387,
       0.00000000383,    -0.00000000320,
      -0.00000000041,    -0.00000000064,
       0.00000001877,    -0.00000004329,
      -0.00000000618,     0.00000000097,
       0.00000000169,     0.00000000184,
       0.00000001414,     0.00000002280,
       0.00000000396,    -0.00000000314,
      -0.00000000074,    -0.00000000093,
       0.00000002258,    -0.00000001402,
      -0.00000000314,    -0.00000000381,
      -0.00000000097,     0.00000000072,
      -0.00000000024,    -0.00000000104,
       0.00000000022,     0.00000000141,
      -0.00000000050,     0.00000000063,
       0.00000000063,     0.00000000057,
      -0.00000000013,     0.00000000091,
       0.00000001700,     0.00000000940,
      -0.00000000029,     0.00000000242,
       0.00000000554,    -0.00000000745,
       0.00000000150,     0.00000000115,
      -0.00000000101,    -0.00000000014,
       0.00000000002,    -0.00000000099,
       0.00000007347,    -0.00000001637,
       0.00000000140,     0.00000000021,
      -0.00000000013,     0.00000000205,
      -0.00000004699,    -0.00000008698,
       0.00000001332,     0.00000000111,
       0.00000000570,    -0.00000000354,
       0.00000000108,    -0.00000000036,
       0.00000000055,     0.00000000124,
      -0.00000005546,    -0.00000001216,
       0.00000000018,    -0.00000000214,
      -0.00000000095,     0.00000000241,
      -0.00000003672,    -0.00000004726,
       0.00000000144,    -0.00000000115,
       0.00000000065,    -0.00000000071,
       0.00000000370,     0.00000002267,
       0.00000001697,     0.00000000736,
       0.00000000046,    -0.00000000180,
      -0.00000000083,     0.00000000436,
       0.00000000429,     0.00000000138,
      -0.00000019096,    -0.00000005222,
      -0.00000001102,     0.00000001580,
       0.00000000168,    -0.00000000589,
       0.00000000432,     0.00000000093,
       0.00000000095,    -0.00000000422,
      -0.00000036357,     0.00000019467,
      -0.00000000193,     0.00000005622,
      -0.00000000350,     0.00000000212,
       0.00000000104,    -0.00000000005,
      -0.00000035683,    -0.00000008021,
      -0.00000002408,     0.00000002759,
       0.00000000148,     0.00000000124,
      -0.00000008015,     0.00000035680,
       0.00000002759,     0.00000002406,
       0.00000000125,    -0.00000000150,
      -0.00000000038,     0.00000000213,
       0.00000000213,     0.00000000038,
       0.00000000177,     0.00000000290,
       0.00000000136,     0.00000000013,
      -0.00000000011,    -0.00000000153,
      -0.00000000165,    -0.00000000002,
      -0.00000000520,    -0.00000000570,
      -0.00000000428,    -0.00000000042,
      -0.00000000086,     0.00000000096,
       0.00000000096,     0.00000000086,
      -0.00000000046,     0.00000000111,
      -0.00000000520,    -0.00000000147,
      -0.00000000149,     0.00000000532,
      -0.00000001260,     0.00000000741,
      -0.00000000009,    -0.00000000102,
      -0.00000000224,     0.00000000003,
      -0.00000000003,     0.00000000225,
       0.00000002379,     0.00000001038,
       0.00000000044,    -0.00000000130,
      -0.00000000863,    -0.00000000297,
      -0.00000000296,     0.00000000862,
       0.00000005879,     0.00000000160,
      -0.00000000011,     0.00000000100,
      -0.00000059749,     0.00000013702,
      -0.00000000808,    -0.00000001206,
      -0.00000000272,    -0.00000000845,
       0.00000000752,     0.00000003869,
      -0.00000000440,    -0.00000000170,
       0.00000003934,    -0.00000001756,
      -0.00000000190,     0.00000000448,
       0.00000057902,     0.00000133659,
      -0.00000016203,    -0.00000004386,
      -0.00000005368,     0.00000002227,
       0.00000000169,    -0.00000000607,
      -0.00000000207,     0.00000000372,
       0.00000000065,    -0.00000000274,
       0.00000000074,     0.00000000137,
      -0.00000000138,     0.00000000074,
       0.00000018528,    -0.00000000378,
      -0.00000000052,    -0.00000000829,
      -0.00000000099,    -0.00000000008,
       0.00000000003,    -0.00000000100,
       0.00000003654,    -0.00000001338,
       0.00000000061,    -0.00000000380,
       0.00000000716,     0.00000000061,
      -0.00000000080,     0.00000000680,
       0.00017526505,    -0.00000011320,
       0.00000002734,    -0.00000000356,
      -0.00000000028,    -0.00000003746,
       0.00000000139,    -0.00000122359,
       0.00000004959,    -0.00000004001,
       0.00000000461,     0.00000000528,
      -0.00000002817,    -0.00000003928,
       0.00000000537,    -0.00000000469,
      -0.00000000360,    -0.00000000289,
      -0.00000000296,     0.00000000358,
       0.00000000029,    -0.00000000311,
      -0.00000000311,    -0.00000000030,
       0.00000001021,    -0.00000007628,
       0.00000000155,    -0.00000000005,
       0.00000000141,     0.00000002429,
       0.00000000261,    -0.00000000303,
       0.00000002426,    -0.00000000143,
      -0.00000000301,    -0.00000000258,
       0.00000004249,     0.00000053972,
       0.00000005656,    -0.00000006567,
      -0.00000002475,     0.00000000306,
       0.00000000348,     0.00000000283,
       0.00000000671,    -0.00000001486,
      -0.00000000195,     0.00000000083,
      -0.00000001450,    -0.00000000659,
       0.00000000077,     0.00000000187,
      -0.00000000755,     0.00000000218,
      -0.00000000113,    -0.00000000236,
      -0.00000012030,    -0.00000024288,
      -0.00000002282,     0.00000001747,
       0.00000002810,     0.00000000881,
      -0.00000000154,    -0.00000000336,
       0.00000001304,     0.00000000078,
       0.00000000007,    -0.00000000139,
       0.00000119418,     0.00000007568,
      -0.00000000443,    -0.00000001960,
       0.00000000097,    -0.00000001026,
       0.00000000045,    -0.00000000236,
       0.00000000184,    -0.00000000027,
      -0.00000000066,    -0.00000000049,
       0.00000000169,    -0.00000000003,
       0.00000000003,     0.00000000169,
      -0.00000060914,     0.00000910300,
      -0.00000088046,    -0.00000061520,
       0.00000005532,    -0.00000005354,
       0.00000006612,     0.00000000050,
      -0.00000000435,     0.00000000612,
      -0.00000003055,     0.00000001580,
      -0.00000000775,    -0.00000000333,
       0.00000000027,    -0.00000000109,
       0.00000001517,     0.00000002401,
      -0.00000000334,     0.00000000770,
      -0.00000000109,    -0.00000000027,
       0.00000000653,    -0.00000002182,
       0.00000000122,     0.00000000010,
      -0.00000003587,     0.00000003355,
       0.00000000091,     0.00000000059,
       0.00000000110,     0.00000000011,
       0.00000011207,    -0.00000003747,
      -0.00000000033,    -0.00000000231,
      -0.00000019020,    -0.00000002293,
      -0.00000001155,     0.00000006208,
       0.00000000074,    -0.00000000026,
      -0.00000052692,    -0.00000007315,
       0.00000008277,    -0.00000009432,
      -0.00000000030,     0.00000000423,
       0.00000001400,     0.00000000035,
      -0.00000000112,     0.00000000504,
      -0.00000000097,    -0.00000000036,
       0.00000000059,    -0.00000001393,
       0.00000000502,     0.00000000115,
      -0.00000000036,     0.00000000097,
      -0.00000002181,     0.00000001473,
       0.00000006432,    -0.00000001972,
      -0.00000010073,    -0.00000000402,
       0.00000000019,    -0.00000000086,
       0.00000000309,    -0.00000003758,
      -0.00000001396,     0.00000000895,
       0.00000003693,    -0.00000001101,
      -0.00000005044,    -0.00000000137,
       0.00000027769,     0.00000010179,
      -0.00000004842,     0.00000011884,
      -0.00000002651,    -0.00000001134,
       0.00000000064,    -0.00000000095,
       0.00000002102,     0.00000001674,
      -0.00000000752,     0.00000001019,
      -0.00000000261,    -0.00000000174,
       0.00000001693,    -0.00000002230,
       0.00000001069,     0.00000000767,
      -0.00000000179,     0.00000000271,
       0.00000002098,    -0.00000000661,
      -0.00000002463,     0.00000000076,
      -0.00000005713,    -0.00000001180,
      -0.00000001450,     0.00000007022,
       0.00000000850,     0.00000000200,
       0.00000001442,     0.00000000011,
       0.00000000012,    -0.00000001441,
       0.00000002279,    -0.00000000059,
       0.00000000010,    -0.00000000273,
      -0.00000000648,     0.00000000046,
       0.00000000048,     0.00000000637,
      -0.00000000107,    -0.00000000015,
      -0.00000000015,     0.00000000107,
      -0.00000005412,     0.00000001551,
       0.00000000001,     0.00000000657,
       0.00000001023,    -0.00000000685,
       0.00000000039,    -0.00000000110,
      -0.00000000684,    -0.00000001009,
      -0.00000000113,    -0.00000000040,
       0.00000000007,     0.00000000197,
       0.00000000448,     0.00000000053,
       0.00000000055,    -0.00000000454,
      -0.00000008132,    -0.00000018922,
      -0.00000000149,    -0.00000002232,
      -0.00000000210,    -0.00000000046,
       0.00000009551,     0.00000006046,
      -0.00000000174,    -0.00000000333,
       0.00000006046,    -0.00000009577,
      -0.00000000331,     0.00000000176,
       0.00000000811,    -0.00000000108,
      -0.00000000063,    -0.00000000721,
      -0.00000001692,    -0.00000000149,
      -0.00000000155,     0.00000001692,
      -0.00000000131,    -0.00000003414,
       0.00000000361,    -0.00000000038,
      -0.00000000195,     0.00000000092,
       0.00000000091,     0.00000000189,
       0.00000413314,     0.00000168571,
      -0.00000003185,    -0.00000018996,
       0.00000000535,     0.00000008852,
       0.00000000651,     0.00000000134,
      -0.00000179639,    -0.00000037945,
       0.00000001397,     0.00000005553,
       0.00000000002,    -0.00000000201,
      -0.00000037878,     0.00000179690,
       0.00000005562,    -0.00000001394,
      -0.00000000201,    -0.00000000002,
      -0.00000000058,    -0.00000000060,
      -0.00000000056,     0.00000000055,
      -0.00000002407,     0.00000000360,
       0.00000000258,     0.00000001896,
       0.00000000101,     0.00000000028,
       0.00000004237,     0.00000000242,
       0.00000000252,    -0.00000004241,
      -0.00000000491,     0.00000001017,
      -0.00005894913,    -0.00001128562,
       0.00000038742,     0.00000190222,
       0.00000000926,    -0.00000005759,
       0.00000009419,    -0.00000087316,
      -0.00000003350,     0.00000000106,
       0.00001715821,     0.00000093627,
      -0.00000007906,    -0.00000026882,
       0.00000000088,     0.00000000798,
       0.00000092981,    -0.00001714504,
      -0.00000026832,     0.00000007892,
       0.00000000793,    -0.00000000089,
       0.00000000139,    -0.00000000144,
      -0.00000000157,    -0.00000000129,
      -0.00000002967,    -0.00000004154,
       0.00000002206,    -0.00000002955,
      -0.00000002955,    -0.00000002207,
       0.00000002743,     0.00000000183,
       0.00000002859,     0.00000000192,
      -0.00000000002,     0.00000000208,
       0.00000000011,    -0.00000000319,
       0.00000000074,    -0.00000000031,
       0.00000852844,    -0.00000001907,
       0.00000001893,     0.00000892049,
      -0.00000000013,    -0.00000000126,
      -0.00000000139,     0.00000000000,
      -0.00000026823,    -0.00000002485,
       0.00000000110,     0.00000000438,
       0.00000002348,    -0.00000027467,
      -0.00000000430,     0.00000000116,
      -0.00000001205,     0.00000001373,
       0.00000001422,     0.00000001161,
       0.00001749732,     0.00000003329,
      -0.00000001727,    -0.00000313387,
       0.00000000015,     0.00000000086,
       0.00000022773,     0.00000001945,
      -0.00000000098,    -0.00000000359,
      -0.00000001885,     0.00000022469,
       0.00000000364,    -0.00000000097,
      -0.00000000825,     0.00000000056,
       0.00000000030,     0.00000000893,
       0.00000000111,     0.00000000000,
       0.00000000001,     0.00000000113,
      -0.00000123270,     0.00000021964,
       0.00000000856,     0.00000005156,
       0.00000072156,     0.00000000847,
       0.00000000148,    -0.00000000008,
      -0.00000000767,     0.00000072379,
       0.00000000004,     0.00000000146,
       0.00000002507,    -0.00000002810,
       0.00000002828,     0.00000002521,
      -0.00000121686,     0.00000014982,
       0.00000000121,     0.00000000686,
       0.00000000449,    -0.00000000067,
       0.00000000065,     0.00000000482,
       0.00000000099,    -0.00000000015,
       0.00000000011,     0.00000000090,
       0.00000006338,    -0.00000000623,
       0.00000000008,    -0.00000000267,
      -0.00000002316,     0.00000000458,
      -0.00000000460,    -0.00000002318,
       0.00000000190,    -0.00000000221,
       0.00000000220,     0.00000000190,
       0.00000004160,     0.00000000039,
       0.00000000000,     0.00000000068,
       0.00000019918,     0.00000000135,
       0.00000000087,    -0.00000017827,
      -0.00000000374,    -0.00000000154,
      -0.00000000377,     0.00000000428,
       0.00000005183,     0.00000000007,
       0.00000000011,     0.00000001068,
       0.00000596929,    -0.00000002114,
       0.00000000110,     0.00000034601,
      -0.00000000111,    -0.00000000077,
       0.00000000001,    -0.00000000678,
      -0.00000000111,     0.00000000063,
       0.00000000080,     0.00000000074,
       0.14818171781,    -0.00003166756,
      -0.00001089408,     0.00000187707,
       0.00000016265,     0.00002216235,
       0.00000002305,    -0.00000002299,
       0.00000070450,     0.00441710230,
      -0.00000004095,    -0.00000032360,
      -0.00000065982,     0.00000000381,
       0.00000000064,     0.00000000069,
      -0.00000706534,    -0.00001642288,
      -0.00000000079,    -0.00000018640,
       0.00000000310,    -0.00000000282,
      -0.00000816026,     0.00001465203,
      -0.00000008010,     0.00000001760,
      -0.00000000241,    -0.00000000391,
       0.00001009225,    -0.00000041300,
      -0.00000038425,    -0.00000031896,
      -0.00000002024,     0.00000002689,
       0.00000000098,     0.00000000055,
      -0.00000041066,    -0.00001008508,
      -0.00000031924,     0.00000038392,
       0.00000002690,     0.00000002023,
       0.00000000055,    -0.00000000098,
       0.00000000102,    -0.00000000076,
       0.00000000075,     0.00000000101,
      -0.00004152784,    -0.00001846775,
      -0.00000017206,    -0.00000319571,
       0.00000029008,    -0.00000025663,
      -0.00000087840,     0.00000137525,
      -0.00000011706,     0.00000002020,
      -0.00000000868,    -0.00000001093,
      -0.00000312644,    -0.00000356683,
      -0.00000000121,    -0.00000003785,
       0.00000000098,    -0.00000000066,
       0.00000356356,    -0.00000312742,
       0.00000003756,    -0.00000000132,
       0.00000000066,     0.00000000095,
       0.00000873391,    -0.00000035552,
      -0.00000033248,    -0.00000027629,
      -0.00000001752,     0.00000002327,
       0.00000000085,     0.00000000048,
       0.00000035777,     0.00000873359,
       0.00000027621,    -0.00000033254,
      -0.00000002327,    -0.00000001752,
      -0.00000000048,     0.00000000085,
      -0.00000021909,    -0.00000000052,
      -0.00000000355,    -0.00000001924,
      -0.00000000091,     0.00000001271,
      -0.00000000123,     0.00000000028,
       0.00000012735,    -0.00000002489,
      -0.00000000080,     0.00000000410,
       0.00000002496,     0.00000012742,
      -0.00000000412,    -0.00000000080,
      -0.00000002432,     0.00000002989,
      -0.00000000144,    -0.00000000112,
      -0.00000002987,    -0.00000002431,
       0.00000000112,    -0.00000000144,
      -0.00000013141,    -0.00000036324,
       0.00000001965,    -0.00000002392,
      -0.00000000576,     0.00000000063,
      -0.00000031678,    -0.00000015320,
      -0.00000000118,    -0.00000002465,
       0.00000000223,    -0.00000000196,
       0.00000015320,    -0.00000031679,
       0.00000002465,    -0.00000000118,
       0.00000000196,     0.00000000222,
       0.00000002538,     0.00000002596,
      -0.00000000002,    -0.00000000148,
      -0.00000002595,     0.00000002538,
       0.00000000148,    -0.00000000002,
       0.00000000013,    -0.00000000121,
       0.00000002241,     0.00000010028,
       0.00000004722,    -0.00000002254,
      -0.00000002655,    -0.00000000243,
       0.00000000019,    -0.00000000212,
      -0.00000000081,     0.00000000052,
      -0.00000000161,    -0.00000000013,
       0.00000000008,    -0.00000000167,
      -0.00000073858,     0.00000011310,
       0.00000008315,     0.00000028650,
       0.00000005627,    -0.00000002513,
      -0.00000000077,    -0.00000000513,
      -0.00000000202,     0.00000000058,
       0.00000000129,     0.00000000110,
       0.00000002522,    -0.00000002664,
      -0.00000001236,    -0.00000001065,
      -0.00000000085,    -0.00000000689,
      -0.00000000065,     0.00000000051,
      -0.00000002616,    -0.00000002174,
      -0.00000003010,     0.00000000470,
       0.00000000485,     0.00000000593,
       0.00000000062,    -0.00000000117,
      -0.00000000470,    -0.00000003010,
      -0.00000000593,     0.00000000485,
       0.00000000117,     0.00000000062,
      -0.00000000906,     0.00000002114,
      -0.00000000128,     0.00000000050,
       0.00000000092,    -0.00000000138,
      -0.00000000066,    -0.00000000072,
       0.00000000133,     0.00000000080,
       0.00000000072,    -0.00000000066,
       0.00000000020,     0.00000000098,
      -0.00000000102,     0.00000000022,
       0.00000003420,     0.00000003530,
      -0.00000005070,    -0.00000004630,
       0.00000000350,    -0.00000001240,
       0.00000001250,     0.00000000340,
      -0.00000000170,    -0.00000000149,
       0.00000000197,    -0.00000000031,
       0.00000000028,     0.00000000196,
       0.00000000101,    -0.00000000029,
       0.00000000062,    -0.00000000097,
      -0.00000000100,    -0.00000000058,
       0.00000164257,    -0.00000032520,
      -0.00000022093,    -0.00000060771,
      -0.00000011299,     0.00000006291,
       0.00000001088,     0.00000001353,
      -0.00000001423,    -0.00000007177,
      -0.00000002655,     0.00000000967,
       0.00000000274,     0.00000000494,
       0.00000000062,    -0.00000000050,
       0.00000001090,    -0.00000000155,
      -0.00000000119,    -0.00000000433,
      -0.00000000085,     0.00000000037,
       0.00000000168,     0.00000001189,
       0.00000000460,    -0.00000000133,
      -0.00000000041,    -0.00000000091,
       0.00000338134,     0.00000235649,
       0.00000079662,    -0.00000140662,
      -0.00000030138,    -0.00000013194,
      -0.00000001213,     0.00000004282,
       0.00000003315,    -0.00000001821,
      -0.00000000976,    -0.00000001157,
      -0.00000000185,     0.00000000233,
       0.00000371349,    -0.00000057149,
      -0.00000041994,    -0.00000144074,
      -0.00000028329,     0.00000012547,
       0.00000002285,     0.00000003595,
       0.00000000324,    -0.00000000295,
       0.00000057106,     0.00000371720,
       0.00000144101,    -0.00000041978,
      -0.00000012548,    -0.00000028330,
      -0.00000003595,     0.00000002286,
       0.00000000295,     0.00000000324,
       0.00000000209,    -0.00000000035,
      -0.00000000024,    -0.00000000081,
      -0.00000000035,    -0.00000000209,
      -0.00000000081,     0.00000000024,
       0.00000002453,    -0.00000000377,
      -0.00000000381,    -0.00000002526,
      -0.00000000981,     0.00000000283,
       0.00000000085,     0.00000000193,
      -0.00000000103,     0.00000000094,
       0.00000000094,     0.00000000104,
       0.00000004962,    -0.00000000773,
       0.00000000134,     0.00000000913,
       0.00000000354,    -0.00000000101,
      -0.00000000031,    -0.00000000070,
       0.00000000029,     0.00000000079,
       0.00000000080,    -0.00000000029,
      -0.00000015877,     0.00000004243,
       0.00000002696,     0.00000005507,
       0.00000000181,     0.00000000300,
       0.00000000095,    -0.00000000072,
       0.00000004812,    -0.00000000987,
      -0.00000000656,    -0.00000001776,
      -0.00000000329,     0.00000000184,
       0.00000000990,     0.00000004810,
       0.00000001778,    -0.00000000658,
      -0.00000000183,    -0.00000000329,
      -0.00000002985,     0.00000000670,
       0.00000000889,     0.00000000623,
       0.00000000054,     0.00000000424,
      -0.00000005515,    -0.00000000221,
      -0.00000000023,    -0.00000000644,
      -0.00000000450,     0.00000000064,
       0.00000000071,     0.00000000161,
      -0.00000000064,    -0.00000000448,
      -0.00000000160,     0.00000000069,
      -0.00000000698,     0.00000002011,
      -0.00000000211,    -0.00000000024,
      -0.00000000492,    -0.00000000028,
       0.00000000026,    -0.00000000504,
       0.00000000024,    -0.00000000080,
       0.00000002126,     0.00000000006,
       0.00000000012,     0.00000000205,
      -0.00000000058,     0.00000000178,
      -0.00000000184,    -0.00000000055,
      -0.00000000059,     0.00000000052,
       0.00000000206,     0.00000000001,
      -0.00000000020,     0.00000000211,
       0.00000000511,     0.00000001566,
      -0.00000000769,     0.00000000251,
      -0.00000000263,    -0.00000000288,
       0.00000000290,    -0.00000000264,
      -0.00000000354,     0.00000000023,
       0.00000000019,     0.00000000347,
       0.00000006489,    -0.00000002547,
      -0.00000000467,     0.00000002195,
       0.00000000160,     0.00000000128,
       0.00000000949,    -0.00000000138,
      -0.00000000037,    -0.00000000865,
      -0.00000110404,    -0.00000021123,
      -0.00000007252,     0.00000008705,
       0.00000001177,    -0.00000006537,
      -0.00000000465,    -0.00000000397,
      -0.00000003266,    -0.00000000526,
       0.00000000404,     0.00000000654,
       0.00000000068,    -0.00000000080,
       0.00000000622,    -0.00000002982,
      -0.00000000656,     0.00000000408,
       0.00000000080,     0.00000000068,
      -0.00000001212,     0.00000003506,
       0.00000000935,     0.00000000155,
       0.00000000232,     0.00000000081,
      -0.00000000081,    -0.00000000051,
      -0.00000000048,     0.00000000054,
       0.00000000015,     0.00000000239,
       0.00000000170,    -0.00000000034,
      -0.00000000027,    -0.00000000171,
      -0.00000008545,    -0.00000000644,
       0.00000000057,    -0.00000001011,
      -0.00000000058,    -0.00000000048,
      -0.00000000163,     0.00000000217,
      -0.00000000311,    -0.00000000121,
      -0.00000000169,     0.00000000097,
      -0.00000000555,    -0.00000002469,
       0.00000000351,    -0.00000000417,
      -0.00000000053,     0.00000000073,
       0.00000000396,     0.00000000368,
       0.00000000291,    -0.00000000388,
      -0.00000000109,     0.00000000121,
      -0.00000003533,     0.00000003928,
      -0.00000000529,    -0.00000000230,
      -0.00000001056,    -0.00000000059,
       0.00000000064,    -0.00000001079,
      -0.00000000062,    -0.00000000050,
      -0.00000001389,    -0.00000000140,
       0.00000000056,    -0.00000001739,
      -0.00000000069,     0.00000000323,
       0.00000000227,    -0.00000000101,
       0.00000010968,     0.00000000506,
       0.00000000062,     0.00000003143,
      -0.00000000081,    -0.00000000292,
      -0.00000000527,     0.00000000949,
      -0.00000000988,    -0.00000000542,
      -0.00000000047,    -0.00000000058,
       0.00000000058,    -0.00000000047,
       0.00000000044,    -0.00000000085,
      -0.00000000230,     0.00000000007,
       0.00000000020,    -0.00000000238,
      -0.00000001961,     0.00000000503,
      -0.00000002080,    -0.00000007499,
       0.00000000269,    -0.00000001790,
      -0.00000001117,    -0.00000000569,
       0.00000053859,    -0.00000007269,
       0.00000006845,     0.00000028004,
      -0.00000000047,     0.00000000269,
       0.00000002126,     0.00000002446,
       0.00000000194,    -0.00000000241,
       0.00000001169,     0.00000004865,
       0.00000000233,     0.00000000214,
       0.00000008091,    -0.00000002806,
      -0.00000000105,    -0.00000001135,
       0.00000000115,    -0.00000000053,
       0.00000000058,     0.00000000145,
      -0.00000260317,     0.00000063441,
      -0.00000005386,    -0.00000026732,
       0.00000000118,    -0.00000000303,
       0.00000000112,     0.00000000002,
       0.00000020177,    -0.00000005375,
       0.00000000454,    -0.00000000925,
       0.00000005504,     0.00000020163,
       0.00000000906,     0.00000000464,
       0.00000000168,     0.00000000103,
      -0.00000000293,     0.00000000098,
      -0.00000000055,     0.00000000242,
      -0.00000002583,    -0.00000000688,
      -0.00000000469,    -0.00000000258,
       0.00000000400,     0.00000000079,
       0.00000000137,     0.00000000027,
      -0.00000000150,    -0.00000000503,
      -0.00000000027,     0.00000000137,
      -0.00000000463,    -0.00000001176,
       0.00000000934,     0.00000002921,
      -0.00000000195,    -0.00000000017,
       0.00000004912,    -0.00000002157,
      -0.00000000128,    -0.00000000514,
       0.00000000674,    -0.00000000103,
       0.00000000087,     0.00000000691,
       0.00000000280,    -0.00000000305,
      -0.00000000004,    -0.00000000098,
      -0.00000000127,    -0.00000000024,
       0.00000002274,     0.00000000794,
      -0.00000000734,     0.00000000878,
       0.00000000037,    -0.00000000087,
       0.00000000168,    -0.00000000207,
       0.00000000075,     0.00000000084,
      -0.00000000131,     0.00000000079,
       0.00000000142,    -0.00000000055,
       0.00000000062,     0.00000000144,
       0.00000000348,     0.00000000369,
       0.00000000145,     0.00000000001,
      -0.00000000001,     0.00000000145,
       0.00000001025,     0.00000005958,
      -0.00000001173,    -0.00000000931,
       0.00000000194,    -0.00000000224,
       0.00000000370,     0.00000000094,
       0.00000000062,     0.00000000270,
      -0.00000003315,     0.00000000037,
      -0.00000000008,    -0.00000000278,
       0.00000003497,     0.00000005738,
       0.00000000554,     0.00000000139,
      -0.00000000083,     0.00000000095,
       0.00000001595,    -0.00000000017,
       0.00000000016,     0.00000001596,
       0.00000000100,    -0.00000000018,
      -0.00000000013,    -0.00000000102,
      -0.00000000238,     0.00000000137,
       0.00000000101,     0.00000000231,
      -0.00000000097,    -0.00000000050,
       0.00000000244,    -0.00000000585,
       0.00000000190,    -0.00000000006,
      -0.00000000031,    -0.00000000138,
       0.00000000116,    -0.00000000005,
      -0.00000000032,    -0.00000000249,
       0.00000000075,     0.00000000051,
       0.00000005481,    -0.00000000181,
       0.00000000078,     0.00000000714,
       0.00000000049,    -0.00000000139,
       0.00000000099,    -0.00000000081,
      -0.00000000019,    -0.00000000239,
       0.00000000370,     0.00000000774,
      -0.00000000163,     0.00000000012,
       0.00000000116,    -0.00000000004,
       0.00000000011,     0.00000000101,
       0.00000003782,     0.00000001583,
       0.00000000062,    -0.00000000083,
      -0.00000000179,     0.00000000015,
      -0.00000000009,    -0.00000000161,
       0.00000000535,    -0.00000004093,
       0.00000000409,     0.00000000561,
       0.00000000246,    -0.00000000002,
      -0.00000001416,    -0.00000001253,
       0.00000000131,     0.00000000010,
      -0.00000000088,     0.00000000137,
       0.00000000095,    -0.00000000046,
      -0.00000006225,    -0.00000026738,
       0.00000002870,     0.00000002422,
       0.00000000505,    -0.00000000115,
      -0.00000000047,     0.00000000065,
       0.00000000469,     0.00000000166,
       0.00000000105,     0.00000000507,
      -0.00000002883,     0.00000003410,
       0.00000000076,    -0.00000000056,
       0.00000000454,     0.00000000937,
      -0.00000000079,    -0.00000000041,
      -0.00000000941,     0.00000000450,
       0.00000000043,    -0.00000000079,
      -0.00000000262,     0.00000000007,
      -0.00000000007,    -0.00000000262,
       0.00000000326,     0.00000000285,
       0.00000000203,    -0.00000000356,
      -0.00000000352,    -0.00000000228,
      -0.00000000154,    -0.00000001909,
      -0.00000000712,    -0.00000000781,
      -0.00000000148,     0.00000000170,
       0.00000000105,    -0.00000000176,
      -0.00000000181,    -0.00000000104,
       0.00000000014,     0.00000000217,
       0.00000000064,    -0.00000000078,
       0.00000000118,     0.00000000009,
       0.00000000009,    -0.00000000117,
      -0.00000000045,    -0.00000000066,
      -0.00000000064,     0.00000000052,
       0.00000006026,     0.00000001166,
      -0.00000000008,    -0.00000000146,
       0.00000000470,     0.00000000019,
       0.00000000019,    -0.00000000481,
       0.00000001598,     0.00000000557,
       0.00000001066,    -0.00000000895,
      -0.00000000063,     0.00000000712,
      -0.00000000332,     0.00000000427,
       0.00000000409,     0.00000000374,
      -0.00000002776,    -0.00000000290,
       0.00000001327,     0.00000000257,
      -0.00000000056,     0.00000000072,
       0.00000000101,     0.00000000010,
       0.00000000967,     0.00000016353,
      -0.00000001511,    -0.00000001218,
      -0.00000000107,     0.00000000081,
       0.00000000295,    -0.00000000238,
      -0.00000000226,     0.00000000139,
       0.00000006175,    -0.00000000094,
      -0.00000002390,     0.00000027693,
      -0.00000002859,    -0.00000003633,
      -0.00000001000,    -0.00000000101,
       0.00000000131,    -0.00000000106,
      -0.00000000068,    -0.00000000115,
       0.00000000158,    -0.00000000068,
      -0.00000000871,     0.00000001913,
       0.00000000181,    -0.00000000502,
      -0.00000000503,    -0.00000000180,
       0.00000005606,    -0.00000015541,
      -0.00000001931,     0.00000000699,
       0.00000000288,     0.00000000102,
      -0.00000001850,     0.00000001946,
       0.00000000336,     0.00000000096,
       0.00000001938,     0.00000001847,
       0.00000000095,    -0.00000000333,
      -0.00000046680,     0.00000076102,
       0.00000010941,     0.00000000434,
      -0.00000003455,    -0.00000002382,
      -0.00000000039,     0.00000000545,
       0.00000000409,     0.00000001949,
       0.00000000156,    -0.00000000149,
       0.00000002037,    -0.00000000368,
      -0.00000000152,    -0.00000000169,
      -0.00000001197,     0.00000004364,
       0.00000000082,    -0.00000000100,
      -0.00000000094,    -0.00000000083,
      -0.00000000873,    -0.00000000584,
       0.00000000145,    -0.00000000281,
       0.00000000219,    -0.00000000086,
      -0.00000000125,    -0.00000000239,
      -0.00000015997,     0.00000021386,
       0.00000001260,    -0.00000001448,
      -0.00000003463,     0.00000001790,
       0.00000000261,     0.00000000268,
       0.00000000140,    -0.00000000117,
       0.00000000118,    -0.00000000029,
      -0.00000001276,     0.00000000090,
       0.00000000009,     0.00000000128,
      -0.00000125792,     0.00000006759,
       0.00000000509,    -0.00000001965,
       0.00000000033,     0.00000001125,
       0.00000000041,     0.00000000238,
      -0.00000000174,     0.00000000136,
      -0.00000002011,     0.00000001609,
      -0.00000000731,     0.00000000044,
      -0.00000000040,    -0.00000000728,
       0.00000008640,     0.00000001041,
      -0.00000000840,     0.00000001634,
       0.00000000097,    -0.00000000084,
      -0.00000000082,    -0.00000000095,
       0.00000003289,    -0.00000002527,
       0.00000000189,     0.00000000193,
      -0.00000005980,     0.00000000647,
       0.00000000025,     0.00000000155,
       0.00000004505,     0.00000005433,
       0.00000000055,     0.00000000109,
       0.00000000001,    -0.00000000104,
      -0.00000000096,    -0.00000000014,
       0.00000000826,    -0.00000000679,
      -0.00000000136,    -0.00000000237,
      -0.00000000058,     0.00000000425,
       0.00000000447,     0.00000000017,
       0.00000019095,    -0.00000002198,
       0.00000000017,     0.00000000114,
       0.00000000145,    -0.00000006640,
       0.00000001295,    -0.00000000774,
      -0.00000002045,    -0.00000000205,
       0.00000000489,     0.00000002551,
       0.00000010294,    -0.00000000523,
       0.00000000053,    -0.00000000678,
      -0.00000000256,    -0.00000000023,
      -0.00000000249,     0.00000003033,
       0.00000000576,     0.00000000048,
      -0.00000000011,     0.00000000122,
      -0.00000010503,    -0.00000000854,
      -0.00000000645,     0.00000007936,
       0.00000002158,    -0.00000000105,
      -0.00000000105,    -0.00000002116,
       0.00000005035,    -0.00000000629,
      -0.00000000019,    -0.00000000645,
      -0.00000000905,     0.00000000230,
       0.00000000229,     0.00000000907,
       0.00000002823,    -0.00000000151,
      -0.00000000190,    -0.00000000015,
      -0.00000000016,     0.00000000199,
      -0.00000005621,     0.00000006177,
       0.00000000148,     0.00000001368,
       0.00000000111,     0.00000000036,
      -0.00000001210,    -0.00000001313,
      -0.00000001306,     0.00000001233,
       0.00000005326,    -0.00000000291,
       0.00000000046,     0.00000000335,
       0.00000000614,     0.00000000031,
       0.00000000034,    -0.00000000612,
      -0.00000101988,    -0.00000043661,
       0.00000001598,     0.00000001868,
      -0.00000000648,    -0.00000003513,
      -0.00000000221,    -0.00000000096,
       0.00000035359,     0.00000004272,
      -0.00000000085,    -0.00000000481,
       0.00000004243,    -0.00000035408,
      -0.00000000485,     0.00000000084,
       0.00000000809,    -0.00000000218,
      -0.00000000196,    -0.00000000726,
      -0.00000001005,    -0.00000000043,
      -0.00000000048,     0.00000001014,
       0.00000004159,    -0.00000006983,
      -0.00000001051,     0.00000000073,
       0.00000000672,     0.00000000374,
      -0.00000000358,    -0.00000000213,
      -0.00000000188,     0.00000000335,
       0.00001602878,     0.00000193045,
      -0.00000004048,    -0.00000022917,
       0.00000000983,     0.00000009886,
       0.00000000308,     0.00000000181,
      -0.00000233145,     0.00000016929,
      -0.00000002577,     0.00000000267,
       0.00000000112,    -0.00000000012,
       0.00000017128,     0.00000232872,
       0.00000000266,     0.00000002585,
      -0.00000000013,    -0.00000000109,
       0.00000003877,     0.00000001402,
       0.00000000027,    -0.00000000103,
       0.00000001404,    -0.00000000629,
      -0.00000000470,    -0.00000000903,
      -0.00000004230,     0.00000000237,
       0.00000000232,     0.00000004150,
      -0.00000000535,    -0.00000000979,
      -0.00013070739,     0.00000425429,
      -0.00000073855,    -0.00000000426,
       0.00000002851,    -0.00000001607,
       0.00000000926,    -0.00000149529,
       0.00000000579,    -0.00000000424,
      -0.00001776005,     0.00000101233,
       0.00000008183,    -0.00000027938,
      -0.00000000118,     0.00000000358,
       0.00000099918,     0.00001785380,
      -0.00000027807,    -0.00000008228,
       0.00000000352,     0.00000000120,
       0.00000000387,    -0.00000000457,
      -0.00000000467,    -0.00000000392,
      -0.00000316686,     0.00000000850,
      -0.00000000001,     0.00000003299,
      -0.00000000416,    -0.00000000123,
      -0.00000000119,     0.00000000480,
      -0.00000002850,     0.00000000190,
      -0.00000002930,     0.00000000193,
      -0.00000000686,     0.00000000055,
       0.00000000047,     0.00000000657,
       0.00000000127,     0.00000000008,
       0.00000000007,     0.00000000181,
      -0.00061824399,     0.00003456915,
       0.00000287998,    -0.00000967896,
      -0.00000004278,     0.00000011825,
      -0.00000144282,    -0.00002588154,
       0.00000040484,     0.00000012132,
      -0.00000000496,    -0.00000000182,
      -0.00000144627,     0.00000010621,
       0.00000000122,    -0.00000001835,
       0.00000008439,    -0.00000033649,
      -0.00000001825,    -0.00000000530,
       0.00000002482,    -0.00000003677,
       0.00000000199,     0.00000000146,
      -0.00000003708,    -0.00000002478,
       0.00000000147,    -0.00000000199,
       0.00002841345,     0.00000165942,
      -0.00000008254,     0.00000045062,
      -0.00000002037,     0.00000033808,
      -0.00000000561,    -0.00000000108,
       0.00000054288,     0.00000007612,
      -0.00000000599,     0.00000001840,
       0.00000007812,    -0.00000068150,
       0.00000001840,     0.00000000604,
       0.00000000113,     0.00000000008,
       0.00000000003,    -0.00000000102,
       0.00000000186,    -0.00000002270,
      -0.00000027145,     0.00000002464,
       0.00000000116,    -0.00000000431,
      -0.00000002433,    -0.00000027149,
       0.00000000434,     0.00000000117,
      -0.00000001183,     0.00000001398,
      -0.00000001375,    -0.00000001142,
       0.00000023526,     0.00000002277,
      -0.00000000026,    -0.00000000685,
       0.00000023639,    -0.00000001971,
      -0.00000000102,     0.00000000376,
       0.00000001974,     0.00000023674,
      -0.00000000377,    -0.00000000103,
      -0.00000000888,     0.00000000049,
      -0.00000000051,    -0.00000000882,
      -0.00000000646,     0.00000000079,
      -0.00000000079,    -0.00000000645,
       0.00000000069,    -0.00000000118,
       0.00000000118,     0.00000000069,
      -0.00000000143,    -0.00000000420,
       0.00000000420,    -0.00000000143,
       0.00000104352,     0.00000196894,
      -0.00000010903,     0.00000004069,
      -0.00000000119,    -0.00000000259,
       0.00000149540,     0.00000705217,
      -0.00000019722,     0.00000004026,
      -0.00000000106,    -0.00000000239,
      -0.00000707661,     0.00000146249,
      -0.00000004201,    -0.00000019846,
       0.00000000243,    -0.00000000114,
       0.00000029419,     0.00000026642,
       0.00000000764,    -0.00000000919,
       0.00000000003,    -0.00000000019,
      -0.00000028764,     0.00000037243,
       0.00000000699,     0.00000000685,
       0.00000000026,    -0.00000000003,
       0.00000015054,    -0.00000023605,
       0.00000003516,    -0.00000001053,
       0.00000000063,     0.00000000060,
      -0.00000088513,     0.00000035686,
       0.00000001884,    -0.00000005490,
       0.00000000112,     0.00000000099,
      -0.00000035443,    -0.00000088583,
       0.00000005558,     0.00000001873,
      -0.00000000099,     0.00000000111,
       0.00000016809,     0.00000034128,
       0.00000003197,    -0.00000000327,
       0.00000000007,    -0.00000000159,
      -0.00000033678,     0.00000016786,
       0.00000000361,     0.00000003205,
       0.00000000160,     0.00000000006,
       0.00000008930,    -0.00000036456,
       0.00000004861,    -0.00000002736,
       0.00000000241,     0.00000000169,
      -0.00000164641,     0.00000224569,
      -0.00000015227,    -0.00000014580,
       0.00000000533,    -0.00000000659,
      -0.00000224892,    -0.00000165098,
       0.00000014680,    -0.00000015315,
       0.00000000666,     0.00000000537,
       0.00000006807,    -0.00000036399,
      -0.00000002602,     0.00000001087,
       0.00000000096,     0.00000000088,
       0.00000036041,     0.00000006684,
      -0.00000001089,    -0.00000002624,
      -0.00000000086,     0.00000000094,
      -0.00000252652,    -0.00000211648,
       0.00000028970,    -0.00000048484,
       0.00000003967,     0.00000002284,
       0.00002273842,    -0.00001240093,
       0.00000106641,     0.00000175561,
      -0.00000007029,     0.00000004502,
       0.00001238631,     0.00002277210,
      -0.00000175946,     0.00000106556,
      -0.00000004513,    -0.00000007055,
       0.00000085966,     0.00000005255,
       0.00000003323,    -0.00000005678,
      -0.00000000173,    -0.00000000245,
      -0.00000007460,     0.00000084544,
       0.00000005746,     0.00000003028,
       0.00000000258,    -0.00000000173,
      -0.00000212973,    -0.00000532872,
       0.00000141350,    -0.00000025992,
       0.00000000593,     0.00000021095,
      -0.00000001794,    -0.00000000322,
      -0.00000205958,     0.00000084938,
      -0.00000011091,    -0.00000052849,
       0.00000008013,    -0.00000000590,
      -0.00000000075,     0.00000000671,
       0.00000086047,    -0.00000018099,
      -0.00000004153,     0.00000029184,
      -0.00000005411,    -0.00000001887,
       0.00000000346,    -0.00000000686,
       0.00000036732,    -0.00000050990,
       0.00000023681,     0.00000010514,
      -0.00000002257,     0.00000005068,
      -0.00000000673,    -0.00000000362,
       0.00000021432,    -0.00000029418,
      -0.00000004010,    -0.00000002985,
      -0.00000000145,     0.00000000225,
       0.00000029096,     0.00000022139,
       0.00000002892,    -0.00000004048,
      -0.00000000223,    -0.00000000160,
       0.00001208759,     0.00002787314,
      -0.00000791765,     0.00000155130,
      -0.00000002160,    -0.00000112436,
       0.00000010067,     0.00000001795,
       0.00001238609,     0.00000031176,
      -0.00000027085,     0.00000221119,
      -0.00000028997,    -0.00000002820,
       0.00000000405,    -0.00000002421,
       0.00008705486,     0.00002324658,
       0.00000244653,     0.00000295819,
       0.00000001806,     0.00000030372,
      -0.00000002491,     0.00000002093,
      -0.00000000251,    -0.00000000200,
      -0.00002412000,     0.00009312439,
      -0.00000529732,     0.00000178456,
      -0.00000012024,    -0.00000044398,
       0.00000003904,     0.00000000616,
      -0.00000000158,     0.00000000327,
      -0.00000052786,     0.00000073883,
       0.00000010065,     0.00000006926,
       0.00000000330,    -0.00000000633,
      -0.00000077494,    -0.00000052663,
      -0.00000006399,     0.00000010120,
       0.00000000635,     0.00000000354,
       0.00000035886,     0.00000032726,
       0.00000002257,     0.00000007096,
      -0.00000117541,     0.00000916940,
      -0.00000052013,     0.00000019073,
      -0.00000002335,    -0.00000002554,
       0.00000000118,    -0.00000000112,
       0.00012898692,     0.00003288876,
       0.00000308734,     0.00000595344,
      -0.00000030968,     0.00000030802,
      -0.00000001343,    -0.00000001319,
      -0.00003291435,     0.00012902526,
      -0.00000596102,     0.00000308827,
      -0.00000030839,    -0.00000031042,
       0.00000001323,    -0.00000001348,
      -0.00000001030,    -0.00000000217,
      -0.00000000170,    -0.00000000087,
       0.00000002773,     0.00000001732,
      -0.00000000075,     0.00000000148,
       0.00000207013,     0.00000111977,
       0.00000006551,     0.00000024101,
       0.00000199271,    -0.00000104035,
       0.00000012602,     0.00000004257,
       0.00000000027,     0.00000000856,
       0.00016147141,     0.00004115275,
       0.00000387034,     0.00000745837,
      -0.00000038781,     0.00000038598,
      -0.00000001683,    -0.00000001652,
       0.00000000060,    -0.00000000041,
      -0.00004120135,     0.00016154183,
      -0.00000746762,     0.00000386934,
      -0.00000038655,    -0.00000038895,
       0.00000001658,    -0.00000001690,
       0.00000000042,     0.00000000060,
      -0.00000002547,    -0.00000000103,
       0.00000000068,     0.00000000051,
      -0.00000000969,    -0.00000001931,
      -0.00000000061,     0.00000000061,
       0.00000000057,     0.00000000022,
      -0.00000000002,     0.00000000001,
       0.00000000143,    -0.00000000380,
       0.00000000007,     0.00000000015,
       0.00000000839,    -0.00000000581,
       0.00000000012,     0.00000000049,
       0.00000000563,     0.00000000832,
      -0.00000000047,     0.00000000010,
       0.00000000158,    -0.00000000270,
       0.00000000275,     0.00000000157,
       0.00000000108,    -0.00000000157,
       0.00000000157,     0.00000000108,
      -0.00000000002,     0.00000000000,
      -0.00000000009,    -0.00000000043,
       0.00000000002,     0.00000000011,
       0.00000000012,    -0.00000000002,
       0.00000000000,    -0.00000000002,
      -0.00000000048,     0.00000000010,
       0.00000000001,    -0.00000000004,
      -0.00000000004,    -0.00000000001,
       0.00000000000,     0.00000000002,
       0.00000000131,     0.00000000026,
       0.00000000000,    -0.00000000007,
       0.00000000015,     0.00000000000,
       0.00000000000,    -0.00000000016,
       0.00000000000,    -0.00000000005,
      -0.00000000005,    -0.00000000001,
       0.00000000000,     0.00000000004,
       0.00000000011,     0.00000000000,
       0.00000000026,     0.00000000016,
      -0.00000000016,     0.00000000026,
       0.00000000009,     0.00000000007,
      -0.00000000007,     0.00000000009,
       0.00000000001,    -0.00000000005,
      -0.00000000005,    -0.00000000001,
      -0.00000000001,     0.00000000005,
       0.00000000255,     0.00000000055,
       0.00000000001,    -0.00000000009,
       0.00000000016,    -0.00000000003,
       0.00000000003,    -0.00000000028,
       0.00000000001,    -0.00000000027,
      -0.00000000003,     0.00000000005,
      -0.00000000023,    -0.00000000022,
      -0.00000000002,     0.00000000004,
       0.00000000002,    -0.00000000003,
      -0.00000000008,     0.00000000013,
      -0.00000000001,     0.00000000004,
       0.00000000005,     0.00000000001,
       0.00000000010,     0.00000000003,
       0.00000000005,    -0.00000000002,
      -0.00000000005,    -0.00000000001,
      -0.00000000017,     0.00000000010,
       0.00000000003,     0.00000000002,
       0.00000000032,    -0.00000000022,
      -0.00000000509,    -0.00000000399,
      -0.00000000028,    -0.00000000087,
       0.00000000401,    -0.00000000505,
       0.00000000087,    -0.00000000028,
       0.00000000047,     0.00000000113,
      -0.00000000113,     0.00000000047,
      -0.00000000083,    -0.00000000173,
       0.00000000173,    -0.00000000083,
       0.00000000030,     0.00000000137,
      -0.00000000137,     0.00000000030,
       0.00000000000,    -0.00000000133,
      -0.00000000001,    -0.00000000141,
       0.00000000009,    -0.00000000051,
       0.00000000176,     0.00000000000,
       0.00000000000,     0.00000000176,
      -0.00000000042,     0.00000000011,
      -0.00000000148,     0.00000000000,
       0.00000000000,    -0.00000000147,
       0.00000000155,    -0.00000000013,
      -0.00000000116,    -0.00000000174,
      -0.00000000022,     0.00000000019,
       0.00000000221,    -0.00000000147,
      -0.00000000025,    -0.00000000028,
       0.00000000113,     0.00000000060,
      -0.00000000126,     0.00000000229,
      -0.00000000131,    -0.00000000115,
       0.00000000114,    -0.00000000021,
       0.00000000084,    -0.00000000004,
      -0.00000000005,     0.00000000220,
      -0.00000000039,    -0.00000000026,
       0.00000000025,    -0.00000000083,
       0.00000000027,     0.00000000051,
       0.00000008948,     0.00000002298,
       0.00000000027,    -0.00000000104,
       0.00000000103,     0.00000000026,
      -0.00000000026,     0.00000000102,
       0.00000000157,     0.00000000040,
      -0.00000000040,     0.00000000157,
       0.00000000000,    -0.00000000136,
       0.00000000086,     0.00000000000,
       0.00000000000,     0.00000000322,
       0.00000000111,     0.00000000080,
       0.00000000084,     0.00000000000,
       0.00000000000,     0.00000000084,
      -0.00000000057,    -0.00000000042,
      -0.00000000124,     0.00000000000,
       0.00000000000,    -0.00000000124,
       0.00000000029,    -0.00000000024,
      -0.00000000121,     0.00000000103,
      -0.00000000058,    -0.00000000027,
      -0.00000000026,     0.00000000057,
      -0.00000000065,     0.00000000000,
       0.00000000000,     0.00000000064,
       0.00000000057,     0.00000000000,
       0.00000000000,    -0.00000000057,
       0.00000000385,     0.00000000007,
       0.00000000498,    -0.00000029223,
      -0.00000000029,     0.00000000037,
      -0.00000000106,     0.00000000320,
       0.00000000086,     0.00000000111,
      -0.00000000003,     0.00000000065,
      -0.00000000603,    -0.00000000011,
      -0.00000000366,     0.00000020239,
       0.00000000083,    -0.00000000360,
       0.00000000030,    -0.00000000092,
      -0.00000000078,     0.00000000000,
       0.00000000000,     0.00000000078,
       0.00000000081,    -0.00000000002,
       0.00000000071,     0.00000000000,
       0.00000000000,    -0.00000000071,
       0.00000000000,    -0.00000000206,
       0.00000000000,     0.00000000114,
      -0.00000000075,     0.00000000004,
       0.00000000080,     0.00000000000,
       0.00000000000,    -0.00000000080,
       0.00000000082,    -0.00000000003,
      -0.00000000074,     0.00000000000,
       0.00000000000,     0.00000000074,
       0.00000000000,    -0.00000000117,
       0.00000000000,     0.00000000113,
      -0.00000000024,     0.00000000229,
       0.00000000002,     0.00000000068,
       0.00000000067,    -0.00000000002,
       0.00000000028,    -0.00000000162,
      -0.00000000002,    -0.00000000061,
      -0.00000000061,     0.00000000002,
       0.00000000054,     0.00000000000,
       0.00000000000,    -0.00000000054,
       0.00000000037,    -0.00000000033,
      -0.00000000101,    -0.00000000123,
       0.00000000288,    -0.00000000182,
      -0.00000001395,    -0.00000000443,
      -0.00000000083,    -0.00000000186,
       0.00000000441,    -0.00000001393,
       0.00000000186,    -0.00000000080,
       0.00000000008,    -0.00000000242,
       0.00000000241,     0.00000000009,
      -0.00000000104,    -0.00000000065,
       0.00000000065,    -0.00000000104,
       0.00000000033,    -0.00000000072,
       0.00000000072,     0.00000000033,
      -0.00000000198,    -0.00000000190,
       0.00000000311,    -0.00000000365,
      -0.00000000122,     0.00000000019,
      -0.00000000048,    -0.00000000835,
      -0.00000000079,    -0.00000000157,
       0.00000000089,    -0.00000000201,
      -0.00000000060,    -0.00000000135,
      -0.00000000110,     0.00000000000,
       0.00000000000,    -0.00000000111,
      -0.00000000148,     0.00000000107,
       0.00000000917,     0.00000001409,
      -0.00000000102,     0.00000000068,
       0.00000000885,    -0.00000000002,
       0.00000000019,     0.00000000889,
       0.00000009508,     0.00000002439,
       0.00000000028,    -0.00000000107,
       0.00000000103,     0.00000000026,
      -0.00000000026,     0.00000000102,
       0.00000000157,     0.00000000040,
      -0.00000000040,     0.00000000157,
       0.00000000090,    -0.00000000163,
       0.00000000134,    -0.00000000083,
      -0.00000000161,    -0.00000000088,
       0.00000000275,    -0.00000000498,
      -0.00000000040,    -0.00000000024,
       0.00000000182,    -0.00000000038,
       0.00000000036,    -0.00000000229,
       0.00000000191,    -0.00000000002,
      -0.00000000069,    -0.00000007059,
       0.00000000066,    -0.00000000096,
      -0.00000000036,    -0.00000000176,
      -0.00000004293,     0.00000000040,
       0.00000000462,     0.00000055402,
      -0.00000000035,     0.00000000186,
       0.00000000481,     0.00000001682,
       0.00000000021,    -0.00000000230,
      -0.00000000193,     0.00000002950,
       0.00000000034,     0.00000000144,
      -0.00000000123,     0.00000000068,
       0.00000000144,     0.00000000266,
      -0.00000000053,     0.00000000001,
      -0.00000000001,    -0.00000000053,
       0.00000000020,     0.00000000032,
       0.00000000032,    -0.00000000020,
       0.00000000143,    -0.00000000088,
       0.00000000080,     0.00000000045,
       0.00000000043,    -0.00000000080,
      -0.00000000224,    -0.00000000123,
      -0.00000000011,     0.00000000016,
       0.00000008423,    -0.00000015319,
      -0.00000001130,    -0.00000000783,
      -0.00000000352,     0.00000000000,
      -0.00000000001,     0.00000000355,
       0.00000005254,     0.00000002891,
       0.00000000269,    -0.00000000389,
      -0.00000069325,     0.00000126118,
       0.00000009325,     0.00000006447,
       0.00000003563,    -0.00000001109,
      -0.00000000160,     0.00000000106,
       0.00000000998,    -0.00000004944,
      -0.00000000084,    -0.00000000188,
       0.00000000021,    -0.00000000051,
      -0.00000000027,     0.00000000038,
       0.00000000000,    -0.00000000067,
      -0.00000000300,    -0.00000000001,
      -0.00000000002,     0.00000000599,
      -0.00000000122,     0.00000000029,
       0.00000000071,     0.00000000029,
      -0.00000000071,     0.00000000039,
       0.00000000055,     0.00000000100,
       0.00000000028,     0.00000000046,
       0.00000000045,    -0.00000000029,
       0.00000000488,    -0.00000000300,
      -0.00000000046,    -0.00000000086,
      -0.00000000039,    -0.00000000022,
      -0.00000000020,     0.00000000039,
       0.00000000459,     0.00000000736,
       0.00000000122,    -0.00000000063,
      -0.00000004294,     0.00000002692,
       0.00000000365,     0.00000000710,
       0.00000000698,     0.00000000253,
       0.00000000005,    -0.00000000049,
       0.00000000502,    -0.00000000711,
      -0.00000000044,    -0.00000000063,
       0.00000000023,    -0.00000000043,
      -0.00000000182,     0.00000000008,
       0.00000000192,     0.00000000270,
       0.00000000036,     0.00000000026,
      -0.00000000206,    -0.00000000038,
      -0.00000000013,     0.00000000181,
      -0.00000000215,    -0.00000000044,
      -0.00000000055,     0.00000000248,
       0.00000000082,    -0.00000000005,
       0.00000000017,     0.00000000256,
       0.00000000061,     0.00000000002,
      -0.00000000865,     0.00000000062,
      -0.00000000005,     0.00000000200,
       0.00000000132,     0.00000000181,
       0.00000000214,    -0.00000000116,
      -0.00000000012,    -0.00000000041,
       0.00000000033,    -0.00000000020,
      -0.00000000097,    -0.00000000154,
      -0.00000000283,     0.00000000177,
       0.00000000025,     0.00000000052,
      -0.00000000153,    -0.00000000090,
      -0.00000000079,     0.00000000153,
      -0.00000000137,    -0.00000000076,
      -0.00000000080,     0.00000000144,
      -0.00000000096,    -0.00000000006,
       0.00000000007,     0.00000000098,
      -0.00000000119,    -0.00000000001,
       0.00000000000,     0.00000000130,
       0.00000000034,     0.00000000017,
      -0.00000000035,     0.00000000072,
       0.00000000025,     0.00000000015,
      -0.00000000139,    -0.00000000066,
      -0.00000000037,     0.00000000059,
       0.00000000003,     0.00000000087,
       0.00000000086,    -0.00000000006,
      -0.00000000006,    -0.00000000122,
      -0.00000000273,     0.00000000013,
       0.00000000000,     0.00000000062,
      -0.00000000052,    -0.00000000088,
      -0.00000000083,     0.00000000055,
      -0.00000000046,    -0.00000000073,
      -0.00000000092,     0.00000000057,
      -0.00000000060,    -0.00000000037,
      -0.00000000029,     0.00000000060,
      -0.00000000083,    -0.00000000046,
      -0.00000000041,     0.00000000074,
      -0.00000000086,    -0.00000000001,
      -0.00000000001,     0.00000000080,
      -0.00000000030,    -0.00000000043,
      -0.00000000022,     0.00000000044,
       0.00000000044,     0.00000000021,
       0.00000000032,    -0.00000000065,
      -0.00000000171,    -0.00000000084,
      -0.00000000028,     0.00000000045,
      -0.00000000002,    -0.00000000075,
      -0.00000000073,     0.00000000006,
      -0.00000000004,    -0.00000000074,
      -0.00000000049,     0.00000000003,
      -0.00000000021,    -0.00000000036,
      -0.00000000033,     0.00000000023,
      -0.00000000059,     0.00000000036,
      -0.00000000038,    -0.00000000025,
      -0.00000000017,     0.00000000038,
      -0.00000000056,    -0.00000000032,
      -0.00000000023,     0.00000000042,
       0.00000000014,    -0.00000000026,
      -0.00000000026,     0.00000000061,
       0.00000000033,     0.00000000017,
      -0.00000000078,    -0.00000000111,
      -0.00000000050,     0.00000000032,
       0.00000000023,    -0.00000000044,
      -0.00000000044,    -0.00000000020,
      -0.00000000036,     0.00000000022,
      -0.00000000013,    -0.00000000133,
      -0.00000000068,     0.00000000002,
      -0.00000000016,     0.00000000037,
       0.00000000071,    -0.00000000168,
      -0.00000000105,    -0.00000000053,
       0.00000000047,    -0.00000000004,
      -0.00000000006,    -0.00000000046,
      -0.00000003151,     0.00000002496,
       0.00000001598,     0.00000002419,
      -0.00000000035,    -0.00000000016,
       0.00000000123,     0.00000000156,
       0.00000000115,    -0.00000000076,
       0.00000024423,    -0.00000019344,
      -0.00000012364,    -0.00000018711,
       0.00000000238,     0.00000000108,
       0.00000000067,    -0.00000000137,
       0.00000000098,    -0.00000000237,
      -0.00000000140,    -0.00000000072,
      -0.00000000991,     0.00000000570,
      -0.00000000210,    -0.00000000128,
       0.00000000008,    -0.00000000029,
       0.00000000726,     0.00000001403,
      -0.00000000175,     0.00000000290,
      -0.00000000045,    -0.00000000010,
      -0.00000006064,     0.00000005656,
      -0.00000001369,    -0.00000000773,
       0.00000000068,    -0.00000000108,
      -0.00000005685,    -0.00000006062,
       0.00000000770,    -0.00000001381,
       0.00000000109,     0.00000000067,
      -0.00000000855,     0.00000001151,
      -0.00000001153,    -0.00000000853,
      -0.00000000131,     0.00000000103,
      -0.00000000037,    -0.00000000015,
      -0.00000000104,    -0.00000000131,
       0.00000000015,    -0.00000000037,
       0.00000000009,     0.00000000052,
      -0.00000000052,     0.00000000009,
      -0.00000001791,    -0.00000000716,
       0.00000000041,    -0.00000000102,
       0.00000000745,    -0.00000001765,
       0.00000000106,     0.00000000048,
       0.00000002797,     0.00000001065,
      -0.00000000057,     0.00000000168,
       0.00000000075,     0.00000000029,
       0.00000000060,     0.00000000071,
      -0.00000003431,    -0.00000001157,
       0.00000000066,    -0.00000000198,
      -0.00000000085,    -0.00000000019,
       0.00000000131,     0.00000000055,
      -0.00000000029,     0.00000000096,
       0.00000001841,     0.00000001173,
      -0.00000000068,     0.00000000100,
      -0.00000000147,    -0.00000000054,
       0.00000000039,    -0.00000000137,
       0.00000000077,     0.00000000051,
      -0.00000000045,     0.00000000075,
       0.00000001206,    -0.00000003190,
       0.00000000196,     0.00000000080,
       0.00000000016,    -0.00000000101,
       0.00000000038,    -0.00000000095,
       0.00000005817,     0.00000001611,
       0.00000000756,    -0.00000002553,
      -0.00000000086,     0.00000000355,
       0.00000000154,     0.00000000041,
      -0.00000000008,    -0.00000000152,
       0.00000000112,     0.00000000125,
       0.00000000041,    -0.00000000064,
      -0.00000000023,     0.00000000085,
      -0.00000006484,    -0.00000001811,
      -0.00000000429,     0.00000002776,
       0.00000000105,    -0.00000000378,
      -0.00000000161,    -0.00000000025,
       0.00000000270,     0.00000000084,
       0.00000000035,    -0.00000000119,
      -0.00000000043,     0.00000000209,
       0.00000000090,     0.00000000025,
       0.00000003274,     0.00000001886,
       0.00000000430,    -0.00000001516,
      -0.00000000111,     0.00000000180,
       0.00000000084,     0.00000000025,
      -0.00000000278,    -0.00000000084,
      -0.00000000020,     0.00000000120,
       0.00000000062,    -0.00000000261,
      -0.00000000110,    -0.00000000014,
      -0.00000000999,    -0.00000001092,
       0.00000000137,     0.00000000082,
      -0.00000000074,     0.00000000134,
      -0.00000000079,    -0.00000000028,
       0.00000001911,    -0.00000005636,
      -0.00000002904,    -0.00000000247,
       0.00000000352,     0.00000000131,
       0.00000000020,    -0.00000000181,
       0.00000000055,    -0.00000000177,
      -0.00000000088,    -0.00000000017,
      -0.00000001279,    -0.00000000245,
       0.00000000025,    -0.00000000169,
       0.00000011896,     0.00000002267,
       0.00000000937,    -0.00000005119,
      -0.00000000118,     0.00000000737,
       0.00000000314,     0.00000000049,
      -0.00000000023,    -0.00000000266,
      -0.00000000128,     0.00000000035,
       0.00000000206,     0.00000000219,
       0.00000000068,    -0.00000000116,
       0.00000001181,     0.00000000259,
      -0.00000000035,     0.00000000159,
      -0.00000012075,    -0.00000002706,
      -0.00000000472,     0.00000005111,
       0.00000000162,    -0.00000000714,
      -0.00000000301,    -0.00000000028,
       0.00000000546,     0.00000000121,
       0.00000000045,    -0.00000000237,
      -0.00000000057,     0.00000000442,
       0.00000000188,     0.00000000030,
       0.00000000038,    -0.00000000069,
       0.00000005722,     0.00000002952,
       0.00000000601,    -0.00000002605,
      -0.00000000178,     0.00000000318,
       0.00000000146,     0.00000000038,
      -0.00000000520,    -0.00000000127,
      -0.00000000023,     0.00000000222,
       0.00000000093,    -0.00000000493,
      -0.00000000206,    -0.00000000014,
      -0.00000001683,    -0.00000001669,
       0.00000000095,    -0.00000000086,
       0.00000000241,     0.00000000130,
       0.00000000027,    -0.00000000111,
      -0.00000000117,     0.00000000236,
       0.00000000107,     0.00000000024,
      -0.00000000069,    -0.00000000072,
       0.00000000068,    -0.00000000069,
      -0.00000001741,    -0.00000000900,
       0.00000000068,    -0.00000000107,
      -0.00000000110,    -0.00000000032,
      -0.00000000358,     0.00000001174,
      -0.00000000139,    -0.00000000046,
      -0.00000000110,    -0.00000000004,
      -0.00000000016,     0.00000000127,
       0.00000002955,    -0.00000009702,
      -0.00000004954,    -0.00000000209,
       0.00000000618,     0.00000000213,
       0.00000000023,    -0.00000000315,
       0.00000000186,    -0.00000000029,
       0.00000000074,    -0.00000000326,
      -0.00000000156,    -0.00000000018,
      -0.00000002574,    -0.00000000306,
      -0.00000000096,     0.00000001086,
       0.00000000028,    -0.00000000347,
      -0.00000000145,    -0.00000000006,
       0.00000000070,     0.00000000064,
      -0.00000000030,    -0.00000000106,
       0.00000023954,     0.00000002827,
       0.00000000875,    -0.00000010108,
      -0.00000002264,    -0.00000000030,
      -0.00000000139,     0.00000001513,
       0.00000000635,     0.00000000038,
      -0.00000000001,    -0.00000000141,
      -0.00000000053,    -0.00000000454,
      -0.00000000216,     0.00000000069,
       0.00000000370,     0.00000000374,
       0.00000000113,    -0.00000000206,
       0.00000002144,     0.00000000354,
      -0.00000000050,     0.00000000296,
       0.00000000124,     0.00000000003,
      -0.00000000141,    -0.00000000031,
      -0.00000022146,    -0.00000003769,
      -0.00000000315,     0.00000009257,
       0.00000001948,    -0.00000000229,
       0.00000000232,    -0.00000001327,
      -0.00000000555,    -0.00000000019,
       0.00000000013,     0.00000000115,
       0.00000001094,     0.00000000157,
       0.00000000043,    -0.00000000465,
      -0.00000000103,     0.00000000000,
      -0.00000000061,     0.00000000921,
       0.00000000385,     0.00000000024,
      -0.00000000846,    -0.00000000387,
       0.00000000060,    -0.00000000120,
       0.00000000111,     0.00000000027,
      -0.00000000009,     0.00000000096,
       0.00000009829,     0.00000004484,
       0.00000000786,    -0.00000004391,
      -0.00000000277,     0.00000000552,
       0.00000000249,     0.00000000053,
      -0.00000000960,    -0.00000000179,
      -0.00000000018,     0.00000000404,
       0.00000000130,    -0.00000000917,
      -0.00000000380,    -0.00000000005,
      -0.00000002793,    -0.00000002494,
      -0.00000000574,     0.00000001422,
       0.00000000146,    -0.00000000145,
      -0.00000000076,    -0.00000000035,
       0.00000000415,     0.00000000199,
       0.00000000036,    -0.00000000188,
      -0.00000000180,     0.00000000409,
       0.00000000183,     0.00000000032,
      -0.00000000115,    -0.00000000108,
       0.00000000102,    -0.00000000114,
      -0.00000000017,     0.00000000089,
       0.00000000090,     0.00000000019,
      -0.00000002471,    -0.00000001176,
       0.00000000093,    -0.00000000157,
      -0.00000000106,    -0.00000000018,
       0.00000000025,    -0.00000000159,
      -0.00000000175,    -0.00000000044,
      -0.00000000546,     0.00000001990,
       0.00000001008,     0.00000000004,
      -0.00000000236,    -0.00000000074,
      -0.00000000002,     0.00000000119,
      -0.00000000217,     0.00000000002,
      -0.00000000016,     0.00000000242,
       0.00000000101,    -0.00000000002,
       0.00000004434,    -0.00000016190,
      -0.00000008197,    -0.00000000036,
       0.00000000550,     0.00000002099,
       0.00000001054,     0.00000000339,
       0.00000000022,    -0.00000000534,
      -0.00000000137,     0.00000000031,
       0.00000000391,    -0.00000000057,
      -0.00000000022,    -0.00000000133,
       0.00000000088,    -0.00000000591,
      -0.00000000272,    -0.00000000012,
      -0.00000005095,    -0.00000000293,
      -0.00000000009,     0.00000002113,
       0.00000000018,    -0.00000000704,
      -0.00000000290,     0.00000000010,
       0.00000000122,     0.00000000103,
      -0.00000000058,    -0.00000000182,
      -0.00000000066,     0.00000000042,
       0.00000047434,     0.00000002687,
       0.00000000056,    -0.00000019671,
      -0.00000004287,     0.00000000372,
      -0.00000000107,     0.00000003070,
       0.00000001265,    -0.00000000026,
      -0.00000000030,    -0.00000000273,
      -0.00000000112,    -0.00000000752,
      -0.00000000356,     0.00000000127,
       0.00000000051,     0.00000000086,
       0.00000000644,     0.00000000631,
       0.00000000184,    -0.00000000356,
      -0.00000000096,    -0.00000000014,
       0.00000003819,     0.00000000421,
      -0.00000000041,    -0.00000001572,
      -0.00000000064,     0.00000000545,
       0.00000000225,    -0.00000000007,
      -0.00000000270,    -0.00000000040,
      -0.00000000006,     0.00000000115,
      -0.00000000012,    -0.00000000172,
      -0.00000039946,    -0.00000004629,
       0.00000000362,     0.00000016475,
       0.00000003397,    -0.00000000611,
       0.00000000298,    -0.00000002438,
      -0.00000001007,     0.00000000019,
       0.00000000037,     0.00000000207,
       0.00000002168,     0.00000000166,
       0.00000000009,    -0.00000000903,
      -0.00000000196,     0.00000000018,
      -0.00000000027,     0.00000001889,
       0.00000000777,    -0.00000000016,
      -0.00000000016,    -0.00000000169,
      -0.00000001387,    -0.00000000549,
       0.00000000091,    -0.00000000205,
      -0.00000000091,    -0.00000000014,
       0.00000000200,     0.00000000034,
      -0.00000000011,     0.00000000179,
       0.00000016603,     0.00000006544,
       0.00000000929,    -0.00000007245,
      -0.00000000419,     0.00000000946,
       0.00000000418,     0.00000000068,
      -0.00000001750,    -0.00000000225,
       0.00000000011,     0.00000000725,
       0.00000000150,    -0.00000000028,
       0.00000000154,    -0.00000001682,
      -0.00000000689,     0.00000000026,
       0.00000000027,     0.00000000141,
      -0.00000000081,    -0.00000000036,
       0.00000000029,    -0.00000000077,
      -0.00000004576,    -0.00000003629,
      -0.00000000762,     0.00000002209,
       0.00000000218,    -0.00000000241,
      -0.00000000119,    -0.00000000049,
       0.00000000709,     0.00000000294,
       0.00000000044,    -0.00000000313,
      -0.00000000268,     0.00000000698,
       0.00000000305,     0.00000000038,
       0.00000000928,     0.00000001372,
      -0.00000000191,    -0.00000000158,
      -0.00000000033,     0.00000000095,
       0.00000000150,    -0.00000000189,
      -0.00000000093,    -0.00000000032,
      -0.00000000023,     0.00000000141,
       0.00000000143,     0.00000000026,
      -0.00000003285,    -0.00000001418,
      -0.00000000111,     0.00000002128,
       0.00000000121,    -0.00000000216,
      -0.00000000145,    -0.00000000022,
       0.00000000034,    -0.00000000248,
      -0.00000000120,     0.00000000016,
      -0.00000000265,    -0.00000000055,
       0.00000000008,     0.00000000136,
      -0.00000000822,     0.00000003058,
       0.00000001636,    -0.00000000048,
      -0.00000000391,    -0.00000000116,
       0.00000000000,     0.00000000196,
      -0.00000000418,     0.00000000019,
       0.00000000020,     0.00000000165,
      -0.00000000004,     0.00000000451,
       0.00000000187,    -0.00000000015,
       0.00000006391,    -0.00000025948,
      -0.00000013032,     0.00000000388,
       0.00000001000,     0.00000003286,
       0.00000001740,     0.00000000526,
       0.00000000014,    -0.00000000877,
      -0.00000000222,     0.00000000059,
       0.00000000784,    -0.00000000111,
      -0.00000000057,    -0.00000000279,
       0.00000000083,    -0.00000001054,
      -0.00000000466,     0.00000000012,
       0.00000000019,     0.00000000109,
      -0.00000009915,    -0.00000000057,
       0.00000000277,     0.00000004048,
       0.00000000860,    -0.00000000150,
      -0.00000000029,    -0.00000001417,
      -0.00000000574,     0.00000000058,
       0.00000000027,     0.00000000120,
       0.00000000206,     0.00000000156,
       0.00000000061,    -0.00000000099,
      -0.00000000110,    -0.00000000303,
      -0.00000000108,     0.00000000074,
       0.00000092349,     0.00000000443,
      -0.00000002623,    -0.00000037696,
      -0.00000008009,     0.00000001407,
       0.00000000082,     0.00000006159,
       0.00000002499,    -0.00000000225,
      -0.00000000105,    -0.00000000528,
      -0.00000000215,    -0.00000001195,
      -0.00000000565,     0.00000000222,
       0.00000000086,     0.00000000135,
       0.00000001083,     0.00000001038,
       0.00000000295,    -0.00000000591,
      -0.00000000157,    -0.00000000020,
       0.00000006656,     0.00000000347,
      -0.00000000222,    -0.00000002700,
      -0.00000000060,     0.00000000988,
       0.00000000402,    -0.00000000033,
      -0.00000000021,    -0.00000000081,
      -0.00000000515,    -0.00000000040,
       0.00000000005,     0.00000000212,
      -0.00000000031,    -0.00000000366,
      -0.00000000146,     0.00000000016,
      -0.00000070728,    -0.00000004127,
       0.00000002236,     0.00000028774,
       0.00000005816,    -0.00000001392,
       0.00000000290,    -0.00000004418,
      -0.00000001800,     0.00000000132,
       0.00000000085,     0.00000000363,
       0.00000004249,     0.00000000085,
      -0.00000000110,    -0.00000001737,
      -0.00000000369,     0.00000000066,
       0.00000000109,     0.00000003807,
       0.00000001545,    -0.00000000139,
      -0.00000000061,    -0.00000000328,
      -0.00000002214,    -0.00000000731,
      -0.00000000076,     0.00000000954,
       0.00000000131,    -0.00000000344,
      -0.00000000150,    -0.00000000016,
       0.00000000358,     0.00000000036,
      -0.00000000007,    -0.00000000148,
      -0.00000000005,     0.00000000327,
       0.00000000132,    -0.00000000013,
       0.00000027531,     0.00000009020,
       0.00000000871,    -0.00000011687,
      -0.00000002360,     0.00000000377,
      -0.00000000607,     0.00000001599,
       0.00000000689,     0.00000000076,
      -0.00000000018,    -0.00000000139,
      -0.00000003144,    -0.00000000216,
       0.00000000092,     0.00000001282,
       0.00000000260,    -0.00000000063,
       0.00000000117,    -0.00000003038,
      -0.00000001230,     0.00000000113,
       0.00000000062,     0.00000000248,
      -0.00000000066,     0.00000000084,
      -0.00000000136,    -0.00000000051,
       0.00000000040,    -0.00000000131,
      -0.00000007423,    -0.00000005115,
      -0.00000000943,     0.00000003352,
       0.00000000317,    -0.00000000401,
      -0.00000000185,    -0.00000000065,
       0.00000001192,     0.00000000416,
       0.00000000044,    -0.00000000512,
      -0.00000000104,     0.00000000016,
      -0.00000000378,     0.00000001172,
       0.00000000500,     0.00000000037,
      -0.00000000016,    -0.00000000102,
       0.00000001552,     0.00000001981,
      -0.00000000116,     0.00000000079,
      -0.00000000313,    -0.00000000227,
      -0.00000000042,     0.00000000145,
       0.00000000215,    -0.00000000312,
      -0.00000000142,    -0.00000000041,
       0.00000000066,     0.00000000087,
      -0.00000000084,     0.00000000066,
      -0.00000001848,    -0.00000002332,
      -0.00000000069,     0.00000000054,
       0.00000000015,     0.00000000099,
       0.00000000092,    -0.00000000138,
      -0.00000000093,    -0.00000000039,
      -0.00000000171,    -0.00000000122,
      -0.00000000064,     0.00000000136,
       0.00000000156,    -0.00000000220,
      -0.00000000164,    -0.00000000080,
      -0.00000000220,    -0.00000000156,
      -0.00000000081,     0.00000000164,
       0.00000000196,    -0.00000000276,
      -0.00000000206,    -0.00000000101,
      -0.00000000276,    -0.00000000197,
      -0.00000000101,     0.00000000206,
       0.00000008157,    -0.00000011644,
      -0.00000008806,    -0.00000004175,
      -0.00000000906,     0.00000002678,
      -0.00000000366,    -0.00000000257,
      -0.00000000131,     0.00000000276,
       0.00000000084,     0.00000000028,
      -0.00000000029,     0.00000000213,
       0.00000000103,    -0.00000000016,
       0.00000000214,     0.00000000031,
      -0.00000000015,    -0.00000000106,
      -0.00000000598,    -0.00000005895,
      -0.00000003365,     0.00000000813,
      -0.00000000063,    -0.00000000433,
      -0.00000000265,     0.00000000134,
       0.00000000041,    -0.00000000360,
      -0.00000000175,     0.00000000029,
      -0.00000000376,    -0.00000000063,
       0.00000000020,     0.00000000189,
      -0.00000001127,     0.00000005085,
       0.00000002536,    -0.00000000152,
      -0.00000000618,    -0.00000000174,
       0.00000000005,     0.00000000309,
       0.00000000077,    -0.00000000025,
      -0.00000000777,     0.00000000060,
       0.00000000054,     0.00000000304,
       0.00000000031,     0.00000000819,
       0.00000000333,    -0.00000000047,
       0.00000008464,    -0.00000039149,
      -0.00000019463,     0.00000001298,
       0.00000001674,     0.00000004885,
       0.00000002743,     0.00000000789,
      -0.00000000008,    -0.00000001375,
      -0.00000000344,     0.00000000100,
       0.00000001500,    -0.00000000211,
      -0.00000000129,    -0.00000000551,
      -0.00000000103,     0.00000000033,
       0.00000000030,    -0.00000001831,
      -0.00000000783,     0.00000000071,
       0.00000000042,     0.00000000175,
      -0.00000018873,     0.00000000760,
       0.00000000994,     0.00000007620,
       0.00000001585,    -0.00000000399,
      -0.00000000168,    -0.00000002825,
      -0.00000001130,     0.00000000177,
       0.00000000068,     0.00000000233,
       0.00000000334,     0.00000000222,
       0.00000000088,    -0.00000000161,
      -0.00000000198,    -0.00000000483,
      -0.00000000171,     0.00000000125,
       0.00000176495,    -0.00000006908,
      -0.00000009330,    -0.00000071003,
      -0.00000014750,     0.00000003744,
       0.00000000846,     0.00000002025,
       0.00000000652,     0.00000012226,
       0.00000004895,    -0.00000000726,
      -0.00000000277,    -0.00000001011,
      -0.00000000139,     0.00000000060,
      -0.00000000383,    -0.00000001798,
      -0.00000000851,     0.00000000367,
       0.00000000138,     0.00000000202,
       0.00000001745,     0.00000001657,
       0.00000000461,    -0.00000000939,
      -0.00000000246,    -0.00000000028,
       0.00000011304,    -0.00000000144,
      -0.00000000636,    -0.00000004512,
      -0.00000000894,     0.00000000268,
       0.00000000004,     0.00000001770,
       0.00000000709,    -0.00000000098,
      -0.00000000043,    -0.00000000140,
      -0.00000000978,    -0.00000000017,
       0.00000000036,     0.00000000395,
       0.00000000083,    -0.00000000020,
      -0.00000000074,    -0.00000000760,
      -0.00000000301,     0.00000000049,
      -0.00000122635,     0.00000000714,
       0.00000006664,     0.00000049111,
       0.00000009735,    -0.00000002905,
       0.00000000011,    -0.00000007901,
      -0.00000003170,     0.00000000411,
       0.00000000184,     0.00000000629,
       0.00000000080,    -0.00000000042,
       0.00000008237,    -0.00000000229,
      -0.00000000422,    -0.00000003314,
      -0.00000000688,     0.00000000175,
       0.00000000041,     0.00000000093,
       0.00000000484,     0.00000007556,
       0.00000003030,    -0.00000000450,
      -0.00000000166,    -0.00000000630,
      -0.00000000086,     0.00000000037,
      -0.00000003421,    -0.00000000912,
      -0.00000000033,     0.00000001436,
       0.00000000178,    -0.00000000574,
      -0.00000000241,    -0.00000000013,
       0.00000000635,     0.00000000017,
      -0.00000000028,    -0.00000000256,
       0.00000000024,     0.00000000589,
       0.00000000236,    -0.00000000035,
       0.00000044866,     0.00000011390,
       0.00000000273,    -0.00000018409,
      -0.00000003623,     0.00000000773,
      -0.00000000825,     0.00000002675,
       0.00000001113,     0.00000000061,
      -0.00000000039,    -0.00000000220,
      -0.00000005572,    -0.00000000014,
       0.00000000293,     0.00000002231,
       0.00000000444,    -0.00000000131,
      -0.00000000121,    -0.00000005406,
      -0.00000002158,     0.00000000317,
       0.00000000130,     0.00000000428,
       0.00000000796,     0.00000000461,
      -0.00000000094,     0.00000000140,
      -0.00000000226,    -0.00000000066,
       0.00000000052,    -0.00000000218,
      -0.00000012024,    -0.00000006989,
      -0.00000001046,     0.00000004985,
       0.00000000454,    -0.00000000675,
      -0.00000000284,    -0.00000000080,
       0.00000001979,     0.00000000545,
       0.00000000024,    -0.00000000822,
      -0.00000000163,     0.00000000034,
      -0.00000000492,     0.00000001945,
       0.00000000806,     0.00000000015,
      -0.00000000034,    -0.00000000159,
       0.00000002646,     0.00000002894,
      -0.00000000179,     0.00000000144,
      -0.00000000519,    -0.00000000319,
      -0.00000000050,     0.00000000221,
       0.00000000300,    -0.00000000515,
      -0.00000000216,    -0.00000000048,
       0.00000000117,     0.00000000132,
      -0.00000000127,     0.00000000117,
      -0.00000000647,    -0.00000002661,
       0.00000000109,    -0.00000000027,
      -0.00000000015,    -0.00000000103,
       0.00000007377,    -0.00000001130,
      -0.00000000805,    -0.00000002867,
       0.00000000045,     0.00000000300,
       0.00000000117,    -0.00000000032,
       0.00000000201,    -0.00000000030,
       0.00000000031,     0.00000000200,
      -0.00000000002,     0.00000000010,
       0.00000000000,     0.00000000020,
       0.00000000020,     0.00000000000,
      -0.00000000071,    -0.00000000015,
      -0.00000000014,     0.00000000071,
      -0.00000000157,     0.00000000258,
       0.00000000224,     0.00000000105,
       0.00000000252,     0.00000000143,
       0.00000000091,    -0.00000000215,
      -0.00000000413,     0.00000000267,
       0.00000000251,     0.00000000325,
       0.00000000130,    -0.00000000117,
       0.00000000267,     0.00000000413,
       0.00000000325,    -0.00000000251,
      -0.00000000117,    -0.00000000130,
      -0.00000000517,     0.00000000334,
       0.00000000314,     0.00000000407,
       0.00000000163,    -0.00000000147,
       0.00000000335,     0.00000000518,
       0.00000000407,    -0.00000000314,
      -0.00000000147,    -0.00000000163,
      -0.00000000006,     0.00000000001,
      -0.00000000004,     0.00000000001,
       0.00000000001,     0.00000000004,
       0.00000000001,     0.00000000004,
      -0.00000000063,     0.00000000021,
       0.00000000022,     0.00000000063,
       0.00000000023,     0.00000000067,
       0.00000000066,     0.00000000024,
       0.00000000066,    -0.00000000176,
      -0.00000000285,     0.00000000041,
       0.00000000042,     0.00000000282,
      -0.00000000226,     0.00000000028,
       0.00000000041,     0.00000000298,
       0.00000000066,    -0.00000000046,
      -0.00000000053,    -0.00000000040,
       0.00000000055,    -0.00000000056,
      -0.00000000035,    -0.00000000025,
      -0.00000000061,    -0.00000000061,
      -0.00000000023,     0.00000000037,
      -0.00000000004,     0.00000000047,
       0.00000000039,    -0.00000000050,
       0.00000000034,     0.00000000062,
       0.00000000052,    -0.00000000031,
      -0.00000012187,     0.00000007655,
       0.00000007395,     0.00000009485,
       0.00000003734,    -0.00000003510,
      -0.00000001100,    -0.00000000958,
       0.00000000343,     0.00000000532,
       0.00000000418,    -0.00000000322,
      -0.00000000151,    -0.00000000168,
      -0.00000000159,     0.00000000103,
       0.00000000097,     0.00000000125,
      -0.00000000098,    -0.00000000152,
      -0.00000000119,     0.00000000092,
       0.00000004646,    -0.00000005332,
      -0.00000004262,    -0.00000002497,
      -0.00000000607,     0.00000001443,
      -0.00000000416,    -0.00000000184,
      -0.00000000128,     0.00000000285,
      -0.00000000031,     0.00000000294,
       0.00000000143,    -0.00000000025,
       0.00000000366,     0.00000000080,
      -0.00000000024,    -0.00000000145,
      -0.00000004581,    -0.00000001003,
       0.00000000356,     0.00000002916,
       0.00000000145,    -0.00000000304,
      -0.00000000202,    -0.00000000023,
       0.00000000045,    -0.00000000470,
      -0.00000000229,     0.00000000042,
      -0.00000000481,    -0.00000000063,
       0.00000000034,     0.00000000239,
      -0.00000000062,    -0.00000000062,
      -0.00000000063,     0.00000000062,
      -0.00000001472,     0.00000007440,
       0.00000003688,    -0.00000000323,
      -0.00000000341,    -0.00000000907,
      -0.00000000918,    -0.00000000249,
       0.00000000014,     0.00000000457,
       0.00000000113,    -0.00000000038,
      -0.00000001385,     0.00000000143,
       0.00000000119,     0.00000000540,
       0.00000000107,    -0.00000000038,
       0.00000000109,     0.00000001432,
       0.00000000573,    -0.00000000110,
      -0.00000000039,    -0.00000000115,
       0.00000011124,    -0.00000055307,
      -0.00000027462,     0.00000002252,
       0.00000002504,     0.00000006750,
       0.00000004055,     0.00000001121,
      -0.00000000047,    -0.00000002023,
      -0.00000000502,     0.00000000156,
       0.00000000053,     0.00000000080,
       0.00000002710,    -0.00000000387,
      -0.00000000261,    -0.00000001018,
      -0.00000000193,     0.00000000074,
      -0.00000000115,    -0.00000003066,
      -0.00000001269,     0.00000000190,
       0.00000000082,     0.00000000272,
      -0.00000035325,     0.00000002617,
       0.00000002562,     0.00000014048,
       0.00000002865,    -0.00000000914,
      -0.00000000509,    -0.00000005587,
      -0.00000002210,     0.00000000450,
       0.00000000155,     0.00000000448,
       0.00000000509,     0.00000000282,
       0.00000000115,    -0.00000000241,
      -0.00000000343,    -0.00000000734,
      -0.00000000255,     0.00000000199,
       0.00000329392,    -0.00000024718,
      -0.00000024009,    -0.00000130971,
      -0.00000026706,     0.00000008553,
       0.00000001796,     0.00000003559,
       0.00000002068,     0.00000024021,
       0.00000009513,    -0.00000001872,
      -0.00000000650,    -0.00000001930,
      -0.00000000257,     0.00000000132,
      -0.00000000626,    -0.00000002484,
      -0.00000001189,     0.00000000558,
       0.00000000205,     0.00000000283,
       0.00000002654,     0.00000002541,
       0.00000000705,    -0.00000001403,
      -0.00000000361,    -0.00000000045,
      -0.00000000061,    -0.00000000111,
       0.00000000037,    -0.00000000083,
       0.00000018586,    -0.00000001691,
      -0.00000001509,    -0.00000007290,
      -0.00000001415,     0.00000000515,
       0.00000000251,     0.00000003132,
       0.00000001233,    -0.00000000247,
      -0.00000000087,    -0.00000000239,
      -0.00000001841,     0.00000000062,
       0.00000000112,     0.00000000735,
       0.00000000150,    -0.00000000046,
      -0.00000000171,    -0.00000001536,
      -0.00000000609,     0.00000000123,
       0.00000000038,     0.00000000123,
      -0.00000207697,     0.00000017116,
       0.00000016382,     0.00000081647,
       0.00000015861,    -0.00000005713,
      -0.00000001174,    -0.00000001979,
      -0.00000001074,    -0.00000013956,
      -0.00000005496,     0.00000001068,
       0.00000000378,     0.00000001068,
       0.00000000134,    -0.00000000077,
       0.00000015782,    -0.00000001065,
      -0.00000001136,    -0.00000006264,
      -0.00000001275,     0.00000000410,
       0.00000000088,     0.00000000168,
       0.00000001369,     0.00000014759,
       0.00000005862,    -0.00000001149,
      -0.00000000394,    -0.00000001197,
      -0.00000000158,     0.00000000082,
      -0.00000005134,    -0.00000000951,
       0.00000000095,     0.00000002075,
       0.00000000216,    -0.00000000953,
      -0.00000000383,     0.00000000006,
       0.00000001114,    -0.00000000063,
      -0.00000000080,    -0.00000000437,
      -0.00000000085,     0.00000000029,
       0.00000000115,     0.00000001045,
       0.00000000411,    -0.00000000088,
      -0.00000000030,    -0.00000000082,
       0.00000072185,     0.00000012242,
      -0.00000001614,    -0.00000028293,
      -0.00000005399,     0.00000001496,
      -0.00000001010,     0.00000004475,
       0.00000001777,    -0.00000000023,
      -0.00000000080,    -0.00000000342,
      -0.00000009747,     0.00000000735,
       0.00000000756,     0.00000003825,
       0.00000000745,    -0.00000000266,
      -0.00000000056,    -0.00000000092,
      -0.00000000917,    -0.00000009485,
      -0.00000003726,     0.00000000781,
       0.00000000265,     0.00000000724,
       0.00000000090,    -0.00000000053,
       0.00000001228,     0.00000000565,
      -0.00000000132,     0.00000000243,
       0.00000000092,     0.00000000017,
      -0.00000000377,    -0.00000000076,
       0.00000000005,     0.00000000150,
       0.00000000055,    -0.00000000362,
      -0.00000000150,     0.00000000007,
      -0.00000020038,    -0.00000009558,
      -0.00000000907,     0.00000007342,
       0.00000000647,    -0.00000001178,
      -0.00000000437,    -0.00000000086,
       0.00000003271,     0.00000000627,
      -0.00000000053,    -0.00000001302,
      -0.00000000250,     0.00000000069,
      -0.00000000547,     0.00000003214,
       0.00000001277,    -0.00000000062,
      -0.00000000067,    -0.00000000245,
       0.00000000056,    -0.00000000054,
       0.00000000103,     0.00000000051,
      -0.00000000046,     0.00000000101,
       0.00000004739,     0.00000004452,
       0.00000000428,    -0.00000001671,
      -0.00000000297,     0.00000000280,
       0.00000000095,     0.00000000033,
      -0.00000000886,    -0.00000000442,
      -0.00000000049,     0.00000000335,
       0.00000000412,    -0.00000000879,
      -0.00000000328,    -0.00000000047,
      -0.00000000819,    -0.00000001749,
       0.00000000117,    -0.00000000048,
       0.00000000219,     0.00000000212,
       0.00000000023,    -0.00000000077,
      -0.00000000203,     0.00000000220,
      -0.00000000040,    -0.00000000087,
       0.00000000088,    -0.00000000040,
      -0.00000023175,     0.00000014543,
       0.00000014052,     0.00000018043,
       0.00000007098,    -0.00000006665,
      -0.00000002105,    -0.00000001829,
       0.00000000438,     0.00000000677,
       0.00000000532,    -0.00000000411,
      -0.00000000191,    -0.00000000213,
      -0.00000000056,     0.00000000060,
      -0.00000000086,     0.00000000022,
      -0.00000000145,     0.00000000020,
       0.00000000022,     0.00000000151,
      -0.00000001433,    -0.00000005930,
      -0.00000002800,     0.00000001378,
       0.00000000255,    -0.00000000064,
      -0.00000000060,    -0.00000000121,
       0.00000000110,    -0.00000000012,
      -0.00000000027,    -0.00000000152,
      -0.00000002154,     0.00000000323,
       0.00000000235,     0.00000000837,
      -0.00000000046,    -0.00000000311,
      -0.00000000120,     0.00000000034,
       0.00000021284,    -0.00000003202,
      -0.00000002329,    -0.00000008269,
       0.00000000134,     0.00000000899,
       0.00000000349,    -0.00000000098,
      -0.00000000045,    -0.00000000162,
      -0.00000000076,     0.00000000041,
       0.00000000163,    -0.00000000039,
      -0.00000000037,    -0.00000000077,
       0.00000000592,    -0.00000000086,
      -0.00000000064,    -0.00000000229,
       0.00000000091,     0.00000000589,
       0.00000000228,    -0.00000000065,
       0.00000000001,    -0.00000000007,
      -0.00000000002,    -0.00000000024,
      -0.00000000012,     0.00000000006,
      -0.00000000024,     0.00000000002,
       0.00000000006,     0.00000000012,
       0.00000000143,    -0.00000000019,
      -0.00000000022,    -0.00000000072,
      -0.00000000019,    -0.00000000143,
      -0.00000000072,     0.00000000022,
       0.00000000281,    -0.00000000378,
      -0.00000000292,    -0.00000000167,
      -0.00000000047,     0.00000000111,
      -0.00000000391,    -0.00000000288,
      -0.00000000166,     0.00000000312,
       0.00000000116,     0.00000000051,
      -0.00000003020,     0.00000003731,
       0.00000003273,     0.00000002165,
       0.00000000769,    -0.00000001443,
       0.00000000219,     0.00000000168,
       0.00000000130,    -0.00000000189,
      -0.00000001088,     0.00000000657,
       0.00000000576,     0.00000000781,
       0.00000000293,    -0.00000000255,
       0.00000000753,     0.00000001079,
       0.00000000760,    -0.00000000675,
      -0.00000000308,    -0.00000000275,
      -0.00000000111,     0.00000000027,
       0.00000000029,     0.00000000110,
      -0.00000000120,    -0.00000000377,
      -0.00000000232,     0.00000000173,
       0.00000000089,     0.00000000064,
      -0.00000000377,     0.00000000120,
       0.00000000173,     0.00000000232,
       0.00000000064,    -0.00000000088,
      -0.00000000150,    -0.00000000472,
      -0.00000000291,     0.00000000217,
       0.00000000111,     0.00000000080,
      -0.00000000472,     0.00000000150,
       0.00000000217,     0.00000000291,
       0.00000000080,    -0.00000000111,
       0.00000000052,    -0.00000000013,
      -0.00000000013,    -0.00000000051,
       0.00000000081,    -0.00000000021,
      -0.00000000021,    -0.00000000081,
      -0.00000000014,    -0.00000000057,
       0.00000000046,    -0.00000000009,
      -0.00000000088,     0.00000000010,
       0.00000000013,     0.00000000044,
       0.00000000036,     0.00000000227,
       0.00000000114,    -0.00000000038,
       0.00000000270,    -0.00000000249,
      -0.00000000139,    -0.00000000098,
      -0.00000000250,    -0.00000000267,
      -0.00000000096,     0.00000000139,
       0.00000000229,    -0.00000000205,
      -0.00000000113,    -0.00000000083,
      -0.00000000291,    -0.00000000282,
      -0.00000000100,     0.00000000156,
       0.00000000101,     0.00000000238,
       0.00000000084,    -0.00000000053,
       0.00000000233,    -0.00000000126,
      -0.00000000066,    -0.00000000078,
       0.00000000151,     0.00000000290,
       0.00000000099,    -0.00000000077,
       0.00000000304,    -0.00000000162,
      -0.00000000082,    -0.00000000104,
      -0.00000000055,    -0.00000000002,
       0.00000000052,     0.00000000105,
       0.00000000031,    -0.00000000025,
      -0.00000000110,     0.00000000048,
       0.00000000017,     0.00000000029,
       0.00000000036,     0.00000000097,
      -0.00000000025,     0.00000000026,
      -0.00000003434,    -0.00000011099,
      -0.00000006858,     0.00000004892,
       0.00000002501,     0.00000001920,
      -0.00000000476,     0.00000000145,
       0.00000000214,     0.00000000296,
       0.00000000080,    -0.00000000111,
      -0.00000000025,     0.00000000354,
       0.00000000172,    -0.00000000035,
       0.00000000355,     0.00000000027,
      -0.00000000035,    -0.00000000173,
      -0.00000002922,    -0.00000000427,
       0.00000000366,     0.00000001712,
       0.00000000123,    -0.00000000274,
      -0.00000000183,    -0.00000000023,
      -0.00000000016,    -0.00000000672,
      -0.00000000346,     0.00000000124,
       0.00000000064,     0.00000000085,
      -0.00000000370,    -0.00000000095,
      -0.00000000021,     0.00000000163,
      -0.00000000111,    -0.00000000106,
      -0.00000000109,     0.00000000111,
       0.00000003522,    -0.00000003262,
      -0.00000002625,    -0.00000001783,
      -0.00000000355,     0.00000000873,
       0.00000001525,     0.00000001637,
       0.00000000832,    -0.00000001228,
      -0.00000000410,    -0.00000000165,
      -0.00000000032,     0.00000000080,
      -0.00000016801,     0.00000025024,
       0.00000016820,     0.00000006945,
       0.00000000985,    -0.00000005126,
      -0.00000004580,    -0.00000003936,
      -0.00000001808,     0.00000003314,
       0.00000001052,     0.00000000310,
      -0.00000002089,     0.00000000062,
       0.00000000031,     0.00000000786,
       0.00000000177,    -0.00000000072,
       0.00000000671,     0.00000002462,
       0.00000000954,    -0.00000000499,
      -0.00000000071,    -0.00000000185,
       0.00000011931,    -0.00000068499,
      -0.00000033758,     0.00000003875,
       0.00000003359,     0.00000008223,
       0.00000001268,    -0.00000000966,
       0.00000005390,     0.00000001467,
      -0.00000000084,    -0.00000002681,
      -0.00000000660,     0.00000000214,
       0.00000000071,     0.00000000103,
       0.00000004487,    -0.00000000608,
      -0.00000000437,    -0.00000001718,
      -0.00000000335,     0.00000000132,
      -0.00000000363,    -0.00000004781,
      -0.00000001927,     0.00000000369,
       0.00000000137,     0.00000000400,
      -0.00000001691,     0.00000001645,
       0.00000000058,     0.00000000058,
      -0.00000063964,     0.00000006624,
       0.00000005648,     0.00000025170,
       0.00000005053,    -0.00000001886,
      -0.00000000378,    -0.00000000656,
      -0.00000001271,    -0.00000010960,
      -0.00000004298,     0.00000001033,
       0.00000000339,     0.00000000858,
       0.00000000108,    -0.00000000069,
       0.00000000711,     0.00000000287,
       0.00000000130,    -0.00000000328,
      -0.00000000073,    -0.00000000027,
      -0.00000000561,    -0.00000001041,
      -0.00000000354,     0.00000000297,
       0.00000000077,     0.00000000051,
       0.00000595575,    -0.00000061341,
      -0.00000052165,    -0.00000234607,
      -0.00000047138,     0.00000017375,
       0.00000003494,     0.00000006139,
       0.00000005297,     0.00000046770,
       0.00000018351,    -0.00000004351,
      -0.00000001429,    -0.00000003667,
      -0.00000000476,     0.00000000279,
      -0.00000000930,    -0.00000002967,
      -0.00000001464,     0.00000000755,
       0.00000000269,     0.00000000353,
       0.00000000053,    -0.00000000058,
       0.00000003713,     0.00000003670,
       0.00000001034,    -0.00000001912,
      -0.00000000484,    -0.00000000078,
      -0.00000000115,    -0.00000000197,
      -0.00000000064,     0.00000000045,
       0.00000000067,    -0.00000000155,
       0.00000029433,    -0.00000005555,
      -0.00000003231,    -0.00000011258,
      -0.00000002139,     0.00000000935,
       0.00000001014,     0.00000005510,
       0.00000002116,    -0.00000000596,
      -0.00000000174,    -0.00000000401,
      -0.00000003453,     0.00000000261,
       0.00000000277,     0.00000001357,
       0.00000000271,    -0.00000000100,
      -0.00000000377,    -0.00000003041,
      -0.00000001201,     0.00000000280,
       0.00000000086,     0.00000000239,
      -0.00000342630,     0.00000062013,
       0.00000036855,     0.00000131556,
       0.00000025007,    -0.00000010773,
      -0.00000002034,    -0.00000003045,
      -0.00000004430,    -0.00000024495,
      -0.00000009410,     0.00000002596,
       0.00000000761,     0.00000001788,
       0.00000000218,    -0.00000000141,
       0.00000029853,    -0.00000002953,
      -0.00000002627,    -0.00000011717,
      -0.00000002348,     0.00000000881,
       0.00000000179,     0.00000000303,
       0.00000003271,     0.00000028341,
       0.00000011174,    -0.00000002598,
      -0.00000000853,    -0.00000002249,
      -0.00000000290,     0.00000000170,
       0.00000000065,     0.00000000096,
      -0.00000000026,     0.00000000092,
      -0.00000007506,    -0.00000000674,
       0.00000000394,     0.00000002868,
       0.00000000199,    -0.00000001618,
      -0.00000000609,     0.00000000064,
       0.00000000038,     0.00000000113,
       0.00000001949,    -0.00000000312,
      -0.00000000200,    -0.00000000744,
      -0.00000000142,     0.00000000060,
       0.00000000381,     0.00000001842,
       0.00000000712,    -0.00000000206,
      -0.00000000061,    -0.00000000137,
       0.00000116364,     0.00000007942,
      -0.00000006409,    -0.00000042729,
      -0.00000007811,     0.00000002829,
      -0.00000000965,     0.00000007672,
       0.00000002851,    -0.00000000291,
      -0.00000000166,    -0.00000000528,
      -0.00000016935,     0.00000002972,
       0.00000001806,     0.00000006480,
       0.00000001232,    -0.00000000531,
      -0.00000000102,    -0.00000000148,
      -0.00000003226,    -0.00000016512,
      -0.00000006346,     0.00000001828,
       0.00000000527,     0.00000001207,
       0.00000000146,    -0.00000000097,
       0.00000000026,    -0.00000000097,
       0.00000002124,     0.00000000825,
      -0.00000000187,     0.00000000456,
       0.00000000150,     0.00000000011,
      -0.00000000644,    -0.00000000065,
       0.00000000032,     0.00000000240,
       0.00000000032,    -0.00000000618,
      -0.00000000237,     0.00000000036,
      -0.00000034281,    -0.00000012597,
      -0.00000000215,     0.00000011032,
       0.00000001923,    -0.00000000592,
       0.00000000958,    -0.00000002257,
      -0.00000000712,    -0.00000000062,
       0.00000000035,     0.00000000124,
       0.00000005529,     0.00000000515,
      -0.00000000271,    -0.00000002070,
      -0.00000000381,     0.00000000137,
      -0.00000000384,     0.00000005427,
       0.00000002032,    -0.00000000278,
      -0.00000000133,    -0.00000000375,
      -0.00000000552,    -0.00000000468,
       0.00000000106,    -0.00000000115,
       0.00000000191,     0.00000000075,
      -0.00000000065,     0.00000000191,
       0.00000009178,     0.00000007676,
       0.00000000285,    -0.00000002711,
      -0.00000000580,     0.00000000627,
       0.00000000173,     0.00000000030,
      -0.00000001660,    -0.00000000634,
      -0.00000000024,     0.00000000540,
       0.00000000098,    -0.00000000031,
       0.00000000580,    -0.00000001646,
      -0.00000000528,    -0.00000000022,
       0.00000000031,     0.00000000099,
      -0.00000000057,    -0.00000000046,
       0.00000000054,    -0.00000000058,
      -0.00000001840,    -0.00000003617,
       0.00000000290,    -0.00000000128,
       0.00000000487,     0.00000000413,
       0.00000000019,    -0.00000000141,
      -0.00000000395,     0.00000000489,
       0.00000000142,     0.00000000017,
      -0.00000000128,    -0.00000000008,
      -0.00000000112,    -0.00000000231,
       0.00000000224,    -0.00000000112,
       0.00000000072,     0.00000000062,
      -0.00000000014,     0.00000000142,
      -0.00000000142,    -0.00000000014,
       0.00000000880,    -0.00000000690,
      -0.00000002150,     0.00000001675,
       0.00000000062,    -0.00000000235,
      -0.00000006310,    -0.00000020910,
      -0.00000012975,     0.00000009120,
       0.00000004691,     0.00000003648,
      -0.00000000614,     0.00000000196,
       0.00000000282,     0.00000000378,
       0.00000000104,    -0.00000000144,
       0.00000000384,     0.00000001591,
       0.00000000751,    -0.00000000372,
      -0.00000000185,     0.00000000047,
       0.00000000044,     0.00000000087,
      -0.00000000391,     0.00000000056,
       0.00000000042,     0.00000000150,
       0.00000000057,     0.00000000398,
       0.00000000155,    -0.00000000044,
      -0.00000002833,    -0.00000011822,
      -0.00000005583,     0.00000002744,
       0.00000000544,    -0.00000000137,
      -0.00000000129,    -0.00000000257,
      -0.00000000061,     0.00000000048,
       0.00000000312,    -0.00000000041,
      -0.00000000027,    -0.00000000116,
      -0.00000000058,    -0.00000000367,
      -0.00000000147,     0.00000000046,
      -0.00000006047,     0.00000000900,
       0.00000000663,     0.00000002348,
      -0.00000000138,    -0.00000000934,
      -0.00000000362,     0.00000000102,
       0.00000000032,     0.00000000071,
       0.00000059807,    -0.00000008920,
      -0.00000006571,    -0.00000023224,
      -0.00000004565,     0.00000002064,
       0.00000000395,     0.00000002665,
       0.00000001035,    -0.00000000292,
      -0.00000000091,    -0.00000000203,
      -0.00000000101,    -0.00000000323,
      -0.00000000154,     0.00000000087,
       0.00000000346,    -0.00000000075,
      -0.00000000076,    -0.00000000162,
      -0.00000000183,     0.00000000022,
      -0.00000000031,    -0.00000000177,
      -0.00000001305,     0.00000002559,
      -0.00000000096,    -0.00000000075,
       0.00000001710,    -0.00000000248,
      -0.00000000186,    -0.00000000663,
      -0.00000000129,     0.00000000058,
       0.00000000261,     0.00000001701,
       0.00000000661,    -0.00000000188,
      -0.00000000058,    -0.00000000129,
      -0.00000000038,     0.00000000070,
      -0.00000000071,    -0.00000000039,
       0.00000000065,     0.00000000017,
       0.00000000001,    -0.00000000028,
       0.00000000219,    -0.00000000005,
      -0.00000000019,    -0.00000000082,
      -0.00000000003,    -0.00000000216,
      -0.00000000082,     0.00000000019,
       0.00000000007,     0.00000000007,
       0.00000000009,    -0.00000000008,
      -0.00000000003,    -0.00000000003,
       0.00000000054,    -0.00000000026,
      -0.00000000026,    -0.00000000054,
      -0.00000000110,    -0.00000000095,
      -0.00000000044,     0.00000000063,
      -0.00000000109,     0.00000001001,
       0.00000000436,    -0.00000000014,
       0.00000000999,     0.00000000109,
      -0.00000000013,    -0.00000000435,
       0.00000000613,    -0.00000000039,
      -0.00000000104,    -0.00000000480,
      -0.00000000213,     0.00000000297,
       0.00000000221,     0.00000000111,
       0.00000001863,     0.00000002631,
       0.00000001414,    -0.00000001195,
      -0.00000000376,    -0.00000000409,
       0.00000002631,    -0.00000001855,
      -0.00000001200,    -0.00000001415,
      -0.00000000417,     0.00000000385,
      -0.00000000062,     0.00000000239,
       0.00000000239,     0.00000000056,
      -0.00000014340,     0.00000008560,
       0.00000007717,     0.00000010310,
       0.00000003752,    -0.00000003467,
      -0.00000001031,    -0.00000000910,
       0.00000000748,    -0.00000000187,
      -0.00000000211,    -0.00000000490,
      -0.00000000160,     0.00000000100,
       0.00000000098,    -0.00000011159,
      -0.00000007288,     0.00000000536,
       0.00000000409,     0.00000002425,
       0.00000000526,    -0.00000000152,
      -0.00000011083,     0.00000000052,
       0.00000000658,     0.00000007214,
       0.00000002386,    -0.00000000459,
      -0.00000000158,    -0.00000000524,
       0.00000000146,    -0.00000000582,
      -0.00000000241,    -0.00000000019,
      -0.00000000581,    -0.00000000135,
      -0.00000000014,     0.00000000241,
      -0.00000000036,    -0.00000000316,
      -0.00000000180,     0.00000000081,
       0.00000004289,    -0.00000001220,
      -0.00000001453,    -0.00000002385,
      -0.00000000618,     0.00000000653,
       0.00000000173,     0.00000000095,
      -0.00000001216,    -0.00000004288,
      -0.00000002386,     0.00000001449,
       0.00000000652,     0.00000000617,
       0.00000000095,    -0.00000000172,
       0.00000000068,     0.00000000035,
       0.00000005383,    -0.00000001516,
      -0.00000001819,    -0.00000002991,
      -0.00000000774,     0.00000000818,
       0.00000000217,     0.00000000119,
      -0.00000001517,    -0.00000005381,
      -0.00000002989,     0.00000001818,
       0.00000000817,     0.00000000773,
       0.00000000119,    -0.00000000216,
      -0.00000000002,     0.00000000000,
       0.00000000000,     0.00000000019,
       0.00000000006,    -0.00000000001,
       0.00000000003,     0.00000000012,
       0.00000000004,    -0.00000000002,
       0.00000000013,    -0.00000000003,
      -0.00000000001,    -0.00000000004,
       0.00000000001,     0.00000000003,
       0.00000000013,    -0.00000000004,
      -0.00000000002,    -0.00000000004,
       0.00000000001,     0.00000000004,
       0.00000000021,     0.00000000107,
       0.00000000041,    -0.00000000018,
       0.00000000107,    -0.00000000021,
      -0.00000000018,    -0.00000000041,
       0.00000001247,    -0.00000000327,
       0.00000000162,     0.00000000803,
       0.00000000279,    -0.00000000123,
       0.00000000804,    -0.00000000162,
      -0.00000000123,    -0.00000000279,
       0.00000000011,     0.00000000056,
       0.00000000019,    -0.00000000008,
       0.00000000135,    -0.00000000026,
      -0.00000000020,    -0.00000000047,
       0.00000000065,     0.00000000321,
       0.00000000113,    -0.00000000049,
       0.00000000320,    -0.00000000065,
      -0.00000000049,    -0.00000000113,
       0.00000001370,    -0.00000000351,
       0.00000000176,     0.00000000881,
       0.00000000305,    -0.00000000134,
       0.00000000881,    -0.00000000177,
      -0.00000000134,    -0.00000000306,
       0.00000000040,     0.00000000201,
       0.00000000070,    -0.00000000031,
       0.00000000359,    -0.00000000071,
      -0.00000000054,    -0.00000000125,
      -0.00000000098,     0.00000000016,
       0.00000000014,     0.00000000093,
      -0.00000000093,     0.00000000014,
       0.00000000007,     0.00000000027,
       0.00000000021,     0.00000000140,
       0.00000000040,    -0.00000000011,
       0.00000000046,    -0.00000000415,
      -0.00000000187,     0.00000000007,
      -0.00000001018,    -0.00000000095,
       0.00000000021,     0.00000000459,
       0.00000000678,     0.00000001465,
       0.00000000493,    -0.00000000347,
       0.00000001448,    -0.00000000684,
      -0.00000000348,    -0.00000000482,
       0.00000000586,     0.00000001297,
       0.00000000430,    -0.00000000299,
       0.00000001599,    -0.00000000796,
      -0.00000000395,    -0.00000000529,
      -0.00000000564,     0.00000000052,
       0.00000000026,     0.00000000167,
       0.00000000173,     0.00000000570,
       0.00000000161,    -0.00000000079,
      -0.00000000705,     0.00000000198,
       0.00000000080,     0.00000000195,
       0.00000000174,     0.00000000728,
       0.00000000207,    -0.00000000077,
       0.00000000044,     0.00000000152,
       0.00000000049,    -0.00000000030,
      -0.00000000109,     0.00000000104,
       0.00000000042,     0.00000000038,
      -0.00000000079,     0.00000000058,
       0.00000000043,     0.00000000052,
      -0.00000000071,     0.00000000026,
      -0.00000000022,    -0.00000000069,
       0.00000000175,    -0.00000000267,
      -0.00000000267,    -0.00000000175,
       0.00000000248,    -0.00000000380,
      -0.00000000380,    -0.00000000248,
       0.00000000311,    -0.00000000476,
      -0.00000000476,    -0.00000000311,
      -0.00000002068,    -0.00000005899,
      -0.00000000177,     0.00000000062,
      -0.00000000159,     0.00000000079,
      -0.00000000079,    -0.00000000159,
      -0.00000002522,     0.00000000533,
      -0.00000000025,    -0.00000000123,
      -0.00000000100,     0.00000000021,
      -0.00000000021,    -0.00000000100,
       0.00000013506,    -0.00000016270,
      -0.00000000691,    -0.00000000574,
      -0.00000000084,    -0.00000000031,
      -0.00000000032,     0.00000000086,
       0.00000123006,    -0.00000035793,
      -0.00000042970,    -0.00000066906,
      -0.00000017202,     0.00000019006,
       0.00000004986,     0.00000002615,
      -0.00000001403,    -0.00000005357,
      -0.00000003001,     0.00000001735,
       0.00000000795,     0.00000000783,
       0.00000000123,    -0.00000000211,
       0.00000000108,    -0.00000000402,
      -0.00000000211,    -0.00000000008,
      -0.00000000368,     0.00000000117,
       0.00000000123,     0.00000000147,
      -0.00000000210,    -0.00000000190,
      -0.00000000051,     0.00000000083,
      -0.00000000156,     0.00000000162,
       0.00000000009,     0.00000000116,
      -0.00000001559,     0.00000002993,
       0.00000000127,     0.00000000066,
      -0.00000001524,     0.00000010899,
       0.00000005333,    -0.00000000793,
      -0.00000000570,    -0.00000001285,
      -0.00000001350,    -0.00000000356,
       0.00000000028,     0.00000000668,
       0.00000000162,    -0.00000000056,
      -0.00000003461,     0.00000000517,
       0.00000000388,     0.00000001339,
       0.00000000261,    -0.00000000116,
       0.00000000494,     0.00000003491,
       0.00000001359,    -0.00000000377,
      -0.00000000115,    -0.00000000266,
       0.00000003295,    -0.00000062425,
      -0.00000029254,     0.00000007667,
       0.00000004103,     0.00000006629,
       0.00000000902,    -0.00000001019,
       0.00000005783,     0.00000001671,
      -0.00000000050,    -0.00000002886,
      -0.00000000711,     0.00000000227,
       0.00000000075,     0.00000000111,
       0.00000008361,    -0.00000001450,
      -0.00000001337,    -0.00000003520,
      -0.00000000754,     0.00000000496,
       0.00000000117,     0.00000000105,
      -0.00000000234,    -0.00000004749,
      -0.00000001592,    -0.00000000013,
      -0.00000000076,     0.00000000265,
       0.00000000072,     0.00000000057,
       0.00000000050,    -0.00000000092,
       0.00000000001,    -0.00000000004,
       0.00000000000,    -0.00000000013,
      -0.00000000013,     0.00000000000,
       0.00000000050,     0.00000000014,
       0.00000000014,    -0.00000000049,
       0.00000000441,     0.00000002722,
       0.00000001323,    -0.00000000655,
      -0.00000001700,     0.00000000235,
       0.00000000376,     0.00000000842,
       0.00000000152,    -0.00000000124,
       0.00000000958,    -0.00000000721,
      -0.00000000562,    -0.00000000432,
      -0.00000000080,     0.00000000196,
       0.00000000626,     0.00000000472,
       0.00000000296,    -0.00000000520,
      -0.00000000184,    -0.00000000072,
      -0.00000112464,    -0.00000002962,
       0.00000002820,     0.00000046911,
       0.00000009902,    -0.00000002042,
      -0.00000000748,    -0.00000001163,
       0.00000001091,    -0.00000022156,
      -0.00000009247,     0.00000000253,
       0.00000000323,     0.00000001969,
       0.00000000206,    -0.00000000138,
      -0.00000003374,     0.00000002578,
       0.00000002365,     0.00000001667,
       0.00000000343,    -0.00000000822,
      -0.00000004239,    -0.00000003027,
      -0.00000001388,     0.00000002782,
       0.00000000868,     0.00000000240,
       0.00001033950,    -0.00000132208,
      -0.00000104416,    -0.00000403942,
      -0.00000080141,     0.00000033341,
       0.00000006439,     0.00000010220,
       0.00000011997,     0.00000089975,
       0.00000035089,    -0.00000009317,
      -0.00000002944,    -0.00000006938,
      -0.00000000888,     0.00000000564,
       0.00000000078,     0.00000000079,
      -0.00000001636,    -0.00000002836,
      -0.00000001411,     0.00000001095,
       0.00000000343,     0.00000000329,
       0.00000000048,    -0.00000000067,
       0.00000004096,     0.00000004964,
       0.00000001531,    -0.00000002065,
      -0.00000000518,    -0.00000000170,
      -0.00000000209,    -0.00000000342,
      -0.00000000112,     0.00000000080,
       0.00000000115,    -0.00000000277,
      -0.00000000104,    -0.00000000011,
      -0.00000000354,    -0.00000000484,
      -0.00000000126,     0.00000000235,
      -0.00000000483,     0.00000000354,
       0.00000000235,     0.00000000126,
       0.00000000045,    -0.00000000009,
      -0.00000000010,    -0.00000000044,
      -0.00000000011,    -0.00000000049,
      -0.00000016133,    -0.00000020996,
      -0.00000005406,     0.00000010480,
      -0.00000000612,     0.00000000450,
       0.00000000298,     0.00000000160,
      -0.00000000143,     0.00000000019,
      -0.00000000019,    -0.00000000143,
       0.00000044308,    -0.00000014541,
      -0.00000006505,    -0.00000016403,
      -0.00000003043,     0.00000001625,
       0.00000003300,     0.00000009836,
       0.00000003642,    -0.00000001458,
      -0.00000000356,    -0.00000000671,
      -0.00000000078,     0.00000000055,
      -0.00000006551,     0.00000000580,
       0.00000000628,     0.00000002526,
       0.00000000498,    -0.00000000208,
      -0.00000000040,    -0.00000000062,
      -0.00000000909,    -0.00000005816,
      -0.00000002316,     0.00000000588,
       0.00000000181,     0.00000000459,
      -0.00000564042,     0.00000160481,
       0.00000075743,     0.00000212624,
       0.00000039654,    -0.00000019521,
      -0.00000003398,    -0.00000004749,
      -0.00000015209,    -0.00000043172,
      -0.00000015854,     0.00000006579,
       0.00000001584,     0.00000002904,
       0.00000000348,    -0.00000000261,
       0.00000056050,    -0.00000006921,
      -0.00000005644,    -0.00000021792,
      -0.00000004308,     0.00000001803,
       0.00000000348,     0.00000000547,
       0.00000007207,     0.00000053744,
       0.00000021077,    -0.00000005516,
      -0.00000001755,    -0.00000004193,
      -0.00000000532,     0.00000000337,
       0.00000000138,     0.00000000153,
      -0.00000000021,     0.00000000167,
      -0.00000011215,    -0.00000000038,
       0.00000000902,     0.00000003894,
       0.00000000671,    -0.00000000314,
      -0.00000000025,    -0.00000003030,
      -0.00000001033,     0.00000000249,
       0.00000000083,     0.00000000182,
       0.00000003567,    -0.00000000998,
      -0.00000000497,    -0.00000001284,
      -0.00000000240,     0.00000000129,
       0.00000001229,     0.00000003233,
       0.00000001240,    -0.00000000501,
      -0.00000000126,    -0.00000000232,
       0.00000198985,    -0.00000008672,
      -0.00000018369,    -0.00000066400,
      -0.00000011365,     0.00000005525,
       0.00000000981,     0.00000001213,
      -0.00000000082,     0.00000014572,
       0.00000004888,    -0.00000001139,
      -0.00000000371,    -0.00000000847,
      -0.00000000088,     0.00000000065,
      -0.00000030196,     0.00000009121,
       0.00000004231,     0.00000011221,
       0.00000002070,    -0.00000001071,
      -0.00000000186,    -0.00000000240,
      -0.00000009512,    -0.00000029438,
      -0.00000011025,     0.00000004240,
       0.00000001060,     0.00000002041,
       0.00000000236,    -0.00000000180,
       0.00000000049,    -0.00000000191,
      -0.00000000083,    -0.00000000018,
       0.00000000152,     0.00000000023,
      -0.00000000063,     0.00000000054,
       0.00000003946,     0.00000001205,
      -0.00000000417,     0.00000001250,
       0.00000000300,    -0.00000000028,
      -0.00000001223,     0.00000000024,
       0.00000000109,     0.00000000417,
       0.00000000079,    -0.00000000040,
      -0.00000000131,    -0.00000001126,
      -0.00000000413,     0.00000000112,
       0.00000000041,     0.00000000085,
      -0.00000069761,    -0.00000018959,
       0.00000002136,     0.00000018449,
       0.00000002851,    -0.00000001366,
       0.00000001731,    -0.00000005481,
      -0.00000001419,     0.00000000080,
       0.00000000096,     0.00000000217,
       0.00000010453,    -0.00000000199,
      -0.00000000908,    -0.00000003564,
      -0.00000000616,     0.00000000296,
       0.00000000060,     0.00000000068,
       0.00000000482,     0.00000010217,
       0.00000003503,    -0.00000000911,
      -0.00000000293,    -0.00000000604,
      -0.00000000070,     0.00000000062,
      -0.00000000035,     0.00000000137,
      -0.00000000129,    -0.00000000021,
      -0.00000000061,     0.00000000058,
      -0.00000001624,    -0.00000001334,
       0.00000000545,    -0.00000000636,
       0.00000000488,     0.00000000134,
      -0.00000000017,    -0.00000000125,
      -0.00000000117,     0.00000000491,
       0.00000000145,    -0.00000000016,
       0.00000025476,     0.00000019360,
      -0.00000000436,    -0.00000005916,
      -0.00000001854,     0.00000002215,
       0.00000000523,    -0.00000000016,
      -0.00000000030,    -0.00000000073,
      -0.00000004232,    -0.00000001176,
       0.00000000101,     0.00000001108,
       0.00000000176,    -0.00000000088,
       0.00000001056,    -0.00000004204,
      -0.00000001082,     0.00000000103,
       0.00000000089,     0.00000000178,
      -0.00000000216,    -0.00000000027,
       0.00000000203,    -0.00000000051,
       0.00000000510,     0.00000001211,
      -0.00000000989,     0.00000000408,
      -0.00000000263,    -0.00000000186,
       0.00000000220,    -0.00000000260,
      -0.00000008634,    -0.00000015830,
      -0.00000000069,     0.00000003314,
       0.00000002740,    -0.00000001403,
      -0.00000000549,    -0.00000000017,
       0.00000002421,     0.00000001850,
      -0.00000000033,    -0.00000000539,
      -0.00000001788,     0.00000002433,
       0.00000000545,    -0.00000000039,
      -0.00000000103,    -0.00000000030,
       0.00000000074,     0.00000000058,
      -0.00000000031,     0.00000000167,
      -0.00000000169,    -0.00000000032,
       0.00000000802,    -0.00000000083,
      -0.00000000058,    -0.00000006834,
      -0.00000000678,     0.00000001513,
      -0.00000001201,     0.00000000028,
       0.00000000338,     0.00000000152,
       0.00000000544,     0.00000001144,
       0.00000000008,    -0.00000000234,
      -0.00000001123,     0.00000000571,
       0.00000000232,     0.00000000007,
       0.00000229177,    -0.00000064533,
      -0.00000078802,    -0.00000124994,
      -0.00000032228,     0.00000035066,
       0.00000009249,     0.00000004929,
      -0.00000001960,    -0.00000006951,
      -0.00000003791,     0.00000002391,
       0.00000001056,     0.00000000998,
       0.00000000153,    -0.00000000278,
      -0.00000016284,     0.00000002428,
       0.00000001799,     0.00000006319,
       0.00000001240,    -0.00000000558,
      -0.00000000416,    -0.00000002783,
      -0.00000001081,     0.00000000308,
       0.00000000094,     0.00000000210,
      -0.00000000035,     0.00000000068,
      -0.00000000049,     0.00000000096,
       0.00000000096,     0.00000000049,
       0.00000161126,    -0.00000023968,
      -0.00000017760,    -0.00000062527,
      -0.00000012277,     0.00000005524,
       0.00000001164,     0.00000007814,
       0.00000003032,    -0.00000000863,
      -0.00000000268,    -0.00000000594,
      -0.00000000075,     0.00000000051,
      -0.00000000243,    -0.00000000383,
      -0.00000000248,     0.00000000142,
       0.00000000049,     0.00000000060,
       0.00000000738,    -0.00000000026,
      -0.00000000108,    -0.00000000283,
      -0.00000000066,     0.00000000044,
       0.00000000503,    -0.00000001023,
       0.00000000187,     0.00000000115,
      -0.00000000523,     0.00000000065,
       0.00000000056,     0.00000000200,
      -0.00000000087,    -0.00000000504,
      -0.00000000198,     0.00000000057,
      -0.00000007217,     0.00000015328,
       0.00000004328,     0.00000002183,
      -0.00000000469,    -0.00000000308,
      -0.00000000114,     0.00000000163,
       0.00000004828,    -0.00000000704,
      -0.00000000531,    -0.00000001871,
      -0.00000000366,     0.00000000165,
       0.00000000740,     0.00000004801,
       0.00000001866,    -0.00000000534,
      -0.00000000165,    -0.00000000366,
       0.00000002036,     0.00000003134,
      -0.00000000138,     0.00000000097,
      -0.00000000189,     0.00000000370,
       0.00000000112,     0.00000000061,
      -0.00000000371,    -0.00000000190,
      -0.00000000060,     0.00000000113,
       0.00000000060,     0.00000000091,
      -0.00000000090,     0.00000000060,
       0.00000000286,    -0.00000000311,
       0.00000000310,     0.00000000286,
      -0.00000000124,     0.00000000373,
       0.00000000132,     0.00000000028,
      -0.00000000091,     0.00000001226,
       0.00000000387,    -0.00000000011,
       0.00000001227,     0.00000000087,
      -0.00000000015,    -0.00000000385,
      -0.00000000053,     0.00000000043,
       0.00000000014,     0.00000000013,
       0.00000000045,     0.00000000066,
       0.00000000017,    -0.00000000015,
       0.00000000009,     0.00000000045,
       0.00000000064,     0.00000000018,
       0.00000000018,    -0.00000000064,
      -0.00000000039,     0.00000000031,
       0.00000000031,     0.00000000039,
       0.00000000041,     0.00000000106,
       0.00000000037,    -0.00000000025,
       0.00000000124,     0.00000000282,
       0.00000000089,    -0.00000000059,
       0.00000000282,    -0.00000000125,
      -0.00000000059,    -0.00000000089,
      -0.00000000058,    -0.00000000090,
      -0.00000000092,     0.00000000059,
       0.00000000668,    -0.00000000312,
      -0.00000000157,    -0.00000000295,
      -0.00000004506,    -0.00000001559,
      -0.00000000444,     0.00000001689,
      -0.00000001562,     0.00000004492,
       0.00000001681,     0.00000000448,
      -0.00000000199,    -0.00000000051,
      -0.00000000039,     0.00000000199,
      -0.00000000631,     0.00000001433,
       0.00000001097,     0.00000000215,
      -0.00000001105,    -0.00000001303,
      -0.00000000611,     0.00000000681,
       0.00000000208,     0.00000000143,
      -0.00000011951,     0.00000006867,
       0.00000002991,     0.00000004645,
       0.00000000869,    -0.00000000635,
       0.00000007400,     0.00000013937,
       0.00000006158,    -0.00000003639,
      -0.00000000991,    -0.00000001437,
      -0.00000000236,     0.00000000191,
      -0.00000000544,    -0.00000000383,
      -0.00000000085,     0.00000000145,
      -0.00000000352,     0.00000000545,
       0.00000000148,     0.00000000068,
      -0.00000001815,    -0.00000121196,
      -0.00000079083,     0.00000008332,
       0.00000005460,     0.00000026072,
       0.00000005717,    -0.00000001884,
      -0.00000000433,    -0.00000000945,
      -0.00000001072,     0.00000005414,
       0.00000002934,     0.00000000116,
      -0.00000000074,    -0.00000000788,
       0.00000056546,     0.00000047155,
       0.00000023374,    -0.00000031529,
      -0.00000008788,    -0.00000005684,
      -0.00000000929,     0.00000001589,
       0.00000000217,     0.00000000096,
       0.00000043998,    -0.00000064972,
      -0.00000037891,    -0.00000020037,
      -0.00000004024,     0.00000011196,
       0.00000002202,     0.00000000424,
      -0.00000000014,    -0.00000000315,
       0.00000001370,     0.00000001002,
       0.00000000208,    -0.00000000341,
       0.00000000952,    -0.00000001403,
      -0.00000000353,    -0.00000000181,
       0.00000003387,    -0.00000000333,
      -0.00000000562,    -0.00000001736,
      -0.00000000426,     0.00000000258,
       0.00000000064,     0.00000000065,
       0.00000010790,     0.00000046382,
       0.00000021814,    -0.00000010608,
      -0.00000003964,    -0.00000005065,
      -0.00000000759,     0.00000000884,
       0.00000000139,     0.00000000076,
       0.00000046357,    -0.00000010760,
      -0.00000010589,    -0.00000021802,
      -0.00000005060,     0.00000003954,
       0.00000000881,     0.00000000760,
       0.00000000076,    -0.00000000138,
      -0.00000000254,    -0.00000000276,
      -0.00000000082,     0.00000000095,
      -0.00000000251,     0.00000000280,
       0.00000000102,     0.00000000067,
      -0.00000000407,     0.00000000710,
       0.00000000400,     0.00000000127,
       0.00000000013,    -0.00000000110,
       0.00000013347,     0.00000058377,
       0.00000027469,    -0.00000013257,
      -0.00000004968,    -0.00000006380,
      -0.00000000952,     0.00000001107,
       0.00000000175,     0.00000000095,
       0.00000058337,    -0.00000013340,
      -0.00000013245,    -0.00000027447,
      -0.00000006377,     0.00000004962,
       0.00000001104,     0.00000000951,
       0.00000000095,    -0.00000000174,
       0.00000000001,    -0.00000000009,
      -0.00000000003,     0.00000000000,
      -0.00000000079,    -0.00000000005,
       0.00000000001,     0.00000000023,
      -0.00000000055,     0.00000000010,
       0.00000000005,     0.00000000015,
       0.00000000009,     0.00000000059,
       0.00000000016,    -0.00000000004,
       0.00000000237,     0.00000001163,
      -0.00000000561,     0.00000000082,
       0.00000000045,     0.00000000159,
       0.00000000083,     0.00000000560,
       0.00000000159,    -0.00000000045,
      -0.00000000014,     0.00000000002,
       0.00000000001,     0.00000000004,
       0.00000000010,     0.00000000063,
       0.00000000018,    -0.00000000005,
      -0.00000000001,    -0.00000000008,
      -0.00000000038,    -0.00000000098,
      -0.00000000098,     0.00000000038,
      -0.00000000001,    -0.00000000002,
      -0.00000000007,     0.00000000003,
      -0.00000000001,    -0.00000000005,
      -0.00000000398,     0.00000000062,
       0.00000000033,     0.00000000117,
       0.00000000061,     0.00000000397,
       0.00000000117,    -0.00000000033,
       0.00000002674,     0.00000013231,
       0.00000004568,    -0.00000002016,
      -0.00000000483,     0.00000000098,
       0.00000000075,     0.00000000168,
      -0.00000006361,     0.00000000929,
       0.00000000493,     0.00000001810,
       0.00000000931,     0.00000006370,
       0.00000001809,    -0.00000000495,
      -0.00000000189,     0.00000000029,
       0.00000000015,     0.00000000055,
       0.00000000071,     0.00000000458,
       0.00000000135,    -0.00000000038,
      -0.00000000079,     0.00000000036,
      -0.00000001124,     0.00000000174,
       0.00000000089,     0.00000000323,
       0.00000000165,     0.00000001138,
       0.00000000323,    -0.00000000088,
       0.00000002886,     0.00000014355,
       0.00000004952,    -0.00000002183,
      -0.00000000497,     0.00000000101,
       0.00000000077,     0.00000000173,
      -0.00000006889,     0.00000001001,
       0.00000000532,     0.00000001961,
       0.00000001003,     0.00000006900,
       0.00000001959,    -0.00000000534,
      -0.00000000715,     0.00000000110,
       0.00000000057,     0.00000000207,
       0.00000000197,     0.00000001277,
       0.00000000369,    -0.00000000101,
       0.00000000007,    -0.00000000102,
      -0.00000000102,    -0.00000000007,
      -0.00000000104,     0.00000000016,
      -0.00000000016,    -0.00000000104,
       0.00000001209,     0.00000000394,
       0.00000000107,    -0.00000000443,
       0.00000000957,    -0.00000002802,
      -0.00000001028,    -0.00000000261,
      -0.00000004342,     0.00000000618,
       0.00000000328,     0.00000001246,
       0.00000000634,     0.00000004289,
       0.00000001217,    -0.00000000337,
       0.00000000254,     0.00000001290,
      -0.00000000617,     0.00000000095,
       0.00000000049,     0.00000000176,
       0.00000000095,     0.00000000618,
       0.00000000177,    -0.00000000049,
      -0.00000003988,     0.00000000580,
       0.00000000309,     0.00000001146,
       0.00000000693,     0.00000004764,
       0.00000001365,    -0.00000000368,
       0.00000000081,     0.00000000104,
       0.00000000026,    -0.00000000046,
       0.00000000018,     0.00000000260,
       0.00000000082,    -0.00000000004,
      -0.00000000277,     0.00000000171,
       0.00000000046,     0.00000000059,
       0.00000000119,     0.00000000196,
       0.00000000041,    -0.00000000034,
      -0.00000000392,     0.00000000064,
       0.00000000033,     0.00000000117,
      -0.00000000052,    -0.00000000387,
      -0.00000000116,     0.00000000030,
      -0.00000000042,     0.00000000031,
      -0.00000000030,    -0.00000000042,
       0.00000000101,     0.00000000061,
       0.00000009068,    -0.00000013850,
      -0.00000001070,    -0.00000000693,
       0.00000000016,     0.00000000310,
       0.00000000164,    -0.00000000048,
       0.00000000312,    -0.00000000017,
      -0.00000000049,    -0.00000000164,
       0.00000000174,    -0.00000000160,
      -0.00000000089,    -0.00000000053,
      -0.00000000160,    -0.00000000174,
      -0.00000000053,     0.00000000089,
      -0.00000003895,    -0.00000012396,
      -0.00000005500,     0.00000003300,
      -0.00000000381,     0.00000000103,
       0.00000000094,     0.00000000169,
       0.00000000019,     0.00000000087,
      -0.00000003752,     0.00000000627,
      -0.00000000051,    -0.00000000235,
      -0.00000000087,     0.00000000032,
      -0.00000000069,    -0.00000000236,
      -0.00000000108,     0.00000000060,
       0.00000000231,    -0.00000000073,
      -0.00000000060,    -0.00000000108,
      -0.00000000204,     0.00000000039,
       0.00000000027,     0.00000000075,
      -0.00000000040,    -0.00000000202,
      -0.00000000075,     0.00000000027,
      -0.00000003071,    -0.00000001377,
      -0.00000000511,     0.00000001461,
       0.00000000441,    -0.00000000491,
      -0.00000000266,    -0.00000000186,
      -0.00000000040,     0.00000000072,
       0.00000000684,    -0.00000000110,
      -0.00000000065,    -0.00000000266,
      -0.00000000122,    -0.00000000680,
      -0.00000000262,     0.00000000078,
       0.00000333262,     0.00001409960,
       0.00000662802,    -0.00000324693,
      -0.00000120910,    -0.00000153820,
      -0.00000022553,     0.00000027091,
       0.00000004220,     0.00000002267,
       0.00000054747,    -0.00000010897,
      -0.00000011755,    -0.00000025848,
      -0.00000006028,     0.00000004516,
       0.00000001018,     0.00000000904,
       0.00000000085,    -0.00000000155,
       0.00000006130,     0.00000000007,
      -0.00000000211,    -0.00000002481,
      -0.00000000525,     0.00000000083,
      -0.00000002495,    -0.00000005914,
      -0.00000002081,     0.00000001448,
       0.00000000398,     0.00000000347,
       0.00000000367,    -0.00000000208,
      -0.00000000120,    -0.00000000106,
       0.00000000309,     0.00000000070,
       0.00000000021,    -0.00000000150,
       0.00000003945,     0.00000002101,
       0.00000000035,     0.00000000214,
       0.00000000085,    -0.00000000026,
       0.00000000057,     0.00000000238,
       0.00000000119,    -0.00000000056,
      -0.00000000246,     0.00000000014,
       0.00000000041,     0.00000000124,
       0.00000002327,     0.00000013650,
       0.00000006715,    -0.00000002815,
       0.00000000388,    -0.00000000058,
      -0.00000000075,    -0.00000000188,
       0.00000000196,     0.00000000012,
      -0.00000000010,     0.00000000196,
      -0.00000000053,    -0.00000000108,
       0.00000000122,     0.00000000000,
       0.00000000090,    -0.00000000218,
      -0.00000000219,    -0.00000000086,
      -0.00000173797,     0.00000026072,
       0.00000019394,     0.00000067406,
       0.00000013231,    -0.00000005957,
      -0.00000001137,    -0.00000001682,
      -0.00000006474,    -0.00000042115,
      -0.00000016318,     0.00000004770,
       0.00000001441,     0.00000003161,
       0.00000000392,    -0.00000000270,
      -0.00000000160,     0.00000000727,
       0.00000000295,    -0.00000000040,
      -0.00000000027,    -0.00000000384,
      -0.00000000182,     0.00000000083,
      -0.00000000016,     0.00000000133,
       0.00000000081,     0.00000000077,
       0.00001599560,    -0.00000318540,
      -0.00000214102,    -0.00000609011,
      -0.00000116430,     0.00000062730,
       0.00000011406,     0.00000014068,
       0.00000001148,    -0.00000001465,
       0.00000025569,     0.00000171705,
       0.00000066686,    -0.00000019096,
      -0.00000005870,    -0.00000013068,
      -0.00000001653,     0.00000001102,
       0.00000000150,     0.00000000140,
       0.00000004095,     0.00000023888,
       0.00000010783,    -0.00000004797,
      -0.00000001826,    -0.00000002477,
      -0.00000000358,     0.00000000418,
       0.00000000069,     0.00000000038,
      -0.00000019349,     0.00000010740,
       0.00000007072,     0.00000009281,
       0.00000002194,    -0.00000002224,
      -0.00000000460,    -0.00000000316,
      -0.00000000038,     0.00000000069,
      -0.00000000359,    -0.00000000561,
      -0.00000000187,     0.00000000131,
       0.00000000169,    -0.00000000462,
      -0.00000000173,    -0.00000000014,
      -0.00000000098,     0.00000000050,
      -0.00000000046,    -0.00000000106,
      -0.00000001321,    -0.00000000759,
      -0.00000000067,     0.00000000621,
       0.00000000760,    -0.00000001585,
      -0.00000000709,    -0.00000000092,
       0.00000000139,     0.00000000899,
       0.00000000461,    -0.00000000225,
      -0.00000000090,    -0.00000000079,
      -0.00000000964,     0.00000000220,
       0.00000000252,     0.00000000468,
       0.00000000097,    -0.00000000088,
       0.00000062345,    -0.00000031945,
      -0.00000012725,    -0.00000022152,
      -0.00000003698,     0.00000002847,
       0.00000009649,     0.00000023101,
       0.00000008359,    -0.00000003989,
      -0.00000000913,    -0.00000001407,
      -0.00000000129,     0.00000000109,
      -0.00000006115,    -0.00000003239,
      -0.00000001687,     0.00000004273,
       0.00000001125,     0.00000000180,
      -0.00000000074,    -0.00000000116,
       0.00000001250,    -0.00000019159,
      -0.00000006956,     0.00000000463,
       0.00000000277,     0.00000001395,
       0.00000000111,    -0.00000000077,
      -0.00000000028,    -0.00000000077,
      -0.00000890681,     0.00000506145,
       0.00000187262,     0.00000309276,
       0.00000054097,    -0.00000038956,
      -0.00000005770,    -0.00000006149,
      -0.00000048542,    -0.00000083746,
      -0.00000029044,     0.00000017635,
       0.00000003623,     0.00000005082,
       0.00000000583,    -0.00000000534,
      -0.00000000068,    -0.00000000047,
       0.00000109085,    -0.00000017687,
      -0.00000013192,    -0.00000041099,
      -0.00000007883,     0.00000003924,
       0.00000000707,     0.00000000974,
       0.00000000086,    -0.00000000087,
       0.00000016578,     0.00000096928,
       0.00000038414,    -0.00000011500,
      -0.00000003553,    -0.00000007589,
      -0.00000000950,     0.00000000676,
       0.00000000089,     0.00000000089,
       0.00000000320,     0.00000000241,
       0.00000000078,    -0.00000000095,
       0.00000000033,     0.00000000333,
       0.00000000104,    -0.00000000024,
       0.00000000067,     0.00000000034,
       0.00000005432,    -0.00000000845,
      -0.00000000874,    -0.00000001116,
      -0.00000000109,     0.00000000181,
      -0.00000000839,    -0.00000005428,
      -0.00000001114,     0.00000000870,
       0.00000000180,     0.00000000110,
       0.00000001318,    -0.00000000287,
       0.00000000029,     0.00000000131,
       0.00000001410,    -0.00000000306,
       0.00000000030,     0.00000000138,
       0.00000257743,    -0.00000039922,
      -0.00000041386,    -0.00000052934,
      -0.00000005175,     0.00000008543,
      -0.00000001024,    -0.00000006630,
      -0.00000001362,     0.00000001065,
       0.00000000220,     0.00000000132,
       0.00000000023,     0.00000000108,
      -0.00000000108,     0.00000000023,
      -0.00000000318,     0.00000000036,
      -0.00000000168,    -0.00000000130,
      -0.00000000130,     0.00000000180,
      -0.00000025019,     0.00000000408,
       0.00000002765,     0.00000006963,
       0.00000000984,    -0.00000000800,
      -0.00000000527,    -0.00000010383,
      -0.00000002848,     0.00000001174,
       0.00000000303,     0.00000000402,
       0.00000008013,    -0.00000003396,
      -0.00000001472,    -0.00000002508,
      -0.00000000444,     0.00000000319,
       0.00000004587,     0.00000006126,
       0.00000002446,    -0.00000001482,
      -0.00000000306,    -0.00000000440,
      -0.00000000058,     0.00000000050,
      -0.00000000043,     0.00000000096,
       0.00000754931,    -0.00000129836,
      -0.00000111080,    -0.00000191550,
      -0.00000025635,     0.00000023625,
       0.00000002947,     0.00000002204,
       0.00000004867,     0.00000038114,
       0.00000011092,    -0.00000004743,
      -0.00000001113,    -0.00000001694,
      -0.00000000156,     0.00000000159,
      -0.00000061390,     0.00000034870,
       0.00000012871,     0.00000021278,
       0.00000003724,    -0.00000002674,
      -0.00000000395,    -0.00000000424,
      -0.00000035577,    -0.00000059794,
      -0.00000020931,     0.00000012840,
       0.00000002634,     0.00000003682,
       0.00000000419,    -0.00000000387,
       0.00000000042,    -0.00000000499,
      -0.00000000192,    -0.00000000021,
       0.00000000396,    -0.00000000010,
       0.00000000010,    -0.00000000159,
      -0.00000000012,    -0.00000000136,
       0.00000000138,    -0.00000000004,
      -0.00000000980,    -0.00000016200,
      -0.00000005189,     0.00000000358,
      -0.00000000414,     0.00000000028,
       0.00000000009,     0.00000000134,
      -0.00000000505,    -0.00000000382,
       0.00000000902,    -0.00000001170,
      -0.00000000065,    -0.00000000217,
       0.00000000064,     0.00000000207,
       0.00000000218,    -0.00000000107,
       0.00000017920,     0.00000005973,
       0.00000000102,    -0.00000002519,
      -0.00000003867,     0.00000012332,
       0.00000001900,    -0.00000000046,
      -0.00000000142,    -0.00000000154,
      -0.00000004590,    -0.00000000306,
       0.00000000573,     0.00000001235,
       0.00000000218,    -0.00000000168,
      -0.00000001794,    -0.00000003662,
      -0.00000001230,     0.00000000590,
       0.00000000180,     0.00000000234,
       0.00000000021,    -0.00000000124,
      -0.00000306133,    -0.00000037653,
       0.00000023778,     0.00000062582,
       0.00000007630,    -0.00000006819,
       0.00000007729,    -0.00000043379,
      -0.00000008692,     0.00000002627,
       0.00000000903,     0.00000001074,
       0.00000000103,    -0.00000000157,
       0.00000050763,    -0.00000008247,
      -0.00000007184,    -0.00000013751,
      -0.00000001963,     0.00000001595,
       0.00000000247,     0.00000000193,
       0.00000009098,     0.00000050094,
       0.00000013571,    -0.00000007167,
      -0.00000001589,    -0.00000001935,
      -0.00000000197,     0.00000000260,
      -0.00000000039,     0.00000000569,
       0.00000000222,     0.00000000026,
      -0.00000000536,     0.00000000015,
      -0.00000000018,     0.00000000210,
       0.00000000114,    -0.00000000481,
      -0.00000000472,    -0.00000000132,
      -0.00000000007,    -0.00000000213,
      -0.00000000262,    -0.00000000128,
       0.00000000074,    -0.00000000291,
      -0.00000000290,    -0.00000000074,
       0.00000001839,    -0.00000009451,
      -0.00000000240,    -0.00000000047,
      -0.00000001361,    -0.00000001254,
      -0.00000013147,     0.00000014355,
       0.00000024233,     0.00000020109,
      -0.00000000877,    -0.00000003587,
       0.00000044338,    -0.00000053060,
      -0.00000007999,     0.00000001955,
      -0.00000006477,    -0.00000001130,
       0.00000000300,     0.00000000949,
       0.00000000983,    -0.00000003520,
      -0.00000001102,     0.00000000313,
      -0.00000237041,    -0.00000163842,
       0.00000011220,     0.00000043076,
       0.00000004359,    -0.00000003305,
      -0.00000053511,     0.00000079722,
       0.00000014521,    -0.00000003937,
      -0.00000001284,    -0.00000001694,
       0.00000044647,     0.00000006899,
      -0.00000002858,    -0.00000008893,
      -0.00000001150,     0.00000001014,
      -0.00000005852,     0.00000044504,
       0.00000008687,    -0.00000002827,
      -0.00000001074,    -0.00000001198,
      -0.00000000220,     0.00000000579,
      -0.00000000572,    -0.00000000152,
       0.00000000068,    -0.00000000190,
      -0.00000000196,    -0.00000000078,
       0.00000000037,    -0.00000000098,
      -0.00000000098,    -0.00000000037,
       0.00000000046,    -0.00000000122,
      -0.00000000122,    -0.00000000046,
       0.00000001032,    -0.00000002412,
       0.00000006317,    -0.00000018642,
      -0.00000000625,    -0.00000000114,
      -0.00000000041,    -0.00000000087,
       0.00000000133,    -0.00000000042,
      -0.00000000467,    -0.00000001230,
      -0.00000000367,     0.00000000144,
       0.00000000178,     0.00000000119,
      -0.00000000135,     0.00000000172,
       0.00000005066,     0.00000009488,
      -0.00000000156,    -0.00000001772,
       0.00000000157,    -0.00000000149,
      -0.00000001526,    -0.00000001075,
       0.00000000062,     0.00000000276,
       0.00000001046,    -0.00000001533,
      -0.00000000280,     0.00000000065,
       0.00000003184,    -0.00000008752,
      -0.00000000306,    -0.00000000100,
      -0.00000000044,     0.00000002898,
      -0.00000000070,     0.00000000129,
       0.00000000178,    -0.00000000438,
       0.00000000432,     0.00000000167,
       0.00000000030,     0.00000000266,
       0.00000000089,     0.00000000198,
      -0.00000000194,     0.00000000094,
      -0.00000000030,     0.00000000119,
       0.00000000075,    -0.00000000185,
       0.00000000186,     0.00000000057,
      -0.00000000501,     0.00000001833,
       0.00000000888,     0.00000000243,
      -0.00000000246,    -0.00000000276,
       0.00000000278,    -0.00000000246,
       0.00000000004,    -0.00000000119,
      -0.00000000072,    -0.00000000042,
      -0.00000000042,     0.00000000072,
       0.00000000157,    -0.00000000002,
      -0.00000000002,    -0.00000000157,
       0.00000000087,    -0.00000000015,
      -0.00000000014,    -0.00000000088,
       0.00000583578,     0.00002554545,
       0.00001201946,    -0.00000579774,
      -0.00000217228,    -0.00000279179,
      -0.00000040980,     0.00000048804,
       0.00000007629,     0.00000004142,
       0.00000074124,    -0.00000017012,
      -0.00000016862,    -0.00000034879,
      -0.00000008101,     0.00000006315,
       0.00000001405,     0.00000001208,
       0.00000000120,    -0.00000000221,
       0.00000000842,    -0.00000000241,
      -0.00000000130,    -0.00000000286,
       0.00000000058,    -0.00000001350,
      -0.00000000570,     0.00000000034,
       0.00000000020,     0.00000000124,
       0.00000000190,     0.00000000000,
       0.00000000006,     0.00000000183,
      -0.00000000463,     0.00000000830,
       0.00000001285,     0.00000000676,
       0.00000000144,    -0.00000000273,
      -0.00000001461,     0.00000000185,
       0.00000000157,     0.00000000559,
       0.00000000109,    -0.00000000049,
      -0.00000000242,    -0.00000001410,
      -0.00000000548,     0.00000000162,
       0.00000000049,     0.00000000109,
      -0.00000027848,     0.00000055946,
       0.00000016266,     0.00000007321,
       0.00000000832,    -0.00000002541,
      -0.00000002752,    -0.00000001517,
      -0.00000000414,     0.00000000797,
       0.00000000123,     0.00000000051,
       0.00000012985,    -0.00000001656,
      -0.00000001299,    -0.00000005037,
      -0.00000000992,     0.00000000405,
       0.00000000078,     0.00000000125,
       0.00000001759,     0.00000012898,
       0.00000005021,    -0.00000001308,
      -0.00000000403,    -0.00000000990,
      -0.00000000125,     0.00000000078,
      -0.00000000534,    -0.00000000625,
       0.00000000147,    -0.00000000111,
       0.00000000096,    -0.00000000195,
       0.00000000197,     0.00000000093,
      -0.00000000002,     0.00000000126,
       0.00000006778,     0.00000010151,
      -0.00000000488,     0.00000000358,
       0.00000000065,     0.00000000045,
      -0.00000000870,     0.00000001758,
       0.00000000512,     0.00000000226,
       0.00000000025,    -0.00000000080,
      -0.00000001762,    -0.00000000863,
      -0.00000000225,     0.00000000514,
       0.00000000080,     0.00000000025,
       0.00000000215,     0.00000000316,
      -0.00000000313,     0.00000000216,
      -0.00000000225,     0.00000000031,
       0.00000000034,     0.00000000078,
      -0.00000000058,    -0.00000000083,
      -0.00000000017,     0.00000000015,
      -0.00000000088,     0.00000000021,
       0.00000000003,     0.00000000021,
       0.00000000004,    -0.00000000007,
      -0.00000000014,     0.00000000000,
      -0.00000000692,     0.00000000134,
       0.00000000105,     0.00000000241,
       0.00000000306,    -0.00000000049,
      -0.00000000025,    -0.00000000085,
      -0.00000000052,    -0.00000000306,
      -0.00000000084,     0.00000000025,
      -0.00000001087,     0.00000000209,
       0.00000000160,     0.00000000375,
       0.00000000636,    -0.00000000122,
      -0.00000000058,    -0.00000000186,
      -0.00000000122,    -0.00000000643,
      -0.00000000183,     0.00000000060,
      -0.00000002495,     0.00000001537,
       0.00000000732,     0.00000000831,
       0.00000002418,    -0.00000000602,
      -0.00000000263,    -0.00000000694,
      -0.00000000415,    -0.00000002390,
      -0.00000000695,     0.00000000175,
       0.00000053222,     0.00000057698,
       0.00000026522,    -0.00000031174,
      -0.00000009252,    -0.00000006197,
      -0.00000000911,     0.00000001818,
      -0.00000040573,     0.00000029119,
       0.00000018555,     0.00000019256,
       0.00000004491,    -0.00000005717,
      -0.00000001166,    -0.00000000684,
      -0.00000122987,     0.00000019234,
       0.00000014925,     0.00000048142,
       0.00000009496,    -0.00000004735,
      -0.00000000930,    -0.00000001167,
       0.00000012989,     0.00000119970,
       0.00000046951,    -0.00000010419,
      -0.00000003144,    -0.00000009351,
      -0.00000001201,     0.00000000565,
      -0.00000007168,     0.00000000680,
       0.00000000477,     0.00000000999,
       0.00000000077,    -0.00000000080,
       0.00000000618,     0.00000007127,
       0.00000000981,    -0.00000000441,
      -0.00000000083,    -0.00000000076,
       0.00000038067,    -0.00000194750,
      -0.00000085087,    -0.00000007285,
       0.00000002056,     0.00000021693,
       0.00000003620,    -0.00000001078,
       0.00000058797,    -0.00000082333,
      -0.00000046026,    -0.00000028184,
      -0.00000006562,     0.00000013024,
       0.00000002457,     0.00000000896,
       0.00000000089,    -0.00000000341,
      -0.00000157010,     0.00000007913,
       0.00000009175,     0.00000061391,
       0.00000011936,    -0.00000002508,
      -0.00000000316,    -0.00000001631,
       0.00000036028,     0.00000165266,
       0.00000063905,    -0.00000028611,
      -0.00000009184,    -0.00000011772,
      -0.00000001367,     0.00000001883,
       0.00000000230,     0.00000000141,
       0.00000017554,    -0.00000001960,
      -0.00000001325,    -0.00000002385,
      -0.00000000169,     0.00000000219,
      -0.00000001548,    -0.00000017536,
      -0.00000002395,     0.00000001109,
       0.00000000156,     0.00000000176,
       0.00000003115,     0.00000002482,
       0.00000001074,    -0.00000001696,
      -0.00000003916,     0.00000029376,
       0.00000012679,     0.00000000232,
      -0.00000000283,    -0.00000002789,
      -0.00000000405,     0.00000000098,
      -0.00000434331,     0.00000067550,
       0.00000049198,     0.00000168334,
       0.00000033075,    -0.00000014880,
      -0.00000002759,    -0.00000004179,
      -0.00000000372,     0.00000000360,
       0.00000067482,     0.00000433898,
       0.00000168072,    -0.00000049140,
      -0.00000014830,    -0.00000032998,
      -0.00000004165,     0.00000002745,
       0.00000000354,     0.00000000373,
       0.00000000571,     0.00000000226,
       0.00000000056,    -0.00000000177,
      -0.00000000013,    -0.00000000289,
      -0.00000006818,    -0.00000003858,
      -0.00000001179,     0.00000002960,
       0.00000000654,     0.00000000153,
       0.00000000004,    -0.00000000099,
      -0.00000539355,     0.00000082852,
       0.00000060799,     0.00000209098,
       0.00000041043,    -0.00000018424,
      -0.00000003429,    -0.00000005197,
      -0.00000000466,     0.00000000449,
       0.00000082762,     0.00000538863,
       0.00000208857,    -0.00000060673,
      -0.00000018364,    -0.00000040983,
      -0.00000005187,     0.00000003411,
       0.00000000445,     0.00000000466,
      -0.00000010353,     0.00000001569,
       0.00000000821,     0.00000002945,
      -0.00000000167,    -0.00000001089,
      -0.00000000310,     0.00000000087,
       0.00000000042,    -0.00000000060,
      -0.00000000074,    -0.00000000039,
      -0.00000000766,    -0.00000001976,
       0.00000000190,    -0.00000000075,
      -0.00000115929,     0.00000017363,
       0.00000009097,     0.00000033039,
      -0.00000001819,    -0.00000011977,
      -0.00000003415,     0.00000000948,
       0.00000000472,    -0.00000000677,
      -0.00000000123,    -0.00000000095,
      -0.00000000835,    -0.00000000444,
      -0.00000000079,     0.00000000172,
       0.00000000427,    -0.00000000657,
       0.00000000064,     0.00000000041,
       0.00000000013,     0.00000000077,
       0.00000000018,    -0.00000000044,
      -0.00000124037,     0.00000018494,
       0.00000009688,     0.00000035355,
      -0.00000001906,    -0.00000012578,
      -0.00000003580,     0.00000000997,
       0.00000000503,    -0.00000000717,
      -0.00000000132,    -0.00000000102,
      -0.00000000912,    -0.00000000497,
      -0.00000000085,     0.00000000184,
      -0.00000000060,    -0.00000000025,
       0.00000000159,    -0.00000002050,
       0.00000000197,     0.00000000014,
      -0.00000000209,     0.00000000051,
       0.00000000019,     0.00000000062,
       0.00000000123,     0.00000000603,
       0.00000000179,    -0.00000000055,
       0.00000000097,    -0.00000000205,
      -0.00000000052,    -0.00000000016,
      -0.00000000128,    -0.00000000122,
      -0.00000000033,     0.00000000022,
      -0.00000010916,     0.00000001641,
       0.00000000862,     0.00000003092,
      -0.00000000165,    -0.00000001072,
      -0.00000000306,     0.00000000085,
      -0.00000000086,    -0.00000000038,
       0.00000000108,    -0.00000000155,
      -0.00000000037,    -0.00000000020,
      -0.00000000193,    -0.00000000130,
      -0.00000000024,     0.00000000039,
      -0.00000000835,     0.00000000535,
      -0.00000000051,    -0.00000000079,
      -0.00000001992,     0.00000000377,
      -0.00000000029,    -0.00000000153,
       0.00000000018,     0.00000000097,
      -0.00000002989,     0.00000000570,
      -0.00000000046,    -0.00000000240,
       0.00000003178,    -0.00000000608,
       0.00000000028,     0.00000000152,
      -0.00000000102,     0.00000000019,
       0.00000000101,    -0.00000000019,
       0.00000000018,     0.00000000098,
      -0.00000004426,     0.00000000849,
      -0.00000000073,    -0.00000000375,
      -0.00000000085,     0.00000000061,
       0.00000004810,    -0.00000000907,
       0.00000000045,     0.00000000236,
      -0.00000000015,    -0.00000000166,
      -0.00000000155,     0.00000000031,
      -0.00000000027,    -0.00000000142,
       0.00000000009,     0.00000000107,
       0.00000000158,    -0.00000000030,
       0.00000000029,     0.00000000151,
       0.00000000019,     0.00000000104,
       0.00000000104,    -0.00000000018,
      -0.00000000011,    -0.00000000145,
      -0.00000006443,     0.00000001253,
       0.00000001018,     0.00000001409,
      -0.00000000116,    -0.00000000587,
      -0.00000000133,     0.00000000096,
       0.00000011607,    -0.00000001333,
      -0.00000001282,    -0.00000003217,
       0.00000000074,     0.00000000233,
       0.00000000081,    -0.00000000058,
      -0.00000002527,     0.00000000151,
      -0.00000000025,    -0.00000000245,
      -0.00000000238,     0.00000000048,
      -0.00000000043,    -0.00000000223,
       0.00000003067,    -0.00000000270,
       0.00000000014,     0.00000000158,
       0.00000000347,    -0.00000000043,
       0.00000000043,     0.00000000328,
      -0.00000000447,    -0.00000000237,
       0.00000000296,    -0.00000000440,
       0.00000000448,     0.00000000280,
       0.00000002738,    -0.00000000918,
       0.00000001135,     0.00000000484,
       0.00000000034,     0.00000000121,
       0.00000000314,    -0.00000000436,
       0.00000000455,     0.00000000305,
      -0.00000000008,    -0.00000000113,
       0.00000000106,    -0.00000000010,
       0.00000000008,     0.00000000104,
      -0.00000000128,     0.00000000122,
       0.00000000066,     0.00000000038,
       0.00000000122,     0.00000000129,
       0.00000000035,    -0.00000000066,
      -0.00000001542,     0.00000003328,
       0.00000001680,     0.00000000770,
       0.00000000018,    -0.00000000249,
      -0.00000000188,     0.00000000000,
      -0.00000001947,    -0.00000000785,
      -0.00000000267,     0.00000000911,
       0.00000000222,     0.00000000048,
      -0.00000000781,     0.00000001982,
       0.00000000931,     0.00000000257,
       0.00000000040,    -0.00000000225,
       0.00000001039,     0.00000001007,
       0.00000000274,    -0.00000000398,
      -0.00000000079,    -0.00000000037,
       0.00000001002,    -0.00000001041,
      -0.00000000398,    -0.00000000273,
      -0.00000000037,     0.00000000079,
      -0.00000000107,     0.00000000032,
       0.00000123814,    -0.00000028667,
      -0.00000017501,    -0.00000045409,
      -0.00000008424,     0.00000004734,
      -0.00000000756,    -0.00000003305,
      -0.00000001244,     0.00000000479,
       0.00000000130,     0.00000000230,
       0.00000000188,    -0.00000000017,
       0.00000000016,     0.00000000175,
      -0.00000013854,     0.00000001820,
       0.00000001757,     0.00000003733,
      -0.00000000196,    -0.00000000737,
      -0.00000000209,     0.00000000152,
       0.00000013267,    -0.00000003060,
      -0.00000001757,    -0.00000003330,
       0.00000000305,     0.00000000654,
       0.00000000127,    -0.00000000091,
       0.00000002431,    -0.00000000537,
      -0.00000000339,    -0.00000000895,
      -0.00000000165,     0.00000000093,
       0.00000000534,     0.00000002435,
       0.00000000894,    -0.00000000338,
      -0.00000000093,    -0.00000000165,
      -0.00000002918,     0.00000000284,
      -0.00000000044,    -0.00000000373,
      -0.00000000080,     0.00000000053,
      -0.00000000371,     0.00000000077,
       0.00000000061,     0.00000000083,
      -0.00000000070,    -0.00000000353,
      -0.00000000082,     0.00000000060,
       0.00000003425,    -0.00000001005,
       0.00000000023,     0.00000000240,
       0.00000000513,    -0.00000000204,
      -0.00000000060,    -0.00000000083,
       0.00000000201,     0.00000000502,
       0.00000000081,    -0.00000000059,
      -0.00000000008,    -0.00000000118,
      -0.00000000168,     0.00000000015,
      -0.00000000015,    -0.00000000181,
       0.00000000167,    -0.00000000018,
       0.00000000015,     0.00000000165,
      -0.00000000005,     0.00000000106,
      -0.00000000090,     0.00000000232,
       0.00000000232,     0.00000000098,
      -0.00000021887,     0.00000003816,
       0.00000002758,     0.00000008404,
       0.00000001623,    -0.00000000831,
      -0.00000002418,    -0.00000012647,
      -0.00000004897,     0.00000001665,
       0.00000000459,     0.00000000902,
       0.00000000100,    -0.00000000079,
      -0.00000000565,    -0.00000000420,
      -0.00000000106,     0.00000000277,
       0.00000000097,    -0.00000000027,
       0.00000000042,     0.00000000119,
       0.00000000040,     0.00000000063,
      -0.00014589156,     0.00002252466,
       0.00001645578,     0.00005654364,
       0.00001109676,    -0.00000497909,
      -0.00000092404,    -0.00000140272,
      -0.00000012645,     0.00000012163,
       0.00000083166,     0.00000526410,
       0.00000203896,    -0.00000060319,
      -0.00000018183,    -0.00000039969,
      -0.00000005057,     0.00000003359,
       0.00000000438,     0.00000000445,
      -0.00000007963,     0.00000001478,
       0.00000001056,     0.00000003398,
       0.00000000705,    -0.00000000376,
      -0.00000000088,    -0.00000000095,
      -0.00000003232,    -0.00000019498,
      -0.00000007327,     0.00000002731,
       0.00000000836,     0.00000001372,
       0.00000000161,    -0.00000000141,
       0.00000002194,     0.00000001721,
       0.00000000414,    -0.00000000841,
      -0.00000000164,    -0.00000000043,
      -0.00000002366,     0.00000002101,
       0.00000000788,     0.00000000694,
       0.00000000107,    -0.00000000155,
      -0.00000025169,     0.00000002913,
       0.00000002533,     0.00000009824,
      -0.00000002327,     0.00000000099,
       0.00000000240,     0.00000001000,
       0.00000000204,    -0.00000000081,
      -0.00000000352,    -0.00000002606,
      -0.00000001001,     0.00000000238,
       0.00000000081,     0.00000000204,
      -0.00000000097,     0.00000000025,
       0.00000000025,     0.00000000097,
      -0.00000000103,     0.00000000026,
       0.00000000026,     0.00000000103,
      -0.00000134651,     0.00000012745,
       0.00000013067,     0.00000054655,
       0.00000011171,    -0.00000004509,
       0.00000000332,     0.00000003441,
       0.00000001359,    -0.00000000324,
      -0.00000000110,    -0.00000000277,
       0.00000000021,     0.00000000092,
      -0.00000000870,     0.00000001952,
      -0.00000000257,    -0.00000000150,
      -0.00000000049,     0.00000000071,
       0.00000007683,    -0.00000007000,
      -0.00000002312,    -0.00000001870,
      -0.00000000096,    -0.00000000010,
      -0.00000000404,     0.00000000142,
       0.00000000057,     0.00000000150,
      -0.00000000142,    -0.00000000403,
      -0.00000000150,     0.00000000057,
      -0.00000005277,     0.00000000836,
      -0.00000000257,    -0.00000001476,
      -0.00000000293,     0.00000000222,
      -0.00000001180,     0.00000000266,
       0.00000000204,     0.00000000266,
      -0.00000000251,    -0.00000001189,
      -0.00000000276,     0.00000000206,
       0.00000008551,    -0.00000001071,
       0.00000000115,     0.00000000958,
       0.00000000204,    -0.00000000138,
       0.00000001196,    -0.00000000266,
      -0.00000000189,    -0.00000000257,
       0.00000000259,     0.00000001184,
       0.00000000259,    -0.00000000190,
       0.00000000020,     0.00000001240,
      -0.00000000260,    -0.00000004130,
      -0.00000006291,     0.00000000617,
       0.00000000545,     0.00000005540,
       0.00000002106,    -0.00000000992,
      -0.00000000202,    -0.00000000290,
       0.00000000929,     0.00000002445,
       0.00000000290,    -0.00000000202,
      -0.00000005090,     0.00000000224,
       0.00000000017,     0.00000001066,
       0.00000000303,    -0.00000000213,
      -0.00000001082,     0.00000000140,
       0.00000000185,     0.00000000263,
      -0.00000000092,    -0.00000001045,
      -0.00000000263,     0.00000000184,
       0.00000001557,    -0.00000003862,
      -0.00000001069,    -0.00000000246,
       0.00000007077,     0.00000003222,
       0.00000000508,    -0.00000001997,
      -0.00000000302,     0.00000000023,
       0.00000002489,     0.00000003526,
       0.00000000395,    -0.00000000414,
       0.00000003388,    -0.00000002500,
      -0.00000000442,    -0.00000000385,
       0.00000046374,    -0.00000052277,
      -0.00000015062,    -0.00000014303,
      -0.00000002297,     0.00000002441,
       0.00000049463,     0.00000038401,
       0.00000011319,    -0.00000014154,
      -0.00000002106,    -0.00000001647,
      -0.00000000103,     0.00000000207,
      -0.00000038767,    -0.00000010318,
       0.00000001435,     0.00000012741,
       0.00000002198,    -0.00000001019,
      -0.00000000154,    -0.00000000265,
      -0.00000018571,    -0.00000018168,
      -0.00000009207,     0.00000004869,
       0.00000000973,     0.00000002148,
       0.00000000274,    -0.00000000188,
      -0.00000000230,    -0.00000000302,
      -0.00000000306,     0.00000000034,
      -0.00000254426,     0.00001409012,
       0.00000328462,    -0.00000009666,
      -0.00000015635,    -0.00000033973,
      -0.00000001196,     0.00000002907,
      -0.00000298415,    -0.00000206509,
      -0.00000056113,     0.00000085771,
       0.00000013096,     0.00000007637,
       0.00000000693,    -0.00000001383,
      -0.00000000121,    -0.00000000031,
      -0.00000215160,     0.00000034767,
       0.00000024584,     0.00000082252,
       0.00000016052,    -0.00000007234,
      -0.00000001297,    -0.00000002010,
      -0.00000000193,     0.00000000174,
      -0.00000029340,    -0.00000201726,
      -0.00000079237,     0.00000021801,
       0.00000006768,     0.00000015664,
       0.00000002022,    -0.00000001307,
      -0.00000000168,    -0.00000000180,
       0.00000001016,     0.00000000425,
       0.00000000140,    -0.00000000267,
       0.00000000541,     0.00000000775,
       0.00000000183,    -0.00000000175,
       0.00000001203,    -0.00000002024,
       0.00000000281,     0.00000000163,
      -0.00000000609,     0.00000000060,
       0.00000000092,     0.00000000129,
       0.00000000181,     0.00000000630,
       0.00000000122,    -0.00000000091,
      -0.00000003916,     0.00000000991,
      -0.00000000049,    -0.00000000191,
      -0.00000004143,     0.00000001048,
      -0.00000000052,    -0.00000000201,
      -0.00000005841,     0.00000001783,
       0.00000001091,     0.00000002391,
       0.00000000041,     0.00000000139,
      -0.00000000176,    -0.00000000151,
       0.00000001638,    -0.00000001445,
       0.00000000095,     0.00000000005,
      -0.00000000084,     0.00000000020,
      -0.00000000023,    -0.00000000089,
      -0.00000004166,     0.00000001502,
      -0.00000000513,    -0.00000001470,
      -0.00000000351,     0.00000000289,
      -0.00000000101,     0.00000000158,
       0.00000000125,     0.00000000072,
      -0.00000000130,    -0.00000000151,
      -0.00000000056,     0.00000000086,
      -0.00000030974,     0.00000005344,
       0.00000005010,     0.00000006977,
       0.00000000459,     0.00000001387,
       0.00000000303,    -0.00000000229,
       0.00000000119,    -0.00000000214,
       0.00000000205,     0.00000000114,
      -0.00000000090,     0.00000000018,
      -0.00000000060,    -0.00000000143,
      -0.00000000005,    -0.00000000049,
      -0.00000000019,    -0.00000000005,
      -0.00000000189,    -0.00000000331,
       0.00000000027,    -0.00000000103,
      -0.00000000089,    -0.00000000021,
      -0.00000000327,    -0.00000000587,
       0.00000000065,    -0.00000000225,
      -0.00000000209,    -0.00000000049,
      -0.00000000294,    -0.00000000871,
       0.00000000225,    -0.00000000673,
      -0.00000000623,    -0.00000000106,
      -0.00000123265,     0.00000022888,
       0.00000021001,     0.00000024313,
       0.00000002394,    -0.00000004485,
       0.00000715974,     0.00004591314,
       0.00000903064,    -0.00000733246,
      -0.00000174197,    -0.00000092995,
       0.00000043326,    -0.00000100587,
      -0.00000015297,    -0.00000002386,
      -0.00000107643,    -0.00000020576,
       0.00000013056,     0.00000015154,
       0.00000001850,    -0.00000000989,
      -0.00000002758,     0.00000000273,
       0.00001779589,    -0.00000344001,
      -0.00000304355,    -0.00000361979,
      -0.00000034234,     0.00000057043,
       0.00000002686,     0.00000001184,
      -0.00002352879,    -0.00014410756,
      -0.00002939355,     0.00002355185,
       0.00000470232,     0.00000283645,
       0.00000010171,    -0.00000022786,
      -0.00000363963,     0.00000596391,
       0.00000118701,     0.00000054015,
       0.00000009557,    -0.00000016462,
      -0.00000001260,    -0.00000001236,
       0.00000256470,    -0.00000094307,
      -0.00000102057,    -0.00000000272,
       0.00000008633,     0.00000012973,
       0.00000001237,    -0.00000001258,
      -0.00000000923,    -0.00000000368,
       0.00000003238,     0.00000000267,
      -0.00010764343,     0.00002360911,
       0.00001778971,     0.00002366959,
       0.00000264395,    -0.00000328290,
      -0.00000043946,    -0.00000019755,
       0.00002768911,     0.00014277808,
       0.00003130602,    -0.00002291044,
      -0.00000412347,    -0.00000346618,
      -0.00000027873,     0.00000061790,
       0.00000996990,    -0.00001393331,
      -0.00000402559,    -0.00000269784,
      -0.00000036439,     0.00000061056,
       0.00000007337,     0.00000003762,
       0.00001418015,     0.00000954471,
       0.00000263770,    -0.00000403768,
      -0.00000060721,    -0.00000035970,
      -0.00000003618,     0.00000007196,
      -0.00000001182,     0.00000011474,
       0.00000005440,     0.00000000831,
      -0.00000007269,     0.00000000638,
      -0.00000000663,     0.00000004107,
      -0.00000000831,    -0.00000001031,
      -0.00000000272,     0.00000000241,
      -0.00000000557,     0.00000000865,
      -0.00000000894,    -0.00000000480,
       0.00000004126,    -0.00000001307,
      -0.00000000850,    -0.00000001672,
      -0.00000000044,    -0.00000000134,
      -0.00000000441,    -0.00000000158,
      -0.00000000114,     0.00000000189,
       0.00000000144,    -0.00000000024,
      -0.00000000085,    -0.00000000057,
      -0.00000002458,     0.00000007060,
       0.00000000421,     0.00000000198,
       0.00000000057,    -0.00000000091,
       0.00000000133,     0.00000000066,
       0.00000001249,    -0.00000002149,
       0.00000001897,     0.00000001124,
       0.00000002450,    -0.00000004192,
      -0.00000000308,    -0.00000000176,
      -0.00000000606,    -0.00000000075,
      -0.00000000073,     0.00000000540,
      -0.00000025897,    -0.00000013593,
      -0.00000000939,     0.00000002536,
      -0.00000005914,     0.00000010624,
       0.00000000919,     0.00000000546,
       0.00000005897,    -0.00000000458,
      -0.00000000864,    -0.00000000825,
      -0.00000000108,     0.00000000157,
       0.00000001337,     0.00000000949,
       0.00000000835,    -0.00000000409,
      -0.00000000166,    -0.00000000116,
       0.00000000009,     0.00000000107,
      -0.00000000095,     0.00000000032,
       0.00000183187,     0.00000049085,
      -0.00000007461,    -0.00000027098,
      -0.00000002664,     0.00000003059,
       0.00000000035,    -0.00000002324,
      -0.00000000409,     0.00000000313,
       0.00000000065,     0.00000000057,
      -0.00000017064,     0.00000004316,
       0.00000003013,     0.00000003898,
       0.00000000450,    -0.00000000594,
      -0.00000000126,    -0.00000000062,
      -0.00000004995,    -0.00000016564,
      -0.00000003803,     0.00000002990,
       0.00000000584,     0.00000000441,
       0.00000000062,    -0.00000000126,
      -0.00000000089,    -0.00000000484,
      -0.00000000164,     0.00000000028,
       0.00000000447,    -0.00000000135,
      -0.00000000034,    -0.00000000155,
       0.00000000622,    -0.00000000119,
      -0.00000000097,    -0.00000000132,
       0.00000000123,     0.00000000629,
       0.00000000132,    -0.00000000098,
       0.00000001797,    -0.00000001946,
      -0.00000000092,     0.00000000512,
      -0.00000000045,    -0.00000000079,
       0.00000000601,     0.00000000306,
       0.00000000091,    -0.00000000049,
      -0.00000000041,     0.00000000098,
       0.00000000103,     0.00000000041,
       0.00000000176,    -0.00000000438,
      -0.00000000465,    -0.00000000174,
      -0.00000000032,     0.00000000085,
      -0.00000001581,     0.00000004720,
       0.00000000145,    -0.00000000015,
      -0.00000000164,     0.00000000316,
       0.00000000084,     0.00000000041,
       0.00000000390,     0.00000000171,
       0.00000000039,    -0.00000000082,
       0.00000002574,    -0.00000000280,
      -0.00000000043,    -0.00000000740,
      -0.00000000107,     0.00000000055,
      -0.00000000074,    -0.00000000051,
       0.00000033391,    -0.00000007292,
      -0.00000005648,    -0.00000007502,
       0.00000000123,     0.00000001013,
       0.00000000218,    -0.00000000154,
      -0.00000000287,    -0.00000000517,
      -0.00000000644,     0.00000000357,
       0.00000000114,     0.00000000121,
       0.00000000093,    -0.00000000089,
       0.00000003592,     0.00000005588,
       0.00000001474,    -0.00000000728,
      -0.00000000098,    -0.00000000056,
      -0.00000001028,    -0.00000000598,
       0.00000000112,     0.00000000125,
      -0.00000000394,    -0.00000000122,
      -0.00000000086,     0.00000000039,
      -0.00000022846,    -0.00000019825,
       0.00000000751,     0.00000003042,
       0.00000000474,    -0.00000000323,
       0.00000003252,     0.00000001124,
      -0.00000000055,    -0.00000000428,
      -0.00000000061,     0.00000000081,
      -0.00000001047,     0.00000003246,
       0.00000000416,    -0.00000000051,
      -0.00000000087,    -0.00000000063,
      -0.00000000136,     0.00000000371,
       0.00000000375,     0.00000000129,
       0.00000000074,    -0.00000000195,
      -0.00000000210,    -0.00000000075,
      -0.00000000380,     0.00000001600,
       0.00000254170,     0.00000061520,
      -0.00000000870,    -0.00000001200,
      -0.00000001190,     0.00000000870,
       0.00000005590,    -0.00000022030,
      -0.00000765002,    -0.00000193740,
       0.00000002750,     0.00000004230,
       0.00000005430,    -0.00000004530,
      -0.00000019415,     0.00000098102,
       0.00000018565,     0.00000018303,
       0.00000483909,     0.00000004725,
       0.00000110285,    -0.00000112443,
      -0.00000003708,     0.00000014157,
       0.00000007004,     0.00000001031,
       0.00000047198,    -0.00000123395,
      -0.00000248378,    -0.00000095240,
       0.00000002391,    -0.00000011895,
      -0.00000011721,    -0.00000002272,
      -0.00000000849,     0.00000005466,
       0.00000001530,     0.00000000586,
       0.00000001502,     0.00000000268,
       0.00000000157,    -0.00000000410,
       0.00000000794,    -0.00000000007,
      -0.00000000139,    -0.00000000185,
      -0.00000000289,    -0.00000000896,
      -0.00000000194,     0.00000000141,
       0.00000002388,    -0.00000014062,
      -0.00000002843,    -0.00000000983,
      -0.00000000928,    -0.00000000250,
      -0.00000000094,     0.00000000231,
      -0.00000000866,     0.00000000248,
       0.00000000128,     0.00000000178,
       0.00000000050,     0.00000000801,
       0.00000000187,    -0.00000000140,
      -0.00000002937,     0.00000001155,
       0.00000000019,    -0.00000000103,
      -0.00000001506,    -0.00000000455,
      -0.00000000071,     0.00000000397,
       0.00000000066,    -0.00000000189,
      -0.00000000272,     0.00000000052,
       0.00000001874,     0.00000004062,
      -0.00000000560,    -0.00000000532,
       0.00000000035,     0.00000000093,
       0.00000000525,    -0.00000000562,
      -0.00000000096,     0.00000000035,
       0.00000000030,    -0.00000000089,
      -0.00000000089,    -0.00000000030,
       0.00000109110,     0.00000041150,
      -0.00000001870,     0.00000004990,
      -0.00000321530,    -0.00000120700,
      -0.00000001360,    -0.00000001180,
      -0.00000001170,     0.00000001360,
       0.00000016160,    -0.00000047390,
       0.00000528904,     0.00000180825,
       0.00000002267,    -0.00000007237,
      -0.00000006982,    -0.00000002321,
      -0.00000030751,     0.00000044443,
       0.00000012784,     0.00000009072,
      -0.00000250682,    -0.00000175029,
      -0.00000032973,     0.00000046466,
      -0.00000001908,     0.00000007071,
       0.00000007259,     0.00000002315,
       0.00000000154,     0.00000000042,
       0.00000000055,    -0.00000000068,
      -0.00000000036,     0.00000000314,
       0.00000000150,    -0.00000000026,
       0.00000001522,    -0.00000004762,
      -0.00000000119,    -0.00000000058,
      -0.00000000004,    -0.00000000146,
      -0.00000000261,    -0.00000000017,
      -0.00000000559,     0.00000003803,
       0.00000000421,     0.00000000005,
       0.00000000028,    -0.00000000107,
       0.00000000663,    -0.00000000633,
      -0.00000000079,    -0.00000000107,
       0.00000000449,    -0.00000000316,
      -0.00000000111,     0.00000000080,
      -0.00000000048,    -0.00000003415,
      -0.00000000237,    -0.00000000003,
      -0.00000000471,     0.00000000037,
       0.00000000077,     0.00000000106,
       0.00000000157,     0.00000000515,
       0.00000000110,    -0.00000000083,
       0.00000010536,    -0.00000002112,
      -0.00000001702,    -0.00000002303,
      -0.00000000231,    -0.00000001201,
      -0.00000000258,     0.00000000190,
      -0.00000011283,     0.00000002421,
       0.00000001832,     0.00000002461,
       0.00000000140,     0.00000000750,
       0.00000000165,    -0.00000000123,
       0.00000000043,     0.00000000103,
      -0.00000000101,     0.00000000045,
       0.00000000940,     0.00000000640,
      -0.00000002770,    -0.00000001740,
       0.00000000939,    -0.00000001827,
       0.00000005329,     0.00000002730,
      -0.00000000816,     0.00000002722,
      -0.00000002575,    -0.00000000775,
       0.00000000086,     0.00000000030,
       0.00000000130,    -0.00000000224,
       0.00000000352,     0.00000000123,
      -0.00000000006,    -0.00000000114,
       0.00000000165,    -0.00000000061,
       0.00000000299,    -0.00000000005,
      -0.00000000051,    -0.00000000069,
      -0.00000000120,    -0.00000000331,
      -0.00000000071,     0.00000000052,
      -0.00000000094,     0.00000000038,
      -0.00000000317,     0.00000000055,
       0.00000000050,     0.00000000069,
       0.00000000073,     0.00000000324,
       0.00000000072,    -0.00000000053,
       0.00000007939,    -0.00000001785,
      -0.00000001305,    -0.00000001752,
      -0.00000000157,    -0.00000000792,
      -0.00000000172,     0.00000000129,
      -0.00000008509,     0.00000001673,
       0.00000001385,     0.00000001865,
       0.00000000098,     0.00000000498,
       0.00000000109,    -0.00000000082,
      -0.00000000113,    -0.00000000040,
       0.00000000083,     0.00000000028,
       0.00000000073,    -0.00000000067,
       0.00000000201,    -0.00000000012,
      -0.00000000073,    -0.00000000214,
      -0.00000000212,     0.00000000040,
       0.00000000046,     0.00000000213,
       0.00000005764,    -0.00000001171,
      -0.00000000952,    -0.00000001270,
      -0.00000000106,    -0.00000000521,
      -0.00000000113,     0.00000000084,
      -0.00000006144,     0.00000001215,
       0.00000000066,     0.00000000329,
       0.00000000072,    -0.00000000055,
       0.00000000135,    -0.00000000014,
      -0.00000000042,    -0.00000000138,
      -0.00000000142,     0.00000000029,
       0.00000000029,     0.00000000140,
       0.00000004047,    -0.00000000831,
      -0.00000000071,    -0.00000000341,
      -0.00000000075,     0.00000000056,
      -0.00000004314,     0.00000000879,
       0.00000000044,     0.00000000217,
      -0.00000000094,     0.00000000020,
       0.00000000018,     0.00000000093,
       0.00000002787,    -0.00000000575,
      -0.00000000048,    -0.00000000224,
      -0.00000002970,     0.00000000612,
       0.00000000029,     0.00000000142,
       0.00000001891,    -0.00000000394,
      -0.00000000031,    -0.00000000146,
       0.00000000019,     0.00000000093,
      -0.00000008092,     0.00000001212,
       0.00000001915,     0.00000003999,
      -0.00000000094,    -0.00000000392,
      -0.00000000188,     0.00000000101,
       0.00000000337,    -0.00000000695,
      -0.00000000304,    -0.00000000092,
      -0.00000000696,    -0.00000000338,
      -0.00000000092,     0.00000000305,
      -0.00000000053,     0.00000001264,
       0.00000000397,     0.00000000055,
       0.00000001264,     0.00000000053,
       0.00000000055,    -0.00000000397,
      -0.00000003252,     0.00000000607,
       0.00000000420,     0.00000001294,
      -0.00000000279,    -0.00000001531,
      -0.00000000598,     0.00000000182,
       0.00000000056,     0.00000000118,
       0.00000000075,    -0.00000000176,
      -0.00000000078,    -0.00000000045,
      -0.00025373837,     0.00003896836,
       0.00002858716,     0.00009835813,
       0.00001930407,    -0.00000865805,
      -0.00000160696,    -0.00000243963,
      -0.00000021974,     0.00000021123,
       0.00000001889,     0.00000001327,
       0.00000101461,     0.00000659535,
       0.00000255664,    -0.00000074356,
      -0.00000022511,    -0.00000050173,
      -0.00000006352,     0.00000004185,
       0.00000000546,     0.00000000571,
       0.00000000439,    -0.00000002132,
      -0.00000000953,    -0.00000000139,
      -0.00000000017,     0.00000000209,
       0.00000002288,     0.00000002815,
       0.00000001006,    -0.00000000928,
      -0.00000000199,    -0.00000000185,
      -0.00000000079,     0.00000001785,
       0.00000000559,     0.00000000080,
       0.00000000022,    -0.00000000106,
      -0.00000001802,    -0.00000000071,
      -0.00000000077,     0.00000000570,
       0.00000000106,     0.00000000022,
      -0.00000000031,     0.00000000143,
      -0.00000000109,     0.00000000122,
      -0.00000000130,     0.00000000003,
      -0.00000002638,    -0.00000002358,
       0.00000000755,    -0.00000000933,
      -0.00000000107,    -0.00000000046,
       0.00000000731,    -0.00000001112,
      -0.00000000351,    -0.00000000126,
       0.00000001038,     0.00000000246,
       0.00000000081,    -0.00000000232,
      -0.00000012593,     0.00000002572,
       0.00000002009,     0.00000002758,
      -0.00000000310,    -0.00000001475,
      -0.00000000335,     0.00000000247,
       0.00000043927,     0.00000037318,
       0.00000000284,    -0.00000008144,
      -0.00000002529,     0.00000003013,
       0.00000000580,     0.00000000065,
      -0.00000009042,     0.00000007460,
       0.00000001964,     0.00000002812,
       0.00000000471,    -0.00000000276,
      -0.00000007478,    -0.00000008983,
      -0.00000002809,     0.00000001970,
       0.00000000276,     0.00000000468,
       0.00000000009,    -0.00000000222,
       0.00000001371,    -0.00000000031,
      -0.00000000021,     0.00000000641,
      -0.00000000159,    -0.00000000219,
       0.00000000219,    -0.00000000159,
      -0.00000004160,     0.00000001640,
      -0.00000000315,    -0.00000000803,
      -0.00000000131,     0.00000000090,
      -0.00000000749,     0.00000000138,
       0.00000000123,     0.00000000251,
      -0.00000000133,    -0.00000000740,
      -0.00000000135,     0.00000000100,
      -0.00000002406,    -0.00000001380,
       0.00000000122,    -0.00000000145,
       0.00000000087,    -0.00000000056,
       0.00000001972,     0.00000001720,
       0.00000000034,    -0.00000000375,
      -0.00000001704,     0.00000001966,
       0.00000000372,     0.00000000037,
      -0.00000000141,     0.00000000251,
      -0.00000000030,    -0.00000000322,
      -0.00000000343,     0.00000000031,
       0.00000000057,     0.00000000084,
      -0.00000000033,    -0.00000000382,
      -0.00000000084,     0.00000000057,
       0.00000000548,     0.00000002488,
      -0.00000000259,     0.00000000092,
      -0.00000000128,    -0.00000000102,
      -0.00000000055,    -0.00000000080,
       0.00000000095,    -0.00000000140,
       0.00000000080,    -0.00000000055,
      -0.00000000316,     0.00000000245,
      -0.00000000039,    -0.00000000214,
       0.00000000239,     0.00000000176,
       0.00000008133,     0.00000022844,
       0.00000006308,    -0.00000002645,
       0.00000000016,     0.00000006122,
       0.00000001486,    -0.00000000179,
       0.00000003042,     0.00000000722,
       0.00000000069,    -0.00000000518,
       0.00000000003,    -0.00000001295,
      -0.00000000126,     0.00000000128,
      -0.00000000523,     0.00000000425,
       0.00000000270,    -0.00000000104,
       0.00000026991,     0.00000049274,
       0.00000014244,    -0.00000008299,
      -0.00000007240,     0.00000015780,
       0.00000003377,     0.00000001032,
       0.00000016152,     0.00000007976,
       0.00000001309,    -0.00000003494,
       0.00000000395,    -0.00000002732,
      -0.00000000168,     0.00000000212,
      -0.00000001750,    -0.00000000725,
       0.00000000161,    -0.00000000032,
       0.00000046532,     0.00000090478,
       0.00000025081,    -0.00000014463,
      -0.00000002190,    -0.00000003420,
      -0.00000016650,     0.00000037839,
       0.00000008065,     0.00000002275,
       0.00000037919,     0.00000018837,
       0.00000003057,    -0.00000008083,
       0.00000001796,    -0.00000002717,
       0.00000000239,    -0.00000000030,
      -0.00000003219,    -0.00000001763,
       0.00000000056,    -0.00000000132,
       0.00000058934,     0.00000125090,
       0.00000036355,    -0.00000017351,
      -0.00000002560,    -0.00000005280,
      -0.00000000510,     0.00000000250,
      -0.00000057415,     0.00000111387,
       0.00000023568,     0.00000009376,
       0.00000000710,    -0.00000002470,
       0.00000113667,     0.00000046960,
       0.00000005635,    -0.00000024136,
      -0.00000002490,    -0.00000000170,
       0.00000003063,    -0.00000006939,
       0.00000000107,     0.00000000382,
      -0.00000005832,    -0.00000002592,
       0.00000000506,    -0.00000000303,
      -0.00025260528,     0.00003894248,
       0.00002823935,     0.00009777997,
       0.00001923539,    -0.00000843562,
      -0.00000154462,    -0.00000244482,
      -0.00000022326,     0.00000020098,
       0.00000002011,     0.00000001139,
      -0.00089235612,    -0.00566521210,
      -0.00219566446,     0.00064610882,
       0.00019271697,     0.00043113277,
       0.00005464241,    -0.00003520953,
      -0.00000455085,    -0.00000491064,
      -0.00000024452,     0.00000043182,
       0.00016083469,    -0.00034435270,
      -0.00009834200,    -0.00003075536,
      -0.00000202081,     0.00001428373,
       0.00000135255,    -0.00000009593,
      -0.00000002363,    -0.00000010031,
      -0.00034105272,    -0.00016040679,
      -0.00003096601,     0.00009681265,
       0.00001392619,     0.00000218521,
      -0.00000005751,    -0.00000131626,
      -0.00000009655,     0.00000003324,
       0.00000412451,    -0.00000750700,
       0.00000001248,     0.00000009174,
       0.00000000330,    -0.00000002377,
      -0.00000782894,    -0.00000334187,
       0.00000028882,     0.00000005175,
      -0.00000001662,    -0.00000003349,
       0.00285693187,    -0.00044004347,
      -0.00031927642,    -0.00110592096,
      -0.00021754469,     0.00009539853,
       0.00001749679,     0.00002764778,
       0.00000253236,    -0.00000229074,
      -0.00000020767,    -0.00000016068,
       0.00219678075,     0.01394457457,
       0.00540436797,    -0.00159042271,
      -0.00047446996,    -0.00106110851,
      -0.00013451043,     0.00008683609,
       0.00001122948,     0.00001220933,
       0.00000075447,    -0.00000097509,
      -0.00063653258,     0.00141684802,
       0.00039982852,     0.00012277317,
       0.00000831777,    -0.00005786103,
      -0.00000544667,     0.00000031052,
       0.00000009728,     0.00000036769,
       0.00140901059,     0.00063511675,
       0.00012319663,    -0.00039630663,
      -0.00005710566,    -0.00000865504,
       0.00000024273,     0.00000535428,
       0.00000036674,    -0.00000011169,
      -0.00000971448,     0.00001815071,
      -0.00000023320,    -0.00000034418,
      -0.00000001886,     0.00000011205,
       0.00001870831,     0.00000869841,
      -0.00000059473,     0.00000012539,
       0.00000010262,     0.00000004533,
       0.00000044289,     0.00000097310,
       0.00000025372,    -0.00000032964,
      -0.00000008270,    -0.00000001932,
      -0.00000025697,    -0.00000119138,
      -0.00000040481,     0.00000016740,
       0.00000003895,     0.00000007509,
       0.00000001714,    -0.00000000918,
       0.00000044116,    -0.00000104741,
      -0.00000030635,    -0.00000006156,
       0.00000000159,     0.00000004565,
      -0.00000102912,    -0.00000037513,
      -0.00000003454,     0.00000028603,
       0.00000003970,    -0.00000000549,
      -0.00000002180,     0.00000001251,
       0.00000000599,     0.00000000380,
      -0.00000001309,     0.00000002951,
       0.00000000898,     0.00000000066,
      -0.00000001248,    -0.00000023708,
      -0.00000009276,     0.00000002223,
       0.00000000759,     0.00000001997,
       0.00000000266,    -0.00000000138,
       0.00000004684,    -0.00000010977,
      -0.00000003129,    -0.00000000876,
      -0.00000000046,     0.00000000581,
       0.00000000340,    -0.00000000009,
      -0.00000010655,    -0.00000004981,
      -0.00000000993,     0.00000002928,
       0.00000000546,     0.00000000069,
       0.00000000017,    -0.00000000323,
      -0.00000000048,     0.00000000329,
       0.00000000259,     0.00000000114,
       0.00000000004,     0.00000000001,
      -0.00000000497,    -0.00000000225,
      -0.00000000036,     0.00000000111,
       0.00000001098,    -0.00000002221,
       0.00000002981,     0.00000001331,
       0.00000000205,    -0.00000000645,
      -0.00000000158,    -0.00000000071,
      -0.00000000513,    -0.00000000234,
      -0.00000000038,     0.00000000117,
       0.00000001200,    -0.00000002398,
       0.00000003140,     0.00000001413,
       0.00000000214,    -0.00000000674,
       0.00000000123,    -0.00000000275,
      -0.00000000058,    -0.00000000018,
      -0.00000000872,    -0.00000000401,
      -0.00000000055,     0.00000000184,
      -0.00000000047,     0.00000000084,
      -0.00000000013,    -0.00000000062,
       0.00000000242,     0.00000000108,
       0.00000000046,     0.00000000060,
       0.00000000050,    -0.00000000028,
      -0.00000000063,    -0.00000000211,
      -0.00000000079,     0.00000000071,
      -0.00000000222,     0.00000000061,
       0.00000000065,     0.00000000083,
      -0.00000001080,    -0.00000000560,
      -0.00000000014,     0.00000000109,
       0.00000000104,    -0.00000000004,
       0.00000001006,     0.00000003478,
       0.00000001482,    -0.00000000962,
      -0.00000000187,     0.00000000196,
       0.00000000125,     0.00000000076,
       0.00000000188,     0.00000000195,
       0.00000000077,    -0.00000000128,
       0.00000000087,     0.00000000026,
       0.00000002940,     0.00000001290,
      -0.00000000400,     0.00000000900,
      -0.00000007240,    -0.00000003250,
      -0.00000001362,     0.00000000873,
      -0.00000001226,    -0.00000001913,
       0.00000000737,     0.00000000556,
       0.00000000100,    -0.00000000047,
      -0.00000000038,    -0.00000000147,
       0.00000000131,    -0.00000000056,
      -0.00000000729,    -0.00000000644,
      -0.00000000137,     0.00000000028,
       0.00000000023,     0.00000000334,
      -0.00000000426,    -0.00000000063,
       0.00000000049,    -0.00000000155,
      -0.00000000075,     0.00000000079,
      -0.00000000178,    -0.00000000219,
      -0.00000000083,     0.00000000184,
      -0.00000000223,     0.00000000227,
       0.00000000192,     0.00000000059,
      -0.00000000118,     0.00000000007,
       0.00000000013,     0.00000000114,
      -0.00000002497,     0.00000004548,
       0.00000003148,     0.00000000675,
       0.00000000628,     0.00000000807,
       0.00000000265,    -0.00000000464,
      -0.00000000137,    -0.00000000022,
       0.00000003223,     0.00000001996,
       0.00000000340,    -0.00000001517,
      -0.00000000347,     0.00000000027,
       0.00000001779,    -0.00000003203,
      -0.00000001459,    -0.00000000233,
       0.00000000073,     0.00000000328,
       0.00000000682,    -0.00000000219,
      -0.00000000068,    -0.00000000138,
      -0.00000000208,    -0.00000000706,
      -0.00000000147,     0.00000000073,
      -0.00000001552,     0.00000004957,
      -0.00000000109,    -0.00000000042,
      -0.00000000072,     0.00000000062,
      -0.00000000089,     0.00000000235,
      -0.00000000234,    -0.00000000084,
       0.00000000853,    -0.00000002029,
      -0.00000000541,    -0.00000000140,
      -0.00000004839,    -0.00000001849,
      -0.00000000292,     0.00000001120,
       0.00000000202,     0.00000000014,
       0.00000001172,     0.00000001768,
       0.00000000191,    -0.00000000191,
       0.00000001740,    -0.00000001179,
      -0.00000000154,    -0.00000000221,
      -0.00000008998,     0.00000021149,
       0.00000005718,     0.00000001721,
       0.00000002016,     0.00000000756,
       0.00000000215,    -0.00000000376,
      -0.00000006994,    -0.00000006790,
      -0.00000000467,     0.00000001277,
      -0.00000006857,     0.00000005635,
       0.00000000755,     0.00000000665,
       0.00000000180,     0.00000000062,
       0.00000000192,     0.00000000032,
       0.00000476960,    -0.00001037184,
      -0.00000292296,    -0.00000094672,
      -0.00000007112,     0.00000042369,
       0.00000003992,    -0.00000000095,
       0.00000058049,     0.00000023769,
       0.00000003699,    -0.00000016471,
      -0.00000002359,    -0.00000000058,
       0.00000000044,     0.00000000209,
      -0.00000002269,    -0.00000000137,
       0.00000000528,    -0.00000000329,
      -0.00000000256,    -0.00000000073,
      -0.00000001176,     0.00000010824,
       0.00000003138,    -0.00000000817,
      -0.00000000276,    -0.00000000535,
      -0.00000001115,    -0.00000000268,
      -0.00000000060,     0.00000000260,
      -0.00000000788,    -0.00000000494,
      -0.00000000060,     0.00000000228,
       0.00000000079,    -0.00000000150,
       0.00000000149,     0.00000000069,
       0.00000002456,    -0.00000004664,
      -0.00000000127,     0.00000000039,
       0.00000000039,     0.00000000127,
      -0.00000006267,    -0.00000008962,
      -0.00000001004,     0.00000001066,
      -0.00000011834,     0.00000008728,
       0.00000001415,     0.00000001342,
       0.00000005922,    -0.00000000968,
      -0.00000000130,    -0.00000000044,
      -0.00000000160,    -0.00000005402,
       0.00000000046,    -0.00000000137,
       0.00000061116,     0.00000089284,
       0.00000010176,    -0.00000010489,
      -0.00000000991,    -0.00000000446,
       0.00000028247,    -0.00000019141,
      -0.00000003103,    -0.00000003513,
      -0.00000000148,     0.00000000356,
      -0.00000026576,     0.00000007969,
       0.00000001947,     0.00000000361,
       0.00000000020,    -0.00000000252,
      -0.00000003420,     0.00000021429,
      -0.00000000605,     0.00000001306,
       0.00000000261,     0.00000000026,
      -0.00000472964,     0.00000033029,
       0.00000069098,     0.00000094755,
       0.00000009029,    -0.00000013658,
      -0.00000000130,    -0.00000019533,
      -0.00000003912,     0.00000002603,
       0.00000000586,     0.00000000374,
       0.00000012540,    -0.00000029300,
      -0.00000008162,    -0.00000002251,
      -0.00000000121,     0.00000001178,
       0.00000000175,    -0.00000000013,
       0.00000029886,     0.00000012383,
       0.00000002350,    -0.00000008353,
      -0.00000001196,    -0.00000000135,
       0.00000000005,     0.00000000180,
       0.00000000160,    -0.00000000065,
       0.00000000073,     0.00000000154,
       0.00000000041,     0.00000000070,
       0.00000002774,    -0.00000000140,
       0.00000000081,     0.00000001601,
      -0.00000000607,    -0.00000000087,
      -0.00000000064,     0.00000000682,
      -0.00000004468,     0.00000000878,
      -0.00000000021,    -0.00000000108,
       0.00000003297,    -0.00000010998,
      -0.00000001496,     0.00000000719,
      -0.00000010585,    -0.00000002885,
       0.00000000716,     0.00000001427,
       0.00000004245,     0.00000003486,
      -0.00000000224,    -0.00000000295,
       0.00000003457,    -0.00000003129,
      -0.00000000056,     0.00000000073,
      -0.00000029349,     0.00000101158,
       0.00000013580,    -0.00000006323,
      -0.00000000493,    -0.00000000906,
       0.00000026151,     0.00000006571,
      -0.00000001595,    -0.00000003523,
      -0.00000000230,     0.00000000193,
      -0.00000023024,    -0.00000013701,
       0.00000001794,     0.00000002366,
       0.00000000184,    -0.00000000216,
      -0.00000015909,     0.00000007018,
      -0.00000000955,     0.00000000683,
       0.00000000349,     0.00000000181,
       0.00000000100,    -0.00000000021,
       0.00000059880,     0.00000022055,
       0.00000000274,    -0.00000005087,
      -0.00000001062,     0.00000002495,
       0.00000000209,     0.00000000041,
      -0.00000009994,    -0.00000000003,
       0.00000001376,     0.00000001942,
       0.00000000177,    -0.00000000364,
      -0.00000000156,    -0.00000009904,
      -0.00000001930,     0.00000001367,
       0.00000000366,     0.00000000175,
      -0.00000000099,    -0.00000000179,
       0.00000000160,    -0.00000000120,
      -0.00000000047,     0.00000000060,
      -0.00000005526,     0.00000000950,
       0.00000000917,     0.00000001122,
       0.00000000107,    -0.00000000182,
      -0.00000000941,    -0.00000005533,
      -0.00000001122,     0.00000000917,
       0.00000000182,     0.00000000107,
       0.00000001857,    -0.00000000656,
       0.00000000065,    -0.00000000219,
      -0.00000000127,     0.00000000414,
       0.00000000403,     0.00000000130,
      -0.00000000034,     0.00000000084,
       0.00000000084,     0.00000000033,
      -0.00000261105,     0.00000044650,
       0.00000043268,     0.00000052939,
       0.00000005055,    -0.00000008600,
      -0.00000001160,    -0.00000006769,
      -0.00000001373,     0.00000001122,
       0.00000000223,     0.00000000131,
      -0.00000001056,    -0.00000000237,
      -0.00000000170,     0.00000000798,
      -0.00000000080,    -0.00000000132,
      -0.00000000135,     0.00000000087,
       0.00000010228,     0.00000002412,
      -0.00000001011,     0.00000000042,
       0.00000000307,    -0.00000001783,
       0.00000000009,     0.00000000207,
       0.00000000013,     0.00000000457,
       0.00000000601,    -0.00000000621,
      -0.00000011965,    -0.00000009708,
       0.00000000421,    -0.00000000351,
       0.00000001993,     0.00000000707,
      -0.00000000048,    -0.00000000190,
      -0.00000000679,     0.00000002007,
       0.00000000187,    -0.00000000044,
       0.00000000067,    -0.00000000122,
      -0.00000000060,     0.00000000208,
       0.00000000185,     0.00000000075,
      -0.00000005529,     0.00000019067,
      -0.00000000682,    -0.00000000186,
       0.00000000058,     0.00000000135,
       0.00000000135,    -0.00000000058,
       0.00000003727,     0.00000016069,
       0.00000004774,    -0.00000000999,
      -0.00000000420,     0.00000000095,
       0.00000000025,     0.00000000124,
       0.00000000308,     0.00000000353,
       0.00000000223,    -0.00000000203,
      -0.00000002508,    -0.00000002906,
      -0.00000000587,     0.00000000513,
       0.00000000062,    -0.00000000084,
       0.00000000017,    -0.00000000104,
      -0.00000000139,     0.00000000040,
       0.00000001310,     0.00000002419,
      -0.00000000384,    -0.00000000335,
       0.00000000276,    -0.00000000336,
      -0.00000003127,     0.00000008940,
      -0.00000000325,    -0.00000000118,
      -0.00000001723,     0.00000009120,
      -0.00000000240,    -0.00000000045,
       0.00000000076,    -0.00000000038,
      -0.00000000150,     0.00000000472,
      -0.00000000490,    -0.00000000095,
      -0.00000000060,     0.00000000237,
      -0.00000000255,    -0.00000000062,
      -0.00000000073,     0.00000000291,
      -0.00000000291,    -0.00000000073,
       0.00000000106,     0.00000000001,
       0.00000000043,    -0.00000001250,
      -0.00000000274,    -0.00000000005,
      -0.00000001033,     0.00000002346,
      -0.00000000087,     0.00000000192,
      -0.00000000204,    -0.00000000066,
      -0.00000000036,     0.00000000097,
      -0.00000000097,    -0.00000000036,
      -0.00000000046,     0.00000000122,
      -0.00000000122,    -0.00000000046,
       0.00000000139,    -0.00000000104,
       0.00000003030,     0.00000005012,
       0.00000002262,    -0.00000002147,
       0.00000000193,    -0.00000000104,
      -0.00000000063,    -0.00000000080,
       0.00000000889,    -0.00000000075,
      -0.00000000068,    -0.00000000382,
      -0.00000000076,     0.00000000029,
      -0.00000000073,    -0.00000000891,
      -0.00000000380,     0.00000000064,
       0.00000000443,    -0.00000000942,
      -0.00000004580,    -0.00000002164,
      -0.00000000531,     0.00000001850,
       0.00000000555,     0.00000000833,
       0.00000000881,    -0.00000000497,
      -0.00000005165,     0.00000010649,
       0.00000002945,     0.00000000892,
       0.00000009151,     0.00000004429,
       0.00000000763,    -0.00000002640,
      -0.00000002353,    -0.00000003587,
      -0.00000000367,     0.00000000350,
      -0.00000003574,     0.00000002048,
       0.00000000278,     0.00000000413,
       0.00000400477,    -0.00000920032,
      -0.00000257645,    -0.00000076874,
      -0.00000005171,     0.00000037094,
       0.00000003557,    -0.00000000483,
       0.00000014696,     0.00000006445,
       0.00000001252,    -0.00000004120,
      -0.00000000616,    -0.00000000057,
       0.00000000748,    -0.00000000519,
      -0.00000000105,    -0.00000000421,
      -0.00000000085,     0.00000000035,
      -0.00000000681,     0.00000001851,
       0.00000000705,    -0.00000000031,
      -0.00000000040,    -0.00000000121,
      -0.00000000178,    -0.00000000106,
      -0.00000000014,    -0.00000000130,
      -0.00000000079,    -0.00000000559,
      -0.00000003125,     0.00000000746,
       0.00000000235,     0.00000001185,
      -0.00000000018,     0.00000000463,
      -0.00000000086,    -0.00000000074,
       0.00000000283,     0.00000000290,
       0.00000013072,     0.00000000920,
      -0.00000000058,     0.00000000360,
      -0.00000001181,    -0.00000001664,
      -0.00000000137,     0.00000000218,
       0.00000001644,    -0.00000001187,
      -0.00000000217,    -0.00000000137,
       0.00000000096,    -0.00000000019,
       0.00000000275,     0.00000000019,
      -0.00000000019,     0.00000000275,
       0.00000000986,     0.00000000387,
       0.00000000043,    -0.00000000210,
       0.00000000044,     0.00000000055,
       0.00000000015,    -0.00000000003,
      -0.00000000075,    -0.00000000089,
      -0.00000000024,     0.00000000026,
      -0.00000000016,    -0.00000000005,
       0.00000000000,     0.00000000005,
       0.00000000039,     0.00000000014,
       0.00000000003,    -0.00000000009,
       0.00000002639,     0.00000000918,
       0.00000000083,    -0.00000000569,
      -0.00000000320,    -0.00000000021,
       0.00000000013,     0.00000000086,
      -0.00000000101,    -0.00000000233,
      -0.00000000072,     0.00000000026,
       0.00000004196,     0.00000001455,
       0.00000000129,    -0.00000000906,
      -0.00000000749,    -0.00000000110,
       0.00000000030,     0.00000000189,
      -0.00000000302,    -0.00000000484,
      -0.00000000160,     0.00000000067,
       0.00000000063,     0.00000000028,
       0.00000008985,     0.00000003625,
       0.00000000389,    -0.00000001954,
      -0.00000002360,    -0.00000000245,
       0.00000000146,     0.00000000617,
      -0.00000001438,    -0.00000001647,
      -0.00000000489,     0.00000000337,
       0.00000000067,     0.00000000029,
       0.00000000162,     0.00000000054,
      -0.00002787902,     0.00005921575,
       0.00001676342,     0.00000536681,
       0.00000039465,    -0.00000241639,
      -0.00000022663,     0.00000000506,
       0.00000000423,     0.00000001577,
       0.00006334507,     0.00002505476,
       0.00000492176,    -0.00001758085,
      -0.00000251653,    -0.00000035654,
       0.00000000845,     0.00000023472,
       0.00000001632,    -0.00000000452,
      -0.00003438791,    -0.00005509354,
      -0.00000681224,     0.00000580070,
       0.00000055215,     0.00000036930,
       0.00000000641,    -0.00000003886,
      -0.00005466305,     0.00003332743,
       0.00000536412,     0.00000676220,
       0.00000038168,    -0.00000046273,
      -0.00000002710,    -0.00000001043,
      -0.00000030569,    -0.00000061795,
       0.00000009014,    -0.00000008245,
       0.00000000139,     0.00000000168,
      -0.00000053414,     0.00000048653,
      -0.00000005969,    -0.00000010122,
       0.00000000047,    -0.00000000288,
       0.00022720221,    -0.00049301066,
      -0.00013866185,    -0.00004385682,
      -0.00000323223,     0.00001995180,
       0.00000187076,    -0.00000003981,
      -0.00000003405,    -0.00000013141,
      -0.00000000637,     0.00000000391,
      -0.00015494480,    -0.00006171989,
      -0.00001220037,     0.00004302931,
       0.00000615960,     0.00000090257,
      -0.00000001632,    -0.00000057388,
      -0.00000004026,     0.00000001016,
       0.00015252409,     0.00022608630,
       0.00002764284,    -0.00002639717,
      -0.00000272204,    -0.00000132778,
       0.00000001065,     0.00000021508,
       0.00000001361,    -0.00000000687,
       0.00022765849,    -0.00013602529,
      -0.00001998718,    -0.00002922883,
      -0.00000183834,     0.00000146486,
       0.00000005443,     0.00000008529,
       0.00000000466,     0.00000000121,
       0.00000076087,     0.00000139754,
      -0.00000025838,     0.00000022326,
       0.00000000235,     0.00000000329,
       0.00000000187,    -0.00000000258,
       0.00000131702,    -0.00000103383,
       0.00000018955,     0.00000026891,
       0.00000000307,    -0.00000000089,
      -0.00000000268,    -0.00000000182,
      -0.00000024821,     0.00000030217,
       0.00000010785,     0.00000003869,
       0.00000034992,     0.00000035761,
       0.00000010524,    -0.00000012281,
      -0.00000002272,    -0.00000001609,
      -0.00000000158,     0.00000000285,
       0.00000427758,    -0.00000064735,
      -0.00000048505,    -0.00000164684,
      -0.00000032191,     0.00000014301,
       0.00000002567,     0.00000004091,
       0.00000000371,    -0.00000000323,
       0.00000069709,     0.00000415655,
       0.00000162288,    -0.00000048046,
      -0.00000014071,    -0.00000032033,
      -0.00000004081,     0.00000002551,
       0.00000000325,     0.00000000370,
      -0.00000000410,     0.00000000316,
       0.00000000154,     0.00000000084,
      -0.00000000281,    -0.00000000004,
       0.00000000041,     0.00000000082,
      -0.00000003925,     0.00000003711,
       0.00000008109,    -0.00000004209,
      -0.00000002215,    -0.00000002561,
      -0.00000000401,     0.00000000522,
       0.00000000081,     0.00000000039,
       0.00000526020,    -0.00000080361,
      -0.00000059716,    -0.00000203154,
      -0.00000039860,     0.00000017638,
       0.00000003179,     0.00000005056,
       0.00000000465,    -0.00000000406,
       0.00000087810,     0.00000518957,
       0.00000201810,    -0.00000060355,
      -0.00000017573,    -0.00000039746,
      -0.00000005058,     0.00000003174,
       0.00000000406,     0.00000000465,
       0.00000003176,     0.00000004940,
       0.00000000166,    -0.00000000106,
       0.00000000066,    -0.00000000385,
      -0.00000000090,    -0.00000000016,
       0.00000000385,     0.00000000185,
       0.00000000016,    -0.00000000090,
       0.00000003466,     0.00000005330,
       0.00000000183,    -0.00000000117,
       0.00000000076,    -0.00000000397,
      -0.00000000094,    -0.00000000017,
       0.00000000401,     0.00000000180,
       0.00000000016,    -0.00000000094,
      -0.00000000018,    -0.00000000068,
      -0.00000000177,     0.00000000034,
       0.00000000066,    -0.00000000125,
       0.00000000135,     0.00000000028,
       0.00000000082,     0.00000000208,
      -0.00000000034,    -0.00000000090,
       0.00000000113,     0.00000000283,
       0.00000000105,    -0.00000000115,
       0.00000000284,    -0.00000000117,
      -0.00000000124,    -0.00000000101,
      -0.00000000103,    -0.00000000012,
       0.00000001822,    -0.00000001727,
       0.00000000662,     0.00000000614,
       0.00000000281,    -0.00000000203,
      -0.00000000116,    -0.00000000084,
       0.00000000081,    -0.00000000033,
      -0.00000000074,    -0.00000000044,
      -0.00000000207,     0.00000000544,
       0.00000000263,    -0.00000000009,
       0.00000000547,     0.00000000200,
      -0.00000000014,    -0.00000000263,
       0.00000005455,     0.00000012119,
       0.00000004565,    -0.00000005639,
      -0.00000002122,    -0.00000000417,
      -0.00000000215,     0.00000000059,
       0.00000000053,     0.00000000103,
       0.00000000209,    -0.00000004327,
      -0.00000002059,     0.00000000770,
       0.00000000435,     0.00000000441,
       0.00000000050,    -0.00000000090,
      -0.00000004341,    -0.00000000213,
       0.00000000770,     0.00000002065,
       0.00000000432,    -0.00000000451,
      -0.00000000090,    -0.00000000050,
      -0.00000000145,    -0.00000000057,
      -0.00000000061,     0.00000000143,
       0.00000000222,     0.00000000229,
      -0.00000000097,    -0.00000000043,
      -0.00000001468,     0.00000000665,
       0.00000000082,    -0.00000000134,
      -0.00000000022,     0.00000000098,
       0.00000000102,    -0.00000000011,
       0.00000000754,     0.00000003191,
       0.00000001507,    -0.00000000923,
      -0.00000000389,    -0.00000000316,
      -0.00000026426,     0.00000005933,
       0.00000007498,     0.00000012336,
       0.00000002661,    -0.00000003275,
      -0.00000004767,    -0.00000001451,
       0.00000000421,     0.00000002059,
      -0.00000001472,     0.00000004679,
       0.00000002086,    -0.00000000296,
      -0.00000002571,    -0.00000016272,
      -0.00000007701,     0.00000003937,
       0.00000001373,     0.00000001562,
       0.00000035025,    -0.00000005948,
      -0.00000009009,    -0.00000016562,
      -0.00000003472,     0.00000003243,
      -0.00000000950,    -0.00000000251,
       0.00000036688,    -0.00000008188,
      -0.00000008488,    -0.00000013104,
      -0.00000001863,     0.00000002962,
       0.00000000377,     0.00000000012,
      -0.00000008094,    -0.00000036939,
      -0.00000013112,     0.00000008585,
       0.00000002874,     0.00000001737,
       0.00000000012,    -0.00000000377,
      -0.00000000416,    -0.00000000477,
      -0.00000000524,     0.00000000376,
      -0.00000116755,    -0.00000539502,
      -0.00000243029,     0.00000152172,
       0.00000061051,     0.00000048920,
       0.00000004669,    -0.00000013898,
      -0.00000001970,    -0.00000000027,
       0.00000144794,    -0.00000047163,
      -0.00000047157,    -0.00000064420,
      -0.00000012868,     0.00000017801,
       0.00000003807,     0.00000001234,
       0.00000000009,    -0.00000000632,
      -0.00000030735,     0.00000114373,
       0.00000057602,    -0.00000002329,
      -0.00000004630,    -0.00000014043,
      -0.00000002263,     0.00000001420,
       0.00000000300,     0.00000000205,
       0.00000114253,     0.00000031422,
      -0.00000001927,    -0.00000057698,
      -0.00000014154,     0.00000004600,
       0.00000001453,     0.00000002221,
       0.00000000205,    -0.00000000299,
       0.00000002053,     0.00000002382,
       0.00000000615,    -0.00000000872,
       0.00000002478,    -0.00000001902,
      -0.00000000839,    -0.00000000666,
      -0.00000000098,    -0.00000000008,
       0.00000000231,    -0.00000000198,
       0.00000000117,     0.00000000027,
      -0.00000004238,    -0.00000000123,
      -0.00000000010,    -0.00000000146,
       0.00000000592,     0.00000000056,
      -0.00000000104,    -0.00000000437,
      -0.00000000131,    -0.00000000223,
      -0.00000000204,     0.00000000126,
      -0.00000001173,     0.00000037314,
       0.00000018223,    -0.00000004390,
      -0.00000002308,    -0.00000004438,
      -0.00000000739,     0.00000000646,
       0.00000029724,     0.00000002487,
      -0.00000002845,    -0.00000014603,
      -0.00000003569,     0.00000001704,
       0.00000000503,     0.00000000586,
       0.00000000042,    -0.00000000067,
       0.00000038563,    -0.00000004297,
      -0.00000003794,    -0.00000014976,
      -0.00000002905,     0.00000001200,
       0.00000000231,     0.00000000405,
      -0.00000004561,    -0.00000038869,
      -0.00000015106,     0.00000003954,
       0.00000001228,     0.00000002903,
       0.00000000363,    -0.00000000244,
       0.00000000071,    -0.00000000061,
      -0.00000000085,    -0.00000000221,
      -0.00000841465,    -0.00000694407,
      -0.00000292565,     0.00000438743,
       0.00000117410,     0.00000061901,
       0.00000008284,    -0.00000021196,
      -0.00000002816,    -0.00000000683,
      -0.00000140107,    -0.00000014458,
       0.00000012334,     0.00000069101,
       0.00000016724,    -0.00000007717,
      -0.00000002065,    -0.00000002570,
      -0.00000000300,     0.00000000344,
       0.00000189000,    -0.00000036942,
      -0.00000023593,    -0.00000072545,
      -0.00000014159,     0.00000006586,
       0.00000001160,     0.00000001784,
       0.00000000136,    -0.00000000142,
      -0.00000037647,    -0.00000187366,
      -0.00000071664,     0.00000023733,
       0.00000006564,     0.00000013906,
       0.00000001741,    -0.00000001142,
      -0.00000000154,    -0.00000000166,
      -0.00000000220,     0.00000000454,
       0.00000000138,    -0.00000000104,
       0.00000000408,     0.00000000596,
       0.00000000032,    -0.00000000187,
       0.00000000082,    -0.00000000024,
      -0.00000000012,    -0.00000000088,
       0.00000005888,     0.00000000071,
       0.00000000000,     0.00000000138,
       0.00000000142,    -0.00000000080,
      -0.00000000148,     0.00000000038,
       0.00000000036,     0.00000000122,
       0.00000027185,    -0.00000002341,
      -0.00000002546,    -0.00000010720,
      -0.00000002268,     0.00000000830,
      -0.00000001368,    -0.00000016500,
      -0.00000006585,     0.00000001541,
       0.00000000522,     0.00000001359,
       0.00000000183,    -0.00000000117,
       0.00000001789,    -0.00000001904,
      -0.00000000691,    -0.00000000374,
      -0.00000000065,     0.00000000108,
      -0.00000001477,    -0.00000001094,
      -0.00000000140,     0.00000000440,
       0.00000000007,     0.00000000149,
      -0.00000000006,     0.00000000145,
       0.00014289486,    -0.00002331380,
      -0.00001650221,    -0.00005533338,
      -0.00001086861,     0.00000482922,
       0.00000086892,     0.00000138261,
       0.00000012685,    -0.00000011103,
       0.00000076562,     0.00000525962,
       0.00000204679,    -0.00000058186,
      -0.00000017462,    -0.00000040326,
      -0.00000005139,     0.00000003190,
       0.00000000407,     0.00000000478,
       0.00000002906,     0.00000007117,
       0.00000001171,    -0.00000001876,
      -0.00000000390,     0.00000000061,
       0.00000008783,    -0.00000014516,
      -0.00000006892,    -0.00000001320,
       0.00000000109,     0.00000001523,
       0.00000000201,    -0.00000000068,
      -0.00000001786,     0.00000002567,
       0.00000001124,     0.00000000596,
       0.00000000110,    -0.00000000294,
       0.00000003192,     0.00000001194,
       0.00000000347,    -0.00000001287,
      -0.00000000314,    -0.00000000067,
       0.00000022514,    -0.00000006570,
      -0.00000003778,    -0.00000008471,
      -0.00000000039,     0.00000000062,
       0.00000002363,    -0.00000000493,
      -0.00000000325,    -0.00000000872,
      -0.00000000162,     0.00000000089,
      -0.00000000495,    -0.00000002364,
      -0.00000000873,     0.00000000328,
       0.00000000089,     0.00000000162,
       0.00000000270,    -0.00000000145,
      -0.00000000076,    -0.00000000104,
      -0.00000000145,    -0.00000000271,
      -0.00000000104,     0.00000000076,
       0.00000128503,    -0.00000025755,
      -0.00000017432,    -0.00000047457,
      -0.00000008838,     0.00000004905,
       0.00000000654,     0.00000003236,
       0.00000001160,    -0.00000000427,
      -0.00000000123,    -0.00000000218,
       0.00000000015,     0.00000000099,
       0.00000001537,     0.00000002248,
       0.00000001739,    -0.00000001296,
      -0.00000000188,    -0.00000000205,
      -0.00000001318,     0.00000000032,
       0.00000000050,     0.00000001302,
      -0.00000018842,    -0.00000017512,
      -0.00000001468,     0.00000003018,
      -0.00000004049,     0.00000004376,
       0.00000000727,     0.00000000287,
       0.00000005382,     0.00000000526,
       0.00000000154,    -0.00000000039,
      -0.00000000741,    -0.00000005441,
       0.00000000036,     0.00000000151,
       0.00000069372,     0.00000048381,
       0.00000001318,    -0.00000014753,
      -0.00000001609,     0.00000000518,
       0.00000003884,    -0.00000009779,
      -0.00000001867,     0.00000000426,
       0.00000000181,     0.00000000127,
       0.00000002947,    -0.00000005176,
      -0.00000001542,    -0.00000000596,
      -0.00000000065,     0.00000000246,
       0.00000004459,     0.00000002976,
       0.00000000788,    -0.00000001489,
      -0.00000000247,    -0.00000000089,
      -0.00000000747,     0.00000001421,
       0.00000000569,     0.00000000222,
       0.00000000042,    -0.00000000144,
      -0.00000001357,    -0.00000000711,
      -0.00000000223,     0.00000000552,
       0.00000000141,     0.00000000042,
      -0.00000000023,     0.00000000112,
       0.00000000112,     0.00000000023,
       0.00000004080,    -0.00000001718,
      -0.00000000856,    -0.00000001744,
       0.00000000032,     0.00000000227,
       0.00000000085,    -0.00000000042,
       0.00000000089,     0.00000000066,
       0.00000000069,    -0.00000000096,
      -0.00000001253,     0.00000003241,
      -0.00000002413,     0.00000000255,
       0.00000000242,     0.00000000978,
       0.00000000201,    -0.00000000079,
       0.00000000257,     0.00000002414,
       0.00000000978,    -0.00000000242,
      -0.00000000079,    -0.00000000201,
      -0.00000122902,     0.00000012673,
       0.00000012204,     0.00000049700,
       0.00000010148,    -0.00000004060,
      -0.00000000335,    -0.00000003219,
      -0.00000001345,     0.00000000330,
       0.00000000110,     0.00000000275,
      -0.00000000174,     0.00000000041,
       0.00000000028,     0.00000000170,
      -0.00000000765,     0.00000000248,
       0.00000015516,    -0.00000001189,
      -0.00000000845,    -0.00000010359,
       0.00000000049,     0.00000000112,
       0.00000000120,     0.00000000920,
      -0.00000000088,    -0.00000000052,
       0.00000000773,    -0.00000000224,
      -0.00000000060,     0.00000000077,
      -0.00000123154,     0.00000008673,
       0.00000001242,    -0.00000000540,
       0.00000002025,     0.00000023696,
      -0.00000000116,    -0.00000000282,
       0.00000000158,    -0.00000003263,
       0.00000000377,     0.00000000232,
      -0.00000003634,     0.00000000648,
       0.00000000281,    -0.00000000342,
       0.00000018659,    -0.00000012045,
      -0.00000001099,     0.00000004137,
       0.00000000307,     0.00000001852,
      -0.00000000112,    -0.00000000022,
      -0.00000000442,     0.00000000401,
       0.00000000158,     0.00000000034,
      -0.00000000503,    -0.00000000185,
      -0.00000000029,     0.00000000127,
      -0.00000000068,    -0.00000000097,
       0.00000000089,    -0.00000000086,
      -0.00000000087,     0.00000000034,
       0.00000000034,     0.00000000087,
      -0.00000004322,     0.00000001559,
      -0.00000000044,    -0.00000000125,
       0.00000000115,     0.00000000055,
       0.00000000050,    -0.00000000176,
      -0.00000000186,    -0.00000000054,
      -0.00000000033,    -0.00000000137,
      -0.00000000003,    -0.00000000003,
      -0.00000000005,    -0.00000000012,
       0.00000000012,    -0.00000000005,
      -0.00000000039,     0.00000000032,
      -0.00000000032,    -0.00000000039,
       0.00000014130,     0.00000011827,
      -0.00000000711,    -0.00000000593,
       0.00000007158,    -0.00000008360,
      -0.00000000372,     0.00000000433,
       0.00000000172,    -0.00000001955,
       0.00000000184,     0.00000000304,
      -0.00000001992,    -0.00000000769,
       0.00000000245,    -0.00000000107,
      -0.00000092290,    -0.00000081126,
       0.00000004607,     0.00000004070,
      -0.00000016870,     0.00000019726,
       0.00000000875,    -0.00000001005,
      -0.00000001249,     0.00000008744,
      -0.00000000692,    -0.00000001129,
       0.00000009072,     0.00000001844,
      -0.00000001185,     0.00000000580,
      -0.00000012249,    -0.00000005603,
       0.00000000260,    -0.00000000304,
       0.00000000914,    -0.00000000127,
       0.00000000149,     0.00000000921,
      -0.00000000189,    -0.00000000318,
      -0.00000000019,     0.00000000114,
       0.00000000317,    -0.00000000189,
      -0.00000000114,    -0.00000000019,
       0.00000000037,    -0.00000000090,
      -0.00000000092,    -0.00000000029,
       0.00000001109,    -0.00000002136,
      -0.00000000234,    -0.00000000069,
       0.00000000890,    -0.00000000699,
       0.00000000048,     0.00000000068,
       0.00000000137,    -0.00000000024,
      -0.00000000024,    -0.00000000138,
      -0.00000007287,    -0.00000013532,
      -0.00000000889,     0.00000004868,
       0.00000000402,    -0.00000000239,
      -0.00000000145,    -0.00000000025,
      -0.00000000302,    -0.00000000521,
      -0.00000000281,     0.00000000163,
       0.00000001236,     0.00000002618,
       0.00000000594,    -0.00000000331,
      -0.00000000077,     0.00000000064,
      -0.00000000057,     0.00000000059,
       0.00000001936,     0.00000002490,
      -0.00000000171,    -0.00000000156,
       0.00000000154,    -0.00000000173,
      -0.00000000114,    -0.00000000043,
      -0.00000000143,     0.00000000004,
       0.00000000083,    -0.00000000056,
       0.00000000057,    -0.00000000089,
      -0.00000000085,    -0.00000000056,
       0.00000002033,    -0.00000003151,
       0.00000000149,     0.00000000096,
      -0.00000000122,    -0.00000000019,
       0.00000000088,    -0.00000000072,
      -0.00000000065,    -0.00000000146,
      -0.00000000093,     0.00000000041,
       0.00000000067,    -0.00000000292,
      -0.00000000184,    -0.00000000030,
      -0.00000000292,    -0.00000000067,
      -0.00000000031,     0.00000000185,
       0.00000000215,     0.00000000284,
       0.00000000140,    -0.00000000139,
       0.00000000284,    -0.00000000215,
      -0.00000000139,    -0.00000000140,
       0.00000000786,    -0.00000007891,
      -0.00000004484,    -0.00000000347,
      -0.00000000316,    -0.00000000147,
      -0.00000000071,     0.00000000230,
       0.00000001041,    -0.00000006974,
      -0.00000003157,     0.00000000250,
       0.00000000229,     0.00000000717,
       0.00000000107,    -0.00000000062,
      -0.00000006977,    -0.00000001040,
       0.00000000247,     0.00000003159,
       0.00000000717,    -0.00000000229,
      -0.00000000062,    -0.00000000107,
       0.00000003650,    -0.00000000616,
      -0.00000000455,    -0.00000001525,
      -0.00000000327,    -0.00000002040,
      -0.00000000789,     0.00000000240,
       0.00000000067,     0.00000000152,
       0.00000000064,    -0.00000000230,
      -0.00000000208,    -0.00000000086,
       0.00024811183,    -0.00003991884,
      -0.00002850347,    -0.00009615067,
      -0.00001889730,     0.00000837383,
       0.00000151098,     0.00000240487,
       0.00000022038,    -0.00000019290,
      -0.00000001667,    -0.00000001418,
       0.00000102410,     0.00000636605,
       0.00000246714,    -0.00000073141,
      -0.00000021479,    -0.00000048454,
      -0.00000006167,     0.00000003876,
       0.00000000497,     0.00000000567,
      -0.00000002118,    -0.00000001662,
      -0.00000000736,     0.00000000868,
       0.00000000193,     0.00000000157,
       0.00000000465,     0.00000003027,
       0.00000001090,    -0.00000000203,
      -0.00000000052,    -0.00000000203,
       0.00000000361,     0.00000002121,
       0.00000000794,    -0.00000000245,
      -0.00000000050,    -0.00000000204,
       0.00000002151,    -0.00000000360,
      -0.00000000244,    -0.00000000803,
      -0.00000000206,     0.00000000049,
       0.00000000363,     0.00000000590,
       0.00000000493,    -0.00000000327,
      -0.00000000348,    -0.00000000006,
      -0.00000000001,     0.00000000344,
      -0.00000003355,    -0.00000005336,
      -0.00000000640,     0.00000000540,
      -0.00000001293,     0.00000000858,
       0.00000000133,     0.00000000158,
       0.00000001392,     0.00000000138,
      -0.00000000091,    -0.00000001474,
       0.00000113294,     0.00000119388,
       0.00000010122,    -0.00000021626,
       0.00000005562,    -0.00000005282,
      -0.00000001015,    -0.00000000479,
      -0.00000000059,     0.00000000073,
      -0.00000000184,    -0.00000001127,
      -0.00000000344,     0.00000000180,
       0.00000000051,     0.00000000053,
       0.00000001131,    -0.00000000280,
      -0.00000000181,    -0.00000000343,
      -0.00000000053,     0.00000000050,
       0.00000000248,     0.00000001471,
       0.00000000550,    -0.00000000168,
      -0.00000000034,    -0.00000000141,
      -0.00000001469,     0.00000000249,
       0.00000000168,     0.00000000550,
       0.00000000141,    -0.00000000034,
      -0.00000000079,    -0.00000000038,
      -0.00000001837,     0.00000004822,
      -0.00000000637,    -0.00000000433,
       0.00000000271,     0.00000000004,
      -0.00000000004,     0.00000000272,
       0.00000001325,    -0.00000000391,
      -0.00000000023,    -0.00000000181,
       0.00000000607,     0.00000000965,
       0.00000000257,    -0.00000000168,
      -0.00000001098,     0.00000000445,
       0.00000000151,     0.00000000269,
       0.00000000034,     0.00000000039,
       0.00000000005,    -0.00000000019,
       0.00000000020,     0.00000000049,
       0.00000000021,     0.00000000001,
       0.00000002755,    -0.00000001633,
      -0.00000000193,    -0.00000000298,
       0.00000000029,    -0.00000000030,
       0.00000000061,    -0.00000000065,
       0.00000000032,     0.00000000030,
      -0.00000000018,     0.00000000049,
       0.00000004652,    -0.00000002742,
      -0.00000000326,    -0.00000000491,
       0.00000000239,     0.00000000081,
       0.00000000030,    -0.00000000024,
      -0.00000000039,    -0.00000000123,
       0.00000000044,    -0.00000000109,
      -0.00000000044,    -0.00000000009,
       0.00000000108,     0.00000000030,
       0.00000008677,    -0.00000005932,
      -0.00000000718,    -0.00000000900,
       0.00000003870,    -0.00000003034,
      -0.00000000730,    -0.00000000742,
       0.00000002829,     0.00000003832,
       0.00000000760,    -0.00000000737,
       0.00000000180,    -0.00000000111,
      -0.00000000041,    -0.00000000065,
       0.00000000179,     0.00000000133,
       0.00000000057,    -0.00000000051,
      -0.00011446916,    -0.00018569682,
      -0.00002291845,     0.00001896195,
       0.00000165832,     0.00000130390,
       0.00000003440,    -0.00000010250,
      -0.00000000462,     0.00000000059,
      -0.00011002622,     0.00007633579,
       0.00001175264,     0.00001399370,
       0.00000079271,    -0.00000101216,
      -0.00000006278,    -0.00000002114,
       0.00000000028,     0.00000000281,
       0.00010740407,     0.00000098364,
       0.00000003166,    -0.00000025882,
      -0.00000002836,     0.00000015602,
       0.00000000153,     0.00000000228,
       0.00000038656,    -0.00010468004,
       0.00000025114,    -0.00000007896,
       0.00000013183,    -0.00000002446,
      -0.00000000183,     0.00000000182,
      -0.00000016134,     0.00000026714,
      -0.00000001202,     0.00000003124,
       0.00000000306,     0.00000000108,
      -0.00000017846,     0.00000006796,
       0.00000003380,    -0.00000000004,
      -0.00000000291,    -0.00000000474,
       0.00081310182,     0.00125946288,
       0.00015810802,    -0.00013013178,
      -0.00001127047,    -0.00000902830,
      -0.00000024681,     0.00000069716,
       0.00000003120,    -0.00000000311,
       0.00025948548,    -0.00017774861,
      -0.00002754090,    -0.00003293221,
      -0.00000187090,     0.00000237486,
       0.00000014749,     0.00000005035,
      -0.00000000068,    -0.00000000690,
      -0.00045172582,    -0.00000876450,
      -0.00000032530,     0.00000055420,
       0.00000005657,    -0.00000064098,
      -0.00000000617,    -0.00000000256,
      -0.00001126286,     0.00044681418,
      -0.00000028045,     0.00000128781,
      -0.00000048460,     0.00000000819,
      -0.00000000090,    -0.00000000920,
       0.00000066976,    -0.00000039244,
       0.00000002222,    -0.00000015826,
      -0.00000001979,    -0.00000000138,
      -0.00000000066,     0.00000000181,
       0.00000008750,    -0.00000053285,
      -0.00000016771,     0.00000001172,
       0.00000000902,     0.00000002200,
       0.00000000191,    -0.00000000125,
       0.00000037405,     0.00000001423,
      -0.00000001984,    -0.00000005430,
      -0.00000008960,    -0.00000004062,
      -0.00000000251,     0.00000002710,
       0.00000000377,    -0.00000000099,
      -0.00000044439,     0.00000131513,
       0.00000038114,     0.00000002230,
      -0.00000001096,    -0.00000005448,
      -0.00000000488,     0.00000000284,
      -0.00000132760,    -0.00000037539,
      -0.00000002230,     0.00000038237,
       0.00000005452,    -0.00000001095,
      -0.00000000284,    -0.00000000488,
       0.00000000139,    -0.00000000453,
      -0.00000000194,    -0.00000000045,
       0.00000000393,     0.00000000237,
       0.00000000076,    -0.00000000179,
       0.00000029098,     0.00000015268,
       0.00000003459,    -0.00000002306,
       0.00000001148,     0.00000001990,
       0.00000000496,    -0.00000000414,
      -0.00000000072,    -0.00000000056,
      -0.00000055468,     0.00000165398,
       0.00000047941,     0.00000002700,
      -0.00000001400,    -0.00000006857,
      -0.00000000612,     0.00000000356,
      -0.00000167024,    -0.00000047062,
      -0.00000002699,     0.00000048079,
       0.00000006854,    -0.00000001401,
      -0.00000000356,    -0.00000000611,
      -0.00000000003,    -0.00000000003,
      -0.00000000003,    -0.00000000004,
       0.00000000004,    -0.00000000002,
       0.00000000788,    -0.00000000003,
      -0.00000010509,    -0.00000000414,
      -0.00000000010,     0.00000000134,
      -0.00000000048,    -0.00000000074,
       0.00000000074,    -0.00000000048,
       0.00000000850,    -0.00000000003,
      -0.00000011223,    -0.00000000441,
      -0.00000000011,     0.00000000171,
      -0.00000000083,    -0.00000000123,
       0.00000000770,     0.00000001026,
      -0.00000000044,     0.00000000013,
      -0.00000000163,    -0.00000000278,
       0.00000000272,    -0.00000000167,
      -0.00000000959,    -0.00000000022,
      -0.00000000093,    -0.00000000111,
      -0.00000000111,     0.00000000093,
      -0.00000000415,     0.00000000350,
       0.00000000286,     0.00000000060,
      -0.00000000063,    -0.00000000532,
      -0.00000000206,     0.00000000155,
      -0.00000000533,     0.00000000063,
       0.00000000155,     0.00000000207,
      -0.00000007638,    -0.00000009769,
      -0.00000001858,     0.00000006550,
       0.00000002120,    -0.00000000825,
       0.00000001611,    -0.00000001090,
      -0.00000000978,    -0.00000000338,
       0.00000000165,     0.00000000423,
       0.00000000614,     0.00000004319,
       0.00000001690,    -0.00000001366,
      -0.00000000596,    -0.00000000177,
       0.00000004329,    -0.00000000586,
      -0.00000001399,    -0.00000001675,
      -0.00000000177,     0.00000000597,
       0.00000000332,     0.00000002973,
       0.00000001169,    -0.00000000880,
       0.00000002598,    -0.00000000475,
      -0.00000000881,    -0.00000000994,
      -0.00000000099,     0.00000000302,
       0.00000000990,    -0.00000002281,
      -0.00000001121,    -0.00000000019,
       0.00000000118,     0.00000000258,
      -0.00000002284,    -0.00000000996,
       0.00000000019,     0.00000001117,
       0.00000000250,    -0.00000000138,
      -0.00000007342,    -0.00000067202,
      -0.00000027101,     0.00000020119,
       0.00000008753,     0.00000003542,
      -0.00000000368,    -0.00000001900,
      -0.00000011755,     0.00000001989,
       0.00000003876,     0.00000004573,
       0.00000000534,    -0.00000001622,
      -0.00000000363,     0.00000000086,
      -0.00000007854,     0.00000019465,
       0.00000009531,    -0.00000000353,
      -0.00000001163,    -0.00000002138,
      -0.00000000266,     0.00000000355,
       0.00000019465,     0.00000007947,
      -0.00000000302,    -0.00000009568,
      -0.00000002131,     0.00000001155,
       0.00000000369,     0.00000000244,
       0.00000000432,     0.00000000150,
      -0.00000000004,    -0.00000000155,
       0.00000000165,    -0.00000000424,
      -0.00000000156,    -0.00000000010,
       0.00000004079,    -0.00000009277,
      -0.00000004623,     0.00000000045,
       0.00000000529,     0.00000001025,
      -0.00000004154,    -0.00000001314,
       0.00000000246,     0.00000002002,
       0.00000000431,    -0.00000000288,
      -0.00000004527,     0.00000003551,
       0.00000001944,     0.00000001423,
       0.00000000190,    -0.00000000508,
       0.00000003532,     0.00000004581,
       0.00000001451,    -0.00000001946,
      -0.00000000511,    -0.00000000171,
       0.00000000080,     0.00000000075,
       0.00000000088,    -0.00000000063,
      -0.00000072277,     0.00000179225,
       0.00000088023,    -0.00000003102,
      -0.00000009984,    -0.00000019544,
      -0.00000002394,     0.00000003206,
       0.00000018634,     0.00000006129,
      -0.00000000983,    -0.00000008934,
      -0.00000001914,     0.00000001163,
       0.00000000371,     0.00000000225,
       0.00000037980,    -0.00000029644,
      -0.00000016493,    -0.00000011904,
      -0.00000001436,     0.00000004319,
       0.00000000721,     0.00000000000,
      -0.00000000035,    -0.00000000087,
      -0.00000029499,    -0.00000038284,
      -0.00000012058,     0.00000016463,
       0.00000004341,     0.00000001480,
       0.00000000002,    -0.00000000721,
      -0.00000000081,     0.00000000041,
      -0.00000000375,    -0.00000000388,
      -0.00000000100,     0.00000000159,
      -0.00000000410,     0.00000000333,
       0.00000000148,     0.00000000113,
       0.00000001774,     0.00000000608,
      -0.00000000038,    -0.00000000730,
       0.00000000511,    -0.00000001965,
      -0.00000000773,     0.00000000085,
       0.00000000049,     0.00000000110,
      -0.00000001466,     0.00000001107,
       0.00000000450,     0.00000000186,
       0.00000001111,     0.00000001447,
       0.00000000182,    -0.00000000439,
      -0.00000053916,     0.00000023431,
       0.00000016350,     0.00000018553,
       0.00000002747,    -0.00000004740,
      -0.00000000872,    -0.00000000091,
       0.00000008451,     0.00000015961,
       0.00000005308,    -0.00000005587,
      -0.00000001583,    -0.00000000726,
       0.00000000026,     0.00000000230,
       0.00000019217,    -0.00000006922,
      -0.00000003340,    -0.00000005927,
      -0.00000000977,     0.00000000691,
       0.00000000081,     0.00000000140,
      -0.00000006794,    -0.00000019410,
      -0.00000006025,     0.00000003308,
       0.00000000662,     0.00000001031,
       0.00000000129,    -0.00000000079,
       0.00000535890,    -0.00000405790,
      -0.00000227397,    -0.00000169626,
      -0.00000020991,     0.00000059976,
       0.00000009986,    -0.00000000009,
      -0.00000042741,    -0.00000049268,
      -0.00000014843,     0.00000023404,
       0.00000006075,     0.00000001590,
      -0.00000000082,    -0.00000001007,
      -0.00000000121,     0.00000000050,
      -0.00000112765,     0.00000012629,
       0.00000011232,     0.00000044407,
       0.00000008849,    -0.00000003504,
      -0.00000000651,    -0.00000001136,
      -0.00000000112,     0.00000000089,
       0.00000011427,     0.00000113070,
       0.00000044804,    -0.00000010736,
      -0.00000003422,    -0.00000009002,
      -0.00000001172,     0.00000000648,
       0.00000000086,     0.00000000108,
       0.00000000020,     0.00000000253,
       0.00000000096,    -0.00000000053,
       0.00000000226,     0.00000000148,
       0.00000000005,    -0.00000000111,
      -0.00000000142,     0.00000000161,
       0.00000000276,    -0.00000000094,
      -0.00000000106,    -0.00000000280,
       0.00000195347,    -0.00000027474,
      -0.00000021021,    -0.00000076161,
      -0.00000015061,     0.00000006271,
       0.00000001114,     0.00000001892,
      -0.00000006450,    -0.00000046211,
      -0.00000018030,     0.00000004962,
       0.00000001511,     0.00000003568,
       0.00000000461,    -0.00000000275,
      -0.00000000776,     0.00000001993,
       0.00000000764,    -0.00000000117,
      -0.00000000116,    -0.00000000099,
       0.00000000725,     0.00000001354,
       0.00000000304,    -0.00000000146,
       0.00000000072,     0.00000000134,
      -0.00000000056,     0.00000000131,
      -0.00001664728,     0.00000058432,
       0.00000129025,     0.00000672768,
       0.00000136181,    -0.00000047004,
      -0.00000009260,    -0.00000017806,
      -0.00000001688,     0.00000001224,
       0.00000029508,     0.00000189357,
       0.00000073559,    -0.00000021220,
      -0.00000006182,    -0.00000014495,
      -0.00000001864,     0.00000001103,
       0.00000000137,     0.00000000178,
      -0.00000021219,     0.00000061597,
       0.00000017591,     0.00000001808,
      -0.00000000379,    -0.00000002478,
      -0.00000000207,     0.00000000098,
       0.00000065818,     0.00000023462,
       0.00000002192,    -0.00000019750,
      -0.00000003012,     0.00000000447,
       0.00000000130,     0.00000000293,
       0.00000000192,    -0.00000000702,
      -0.00000000302,    -0.00000000051,
      -0.00000000374,    -0.00000000529,
      -0.00000000192,     0.00000000215,
       0.00000000034,     0.00000000177,
       0.00000000014,    -0.00000000189,
      -0.00000000188,     0.00000000007,
       0.00000007465,    -0.00000007361,
      -0.00000002653,    -0.00000001676,
      -0.00000001800,    -0.00000001482,
      -0.00000000274,     0.00000000615,
      -0.00000000942,     0.00000001021,
       0.00000000234,     0.00000000246,
       0.00000000739,    -0.00000000544,
      -0.00000000317,    -0.00000000023,
      -0.00001267009,     0.00004048697,
       0.00001171191,     0.00000071156,
      -0.00000033257,    -0.00000167456,
      -0.00000014846,     0.00000008230,
      -0.00000153606,    -0.00000045579,
      -0.00000001894,     0.00000044148,
       0.00000006265,    -0.00000001400,
      -0.00000000341,    -0.00000000554,
      -0.00000004815,     0.00000003752,
       0.00000001165,     0.00000002001,
       0.00000000404,    -0.00000000252,
       0.00000003778,    -0.00000006840,
      -0.00000002819,    -0.00000000423,
       0.00000000059,     0.00000000572,
       0.00000000078,    -0.00000000030,
      -0.00000000424,    -0.00000000444,
      -0.00000000188,     0.00000000225,
      -0.00000000659,     0.00000000436,
       0.00000000284,     0.00000000272,
       0.00000000040,    -0.00000000064,
      -0.00000005279,     0.00000007537,
       0.00000002520,     0.00000001301,
       0.00000000004,     0.00000000217,
       0.00000000084,    -0.00000000019,
      -0.00000000172,     0.00000000705,
       0.00000000190,    -0.00000000001,
       0.00000000704,     0.00000000120,
      -0.00000000022,    -0.00000000187,
      -0.00000000142,     0.00000000023,
       0.00000000027,     0.00000000144,
      -0.00000000213,     0.00000000129,
       0.00000000052,     0.00000000073,
       0.00000000129,     0.00000000214,
       0.00000000073,    -0.00000000053,
      -0.00000010108,     0.00000038821,
       0.00000010609,    -0.00000000308,
      -0.00000001128,    -0.00000000242,
       0.00000000016,     0.00000000301,
      -0.00000000792,     0.00000002073,
      -0.00000000091,    -0.00000000034,
      -0.00000005034,     0.00000000126,
       0.00000000070,     0.00000002716,
       0.00000029464,    -0.00000000285,
      -0.00000000140,    -0.00000006421,
      -0.00000000212,    -0.00000000010,
       0.00000000075,     0.00000000038,
      -0.00000067407,    -0.00000012747,
      -0.00000000963,     0.00000005468,
       0.00000000019,     0.00000000093,
      -0.00000000239,     0.00000000636,
       0.00000000202,    -0.00000000038,
      -0.00000000664,    -0.00000000295,
       0.00000000038,     0.00000000214,
      -0.00000000310,    -0.00000000356,
      -0.00000000146,     0.00000000181,
       0.00000000349,    -0.00000000318,
      -0.00000000181,    -0.00000000147,
       0.00000000216,    -0.00000000038,
      -0.00000000025,    -0.00000000089,
      -0.00000000037,    -0.00000000216,
      -0.00000000089,     0.00000000024,
       0.00000000010,     0.00000000097,
      -0.00000000114,     0.00000002048,
       0.00000003819,    -0.00000001094,
      -0.00000000023,    -0.00000000257,
      -0.00000000104,     0.00000000024,
       0.00000000227,    -0.00000000684,
      -0.00000000209,    -0.00000000018,
      -0.00000000688,    -0.00000000225,
      -0.00000000017,     0.00000000214,
      -0.00000000243,     0.00000000059,
       0.00000000059,     0.00000000243,
       0.00000013057,    -0.00000035082,
      -0.00000010885,    -0.00000001195,
       0.00000001098,     0.00000000347,
       0.00000000025,    -0.00000000334,
       0.00000000451,     0.00000003721,
      -0.00000000390,    -0.00000000273,
       0.00000001875,    -0.00000000189,
      -0.00000000140,     0.00000000200,
      -0.00000000115,    -0.00000000041,
      -0.00000000022,    -0.00000000103,
       0.00000005514,    -0.00000016777,
       0.00000002064,     0.00000001739,
      -0.00000004372,     0.00000000416,
       0.00000000325,    -0.00000000447,
      -0.00000000502,     0.00000000417,
       0.00000000299,    -0.00000000602,
      -0.00000007726,    -0.00000005616,
       0.00000000145,     0.00000000020,
      -0.00000000626,    -0.00000000070,
       0.00000000086,    -0.00000000625,
      -0.00000000133,     0.00000000026,
      -0.00000000026,    -0.00000000133,
      -0.00000000156,    -0.00000000040,
      -0.00000008178,     0.00000002105,
      -0.00000000083,    -0.00000000346,
      -0.00000005848,     0.00000001109,
      -0.00000000032,    -0.00000000172,
       0.00000001583,    -0.00000008630,
       0.00000000655,     0.00000001135,
      -0.00000003965,    -0.00000000695,
       0.00000000516,    -0.00000000306,
      -0.00000000098,    -0.00000000117,
      -0.00000000050,    -0.00000000096,
      -0.00000001086,     0.00000044915,
      -0.00000003364,    -0.00000004909,
       0.00000008574,     0.00000001631,
      -0.00000001139,     0.00000000663,
      -0.00000000092,    -0.00000000655,
      -0.00000000008,    -0.00000000723,
       0.00000002927,     0.00000003493,
       0.00000000119,    -0.00000000117,
      -0.00000000134,    -0.00000000173,
       0.00000000167,    -0.00000000134,
       0.00000000146,     0.00000000010,
       0.00000000010,    -0.00000000146,
       0.00000008695,     0.00000000321,
      -0.00000000011,     0.00000000258,
       0.00000000082,     0.00000000102,
       0.00000000102,    -0.00000000082,
       0.00000000019,     0.00000000152,
      -0.00000000093,    -0.00000000056,
      -0.00000000710,    -0.00000000920,
      -0.00000000458,     0.00000000464,
       0.00000000148,     0.00000000111,
      -0.00000000920,     0.00000000712,
       0.00000000463,     0.00000000459,
       0.00000000111,    -0.00000000147,
       0.00000001474,     0.00000001810,
       0.00000000927,    -0.00000000972,
       0.00000000550,    -0.00000000371,
      -0.00000000247,    -0.00000000277,
      -0.00000000068,     0.00000000080,
       0.00000001520,    -0.00000000245,
      -0.00000000174,    -0.00000000591,
      -0.00000000118,     0.00000000053,
      -0.00000000253,    -0.00000001528,
      -0.00000000598,     0.00000000177,
       0.00000000050,     0.00000000117,
      -0.00000021824,    -0.00000029119,
      -0.00000014533,     0.00000014471,
       0.00000004572,     0.00000003525,
      -0.00000001630,     0.00000001076,
       0.00000000724,     0.00000000800,
       0.00000000192,    -0.00000000232,
      -0.00000013389,     0.00000003167,
       0.00000001814,     0.00000005207,
       0.00000001029,    -0.00000000487,
      -0.00000000085,    -0.00000000134,
       0.00000003187,     0.00000013406,
       0.00000005215,    -0.00000001828,
      -0.00000000491,    -0.00000001031,
      -0.00000000135,     0.00000000085,
       0.00000000258,    -0.00000001079,
      -0.00000000234,    -0.00000000051,
      -0.00000000021,     0.00000000153,
       0.00000000121,    -0.00000000166,
      -0.00002248577,     0.00007293572,
       0.00002106889,     0.00000118170,
      -0.00000061516,    -0.00000300769,
      -0.00000026612,     0.00000015032,
       0.00000001844,     0.00000001590,
      -0.00000210908,    -0.00000065165,
      -0.00000003460,     0.00000060917,
       0.00000008695,    -0.00000001769,
      -0.00000000450,    -0.00000000776,
      -0.00000000165,     0.00000000180,
       0.00000000052,     0.00000000232,
      -0.00000000710,    -0.00000001971,
      -0.00000000599,     0.00000000345,
       0.00000000074,     0.00000000102,
      -0.00000000387,     0.00000000241,
      -0.00000000023,     0.00000000192,
       0.00000000244,     0.00000000391,
       0.00000000192,     0.00000000023,
      -0.00000001265,    -0.00000000019,
      -0.00000000012,     0.00000000709,
       0.00000008035,     0.00000000221,
       0.00000000030,    -0.00000001642,
      -0.00000148256,    -0.00000030156,
      -0.00000001047,     0.00000006242,
       0.00000000013,     0.00000000102,
       0.00000000321,    -0.00000000188,
       0.00000000039,     0.00000000287,
      -0.00000000323,     0.00000000202,
      -0.00000000020,     0.00000000160,
      -0.00000000202,    -0.00000000324,
      -0.00000000160,    -0.00000000019,
      -0.00000000075,     0.00000000048,
      -0.00000000048,    -0.00000000076,
      -0.00000000069,    -0.00000000076,
       0.00000000031,     0.00000000112,
       0.00000000028,    -0.00000000096,
      -0.00000000160,    -0.00000002727,
       0.00000000009,     0.00000000020,
      -0.00000000580,    -0.00000000420,
       0.00000000026,     0.00000000093,
       0.00000000039,    -0.00000000510,
      -0.00000000093,     0.00000000015,
      -0.00000000019,     0.00000000082,
       0.00000000006,     0.00000000004,
      -0.00000000025,    -0.00000000115,
      -0.00000000004,     0.00000000007,
       0.00000000148,    -0.00000007025,
      -0.00000000236,    -0.00000000034,
       0.00000000040,    -0.00000000349,
      -0.00000000001,     0.00000000114,
       0.00000000000,    -0.00000000057,
       0.00000000228,    -0.00000011858,
      -0.00000000680,    -0.00000000288,
      -0.00000000028,     0.00000000069,
       0.00000000164,    -0.00000000962,
      -0.00000000054,    -0.00000000040,
       0.00000000019,    -0.00000000118,
       0.00000001049,    -0.00000023578,
      -0.00000000023,    -0.00000000040,
      -0.00000001314,     0.00000002940,
       0.00000000174,    -0.00000000152,
      -0.00000002002,     0.00000000064,
       0.00000000040,     0.00000000244,
       0.00000000012,     0.00000000127,
      -0.00000000092,    -0.00000000238,
       0.00069068268,     0.00000509943,
       0.00000021301,    -0.00000004570,
      -0.00000000817,     0.00000067916,
       0.00000266758,    -0.00032250085,
      -0.00000002656,    -0.00000010154,
       0.00000031522,     0.00000000451,
      -0.00000184183,     0.00000043556,
       0.00000037114,     0.00000028589,
      -0.00000001558,     0.00000001417,
      -0.00000460336,    -0.00000516705,
       0.00000016387,    -0.00000022383,
       0.00000002730,     0.00000000985,
      -0.00000019263,    -0.00000068288,
       0.00000004140,    -0.00000007626,
       0.00000000126,     0.00000000424,
       0.00000028787,    -0.00000050349,
       0.00000003322,     0.00000009517,
      -0.00000000127,     0.00000000302,
      -0.00308241590,    -0.00011647970,
      -0.00000698153,     0.00000173833,
       0.00000020701,    -0.00000241847,
       0.00000001359,    -0.00000001553,
      -0.00000500640,     0.00071178195,
       0.00000004189,     0.00000014053,
      -0.00000070198,    -0.00000000868,
       0.00005426918,     0.00002439356,
       0.00000103300,     0.00000102439,
      -0.00000014517,     0.00000005361,
      -0.00000000423,     0.00000000007,
       0.00001640214,    -0.00003986716,
       0.00000241456,     0.00000041305,
       0.00000013460,     0.00000004521,
      -0.00000000594,     0.00000000161,
      -0.00000062567,     0.00000167126,
       0.00000007931,     0.00000019144,
       0.00000000686,    -0.00000001167,
       0.00000070978,     0.00000126687,
       0.00000005640,    -0.00000023306,
      -0.00000000886,    -0.00000000644,
       0.00000034978,    -0.00000024229,
      -0.00000000161,     0.00000004016,
       0.00000000197,     0.00000000433,
       0.00000073564,     0.00000012407,
       0.00000006905,    -0.00000003652,
      -0.00000000141,    -0.00000000612,
      -0.00000012960,     0.00000076532,
       0.00000003510,     0.00000006995,
       0.00000000602,    -0.00000000149,
       0.00000061758,    -0.00000059337,
      -0.00000003840,     0.00000000983,
       0.00000001085,    -0.00000000916,
       0.00000000015,    -0.00000000128,
       0.00000092159,     0.00000015625,
       0.00000008670,    -0.00000004562,
      -0.00000000177,    -0.00000000768,
      -0.00000016208,     0.00000095647,
       0.00000004377,     0.00000008793,
       0.00000000756,    -0.00000000187,
       0.00000000000,    -0.00000000002,
      -0.00000000019,     0.00000000004,
      -0.00000000014,     0.00000000000,
       0.00000000000,    -0.00000000014,
      -0.00000000097,     0.00000000000,
       0.00000000001,    -0.00000000097,
       0.00000002635,     0.00000000823,
       0.00000000044,     0.00000000016,
      -0.00000000256,    -0.00000000002,
       0.00000000002,    -0.00000000256,
       0.00000002797,     0.00000000864,
      -0.00000000075,    -0.00000000022,
       0.00000000201,     0.00000000125,
      -0.00000000207,     0.00000000131,
       0.00000000284,     0.00000000383,
      -0.00000000033,     0.00000000024,
      -0.00000000973,    -0.00000000024,
       0.00000000024,    -0.00000000972,
       0.00000000670,     0.00000000500,
      -0.00000004880,    -0.00000000670,
      -0.00000000740,     0.00000000500,
       0.00000007680,    -0.00000004950,
      -0.00000000910,     0.00000000570,
      -0.00000002089,    -0.00000001104,
      -0.00000000188,     0.00000000345,
       0.00000000234,    -0.00000000056,
      -0.00000000083,    -0.00000000043,
       0.00000000442,     0.00000000650,
       0.00000000134,    -0.00000000408,
       0.00000000651,    -0.00000000443,
      -0.00000000409,    -0.00000000134,
      -0.00000000073,     0.00000000048,
       0.00000000009,     0.00000000219,
       0.00000000089,    -0.00000000059,
       0.00000000224,    -0.00000000019,
      -0.00000000066,    -0.00000000087,
       0.00000003305,     0.00000005193,
       0.00000001164,    -0.00000003055,
       0.00000000474,    -0.00000000381,
      -0.00000000321,    -0.00000000085,
       0.00000000044,     0.00000000104,
      -0.00000000141,    -0.00000001823,
      -0.00000000731,     0.00000000515,
       0.00000000237,     0.00000000094,
      -0.00000001824,     0.00000000126,
       0.00000000518,     0.00000000730,
       0.00000000089,    -0.00000000237,
      -0.00000000006,     0.00000001436,
       0.00000000589,    -0.00000000394,
       0.00000000464,    -0.00000000092,
      -0.00000000162,    -0.00000000177,
       0.00000000340,    -0.00000000754,
      -0.00000000371,    -0.00000000013,
       0.00000000035,     0.00000000085,
      -0.00000000754,    -0.00000000350,
      -0.00000000002,     0.00000000371,
       0.00000000083,    -0.00000000043,
      -0.00000001656,    -0.00000027343,
      -0.00000011042,     0.00000007476,
       0.00000003313,     0.00000001500,
      -0.00000002152,     0.00000000394,
       0.00000000724,     0.00000000827,
       0.00000000090,    -0.00000000297,
      -0.00000002906,     0.00000006433,
       0.00000003153,     0.00000000032,
      -0.00000000330,    -0.00000000705,
      -0.00000000092,     0.00000000109,
       0.00000006417,     0.00000002949,
       0.00000000060,    -0.00000003163,
      -0.00000000707,     0.00000000326,
       0.00000000111,     0.00000000086,
       0.00000000110,     0.00000000035,
       0.00000000040,    -0.00000000107,
       0.00000003506,    -0.00000007164,
      -0.00000003555,    -0.00000000153,
       0.00000000350,     0.00000000787,
      -0.00000001955,    -0.00000000675,
       0.00000000089,     0.00000000942,
       0.00000000205,    -0.00000000123,
      -0.00000002305,     0.00000001830,
       0.00000000984,     0.00000000722,
       0.00000000097,    -0.00000000254,
       0.00000001788,     0.00000002346,
       0.00000000751,    -0.00000000978,
      -0.00000000254,    -0.00000000097,
      -0.00000052109,     0.00000112144,
       0.00000055093,     0.00000001372,
      -0.00000005359,    -0.00000012311,
      -0.00000001534,     0.00000001813,
       0.00000008950,     0.00000003128,
      -0.00000000378,    -0.00000004297,
      -0.00000000924,     0.00000000535,
       0.00000000161,     0.00000000110,
       0.00000020106,    -0.00000015750,
      -0.00000008580,    -0.00000006306,
      -0.00000000788,     0.00000002219,
       0.00000000368,     0.00000000010,
      -0.00000015548,    -0.00000020251,
      -0.00000006425,     0.00000008523,
       0.00000002225,     0.00000000828,
       0.00000000013,    -0.00000000369,
      -0.00000000159,    -0.00000000153,
      -0.00000000039,     0.00000000067,
      -0.00000000163,     0.00000000135,
       0.00000000060,     0.00000000047,
      -0.00000030420,     0.00000023014,
       0.00000012645,     0.00000009651,
       0.00000001218,    -0.00000003259,
       0.00000005499,     0.00000006419,
       0.00000001956,    -0.00000002978,
      -0.00000000772,    -0.00000000219,
       0.00000000008,     0.00000000127,
       0.00000006596,    -0.00000001463,
      -0.00000000860,    -0.00000002510,
      -0.00000000486,     0.00000000224,
       0.00000000039,     0.00000000067,
      -0.00000001130,    -0.00000006632,
      -0.00000002592,     0.00000000716,
       0.00000000201,     0.00000000522,
       0.00000000066,    -0.00000000037,
       0.00000373432,    -0.00000286997,
      -0.00000156456,    -0.00000118219,
      -0.00000015115,     0.00000040513,
       0.00000006668,     0.00000000215,
      -0.00000024250,    -0.00000028200,
      -0.00000008580,     0.00000013125,
       0.00000003384,     0.00000000968,
      -0.00000000024,    -0.00000000555,
      -0.00000058588,     0.00000011601,
       0.00000007152,     0.00000022558,
       0.00000004424,    -0.00000001948,
      -0.00000000330,    -0.00000000567,
       0.00000010392,     0.00000058569,
       0.00000022797,    -0.00000006614,
      -0.00000001843,    -0.00000004543,
      -0.00000000599,     0.00000000320,
       0.00000000041,     0.00000000152,
       0.00000000134,     0.00000000062,
      -0.00000003582,     0.00000002782,
       0.00000001112,     0.00000000471,
       0.00000001897,     0.00000002150,
       0.00000000266,    -0.00000000702,
      -0.00000000137,    -0.00000002163,
      -0.00000000192,     0.00000000252,
      -0.00000002137,     0.00000000141,
       0.00000000243,     0.00000000176,
       0.00000141958,    -0.00000041639,
      -0.00000021614,    -0.00000048526,
      -0.00000008875,     0.00000005037,
       0.00000000728,     0.00000001093,
      -0.00000008566,    -0.00000027921,
      -0.00000009493,     0.00000004320,
       0.00000001010,     0.00000001724,
       0.00000000227,    -0.00000000131,
       0.00000000124,     0.00000010733,
       0.00000001402,    -0.00000001060,
      -0.00000000048,    -0.00000000123,
       0.00000009481,    -0.00000001069,
      -0.00000001283,    -0.00000000741,
       0.00000000047,     0.00000000077,
      -0.00000000034,     0.00000000076,
      -0.00001083467,     0.00000214554,
       0.00000132215,     0.00000418928,
       0.00000082582,    -0.00000035957,
      -0.00000006133,    -0.00000010630,
       0.00000017526,     0.00000096073,
       0.00000037086,    -0.00000011250,
      -0.00000003112,    -0.00000007316,
      -0.00000000946,     0.00000000537,
       0.00000000066,     0.00000000090,
      -0.00000001456,    -0.00000010287,
      -0.00000003159,     0.00000000727,
       0.00000000246,     0.00000000484,
       0.00000000055,    -0.00000000051,
      -0.00000006310,     0.00000004752,
       0.00000001586,     0.00000001079,
      -0.00000000036,    -0.00000000289,
       0.00000000090,    -0.00000000419,
      -0.00000000179,    -0.00000000026,
      -0.00000000237,    -0.00000000286,
      -0.00000000108,     0.00000000132,
       0.00000000059,    -0.00000000049,
       0.00000000047,     0.00000000059,
       0.00000002594,    -0.00000002112,
       0.00000000060,     0.00000000073,
       0.00000000723,    -0.00000000193,
      -0.00000000144,    -0.00000000407,
      -0.00000000074,     0.00000000068,
      -0.00000000066,     0.00000000418,
       0.00000000413,     0.00000000054,
      -0.00000006623,     0.00000026083,
       0.00000007520,    -0.00000000730,
      -0.00000000504,    -0.00000001016,
       0.00000003325,     0.00000000919,
      -0.00000000106,    -0.00000000960,
      -0.00000000102,     0.00000000086,
       0.00000004034,    -0.00000002261,
      -0.00000000538,    -0.00000001261,
      -0.00000000269,     0.00000000111,
      -0.00000001159,     0.00000003191,
       0.00000001512,    -0.00000000234,
      -0.00000000116,    -0.00000000270,
      -0.00000098500,    -0.00000179586,
      -0.00000049497,     0.00000016342,
       0.00000003699,     0.00000006052,
      -0.00000012489,    -0.00000001384,
       0.00000001109,     0.00000003210,
       0.00000000410,    -0.00000000364,
       0.00000024464,     0.00000006129,
       0.00000003506,     0.00000001034,
       0.00000000445,    -0.00000000452,
      -0.00000000052,    -0.00000000072,
       0.00000003528,    -0.00000037919,
      -0.00000004159,    -0.00000001888,
      -0.00000000008,     0.00000000577,
       0.00000000033,     0.00000000087,
      -0.00000000101,     0.00000000018,
       0.00000000018,     0.00000000102,
       0.00000000144,    -0.00000000656,
      -0.00000000320,    -0.00000000030,
       0.00000000207,     0.00000000007,
      -0.00000000004,    -0.00000000211,
      -0.00000001563,     0.00000005076,
       0.00000000825,    -0.00000000149,
       0.00000000820,     0.00000000298,
      -0.00000000069,    -0.00000000097,
      -0.00000000973,    -0.00000000192,
       0.00000000091,     0.00000001215,
       0.00002155283,     0.00000360887,
       0.00000197660,    -0.00000104220,
      -0.00000004358,    -0.00000017118,
      -0.00000016369,     0.00000093451,
       0.00000004323,     0.00000008929,
       0.00000000786,    -0.00000000165,
       0.00000002066,    -0.00000001075,
      -0.00000000162,    -0.00000000017,
       0.00000001131,    -0.00000003162,
      -0.00000000190,    -0.00000000450,
       0.00000000012,    -0.00000000131,
      -0.00000000139,    -0.00000000058,
       0.00000000361,     0.00000000262,
       0.00000000039,    -0.00000000093,
       0.00000003728,    -0.00000000047,
       0.00000000429,     0.00000000036,
       0.00000000054,    -0.00000000420,
       0.00000000022,     0.00000000119,
       0.00000000028,     0.00000000117,
       0.00000021414,     0.00000002302,
      -0.00000000070,     0.00000000628,
      -0.00000003060,    -0.00000000044,
       0.00000000173,    -0.00000000114,
       0.00000000079,    -0.00000000063,
      -0.00000000062,     0.00000000134,
      -0.00000057256,    -0.00000011585,
      -0.00000001453,    -0.00000002625,
      -0.00000000490,     0.00000002635,
      -0.00000000159,     0.00000000083,
       0.00000000490,    -0.00000000091,
       0.00000000147,     0.00000000364,
      -0.00000000145,     0.00000000135,
      -0.00000000137,    -0.00000000148,
       0.00000000106,    -0.00000000011,
      -0.00000000011,    -0.00000000106,
      -0.00000000013,    -0.00000000130,
      -0.00000000414,    -0.00000000091,
      -0.00000000091,     0.00000000414,
      -0.00000019376,    -0.00000004346,
       0.00000000138,    -0.00000000608,
      -0.00000000061,    -0.00000000047,
      -0.00000000108,     0.00000000053,
       0.00000000898,     0.00000004024,
       0.00000000022,    -0.00000000135,
       0.00000000040,     0.00000000294,
       0.00000009576,     0.00000007261,
       0.00000000194,    -0.00000000212,
      -0.00000000726,    -0.00000000126,
       0.00000000214,    -0.00000000699,
      -0.00000000103,    -0.00000000022,
       0.00000003775,    -0.00000007336,
       0.00000000476,     0.00000000943,
       0.00000000932,     0.00000000383,
      -0.00000000156,     0.00000000134,
      -0.00000000329,     0.00000000094,
      -0.00000001361,    -0.00000002624,
       0.00000000137,     0.00000000058,
      -0.00000000198,     0.00000000098,
       0.00000000582,    -0.00000002240,
      -0.00000000314,    -0.00000000388,
      -0.00000000194,     0.00000000202,
       0.00000000063,     0.00000000047,
      -0.00000000389,     0.00000000314,
       0.00000000202,     0.00000000194,
       0.00000000047,    -0.00000000063,
       0.00000000825,     0.00000000994,
       0.00000000496,    -0.00000000527,
       0.00000000221,    -0.00000000157,
      -0.00000000103,    -0.00000000110,
       0.00000000526,    -0.00000000082,
      -0.00000000060,    -0.00000000206,
      -0.00000000088,    -0.00000000532,
      -0.00000000207,     0.00000000060,
      -0.00000010807,    -0.00000013009,
      -0.00000006507,     0.00000006899,
       0.00000002137,     0.00000001584,
      -0.00000000657,     0.00000000462,
       0.00000000306,     0.00000000324,
       0.00000000078,    -0.00000000097,
      -0.00000004954,     0.00000000784,
       0.00000000556,     0.00000001919,
       0.00000000379,    -0.00000000160,
       0.00000000796,     0.00000004963,
       0.00000001923,    -0.00000000565,
      -0.00000000163,    -0.00000000380,
       0.00000016996,    -0.00000002666,
      -0.00000001897,    -0.00000006531,
      -0.00000001287,     0.00000000544,
      -0.00000000466,    -0.00000002897,
      -0.00000001120,     0.00000000329,
       0.00000000097,     0.00000000224,
      -0.00000000003,     0.00000000263,
       0.00000000093,    -0.00000000028,
       0.00000000109,     0.00000000086,
      -0.00000165279,     0.00000025896,
       0.00000018683,     0.00000064052,
       0.00000012626,    -0.00000005346,
       0.00000001297,     0.00000008051,
       0.00000003118,    -0.00000000913,
      -0.00000000261,    -0.00000000614,
      -0.00000000079,     0.00000000046,
       0.00000000729,    -0.00000001883,
      -0.00000000434,     0.00000000043,
      -0.00000001708,    -0.00000000828,
      -0.00000000027,     0.00000000348,
       0.00000000374,     0.00000000752,
       0.00000000127,    -0.00000000072,
      -0.00000000212,    -0.00000000013,
      -0.00000000009,     0.00000000215,
       0.00004001285,     0.00000673152,
       0.00000371719,    -0.00000190089,
      -0.00000007708,    -0.00000032432,
      -0.00000002425,     0.00000000769,
      -0.00000020593,     0.00000121339,
       0.00000005780,     0.00000011280,
       0.00000000984,    -0.00000000236,
       0.00000000143,    -0.00000000763,
      -0.00000000359,     0.00000000155,
       0.00000000279,     0.00000000082,
       0.00000000083,    -0.00000000279,
      -0.00000070148,    -0.00000012442,
      -0.00000002458,    -0.00000003736,
      -0.00000000427,     0.00000002491,
      -0.00000000153,     0.00000000093,
       0.00000000060,    -0.00000000306,
       0.00000000301,     0.00000000063,
       0.00000000247,     0.00000000073,
      -0.00000000073,     0.00000000247,
       0.00000000016,     0.00000000052,
      -0.00000000006,     0.00000000002,
       0.00000000063,    -0.00000000076,
      -0.00000000011,     0.00000000005,
       0.00000000085,     0.00000000064,
      -0.00000000005,    -0.00000000011,
      -0.00000000100,     0.00000000252,
       0.00000000181,     0.00000000004,
      -0.00000000005,     0.00000000182,
      -0.00000000141,     0.00000000401,
       0.00000000430,     0.00000000013,
      -0.00000000011,     0.00000000435,
      -0.00000000497,     0.00000000218,
       0.00000000774,    -0.00000000052,
       0.00000000028,     0.00000000021,
       0.00000000073,     0.00000000724,
      -0.00000812575,    -0.00000865136,
       0.00000053042,    -0.00000003919,
       0.00000000204,     0.00000002451,
      -0.00000241309,     0.00000423720,
      -0.00000013799,    -0.00000006346,
      -0.00000000215,    -0.00000000901,
      -0.00000580494,    -0.00000075709,
       0.00000021837,    -0.00000003721,
       0.00000000113,     0.00000002812,
      -0.00000000285,    -0.00000000028,
      -0.00000093732,    -0.00000514939,
      -0.00000003704,    -0.00000021112,
       0.00000003683,    -0.00000000104,
      -0.00000000043,     0.00000000277,
      -0.00000006946,     0.00000007202,
       0.00000000604,     0.00000002597,
       0.00000000234,    -0.00000000051,
      -0.00000003509,    -0.00000005754,
      -0.00000002533,     0.00000000239,
       0.00000000132,     0.00000000198,
       0.00142227275,     0.00036422562,
       0.00003354534,     0.00006641215,
      -0.00000363329,     0.00000384708,
      -0.00000016323,    -0.00000013831,
      -0.00002264450,     0.00009887850,
      -0.00000467315,     0.00000272200,
      -0.00000029259,    -0.00000024764,
       0.00000000934,    -0.00000001432,
      -0.00001683167,    -0.00000065422,
      -0.00000060112,     0.00000040004,
      -0.00000001944,    -0.00000009985,
       0.00000000849,     0.00000000096,
       0.00000135152,    -0.00002149782,
       0.00000037558,     0.00000046842,
      -0.00000008767,     0.00000002151,
       0.00000000105,    -0.00000000816,
       0.00000050161,    -0.00000041898,
      -0.00000002945,    -0.00000010246,
      -0.00000000761,     0.00000000213,
      -0.00000017982,    -0.00000019899,
       0.00000002482,     0.00000000519,
      -0.00000000213,    -0.00000000252,
       0.00000515745,     0.00000197603,
       0.00000005965,     0.00000018486,
       0.00000003753,    -0.00000003880,
      -0.00000000272,     0.00000000430,
      -0.00000057546,    -0.00000007211,
       0.00000000256,     0.00000000623,
      -0.00000001439,    -0.00000058804,
       0.00000000124,    -0.00000000646,
       0.00001130184,     0.00000288617,
       0.00000027291,     0.00000052316,
      -0.00000002825,     0.00000003129,
      -0.00000000925,     0.00000001235,
      -0.00000073138,    -0.00000003500,
      -0.00000000260,     0.00000000304,
       0.00000003498,    -0.00000073190,
      -0.00000000266,    -0.00000000204,
      -0.00000000039,     0.00000000027,
       0.00000000015,     0.00000000039,
       0.00000000990,     0.00000000350,
      -0.00000002395,    -0.00000000871,
       0.00000000334,     0.00000000295,
       0.00000000060,    -0.00000000085,
       0.00000000115,     0.00000000187,
       0.00000000042,    -0.00000000108,
       0.00000000188,    -0.00000000115,
      -0.00000000110,    -0.00000000040,
       0.00000000000,     0.00000000118,
       0.00000000119,    -0.00000000005,
       0.00000001932,     0.00000003289,
       0.00000000808,    -0.00000001842,
       0.00000000209,    -0.00000000165,
      -0.00000000139,    -0.00000000039,
      -0.00000000050,    -0.00000001001,
      -0.00000000407,     0.00000000268,
       0.00000000122,     0.00000000054,
      -0.00000001001,     0.00000000041,
       0.00000000266,     0.00000000407,
       0.00000000055,    -0.00000000121,
      -0.00000000124,     0.00000001302,
       0.00000000275,    -0.00000000046,
      -0.00000000092,    -0.00000000106,
       0.00000000209,    -0.00000000455,
      -0.00000000222,    -0.00000000007,
      -0.00000000453,    -0.00000000218,
      -0.00000000006,     0.00000000222,
      -0.00000000695,    -0.00000019442,
      -0.00000007910,     0.00000004982,
       0.00000002246,     0.00000001114,
      -0.00000001294,     0.00000000206,
       0.00000000411,     0.00000000495,
       0.00000000057,    -0.00000000168,
      -0.00000001826,     0.00000003944,
       0.00000001926,     0.00000000050,
      -0.00000000186,    -0.00000000428,
      -0.00000000053,     0.00000000063,
       0.00000003924,     0.00000001862,
       0.00000000076,    -0.00000001928,
      -0.00000000431,     0.00000000183,
       0.00000000064,     0.00000000052,
       0.00000002837,    -0.00000005732,
      -0.00000002745,    -0.00000000132,
      -0.00000001221,    -0.00000000447,
       0.00000000040,     0.00000000588,
       0.00000000128,    -0.00000000071,
      -0.00000001336,     0.00000001090,
       0.00000000575,     0.00000000415,
       0.00000000056,    -0.00000000146,
       0.00000001039,     0.00000001366,
       0.00000000442,    -0.00000000561,
      -0.00000000146,    -0.00000000059,
      -0.00000037574,     0.00000078898,
       0.00000038637,     0.00000001596,
      -0.00000003515,    -0.00000008630,
      -0.00000001094,     0.00000001216,
       0.00000005606,     0.00000002066,
      -0.00000000172,    -0.00000002685,
      -0.00000000581,     0.00000000314,
       0.00000000096,     0.00000000070,
       0.00000011861,    -0.00000009416,
      -0.00000005051,    -0.00000003720,
      -0.00000000473,     0.00000001292,
       0.00000000213,     0.00000000011,
      -0.00000009221,    -0.00000011960,
      -0.00000003817,     0.00000004988,
       0.00000001292,     0.00000000509,
       0.00000000014,    -0.00000000214,
      -0.00000000081,    -0.00000000075,
      -0.00000000081,     0.00000000065,
      -0.00000021139,     0.00000016389,
       0.00000008819,     0.00000006645,
       0.00000000864,    -0.00000002236,
       0.00000003238,     0.00000003797,
       0.00000001166,    -0.00000001734,
      -0.00000000448,    -0.00000000135,
       0.00000003391,    -0.00000000938,
      -0.00000000497,    -0.00000001268,
      -0.00000000243,     0.00000000123,
      -0.00000000615,    -0.00000003364,
      -0.00000001321,     0.00000000348,
       0.00000000095,     0.00000000271,
       0.00000242288,    -0.00000188652,
      -0.00000101385,    -0.00000076867,
      -0.00000010049,     0.00000026017,
       0.00000004271,     0.00000000231,
      -0.00000014401,    -0.00000016781,
      -0.00000005144,     0.00000007691,
       0.00000001965,     0.00000000611,
      -0.00000000003,    -0.00000000321,
      -0.00000030439,     0.00000007130,
       0.00000003981,     0.00000011588,
       0.00000002263,    -0.00000001018,
      -0.00000000165,    -0.00000000291,
       0.00000006080,     0.00000030298,
       0.00000011732,    -0.00000003499,
      -0.00000000921,    -0.00000002351,
      -0.00000000314,     0.00000000154,
       0.00000000033,     0.00000000095,
       0.00000000082,     0.00000000031,
       0.00000066798,    -0.00000015234,
      -0.00000008529,    -0.00000025413,
      -0.00000004991,     0.00000002174,
       0.00000000354,     0.00000000648,
      -0.00000002386,    -0.00000011465,
      -0.00000004406,     0.00000001396,
       0.00000000368,     0.00000000870,
       0.00000000115,    -0.00000000061,
      -0.00000000453,     0.00000001441,
       0.00000000496,     0.00000000023,
      -0.00000000026,    -0.00000000099,
       0.00000000289,    -0.00000000587,
      -0.00000000259,     0.00000000102,
      -0.00000614378,     0.00000141829,
       0.00000078719,     0.00000235127,
       0.00000046350,    -0.00000020203,
      -0.00000003311,    -0.00000006022,
       0.00000010302,     0.00000048569,
       0.00000018660,    -0.00000005954,
      -0.00000001555,    -0.00000003687,
      -0.00000000482,     0.00000000258,
       0.00000000175,    -0.00000007776,
      -0.00000002322,     0.00000000564,
       0.00000000208,     0.00000000379,
      -0.00000004779,     0.00000002637,
       0.00000001354,     0.00000000751,
      -0.00000000017,    -0.00000000276,
       0.00000000043,    -0.00000000233,
      -0.00000000101,    -0.00000000012,
      -0.00000000133,    -0.00000000155,
      -0.00000000060,     0.00000000074,
      -0.00000000951,    -0.00000009458,
      -0.00000000806,     0.00000001142,
      -0.00000004503,     0.00000000471,
       0.00000000545,     0.00000000374,
      -0.00000000181,     0.00000000062,
      -0.00000000004,     0.00000000268,
       0.00000003330,     0.00000060932,
       0.00000008531,    -0.00000006920,
      -0.00000001089,    -0.00000000631,
       0.00000012403,    -0.00000001350,
      -0.00000001499,    -0.00000001430,
      -0.00000000076,     0.00000000079,
       0.00000003155,    -0.00000001166,
      -0.00000000260,    -0.00000000884,
      -0.00000000178,     0.00000000073,
      -0.00000000096,     0.00000001679,
       0.00000000917,    -0.00000000306,
      -0.00000000079,    -0.00000000180,
      -0.00000012918,    -0.00000134323,
      -0.00000034223,     0.00000014294,
       0.00000003831,     0.00000004334,
      -0.00000010160,     0.00000000991,
       0.00000001275,     0.00000002623,
       0.00000000323,    -0.00000000339,
      -0.00000002150,     0.00000001491,
       0.00000000421,     0.00000001671,
       0.00000000352,    -0.00000000129,
      -0.00000000565,    -0.00000007315,
      -0.00000001955,     0.00000000765,
       0.00000000180,     0.00000000366,
       0.00000000038,     0.00000000069,
      -0.00000000373,     0.00000001589,
       0.00000000723,     0.00000000180,
       0.00000004325,    -0.00000006385,
      -0.00000001421,     0.00000000296,
       0.00000000163,     0.00000000104,
      -0.00000000282,     0.00000000873,
       0.00000000185,    -0.00000000055,
      -0.00000000581,     0.00000000160,
       0.00000000058,     0.00000000184,
       0.00000098663,     0.00000019680,
      -0.00000003015,    -0.00000004974,
       0.00000000110,    -0.00000000030,
      -0.00000000227,     0.00000000236,
      -0.00000023179,    -0.00000003494,
      -0.00000000260,     0.00000000268,
      -0.00000001464,     0.00000023182,
       0.00000000097,    -0.00000000278,
       0.00000000073,    -0.00000000146,
       0.00000000149,     0.00000000083,
       0.00000000786,     0.00000000361,
       0.00000000149,    -0.00000000301,
      -0.00000000015,    -0.00000000088,
      -0.00000000086,     0.00000000016,
      -0.00000009666,    -0.00000000650,
       0.00000000011,     0.00000000237,
       0.00000000119,     0.00000000081,
       0.00000000040,    -0.00000000116,
      -0.00001640472,    -0.00000112062,
      -0.00000004189,     0.00000007942,
       0.00000001939,    -0.00000075110,
      -0.00000000219,    -0.00000000360,
       0.00000002023,     0.00000004614,
      -0.00000000275,    -0.00000000171,
       0.00000004439,    -0.00000001013,
      -0.00000000157,     0.00000000266,
       0.00000000107,    -0.00000000118,
      -0.00000000118,    -0.00000000107,
      -0.00000000257,     0.00000000008,
       0.00000000007,     0.00000000257,
      -0.00000015030,     0.00000000271,
      -0.00000000011,    -0.00000000472,
       0.00000000143,     0.00000000027,
      -0.00000013570,    -0.00000136822,
       0.00000014598,     0.00000007400,
      -0.00000000679,     0.00000001326,
       0.00000001658,    -0.00000000327,
      -0.00000000010,     0.00000000260,
       0.00000001255,     0.00000003697,
      -0.00000000212,    -0.00000000265,
       0.00000001624,     0.00000003775,
      -0.00000000506,     0.00000000316,
       0.00000000094,    -0.00000000094,
       0.00000000106,     0.00000000084,
       0.00000000256,     0.00000000022,
       0.00000000022,    -0.00000000256,
       0.00000014520,     0.00000001557,
      -0.00000000039,     0.00000000458,
      -0.00000001632,     0.00000000067,
       0.00000000012,     0.00000000231,
      -0.00000049686,     0.00000115806,
      -0.00000011604,    -0.00000016041,
       0.00000001770,    -0.00000000657,
      -0.00000004985,    -0.00000002174,
       0.00000000670,    -0.00000000487,
      -0.00000000128,    -0.00000000352,
       0.00000000541,    -0.00000000063,
       0.00000000080,    -0.00000000046,
      -0.00000000723,    -0.00000001216,
      -0.00000000187,     0.00000000111,
       0.00000000099,     0.00000000063,
      -0.00000000083,     0.00000000140,
      -0.00000000140,    -0.00000000084,
      -0.00000000131,    -0.00000000159,
      -0.00000000080,     0.00000000084,
      -0.00000000159,     0.00000000131,
       0.00000000084,     0.00000000080,
       0.00000000087,    -0.00000000064,
       0.00000000180,    -0.00000000027,
      -0.00000000029,    -0.00000000183,
      -0.00000004683,    -0.00000005569,
      -0.00000002792,     0.00000002976,
      -0.00000000260,     0.00000000189,
       0.00000000124,     0.00000000129,
      -0.00000001716,     0.00000000264,
       0.00000000187,     0.00000000664,
       0.00000000131,    -0.00000000054,
       0.00000000270,     0.00000001719,
       0.00000000666,    -0.00000000192,
      -0.00000000055,    -0.00000000132,
       0.00000006100,    -0.00000000894,
      -0.00000000671,    -0.00000002368,
      -0.00000000150,    -0.00000000947,
      -0.00000000366,     0.00000000106,
       0.00000000031,     0.00000000073,
       0.00000000005,     0.00000000121,
      -0.00000060125,     0.00000009274,
       0.00000006604,     0.00000023290,
       0.00000004603,    -0.00000001870,
       0.00000000424,     0.00000002683,
       0.00000001039,    -0.00000000298,
      -0.00000000084,    -0.00000000206,
       0.00000000122,    -0.00000000757,
      -0.00000000222,     0.00000000029,
      -0.00000000678,    -0.00000000171,
      -0.00000000005,     0.00000000183,
      -0.00000000830,     0.00000003580,
       0.00000001022,    -0.00000000036,
       0.00000000396,     0.00000000077,
      -0.00000000008,    -0.00000000112,
       0.00000000354,    -0.00000000137,
      -0.00000000044,    -0.00000000150,
      -0.00000000008,     0.00000000426,
       0.00000000154,    -0.00000000046,
       0.00000004497,    -0.00000024545,
      -0.00000007037,     0.00000000543,
      -0.00000001161,    -0.00000000228,
       0.00000000022,     0.00000000327,
      -0.00000000507,     0.00000000369,
       0.00000000031,     0.00000000116,
       0.00000000260,    -0.00000000172,
      -0.00000000147,     0.00000000044,
      -0.00000002572,    -0.00000000114,
      -0.00002979066,    -0.00000152164,
      -0.00000010365,     0.00000012694,
       0.00000004559,    -0.00000095564,
      -0.00000000391,    -0.00000000337,
       0.00000003454,     0.00000001196,
       0.00000000097,     0.00000000204,
       0.00000001076,    -0.00000003813,
       0.00000000191,    -0.00000000107,
      -0.00000000167,    -0.00000000036,
      -0.00000000036,     0.00000000166,
       0.00000005083,     0.00000011864,
       0.00000000097,    -0.00000000035,
       0.00000003849,     0.00000001046,
       0.00000000107,     0.00000000205,
      -0.00000001048,     0.00000003851,
      -0.00000000205,     0.00000000107,
      -0.00000000152,    -0.00000000033,
       0.00000000033,    -0.00000000152,
       0.00000000015,     0.00000000011,
       0.00000000045,     0.00000000052,
      -0.00000000052,     0.00000000045,
       0.00000000117,    -0.00000000146,
       0.00000000146,     0.00000000118,
       0.00000549260,     0.00000066928,
      -0.00000056629,     0.00000212333,
      -0.00000041973,    -0.00000016550,
       0.00000002868,    -0.00000005428,
       0.00000000516,     0.00000000349,
       0.00000025890,    -0.00000182857,
       0.00000070324,     0.00000019931,
      -0.00000005695,     0.00000013896,
      -0.00000001792,    -0.00000000977,
       0.00000000118,    -0.00000000171,
      -0.00000001860,    -0.00000017721,
       0.00000005633,     0.00000001460,
      -0.00000000647,     0.00000000858,
      -0.00000000093,    -0.00000000126,
       0.00000013812,    -0.00000012760,
       0.00000004197,     0.00000003699,
      -0.00000000392,     0.00000000864,
       0.00000001946,     0.00000002943,
       0.00000000170,    -0.00000000175,
      -0.00000003176,     0.00000001660,
       0.00000000258,     0.00000000125,
      -0.00002492775,    -0.00000294363,
       0.00000253976,    -0.00000972347,
       0.00000191949,     0.00000074243,
      -0.00000013019,     0.00000024820,
      -0.00000002369,    -0.00000001576,
      -0.00000058568,     0.00000414364,
      -0.00000159463,    -0.00000045005,
       0.00000012814,    -0.00000031570,
       0.00000004092,     0.00000002214,
      -0.00000000271,     0.00000000394,
       0.00000010907,    -0.00000048273,
      -0.00000001139,    -0.00000003636,
       0.00000002361,    -0.00000000689,
       0.00000000153,     0.00000000492,
       0.00000064402,     0.00000052054,
      -0.00000017854,     0.00000006812,
      -0.00000001260,    -0.00000003412,
       0.00000000555,    -0.00000000159,
       0.00000015803,    -0.00000024367,
      -0.00000001006,    -0.00000002332,
       0.00000025961,     0.00000017534,
       0.00000001794,    -0.00000000683,
      -0.00000000009,    -0.00000000198,
       0.00000000955,    -0.00000001047,
       0.00000000306,     0.00000000108,
       0.00000001048,     0.00000000959,
      -0.00000000110,     0.00000000306,
       0.00000000791,    -0.00000001422,
       0.00000000379,     0.00000000117,
       0.00000001422,     0.00000000797,
      -0.00000000120,     0.00000000379,
       0.00000000050,     0.00000000090,
       0.00000000091,    -0.00000000049,
       0.00000001048,     0.00000001973,
       0.00000000105,    -0.00000000075,
      -0.00000000014,    -0.00000000472,
      -0.00000000192,     0.00000000117,
      -0.00000000470,     0.00000000009,
       0.00000000115,     0.00000000192,
       0.00000000131,    -0.00000000016,
       0.00000000083,    -0.00000000189,
      -0.00000000185,    -0.00000000090,
      -0.00000000179,    -0.00000010622,
      -0.00000004259,     0.00000002505,
      -0.00000000616,     0.00000000076,
       0.00000000179,     0.00000000239,
       0.00000000030,    -0.00000000074,
      -0.00000000760,     0.00000001663,
       0.00000000800,     0.00000000027,
      -0.00000000072,    -0.00000000177,
       0.00000001643,     0.00000000783,
       0.00000000045,    -0.00000000797,
      -0.00000000179,     0.00000000068,
       0.00000001410,    -0.00000002950,
      -0.00000001421,    -0.00000000079,
      -0.00000000508,    -0.00000000197,
       0.00000000009,     0.00000000242,
      -0.00000000442,     0.00000000398,
       0.00000000201,     0.00000000134,
       0.00000000353,     0.00000000457,
       0.00000000152,    -0.00000000185,
      -0.00000017330,     0.00000037612,
       0.00000017873,     0.00000000827,
      -0.00000001488,    -0.00000003940,
       0.00000002337,     0.00000000901,
      -0.00000000036,    -0.00000001110,
      -0.00000000241,     0.00000000116,
       0.00000004014,    -0.00000003625,
      -0.00000001751,    -0.00000001250,
      -0.00000000164,     0.00000000435,
      -0.00000003481,    -0.00000004058,
      -0.00000001307,     0.00000001701,
       0.00000000431,     0.00000000186,
      -0.00000008304,     0.00000006860,
       0.00000003533,     0.00000002617,
       0.00000000353,    -0.00000000881,
       0.00000001118,     0.00000001275,
       0.00000000394,    -0.00000000580,
      -0.00000000146,    -0.00000000049,
       0.00000000888,    -0.00000000382,
      -0.00000000170,    -0.00000000316,
      -0.00000000174,    -0.00000000839,
      -0.00000000335,     0.00000000074,
       0.00000089593,    -0.00000077996,
      -0.00000038834,    -0.00000027945,
      -0.00000003769,     0.00000009581,
      -0.00000005234,    -0.00000005657,
      -0.00000001752,     0.00000002585,
       0.00000000646,     0.00000000223,
       0.00000000005,    -0.00000000105,
      -0.00000007909,     0.00000003548,
       0.00000001328,     0.00000002854,
       0.00000000574,    -0.00000000275,
      -0.00000000040,    -0.00000000075,
       0.00000002938,     0.00000007755,
       0.00000002903,    -0.00000001049,
      -0.00000000216,    -0.00000000612,
      -0.00000000085,     0.00000000032,
       0.00000018836,    -0.00000006052,
      -0.00000002723,    -0.00000007053,
      -0.00000001393,     0.00000000608,
      -0.00000000810,    -0.00000002828,
      -0.00000001071,     0.00000000384,
       0.00000000088,     0.00000000213,
      -0.00000000186,     0.00000000593,
       0.00000000220,     0.00000000019,
       0.00000000044,    -0.00000000357,
      -0.00000000139,     0.00000000061,
      -0.00000172908,     0.00000068178,
       0.00000026716,     0.00000064071,
       0.00000012787,    -0.00000005696,
      -0.00000000828,    -0.00000001719,
       0.00000004453,     0.00000012091,
       0.00000004503,    -0.00000001765,
      -0.00000000377,    -0.00000000916,
      -0.00000000122,     0.00000000056,
      -0.00000000157,    -0.00000005727,
      -0.00000001226,     0.00000000560,
       0.00000000097,     0.00000000167,
      -0.00000004325,     0.00000001539,
       0.00000000964,     0.00000000521,
      -0.00000000007,    -0.00000000143,
       0.00000000067,     0.00000000628,
       0.00000000233,    -0.00000000024,
       0.00000001236,     0.00000008055,
       0.00000002020,    -0.00000001199,
       0.00000000956,    -0.00000000208,
      -0.00000000160,    -0.00000000241,
       0.00000000723,    -0.00000000200,
      -0.00000000092,    -0.00000000283,
       0.00000000254,     0.00000000824,
       0.00000000300,    -0.00000000120,
      -0.00000014406,    -0.00000088872,
      -0.00000018164,     0.00000012112,
       0.00000002533,     0.00000002029,
      -0.00000006320,     0.00000001202,
       0.00000000957,     0.00000001235,
       0.00000000130,    -0.00000000171,
      -0.00000001761,     0.00000004058,
       0.00000000396,     0.00000000290,
       0.00000000122,    -0.00000000034,
       0.00000003001,    -0.00000001495,
      -0.00000000933,     0.00000000068,
       0.00000000070,     0.00000000123,
       0.00000000655,    -0.00000000645,
      -0.00000000102,    -0.00000000095,
       0.00000000037,     0.00000000167,
       0.00000000169,    -0.00000000049,
      -0.00000000639,     0.00000000344,
       0.00000000060,     0.00000000168,
       0.00000000017,     0.00000000330,
       0.00000000088,    -0.00000000042,
      -0.00000000338,     0.00000000060,
       0.00000000046,     0.00000000088,
       0.00000014855,     0.00000098601,
       0.00000005024,    -0.00000009066,
      -0.00000004903,    -0.00000001047,
       0.00000000245,     0.00000000490,
       0.00000001267,    -0.00000002879,
      -0.00000000321,     0.00000000124,
      -0.00000001597,    -0.00000001406,
       0.00000000748,     0.00000002542,
      -0.00000000351,    -0.00000000353,
       0.00000000926,    -0.00000000359,
      -0.00000000115,     0.00000000130,
      -0.00000000753,    -0.00000000139,
       0.00000000123,    -0.00000000142,
      -0.00000000134,     0.00000000699,
      -0.00000000148,    -0.00000000121,
      -0.00000007691,    -0.00000014177,
       0.00000002010,     0.00000001996,
      -0.00000002430,     0.00000000104,
       0.00000000318,    -0.00000000298,
       0.00000003216,     0.00000000655,
      -0.00000000541,     0.00000000623,
      -0.00000000060,    -0.00000000106,
       0.00000000565,    -0.00000003152,
       0.00000000644,     0.00000000551,
      -0.00000000103,     0.00000000057,
      -0.00000041750,    -0.00000052296,
      -0.00000003332,    -0.00000000554,
       0.00000005937,     0.00000002772,
      -0.00000000175,    -0.00000000408,
      -0.00000000092,    -0.00000000121,
       0.00000000219,     0.00000000232,
      -0.00000002305,     0.00000001672,
      -0.00000000052,    -0.00000000072,
       0.00000003501,     0.00000000713,
       0.00000000073,    -0.00000000476,
       0.00000362030,     0.00000095019,
       0.00000008578,     0.00000016637,
       0.00000000922,    -0.00000003290,
       0.00000000105,    -0.00000000058,
       0.00000000342,    -0.00000001086,
       0.00000000066,     0.00000000051,
       0.00000000288,    -0.00000000118,
       0.00000000020,    -0.00000000159,
       0.00000019087,    -0.00000028149,
       0.00000007837,     0.00000002787,
       0.00000001454,     0.00000000513,
      -0.00000000107,     0.00000000383,
       0.00000004303,     0.00000001108,
       0.00000000097,     0.00000000171,
      -0.00000001163,     0.00000004276,
      -0.00000000167,     0.00000000081,
       0.00000008492,     0.00000001583,
      -0.00000000194,     0.00000000027,
       0.00000000029,     0.00000000195,
      -0.00000000015,    -0.00000000100,
      -0.00000007152,     0.00000001009,
       0.00000000724,     0.00000002770,
       0.00000000043,     0.00000000291,
       0.00000000113,    -0.00000000030,
       0.00000000015,    -0.00000000132,
      -0.00000000119,    -0.00000000024,
       0.00000000676,    -0.00000004627,
      -0.00000000191,    -0.00000000029,
       0.00000002788,    -0.00000001522,
      -0.00000007808,    -0.00000002443,
       0.00000061402,    -0.00000014053,
      -0.00000000047,    -0.00000000141,
       0.00000037096,    -0.00000054965,
       0.00000015135,     0.00000005428,
       0.00000001659,     0.00000001216,
      -0.00000000165,     0.00000000494,
       0.00000006059,     0.00000001526,
       0.00000000125,     0.00000000236,
      -0.00000001523,     0.00000006073,
      -0.00000000239,     0.00000000126,
      -0.00000000016,    -0.00000000015,
      -0.00000000041,    -0.00000000050,
       0.00000000004,    -0.00000000004,
       0.00000000050,    -0.00000000041,
       0.00000000004,     0.00000000004,
      -0.00000000004,    -0.00000000002,
       0.00000000002,    -0.00000000005,
      -0.00000000141,     0.00000000060,
      -0.00000000061,    -0.00000000140,
       0.00000049127,    -0.00000003148,
       0.00000001679,     0.00000023701,
      -0.00000005979,     0.00000000434,
      -0.00000000072,    -0.00000001027,
      -0.00000001813,    -0.00000011281,
       0.00000005619,    -0.00000000956,
       0.00000000237,     0.00000001451,
      -0.00000000256,     0.00000000042,
       0.00000025323,     0.00000002275,
      -0.00000002054,     0.00000009954,
      -0.00000001988,    -0.00000000578,
       0.00000000091,    -0.00000000264,
      -0.00000004362,     0.00000025411,
      -0.00000009842,    -0.00000003430,
       0.00000001038,    -0.00000001899,
       0.00000000232,     0.00000000189,
      -0.00000000647,     0.00000000008,
       0.00000000038,    -0.00000000087,
       0.00000000002,    -0.00000000642,
       0.00000000085,     0.00000000040,
      -0.00000154663,     0.00000002448,
      -0.00000001722,    -0.00000073305,
       0.00000018233,    -0.00000000723,
       0.00000000178,     0.00000003104,
       0.00000004108,     0.00000026579,
      -0.00000013269,     0.00000002191,
      -0.00000000543,    -0.00000003421,
       0.00000000599,    -0.00000000084,
      -0.00000026581,     0.00000001607,
      -0.00000000249,    -0.00000011190,
       0.00000002326,    -0.00000000153,
       0.00000000070,     0.00000000341,
       0.00000007013,    -0.00000027400,
       0.00000010955,     0.00000005527,
      -0.00000001798,     0.00000002008,
      -0.00000000219,    -0.00000000366,
       0.00000001855,     0.00000000454,
      -0.00000000109,     0.00000000177,
      -0.00000000390,     0.00000001887,
      -0.00000000187,    -0.00000000079,
      -0.00000000261,     0.00000000317,
      -0.00000000217,     0.00000000002,
      -0.00000000317,    -0.00000000261,
      -0.00000000003,    -0.00000000216,
      -0.00000000321,     0.00000000384,
      -0.00000000269,     0.00000000004,
      -0.00000000384,    -0.00000000321,
      -0.00000000005,    -0.00000000269,
       0.00000000775,     0.00000001496,
       0.00000000077,    -0.00000000053,
      -0.00000000010,    -0.00000000324,
      -0.00000000130,     0.00000000079,
      -0.00000000323,     0.00000000006,
       0.00000000078,     0.00000000131,
       0.00000000050,    -0.00000000121,
      -0.00000000117,    -0.00000000056,
      -0.00000000125,    -0.00000007380,
      -0.00000002995,     0.00000001736,
      -0.00000000422,     0.00000000048,
       0.00000000119,     0.00000000164,
      -0.00000000471,     0.00000001061,
       0.00000000503,     0.00000000015,
      -0.00000000043,    -0.00000000111,
       0.00000001042,     0.00000000489,
       0.00000000029,    -0.00000000500,
      -0.00000000112,     0.00000000040,
       0.00000000931,    -0.00000001998,
      -0.00000000951,    -0.00000000051,
      -0.00000000321,    -0.00000000124,
       0.00000000004,     0.00000000153,
      -0.00000000247,     0.00000000237,
       0.00000000116,     0.00000000073,
       0.00000000202,     0.00000000257,
       0.00000000086,    -0.00000000104,
      -0.00000011063,     0.00000024115,
       0.00000011446,     0.00000000573,
      -0.00000000935,    -0.00000002528,
       0.00000001476,     0.00000000568,
      -0.00000000017,    -0.00000000695,
      -0.00000000151,     0.00000000071,
       0.00000002269,    -0.00000002020,
      -0.00000001009,    -0.00000000700,
      -0.00000000092,     0.00000000246,
      -0.00000001911,    -0.00000002298,
      -0.00000000740,     0.00000000969,
       0.00000000241,     0.00000000108,
      -0.00000004897,     0.00000004228,
       0.00000002120,     0.00000001537,
       0.00000000647,     0.00000000716,
       0.00000000221,    -0.00000000328,
      -0.00000000082,    -0.00000000028,
       0.00000000448,    -0.00000000238,
      -0.00000000100,    -0.00000000154,
      -0.00000000090,    -0.00000000407,
      -0.00000000165,     0.00000000031,
       0.00000052001,    -0.00000045153,
      -0.00000022524,    -0.00000016312,
      -0.00000002242,     0.00000005525,
      -0.00000002910,    -0.00000003191,
      -0.00000000987,     0.00000001464,
       0.00000000360,     0.00000000130,
      -0.00000003986,     0.00000001638,
       0.00000000682,     0.00000001454,
       0.00000000282,    -0.00000000141,
       0.00000001219,     0.00000003869,
       0.00000001482,    -0.00000000489,
      -0.00000000100,    -0.00000000307,
       0.00000009633,    -0.00000003680,
      -0.00000001500,    -0.00000003567,
      -0.00000000707,     0.00000000309,
      -0.00000000470,    -0.00000001385,
      -0.00000000520,     0.00000000201,
       0.00000000042,     0.00000000104,
      -0.00000000112,     0.00000000364,
       0.00000000135,     0.00000000013,
       0.00000000014,    -0.00000000234,
      -0.00000000089,     0.00000000041,
      -0.00000089542,     0.00000034550,
       0.00000013959,     0.00000033041,
       0.00000006574,    -0.00000002867,
       0.00000002100,     0.00000005995,
       0.00000002246,    -0.00000000878,
      -0.00000000182,    -0.00000000449,
      -0.00000000027,    -0.00000002116,
      -0.00000000613,     0.00000000198,
       0.00000000060,     0.00000000101,
      -0.00000001247,     0.00000000893,
       0.00000000454,     0.00000000180,
       0.00000000993,     0.00000004741,
       0.00000001187,    -0.00000000802,
       0.00000000564,    -0.00000000155,
      -0.00000000104,    -0.00000000140,
       0.00000000378,    -0.00000000125,
      -0.00000000051,    -0.00000000148,
       0.00000000164,     0.00000000448,
       0.00000000160,    -0.00000000071,
      -0.00000007530,    -0.00000039481,
      -0.00000009548,     0.00000006645,
       0.00000001482,     0.00000001176,
      -0.00000002361,     0.00000000685,
       0.00000000462,     0.00000000623,
       0.00000000076,    -0.00000000107,
      -0.00000000718,     0.00000000138,
       0.00000000045,     0.00000000287,
      -0.00000000478,    -0.00000001051,
      -0.00000000373,     0.00000000218,
       0.00000000042,     0.00000000065,
       0.00000000094,    -0.00000000036,
      -0.00000000701,     0.00000000313,
       0.00000000046,     0.00000000115,
       0.00000000025,     0.00000000205,
      -0.00000000212,     0.00000000050,
      -0.00000001370,    -0.00000001414,
      -0.00000000138,    -0.00000000234,
      -0.00000000072,     0.00000000122,
       0.00000000249,    -0.00000000352,
      -0.00000000098,     0.00000000044,
       0.00000000428,    -0.00000000371,
      -0.00000000068,    -0.00000000094,
      -0.00000000100,     0.00000000025,
       0.00000000364,     0.00000001373,
       0.00000000227,    -0.00000000059,
      -0.00000001291,    -0.00000000011,
       0.00000019383,    -0.00000001241,
      -0.00000000025,    -0.00000000456,
       0.00000000336,    -0.00000000083,
      -0.00000000069,    -0.00000000243,
      -0.00000005953,    -0.00000000995,
       0.00000000956,    -0.00000001162,
      -0.00000000354,     0.00000001887,
      -0.00000000375,    -0.00000000323,
      -0.00000000047,     0.00000000112,
      -0.00000000032,     0.00000000135,
       0.00000027375,     0.00000004648,
      -0.00000004323,     0.00000005320,
      -0.00000000510,    -0.00000000859,
       0.00000000815,    -0.00000004389,
       0.00000000864,     0.00000000737,
      -0.00000000144,     0.00000000083,
       0.00000000220,    -0.00000000244,
       0.00000000030,     0.00000000162,
      -0.00000000085,    -0.00000000501,
       0.00000000140,     0.00000000009,
      -0.00000028811,    -0.00000014491,
      -0.00000000087,    -0.00000000652,
       0.00000000102,    -0.00000000029,
      -0.00000006008,     0.00000018076,
      -0.00000002130,    -0.00000001334,
       0.00000000178,    -0.00000000013,
      -0.00000000046,     0.00000000087,
       0.00000000137,     0.00000000140,
      -0.00000007123,     0.00000008400,
      -0.00000005791,     0.00000000090,
      -0.00000000385,    -0.00000000336,
      -0.00000000005,    -0.00000000278,
      -0.00000000059,     0.00000000162,
      -0.00000000145,    -0.00000000048,
       0.00000000013,     0.00000000094,
       0.00000000076,     0.00000000033,
      -0.00000000045,    -0.00000000072,
      -0.00000000071,     0.00000000052,
      -0.00000008184,     0.00000000314,
      -0.00000013714,     0.00000016137,
      -0.00000011138,     0.00000000003,
      -0.00000000508,    -0.00000000430,
      -0.00000000005,    -0.00000000354,
      -0.00000000116,     0.00000000324,
      -0.00000000324,    -0.00000000121,
       0.00000000001,     0.00000000004,
       0.00000000000,     0.00000000021,
      -0.00000000021,    -0.00000000001,
       0.00000000059,    -0.00000000017,
       0.00000000017,     0.00000000059,
       0.00000000465,    -0.00000006552,
       0.00000004472,     0.00000000757,
      -0.00000000411,     0.00000001533,
      -0.00000002093,     0.00000000203,
       0.00000000048,    -0.00000001390,
       0.00000000469,     0.00000000071,
       0.00000002709,     0.00000000250,
      -0.00000000200,     0.00000001390,
      -0.00000000383,    -0.00000000085,
      -0.00000000087,     0.00000001833,
      -0.00000000722,     0.00000000029,
      -0.00000000044,    -0.00000000129,
       0.00000000171,    -0.00000000204,
       0.00000000201,     0.00000000176,
       0.00000005893,     0.00000032501,
      -0.00000022585,     0.00000000514,
       0.00000000922,    -0.00000007768,
       0.00000001747,     0.00000000454,
       0.00000005113,    -0.00000000481,
      -0.00000000112,     0.00000003368,
      -0.00000001134,    -0.00000000171,
       0.00000000067,    -0.00000000251,
      -0.00000003520,    -0.00000000132,
       0.00000000471,    -0.00000002183,
       0.00000000729,     0.00000000277,
      -0.00000000101,     0.00000000174,
      -0.00000000188,     0.00000000013,
      -0.00000000580,    -0.00000000377,
       0.00000000258,    -0.00000000342,
       0.00000000100,     0.00000000093,
      -0.00000000431,     0.00000000519,
      -0.00000000109,    -0.00000000141,
      -0.00000000505,    -0.00000000460,
       0.00000000154,    -0.00000000098,
       0.00000000362,     0.00000000121,
      -0.00000000060,     0.00000000145,
      -0.00000000121,     0.00000000362,
      -0.00000000145,    -0.00000000060,
       0.00000000454,     0.00000000151,
      -0.00000000075,     0.00000000182,
      -0.00000000152,     0.00000000454,
      -0.00000000182,    -0.00000000075,
      -0.00000000007,    -0.00000000217,
      -0.00000000086,     0.00000000053,
      -0.00000000216,     0.00000000004,
       0.00000000051,     0.00000000087,
       0.00000000030,    -0.00000000075,
      -0.00000000072,    -0.00000000034,
      -0.00000000103,    -0.00000005054,
      -0.00000002035,     0.00000001178,
      -0.00000000283,     0.00000000031,
       0.00000000079,     0.00000000109,
      -0.00000000283,     0.00000000663,
       0.00000000310,     0.00000000007,
       0.00000000648,     0.00000000296,
       0.00000000018,    -0.00000000307,
       0.00000000587,    -0.00000001306,
      -0.00000000199,    -0.00000000076,
      -0.00000000135,     0.00000000141,
       0.00000000067,     0.00000000040,
       0.00000000113,     0.00000000141,
       0.00000000048,    -0.00000000057,
      -0.00000006791,     0.00000015247,
       0.00000007173,     0.00000000326,
      -0.00000000573,    -0.00000001574,
       0.00000000915,     0.00000000347,
      -0.00000000009,    -0.00000000426,
      -0.00000000092,     0.00000000041,
       0.00000001252,    -0.00000001180,
      -0.00000000572,    -0.00000000383,
      -0.00000000052,     0.00000000137,
      -0.00000001099,    -0.00000001271,
      -0.00000000411,     0.00000000542,
       0.00000000133,     0.00000000061,
      -0.00000002795,     0.00000002541,
       0.00000001237,     0.00000000873,
       0.00000000370,     0.00000000395,
       0.00000000121,    -0.00000000183,
       0.00000000223,    -0.00000000146,
      -0.00000000059,    -0.00000000075,
      -0.00000000045,    -0.00000000193,
       0.00000029181,    -0.00000026573,
      -0.00000012934,    -0.00000009119,
      -0.00000001276,     0.00000003114,
      -0.00000001668,    -0.00000001758,
      -0.00000000543,     0.00000000816,
       0.00000000197,     0.00000000074,
      -0.00000001965,     0.00000000977,
       0.00000000375,     0.00000000704,
       0.00000000136,    -0.00000000072,
       0.00000000698,     0.00000001884,
       0.00000000720,    -0.00000000247,
      -0.00000000043,    -0.00000000151,
       0.00000004805,    -0.00000002184,
      -0.00000000814,    -0.00000001762,
      -0.00000000271,    -0.00000000671,
      -0.00000000249,     0.00000000105,
      -0.00000000063,     0.00000000217,
       0.00000000000,    -0.00000000145,
      -0.00000044572,     0.00000020365,
       0.00000007566,     0.00000016376,
       0.00000003268,    -0.00000001425,
       0.00000001217,     0.00000002911,
       0.00000001081,    -0.00000000459,
      -0.00000000086,    -0.00000000217,
      -0.00000000101,    -0.00000001203,
      -0.00000000356,     0.00000000124,
      -0.00000000674,     0.00000000617,
       0.00000000278,     0.00000000098,
       0.00000000785,     0.00000002754,
       0.00000000674,    -0.00000000508,
       0.00000000322,    -0.00000000107,
      -0.00000000064,    -0.00000000079,
       0.00000000190,    -0.00000000076,
      -0.00000000028,    -0.00000000076,
       0.00000000103,     0.00000000237,
       0.00000000082,    -0.00000000041,
      -0.00000006707,    -0.00000022516,
      -0.00000005462,     0.00000004136,
      -0.00000001434,     0.00000000486,
       0.00000000285,     0.00000000351,
       0.00000000043,    -0.00000000064,
      -0.00000000247,     0.00000000076,
       0.00000000014,     0.00000000144,
      -0.00000000309,    -0.00000000673,
      -0.00000000196,     0.00000000134,
       0.00000000034,     0.00000000075,
       0.00000000022,     0.00000000121,
      -0.00000000127,     0.00000000037,
       0.00000003863,    -0.00000001278,
      -0.00000001103,    -0.00000000840,
      -0.00000000156,    -0.00000000338,
      -0.00000000043,     0.00000000085,
      -0.00000000019,    -0.00000000205,
       0.00000000257,    -0.00000000097,
      -0.00000000048,    -0.00000000056,
       0.00000006461,     0.00000000511,
       0.00000000022,    -0.00000000113,
      -0.00000000110,    -0.00000000020,
      -0.00000014164,    -0.00000006153,
      -0.00000000169,    -0.00000000054,
      -0.00000000735,     0.00000000161,
       0.00000000067,     0.00000000189,
      -0.00000000245,    -0.00000000105,
      -0.00000000008,    -0.00000000271,
       0.00000003033,    -0.00000000644,
      -0.00000000129,     0.00000001122,
      -0.00000000154,    -0.00000000455,
       0.00000000173,    -0.00000000002,
       0.00000000241,     0.00000000292,
      -0.00000000169,    -0.00000000011,
       0.00000000196,     0.00000000364,
      -0.00000000135,     0.00000000067,
       0.00000000132,     0.00000000043,
       0.00000000045,    -0.00000000131,
       0.00000009703,     0.00000003237,
      -0.00000001613,     0.00000003884,
      -0.00000000158,     0.00000000473,
      -0.00000000188,    -0.00000000078,
      -0.00000000456,    -0.00000000648,
      -0.00000000099,     0.00000000068,
      -0.00000002626,    -0.00000001003,
       0.00000018664,     0.00000006222,
      -0.00000003099,     0.00000007473,
      -0.00000000199,     0.00000000598,
      -0.00000000239,    -0.00000000099,
       0.00000003583,     0.00000000580,
      -0.00000000887,     0.00000002744,
      -0.00000001049,    -0.00000000508,
       0.00000000211,    -0.00000001019,
       0.00000000775,     0.00000000284,
      -0.00000000156,     0.00000000295,
      -0.00000000025,    -0.00000000324,
       0.00000000213,     0.00000000031,
       0.00000000305,    -0.00000000156,
       0.00000000088,     0.00000000204,
      -0.00000011679,     0.00000003225,
       0.00000001327,    -0.00000009902,
       0.00000004012,     0.00000001692,
      -0.00000000697,     0.00000001044,
      -0.00000001017,     0.00000002777,
      -0.00000002035,    -0.00000000901,
       0.00000000416,    -0.00000000757,
       0.00000000186,     0.00000000128,
      -0.00000000158,     0.00000000359,
      -0.00000000263,    -0.00000000203,
       0.00000000113,    -0.00000000089,
      -0.00000000303,     0.00000000349,
      -0.00000000266,    -0.00000000237,
       0.00000000089,    -0.00000000109,
      -0.00000000268,    -0.00000003732,
       0.00000000176,    -0.00000000013,
      -0.00000000007,    -0.00000000143,
      -0.00000000141,     0.00000000004,
      -0.00000000097,    -0.00000003363,
      -0.00000001342,     0.00000000780,
      -0.00000000186,     0.00000000020,
       0.00000000051,     0.00000000072,
      -0.00000000165,     0.00000000407,
       0.00000000187,     0.00000000003,
       0.00000000395,     0.00000000174,
       0.00000000010,    -0.00000000185,
       0.00000000356,    -0.00000000830,
      -0.00000000121,    -0.00000000045,
      -0.00000000073,     0.00000000083,
       0.00000000063,     0.00000000077,
      -0.00000004032,     0.00000009458,
       0.00000004387,     0.00000000171,
       0.00000000556,     0.00000000205,
      -0.00000000005,    -0.00000000256,
       0.00000000676,    -0.00000000681,
      -0.00000000320,    -0.00000000204,
      -0.00000000027,     0.00000000075,
      -0.00000000624,    -0.00000000689,
      -0.00000000223,     0.00000000299,
       0.00000000072,     0.00000000034,
      -0.00000001552,     0.00000001497,
       0.00000000706,     0.00000000482,
       0.00000000209,     0.00000000213,
       0.00000000065,    -0.00000000101,
       0.00000000109,    -0.00000000088,
      -0.00000000022,    -0.00000000090,
       0.00000016000,    -0.00000015457,
      -0.00000007288,    -0.00000004976,
      -0.00000000710,     0.00000001721,
      -0.00000000945,    -0.00000000949,
      -0.00000000293,     0.00000000449,
       0.00000000107,     0.00000000041,
      -0.00000000953,     0.00000000575,
       0.00000000205,     0.00000000335,
       0.00000000064,    -0.00000000037,
       0.00000000395,     0.00000000900,
       0.00000000344,    -0.00000000123,
       0.00000002351,    -0.00000001281,
      -0.00000000436,    -0.00000000852,
      -0.00000000155,    -0.00000000321,
      -0.00000000117,     0.00000000054,
      -0.00000000034,     0.00000000126,
      -0.00000021788,     0.00000011899,
       0.00000004053,     0.00000007917,
       0.00000001594,    -0.00000000693,
       0.00000000698,     0.00000001392,
       0.00000000511,    -0.00000000239,
      -0.00000000041,    -0.00000000103,
      -0.00000000083,    -0.00000000686,
      -0.00000000201,     0.00000000076,
      -0.00000000371,     0.00000000380,
       0.00000000164,     0.00000000050,
       0.00000000552,     0.00000001540,
       0.00000000179,    -0.00000000070,
       0.00000000093,    -0.00000000044,
       0.00000000063,     0.00000000122,
      -0.00000004620,    -0.00000012879,
      -0.00000003055,     0.00000002546,
      -0.00000000799,     0.00000000318,
       0.00000000169,     0.00000000193,
      -0.00000000099,     0.00000000036,
      -0.00000000197,    -0.00000000368,
      -0.00000000100,     0.00000000081,
       0.00000002565,    -0.00000000921,
      -0.00000000109,    -0.00000000211,
      -0.00000000016,    -0.00000000117,
       0.00000000155,    -0.00000000069,
       0.00000003609,    -0.00000000153,
      -0.00000006642,    -0.00000001484,
       0.00000000095,    -0.00000000474,
      -0.00000000143,    -0.00000000028,
       0.00000000038,    -0.00000000081,
      -0.00000000589,     0.00000002567,
      -0.00000001222,    -0.00000000726,
       0.00000000339,     0.00000000068,
      -0.00000000087,     0.00000000167,
      -0.00000000185,    -0.00000000214,
       0.00000000159,    -0.00000000098,
      -0.00000000190,     0.00000000253,
      -0.00000000125,    -0.00000000163,
       0.00000000004,     0.00000000019,
      -0.00000000006,     0.00000000001,
      -0.00000000004,     0.00000000068,
      -0.00000000024,    -0.00000000002,
      -0.00000000068,    -0.00000000004,
       0.00000000002,    -0.00000000024,
      -0.00000000003,     0.00000000003,
       0.00000000121,     0.00000000112,
      -0.00000000043,     0.00000000047,
      -0.00000000112,     0.00000000121,
      -0.00000000047,    -0.00000000043,
       0.00000000543,     0.00000000014,
      -0.00000000004,    -0.00000000124,
       0.00000000180,     0.00000000023,
      -0.00000000034,     0.00000000139,
      -0.00000000043,     0.00000000174,
      -0.00000000130,    -0.00000000055,
      -0.00000001970,    -0.00000000138,
       0.00000000221,    -0.00000001709,
       0.00000000026,     0.00000000318,
      -0.00000000277,     0.00000000005,
      -0.00000000018,     0.00000000271,
      -0.00000000115,    -0.00000000063,
      -0.00000000183,     0.00000000006,
       0.00000000295,     0.00000000380,
      -0.00000000179,     0.00000000168,
      -0.00000000380,     0.00000000295,
      -0.00000000168,    -0.00000000179,
       0.00000000369,     0.00000000475,
      -0.00000000224,     0.00000000211,
      -0.00000000063,    -0.00000000054,
      -0.00000000476,     0.00000000369,
      -0.00000000211,    -0.00000000224,
       0.00000000054,    -0.00000000063,
      -0.00000000086,    -0.00000002185,
      -0.00000000119,     0.00000000014,
      -0.00000000094,     0.00000000246,
       0.00000000112,     0.00000000001,
       0.00000000237,     0.00000000100,
       0.00000000006,    -0.00000000109,
      -0.00000002326,     0.00000005752,
       0.00000002621,     0.00000000078,
       0.00000000333,     0.00000000118,
      -0.00000000004,    -0.00000000151,
       0.00000000359,    -0.00000000389,
      -0.00000000176,    -0.00000000107,
      -0.00000000351,    -0.00000000366,
      -0.00000000119,     0.00000000162,
      -0.00000000840,     0.00000000868,
       0.00000000117,     0.00000000113,
       0.00000000053,    -0.00000000053,
       0.00000008577,    -0.00000008860,
      -0.00000004035,    -0.00000002653,
      -0.00000000530,    -0.00000000503,
      -0.00000000155,     0.00000000243,
      -0.00000000454,     0.00000000334,
       0.00000000111,     0.00000000156,
       0.00000000221,     0.00000000422,
       0.00000000161,    -0.00000000060,
       0.00000001125,    -0.00000000740,
      -0.00000000088,    -0.00000000151,
      -0.00000010415,     0.00000006867,
       0.00000002146,     0.00000003744,
       0.00000000396,     0.00000000654,
       0.00000000238,    -0.00000000123,
      -0.00000000062,    -0.00000000383,
      -0.00000000111,     0.00000000045,
      -0.00000000198,     0.00000000227,
       0.00000000094,     0.00000000026,
       0.00000000357,     0.00000000847,
       0.00000000098,    -0.00000000044,
      -0.00000003010,    -0.00000007121,
      -0.00000001661,     0.00000001513,
      -0.00000000435,     0.00000000200,
       0.00000000098,     0.00000000103,
      -0.00000000123,    -0.00000000197,
       0.00000001522,    -0.00000000870,
      -0.00000000073,    -0.00000000127,
       0.00000000091,    -0.00000000047,
       0.00000002614,     0.00000000514,
      -0.00000000145,    -0.00000000065,
      -0.00000000064,     0.00000000143,
      -0.00000006131,    -0.00000002054,
      -0.00000000085,     0.00000000187,
       0.00000000447,     0.00000000133,
      -0.00000000067,     0.00000000156,
       0.00000000128,    -0.00000000431,
       0.00000000153,     0.00000000064,
       0.00000010440,     0.00000002651,
      -0.00000001539,     0.00000003714,
       0.00000000190,    -0.00000000626,
       0.00000000220,     0.00000000093,
      -0.00000000820,    -0.00000000099,
       0.00000000084,    -0.00000000314,
      -0.00000000074,     0.00000000762,
      -0.00000000298,    -0.00000000075,
       0.00000000014,    -0.00000000122,
      -0.00000018561,    -0.00000002035,
       0.00000001861,    -0.00000007192,
       0.00000001409,     0.00000000587,
      -0.00000000143,     0.00000001255,
      -0.00000000486,    -0.00000000128,
       0.00000000040,    -0.00000000095,
      -0.00000000164,    -0.00000000049,
       0.00000000028,    -0.00000000075,
      -0.00000000101,     0.00000000335,
      -0.00000000135,    -0.00000000053,
       0.00000000637,     0.00000000471,
      -0.00000000369,     0.00000000314,
       0.00000000121,    -0.00000000134,
       0.00000000068,     0.00000000099,
      -0.00000002247,    -0.00000001980,
       0.00000001595,    -0.00000001148,
      -0.00000000278,     0.00000000308,
      -0.00000000160,    -0.00000000228,
      -0.00000005178,    -0.00000001779,
       0.00000000912,    -0.00000002183,
      -0.00000000123,     0.00000000382,
      -0.00000000160,    -0.00000000064,
      -0.00000000206,    -0.00000000046,
       0.00000000030,    -0.00000000083,
      -0.00000000005,    -0.00000000138,
       0.00000000065,     0.00000000120,
       0.00000000149,    -0.00000000143,
       0.00000000076,     0.00000000070,
       0.00000007822,     0.00000010084,
      -0.00000004743,     0.00000004461,
      -0.00000001336,    -0.00000001142,
      -0.00000000497,     0.00000000385,
      -0.00000000220,    -0.00000000234,
       0.00000000056,    -0.00000000066,
      -0.00000000014,     0.00000000104,
       0.00000015108,     0.00000019467,
      -0.00000009157,     0.00000008620,
      -0.00000002584,    -0.00000002206,
      -0.00000000629,     0.00000000488,
      -0.00000000278,    -0.00000000295,
       0.00000000071,    -0.00000000083,
      -0.00000000052,     0.00000000146,
       0.00000000140,     0.00000000056,
      -0.00000001304,     0.00000003436,
       0.00000000196,     0.00000000066,
       0.00000000186,    -0.00000000220,
      -0.00000000096,    -0.00000000055,
      -0.00000000195,    -0.00000000191,
      -0.00000000062,     0.00000000086,
       0.00000000065,     0.00000000059,
       0.00000004506,    -0.00000005020,
      -0.00000002201,    -0.00000001384,
      -0.00000000295,    -0.00000000262,
      -0.00000000080,     0.00000000131,
      -0.00000000212,     0.00000000191,
       0.00000000059,     0.00000000072,
       0.00000000122,     0.00000000193,
       0.00000000074,    -0.00000000029,
      -0.00000000049,    -0.00000000070,
      -0.00000004859,     0.00000003906,
       0.00000001126,     0.00000001730,
       0.00000000222,     0.00000000301,
       0.00000000109,    -0.00000000063,
      -0.00000000043,    -0.00000000210,
      -0.00000000103,     0.00000000133,
      -0.00000001935,    -0.00000003844,
      -0.00000000233,     0.00000000122,
       0.00000000056,     0.00000000054,
      -0.00000000075,    -0.00000000104,
      -0.00000000711,     0.00000002021,
       0.00000000114,     0.00000000036,
       0.00000000095,    -0.00000000123,
      -0.00000000107,    -0.00000000098,
       0.00000002318,    -0.00000002810,
      -0.00000000163,    -0.00000000134,
      -0.00000000041,     0.00000000070,
      -0.00000000096,     0.00000000108,
       0.00000000067,     0.00000000085,
      -0.00000002203,     0.00000002192,
       0.00000000123,     0.00000000135,
      -0.00000000029,    -0.00000000113,
      -0.00000000052,     0.00000000076,
      -0.00000001184,    -0.00000002048,
      -0.00000000122,     0.00000000073,
       0.00000001168,    -0.00000001557,
      -0.00000000089,    -0.00000000067,
      -0.00000000962,     0.00000001213,
       0.00000000067,     0.00000000059,
       0.00000000071,     0.00000000125,
       0.00000000125,    -0.00000000072,
       0.00000001958,     0.00000002637,
       0.00000000678,    -0.00000001450,
       0.00000000143,    -0.00000000106,
      -0.00000000090,    -0.00000000027,
      -0.00000000186,    -0.00000000692,
      -0.00000000276,     0.00000000173,
       0.00000000079,     0.00000000038,
      -0.00000000692,     0.00000000180,
       0.00000000171,     0.00000000277,
       0.00000000039,    -0.00000000079,
       0.00000000013,     0.00000000961,
       0.00000000189,    -0.00000000027,
      -0.00000000059,    -0.00000000073,
       0.00000000133,    -0.00000000293,
      -0.00000000143,    -0.00000000005,
      -0.00000000290,    -0.00000000142,
      -0.00000000006,     0.00000000143,
      -0.00000005574,    -0.00000014673,
      -0.00000006113,     0.00000004309,
       0.00000001612,     0.00000000849,
      -0.00000000886,     0.00000000122,
       0.00000000268,     0.00000000341,
       0.00000000041,    -0.00000000111,
      -0.00000000761,     0.00000002641,
       0.00000001245,     0.00000000041,
      -0.00000000115,    -0.00000000276,
       0.00000002618,     0.00000000794,
       0.00000000063,    -0.00000001245,
      -0.00000000279,     0.00000000111,
       0.00000000165,     0.00000000020,
       0.00000000020,    -0.00000000162,
       0.00000002024,    -0.00000004173,
      -0.00000002028,    -0.00000000111,
      -0.00000000791,    -0.00000000301,
       0.00000000019,     0.00000000379,
       0.00000000082,    -0.00000000043,
      -0.00000000775,     0.00000000660,
       0.00000000341,     0.00000000238,
       0.00000000032,    -0.00000000085,
       0.00000000609,     0.00000000796,
       0.00000000261,    -0.00000000324,
      -0.00000000084,    -0.00000000035,
      -0.00000015381,     0.00000056449,
       0.00000026948,     0.00000000066,
      -0.00000002386,    -0.00000005972,
       0.00000003700,     0.00000000769,
      -0.00000000078,    -0.00000001736,
      -0.00000000377,     0.00000000192,
       0.00000000060,     0.00000000046,
       0.00000003813,    -0.00000006102,
      -0.00000003056,    -0.00000001889,
      -0.00000000281,     0.00000000756,
       0.00000000123,     0.00000000008,
      -0.00000005940,    -0.00000003933,
      -0.00000001966,     0.00000002998,
       0.00000000752,     0.00000000310,
       0.00000000012,    -0.00000000125,
       0.00000000199,     0.00000000012,
       0.00000000013,    -0.00000000197,
       0.00000004552,     0.00000000597,
       0.00000000022,    -0.00000000212,
      -0.00000000469,    -0.00000000051,
      -0.00000000051,     0.00000000462,
       0.00000000539,     0.00000000031,
       0.00000000028,    -0.00000000290,
      -0.00000000541,    -0.00000000053,
      -0.00000000054,     0.00000000536,
      -0.00000013483,     0.00000010985,
       0.00000005676,     0.00000004230,
       0.00000000566,    -0.00000001438,
       0.00000001912,     0.00000002222,
       0.00000000686,    -0.00000001009,
      -0.00000000258,    -0.00000000083,
       0.00000001742,    -0.00000000601,
      -0.00000000291,    -0.00000000637,
      -0.00000000121,     0.00000000068,
      -0.00000000330,    -0.00000001693,
      -0.00000000669,     0.00000000166,
       0.00000000042,     0.00000000141,
       0.00000079288,    -0.00000129358,
      -0.00000064872,    -0.00000041025,
      -0.00000005959,     0.00000016218,
       0.00000002620,     0.00000000197,
      -0.00000008891,    -0.00000006233,
      -0.00000002693,     0.00000004567,
       0.00000001136,     0.00000000374,
       0.00000000004,    -0.00000000185,
       0.00000004434,     0.00000006467,
       0.00000002476,     0.00000004489,
       0.00000001147,    -0.00000000531,
      -0.00000000082,    -0.00000000149,
       0.00000005636,    -0.00000004549,
       0.00000004583,    -0.00000002093,
      -0.00000000451,    -0.00000001207,
      -0.00000000164,     0.00000000072,
       0.00000004855,     0.00000000308,
       0.00000000011,    -0.00000000249,
      -0.00000000556,    -0.00000000033,
      -0.00000000034,     0.00000000547,
      -0.00000012276,    -0.00000001511,
      -0.00000000062,     0.00000000647,
       0.00000000615,     0.00000000089,
       0.00000000087,    -0.00000000581,
       0.00000001998,     0.00000000603,
       0.00000002753,     0.00000000973,
      -0.00000004224,    -0.00000000365,
      -0.00000000146,     0.00000002023,
       0.00000000144,     0.00000000027,
       0.00000003065,     0.00000000205,
       0.00000000020,    -0.00000000144,
       0.00000000211,    -0.00000003018,
      -0.00000000139,    -0.00000000020,
       0.00000035928,    -0.00000009973,
      -0.00000004883,    -0.00000013538,
      -0.00000002678,     0.00000001168,
      -0.00000001389,    -0.00000005715,
      -0.00000002181,     0.00000000731,
       0.00000000181,     0.00000000432,
      -0.00000000285,     0.00000000925,
       0.00000000343,     0.00000000026,
       0.00000000103,    -0.00000000506,
      -0.00000000201,     0.00000000082,
       0.00000114114,     0.00000133187,
       0.00000050997,     0.00000094953,
       0.00000023455,    -0.00000011396,
      -0.00000001728,    -0.00000003206,
       0.00000007813,     0.00000000380,
       0.00000007590,    -0.00000003486,
      -0.00000000806,    -0.00000001763,
      -0.00000000244,     0.00000000121,
      -0.00000116414,    -0.00000014014,
      -0.00000002411,     0.00000005808,
       0.00000000335,     0.00000000317,
      -0.00000011919,     0.00000117943,
       0.00000006355,     0.00000001311,
       0.00000000043,    -0.00000000393,
       0.00000000103,     0.00000000022,
       0.00000000073,    -0.00000000162,
      -0.00000000314,    -0.00000000018,
      -0.00000000019,     0.00000000312,
      -0.00000012857,    -0.00000001166,
      -0.00000000024,     0.00000000695,
       0.00000000609,     0.00000000114,
       0.00000000113,    -0.00000000570,
       0.00000000081,    -0.00000000076,
      -0.00000000084,    -0.00000000080,
       0.00000024090,     0.00000005435,
       0.00000000098,    -0.00000000938,
       0.00000000444,     0.00000000076,
       0.00000000086,    -0.00000000555,
      -0.00000004053,    -0.00000001019,
       0.00000003800,     0.00000001446,
      -0.00000000033,     0.00000000092,
      -0.00000002595,    -0.00000001248,
       0.00000000043,    -0.00000000122,
      -0.00000000084,    -0.00000000054,
      -0.00000000074,     0.00000000146,
       0.00000030082,     0.00000002045,
       0.00000000187,    -0.00000001380,
       0.00000000641,    -0.00000013129,
      -0.00000000631,    -0.00000000103,
      -0.00000014301,    -0.00000000459,
      -0.00000000033,     0.00000000347,
      -0.00000000518,     0.00000013957,
       0.00000000312,     0.00000000025,
       0.00000000983,     0.00000012861,
       0.00000003308,    -0.00000001667,
       0.00000001610,    -0.00000000195,
      -0.00000000230,    -0.00000000396,
      -0.00000000049,     0.00000000059,
       0.00000001321,    -0.00000000309,
      -0.00000000159,    -0.00000000517,
      -0.00000000105,     0.00000000043,
       0.00000000380,     0.00000001458,
       0.00000000542,    -0.00000000196,
      -0.00000000047,    -0.00000000106,
      -0.00002489021,    -0.00000267365,
      -0.00000043105,     0.00000127144,
       0.00000007334,     0.00000004341,
      -0.00000013901,     0.00000149276,
       0.00000008109,     0.00000002902,
       0.00000000307,    -0.00000000516,
       0.00000566865,     0.00000031561,
       0.00000001920,    -0.00000011418,
      -0.00000000129,    -0.00000000052,
       0.00000030128,    -0.00000568899,
      -0.00000013328,    -0.00000001105,
       0.00000000142,     0.00000000548,
      -0.00000000154,    -0.00000000423,
      -0.00000000387,     0.00000000129,
      -0.00000005566,    -0.00000000403,
      -0.00000000012,     0.00000000351,
       0.00000000760,     0.00000000049,
       0.00000000050,    -0.00000000747,
       0.00000027460,     0.00000006784,
       0.00000000067,    -0.00000000531,
      -0.00000000437,    -0.00000000139,
      -0.00000000134,     0.00000000325,
      -0.00000007277,    -0.00000001832,
      -0.00000000029,     0.00000000110,
      -0.00000000032,     0.00000000146,
      -0.00000028469,    -0.00000010904,
      -0.00000000692,    -0.00000002188,
       0.00000000514,    -0.00000001943,
      -0.00000002162,    -0.00000001150,
       0.00000000014,    -0.00000000189,
      -0.00000001238,     0.00000002552,
      -0.00000000207,    -0.00000000008,
      -0.00000002853,    -0.00000001049,
      -0.00000000087,    -0.00000000023,
      -0.00000005056,    -0.00000002684,
       0.00000000119,    -0.00000000224,
      -0.00000185507,    -0.00000006883,
      -0.00000000444,     0.00000004326,
      -0.00000001961,     0.00000074111,
       0.00000001778,     0.00000000227,
       0.00000043315,     0.00000000085,
       0.00000000478,    -0.00000041071,
       0.00000000061,     0.00000000057,
      -0.00000000038,    -0.00000000063,
      -0.00000001835,    -0.00000004876,
       0.00000000647,     0.00000000340,
       0.00000000944,     0.00000000191,
      -0.00000000123,    -0.00000000126,
      -0.00000001076,    -0.00000001100,
       0.00000000405,     0.00000000171,
      -0.00000002363,     0.00000001000,
       0.00000000299,    -0.00000000119,
       0.00010992607,     0.00000679037,
       0.00000047595,    -0.00000232847,
      -0.00000008283,     0.00000000708,
       0.00000015901,    -0.00000773201,
      -0.00000019002,    -0.00000002212,
      -0.00000000079,     0.00000000464,
      -0.00002041873,    -0.00000139121,
      -0.00000006483,    -0.00000010975,
       0.00000000760,    -0.00000001090,
      -0.00000137721,     0.00002018920,
      -0.00000012353,     0.00000005794,
      -0.00000001115,    -0.00000000709,
      -0.00000000467,     0.00000000652,
       0.00000000645,     0.00000000668,
      -0.00000000260,     0.00000000001,
       0.00000000001,     0.00000000259,
       0.00000000217,     0.00000000012,
       0.00000000012,    -0.00000000216,
      -0.00000003524,     0.00000000908,
      -0.00000000023,    -0.00000000104,
       0.00000012237,     0.00000001818,
      -0.00000000005,    -0.00000000744,
      -0.00000000841,    -0.00000000219,
      -0.00000000218,     0.00000000799,
      -0.00000012427,    -0.00000004237,
       0.00000000210,    -0.00000000635,
       0.00000000256,     0.00000000065,
      -0.00000000075,     0.00000000297,
      -0.00000005757,    -0.00000002396,
      -0.00000000021,     0.00000000084,
       0.00000000106,    -0.00000000313,
      -0.00000098415,    -0.00000045841,
       0.00000000280,    -0.00000008030,
       0.00000002071,    -0.00000004444,
       0.00000000362,     0.00000000013,
      -0.00000000335,    -0.00000000117,
       0.00000000058,    -0.00000000156,
       0.00000011543,     0.00000004585,
      -0.00000000069,     0.00000000167,
       0.00000000091,     0.00000000051,
       0.00000000089,    -0.00000000203,
       0.00000008663,     0.00000004284,
      -0.00000000124,     0.00000000248,
       0.00000805974,     0.00000000919,
       0.00000000391,    -0.00000295617,
       0.00000000030,     0.00000000085,
       0.00000000353,     0.00000002056,
      -0.00000000053,     0.00000000422,
      -0.00000000720,    -0.00000010012,
      -0.00000000047,    -0.00000000201,
       0.00000000148,     0.00000000229,
      -0.00000000187,     0.00000000186,
      -0.00000006160,     0.00000002436,
      -0.00000000223,    -0.00000000108,
       0.00000000123,    -0.00000000129,
       0.00000000038,     0.00000000285,
      -0.00034262602,    -0.00006599541,
      -0.00000311993,    -0.00000556709,
       0.00000036097,    -0.00000035931,
       0.00000000713,     0.00000001491,
       0.00000091713,     0.00002487227,
       0.00000007738,    -0.00000005617,
      -0.00000000295,     0.00000000543,
       0.00004883957,     0.00001254812,
       0.00000117982,     0.00000223400,
      -0.00000011791,     0.00000011593,
      -0.00000000509,    -0.00000000500,
       0.00001254685,    -0.00004770735,
       0.00000226321,    -0.00000117196,
       0.00000011625,     0.00000011685,
      -0.00000000503,     0.00000000506,
       0.00000002820,    -0.00000003465,
       0.00000000143,     0.00000000099,
      -0.00000002864,    -0.00000003166,
       0.00000000092,    -0.00000000164,
      -0.00000006679,    -0.00000002159,
       0.00000000070,    -0.00000000157,
       0.00000000016,     0.00000000834,
      -0.00000004282,     0.00000000124,
       0.00000000011,     0.00000000252,
       0.00000000949,    -0.00000000035,
      -0.00000000034,    -0.00000000941,
       0.00000003136,     0.00000000299,
      -0.00000000001,    -0.00000000186,
      -0.00000000515,    -0.00000000052,
      -0.00000000053,     0.00000000508,
       0.00000018113,    -0.00000003672,
       0.00000000110,     0.00000000594,
      -0.00000028463,    -0.00000007445,
      -0.00000000669,    -0.00000001317,
       0.00000000298,    -0.00000001142,
      -0.00000027333,    -0.00000010255,
      -0.00000000101,     0.00000000636,
       0.00000001557,     0.00000000519,
       0.00000000028,     0.00000000107,
       0.00000000542,    -0.00000001603,
       0.00000000109,    -0.00000000030,
       0.00000153699,     0.00000063692,
      -0.00000000044,     0.00000012296,
      -0.00000001894,     0.00000004552,
      -0.00000000364,    -0.00000000002,
       0.00000000432,     0.00000000123,
      -0.00000000156,     0.00000000447,
      -0.00000000098,    -0.00000000032,
       0.00000000032,    -0.00000000098,
       0.00000000087,     0.00000000028,
      -0.00000000029,     0.00000000087,
       0.00000015658,    -0.00000002749,
      -0.00000001008,    -0.00000005037,
      -0.00000002716,     0.00000000055,
      -0.00000000055,    -0.00000002886,
       0.00000001580,     0.00000000469,
      -0.00000000093,     0.00000000254,
      -0.00000000450,     0.00000000000,
       0.00000000026,    -0.00000000454,
       0.00353806375,     0.00090470455,
       0.00008486764,     0.00016358543,
      -0.00000851116,     0.00000843784,
      -0.00000036857,    -0.00000036331,
       0.00000001326,    -0.00000000911,
      -0.00004223570,     0.00016600353,
      -0.00000765769,     0.00000396501,
      -0.00000039453,    -0.00000039794,
       0.00000001699,    -0.00000001723,
       0.00000000042,     0.00000000061,
       0.00000225218,     0.00000002430,
       0.00000004920,     0.00000015239,
      -0.00000000696,     0.00000000647,
       0.00000082383,    -0.00000429062,
       0.00000021978,    -0.00000008717,
       0.00000001024,     0.00000001134,
      -0.00000023746,     0.00000017269,
      -0.00000002164,    -0.00000001067,
      -0.00000000092,    -0.00000000084,
       0.00000016569,     0.00000023307,
      -0.00000001077,     0.00000002154,
      -0.00000000084,     0.00000000093,
       0.00000877806,     0.00000085468,
       0.00000015453,     0.00000030761,
      -0.00000000742,     0.00000007801,
      -0.00000000251,     0.00000000143,
       0.00000059083,     0.00000011665,
       0.00000001066,     0.00000003498,
      -0.00000000176,     0.00000000143,
       0.00000011731,    -0.00000060589,
       0.00000003502,    -0.00000001066,
       0.00000000143,     0.00000000177,
       0.00000013021,    -0.00000000823,
      -0.00000000076,    -0.00000000896,
      -0.00000002343,     0.00000000090,
       0.00000000087,     0.00000002313,
      -0.00000000210,    -0.00000000006,
      -0.00000000006,     0.00000000210,
       0.00000000533,    -0.00000000044,
      -0.00000000016,    -0.00000000146,
       0.00000000108,    -0.00000000005,
      -0.00000000006,    -0.00000000112,
      -0.00000012044,    -0.00000001755,
       0.00000000042,    -0.00000000177,
       0.00000000798,     0.00000000204,
       0.00000000202,    -0.00000000777,
       0.00000009057,    -0.00000012065,
       0.00000000100,    -0.00000000252,
       0.00000002245,     0.00000000996,
       0.00000000231,     0.00000000217,
       0.00000000996,    -0.00000002238,
       0.00000000218,    -0.00000000230,
      -0.00000000039,     0.00000000133,
       0.00003608292,     0.00000708241,
       0.00000064217,     0.00000211286,
      -0.00000010608,     0.00000008581,
      -0.00000021690,     0.00000110461,
      -0.00000006441,     0.00000001957,
      -0.00000000262,    -0.00000000324,
       0.00000003240,     0.00000000044,
       0.00000000063,     0.00000000094,
      -0.00000000722,     0.00000003657,
      -0.00000000142,     0.00000000038,
       0.00000000251,     0.00000000036,
       0.00000000036,    -0.00000000253,
       0.00000000063,     0.00000000106,
      -0.00000000106,     0.00000000063,
      -0.00000000209,    -0.00000000035,
       0.00000000035,    -0.00000000209,
      -0.00000132346,    -0.00000030701,
      -0.00000003068,    -0.00000015203,
      -0.00000001604,     0.00000005115,
      -0.00000000686,     0.00000000135,
       0.00000095241,     0.00000012090,
       0.00000001997,     0.00000005415,
      -0.00000000266,     0.00000000245,
      -0.00000012068,     0.00000095343,
      -0.00000005421,     0.00000001999,
      -0.00000000245,    -0.00000000266,
      -0.00000020578,     0.00000014925,
      -0.00000001886,    -0.00000000937,
      -0.00000000081,    -0.00000000073,
      -0.00000014914,    -0.00000020569,
       0.00000000937,    -0.00000001885,
       0.00000000073,    -0.00000000080,
       0.00000000183,     0.00000000014,
       0.00000000015,    -0.00000000175,
      -0.00000000096,    -0.00000000128,
      -0.00000000128,     0.00000000096,
      -0.00000000117,    -0.00000000006,
      -0.00000000006,     0.00000000117,
      -0.00000010101,     0.00000004515,
      -0.00000000087,     0.00000000205,
       0.00000003748,     0.00000000919,
      -0.00000000073,    -0.00000000071,
       0.00000000917,    -0.00000003748,
      -0.00000000071,     0.00000000073,
      -0.00000002330,    -0.00000000139,
      -0.00000000046,     0.00000000843,
       0.00000114675,     0.00000012750,
       0.00000003154,     0.00000001167,
      -0.00000000610,     0.00000005846,
      -0.00000000054,     0.00000000154,
       0.00000001336,     0.00000000212,
       0.00000000227,    -0.00000001424,
      -0.00000308169,     0.00000019524,
      -0.00000001349,     0.00000004201,
      -0.00000000203,    -0.00000003464,
      -0.00000057505,    -0.00000018128,
      -0.00000001224,    -0.00000001793,
       0.00000000115,    -0.00000000117,
      -0.00000018139,     0.00000057448,
      -0.00000001797,     0.00000001223,
      -0.00000000117,    -0.00000000115,
       0.00000017136,     0.00000002873,
      -0.00000000162,     0.00000000874,
      -0.00000000124,    -0.00000000037,
      -0.00000000030,     0.00000000127,
      -0.00003228564,    -0.00001020872,
      -0.00000068458,    -0.00000100905,
       0.00000006410,    -0.00000006509,
       0.00000032959,    -0.00000104377,
       0.00000003271,    -0.00000002224,
       0.00000000211,     0.00000000207,
      -0.00000006140,    -0.00000000885,
      -0.00000000124,    -0.00000000195,
      -0.00000000894,     0.00000003564,
      -0.00000000233,     0.00000000125,
      -0.00000000236,    -0.00000000056,
      -0.00000000056,     0.00000000237,
       0.00000000110,     0.00000000028,
      -0.00000000028,     0.00000000109,
      -0.00000001845,    -0.00000001345,
      -0.00000000064,     0.00000000103,
      -0.00000000788,    -0.00000000292,
       0.00000000291,    -0.00000000787,
       0.00000000213,     0.00000000066,
      -0.00000000066,     0.00000000213,
       0.00000004576,     0.00000000716,
      -0.00000002625,    -0.00000000354,
      -0.00000000058,    -0.00000000251,
       0.00000000353,    -0.00000002627,
       0.00000000251,    -0.00000000058,
       0.00000000406,    -0.00000000382,
       0.00000000382,     0.00000000406,
       0.00000000136,    -0.00000000007,
      -0.00000000007,    -0.00000000136,
       0.00000006291,     0.00000000252,
      -0.00000000010,     0.00000000338,
       0.00000000119,     0.00000000010,
       0.00000000010,    -0.00000000126,
      -0.00000009959,     0.00000003497,
      -0.00000000035,    -0.00000000096,
      -0.00000002147,    -0.00000000098,
      -0.00000000099,     0.00000002145,
       0.00000002281,     0.00000000192,
      -0.00000000012,     0.00000000115,
      -0.00000127605,    -0.00000008072,
      -0.00000003559,    -0.00000001104,
       0.00000000296,    -0.00000004164,
       0.00000000039,    -0.00000000119,
      -0.00000001669,    -0.00000000257,
      -0.00000000281,     0.00000001577,
       0.00000005719,     0.00000000800,
      -0.00000000498,     0.00000003161,
      -0.00000000085,     0.00000000090,
      -0.00000030528,    -0.00000004654,
       0.00000000777,    -0.00000004498,
       0.00000000122,    -0.00000000125,
       0.00000018934,     0.00000004803,
       0.00000000187,    -0.00000000738,
      -0.00000048623,    -0.00000012223,
      -0.00000001157,    -0.00000002238,
      -0.00000000147,     0.00000000573,
      -0.00000000493,    -0.00000000093,
       0.00000000097,    -0.00000000490,
       0.00000000083,     0.00000000043,
      -0.00000000107,     0.00000000001,
       0.00000000000,     0.00000000107,
      -0.00000006481,     0.00000000023,
       0.00000000003,    -0.00000000215,
      -0.00000000140,    -0.00000000009,
      -0.00000000010,     0.00000000135,
      -0.00000000021,     0.00000000272,
      -0.00000003980,    -0.00000000138,
       0.00000000032,    -0.00000000387,
       0.00000003062,     0.00000000374,
       0.00000000017,    -0.00000000128,
      -0.00000010544,    -0.00000001165,
      -0.00000000015,     0.00000000106,
       0.00000009241,     0.00000002466,
       0.00000000045,    -0.00000000162,
       0.00000000278,     0.00000000046,
      -0.00000000045,     0.00000000272,
      -0.00000023253,    -0.00000005860,
      -0.00000000029,     0.00000000116,
      -0.00000000052,    -0.00000000063,
      -0.00000000063,     0.00000000052,
      -0.00000003120,    -0.00000000294,
       0.00000001759,     0.00000000450,
      -0.00000003809,    -0.00000000973,
      -0.00000000142,     0.00000000017,
       0.00000000016,     0.00000000140,
      -0.00000001909,    -0.00000002261,
      -0.00000000100,     0.00000000075,
      -0.00000000582,     0.00000000086,
       0.00000000061,     0.00000000225,
       0.00000000089,     0.00000000584,
       0.00000000226,    -0.00000000064,
      -0.00000005470,    -0.00000000544,
       0.00000003070,     0.00000000780,
      -0.00000006735,    -0.00000001710,
       0.00000002134,    -0.00000000317,
      -0.00000000226,    -0.00000000827,
      -0.00000000048,    -0.00000000309,
      -0.00000000120,     0.00000000034,
      -0.00000021230,     0.00000003042,
       0.00000002227,     0.00000008134,
       0.00000000136,     0.00000000887,
       0.00000000344,    -0.00000000095,
       0.00000000082,    -0.00000000483,
      -0.00000000096,     0.00000000017,
      -0.00000000450,    -0.00000000104,
       0.00000005454,     0.00000001381,
      -0.00000012211,    -0.00000003083,
       0.00000000560,    -0.00000000014,
      -0.00000000026,    -0.00000000213,
      -0.00000000110,    -0.00000000171,
      -0.00000000171,     0.00000000110,
      -0.00000000231,     0.00000001407,
       0.00000000164,     0.00000000028,
       0.00000000138,    -0.00000000024,
       0.00000000023,     0.00000000149,
      -0.00000025333,    -0.00000001630,
       0.00000000287,     0.00000001800,
       0.00000003251,     0.00000000466,
       0.00000000096,     0.00000000037,
       0.00000000476,    -0.00000003164,
       0.00000000044,    -0.00000000096,
       0.00000002325,    -0.00000028571,
      -0.00000004109,     0.00000000162,
      -0.00000000794,    -0.00000000053,
       0.00000000015,     0.00000000137,
       0.00000004800,     0.00000005855,
       0.00000000360,    -0.00000000576,
       0.00000005813,    -0.00000005058,
      -0.00000000576,    -0.00000000360,
      -0.00000002164,    -0.00000003603,
      -0.00000001192,     0.00000001325,
       0.00000000121,     0.00000000072,
       0.00000007389,     0.00000000006,
      -0.00000000047,    -0.00000007366,
      -0.00000000474,    -0.00000000482,
      -0.00000000090,     0.00000000036,
      -0.00000000020,     0.00000000094,
      -0.00000000094,    -0.00000000009,
      -0.00000002118,    -0.00000000491,
       0.00000000467,     0.00000000088,
       0.00000000092,    -0.00000000465,
       0.00000656534,    -0.00000343877,
       0.00000014717,    -0.00000028797,
       0.00000005962,    -0.00000012094,
      -0.00000000903,    -0.00000000512,
      -0.00000237892,    -0.00000046241,
       0.00000004412,     0.00000000360,
       0.00000000039,     0.00000000159,
      -0.00000046146,     0.00000237848,
       0.00000000355,    -0.00000004421,
       0.00000000158,    -0.00000000038,
       0.00000227728,    -0.00000000053,
      -0.00000000015,    -0.00000079761,
      -0.00000000563,     0.00000000607,
      -0.00000000012,     0.00000000116,
      -0.00000000188,    -0.00000002487,
       0.00000000004,    -0.00000000098,
      -0.00000001057,     0.00000000177,
       0.00000000078,    -0.00000000022,
       0.00000000020,     0.00000000089,
       0.00002534959,     0.00000803764,
       0.00000053835,     0.00000079484,
      -0.00000005038,     0.00000005203,
      -0.00000038222,     0.00000123688,
      -0.00000003797,     0.00000002562,
      -0.00000000250,    -0.00000000240,
       0.00000016967,     0.00000003936,
       0.00000000392,     0.00000000788,
       0.00000004690,    -0.00000018318,
       0.00000000824,    -0.00000000413,
      -0.00000000177,     0.00000000105,
       0.00000000101,     0.00000000173,
       0.00031587069,    -0.00003604104,
       0.00000334310,     0.00000029051,
       0.00000003108,     0.00000027222,
       0.00000038583,     0.00000310812,
      -0.00000000311,     0.00000003467,
      -0.00000000257,     0.00000000035,
       0.00005774875,     0.00001472929,
       0.00000138320,     0.00000266522,
      -0.00000013886,     0.00000013807,
      -0.00000000603,    -0.00000000591,
       0.00001474374,    -0.00005769557,
       0.00000266781,    -0.00000138182,
       0.00000013817,     0.00000013874,
      -0.00000000592,     0.00000000603,
      -0.00000001751,    -0.00000000050,
       0.00000000074,     0.00000000062,
      -0.00000000053,     0.00000001812,
       0.00000000060,    -0.00000000076,
       0.00000021351,     0.00000005438,
       0.00000000041,    -0.00000000160,
       0.00000022612,     0.00000005781,
       0.00000000041,    -0.00000000162,
      -0.00000002967,     0.00000000151,
      -0.00000000011,    -0.00000000192,
      -0.00000000142,     0.00000000003,
      -0.00000000009,    -0.00000000146,
      -0.00000005011,     0.00000000378,
       0.00000000380,     0.00000002048,
      -0.00000000027,    -0.00000000335,
      -0.00000000136,     0.00000000026,
      -0.00000000038,     0.00000000097,
      -0.00000000247,     0.00000000014,
       0.00000000017,     0.00000000101,
      -0.00000000019,    -0.00000000257,
      -0.00000000105,     0.00000000020,
      -0.00000000027,     0.00000000094,
      -0.00000004229,    -0.00000008061,
      -0.00000000343,     0.00000000180,
      -0.00000000108,    -0.00000000216,
       0.00000000216,    -0.00000000108,
      -0.00000000103,    -0.00000000031,
       0.00000000013,     0.00000000131,
      -0.00000008145,     0.00000000757,
       0.00000000721,     0.00000003305,
      -0.00000000054,    -0.00000000565,
      -0.00000000229,     0.00000000049,
       0.00000001801,     0.00000000584,
      -0.00000000044,     0.00000000120,
      -0.00000000421,     0.00000000031,
       0.00000000035,     0.00000000170,
      -0.00000000039,    -0.00000000438,
      -0.00000000177,     0.00000000039,
       0.00000000119,     0.00000000030,
      -0.00000000030,     0.00000000128,
      -0.00000000815,    -0.00000000900,
      -0.00000000068,     0.00000000063,
      -0.00000000121,    -0.00000000036,
       0.00000001348,    -0.00000000140,
       0.00000000022,     0.00000000214,
       0.00000000087,    -0.00000000021,
      -0.00000012585,     0.00000001307,
       0.00000001223,     0.00000005085,
      -0.00000000096,    -0.00000000919,
      -0.00000000371,     0.00000000088,
       0.00000000032,     0.00000000076,
      -0.00000000037,     0.00000000103,
      -0.00000000090,    -0.00000000040,
       0.00000001695,     0.00000000439,
      -0.00000000038,     0.00000000123,
      -0.00000000694,     0.00000000061,
       0.00000000063,     0.00000000279,
      -0.00000000069,    -0.00000000721,
      -0.00000000292,     0.00000000069,
       0.00000000147,     0.00000000021,
      -0.00000000020,     0.00000000159,
      -0.00000000104,     0.00000000028,
      -0.00000000120,    -0.00000000038,
       0.00000001939,    -0.00000000211,
      -0.00000000198,    -0.00000000780,
       0.00000000035,     0.00000000333,
       0.00000000134,    -0.00000000034,
      -0.00000018123,     0.00000002101,
       0.00000001892,     0.00000007285,
       0.00000001471,    -0.00000000641,
      -0.00000000146,    -0.00000001420,
      -0.00000000573,     0.00000000143,
       0.00000000050,     0.00000000117,
      -0.00000000054,     0.00000000103,
      -0.00000000096,    -0.00000000036,
       0.00000000121,    -0.00000000006,
       0.00000000018,     0.00000000128,
      -0.00000001100,     0.00000000101,
       0.00000000106,     0.00000000442,
       0.00000000091,    -0.00000000037,
      -0.00000000115,    -0.00000001134,
      -0.00000000458,     0.00000000112,
       0.00000000038,     0.00000000093,
       0.00000002670,    -0.00000000424,
       0.00000000006,     0.00000000173,
       0.00000000157,    -0.00000000023,
       0.00000000028,     0.00000000173,
      -0.00000000121,     0.00000000016,
       0.00000002150,    -0.00000000402,
      -0.00000000270,    -0.00000000879,
       0.00000000083,     0.00000000384,
       0.00000000195,    -0.00000000049,
      -0.00000000040,     0.00000000146,
      -0.00000000024,     0.00000000081,
      -0.00000023364,     0.00000002502,
       0.00000002384,     0.00000009413,
       0.00000001911,    -0.00000000813,
      -0.00000000212,    -0.00000002031,
      -0.00000000821,     0.00000000203,
       0.00000000071,     0.00000000167,
      -0.00000000104,     0.00000000075,
       0.00000000187,    -0.00000000003,
       0.00000000033,     0.00000000190,
      -0.00000003225,    -0.00000001761,
       0.00000000172,    -0.00000000017,
      -0.00000001653,     0.00000000144,
       0.00000000157,     0.00000000666,
       0.00000000137,    -0.00000000054,
      -0.00000000179,    -0.00000001691,
      -0.00000000683,     0.00000000168,
       0.00000000057,     0.00000000139,
       0.00000004330,    -0.00000001225,
      -0.00000000510,    -0.00000001794,
       0.00000000031,     0.00000000299,
       0.00000000128,    -0.00000000023,
       0.00000000104,    -0.00000000154,
       0.00000000171,     0.00000000124,
      -0.00000001805,     0.00000000398,
       0.00000000172,    -0.00000000006,
       0.00000000004,     0.00000000155,
       0.00000000023,     0.00000000104,
       0.00000000046,    -0.00000000126,
       0.00000000088,    -0.00000000047,
       0.00000000043,     0.00000000090,
      -0.00000001966,    -0.00000010985,
      -0.00000005398,     0.00000002303,
       0.00000000915,     0.00000001325,
      -0.00000000422,     0.00000000049,
       0.00000000076,     0.00000000209,
       0.00000002633,    -0.00000000229,
      -0.00000000247,    -0.00000001065,
       0.00000000056,     0.00000000619,
       0.00000000251,    -0.00000000059,
      -0.00000024843,    -0.00000000193,
       0.00000001169,     0.00000010388,
       0.00000002220,    -0.00000000515,
      -0.00000000253,    -0.00000002548,
      -0.00000001031,     0.00000000251,
       0.00000000083,     0.00000000212,
      -0.00000000110,    -0.00000000357,
      -0.00000000166,     0.00000000089,
       0.00000000324,    -0.00000000071,
      -0.00000000075,    -0.00000000164,
      -0.00000000063,     0.00000000148,
       0.00000000148,    -0.00000000578,
      -0.00000000202,    -0.00000000057,
      -0.00000000478,     0.00000000237,
       0.00000000078,     0.00000000142,
      -0.00000005139,    -0.00000005718,
      -0.00000000933,     0.00000002466,
       0.00000000574,    -0.00000000419,
      -0.00000000209,    -0.00000000103,
      -0.00000002368,    -0.00000000172,
       0.00000000099,     0.00000000962,
       0.00000000193,    -0.00000000072,
      -0.00000000511,    -0.00000002349,
      -0.00000000956,     0.00000000305,
       0.00000000079,     0.00000000195,
      -0.00000000030,    -0.00000000375,
      -0.00000000114,     0.00000000009,
      -0.00000000374,     0.00000000028,
       0.00000000008,     0.00000000114,
      -0.00000000190,    -0.00000000103,
       0.00000003980,    -0.00000032594,
      -0.00000007052,    -0.00000002429,
       0.00000000164,     0.00000000603,
       0.00000000268,    -0.00000000065,
       0.00000000108,    -0.00000000674,
      -0.00000000162,     0.00000000009,
       0.00000000729,     0.00000000140,
       0.00000000002,    -0.00000000168,
       0.00000000099,    -0.00000000008,
      -0.00000000093,    -0.00000000032,
      -0.00000005943,     0.00000008178,
       0.00000002337,     0.00000002399,
      -0.00000000805,    -0.00000000694,
      -0.00000000315,     0.00000000261,
       0.00000000176,    -0.00000001393,
      -0.00000000314,    -0.00000000142,
       0.00000001391,     0.00000000124,
       0.00000000138,    -0.00000000309,
       0.00000000298,     0.00000000573,
       0.00000001356,    -0.00000000366,
      -0.00000000906,    -0.00000003356,
      -0.00000012465,     0.00000005066,
       0.00000000873,     0.00000002156,
       0.00000001901,     0.00000004805,
       0.00000000930,    -0.00000000377,
       0.00000000509,    -0.00000001208,
      -0.00000000228,    -0.00000000279,
       0.00000001240,     0.00000000448,
       0.00000000280,    -0.00000000230,
       0.00000000027,    -0.00000000311,
       0.00000113017,    -0.00000010643,
      -0.00000011102,    -0.00000045881,
      -0.00000009374,     0.00000003830,
       0.00000000760,     0.00000001226,
      -0.00000000372,    -0.00000004001,
      -0.00000001567,     0.00000000366,
       0.00000000121,     0.00000000318,
       0.00000000288,    -0.00000000069,
      -0.00000000024,    -0.00000000120,
      -0.00000000044,    -0.00000000088,
      -0.00000005062,     0.00000000764,
      -0.00000000096,     0.00000000031,
      -0.00000000035,    -0.00000000133,
      -0.00000000102,    -0.00000000047,
      -0.00000000441,     0.00000000707,
      -0.00000000603,    -0.00000000358,
      -0.00000000079,     0.00000000174,
       0.00000000559,     0.00000000171,
      -0.00000000021,    -0.00000000161,
       0.00000000241,     0.00000000298,
       0.00000000160,    -0.00000000054,
       0.00000038329,    -0.00000025062,
      -0.00000009223,    -0.00000013493,
      -0.00000002334,     0.00000001733,
       0.00000003544,     0.00000001778,
       0.00000000374,    -0.00000001071,
      -0.00000000144,     0.00000000000,
       0.00000003199,     0.00000000281,
      -0.00000000198,    -0.00000001262,
      -0.00000000273,     0.00000000114,
       0.00000000044,     0.00000000072,
       0.00000000923,     0.00000003320,
       0.00000001369,    -0.00000000439,
      -0.00000000114,    -0.00000000271,
      -0.00000000072,     0.00000000044,
       0.00000000165,     0.00000000041,
      -0.00000000084,     0.00000000153,
       0.00000000207,     0.00000000098,
       0.00000000419,     0.00000006647,
       0.00000416450,    -0.00000025971,
      -0.00000004623,    -0.00000003260,
      -0.00000003199,     0.00000004585,
      -0.00000005904,    -0.00000087460,
      -0.00000020565,     0.00000001222,
      -0.00001084158,     0.00000074560,
       0.00000016405,     0.00000280457,
       0.00000018808,     0.00000006877,
       0.00000017178,    -0.00000018229,
      -0.00000034023,     0.00000565257,
       0.00000100794,     0.00000018407,
       0.00000003783,    -0.00000018319,
       0.00001239841,     0.00000070524,
       0.00000037010,    -0.00000201476,
      -0.00000043773,    -0.00000009183,
      -0.00000019372,     0.00000044770,
       0.00000011612,     0.00000004361,
       0.00000001312,    -0.00000001870,
      -0.00000046671,    -0.00000035515,
      -0.00000008305,     0.00000014597,
       0.00000001658,     0.00000001163,
       0.00000065798,    -0.00000138123,
      -0.00000036124,    -0.00000019852,
      -0.00000003594,     0.00000007750,
      -0.00000106460,    -0.00000049821,
      -0.00000014156,     0.00000024810,
       0.00000006635,     0.00000003070,
      -0.00000001251,    -0.00000000053,
      -0.00000000224,     0.00000001480,
       0.00000000085,    -0.00000000051,
      -0.00000000013,    -0.00000000144,
       0.00000000153,     0.00000000000,
       0.00000015079,    -0.00000004582,
      -0.00000003079,    -0.00000003588,
      -0.00000000419,    -0.00000001894,
      -0.00000000385,     0.00000000278,
       0.00000000129,    -0.00000000003,
       0.00000000005,     0.00000000132,
      -0.00000013611,     0.00000002445,
       0.00000002121,     0.00000002948,
       0.00000000201,     0.00000001123,
       0.00000000245,    -0.00000000180,
      -0.00000000052,    -0.00000000060,
       0.00000000092,     0.00000000061,
       0.00000000121,    -0.00000000110,
       0.00000000111,     0.00000000115,
       0.00000000110,     0.00000001160,
      -0.00000000230,    -0.00000002760,
       0.00000003641,    -0.00000000311,
       0.00000000210,     0.00000002465,
      -0.00000000944,    -0.00000000268,
      -0.00000000261,     0.00000000783,
       0.00000000029,     0.00000000132,
      -0.00000002251,     0.00000006177,
       0.00000001977,     0.00000000446,
      -0.00000000256,    -0.00000000478,
      -0.00000000120,     0.00000000077,
      -0.00000000131,     0.00000000481,
       0.00000000131,     0.00000000052,
       0.00000000500,     0.00000000015,
       0.00000000003,    -0.00000000112,
       0.00000000158,    -0.00000000091,
      -0.00000000800,    -0.00000001523,
      -0.00000000521,     0.00000000288,
       0.00000000424,    -0.00000000056,
       0.00000000117,    -0.00000000325,
      -0.00000001480,     0.00000029664,
       0.00000006716,     0.00000000415,
      -0.00000001127,    -0.00000000060,
      -0.00000000007,     0.00000000263,
      -0.00000000237,     0.00000000450,
       0.00000000198,     0.00000000065,
      -0.00000000399,    -0.00000000194,
      -0.00000000059,     0.00000000181,
      -0.00000001727,     0.00000003669,
      -0.00000002467,     0.00000000424,
       0.00000000410,     0.00000000500,
       0.00000000050,    -0.00000000084,
       0.00000000425,     0.00000002466,
       0.00000000500,    -0.00000000409,
      -0.00000000084,    -0.00000000050,
      -0.00000000160,     0.00000000288,
       0.00000000137,    -0.00000000065,
       0.00000000013,    -0.00000000131,
       0.00000002733,     0.00000001609,
       0.00000000471,    -0.00000000775,
      -0.00000000084,    -0.00000000082,
      -0.00000000484,     0.00000000724,
       0.00000000140,    -0.00000000027,
      -0.00000000524,     0.00000000576,
       0.00000000002,     0.00000000232,
       0.00000000070,     0.00000000531,
       0.00000000137,    -0.00000000020,
      -0.00000000527,     0.00000000073,
       0.00000000019,     0.00000000141,
       0.00000000032,     0.00000000382,
       0.00000000114,    -0.00000000010,
      -0.00000000379,     0.00000000029,
       0.00000000010,     0.00000000115,
       0.00000001377,     0.00000004730,
       0.00000002260,    -0.00000001215,
       0.00000002147,    -0.00000000679,
      -0.00000000500,    -0.00000001015,
      -0.00000000230,     0.00000000191,
       0.00000004587,    -0.00000000748,
      -0.00000000533,    -0.00000001779,
      -0.00000000355,     0.00000000159,
      -0.00000000763,    -0.00000004598,
      -0.00000001785,     0.00000000530,
       0.00000000159,     0.00000000355,
       0.00000105097,    -0.00000011581,
      -0.00000010610,    -0.00000042599,
      -0.00000008690,     0.00000003511,
       0.00000000321,     0.00000004058,
       0.00000001719,    -0.00000000389,
      -0.00000000134,    -0.00000000358,
       0.00000000721,     0.00000000047,
      -0.00000000057,    -0.00000000284,
       0.00000000064,    -0.00000000797,
      -0.00000000318,     0.00000000035,
       0.00000118977,    -0.00000313308,
      -0.00000138328,    -0.00000025295,
       0.00000000387,     0.00000030821,
       0.00000004494,    -0.00000001066,
      -0.00000004890,     0.00000002323,
       0.00000001715,     0.00000002356,
       0.00000000558,    -0.00000000573,
      -0.00000000123,    -0.00000000082,
       0.00000235592,    -0.00000037903,
      -0.00000027050,    -0.00000091232,
      -0.00000017930,     0.00000007946,
       0.00000001431,     0.00000002284,
       0.00000000210,    -0.00000000182,
      -0.00000037866,    -0.00000235567,
      -0.00000091216,     0.00000027024,
       0.00000007940,     0.00000017927,
       0.00000002281,    -0.00000001433,
      -0.00000000182,    -0.00000000210,
       0.00000000005,    -0.00000000140,
      -0.00000000140,    -0.00000000006,
      -0.00000000815,     0.00000000286,
       0.00000000255,     0.00000000291,
       0.00000000419,    -0.00000000261,
      -0.00000000272,    -0.00000000413,
      -0.00000001755,    -0.00000011551,
      -0.00000005228,     0.00000003003,
       0.00000001336,     0.00000000967,
       0.00000002788,    -0.00000000695,
      -0.00000000791,    -0.00000001181,
      -0.00000000287,     0.00000000396,
      -0.00000000865,     0.00000002390,
       0.00000001229,     0.00000000084,
      -0.00000000137,    -0.00000000291,
       0.00000002364,     0.00000000871,
      -0.00000000013,    -0.00000001284,
      -0.00000000291,     0.00000000137,
       0.00000000541,     0.00000000041,
      -0.00000000030,    -0.00000000193,
       0.00000000592,    -0.00000000038,
      -0.00000000037,    -0.00000000244,
      -0.00000000055,    -0.00000000594,
      -0.00000000245,     0.00000000069,
      -0.00000011692,    -0.00000011858,
      -0.00000005142,     0.00000006493,
       0.00000001818,     0.00000001085,
      -0.00000002228,    -0.00000000416,
       0.00000000119,     0.00000001160,
       0.00000000298,    -0.00000000150,
       0.00000002679,    -0.00000000465,
      -0.00000000323,    -0.00000001067,
      -0.00000000214,     0.00000000091,
       0.00000000050,     0.00000000086,
      -0.00000000485,    -0.00000002663,
      -0.00000001067,     0.00000000326,
       0.00000000092,     0.00000000222,
       0.00000000086,    -0.00000000050,
       0.00000001859,     0.00000000969,
       0.00000000332,     0.00000001179,
       0.00000000455,    -0.00000000174,
      -0.00000000046,    -0.00000000100,
       0.00000000019,    -0.00000000249,
      -0.00000000254,    -0.00000000115,
       0.00000002830,    -0.00000000447,
      -0.00000000083,    -0.00000000501,
      -0.00000000195,     0.00000000058,
       0.00000000091,    -0.00000002778,
      -0.00000000353,    -0.00000000034,
      -0.00000000010,     0.00000000164,
       0.00000001508,     0.00000000205,
       0.00000000042,    -0.00000000312,
       0.00000003699,    -0.00000000717,
      -0.00000000042,    -0.00000000233,
      -0.00000000075,    -0.00000000124,
      -0.00000000129,     0.00000000077,
       0.00000000485,    -0.00000000145,
      -0.00000000084,    -0.00000000289,
      -0.00000000098,     0.00000000039,
      -0.00000003268,     0.00000000860,
       0.00000000184,     0.00000000693,
       0.00000000155,     0.00000000302,
       0.00000000372,    -0.00000000216,
       0.00000043750,    -0.00000007279,
      -0.00000005067,    -0.00000016964,
      -0.00000003301,     0.00000001445,
      -0.00000001455,    -0.00000008801,
      -0.00000003456,     0.00000001031,
       0.00000000299,     0.00000000681,
       0.00000000088,    -0.00000000054,
      -0.00000000132,     0.00000000547,
       0.00000000178,    -0.00000000031,
       0.00000000200,     0.00000000200,
      -0.00000011278,     0.00000030068,
       0.00000009315,     0.00000001121,
      -0.00000001052,    -0.00000000348,
      -0.00000000002,     0.00000000332,
      -0.00000000067,     0.00000000087,
       0.00000000097,    -0.00000000035,
      -0.00000354395,     0.00000128938,
       0.00000060612,     0.00000137898,
       0.00000027379,    -0.00000014803,
      -0.00000002420,    -0.00000003535,
       0.00000003403,     0.00000024129,
       0.00000009338,    -0.00000002626,
      -0.00000000773,    -0.00000001819,
      -0.00000000232,     0.00000000140,
      -0.00000020238,     0.00000065511,
       0.00000018890,     0.00000001093,
      -0.00000000596,    -0.00000002722,
      -0.00000000238,     0.00000000137,
       0.00000065861,     0.00000020080,
       0.00000000974,    -0.00000019059,
      -0.00000002761,     0.00000000636,
       0.00000000145,     0.00000000244,
       0.00000000043,     0.00000000149,
       0.00000000149,    -0.00000000043,
       0.00000000091,    -0.00000000017,
      -0.00000000153,    -0.00000002634,
      -0.00000000489,     0.00000000039,
       0.00000000180,     0.00000000218,
      -0.00000000361,     0.00000000734,
       0.00000000384,     0.00000000004,
       0.00000000753,     0.00000000336,
      -0.00000000013,    -0.00000000388,
      -0.00000000080,    -0.00000000038,
      -0.00000000151,     0.00000000093,
       0.00000000053,     0.00000000056,
       0.00000000109,     0.00000000148,
       0.00000000051,    -0.00000000059,
      -0.00000002513,     0.00000005364,
       0.00000002783,     0.00000000095,
       0.00000000569,     0.00000000225,
      -0.00000000022,    -0.00000000296,
      -0.00000000076,     0.00000000046,
       0.00000001223,    -0.00000000834,
      -0.00000000507,    -0.00000000411,
      -0.00000000057,     0.00000000135,
      -0.00000000818,    -0.00000001233,
      -0.00000000410,     0.00000000504,
       0.00000000143,     0.00000000041,
      -0.00000001191,     0.00000000506,
       0.00000000214,     0.00000000270,
       0.00000000085,    -0.00000000127,
       0.00000000390,    -0.00000000110,
      -0.00000000020,    -0.00000000127,
      -0.00000000097,    -0.00000000398,
      -0.00000000124,     0.00000000032,
       0.00000012841,    -0.00000008484,
      -0.00000005148,    -0.00000004390,
      -0.00000000620,     0.00000001447,
      -0.00000000919,    -0.00000001195,
      -0.00000000389,     0.00000000541,
       0.00000000154,     0.00000000053,
      -0.00000002592,     0.00000000085,
       0.00000000203,     0.00000001074,
       0.00000000225,    -0.00000000085,
       0.00000000053,     0.00000002594,
       0.00000001085,    -0.00000000190,
      -0.00000000079,    -0.00000000224,
       0.00000003025,    -0.00000000212,
      -0.00000000259,    -0.00000001210,
      -0.00000000053,    -0.00000000721,
      -0.00000000298,     0.00000000060,
      -0.00000025587,    -0.00000001023,
       0.00000001326,     0.00000010900,
       0.00000002298,    -0.00000000630,
       0.00000000245,     0.00000002944,
       0.00000001203,    -0.00000000258,
      -0.00000000086,    -0.00000000247,
      -0.00000000447,     0.00000001002,
       0.00000000301,     0.00000000080,
       0.00000001064,     0.00000000489,
       0.00000000090,    -0.00000000339,
      -0.00000000612,    -0.00000000157,
      -0.00000000010,     0.00000000178,
       0.00000000163,     0.00000000042,
       0.00000000058,     0.00000000057,
      -0.00000000150,     0.00000000085,
       0.00000001039,     0.00000001407,
       0.00000000326,    -0.00000000180,
      -0.00000001947,     0.00000007413,
       0.00000002110,     0.00000000002,
       0.00000000868,     0.00000000200,
      -0.00000000005,    -0.00000000246,
       0.00000001004,    -0.00000000279,
      -0.00000000111,    -0.00000000374,
      -0.00000000075,     0.00000000033,
       0.00000000048,     0.00000000933,
       0.00000000379,    -0.00000000114,
      -0.00000000033,    -0.00000000074,
       0.00000015818,     0.00000002439,
       0.00000001187,    -0.00000000997,
      -0.00000000255,     0.00000000670,
       0.00000000074,     0.00000000056,
       0.00000000131,     0.00000000017,
       0.00000000021,    -0.00000000139,
       0.00000039940,    -0.00000060956,
      -0.00000009304,    -0.00000004184,
      -0.00000002171,    -0.00000000315,
       0.00000000004,     0.00000000684,
       0.00000000095,    -0.00000000026,
       0.00000034244,     0.00000006083,
       0.00000003355,    -0.00000001356,
      -0.00000000006,    -0.00000000312,
       0.00000005823,    -0.00000035837,
      -0.00000001975,    -0.00000003173,
      -0.00000000259,     0.00000000129,
      -0.00000000147,     0.00000000286,
       0.00000000147,     0.00000000011,
       0.00000000291,     0.00000000147,
       0.00000000004,    -0.00000000150,
      -0.00000000073,    -0.00000000029,
      -0.00000000095,     0.00000000057,
       0.00000000066,     0.00000000093,
      -0.00000002210,     0.00000004115,
       0.00000002155,     0.00000000188,
       0.00000000329,     0.00000000139,
      -0.00000000006,    -0.00000000170,
       0.00000000799,    -0.00000000543,
      -0.00000000322,    -0.00000000271,
      -0.00000000038,     0.00000000085,
      -0.00000000530,    -0.00000000806,
      -0.00000000273,     0.00000000317,
       0.00000000088,     0.00000000034,
      -0.00000000955,     0.00000000641,
       0.00000000153,     0.00000000204,
       0.00000000068,    -0.00000000091,
       0.00000000199,    -0.00000000025,
      -0.00000000018,    -0.00000000201,
       0.00000011861,    -0.00000007893,
      -0.00000004656,    -0.00000004010,
      -0.00000000679,    -0.00000000905,
      -0.00000000299,     0.00000000395,
       0.00000000109,     0.00000000039,
      -0.00000001776,     0.00000000199,
       0.00000000169,     0.00000000717,
       0.00000000150,    -0.00000000054,
       0.00000000158,     0.00000001773,
       0.00000000726,    -0.00000000151,
      -0.00000000050,    -0.00000000151,
       0.00000003139,    -0.00000000604,
      -0.00000000391,    -0.00000001146,
      -0.00000000126,    -0.00000000624,
      -0.00000000216,     0.00000000049,
      -0.00000000001,     0.00000000155,
       0.00000000156,    -0.00000000001,
      -0.00000024776,     0.00000002888,
       0.00000002395,     0.00000010068,
       0.00000002069,    -0.00000000741,
       0.00000000222,     0.00000002200,
       0.00000000896,    -0.00000000200,
      -0.00000000063,    -0.00000000184,
      -0.00000000082,    -0.00000000160,
      -0.00000000052,     0.00000000059,
      -0.00000000073,     0.00000000166,
      -0.00000002210,    -0.00000002062,
      -0.00000000133,     0.00000000125,
       0.00000000360,     0.00000000038,
       0.00000000012,    -0.00000000572,
      -0.00000000095,     0.00000000013,
       0.00000000014,     0.00000000096,
      -0.00000000046,     0.00000000285,
       0.00000000016,    -0.00000000103,
       0.00000000985,    -0.00000000541,
      -0.00000000047,     0.00000000098,
      -0.00000000054,     0.00000000205,
      -0.00000000205,    -0.00000000022,
      -0.00000012279,    -0.00000001625,
       0.00000000026,    -0.00000000508,
      -0.00000030600,     0.00000017558,
      -0.00000000177,    -0.00000000217,
      -0.00000026642,    -0.00000001417,
      -0.00000000109,     0.00000000100,
      -0.00000001163,     0.00000026688,
       0.00000000094,     0.00000000038,
      -0.00000000109,     0.00000000204,
       0.00000000103,     0.00000000011,
       0.00000000204,     0.00000000110,
       0.00000000008,    -0.00000000105,
      -0.00000000066,     0.00000000041,
       0.00000000046,     0.00000000065,
      -0.00000001919,     0.00000003466,
       0.00000001805,     0.00000000195,
       0.00000000246,     0.00000000110,
       0.00000000000,    -0.00000000128,
       0.00000000563,    -0.00000000386,
      -0.00000000226,    -0.00000000192,
      -0.00000000373,    -0.00000000568,
      -0.00000000194,     0.00000000220,
      -0.00000000824,     0.00000000559,
       0.00000000111,     0.00000000151,
       0.00000000050,    -0.00000000065,
       0.00000000127,    -0.00000000022,
      -0.00000000012,    -0.00000000127,
       0.00000009599,    -0.00000006440,
      -0.00000003753,    -0.00000003267,
      -0.00000000496,    -0.00000000668,
      -0.00000000224,     0.00000000286,
       0.00000000079,     0.00000000031,
      -0.00000001141,     0.00000000166,
       0.00000000117,     0.00000000457,
       0.00000000095,    -0.00000000034,
       0.00000000123,     0.00000001133,
       0.00000000463,    -0.00000000096,
      -0.00000000030,    -0.00000000097,
       0.00000002003,    -0.00000000287,
      -0.00000000201,    -0.00000000805,
      -0.00000000040,    -0.00000000347,
      -0.00000000143,     0.00000000032,
      -0.00000018575,     0.00000002635,
       0.00000001867,     0.00000007468,
       0.00000001539,    -0.00000000543,
       0.00000000182,     0.00000001470,
       0.00000000594,    -0.00000000137,
      -0.00000000040,    -0.00000000122,
      -0.00000000069,    -0.00000000193,
      -0.00000000063,     0.00000000050,
      -0.00000000108,     0.00000000162,
       0.00000000137,     0.00000001086,
       0.00000000174,    -0.00000000002,
      -0.00000001301,    -0.00000002494,
      -0.00000000186,     0.00000000124,
       0.00000000070,     0.00000000044,
      -0.00000000086,    -0.00000000036,
      -0.00000000067,    -0.00000000131,
      -0.00000000342,    -0.00000000055,
      -0.00000000055,     0.00000000342,
       0.00000000031,    -0.00000000243,
      -0.00000000049,    -0.00000000104,
      -0.00000000106,     0.00000000058,
      -0.00000002859,    -0.00000000535,
      -0.00000000025,     0.00000000167,
       0.00000000186,    -0.00000000047,
       0.00000001045,     0.00000006566,
      -0.00000000279,     0.00000000039,
       0.00000800927,     0.00000201275,
       0.00000019666,     0.00000037114,
      -0.00000001852,     0.00000001654,
       0.00000001464,    -0.00000005725,
       0.00000000211,    -0.00000000108,
       0.00000000289,    -0.00000000500,
       0.00000000135,     0.00000000043,
      -0.00000000463,    -0.00000000292,
       0.00000000043,    -0.00000000132,
      -0.00000000059,     0.00000000109,
       0.00000000108,     0.00000000060,
      -0.00000001185,     0.00000002134,
       0.00000000136,     0.00000000064,
       0.00000000254,    -0.00000000184,
      -0.00000000102,    -0.00000000086,
      -0.00000000173,    -0.00000000256,
      -0.00000000089,     0.00000000098,
       0.00000000052,     0.00000000071,
       0.00000004953,    -0.00000003689,
      -0.00000001955,    -0.00000001692,
      -0.00000000237,    -0.00000000315,
      -0.00000000106,     0.00000000133,
      -0.00000000412,     0.00000000098,
       0.00000000051,     0.00000000162,
       0.00000000065,     0.00000000402,
       0.00000000163,    -0.00000000034,
       0.00000000839,    -0.00000000185,
      -0.00000000022,    -0.00000000127,
      -0.00000007756,     0.00000002198,
       0.00000000948,     0.00000003014,
       0.00000000104,     0.00000000543,
       0.00000000217,    -0.00000000055,
      -0.00000000035,    -0.00000000137,
      -0.00000000078,     0.00000000101,
      -0.00000001035,    -0.00000002946,
      -0.00000000140,     0.00000000083,
      -0.00000000028,     0.00000000123,
       0.00000000081,    -0.00000000049,
       0.00000000370,     0.00000002076,
      -0.00000000120,    -0.00000000015,
       0.00000000117,     0.00000000072,
       0.00000005178,     0.00000001723,
      -0.00000014879,     0.00000041491,
      -0.00000005712,    -0.00000003084,
       0.00000000376,     0.00000000126,
      -0.00000000115,     0.00000000131,
       0.00000000139,     0.00000000113,
      -0.00000000040,     0.00000000077,
      -0.00000000851,     0.00000001562,
       0.00000000097,     0.00000000046,
       0.00000000162,    -0.00000000122,
      -0.00000000065,    -0.00000000054,
      -0.00000000113,    -0.00000000162,
      -0.00000000056,     0.00000000062,
       0.00000003269,    -0.00000002410,
      -0.00000001305,    -0.00000001114,
      -0.00000000156,    -0.00000000202,
      -0.00000000068,     0.00000000086,
      -0.00000000235,     0.00000000071,
       0.00000000032,     0.00000000092,
       0.00000000044,     0.00000000225,
       0.00000000091,    -0.00000000019,
      -0.00000004615,     0.00000001257,
       0.00000000564,     0.00000001804,
       0.00000000074,     0.00000000309,
       0.00000000122,    -0.00000000033,
      -0.00000000024,    -0.00000000098,
      -0.00000000055,     0.00000000073,
      -0.00000000670,    -0.00000001591,
      -0.00000000101,     0.00000000060,
       0.00000000158,     0.00000000053,
       0.00000000053,    -0.00000000158,
       0.00000000099,    -0.00000000079,
      -0.00000000072,    -0.00000000099,
       0.00000002068,    -0.00000001599,
      -0.00000000101,    -0.00000000125,
      -0.00000000130,     0.00000000049,
       0.00000000029,     0.00000000122,
      -0.00000002623,     0.00000000885,
       0.00000000051,     0.00000000170,
       0.00000001261,    -0.00000001031,
      -0.00000000063,    -0.00000000075,
       0.00000000034,     0.00000000091,
       0.00000002532,     0.00000001804,
       0.00000000123,     0.00000000162,
      -0.00000000076,     0.00000000071,
       0.00000000166,    -0.00000000132,
       0.00000000074,     0.00000000078,
      -0.00000000081,     0.00000000151,
       0.00000000150,     0.00000000083,
      -0.00000001010,     0.00000002910,
       0.00000000185,     0.00000000085,
       0.00000000224,    -0.00000000300,
      -0.00000000154,    -0.00000000132,
      -0.00000000288,    -0.00000000227,
      -0.00000000134,     0.00000000150,
      -0.00000000638,     0.00000000442,
       0.00000000078,     0.00000000106,
       0.00000003947,    -0.00000005485,
      -0.00000002865,    -0.00000002096,
      -0.00000000374,    -0.00000000308,
      -0.00000000157,     0.00000000200,
       0.00000000195,     0.00000000273,
       0.00000000077,     0.00000000278,
       0.00000000234,    -0.00000000203,
       0.00000000281,    -0.00000000058,
       0.00000000112,     0.00000000014,
       0.00000000015,    -0.00000000115,
       0.00000001343,    -0.00000000239,
      -0.00000000030,    -0.00000000215,
      -0.00000000087,     0.00000000021,
       0.00000004188,     0.00000004773,
       0.00000001556,     0.00000003553,
       0.00000000258,     0.00000000014,
       0.00000000367,    -0.00000000088,
      -0.00000004381,    -0.00000000748,
      -0.00000000025,     0.00000000266,
      -0.00000000673,     0.00000004443,
       0.00000000265,     0.00000000025,
       0.00000000897,     0.00000000115,
       0.00000000042,    -0.00000000392,
      -0.00000000427,    -0.00000000038,
      -0.00000000039,     0.00000000416,
      -0.00000075035,    -0.00000011857,
      -0.00000001122,     0.00000005283,
      -0.00000000631,     0.00000004508,
       0.00000000346,     0.00000000073,
       0.00000016899,     0.00000001895,
       0.00000000006,    -0.00000000634,
       0.00000001836,    -0.00000016892,
      -0.00000000624,    -0.00000000004,
      -0.00000004172,    -0.00000000388,
      -0.00000000138,     0.00000001663,
       0.00000000958,     0.00000000046,
       0.00000000054,    -0.00000000908,
       0.00000245979,     0.00000028838,
       0.00000000305,    -0.00000009069,
       0.00000001367,    -0.00000017386,
      -0.00000000682,    -0.00000000006,
      -0.00000045010,    -0.00000005660,
       0.00000000044,     0.00000000457,
      -0.00000005636,     0.00000044491,
       0.00000000419,    -0.00000000050,
       0.00000000162,    -0.00000000031,
      -0.00000000031,    -0.00000000166,
       0.00000012071,     0.00000000692,
       0.00000000260,    -0.00000004431,
       0.00000000090,     0.00000000031,
      -0.00000000003,    -0.00000000236,
      -0.00000502495,    -0.00000127738,
      -0.00000003126,    -0.00000000287,
      -0.00000000760,     0.00000037619,
       0.00000000661,    -0.00000000270,
       0.00000068020,     0.00000022418,
       0.00000001517,     0.00000002175,
      -0.00000000142,     0.00000000144,
       0.00000022317,    -0.00000066303,
       0.00000002244,    -0.00000001511,
       0.00000000144,     0.00000000140,
      -0.00000000955,     0.00000000024,
       0.00000000026,     0.00000000968,
       0.00000037628,     0.00000009753,
       0.00000000923,     0.00000001765,
      -0.00000010656,     0.00000042118,
      -0.00000001936,     0.00000001006,
      -0.00000000100,    -0.00000000101,
      -0.00000000667,    -0.00000000438,
       0.00000000444,    -0.00000000715,
      -0.00000000073,     0.00000000050,
       0.00000000052,     0.00000000071,
       0.00000079766,     0.00000020544,
       0.00000001920,     0.00000003706,
       0.00000003668,    -0.00000014701,
       0.00000000686,    -0.00000000358,
       0.00000000900,     0.00000000368,
      -0.00000000359,     0.00000000872,
      -0.00000006845,    -0.00000000723,
      -0.00000000018,     0.00000000198,
       0.00000003672,     0.00000000975,
       0.00000000094,     0.00000000171,
      -0.00000000974,     0.00000003687,
      -0.00000000172,     0.00000000094,
       0.00000000160,    -0.00000000105,
       0.00000000106,     0.00000000161,
      -0.00000005742,    -0.00000000727,
      -0.00000000155,    -0.00000000016,
       0.00000000016,    -0.00000000156,
      -0.00000001106,    -0.00000000828,
      -0.00000000291,     0.00000000415,
      -0.00000000468,     0.00000000000,
      -0.00000000010,    -0.00000000464,
      -0.00000000137,    -0.00000000001,
       0.00000000002,    -0.00000000138,
       0.00000026705,     0.00000006712,
      -0.00000000407,     0.00000001597,
      -0.00000000074,     0.00000000038,
      -0.00000000184,    -0.00000000006,
       0.00000000009,    -0.00000000212,
       0.00672759309,     0.00171618598,
       0.00016118288,     0.00031087081,
      -0.00001617777,     0.00001609639,
      -0.00000070267,    -0.00000068947,
       0.00000002510,    -0.00000001749,
      -0.00005392286,     0.00021128940,
      -0.00000976155,     0.00000506147,
      -0.00000050540,    -0.00000050809,
       0.00000002165,    -0.00000002206,
       0.00000000055,     0.00000000078,
       0.00000228374,    -0.00000082957,
       0.00000002813,    -0.00000004151,
       0.00000000303,     0.00000000145,
      -0.00000054143,     0.00000299030,
      -0.00000004384,    -0.00000000182,
      -0.00000000155,    -0.00000000278,
       0.00000046346,     0.00000009843,
      -0.00000000188,     0.00000000175,
      -0.00000000203,     0.00000000091,
       0.00000009875,    -0.00000046282,
       0.00000000179,     0.00000000187,
       0.00000000091,     0.00000000203,
      -0.00000133669,    -0.00000098593,
       0.00000002273,    -0.00000017937,
       0.00000002219,    -0.00000001072,
      -0.00000004205,     0.00000003639,
      -0.00000000625,    -0.00000000076,
      -0.00000000032,    -0.00000000076,
      -0.00000010544,    -0.00000021117,
       0.00000000485,    -0.00000001281,
       0.00000021107,    -0.00000010553,
       0.00000001280,     0.00000000484,
       0.00000043073,     0.00000009157,
      -0.00000000175,     0.00000000166,
      -0.00000000188,     0.00000000084,
      -0.00000009145,     0.00000043074,
      -0.00000000166,    -0.00000000175,
      -0.00000000084,    -0.00000000188,
       0.00000000703,     0.00000000046,
      -0.00000000045,     0.00000000704,
      -0.00000000164,     0.00000000122,
      -0.00000000121,    -0.00000000164,
      -0.00000001051,    -0.00000000844,
       0.00000000022,    -0.00000000146,
       0.00000000844,    -0.00000001051,
       0.00000000146,     0.00000000022,
       0.00000000093,     0.00000000165,
      -0.00000000165,     0.00000000093,
       0.00000002641,    -0.00000000035,
       0.00000000005,     0.00000000170,
      -0.00000002633,    -0.00000000120,
       0.00000000009,    -0.00000000161,
       0.00000000137,     0.00000000001,
       0.00000000008,     0.00000000115,
      -0.00000000129,    -0.00000000007,
       0.00000000002,    -0.00000000121,
       0.00000000081,     0.00000000057,
       0.00000004539,    -0.00000000283,
      -0.00000000266,    -0.00000001768,
       0.00000000025,     0.00000000301,
       0.00000000117,    -0.00000000019,
       0.00000000062,     0.00000000078,
      -0.00000004143,     0.00000000040,
      -0.00000000001,    -0.00000000258,
      -0.00000000101,     0.00000000016,
       0.00000000239,    -0.00000000012,
      -0.00000000016,    -0.00000000094,
       0.00000000022,     0.00000000210,
       0.00000001904,     0.00000000470,
      -0.00000000032,     0.00000000111,
      -0.00000000209,     0.00000000000,
      -0.00000000007,    -0.00000000199,
       0.00000000070,    -0.00000000101,
      -0.00000000077,    -0.00000000036,
       0.00000000101,     0.00000000070,
       0.00000000036,    -0.00000000077,
       0.00000000965,    -0.00000001543,
       0.00000000109,     0.00000000083,
      -0.00000000016,    -0.00000000120,
       0.00000007502,    -0.00000000769,
      -0.00000000613,    -0.00000002874,
       0.00000000061,     0.00000000518,
       0.00000000197,    -0.00000000044,
       0.00000000087,     0.00000000113,
      -0.00000006287,     0.00000000453,
       0.00000000527,     0.00000002403,
      -0.00000000027,    -0.00000000404,
      -0.00000000155,     0.00000000032,
       0.00000000408,    -0.00000000038,
      -0.00000000036,    -0.00000000156,
       0.00000000049,     0.00000000369,
       0.00000000138,    -0.00000000034,
       0.00000002767,     0.00000000483,
      -0.00000000037,     0.00000000166,
      -0.00000000328,     0.00000000021,
       0.00000000028,     0.00000000124,
      -0.00000000031,    -0.00000000314,
      -0.00000000120,     0.00000000028,
       0.00000000139,     0.00000000026,
      -0.00000000023,     0.00000000135,
      -0.00000000186,     0.00000000120,
       0.00000000113,     0.00000000146,
       0.00000000058,    -0.00000000052,
      -0.00000000120,    -0.00000000186,
      -0.00000000146,     0.00000000113,
       0.00000000052,     0.00000000058,
       0.00000001239,    -0.00000001769,
       0.00000000133,     0.00000000116,
      -0.00000001267,     0.00000000169,
      -0.00000000032,    -0.00000000201,
       0.00000011787,    -0.00000001585,
      -0.00000001182,    -0.00000004458,
       0.00000000125,     0.00000000859,
       0.00000000324,    -0.00000000088,
       0.00000000038,    -0.00000000098,
       0.00000000113,     0.00000000157,
       0.00000000049,    -0.00000000059,
       0.00000000020,     0.00000000138,
      -0.00000009119,     0.00000001350,
       0.00000000970,     0.00000003421,
      -0.00000000089,    -0.00000000611,
      -0.00000000230,     0.00000000062,
       0.00000000677,    -0.00000000087,
      -0.00000000069,    -0.00000000254,
       0.00000000097,     0.00000000624,
       0.00000000233,    -0.00000000068,
       0.00000003918,     0.00000000359,
      -0.00000000034,     0.00000000244,
       0.00000000092,    -0.00000000011,
      -0.00000000500,     0.00000000070,
       0.00000000055,     0.00000000187,
      -0.00000000081,    -0.00000000482,
      -0.00000000181,     0.00000000055,
       0.00000000205,     0.00000000024,
      -0.00000000018,     0.00000000199,
       0.00000000023,     0.00000000099,
       0.00000000099,    -0.00000000023,
      -0.00000000069,    -0.00000000233,
      -0.00000000134,     0.00000000093,
       0.00000000111,    -0.00000000041,
      -0.00000000065,    -0.00000000077,
       0.00000001408,    -0.00000001686,
       0.00000000138,     0.00000000146,
       0.00000000040,    -0.00000000069,
       0.00000000104,    -0.00000000030,
      -0.00000001870,     0.00000000265,
      -0.00000000056,    -0.00000000319,
      -0.00000000116,     0.00000000040,
       0.00000017236,    -0.00000002721,
      -0.00000001966,    -0.00000006451,
       0.00000000224,     0.00000001350,
       0.00000000505,    -0.00000000157,
      -0.00000000047,    -0.00000000097,
       0.00000000063,    -0.00000000098,
       0.00000000130,     0.00000000201,
       0.00000000065,    -0.00000000069,
       0.00000001083,    -0.00000000277,
       0.00000000052,     0.00000000202,
      -0.00000000127,     0.00000000017,
      -0.00000000020,    -0.00000000107,
      -0.00000012514,     0.00000003112,
       0.00000001676,     0.00000004580,
      -0.00000000224,    -0.00000000893,
      -0.00000000327,     0.00000000117,
       0.00000001078,    -0.00000000168,
      -0.00000000126,    -0.00000000399,
      -0.00000000077,     0.00000000039,
       0.00000000175,     0.00000001010,
       0.00000000374,    -0.00000000121,
      -0.00000000037,    -0.00000000071,
       0.00000000085,    -0.00000000022,
       0.00000000021,     0.00000000081,
       0.00000005445,    -0.00000000046,
      -0.00000000471,    -0.00000001854,
      -0.00000000014,     0.00000000362,
       0.00000000126,    -0.00000000027,
      -0.00000000746,     0.00000000181,
       0.00000000101,     0.00000000272,
      -0.00000000194,    -0.00000000720,
      -0.00000000265,     0.00000000101,
      -0.00000001962,    -0.00000000570,
       0.00000000047,    -0.00000000132,
       0.00000000305,     0.00000000005,
      -0.00000000024,    -0.00000000107,
       0.00000000004,     0.00000000297,
       0.00000000111,    -0.00000000025,
      -0.00000000108,    -0.00000000032,
       0.00000000031,    -0.00000000109,
       0.00000000698,     0.00000002916,
       0.00000001376,    -0.00000000678,
      -0.00000000343,     0.00000000087,
       0.00000000082,     0.00000000162,
      -0.00000000963,     0.00000000139,
       0.00000000104,     0.00000000370,
       0.00000000072,    -0.00000000032,
       0.00000000141,     0.00000000969,
       0.00000000375,    -0.00000000106,
      -0.00000000032,    -0.00000000072,
      -0.00000001886,    -0.00000013339,
      -0.00000006475,     0.00000002130,
       0.00000000954,    -0.00000000259,
      -0.00000000239,    -0.00000000449,
      -0.00000000104,     0.00000000088,
       0.00000002828,    -0.00000000648,
      -0.00000000763,    -0.00000001421,
      -0.00000000348,     0.00000000331,
       0.00000000086,     0.00000000052,
       0.00000000418,     0.00000001259,
       0.00000000812,    -0.00000000591,
      -0.00000000279,    -0.00000000229,
      -0.00000000037,     0.00000000076,
       0.00000001096,    -0.00000001040,
       0.00000000106,     0.00000000154,
       0.00000000125,    -0.00000000034,
       0.00000000095,    -0.00000000009,
       0.00000000011,     0.00000000095,
      -0.00000002430,     0.00000000491,
       0.00000000302,     0.00000000905,
      -0.00000000089,    -0.00000000473,
      -0.00000000173,     0.00000000061,
       0.00000022657,    -0.00000004004,
      -0.00000002843,    -0.00000008415,
      -0.00000001577,     0.00000000831,
       0.00000000356,     0.00000001977,
       0.00000000735,    -0.00000000249,
      -0.00000000073,    -0.00000000138,
       0.00000000085,    -0.00000000070,
       0.00000000125,     0.00000000227,
       0.00000000077,    -0.00000000067,
       0.00000001751,    -0.00000000597,
       0.00000000135,     0.00000000377,
       0.00000000100,    -0.00000000048,
      -0.00000000192,     0.00000000030,
      -0.00000000032,    -0.00000000169,
      -0.00000016311,     0.00000006070,
       0.00000002678,     0.00000005842,
      -0.00000000525,    -0.00000001259,
      -0.00000000444,     0.00000000217,
       0.00000000053,     0.00000000078,
       0.00000001633,    -0.00000000282,
      -0.00000000207,    -0.00000000602,
      -0.00000000113,     0.00000000062,
       0.00000000284,     0.00000001546,
       0.00000000573,    -0.00000000198,
      -0.00000000057,    -0.00000000106,
      -0.00000003975,     0.00000000339,
      -0.00000000036,    -0.00000000240,
      -0.00000000140,    -0.00000000188,
      -0.00000000051,     0.00000000095,
       0.00000000194,    -0.00000000139,
      -0.00000000095,    -0.00000000051,
      -0.00000000013,    -0.00000000127,
       0.00000000126,    -0.00000000051,
       0.00000000051,     0.00000000122,
       0.00000007760,    -0.00000000930,
      -0.00000000932,    -0.00000002360,
       0.00000000044,     0.00000000578,
       0.00000000180,    -0.00000000065,
      -0.00000001098,     0.00000000419,
       0.00000000190,     0.00000000385,
       0.00000000065,    -0.00000000044,
      -0.00000000437,    -0.00000001061,
      -0.00000000377,     0.00000000190,
       0.00000000046,     0.00000000068,
      -0.00000003384,    -0.00000000687,
       0.00000000065,    -0.00000000271,
       0.00000000498,    -0.00000000051,
      -0.00000000060,    -0.00000000162,
       0.00000000063,     0.00000000485,
       0.00000000168,    -0.00000000062,
       0.00000001405,     0.00000000978,
      -0.00000000106,     0.00000000140,
      -0.00000000245,    -0.00000000051,
       0.00000000045,    -0.00000000244,
       0.00000000126,    -0.00000000053,
       0.00000000157,     0.00000000121,
      -0.00000000119,     0.00000000157,
      -0.00000040537,     0.00000006065,
       0.00000004502,     0.00000015779,
       0.00000003087,    -0.00000001379,
      -0.00000001268,    -0.00000008358,
      -0.00000003203,     0.00000000921,
       0.00000000281,     0.00000000627,
       0.00000000079,    -0.00000000053,
       0.00000000167,     0.00000000013,
      -0.00000000188,    -0.00000000107,
      -0.00000000030,     0.00000000091,
       0.00000003426,     0.00000010125,
       0.00000004501,    -0.00000002815,
      -0.00000000953,    -0.00000000988,
       0.00000000418,    -0.00000000010,
      -0.00000000063,    -0.00000000192,
       0.00000000063,     0.00000000057,
       0.00000354980,    -0.00000034592,
      -0.00000029205,    -0.00000138392,
      -0.00000027463,     0.00000009315,
       0.00000001770,     0.00000003547,
       0.00000003546,     0.00000022487,
       0.00000008718,    -0.00000002588,
      -0.00000000784,    -0.00000001692,
      -0.00000000213,     0.00000000145,
       0.00000005033,     0.00000023231,
       0.00000010884,    -0.00000005144,
      -0.00000001910,    -0.00000002561,
      -0.00000000376,     0.00000000433,
       0.00000000070,     0.00000000038,
      -0.00000022767,     0.00000005355,
       0.00000005239,     0.00000010753,
       0.00000002540,    -0.00000001924,
      -0.00000000435,    -0.00000000374,
      -0.00000000038,     0.00000000070,
       0.00000000165,    -0.00000000029,
      -0.00000000034,    -0.00000000082,
      -0.00000002609,     0.00000000484,
       0.00000000345,     0.00000000966,
      -0.00000000128,    -0.00000000623,
      -0.00000000228,     0.00000000087,
       0.00000023529,    -0.00000005606,
      -0.00000003674,    -0.00000008543,
      -0.00000001543,     0.00000001005,
       0.00000000488,     0.00000002557,
       0.00000000943,    -0.00000000340,
      -0.00000000100,    -0.00000000174,
       0.00000000193,     0.00000000342,
       0.00000000145,    -0.00000000123,
      -0.00000000261,     0.00000000295,
       0.00000000159,     0.00000000109,
       0.00000001279,    -0.00000000894,
       0.00000000288,     0.00000000412,
       0.00000000131,    -0.00000000091,
      -0.00000000276,     0.00000000048,
       0.00000000036,     0.00000000095,
      -0.00000000048,    -0.00000000253,
      -0.00000000098,     0.00000000035,
      -0.00000019252,     0.00000012715,
       0.00000004507,     0.00000006361,
       0.00000001052,    -0.00000000900,
      -0.00000001232,    -0.00000001823,
      -0.00000000603,     0.00000000432,
       0.00000000094,     0.00000000102,
       0.00000002372,    -0.00000000473,
      -0.00000000341,    -0.00000000853,
      -0.00000000158,     0.00000000089,
       0.00000000454,     0.00000002161,
       0.00000000801,    -0.00000000305,
      -0.00000000084,    -0.00000000152,
      -0.00000000082,    -0.00000000311,
       0.00000000209,    -0.00000000139,
      -0.00000000044,    -0.00000000065,
       0.00000000139,     0.00000000208,
       0.00000000079,    -0.00000000050,
       0.00000019161,    -0.00000004539,
      -0.00000003219,    -0.00000004549,
       0.00000000256,     0.00000001231,
       0.00000000325,    -0.00000000191,
      -0.00000000053,    -0.00000000059,
      -0.00000001772,     0.00000001143,
       0.00000000415,     0.00000000586,
       0.00000000091,    -0.00000000081,
      -0.00000001177,    -0.00000001708,
      -0.00000000574,     0.00000000417,
       0.00000000085,     0.00000000097,
       0.00000001894,     0.00000003681,
      -0.00000000119,    -0.00000000026,
       0.00000002445,    -0.00000000377,
      -0.00000000393,    -0.00000000502,
      -0.00000000052,     0.00000000085,
       0.00000000376,     0.00000002446,
       0.00000000502,    -0.00000000393,
      -0.00000000085,    -0.00000000052,
      -0.00000000109,     0.00000000393,
      -0.00000000200,     0.00000000051,
      -0.00000000040,    -0.00000000216,
      -0.00000011785,    -0.00000001086,
       0.00000001113,     0.00000002308,
       0.00000000212,    -0.00000001738,
      -0.00000000354,     0.00000000156,
       0.00000001854,    -0.00000000428,
      -0.00000000320,    -0.00000000489,
       0.00000000457,     0.00000001815,
       0.00000000510,    -0.00000000329,
      -0.00000001498,     0.00000001488,
       0.00000000663,     0.00000000772,
       0.00000001567,    -0.00000001345,
      -0.00000010134,    -0.00000006350,
       0.00000000522,     0.00000002064,
      -0.00000001824,     0.00000002970,
       0.00000000730,    -0.00000000184,
       0.00000001752,     0.00000000279,
      -0.00000000104,    -0.00000000344,
      -0.00000000253,     0.00000001762,
       0.00000000341,    -0.00000000102,
      -0.00000000079,    -0.00000000060,
       0.00000000059,    -0.00000000079,
       0.00000000135,     0.00000000074,
       0.00000000117,     0.00000000169,
       0.00000000178,    -0.00000000112,
       0.00000006904,    -0.00000010824,
      -0.00000003055,    -0.00000001831,
       0.00000005067,     0.00000003117,
       0.00000000884,    -0.00000001568,
      -0.00000000232,    -0.00000000101,
      -0.00000004504,    -0.00000000142,
       0.00000000366,     0.00000001626,
       0.00000000301,    -0.00000000135,
      -0.00000001345,    -0.00000003431,
      -0.00000001457,     0.00000000536,
       0.00000000135,     0.00000000300,
      -0.00000105521,     0.00000021740,
       0.00000014137,     0.00000038893,
       0.00000007231,    -0.00000004071,
       0.00000000763,     0.00000003894,
       0.00000001492,    -0.00000000547,
      -0.00000000156,    -0.00000000275,
      -0.00000000853,     0.00000000194,
       0.00000000100,     0.00000000289,
      -0.00000000105,    -0.00000000797,
      -0.00000000305,     0.00000000088,
      -0.00000357686,     0.00000075356,
       0.00000011960,     0.00000127103,
       0.00000024211,    -0.00000000802,
      -0.00000000152,    -0.00000003177,
      -0.00000016898,    -0.00000008652,
      -0.00000002019,     0.00000004796,
       0.00000000709,     0.00000000208,
      -0.00000239667,     0.00000036907,
       0.00000027004,     0.00000092929,
       0.00000018277,    -0.00000008195,
      -0.00000001523,    -0.00000002313,
      -0.00000000208,     0.00000000199,
      -0.00000036596,    -0.00000239967,
      -0.00000092988,     0.00000026982,
       0.00000008195,     0.00000018285,
       0.00000002312,    -0.00000001523,
      -0.00000000199,    -0.00000000208,
      -0.00000001710,     0.00000000702,
       0.00000000176,     0.00000001294,
       0.00000000502,    -0.00000000138,
      -0.00000000046,    -0.00000000106,
      -0.00000000077,    -0.00000000518,
      -0.00000000201,     0.00000000057,
       0.00000000128,     0.00000000036,
       0.00000000709,    -0.00000000942,
       0.00000000786,     0.00000000550,
       0.00000000163,    -0.00000000279,
      -0.00000000598,    -0.00000000123,
       0.00000000049,     0.00000000155,
      -0.00000000271,    -0.00000000257,
      -0.00000000165,     0.00000000063,
      -0.00000018623,     0.00000026604,
       0.00000007536,     0.00000002948,
      -0.00000005437,    -0.00000005348,
      -0.00000001275,     0.00000001747,
       0.00000000207,     0.00000000068,
      -0.00000003000,     0.00000000718,
       0.00000000384,     0.00000001094,
       0.00000000205,    -0.00000000114,
      -0.00000000706,    -0.00000003141,
      -0.00000001144,     0.00000000414,
       0.00000000128,     0.00000000216,
       0.00000017224,    -0.00000003488,
      -0.00000002726,    -0.00000003727,
       0.00000000320,     0.00000001472,
       0.00000000322,    -0.00000000235,
       0.00000000139,    -0.00000000160,
       0.00000000156,     0.00000000120,
      -0.00000001710,     0.00000000331,
       0.00000000522,     0.00000000350,
       0.00000006625,     0.00000046401,
       0.00000009868,    -0.00000014719,
       0.00000000463,    -0.00000001686,
      -0.00000001640,    -0.00000000320,
       0.00000026930,    -0.00000005579,
      -0.00000005498,    -0.00000004602,
      -0.00000028984,    -0.00000161669,
      -0.00000028377,     0.00000033902,
      -0.00000004159,     0.00000009428,
       0.00000000961,     0.00000000726,
       0.00000002173,    -0.00000002007,
      -0.00000000952,     0.00000001095,
      -0.00000165318,     0.00000048912,
       0.00000030579,     0.00000032323,
       0.00000004069,    -0.00000006045,
       0.00000045187,     0.00000162442,
       0.00000030977,    -0.00000028609,
      -0.00000006001,    -0.00000003936,
       0.00000019272,    -0.00000025350,
      -0.00000007672,    -0.00000004765,
      -0.00000000387,     0.00000000809,
       0.00000026053,     0.00000018448,
       0.00000004791,    -0.00000007565,
      -0.00000001012,    -0.00000000477,
      -0.00000018975,     0.00000004178,
       0.00000003252,     0.00000003003,
       0.00000001194,     0.00000005144,
       0.00000000731,    -0.00000000917,
      -0.00000001702,     0.00000000512,
       0.00000000280,     0.00000000351,
      -0.00000000481,    -0.00000001689,
      -0.00000000398,     0.00000000315,
       0.00000000050,    -0.00000000117,
      -0.00000000806,    -0.00000000356,
      -0.00000000148,     0.00000000336,
       0.00000000109,    -0.00000000046,
       0.00000000035,     0.00000000126,
       0.00000005656,     0.00000001091,
      -0.00000000587,     0.00000000184,
       0.00000000166,     0.00000000174,
      -0.00000000203,    -0.00000000556,
      -0.00000000194,     0.00000000171,
       0.00000000160,     0.00000000010,
      -0.00000000009,     0.00000000161,
      -0.00000001202,    -0.00000001844,
      -0.00000004477,     0.00000002996,
       0.00000000494,     0.00000000509,
       0.00000001329,    -0.00000000357,
      -0.00000000087,    -0.00000000029,
       0.00000000213,    -0.00000001046,
       0.00000000022,    -0.00000000084,
       0.00000013061,     0.00000019889,
       0.00000002314,    -0.00000002260,
       0.00000010360,    -0.00000006605,
      -0.00000001164,    -0.00000001287,
      -0.00000006457,     0.00000003325,
       0.00000000886,     0.00000000268,
       0.00000000022,    -0.00000000111,
      -0.00000002894,     0.00000003668,
      -0.00000000269,     0.00000000883,
       0.00000000127,     0.00000000011,
       0.00000003863,    -0.00000008388,
      -0.00000002361,    -0.00000000717,
       0.00000000456,     0.00000000209,
       0.00000000023,    -0.00000000114,
      -0.00000061128,    -0.00000086147,
      -0.00000008340,     0.00000011435,
       0.00000002116,    -0.00000001745,
      -0.00000000324,    -0.00000000156,
       0.00000006579,    -0.00000014820,
      -0.00000004142,    -0.00000001244,
      -0.00000000084,     0.00000000609,
       0.00000014855,     0.00000006379,
       0.00000001237,    -0.00000004154,
      -0.00000000608,    -0.00000000080,
       0.00000000599,    -0.00000001636,
       0.00000043631,     0.00000016082,
       0.00000001403,     0.00000001940,
       0.00000001982,    -0.00000001206,
      -0.00000008554,     0.00000022003,
       0.00000005764,     0.00000001187,
      -0.00000137473,    -0.00000052555,
      -0.00000007043,     0.00000034202,
      -0.00000004290,    -0.00000005546,
      -0.00000006064,     0.00000004559,
      -0.00000003242,     0.00000005847,
      -0.00000003330,    -0.00000002115,
       0.00000000092,    -0.00000000100,
      -0.00000000280,    -0.00000000259,
      -0.00000000486,     0.00000000526,
       0.00000000969,     0.00000000501,
       0.00000002030,    -0.00000006131,
       0.00000002611,     0.00000000866,
       0.00000000145,    -0.00000000504,
      -0.00000000137,     0.00000000117,
       0.00000001029,     0.00000001338,
       0.00000000461,    -0.00000000363,
      -0.00000000395,     0.00000000163,
      -0.00000000043,     0.00000000327,
      -0.00000003643,     0.00000000050,
      -0.00000000020,    -0.00000000173,
       0.00000000183,    -0.00000000465,
      -0.00000000184,    -0.00000000037,
       0.00000000492,     0.00000000170,
       0.00000000039,    -0.00000000191,
      -0.00000000161,    -0.00000000038,
       0.00000000071,     0.00000000064,
       0.00000000062,    -0.00000000092,
      -0.00000000507,     0.00000002124,
       0.00000000538,     0.00000000129,
      -0.00000000057,    -0.00000000067,
      -0.00000000437,    -0.00000000254,
      -0.00000000288,     0.00000000243,
      -0.00000000193,     0.00000000043,
      -0.00000000047,    -0.00000000187,
       0.00000001947,    -0.00000000075,
      -0.00000000065,    -0.00000001518,
      -0.00000000055,     0.00000000158,
       0.00000000161,     0.00000000030,
      -0.00000017122,     0.00000000622,
       0.00000000163,     0.00000003420,
       0.00000000269,    -0.00000000618,
      -0.00000000721,    -0.00000000165,
       0.00000008319,     0.00000001780,
      -0.00000000036,     0.00000000447,
       0.00000000683,     0.00000000709,
       0.00000000060,    -0.00000000131,
      -0.00000000718,     0.00000000681,
       0.00000000132,     0.00000000058,
       0.00000000157,    -0.00000000109,
      -0.00000000071,    -0.00000000067,
      -0.00000000115,    -0.00000000156,
      -0.00000000057,     0.00000000085,
       0.00000000345,    -0.00000000800,
       0.00000001674,     0.00000000735,
      -0.00000000065,     0.00000000737,
       0.00000001290,     0.00000000980,
       0.00000000090,    -0.00000000341,
       0.00000000910,    -0.00000001262,
       0.00000011750,    -0.00000019477,
      -0.00000011054,    -0.00000003176,
       0.00000000301,     0.00000002611,
       0.00000005793,     0.00000002144,
       0.00000000361,    -0.00000003177,
      -0.00000000628,     0.00000000217,
      -0.00000002146,    -0.00000002703,
      -0.00000001069,     0.00000001350,
       0.00000000300,     0.00000000266,
      -0.00000002672,     0.00000002170,
       0.00000001298,     0.00000001037,
       0.00000000161,    -0.00000000243,
      -0.00000000374,     0.00000002797,
      -0.00000000578,    -0.00000000068,
       0.00000000045,     0.00000000283,
      -0.00000000454,    -0.00000000582,
      -0.00000000471,     0.00000000372,
       0.00000000158,     0.00000000139,
      -0.00000000558,     0.00000000142,
       0.00000000074,     0.00000000213,
       0.00000000124,     0.00000000533,
       0.00000000212,    -0.00000000078,
      -0.00000018034,     0.00000015539,
       0.00000007621,     0.00000005307,
       0.00000000565,    -0.00000001792,
       0.00000001940,    -0.00000001900,
      -0.00000001126,    -0.00000000781,
      -0.00000000136,     0.00000000332,
      -0.00000002795,     0.00000000557,
       0.00000000393,     0.00000001036,
       0.00000000196,    -0.00000000093,
      -0.00000000049,    -0.00000000061,
       0.00000000536,     0.00000002851,
       0.00000001060,    -0.00000000379,
      -0.00000000121,    -0.00000000183,
      -0.00000000062,     0.00000000050,
       0.00000000088,     0.00000000156,
      -0.00000108571,     0.00000024154,
       0.00000015276,     0.00000040010,
       0.00000007432,    -0.00000004073,
      -0.00000001102,    -0.00000004029,
      -0.00000001394,     0.00000000594,
       0.00000000157,     0.00000000252,
      -0.00000000315,    -0.00000000029,
       0.00000000018,     0.00000000116,
      -0.00000000095,    -0.00000000095,
       0.00000005131,    -0.00000000569,
      -0.00000000005,    -0.00000000148,
      -0.00000000218,     0.00000000179,
       0.00000000129,     0.00000000006,
      -0.00000000028,    -0.00000000138,
       0.00000001323,     0.00000001764,
       0.00000000565,    -0.00000000419,
      -0.00000000076,    -0.00000000072,
      -0.00000000592,     0.00000000138,
      -0.00000000028,     0.00000000534,
      -0.00000000089,     0.00000000081,
      -0.00000000025,    -0.00000000153,
      -0.00000001915,     0.00000000288,
       0.00000000063,     0.00000000401,
       0.00000000314,     0.00000000221,
       0.00000000122,    -0.00000000167,
      -0.00000002042,    -0.00000001491,
      -0.00000000310,     0.00000000429,
      -0.00000000008,     0.00000000165,
       0.00000000155,     0.00000000018,
      -0.00000000214,     0.00000000629,
       0.00000000346,     0.00000000120,
       0.00000002243,    -0.00000003467,
      -0.00000000760,    -0.00000000363,
       0.00000000069,    -0.00000000081,
      -0.00000000079,     0.00000000046,
      -0.00000001627,    -0.00000004642,
       0.00000000021,     0.00000000087,
      -0.00000001149,    -0.00000000211,
       0.00000000211,    -0.00000001148,
       0.00000000392,     0.00000000002,
      -0.00000000101,    -0.00000000137,
      -0.00000000234,     0.00000000360,
       0.00000000360,     0.00000000235,
       0.00000000168,     0.00000000101,
       0.00000000019,    -0.00000000091,
       0.00000000145,    -0.00000000031,
      -0.00000000031,    -0.00000000146,
       0.00000003575,    -0.00000005351,
      -0.00000002930,    -0.00000000679,
      -0.00000000946,    -0.00000000591,
      -0.00000000106,     0.00000000527,
       0.00000000086,    -0.00000000021,
      -0.00000001093,     0.00000000353,
       0.00000000252,     0.00000000470,
       0.00000000071,    -0.00000000080,
       0.00000000343,     0.00000001098,
       0.00000000457,    -0.00000000284,
      -0.00000000081,    -0.00000000071,
      -0.00000000103,    -0.00000000234,
      -0.00000000060,     0.00000000045,
      -0.00000000113,    -0.00000000077,
      -0.00000000079,     0.00000000114,
      -0.00000010765,     0.00000004639,
       0.00000003101,     0.00000004210,
       0.00000000496,     0.00000001093,
       0.00000000432,    -0.00000000330,
      -0.00000000107,    -0.00000000079,
       0.00000000928,     0.00000000673,
       0.00000000272,    -0.00000000494,
      -0.00000000125,    -0.00000000058,
       0.00000000692,    -0.00000000928,
      -0.00000000484,    -0.00000000281,
      -0.00000000053,     0.00000000126,
      -0.00000001571,    -0.00000000968,
      -0.00000000290,     0.00000000799,
      -0.00000000329,     0.00000000497,
       0.00000000173,     0.00000000057,
       0.00000000042,     0.00000000114,
       0.00000000113,    -0.00000000039,
       0.00000016510,     0.00000006774,
       0.00000002490,    -0.00000007766,
      -0.00000001875,    -0.00000000407,
       0.00000000658,    -0.00000001609,
      -0.00000000758,    -0.00000000252,
      -0.00000000045,     0.00000000191,
       0.00000002594,    -0.00000000691,
      -0.00000000372,    -0.00000000945,
      -0.00000000177,     0.00000000098,
      -0.00000000730,    -0.00000002630,
      -0.00000000961,     0.00000000395,
       0.00000000099,     0.00000000182,
      -0.00000002934,     0.00000000679,
       0.00000000412,     0.00000001044,
       0.00000000163,     0.00000000690,
       0.00000000255,    -0.00000000099,
       0.00000028200,    -0.00000012907,
      -0.00000005657,    -0.00000009717,
      -0.00000001716,     0.00000001262,
      -0.00000000581,    -0.00000002773,
      -0.00000001023,     0.00000000380,
       0.00000000110,     0.00000000188,
       0.00000000281,    -0.00000001012,
      -0.00000000271,    -0.00000000022,
      -0.00000001052,    -0.00000000176,
       0.00000000032,     0.00000000281,
       0.00000007436,    -0.00000031922,
      -0.00000008565,     0.00000000360,
       0.00000001170,     0.00000000297,
      -0.00000000018,    -0.00000000322,
      -0.00000000017,    -0.00000000259,
      -0.00000003115,     0.00000000182,
       0.00000000042,     0.00000000657,
       0.00000000152,     0.00000000015,
      -0.00000000572,    -0.00000000105,
       0.00000000105,    -0.00000000572,
       0.00000000108,    -0.00000000126,
      -0.00000000132,    -0.00000000105,
       0.00000001600,    -0.00000002005,
      -0.00000000170,    -0.00000000120,
      -0.00000000029,     0.00000000098,
      -0.00000000367,     0.00000000077,
       0.00000000066,     0.00000000164,
       0.00000000070,     0.00000000369,
       0.00000000165,    -0.00000000068,
      -0.00000000035,    -0.00000000110,
      -0.00000006730,     0.00000002157,
       0.00000001520,     0.00000002780,
       0.00000000174,     0.00000000512,
       0.00000000212,    -0.00000000127,
       0.00000000353,     0.00000000431,
       0.00000000200,    -0.00000000212,
      -0.00000000059,    -0.00000000047,
       0.00000000450,    -0.00000000345,
      -0.00000000207,    -0.00000000209,
      -0.00000000046,     0.00000000061,
      -0.00000000741,    -0.00000000444,
      -0.00000000118,     0.00000000158,
       0.00000000083,     0.00000000052,
      -0.00000000192,     0.00000000032,
       0.00000000042,     0.00000000202,
       0.00000000077,    -0.00000000027,
       0.00000008952,     0.00000005510,
       0.00000002447,    -0.00000004512,
       0.00000000505,    -0.00000000718,
      -0.00000000371,    -0.00000000224,
      -0.00000000047,     0.00000000099,
       0.00000001723,    -0.00000000323,
      -0.00000000216,    -0.00000000639,
      -0.00000000120,     0.00000000063,
      -0.00000000359,    -0.00000001779,
      -0.00000000664,     0.00000000236,
       0.00000000066,     0.00000000126,
       0.00000000009,     0.00000000112,
       0.00000000048,    -0.00000000075,
      -0.00000000074,    -0.00000000052,
      -0.00000001339,     0.00000000436,
       0.00000000081,     0.00000000227,
       0.00000000189,    -0.00000000068,
      -0.00000000201,     0.00000000302,
       0.00000000301,     0.00000000198,
       0.00000024504,    -0.00000004191,
      -0.00000002983,    -0.00000008978,
      -0.00000001682,     0.00000000833,
      -0.00000000389,    -0.00000002116,
      -0.00000000785,     0.00000000268,
       0.00000000073,     0.00000000147,
       0.00000000091,     0.00000000198,
       0.00000000048,    -0.00000000055,
       0.00000000141,     0.00000000026,
      -0.00000003501,     0.00000001730,
       0.00000000190,     0.00000000006,
      -0.00000000395,    -0.00000000063,
      -0.00000000060,     0.00000000560,
      -0.00000017107,    -0.00000001825,
      -0.00000001336,     0.00000000674,
       0.00000000086,    -0.00000000684,
       0.00000000062,    -0.00000000061,
      -0.00000000064,    -0.00000000061,
       0.00000001176,    -0.00000001277,
      -0.00000000091,    -0.00000000074,
      -0.00000000213,     0.00000000018,
       0.00000000026,     0.00000000100,
       0.00000000014,     0.00000000214,
       0.00000000101,    -0.00000000026,
      -0.00000004488,     0.00000000968,
       0.00000000788,     0.00000001965,
       0.00000000069,     0.00000000308,
       0.00000000137,    -0.00000000059,
       0.00000000123,     0.00000000311,
       0.00000000152,    -0.00000000098,
       0.00000000327,    -0.00000000111,
      -0.00000000095,    -0.00000000160,
      -0.00000000090,     0.00000000066,
      -0.00000000120,     0.00000000021,
       0.00000000027,     0.00000000135,
       0.00000004078,     0.00000004781,
       0.00000002287,    -0.00000002404,
       0.00000000388,    -0.00000000304,
      -0.00000000183,    -0.00000000184,
       0.00000001113,    -0.00000000207,
      -0.00000000137,    -0.00000000413,
      -0.00000000080,     0.00000000039,
      -0.00000000231,    -0.00000001167,
      -0.00000000437,     0.00000000150,
       0.00000000042,     0.00000000084,
      -0.00000001963,     0.00000000340,
       0.00000000241,     0.00000000733,
       0.00000000062,     0.00000000342,
       0.00000000129,    -0.00000000043,
       0.00000018356,    -0.00000003135,
      -0.00000002226,    -0.00000006820,
      -0.00000000256,    -0.00000001443,
      -0.00000000539,     0.00000000175,
       0.00000000047,     0.00000000101,
       0.00000000071,     0.00000000215,
       0.00000000056,    -0.00000000052,
       0.00000000155,     0.00000000006,
       0.00000000208,    -0.00000000325,
      -0.00000000156,    -0.00000000096,
      -0.00000001001,     0.00000001051,
       0.00000000342,     0.00000000218,
      -0.00000000325,     0.00000002744,
       0.00000000227,    -0.00000000020,
       0.00000000024,     0.00000000124,
       0.00000002499,     0.00000000535,
       0.00000000326,    -0.00000000044,
      -0.00000000043,    -0.00000000328,
       0.00000011751,    -0.00000001115,
      -0.00000000036,     0.00000000562,
      -0.00000000017,     0.00000000148,
       0.00000000157,     0.00000000026,
       0.00000000254,     0.00000002452,
       0.00000000193,    -0.00000000017,
      -0.00000000036,    -0.00000000098,
       0.00000000408,    -0.00000000089,
      -0.00000000051,    -0.00000000151,
      -0.00000000093,    -0.00000000439,
      -0.00000000165,     0.00000000053,
      -0.00000000855,     0.00000000169,
       0.00000000025,     0.00000000129,
       0.00000007733,    -0.00000000950,
      -0.00000000853,    -0.00000002973,
      -0.00000000107,    -0.00000000557,
      -0.00000000208,     0.00000000064,
       0.00000000104,     0.00000000095,
       0.00000000053,    -0.00000000087,
       0.00000000998,     0.00000001350,
       0.00000000160,    -0.00000000031,
      -0.00000000104,     0.00000000039,
       0.00000000053,     0.00000000180,
      -0.00000003287,     0.00000002804,
      -0.00000000153,    -0.00000000166,
       0.00000000034,     0.00000000067,
       0.00000001071,    -0.00000000506,
      -0.00000000065,    -0.00000000181,
      -0.00000000021,     0.00000000235,
       0.00000000235,     0.00000000016,
      -0.00000000182,     0.00000000051,
      -0.00000005689,    -0.00000001104,
      -0.00000000026,     0.00000000097,
       0.00000000104,     0.00000000034,
      -0.00000000288,     0.00000001886,
       0.00000000129,     0.00000000015,
       0.00000000235,    -0.00000000058,
      -0.00000000030,    -0.00000000086,
      -0.00000000058,    -0.00000000255,
      -0.00000000096,     0.00000000031,
       0.00000004760,    -0.00000001014,
      -0.00000000578,    -0.00000001784,
      -0.00000000069,    -0.00000000325,
      -0.00000000121,     0.00000000037,
       0.00000000028,     0.00000000103,
       0.00000000533,     0.00000001641,
       0.00000000115,    -0.00000000027,
      -0.00000000141,    -0.00000000012,
      -0.00000000151,     0.00000001937,
       0.00000000322,     0.00000000025,
       0.00000000131,    -0.00000000038,
      -0.00000000036,    -0.00000000144,
       0.00000002816,    -0.00000000724,
      -0.00000000044,    -0.00000000184,
      -0.00000000028,    -0.00000000101,
      -0.00000000129,    -0.00000000006,
      -0.00000000009,     0.00000000129,
       0.00000000020,     0.00000000189,
       0.00000000089,    -0.00000000025,
      -0.00000000263,     0.00000000187,
       0.00000000110,    -0.00000000042,
       0.00000000200,     0.00000000271,
      -0.00000000039,    -0.00000000117,
       0.00000000017,     0.00000000086,
      -0.00000005466,     0.00000003000,
       0.00000001759,    -0.00000000700,
       0.00000000250,     0.00000000237,
      -0.00000000085,    -0.00000000138,
       0.00000001953,    -0.00000000019,
      -0.00000000084,    -0.00000000255,
      -0.00000000030,    -0.00000001986,
      -0.00000000274,     0.00000000091,
      -0.00000000009,     0.00000000133,
       0.00000000107,     0.00000000006,
       0.00000000007,    -0.00000000103,
      -0.00000001325,     0.00000000221,
       0.00000000039,     0.00000000215,
       0.00000000082,    -0.00000000026,
       0.00000043321,     0.00000000312,
      -0.00000001194,    -0.00000005939,
      -0.00000000058,    -0.00000002514,
      -0.00000000343,     0.00000000108,
      -0.00000004049,    -0.00000000139,
      -0.00000000193,     0.00000004041,
       0.00000001276,     0.00000000068,
       0.00000000027,    -0.00000000517,
      -0.00000102477,    -0.00000004955,
      -0.00000000099,     0.00000005796,
       0.00000000101,     0.00000000034,
       0.00000004827,     0.00000000943,
       0.00000000099,     0.00000000176,
       0.00000000964,    -0.00000004483,
       0.00000000187,    -0.00000000096,
      -0.00000002547,    -0.00000000079,
      -0.00000000060,     0.00000000917,
      -0.00000000968,     0.00000000046,
       0.00000000055,     0.00000000997,
       0.00000000121,    -0.00000000058,
      -0.00000000077,    -0.00000000131,
       0.00000180750,     0.00000033801,
       0.00000002090,     0.00000004380,
       0.00000000558,    -0.00000009093,
       0.00000000146,    -0.00000000034,
       0.00000028765,    -0.00000003885,
      -0.00000000519,     0.00000000183,
      -0.00000003858,    -0.00000029665,
       0.00000000164,     0.00000000512,
      -0.00000012027,     0.00000000669,
       0.00000000245,     0.00000004401,
       0.00000000019,     0.00000000169,
      -0.00000290070,    -0.00000136789,
      -0.00000017965,    -0.00000021632,
       0.00000001338,    -0.00000001458,
      -0.00000000225,    -0.00000049432,
       0.00000001030,    -0.00000000072,
      -0.00000074093,    -0.00000014894,
      -0.00000001351,    -0.00000004394,
       0.00000000220,    -0.00000000176,
      -0.00000015178,     0.00000075982,
      -0.00000004428,     0.00000001357,
      -0.00000000178,    -0.00000000221,
      -0.00000014307,    -0.00000003593,
      -0.00000000050,     0.00000000194,
      -0.00002765789,    -0.00000544619,
      -0.00000049020,    -0.00000161872,
       0.00000008072,    -0.00000006458,
       0.00000025049,    -0.00000127800,
       0.00000007437,    -0.00000002242,
       0.00000000295,     0.00000000371,
      -0.00000006474,    -0.00000001134,
      -0.00000000146,    -0.00000000360,
       0.00000000736,    -0.00000001401,
       0.00000000146,    -0.00000000130,
      -0.00000000139,    -0.00000000156,
       0.00000136558,     0.00000040994,
       0.00000001658,     0.00000008225,
      -0.00000000550,     0.00000001749,
      -0.00000000111,     0.00000000024,
       0.00000002280,     0.00000000938,
      -0.00000000006,     0.00000000193,
       0.00000001124,    -0.00000002951,
       0.00000000224,    -0.00000000010,
       0.00000000036,    -0.00000000100,
      -0.00000001319,    -0.00000000214,
      -0.00000000025,    -0.00000000075,
       0.00000000215,    -0.00000001318,
       0.00000000076,    -0.00000000025,
      -0.00000000069,     0.00000000049,
      -0.00000000049,    -0.00000000067,
       0.00000001123,     0.00000000189,
      -0.00000000189,     0.00000001124,
       0.00000009742,     0.00000021558,
      -0.00000001871,     0.00000001575,
      -0.00000000109,    -0.00000000075,
       0.00000011469,     0.00000074427,
      -0.00000005094,     0.00000003345,
      -0.00000000216,    -0.00000000172,
      -0.00000074588,     0.00000011105,
      -0.00000003346,    -0.00000005119,
       0.00000000173,    -0.00000000217,
       0.00000003097,     0.00000002634,
       0.00000000050,     0.00000000137,
      -0.00000002905,     0.00000003925,
      -0.00000000208,     0.00000000061,
      -0.00000000001,    -0.00000000006,
       0.00000001135,    -0.00000000778,
       0.00000000300,    -0.00000000009,
      -0.00000008727,     0.00000002144,
      -0.00000000051,    -0.00000000789,
      -0.00000002127,    -0.00000008721,
       0.00000000793,    -0.00000000050,
       0.00000000986,     0.00000004259,
       0.00000000237,     0.00000000224,
      -0.00000004236,     0.00000000990,
      -0.00000000222,     0.00000000239,
       0.00000000294,    -0.00000001182,
       0.00000000624,    -0.00000000349,
      -0.00000019113,     0.00000023131,
      -0.00000002875,    -0.00000001788,
       0.00000000067,    -0.00000000200,
      -0.00000023143,    -0.00000019132,
       0.00000001794,    -0.00000002876,
       0.00000000201,     0.00000000067,
       0.00000001556,    -0.00000003957,
      -0.00000000105,    -0.00000000006,
       0.00000003942,     0.00000001551,
       0.00000000005,    -0.00000000108,
       0.00000000038,    -0.00000000038,
      -0.00000009932,    -0.00000029211,
       0.00000004763,    -0.00000003125,
       0.00000000309,     0.00000000365,
       0.00000225208,     0.00000018321,
       0.00000004868,     0.00000022451,
      -0.00000000955,     0.00000000541,
      -0.00000018428,     0.00000225323,
      -0.00000022466,     0.00000004863,
      -0.00000000541,    -0.00000000956,
       0.00000004917,     0.00000004940,
       0.00000000536,    -0.00000000047,
      -0.00000005030,     0.00000004670,
       0.00000000069,     0.00000000509,
       0.00000326707,     0.00000030141,
      -0.00000030860,     0.00000126664,
      -0.00000025111,    -0.00000009161,
       0.00000001583,    -0.00000003267,
       0.00000011512,    -0.00000099392,
       0.00000038357,     0.00000010129,
      -0.00000002923,     0.00000007599,
      -0.00000000984,    -0.00000000499,
      -0.00000001653,    -0.00000015261,
       0.00000004649,     0.00000001251,
      -0.00000000522,     0.00000000734,
       0.00000012324,    -0.00000010630,
       0.00000003357,     0.00000003213,
      -0.00000000382,     0.00000000699,
       0.00000001351,     0.00000000658,
       0.00000000088,    -0.00000000154,
      -0.00000000792,     0.00000001201,
       0.00000000197,     0.00000000061,
      -0.00001493155,    -0.00000141659,
       0.00000142380,    -0.00000582197,
       0.00000115262,     0.00000042467,
      -0.00000007302,     0.00000015008,
      -0.00000001450,    -0.00000000866,
       0.00000010642,     0.00000250442,
      -0.00000091676,    -0.00000016603,
       0.00000005790,    -0.00000017784,
       0.00000002285,     0.00000001041,
      -0.00000000131,     0.00000000222,
       0.00000439989,     0.00000267388,
       0.00000003063,     0.00000042762,
      -0.00000001243,     0.00000001251,
      -0.00000000040,     0.00000000262,
      -0.00000255174,     0.00000476132,
      -0.00000061286,     0.00000009014,
      -0.00000002735,    -0.00000006157,
       0.00000000617,    -0.00000000282,
      -0.00000003869,    -0.00000001686,
      -0.00000000047,     0.00000000184,
       0.00000002332,    -0.00000002795,
      -0.00000000491,     0.00000000129,
       0.00000001251,     0.00000002320,
      -0.00000018583,     0.00000046827,
      -0.00000005394,     0.00000001183,
      -0.00000000296,    -0.00000000398,
       0.00000637687,     0.00000347760,
       0.00000016666,     0.00000071334,
      -0.00000005119,     0.00000003845,
      -0.00000000326,    -0.00000000251,
      -0.00000347887,     0.00000637819,
      -0.00000071374,     0.00000016664,
      -0.00000003847,    -0.00000005124,
       0.00000000252,    -0.00000000326,
      -0.00000000108,     0.00000000030,
       0.00000000068,     0.00000000072,
       0.00000009258,     0.00000008646,
       0.00000011934,    -0.00000002815,
       0.00000001114,     0.00000000749,
      -0.00000000017,     0.00000000118,
       0.00000797501,     0.00000435004,
       0.00000020876,     0.00000089278,
      -0.00000006407,     0.00000004815,
      -0.00000000408,    -0.00000000315,
      -0.00000435281,     0.00000797760,
      -0.00000089332,     0.00000020860,
      -0.00000004818,    -0.00000006416,
       0.00000000315,    -0.00000000409,
      -0.00000000163,    -0.00000000047,
      -0.00000000018,    -0.00000000145,
       0.00000000008,     0.00000000006,
       0.00000000034,    -0.00000000058,
       0.00000000005,     0.00000000003,
       0.00000000139,    -0.00000000083,
       0.00000000009,     0.00000000011,
       0.00000000080,     0.00000000137,
      -0.00000000010,     0.00000000008,
       0.00000000028,    -0.00000000040,
       0.00000000041,     0.00000000028,
       0.00000000002,    -0.00000000013,
       0.00000000018,     0.00000000005,
       0.00000000005,     0.00000000001,
       0.00000000001,    -0.00000000005,
       0.00000000004,     0.00000000004,
      -0.00000000004,     0.00000000004,
       0.00000000027,     0.00000000005,
      -0.00000000007,    -0.00000000002,
      -0.00000000018,    -0.00000000044,
       0.00000000044,    -0.00000000017,
      -0.00000000181,    -0.00000000145,
       0.00000000007,    -0.00000000065,
      -0.00000000065,    -0.00000000008,
      -0.00000000037,     0.00000000044,
       0.00000000950,     0.00000000761,
      -0.00000000020,     0.00000000109,
       0.00000000110,     0.00000000021,
      -0.00000000020,    -0.00000000062,
      -0.00000000061,     0.00000000020,
      -0.00000000056,    -0.00000000172,
       0.00000000079,    -0.00000000026,
      -0.00000000040,    -0.00000000066,
       0.00000000070,    -0.00000000044,
      -0.00000000023,    -0.00000000035,
       0.00000000037,    -0.00000000023,
      -0.00000000058,     0.00000000035,
       0.00000000008,    -0.00000000045,
      -0.00000000018,     0.00000000100,
       0.00000000123,    -0.00000000731,
       0.00000000035,    -0.00000000019,
      -0.00000000089,     0.00000000521,
      -0.00000000059,    -0.00000000069,
      -0.00000000046,    -0.00000000008,
      -0.00000000008,     0.00000000046,
      -0.00000000001,     0.00000000069,
       0.00000000011,    -0.00000000060,
       0.00000000000,     0.00000000072,
      -0.00000000014,    -0.00000000034,
      -0.00000018808,     0.00000008091,
       0.00000000293,     0.00000000987,
       0.00000000189,    -0.00000000076,
      -0.00000000075,    -0.00000000188,
       0.00000000023,     0.00000000053,
       0.00000013199,    -0.00000005623,
      -0.00000000228,    -0.00000000691,
      -0.00000000202,     0.00000000071,
       0.00000000069,     0.00000000201,
       0.00000000048,    -0.00000000017,
      -0.00000000071,    -0.00000000032,
       0.00000000032,    -0.00000000071,
      -0.00000000127,    -0.00000000183,
       0.00000000236,    -0.00000000160,
      -0.00000000273,    -0.00000000152,
       0.00000000210,    -0.00000000412,
      -0.00000000098,    -0.00000000161,
       0.00000000164,    -0.00000000095,
      -0.00000000013,     0.00000000056,
      -0.00000000203,    -0.00000000062,
       0.00000000257,    -0.00000000515,
      -0.00000000384,    -0.00000000200,
       0.00000000203,    -0.00000000363,
       0.00000000137,    -0.00000000028,
      -0.00000000055,    -0.00000000262,
       0.00000000247,     0.00000000110,
      -0.00000000112,     0.00000000245,
       0.00000000042,    -0.00000000025,
      -0.00000000137,    -0.00000000175,
       0.00000000283,    -0.00000000201,
      -0.00000000047,     0.00000000051,
      -0.00000000086,    -0.00000000158,
       0.00000000406,    -0.00000000244,
      -0.00000000158,    -0.00000000265,
       0.00000000326,    -0.00000000204,
       0.00000000035,     0.00000000082,
      -0.00000000038,    -0.00000000036,
       0.00000000036,    -0.00000000039,
       0.00000000206,    -0.00000000061,
      -0.00000000374,    -0.00000000508,
       0.00000000312,     0.00000000438,
      -0.00000000446,     0.00000000375,
       0.00000000044,     0.00000000011,
      -0.00000000011,     0.00000000044,
       0.00000000117,     0.00000000269,
       0.00000000115,    -0.00000000009,
       0.00000000191,    -0.00000000016,
      -0.00000000036,    -0.00000000116,
       0.00000000117,    -0.00000000039,
       0.00000000044,    -0.00000000094,
      -0.00000001530,     0.00000000404,
       0.00000000048,     0.00000000189,
      -0.00000000100,     0.00000000258,
      -0.00000000561,     0.00000000185,
      -0.00000000059,     0.00000000046,
       0.00000000045,     0.00000000039,
      -0.00000000039,     0.00000000045,
       0.00000000022,     0.00000000168,
      -0.00000000035,     0.00000000022,
       0.00000000023,     0.00000000035,
      -0.00000000022,    -0.00000000067,
       0.00000000080,    -0.00000000027,
      -0.00000000114,    -0.00000000025,
       0.00000000023,     0.00000000072,
       0.00000000070,    -0.00000000024,
      -0.00000000044,    -0.00000000057,
      -0.00000000154,     0.00000000124,
      -0.00000000024,    -0.00000000071,
       0.00000000071,    -0.00000000022,
       0.00000000013,    -0.00000000058,
      -0.00000000058,    -0.00000000014,
      -0.00000000085,    -0.00000000054,
       0.00000000039,     0.00000000112,
       0.00000000115,    -0.00000000037,
       0.00000000093,    -0.00000000177,
       0.00000000584,     0.00000000600,
      -0.00000000063,     0.00000000201,
      -0.00000000119,    -0.00000000283,
      -0.00000000422,     0.00000000110,
      -0.00000000091,    -0.00000000264,
       0.00000000100,    -0.00000000001,
       0.00000000029,    -0.00000000035,
      -0.00000000035,    -0.00000000029,
      -0.00000000115,    -0.00000000096,
      -0.00000000010,     0.00000000051,
       0.00000000051,     0.00000000010,
       0.00000000203,    -0.00000000229,
       0.00000000740,     0.00000000654,
       0.00000000118,    -0.00000000525,
      -0.00000000513,    -0.00000000126,
      -0.00000000068,    -0.00000000014,
      -0.00000000058,     0.00000000206,
       0.00000000476,     0.00000000119,
       0.00000000078,     0.00000000236,
       0.00000000235,    -0.00000000080,
       0.00000000039,     0.00000000115,
       0.00000000202,    -0.00000000067,
       0.00000000123,    -0.00000000035,
       0.00000000066,     0.00000000248,
       0.00000000145,    -0.00000000170,
      -0.00000000171,    -0.00000000143,
      -0.00000000044,    -0.00000000038,
      -0.00000000127,     0.00000000151,
       0.00000000341,     0.00000000292,
      -0.00000000050,     0.00000000247,
       0.00000000247,     0.00000000045,
      -0.00000000028,     0.00000000144,
       0.00000000208,     0.00000000042,
       0.00000000034,     0.00000000110,
       0.00000000107,    -0.00000000038,
       0.00000000030,     0.00000000092,
       0.00000000129,    -0.00000000042,
      -0.00000000010,     0.00000000077,
       0.00000000081,    -0.00000000030,
      -0.00000000031,    -0.00000000081,
      -0.00000000016,    -0.00000000042,
      -0.00000000127,     0.00000000048,
       0.00000000121,     0.00000000325,
      -0.00000000122,     0.00000000142,
       0.00000000143,     0.00000000118,
      -0.00000000068,     0.00000000081,
       0.00000000111,     0.00000000095,
      -0.00000000023,     0.00000000110,
       0.00000000109,     0.00000000019,
      -0.00000000021,     0.00000000106,
       0.00000000127,     0.00000000025,
       0.00000000018,     0.00000000064,
       0.00000000059,    -0.00000000023,
       0.00000000022,     0.00000000068,
       0.00000000082,    -0.00000000027,
      -0.00000000088,    -0.00000000013,
      -0.00000000040,     0.00000000268,
      -0.00000000119,     0.00000000044,
       0.00000000046,     0.00000000117,
      -0.00000000060,     0.00000000023,
       0.00000000030,     0.00000000081,
      -0.00000000050,     0.00000000058,
       0.00000000059,     0.00000000047,
      -0.00000000049,     0.00000000059,
       0.00000000065,     0.00000000056,
      -0.00000000015,     0.00000000067,
       0.00000000066,     0.00000000011,
      -0.00000000015,     0.00000000079,
       0.00000000084,     0.00000000017,
       0.00000000051,    -0.00000000017,
      -0.00000000047,    -0.00000000036,
      -0.00000000146,     0.00000000190,
      -0.00000000086,    -0.00000000013,
      -0.00000000011,     0.00000000085,
      -0.00000000007,     0.00000000048,
      -0.00000000042,     0.00000000015,
       0.00000000016,     0.00000000041,
       0.00000000017,     0.00000000046,
      -0.00000000032,     0.00000000036,
       0.00000000036,     0.00000000029,
      -0.00000000038,     0.00000000045,
       0.00000000044,     0.00000000038,
      -0.00000000011,     0.00000000044,
       0.00000000057,     0.00000000011,
       0.00000000036,    -0.00000000016,
      -0.00000000274,     0.00000000120,
      -0.00000000054,    -0.00000000042,
      -0.00000000040,     0.00000000054,
       0.00000000032,     0.00000000027,
       0.00000000210,     0.00000000019,
      -0.00000001514,    -0.00000000162,
      -0.00000000004,    -0.00000000050,
      -0.00000000043,    -0.00000000101,
      -0.00000000099,     0.00000000045,
      -0.00000000205,    -0.00000000138,
       0.00000000015,    -0.00000000016,
      -0.00000000004,    -0.00000000001,
      -0.00000000025,    -0.00000000072,
       0.00000000033,     0.00000000101,
       0.00000000101,    -0.00000000033,
       0.00000000017,     0.00000000049,
      -0.00000000029,    -0.00000000090,
      -0.00000000090,     0.00000000029,
      -0.00000000055,     0.00000000003,
       0.00000000042,    -0.00000000014,
       0.00000000087,    -0.00000000349,
      -0.00000000060,     0.00000000238,
       0.00000000000,     0.00000000005,
      -0.00000000017,    -0.00000000051,
      -0.00000000051,     0.00000000017,
       0.00000000060,    -0.00000000020,
      -0.00000000038,     0.00000000012,
      -0.00000000044,     0.00000000014,
      -0.00000000362,    -0.00000001161,
      -0.00000000149,     0.00000000005,
       0.00003277491,    -0.00000968414,
       0.00000019178,    -0.00000424825,
      -0.00000027264,    -0.00000002136,
      -0.00000000045,     0.00000001287,
       0.00000000041,     0.00000000001,
      -0.00000000979,    -0.00000001584,
      -0.00000000131,     0.00000000037,
      -0.00000001579,     0.00000000829,
       0.00000000031,     0.00000000112,
      -0.00000000057,     0.00000000008,
       0.00000008409,     0.00000027017,
       0.00000003477,    -0.00000000118,
      -0.00026608044,     0.00007861222,
      -0.00000155762,     0.00003448923,
       0.00000221345,     0.00000017343,
       0.00000000364,    -0.00000010447,
      -0.00000000337,     0.00000000008,
       0.00000015787,     0.00000022388,
       0.00000001938,    -0.00000001164,
       0.00000012504,    -0.00000004907,
       0.00000000631,    -0.00000000530,
       0.00000000312,    -0.00000000089,
      -0.00000000163,    -0.00000000124,
       0.00000000320,    -0.00000000113,
      -0.00000000114,     0.00000000053,
      -0.00000000024,    -0.00000000020,
       0.00000000071,     0.00000000159,
      -0.00000000028,     0.00000000031,
      -0.00000001061,     0.00000000287,
      -0.00000000120,    -0.00000000177,
      -0.00000000291,    -0.00000001063,
       0.00000000177,    -0.00000000121,
      -0.00000000183,     0.00000000035,
      -0.00000000034,    -0.00000000182,
       0.00000003264,    -0.00000000225,
       0.00000000027,     0.00000000261,
       0.00000000121,    -0.00000000009,
       0.00000000015,    -0.00000000140,
       0.00000000174,    -0.00000000022,
       0.00000000020,     0.00000000161,
       0.00000002470,    -0.00000000770,
       0.00000000850,     0.00000000220,
      -0.00000005460,    -0.00000001580,
      -0.00000000760,    -0.00000000220,
       0.00000005700,     0.00000006350,
      -0.00000001490,    -0.00000000850,
      -0.00000000151,    -0.00000000097,
      -0.00000000069,     0.00000000145,
      -0.00000000103,     0.00000000143,
       0.00000000136,     0.00000000077,
       0.00000000670,     0.00000000403,
       0.00000000278,    -0.00000000645,
      -0.00000000327,    -0.00000000083,
       0.00000000409,    -0.00000000654,
      -0.00000000627,    -0.00000000290,
      -0.00000000096,     0.00000000308,
       0.00000000094,    -0.00000000041,
       0.00000001589,    -0.00000001428,
      -0.00000001125,    -0.00000000774,
      -0.00000000138,     0.00000000321,
      -0.00000001346,    -0.00000001615,
      -0.00000000817,     0.00000001052,
       0.00000000296,     0.00000000142,
       0.00000000212,    -0.00000000059,
      -0.00000000041,    -0.00000000127,
       0.00000002113,    -0.00000001804,
      -0.00000001419,    -0.00000001060,
      -0.00000000199,     0.00000000413,
      -0.00000001805,    -0.00000002114,
      -0.00000001061,     0.00000001420,
       0.00000000414,     0.00000000200,
       0.00000002663,    -0.00000002267,
      -0.00000001786,    -0.00000001337,
      -0.00000000252,     0.00000000521,
      -0.00000002269,    -0.00000002664,
      -0.00000001338,     0.00000001787,
       0.00000000522,     0.00000000252,
       0.00000114929,    -0.00000100656,
      -0.00000081396,    -0.00000057816,
      -0.00000010915,     0.00000025186,
      -0.00000002999,    -0.00000003427,
      -0.00000001724,     0.00000002422,
       0.00000000747,     0.00000000325,
      -0.00000000027,    -0.00000000107,
       0.00000037538,    -0.00000022150,
      -0.00000018532,    -0.00000018239,
      -0.00000002966,     0.00000004692,
      -0.00000000862,    -0.00000000946,
      -0.00000000482,     0.00000000767,
       0.00000000205,     0.00000000106,
      -0.00000000014,     0.00000000110,
      -0.00000000340,    -0.00000003931,
      -0.00000002427,     0.00000000825,
      -0.00000000116,     0.00000000010,
      -0.00000002455,    -0.00000000706,
       0.00000000007,    -0.00000000232,
      -0.00000000108,     0.00000000003,
       0.00000000351,     0.00000000077,
      -0.00000000026,    -0.00000000131,
      -0.00000000085,     0.00000000330,
       0.00000000130,    -0.00000000025,
      -0.00000000133,     0.00000000009,
       0.00000000017,    -0.00000000132,
      -0.00000000133,    -0.00000000007,
      -0.00000001536,    -0.00000001548,
       0.00000000130,    -0.00000000133,
       0.00000000089,    -0.00000000064,
       0.00000001524,     0.00000001955,
      -0.00000000231,     0.00000000182,
      -0.00000000107,    -0.00000000115,
       0.00000000092,    -0.00000000100,
       0.00000000310,     0.00000000870,
      -0.00000003120,     0.00000001160,
      -0.00000002810,    -0.00000008244,
       0.00000005104,    -0.00000001740,
       0.00000000553,     0.00000000637,
      -0.00000000635,     0.00000000552,
      -0.00000000001,    -0.00000000008,
      -0.00000000002,    -0.00000000006,
      -0.00000000006,     0.00000000002,
       0.00000000104,    -0.00000000044,
      -0.00000000041,    -0.00000000102,
      -0.00000000032,     0.00000000213,
       0.00000000194,     0.00000000023,
       0.00000001753,    -0.00000000700,
      -0.00000000726,    -0.00000001320,
      -0.00000000499,     0.00000000354,
       0.00000000140,     0.00000000143,
      -0.00000000693,    -0.00000001756,
      -0.00000001317,     0.00000000725,
       0.00000000350,     0.00000000500,
       0.00000000149,    -0.00000000125,
       0.00000000110,    -0.00000000029,
      -0.00000000032,    -0.00000000108,
       0.00000009448,     0.00000004587,
       0.00000002932,    -0.00000008803,
      -0.00000004131,    -0.00000000847,
      -0.00000000104,     0.00000001284,
      -0.00000000219,    -0.00000000407,
      -0.00000000318,     0.00000000203,
       0.00000000089,     0.00000000136,
      -0.00000010442,     0.00000005258,
       0.00000004999,     0.00000007627,
       0.00000002808,    -0.00000002345,
      -0.00000000719,    -0.00000000662,
      -0.00000000101,     0.00000000155,
       0.00000005261,     0.00000010436,
       0.00000007620,    -0.00000005004,
      -0.00000002344,    -0.00000002800,
      -0.00000000660,     0.00000000714,
       0.00000000145,     0.00000000099,
      -0.00000000290,     0.00000000073,
       0.00000000067,     0.00000000144,
       0.00000000083,     0.00000000287,
       0.00000000141,    -0.00000000073,
       0.00000000081,     0.00000000416,
       0.00000000331,    -0.00000000114,
      -0.00000000065,    -0.00000000129,
      -0.00000005878,     0.00000001935,
       0.00000002141,     0.00000004455,
       0.00000001678,    -0.00000001101,
      -0.00000000363,    -0.00000000411,
      -0.00000000074,     0.00000000091,
       0.00000001938,     0.00000005881,
       0.00000004458,    -0.00000002144,
      -0.00000001102,    -0.00000001680,
      -0.00000000412,     0.00000000363,
       0.00000000091,     0.00000000074,
      -0.00000000095,    -0.00000000042,
      -0.00000000025,     0.00000000080,
      -0.00000007365,     0.00000002423,
       0.00000002681,     0.00000005583,
       0.00000002103,    -0.00000001379,
      -0.00000000455,    -0.00000000516,
      -0.00000000092,     0.00000000114,
       0.00000002427,     0.00000007369,
       0.00000005587,    -0.00000002684,
      -0.00000001380,    -0.00000002105,
      -0.00000000516,     0.00000000455,
       0.00000000114,     0.00000000092,
       0.00000000036,     0.00000000079,
      -0.00000170521,     0.00000055859,
       0.00000061934,     0.00000129423,
       0.00000048819,    -0.00000031891,
      -0.00000010598,    -0.00000012012,
      -0.00000002153,     0.00000002657,
       0.00000002409,     0.00000007315,
       0.00000005544,    -0.00000002666,
      -0.00000001369,    -0.00000002088,
      -0.00000000513,     0.00000000453,
       0.00000000114,     0.00000000092,
      -0.00000000119,     0.00000000017,
       0.00000000027,     0.00000000097,
      -0.00000000049,    -0.00000000209,
      -0.00000000161,     0.00000000061,
       0.00000000037,     0.00000000064,
      -0.00000001093,    -0.00000001247,
      -0.00000000094,     0.00000000083,
       0.00000007639,    -0.00000002904,
      -0.00000003116,    -0.00000005700,
      -0.00000002116,     0.00000001581,
      -0.00000000184,    -0.00000000159,
      -0.00000000101,     0.00000000142,
      -0.00000002476,     0.00000000814,
       0.00000000901,     0.00000001877,
       0.00000000708,    -0.00000000464,
      -0.00000000153,    -0.00000000173,
      -0.00000000821,    -0.00000002483,
      -0.00000001879,     0.00000000908,
       0.00000000467,     0.00000000708,
       0.00000000172,    -0.00000000151,
       0.00000000506,    -0.00000000418,
      -0.00000000348,    -0.00000000265,
       0.00000000589,     0.00000000610,
       0.00000000317,    -0.00000000472,
      -0.00000000159,    -0.00000000065,
      -0.00000000026,    -0.00000000288,
      -0.00000000164,     0.00000000056,
      -0.00000000318,     0.00000000017,
       0.00000000071,     0.00000000163,
       0.00000091533,    -0.00000081956,
      -0.00000066732,    -0.00000046780,
      -0.00000009501,     0.00000022574,
       0.00000002926,     0.00000000489,
      -0.00000006328,    -0.00000007176,
      -0.00000003659,     0.00000005129,
       0.00000001688,     0.00000000708,
       0.00000000038,    -0.00000000218,
       0.00000000627,     0.00000001202,
       0.00000000548,    -0.00000000561,
      -0.00000000199,    -0.00000000088,
       0.00000000877,    -0.00000000037,
      -0.00000000052,    -0.00000000398,
       0.00000008861,    -0.00000002715,
      -0.00000001348,    -0.00000003846,
       0.00000000185,     0.00000000799,
       0.00000000361,    -0.00000000085,
      -0.00000000022,    -0.00000000086,
       0.00000000148,     0.00000000006,
      -0.00000000023,    -0.00000000118,
       0.00000000003,     0.00000000145,
       0.00000000112,    -0.00000000025,
      -0.00000003712,     0.00000000544,
       0.00000000388,     0.00000001689,
      -0.00000000025,    -0.00000000398,
      -0.00000000184,     0.00000000024,
       0.00000000524,    -0.00000000132,
      -0.00000000063,    -0.00000000223,
       0.00000000112,     0.00000000480,
       0.00000000222,    -0.00000000057,
      -0.00000000049,    -0.00000000086,
       0.00000002356,    -0.00000001280,
       0.00000000112,     0.00000000215,
       0.00000000065,    -0.00000000038,
      -0.00000000218,     0.00000000051,
       0.00000000015,     0.00000000101,
       0.00000000023,    -0.00000000228,
      -0.00000000105,     0.00000000007,
       0.00000000021,    -0.00000000958,
       0.00000000317,     0.00000000033,
       0.00000000166,     0.00000006271,
      -0.00000000757,    -0.00000000009,
       0.00000000024,     0.00000000123,
       0.00000000116,    -0.00000000108,
       0.00000000046,     0.00000000198,
       0.00000001102,    -0.00000001303,
       0.00000002584,     0.00000002300,
      -0.00000003757,     0.00000004221,
      -0.00000018796,    -0.00000015551,
      -0.00000002693,     0.00000001402,
       0.00000005307,    -0.00000006486,
      -0.00000000399,    -0.00000000767,
       0.00000000133,     0.00000001319,
      -0.00000001104,    -0.00000000057,
      -0.00000000132,     0.00000000110,
       0.00000000427,     0.00000000617,
       0.00000000380,    -0.00000000262,
      -0.00000003517,    -0.00000003989,
      -0.00000000231,     0.00000000202,
       0.00000000299,     0.00000000246,
      -0.00000000207,     0.00000000286,
      -0.00000000063,    -0.00000000067,
       0.00000000069,    -0.00000000063,
      -0.00000315133,     0.00000103118,
       0.00000114415,     0.00000239245,
       0.00000090246,    -0.00000058948,
      -0.00000019472,    -0.00000022147,
      -0.00000003980,     0.00000004907,
       0.00000003117,     0.00000009470,
       0.00000007179,    -0.00000003449,
      -0.00000001774,    -0.00000002705,
      -0.00000000663,     0.00000000585,
       0.00000000147,     0.00000000119,
       0.00000000003,     0.00000000003,
       0.00000000003,    -0.00000000003,
       0.00000003108,    -0.00000005689,
      -0.00000004206,    -0.00000001472,
      -0.00000000302,     0.00000001543,
      -0.00000000444,    -0.00000000785,
      -0.00000000727,     0.00000000455,
       0.00000000246,     0.00000000294,
      -0.00000004062,     0.00000017905,
       0.00000012207,     0.00000001245,
      -0.00000000091,    -0.00000004193,
      -0.00000000939,     0.00000000158,
       0.00000017839,     0.00000004144,
       0.00000001301,    -0.00000012168,
      -0.00000004185,     0.00000000100,
       0.00000000149,     0.00000000934,
      -0.00000000178,     0.00000000609,
       0.00000000241,     0.00000000034,
       0.00000000613,     0.00000000137,
      -0.00000000002,    -0.00000000248,
      -0.00000228376,     0.00000115819,
       0.00000110144,     0.00000166313,
       0.00000060710,    -0.00000051270,
      -0.00000015706,    -0.00000014486,
      -0.00000002498,     0.00000003585,
       0.00000003857,     0.00000010966,
       0.00000008701,    -0.00000004279,
      -0.00000002241,    -0.00000003392,
      -0.00000000863,     0.00000000746,
       0.00000000181,     0.00000000151,
      -0.00000004833,    -0.00000064644,
      -0.00000041027,     0.00000006480,
       0.00000003389,     0.00000013307,
       0.00000002876,    -0.00000001068,
      -0.00000000236,    -0.00000000427,
      -0.00000063774,     0.00000004519,
       0.00000006064,     0.00000040298,
       0.00000013016,    -0.00000003148,
      -0.00000000994,    -0.00000002804,
      -0.00000000372,     0.00000000218,
       0.00000000402,    -0.00000001484,
      -0.00000000608,    -0.00000000051,
      -0.00000001463,    -0.00000000402,
      -0.00000000060,     0.00000000597,
       0.00000000213,    -0.00000000015,
      -0.00000000055,    -0.00000000120,
      -0.00000000013,    -0.00000000217,
      -0.00000000121,     0.00000000052,
       0.00000000269,    -0.00000000017,
      -0.00000000067,    -0.00000000151,
      -0.00000000017,    -0.00000000269,
      -0.00000000151,     0.00000000067,
      -0.00000000222,     0.00000000340,
       0.00000000346,     0.00000000226,
      -0.00000000109,     0.00000000167,
       0.00000000166,     0.00000000109,
      -0.00000000136,     0.00000000208,
       0.00000000208,     0.00000000136,
      -0.00000010191,     0.00000013395,
       0.00000000346,     0.00000000263,
      -0.00000000132,     0.00000000161,
      -0.00000000161,    -0.00000000132,
      -0.00000002084,     0.00000004333,
       0.00000000159,     0.00000000075,
       0.00000006860,    -0.00000000330,
      -0.00000001623,    -0.00000003895,
      -0.00000000041,    -0.00000000237,
      -0.00000000120,     0.00000000088,
       0.00000001005,    -0.00000002135,
       0.00000000084,    -0.00000000110,
       0.00000000065,    -0.00000000050,
      -0.00000000551,     0.00000000520,
      -0.00000000042,    -0.00000000383,
      -0.00000000166,     0.00000000093,
       0.00000012873,    -0.00000004031,
      -0.00000001851,    -0.00000005618,
       0.00000000474,     0.00000001572,
       0.00000000664,    -0.00000000253,
      -0.00000000062,    -0.00000000152,
      -0.00000000033,    -0.00000000073,
      -0.00000000005,     0.00000000160,
      -0.00000002378,     0.00000001036,
       0.00000000958,     0.00000001118,
       0.00000000253,    -0.00000000310,
      -0.00000003199,    -0.00000003354,
      -0.00000001442,     0.00000002279,
       0.00000000597,     0.00000000312,
       0.00000005652,    -0.00000000646,
      -0.00000001682,    -0.00000002754,
      -0.00000000553,     0.00000000753,
       0.00000004555,    -0.00000000038,
       0.00000000111,    -0.00000002516,
      -0.00000000745,    -0.00000000068,
       0.00000032617,     0.00000018497,
       0.00000003950,    -0.00000020695,
      -0.00000005631,     0.00000000487,
       0.00000007730,     0.00000008383,
       0.00000003452,    -0.00000005467,
      -0.00000001550,    -0.00000000673,
      -0.00000000080,     0.00000000152,
      -0.00000005091,    -0.00000004529,
      -0.00000002237,     0.00000001437,
       0.00000000272,     0.00000000616,
       0.00000000815,     0.00000011150,
       0.00000004678,    -0.00000002255,
      -0.00000000839,    -0.00000000944,
      -0.00000000048,    -0.00000000106,
      -0.00000000068,     0.00000000089,
      -0.00000008172,     0.00000001437,
       0.00000000835,     0.00000003601,
      -0.00000000215,    -0.00000001365,
      -0.00000000538,     0.00000000131,
       0.00000000020,     0.00000000084,
       0.00000001110,    -0.00000000400,
      -0.00000000175,    -0.00000000464,
      -0.00000000087,     0.00000000038,
       0.00000000259,     0.00000001020,
       0.00000000455,    -0.00000000156,
      -0.00000000033,    -0.00000000091,
       0.00000000365,    -0.00000000257,
       0.00000001271,     0.00000001799,
       0.00000000870,    -0.00000000809,
      -0.00000000076,     0.00000000287,
      -0.00000004105,     0.00000003054,
       0.00000001118,     0.00000001200,
      -0.00000003510,    -0.00000005000,
      -0.00000001378,     0.00000001223,
      -0.00000000104,    -0.00000000175,
       0.00000000133,    -0.00000001277,
      -0.00000001337,     0.00000000311,
       0.00000019186,    -0.00000012447,
      -0.00000004447,    -0.00000005503,
       0.00000003181,     0.00000005177,
       0.00000001630,    -0.00000001174,
      -0.00000000241,    -0.00000000230,
      -0.00000001430,     0.00000000491,
       0.00000000200,     0.00000000613,
       0.00000000034,    -0.00000001538,
      -0.00000000636,     0.00000000138,
      -0.00000001441,    -0.00000000117,
      -0.00000000117,     0.00000000303,
      -0.00000000245,     0.00000004433,
       0.00000000860,     0.00000000200,
       0.00000003376,     0.00000000006,
       0.00000000154,    -0.00000000664,
      -0.00000000207,    -0.00000000401,
      -0.00000000402,     0.00000000312,
      -0.00000001454,    -0.00000020299,
      -0.00000003269,     0.00000000125,
      -0.00000001437,     0.00000000202,
       0.00000000011,     0.00000000229,
      -0.00000000455,     0.00000000455,
       0.00000000123,     0.00000000069,
      -0.00000000281,    -0.00000000755,
      -0.00000000077,     0.00000000115,
      -0.00000000037,    -0.00000000121,
       0.00000000149,    -0.00000000069,
      -0.00000000056,    -0.00000000155,
       0.00000001197,     0.00000000115,
       0.00000000077,     0.00000000174,
      -0.00000000311,     0.00000000269,
       0.00000000346,     0.00000000315,
       0.00000000081,    -0.00000000080,
      -0.00000000993,    -0.00000001056,
      -0.00000000269,     0.00000000227,
       0.00000000173,     0.00000000146,
       0.00000000147,    -0.00000000180,
       0.00000005476,     0.00000003536,
      -0.00000000069,    -0.00000000338,
       0.00000000290,    -0.00000000021,
      -0.00000000065,     0.00000000081,
       0.00000000134,     0.00000000128,
       0.00000000079,    -0.00000000073,
      -0.00000001462,    -0.00000001482,
       0.00000000146,     0.00000000090,
      -0.00000000079,     0.00000000141,
       0.00000012058,    -0.00000000783,
      -0.00000003019,    -0.00000006801,
      -0.00000001809,     0.00000001622,
      -0.00000000025,    -0.00000000338,
      -0.00000000189,     0.00000000087,
       0.00000000055,    -0.00000000125,
      -0.00000000123,    -0.00000000062,
       0.00000000136,    -0.00000000112,
      -0.00000000053,     0.00000000055,
       0.00000002437,    -0.00000004711,
      -0.00000000101,    -0.00000000049,
      -0.00000000045,    -0.00000000100,
       0.00000000100,    -0.00000000045,
       0.00000000009,    -0.00000000012,
      -0.00000000001,    -0.00000000006,
      -0.00000000010,    -0.00000000001,
       0.00000000071,     0.00000000049,
       0.00000000047,    -0.00000000069,
       0.00000077037,     0.00000081360,
       0.00000040088,    -0.00000047938,
      -0.00000014975,    -0.00000010025,
      -0.00000001644,     0.00000003127,
       0.00000000495,     0.00000000185,
      -0.00000061984,     0.00000075769,
       0.00000046205,     0.00000026854,
       0.00000005533,    -0.00000014015,
      -0.00000002867,    -0.00000000678,
      -0.00000000067,     0.00000000493,
      -0.00000167826,    -0.00000179154,
      -0.00000084043,     0.00000101494,
       0.00000030610,     0.00000018999,
       0.00000002548,    -0.00000006102,
      -0.00000000797,    -0.00000000263,
      -0.00000179784,     0.00000164948,
       0.00000099479,     0.00000084769,
       0.00000019376,    -0.00000029873,
      -0.00000005910,    -0.00000002718,
      -0.00000000192,     0.00000000877,
      -0.00000002152,    -0.00000002022,
      -0.00000000439,     0.00000000504,
      -0.00000001564,     0.00000002443,
       0.00000000672,     0.00000000228,
      -0.00000578542,    -0.00001302156,
      -0.00000747190,     0.00000400174,
       0.00000140485,     0.00000221660,
       0.00000044254,    -0.00000033292,
      -0.00000005936,    -0.00000006502,
      -0.00000000784,     0.00000000768,
       0.00000156011,    -0.00000191131,
      -0.00000116637,    -0.00000067542,
      -0.00000013810,     0.00000035388,
       0.00000007034,     0.00000001536,
      -0.00000000042,    -0.00000001006,
       0.00000536815,     0.00000118144,
       0.00000056320,    -0.00000270033,
      -0.00000070825,    -0.00000011700,
      -0.00000001237,     0.00000012578,
       0.00000001673,    -0.00000000052,
       0.00000113759,    -0.00000537192,
      -0.00000270052,    -0.00000053151,
      -0.00000010533,     0.00000070803,
       0.00000012544,     0.00000000951,
      -0.00000000079,    -0.00000001616,
       0.00000005020,     0.00000004680,
       0.00000001014,    -0.00000001085,
      -0.00000000058,    -0.00000000096,
       0.00000004186,    -0.00000005748,
      -0.00000001435,    -0.00000000720,
      -0.00000000011,     0.00000000150,
      -0.00000000015,     0.00000000116,
       0.00000000081,    -0.00000000023,
      -0.00000000632,     0.00000000626,
       0.00000000296,     0.00000000233,
       0.00000000627,     0.00000000633,
       0.00000000231,    -0.00000000310,
      -0.00000000738,     0.00000000752,
       0.00000000354,     0.00000000263,
       0.00000000042,    -0.00000000083,
       0.00000000750,     0.00000000737,
       0.00000000264,    -0.00000000355,
      -0.00000000083,    -0.00000000044,
      -0.00000000079,     0.00000000065,
       0.00000000065,     0.00000000079,
       0.00000000101,    -0.00000000092,
      -0.00000000092,    -0.00000000101,
      -0.00000276630,    -0.00000180710,
      -0.00000000490,     0.00000000760,
       0.00000547900,     0.00000357930,
       0.00000128390,     0.00000155510,
       0.00000005120,    -0.00000004840,
      -0.00000197170,    -0.00000208760,
      -0.00000000100,     0.00000000091,
      -0.00000000091,    -0.00000000100,
      -0.00000003220,    -0.00000002850,
       0.00000008640,     0.00000006040,
      -0.00000002570,    -0.00000001660,
       0.00000000079,    -0.00000000065,
       0.00000000065,     0.00000000079,
       0.00000000730,    -0.00000000046,
      -0.00000000129,    -0.00000000236,
      -0.00000000410,     0.00000001495,
      -0.00000001476,    -0.00000000292,
      -0.00000000098,     0.00000000536,
      -0.00000000617,     0.00000000113,
       0.00000000114,     0.00000000619,
      -0.00000021499,     0.00000022575,
       0.00000010620,     0.00000007761,
       0.00000001229,    -0.00000002738,
       0.00000001480,     0.00000001761,
       0.00000000651,    -0.00000000697,
      -0.00000000132,    -0.00000000126,
      -0.00000000153,    -0.00000000066,
       0.00000001980,    -0.00000000294,
      -0.00000000244,    -0.00000000761,
      -0.00000001622,    -0.00000009642,
      -0.00000003528,     0.00000001109,
       0.00000000724,     0.00000001257,
      -0.00000001034,     0.00000002081,
       0.00000000539,     0.00000000218,
       0.00000002089,     0.00000000926,
       0.00000000100,    -0.00000000617,
      -0.00000027026,     0.00000004352,
       0.00000003383,     0.00000010431,
       0.00000002117,    -0.00000001221,
       0.00000004069,     0.00000024665,
       0.00000009409,    -0.00000002979,
      -0.00000001225,    -0.00000002093,
       0.00000000128,    -0.00000000082,
       0.00000004067,    -0.00000009255,
      -0.00000002477,    -0.00000000791,
      -0.00000009157,    -0.00000003481,
      -0.00000000391,     0.00000002647,
       0.00000028250,    -0.00000006139,
      -0.00000003946,    -0.00000011214,
      -0.00000002061,     0.00000001024,
      -0.00000000235,    -0.00000001996,
      -0.00000000777,     0.00000000175,
      -0.00000000939,     0.00000000803,
       0.00000000362,     0.00000000340,
       0.00000000117,     0.00000000078,
      -0.00000000586,    -0.00000000706,
      -0.00000000303,     0.00000000299,
      -0.00000000069,     0.00000000129,
      -0.00000004718,     0.00000002054,
       0.00000001042,     0.00000001122,
       0.00000002588,     0.00000009464,
       0.00000002541,    -0.00000001517,
      -0.00000000390,    -0.00000000329,
       0.00000008826,    -0.00000002713,
      -0.00000002198,    -0.00000001500,
      -0.00000000054,     0.00000000477,
       0.00000000499,    -0.00000009246,
      -0.00000002203,     0.00000000556,
       0.00000000129,     0.00000000325,
       0.00000015746,    -0.00000056742,
      -0.00000019034,     0.00000007433,
       0.00000002844,     0.00000002581,
      -0.00000006171,    -0.00000022468,
      -0.00000006177,     0.00000003670,
       0.00000000874,     0.00000000892,
       0.00000000130,    -0.00000000072,
       0.00000015246,     0.00000000649,
      -0.00000000572,    -0.00000003347,
      -0.00000000378,     0.00000000029,
      -0.00000006723,    -0.00000016876,
      -0.00000003236,     0.00000004714,
       0.00000001004,     0.00000000204,
       0.00000000077,    -0.00000000029,
      -0.00000000071,    -0.00000000079,
      -0.00000008805,     0.00000006199,
       0.00000002169,     0.00000002632,
      -0.00000000108,    -0.00000000075,
       0.00000000393,    -0.00000000194,
      -0.00000000027,    -0.00000000140,
       0.00000000003,     0.00000000445,
       0.00000000143,    -0.00000000027,
      -0.00000000361,     0.00000000014,
       0.00000000095,     0.00000000091,
       0.00000000086,    -0.00000000112,
       0.00000000540,     0.00000003588,
       0.00000000836,    -0.00000000008,
      -0.00000000012,    -0.00000000204,
      -0.00000000457,    -0.00000000368,
      -0.00000000333,     0.00000000563,
      -0.00000000349,    -0.00000006832,
       0.00000000239,    -0.00000000025,
      -0.00000000139,     0.00000000162,
      -0.00000000126,    -0.00000000242,
       0.00000000061,    -0.00000000050,
      -0.00000001245,    -0.00000001165,
      -0.00000000186,     0.00000000109,
       0.00000000064,     0.00000000218,
       0.00000000198,    -0.00000000067,
       0.00000003843,     0.00000001750,
      -0.00000000070,     0.00000000167,
      -0.00000000063,    -0.00000000218,
       0.00000000192,    -0.00000000037,
       0.00000000512,     0.00000000186,
       0.00000000133,    -0.00000000574,
      -0.00000000560,    -0.00000000157,
      -0.00000001423,    -0.00000002939,
      -0.00000001255,    -0.00000000535,
      -0.00000000134,     0.00000000067,
       0.00000000131,    -0.00000000594,
      -0.00000000612,    -0.00000000140,
       0.00000001823,     0.00000001622,
       0.00000000136,     0.00000000063,
      -0.00000000057,     0.00000000134,
       0.00000000183,     0.00000000094,
       0.00000000080,    -0.00000000232,
      -0.00000000228,    -0.00000000108,
      -0.00000000428,    -0.00000001368,
      -0.00000000532,    -0.00000000350,
       0.00000000094,    -0.00000000261,
      -0.00000000251,    -0.00000000079,
      -0.00000000088,     0.00000000015,
      -0.00000038907,     0.00000039935,
       0.00000018798,     0.00000013921,
       0.00000002290,    -0.00000004541,
       0.00000000877,     0.00000000885,
       0.00000000320,    -0.00000000409,
      -0.00000000097,    -0.00000000055,
       0.00000000621,    -0.00000001057,
       0.00000059129,     0.00000034477,
       0.00000005200,    -0.00000013130,
      -0.00000007539,     0.00000012726,
       0.00000003000,     0.00000001220,
      -0.00000157933,    -0.00000091560,
      -0.00000014013,     0.00000034385,
      -0.00000002314,    -0.00000003582,
      -0.00000003734,     0.00000002073,
      -0.00000001598,     0.00000002271,
      -0.00000004527,    -0.00000003184,
       0.00000027894,    -0.00000043191,
      -0.00000010029,    -0.00000003817,
       0.00000032405,     0.00000020433,
       0.00000002476,    -0.00000006505,
      -0.00000001552,     0.00000000406,
       0.00000000360,     0.00000001100,
      -0.00000000018,    -0.00000002069,
      -0.00000001140,     0.00000000300,
      -0.00000000416,     0.00000000239,
       0.00000000438,     0.00000000886,
       0.00000000840,    -0.00000000416,
      -0.00000002163,    -0.00000003623,
      -0.00000000043,     0.00000000100,
      -0.00000000098,    -0.00000000082,
       0.00000000000,     0.00000000132,
       0.00000000004,     0.00000000021,
      -0.00000000007,    -0.00000000007,
      -0.00000000006,     0.00000000006,
       0.00000000001,     0.00000000006,
       0.00000000052,     0.00000000075,
       0.00000000084,     0.00000000147,
      -0.00000000035,    -0.00000000019,
      -0.00000000094,     0.00000000016,
       0.00000000017,     0.00000000095,
       0.00000000191,     0.00000000179,
      -0.00000000040,    -0.00000000041,
      -0.00000000150,     0.00000000015,
       0.00000000018,     0.00000000129,
       0.00000388670,    -0.00000307275,
      -0.00000144510,    -0.00000149855,
      -0.00000028740,     0.00000034951,
       0.00000005597,     0.00000003453,
       0.00000000287,    -0.00000000711,
      -0.00000502389,    -0.00000800623,
      -0.00000324664,     0.00000235511,
       0.00000057506,     0.00000066975,
       0.00000009216,    -0.00000009362,
      -0.00000001131,    -0.00000000457,
      -0.00000748581,     0.00000139673,
       0.00000091727,     0.00000287942,
       0.00000056333,    -0.00000027094,
      -0.00000004988,    -0.00000007115,
      -0.00000000624,     0.00000000647,
       0.00000132346,     0.00000746146,
       0.00000286924,    -0.00000087673,
      -0.00000025942,    -0.00000056170,
      -0.00000007065,     0.00000004764,
       0.00000000603,     0.00000000663,
       0.00000013559,    -0.00000002609,
      -0.00000001391,    -0.00000002082,
      -0.00000000251,     0.00000000288,
      -0.00000000724,    -0.00000014278,
      -0.00000002373,     0.00000000681,
       0.00000000135,     0.00000000285,
      -0.00002303941,     0.00007280618,
       0.00003423240,     0.00000249545,
      -0.00000144208,    -0.00000807212,
      -0.00000122456,     0.00000058534,
       0.00000011503,     0.00000012953,
       0.00000000844,    -0.00000001598,
       0.00001239422,     0.00001900289,
       0.00000765425,    -0.00000580745,
      -0.00000141333,    -0.00000156486,
      -0.00000020870,     0.00000023057,
       0.00000002756,     0.00000001959,
      -0.00002972358,     0.00000372205,
       0.00000308022,     0.00001158241,
       0.00000227431,    -0.00000098505,
      -0.00000019018,    -0.00000028602,
      -0.00000002506,     0.00000002547,
       0.00000000170,     0.00000000205,
       0.00000368027,     0.00002993802,
       0.00001170231,    -0.00000307839,
      -0.00000099089,    -0.00000230675,
      -0.00000029156,     0.00000019017,
       0.00000002647,     0.00000002545,
       0.00000000098,    -0.00000000358,
      -0.00000034055,     0.00000007151,
       0.00000004107,     0.00000005132,
       0.00000000547,    -0.00000000874,
       0.00000003396,     0.00000033688,
       0.00000005162,    -0.00000002535,
      -0.00000000524,    -0.00000000533,
       0.00000001492,     0.00000009887,
       0.00000004609,    -0.00000001970,
      -0.00000000166,    -0.00000000109,
       0.00000000060,    -0.00000000042,
       0.00000000052,     0.00000000048,
      -0.00000000757,     0.00000000085,
       0.00000000065,     0.00000000287,
       0.00000000093,    -0.00000000043,
       0.00000000136,     0.00000000692,
       0.00000000244,    -0.00000000086,
      -0.00000000056,    -0.00000000115,
       0.00000004469,     0.00000021769,
       0.00000010214,    -0.00000004678,
      -0.00000001751,    -0.00000002423,
      -0.00000000899,     0.00000000080,
       0.00000000075,     0.00000000365,
       0.00000000077,    -0.00000000027,
       0.00000000166,     0.00000000808,
       0.00000000306,    -0.00000000113,
      -0.00000001107,     0.00000000151,
      -0.00000001216,     0.00000000172,
      -0.00000000096,    -0.00000000057,
      -0.00000000033,     0.00000000081,
       0.00000002094,    -0.00000004319,
       0.00000000159,     0.00000000077,
       0.00000000132,    -0.00000000161,
      -0.00000000161,    -0.00000000132,
       0.00000010193,    -0.00000013405,
       0.00000000346,     0.00000000263,
       0.00000000224,    -0.00000000343,
       0.00000000343,     0.00000000226,
       0.00000000109,    -0.00000000166,
       0.00000000166,     0.00000000109,
       0.00000000136,    -0.00000000208,
       0.00000000208,     0.00000000136,
      -0.00000001429,    -0.00000001256,
       0.00000000349,    -0.00000000192,
      -0.00000000168,    -0.00000000159,
       0.00000000055,    -0.00000000097,
       0.00000000131,    -0.00000000099,
      -0.00000000181,    -0.00000000042,
      -0.00000000133,    -0.00000000187,
      -0.00000000097,     0.00000000167,
      -0.00000000995,     0.00000002095,
       0.00000000045,     0.00000000266,
       0.00000000093,    -0.00000000214,
      -0.00000000213,    -0.00000000093,
       0.00000003399,    -0.00000000440,
      -0.00000000429,    -0.00000001311,
      -0.00000000052,    -0.00000000481,
      -0.00000000246,     0.00000000079,
      -0.00000000392,     0.00000000825,
       0.00000000229,     0.00000000064,
       0.00000000838,     0.00000000376,
       0.00000000055,    -0.00000000222,
       0.00000005650,    -0.00000001280,
      -0.00000000784,    -0.00000002124,
       0.00000000210,     0.00000001240,
       0.00000000484,    -0.00000000152,
      -0.00000000037,    -0.00000000072,
      -0.00000000033,     0.00000000088,
       0.00000000077,    -0.00000000024,
       0.00000000045,     0.00000000066,
      -0.00000000700,     0.00000001637,
       0.00000000460,     0.00000000111,
       0.00000001471,     0.00000000535,
       0.00000000086,    -0.00000000409,
      -0.00000000681,    -0.00000000982,
      -0.00000000115,     0.00000000106,
      -0.00000000941,     0.00000000722,
       0.00000000134,     0.00000000099,
       0.00000006167,    -0.00000015891,
      -0.00000004504,    -0.00000000780,
      -0.00000003632,    -0.00000001384,
      -0.00000000286,     0.00000001035,
       0.00000000155,     0.00000000000,
       0.00000002901,     0.00000004098,
       0.00000000487,    -0.00000000453,
       0.00000004028,    -0.00000003081,
      -0.00000000585,    -0.00000000434,
      -0.00000007456,     0.00000004868,
       0.00000001748,     0.00000002421,
      -0.00000000146,    -0.00000000111,
       0.00000000289,    -0.00000000103,
      -0.00000000040,    -0.00000000117,
      -0.00000000006,     0.00000000311,
       0.00000000120,    -0.00000000024,
       0.00000000110,    -0.00000000639,
      -0.00000000134,     0.00000000056,
       0.00000000529,    -0.00000001247,
      -0.00000000172,    -0.00000000048,
      -0.00000001412,    -0.00000000163,
       0.00000000182,     0.00000000205,
       0.00000342603,    -0.00000053306,
      -0.00000055008,    -0.00000069936,
      -0.00000006950,     0.00000011793,
       0.00000002599,     0.00000020787,
       0.00000004221,    -0.00000003235,
      -0.00000000680,    -0.00000000405,
      -0.00000001890,     0.00000003427,
       0.00000000476,     0.00000000225,
       0.00000004435,     0.00000000820,
      -0.00000000394,    -0.00000000708,
       0.00000002202,    -0.00000005489,
      -0.00000001235,    -0.00000000750,
       0.00000000250,    -0.00000000180,
       0.00000000036,     0.00000000185,
      -0.00000000066,    -0.00000000374,
      -0.00000000079,     0.00000000026,
       0.00000000382,     0.00000000247,
       0.00000000152,    -0.00000000237,
      -0.00000000098,    -0.00000000011,
      -0.00000003843,    -0.00000002413,
      -0.00000000353,     0.00000000542,
       0.00000000379,     0.00000000056,
       0.00000000018,    -0.00000000364,
       0.00000009524,     0.00000002834,
      -0.00000000097,     0.00000000354,
      -0.00000000069,    -0.00000000181,
       0.00000000093,     0.00000000091,
      -0.00000000104,     0.00000000041,
       0.00000000669,     0.00000001601,
       0.00000000267,    -0.00000000101,
      -0.00000000141,    -0.00000000139,
      -0.00000000117,     0.00000000109,
      -0.00000004418,    -0.00000003283,
       0.00000000123,    -0.00000000157,
       0.00000000146,     0.00000000049,
      -0.00000000045,     0.00000000145,
      -0.00000000114,    -0.00000000034,
       0.00000004298,    -0.00000017212,
       0.00000001291,     0.00000000330,
      -0.00000000056,    -0.00000000047,
      -0.00000000102,     0.00000000023,
       0.00000002708,    -0.00000007235,
       0.00000000561,     0.00000000218,
       0.00000001188,    -0.00000001857,
       0.00000000142,     0.00000000093,
      -0.00000002440,    -0.00000002261,
      -0.00000000060,     0.00000000042,
       0.00000000929,    -0.00000000072,
      -0.00000000018,    -0.00000000231,
      -0.00000000124,     0.00000000243,
       0.00000000246,     0.00000000108,
      -0.00000018585,     0.00000002807,
       0.00000002079,     0.00000007237,
       0.00000000188,     0.00000001292,
       0.00000000500,    -0.00000000140,
      -0.00000000048,    -0.00000000109,
      -0.00000000013,     0.00000000114,
       0.00000000190,    -0.00000000100,
      -0.00000000193,     0.00000000138,
      -0.00000000127,     0.00000000071,
      -0.00000000052,     0.00000000108,
      -0.00000000107,    -0.00000000062,
       0.00000000003,     0.00000000028,
       0.00000000011,    -0.00000000002,
       0.00000000004,    -0.00000000002,
      -0.00000000003,    -0.00000000004,
       0.00000000003,     0.00000000077,
       0.00000000055,    -0.00000000017,
       0.00000000001,     0.00000000253,
       0.00000000107,    -0.00000000031,
      -0.00000000023,     0.00000000062,
       0.00000000064,     0.00000000009,
       0.00000592681,    -0.00000044217,
      -0.00000052750,    -0.00000233773,
      -0.00000046140,     0.00000018227,
       0.00000003586,     0.00000005865,
       0.00000000536,    -0.00000000496,
      -0.00000073345,    -0.00000353503,
      -0.00000137730,     0.00000044835,
       0.00000012823,     0.00000027076,
       0.00000003424,    -0.00000002355,
      -0.00000000329,    -0.00000000311,
      -0.00000358955,     0.00000703109,
       0.00000201144,     0.00000070791,
       0.00000005248,    -0.00000029142,
      -0.00000002763,     0.00000000073,
       0.00000711716,     0.00000344551,
       0.00000062873,    -0.00000205284,
      -0.00000030054,    -0.00000003023,
       0.00000000474,     0.00000002888,
       0.00000000229,    -0.00000000140,
      -0.00000006772,     0.00000013406,
       0.00000000394,    -0.00000000107,
       0.00000013133,     0.00000006239,
      -0.00000000216,    -0.00000000323,
      -0.00039953998,     0.00005795591,
       0.00004399313,     0.00015521642,
       0.00003047491,    -0.00001354233,
      -0.00000253790,    -0.00000384880,
      -0.00000034387,     0.00000033612,
       0.00000003173,     0.00000001936,
      -0.00000122279,    -0.00001041524,
      -0.00000401656,     0.00000108316,
       0.00000034618,     0.00000078525,
       0.00000009796,    -0.00000006593,
      -0.00000000870,    -0.00000000861,
       0.00001600001,    -0.00003103257,
      -0.00000881095,    -0.00000324975,
      -0.00000026148,     0.00000127195,
       0.00000011985,     0.00000000200,
      -0.00000000106,    -0.00000000821,
      -0.00003173177,    -0.00001577700,
      -0.00000306258,     0.00000913775,
       0.00000134745,     0.00000019550,
      -0.00000001234,    -0.00000013073,
      -0.00000000975,     0.00000000448,
       0.00000012056,    -0.00000026015,
       0.00000004249,     0.00000002617,
       0.00000000432,    -0.00000001319,
      -0.00000000253,    -0.00000000038,
      -0.00000024599,    -0.00000012725,
       0.00000002353,    -0.00000004579,
      -0.00000001360,    -0.00000000391,
      -0.00000000037,     0.00000000252,
      -0.00000128785,     0.00000029025,
       0.00000016850,     0.00000048329,
       0.00000009225,    -0.00000004588,
      -0.00000000828,    -0.00000001165,
      -0.00000000142,    -0.00000000194,
      -0.00000000075,     0.00000000050,
       0.00000004248,    -0.00000003658,
      -0.00000001324,    -0.00000001162,
      -0.00000000149,     0.00000000179,
      -0.00000002665,    -0.00000003541,
      -0.00000000931,     0.00000000865,
       0.00000000197,     0.00000000134,
      -0.00000244338,     0.00000038011,
       0.00000027681,     0.00000094706,
       0.00000018586,    -0.00000008407,
      -0.00000001586,    -0.00000002361,
      -0.00000000011,    -0.00000000098,
       0.00000000673,    -0.00000000941,
      -0.00000000384,    -0.00000000200,
       0.00000000317,     0.00000000291,
       0.00000000120,    -0.00000000204,
       0.00000000392,    -0.00000000721,
       0.00000000446,    -0.00000000800,
       0.00000000016,     0.00000000294,
       0.00000000147,    -0.00000000127,
       0.00000002257,    -0.00000000540,
      -0.00000000866,    -0.00000001159,
      -0.00000000037,    -0.00000000663,
      -0.00000000401,     0.00000000194,
       0.00000000159,     0.00000000096,
       0.00000000227,     0.00000001150,
       0.00000000459,    -0.00000000295,
       0.00000001163,    -0.00000000238,
      -0.00000000292,    -0.00000000458,
      -0.00000000123,     0.00000000215,
       0.00000000019,    -0.00000000186,
      -0.00000000186,    -0.00000000009,
       0.00000000222,    -0.00000000242,
      -0.00000000160,    -0.00000000037,
      -0.00000000243,    -0.00000000225,
      -0.00000000038,     0.00000000151,
      -0.00000000107,    -0.00000000059,
      -0.00000000058,     0.00000000109,
       0.00000017445,    -0.00000016304,
      -0.00000007846,    -0.00000006059,
      -0.00000000964,     0.00000001920,
       0.00000000614,     0.00000000447,
       0.00000000146,    -0.00000000312,
      -0.00000001022,     0.00000000316,
       0.00000000154,     0.00000000359,
       0.00000000100,    -0.00000000056,
       0.00000000307,     0.00000001012,
       0.00000000355,    -0.00000000151,
      -0.00000000056,    -0.00000000101,
       0.00000000267,    -0.00000000507,
      -0.00000000334,    -0.00000000134,
       0.00000000191,     0.00000000309,
       0.00000000300,    -0.00000000201,
      -0.00000002164,     0.00000004569,
       0.00000001320,     0.00000000359,
       0.00000000796,     0.00000000349,
       0.00000000057,    -0.00000000234,
      -0.00000000777,    -0.00000001276,
      -0.00000000141,     0.00000000127,
      -0.00000001261,     0.00000000795,
       0.00000000136,     0.00000000144,
       0.00000153191,    -0.00000135577,
      -0.00000045281,    -0.00000042123,
      -0.00000005741,     0.00000007340,
       0.00000001432,     0.00000001824,
       0.00000000499,    -0.00000000473,
      -0.00000000077,    -0.00000000073,
      -0.00000014602,     0.00000002513,
       0.00000002234,     0.00000003466,
       0.00000000422,    -0.00000000396,
      -0.00000000084,    -0.00000000044,
       0.00000002269,     0.00000013272,
       0.00000002820,    -0.00000002037,
      -0.00000000336,    -0.00000000293,
      -0.00000000028,     0.00000000073,
      -0.00000002444,    -0.00000003419,
      -0.00000000377,     0.00000000423,
      -0.00000001901,     0.00000001464,
       0.00000000237,     0.00000000212,
       0.00000001437,    -0.00000000147,
      -0.00000000210,    -0.00000001423,
       0.00000020552,     0.00000029228,
       0.00000003251,    -0.00000003585,
       0.00000004490,    -0.00000003525,
      -0.00000000600,    -0.00000000535,
      -0.00000006068,     0.00000000621,
       0.00000000093,    -0.00000000036,
       0.00000000848,     0.00000006085,
      -0.00000000041,    -0.00000000122,
      -0.00000420620,     0.00000076109,
       0.00000064262,     0.00000097606,
       0.00000011250,    -0.00000010120,
      -0.00000002612,    -0.00000001221,
      -0.00000003054,    -0.00000017179,
      -0.00000003972,     0.00000002648,
       0.00000000389,     0.00000000439,
       0.00000000047,    -0.00000000102,
       0.00000000301,    -0.00000000617,
      -0.00000000184,    -0.00000000117,
       0.00000000804,     0.00000000992,
       0.00000000262,    -0.00000000311,
       0.00000000058,    -0.00000000112,
      -0.00000000112,    -0.00000000058,
       0.00000004741,    -0.00000000732,
       0.00000000021,     0.00000000142,
      -0.00000000010,    -0.00000000008,
      -0.00000000031,    -0.00000000038,
      -0.00000000024,    -0.00000000031,
      -0.00000000108,     0.00000000026,
      -0.00000000022,    -0.00000000107,
       0.00000001482,    -0.00000005170,
      -0.00000000722,     0.00000000303,
      -0.00000002667,    -0.00000000648,
       0.00000000172,     0.00000000366,
       0.00000000959,     0.00000001189,
       0.00000001114,    -0.00000001171,
      -0.00000000080,     0.00000000055,
      -0.00000012586,     0.00000040693,
       0.00000005715,    -0.00000002320,
       0.00000008155,     0.00000001829,
      -0.00000000527,    -0.00000001114,
       0.00000000244,    -0.00000005693,
      -0.00000000673,    -0.00000000707,
      -0.00000000087,     0.00000000206,
      -0.00000004112,     0.00000009239,
       0.00000001152,    -0.00000000873,
      -0.00000000173,    -0.00000000092,
       0.00000004679,     0.00000000857,
       0.00000000116,     0.00000000548,
       0.00000000053,    -0.00000000079,
       0.00000005908,    -0.00000001109,
      -0.00000000973,    -0.00000001167,
      -0.00000000120,     0.00000000235,
       0.00000001099,     0.00000005924,
       0.00000001167,    -0.00000000974,
      -0.00000000235,    -0.00000000120,
       0.00000000068,    -0.00000000084,
       0.00000007907,    -0.00000001220,
      -0.00000001268,    -0.00000001557,
      -0.00000000161,     0.00000000304,
       0.00000001226,     0.00000007918,
       0.00000001558,    -0.00000001269,
      -0.00000000305,    -0.00000000161,
       0.00000001668,    -0.00000003276,
      -0.00000000359,     0.00000001041,
       0.00000001041,     0.00000000359,
       0.00000340760,    -0.00000052922,
      -0.00000054679,    -0.00000067149,
      -0.00000006946,     0.00000013111,
       0.00000001579,     0.00000010165,
       0.00000002003,    -0.00000001631,
      -0.00000000391,    -0.00000000207,
      -0.00000000457,    -0.00000000069,
      -0.00000000030,     0.00000000217,
      -0.00000000058,     0.00000000047,
       0.00000003529,     0.00000000599,
       0.00000000082,    -0.00000000542,
       0.00000000107,     0.00000000196,
       0.00000000233,    -0.00000000181,
       0.00000000301,    -0.00000000028,
       0.00000000031,     0.00000000299,
       0.00000000927,    -0.00000001864,
      -0.00000000193,     0.00000000569,
       0.00000000568,     0.00000000193,
      -0.00000011097,     0.00000030953,
      -0.00000001290,    -0.00000000464,
       0.00000000024,    -0.00000000403,
      -0.00000000403,    -0.00000000024,
       0.00000000649,    -0.00000038477,
       0.00000001139,     0.00000000012,
       0.00000000073,    -0.00000000072,
      -0.00000001074,    -0.00000001160,
      -0.00000000180,     0.00000000169,
      -0.00000005690,     0.00000016989,
      -0.00000000728,    -0.00000000244,
       0.00000000165,    -0.00000000492,
      -0.00000000492,    -0.00000000165,
       0.00000011761,    -0.00000037577,
       0.00000001110,     0.00000000347,
       0.00000000054,    -0.00000000257,
       0.00000000247,     0.00000000060,
       0.00000000082,    -0.00000000342,
       0.00000000342,     0.00000000082,
       0.00000000107,    -0.00000000440,
       0.00000000441,     0.00000000107,
      -0.00000000115,    -0.00000000001,
       0.00000002598,    -0.00000003950,
       0.00000000167,     0.00000000110,
       0.00000000108,    -0.00000000317,
      -0.00000000317,    -0.00000000108,
       0.00000007643,    -0.00000021845,
       0.00000000655,     0.00000000229,
       0.00000000040,    -0.00000000106,
       0.00000000106,     0.00000000040,
       0.00000000057,    -0.00000000152,
       0.00000000152,     0.00000000057,
       0.00000000070,    -0.00000000186,
       0.00000000186,     0.00000000070,
       0.00000000046,    -0.00000000071,
      -0.00000000071,    -0.00000000046,
       0.00000005102,    -0.00000007997,
       0.00000000238,     0.00000000152,
       0.00000000066,    -0.00000000104,
       0.00000000104,     0.00000000067,
       0.00000000094,    -0.00000000147,
       0.00000000147,     0.00000000094,
       0.00000000118,    -0.00000000184,
       0.00000000184,     0.00000000118,
       0.00000020585,    -0.00000026700,
      -0.00000012540,    -0.00000006753,
      -0.00000000948,     0.00000003000,
       0.00000000799,     0.00000000612,
       0.00000000201,    -0.00000000373,
      -0.00000000100,    -0.00000000009,
      -0.00000000770,     0.00000000118,
       0.00000000090,     0.00000000306,
       0.00000000100,    -0.00000000045,
       0.00000000121,     0.00000000770,
       0.00000000306,    -0.00000000091,
      -0.00000000045,    -0.00000000101,
      -0.00000000087,    -0.00000000038,
       0.00000000042,     0.00000000076,
       0.00000000078,    -0.00000000049,
      -0.00000000710,     0.00000001382,
       0.00000000195,     0.00000000089,
      -0.00000000185,    -0.00000000323,
      -0.00000000320,     0.00000000195,
       0.00000004266,    -0.00000011377,
      -0.00000003343,    -0.00000000506,
      -0.00000000442,    -0.00000000248,
      -0.00000000050,     0.00000000128,
      -0.00000000821,     0.00000000122,
       0.00000000094,     0.00000000319,
       0.00000000089,    -0.00000000041,
      -0.00000000134,    -0.00000000810,
      -0.00000000317,     0.00000000094,
       0.00000000041,     0.00000000089,
       0.00000000005,     0.00000000160,
       0.00000000024,    -0.00000000094,
      -0.00000000071,    -0.00000000032,
      -0.00000001795,    -0.00000000328,
       0.00000000011,     0.00000000378,
       0.00000000133,    -0.00000000013,
      -0.00000000003,    -0.00000000060,
       0.00000000035,     0.00000000425,
       0.00000000121,    -0.00000000018,
       0.00000000049,     0.00000000004,
      -0.00000000003,    -0.00000000011,
      -0.00000000049,     0.00000000002,
       0.00000000005,     0.00000000008,
      -0.00000005541,    -0.00000000459,
       0.00000000160,     0.00000001209,
       0.00000001779,    -0.00000000263,
      -0.00000000140,    -0.00000000509,
       0.00000000190,     0.00000001760,
       0.00000000507,    -0.00000000124,
       0.00000000075,     0.00000000007,
      -0.00000009092,    -0.00000000812,
       0.00000000249,     0.00000001977,
       0.00000004015,    -0.00000000591,
      -0.00000000308,    -0.00000001151,
       0.00000000460,     0.00000004027,
       0.00000001161,    -0.00000000282,
      -0.00000000078,    -0.00000000009,
      -0.00000023410,    -0.00000007380,
      -0.00000000606,     0.00000005267,
       0.00000013226,    -0.00000002062,
      -0.00000001132,    -0.00000003736,
       0.00000003115,     0.00000012670,
       0.00000003600,    -0.00000001260,
       0.00000000083,     0.00000000019,
      -0.00000000016,    -0.00000000038,
      -0.00000000270,    -0.00000000039,
       0.00000000004,     0.00000000063,
      -0.00001437247,     0.00002816075,
       0.00000810546,     0.00000273826,
       0.00000018423,    -0.00000117928,
      -0.00000011175,     0.00000000667,
       0.00000000269,     0.00000000776,
       0.00001583966,     0.00000651183,
       0.00000127258,    -0.00000444054,
      -0.00000063859,    -0.00000008263,
       0.00000000454,     0.00000006006,
       0.00000000427,    -0.00000000158,
      -0.00001030456,    -0.00001754822,
      -0.00000219465,     0.00000167652,
       0.00000015271,     0.00000012511,
       0.00000000374,    -0.00000000946,
      -0.00001726988,     0.00001072348,
       0.00000184009,     0.00000208376,
       0.00000010044,    -0.00000018579,
      -0.00000001353,     0.00000000026,
      -0.00000005487,    -0.00000014449,
       0.00000002147,    -0.00000002247,
      -0.00000011647,     0.00000012446,
      -0.00000001334,    -0.00000002648,
       0.00011099106,    -0.00021809606,
      -0.00006266831,    -0.00002119277,
      -0.00000143992,     0.00000911180,
       0.00000086402,    -0.00000004902,
      -0.00000002029,    -0.00000006036,
      -0.00003504577,    -0.00001396506,
      -0.00000268107,     0.00000981732,
       0.00000140944,     0.00000015997,
      -0.00000001273,    -0.00000013292,
      -0.00000000923,     0.00000000339,
       0.00003894594,     0.00007496054,
       0.00000983610,    -0.00000517363,
      -0.00000027070,    -0.00000069161,
      -0.00000004534,    -0.00000000801,
      -0.00000000272,     0.00000000436,
       0.00007269678,    -0.00005068474,
      -0.00000971296,    -0.00000837964,
      -0.00000027381,     0.00000116056,
       0.00000010450,    -0.00000003005,
      -0.00000000572,    -0.00000000730,
       0.00000013635,     0.00000028282,
      -0.00000007516,     0.00000006573,
       0.00000000345,     0.00000000380,
       0.00000025466,    -0.00000022413,
       0.00000005277,     0.00000008037,
       0.00000000439,    -0.00000000216,
       0.00000006566,    -0.00000020045,
      -0.00000005871,    -0.00000000926,
      -0.00000016749,    -0.00000052477,
      -0.00000018952,     0.00000009254,
       0.00000002368,     0.00000003438,
       0.00000000393,    -0.00000000384,
      -0.00000775782,     0.00000125756,
       0.00000089408,     0.00000300565,
       0.00000058957,    -0.00000026646,
      -0.00000004882,    -0.00000007459,
      -0.00000000666,     0.00000000635,
      -0.00000122774,    -0.00000775055,
      -0.00000300327,     0.00000088796,
       0.00000026567,     0.00000058952,
       0.00000007459,    -0.00000004876,
      -0.00000000635,    -0.00000000666,
       0.00000000166,    -0.00000000293,
      -0.00000000075,    -0.00000000028,
       0.00000005276,    -0.00000042425,
      -0.00000011888,     0.00000001330,
      -0.00000008624,     0.00000010288,
       0.00000004409,     0.00000002918,
       0.00000000475,    -0.00000000965,
      -0.00000000139,    -0.00000000045,
      -0.00000971580,     0.00000157893,
       0.00000112083,     0.00000376365,
       0.00000073816,    -0.00000033388,
      -0.00000006114,    -0.00000009337,
      -0.00000000833,     0.00000000795,
       0.00000000079,     0.00000000041,
      -0.00000154046,    -0.00000970790,
      -0.00000376236,     0.00000111327,
       0.00000033299,     0.00000073825,
       0.00000009339,    -0.00000006108,
      -0.00000000795,    -0.00000000834,
      -0.00000000041,     0.00000000079,
       0.00000000001,     0.00000000012,
      -0.00000000001,     0.00000000008,
      -0.00000000008,    -0.00000000002,
      -0.00000000004,     0.00000000065,
      -0.00000000064,    -0.00000000005,
       0.00000001209,     0.00000002137,
      -0.00000000012,     0.00000000168,
      -0.00000000167,    -0.00000000014,
       0.00000001299,     0.00000002282,
       0.00000000018,     0.00000000048,
      -0.00000000013,     0.00000000062,
       0.00000000074,     0.00000000181,
       0.00000000386,    -0.00000000173,
      -0.00000000037,    -0.00000000034,
      -0.00000000244,     0.00000000750,
       0.00000000170,     0.00000000013,
      -0.00000000747,    -0.00000000166,
      -0.00000000012,     0.00000000170,
       0.00000000260,     0.00000004140,
      -0.00000000620,    -0.00000009620,
      -0.00000002160,     0.00000000200,
      -0.00000003550,    -0.00000018980,
      -0.00000003860,     0.00000000480,
      -0.00000000067,     0.00000000071,
       0.00000000068,     0.00000000063,
       0.00000001149,     0.00000001552,
       0.00000000057,     0.00000000214,
       0.00000000095,    -0.00000000032,
      -0.00000001011,    -0.00000000372,
      -0.00000000071,     0.00000000429,
       0.00000000095,    -0.00000000007,
      -0.00000000372,     0.00000001016,
       0.00000000426,     0.00000000071,
      -0.00000000007,    -0.00000000095,
       0.00000002547,    -0.00000000191,
       0.00000000174,     0.00000000008,
       0.00000000005,    -0.00000000189,
      -0.00000000205,     0.00000000164,
       0.00000000077,     0.00000000118,
       0.00000000532,     0.00000000095,
      -0.00000000031,    -0.00000000228,
       0.00000000093,    -0.00000000532,
      -0.00000000233,     0.00000000032,
       0.00000001343,     0.00000006013,
       0.00000002783,    -0.00000001697,
      -0.00000000676,    -0.00000000578,
       0.00000000342,    -0.00000000090,
      -0.00000000089,    -0.00000000146,
      -0.00000001277,    -0.00000000591,
      -0.00000000010,     0.00000000532,
      -0.00000000584,     0.00000001293,
       0.00000000531,     0.00000000001,
      -0.00000056985,    -0.00000009996,
      -0.00000000211,     0.00000024516,
       0.00000005181,    -0.00000000913,
      -0.00000000304,    -0.00000001087,
      -0.00000000488,     0.00000000165,
       0.00000000044,     0.00000000116,
       0.00000023633,    -0.00000000174,
      -0.00000001484,    -0.00000009637,
      -0.00000001975,     0.00000000626,
       0.00000000138,     0.00000000262,
      -0.00000000183,    -0.00000023636,
      -0.00000009636,     0.00000001489,
       0.00000000627,     0.00000001976,
       0.00000000260,    -0.00000000137,
       0.00000006938,     0.00000000035,
      -0.00000000314,    -0.00000002502,
      -0.00000000001,     0.00000000280,
       0.00000000104,    -0.00000000014,
       0.00000000439,    -0.00000000104,
      -0.00000000125,    -0.00000000353,
      -0.00000000151,     0.00000000062,
      -0.00000000823,     0.00000000135,
       0.00000000097,     0.00000000320,
       0.00000000133,     0.00000000822,
       0.00000000318,    -0.00000000094,
       0.00000000203,     0.00000000167,
       0.00000000158,    -0.00000000205,
       0.00001164385,    -0.00000030522,
      -0.00000080320,    -0.00000476006,
      -0.00000097983,     0.00000032017,
       0.00000006890,     0.00000012962,
       0.00000001215,    -0.00000000998,
       0.00000001875,     0.00000016602,
       0.00000006865,    -0.00000001591,
      -0.00000000528,    -0.00000001438,
      -0.00000000197,     0.00000000105,
      -0.00000329417,     0.00000052821,
       0.00000037853,     0.00000127623,
       0.00000025051,    -0.00000011290,
      -0.00000002069,    -0.00000003170,
      -0.00000000284,     0.00000000271,
       0.00000052885,     0.00000329138,
       0.00000127506,    -0.00000037861,
      -0.00000011289,    -0.00000025024,
      -0.00000003165,     0.00000002069,
       0.00000000270,     0.00000000282,
       0.00000001872,    -0.00000000298,
      -0.00000000120,     0.00000000013,
       0.00000000014,     0.00000000124,
      -0.00000010066,    -0.00000000039,
       0.00000000629,     0.00000003683,
       0.00000000001,    -0.00000000283,
      -0.00000000102,     0.00000000018,
       0.00000000031,     0.00000000305,
       0.00000000120,    -0.00000000030,
      -0.00000000071,     0.00000000047,
       0.00000000193,    -0.00000000072,
      -0.00000000071,    -0.00000000193,
      -0.00022707480,     0.00003641060,
       0.00002609715,     0.00008799628,
       0.00001727137,    -0.00000778955,
      -0.00000142775,    -0.00000218531,
      -0.00000019527,     0.00000018586,
       0.00000001843,     0.00000000957,
      -0.00000153974,    -0.00000961933,
      -0.00000372799,     0.00000110482,
       0.00000032994,     0.00000073147,
       0.00000009253,    -0.00000006046,
      -0.00000000787,    -0.00000000827,
      -0.00000000041,     0.00000000081,
      -0.00000013193,     0.00000005309,
       0.00000002790,     0.00000004506,
       0.00000000761,    -0.00000000678,
      -0.00000000104,    -0.00000000076,
       0.00000005698,     0.00000026809,
       0.00000010073,    -0.00000003742,
      -0.00000001056,    -0.00000001908,
      -0.00000000231,     0.00000000183,
       0.00000001339,    -0.00000002380,
      -0.00000000609,    -0.00000000246,
      -0.00000000028,     0.00000000114,
      -0.00000002388,    -0.00000001330,
      -0.00000000242,     0.00000000611,
       0.00000000114,     0.00000000028,
      -0.00000036829,     0.00000009091,
       0.00000005458,     0.00000014094,
       0.00000002725,    -0.00000001502,
      -0.00000000068,    -0.00000000257,
      -0.00000000098,     0.00000000041,
      -0.00000003733,     0.00000000791,
       0.00000000515,     0.00000001377,
       0.00000000255,    -0.00000000145,
       0.00000000791,     0.00000003736,
       0.00000001378,    -0.00000000515,
      -0.00000000145,    -0.00000000255,
      -0.00000000425,     0.00000000133,
       0.00000000075,     0.00000000162,
       0.00000000133,     0.00000000426,
       0.00000000163,    -0.00000000075,
      -0.00000217778,     0.00000043985,
       0.00000029480,     0.00000080543,
       0.00000014930,    -0.00000008457,
      -0.00000001441,    -0.00000001752,
      -0.00000001268,    -0.00000006273,
      -0.00000002321,     0.00000000849,
       0.00000000244,     0.00000000431,
      -0.00000000114,     0.00000000041,
      -0.00000000013,    -0.00000000152,
       0.00000000796,     0.00000001287,
       0.00000000615,    -0.00000000423,
      -0.00000000065,    -0.00000000071,
      -0.00000000495,     0.00000000009,
       0.00000000033,     0.00000000493,
      -0.00000005921,    -0.00000009302,
      -0.00000001063,     0.00000001008,
      -0.00000001501,     0.00000001024,
       0.00000000161,     0.00000000173,
       0.00000002095,    -0.00000000047,
      -0.00000000119,    -0.00000002102,
      -0.00000006187,     0.00000030416,
       0.00000006103,     0.00000000823,
       0.00000002144,     0.00000000435,
       0.00000000043,    -0.00000000454,
      -0.00000004961,     0.00000001663,
       0.00000000891,     0.00000001784,
       0.00000000324,    -0.00000000226,
      -0.00000001674,    -0.00000004961,
      -0.00000001787,     0.00000000894,
       0.00000000227,     0.00000000324,
       0.00000001001,    -0.00000001788,
      -0.00000000457,    -0.00000000183,
      -0.00000000021,     0.00000000086,
       0.00000001787,     0.00000001001,
       0.00000000183,    -0.00000000457,
      -0.00000000086,    -0.00000000021,
       0.00000000050,    -0.00000000071,
      -0.00000000071,    -0.00000000050,
      -0.00000007084,     0.00000002176,
       0.00000001328,     0.00000002873,
      -0.00000000096,    -0.00000000318,
      -0.00000000129,     0.00000000060,
       0.00000002389,    -0.00000003341,
       0.00000003797,    -0.00000000400,
      -0.00000000380,    -0.00000001537,
      -0.00000000314,     0.00000000128,
      -0.00000000400,    -0.00000003797,
      -0.00000001537,     0.00000000379,
       0.00000000128,     0.00000000314,
       0.00000207053,    -0.00000021543,
      -0.00000020577,    -0.00000083656,
      -0.00000017070,     0.00000006951,
       0.00000001384,     0.00000002240,
       0.00000000668,     0.00000006382,
       0.00000002580,    -0.00000000635,
      -0.00000000214,    -0.00000000526,
      -0.00000000068,     0.00000000042,
       0.00000000300,    -0.00000000075,
      -0.00000000046,    -0.00000000115,
      -0.00000000062,    -0.00000000291,
      -0.00000000112,     0.00000000040,
       0.00000007874,    -0.00000000840,
      -0.00000000390,    -0.00000003530,
       0.00000000038,     0.00000000294,
       0.00000000406,    -0.00000000176,
      -0.00000053997,     0.00000005426,
       0.00000000831,    -0.00000000296,
       0.00000000942,     0.00000008341,
      -0.00000000052,    -0.00000000138,
      -0.00000000236,    -0.00000001331,
       0.00000000166,     0.00000000075,
      -0.00000001677,     0.00000000611,
       0.00000000119,    -0.00000000128,
       0.00000002940,    -0.00000000248,
      -0.00000000096,     0.00000000222,
      -0.00000000010,     0.00000000256,
      -0.00000000264,    -0.00000000010,
       0.00000000133,    -0.00000000051,
      -0.00000000051,    -0.00000000133,
       0.00000007263,    -0.00000002603,
      -0.00000001519,    -0.00000002951,
       0.00000000083,     0.00000000235,
       0.00000000093,    -0.00000000048,
       0.00000000099,    -0.00000000025,
      -0.00000000025,    -0.00000000099,
      -0.00000000047,    -0.00000000177,
      -0.00000000072,     0.00000000029,
       0.00000000063,     0.00000000260,
       0.00000000104,    -0.00000000040,
       0.00000002977,    -0.00000000464,
       0.00000007079,     0.00000006522,
      -0.00000000276,    -0.00000000314,
       0.00000002868,    -0.00000003000,
      -0.00000000144,     0.00000000122,
       0.00000000284,    -0.00000001084,
       0.00000000059,     0.00000000154,
      -0.00000000807,    -0.00000000284,
       0.00000000126,    -0.00000000073,
      -0.00000042694,    -0.00000040045,
       0.00000001604,     0.00000001912,
      -0.00000006781,     0.00000007081,
       0.00000000333,    -0.00000000283,
      -0.00000001160,     0.00000004433,
      -0.00000000256,    -0.00000000638,
       0.00000003826,     0.00000001185,
      -0.00000000572,     0.00000000323,
       0.00000000152,     0.00000000020,
      -0.00000000019,     0.00000000152,
      -0.00000000127,     0.00000000080,
       0.00000001160,     0.00000001755,
       0.00000000290,    -0.00000000183,
      -0.00000000229,     0.00000000067,
       0.00000000043,     0.00000000096,
       0.00000000066,     0.00000000227,
       0.00000000096,    -0.00000000043,
       0.00000001202,     0.00000004508,
      -0.00000003183,     0.00000004639,
       0.00000001997,     0.00000001082,
       0.00000000179,    -0.00000000439,
       0.00000004639,     0.00000003183,
       0.00000001082,    -0.00000001997,
      -0.00000000439,    -0.00000000179,
      -0.00000000208,     0.00000000034,
       0.00000000024,     0.00000000080,
       0.00000000034,     0.00000000208,
       0.00000000080,    -0.00000000024,
      -0.00000167345,     0.00000017477,
       0.00000016744,     0.00000067903,
       0.00000013848,    -0.00000005652,
      -0.00000001129,    -0.00000001817,
      -0.00000000755,    -0.00000007275,
      -0.00000002953,     0.00000000725,
       0.00000000245,     0.00000000602,
       0.00000000082,    -0.00000000051,
      -0.00000001080,     0.00000000195,
       0.00000000133,     0.00000000415,
       0.00000000082,    -0.00000000040,
       0.00000000191,     0.00000001178,
       0.00000000460,    -0.00000000138,
      -0.00000000043,    -0.00000000092,
      -0.00000228803,     0.00000346021,
       0.00000148701,     0.00000076761,
       0.00000012430,    -0.00000032571,
      -0.00000004715,    -0.00000001130,
      -0.00000002977,    -0.00000001567,
      -0.00000000475,     0.00000001269,
       0.00000000276,     0.00000000062,
      -0.00000370608,     0.00000059526,
       0.00000042647,     0.00000143547,
       0.00000028171,    -0.00000012724,
      -0.00000002331,    -0.00000003564,
      -0.00000000318,     0.00000000303,
       0.00000059511,     0.00000370613,
       0.00000143550,    -0.00000042641,
      -0.00000012723,    -0.00000028172,
      -0.00000003564,     0.00000002330,
       0.00000000303,     0.00000000318,
       0.00000000526,     0.00000000026,
      -0.00000000024,    -0.00000000228,
       0.00000000026,    -0.00000000525,
      -0.00000000225,     0.00000000024,
       0.00000017175,     0.00000000509,
      -0.00000000875,    -0.00000007358,
      -0.00000001580,     0.00000000450,
       0.00000000013,     0.00000000256,
       0.00000000114,    -0.00000000020,
      -0.00000004894,     0.00000000515,
       0.00000000490,     0.00000001987,
       0.00000000407,    -0.00000000167,
       0.00000000517,     0.00000004892,
       0.00000001984,    -0.00000000491,
      -0.00000000166,    -0.00000000405,
      -0.00000002410,     0.00000000394,
      -0.00000000410,    -0.00000002504,
      -0.00000000973,     0.00000000291,
       0.00000000087,     0.00000000191,
      -0.00000004925,     0.00000000773,
       0.00000000150,     0.00000000905,
       0.00000000350,    -0.00000000106,
      -0.00000000032,    -0.00000000069,
      -0.00000000203,     0.00000000028,
       0.00000000023,     0.00000000080,
      -0.00000000030,    -0.00000000203,
      -0.00000000081,     0.00000000023,
       0.00000000009,    -0.00000000135,
      -0.00000000020,    -0.00000000102,
      -0.00000001444,     0.00000000294,
       0.00000000045,     0.00000000245,
       0.00000000090,     0.00000000172,
       0.00000000175,    -0.00000000070,
      -0.00041878494,     0.00006725208,
       0.00004816592,     0.00016226445,
       0.00003184525,    -0.00001437567,
      -0.00000263462,    -0.00000402885,
      -0.00000035968,     0.00000034261,
       0.00000003398,     0.00000001764,
      -0.00000200151,    -0.00001246669,
      -0.00000483064,     0.00000143368,
       0.00000042796,     0.00000094768,
       0.00000011988,    -0.00000007843,
      -0.00000001020,    -0.00000001070,
      -0.00000000052,     0.00000000100,
       0.00000002818,     0.00000004145,
       0.00000001529,    -0.00000001269,
      -0.00000000292,    -0.00000000284,
       0.00000001467,    -0.00000004587,
      -0.00000001862,    -0.00000000377,
      -0.00000000031,     0.00000000387,
      -0.00000002367,    -0.00000000381,
       0.00000000494,     0.00000001188,
       0.00000000244,    -0.00000000171,
      -0.00000000382,     0.00000002365,
       0.00000001187,    -0.00000000494,
      -0.00000000171,    -0.00000000244,
       0.00000000152,    -0.00000000098,
      -0.00000000122,    -0.00000000005,
       0.00000000003,     0.00000000121,
      -0.00000001410,    -0.00000002413,
      -0.00000000385,     0.00000000240,
       0.00000000513,     0.00000000017,
      -0.00000000005,    -0.00000000519,
      -0.00000014652,     0.00000071604,
       0.00000014168,     0.00000002363,
       0.00000002691,     0.00000000604,
       0.00000000059,    -0.00000000605,
       0.00000001071,     0.00000000779,
       0.00000000267,    -0.00000000452,
      -0.00000000098,    -0.00000000042,
      -0.00000000774,     0.00000001080,
       0.00000000453,     0.00000000267,
       0.00000000042,    -0.00000000098,
      -0.00000002045,    -0.00000000330,
       0.00000000427,     0.00000001026,
       0.00000000211,    -0.00000000148,
       0.00000000329,    -0.00000002045,
      -0.00000001026,     0.00000000427,
       0.00000000148,     0.00000000211,
       0.00000000884,    -0.00000000047,
      -0.00000000025,    -0.00000000423,
      -0.00000006743,     0.00000000345,
       0.00000000059,     0.00000001059,
       0.00000000091,    -0.00000000221,
      -0.00000000276,    -0.00000000037,
      -0.00000000112,     0.00000000553,
       0.00000000124,     0.00000000008,
      -0.00000000554,    -0.00000000112,
      -0.00000000008,     0.00000000124,
      -0.00000000500,    -0.00000003800,
      -0.00000004603,     0.00000000169,
       0.00000000328,     0.00000001872,
       0.00000000654,    -0.00000000152,
      -0.00000000087,    -0.00000000252,
      -0.00000000153,    -0.00000000653,
      -0.00000000252,     0.00000000087,
       0.00000036789,    -0.00000007013,
      -0.00000004481,    -0.00000014236,
      -0.00000002791,     0.00000001274,
       0.00000000068,     0.00000000453,
       0.00000000174,    -0.00000000048,
       0.00000004991,    -0.00000001128,
      -0.00000000709,    -0.00000001830,
      -0.00000000339,     0.00000000193,
      -0.00000001128,    -0.00000005018,
      -0.00000001843,     0.00000000711,
       0.00000000193,     0.00000000340,
       0.00000174013,    -0.00000040100,
      -0.00000024911,    -0.00000063788,
      -0.00000011797,     0.00000006665,
       0.00000001118,     0.00000001370,
       0.00000001666,     0.00000007199,
       0.00000002637,    -0.00000001033,
      -0.00000000277,    -0.00000000489,
      -0.00000000058,     0.00000000048,
       0.00000000402,    -0.00000000091,
      -0.00000000059,    -0.00000000150,
      -0.00000008061,     0.00000000824,
       0.00000000744,     0.00000003015,
      -0.00000000010,    -0.00000000101,
      -0.00000000158,     0.00000000001,
       0.00000000008,     0.00000000203,
      -0.00000000055,     0.00000000048,
       0.00000000610,     0.00000000862,
       0.00000000124,    -0.00000000096,
      -0.00000000161,    -0.00000000035,
       0.00000000078,     0.00000000201,
       0.00000000102,    -0.00000000095,
       0.00000000095,     0.00000000104,
      -0.00000000032,    -0.00000000078,
       0.00000000078,    -0.00000000032,
      -0.00000000862,     0.00000000151,
       0.00000000025,     0.00000000135,
       0.00000000050,    -0.00000000062,
      -0.00000000970,    -0.00000000742,
      -0.00000000127,     0.00000000161,
      -0.00000000180,     0.00000000101,
       0.00000000013,     0.00000000016,
      -0.00000000029,     0.00000000018,
       0.00000000003,     0.00000000003,
       0.00000000002,    -0.00000000007,
       0.00000000006,    -0.00000000004,
      -0.00000000006,     0.00000000003,
      -0.00000000406,     0.00000000279,
       0.00000000031,     0.00000000048,
      -0.00000000739,     0.00000000446,
       0.00000000052,     0.00000000084,
      -0.00000000061,     0.00000000060,
      -0.00000000072,     0.00000000011,
      -0.00000001466,     0.00000000856,
       0.00000000108,     0.00000000144,
       0.00000000047,     0.00000000194,
       0.00000000047,     0.00000000018,
      -0.00000000176,    -0.00000000202,
      -0.00000000018,     0.00000000048,
      -0.00006101843,    -0.00010142577,
      -0.00001237807,     0.00001032025,
       0.00000096013,     0.00000066827,
       0.00000001250,    -0.00000006250,
      -0.00004294219,     0.00002874004,
       0.00000453380,     0.00000539719,
       0.00000029018,    -0.00000041498,
      -0.00000002714,    -0.00000000567,
       0.00003765050,     0.00000077980,
       0.00000020187,    -0.00000000240,
      -0.00000000164,     0.00000004744,
      -0.00000000214,     0.00000000047,
      -0.00000072969,    -0.00003748473,
       0.00000000701,     0.00000017488,
       0.00000009622,    -0.00000000075,
      -0.00000000031,    -0.00000000213,
      -0.00000008463,     0.00000011879,
      -0.00000000276,     0.00000001230,
       0.00000000161,     0.00000000015,
      -0.00000009698,     0.00000002380,
       0.00000001318,     0.00000000085,
      -0.00000000079,    -0.00000000209,
       0.00037964021,     0.00062833946,
       0.00007673726,    -0.00006409021,
      -0.00000595950,    -0.00000413709,
      -0.00000007694,     0.00000038824,
       0.00000001781,    -0.00000000533,
       0.00010162733,    -0.00006838957,
      -0.00001075268,    -0.00001279009,
      -0.00000068735,     0.00000098330,
       0.00000006453,     0.00000001349,
      -0.00000000070,    -0.00000000305,
      -0.00016223369,    -0.00000262411,
      -0.00000066273,    -0.00000004007,
       0.00000000369,    -0.00000023268,
       0.00000000702,    -0.00000000144,
       0.00000163758,     0.00016256129,
       0.00000004346,    -0.00000044125,
      -0.00000038313,    -0.00000000257,
       0.00000000095,     0.00000000664,
       0.00000034795,    -0.00000016666,
      -0.00000000310,    -0.00000007695,
      -0.00000001096,     0.00000000264,
       0.00000010282,    -0.00000024505,
      -0.00000007121,    -0.00000000723,
       0.00000000033,     0.00000000979,
       0.00000000796,     0.00000000300,
       0.00000000039,    -0.00000000197,
       0.00000004412,    -0.00000011155,
      -0.00000003162,    -0.00000000985,
      -0.00000000069,     0.00000000437,
       0.00000011396,     0.00000005758,
       0.00000000982,    -0.00000003182,
      -0.00000000443,    -0.00000000068,
       0.00000001846,     0.00000001291,
      -0.00000000086,    -0.00000000196,
       0.00000005502,    -0.00000013951,
      -0.00000003955,    -0.00000001221,
      -0.00000000085,     0.00000000546,
       0.00000014283,     0.00000007168,
       0.00000001220,    -0.00000003987,
      -0.00000000554,    -0.00000000084,
      -0.00000004491,     0.00000000014,
      -0.00000004783,     0.00000000011,
       0.00000000039,     0.00000000061,
      -0.00000000061,     0.00000000039,
       0.00000000029,    -0.00000000189,
      -0.00000000098,     0.00000000012,
      -0.00000000189,    -0.00000000030,
       0.00000000012,     0.00000000097,
       0.00000000994,     0.00000000405,
       0.00000000176,    -0.00000000564,
      -0.00000000232,     0.00000000015,
      -0.00000000569,     0.00000000385,
       0.00000000160,     0.00000000080,
       0.00000000399,     0.00000000555,
       0.00000000073,    -0.00000000163,
      -0.00000007898,    -0.00000003139,
       0.00000000012,     0.00000003358,
      -0.00000000436,     0.00000001294,
       0.00000000521,    -0.00000000011,
      -0.00000000043,    -0.00000000111,
       0.00000002397,    -0.00000001601,
      -0.00000000659,    -0.00000000360,
      -0.00000001699,    -0.00000002359,
      -0.00000000320,     0.00000000705,
      -0.00000001983,    -0.00000004993,
      -0.00000002160,     0.00000001449,
       0.00000000391,     0.00000000058,
       0.00000000000,    -0.00000000172,
      -0.00000000168,     0.00000000323,
       0.00000000105,     0.00000000028,
       0.00000000258,     0.00000000181,
       0.00000000045,    -0.00000000073,
      -0.00000000060,     0.00000000076,
       0.00000000132,    -0.00000000056,
      -0.00000000051,    -0.00000000135,
       0.00000001544,     0.00000000982,
       0.00000000142,    -0.00000000259,
      -0.00000000118,    -0.00000000023,
      -0.00000000570,     0.00000000209,
       0.00000000114,     0.00000000104,
       0.00000000211,     0.00000000570,
       0.00000000111,    -0.00000000102,
       0.00000024048,     0.00000009444,
       0.00000000739,    -0.00000010826,
      -0.00000002344,     0.00000000414,
      -0.00000000245,    -0.00000000401,
      -0.00000000060,     0.00000000078,
       0.00000002015,    -0.00000004525,
      -0.00000001226,    -0.00000000366,
      -0.00000000024,     0.00000000181,
      -0.00000004465,    -0.00000002024,
      -0.00000000381,     0.00000001194,
       0.00000000172,     0.00000000030,
       0.00000000027,     0.00000000080,
      -0.00000000008,    -0.00000000095,
      -0.00000000096,     0.00000000006,
       0.00000001285,    -0.00000000384,
      -0.00000000067,    -0.00000000200,
       0.00000000029,     0.00000000417,
       0.00000000420,    -0.00000000027,
       0.00000142677,    -0.00000317631,
      -0.00000089302,    -0.00000027538,
      -0.00000001976,     0.00000012912,
       0.00000014237,     0.00000006369,
       0.00000001222,    -0.00000004004,
      -0.00000000556,    -0.00000000083,
       0.00000000422,    -0.00000000274,
      -0.00000000085,    -0.00000000112,
      -0.00000000352,     0.00000000174,
       0.00000000091,     0.00000000071,
       0.00000001251,    -0.00000003213,
       0.00000000092,     0.00000000036,
      -0.00000002963,     0.00000000122,
       0.00000000044,     0.00000001207,
       0.00000016506,    -0.00000000664,
      -0.00000000097,    -0.00000002770,
       0.00000000072,    -0.00000000152,
      -0.00000015512,     0.00000002609,
       0.00000000185,     0.00000000975,
      -0.00000001455,     0.00000002837,
      -0.00000000086,    -0.00000000044,
       0.00000000692,     0.00000002300,
       0.00000000898,    -0.00000000220,
      -0.00000000061,     0.00000000091,
       0.00000000234,    -0.00000000035,
       0.00000000014,     0.00000000228,
      -0.00000003366,    -0.00000011002,
       0.00000001059,     0.00000000703,
      -0.00000002070,     0.00000000495,
       0.00000000133,    -0.00000000203,
      -0.00000000439,     0.00000000114,
       0.00000000015,    -0.00000000464,
      -0.00000000191,     0.00000000034,
      -0.00000000033,    -0.00000000183,
       0.00000001653,    -0.00000005685,
       0.00000000416,     0.00000000837,
      -0.00000002156,    -0.00000000578,
       0.00000000309,    -0.00000000158,
       0.00000000231,     0.00000000148,
      -0.00000000297,     0.00000000238,
      -0.00000007025,     0.00000026162,
      -0.00000001968,    -0.00000003774,
       0.00000004898,     0.00000001310,
      -0.00000000701,     0.00000000353,
      -0.00000000408,    -0.00000000066,
       0.00000000685,    -0.00000000457,
      -0.00000000043,     0.00000000092,
      -0.00000001999,    -0.00000001672,
       0.00000000052,    -0.00000000062,
      -0.00000000033,    -0.00000000072,
      -0.00000000072,     0.00000000033,
       0.00000000040,     0.00000000074,
       0.00000000073,    -0.00000000039,
       0.00000268054,    -0.00000596561,
      -0.00000167819,    -0.00000051674,
      -0.00000003683,     0.00000024232,
       0.00000018303,     0.00000008233,
       0.00000001589,    -0.00000005148,
      -0.00000000713,    -0.00000000110,
       0.00000000458,    -0.00000000085,
      -0.00000000043,    -0.00000000134,
       0.00000000132,     0.00000000587,
       0.00000000165,    -0.00000000059,
      -0.00000000723,    -0.00000000003,
      -0.00000000003,     0.00000000298,
       0.00000004247,     0.00000000017,
       0.00000000005,    -0.00000000701,
      -0.00000031842,     0.00000006340,
       0.00000000225,     0.00000001140,
       0.00000000054,    -0.00000000257,
      -0.00000000007,    -0.00000000005,
       0.00000000018,    -0.00000000100,
      -0.00000000007,    -0.00000000003,
       0.00000000072,     0.00000000016,
       0.00000000003,    -0.00000000005,
       0.00000000001,     0.00000000006,
       0.00000000005,    -0.00000000008,
       0.00000000216,    -0.00000000542,
      -0.00000000107,    -0.00000000167,
       0.00000000169,    -0.00000000107,
       0.00000000351,    -0.00000000943,
      -0.00000000225,    -0.00000000384,
      -0.00000000039,     0.00000000029,
       0.00000000395,    -0.00000000246,
      -0.00000000029,    -0.00000000039,
       0.00000000346,    -0.00000001199,
       0.00000000014,    -0.00000000049,
      -0.00000000913,    -0.00000001175,
      -0.00000000125,     0.00000000103,
       0.00000001225,    -0.00000000840,
      -0.00000000108,    -0.00000000123,
       0.00031138231,    -0.00000158415,
      -0.00000016597,     0.00000004587,
       0.00000000379,     0.00000049241,
       0.00000011428,    -0.00011730584,
       0.00000000540,     0.00000002557,
       0.00000018083,     0.00000000036,
      -0.00000104412,    -0.00000406615,
      -0.00000016684,     0.00000055043,
       0.00000001876,     0.00000003921,
      -0.00000000061,    -0.00000000222,
       0.00000138436,    -0.00000476803,
      -0.00000015927,    -0.00000062449,
      -0.00000000360,     0.00000003980,
       0.00000000159,     0.00000000165,
      -0.00000008401,    -0.00000028224,
       0.00000001985,    -0.00000003349,
       0.00000000046,     0.00000000143,
       0.00000012607,    -0.00000020846,
       0.00000001622,     0.00000004171,
      -0.00142073546,     0.00000646698,
       0.00000100703,    -0.00000014943,
      -0.00000001547,    -0.00000224324,
      -0.00000037343,     0.00026548064,
      -0.00000001149,    -0.00000004767,
      -0.00000040956,    -0.00000000159,
      -0.00000133405,     0.00001013617,
      -0.00000013859,    -0.00000134003,
      -0.00000001674,    -0.00000010223,
       0.00000000328,     0.00000000448,
      -0.00000017155,     0.00001116863,
      -0.00000005610,     0.00000156124,
      -0.00000002680,    -0.00000009111,
      -0.00000000238,    -0.00000000490,
      -0.00000027913,     0.00000068984,
       0.00000003366,     0.00000008147,
       0.00000000309,    -0.00000000443,
       0.00000032624,     0.00000051228,
       0.00000002091,    -0.00000010127,
      -0.00000000479,    -0.00000000168,
       0.00000008407,     0.00000001016,
      -0.00000003000,     0.00000002511,
       0.00000000365,     0.00000000301,
       0.00000030366,     0.00000047279,
       0.00000005891,    -0.00000004958,
      -0.00000000431,    -0.00000000326,
      -0.00000047397,     0.00000030732,
       0.00000004937,     0.00000005907,
       0.00000000324,    -0.00000000433,
       0.00000015158,     0.00000002235,
       0.00000000912,     0.00000000213,
       0.00000000023,    -0.00000000122,
       0.00000038080,     0.00000059240,
       0.00000007372,    -0.00000006215,
      -0.00000000540,    -0.00000000407,
      -0.00000059344,     0.00000038524,
       0.00000006187,     0.00000007399,
       0.00000000406,    -0.00000000542,
      -0.00000000076,    -0.00000000001,
       0.00000000002,    -0.00000000076,
      -0.00000001124,     0.00000001124,
       0.00000004270,    -0.00000002036,
      -0.00000001166,    -0.00000001545,
      -0.00000002593,     0.00000001884,
       0.00000000770,     0.00000000359,
       0.00000000887,     0.00000001105,
       0.00000000146,    -0.00000000341,
      -0.00000000005,    -0.00000000980,
      -0.00000000092,     0.00000000102,
      -0.00000000975,     0.00000000051,
       0.00000000116,     0.00000000080,
       0.00000016329,    -0.00000011665,
      -0.00000004817,    -0.00000002326,
      -0.00000002104,    -0.00000002645,
      -0.00000000350,     0.00000000816,
       0.00000000072,     0.00000004229,
       0.00000000394,    -0.00000000448,
       0.00000004215,    -0.00000000207,
      -0.00000000487,    -0.00000000377,
      -0.00000008527,     0.00000003445,
       0.00000001206,     0.00000003570,
      -0.00000000080,     0.00000000095,
      -0.00000000595,    -0.00000001310,
      -0.00000000184,     0.00000000122,
      -0.00000001307,     0.00000000584,
       0.00000000117,     0.00000000183,
       0.00000000573,    -0.00000000278,
      -0.00000000103,    -0.00000000248,
      -0.00000000022,     0.00000000224,
       0.00000000227,     0.00000000018,
      -0.00000003767,     0.00000001148,
       0.00000000669,     0.00000000716,
       0.00000000240,     0.00000000587,
       0.00000000110,    -0.00000000116,
       0.00000000129,    -0.00000000971,
      -0.00000000097,     0.00000000077,
      -0.00000000979,    -0.00000000122,
       0.00000000057,     0.00000000105,
      -0.00000037197,    -0.00000052017,
      -0.00000004933,     0.00000006829,
       0.00000001357,    -0.00000000593,
      -0.00000000122,    -0.00000000219,
       0.00000011729,     0.00000018206,
       0.00000002249,    -0.00000001879,
      -0.00000000167,    -0.00000000126,
       0.00000018193,    -0.00000011752,
      -0.00000001881,    -0.00000002247,
      -0.00000000126,     0.00000000166,
      -0.00000000191,     0.00000000008,
       0.00000000115,     0.00000000001,
      -0.00000000002,    -0.00000000119,
       0.00000000090,     0.00000002669,
       0.00000000450,    -0.00000000028,
      -0.00000000491,    -0.00000000002,
       0.00000000008,     0.00000000505,
       0.00000845382,     0.00001310058,
       0.00000163203,    -0.00000136972,
      -0.00000012070,    -0.00000009049,
      -0.00000060524,     0.00000039095,
       0.00000006329,     0.00000007538,
       0.00000000415,    -0.00000000552,
       0.00000000717,     0.00000000857,
       0.00000000081,    -0.00000000091,
       0.00000001394,    -0.00000001214,
      -0.00000000175,    -0.00000000139,
      -0.00000000123,    -0.00000000001,
      -0.00000000001,     0.00000000122,
       0.00000001506,     0.00000002435,
       0.00000000153,     0.00000000209,
       0.00000000210,    -0.00000000152,
       0.00000009154,     0.00000012634,
      -0.00000000379,     0.00000000273,
       0.00000000059,    -0.00000000094,
       0.00000000134,    -0.00000000046,
      -0.00000000030,     0.00000000176,
      -0.00000005220,    -0.00000001885,
      -0.00000000070,     0.00000000273,
       0.00000000298,     0.00000000330,
      -0.00000000282,     0.00000000306,
      -0.00000000105,    -0.00000000001,
       0.00000000001,    -0.00000000105,
      -0.00000000128,    -0.00000000224,
      -0.00000000224,     0.00000000128,
      -0.00000007148,    -0.00000012532,
      -0.00000001721,     0.00000001346,
       0.00000000401,    -0.00000000228,
       0.00000000748,     0.00000000855,
       0.00000000079,    -0.00000000046,
      -0.00000000139,    -0.00000000062,
       0.00000000099,     0.00000000075,
       0.00000000082,    -0.00000000493,
      -0.00000000138,    -0.00000000016,
       0.00000000045,    -0.00000000176,
       0.00000000214,     0.00000000032,
      -0.00000000324,     0.00000002391,
       0.00000000322,     0.00000000046,
      -0.00000000057,     0.00000000071,
      -0.00000000293,    -0.00000000053,
       0.00000000367,     0.00000000077,
       0.00000000078,    -0.00000000367,
       0.00000000075,    -0.00000000046,
       0.00001601663,     0.00002479349,
       0.00000308854,    -0.00000259296,
      -0.00000022875,    -0.00000017099,
      -0.00000077393,     0.00000049977,
       0.00000008092,     0.00000009641,
       0.00000000531,    -0.00000000706,
       0.00000000162,    -0.00000000042,
      -0.00000000582,    -0.00000000108,
       0.00000000114,     0.00000000179,
       0.00000000179,    -0.00000000114,
      -0.00000008222,    -0.00000002033,
      -0.00000000074,     0.00000000324,
       0.00000000084,    -0.00000000093,
       0.00000000090,     0.00000000083,
       0.00000000105,     0.00000000166,
      -0.00000000165,     0.00000000105,
      -0.00000000014,    -0.00000000039,
       0.00000000014,    -0.00000000004,
      -0.00000000085,     0.00000000437,
       0.00000000021,     0.00000000018,
      -0.00000000464,    -0.00000000083,
      -0.00000000019,     0.00000000019,
       0.00000000014,     0.00000000026,
      -0.00000000027,     0.00000000007,
       0.00000000119,    -0.00000000570,
      -0.00000000504,     0.00000000005,
      -0.00000000003,    -0.00000000509,
       0.00000000133,    -0.00000000922,
       0.00000000036,    -0.00000000026,
      -0.00000001172,     0.00000000096,
      -0.00000000102,    -0.00000001186,
       0.00000001050,    -0.00000000394,
      -0.00000000040,     0.00000000011,
      -0.00000001587,    -0.00000000504,
       0.00000000076,    -0.00000000008,
       0.00000000507,    -0.00000001499,
      -0.00000000002,     0.00000000078,
       0.00000000048,     0.00000000010,
      -0.00000000012,     0.00000000047,
       0.00002050039,    -0.00000450586,
       0.00000149815,     0.00000152905,
      -0.00000008596,     0.00000012844,
      -0.00000000729,    -0.00000000295,
      -0.00000174519,    -0.00000453477,
       0.00000038443,    -0.00000040979,
       0.00000003232,     0.00000002129,
      -0.00000000077,     0.00000000200,
       0.00001471775,    -0.00000034275,
       0.00000012683,    -0.00000001615,
       0.00000000016,     0.00000003584,
      -0.00000065522,     0.00001500328,
      -0.00000001335,    -0.00000013201,
       0.00000000295,    -0.00000000045,
      -0.00000005931,     0.00000001951,
       0.00000000384,     0.00000001418,
       0.00000000206,    -0.00000000020,
      -0.00000001016,    -0.00000005547,
      -0.00000001431,     0.00000000214,
       0.00000000059,     0.00000000188,
      -0.00005070345,     0.00000567191,
      -0.00000404578,    -0.00000362065,
       0.00000017756,    -0.00000032312,
       0.00000002191,     0.00000000515,
       0.00000428400,     0.00001025179,
      -0.00000087895,     0.00000095216,
      -0.00000007501,    -0.00000004849,
       0.00000000175,    -0.00000000458,
      -0.00001300247,     0.00000327473,
      -0.00000047251,     0.00000029961,
      -0.00000001934,    -0.00000008540,
       0.00000000563,     0.00000000024,
       0.00000306629,    -0.00001840589,
       0.00000030873,     0.00000052227,
      -0.00000004707,     0.00000001798,
       0.00000000007,    -0.00000000569,
       0.00000012545,    -0.00000003851,
      -0.00000000805,    -0.00000003860,
      -0.00000000436,     0.00000000057,
       0.00000005179,     0.00000012655,
       0.00000004085,    -0.00000000412,
      -0.00000000161,    -0.00000000416,
       0.00000001343,    -0.00000001352,
       0.00000000791,     0.00000009231,
      -0.00000000135,     0.00000000014,
       0.00000132418,     0.00000004523,
       0.00000000863,    -0.00000000073,
      -0.00000000015,     0.00000000127,
      -0.00000004568,     0.00000132540,
       0.00000000056,     0.00000000872,
      -0.00000000129,    -0.00000000016,
       0.00000005901,    -0.00000001820,
       0.00000001719,    -0.00000001456,
       0.00000165884,     0.00000005614,
       0.00000001081,    -0.00000000089,
      -0.00000000018,     0.00000000159,
      -0.00000005688,     0.00000166076,
       0.00000000069,     0.00000001090,
      -0.00000000161,    -0.00000000020,
      -0.00000000044,    -0.00000000007,
       0.00000000007,    -0.00000000044,
       0.00000000058,    -0.00000000120,
      -0.00000000411,     0.00000000947,
       0.00000000022,     0.00000000068,
       0.00000000081,    -0.00000000026,
      -0.00000000042,    -0.00000000087,
       0.00000000088,    -0.00000000042,
       0.00000003391,    -0.00000001557,
      -0.00000000854,    -0.00000001165,
      -0.00000008186,     0.00000001386,
       0.00000000877,     0.00000002621,
       0.00000000506,     0.00000000084,
       0.00000000089,    -0.00000000500,
      -0.00000000521,    -0.00000006206,
      -0.00000000545,     0.00000000731,
      -0.00000002362,     0.00000000153,
       0.00000000272,     0.00000000205,
      -0.00000000215,     0.00000000066,
      -0.00000000061,     0.00000000121,
       0.00000002201,     0.00000028002,
       0.00000002375,    -0.00000003347,
       0.00000005366,    -0.00000000339,
      -0.00000000621,    -0.00000000465,
       0.00000000684,    -0.00000000112,
       0.00000000105,    -0.00000000485,
       0.00000014326,     0.00000001930,
       0.00000000020,    -0.00000000302,
      -0.00000004426,    -0.00000000725,
      -0.00000000064,     0.00000000119,
      -0.00000000714,     0.00000004414,
       0.00000000114,     0.00000000061,
       0.00000000214,     0.00000001415,
       0.00000000502,    -0.00000000076,
       0.00000001301,    -0.00000005480,
      -0.00000001064,    -0.00000000122,
       0.00000000079,     0.00000000112,
       0.00000000167,    -0.00000000017,
      -0.00000000052,    -0.00000000221,
      -0.00000182796,    -0.00000027668,
      -0.00000000592,     0.00000003421,
       0.00000000202,    -0.00000001451,
       0.00000000044,     0.00000000064,
       0.00000049730,     0.00000001714,
       0.00000000328,    -0.00000000027,
       0.00000001703,    -0.00000049624,
      -0.00000000020,    -0.00000000327,
       0.00000000609,    -0.00000000012,
      -0.00000000001,    -0.00000000226,
      -0.00000002748,     0.00000000074,
       0.00000000002,     0.00000000512,
       0.00003619945,     0.00000124187,
       0.00000023845,    -0.00000001804,
      -0.00000000423,     0.00000003460,
      -0.00000005851,     0.00000170686,
       0.00000000083,     0.00000001119,
      -0.00000000162,    -0.00000000020,
       0.00000002215,    -0.00000000469,
      -0.00000000149,    -0.00000004313,
      -0.00000000191,     0.00000000212,
       0.00000000212,     0.00000000191,
       0.00000006302,    -0.00000000248,
       0.00000000608,    -0.00000000021,
      -0.00000000007,    -0.00000000606,
       0.00000036522,    -0.00000000805,
       0.00000000025,     0.00000001123,
      -0.00000003377,    -0.00000000806,
      -0.00000000047,     0.00000000148,
       0.00000000966,    -0.00000000079,
       0.00000000115,     0.00000000962,
      -0.00000000168,     0.00000000187,
      -0.00000000187,    -0.00000000168,
      -0.00000000600,    -0.00000000053,
      -0.00000000053,     0.00000000600,
      -0.00000033572,    -0.00000003016,
       0.00000000098,    -0.00000001099,
      -0.00000001056,     0.00000000203,
       0.00000000002,     0.00000000134,
      -0.00000000027,    -0.00000001187,
      -0.00000000078,     0.00000000079,
      -0.00000000034,     0.00000000074,
       0.00000000427,    -0.00000000621,
      -0.00000000621,    -0.00000000429,
       0.00006891693,     0.00000234198,
       0.00000045404,    -0.00000003333,
      -0.00000000802,     0.00000006653,
      -0.00000007420,     0.00000217449,
       0.00000000104,     0.00000001422,
      -0.00000000209,    -0.00000000025,
      -0.00000000140,    -0.00000000746,
      -0.00000000460,     0.00000000820,
       0.00000000475,    -0.00000000008,
      -0.00000000008,    -0.00000000475,
      -0.00000003929,    -0.00000001624,
      -0.00000000069,     0.00000000135,
      -0.00000000138,    -0.00000000196,
       0.00000000196,    -0.00000000138,
       0.00000000445,    -0.00000000008,
       0.00000000008,     0.00000000445,
      -0.00000000045,    -0.00000000089,
      -0.00000000058,    -0.00000000322,
       0.00000000323,    -0.00000000056,
      -0.00000000013,    -0.00000000012,
       0.00000000013,    -0.00000000016,
       0.00000000074,    -0.00000000087,
       0.00000000088,     0.00000000074,
       0.00000000095,     0.00000000083,
      -0.00000001015,     0.00000000569,
      -0.00000000054,    -0.00000000081,
      -0.00000000568,    -0.00000001017,
       0.00000000083,    -0.00000000054,
      -0.00000145309,    -0.00000382911,
       0.00000093683,    -0.00000015375,
       0.00000000352,     0.00000013841,
      -0.00000001167,    -0.00000000195,
      -0.00000121587,     0.00000049905,
      -0.00000005943,    -0.00000029858,
       0.00000004506,    -0.00000000319,
      -0.00000000038,     0.00000000374,
       0.00000020889,    -0.00000034576,
      -0.00000000699,     0.00000015649,
      -0.00000002936,    -0.00000000883,
       0.00000000164,    -0.00000000375,
       0.00000042237,    -0.00000052620,
       0.00000012817,     0.00000006494,
      -0.00000001177,     0.00000002722,
      -0.00000000363,    -0.00000000186,
       0.00000010792,    -0.00000014748,
      -0.00000002034,    -0.00000001478,
      -0.00000000061,     0.00000000119,
       0.00000014639,     0.00000011154,
       0.00000001432,    -0.00000002055,
      -0.00000000117,    -0.00000000070,
       0.00007251626,     0.00005343563,
      -0.00000321611,     0.00000818002,
      -0.00000054929,    -0.00000028923,
       0.00000002604,    -0.00000001475,
      -0.00000014810,     0.00000437741,
      -0.00000048133,     0.00000085698,
      -0.00000014361,    -0.00000003798,
       0.00000000307,    -0.00000001207,
      -0.00000225652,    -0.00000029672,
       0.00000014660,    -0.00000066709,
       0.00000012728,     0.00000004639,
      -0.00000000800,     0.00000001647,
      -0.00000000159,    -0.00000000087,
      -0.00000036298,     0.00000083485,
      -0.00000056954,    -0.00000023175,
       0.00000005382,    -0.00000012140,
       0.00000001627,     0.00000000828,
      -0.00000000093,     0.00000000164,
      -0.00000024210,     0.00000035804,
       0.00000004880,     0.00000003395,
       0.00000000155,    -0.00000000319,
      -0.00000037011,    -0.00000029957,
      -0.00000003482,     0.00000004839,
       0.00000000311,     0.00000000195,
       0.00000025502,     0.00000009191,
       0.00000000914,     0.00000003444,
      -0.00000000224,    -0.00000000468,
       0.00000000078,    -0.00000000026,
      -0.00000005846,    -0.00000001938,
      -0.00000000103,    -0.00000000263,
       0.00000001545,    -0.00000005991,
       0.00000000296,    -0.00000000157,
       0.00000055338,     0.00000030236,
       0.00000001456,     0.00000006203,
      -0.00000000037,     0.00000000067,
      -0.00000007473,    -0.00000002428,
      -0.00000000149,    -0.00000000341,
       0.00000002429,    -0.00000007469,
       0.00000000338,    -0.00000000149,
       0.00000000025,    -0.00000000045,
       0.00000002016,    -0.00000001104,
      -0.00000003499,     0.00000001163,
       0.00000000055,     0.00000000051,
      -0.00000000069,     0.00000000122,
       0.00000000118,     0.00000000059,
      -0.00000000771,    -0.00000000107,
      -0.00000000027,     0.00000000254,
       0.00000000062,     0.00000000067,
       0.00000000460,    -0.00000000082,
       0.00000000713,     0.00000001676,
       0.00000000083,    -0.00000000654,
      -0.00000000324,    -0.00000000957,
       0.00000000171,     0.00000000127,
      -0.00000001373,     0.00000000356,
       0.00000000180,    -0.00000000140,
       0.00000001137,     0.00000002850,
       0.00000000256,     0.00000000172,
       0.00000000384,    -0.00000000365,
      -0.00000000310,    -0.00000000385,
      -0.00000000424,     0.00000000195,
       0.00000000044,     0.00000000060,
       0.00000001571,    -0.00000000093,
      -0.00000000037,     0.00000000105,
       0.00000003121,    -0.00000006031,
      -0.00000000555,    -0.00000000497,
      -0.00000002065,     0.00000000210,
      -0.00000000126,    -0.00000000170,
       0.00000000182,     0.00000002041,
      -0.00000000169,     0.00000000125,
      -0.00000000111,     0.00000000021,
      -0.00000155959,    -0.00000033280,
      -0.00000004748,    -0.00000008628,
       0.00000003697,    -0.00000007974,
       0.00000000263,    -0.00000000045,
       0.00000000102,    -0.00000000197,
      -0.00000000135,     0.00000000007,
       0.00000000687,    -0.00000006464,
       0.00000000128,     0.00000000037,
       0.00000000009,     0.00000000172,
       0.00000000067,     0.00000000207,
      -0.00000004082,     0.00000005321,
      -0.00000000229,    -0.00000000179,
      -0.00000299565,    -0.00000098347,
      -0.00000005513,    -0.00000013348,
       0.00000003199,    -0.00000009848,
       0.00000000448,    -0.00000000197,
       0.00000000135,     0.00000000137,
       0.00000000134,    -0.00000000271,
       0.00000000209,     0.00000000122,
      -0.00000000122,     0.00000000209,
       0.00000000012,     0.00000000003,
      -0.00000000003,     0.00000000012,
       0.00000043811,    -0.00000004805,
       0.00000002358,     0.00000021283,
      -0.00000005407,     0.00000000601,
      -0.00000000102,    -0.00000000933,
      -0.00000001946,    -0.00000010141,
       0.00000005056,    -0.00000000972,
       0.00000000242,     0.00000001310,
      -0.00000000232,     0.00000000040,
       0.00000017796,     0.00000001063,
      -0.00000001251,     0.00000006993,
      -0.00000001407,    -0.00000000359,
       0.00000000057,    -0.00000000189,
      -0.00000002610,     0.00000017643,
      -0.00000006792,    -0.00000002272,
       0.00000000700,    -0.00000001309,
       0.00000000162,     0.00000000126,
      -0.00000000308,     0.00000000042,
      -0.00000000042,    -0.00000000308,
      -0.00000153271,     0.00000012538,
      -0.00000006208,    -0.00000073610,
       0.00000018494,    -0.00000001777,
       0.00000000336,     0.00000003175,
       0.00000004665,     0.00000023737,
      -0.00000011875,     0.00000002335,
      -0.00000000581,    -0.00000003081,
       0.00000000542,    -0.00000000094,
      -0.00000018733,    -0.00000002514,
       0.00000000543,    -0.00000008221,
       0.00000001769,    -0.00000000006,
       0.00000000045,     0.00000000266,
       0.00000009026,    -0.00000018426,
       0.00000007562,     0.00000004867,
      -0.00000001448,     0.00000001403,
      -0.00000000151,    -0.00000000278,
       0.00000002177,    -0.00000000748,
       0.00000000835,     0.00000002220,
       0.00000000159,    -0.00000000088,
       0.00000000088,     0.00000000160,
       0.00000000185,    -0.00000000118,
       0.00000000118,     0.00000000185,
       0.00000000638,    -0.00000001968,
       0.00000000017,     0.00000000264,
       0.00000000260,    -0.00000000016,
      -0.00000000381,     0.00000005280,
       0.00000000302,    -0.00000000023,
      -0.00000000046,    -0.00000000551,
      -0.00000000049,     0.00000000065,
      -0.00000000536,     0.00000000045,
       0.00000000063,     0.00000000047,
      -0.00000002186,    -0.00000009549,
      -0.00000000840,     0.00000001147,
      -0.00000000676,     0.00000000058,
       0.00000000080,     0.00000000059,
      -0.00000000052,     0.00000000115,
       0.00000000064,     0.00000000056,
       0.00000004105,     0.00000002023,
       0.00000000150,     0.00000000100,
      -0.00000000039,     0.00000000124,
       0.00000000225,     0.00000000024,
       0.00000000232,     0.00000000820,
       0.00000000125,    -0.00000000035,
      -0.00000011592,     0.00000004848,
      -0.00000000268,    -0.00000000222,
       0.00000000080,    -0.00000000080,
      -0.00000000098,    -0.00000000073,
      -0.00000003433,    -0.00000000583,
       0.00000000578,    -0.00000000694,
      -0.00000000189,     0.00000001037,
      -0.00000000207,    -0.00000000177,
      -0.00000000034,     0.00000000098,
      -0.00000000045,     0.00000000105,
       0.00000015548,     0.00000002521,
      -0.00000002620,     0.00000003170,
       0.00000000428,    -0.00000002445,
       0.00000000480,     0.00000000408,
       0.00000000224,    -0.00000000210,
       0.00000000022,     0.00000000125,
      -0.00000000009,    -0.00000000344,
      -0.00000000226,    -0.00000002452,
       0.00000000282,     0.00000000269,
       0.00000017475,     0.00000009692,
       0.00000000462,     0.00000001932,
       0.00000000079,    -0.00000000137,
       0.00000000066,    -0.00000000065,
       0.00000003953,    -0.00000002400,
       0.00000000127,     0.00000000176,
       0.00000000192,     0.00000000101,
      -0.00000000107,     0.00000000185,
       0.00000006214,     0.00000000874,
       0.00000007559,    -0.00000004646,
       0.00000001393,     0.00000002220,
       0.00000000146,     0.00000000246,
      -0.00000000071,     0.00000000044,
       0.00000000266,     0.00000000141,
      -0.00000000141,     0.00000000268,
      -0.00000000003,    -0.00000000007,
       0.00000000007,    -0.00000000003,
       0.00000001586,    -0.00000005225,
       0.00000003566,     0.00000001256,
      -0.00000000505,     0.00000001223,
      -0.00000001556,    -0.00000000103,
       0.00000000173,    -0.00000001034,
       0.00000000348,     0.00000000091,
       0.00000002635,     0.00000000043,
      -0.00000000071,     0.00000001329,
      -0.00000000358,    -0.00000000040,
       0.00000000019,     0.00000002062,
      -0.00000000890,     0.00000000045,
      -0.00000000032,    -0.00000000189,
       0.00000000105,    -0.00000000130,
       0.00000000129,     0.00000000109,
      -0.00000000549,     0.00000025209,
      -0.00000017284,    -0.00000002219,
       0.00000001391,    -0.00000005925,
       0.00000001330,     0.00000000463,
       0.00000003720,     0.00000000311,
      -0.00000000444,     0.00000002474,
      -0.00000000838,    -0.00000000225,
       0.00000000068,    -0.00000000188,
      -0.00000003396,    -0.00000000168,
       0.00000000313,    -0.00000001972,
       0.00000000616,     0.00000000176,
      -0.00000000063,     0.00000000137,
      -0.00000000192,    -0.00000000835,
       0.00000000011,    -0.00000000265,
       0.00000000163,    -0.00000000134,
      -0.00000000247,     0.00000000366,
      -0.00000000072,    -0.00000000087,
      -0.00000000353,    -0.00000000267,
       0.00000000097,    -0.00000000063,
       0.00000000134,    -0.00000000011,
      -0.00000000009,    -0.00000000150,
       0.00000003871,    -0.00000000803,
      -0.00000000013,    -0.00000000224,
       0.00000002615,     0.00000000289,
      -0.00000000525,     0.00000001996,
      -0.00000000762,    -0.00000000316,
       0.00000000104,    -0.00000000691,
       0.00000000526,     0.00000000163,
      -0.00000000094,     0.00000000201,
       0.00000000124,    -0.00000000226,
       0.00000000160,     0.00000000102,
       0.00000000209,     0.00000000011,
       0.00000000001,     0.00000000152,
      -0.00000011588,    -0.00000001413,
       0.00000002488,    -0.00000008945,
       0.00000003440,     0.00000001491,
      -0.00000000518,     0.00000000871,
      -0.00000000222,     0.00000001608,
      -0.00000001225,    -0.00000000366,
       0.00000000215,    -0.00000000469,
       0.00000000118,     0.00000000073,
      -0.00000000199,     0.00000000364,
      -0.00000000248,    -0.00000000214,
       0.00000000109,    -0.00000000081,
      -0.00000000258,     0.00000000255,
      -0.00000000203,    -0.00000000196,
       0.00000000075,    -0.00000000086,
      -0.00000000007,    -0.00000000115,
       0.00000000143,     0.00000000007,
      -0.00000000019,     0.00000000112,
      -0.00000000024,     0.00000000135,
      -0.00000000101,    -0.00000000037,
      -0.00000001715,     0.00000000442,
      -0.00000000109,    -0.00000001529,
      -0.00000000025,     0.00000000294,
      -0.00000000258,    -0.00000000020,
      -0.00000000185,     0.00000000011,
       0.00000000002,    -0.00000000153,
       0.00000000062,    -0.00000000157,
       0.00000000113,     0.00000000078,
      -0.00000000001,     0.00000000006,
      -0.00000000006,    -0.00000000001,
       0.00000000148,     0.00000000040,
       0.00000000037,    -0.00000000143,
       0.00000003460,     0.00000000885,
       0.00000000055,    -0.00000000206,
       0.00000000072,     0.00000000029,
      -0.00000000225,    -0.00000000022,
       0.00000000022,    -0.00000000087,
      -0.00000000013,     0.00000000204,
      -0.00000005194,    -0.00000000442,
       0.00000000481,    -0.00000002010,
      -0.00000000030,     0.00000000335,
      -0.00000000130,    -0.00000000032,
      -0.00000000038,     0.00000000129,
      -0.00000002127,    -0.00000000686,
      -0.00000000045,     0.00000000146,
       0.00000000180,     0.00000000030,
       0.00000000030,    -0.00000000179,
       0.00000003745,     0.00000001117,
       0.00000000031,    -0.00000000215,
      -0.00000001177,    -0.00000000161,
      -0.00000000023,     0.00000000109,
      -0.00000000161,     0.00000001171,
       0.00000000108,     0.00000000023,
       0.00000000012,    -0.00000000109,
      -0.00000000182,    -0.00000000018,
      -0.00000000019,     0.00000000178,
      -0.00000027144,    -0.00000004667,
      -0.00000000926,     0.00000002173,
      -0.00000000149,     0.00000001356,
       0.00000000128,     0.00000000030,
       0.00000006873,     0.00000000707,
       0.00000000084,    -0.00000000481,
       0.00000000704,    -0.00000006838,
      -0.00000000475,    -0.00000000083,
      -0.00000000136,    -0.00000000004,
      -0.00000000004,     0.00000000132,
      -0.00000001917,    -0.00000000191,
      -0.00000000056,     0.00000000705,
       0.00000000917,     0.00000000062,
       0.00000000061,    -0.00000000878,
       0.00000162941,     0.00000018104,
       0.00000002284,    -0.00000010725,
       0.00000000680,    -0.00000008362,
      -0.00000000595,    -0.00000000108,
      -0.00000034920,    -0.00000002400,
      -0.00000000214,     0.00000001634,
      -0.00000002384,     0.00000034633,
       0.00000001606,     0.00000000207,
      -0.00000003505,    -0.00000000085,
      -0.00000000005,     0.00000000184,
       0.00000011492,     0.00000000824,
       0.00000000067,    -0.00000000534,
       0.00000000214,    -0.00000003993,
      -0.00000000190,    -0.00000000028,
      -0.00000003657,    -0.00000000116,
      -0.00000000007,     0.00000000094,
      -0.00000000109,     0.00000003366,
      -0.00000826274,    -0.00000059300,
      -0.00000004922,     0.00000038391,
      -0.00000002413,     0.00000045283,
       0.00000002167,     0.00000000314,
       0.00000139701,     0.00000004506,
       0.00000000230,    -0.00000003336,
       0.00000004441,    -0.00000137903,
      -0.00000003203,    -0.00000000204,
      -0.00000000014,    -0.00000000098,
      -0.00000056121,    -0.00000001879,
      -0.00000000078,     0.00000001317,
      -0.00000000470,     0.00000018421,
       0.00000000444,     0.00000000042,
       0.00000009182,    -0.00000000084,
      -0.00000000112,    -0.00000007190,
       0.00003236242,     0.00000111251,
       0.00000004561,    -0.00000075648,
      -0.00000001602,     0.00000000497,
       0.00000004933,    -0.00000198486,
      -0.00000004794,    -0.00000000444,
       0.00000000018,     0.00000000110,
      -0.00000336947,     0.00000003872,
       0.00000000446,     0.00000000605,
      -0.00000000026,    -0.00000000118,
       0.00000004090,     0.00000327051,
       0.00000000108,    -0.00000000531,
      -0.00000000125,     0.00000000044,
      -0.00000000139,     0.00000000193,
       0.00000000146,     0.00000000166,
       0.00000173418,    -0.00000000052,
      -0.00000000510,    -0.00000000415,
       0.00000000211,    -0.00000054094,
      -0.00000000108,     0.00000000148,
      -0.00000004280,     0.00000000174,
       0.00000000075,     0.00000000142,
       0.00000000243,    -0.00000007339,
      -0.00000000158,    -0.00000000116,
      -0.00000003829,    -0.00000010253,
       0.00000001364,     0.00000001376,
      -0.00000001384,     0.00000000493,
       0.00000000185,    -0.00000000199,
       0.00000001672,     0.00000000336,
      -0.00000000293,     0.00000000341,
       0.00000000300,    -0.00000001763,
       0.00000000354,     0.00000000300,
      -0.00007565664,     0.00000058678,
       0.00000009263,     0.00000005751,
      -0.00000000608,    -0.00000002959,
       0.00000003668,     0.00000546727,
       0.00000000289,    -0.00000000709,
      -0.00000000216,     0.00000000035,
      -0.00000060290,    -0.00000020424,
      -0.00000001227,    -0.00000004645,
       0.00000000217,    -0.00000000194,
      -0.00000020813,     0.00000103895,
      -0.00000003539,     0.00000001375,
      -0.00000000194,    -0.00000000246,
       0.00000000380,    -0.00000000564,
      -0.00000000422,    -0.00000000430,
      -0.00000000113,    -0.00000000003,
       0.00000000004,    -0.00000000114,
      -0.00000000638,     0.00000000128,
       0.00000000049,     0.00000000154,
      -0.00000000173,    -0.00000000064,
       0.00000000001,    -0.00000000186,
       0.00000002623,    -0.00000000512,
      -0.00000000106,     0.00000000934,
      -0.00000000112,    -0.00000000369,
       0.00000000138,     0.00000000002,
       0.00000000184,     0.00000000173,
       0.00000000099,     0.00000000246,
      -0.00000001182,    -0.00000001248,
      -0.00000000311,     0.00000001527,
      -0.00000000732,    -0.00000000423,
       0.00000000192,     0.00000000035,
      -0.00000000048,     0.00000000094,
      -0.00000000081,    -0.00000000111,
       0.00000000083,    -0.00000000048,
      -0.00000000097,     0.00000000151,
      -0.00000000074,    -0.00000000087,
       0.00000000063,    -0.00000000071,
      -0.00000001243,    -0.00000001043,
       0.00000000862,    -0.00000000645,
      -0.00000000147,     0.00000000167,
      -0.00000000087,    -0.00000000121,
       0.00000000306,    -0.00000000014,
      -0.00000000014,    -0.00000000302,
      -0.00000004679,    -0.00000000155,
       0.00000000005,    -0.00000000149,
      -0.00000004484,    -0.00000003671,
       0.00000000172,    -0.00000000210,
       0.00000042445,     0.00000007983,
       0.00000000789,     0.00000001732,
       0.00000001664,    -0.00000009854,
       0.00000000387,    -0.00000000170,
       0.00000023598,     0.00000000085,
       0.00000000009,     0.00000023301,
       0.00000000216,    -0.00000000021,
      -0.00000000028,     0.00000000226,
      -0.00000495899,    -0.00000398619,
      -0.00000001422,    -0.00000060611,
       0.00000004701,    -0.00000002458,
      -0.00000004347,     0.00000075279,
      -0.00000002041,     0.00000001148,
      -0.00000000085,    -0.00000000039,
       0.00000169592,     0.00000126413,
       0.00000006093,     0.00000025813,
      -0.00000001855,     0.00000001364,
      -0.00000000119,    -0.00000000092,
       0.00000125828,    -0.00000288722,
       0.00000025759,    -0.00000005943,
       0.00000001413,     0.00000001848,
      -0.00000000092,     0.00000000119,
      -0.00000000230,    -0.00000000255,
       0.00000000327,    -0.00000000007,
      -0.00000000014,     0.00000000433,
       0.00000005363,    -0.00000000305,
      -0.00000000022,    -0.00000000329,
      -0.00000000532,     0.00000000057,
       0.00000000055,     0.00000000513,
       0.00000015959,     0.00000000194,
      -0.00000000002,     0.00000000484,
      -0.00000001353,    -0.00000000750,
       0.00000000070,     0.00000000048,
       0.00000000052,    -0.00000000088,
       0.00000007183,     0.00000005378,
      -0.00000000163,     0.00000000217,
       0.00000000862,     0.00000000034,
       0.00000000002,    -0.00000000165,
       0.00000000414,    -0.00000000089,
       0.00000000091,     0.00000000409,
       0.00017092697,     0.00009340380,
       0.00000446883,     0.00001914717,
      -0.00000137411,     0.00000103082,
      -0.00000008741,    -0.00000006760,
      -0.00000452256,     0.00000828682,
      -0.00000092762,     0.00000021655,
      -0.00000004992,    -0.00000006656,
       0.00000000327,    -0.00000000424,
       0.00000022099,     0.00000006057,
       0.00000000598,     0.00000001899,
      -0.00000000130,     0.00000000109,
       0.00000006943,    -0.00000010607,
       0.00000001872,    -0.00000000228,
       0.00000000098,     0.00000000149,
      -0.00000001050,     0.00000000097,
      -0.00000000148,    -0.00000000107,
       0.00000001025,     0.00000001821,
      -0.00000000131,     0.00000000185,
       0.00000061662,     0.00000019495,
       0.00000001592,     0.00000005041,
      -0.00000000083,     0.00000000108,
       0.00000002793,     0.00000001375,
       0.00000000054,     0.00000000348,
       0.00000001411,    -0.00000003080,
       0.00000000357,    -0.00000000057,
      -0.00000008497,     0.00000000988,
       0.00000000071,     0.00000000616,
      -0.00000000169,    -0.00000000009,
      -0.00000000004,     0.00000000221,
      -0.00000000135,     0.00000000029,
       0.00000000029,     0.00000000135,
       0.00000000505,    -0.00000000024,
      -0.00000000008,    -0.00000000147,
      -0.00000018883,     0.00000000669,
       0.00000000086,     0.00000000867,
       0.00000000184,     0.00000000106,
       0.00000000103,    -0.00000000226,
       0.00000179037,     0.00000085093,
       0.00000003365,     0.00000021497,
      -0.00000002644,     0.00000005562,
      -0.00000000667,     0.00000000104,
       0.00000000233,     0.00000000058,
      -0.00000000099,     0.00000000245,
      -0.00000006278,    -0.00000002780,
      -0.00000000108,     0.00000000217,
       0.00000004985,     0.00000001976,
       0.00000000127,     0.00000000576,
      -0.00000001977,     0.00000004991,
      -0.00000000577,     0.00000000127,
      -0.00000001314,     0.00000000517,
      -0.00000000153,    -0.00000000109,
      -0.00000000516,    -0.00000001313,
       0.00000000109,    -0.00000000152,
       0.00000000871,     0.00000000044,
       0.00000000043,    -0.00000000871,
      -0.00000000517,    -0.00000000025,
      -0.00000000008,     0.00000000154,
       0.00000021889,     0.00000001662,
       0.00000000056,    -0.00000001048,
       0.00000000106,     0.00000000055,
       0.00000000060,    -0.00000000114,
       0.00000005678,     0.00000002177,
      -0.00000000111,     0.00000000297,
      -0.00000031901,    -0.00000008235,
      -0.00000000061,     0.00000000058,
      -0.00000002648,    -0.00000001688,
      -0.00000000068,    -0.00000000269,
      -0.00000001699,     0.00000002813,
      -0.00000000267,     0.00000000067,
      -0.00000154525,    -0.00000096177,
      -0.00000003814,    -0.00000015231,
       0.00000003172,    -0.00000005094,
       0.00000000503,    -0.00000000126,
      -0.00000000367,    -0.00000000147,
      -0.00000000075,     0.00000000116,
      -0.00000000136,    -0.00000000049,
       0.00000000048,    -0.00000000134,
      -0.00000000107,    -0.00000000034,
      -0.00000000033,     0.00000000113,
      -0.00000006621,    -0.00000002138,
       0.00000000071,    -0.00000000211,
      -0.00000000087,    -0.00000000036,
      -0.00000000036,     0.00000000076,
      -0.00000000069,     0.00000000160,
       0.00000000102,    -0.00000000228,
      -0.00000002408,    -0.00000001295,
      -0.00000000055,    -0.00000000075,
      -0.00000000058,     0.00000000044,
      -0.00000000044,    -0.00000000057,
       0.00000002109,    -0.00000000199,
       0.00000000007,     0.00000000147,
       0.00000000748,     0.00000001212,
       0.00000000054,    -0.00000000089,
       0.00000001211,    -0.00000000748,
      -0.00000000089,    -0.00000000054,
      -0.00000000364,    -0.00000000759,
      -0.00000000211,     0.00000000198,
       0.00000000977,    -0.00000000003,
      -0.00000000011,    -0.00000000974,
       0.00000000011,     0.00000000160,
       0.00000000115,     0.00000000072,
       0.00000000061,    -0.00000000202,
       0.00000028253,     0.00000044223,
       0.00000002067,    -0.00000003383,
       0.00000001514,    -0.00000001468,
      -0.00000000141,    -0.00000000084,
      -0.00000041539,     0.00000000178,
       0.00000000193,     0.00000041535,
       0.00000032686,    -0.00000000192,
      -0.00000000042,    -0.00000009712,
       0.00000000139,    -0.00000000426,
      -0.00001385052,    -0.00000030287,
       0.00000004313,    -0.00000000009,
       0.00000000138,     0.00000066082,
      -0.00000004399,    -0.00000003278,
       0.00000000206,    -0.00000000257,
      -0.00000003573,     0.00000004927,
      -0.00000000215,    -0.00000000182,
      -0.00000000096,     0.00000000005,
       0.00000000005,     0.00000000104,
       0.00000010118,     0.00000002416,
       0.00000000243,     0.00000000471,
       0.00000000496,    -0.00000002132,
       0.00000000100,    -0.00000000052,
       0.00000006958,     0.00000000010,
       0.00000000011,     0.00000006913,
       0.00000120861,     0.00000075088,
       0.00000002981,     0.00000011891,
      -0.00000003716,     0.00000006033,
      -0.00000000589,     0.00000000147,
       0.00000000928,     0.00000000454,
       0.00000000024,     0.00000000094,
       0.00000000458,    -0.00000000783,
       0.00000000094,    -0.00000000019,
       0.00003328841,     0.00000656237,
       0.00000103898,     0.00000152829,
      -0.00000007315,     0.00000010563,
       0.00000004255,    -0.00000006593,
       0.00000000299,     0.00000000125,
       0.00000271004,     0.00000152091,
       0.00000007298,     0.00000031200,
      -0.00000002240,     0.00000001679,
      -0.00000000143,    -0.00000000110,
       0.00000152187,    -0.00000286298,
       0.00000031201,    -0.00000007272,
       0.00000001686,     0.00000002240,
      -0.00000000110,     0.00000000143,
      -0.00000000119,    -0.00000000032,
      -0.00000000033,     0.00000000127,
      -0.00000000581,    -0.00000001511,
       0.00000000039,     0.00000000084,
       0.00000000084,    -0.00000000039,
      -0.00000031460,     0.00000024940,
       0.00000021710,    -0.00000023110,
       0.00000002200,     0.00000004650,
       0.00000169110,    -0.00000080260,
      -0.00000000039,    -0.00000000084,
       0.00000000083,    -0.00000000039,
      -0.00000000840,    -0.00000001720,
       0.00000000597,     0.00000001522,
      -0.00000000055,     0.00000000234,
      -0.00000000096,    -0.00000000032,
      -0.00000000055,    -0.00000000658,
      -0.00000000082,    -0.00000000050,
      -0.00000000052,     0.00000000081,
      -0.00000000036,    -0.00000000066,
       0.00000000164,    -0.00000000297,
      -0.00000000142,    -0.00000000046,
      -0.00000000297,    -0.00000000164,
      -0.00000000046,     0.00000000143,
       0.00000000063,    -0.00000000114,
      -0.00000000114,    -0.00000000063,
       0.00000000856,    -0.00000002129,
      -0.00000000299,     0.00000001244,
       0.00000000251,     0.00000000090,
       0.00000001244,     0.00000000311,
       0.00000000095,    -0.00000000251,
      -0.00000000083,     0.00000000046,
       0.00000000096,     0.00000000009,
       0.00000000008,    -0.00000000097,
      -0.00000007308,     0.00000035957,
       0.00000008364,     0.00000002320,
       0.00000000545,    -0.00000002006,
      -0.00000001453,    -0.00000000298,
      -0.00000000093,     0.00000000340,
       0.00000000085,     0.00000000024,
       0.00000001782,    -0.00000002652,
       0.00000003013,    -0.00000000471,
      -0.00000000484,    -0.00000000594,
      -0.00000000063,     0.00000000118,
      -0.00000000471,    -0.00000003013,
      -0.00000000594,     0.00000000484,
       0.00000000118,     0.00000000063,
       0.00000001176,    -0.00000002266,
       0.00000000102,     0.00000000051,
       0.00000068852,    -0.00000017162,
      -0.00000009679,    -0.00000026820,
      -0.00000005295,     0.00000002612,
      -0.00000000023,    -0.00000000523,
      -0.00000000202,     0.00000000043,
       0.00000002333,    -0.00000005194,
      -0.00000001402,    -0.00000000432,
      -0.00000000031,     0.00000000205,
      -0.00000005188,    -0.00000002331,
      -0.00000000431,     0.00000001400,
       0.00000000205,     0.00000000031,
       0.00000005648,     0.00000009910,
       0.00000001359,    -0.00000001064,
      -0.00000000464,     0.00000000264,
       0.00000000050,     0.00000000063,
       0.00000016903,     0.00000037540,
       0.00000009851,    -0.00000003120,
       0.00000000233,     0.00000000337,
       0.00000013786,     0.00000021350,
       0.00000002637,    -0.00000002207,
      -0.00000000196,    -0.00000000147,
       0.00000021347,    -0.00000013793,
      -0.00000002208,    -0.00000002637,
      -0.00000000147,     0.00000000196,
       0.00000000158,     0.00000000276,
       0.00000000274,    -0.00000000156,
      -0.00000000152,     0.00000000099,
       0.00000026420,     0.00000002366,
      -0.00000000114,     0.00000001276,
       0.00000000174,     0.00000000003,
       0.00000000006,    -0.00000000189,
       0.00000061336,    -0.00000040635,
       0.00000002213,    -0.00000001824,
       0.00000000514,     0.00000000500,
       0.00000058947,     0.00000002006,
       0.00000000387,    -0.00000000028,
       0.00000002011,    -0.00000058943,
      -0.00000000028,    -0.00000000387,
      -0.00000000097,    -0.00000000022,
      -0.00000000022,     0.00000000098,
      -0.00000002682,    -0.00000000566,
       0.00000000734,     0.00000000065,
       0.00000000065,    -0.00000000733,
      -0.00000000014,     0.00000000433,
       0.00000000004,    -0.00000000154,
      -0.00000009730,     0.00000000741,
      -0.00000002625,    -0.00000000820,
      -0.00000000052,    -0.00000000120,
      -0.00000000822,     0.00000002624,
      -0.00000000120,     0.00000000052,
       0.00000039349,     0.00000020898,
       0.00000001145,     0.00000004418,
       0.00000000162,    -0.00000000300,
       0.00000000065,    -0.00000000040,
      -0.00000000039,    -0.00000000064,
      -0.00000001372,     0.00000002047,
       0.00000000356,     0.00000000057,
       0.00000000057,    -0.00000000354,
       0.00000007301,     0.00000001207,
       0.00000000052,    -0.00000000372,
      -0.00000001552,    -0.00000000196,
      -0.00000000008,     0.00000000096,
      -0.00000000195,     0.00000001538,
       0.00000000095,     0.00000000008,
       0.00000000017,    -0.00000000149,
      -0.00000000134,    -0.00000000012,
      -0.00000000011,     0.00000000125,
      -0.00000030678,    -0.00000003957,
      -0.00000000152,     0.00000001908,
      -0.00000000187,     0.00000001687,
       0.00000000107,     0.00000000010,
       0.00000005160,     0.00000000463,
      -0.00000000006,    -0.00000000203,
       0.00000000457,    -0.00000005092,
      -0.00000000197,     0.00000000007,
      -0.00000001669,    -0.00000000151,
      -0.00000000046,     0.00000000548,
       0.00000000268,     0.00000000013,
       0.00000000009,    -0.00000000209,
       0.00000096048,     0.00000008812,
      -0.00000000141,    -0.00000003751,
       0.00000000488,    -0.00000005910,
      -0.00000000234,     0.00000000005,
      -0.00000009807,    -0.00000000448,
       0.00000000062,     0.00000000167,
      -0.00000000425,     0.00000009509,
       0.00000000147,    -0.00000000064,
       0.00000003828,     0.00000000196,
       0.00000000060,    -0.00000001191,
       0.00000000017,    -0.00000000198,
      -0.00000166808,    -0.00000008233,
       0.00000001035,     0.00000002689,
      -0.00000000624,     0.00000012077,
       0.00000000192,    -0.00000000077,
      -0.00000002271,    -0.00000000526,
      -0.00000000022,    -0.00000000081,
      -0.00000000592,     0.00000003255,
       0.00000000139,     0.00000000003,
       0.00000000003,    -0.00000000144,
       0.00000001090,     0.00000000140,
       0.00000000033,    -0.00000000259,
       0.00000000356,     0.00000000023,
      -0.00000000019,     0.00000000344,
      -0.00000016532,    -0.00000006365,
      -0.00000000131,     0.00000002420,
       0.00000002267,     0.00000001975,
       0.00000000085,     0.00000000334,
       0.00000002069,    -0.00000004060,
       0.00000000304,    -0.00000000074,
      -0.00000005336,     0.00000000773,
       0.00000000028,     0.00000000174,
      -0.00000000858,     0.00000000035,
       0.00000000033,     0.00000000871,
       0.00000001804,     0.00000001007,
      -0.00000001128,     0.00000002075,
      -0.00000000232,     0.00000000054,
       0.00000003899,     0.00000002141,
       0.00000000397,    -0.00000000736,
       0.00000000082,    -0.00000000019,
       0.00000000185,     0.00000000103,
      -0.00000000104,     0.00000000185,
      -0.00000000020,    -0.00000000085,
      -0.00000000085,     0.00000000020,
      -0.00000000665,    -0.00000002927,
      -0.00000001566,     0.00000000670,
      -0.00000000201,     0.00000000050,
       0.00000000047,     0.00000000109,
      -0.00000000133,    -0.00000000067,
       0.00000000143,    -0.00000000204,
      -0.00000000064,    -0.00000000187,
       0.00000001589,    -0.00000000577,
      -0.00000000344,    -0.00000000753,
       0.00000000321,     0.00000000011,
      -0.00000000020,    -0.00000000144,
      -0.00000000256,     0.00000000389,
       0.00000000098,     0.00000000098,
       0.00000000590,     0.00000000225,
       0.00000000041,    -0.00000000206,
       0.00000000619,    -0.00000000039,
       0.00000000008,    -0.00000000184,
      -0.00000000087,    -0.00000000734,
      -0.00000000164,     0.00000000023,
      -0.00000000719,     0.00000000022,
      -0.00000000010,     0.00000000164,
      -0.00000003500,    -0.00000002750,
      -0.00000000660,     0.00000001016,
      -0.00000001537,     0.00000000020,
      -0.00000000120,     0.00000000550,
      -0.00000000033,    -0.00000000897,
      -0.00000000190,    -0.00000000018,
      -0.00000001008,     0.00000000132,
       0.00000000049,     0.00000000235,
      -0.00000000104,     0.00000000000,
      -0.00000001759,    -0.00000024484,
      -0.00000005241,     0.00000000376,
       0.00000001646,    -0.00000000109,
      -0.00000000024,    -0.00000000334,
       0.00000000063,    -0.00000000072,
       0.00000001059,     0.00000000718,
       0.00000000138,    -0.00000000255,
      -0.00000000183,    -0.00000000400,
       0.00000000423,     0.00000000129,
      -0.00000000035,    -0.00000000552,
       0.00000000552,    -0.00000000036,
      -0.00000000045,    -0.00000000722,
       0.00000000720,    -0.00000000040,
      -0.00000000089,     0.00000000043,
       0.00000000511,     0.00000001082,
       0.00000000199,    -0.00000000094,
       0.00032843328,     0.00017926664,
       0.00000858647,     0.00003677567,
      -0.00000264010,     0.00000198318,
      -0.00000016816,    -0.00000012972,
      -0.00000573459,     0.00001050065,
      -0.00000117571,     0.00000027441,
      -0.00000006339,    -0.00000008441,
       0.00000000415,    -0.00000000538,
       0.00000025138,     0.00000001811,
       0.00000000977,     0.00000000826,
      -0.00000000021,     0.00000000081,
      -0.00000009915,     0.00000027000,
      -0.00000001575,     0.00000000465,
      -0.00000000069,    -0.00000000100,
       0.00000002194,     0.00000001136,
       0.00000000026,     0.00000000134,
       0.00000001130,    -0.00000002381,
       0.00000000128,    -0.00000000018,
      -0.00000004882,    -0.00000005623,
      -0.00000000219,     0.00000000101,
      -0.00000000276,    -0.00000001271,
       0.00000000082,    -0.00000000115,
       0.00000001271,    -0.00000000277,
       0.00000000115,     0.00000000082,
       0.00000002179,     0.00000001078,
       0.00000000021,     0.00000000125,
      -0.00000001078,     0.00000002179,
      -0.00000000125,     0.00000000021,
       0.00000000167,     0.00000000000,
       0.00000000000,    -0.00000000161,
       0.00000003182,    -0.00000000012,
       0.00000000000,    -0.00000000216,
       0.00000001018,    -0.00000000915,
      -0.00000000756,    -0.00000000513,
      -0.00000000098,     0.00000000247,
       0.00000000915,     0.00000001018,
       0.00000000513,    -0.00000000756,
      -0.00000000247,    -0.00000000098,
      -0.00000003029,    -0.00000001268,
      -0.00000000738,     0.00000002470,
      -0.00000002799,     0.00000000921,
       0.00000001019,     0.00000002122,
       0.00000000799,    -0.00000000524,
      -0.00000000173,    -0.00000000196,
      -0.00000000921,    -0.00000002799,
      -0.00000002122,     0.00000001019,
       0.00000000524,     0.00000000799,
       0.00000000196,    -0.00000000173,
       0.00000000211,    -0.00000000358,
       0.00000002867,    -0.00000000700,
       0.00000000050,     0.00000000163,
       0.00000000112,    -0.00000000011,
       0.00000000007,     0.00000000115,
       0.00000000021,     0.00000000137,
      -0.00000000096,     0.00000000040,
      -0.00000000088,    -0.00000000845,
       0.00000001415,    -0.00000000173,
       0.00000000177,     0.00000001426,
       0.00000000466,    -0.00000000115,
       0.00000000066,     0.00000000138,
      -0.00000000359,     0.00000000672,
       0.00000000720,     0.00000000287,
      -0.00000004816,     0.00000000621,
      -0.00000000103,    -0.00000000403,
      -0.00000000190,     0.00000000084,
      -0.00000000365,     0.00000000387,
       0.00000000165,     0.00000000134,
       0.00000000039,    -0.00000000081,
      -0.00000000389,    -0.00000000375,
      -0.00000000140,     0.00000000187,
       0.00000000081,     0.00000000039,
       0.00000000147,     0.00000000559,
       0.00000000263,     0.00000000118,
      -0.00000000091,    -0.00000000136,
      -0.00000000144,     0.00000000089,
       0.00000001006,    -0.00000002132,
      -0.00000000671,    -0.00000000319,
      -0.00000000062,     0.00000000196,
       0.00000000354,     0.00000000572,
       0.00000000061,    -0.00000000062,
       0.00000000590,    -0.00000000433,
      -0.00000000059,    -0.00000000071,
      -0.00000003052,     0.00000005554,
      -0.00000000181,    -0.00000000084,
      -0.00000000575,    -0.00000000549,
      -0.00000000252,     0.00000000000,
       0.00000000082,    -0.00000000221,
      -0.00000000151,    -0.00000000026,
      -0.00000000227,    -0.00000000066,
      -0.00000000026,     0.00000000151,
       0.00000000050,     0.00000000059,
       0.00000000059,    -0.00000000048,
      -0.00000000253,    -0.00000000403,
      -0.00000000248,     0.00000000168,
       0.00000000178,    -0.00000000009,
      -0.00000000012,    -0.00000000174,
       0.00000002441,     0.00000003899,
       0.00000000628,    -0.00000000415,
      -0.00000000078,    -0.00000000078,
      -0.00000000741,     0.00000000028,
       0.00000000049,     0.00000000744,
       0.00000002819,    -0.00000000859,
       0.00000000061,    -0.00000000040,
       0.00000000038,    -0.00000000068,
       0.00000001334,    -0.00000003664,
      -0.00000000989,    -0.00000000159,
       0.00000000181,     0.00000000067,
       0.00000000084,     0.00000000110,
       0.00000000109,    -0.00000000090,
       0.00000000074,    -0.00000000061,
      -0.00000000094,     0.00000000017,
       0.00000000018,     0.00000000096,
      -0.00000001935,     0.00000000731,
       0.00000000150,     0.00000000039,
       0.00000000456,    -0.00000002056,
      -0.00000000343,    -0.00000000093,
      -0.00000000250,     0.00000000050,
      -0.00000000049,    -0.00000000250,
      -0.00000000009,    -0.00000000100,
      -0.00000001568,     0.00000000154,
       0.00000000024,     0.00000000254,
      -0.00000002586,    -0.00000000297,
      -0.00000000012,     0.00000000120,
       0.00000000381,     0.00000000035,
       0.00000000035,    -0.00000000376,
       0.00000010392,     0.00000000895,
       0.00000000039,    -0.00000000502,
      -0.00000000941,    -0.00000000066,
      -0.00000000065,     0.00000000915,
       0.00000000008,    -0.00000000130,
      -0.00000028270,    -0.00000001605,
      -0.00000000088,     0.00000001471,
      -0.00000000263,     0.00000000086,
       0.00000000081,     0.00000000376,
      -0.00000000014,     0.00000000086,
      -0.00000000232,     0.00000000015,
       0.00000000012,     0.00000000302,
       0.00000017652,     0.00000001339,
       0.00000000153,    -0.00000000906,
       0.00000010601,    -0.00000000510,
      -0.00000000048,     0.00000000187,
      -0.00000000494,    -0.00000010958,
       0.00000000179,     0.00000000045,
      -0.00000004338,     0.00000000255,
       0.00000000062,     0.00000001285,
      -0.00000000102,    -0.00000000011,
       0.00000000752,    -0.00000000349,
      -0.00000000051,    -0.00000000103,
      -0.00000000011,     0.00000000128,
       0.00000000130,     0.00000000006,
       0.00000180848,    -0.00000007105,
      -0.00000000812,     0.00000003209,
      -0.00000000646,    -0.00000013046,
       0.00000000216,     0.00000000059,
       0.00000001321,    -0.00000000050,
      -0.00000000091,    -0.00000000874,
       0.00000000113,     0.00000000006,
       0.00000000002,    -0.00000000114,
      -0.00000000001,    -0.00000000135,
      -0.00000000100,     0.00000000135,
       0.00000000138,     0.00000000093,
      -0.00000000077,    -0.00000000052,
      -0.00000000623,     0.00000000969,
       0.00000000178,     0.00000000118,
       0.00000000076,    -0.00000000026,
      -0.00000000049,    -0.00000000086,
      -0.00000000063,     0.00000001170,
       0.00000000178,     0.00000000011,
      -0.00000000640,    -0.00000000084,
      -0.00000000017,     0.00000000149,
      -0.00000000353,     0.00000000019,
      -0.00000000020,    -0.00000000348,
      -0.00000030728,    -0.00000015399,
      -0.00000001050,    -0.00000003293,
       0.00000000360,    -0.00000001773,
       0.00000000114,    -0.00000000028,
      -0.00000002693,    -0.00000001772,
      -0.00000000072,    -0.00000000420,
      -0.00000001696,     0.00000004549,
      -0.00000000452,     0.00000000064,
      -0.00000001315,    -0.00000002255,
      -0.00000000194,    -0.00000000266,
      -0.00000000267,     0.00000000194,
      -0.00000005478,    -0.00000000040,
      -0.00000000764,     0.00000000018,
       0.00000000017,     0.00000000769,
      -0.00000135825,    -0.00000064643,
      -0.00000002540,    -0.00000016301,
       0.00000001230,    -0.00000000801,
       0.00000003075,    -0.00000006464,
       0.00000000775,    -0.00000000121,
      -0.00000000401,    -0.00000000162,
       0.00000000075,    -0.00000000144,
       0.00000007274,     0.00000004220,
      -0.00000000049,     0.00000000075,
       0.00000000105,     0.00000000079,
       0.00000000097,    -0.00000000139,
      -0.00000000951,     0.00000002522,
      -0.00000000108,    -0.00000000041,
      -0.00000007136,    -0.00000009726,
      -0.00000001004,     0.00000001059,
       0.00000000437,    -0.00000000319,
      -0.00000027945,     0.00000000614,
      -0.00000000028,    -0.00000001299,
      -0.00000000125,     0.00000000042,
       0.00000000033,     0.00000000070,
      -0.00000000112,    -0.00000000203,
       0.00000000209,    -0.00000000034,
      -0.00000000126,    -0.00000000108,
      -0.00000000063,    -0.00000000208,
      -0.00000000108,     0.00000000127,
       0.00000003954,    -0.00000000631,
      -0.00000000903,    -0.00000000732,
       0.00000000011,     0.00000000171,
      -0.00000000026,     0.00000000089,
       0.00000000103,     0.00000000014,
      -0.00000000235,     0.00000000857,
       0.00000000164,     0.00000000031,
      -0.00000000055,    -0.00000000098,
      -0.00000000086,     0.00000000154,
       0.00000000085,     0.00000000016,
      -0.00000000111,    -0.00000000009,
      -0.00000003261,    -0.00000000275,
      -0.00000000014,     0.00000000164,
       0.00000002270,     0.00000000001,
       0.00000000003,    -0.00000000136,
       0.00000000273,     0.00000000093,
       0.00000000096,    -0.00000000314,
      -0.00000000063,    -0.00000000059,
       0.00000000020,     0.00000000091,
       0.00000000092,    -0.00000000028,
       0.00000006017,     0.00000001819,
       0.00000000107,    -0.00000000403,
      -0.00000001836,    -0.00000001235,
      -0.00000000068,    -0.00000000081,
      -0.00000000102,     0.00000000142,
       0.00000000679,     0.00000001769,
      -0.00000000216,     0.00000000212,
      -0.00000000020,    -0.00000000010,
       0.00000000580,     0.00000006001,
      -0.00000000678,     0.00000000481,
      -0.00000000046,    -0.00000000030,
      -0.00000006010,     0.00000000549,
      -0.00000000479,    -0.00000000681,
       0.00000000031,    -0.00000000046,
       0.00000000254,     0.00000000199,
       0.00000000003,     0.00000000029,
      -0.00000000224,     0.00000000319,
      -0.00000000039,     0.00000000005,
       0.00000000077,    -0.00000000014,
      -0.00000000721,     0.00000000015,
      -0.00000000014,    -0.00000000097,
      -0.00000000014,    -0.00000000721,
       0.00000000097,    -0.00000000014,
      -0.00000000003,     0.00000000393,
      -0.00000000392,    -0.00000000002,
       0.00000000076,    -0.00000000031,
      -0.00000001937,     0.00000001759,
      -0.00000000352,    -0.00000000215,
      -0.00000001760,    -0.00000001937,
       0.00000000215,    -0.00000000351,
       0.00000000203,    -0.00000000323,
       0.00000000322,     0.00000000203,
       0.00000000367,    -0.00000001990,
       0.00000000413,    -0.00000000082,
       0.00000000017,     0.00000000034,
       0.00000012060,     0.00000008167,
      -0.00000000185,     0.00000001869,
      -0.00000000106,     0.00000000048,
      -0.00000008174,     0.00000012064,
      -0.00000001869,    -0.00000000185,
      -0.00000000048,    -0.00000000106,
       0.00000000024,     0.00000000420,
       0.00000000025,     0.00000000026,
      -0.00000000416,     0.00000000006,
       0.00000034183,    -0.00000004877,
       0.00000002321,     0.00000016674,
      -0.00000004250,     0.00000000596,
      -0.00000000104,    -0.00000000736,
      -0.00000001661,    -0.00000007837,
       0.00000003909,    -0.00000000808,
       0.00000000203,     0.00000001014,
      -0.00000000179,     0.00000000034,
       0.00000011439,     0.00000000256,
      -0.00000000644,     0.00000004499,
      -0.00000000915,    -0.00000000189,
      -0.00000001385,     0.00000011177,
      -0.00000004273,    -0.00000001389,
       0.00000000438,    -0.00000000823,
      -0.00000000144,     0.00000000011,
      -0.00000000014,    -0.00000000146,
      -0.00000126078,     0.00000014520,
      -0.00000007024,    -0.00000060981,
       0.00000015414,    -0.00000001918,
       0.00000000367,     0.00000002662,
       0.00000005706,     0.00000020308,
      -0.00000009567,     0.00000002379,
      -0.00000000561,    -0.00000002430,
       0.00000000422,    -0.00000000090,
       0.00000007394,     0.00000021320,
      -0.00000000438,    -0.00000001369,
       0.00000000801,     0.00000000077,
      -0.00000016570,     0.00000008293,
       0.00000000535,     0.00000002717,
      -0.00000001137,     0.00000000442,
      -0.00000000061,    -0.00000000214,
       0.00000000304,    -0.00000000015,
       0.00000000077,     0.00000000342,
      -0.00000001699,     0.00000002343,
      -0.00000000463,     0.00000000004,
       0.00000030600,     0.00000028423,
      -0.00000000128,     0.00000006325,
      -0.00000000630,     0.00000000313,
      -0.00000028429,     0.00000030605,
      -0.00000006328,    -0.00000000128,
      -0.00000000313,    -0.00000000630,
       0.00000000704,     0.00000000013,
       0.00000000075,     0.00000000085,
       0.00000038237,     0.00000035531,
      -0.00000000159,     0.00000007911,
      -0.00000000789,     0.00000000392,
      -0.00000000055,    -0.00000000050,
      -0.00000035547,     0.00000038247,
      -0.00000007914,    -0.00000000161,
      -0.00000000392,    -0.00000000789,
       0.00000000050,    -0.00000000055,
       0.00000000004,    -0.00000000006,
       0.00000000015,    -0.00000000008,
       0.00000000008,     0.00000000015,
       0.00000000003,    -0.00000000004,
       0.00000000004,     0.00000000003,
      -0.00000000023,    -0.00000000039,
      -0.00000000026,     0.00000000052,
       0.00000000016,    -0.00000000043,
      -0.00000000018,     0.00000000050,
       0.00000000011,    -0.00000000055,
      -0.00000000011,     0.00000000061,
      -0.00000000025,    -0.00000000074,
       0.00000000047,     0.00000000056,
       0.00000000035,    -0.00000000037,
      -0.00000000031,    -0.00000000046,
       0.00000000030,    -0.00000000022,
      -0.00000000078,    -0.00000000004,
      -0.00000000053,     0.00000000020,
      -0.00000000047,    -0.00000000124,
       0.00000000399,    -0.00000000145,
      -0.00000000160,     0.00000000081,
       0.00000000224,    -0.00000000039,
      -0.00000000014,    -0.00000000076,
      -0.00000002002,     0.00000000432,
      -0.00000000138,     0.00000000238,
      -0.00000000011,    -0.00000000078,
       0.00000000064,     0.00000000011,
      -0.00000000272,    -0.00000000018,
      -0.00000000017,     0.00000000047,
      -0.00000000053,    -0.00000000024,
       0.00000340436,     0.00000106161,
      -0.00002819522,    -0.00000824460,
      -0.00000000086,    -0.00000000021,
       0.00000000021,    -0.00000000086,
      -0.00000000007,    -0.00000000119,
      -0.00000000096,     0.00000000016,
       0.00000001597,    -0.00000000531,
       0.00000000034,     0.00000000097,
       0.00000000074,    -0.00000000034,
      -0.00000003199,     0.00000000332,
       0.00000000552,     0.00000002544,
      -0.00000000024,    -0.00000000212,
      -0.00000000168,     0.00000000039,
       0.00000002546,    -0.00000001071,
      -0.00000001007,    -0.00000001904,
       0.00000000070,     0.00000000159,
       0.00000000118,    -0.00000000065,
      -0.00000000145,     0.00000000013,
       0.00000000025,     0.00000000115,
      -0.00000000018,    -0.00000000133,
      -0.00000000104,     0.00000000027,
       0.00000000112,    -0.00000000047,
      -0.00000000044,    -0.00000000083,
       0.00000000050,     0.00000000112,
       0.00000000083,    -0.00000000046,
       0.00000000717,    -0.00000000262,
       0.00000000015,     0.00000000085,
      -0.00000005374,     0.00000000830,
       0.00000001171,     0.00000004222,
       0.00000001663,    -0.00000000692,
      -0.00000000061,    -0.00000000370,
      -0.00000000291,     0.00000000083,
       0.00000000048,     0.00000000114,
       0.00000003936,    -0.00000002058,
      -0.00000001840,    -0.00000002892,
       0.00000000139,     0.00000000254,
       0.00000000186,    -0.00000000122,
      -0.00000000054,    -0.00000000068,
      -0.00000000246,     0.00000000033,
       0.00000000052,     0.00000000194,
       0.00000000076,    -0.00000000032,
      -0.00000000041,    -0.00000000233,
      -0.00000000180,     0.00000000056,
       0.00000000032,     0.00000000070,
       0.00000000177,    -0.00000000092,
      -0.00000000082,    -0.00000000130,
       0.00000000097,     0.00000000179,
       0.00000000131,    -0.00000000086,
      -0.00000000112,    -0.00000000013,
       0.00000000009,     0.00000000099,
       0.00000000932,    -0.00000000184,
      -0.00000000239,    -0.00000000725,
       0.00000000032,     0.00000000146,
       0.00000000111,    -0.00000000042,
      -0.00000008712,     0.00000001739,
       0.00000002245,     0.00000006773,
       0.00000002636,    -0.00000001269,
      -0.00000000131,    -0.00000000632,
      -0.00000000491,     0.00000000166,
       0.00000000092,     0.00000000190,
      -0.00000000102,    -0.00000000061,
      -0.00000000032,     0.00000000088,
       0.00000000046,     0.00000000061,
      -0.00000001082,     0.00000000817,
      -0.00000000162,    -0.00000000221,
      -0.00000000063,     0.00000000052,
       0.00000005890,    -0.00000003799,
      -0.00000003244,    -0.00000004239,
      -0.00000001520,     0.00000001412,
       0.00000000270,     0.00000000396,
       0.00000000283,    -0.00000000226,
      -0.00000000097,    -0.00000000101,
      -0.00000000408,     0.00000000075,
       0.00000000102,     0.00000000320,
       0.00000000123,    -0.00000000060,
      -0.00000000085,    -0.00000000396,
      -0.00000000303,     0.00000000108,
       0.00000000060,     0.00000000116,
       0.00000000275,    -0.00000000178,
      -0.00000000151,    -0.00000000198,
      -0.00000000069,     0.00000000064,
       0.00000000184,     0.00000000276,
       0.00000000199,    -0.00000000157,
      -0.00000000066,    -0.00000000072,
      -0.00000000127,    -0.00000000016,
       0.00000000011,     0.00000000112,
       0.00000001439,    -0.00000000339,
      -0.00000000418,    -0.00000001109,
       0.00000000063,     0.00000000246,
       0.00000000187,    -0.00000000076,
      -0.00000000042,    -0.00000000069,
      -0.00000013089,     0.00000002802,
       0.00000003474,     0.00000010177,
       0.00000003971,    -0.00000001942,
      -0.00000000270,    -0.00000001065,
      -0.00000000816,     0.00000000326,
       0.00000000175,     0.00000000310,
       0.00000000078,    -0.00000000058,
      -0.00000000131,    -0.00000000079,
      -0.00000000042,     0.00000000112,
      -0.00000000721,     0.00000000622,
       0.00000000501,     0.00000000499,
      -0.00000000115,    -0.00000000136,
      -0.00000000093,     0.00000000094,
       0.00000008461,    -0.00000006782,
      -0.00000005538,    -0.00000005932,
      -0.00000002071,     0.00000002327,
       0.00000000514,     0.00000000607,
       0.00000000424,    -0.00000000414,
      -0.00000000172,    -0.00000000145,
      -0.00000000656,     0.00000000143,
       0.00000000182,     0.00000000509,
       0.00000000195,    -0.00000000106,
      -0.00000000154,    -0.00000000642,
      -0.00000000492,     0.00000000189,
       0.00000000105,     0.00000000187,
       0.00000000098,    -0.00000000003,
      -0.00000000117,     0.00000000088,
      -0.00000000090,    -0.00000000116,
       0.00000000124,     0.00000000041,
      -0.00000000016,    -0.00000000088,
       0.00000000419,    -0.00000000336,
      -0.00000000277,    -0.00000000294,
      -0.00000000100,     0.00000000113,
       0.00000000343,     0.00000000417,
       0.00000000294,    -0.00000000281,
      -0.00000000116,    -0.00000000103,
      -0.00000000083,     0.00000000036,
       0.00000000055,     0.00000000074,
       0.00000000036,     0.00000000083,
       0.00000000073,    -0.00000000054,
      -0.00000000104,     0.00000000045,
       0.00000000069,     0.00000000093,
       0.00000000045,     0.00000000104,
       0.00000000093,    -0.00000000068,
      -0.00000000116,    -0.00000000014,
       0.00000000010,     0.00000000101,
      -0.00000000156,     0.00000000051,
       0.00000000064,     0.00000000122,
      -0.00000000064,     0.00000000041,
       0.00000000046,     0.00000000075,
       0.00000001798,    -0.00000001077,
      -0.00000001107,    -0.00000001251,
      -0.00000000415,     0.00000000557,
       0.00000000114,     0.00000000400,
       0.00000000304,    -0.00000000132,
      -0.00000000071,    -0.00000000112,
       0.00000000452,     0.00000000084,
      -0.00000000013,    -0.00000000358,
      -0.00000000129,     0.00000000039,
       0.00000000075,    -0.00000000456,
      -0.00000000362,     0.00000000019,
       0.00000000041,     0.00000000136,
      -0.00000019560,     0.00000005329,
       0.00000006243,     0.00000014917,
       0.00000005684,    -0.00000003349,
      -0.00000001124,    -0.00000001395,
      -0.00000000453,    -0.00000001696,
      -0.00000001299,     0.00000000540,
       0.00000000287,     0.00000000489,
       0.00000000121,    -0.00000000098,
       0.00000000176,     0.00000000132,
       0.00000000078,    -0.00000000169,
      -0.00000000011,    -0.00000000219,
      -0.00000000175,     0.00000000048,
       0.00000000037,     0.00000000069,
       0.00000000164,     0.00000000176,
       0.00000000094,    -0.00000000160,
       0.00000000175,    -0.00000000164,
      -0.00000000159,    -0.00000000094,
       0.00000007341,     0.00000007522,
       0.00000003985,    -0.00000007073,
      -0.00000002912,    -0.00000000720,
       0.00000000222,    -0.00000000207,
      -0.00000000201,    -0.00000000119,
      -0.00000000020,     0.00000000083,
      -0.00000000949,     0.00000000966,
       0.00000000775,     0.00000000653,
      -0.00000000220,    -0.00000000210,
      -0.00000000139,     0.00000000173,
       0.00000000067,     0.00000000045,
       0.00000000128,    -0.00000000038,
      -0.00000000042,    -0.00000000096,
       0.00000000024,     0.00000000103,
       0.00000000080,    -0.00000000029,
       0.00000016371,    -0.00000007357,
      -0.00000006942,    -0.00000012377,
      -0.00000004485,     0.00000003318,
       0.00000001035,     0.00000000999,
       0.00000001179,     0.00000000765,
       0.00000000465,    -0.00000000869,
      -0.00000000327,    -0.00000000139,
      -0.00000000038,     0.00000000084,
      -0.00000001044,     0.00000000276,
       0.00000000331,     0.00000000798,
       0.00000000302,    -0.00000000179,
      -0.00000000065,    -0.00000000074,
      -0.00000000278,    -0.00000001006,
      -0.00000000767,     0.00000000325,
       0.00000000173,     0.00000000291,
       0.00000000071,    -0.00000000054,
      -0.00000000061,    -0.00000000050,
      -0.00000000073,     0.00000000074,
       0.00000000054,     0.00000000049,
      -0.00000000072,    -0.00000000072,
      -0.00000000052,     0.00000000053,
       0.00000000231,     0.00000000144,
       0.00000000018,    -0.00000000170,
       0.00000000722,    -0.00000000567,
      -0.00000000468,    -0.00000000508,
      -0.00000000164,     0.00000000198,
       0.00000000574,     0.00000000706,
       0.00000000504,    -0.00000000470,
      -0.00000000199,    -0.00000000167,
       0.00000000102,    -0.00000000006,
      -0.00000000110,    -0.00000000021,
       0.00000000064,    -0.00000000137,
       0.00000000129,     0.00000000074,
       0.00000000180,     0.00000000037,
      -0.00000000174,     0.00000000016,
      -0.00000000545,    -0.00000000743,
      -0.00000001312,     0.00000000963,
       0.00000001935,     0.00000002555,
       0.00000000786,    -0.00000000596,
      -0.00000004323,     0.00000001885,
       0.00000002851,     0.00000003837,
       0.00000001615,    -0.00000001869,
       0.00000000058,     0.00000000134,
       0.00000000119,    -0.00000000088,
      -0.00000000058,    -0.00000000050,
      -0.00000003060,     0.00000000941,
       0.00000001064,     0.00000002325,
      -0.00000000046,    -0.00000000148,
      -0.00000000113,     0.00000000051,
      -0.00000000092,     0.00000000028,
       0.00000000031,     0.00000000069,
      -0.00000000028,    -0.00000000090,
      -0.00000000002,    -0.00000000016,
      -0.00000000016,     0.00000000002,
       0.00000000067,     0.00000000013,
       0.00000000013,    -0.00000000067,
      -0.00000000176,    -0.00000000430,
      -0.00000000348,     0.00000000230,
       0.00000000135,     0.00000000141,
      -0.00000000445,     0.00000000150,
       0.00000000194,     0.00000000369,
       0.00000000153,    -0.00000000112,
       0.00000000590,    -0.00000001315,
      -0.00000001429,    -0.00000000380,
       0.00000000352,     0.00000000526,
       0.00000000409,    -0.00000000386,
      -0.00000000241,    -0.00000000148,
       0.00000000553,    -0.00000000300,
      -0.00000000327,    -0.00000000448,
      -0.00000000182,     0.00000000188,
      -0.00000000282,    -0.00000000748,
      -0.00000000619,     0.00000000376,
       0.00000000223,     0.00000000252,
       0.00000000063,    -0.00000000081,
      -0.00000000748,     0.00000000281,
       0.00000000374,     0.00000000619,
       0.00000000248,    -0.00000000221,
      -0.00000000080,    -0.00000000063,
      -0.00000000353,    -0.00000000938,
      -0.00000000776,     0.00000000470,
       0.00000000278,     0.00000000313,
       0.00000000080,    -0.00000000102,
      -0.00000000937,     0.00000000353,
       0.00000000469,     0.00000000775,
       0.00000000313,    -0.00000000278,
      -0.00000000102,    -0.00000000080,
       0.00000000093,    -0.00000000031,
      -0.00000000031,    -0.00000000093,
       0.00000000059,    -0.00000000019,
      -0.00000000019,    -0.00000000058,
      -0.00000000020,    -0.00000000062,
      -0.00000008275,    -0.00000021401,
      -0.00000017676,     0.00000010917,
       0.00000006414,     0.00000007107,
       0.00000001809,    -0.00000002353,
      -0.00000000889,     0.00000000303,
       0.00000000419,     0.00000000737,
       0.00000000299,    -0.00000000252,
      -0.00000000091,    -0.00000000072,
       0.00000000056,     0.00000000088,
       0.00000000060,    -0.00000000056,
       0.00000002513,    -0.00000000761,
      -0.00000000866,    -0.00000001933,
      -0.00000000758,     0.00000000481,
       0.00000000196,     0.00000000624,
       0.00000000472,    -0.00000000220,
      -0.00000000114,    -0.00000000172,
      -0.00000024118,     0.00000008694,
       0.00000009470,     0.00000018072,
       0.00000006676,    -0.00000004850,
      -0.00000001557,    -0.00000001587,
      -0.00000000781,    -0.00000002579,
      -0.00000001963,     0.00000000890,
       0.00000000466,     0.00000000736,
       0.00000000178,    -0.00000000158,
      -0.00000000114,    -0.00000000443,
      -0.00000000352,     0.00000000163,
       0.00000000095,     0.00000000138,
       0.00000000232,    -0.00000000223,
      -0.00000000235,    -0.00000000202,
      -0.00000000081,     0.00000000127,
      -0.00000000019,     0.00000000104,
       0.00000000073,    -0.00000000069,
      -0.00000000089,    -0.00000000070,
       0.00000002405,     0.00000002140,
       0.00000001048,    -0.00000002110,
      -0.00000000867,    -0.00000000178,
      -0.00000000594,    -0.00000000587,
      -0.00000000370,     0.00000000463,
       0.00000000133,     0.00000000056,
      -0.00000002055,     0.00000001886,
       0.00000001534,     0.00000001025,
       0.00000000250,    -0.00000000589,
       0.00000002078,     0.00000002374,
       0.00000001262,    -0.00000001717,
      -0.00000000610,    -0.00000000244,
       0.00000015570,    -0.00000022459,
      -0.00000016755,    -0.00000009741,
      -0.00000002980,     0.00000006342,
       0.00000001611,     0.00000000597,
       0.00000002328,     0.00000001592,
       0.00000000982,    -0.00000001729,
      -0.00000000653,    -0.00000000290,
      -0.00000000057,     0.00000000162,
      -0.00000002861,     0.00000001609,
       0.00000001486,     0.00000001854,
       0.00000000583,    -0.00000000625,
      -0.00000000165,    -0.00000000120,
       0.00000000633,    -0.00000000224,
      -0.00000000480,    -0.00000000383,
      -0.00000000052,     0.00000000284,
       0.00000000093,    -0.00000000027,
      -0.00000001578,     0.00000001359,
       0.00000001133,     0.00000000817,
       0.00000000173,    -0.00000000407,
       0.00000001358,     0.00000001577,
       0.00000000816,    -0.00000001131,
      -0.00000000405,    -0.00000000173,
      -0.00000074478,     0.00000063475,
       0.00000052399,     0.00000038539,
       0.00000008174,    -0.00000018590,
      -0.00000003125,    -0.00000000589,
       0.00000001644,     0.00000001929,
       0.00000000997,    -0.00000001358,
      -0.00000000483,    -0.00000000212,
      -0.00000000016,     0.00000000084,
       0.00000000182,     0.00000000185,
       0.00000000101,    -0.00000000145,
      -0.00000000127,     0.00000000179,
       0.00000000127,     0.00000000077,
      -0.00000000162,    -0.00000000119,
      -0.00000000080,     0.00000000120,
      -0.00000062277,     0.00000053874,
       0.00000044352,     0.00000033005,
       0.00000007483,    -0.00000016059,
      -0.00000003005,    -0.00000000773,
       0.00000000817,     0.00000000777,
       0.00000000324,    -0.00000000631,
      -0.00000000186,    -0.00000000025,
       0.00000001031,    -0.00000001452,
      -0.00000001080,    -0.00000000641,
      -0.00000000189,     0.00000000416,
       0.00000000096,     0.00000000037,
       0.00000001419,     0.00000000936,
       0.00000000607,    -0.00000001052,
      -0.00000000402,    -0.00000000195,
      -0.00000000039,     0.00000000095,
       0.00000000061,    -0.00000000195,
       0.00000010300,    -0.00000004538,
      -0.00000004131,    -0.00000005025,
      -0.00000001181,     0.00000001700,
       0.00000000026,     0.00000000960,
       0.00000000479,    -0.00000000118,
      -0.00000000096,    -0.00000000147,
      -0.00000001801,     0.00000001481,
       0.00000001255,     0.00000000988,
       0.00000000246,    -0.00000000453,
      -0.00000001517,    -0.00000001759,
      -0.00000000999,     0.00000001243,
       0.00000000452,     0.00000000250,
       0.00000002160,    -0.00000002010,
       0.00000000880,    -0.00000000530,
       0.00000000504,    -0.00000002608,
      -0.00000000370,    -0.00000000590,
      -0.00000011076,    -0.00000002142,
      -0.00000002560,     0.00000001590,
       0.00000009826,     0.00000010953,
       0.00000001291,    -0.00000006297,
       0.00000007242,    -0.00000006417,
      -0.00000004252,    -0.00000000558,
      -0.00000001981,    -0.00000000149,
       0.00000000747,     0.00000001082,
      -0.00000000126,    -0.00000001761,
      -0.00000001055,     0.00000000726,
       0.00000000166,     0.00000000011,
      -0.00000000382,    -0.00000001768,
       0.00000001847,     0.00000002330,
      -0.00000000069,     0.00000000054,
      -0.00000014659,    -0.00000038857,
      -0.00000032128,     0.00000019512,
       0.00000011529,     0.00000012945,
       0.00000003316,    -0.00000004244,
      -0.00000001191,     0.00000000449,
       0.00000000598,     0.00000000984,
       0.00000000397,    -0.00000000353,
      -0.00000000128,    -0.00000000100,
       0.00000000694,    -0.00000002515,
       0.00000000127,     0.00000000037,
       0.00000000015,    -0.00000000087,
      -0.00000000201,     0.00000000059,
       0.00000000067,     0.00000000152,
      -0.00000000061,    -0.00000000198,
      -0.00000000149,     0.00000000068,
       0.00000000023,    -0.00000000079,
       0.00000000078,     0.00000000022,
      -0.00000003042,    -0.00000002365,
      -0.00000001809,     0.00000003064,
       0.00000001540,     0.00000000664,
       0.00000000140,    -0.00000000497,
       0.00000001311,    -0.00000001002,
      -0.00000001111,    -0.00000001039,
      -0.00000000437,     0.00000000606,
       0.00000000116,     0.00000000075,
       0.00000006806,    -0.00000002151,
      -0.00000002430,    -0.00000005197,
      -0.00000001977,     0.00000001258,
       0.00000000477,     0.00000000533,
      -0.00000002113,    -0.00000006793,
      -0.00000005170,     0.00000002405,
       0.00000001243,     0.00000001963,
       0.00000000532,    -0.00000000444,
       0.00000000301,    -0.00000000077,
      -0.00000000065,    -0.00000000148,
      -0.00000000088,    -0.00000000296,
      -0.00000000143,     0.00000000075,
       0.00000031762,     0.00000027709,
       0.00000021396,    -0.00000032366,
      -0.00000016422,    -0.00000007951,
      -0.00000001753,     0.00000005482,
       0.00000001381,     0.00000000203,
      -0.00000002451,     0.00000002208,
       0.00000002385,     0.00000001962,
       0.00000000731,    -0.00000001276,
      -0.00000000465,    -0.00000000169,
      -0.00000019318,     0.00000006590,
       0.00000007102,     0.00000014626,
       0.00000005541,    -0.00000003633,
      -0.00000001183,    -0.00000001348,
      -0.00000000226,     0.00000000268,
       0.00000006527,     0.00000019216,
       0.00000014540,    -0.00000007026,
      -0.00000003576,    -0.00000005477,
      -0.00000001330,     0.00000001168,
       0.00000000275,     0.00000000183,
      -0.00000000721,     0.00000000179,
       0.00000000163,     0.00000000366,
       0.00000000195,     0.00000000712,
       0.00000000359,    -0.00000000171,
      -0.00000000020,    -0.00000000393,
      -0.00000000316,     0.00000000057,
       0.00000000040,     0.00000000129,
       0.00000005456,    -0.00000001764,
      -0.00000001967,    -0.00000004135,
      -0.00000001560,     0.00000001019,
       0.00000000332,     0.00000000380,
       0.00000000067,    -0.00000000080,
      -0.00000001762,    -0.00000005451,
      -0.00000004133,     0.00000001960,
       0.00000001015,     0.00000001559,
       0.00000000379,    -0.00000000328,
      -0.00000000076,    -0.00000000066,
       0.00000000095,     0.00000000036,
       0.00000000019,    -0.00000000082,
       0.00000006787,    -0.00000002189,
      -0.00000002437,    -0.00000005147,
      -0.00000001939,     0.00000001260,
       0.00000000415,     0.00000000475,
       0.00000000083,    -0.00000000098,
      -0.00000002187,    -0.00000006781,
      -0.00000005141,     0.00000002434,
       0.00000001257,     0.00000001936,
       0.00000000475,    -0.00000000413,
      -0.00000000097,    -0.00000000083,
       0.00000001087,    -0.00000000359,
       0.00000000033,     0.00000000098,
       0.00000001158,    -0.00000000370,
       0.00000000035,     0.00000000103,
      -0.00000000007,     0.00000000110,
      -0.00000000149,     0.00000000023,
       0.00000000002,     0.00000000143,
      -0.00000000878,    -0.00000000437,
      -0.00000000019,    -0.00000000133,
       0.00000000321,    -0.00000000122,
      -0.00000000220,    -0.00000000360,
       0.00000000056,     0.00000000118,
       0.00000000088,    -0.00000000055,
       0.00000181357,    -0.00000058138,
      -0.00000064878,    -0.00000137523,
      -0.00000051797,     0.00000033579,
       0.00000011110,     0.00000012694,
       0.00000002230,    -0.00000002670,
      -0.00000002138,    -0.00000006516,
      -0.00000004940,     0.00000002364,
       0.00000001210,     0.00000001854,
       0.00000000448,    -0.00000000397,
      -0.00000000093,    -0.00000000075,
       0.00000000152,    -0.00000000144,
      -0.00000000123,    -0.00000000128,
      -0.00000000057,     0.00000000053,
      -0.00000000028,     0.00000000191,
       0.00000000157,    -0.00000000012,
      -0.00000022460,     0.00000016240,
       0.00000051280,    -0.00000047910,
      -0.00000006120,     0.00000005170,
      -0.00000000258,    -0.00000000043,
       0.00000000006,     0.00000000165,
      -0.00000000099,    -0.00000000076,
      -0.00000000090,     0.00000000069,
      -0.00000000699,     0.00000001459,
       0.00000001027,     0.00000000385,
      -0.00000001421,    -0.00000000519,
      -0.00000000250,     0.00000000869,
       0.00000000433,     0.00000000041,
       0.00000000672,     0.00000000265,
       0.00000000038,    -0.00000000438,
      -0.00000000068,     0.00000000051,
       0.00000000462,    -0.00000000041,
       0.00000000062,    -0.00000000265,
       0.00000007383,    -0.00000049872,
      -0.00000032798,    -0.00000000438,
       0.00000001465,     0.00000010698,
       0.00000002119,    -0.00000000631,
       0.00000008506,     0.00000002869,
       0.00000001300,    -0.00000005889,
      -0.00000002060,    -0.00000000213,
       0.00000000026,     0.00000000455,
       0.00000000089,    -0.00000000028,
       0.00000003130,    -0.00000001414,
      -0.00000001348,    -0.00000002255,
      -0.00000000822,     0.00000000633,
       0.00000000154,     0.00000000186,
       0.00000000060,    -0.00000000072,
       0.00000000367,     0.00000002213,
       0.00000001795,    -0.00000000596,
      -0.00000000367,    -0.00000000712,
      -0.00000000190,     0.00000000183,
       0.00000000072,     0.00000000060,
       0.00000001806,    -0.00000001957,
      -0.00000000056,    -0.00000000053,
      -0.00000027206,     0.00000020700,
       0.00000017755,     0.00000014307,
       0.00000003043,    -0.00000006332,
      -0.00001312732,    -0.00001546101,
      -0.00000775933,     0.00001037752,
       0.00000304065,     0.00000146140,
       0.00000000339,    -0.00000049211,
      -0.00000027532,     0.00000006567,
       0.00000003117,     0.00000006398,
      -0.00000048364,     0.00000001467,
       0.00000009111,     0.00000024124,
       0.00000005466,    -0.00000003490,
       0.00000320844,    -0.00000242123,
      -0.00000202803,    -0.00000170054,
      -0.00000038317,     0.00000074146,
       0.00000015439,     0.00000003265,
       0.00003373410,     0.00003992466,
       0.00002045835,    -0.00002696040,
      -0.00000924443,    -0.00000431448,
      -0.00000034105,     0.00000161815,
       0.00000000119,     0.00000199850,
       0.00000100648,    -0.00000025659,
      -0.00000009536,    -0.00000018688,
       0.00000180708,    -0.00000001558,
      -0.00000025786,    -0.00000090364,
      -0.00000018898,     0.00000009016,
      -0.00000000350,     0.00000002000,
       0.00000002000,     0.00000000230,
      -0.00000340099,     0.00000337574,
       0.00000281062,     0.00000179112,
       0.00000040150,    -0.00000105974,
      -0.00000022979,    -0.00000006408,
      -0.00000850475,    -0.00000771451,
      -0.00000397209,     0.00000682265,
       0.00000241271,     0.00000088705,
       0.00000012339,    -0.00000043551,
       0.00000024033,    -0.00000076757,
      -0.00000052109,    -0.00000009669,
      -0.00000000865,     0.00000016877,
       0.00000002924,     0.00000000307,
       0.00000076059,     0.00000021853,
       0.00000010045,    -0.00000052266,
      -0.00000017572,    -0.00000001926,
      -0.00000000397,     0.00000003221,
      -0.00000000055,     0.00000000063,
       0.00000000069,    -0.00000000075,
       0.00000000075,     0.00000000069,
      -0.00000000016,     0.00000000207,
       0.00000001237,     0.00000000175,
       0.00000000078,    -0.00000000477,
      -0.00000000219,    -0.00000000038,
      -0.00000000084,     0.00000000068,
       0.00000000063,     0.00000000045,
      -0.00000000055,    -0.00000000069,
      -0.00000000057,     0.00000000055,
      -0.00000006593,    -0.00000000789,
       0.00000000204,     0.00000003230,
       0.00000000136,     0.00000000117,
       0.00000000044,    -0.00000000089,
      -0.00000000048,    -0.00000000060,
       0.00000000053,    -0.00000000051,
       0.00000000083,     0.00000000078,
      -0.00000001749,     0.00000001872,
      -0.00000000082,    -0.00000000074,
      -0.00000000038,     0.00000000063,
      -0.00000000141,    -0.00000000047,
       0.00000000025,    -0.00000000129,
      -0.00000000075,     0.00000000720,
      -0.00000017881,    -0.00000001852,
       0.00000000648,    -0.00000007758,
      -0.00000002853,     0.00000000508,
       0.00000044652,     0.00000003069,
      -0.00000002761,    -0.00000015491,
       0.00000003520,     0.00000002567,
       0.00000002514,    -0.00000003578,
      -0.00000000359,    -0.00000000164,
       0.00000000178,     0.00000000116,
       0.00000000034,    -0.00000000071,
      -0.00000000134,    -0.00000000078,
      -0.00000000036,     0.00000000097,
       0.00000000527,    -0.00000002641,
       0.00000011553,     0.00000002305,
       0.00000001114,     0.00000000755,
       0.00000000905,    -0.00000001091,
       0.00000000085,     0.00000000080,
       0.00000315679,    -0.00000101836,
      -0.00000113358,    -0.00000239377,
      -0.00000090156,     0.00000058561,
       0.00000019283,     0.00000022095,
       0.00000003906,    -0.00000004611,
      -0.00000002688,    -0.00000008319,
      -0.00000006308,     0.00000002990,
       0.00000001544,     0.00000002375,
       0.00000000582,    -0.00000000507,
      -0.00000000119,    -0.00000000101,
      -0.00000000042,     0.00000000086,
      -0.00000000053,     0.00000000076,
       0.00000000082,     0.00000000056,
       0.00000000070,     0.00000000065,
       0.00000000145,    -0.00000000285,
      -0.00000000194,    -0.00000000080,
       0.00000000284,     0.00000000140,
       0.00000000079,    -0.00000000194,
      -0.00000009560,     0.00000004630,
       0.00000000480,     0.00000000980,
       0.00000025040,    -0.00000011220,
      -0.00000001330,    -0.00000001390,
      -0.00000002570,    -0.00000005580,
      -0.00000007660,     0.00000003340,
       0.00000000630,     0.00000000820,
      -0.00000000096,    -0.00000000164,
       0.00000000013,    -0.00000000009,
       0.00000000000,    -0.00000000049,
      -0.00000000029,    -0.00000000003,
      -0.00000000001,     0.00000000013,
       0.00000000007,    -0.00000000003,
      -0.00000000290,    -0.00000000313,
       0.00000000032,    -0.00000000028,
       0.00000000040,    -0.00000000152,
      -0.00000000156,    -0.00000000045,
      -0.00000000484,    -0.00000000607,
       0.00000000060,    -0.00000000047,
       0.00000000068,    -0.00000000308,
      -0.00000000311,    -0.00000000086,
      -0.00000000649,    -0.00000000954,
       0.00000000087,    -0.00000000067,
       0.00000000273,    -0.00000000911,
       0.00000000077,     0.00000000016,
      -0.00000000900,    -0.00000000188,
       0.00000000006,    -0.00000000077,
      -0.00000000017,     0.00000000069,
       0.00000000063,     0.00000000022,
       0.00000454749,    -0.00000152967,
      -0.00000167170,    -0.00000344378,
      -0.00000129722,     0.00000085426,
       0.00000027987,     0.00000031819,
       0.00000005638,    -0.00000006768,
      -0.00000001307,    -0.00000000677,
       0.00001684178,     0.00005026734,
       0.00003808934,    -0.00001850480,
      -0.00000948424,    -0.00001434597,
      -0.00000351764,     0.00000312060,
       0.00000077468,     0.00000062982,
       0.00000007272,    -0.00000014026,
      -0.00000190717,     0.00000685930,
       0.00000472778,     0.00000067638,
       0.00000001505,    -0.00000163130,
      -0.00000036811,     0.00000004981,
       0.00000001768,     0.00000006005,
       0.00000682129,     0.00000188708,
       0.00000066097,    -0.00000469210,
      -0.00000161790,     0.00000000146,
       0.00000005510,     0.00000036380,
       0.00000005964,    -0.00000001804,
      -0.00000004659,     0.00000014825,
       0.00000005896,     0.00000001172,
       0.00000000109,    -0.00000001147,
       0.00000015187,     0.00000002996,
      -0.00000000183,    -0.00000006160,
      -0.00000001150,     0.00000000282,
      -0.00005129167,     0.00001723792,
       0.00001885279,     0.00003884312,
       0.00001462809,    -0.00000963175,
      -0.00000314634,    -0.00000359256,
      -0.00000063684,     0.00000074641,
       0.00000013268,     0.00000007745,
      -0.00004152657,    -0.00012389593,
      -0.00009386556,     0.00004564111,
       0.00002337015,     0.00003532219,
       0.00000865192,    -0.00000763445,
      -0.00000179356,    -0.00000151523,
      -0.00000017276,     0.00000029254,
       0.00000730184,    -0.00002808266,
      -0.00001923908,    -0.00000250846,
       0.00000001011,     0.00000659964,
       0.00000148394,    -0.00000020922,
      -0.00000007202,    -0.00000023644,
      -0.00000002194,     0.00000001066,
      -0.00002802325,    -0.00000718080,
      -0.00000241031,     0.00001918348,
       0.00000657265,    -0.00000005708,
      -0.00000023178,    -0.00000147652,
      -0.00000022588,     0.00000007569,
       0.00000000950,     0.00000002160,
       0.00000010828,    -0.00000035132,
      -0.00000013923,    -0.00000002064,
       0.00000000078,     0.00000002566,
      -0.00000035976,    -0.00000008629,
      -0.00000000727,     0.00000014346,
       0.00000002573,    -0.00000000260,
       0.00000000136,     0.00000001306,
       0.00000001378,    -0.00000000158,
      -0.00000000378,    -0.00000000544,
       0.00000000238,     0.00000001559,
       0.00000001068,    -0.00000000553,
      -0.00000000315,    -0.00000000686,
       0.00000001603,    -0.00000000296,
      -0.00000000433,    -0.00000001060,
      -0.00000000589,     0.00000000204,
       0.00000000014,     0.00000000248,
       0.00000000195,    -0.00000000043,
      -0.00000000040,     0.00000000104,
       0.00000000208,     0.00000000049,
      -0.00000000393,    -0.00000000322,
      -0.00000000287,     0.00000000445,
       0.00000000436,     0.00000000309,
       0.00000000255,     0.00000002882,
       0.00000000903,     0.00000000850,
       0.00000000126,    -0.00000000016,
      -0.00000000274,     0.00000000462,
       0.00000000461,     0.00000000292,
       0.00000000288,    -0.00000000326,
      -0.00000000158,    -0.00000000102,
      -0.00000000080,     0.00000000034,
      -0.00000002090,    -0.00000002083,
      -0.00000000995,     0.00000001211,
      -0.00000000795,     0.00000000727,
       0.00000000501,     0.00000000356,
       0.00000000353,    -0.00000000258,
      -0.00000000074,    -0.00000000142,
      -0.00000000030,    -0.00000000294,
      -0.00000000140,     0.00000000042,
       0.00000050208,    -0.00000044345,
      -0.00000036333,    -0.00000025931,
      -0.00000005532,     0.00000012785,
       0.00000002229,     0.00000000502,
       0.00000001686,     0.00000001931,
       0.00000001008,    -0.00000001396,
      -0.00000000497,    -0.00000000224,
      -0.00000000022,     0.00000000092,
      -0.00000000126,     0.00000000597,
       0.00000000393,     0.00000000032,
      -0.00000000018,    -0.00000000189,
      -0.00000000659,    -0.00000000189,
      -0.00000000080,     0.00000000450,
       0.00000000225,     0.00000000002,
      -0.00000000023,     0.00000002521,
       0.00000001319,    -0.00000000355,
      -0.00000000195,    -0.00000000329,
       0.00000002799,    -0.00000000122,
      -0.00000000462,    -0.00000001445,
      -0.00000000333,     0.00000000213,
      -0.00000001843,    -0.00000001274,
      -0.00000000202,     0.00000000912,
       0.00000000196,    -0.00000000037,
      -0.00000001532,     0.00000001532,
       0.00000000740,     0.00000000429,
       0.00000000065,    -0.00000000139,
       0.00000000291,    -0.00000013486,
      -0.00000007186,     0.00000001795,
       0.00000001062,     0.00000002010,
      -0.00000006720,     0.00000000107,
       0.00000000986,     0.00000003519,
       0.00000000889,    -0.00000000483,
      -0.00000000139,    -0.00000000134,
       0.00000009457,     0.00000003609,
      -0.00000000496,    -0.00000004688,
      -0.00000001056,     0.00000000596,
       0.00000000126,     0.00000000097,
       0.00000007955,    -0.00000004578,
      -0.00000002178,    -0.00000003104,
      -0.00000000707,     0.00000000507,
      -0.00000002089,     0.00000000094,
       0.00000000105,    -0.00000000056,
       0.00000001303,    -0.00000001119,
      -0.00000000940,    -0.00000000665,
      -0.00000000137,     0.00000000344,
       0.00000001131,     0.00000001297,
       0.00000000676,    -0.00000000942,
      -0.00000000345,    -0.00000000137,
       0.00000001484,    -0.00000001327,
      -0.00000001095,    -0.00000000763,
      -0.00000000161,     0.00000000390,
       0.00000001325,     0.00000001480,
       0.00000000762,    -0.00000001094,
      -0.00000000392,    -0.00000000160,
       0.00000070311,    -0.00000062238,
      -0.00000050938,    -0.00000036169,
      -0.00000007581,     0.00000017851,
       0.00000002938,     0.00000000538,
       0.00000001601,     0.00000001808,
       0.00000000932,    -0.00000001309,
      -0.00000000460,    -0.00000000196,
      -0.00000000027,     0.00000000119,
       0.00000000123,     0.00000000013,
       0.00000000079,    -0.00000000027,
       0.00000000054,    -0.00000000075,
      -0.00000000108,     0.00000000030,
       0.00000000044,     0.00000000112,
      -0.00000000773,    -0.00000000630,
      -0.00000000184,     0.00000000247,
       0.00000000473,    -0.00000000165,
      -0.00000000144,    -0.00000000476,
      -0.00000000092,     0.00000000116,
       0.00000000184,    -0.00000000039,
      -0.00000000048,    -0.00000000203,
      -0.00000005323,     0.00000022378,
       0.00000015307,     0.00000001845,
      -0.00000000071,    -0.00000005136,
      -0.00000000218,    -0.00000000051,
      -0.00000000002,     0.00000000139,
       0.00000000127,     0.00000000095,
       0.00000000057,    -0.00000000138,
      -0.00000000123,    -0.00000000073,
      -0.00000000044,    -0.00000001000,
      -0.00000000357,    -0.00000000352,
      -0.00000000113,     0.00000000126,
       0.00000000092,    -0.00000000184,
      -0.00000000120,    -0.00000000059,
      -0.00000000194,    -0.00000000085,
      -0.00000000005,     0.00000000153,
      -0.00000000010,    -0.00000000001,
      -0.00000000065,    -0.00000000015,
       0.00000054154,    -0.00000196970,
      -0.00000135206,    -0.00000019144,
      -0.00000000349,     0.00000046396,
       0.00000010487,    -0.00000001426,
      -0.00000000526,    -0.00000001730,
      -0.00000209957,    -0.00000044227,
      -0.00000012906,     0.00000142723,
       0.00000048710,    -0.00000001691,
      -0.00000001890,    -0.00000010928,
      -0.00000001783,     0.00000000566,
       0.00000129647,     0.00000147034,
       0.00000069260,    -0.00000079045,
      -0.00000024067,    -0.00000015780,
      -0.00000002153,     0.00000004855,
       0.00000000666,     0.00000000214,
       0.00000145111,    -0.00000128847,
      -0.00000078230,    -0.00000068013,
      -0.00000015371,     0.00000023683,
       0.00000004732,     0.00000002114,
       0.00000000153,    -0.00000000701,
       0.00000000870,     0.00000001045,
       0.00000000244,    -0.00000000142,
       0.00000000695,    -0.00000001349,
      -0.00000000370,    -0.00000000033,
      -0.00000448822,     0.00001705156,
       0.00001166497,     0.00000154711,
       0.00000000201,    -0.00000399331,
      -0.00000090169,     0.00000012750,
       0.00000004507,     0.00000014970,
       0.00000001962,    -0.00000000966,
       0.00000514479,     0.00000108531,
       0.00000031845,    -0.00000349707,
      -0.00000119228,     0.00000004003,
       0.00000004589,     0.00000026818,
       0.00000004453,    -0.00000001435,
      -0.00000000303,    -0.00000000604,
      -0.00000547873,    -0.00000596647,
      -0.00000280866,     0.00000331506,
       0.00000100484,     0.00000063739,
       0.00000008762,    -0.00000020086,
      -0.00000002951,    -0.00000000690,
      -0.00000595313,     0.00000540447,
       0.00000325607,     0.00000280908,
       0.00000064212,    -0.00000098182,
      -0.00000019554,    -0.00000008964,
      -0.00000000769,     0.00000002846,
      -0.00000002114,    -0.00000002185,
      -0.00000000337,     0.00000000221,
      -0.00000001795,     0.00000002817,
       0.00000000562,     0.00000000097,
       0.00000000587,    -0.00000001962,
      -0.00000000549,    -0.00000000229,
      -0.00000000075,     0.00000000402,
       0.00000000146,     0.00000000001,
      -0.00000002734,     0.00000001093,
       0.00000001126,     0.00000001967,
       0.00000000686,    -0.00000000552,
      -0.00000000141,    -0.00000000156,
      -0.00000000489,    -0.00000001830,
      -0.00000001489,     0.00000000597,
       0.00000000320,     0.00000000619,
       0.00000000159,    -0.00000000141,
      -0.00000002893,     0.00000000921,
       0.00000001039,     0.00000002179,
       0.00000000789,    -0.00000000554,
      -0.00000000178,    -0.00000000194,
      -0.00000001052,    -0.00000002749,
      -0.00000002092,     0.00000001097,
       0.00000000528,     0.00000000816,
       0.00000000199,    -0.00000000173,
       0.00000000950,    -0.00000000509,
      -0.00000000586,    -0.00000000458,
       0.00000000501,     0.00000000904,
      -0.00000001214,     0.00000000691,
       0.00000000710,     0.00000000498,
      -0.00000000422,     0.00000000014,
       0.00000000013,     0.00000000422,
       0.00000001198,     0.00000002013,
       0.00000001730,    -0.00000001602,
      -0.00000000480,     0.00000000291,
       0.00000000386,     0.00000000409,
       0.00000000137,    -0.00000000220,
       0.00000000011,    -0.00000000501,
      -0.00000000457,     0.00000000099,
       0.00000000100,     0.00000000191,
      -0.00000000532,    -0.00000000002,
       0.00000000088,     0.00000000461,
       0.00000000195,    -0.00000000098,
      -0.00000000189,     0.00000000031,
       0.00000000058,     0.00000000167,
       0.00000000064,    -0.00000000055,
      -0.00000000246,     0.00000000073,
       0.00000000096,     0.00000000193,
       0.00000000094,     0.00000000249,
       0.00000000188,    -0.00000000114,
       0.00000005672,     0.00000003891,
       0.00000002885,    -0.00000005417,
      -0.00000002577,    -0.00000001059,
       0.00000001042,    -0.00000000051,
      -0.00000000257,    -0.00000000901,
      -0.00000000396,     0.00000000221,
       0.00000000097,     0.00000000028,
      -0.00000000949,     0.00000000422,
       0.00000000439,     0.00000000689,
       0.00000000247,    -0.00000000202,
       0.00000000418,     0.00000000939,
       0.00000000698,    -0.00000000430,
      -0.00000000209,    -0.00000000259,
       0.00000000095,     0.00000000068,
      -0.00000009044,     0.00000013872,
      -0.00000001065,    -0.00000000702,
       0.00000000005,     0.00000000108,
      -0.00000077516,     0.00000027752,
       0.00000029770,     0.00000058507,
       0.00000021956,    -0.00000015030,
      -0.00000004845,    -0.00000005328,
      -0.00000000935,     0.00000001111,
      -0.00000000904,    -0.00000002922,
      -0.00000002233,     0.00000001032,
       0.00000000552,     0.00000000842,
       0.00000000207,    -0.00000000181,
       0.00000000019,    -0.00000000220,
      -0.00000000154,     0.00000000048,
      -0.00000000237,     0.00000000040,
       0.00000000086,     0.00000000162,
       0.00000000048,    -0.00000000053,
       0.00000000962,    -0.00000000338,
      -0.00000000058,    -0.00000000245,
      -0.00000000079,     0.00000000109,
      -0.00000006241,     0.00000003923,
       0.00000003564,     0.00000003371,
       0.00000000788,    -0.00000001333,
       0.00000000479,     0.00000000696,
       0.00000000370,    -0.00000000425,
      -0.00000000175,    -0.00000000061,
       0.00000000039,     0.00000000182,
       0.00000000091,    -0.00000000050,
       0.00000000074,    -0.00000000074,
       0.00000000028,     0.00000000116,
      -0.00000000101,     0.00000000189,
       0.00000000189,     0.00000000091,
      -0.00000000087,    -0.00000000291,
      -0.00000000114,     0.00000000066,
       0.00000000415,    -0.00000000813,
      -0.00000000223,    -0.00000000087,
      -0.00000000811,    -0.00000000419,
      -0.00000000062,     0.00000000234,
       0.00000002825,     0.00000007716,
       0.00000003341,    -0.00000003041,
      -0.00000000292,     0.00000000084,
       0.00000000109,     0.00000000127,
      -0.00000004685,    -0.00000003565,
      -0.00000000762,     0.00000002280,
       0.00000000510,     0.00000000014,
      -0.00000002091,     0.00000002999,
       0.00000001414,     0.00000000435,
       0.00000000000,    -0.00000000299,
       0.00000002467,    -0.00000000364,
      -0.00000000427,    -0.00000000522,
      -0.00000000053,     0.00000000104,
      -0.00000000177,    -0.00000002324,
      -0.00000000467,     0.00000000297,
       0.00000031059,     0.00000022654,
       0.00000004816,    -0.00000014889,
      -0.00000003348,    -0.00000000086,
       0.00000004926,    -0.00000006993,
      -0.00000003316,    -0.00000001026,
      -0.00000000008,     0.00000000735,
      -0.00000010315,     0.00000001709,
       0.00000001864,     0.00000002056,
       0.00000000197,    -0.00000000430,
       0.00000000293,     0.00000010290,
       0.00000002274,    -0.00000001111,
      -0.00000000270,    -0.00000000282,
       0.00000000077,     0.00000000234,
       0.00000000107,    -0.00000000093,
      -0.00000000248,     0.00000000085,
       0.00000000092,     0.00000000107,
       0.00000000143,     0.00000000303,
       0.00000000131,    -0.00000000133,
      -0.00000000303,     0.00000000142,
       0.00000000133,     0.00000000131,
       0.00000005961,     0.00000013499,
       0.00000005908,    -0.00000005668,
      -0.00000000385,     0.00000000179,
       0.00000000167,     0.00000000167,
      -0.00000000102,     0.00000001118,
       0.00000000160,     0.00000000023,
      -0.00000000127,    -0.00000000112,
      -0.00000000098,     0.00000000142,
       0.00000000063,     0.00000000173,
       0.00000000038,    -0.00000000147,
      -0.00000000146,    -0.00000000037,
       0.00000002069,    -0.00000000675,
      -0.00000000162,    -0.00000000429,
      -0.00000000166,     0.00000000625,
       0.00000000619,     0.00000000169,
       0.00000000926,    -0.00000000205,
      -0.00000000051,    -0.00000000203,
      -0.00000000103,     0.00000000263,
       0.00000000268,     0.00000000102,
      -0.00000134813,     0.00000047107,
       0.00000051061,     0.00000102092,
       0.00000038321,    -0.00000025879,
      -0.00000008381,    -0.00000009370,
      -0.00000001629,     0.00000001944,
      -0.00000001201,    -0.00000003431,
      -0.00000002598,     0.00000001301,
       0.00000000659,     0.00000000976,
       0.00000000239,    -0.00000000214,
      -0.00000002323,    -0.00000002106,
      -0.00000000097,     0.00000000107,
       0.00000000063,     0.00000000042,
       0.00000002338,    -0.00000004833,
      -0.00000001259,    -0.00000000301,
       0.00000000353,     0.00000000179,
      -0.00000000024,     0.00000000014,
      -0.00000000063,     0.00000000051,
      -0.00000000107,     0.00000000099,
      -0.00000000182,     0.00000000171,
       0.00000336173,     0.00000377358,
       0.00000177227,    -0.00000204443,
      -0.00000061941,    -0.00000040335,
      -0.00000005595,     0.00000012352,
       0.00000001809,     0.00000000464,
       0.00000223388,    -0.00000216990,
      -0.00000129752,    -0.00000105116,
      -0.00000023866,     0.00000038862,
       0.00000007722,     0.00000003298,
       0.00000000277,    -0.00000001078,
      -0.00000195088,     0.00000036798,
       0.00000023702,     0.00000075453,
       0.00000014887,    -0.00000007089,
      -0.00000001324,    -0.00000001894,
      -0.00000000159,     0.00000000173,
       0.00000038170,     0.00000188972,
       0.00000071935,    -0.00000023885,
      -0.00000006962,    -0.00000013902,
      -0.00000001708,     0.00000001279,
       0.00000000161,     0.00000000160,
       0.00000002681,    -0.00000000864,
      -0.00000000421,    -0.00000000476,
       0.00000000212,    -0.00000002597,
      -0.00000000493,    -0.00000000009,
      -0.00002209088,    -0.00002304692,
      -0.00001084539,     0.00001321752,
       0.00000396196,     0.00000246992,
       0.00000034256,    -0.00000078435,
      -0.00000011449,    -0.00000002909,
      -0.00000000102,     0.00000001279,
      -0.00000522091,     0.00000500977,
       0.00000300298,     0.00000245734,
       0.00000055884,    -0.00000090106,
      -0.00000017849,    -0.00000007714,
      -0.00000000614,     0.00000002604,
       0.00000000306,     0.00000000009,
       0.00000814235,    -0.00000135328,
      -0.00000089771,    -0.00000315042,
      -0.00000062254,     0.00000027028,
       0.00000005065,     0.00000007935,
       0.00000000711,    -0.00000000693,
      -0.00000140803,    -0.00000808272,
      -0.00000311416,     0.00000092764,
       0.00000027862,     0.00000061127,
       0.00000007675,    -0.00000005289,
      -0.00000000702,    -0.00000000680,
      -0.00000007158,     0.00000001577,
       0.00000000907,     0.00000001627,
       0.00000000328,    -0.00000000225,
       0.00000000376,     0.00000006975,
       0.00000001615,    -0.00000000436,
      -0.00000000133,    -0.00000000328,
      -0.00000003644,     0.00000003275,
       0.00000002570,     0.00000001846,
       0.00000000320,     0.00000000306,
       0.00000000153,    -0.00000000254,
      -0.00000000087,    -0.00000000030,
       0.00000000276,    -0.00000000582,
      -0.00000000420,    -0.00000000097,
       0.00000000025,     0.00000000153,
       0.00000000575,     0.00000000180,
       0.00000000025,    -0.00000000470,
      -0.00000000156,     0.00000000019,
       0.00000000361,    -0.00000000715,
      -0.00000000518,    -0.00000000133,
       0.00000000031,     0.00000000191,
       0.00000000724,     0.00000000212,
       0.00000000070,    -0.00000000518,
      -0.00000000196,     0.00000000024,
      -0.00000000066,     0.00000000084,
       0.00000000085,     0.00000000035,
       0.00000000097,     0.00000000062,
       0.00000000035,    -0.00000000085,
       0.00000001135,    -0.00000000941,
      -0.00000000286,     0.00000000311,
       0.00000000310,     0.00000000285,
       0.00000000008,    -0.00000000157,
      -0.00000000117,     0.00000000028,
      -0.00000000099,     0.00000000144,
       0.00000000108,     0.00000000034,
       0.00000000140,     0.00000000107,
       0.00000000050,    -0.00000000089,
      -0.00000000985,     0.00000001151,
       0.00000000126,     0.00000000101,
       0.00000000058,    -0.00000000121,
       0.00000000253,    -0.00000000065,
      -0.00000000081,    -0.00000000200,
      -0.00000000080,     0.00000000050,
      -0.00000000061,    -0.00000000253,
      -0.00000000199,     0.00000000077,
       0.00000000052,     0.00000000076,
      -0.00000013502,     0.00000016264,
      -0.00000000691,    -0.00000000574,
       0.00000000031,     0.00000000103,
       0.00000000080,    -0.00000000035,
       0.00000003378,    -0.00000000290,
      -0.00000000714,    -0.00000002816,
      -0.00000000137,    -0.00000000422,
      -0.00000000327,     0.00000000157,
       0.00000000081,     0.00000000124,
       0.00000000139,    -0.00000000338,
      -0.00000000267,     0.00000000007,
       0.00000000019,     0.00000000094,
      -0.00000000353,    -0.00000000146,
      -0.00000000002,     0.00000000281,
       0.00000000095,    -0.00000000017,
       0.00000000159,    -0.00000000079,
      -0.00000000079,    -0.00000000159,
       0.00000002106,     0.00000005863,
      -0.00000000176,     0.00000000063,
      -0.00000000174,     0.00000000268,
      -0.00000000267,    -0.00000000175,
      -0.00000000248,     0.00000000380,
      -0.00000000380,    -0.00000000248,
      -0.00000000310,     0.00000000476,
      -0.00000000476,    -0.00000000311,
       0.00000007411,    -0.00000018894,
      -0.00000013617,    -0.00000002397,
       0.00000000094,     0.00000004730,
       0.00000000559,     0.00000000244,
       0.00000000109,    -0.00000000402,
      -0.00000000154,     0.00000000014,
      -0.00000000178,     0.00000000161,
       0.00000000130,     0.00000000088,
       0.00000000156,     0.00000000201,
       0.00000000109,    -0.00000000133,
       0.00000001544,    -0.00000002975,
       0.00000000126,     0.00000000066,
       0.00000000035,    -0.00000000071,
      -0.00000000073,    -0.00000000033,
      -0.00000000147,     0.00000000286,
       0.00000000288,     0.00000000144,
      -0.00000000013,    -0.00000000135,
      -0.00000000418,     0.00000000800,
       0.00000000413,     0.00000000175,
       0.00000000034,    -0.00000000117,
      -0.00000000257,    -0.00000000371,
      -0.00000000354,     0.00000000283,
       0.00000003265,    -0.00000007095,
      -0.00000002036,    -0.00000000555,
      -0.00000001100,    -0.00000000473,
      -0.00000000094,     0.00000000300,
       0.00000001123,     0.00000001543,
       0.00000000173,    -0.00000000187,
       0.00000001520,    -0.00000001161,
      -0.00000000214,    -0.00000000159,
       0.00000011527,    -0.00000001457,
      -0.00000001763,    -0.00000002367,
      -0.00000000240,     0.00000000398,
      -0.00000000630,    -0.00000005368,
      -0.00000001104,     0.00000000813,
       0.00000000187,     0.00000000114,
       0.00000000169,    -0.00000000545,
      -0.00000000693,    -0.00000000241,
       0.00000000042,     0.00000000120,
      -0.00000047989,     0.00000003980,
       0.00000006704,     0.00000010169,
       0.00000001095,    -0.00000001592,
       0.00000001377,     0.00000011586,
       0.00000002379,    -0.00000001760,
      -0.00000000393,    -0.00000000250,
       0.00000000285,     0.00000002525,
       0.00000000280,    -0.00000000318,
       0.00000002572,    -0.00000000131,
      -0.00000000286,    -0.00000000300,
       0.00000000055,    -0.00000000078,
      -0.00000001478,    -0.00000001047,
      -0.00000000137,     0.00000000200,
       0.00000000157,     0.00000000014,
      -0.00000000009,    -0.00000000154,
       0.00000000231,     0.00000000772,
       0.00000000107,    -0.00000000030,
      -0.00000000064,    -0.00000000060,
       0.00000000179,    -0.00000000694,
      -0.00000000325,    -0.00000000085,
      -0.00000000902,     0.00000003072,
       0.00000000707,     0.00000000188,
      -0.00000000138,    -0.00000000054,
      -0.00000000551,     0.00000001309,
       0.00000000304,     0.00000000119,
       0.00000000077,     0.00000000051,
       0.00000012567,    -0.00000031617,
      -0.00000022756,    -0.00000004455,
      -0.00000000095,     0.00000007958,
       0.00000000912,     0.00000000364,
       0.00000000132,    -0.00000000657,
      -0.00000000247,     0.00000000035,
      -0.00000000035,     0.00000000067,
       0.00000000070,     0.00000000033,
      -0.00000000020,    -0.00000000175,
       0.00000000074,     0.00000000082,
       0.00000000048,    -0.00000000096,
       0.00000000096,     0.00000000049,
       0.00000000063,    -0.00000000119,
       0.00000000121,     0.00000000064,
       0.00000000004,     0.00000000032,
      -0.00000000008,     0.00000000002,
       0.00000000010,     0.00000000068,
       0.00000000019,     0.00000000123,
       0.00000000028,     0.00000000233,
      -0.00000000057,     0.00000000014,
       0.00000000040,    -0.00000000040,
      -0.00000956539,     0.00000185122,
       0.00000117916,     0.00000367722,
       0.00000072004,    -0.00000034487,
      -0.00000006369,    -0.00000009031,
      -0.00000000791,     0.00000000836,
       0.00000062920,     0.00000442909,
       0.00000171228,    -0.00000048040,
      -0.00000015004,    -0.00000033597,
      -0.00000004240,     0.00000002853,
       0.00000000373,     0.00000000378,
      -0.00000152812,     0.00000290806,
       0.00000082826,     0.00000028990,
       0.00000001760,    -0.00000011887,
      -0.00000001104,     0.00000000122,
       0.00000304519,     0.00000160302,
       0.00000030919,    -0.00000089321,
      -0.00000013404,    -0.00000001860,
       0.00000000154,     0.00000001333,
      -0.00000001828,     0.00000005214,
       0.00000000322,    -0.00000000156,
       0.00000004058,     0.00000002706,
       0.00000000119,    -0.00000000075,
       0.00002500560,    -0.00000533253,
      -0.00000291479,    -0.00000957123,
      -0.00000190651,     0.00000079723,
       0.00000014335,     0.00000024625,
       0.00000002281,    -0.00000001845,
      -0.00000130012,    -0.00000912916,
      -0.00000352430,     0.00000099457,
       0.00000031076,     0.00000069116,
       0.00000008656,    -0.00000005921,
      -0.00000000781,    -0.00000000748,
       0.00000560271,    -0.00001318055,
      -0.00000379680,    -0.00000086315,
       0.00000000895,     0.00000055330,
       0.00000005187,    -0.00000001780,
      -0.00000000267,    -0.00000000345,
      -0.00001310717,    -0.00000579551,
      -0.00000096438,     0.00000376157,
       0.00000054567,     0.00000001864,
      -0.00000001270,    -0.00000005091,
      -0.00000000346,     0.00000000249,
       0.00000005834,    -0.00000012649,
      -0.00000001143,    -0.00000000178,
      -0.00000000067,     0.00000000271,
      -0.00000011644,    -0.00000006527,
      -0.00000000355,     0.00000000942,
       0.00000000253,     0.00000000072,
      -0.00000001556,     0.00000000497,
       0.00000000249,     0.00000000697,
       0.00000000123,    -0.00000000082,
      -0.00000000597,    -0.00000001652,
      -0.00000000719,     0.00000000278,
       0.00000000088,     0.00000000127,
      -0.00000001935,     0.00000000608,
       0.00000000298,     0.00000000863,
       0.00000000154,    -0.00000000102,
      -0.00000000727,    -0.00000002052,
      -0.00000000892,     0.00000000334,
       0.00000000111,     0.00000000159,
       0.00000000102,     0.00000000082,
      -0.00000001552,     0.00000000598,
       0.00000000046,     0.00000000124,
       0.00000000091,    -0.00000000049,
       0.00000002649,    -0.00000001530,
      -0.00000001005,    -0.00000001016,
      -0.00000000753,     0.00000000310,
       0.00000000178,     0.00000000342,
       0.00000000300,     0.00000000733,
       0.00000000330,    -0.00000000170,
      -0.00000047320,     0.00000017078,
       0.00000008637,     0.00000020943,
       0.00000004784,    -0.00000002329,
      -0.00000000529,    -0.00000001860,
      -0.00000000808,     0.00000000206,
       0.00000000106,     0.00000000158,
      -0.00000001443,    -0.00000001993,
      -0.00000000817,     0.00000000639,
       0.00000000106,     0.00000000089,
       0.00000000630,    -0.00000000071,
      -0.00000000124,    -0.00000000632,
       0.00000010748,     0.00000015020,
       0.00000001623,    -0.00000001901,
       0.00000002036,    -0.00000001595,
      -0.00000000269,    -0.00000000225,
      -0.00000002709,     0.00000000340,
       0.00000000475,     0.00000002720,
       0.00000000867,    -0.00000003345,
      -0.00000000468,     0.00000000198,
      -0.00000001336,    -0.00000000296,
       0.00000000086,     0.00000000185,
       0.00000000334,     0.00000000557,
       0.00000000501,    -0.00000000524,
       0.00000011483,     0.00000020664,
       0.00000001432,    -0.00000004300,
       0.00000003386,     0.00000002135,
       0.00000000051,    -0.00000000587,
      -0.00000001953,    -0.00000002369,
       0.00000000013,     0.00000000134,
      -0.00000002328,     0.00000001686,
       0.00000001671,     0.00000000150,
       0.00000000015,    -0.00000000228,
       0.00000000055,     0.00000000093,
       0.00000000115,    -0.00000000092,
      -0.00000000566,    -0.00000000553,
      -0.00000000077,     0.00000000080,
       0.00000000447,    -0.00000000853,
      -0.00000085014,     0.00000028584,
       0.00000013533,     0.00000037384,
       0.00000008552,    -0.00000003467,
      -0.00000000862,    -0.00000002575,
      -0.00000001134,     0.00000000407,
       0.00000000138,     0.00000000203,
       0.00000000077,     0.00000000036,
      -0.00000000085,    -0.00000000152,
      -0.00000000150,     0.00000000090,
       0.00000002288,    -0.00000002185,
      -0.00000000124,    -0.00000000129,
       0.00000000019,     0.00000000007,
       0.00000000002,    -0.00000000006,
       0.00000000005,     0.00000000001,
       0.00000000044,     0.00000000019,
       0.00000000073,     0.00000000031,
       0.00000000141,     0.00000000055,
       0.00000000021,    -0.00000000029,
      -0.00000881939,     0.00001690979,
       0.00000489965,     0.00000164249,
       0.00000009441,    -0.00000071705,
      -0.00000006833,     0.00000000792,
       0.00000000214,     0.00000000470,
       0.00000725565,     0.00000307243,
       0.00000058694,    -0.00000204814,
      -0.00000029611,    -0.00000003180,
       0.00000000360,     0.00000002800,
       0.00000000202,    -0.00000000092,
      -0.00000458486,    -0.00000805862,
      -0.00000101291,     0.00000073044,
       0.00000006728,     0.00000005792,
       0.00000000190,    -0.00000000398,
      -0.00000788261,     0.00000497739,
       0.00000088196,     0.00000093436,
       0.00000003912,    -0.00000009646,
      -0.00000000756,     0.00000000133,
      -0.00000001177,    -0.00000005205,
       0.00000000774,    -0.00000000997,
      -0.00000003778,     0.00000005082,
      -0.00000000413,    -0.00000001102,
       0.00004839814,    -0.00012976258,
      -0.00003728660,    -0.00000605337,
       0.00000048432,     0.00000535994,
       0.00000048986,    -0.00000021106,
      -0.00000003018,    -0.00000003080,
      -0.00001698909,    -0.00000862496,
      -0.00000189003,     0.00000480765,
       0.00000070184,     0.00000017269,
       0.00000000423,    -0.00000006833,
      -0.00000000496,     0.00000000073,
       0.00002052216,     0.00003431943,
       0.00000423406,    -0.00000344064,
      -0.00000034883,    -0.00000021822,
      -0.00000000263,     0.00000002442,
       0.00003405359,    -0.00002063619,
      -0.00000347905,    -0.00000414528,
      -0.00000020213,     0.00000035504,
       0.00000002511,     0.00000000033,
       0.00000003334,     0.00000009630,
      -0.00000002827,     0.00000003040,
       0.00000000264,     0.00000000134,
       0.00000008143,    -0.00000007877,
       0.00000002274,     0.00000003252,
       0.00000000223,    -0.00000000160,
      -0.00000009060,    -0.00000001689,
      -0.00000000265,     0.00000003219,
       0.00000000162,     0.00000000230,
       0.00000000063,    -0.00000000054,
      -0.00000000140,     0.00000000035,
       0.00000000031,     0.00000000143,
      -0.00000000005,    -0.00000001205,
      -0.00000000391,     0.00000000103,
      -0.00000011174,    -0.00000009665,
      -0.00000002582,     0.00000004667,
      -0.00000000703,     0.00000000188,
       0.00000000111,     0.00000000309,
       0.00000000100,    -0.00000000902,
      -0.00000000359,     0.00000000062,
       0.00000000606,     0.00000001073,
       0.00000000651,     0.00000001128,
       0.00000001046,    -0.00000001141,
      -0.00000000085,    -0.00000000070,
      -0.00000000042,     0.00000000078,
      -0.00000000143,     0.00000000068,
       0.00000000064,     0.00000000105,
       0.00000000063,     0.00000000143,
       0.00000000106,    -0.00000000060,
      -0.00000002899,     0.00000001368,
       0.00000001293,     0.00000002115,
       0.00000000101,     0.00000000228,
       0.00000000169,    -0.00000000099,
      -0.00000000046,    -0.00000000063,
       0.00000000149,     0.00000000062,
       0.00000000060,    -0.00000000151,
      -0.00000000630,    -0.00000000263,
      -0.00000000003,     0.00000000268,
      -0.00000000262,     0.00000000634,
       0.00000000269,    -0.00000000002,
      -0.00000000060,     0.00000000061,
       0.00000000116,     0.00000000092,
       0.00000000090,    -0.00000000115,
       0.00000000081,     0.00000000054,
      -0.00000000182,    -0.00000000026,
      -0.00000000025,     0.00000000182,
       0.00000000104,    -0.00000000048,
      -0.00000000048,    -0.00000000107,
      -0.00000016574,    -0.00000000057,
       0.00000000939,     0.00000006836,
      -0.00000000113,    -0.00000000877,
      -0.00000000360,     0.00000000109,
       0.00000000033,     0.00000000072,
      -0.00000000288,     0.00000000128,
       0.00000000051,     0.00000000083,
       0.00000000127,     0.00000000298,
       0.00000000088,    -0.00000000051,
       0.00000000441,     0.00000000857,
       0.00000000319,    -0.00000000215,
      -0.00000000236,     0.00000000002,
       0.00000000026,     0.00000000239,
      -0.00000003369,    -0.00000005373,
      -0.00000000600,     0.00000000580,
      -0.00000000778,     0.00000000519,
       0.00000000082,     0.00000000088,
       0.00000001015,    -0.00000000032,
      -0.00000000078,    -0.00000001024,
       0.00000009737,    -0.00000000560,
      -0.00000000567,    -0.00000002549,
       0.00000000222,    -0.00000000249,
      -0.00000000486,    -0.00000000151,
       0.00000000004,     0.00000000126,
      -0.00000000130,     0.00000000390,
       0.00000000091,    -0.00000000011,
       0.00000004404,    -0.00000000608,
      -0.00000000226,    -0.00000001619,
       0.00000000012,     0.00000000110,
       0.00000000228,    -0.00000000130,
      -0.00000027504,     0.00000003664,
       0.00000000537,     0.00000003926,
      -0.00000000153,    -0.00000000609,
      -0.00000000876,     0.00000000416,
      -0.00000007224,    -0.00000008051,
      -0.00000000775,     0.00000002897,
       0.00000000366,    -0.00000000317,
      -0.00000000124,    -0.00000000038,
      -0.00000000062,     0.00000000080,
       0.00000000007,     0.00000000008,
       0.00000000014,     0.00000000032,
      -0.00000000032,     0.00000000014,
       0.00000000093,    -0.00000000072,
       0.00000000072,     0.00000000094,
       0.00000003605,     0.00000003713,
       0.00000001382,    -0.00000001277,
       0.00000000192,    -0.00000000615,
      -0.00000000376,    -0.00000000164,
      -0.00000020223,    -0.00000021110,
       0.00000000467,     0.00000000967,
      -0.00000003193,     0.00000003063,
       0.00000000143,    -0.00000000086,
      -0.00000000293,     0.00000002277,
      -0.00000000181,    -0.00000000426,
       0.00000002115,     0.00000001206,
      -0.00000000252,     0.00000000116,
       0.00000000679,    -0.00000000231,
      -0.00000000089,    -0.00000000097,
       0.00000000232,     0.00000000679,
       0.00000000097,    -0.00000000088,
       0.00000000879,    -0.00000000278,
      -0.00000000110,    -0.00000000127,
       0.00000000279,     0.00000000879,
       0.00000000127,    -0.00000000110,
       0.00000036608,    -0.00000011622,
      -0.00000004583,    -0.00000005278,
       0.00000000365,     0.00000001150,
       0.00000000165,    -0.00000000144,
      -0.00000000063,     0.00000000043,
       0.00000000661,     0.00000000951,
       0.00000000147,    -0.00000000101,
      -0.00000000021,    -0.00000003572,
       0.00000000111,    -0.00000000001,
       0.00000001269,    -0.00000002533,
       0.00000000079,     0.00000000039,
      -0.00000034963,     0.00000002045,
       0.00000002598,     0.00000013964,
       0.00000002830,    -0.00000000930,
      -0.00000000057,    -0.00000001038,
      -0.00000000432,     0.00000000112,
       0.00000000037,     0.00000000085,
       0.00000000077,    -0.00000000048,
      -0.00000000784,    -0.00000001330,
      -0.00000000192,     0.00000000120,
       0.00000000245,     0.00000000008,
      -0.00000000006,    -0.00000000249,
       0.00000003795,     0.00000005461,
       0.00000000195,    -0.00000000134,
      -0.00000000925,    -0.00000000996,
      -0.00000000213,     0.00000000253,
      -0.00000001511,    -0.00000003712,
      -0.00000000947,     0.00000000527,
       0.00000003729,    -0.00000001473,
      -0.00000000517,    -0.00000000951,
      -0.00000000186,    -0.00000000114,
      -0.00000000032,     0.00000000064,
       0.00000000111,    -0.00000000234,
      -0.00000000079,    -0.00000000031,
      -0.00000000458,    -0.00000000092,
      -0.00000000012,     0.00000000087,
       0.00000000305,     0.00000000059,
      -0.00000000028,    -0.00000000071,
      -0.00000000063,     0.00000000302,
       0.00000000071,    -0.00000000027,
      -0.00000000123,    -0.00000000177,
      -0.00000000064,     0.00000000046,
       0.00000000174,    -0.00000000123,
      -0.00000000046,    -0.00000000063,
      -0.00000000717,    -0.00000000138,
      -0.00000000023,     0.00000000129,
       0.00000000333,    -0.00000000608,
      -0.00000000170,    -0.00000000042,
       0.00000000599,     0.00000000336,
       0.00000000043,    -0.00000000168,
      -0.00000000007,     0.00000000210,
       0.00000000074,    -0.00000000011,
      -0.00000000208,    -0.00000000007,
       0.00000000011,     0.00000000074,
       0.00000001652,     0.00000000102,
      -0.00000000061,    -0.00000000182,
      -0.00000012646,     0.00000011363,
       0.00000002827,     0.00000002402,
      -0.00000011285,    -0.00000012596,
      -0.00000002407,     0.00000002820,
      -0.00000000511,     0.00000000099,
       0.00000000019,     0.00000000172,
      -0.00000000084,    -0.00000000506,
      -0.00000000173,     0.00000000014,
      -0.00003439970,    -0.00005752107,
      -0.00000697025,     0.00000587620,
       0.00000057788,     0.00000035503,
       0.00000000343,    -0.00000003922,
      -0.00002042213,     0.00001338154,
       0.00000214620,     0.00000254440,
       0.00000012918,    -0.00000020774,
      -0.00000001419,    -0.00000000143,
       0.00001557276,     0.00000053014,
       0.00000018315,     0.00000002801,
       0.00000000096,     0.00000001394,
      -0.00000000229,     0.00000000046,
      -0.00000085591,    -0.00001577836,
      -0.00000002087,     0.00000018601,
       0.00000006445,     0.00000000045,
      -0.00000000038,    -0.00000000231,
      -0.00000004728,     0.00000006106,
       0.00000000012,     0.00000000597,
      -0.00000005568,     0.00000000954,
       0.00000000584,     0.00000000138,
       0.00019788831,     0.00032967413,
       0.00003994742,    -0.00003378355,
      -0.00000332260,    -0.00000203000,
      -0.00000001924,     0.00000022580,
       0.00000001044,    -0.00000000474,
       0.00004883670,    -0.00003195496,
      -0.00000510043,    -0.00000610067,
      -0.00000031140,     0.00000049232,
       0.00000003362,     0.00000000372,
      -0.00006873029,    -0.00000161942,
      -0.00000058633,     0.00000006098,
       0.00000002969,    -0.00000009361,
       0.00000000661,    -0.00000000559,
       0.00000255568,     0.00006882274,
      -0.00000008301,    -0.00000054016,
      -0.00000024714,     0.00000003061,
       0.00000000545,     0.00000000648,
       0.00000018183,    -0.00000006388,
      -0.00000000044,    -0.00000003895,
      -0.00000000576,     0.00000000095,
       0.00000005025,    -0.00000013757,
      -0.00000003730,    -0.00000000153,
       0.00000000063,     0.00000000550,
       0.00000000534,    -0.00000001644,
       0.00000006818,    -0.00000002978,
      -0.00000001521,    -0.00000001679,
      -0.00000000184,     0.00000000339,
      -0.00000030888,    -0.00000103316,
      -0.00000028563,     0.00000018648,
       0.00000004670,     0.00000003768,
       0.00000000272,    -0.00000000695,
       0.00000103390,    -0.00000030430,
      -0.00000018635,    -0.00000028583,
      -0.00000003771,     0.00000004668,
       0.00000000694,     0.00000000272,
       0.00000000849,    -0.00000004535,
      -0.00000001572,    -0.00000000956,
      -0.00000000156,     0.00000000544,
       0.00000000098,    -0.00000000003,
      -0.00000038657,    -0.00000129443,
      -0.00000035788,     0.00000023342,
       0.00000005844,     0.00000004721,
       0.00000000341,    -0.00000000869,
       0.00000129641,    -0.00000038116,
      -0.00000023329,    -0.00000035820,
      -0.00000004726,     0.00000005845,
       0.00000000870,     0.00000000341,
      -0.00000000004,     0.00000000004,
      -0.00000000004,    -0.00000000004,
      -0.00000002024,     0.00000000037,
       0.00000000026,     0.00000000152,
      -0.00000000018,    -0.00000000106,
      -0.00000002153,     0.00000000032,
      -0.00000000061,    -0.00000000306,
      -0.00000000027,     0.00000000083,
       0.00000000213,    -0.00000000650,
       0.00000000034,    -0.00000000039,
      -0.00000000285,     0.00000000322,
      -0.00000000040,    -0.00000000050,
       0.00000000050,    -0.00000000041,
       0.00000001513,    -0.00000001279,
      -0.00000001183,    -0.00000000871,
      -0.00000000088,    -0.00000000070,
      -0.00000000042,     0.00000000079,
      -0.00000000244,     0.00000000076,
       0.00000000063,     0.00000000091,
       0.00000000069,     0.00000000242,
       0.00000000091,    -0.00000000055,
      -0.00000005271,     0.00000001586,
       0.00000001499,     0.00000003149,
       0.00000000111,     0.00000000337,
       0.00000000143,    -0.00000000093,
       0.00000000674,    -0.00000000050,
      -0.00000000056,    -0.00000000295,
      -0.00000000049,    -0.00000000670,
      -0.00000000295,     0.00000000056,
       0.00000014095,    -0.00000002482,
      -0.00000002106,    -0.00000005993,
      -0.00000001238,     0.00000000721,
      -0.00000000203,    -0.00000000849,
      -0.00000000378,     0.00000000082,
       0.00000000028,     0.00000000086,
      -0.00000003191,     0.00000000202,
       0.00000000260,     0.00000001338,
       0.00000000284,    -0.00000000095,
       0.00000000199,     0.00000003169,
       0.00000001328,    -0.00000000257,
      -0.00000000094,    -0.00000000282,
       0.00000001785,     0.00000000177,
      -0.00000000102,    -0.00000000751,
       0.00000000047,    -0.00000000739,
      -0.00000000306,     0.00000000049,
      -0.00000000565,     0.00000000242,
       0.00000000115,     0.00000000145,
       0.00000000251,     0.00000000544,
       0.00000000131,    -0.00000000116,
      -0.00000005086,    -0.00000002140,
      -0.00000000019,     0.00000002183,
      -0.00000000276,     0.00000000732,
       0.00000000303,     0.00000000002,
       0.00000001332,    -0.00000000841,
      -0.00000000360,    -0.00000000204,
      -0.00000000870,    -0.00000001316,
      -0.00000000192,     0.00000000372,
      -0.00000059242,     0.00000000772,
       0.00000004029,     0.00000025315,
       0.00000005455,    -0.00000001681,
       0.00000000653,     0.00000004416,
       0.00000001822,    -0.00000000485,
      -0.00000000151,    -0.00000000384,
       0.00000010278,     0.00000001393,
      -0.00000000227,    -0.00000004237,
      -0.00000000875,     0.00000000205,
       0.00000000050,     0.00000000116,
       0.00000001414,    -0.00000010147,
      -0.00000004180,     0.00000000212,
       0.00000000200,     0.00000000863,
       0.00000000115,    -0.00000000049,
      -0.00000004783,     0.00000000652,
       0.00000000517,     0.00000001865,
       0.00000000364,    -0.00000000159,
       0.00000000276,     0.00000001709,
       0.00000000661,    -0.00000000198,
      -0.00000000059,    -0.00000000130,
       0.00000000078,    -0.00000000035,
      -0.00000000034,    -0.00000000079,
       0.00000000910,     0.00000000743,
       0.00000000091,    -0.00000000135,
      -0.00000000339,     0.00000000134,
       0.00000000069,     0.00000000062,
       0.00000000139,     0.00000000338,
       0.00000000065,    -0.00000000064,
       0.00000126289,     0.00000127801,
       0.00000031903,    -0.00000061123,
      -0.00000014037,    -0.00000002835,
       0.00000000120,     0.00000002008,
      -0.00000004671,    -0.00000015431,
      -0.00000005800,     0.00000002464,
       0.00000000613,     0.00000001050,
       0.00000000129,    -0.00000000104,
      -0.00000011604,    -0.00000039264,
      -0.00000010870,     0.00000007072,
       0.00000001775,     0.00000001436,
       0.00000000103,    -0.00000000265,
      -0.00000039336,     0.00000010978,
       0.00000006807,     0.00000010931,
       0.00000001456,    -0.00000001718,
      -0.00000000257,    -0.00000000109,
       0.00000000840,    -0.00000000342,
      -0.00000000055,    -0.00000000118,
       0.00000000034,     0.00000000242,
       0.00000000245,    -0.00000000035,
      -0.00000839437,    -0.00002842586,
      -0.00000786486,     0.00000510822,
       0.00000128019,     0.00000103839,
       0.00000007515,    -0.00000019050,
      -0.00000001962,    -0.00000000200,
       0.00000132739,    -0.00000039546,
      -0.00000023997,    -0.00000036675,
      -0.00000004835,     0.00000006004,
       0.00000000893,     0.00000000349,
      -0.00000000904,    -0.00000001588,
      -0.00000000342,     0.00000000404,
       0.00000000082,     0.00000000028,
      -0.00000003327,     0.00000001238,
       0.00000000681,     0.00000000866,
       0.00000000104,    -0.00000000160,
       0.00000000242,     0.00000000057,
       0.00000000061,    -0.00000000239,
      -0.00000003347,    -0.00000004322,
      -0.00000000160,    -0.00000000466,
      -0.00000000120,     0.00000000091,
      -0.00000000467,     0.00000000169,
       0.00000000094,     0.00000000119,
      -0.00000009654,    -0.00000028093,
      -0.00000007204,     0.00000005540,
       0.00000000856,    -0.00000000294,
      -0.00000000169,    -0.00000000219,
      -0.00000001729,     0.00000000100,
       0.00000000029,     0.00000000603,
       0.00000009051,    -0.00000000519,
      -0.00000000069,    -0.00000001403,
       0.00000000045,    -0.00000000123,
      -0.00000004578,     0.00000002278,
       0.00000000110,     0.00000000354,
      -0.00000000306,    -0.00000000708,
      -0.00000000177,     0.00000000158,
       0.00000000710,    -0.00000000318,
      -0.00000000158,    -0.00000000177,
       0.00000000210,     0.00000000051,
      -0.00000000051,     0.00000000210,
       0.00000000113,     0.00000000477,
       0.00000000142,    -0.00000000082,
       0.00000000477,    -0.00000000113,
      -0.00000000082,    -0.00000000142,
       0.00000006347,     0.00000026779,
       0.00000007930,    -0.00000004563,
      -0.00000000868,     0.00000000207,
       0.00000000149,     0.00000000257,
       0.00000000564,     0.00000001345,
       0.00000000455,    -0.00000000159,
       0.00000000185,    -0.00000000025,
       0.00000000022,     0.00000000174,
      -0.00000002720,    -0.00000006432,
       0.00000000596,     0.00000000372,
      -0.00000001046,     0.00000000381,
       0.00000000062,    -0.00000000101,
      -0.00000000312,     0.00000000060,
      -0.00000000025,    -0.00000000282,
       0.00000001272,    -0.00000003563,
       0.00000000252,     0.00000000561,
      -0.00000001195,    -0.00000000397,
       0.00000000182,    -0.00000000083,
       0.00000000166,     0.00000000112,
      -0.00000000234,     0.00000000171,
      -0.00000005603,     0.00000016392,
      -0.00000001193,    -0.00000002549,
       0.00000002730,     0.00000000920,
      -0.00000000421,     0.00000000191,
      -0.00000000250,    -0.00000000011,
       0.00000000505,    -0.00000000275,
      -0.00000003935,     0.00000002325,
       0.00000001161,    -0.00000000579,
      -0.00000000245,    -0.00000000473,
      -0.00000000096,     0.00000000057,
      -0.00000000580,    -0.00000001160,
      -0.00000000473,     0.00000000244,
       0.00000000057,     0.00000000096,
      -0.00001599754,    -0.00005397079,
      -0.00001492741,     0.00000971966,
       0.00000243483,     0.00000196956,
       0.00000014211,    -0.00000036237,
      -0.00000003731,    -0.00000000376,
       0.00000169445,    -0.00000050223,
      -0.00000030529,    -0.00000046836,
      -0.00000006179,     0.00000007646,
       0.00000001137,     0.00000000446,
       0.00000000012,    -0.00000000117,
      -0.00000001881,     0.00000000651,
       0.00000000330,     0.00000000648,
       0.00000000115,    -0.00000000082,
       0.00000000394,    -0.00000001226,
      -0.00000000541,    -0.00000000039,
       0.00000000016,     0.00000000117,
      -0.00000000097,    -0.00000000406,
      -0.00000000103,     0.00000000088,
      -0.00000000406,     0.00000000097,
       0.00000000088,     0.00000000103,
      -0.00000000001,     0.00000000145,
       0.00000002277,    -0.00000000020,
      -0.00000000001,    -0.00000000344,
      -0.00000010392,     0.00000003805,
       0.00000000127,     0.00000000386,
      -0.00000000099,     0.00000000153,
      -0.00000000154,    -0.00000000098,
      -0.00000000090,    -0.00000000378,
      -0.00000000096,     0.00000000082,
       0.00000000378,    -0.00000000090,
      -0.00000000082,    -0.00000000096,
       0.00000000011,    -0.00000000026,
       0.00000000065,     0.00000000018,
       0.00000000001,    -0.00000000011,
      -0.00000000021,     0.00000000069,
       0.00000000011,     0.00000000001,
      -0.00000000056,    -0.00000000041,
       0.00000000028,     0.00000000045,
      -0.00000000046,     0.00000000028,
      -0.00000000075,    -0.00000000093,
       0.00000000056,     0.00000000094,
      -0.00000000094,     0.00000000058,
      -0.00000000083,    -0.00000000385,
       0.00000000115,     0.00000000000,
       0.00000000044,     0.00000000150,
       0.00014630965,    -0.00000250287,
      -0.00000018526,     0.00000005243,
       0.00000000694,     0.00000031839,
      -0.00000034504,    -0.00004844379,
       0.00000000900,     0.00000003553,
       0.00000010225,    -0.00000000078,
      -0.00000055378,    -0.00000306657,
      -0.00000015003,     0.00000040192,
       0.00000001723,     0.00000002813,
      -0.00000000060,    -0.00000000175,
       0.00000144130,    -0.00000332089,
      -0.00000013356,    -0.00000045279,
      -0.00000000242,     0.00000003088,
       0.00000000149,     0.00000000104,
      -0.00000003812,    -0.00000012434,
       0.00000001018,    -0.00000001612,
       0.00000005630,    -0.00000009234,
       0.00000000850,     0.00000002001,
      -0.00067149835,     0.00001191689,
       0.00000100373,    -0.00000024251,
      -0.00000003587,    -0.00000146452,
       0.00000081066,     0.00011053682,
      -0.00000002131,    -0.00000008545,
      -0.00000023406,     0.00000000167,
      -0.00000223264,     0.00000683547,
      -0.00000012818,    -0.00000093508,
      -0.00000001204,    -0.00000007365,
       0.00000000295,     0.00000000306,
      -0.00000024863,     0.00000818403,
      -0.00000009230,     0.00000105885,
      -0.00000002777,    -0.00000006418,
      -0.00000000202,    -0.00000000347,
      -0.00000013616,     0.00000030403,
       0.00000001295,     0.00000003967,
       0.00000000184,    -0.00000000137,
       0.00000016505,     0.00000022556,
       0.00000000703,    -0.00000005001,
      -0.00000000302,     0.00000000001,
       0.00000000571,    -0.00000000378,
      -0.00000000051,    -0.00000000065,
      -0.00000004064,    -0.00000008685,
      -0.00000001183,     0.00000000626,
       0.00000000056,     0.00000000080,
       0.00000008700,    -0.00000003969,
      -0.00000000633,    -0.00000001179,
      -0.00000000081,     0.00000000056,
      -0.00000000151,    -0.00000000070,
      -0.00000005096,    -0.00000010875,
      -0.00000001481,     0.00000000786,
       0.00000000071,     0.00000000100,
       0.00000010900,    -0.00000004977,
      -0.00000000793,    -0.00000001476,
      -0.00000000101,     0.00000000070,
       0.00000000020,     0.00000000042,
      -0.00000000073,    -0.00000000056,
      -0.00000000089,     0.00000000057,
       0.00000000050,     0.00000000064,
       0.00000000049,     0.00000000088,
       0.00000000064,    -0.00000000043,
      -0.00000001948,     0.00000001224,
       0.00000001065,     0.00000001403,
       0.00000000080,     0.00000000136,
       0.00000000099,    -0.00000000071,
      -0.00000001849,     0.00000001249,
       0.00000000525,     0.00000000267,
       0.00000000485,     0.00000000659,
       0.00000000093,    -0.00000000193,
       0.00000000028,    -0.00000000494,
      -0.00000000498,     0.00000000026,
       0.00000010508,    -0.00000007110,
      -0.00000003013,    -0.00000001509,
      -0.00000001157,    -0.00000001576,
      -0.00000000220,     0.00000000461,
      -0.00000000055,     0.00000002174,
       0.00000000210,    -0.00000000207,
       0.00000002184,    -0.00000000063,
      -0.00000000250,    -0.00000000198,
       0.00000000032,     0.00000000327,
       0.00000000327,    -0.00000000061,
       0.00000000480,    -0.00000000218,
      -0.00000000072,    -0.00000000160,
      -0.00000000007,     0.00000000127,
       0.00000000132,     0.00000000003,
      -0.00000002566,     0.00000000872,
       0.00000000490,     0.00000000489,
       0.00000000172,     0.00000000376,
       0.00000000071,    -0.00000000080,
       0.00000000050,    -0.00000000553,
      -0.00000000560,    -0.00000000036,
       0.00000005532,     0.00000009661,
       0.00000001097,    -0.00000001149,
      -0.00000000118,     0.00000000088,
      -0.00000001526,    -0.00000003268,
      -0.00000000439,     0.00000000237,
      -0.00000003254,     0.00000001515,
       0.00000000232,     0.00000000437,
      -0.00000000118,     0.00000000016,
       0.00000000229,     0.00000001858,
       0.00000000280,    -0.00000000043,
      -0.00000000279,     0.00000000004,
       0.00000000012,     0.00000000291,
      -0.00000109235,    -0.00000236134,
      -0.00000031771,     0.00000016816,
       0.00000001523,     0.00000002189,
       0.00000011232,    -0.00000005192,
      -0.00000000814,    -0.00000001525,
      -0.00000000103,     0.00000000072,
      -0.00000000073,    -0.00000000159,
      -0.00000000245,     0.00000000163,
      -0.00000001246,    -0.00000002349,
       0.00000000072,    -0.00000000038,
      -0.00000000094,    -0.00000000069,
       0.00000000084,     0.00000000034,
       0.00000000064,    -0.00000000463,
      -0.00000000122,    -0.00000000008,
       0.00000000053,    -0.00000000144,
       0.00000000163,     0.00000000048,
      -0.00000000365,     0.00000001903,
       0.00000000296,     0.00000000024,
      -0.00000000069,     0.00000000122,
      -0.00000000215,    -0.00000000058,
      -0.00000208527,    -0.00000450845,
      -0.00000060656,     0.00000032105,
       0.00000002911,     0.00000004178,
       0.00000014283,    -0.00000006605,
      -0.00000001036,    -0.00000001939,
      -0.00000000132,     0.00000000092,
       0.00000000116,     0.00000000033,
      -0.00000000054,     0.00000000075,
       0.00000000007,     0.00000000001,
      -0.00000000242,     0.00000000229,
       0.00000000026,     0.00000000018,
      -0.00000000231,    -0.00000000243,
      -0.00000000018,     0.00000000026,
      -0.00000000003,     0.00000000018,
      -0.00000000022,    -0.00000000002,
       0.00000000014,    -0.00000000061,
      -0.00000000073,    -0.00000000023,
       0.00000000023,    -0.00000000074,
       0.00000000019,    -0.00000000100,
      -0.00000000210,    -0.00000000021,
       0.00000000021,    -0.00000000211,
       0.00000000099,    -0.00000000169,
       0.00000000827,     0.00000000541,
      -0.00000000541,     0.00000000834,
       0.00001575340,    -0.00000418357,
       0.00000119806,     0.00000121278,
      -0.00000006658,     0.00000011033,
      -0.00000000640,    -0.00000000181,
      -0.00000142605,    -0.00000363750,
       0.00000031112,    -0.00000032481,
       0.00000002775,     0.00000001680,
      -0.00000000048,     0.00000000170,
       0.00000830509,    -0.00000039125,
       0.00000006677,    -0.00000000226,
       0.00000000033,     0.00000002689,
      -0.00000023388,     0.00000840956,
      -0.00000000644,    -0.00000009885,
      -0.00000000238,     0.00000000030,
      -0.00000004086,     0.00000000416,
       0.00000000220,     0.00000000799,
       0.00000000144,    -0.00000000011,
      -0.00000000299,    -0.00000003989,
      -0.00000000826,     0.00000000138,
       0.00000000031,     0.00000000135,
      -0.00005064284,     0.00001190374,
      -0.00000383619,    -0.00000364859,
       0.00000019039,    -0.00000036447,
       0.00000002214,     0.00000000440,
       0.00000333894,     0.00000860574,
      -0.00000073388,     0.00000076829,
      -0.00000006549,    -0.00000003946,
       0.00000000113,    -0.00000000402,
      -0.00000951927,     0.00000145121,
      -0.00000032484,     0.00000002270,
      -0.00000000112,    -0.00000006834,
       0.00000000430,     0.00000000050,
       0.00000111348,    -0.00001008871,
       0.00000003245,     0.00000037483,
      -0.00000003291,    -0.00000000038,
       0.00000000034,    -0.00000000437,
       0.00000008262,    -0.00000000275,
      -0.00000000400,    -0.00000002267,
      -0.00000000296,     0.00000000033,
       0.00000003103,     0.00000008744,
       0.00000002472,    -0.00000000226,
      -0.00000000089,    -0.00000000285,
      -0.00000000299,     0.00000000809,
       0.00000010838,     0.00000005680,
       0.00000000585,    -0.00000000091,
      -0.00000005688,     0.00000010853,
       0.00000000089,     0.00000000586,
       0.00000000202,    -0.00000000053,
       0.00000013577,     0.00000007109,
       0.00000000731,    -0.00000000113,
      -0.00000007115,     0.00000013597,
       0.00000000111,     0.00000000734,
      -0.00000000076,    -0.00000000039,
      -0.00000001264,     0.00000000926,
       0.00000000058,     0.00000000085,
       0.00000000061,    -0.00000000049,
       0.00000000076,     0.00000000112,
       0.00000000111,    -0.00000000076,
      -0.00000000176,    -0.00000003820,
      -0.00000000344,     0.00000000431,
      -0.00000001262,     0.00000000036,
       0.00000000138,     0.00000000113,
      -0.00000000162,     0.00000000043,
      -0.00000000069,     0.00000000051,
       0.00000000901,     0.00000017166,
       0.00000001472,    -0.00000002006,
       0.00000002885,    -0.00000000074,
      -0.00000000321,    -0.00000000257,
       0.00000000509,    -0.00000000029,
       0.00000000143,    -0.00000000323,
       0.00000001335,     0.00000002342,
       0.00000000134,    -0.00000000089,
      -0.00000000412,    -0.00000000394,
      -0.00000000390,     0.00000000402,
       0.00000000021,     0.00000001030,
       0.00000000292,     0.00000000021,
       0.00000000567,    -0.00000003756,
      -0.00000000671,    -0.00000000033,
       0.00000000052,     0.00000000071,
      -0.00000000051,    -0.00000000118,
      -0.00000012958,    -0.00000007602,
      -0.00000000115,    -0.00000000035,
       0.00000004027,     0.00000002080,
       0.00000000216,    -0.00000000033,
       0.00000002053,    -0.00000004007,
      -0.00000000031,    -0.00000000212,
      -0.00000000003,    -0.00000000143,
      -0.00000001951,     0.00000000063,
       0.00000000007,     0.00000000327,
       0.00000292417,     0.00000153081,
       0.00000015763,    -0.00000002410,
      -0.00000007376,     0.00000014077,
       0.00000000116,     0.00000000760,
       0.00000000224,     0.00000000089,
       0.00000000122,    -0.00000000350,
       0.00000003091,     0.00000001382,
      -0.00000000043,     0.00000000096,
       0.00000000045,    -0.00000000067,
      -0.00000000067,    -0.00000000044,
       0.00000560381,     0.00000293346,
       0.00000030308,    -0.00000004639,
      -0.00000009355,     0.00000017863,
       0.00000000147,     0.00000000964,
      -0.00000000153,     0.00000000088,
       0.00000000116,     0.00000000232,
      -0.00000000013,     0.00000000007,
       0.00000000152,     0.00000000839,
      -0.00000000020,     0.00000000009,
      -0.00000000844,     0.00000000148,
      -0.00000000009,    -0.00000000020,
       0.00000000034,     0.00000000032,
      -0.00000000035,     0.00000000043,
       0.00000000026,    -0.00000000043,
      -0.00000000125,     0.00000000038,
      -0.00000000038,    -0.00000000125,
       0.00000000024,     0.00000000049,
      -0.00000000048,     0.00000000023,
       0.00000000018,    -0.00000000071,
      -0.00000000238,     0.00000000252,
      -0.00000000253,    -0.00000000238,
       0.00000000010,    -0.00000000052,
       0.00000000052,     0.00000000010,
      -0.00000000265,    -0.00000000244,
       0.00000000036,    -0.00000000052,
       0.00000002618,    -0.00000001339,
       0.00000000126,     0.00000000200,
       0.00000001349,     0.00000002635,
      -0.00000000200,     0.00000000126,
       0.00000000127,     0.00000000008,
      -0.00000000011,     0.00000000126,
       0.00000002506,    -0.00000229787,
       0.00000067211,     0.00000002063,
      -0.00000000543,     0.00000010009,
      -0.00000000846,    -0.00000000138,
      -0.00000070684,     0.00000014059,
      -0.00000001488,    -0.00000018954,
       0.00000002850,    -0.00000000060,
      -0.00000000023,     0.00000000237,
       0.00000112707,    -0.00000009940,
       0.00000003954,     0.00000014228,
      -0.00000001912,     0.00000000018,
       0.00000000058,    -0.00000000215,
       0.00000012408,     0.00000072942,
       0.00000001194,     0.00000007297,
      -0.00000001011,     0.00000001155,
      -0.00000000190,    -0.00000000113,
       0.00000005332,    -0.00000007401,
      -0.00000001035,    -0.00000000736,
       0.00000007356,     0.00000005528,
       0.00000000710,    -0.00000001048,
       0.00000305703,     0.00001102611,
      -0.00000319893,     0.00000030375,
       0.00000000397,    -0.00000046175,
       0.00000004020,     0.00000000605,
       0.00000171622,    -0.00000033230,
       0.00000003545,     0.00000045588,
      -0.00000006864,     0.00000000165,
       0.00000000055,    -0.00000000572,
      -0.00000150355,     0.00000011099,
       0.00000003593,    -0.00000037307,
       0.00000006935,     0.00000001958,
      -0.00000000372,     0.00000000895,
      -0.00000008350,     0.00000009202,
      -0.00000027875,    -0.00000012011,
       0.00000002816,    -0.00000006369,
       0.00000000881,     0.00000000415,
      -0.00000013031,     0.00000018585,
       0.00000002436,     0.00000001756,
       0.00000000064,    -0.00000000173,
      -0.00000018228,    -0.00000014056,
      -0.00000001613,     0.00000002524,
       0.00000000162,     0.00000000082,
      -0.00000000149,     0.00000001011,
       0.00000014212,     0.00000003899,
       0.00000000412,     0.00000000690,
      -0.00000003893,     0.00000014221,
      -0.00000000693,     0.00000000412,
       0.00000000223,    -0.00000000112,
       0.00000017791,     0.00000004872,
       0.00000000516,     0.00000000865,
      -0.00000000048,     0.00000000062,
      -0.00000004880,     0.00000017806,
      -0.00000000868,     0.00000000516,
      -0.00000000062,    -0.00000000048,
       0.00000000043,    -0.00000000025,
      -0.00000000362,     0.00000000203,
       0.00000001728,     0.00000000368,
       0.00000000017,    -0.00000000125,
      -0.00000000119,    -0.00000000025,
      -0.00000000026,     0.00000000112,
      -0.00000000654,    -0.00000000108,
      -0.00000000033,     0.00000000210,
       0.00000000328,    -0.00000000045,
       0.00000001827,     0.00000001350,
       0.00000000095,    -0.00000000480,
      -0.00000000176,    -0.00000000528,
       0.00000000113,     0.00000000073,
      -0.00000000917,     0.00000000203,
       0.00000000120,    -0.00000000083,
      -0.00000003860,    -0.00000000522,
      -0.00000000067,     0.00000000282,
      -0.00000000753,    -0.00000000166,
      -0.00000000164,     0.00000000784,
      -0.00000000040,    -0.00000000157,
      -0.00000010304,    -0.00000006725,
       0.00000000100,     0.00000000159,
       0.00000005080,     0.00000001399,
       0.00000000147,     0.00000000254,
       0.00000001413,    -0.00000005147,
       0.00000000247,    -0.00000000153,
       0.00000380239,     0.00000103833,
       0.00000011062,     0.00000018544,
      -0.00000001024,     0.00000001318,
      -0.00000005083,     0.00000018513,
      -0.00000000901,     0.00000000536,
      -0.00000000064,    -0.00000000050,
       0.00000000380,     0.00000000014,
       0.00000000090,    -0.00000000329,
       0.00000003933,     0.00000000844,
      -0.00000000026,     0.00000000122,
       0.00000731414,     0.00000199613,
       0.00000021303,     0.00000035667,
      -0.00000001966,     0.00000002537,
      -0.00000006430,     0.00000023446,
      -0.00000001141,     0.00000000680,
      -0.00000000081,    -0.00000000063,
       0.00000000255,    -0.00000000077,
      -0.00000000074,     0.00000000328,
      -0.00000000005,    -0.00000000016,
      -0.00000000003,    -0.00000000054,
       0.00000000054,    -0.00000000002,
      -0.00000000165,    -0.00000000026,
       0.00000000026,    -0.00000000165,
       0.00000188662,     0.00000008061,
      -0.00000015406,     0.00000073461,
      -0.00000014601,    -0.00000004772,
       0.00000000827,    -0.00000001910,
       0.00000004242,    -0.00000053848,
       0.00000020867,     0.00000004951,
      -0.00000001462,     0.00000004142,
      -0.00000000540,    -0.00000000248,
      -0.00000001978,    -0.00000013750,
       0.00000003591,     0.00000000982,
      -0.00000000394,     0.00000000564,
       0.00000011649,    -0.00000008619,
       0.00000002407,     0.00000002575,
      -0.00000000312,     0.00000000498,
       0.00000000787,    -0.00000000032,
       0.00000000047,    -0.00000000101,
      -0.00000000043,     0.00000000708,
       0.00000000125,     0.00000000033,
      -0.00000562235,     0.00000231743,
       0.00000071212,    -0.00000275270,
       0.00000060741,     0.00000025573,
      -0.00000004298,     0.00000008390,
      -0.00000000846,    -0.00000000486,
      -0.00000034499,     0.00000151373,
      -0.00000053742,    -0.00000011376,
       0.00000003051,    -0.00000010150,
       0.00000001285,     0.00000000527,
      -0.00000007879,     0.00000016790,
      -0.00000006333,    -0.00000005988,
       0.00000001787,    -0.00000001087,
       0.00000000143,     0.00000000329,
      -0.00000009257,     0.00000018167,
      -0.00000007742,    -0.00000002389,
       0.00000000072,    -0.00000001877,
       0.00000000328,    -0.00000000037,
      -0.00000002093,    -0.00000000005,
      -0.00000000030,     0.00000000135,
       0.00000000515,    -0.00000001803,
      -0.00000000308,     0.00000000061,
      -0.00000000465,    -0.00000000339,
       0.00000000322,    -0.00000000476,
       0.00000002642,     0.00000002454,
      -0.00000000593,    -0.00000000408,
       0.00000000409,    -0.00000000592,
       0.00000000062,     0.00000000066,
       0.00000000085,    -0.00000000136,
      -0.00000000132,    -0.00000000082,
       0.00000000253,     0.00000001322,
       0.00000000377,    -0.00000000113,
      -0.00000000245,    -0.00000000041,
      -0.00000000027,     0.00000000181,
      -0.00000002228,    -0.00000006661,
       0.00000000920,     0.00000000915,
      -0.00000000904,     0.00000000275,
       0.00000000118,    -0.00000000126,
       0.00000000968,     0.00000000176,
      -0.00000000157,     0.00000000187,
       0.00000000138,    -0.00000000929,
       0.00000000197,     0.00000000163,
       0.00000000589,    -0.00000003057,
      -0.00000000156,    -0.00000000193,
       0.00000000063,     0.00000000261,
       0.00000000249,    -0.00000000077,
       0.00000000972,     0.00000005704,
       0.00000000321,    -0.00000000084,
      -0.00000000129,    -0.00000000180,
      -0.00000000140,     0.00000000172,
      -0.00000012077,    -0.00000009344,
       0.00000000378,    -0.00000000644,
      -0.00000023586,    -0.00000016387,
       0.00000000540,    -0.00000000784,
       0.00000002045,    -0.00000004071,
       0.00000002776,     0.00000001410,
      -0.00000000509,     0.00000000953,
      -0.00000001147,    -0.00000000261,
       0.00000000227,    -0.00000000762,
       0.00000000257,     0.00000000094,
       0.00000002207,    -0.00000000081,
       0.00000000009,     0.00000001104,
      -0.00000000293,    -0.00000000010,
       0.00000000084,     0.00000001839,
      -0.00000000822,     0.00000000056,
      -0.00000000025,    -0.00000000184,
       0.00000000063,    -0.00000000080,
       0.00000000079,     0.00000000065,
      -0.00000004028,     0.00000018714,
      -0.00000012832,    -0.00000003576,
       0.00000001551,    -0.00000004410,
       0.00000000992,     0.00000000434,
       0.00000002770,     0.00000000651,
      -0.00000000558,     0.00000001838,
      -0.00000000622,    -0.00000000228,
       0.00000000062,    -0.00000000141,
      -0.00000002622,    -0.00000000182,
       0.00000000171,    -0.00000001531,
       0.00000000468,     0.00000000092,
       0.00000000110,    -0.00000000979,
       0.00000000272,    -0.00000000082,
      -0.00000000062,     0.00000000188,
      -0.00000000176,    -0.00000000075,
      -0.00000000011,    -0.00000000154,
      -0.00000000147,     0.00000000010,
      -0.00000000592,    -0.00000002817,
      -0.00000000192,     0.00000000013,
      -0.00000001927,     0.00000000171,
      -0.00000000471,     0.00000000094,
       0.00000000033,     0.00000000113,
      -0.00000000112,    -0.00000000037,
       0.00000000002,    -0.00000000118,
       0.00000001959,    -0.00000000400,
      -0.00000000078,    -0.00000000275,
       0.00000000131,     0.00000000097,
       0.00000000053,     0.00000000158,
       0.00000001816,     0.00000000046,
      -0.00000000241,     0.00000001382,
      -0.00000000527,    -0.00000000168,
       0.00000000032,    -0.00000000451,
       0.00000000344,     0.00000000079,
      -0.00000000051,     0.00000000131,
       0.00000000158,    -0.00000000191,
       0.00000000134,     0.00000000113,
       0.00000000176,     0.00000000068,
      -0.00000000032,     0.00000000126,
      -0.00000007935,    -0.00000000331,
       0.00000001214,    -0.00000006113,
       0.00000002351,     0.00000000823,
      -0.00000000064,     0.00000001058,
      -0.00000000807,    -0.00000000176,
       0.00000000114,    -0.00000000309,
      -0.00000000237,     0.00000000276,
      -0.00000000192,    -0.00000000214,
       0.00000000099,    -0.00000000063,
      -0.00000000214,     0.00000000141,
      -0.00000000129,    -0.00000000159,
      -0.00000000009,    -0.00000000096,
       0.00000000111,     0.00000000001,
      -0.00000000015,     0.00000000103,
      -0.00000001656,     0.00000000042,
       0.00000000056,    -0.00000001456,
       0.00000000021,     0.00000000229,
      -0.00000000204,     0.00000000007,
      -0.00000000143,     0.00000000021,
      -0.00000000008,    -0.00000000118,
       0.00000000040,    -0.00000000111,
       0.00000000079,     0.00000000054,
      -0.00000000470,    -0.00000000065,
      -0.00000000065,     0.00000000466,
      -0.00000011639,    -0.00000001640,
      -0.00000000063,     0.00000000547,
       0.00000002467,     0.00000000256,
       0.00000000028,    -0.00000000172,
       0.00000000254,    -0.00000002444,
      -0.00000000169,    -0.00000000028,
      -0.00000000803,    -0.00000000085,
      -0.00000000022,     0.00000000256,
       0.00000000288,     0.00000000019,
       0.00000000018,    -0.00000000265,
       0.00000060471,     0.00000006401,
       0.00000000670,    -0.00000004206,
       0.00000000257,    -0.00000003018,
      -0.00000000215,    -0.00000000036,
      -0.00000010987,    -0.00000000743,
      -0.00000000059,     0.00000000517,
      -0.00000000732,     0.00000010845,
       0.00000000503,     0.00000000056,
       0.00000004123,     0.00000000287,
       0.00000000070,    -0.00000001260,
      -0.00000000995,    -0.00000000028,
      -0.00000000021,     0.00000000846,
      -0.00000265973,    -0.00000018349,
      -0.00000001341,     0.00000012421,
      -0.00000000783,     0.00000014279,
       0.00000000681,     0.00000000086,
       0.00000037384,     0.00000001059,
       0.00000000029,    -0.00000000914,
       0.00000001015,    -0.00000036595,
      -0.00000000856,    -0.00000000019,
       0.00000000032,    -0.00000000104,
      -0.00000016453,    -0.00000000465,
      -0.00000000112,     0.00000004837,
       0.00000000118,     0.00000000007,
       0.00000002142,    -0.00000000032,
      -0.00000000058,    -0.00000001290,
       0.00000887210,     0.00000025005,
       0.00000000399,    -0.00000021232,
       0.00000001160,    -0.00000052459,
      -0.00000001280,    -0.00000000064,
      -0.00000072828,     0.00000001499,
       0.00000000194,     0.00000000225,
       0.00000001631,     0.00000069040,
       0.00000000039,    -0.00000000220,
       0.00000038516,    -0.00000000926,
      -0.00000000190,    -0.00000010954,
      -0.00000001350,     0.00000000071,
       0.00000000113,    -0.00000002593,
      -0.00001664996,     0.00000040806,
       0.00000005084,     0.00000003449,
       0.00000002158,     0.00000112537,
       0.00000000183,    -0.00000000316,
      -0.00000015686,    -0.00000003187,
      -0.00000000343,    -0.00000001018,
      -0.00000003369,     0.00000029821,
      -0.00000000664,     0.00000000372,
       0.00000000088,    -0.00000000131,
      -0.00000000093,    -0.00000000093,
       0.00000012381,     0.00000002246,
       0.00000000593,     0.00000000159,
       0.00000000439,    -0.00000002690,
       0.00000000013,    -0.00000000151,
       0.00000005754,     0.00000000017,
       0.00000000073,     0.00000005688,
       0.00000009120,     0.00000001309,
      -0.00000001508,     0.00000001872,
       0.00000000221,    -0.00000001323,
       0.00000000268,     0.00000000223,
       0.00000000173,    -0.00000000158,
       0.00000000015,    -0.00000000239,
      -0.00000475106,    -0.00000087295,
      -0.00000008662,    -0.00000019158,
       0.00000000934,    -0.00000001048,
      -0.00000005750,     0.00000035010,
      -0.00000001362,     0.00000000589,
      -0.00000000072,    -0.00000000064,
      -0.00000018773,    -0.00000001455,
      -0.00000000003,    -0.00000000351,
      -0.00000001961,    -0.00000012115,
      -0.00000000393,     0.00000000086,
      -0.00000000062,    -0.00000000061,
      -0.00000000135,     0.00000000885,
       0.00000000106,     0.00000000014,
      -0.00000000049,     0.00000000092,
      -0.00000000690,    -0.00000000552,
      -0.00000000076,     0.00000000089,
      -0.00000000149,     0.00000000016,
       0.00000000015,     0.00000000139,
       0.00000002069,     0.00000000889,
       0.00000000131,    -0.00000000345,
       0.00000001880,     0.00000000382,
       0.00000000036,     0.00000000079,
      -0.00000000377,     0.00000001870,
      -0.00000000078,     0.00000000036,
       0.00000023342,    -0.00000003477,
       0.00000002299,    -0.00000000042,
      -0.00000000160,     0.00000002178,
      -0.00000000063,     0.00000000063,
       0.00000007156,     0.00000009336,
      -0.00000000118,     0.00000002086,
      -0.00000000216,     0.00000000102,
       0.00000010794,    -0.00000014383,
       0.00000002388,    -0.00000000023,
       0.00000000119,     0.00000000231,
       0.00000000099,    -0.00000000112,
       0.00000000105,     0.00000000079,
      -0.00000002584,     0.00000000315,
       0.00000000020,     0.00000000174,
       0.00000810328,     0.00000754119,
      -0.00000003461,     0.00000167771,
      -0.00000016723,     0.00000008301,
      -0.00000001160,    -0.00000001058,
      -0.00000037140,     0.00000039903,
      -0.00000008263,    -0.00000000171,
      -0.00000000409,    -0.00000000824,
       0.00000000052,    -0.00000000057,
       0.00000001573,     0.00000000882,
       0.00000000031,     0.00000000227,
       0.00000000293,    -0.00000000045,
       0.00000000093,     0.00000000029,
       0.00000000024,     0.00000000108,
       0.00000003850,     0.00000002290,
       0.00000000133,     0.00000000114,
       0.00000000122,    -0.00000000151,
      -0.00000000157,     0.00000000009,
       0.00000000008,     0.00000000157,
       0.00000008714,     0.00000007237,
      -0.00000000227,     0.00000000273,
       0.00000000253,     0.00000000183,
      -0.00000000183,     0.00000000253,
      -0.00000002311,     0.00000001035,
       0.00000001551,     0.00000002050,
       0.00000000882,    -0.00000001044,
       0.00000000043,     0.00000000108,
       0.00000000097,    -0.00000000064,
       0.00000000150,     0.00000000010,
      -0.00000000097,    -0.00000000129,
      -0.00000000104,     0.00000000043,
       0.00000000370,     0.00000000043,
      -0.00000000116,    -0.00000000130,
      -0.00000000065,     0.00000000058,
       0.00000000067,    -0.00000000286,
      -0.00000010234,     0.00000021014,
       0.00000014949,     0.00000005374,
       0.00000001215,    -0.00000005296,
      -0.00000001467,     0.00000000125,
       0.00000000307,     0.00000000965,
       0.00000000315,    -0.00000000162,
      -0.00000000049,    -0.00000000112,
       0.00000000645,    -0.00000000666,
      -0.00000000521,    -0.00000000310,
      -0.00000000042,     0.00000000165,
       0.00000000088,     0.00000000021,
      -0.00000000691,    -0.00000000847,
      -0.00000000461,     0.00000000577,
       0.00000000180,     0.00000000113,
       0.00000000020,    -0.00000000081,
      -0.00000000143,    -0.00000000061,
       0.00000000090,     0.00000000148,
       0.00000000144,    -0.00000000095,
      -0.00000001175,     0.00000002403,
       0.00000000348,     0.00000000158,
       0.00000000028,    -0.00000000101,
      -0.00000000373,    -0.00000000620,
      -0.00000000068,     0.00000000061,
      -0.00000000612,     0.00000000384,
       0.00000000064,     0.00000000068,
       0.00000001554,    -0.00000000952,
      -0.00000000082,    -0.00000000117,
       0.00000000015,    -0.00000000095,
       0.00000000098,    -0.00000000029,
      -0.00000000018,     0.00000000110,
       0.00000000092,     0.00000000036,
      -0.00000000031,     0.00000000094,
       0.00000000109,     0.00000000015,
      -0.00000000015,     0.00000000109,
       0.00000000159,     0.00000000007,
       0.00000000007,    -0.00000000159,
       0.00000004826,     0.00000000209,
       0.00000000009,    -0.00000000197,
      -0.00000002289,    -0.00000001305,
      -0.00000000122,    -0.00000000130,
      -0.00000000133,     0.00000000131,
      -0.00000007179,    -0.00000007493,
       0.00000000249,    -0.00000000238,
       0.00000000911,     0.00000002284,
      -0.00000000074,     0.00000000030,
      -0.00000002692,    -0.00000001584,
       0.00000000052,    -0.00000000088,
      -0.00000003512,    -0.00000001178,
       0.00000000039,    -0.00000000116,
       0.00000000128,     0.00000000218,
       0.00000000218,    -0.00000000128,
       0.00000000141,    -0.00000000001,
      -0.00000000003,    -0.00000000141,
       0.00000004669,     0.00000008639,
       0.00000000277,    -0.00000000238,
      -0.00000006068,     0.00000000069,
       0.00000000072,     0.00000006068,
       0.00000004933,    -0.00000000059,
      -0.00000000012,    -0.00000001335,
       0.00000000034,    -0.00000000085,
      -0.00000230082,     0.00000002571,
       0.00000000095,     0.00000009321,
      -0.00000002011,    -0.00000000427,
      -0.00000000048,    -0.00000000101,
      -0.00000000498,     0.00000002102,
      -0.00000000093,     0.00000000053,
       0.00000002191,     0.00000000527,
       0.00000000105,    -0.00000000456,
       0.00000001143,    -0.00000000005,
       0.00000000013,     0.00000001126,
      -0.00000061984,    -0.00000017225,
      -0.00000001131,    -0.00000002925,
      -0.00000000763,     0.00000003251,
      -0.00000000153,     0.00000000077,
      -0.00000001376,    -0.00000000149,
      -0.00000000172,    -0.00000001302,
       0.00000000435,     0.00000000224,
       0.00000000477,     0.00000000116,
      -0.00000000115,     0.00000000475,
       0.00000005624,     0.00000005841,
      -0.00000000293,     0.00000000283,
       0.00000241287,     0.00000119596,
       0.00000008070,     0.00000026749,
      -0.00000001864,     0.00000001688,
       0.00000000760,    -0.00000001089,
       0.00000000156,    -0.00000000014,
       0.00000012816,     0.00000012182,
      -0.00000000064,     0.00000002715,
      -0.00000000272,     0.00000000134,
       0.00000012381,    -0.00000013594,
       0.00000002751,     0.00000000047,
       0.00000000136,     0.00000000274,
       0.00000000812,     0.00000000900,
      -0.00000000070,     0.00000000061,
       0.00000000764,     0.00000001482,
       0.00000000109,     0.00000000216,
       0.00000000216,    -0.00000000108,
       0.00000004239,     0.00000008083,
      -0.00000000343,     0.00000000180,
       0.00000000137,     0.00000000129,
       0.00000000056,    -0.00000000091,
      -0.00000000082,    -0.00000000076,
      -0.00000000040,     0.00000000064,
      -0.00000001280,     0.00000000448,
       0.00000000485,     0.00000000969,
       0.00000000364,    -0.00000000245,
      -0.00000000080,    -0.00000000090,
       0.00000000448,     0.00000001280,
       0.00000000969,    -0.00000000485,
      -0.00000000245,    -0.00000000365,
      -0.00000000090,     0.00000000080,
       0.00000000115,    -0.00000000283,
      -0.00000000230,    -0.00000000001,
      -0.00000000284,    -0.00000000116,
      -0.00000000001,     0.00000000230,
      -0.00000000747,     0.00000000248,
       0.00000000118,     0.00000000330,
       0.00000000248,     0.00000000747,
       0.00000000330,    -0.00000000118,
      -0.00000000308,     0.00000000034,
       0.00000000032,     0.00000000123,
       0.00000000034,     0.00000000308,
       0.00000000123,    -0.00000000032,
      -0.00000001327,     0.00000000213,
       0.00000000153,     0.00000000507,
       0.00000000075,     0.00000000459,
       0.00000000178,    -0.00000000053,
      -0.00000004895,    -0.00000021112,
      -0.00000006258,     0.00000003575,
       0.00000001004,    -0.00000000253,
      -0.00000000178,    -0.00000000296,
      -0.00000000042,    -0.00000000135,
      -0.00000000148,     0.00000000046,
      -0.00000194484,     0.00000000979,
       0.00000013882,     0.00000073459,
       0.00000014126,    -0.00000005393,
      -0.00000000013,    -0.00000001998,
      -0.00000000754,     0.00000000144,
       0.00000000039,     0.00000000107,
      -0.00000013765,    -0.00000046278,
      -0.00000012799,     0.00000008353,
       0.00000002092,     0.00000001688,
       0.00000000123,    -0.00000000313,
      -0.00000046290,     0.00000013735,
       0.00000008340,     0.00000012804,
       0.00000001690,    -0.00000002089,
      -0.00000000313,    -0.00000000123,
       0.00000000226,     0.00000000044,
       0.00000000045,    -0.00000000221,
       0.00000001778,     0.00000001995,
      -0.00000000056,    -0.00000000224,
      -0.00000000089,     0.00000000034,
      -0.00000000123,    -0.00000000591,
      -0.00000000177,     0.00000000095,
      -0.00000000591,     0.00000000112,
       0.00000000090,     0.00000000178,
       0.00000000336,    -0.00000000100,
      -0.00000000061,    -0.00000000093,
      -0.00000000119,     0.00000000036,
      -0.00000000715,    -0.00000001788,
       0.00000000531,    -0.00000009706,
      -0.00000002571,    -0.00000000489,
      -0.00000000034,    -0.00000000079,
      -0.00000001809,    -0.00000003889,
      -0.00000000523,     0.00000000280,
      -0.00000003889,     0.00000001809,
       0.00000000279,     0.00000000523,
       0.00000002104,     0.00000001236,
      -0.00000000062,     0.00000000104,
       0.00000011640,     0.00000009685,
       0.00000001770,    -0.00000001271,
       0.00000000073,     0.00000000038,
       0.00000004825,     0.00000002490,
       0.00000000258,    -0.00000000039,
       0.00000002488,    -0.00000004823,
      -0.00000000039,    -0.00000000258,
       0.00000002724,     0.00000000914,
      -0.00000000046,     0.00000000137,
       0.00000038526,    -0.00000002006,
       0.00000000054,    -0.00000000121,
       0.00000006195,     0.00000001698,
       0.00000000180,     0.00000000302,
       0.00000001702,    -0.00000006204,
       0.00000000302,    -0.00000000180,
      -0.00000000203,    -0.00000000142,
      -0.00000000142,     0.00000000208,
       0.00000001924,     0.00000001687,
       0.00000000145,     0.00000000024,
       0.00000000023,    -0.00000000144,
       0.00000003129,     0.00000000513,
       0.00000000022,    -0.00000000157,
      -0.00000000567,    -0.00000000071,
      -0.00000000070,     0.00000000560,
      -0.00000011807,    -0.00000001494,
      -0.00000000071,     0.00000000634,
       0.00000001647,     0.00000000141,
       0.00000000137,    -0.00000001611,
      -0.00000000607,    -0.00000000052,
      -0.00000000015,     0.00000000179,
       0.00000032706,     0.00000002796,
       0.00000000154,    -0.00000001934,
      -0.00000002610,    -0.00000000095,
      -0.00000000082,     0.00000002467,
       0.00000001117,     0.00000000037,
       0.00000000013,    -0.00000000317,
      -0.00000048170,    -0.00000001559,
      -0.00000000126,     0.00000003252,
      -0.00000000759,    -0.00000000118,
      -0.00000000148,     0.00000001183,
       0.00000000016,    -0.00000000109,
       0.00000000129,     0.00000000007,
      -0.00000000004,     0.00000000122,
      -0.00000014868,    -0.00000002518,
      -0.00000000171,     0.00000001169,
      -0.00000000475,    -0.00000000067,
      -0.00000000038,    -0.00000000209,
      -0.00000000018,     0.00000000112,
       0.00000000068,     0.00000000141,
       0.00000000169,    -0.00000000225,
      -0.00000004841,     0.00000000230,
       0.00000000010,     0.00000000193,
      -0.00000000092,     0.00000000100,
       0.00000000664,    -0.00000000581,
      -0.00000000470,    -0.00000000341,
      -0.00000000072,     0.00000000162,
      -0.00000000581,    -0.00000000664,
      -0.00000000341,     0.00000000470,
       0.00000000162,     0.00000000072,
       0.00000000058,     0.00000000064,
       0.00000000055,     0.00000000120,
       0.00000000053,    -0.00000000054,
       0.00000000123,    -0.00000000057,
      -0.00000000054,    -0.00000000053,
       0.00000000316,    -0.00000000100,
      -0.00000000100,    -0.00000000316,
       0.00000000245,     0.00000001195,
       0.00000000824,    -0.00000000358,
      -0.00000044896,     0.00000010538,
       0.00000011702,     0.00000026981,
      -0.00000001964,    -0.00000000617,
      -0.00000000697,     0.00000001874,
      -0.00000003092,    -0.00000014788,
      -0.00000008851,     0.00000003098,
       0.00000119193,    -0.00000026837,
      -0.00000022144,    -0.00000063232,
       0.00000008678,     0.00000002857,
       0.00000000886,    -0.00000003433,
       0.00000002786,    -0.00000008510,
      -0.00000003472,    -0.00000000632,
       0.00000001433,     0.00000004823,
      -0.00000006135,     0.00000001665,
      -0.00000000060,     0.00000000085,
      -0.00000002702,     0.00000005248,
       0.00000003112,     0.00000001102,
      -0.00000001868,    -0.00000000923,
      -0.00000000336,     0.00000000932,
       0.00000000053,    -0.00000000149,
      -0.00000000118,     0.00000000079,
       0.00000000093,     0.00000000121,
      -0.00000001009,    -0.00000000451,
      -0.00000000143,     0.00000000368,
       0.00000000247,     0.00000000059,
       0.00000000494,    -0.00000000432,
      -0.00000000179,    -0.00000000170,
      -0.00000000284,    -0.00000000528,
      -0.00000000181,     0.00000000115,
      -0.00000000332,     0.00000000222,
       0.00000000137,     0.00000000186,
       0.00000000064,    -0.00000000058,
      -0.00000000016,    -0.00000000194,
      -0.00000000191,     0.00000000012,
       0.00000002202,    -0.00000001520,
      -0.00000000655,    -0.00000000796,
      -0.00000000340,    -0.00000000457,
      -0.00000000177,     0.00000000150,
       0.00000000049,     0.00000000811,
       0.00000000197,    -0.00000000015,
       0.00000000787,    -0.00000000032,
      -0.00000000009,    -0.00000000195,
      -0.00000000059,    -0.00000000911,
      -0.00000000424,     0.00000000025,
       0.00000000135,     0.00000003934,
       0.00000000922,    -0.00000000053,
      -0.00000000010,    -0.00000000192,
       0.00000000078,    -0.00000000041,
       0.00000000632,    -0.00000001018,
       0.00000000156,    -0.00000000075,
       0.00000000010,     0.00000000099,
       0.00000000142,    -0.00000000036,
       0.00000000310,     0.00000000724,
       0.00000000118,    -0.00000000051,
       0.00001565403,     0.00001455622,
      -0.00000006611,     0.00000323949,
      -0.00000032300,     0.00000016049,
      -0.00000002243,    -0.00000002044,
      -0.00000046983,     0.00000050495,
      -0.00000010451,    -0.00000000216,
      -0.00000000517,    -0.00000001041,
       0.00000000066,    -0.00000000072,
       0.00000001881,     0.00000000736,
       0.00000000085,     0.00000000184,
      -0.00000001211,     0.00000001869,
      -0.00000000234,     0.00000000028,
       0.00000000106,     0.00000000094,
       0.00000000096,    -0.00000000116,
       0.00000000107,     0.00000000092,
      -0.00000000092,     0.00000000107,
       0.00000000057,     0.00000000068,
       0.00000000066,     0.00000000071,
       0.00000000038,    -0.00000000064,
      -0.00000000071,     0.00000000066,
       0.00000000064,     0.00000000038,
       0.00000000040,     0.00000000124,
       0.00000000094,    -0.00000000044,
      -0.00000005473,     0.00000001404,
       0.00000001635,     0.00000004190,
      -0.00000000109,    -0.00000000340,
      -0.00000000258,     0.00000000122,
       0.00000000062,     0.00000000099,
      -0.00000000134,    -0.00000000380,
      -0.00000000312,     0.00000000182,
       0.00000000112,     0.00000000126,
       0.00000000357,    -0.00000000142,
      -0.00000000185,    -0.00000000297,
      -0.00000000121,     0.00000000110,
      -0.00000000705,     0.00000000593,
       0.00000000485,     0.00000000366,
       0.00000000077,    -0.00000000168,
      -0.00000000593,    -0.00000000705,
      -0.00000000366,     0.00000000485,
       0.00000000168,     0.00000000077,
      -0.00000000175,    -0.00000000055,
      -0.00000000024,     0.00000000128,
       0.00000005457,    -0.00000006000,
      -0.00000004263,    -0.00000003590,
       0.00000000511,     0.00000000140,
       0.00000000051,    -0.00000000350,
      -0.00000000119,    -0.00000000002,
       0.00000002999,    -0.00000000968,
      -0.00000001078,    -0.00000002275,
      -0.00000000856,     0.00000000558,
       0.00000000182,     0.00000000210,
       0.00000000965,     0.00000003006,
       0.00000002279,    -0.00000001077,
      -0.00000000557,    -0.00000000861,
      -0.00000000210,     0.00000000182,
       0.00000000709,    -0.00000001491,
       0.00000000255,     0.00000000137,
       0.00000000016,    -0.00000000105,
      -0.00000000066,    -0.00000000062,
       0.00000000091,    -0.00000000107,
      -0.00000000494,    -0.00000000536,
      -0.00000000329,     0.00000000297,
       0.00000000145,     0.00000000111,
      -0.00000000091,     0.00000000340,
       0.00000000297,     0.00000000030,
      -0.00000000337,    -0.00000000073,
      -0.00000000029,     0.00000000296,
       0.00000000096,    -0.00000000005,
       0.00000000055,    -0.00000000091,
      -0.00000000100,    -0.00000000044,
       0.00000000415,    -0.00000000808,
      -0.00000000125,    -0.00000000063,
       0.00000000111,     0.00000000178,
       0.00000000175,    -0.00000000120,
       0.00000001148,    -0.00000002644,
       0.00000000079,     0.00000000030,
      -0.00000000092,     0.00000000063,
       0.00000001101,     0.00000001765,
       0.00000000234,    -0.00000000159,
      -0.00000000312,     0.00000000014,
       0.00000000029,     0.00000000316,
       0.00000000487,    -0.00000000035,
      -0.00000000014,    -0.00000000185,
      -0.00000003312,     0.00000000243,
       0.00000000033,     0.00000000459,
       0.00000000038,    -0.00000000102,
      -0.00000000134,    -0.00000000009,
       0.00000000045,     0.00000000071,
      -0.00000000071,     0.00000000045,
       0.00000000153,    -0.00000001085,
      -0.00000000169,    -0.00000000034,
      -0.00000000114,     0.00000000032,
      -0.00000000032,    -0.00000000114,
      -0.00000000192,     0.00000000119,
       0.00000000050,     0.00000000067,
       0.00000000120,     0.00000000197,
       0.00000000068,    -0.00000000051,
      -0.00000000090,     0.00000004647,
      -0.00000000021,     0.00000000244,
       0.00000000095,    -0.00000000012,
       0.00000000205,     0.00000000590,
       0.00000000151,    -0.00000000114,
       0.00000000593,    -0.00000000216,
      -0.00000000120,    -0.00000000152,
       0.00000007921,     0.00000021504,
       0.00000005415,    -0.00000004302,
      -0.00000000985,     0.00000000366,
       0.00000000199,     0.00000000247,
       0.00000000015,     0.00000000118,
       0.00000000118,     0.00000000011,
       0.00000000011,    -0.00000000116,
       0.00000003340,     0.00000000297,
       0.00000000013,    -0.00000000160,
      -0.00000000211,    -0.00000000017,
      -0.00000000016,     0.00000000199,
      -0.00000007001,    -0.00000000409,
      -0.00000000025,     0.00000000355,
      -0.00000000375,     0.00000000024,
       0.00000000021,     0.00000000420,
      -0.00000002736,     0.00000000273,
       0.00000000038,     0.00000000191,
       0.00000002930,    -0.00000000163,
      -0.00000000156,    -0.00000003044,
      -0.00000001313,     0.00000000063,
       0.00000000018,     0.00000000365,
       0.00000056893,    -0.00000002722,
      -0.00000000183,    -0.00000003789,
       0.00000000742,     0.00000000081,
       0.00000000069,    -0.00000000698,
      -0.00000000337,    -0.00000000195,
      -0.00000000144,     0.00000000005,
      -0.00000000007,    -0.00000000140,
       0.00000000098,     0.00000000004,
       0.00000010992,     0.00000001528,
       0.00000000095,    -0.00000000819,
       0.00000000426,    -0.00000000016,
       0.00000000017,     0.00000000344,
      -0.00000000053,     0.00000000067,
       0.00000000070,     0.00000000048,
      -0.00000000388,     0.00000000582,
       0.00000000093,     0.00000000065,
      -0.00000002161,    -0.00000001607,
      -0.00000000119,    -0.00000000136,
      -0.00000000154,     0.00000000229,
      -0.00000006575,    -0.00000005467,
       0.00000000265,    -0.00000000318,
       0.00000000963,     0.00000001793,
      -0.00000000083,     0.00000000045,
      -0.00000002354,    -0.00000001052,
       0.00000000049,    -0.00000000110,
      -0.00000002981,    -0.00000000640,
       0.00000000031,    -0.00000000142,
      -0.00000000054,     0.00000000062,
       0.00000000085,    -0.00000000031,
      -0.00000000015,    -0.00000000086,
      -0.00000000026,    -0.00000000101,
      -0.00000000033,     0.00000000097,
       0.00000000094,     0.00000000037,
       0.00000000113,     0.00000000041,
       0.00000000068,    -0.00000000041,
       0.00000001950,     0.00000004470,
      -0.00000004906,    -0.00000012644,
      -0.00000000100,     0.00000000084,
       0.00000001855,     0.00000005093,
      -0.00000003365,     0.00000003577,
      -0.00000003993,    -0.00000003757,
       0.00000000091,     0.00000000036,
       0.00000000055,    -0.00000000054,
       0.00000000048,    -0.00000000083,
      -0.00000000107,     0.00000000014,
       0.00000000007,     0.00000000104,
       0.00000000006,     0.00000000111,
      -0.00000000062,     0.00000000071,
       0.00000000041,     0.00000000129,
      -0.00000000018,     0.00000000017,
       0.00000000015,     0.00000000431,
      -0.00000000069,     0.00000000047,
      -0.00000000004,    -0.00000000002,
      -0.00000000431,     0.00000000013,
      -0.00000000047,    -0.00000000069,
       0.00000000002,    -0.00000000004,
       0.00000000019,     0.00000000013,
      -0.00000000015,     0.00000000023,
      -0.00000000052,    -0.00000000012,
       0.00000000012,    -0.00000000052,
      -0.00000000175,     0.00000000113,
      -0.00000000029,    -0.00000000024,
      -0.00000000113,    -0.00000000175,
       0.00000000024,    -0.00000000029,
       0.00000000066,    -0.00000000094,
       0.00000000428,     0.00000000705,
      -0.00000000047,     0.00000000129,
      -0.00000000706,     0.00000000428,
      -0.00000000129,    -0.00000000047,
       0.00000002021,    -0.00000003097,
       0.00000002106,     0.00000001317,
      -0.00000000451,     0.00000000723,
      -0.00000000831,    -0.00000000306,
       0.00000000227,    -0.00000000553,
       0.00000000186,     0.00000000084,
       0.00000001671,    -0.00000000140,
       0.00000000050,     0.00000000834,
      -0.00000000220,     0.00000000006,
       0.00000000113,     0.00000001439,
      -0.00000000656,     0.00000000056,
      -0.00000000019,    -0.00000000151,
      -0.00000005217,     0.00000013674,
      -0.00000009379,    -0.00000003829,
       0.00000001457,    -0.00000003228,
       0.00000000730,     0.00000000371,
       0.00000002085,     0.00000000886,
      -0.00000000583,     0.00000001365,
      -0.00000000459,    -0.00000000207,
      -0.00000001178,     0.00000001586,
      -0.00000000093,    -0.00000000820,
       0.00000000297,     0.00000000038,
      -0.00000001510,    -0.00000000141,
       0.00000000026,    -0.00000000159,
      -0.00000000093,     0.00000000125,
      -0.00000000116,    -0.00000000102,
      -0.00000000132,     0.00000000110,
       0.00000001342,     0.00000002078,
      -0.00000000133,     0.00000000482,
      -0.00000002078,     0.00000001342,
      -0.00000000482,    -0.00000000133,
       0.00000001676,     0.00000002596,
      -0.00000000166,     0.00000000603,
      -0.00000000082,     0.00000000021,
      -0.00000002597,     0.00000001676,
      -0.00000000603,    -0.00000000167,
      -0.00000000021,    -0.00000000082,
      -0.00000000038,    -0.00000000029,
      -0.00000000028,     0.00000000056,
       0.00000000015,    -0.00000000039,
      -0.00000000086,    -0.00000000038,
      -0.00000000062,     0.00000000020,
       0.00000000036,    -0.00000000190,
       0.00000000136,     0.00000000031,
      -0.00000000075,    -0.00000000233,
      -0.00000000229,     0.00000000079,
      -0.00000000530,    -0.00000001605,
       0.00000002040,    -0.00000000675,
       0.00000000063,     0.00000000022,
       0.00000000095,    -0.00000000038,
       0.00000000197,     0.00000000195,
      -0.00000000498,     0.00000000522,
      -0.00000000336,    -0.00000000997,
       0.00000000993,    -0.00000000322,
      -0.00000000155,     0.00000000067,
      -0.00000000251,     0.00000000006,
      -0.00000000051,     0.00000000000,
       0.00000000124,     0.00000000020,
       0.00000000024,    -0.00000000078,
       0.00000000036,    -0.00000000018,
      -0.00000000126,     0.00000000032,
       0.00000000113,     0.00000000072,
      -0.00000000918,    -0.00000000564,
      -0.00000001190,     0.00000001127,
      -0.00000000066,     0.00000000055,
      -0.00000000055,    -0.00000000066,
      -0.00000000840,     0.00000000500,
       0.00000000148,    -0.00000000137,
      -0.00000000181,    -0.00000000154,
      -0.00000000079,     0.00000000122,
      -0.00000000133,    -0.00000000147,
      -0.00000000155,     0.00000000177,
       0.00000000115,     0.00000000075,
      -0.00000000051,    -0.00000000069,
      -0.00000000069,     0.00000000051,
      -0.00000000153,     0.00000000260,
       0.00000000112,     0.00000000067,
       0.00000002726,    -0.00000001495,
      -0.00000002207,    -0.00000002974,
      -0.00000001579,     0.00000001580,
      -0.00000000044,    -0.00000000081,
      -0.00000000088,     0.00000000065,
       0.00000000013,    -0.00000000319,
      -0.00000000339,     0.00000000035,
       0.00000000049,     0.00000000181,
      -0.00000000321,    -0.00000000011,
       0.00000000025,     0.00000000337,
       0.00000000182,    -0.00000000034,
       0.00000001719,    -0.00000002027,
      -0.00000002535,    -0.00000001687,
      -0.00000000805,     0.00000001606,
      -0.00000000025,    -0.00000000077,
      -0.00000000098,     0.00000000044,
       0.00000000441,     0.00000001020,
       0.00000000997,    -0.00000000544,
      -0.00000000340,    -0.00000000497,
      -0.00000000124,     0.00000000152,
       0.00000001011,    -0.00000000438,
      -0.00000000534,    -0.00000000985,
      -0.00000000476,     0.00000000318,
       0.00000000127,     0.00000000124,
      -0.00000002025,    -0.00000000026,
       0.00000000355,     0.00000001915,
       0.00000000906,    -0.00000000357,
      -0.00000000297,    -0.00000001867,
      -0.00000001765,     0.00000000656,
       0.00000000571,     0.00000000838,
       0.00000000327,    -0.00000000224,
       0.00000002198,     0.00000001579,
       0.00000001206,    -0.00000002232,
      -0.00000001109,    -0.00000000395,
      -0.00000000073,     0.00000000329,
       0.00000001540,    -0.00000002258,
      -0.00000002281,    -0.00000001179,
      -0.00000000389,     0.00000001141,
       0.00000000332,     0.00000000092,
       0.00000018005,     0.00000007362,
       0.00000004465,    -0.00000017875,
      -0.00000008757,    -0.00000000859,
       0.00000000081,     0.00000002854,
       0.00000000576,     0.00000004838,
       0.00000004494,    -0.00000001562,
      -0.00000001254,    -0.00000002013,
      -0.00000000507,     0.00000000632,
       0.00000000215,     0.00000000042,
      -0.00000004357,     0.00000005469,
       0.00000004892,     0.00000003036,
       0.00000000853,    -0.00000002183,
      -0.00000000628,    -0.00000000168,
       0.00000005487,     0.00000004279,
       0.00000002931,    -0.00000004884,
      -0.00000002199,    -0.00000000891,
      -0.00000000175,     0.00000000655,
       0.00000000025,     0.00000000079,
       0.00000000089,    -0.00000000015,
       0.00000000045,     0.00000000098,
       0.00000000098,    -0.00000000044,
       0.00000003882,     0.00000016637,
       0.00000013895,    -0.00000006135,
      -0.00000003945,    -0.00000005715,
      -0.00000001507,     0.00000001516,
       0.00000000452,     0.00000000303,
       0.00000030929,    -0.00000006774,
      -0.00000010962,    -0.00000025927,
      -0.00000010640,     0.00000007068,
       0.00000002843,     0.00000002776,
       0.00000000521,    -0.00000000777,
      -0.00000003488,    -0.00000000314,
       0.00000000240,     0.00000002880,
       0.00000001176,    -0.00000000259,
      -0.00000000183,    -0.00000000397,
      -0.00000000148,     0.00000003516,
       0.00000002846,    -0.00000000433,
      -0.00000000395,    -0.00000001107,
      -0.00000000339,     0.00000000269,
      -0.00000000761,     0.00000000214,
       0.00000000174,     0.00000000388,
       0.00000000216,     0.00000000753,
       0.00000000377,    -0.00000000198,
      -0.00000085553,    -0.00000240226,
      -0.00000198850,     0.00000115147,
       0.00000068761,     0.00000080550,
       0.00000020704,    -0.00000025638,
      -0.00000006970,    -0.00000003662,
      -0.00000000368,     0.00000001317,
      -0.00000076834,     0.00000018797,
       0.00000028937,     0.00000063993,
       0.00000026203,    -0.00000018371,
      -0.00000007018,    -0.00000006818,
      -0.00000001192,     0.00000001869,
       0.00000000495,     0.00000000138,
       0.00000127191,    -0.00000035272,
      -0.00000041403,    -0.00000097228,
      -0.00000036884,     0.00000022120,
       0.00000007377,     0.00000009134,
       0.00000001627,    -0.00000001777,
      -0.00000000405,    -0.00000000289,
      -0.00000035568,    -0.00000127380,
      -0.00000097356,     0.00000041777,
       0.00000022357,     0.00000036861,
       0.00000009044,    -0.00000007616,
      -0.00000001888,    -0.00000001503,
      -0.00000000213,     0.00000000484,
       0.00000001836,    -0.00000000498,
      -0.00000000458,    -0.00000000933,
      -0.00000000225,     0.00000000177,
      -0.00000000495,    -0.00000001823,
      -0.00000000927,     0.00000000445,
       0.00000000153,     0.00000000229,
      -0.00000035330,     0.00000009193,
       0.00000011125,     0.00000027086,
       0.00000010296,    -0.00000006009,
      -0.00000002040,    -0.00000002546,
      -0.00000000453,     0.00000000503,
       0.00000009122,     0.00000024988,
       0.00000018970,    -0.00000009684,
      -0.00000004910,    -0.00000007139,
      -0.00000001746,     0.00000001602,
       0.00000000310,     0.00000000275,
       0.00000011512,    -0.00000036493,
      -0.00000025239,    -0.00000004456,
      -0.00000000335,     0.00000008744,
       0.00000001985,    -0.00000000218,
      -0.00000000086,    -0.00000000333,
      -0.00000036572,    -0.00000011033,
      -0.00000004003,     0.00000025286,
       0.00000008741,     0.00000000087,
      -0.00000000307,    -0.00000001983,
      -0.00000000344,     0.00000000114,
       0.00000000199,    -0.00000000602,
      -0.00000000243,    -0.00000000053,
      -0.00000000623,    -0.00000000120,
       0.00000000009,     0.00000000257,
       0.00000788716,    -0.00000237183,
      -0.00000270969,    -0.00000600535,
      -0.00000226956,     0.00000142107,
       0.00000047294,     0.00000055829,
       0.00000009890,    -0.00000011425,
      -0.00000002062,    -0.00000001218,
      -0.00000014211,    -0.00000036833,
      -0.00000028026,     0.00000014732,
       0.00000007342,     0.00000010601,
       0.00000002614,    -0.00000002354,
      -0.00000000564,    -0.00000000478,
      -0.00000049743,     0.00000159678,
       0.00000110267,     0.00000019274,
       0.00000001421,    -0.00000038070,
      -0.00000008649,     0.00000000952,
       0.00000000400,     0.00000001418,
       0.00000160586,     0.00000048496,
       0.00000018020,    -0.00000110958,
      -0.00000038315,    -0.00000000778,
       0.00000001211,     0.00000008705,
       0.00000001421,    -0.00000000469,
      -0.00000000409,     0.00000001360,
       0.00000000455,     0.00000000055,
       0.00000001386,     0.00000000337,
       0.00000000009,    -0.00000000466,
       0.00000003404,    -0.00000001111,
      -0.00000001239,    -0.00000002576,
       0.00000000038,     0.00000000006,
      -0.00000000002,    -0.00000000004,
       0.00000000122,     0.00000000012,
       0.00000000203,     0.00000000023,
       0.00000000429,     0.00000000082,
      -0.00000000127,     0.00000000038,
      -0.00000000060,    -0.00000000114,
       0.00000038798,    -0.00000123135,
      -0.00000085220,    -0.00000014661,
      -0.00000000866,     0.00000029421,
       0.00000006679,    -0.00000000818,
      -0.00000000333,    -0.00000001106,
      -0.00000068206,    -0.00000015846,
      -0.00000005005,     0.00000046611,
       0.00000015941,    -0.00000000375,
      -0.00000000612,    -0.00000003586,
      -0.00000000590,     0.00000000191,
       0.00000051210,     0.00000059945,
       0.00000028348,    -0.00000031193,
      -0.00000009505,    -0.00000006513,
      -0.00000000913,     0.00000001916,
       0.00000000250,     0.00000000108,
       0.00000058692,    -0.00000052245,
      -0.00000031924,    -0.00000027320,
      -0.00000006064,     0.00000009766,
       0.00000001966,     0.00000000801,
       0.00000000027,    -0.00000000304,
       0.00000000168,     0.00000000301,
       0.00000000141,    -0.00000000431,
      -0.00000000121,     0.00000000026,
      -0.00000289549,     0.00000932223,
       0.00000644592,     0.00000108710,
       0.00000005977,    -0.00000222406,
      -0.00000050484,     0.00000006272,
       0.00000002481,     0.00000008395,
       0.00000001060,    -0.00000000540,
       0.00000156969,     0.00000035586,
       0.00000010979,    -0.00000107193,
      -0.00000036655,     0.00000001012,
       0.00000001419,     0.00000008279,
       0.00000001403,    -0.00000000461,
      -0.00000000106,    -0.00000000201,
      -0.00000212846,    -0.00000254587,
      -0.00000121302,     0.00000128789,
       0.00000038948,     0.00000028292,
       0.00000004147,    -0.00000007753,
      -0.00000001123,    -0.00000000406,
      -0.00000248156,     0.00000226617,
       0.00000139104,     0.00000114963,
       0.00000025231,    -0.00000042815,
      -0.00000008706,    -0.00000003168,
      -0.00000000170,     0.00000001299,
      -0.00000000409,    -0.00000000471,
      -0.00000000313,     0.00000000722,
       0.00000000069,    -0.00000000104,
       0.00000000290,     0.00000000524,
       0.00000000374,    -0.00000000279,
      -0.00000000133,    -0.00000000131,
       0.00000007905,    -0.00000002735,
      -0.00000002966,    -0.00000005987,
      -0.00000002254,     0.00000001511,
       0.00000000495,     0.00000000552,
       0.00000000098,    -0.00000000123,
       0.00000002646,     0.00000007916,
       0.00000005996,    -0.00000002915,
      -0.00000001496,    -0.00000002257,
      -0.00000000553,     0.00000000493,
       0.00000000121,     0.00000000099,
       0.00000000079,    -0.00000000123,
      -0.00000000105,    -0.00000000049,
       0.00000009904,    -0.00000003433,
      -0.00000003719,    -0.00000007500,
      -0.00000002823,     0.00000001894,
       0.00000000621,     0.00000000691,
       0.00000000123,    -0.00000000153,
       0.00000003317,     0.00000009918,
       0.00000007512,    -0.00000003654,
      -0.00000001875,    -0.00000002828,
      -0.00000000693,     0.00000000617,
       0.00000000153,     0.00000000123,
       0.00000000004,    -0.00000000004,
       0.00000000042,    -0.00000000040,
       0.00000253027,     0.00000290890,
       0.00000136233,    -0.00000154745,
      -0.00000047206,    -0.00000030806,
      -0.00000004196,     0.00000009478,
       0.00000001396,     0.00000000327,
       0.00000122573,    -0.00000114657,
      -0.00000069074,    -0.00000057563,
      -0.00000013016,     0.00000020881,
       0.00000004177,     0.00000001772,
       0.00000000143,    -0.00000000599,
      -0.00000090568,     0.00000017043,
       0.00000010194,     0.00000034856,
       0.00000006876,    -0.00000002930,
      -0.00000000511,    -0.00000000886,
       0.00000022168,     0.00000089297,
       0.00000033854,    -0.00000013466,
      -0.00000003992,    -0.00000006483,
      -0.00000000764,     0.00000000773,
       0.00000001211,    -0.00000000495,
      -0.00000000233,    -0.00000000237,
       0.00000000218,    -0.00000001119,
      -0.00000000236,    -0.00000000050,
      -0.00001477675,    -0.00001671750,
      -0.00000782337,     0.00000900633,
       0.00000274095,     0.00000176687,
       0.00000024008,    -0.00000054938,
      -0.00000008108,    -0.00000001920,
      -0.00000285682,     0.00000269562,
       0.00000162089,     0.00000134245,
       0.00000030358,    -0.00000048906,
      -0.00000009737,    -0.00000004113,
      -0.00000000319,     0.00000001434,
       0.00000392120,    -0.00000075788,
      -0.00000046131,    -0.00000150601,
      -0.00000029588,     0.00000013432,
       0.00000002452,     0.00000003747,
       0.00000000337,    -0.00000000327,
      -0.00000088731,    -0.00000391021,
      -0.00000149246,     0.00000054557,
       0.00000016201,     0.00000028878,
       0.00000003507,    -0.00000003106,
      -0.00000000416,    -0.00000000292,
      -0.00000003414,     0.00000000805,
       0.00000000469,     0.00000000922,
       0.00000000212,    -0.00000000130,
       0.00000000040,     0.00000003262,
       0.00000000899,    -0.00000000162,
      -0.00000000069,    -0.00000000209,
      -0.00000000075,     0.00000000380,
       0.00000000262,     0.00000000024,
      -0.00000000403,    -0.00000000125,
      -0.00000000043,     0.00000000274,
      -0.00000000095,     0.00000000476,
       0.00000000328,     0.00000000031,
      -0.00000000003,    -0.00000000113,
      -0.00000000504,    -0.00000000156,
      -0.00000000053,     0.00000000342,
       0.00000000117,     0.00000000001,
       0.00000000029,     0.00000000026,
      -0.00000000029,     0.00000000025,
      -0.00000652659,     0.00000141761,
       0.00000086806,     0.00000249643,
       0.00000048545,    -0.00000025270,
      -0.00000004684,    -0.00000005995,
      -0.00000000508,     0.00000000618,
       0.00000039265,     0.00000242875,
       0.00000093523,    -0.00000028142,
      -0.00000008731,    -0.00000018288,
      -0.00000002279,     0.00000001662,
       0.00000000218,     0.00000000199,
      -0.00000079202,     0.00000164102,
       0.00000048678,     0.00000012370,
      -0.00000000286,    -0.00000007420,
      -0.00000000744,     0.00000000340,
       0.00000152620,     0.00000095416,
       0.00000021438,    -0.00000043310,
      -0.00000006209,    -0.00000002325,
      -0.00000000169,     0.00000000582,
      -0.00000000712,     0.00000002435,
       0.00000000180,    -0.00000000079,
       0.00000001720,     0.00000001304,
       0.00000000085,    -0.00000000041,
       0.00002100130,    -0.00000586337,
      -0.00000319032,    -0.00000791985,
      -0.00000153245,     0.00000087528,
       0.00000015712,     0.00000018747,
       0.00000001567,    -0.00000002023,
      -0.00000084682,    -0.00000520152,
      -0.00000200043,     0.00000060454,
       0.00000018696,     0.00000039035,
       0.00000004843,    -0.00000003574,
      -0.00000000480,    -0.00000000409,
       0.00000356924,    -0.00000688561,
      -0.00000201419,    -0.00000064130,
      -0.00000002109,     0.00000030199,
       0.00000002963,    -0.00000000774,
      -0.00000000141,    -0.00000000207,
      -0.00000667909,    -0.00000381863,
      -0.00000078495,     0.00000191717,
       0.00000028030,     0.00000006318,
       0.00000000046,    -0.00000002674,
      -0.00000000189,     0.00000000055,
       0.00000002686,    -0.00000005969,
      -0.00000000696,    -0.00000000172,
      -0.00000000056,     0.00000000170,
      -0.00000005421,    -0.00000003062,
      -0.00000000215,     0.00000000615,
       0.00000000174,     0.00000000032,
       0.00000000061,    -0.00000000075,
      -0.00000000932,    -0.00000001056,
      -0.00000000492,     0.00000000571,
       0.00000000174,     0.00000000111,
       0.00000001005,    -0.00000000974,
      -0.00000000581,    -0.00000000477,
      -0.00000000108,     0.00000000175,
      -0.00000001168,    -0.00000001324,
      -0.00000000616,     0.00000000716,
       0.00000000217,     0.00000000139,
       0.00000001258,    -0.00000001220,
      -0.00000000728,    -0.00000000597,
      -0.00000000136,     0.00000000219,
      -0.00000000047,    -0.00000000084,
      -0.00000000083,     0.00000000049,
       0.00000000004,     0.00000000003,
      -0.00000000003,    -0.00000000003,
       0.00000000004,    -0.00000000004,
       0.00000000029,     0.00000000038,
       0.00000000067,     0.00000000025,
       0.00000000006,     0.00000000047,
      -0.00000607742,     0.00001050456,
       0.00000306289,     0.00000119783,
       0.00000008330,    -0.00000045392,
      -0.00000004408,     0.00000000295,
       0.00000387402,     0.00000181762,
       0.00000036053,    -0.00000110140,
      -0.00000016084,    -0.00000002228,
       0.00000000169,     0.00000001542,
      -0.00000264927,    -0.00000400308,
      -0.00000046943,     0.00000050054,
       0.00000006173,     0.00000001590,
      -0.00000000147,    -0.00000000548,
      -0.00000401295,     0.00000217723,
       0.00000031737,     0.00000050373,
       0.00000002940,    -0.00000002552,
      -0.00000000090,    -0.00000000123,
      -0.00000000030,    -0.00000002110,
       0.00000000321,    -0.00000000534,
      -0.00000001349,     0.00000002469,
      -0.00000000108,    -0.00000000534,
       0.00003776006,    -0.00006850945,
      -0.00001990654,    -0.00000723823,
      -0.00000044363,     0.00000293600,
       0.00000028303,    -0.00000003099,
      -0.00000000907,    -0.00000001965,
      -0.00000958476,    -0.00000446509,
      -0.00000088354,     0.00000272410,
       0.00000039812,     0.00000005353,
      -0.00000000431,    -0.00000003818,
      -0.00000000264,     0.00000000111,
       0.00001056299,     0.00001739041,
       0.00000210672,    -0.00000180838,
      -0.00000019692,    -0.00000009513,
       0.00000000120,     0.00000001465,
       0.00001732281,    -0.00001028743,
      -0.00000170084,    -0.00000210729,
      -0.00000010062,     0.00000017546,
       0.00000001205,     0.00000000017,
       0.00000000738,     0.00000003277,
      -0.00000001439,     0.00000001497,
       0.00000000165,     0.00000000107,
       0.00000002436,    -0.00000003557,
       0.00000000880,     0.00000001742,
       0.00000000183,    -0.00000000047,
      -0.00000000039,    -0.00000000158,
      -0.00000002276,     0.00000000398,
       0.00000000250,     0.00000000888,
       0.00000000176,    -0.00000000075,
      -0.00000000287,    -0.00000002345,
      -0.00000000900,     0.00000000236,
       0.00000000075,     0.00000000177,
      -0.00000002851,     0.00000000500,
       0.00000000313,     0.00000001112,
       0.00000000220,    -0.00000000094,
      -0.00000000359,    -0.00000002938,
      -0.00000001128,     0.00000000296,
       0.00000000094,     0.00000000222,
      -0.00000000480,    -0.00000000825,
      -0.00000000109,     0.00000000068,
       0.00000000125,     0.00000000004,
      -0.00000000006,    -0.00000000128,
       0.00000001250,     0.00000001950,
       0.00000000093,    -0.00000000057,
      -0.00000000004,     0.00000000002,
      -0.00000000005,    -0.00000000005,
       0.00000000008,    -0.00000000002,
      -0.00000000036,     0.00000000022,
       0.00000000006,    -0.00000000051,
       0.00000000052,     0.00000000053,
      -0.00001964197,    -0.00003267275,
      -0.00000393043,     0.00000337328,
       0.00000034895,     0.00000018668,
      -0.00000000033,    -0.00000002447,
      -0.00001039003,     0.00000674015,
       0.00000109432,     0.00000128369,
       0.00000006100,    -0.00000011157,
      -0.00000000791,    -0.00000000004,
       0.00000686821,     0.00000033197,
       0.00000013578,     0.00000002832,
       0.00000000108,     0.00000000208,
      -0.00000000181,     0.00000000040,
      -0.00000070995,    -0.00000711771,
      -0.00000002173,     0.00000014766,
       0.00000004166,     0.00000000003,
      -0.00000000040,    -0.00000000188,
      -0.00000002689,     0.00000003333,
       0.00000000088,     0.00000000315,
      -0.00000003251,     0.00000000383,
       0.00000000270,     0.00000000131,
       0.00010801139,     0.00017845876,
       0.00002128599,    -0.00001842993,
      -0.00000188200,    -0.00000097611,
       0.00000000649,     0.00000012888,
       0.00000000552,    -0.00000000410,
       0.00002491388,    -0.00001620283,
      -0.00000263432,    -0.00000309679,
      -0.00000015007,     0.00000027104,
       0.00000001949,     0.00000000052,
      -0.00003067274,    -0.00000026830,
      -0.00000024654,    -0.00000006738,
      -0.00000000442,    -0.00000005961,
       0.00000000341,    -0.00000000046,
       0.00000161557,     0.00003120162,
       0.00000004298,    -0.00000026630,
      -0.00000013185,    -0.00000000070,
       0.00000000054,     0.00000000342,
       0.00000010199,    -0.00000003224,
      -0.00000000190,    -0.00000002149,
      -0.00000000332,     0.00000000097,
       0.00000003277,    -0.00000007482,
      -0.00000001937,    -0.00000000225,
      -0.00000000010,     0.00000000293,
      -0.00000000074,    -0.00000000086,
      -0.00000000520,     0.00000000828,
       0.00000000235,     0.00000000093,
      -0.00000000810,    -0.00000000405,
      -0.00000000101,     0.00000000237,
      -0.00000000663,     0.00000001032,
       0.00000000295,     0.00000000119,
      -0.00000001009,    -0.00000000479,
      -0.00000000120,     0.00000000293,
      -0.00000000945,     0.00000000031,
      -0.00000000999,     0.00000000032,
       0.00000001263,    -0.00000000028,
      -0.00000000003,    -0.00000000178,
      -0.00000002647,     0.00000000396,
       0.00000000016,     0.00000000121,
       0.00000000087,    -0.00000000180,
      -0.00000000044,    -0.00000000006,
       0.00000000437,    -0.00000000456,
      -0.00000000128,    -0.00000000068,
       0.00000000450,     0.00000000442,
       0.00000000069,    -0.00000000128,
       0.00000000004,    -0.00000000030,
      -0.00000000008,     0.00000000003,
       0.00000000036,     0.00000000002,
      -0.00000000004,    -0.00000000009,
      -0.00000000016,    -0.00000000037,
       0.00000000054,     0.00000000063,
      -0.00000000063,     0.00000000054,
       0.00000000020,    -0.00000000032,
       0.00000000032,     0.00000000019,
      -0.00000000031,    -0.00000000047,
       0.00000000232,     0.00000000018,
      -0.00000000014,    -0.00000000045,
      -0.00000000019,     0.00000000232,
       0.00000000045,    -0.00000000014,
      -0.00000000150,     0.00000000285,
      -0.00000002088,    -0.00000002143,
      -0.00000000235,     0.00000000469,
       0.00000002147,    -0.00000002087,
      -0.00000000469,    -0.00000000235,
      -0.00000000008,    -0.00000000100,
       0.00000000099,    -0.00000000005,
       0.00007012767,    -0.00000242693,
      -0.00000016161,     0.00000004569,
       0.00000000789,     0.00000019681,
      -0.00000043091,    -0.00002129496,
       0.00000000842,     0.00000003252,
       0.00000005779,    -0.00000000124,
      -0.00000030594,    -0.00000203009,
      -0.00000010587,     0.00000026257,
       0.00000001311,     0.00000001812,
       0.00000107018,    -0.00000211906,
      -0.00000009424,    -0.00000029396,
      -0.00000000086,     0.00000002115,
      -0.00000001800,    -0.00000005636,
       0.00000000535,    -0.00000000803,
       0.00000002521,    -0.00000004249,
       0.00000000451,     0.00000001001,
      -0.00032177872,     0.00001093093,
       0.00000078711,    -0.00000020469,
      -0.00000003717,    -0.00000090196,
       0.00000107802,     0.00004899910,
      -0.00000001955,    -0.00000008348,
      -0.00000013374,     0.00000000287,
      -0.00000177743,     0.00000427778,
      -0.00000009749,    -0.00000060744,
      -0.00000000906,    -0.00000004667,
       0.00000000253,     0.00000000175,
      -0.00000031845,     0.00000544682,
      -0.00000006601,     0.00000066589,
      -0.00000002387,    -0.00000004116,
      -0.00000000150,    -0.00000000207,
      -0.00000006588,     0.00000014058,
       0.00000000556,     0.00000001937,
       0.00000008202,     0.00000010230,
       0.00000000237,    -0.00000002478,
      -0.00000000166,     0.00000000029,
       0.00000000846,    -0.00000000055,
      -0.00000000050,    -0.00000000190,
       0.00000000755,    -0.00000012209,
      -0.00000002985,     0.00000000293,
       0.00000000162,     0.00000000416,
       0.00000012217,     0.00000000794,
      -0.00000000294,    -0.00000002984,
      -0.00000000416,     0.00000000162,
      -0.00000000129,    -0.00000000164,
       0.00000000949,    -0.00000015282,
      -0.00000003736,     0.00000000364,
       0.00000000203,     0.00000000521,
       0.00000015296,     0.00000000999,
      -0.00000000366,    -0.00000003735,
      -0.00000000521,     0.00000000203,
       0.00000000275,    -0.00000000072,
      -0.00000000192,     0.00000000049,
      -0.00000002707,    -0.00000001482,
       0.00000000041,    -0.00000000071,
       0.00000022051,     0.00000012535,
       0.00000000007,    -0.00000000090,
      -0.00000000061,    -0.00000000034,
       0.00000000307,    -0.00000000030,
      -0.00000000028,    -0.00000000134,
      -0.00000000030,    -0.00000000304,
      -0.00000000133,     0.00000000028,
       0.00000007265,    -0.00000000708,
      -0.00000000657,    -0.00000003171,
      -0.00000000045,    -0.00000000397,
      -0.00000000174,     0.00000000038,
      -0.00000001225,     0.00000000175,
       0.00000000132,     0.00000000504,
       0.00000000106,    -0.00000000042,
       0.00000000173,     0.00000001208,
       0.00000000497,    -0.00000000130,
      -0.00000000041,    -0.00000000104,
       0.00000000498,    -0.00000000064,
      -0.00000000023,    -0.00000000161,
      -0.00000028312,     0.00000003952,
       0.00000003040,     0.00000011660,
       0.00000002439,    -0.00000000968,
       0.00000000257,     0.00000001738,
       0.00000000715,    -0.00000000191,
      -0.00000000061,    -0.00000000150,
       0.00000002971,    -0.00000000621,
      -0.00000000388,    -0.00000001144,
      -0.00000000223,     0.00000000109,
      -0.00000000608,    -0.00000002885,
      -0.00000001106,     0.00000000379,
       0.00000000105,     0.00000000215,
       0.00000000022,     0.00000000111,
      -0.00000002780,     0.00000001113,
       0.00000000545,     0.00000000773,
       0.00000000368,     0.00000000875,
       0.00000000242,    -0.00000000173,
       0.00000000030,    -0.00000000254,
      -0.00000000262,     0.00000000016,
       0.00000006894,    -0.00000004429,
      -0.00000001912,    -0.00000001018,
      -0.00000000659,    -0.00000000967,
      -0.00000000141,     0.00000000272,
      -0.00000000069,     0.00000001140,
       0.00000000111,    -0.00000000102,
       0.00000001158,    -0.00000000014,
      -0.00000000134,    -0.00000000106,
       0.00000066339,    -0.00000012953,
      -0.00000008360,    -0.00000025539,
      -0.00000004985,     0.00000002419,
      -0.00000000922,    -0.00000004785,
      -0.00000001844,     0.00000000602,
       0.00000000176,     0.00000000359,
       0.00000000172,     0.00000000688,
       0.00000000184,    -0.00000000100,
       0.00000000626,    -0.00000000550,
      -0.00000000257,    -0.00000000141,
      -0.00000000107,    -0.00000000042,
      -0.00000000202,     0.00000000032,
       0.00000000024,     0.00000000078,
      -0.00000000036,    -0.00000000201,
      -0.00000000077,     0.00000000024,
      -0.00000001757,     0.00000000690,
       0.00000000121,     0.00000000242,
       0.00000000015,    -0.00000000318,
      -0.00000000326,    -0.00000000001,
      -0.00000003657,     0.00000011175,
       0.00000002776,     0.00000000578,
       0.00000000217,    -0.00000000596,
      -0.00000000218,    -0.00000000025,
       0.00000000816,    -0.00000004520,
      -0.00000001145,    -0.00000000101,
       0.00000000018,     0.00000000170,
      -0.00000004325,     0.00000000222,
       0.00000000300,     0.00000001017,
       0.00000000132,    -0.00000000097,
      -0.00000000003,     0.00000002546,
       0.00000000234,     0.00000001230,
       0.00000000174,    -0.00000000038,
      -0.00000000157,     0.00000000005,
       0.00000000012,     0.00000000168,
       0.00000021586,    -0.00000327138,
      -0.00000080041,     0.00000007679,
       0.00000004329,     0.00000011179,
       0.00000015875,     0.00000001016,
      -0.00000000379,    -0.00000003879,
      -0.00000000541,     0.00000000210,
      -0.00000000086,    -0.00000000294,
      -0.00000000068,     0.00000000044,
      -0.00000000299,    -0.00000000056,
       0.00000000062,    -0.00000003330,
       0.00000000103,     0.00000000002,
      -0.00000000066,    -0.00000000047,
      -0.00000000098,    -0.00000000005,
       0.00000000043,    -0.00000000103,
       0.00000000112,     0.00000000042,
      -0.00000000277,     0.00000001593,
       0.00000000235,     0.00000000016,
      -0.00000000051,     0.00000000102,
      -0.00000000148,    -0.00000000054,
       0.00000000359,    -0.00000000060,
      -0.00000000042,    -0.00000000139,
      -0.00000000060,    -0.00000000359,
      -0.00000000139,     0.00000000042,
       0.00000011817,    -0.00000002187,
      -0.00000001454,    -0.00000004598,
      -0.00000000096,    -0.00000000570,
      -0.00000000222,     0.00000000068,
      -0.00000000370,     0.00000003060,
      -0.00000000101,    -0.00000000012,
       0.00000041133,    -0.00000628173,
      -0.00000153682,     0.00000014850,
       0.00000008338,     0.00000021457,
       0.00000002009,    -0.00000001601,
       0.00000020126,     0.00000001285,
      -0.00000000480,    -0.00000004918,
      -0.00000000686,     0.00000000267,
       0.00000000051,     0.00000000064,
      -0.00000000203,    -0.00000000179,
      -0.00000000048,     0.00000000078,
       0.00000000264,    -0.00000000094,
      -0.00000000055,    -0.00000000072,
       0.00000000010,    -0.00000000018,
       0.00000000047,    -0.00000000041,
      -0.00000000005,    -0.00000000004,
       0.00000000040,     0.00000000050,
       0.00000000004,    -0.00000000005,
       0.00000000001,    -0.00000000005,
       0.00000000005,     0.00000000001,
      -0.00000000226,    -0.00000000116,
       0.00000000116,    -0.00000000224,
       0.00001060151,    -0.00000307072,
       0.00000082724,     0.00000082671,
      -0.00000004353,     0.00000008136,
      -0.00000000486,    -0.00000000082,
      -0.00000095634,    -0.00000244066,
       0.00000020821,    -0.00000021702,
       0.00000001990,     0.00000001072,
       0.00000442428,    -0.00000030635,
       0.00000003947,     0.00000000250,
       0.00000000034,     0.00000001835,
      -0.00000009066,     0.00000444486,
      -0.00000000228,    -0.00000006802,
      -0.00000000258,     0.00000000046,
      -0.00000002606,    -0.00000000045,
       0.00000000127,     0.00000000451,
      -0.00000000049,    -0.00000002602,
      -0.00000000473,     0.00000000087,
      -0.00003759289,     0.00000981086,
      -0.00000289863,    -0.00000277349,
       0.00000014054,    -0.00000029303,
       0.00000001799,     0.00000000209,
       0.00000227169,     0.00000575019,
      -0.00000049142,     0.00000051464,
      -0.00000004703,    -0.00000002521,
       0.00000000050,    -0.00000000298,
      -0.00000529581,     0.00000096872,
      -0.00000020929,     0.00000000189,
      -0.00000000062,    -0.00000004556,
       0.00000000276,     0.00000000021,
       0.00000068800,    -0.00000546440,
       0.00000000955,     0.00000024480,
      -0.00000001997,    -0.00000000064,
       0.00000000007,    -0.00000000284,
       0.00000005096,     0.00000000615,
      -0.00000000218,    -0.00000001312,
      -0.00000000188,     0.00000000021,
       0.00000001884,     0.00000005687,
       0.00000001479,    -0.00000000144,
      -0.00000000052,    -0.00000000182,
       0.00000000088,     0.00000000105,
       0.00000001658,    -0.00000001066,
      -0.00000000111,    -0.00000000001,
       0.00000001068,     0.00000001662,
       0.00000000001,    -0.00000000110,
       0.00000002077,    -0.00000001335,
      -0.00000000139,    -0.00000000002,
       0.00000001338,     0.00000002083,
       0.00000000001,    -0.00000000138,
       0.00000000000,     0.00000000029,
      -0.00000000028,     0.00000000096,
       0.00000000003,     0.00000000003,
      -0.00000000096,    -0.00000000028,
      -0.00000000003,     0.00000000003,
      -0.00000000003,     0.00000000003,
      -0.00000000038,     0.00000000042,
      -0.00000000043,    -0.00000000038,
      -0.00000000033,    -0.00000000026,
       0.00000000260,     0.00000000020,
      -0.00000000021,     0.00000000261,
       0.00000027805,    -0.00000157235,
       0.00000046925,     0.00000004768,
      -0.00000000500,     0.00000006947,
      -0.00000000591,    -0.00000000074,
      -0.00000044828,     0.00000002692,
      -0.00000000109,    -0.00000012312,
       0.00000001831,     0.00000000001,
       0.00000079534,    -0.00000011559,
       0.00000003702,     0.00000009120,
      -0.00000001108,     0.00000000211,
       0.00000011644,     0.00000058011,
      -0.00000000779,     0.00000005091,
      -0.00000000658,     0.00000000546,
       0.00000002608,    -0.00000003735,
      -0.00000000532,    -0.00000000368,
       0.00000003710,     0.00000002718,
       0.00000000352,    -0.00000000540,
       0.00000069449,     0.00000723554,
      -0.00000212335,     0.00000004513,
       0.00000000885,    -0.00000030799,
       0.00000002675,     0.00000000317,
       0.00000107516,    -0.00000004724,
       0.00000000043,     0.00000029744,
      -0.00000004431,    -0.00000000003,
       0.00000000024,    -0.00000000372,
      -0.00000112197,     0.00000007768,
      -0.00000000460,    -0.00000023014,
       0.00000003961,     0.00000000643,
      -0.00000000155,     0.00000000494,
      -0.00000010369,    -0.00000015991,
      -0.00000014380,    -0.00000007280,
       0.00000001476,    -0.00000003462,
       0.00000000487,     0.00000000198,
      -0.00000006403,     0.00000009447,
       0.00000001239,     0.00000000867,
      -0.00000009166,    -0.00000007023,
      -0.00000000776,     0.00000001299,
      -0.00000000041,     0.00000000086,
       0.00000001153,     0.00000000730,
       0.00000000066,     0.00000000079,
      -0.00000000731,     0.00000001155,
      -0.00000000079,     0.00000000066,
       0.00000001444,     0.00000000914,
       0.00000000082,     0.00000000099,
      -0.00000000914,     0.00000001446,
      -0.00000000099,     0.00000000082,
       0.00000000018,     0.00000000040,
       0.00000000020,     0.00000000137,
      -0.00000000006,     0.00000000003,
      -0.00000000137,     0.00000000019,
      -0.00000000003,    -0.00000000006,
       0.00000000007,     0.00000000006,
      -0.00000000007,     0.00000000009,
      -0.00000000047,     0.00000000050,
      -0.00000000050,    -0.00000000047,
      -0.00000000021,    -0.00000000065,
       0.00000000411,     0.00000000028,
      -0.00000000029,     0.00000000411,
       0.00000110560,     0.00000000421,
      -0.00000006742,     0.00000042499,
      -0.00000008380,    -0.00000002258,
       0.00000000400,    -0.00000001096,
       0.00000001094,    -0.00000029413,
       0.00000011369,     0.00000002245,
      -0.00000000694,     0.00000002251,
      -0.00000000294,    -0.00000000118,
       0.00000004133,    -0.00000008206,
       0.00000002827,     0.00000001361,
      -0.00000000320,     0.00000000452,
       0.00000006748,    -0.00000000467,
       0.00000000973,     0.00000002126,
      -0.00000000276,     0.00000000293,
       0.00000000453,    -0.00000000207,
       0.00000000164,     0.00000000413,
      -0.00000492605,    -0.00000006200,
       0.00000033458,    -0.00000192764,
       0.00000038256,     0.00000011003,
      -0.00000001907,     0.00000005029,
      -0.00000001882,     0.00000068320,
      -0.00000026398,    -0.00000005033,
       0.00000001575,    -0.00000005230,
       0.00000000686,     0.00000000269,
      -0.00000007312,     0.00000013619,
      -0.00000004666,    -0.00000004399,
       0.00000001252,    -0.00000000804,
       0.00000000105,     0.00000000231,
      -0.00000006337,     0.00000011238,
      -0.00000005122,    -0.00000001641,
       0.00000000062,    -0.00000001282,
       0.00000000228,    -0.00000000025,
      -0.00000001285,     0.00000000492,
      -0.00000000173,    -0.00000001005,
      -0.00000000184,     0.00000000046,
       0.00000001115,     0.00000000613,
       0.00000000038,     0.00000000130,
      -0.00000000613,     0.00000001115,
      -0.00000000130,     0.00000000038,
       0.00000001395,     0.00000000767,
       0.00000000047,     0.00000000163,
      -0.00000000768,     0.00000001396,
      -0.00000000163,     0.00000000047,
      -0.00000000144,    -0.00000000005,
      -0.00000000006,     0.00000000155,
       0.00000024604,    -0.00000004355,
       0.00000001999,     0.00000012027,
      -0.00000003071,     0.00000000517,
      -0.00000000093,    -0.00000000532,
      -0.00000001290,    -0.00000005569,
       0.00000002777,    -0.00000000614,
       0.00000000156,     0.00000000720,
       0.00000006955,    -0.00000000346,
      -0.00000000246,     0.00000002770,
      -0.00000000569,    -0.00000000080,
      -0.00000000461,     0.00000006675,
      -0.00000002559,    -0.00000000778,
       0.00000000258,    -0.00000000489,
      -0.00000080898,     0.00000034185,
      -0.00000007929,    -0.00000040923,
       0.00000010920,    -0.00000001623,
       0.00000000302,     0.00000001947,
       0.00000001251,     0.00000014418,
      -0.00000007032,     0.00000001367,
      -0.00000000393,    -0.00000001769,
       0.00000000303,    -0.00000000070,
      -0.00000008952,     0.00000001359,
      -0.00000000478,    -0.00000003740,
       0.00000000816,    -0.00000000164,
       0.00000002000,    -0.00000007998,
       0.00000002946,     0.00000001769,
      -0.00000000595,     0.00000000502,
       0.00000000151,     0.00000000026,
       0.00000000038,     0.00000000166,
      -0.00000000184,    -0.00000000026,
      -0.00000000025,     0.00000000182,
      -0.00000004611,    -0.00000000654,
      -0.00000000026,     0.00000000217,
       0.00000000870,     0.00000000090,
       0.00000000088,    -0.00000000859,
       0.00000021598,     0.00000002260,
       0.00000000093,    -0.00000001074,
      -0.00000003428,    -0.00000000220,
      -0.00000000015,     0.00000000163,
      -0.00000000215,     0.00000003367,
       0.00000000157,     0.00000000014,
       0.00000001396,     0.00000000091,
       0.00000000021,    -0.00000000394,
      -0.00000000277,    -0.00000000007,
      -0.00000000003,     0.00000000211,
      -0.00000084082,    -0.00000005472,
      -0.00000000347,     0.00000003957,
      -0.00000000239,     0.00000004460,
       0.00000000214,     0.00000000022,
       0.00000010059,     0.00000000214,
      -0.00000000004,    -0.00000000255,
       0.00000000194,    -0.00000009743,
      -0.00000000233,     0.00000000007,
      -0.00000004685,    -0.00000000091,
      -0.00000000023,     0.00000001284,
       0.00000000530,    -0.00000000011,
      -0.00000000023,    -0.00000000207,
       0.00000242097,     0.00000004921,
      -0.00000000182,    -0.00000005964,
       0.00000000230,    -0.00000013968,
      -0.00000000348,     0.00000000000,
      -0.00000016078,     0.00000000609,
       0.00000000075,     0.00000000086,
       0.00000000662,     0.00000014741,
       0.00000000021,    -0.00000000083,
       0.00000008612,    -0.00000000376,
      -0.00000000077,    -0.00000002316,
      -0.00000000428,     0.00000000021,
       0.00000000033,    -0.00000000803,
      -0.00000371797,     0.00000016351,
       0.00000001958,     0.00000001360,
       0.00000000873,     0.00000024025,
       0.00000000073,    -0.00000000117,
      -0.00000004988,    -0.00000001010,
      -0.00000000110,    -0.00000000316,
      -0.00000001055,     0.00000009257,
      -0.00000000208,     0.00000000112,
       0.00000004480,     0.00000000752,
       0.00000000158,    -0.00000000994,
       0.00000001392,    -0.00000000030,
       0.00000000055,     0.00000001321,
      -0.00000155886,    -0.00000026235,
      -0.00000002726,    -0.00000006010,
      -0.00000001690,     0.00000010737,
      -0.00000000406,     0.00000000176,
      -0.00000005010,    -0.00000000294,
      -0.00000000010,    -0.00000000096,
      -0.00000000540,    -0.00000002551,
      -0.00000000116,     0.00000000046,
       0.00000000568,     0.00000000429,
       0.00000000081,    -0.00000000080,
       0.00000000640,     0.00000000121,
      -0.00000000117,     0.00000000633,
       0.00000001414,    -0.00000000286,
      -0.00000000052,    -0.00000000191,
       0.00000000085,     0.00000000052,
       0.00000000028,     0.00000000093,
      -0.00000024010,    -0.00000009882,
      -0.00000000563,    -0.00000002281,
      -0.00000000677,     0.00000001816,
      -0.00000000163,     0.00000000042,
      -0.00000001517,    -0.00000000378,
      -0.00000000023,    -0.00000000083,
       0.00000000141,    -0.00000001251,
      -0.00000000247,     0.00000000043,
       0.00000000062,     0.00000000196,
       0.00000000081,    -0.00000000040,
       0.00000000043,     0.00000000247,
       0.00000000195,    -0.00000000062,
      -0.00000000040,    -0.00000000081,
      -0.00000000063,     0.00000000075,
       0.00000000072,     0.00000000064,
       0.00000000076,     0.00000000067,
      -0.00000001225,     0.00000000973,
       0.00000000090,     0.00000000038,
       0.00000000038,    -0.00000000090,
      -0.00000000381,    -0.00000000169,
      -0.00000000006,     0.00000000167,
      -0.00000000172,     0.00000000382,
       0.00000000163,     0.00000000006,
       0.00000000518,     0.00000000237,
       0.00000000071,    -0.00000000193,
      -0.00000000182,     0.00000000109,
       0.00000000114,     0.00000000178,
      -0.00000003297,    -0.00000001551,
      -0.00000000076,     0.00000001441,
      -0.00000000178,     0.00000000465,
       0.00000000194,     0.00000000002,
       0.00000000791,    -0.00000000473,
      -0.00000000206,    -0.00000000125,
      -0.00000000486,    -0.00000000783,
      -0.00000000118,     0.00000000213,
      -0.00000000023,    -0.00000002214,
      -0.00000000675,    -0.00000000001,
       0.00000000071,     0.00000000061,
      -0.00000000112,     0.00000000024,
       0.00000000287,     0.00000010079,
       0.00000000871,    -0.00000001148,
       0.00000001554,     0.00000000008,
      -0.00000000169,    -0.00000000141,
       0.00000000331,     0.00000000014,
       0.00000000144,    -0.00000000176,
      -0.00000000466,    -0.00000000122,
      -0.00000000040,     0.00000000139,
       0.00000000215,    -0.00000000024,
       0.00000001385,     0.00000001161,
       0.00000000104,    -0.00000000320,
      -0.00000000097,    -0.00000000312,
      -0.00000000590,     0.00000000110,
       0.00000000077,    -0.00000000050,
      -0.00000001041,    -0.00000001575,
      -0.00000000111,     0.00000000069,
       0.00000000312,     0.00000000857,
       0.00000000239,    -0.00000000062,
      -0.00000000142,    -0.00000000024,
      -0.00000001323,    -0.00000004330,
       0.00000000601,     0.00000000600,
      -0.00000000574,     0.00000000147,
       0.00000000075,    -0.00000000080,
       0.00000000534,     0.00000000092,
      -0.00000000083,     0.00000000103,
       0.00000000054,    -0.00000000499,
       0.00000000111,     0.00000000087,
      -0.00000003521,    -0.00000000160,
      -0.00000000012,     0.00000000247,
      -0.00000001166,    -0.00000000174,
      -0.00000000046,     0.00000000310,
       0.00000005223,     0.00000000628,
      -0.00000000842,     0.00000001089,
       0.00000000103,    -0.00000000724,
       0.00000000149,     0.00000000119,
       0.00000000119,    -0.00000000108,
       0.00000000016,    -0.00000000167,
       0.00000000092,    -0.00000000035,
       0.00000000114,     0.00000000050,
      -0.00000000050,     0.00000000114,
       0.00000003080,     0.00000001690,
       0.00000000291,     0.00000000647,
      -0.00000000051,     0.00000000151,
       0.00000000804,    -0.00000000641,
       0.00000000186,     0.00000000042,
      -0.00000012263,     0.00000002356,
       0.00000003231,     0.00000009649,
       0.00000003737,    -0.00000001889,
      -0.00000000050,    -0.00000000173,
      -0.00000000137,     0.00000000059,
       0.00000003372,    -0.00000001149,
      -0.00000001256,    -0.00000002556,
      -0.00000000962,     0.00000000642,
       0.00000000211,     0.00000000236,
       0.00000000046,    -0.00000000058,
      -0.00000001150,    -0.00000003369,
      -0.00000002552,     0.00000001256,
       0.00000000641,     0.00000000961,
       0.00000000235,    -0.00000000210,
      -0.00000000058,    -0.00000000046,
      -0.00000001756,     0.00000000716,
       0.00000000745,     0.00000001285,
      -0.00000000041,     0.00000000158,
       0.00000000112,     0.00000000014,
       0.00000000159,     0.00000000040,
       0.00000000013,    -0.00000000112,
       0.00000001097,     0.00000001466,
      -0.00000000367,    -0.00000000397,
      -0.00000000187,     0.00000000222,
       0.00000000068,     0.00000000043,
      -0.00000000396,     0.00000000366,
       0.00000000221,     0.00000000186,
       0.00000000043,    -0.00000000068,
       0.00000000073,     0.00000000059,
       0.00000000058,    -0.00000000073,
       0.00000003244,    -0.00000000165,
      -0.00000000868,     0.00000000129,
       0.00000000091,     0.00000000336,
       0.00000000128,     0.00000000867,
       0.00000000336,    -0.00000000091,
       0.00000000601,     0.00000000526,
       0.00000000060,    -0.00000000092,
      -0.00000000209,     0.00000000089,
       0.00000000092,     0.00000000209,
       0.00000001355,    -0.00000001361,
      -0.00000000233,     0.00000000335,
       0.00000000101,     0.00000000053,
       0.00000000333,     0.00000000231,
       0.00000000052,    -0.00000000101,
      -0.00000000021,     0.00000000555,
       0.00000000175,     0.00000000004,
       0.00000000204,    -0.00000002456,
      -0.00000000407,     0.00000000002,
      -0.00000000043,    -0.00000000088,
      -0.00000002386,     0.00000000302,
       0.00000000596,    -0.00000000380,
      -0.00000000386,    -0.00000000601,
      -0.00000000037,    -0.00000000116,
       0.00000000402,     0.00000000247,
       0.00000000273,    -0.00000000418,
       0.00000000366,     0.00000000215,
       0.00000000218,    -0.00000000422,
       0.00000000225,    -0.00000000536,
       0.00000011578,     0.00000004861,
      -0.00000002363,     0.00000005110,
      -0.00000024157,    -0.00000011172,
      -0.00000002048,    -0.00000002826,
      -0.00000002635,     0.00000002009,
      -0.00000000065,     0.00000000051,
       0.00000000102,    -0.00000000001,
       0.00000000001,    -0.00000000106,
       0.00000000470,     0.00000000763,
       0.00000000243,    -0.00000000172,
      -0.00000000436,    -0.00000000021,
       0.00000000015,     0.00000000451,
      -0.00000000005,    -0.00000000167,
      -0.00000000076,     0.00000000117,
       0.00000000115,     0.00000000076,
      -0.00000001974,    -0.00000000009,
       0.00000000014,     0.00000000417,
       0.00000000335,    -0.00000000486,
      -0.00000000488,    -0.00000000330,
      -0.00000000376,     0.00000000565,
       0.00000000265,     0.00000000177,
       0.00000001756,    -0.00000002482,
      -0.00000000579,    -0.00000000389,
      -0.00000000709,     0.00000000655,
       0.00000035261,     0.00000054717,
      -0.00000003511,     0.00000012705,
      -0.00000001731,     0.00000000451,
      -0.00000002721,     0.00000001752,
      -0.00000000632,    -0.00000000175,
      -0.00000000022,    -0.00000000086,
       0.00000000094,     0.00000000089,
       0.00000232755,    -0.00000079003,
      -0.00000086403,    -0.00000176293,
      -0.00000066362,     0.00000044168,
       0.00000014504,     0.00000016223,
       0.00000002889,    -0.00000003579,
       0.00000003333,     0.00000009809,
       0.00000007430,    -0.00000003643,
      -0.00000001864,    -0.00000002796,
      -0.00000000685,     0.00000000612,
       0.00000000152,     0.00000000123,
       0.00000000128,    -0.00000000075,
      -0.00000000077,    -0.00000000089,
      -0.00000000103,    -0.00000000269,
      -0.00000000200,     0.00000000110,
       0.00000000059,     0.00000000076,
       0.00000002211,    -0.00000000883,
      -0.00000002903,     0.00000011470,
       0.00000007843,     0.00000000963,
      -0.00000000024,    -0.00000002684,
      -0.00000000495,    -0.00000000126,
      -0.00000000042,     0.00000000339,
       0.00000000119,    -0.00000000002,
      -0.00000026626,    -0.00000028744,
      -0.00000013496,     0.00000016093,
       0.00000004812,     0.00000003040,
       0.00000001319,    -0.00000001219,
      -0.00000000737,    -0.00000000619,
      -0.00000000140,     0.00000000223,
      -0.00000063041,     0.00000009313,
       0.00000006640,     0.00000024454,
       0.00000004824,    -0.00000002047,
      -0.00000000441,    -0.00000002976,
      -0.00000001151,     0.00000000313,
       0.00000000096,     0.00000000226,
      -0.00000000042,    -0.00000000074,
       0.00000000028,     0.00000000145,
       0.00000000147,    -0.00000000031,
      -0.00000013418,     0.00000022343,
       0.00000006509,     0.00000002906,
      -0.00000001009,    -0.00000000513,
      -0.00000000102,     0.00000000286,
      -0.00000001256,     0.00000000079,
       0.00000000009,     0.00000000196,
       0.00000043304,    -0.00000028404,
      -0.00000002957,    -0.00000000033,
       0.00000001391,     0.00000002162,
       0.00000000002,    -0.00000000144,
       0.00000030597,     0.00000019295,
       0.00000001746,     0.00000002100,
      -0.00000000954,     0.00000001507,
      -0.00000000103,     0.00000000085,
       0.00000029543,     0.00000016249,
       0.00000000995,     0.00000003452,
      -0.00000000804,     0.00000001458,
      -0.00000000170,     0.00000000049,
       0.00000001476,    -0.00000000811,
      -0.00000001195,    -0.00000001609,
      -0.00000000071,     0.00000000053,
      -0.00000000115,     0.00000000022,
       0.00000000036,     0.00000000101,
       0.00000000115,    -0.00000000072,
       0.00000000052,    -0.00000000061,
       0.00000001079,     0.00000001248,
      -0.00000000094,     0.00000000081,
       0.00000001498,    -0.00000005550,
      -0.00000003804,    -0.00000000511,
       0.00000000039,    -0.00000000166,
      -0.00000000102,    -0.00000000040,
      -0.00000000167,    -0.00000000018,
      -0.00000000022,     0.00000000095,
      -0.00000000085,     0.00000000158,
       0.00000000157,     0.00000000084,
      -0.00000000081,    -0.00000000038,
       0.00000000051,     0.00000000082,
       0.00000000079,    -0.00000000052,
      -0.00000000803,     0.00000001482,
       0.00000000198,     0.00000000095,
      -0.00000000210,    -0.00000000346,
      -0.00000000341,     0.00000000215,
       0.00000000336,     0.00000000549,
       0.00000000182,    -0.00000000122,
      -0.00000000119,    -0.00000000001,
       0.00000000020,     0.00000000124,
      -0.00000002075,    -0.00000003307,
      -0.00000000441,     0.00000000294,
       0.00000000523,    -0.00000000020,
      -0.00000000057,    -0.00000000533,
      -0.00000000997,     0.00000000079,
       0.00000000021,     0.00000000314,
       0.00000005087,    -0.00000000392,
      -0.00000000049,    -0.00000000734,
       0.00000000028,    -0.00000000092,
       0.00000000537,     0.00000001660,
       0.00000001225,    -0.00000000768,
       0.00000001750,    -0.00000000593,
      -0.00000000725,    -0.00000001228,
      -0.00000000750,    -0.00000002262,
      -0.00000001633,     0.00000000963,
      -0.00000000100,     0.00000000649,
       0.00000000641,     0.00000000109,
      -0.00000002857,     0.00000001004,
       0.00000001046,     0.00000002177,
       0.00000000099,     0.00000000399,
       0.00000000305,    -0.00000000118,
      -0.00000000062,    -0.00000000109,
      -0.00000000179,     0.00000000805,
       0.00000000205,     0.00000000041,
       0.00000000008,    -0.00000000149,
      -0.00000000161,    -0.00000000151,
      -0.00000000063,     0.00000000086,
      -0.00000000148,     0.00000000163,
       0.00000000084,     0.00000000068,
       0.00000000063,    -0.00000000068,
      -0.00000000984,    -0.00000000934,
      -0.00000000417,     0.00000000578,
      -0.00000000166,     0.00000000180,
       0.00000000111,     0.00000000074,
       0.00000000131,    -0.00000000058,
      -0.00000000060,    -0.00000000131,
      -0.00000031248,     0.00000022040,
       0.00000018463,     0.00000015181,
       0.00000002170,    -0.00000004611,
      -0.00000000892,    -0.00000001293,
      -0.00000000626,     0.00000000752,
       0.00000000187,     0.00000000086,
      -0.00000000049,     0.00000000090,
       0.00000000088,     0.00000000047,
       0.00000000730,    -0.00000000397,
      -0.00000000054,    -0.00000000154,
       0.00000000208,    -0.00000000382,
      -0.00000000108,    -0.00000000043,
      -0.00000000374,    -0.00000000217,
      -0.00000000038,     0.00000000109,
      -0.00000000288,     0.00000000492,
       0.00000000200,     0.00000000095,
      -0.00000000141,    -0.00000000189,
      -0.00000000183,     0.00000000143,
       0.00000002031,    -0.00000003910,
      -0.00000001148,    -0.00000000388,
      -0.00000000526,    -0.00000000254,
      -0.00000000055,     0.00000000144,
       0.00000000602,     0.00000000794,
       0.00000000086,    -0.00000000104,
       0.00000000789,    -0.00000000602,
      -0.00000000113,    -0.00000000081,
      -0.00000000918,    -0.00000001237,
      -0.00000000429,     0.00000000341,
       0.00000000311,    -0.00000000039,
      -0.00000000080,    -0.00000000319,
       0.00000006261,     0.00000008517,
       0.00000000888,    -0.00000001122,
       0.00000001065,    -0.00000000851,
      -0.00000000145,    -0.00000000114,
      -0.00000001360,     0.00000000206,
       0.00000000293,     0.00000001376,
       0.00000002507,    -0.00000000434,
      -0.00000000139,    -0.00000000815,
       0.00000000132,    -0.00000000093,
      -0.00000014847,     0.00000002487,
       0.00000000337,     0.00000001969,
      -0.00000000107,    -0.00000000287,
      -0.00000000469,     0.00000000279,
       0.00000000421,     0.00000000761,
       0.00000000234,    -0.00000000109,
       0.00000000127,    -0.00000000019,
       0.00000000021,     0.00000000116,
      -0.00000001997,    -0.00000003600,
      -0.00000000541,     0.00000000265,
      -0.00000000212,     0.00000000036,
      -0.00000000031,    -0.00000000178,
      -0.00000000044,     0.00000000068,
       0.00000000063,     0.00000000046,
      -0.00000000052,     0.00000000064,
       0.00000000077,     0.00000000042,
       0.00000000064,     0.00000000052,
       0.00000000041,    -0.00000000077,
      -0.00000002242,     0.00000002738,
       0.00000003301,     0.00000001787,
       0.00000000081,     0.00000000066,
       0.00000000053,    -0.00000000097,
       0.00000000036,     0.00000000089,
       0.00000000092,    -0.00000000027,
      -0.00000001454,     0.00000001960,
       0.00000002311,     0.00000001061,
       0.00000000220,    -0.00000001250,
       0.00000000103,     0.00000000089,
       0.00000000072,    -0.00000000126,
      -0.00000000173,    -0.00000000269,
      -0.00000000182,     0.00000000190,
       0.00000000103,     0.00000000062,
      -0.00000000263,     0.00000000163,
       0.00000000175,     0.00000000181,
       0.00000000076,    -0.00000000087,
       0.00000000481,     0.00000000220,
       0.00000000097,    -0.00000000430,
      -0.00000000145,     0.00000000453,
       0.00000000417,     0.00000000034,
       0.00000000009,    -0.00000000246,
      -0.00000001157,    -0.00000000322,
      -0.00000000022,     0.00000000956,
       0.00000000283,    -0.00000000080,
      -0.00000000331,     0.00000001158,
       0.00000000947,     0.00000000054,
      -0.00000000050,    -0.00000000285,
      -0.00000005379,    -0.00000004805,
      -0.00000003270,     0.00000005304,
       0.00000002486,     0.00000000933,
       0.00000000273,    -0.00000001132,
      -0.00000000919,     0.00000000033,
       0.00000000064,     0.00000000350,
       0.00000002862,    -0.00000000436,
      -0.00000000712,    -0.00000002099,
      -0.00000000708,     0.00000000381,
      -0.00000000489,    -0.00000002885,
      -0.00000002107,     0.00000000768,
       0.00000000417,     0.00000000701,
       0.00000001616,    -0.00000001375,
      -0.00000001148,    -0.00000000985,
      -0.00000000285,     0.00000000429,
      -0.00000002472,    -0.00000003333,
      -0.00000002075,     0.00000002084,
       0.00000000774,     0.00000000555,
      -0.00000003159,     0.00000002708,
       0.00000002232,     0.00000001605,
       0.00000000273,    -0.00000000768,
       0.00000002667,     0.00000003145,
       0.00000001646,    -0.00000002167,
      -0.00000000733,    -0.00000000336,
      -0.00000002806,     0.00000024719,
       0.00000016854,    -0.00000002473,
      -0.00000002085,    -0.00000004992,
      -0.00000000592,     0.00000000661,
       0.00000005841,     0.00000007998,
       0.00000005043,    -0.00000004998,
      -0.00000001938,    -0.00000001579,
      -0.00000000133,     0.00000000353,
      -0.00000004412,     0.00000003459,
       0.00000002921,     0.00000002273,
       0.00000000494,    -0.00000001013,
      -0.00000000164,    -0.00000000084,
       0.00000003545,     0.00000004498,
       0.00000002319,    -0.00000003007,
      -0.00000001102,    -0.00000000477,
       0.00000000370,     0.00000000397,
       0.00000000191,    -0.00000000286,
      -0.00000000072,     0.00000000428,
       0.00000000245,    -0.00000000016,
       0.00000000464,    -0.00000000012,
      -0.00000000102,    -0.00000000234,
      -0.00000100443,     0.00000084434,
       0.00000069717,     0.00000051794,
       0.00000010788,    -0.00000024065,
      -0.00000003371,    -0.00000000627,
      -0.00000005850,    -0.00000006705,
      -0.00000003410,     0.00000004770,
       0.00000001646,     0.00000000701,
       0.00000000030,    -0.00000000180,
       0.00000000437,    -0.00000001795,
      -0.00000000974,    -0.00000000005,
       0.00000000029,     0.00000000246,
      -0.00000002029,    -0.00000000278,
       0.00000000172,     0.00000001096,
       0.00000000261,    -0.00000000134,
      -0.00000000125,     0.00000001568,
       0.00000000835,    -0.00000000173,
       0.00000000858,     0.00000000001,
      -0.00000000131,    -0.00000000446,
      -0.00000000105,     0.00000000067,
      -0.00000000616,    -0.00000000632,
      -0.00000000166,     0.00000000306,
      -0.00000000562,     0.00000000682,
       0.00000000339,     0.00000000107,
       0.00000001064,    -0.00000011877,
      -0.00000006349,     0.00000001228,
       0.00000000781,     0.00000001618,
      -0.00000001669,     0.00000000074,
       0.00000000292,     0.00000000868,
       0.00000000228,    -0.00000000143,
       0.00000001191,     0.00000003858,
       0.00000001628,    -0.00000000608,
      -0.00000000171,    -0.00000000342,
       0.00000001225,    -0.00000004292,
      -0.00000002158,     0.00000000449,
       0.00000000276,     0.00000000458,
      -0.00000000147,    -0.00000000139,
      -0.00000000059,     0.00000000107,
      -0.00000002159,     0.00000001827,
       0.00000001441,     0.00000001085,
       0.00000000205,    -0.00000000422,
      -0.00000001826,    -0.00000002164,
      -0.00000001089,     0.00000001441,
       0.00000000422,     0.00000000206,
      -0.00000002694,     0.00000002284,
       0.00000001799,     0.00000001353,
       0.00000000256,    -0.00000000527,
      -0.00000002280,    -0.00000002696,
      -0.00000001355,     0.00000001798,
       0.00000000526,     0.00000000256,
      -0.00000116325,     0.00000101190,
       0.00000081695,     0.00000058529,
       0.00000011082,    -0.00000025226,
      -0.00000003018,    -0.00000003465,
      -0.00000001744,     0.00000002440,
       0.00000000755,     0.00000000330,
      -0.00000003255,    -0.00000002836,
      -0.00000000635,     0.00000001627,
      -0.00000001187,     0.00000001480,
       0.00000000718,     0.00000000259,
       0.00000000006,    -0.00000000159,
       0.00000001155,    -0.00000000100,
      -0.00000000164,    -0.00000000247,
      -0.00000000147,    -0.00000001132,
      -0.00000000221,     0.00000000178,
       0.00000019836,     0.00000017124,
       0.00000003812,    -0.00000009902,
      -0.00000002275,    -0.00000000139,
       0.00000002813,    -0.00000003497,
      -0.00000001693,    -0.00000000610,
      -0.00000000020,     0.00000000383,
      -0.00000005027,     0.00000000440,
       0.00000000697,     0.00000001064,
       0.00000000129,    -0.00000000169,
       0.00000000636,     0.00000004951,
       0.00000001002,    -0.00000000781,
      -0.00000000182,    -0.00000000097,
       0.00000007743,    -0.00000001003,
      -0.00000001197,    -0.00000001577,
      -0.00000000334,    -0.00000002916,
      -0.00000000600,     0.00000000439,
       0.00000000103,     0.00000000062,
       0.00000000038,    -0.00000000403,
      -0.00000000298,    -0.00000000256,
      -0.00000033191,     0.00000004199,
       0.00000005134,     0.00000006729,
       0.00000000675,    -0.00000001196,
       0.00000000745,     0.00000006538,
       0.00000001337,    -0.00000000987,
      -0.00000000226,    -0.00000000136,
      -0.00000000409,     0.00000001553,
       0.00000000243,    -0.00000000117,
       0.00000001361,     0.00000000765,
       0.00000000027,    -0.00000000209,
       0.00000001063,    -0.00000002280,
      -0.00000000787,    -0.00000000296,
       0.00000000031,     0.00000000120,
       0.00000000467,     0.00000000236,
       0.00000000290,     0.00000000095,
      -0.00000005263,     0.00000013293,
       0.00000002016,    -0.00000000419,
       0.00000001925,     0.00000000703,
      -0.00000000063,    -0.00000000290,
      -0.00000001008,    -0.00000001118,
      -0.00000001162,     0.00000000486,
       0.00000001809,     0.00000002086,
       0.00000000701,    -0.00000000569,
       0.00000000124,    -0.00000000362,
      -0.00000000169,    -0.00000000095,
      -0.00000008584,    -0.00000011393,
      -0.00000001673,     0.00000001462,
      -0.00000000511,     0.00000001349,
      -0.00000000056,    -0.00000000201,
       0.00000001019,     0.00000000423,
      -0.00000000187,     0.00000000101,
       0.00000000002,     0.00000000005,
      -0.00000000005,     0.00000000002,
       0.00000000889,    -0.00000002135,
      -0.00000000658,    -0.00000000257,
       0.00000000107,    -0.00000000044,
       0.00000000099,     0.00000000071,
      -0.00000000162,     0.00000000102,
      -0.00000003916,     0.00000009832,
      -0.00000000701,    -0.00000001620,
       0.00000001520,     0.00000000606,
      -0.00000000248,     0.00000000104,
      -0.00000000104,     0.00000000016,
       0.00000000367,    -0.00000000112,
       0.00000003351,    -0.00000000497,
       0.00000000016,     0.00000000107,
      -0.00000001390,     0.00000001622,
       0.00000001950,     0.00000001104,
       0.00000000045,    -0.00000000081,
       0.00000000234,     0.00000000153,
      -0.00000001166,     0.00000001178,
      -0.00000000209,    -0.00000000243,
      -0.00000000110,     0.00000000106,
       0.00000001591,     0.00000003221,
       0.00000001718,    -0.00000001478,
      -0.00000000094,     0.00000000048,
      -0.00000000056,    -0.00000000052,
      -0.00000000047,     0.00000000064,
      -0.00000000734,    -0.00000000560,
      -0.00000000065,     0.00000000088,
      -0.00000000002,    -0.00000000111,
       0.00000000061,    -0.00000000051,
       0.00000000077,    -0.00000000057,
      -0.00000000373,     0.00000000105,
       0.00000000455,     0.00000001693,
       0.00000000268,    -0.00000000072,
      -0.00000000061,    -0.00000000148,
       0.00000000094,    -0.00000000007,
      -0.00000000019,     0.00000000116,
       0.00000000130,    -0.00000000009,
      -0.00000000090,     0.00000001422,
      -0.00000000069,    -0.00000000065,
      -0.00000000056,     0.00000000082,
      -0.00000000053,     0.00000000069,
       0.00000000194,    -0.00000000033,
      -0.00000000088,    -0.00000000200,
       0.00000000032,     0.00000000093,
       0.00000001392,    -0.00000000411,
      -0.00000000080,    -0.00000000226,
      -0.00000000085,     0.00000000305,
       0.00000000311,     0.00000000075,
       0.00000000116,    -0.00000000470,
      -0.00000000178,    -0.00000000045,
      -0.00000000530,     0.00000002150,
       0.00000000397,     0.00000000100,
       0.00000000071,    -0.00000000051,
      -0.00000000051,    -0.00000000071,
       0.00000000088,    -0.00000000064,
      -0.00000000064,    -0.00000000088,
       0.00000003793,    -0.00000002755,
      -0.00000000081,    -0.00000000112,
       0.00000003180,    -0.00000002404,
      -0.00000000186,    -0.00000000247,
      -0.00000000059,    -0.00000000082,
       0.00000000092,    -0.00000000063,
       0.00000000071,     0.00000000080,
       0.00000000207,     0.00000000246,
       0.00000000106,    -0.00000000085,
       0.00000000078,     0.00000000113,
      -0.00000000108,     0.00000000000,
      -0.00000000035,    -0.00000000088,
      -0.00000000025,    -0.00000000107,
      -0.00000000052,     0.00000000129,
       0.00000000133,     0.00000000049,
      -0.00000000075,    -0.00000000029,
      -0.00000000343,     0.00000000923,
       0.00000000171,     0.00000000064,
      -0.00000000922,     0.00000000018,
       0.00000000018,     0.00000000922,
       0.00000000764,    -0.00000000016,
      -0.00000000003,    -0.00000000195,
      -0.00000035699,     0.00000000732,
       0.00000000024,     0.00000001379,
      -0.00000000425,    -0.00000000089,
      -0.00000000104,     0.00000000441,
       0.00000000445,     0.00000000102,
       0.00000000021,    -0.00000000091,
       0.00000000186,    -0.00000000001,
       0.00000000004,     0.00000000181,
      -0.00000016865,    -0.00000003897,
      -0.00000000151,     0.00000000666,
      -0.00000000311,    -0.00000000040,
      -0.00000000047,    -0.00000000134,
       0.00000000104,     0.00000000025,
      -0.00000000024,     0.00000000104,
      -0.00000002016,    -0.00000001211,
      -0.00000000070,     0.00000000138,
      -0.00000000094,    -0.00000000029,
       0.00000000018,    -0.00000000103,
       0.00000014811,     0.00000012919,
       0.00000000075,     0.00000003021,
       0.00000000098,    -0.00000000093,
       0.00000000558,     0.00000000882,
      -0.00000000058,     0.00000000205,
       0.00000000900,    -0.00000000591,
       0.00000000209,     0.00000000057,
      -0.00000001031,     0.00000000922,
       0.00000000762,     0.00000000519,
       0.00000000099,    -0.00000000250,
       0.00000000922,     0.00000001031,
       0.00000000520,    -0.00000000762,
      -0.00000000250,    -0.00000000099,
       0.00000001731,    -0.00000000482,
      -0.00000000575,    -0.00000001346,
       0.00000001791,    -0.00000004057,
      -0.00000003398,    -0.00000001044,
       0.00000003789,    -0.00000001290,
      -0.00000001409,    -0.00000002869,
      -0.00000001080,     0.00000000721,
       0.00000000237,     0.00000000265,
       0.00000000048,    -0.00000000060,
      -0.00000001290,    -0.00000003789,
      -0.00000002869,     0.00000001409,
       0.00000000721,     0.00000001080,
       0.00000000265,    -0.00000000237,
      -0.00000000060,    -0.00000000048,
      -0.00000000046,     0.00000000183,
       0.00000000125,     0.00000000016,
       0.00000000183,     0.00000000046,
       0.00000000016,    -0.00000000125,
      -0.00000000430,    -0.00000000465,
      -0.00000000219,     0.00000000261,
       0.00000000079,     0.00000000050,
      -0.00000000465,     0.00000000430,
       0.00000000261,     0.00000000219,
       0.00000000050,    -0.00000000079,
      -0.00000001028,     0.00000000153,
       0.00000000109,     0.00000000400,
       0.00000000079,    -0.00000000034,
       0.00000000153,     0.00000001028,
       0.00000000400,    -0.00000000109,
      -0.00000000034,    -0.00000000079,
      -0.00000000202,     0.00000000360,
       0.00000000104,     0.00000000042,
       0.00000000360,     0.00000000202,
       0.00000000042,    -0.00000000104,
       0.00000000289,    -0.00000002383,
       0.00000000117,     0.00000000014,
      -0.00000020096,    -0.00000028961,
      -0.00000007666,     0.00000008670,
      -0.00000000080,     0.00000000037,
       0.00000000407,    -0.00000005354,
      -0.00000001313,     0.00000000102,
       0.00000000066,     0.00000000184,
      -0.00000005331,    -0.00000000275,
       0.00000000153,     0.00000001298,
       0.00000000180,    -0.00000000076,
      -0.00000000419,    -0.00000003695,
       0.00000000723,    -0.00000000464,
      -0.00000000465,    -0.00000000724,
       0.00000003523,     0.00000001803,
       0.00000000499,     0.00000000315,
       0.00000000318,    -0.00000000501,
       0.00000005603,     0.00000001260,
       0.00000000478,     0.00000000264,
       0.00000000265,    -0.00000000485,
      -0.00000000202,    -0.00000000025,
      -0.00000000024,     0.00000000198,
      -0.00000004342,    -0.00000000533,
      -0.00000000026,     0.00000000230,
       0.00000000513,     0.00000000040,
       0.00000000038,    -0.00000000497,
       0.00000010637,     0.00000000825,
       0.00000000046,    -0.00000000614,
      -0.00000000680,    -0.00000000013,
      -0.00000000007,     0.00000000620,
      -0.00000013217,    -0.00000000161,
      -0.00000000018,     0.00000000852,
      -0.00000000270,    -0.00000000046,
      -0.00000000057,     0.00000000428,
      -0.00000006034,    -0.00000001015,
      -0.00000000067,     0.00000000433,
      -0.00000000174,    -0.00000000022,
      -0.00000000075,     0.00000000029,
       0.00000000246,    -0.00000000094,
      -0.00000000066,    -0.00000000145,
       0.00000000045,     0.00000000212,
       0.00000000134,    -0.00000000022,
       0.00000000211,    -0.00000000058,
      -0.00000000089,    -0.00000000126,
       0.00000000783,     0.00000003533,
       0.00000001939,    -0.00000000774,
      -0.00000000208,     0.00000000047,
       0.00000000039,     0.00000000092,
      -0.00000000126,    -0.00000000049,
      -0.00000000045,     0.00000000129,
      -0.00000000695,    -0.00000000326,
      -0.00000000038,     0.00000000097,
       0.00000000230,    -0.00000000096,
      -0.00000000199,    -0.00000000205,
      -0.00000000052,     0.00000000100,
       0.00000000022,     0.00000000088,
      -0.00000000088,     0.00000000022,
       0.00000000074,     0.00000000102,
      -0.00000000093,     0.00000000010,
       0.00000001540,    -0.00000001016,
      -0.00000000426,    -0.00000000538,
      -0.00000000179,    -0.00000000247,
      -0.00000000092,     0.00000000076,
       0.00000000019,     0.00000000395,
       0.00000000397,    -0.00000000035,
      -0.00000000048,    -0.00000000615,
      -0.00000000232,     0.00000000015,
       0.00000000225,     0.00000002741,
       0.00000000520,    -0.00000000040,
       0.00000068347,     0.00000105977,
      -0.00000006798,     0.00000024607,
      -0.00000003357,     0.00000000876,
      -0.00000003439,     0.00000002216,
      -0.00000000798,    -0.00000000221,
      -0.00000000028,    -0.00000000109,
       0.00000000119,     0.00000000090,
      -0.00000000117,     0.00000000111,
       0.00000428949,    -0.00000145671,
      -0.00000159248,    -0.00000324880,
      -0.00000122309,     0.00000081453,
       0.00000026768,     0.00000029968,
       0.00000005319,    -0.00000006590,
       0.00000004331,     0.00000012718,
       0.00000009631,    -0.00000004731,
      -0.00000002418,    -0.00000003625,
      -0.00000000888,     0.00000000794,
       0.00000000196,     0.00000000159,
      -0.00000005435,     0.00000021511,
       0.00000014711,     0.00000001804,
      -0.00000000052,    -0.00000005045,
      -0.00000000636,    -0.00000000162,
      -0.00000000055,     0.00000000435,
       0.00000000149,    -0.00000000001,
      -0.00000050394,    -0.00000054411,
      -0.00000025554,     0.00000030453,
       0.00000009185,     0.00000005785,
       0.00000001685,    -0.00000001558,
      -0.00000000942,    -0.00000000792,
      -0.00000000180,     0.00000000285,
      -0.00000119950,     0.00000017754,
       0.00000012595,     0.00000046428,
       0.00000009172,    -0.00000003900,
      -0.00000000562,    -0.00000003792,
      -0.00000001467,     0.00000000399,
       0.00000000123,     0.00000000289,
      -0.00000023284,     0.00000041231,
       0.00000011871,     0.00000004895,
      -0.00000001340,    -0.00000000750,
      -0.00000000156,     0.00000000385,
       0.00000083304,    -0.00000054692,
      -0.00000005689,    -0.00000000060,
       0.00000001762,     0.00000002740,
       0.00000000002,    -0.00000000183,
       0.00000059034,     0.00000037207,
       0.00000003367,     0.00000004052,
      -0.00000001206,     0.00000001906,
      -0.00000000130,     0.00000000108,
       0.00000057079,     0.00000031394,
       0.00000001927,     0.00000006669,
      -0.00000001015,     0.00000001843,
      -0.00000000215,     0.00000000062,
       0.00000000056,     0.00000000088,
       0.00000000087,    -0.00000000059,
       0.00000000621,     0.00000000976,
       0.00000000116,    -0.00000000080,
      -0.00000000154,     0.00000000009,
       0.00000000019,     0.00000000158,
      -0.00000000009,    -0.00000000092,
      -0.00000001783,     0.00000000160,
       0.00000000021,     0.00000000226,
      -0.00000000187,     0.00000000007,
       0.00000000006,     0.00000000203,
      -0.00000002896,     0.00000000064,
       0.00000000010,     0.00000000172,
       0.00000000803,    -0.00000000050,
      -0.00000000048,    -0.00000000834,
       0.00000000005,     0.00000000101,
       0.00000016741,    -0.00000000911,
      -0.00000000056,    -0.00000001064,
       0.00000000276,     0.00000000032,
       0.00000000028,    -0.00000000297,
       0.00000005073,     0.00000000671,
       0.00000000041,    -0.00000000346,
       0.00000000171,     0.00000000005,
      -0.00000001372,    -0.00000000572,
       0.00000000438,    -0.00000001539,
       0.00000003404,     0.00000001319,
      -0.00000000085,     0.00000000077,
       0.00000000129,    -0.00000000021,
      -0.00000000020,    -0.00000000132,
      -0.00000000022,    -0.00000000150,
      -0.00000000085,     0.00000000171,
       0.00000000171,     0.00000000085,
      -0.00000000089,    -0.00000000045,
      -0.00000000422,     0.00000000796,
       0.00000000198,     0.00000000099,
       0.00000000002,     0.00000000009,
      -0.00000000001,     0.00000000029,
      -0.00000000005,     0.00000000003,
      -0.00000000029,    -0.00000000001,
      -0.00000000003,    -0.00000000005,
       0.00000000010,     0.00000000043,
      -0.00000000043,     0.00000000010,
       0.00000000820,    -0.00000000162,
       0.00000000040,     0.00000000616,
      -0.00000000029,    -0.00000000183,
       0.00000000139,     0.00000000000,
       0.00000000139,    -0.00000000125,
       0.00000000085,     0.00000000091,
       0.00000000113,     0.00000000091,
      -0.00000000047,     0.00000000079,
      -0.00000003445,     0.00000000564,
      -0.00000000041,    -0.00000002636,
       0.00000001009,     0.00000000122,
       0.00000000080,     0.00000000440,
      -0.00000000330,     0.00000000009,
      -0.00000000159,     0.00000000267,
      -0.00000000128,    -0.00000000127,
      -0.00000000220,     0.00000000046,
      -0.00000000057,    -0.00000000101,
       0.00000000048,     0.00000000141,
      -0.00000000141,     0.00000000048,
       0.00000000059,     0.00000000177,
      -0.00000000177,     0.00000000059,
      -0.00000000018,    -0.00000000139,
      -0.00000000040,     0.00000000039,
       0.00000000051,    -0.00000000094,
       0.00000000140,    -0.00000000436,
       0.00000000064,    -0.00000000012,
      -0.00000000131,     0.00000000318,
      -0.00000000042,     0.00000000011,
       0.00000000021,    -0.00000000096,
       0.00000000068,     0.00000000016,
      -0.00000000024,    -0.00000000092,
      -0.00000000089,     0.00000000027,
      -0.00000000176,    -0.00000000605,
       0.00000000637,    -0.00000000193,
       0.00000000032,    -0.00000000039,
       0.00000000109,    -0.00000000176,
       0.00000000200,     0.00000000178,
      -0.00000000103,    -0.00000000317,
       0.00000000337,    -0.00000000103,
      -0.00000000140,     0.00000000029,
       0.00000000084,    -0.00000000125,
       0.00000000123,     0.00000000085,
      -0.00000000051,     0.00000000019,
       0.00000002352,     0.00000000372,
       0.00000000130,    -0.00000000351,
      -0.00000019012,    -0.00000002849,
      -0.00000001029,     0.00000002849,
      -0.00000000084,     0.00000000070,
       0.00000000068,     0.00000000092,
      -0.00000002474,     0.00000001911,
       0.00000000049,     0.00000000064,
      -0.00000002302,     0.00000001859,
       0.00000000055,     0.00000000080,
      -0.00000000103,     0.00000000081,
      -0.00000000081,    -0.00000000103,
       0.00000001356,    -0.00000001771,
      -0.00000002100,    -0.00000001045,
      -0.00000037240,    -0.00000051220,
       0.00000002430,    -0.00000001430,
      -0.00000002920,     0.00000002250,
       0.00000106100,     0.00000137620,
      -0.00000005760,     0.00000004340,
      -0.00000000420,    -0.00000002140,
      -0.00000002160,     0.00000000470,
       0.00000009949,    -0.00000008048,
      -0.00000065648,    -0.00000081146,
       0.00000003560,    -0.00000002860,
      -0.00000001040,     0.00000000430,
       0.00000000540,    -0.00000000800,
      -0.00000000192,     0.00000000106,
       0.00000000152,     0.00000000212,
       0.00000000103,    -0.00000000114,
       0.00000000100,     0.00000000195,
       0.00000000218,    -0.00000000143,
      -0.00000000100,    -0.00000000106,
      -0.00000001197,    -0.00000000650,
      -0.00000000661,     0.00000001683,
       0.00000001184,     0.00000000269,
       0.00000000057,    -0.00000000058,
      -0.00000000084,    -0.00000000057,
       0.00000000643,    -0.00000000353,
      -0.00000000519,    -0.00000000704,
      -0.00000000378,     0.00000000400,
       0.00000000168,     0.00000000094,
      -0.00000000355,    -0.00000000638,
      -0.00000000694,     0.00000000514,
       0.00000000389,     0.00000000367,
       0.00000000097,    -0.00000000157,
       0.00000000054,     0.00000000078,
       0.00000000077,    -0.00000000053,
      -0.00000000087,     0.00000000045,
       0.00000000067,     0.00000000096,
       0.00000000052,    -0.00000000048,
       0.00000000045,     0.00000000086,
       0.00000000096,    -0.00000000067,
      -0.00000002337,     0.00000001227,
       0.00000001809,     0.00000002581,
       0.00000001405,    -0.00000001308,
       0.00000000084,    -0.00000000061,
      -0.00000000190,    -0.00000000012,
       0.00000000013,     0.00000000184,
       0.00000000110,    -0.00000000026,
       0.00000000047,     0.00000000067,
      -0.00000000067,     0.00000000047,
      -0.00000000079,    -0.00000000100,
       0.00000002290,    -0.00000001845,
       0.00000000066,     0.00000000085,
       0.00000000089,    -0.00000000066,
       0.00000000067,     0.00000000084,
      -0.00000003958,     0.00000002076,
       0.00000003078,     0.00000004383,
       0.00000002426,    -0.00000002237,
       0.00000000055,     0.00000000106,
       0.00000000117,    -0.00000000082,
      -0.00000000054,    -0.00000000059,
       0.00000000640,    -0.00000000450,
      -0.00000001450,     0.00000001070,
      -0.00000006710,     0.00000003771,
       0.00000005498,     0.00000007334,
       0.00000003906,    -0.00000003868,
      -0.00000001806,    -0.00000001384,
      -0.00000000230,     0.00000000552,
      -0.00000027080,    -0.00000048998,
      -0.00000053646,     0.00000039815,
       0.00000028197,     0.00000028563,
       0.00000010210,    -0.00000013662,
      -0.00000003948,    -0.00000001649,
       0.00000001051,    -0.00000011548,
      -0.00000012341,     0.00000000573,
       0.00000001293,     0.00000006592,
       0.00000002006,    -0.00000000714,
      -0.00000011490,    -0.00000001000,
       0.00000000366,     0.00000012238,
       0.00000006559,    -0.00000001030,
      -0.00000000417,    -0.00000002051,
       0.00000075160,    -0.00000042488,
      -0.00000061541,    -0.00000082934,
      -0.00000044922,     0.00000043542,
       0.00000020034,     0.00000015958,
       0.00000002999,    -0.00000005804,
       0.00000066996,     0.00000120085,
       0.00000132505,    -0.00000097573,
      -0.00000068992,    -0.00000071348,
      -0.00000024787,     0.00000031159,
       0.00000008409,     0.00000004424,
      -0.00000003272,     0.00000046393,
       0.00000049597,    -0.00000002605,
      -0.00000004646,    -0.00000026615,
      -0.00000008823,     0.00000003241,
       0.00000000974,     0.00000001535,
       0.00000046294,     0.00000003003,
      -0.00000002946,    -0.00000049439,
      -0.00000026177,     0.00000004753,
       0.00000003074,     0.00000008759,
       0.00000001530,    -0.00000000737,
      -0.00000000002,    -0.00000000105,
       0.00000000086,    -0.00000000037,
      -0.00000000299,     0.00000004146,
       0.00000004436,    -0.00000000234,
      -0.00000000444,    -0.00000002358,
      -0.00000000823,     0.00000000275,
       0.00000004466,     0.00000000043,
      -0.00000000509,    -0.00000004738,
      -0.00000002549,     0.00000000590,
       0.00000000313,     0.00000000853,
       0.00000000163,    -0.00000000088,
      -0.00000002877,    -0.00000002203,
      -0.00000001697,     0.00000002926,
       0.00000001472,     0.00000000604,
       0.00000000103,    -0.00000000476,
      -0.00000002148,     0.00000002887,
       0.00000002935,     0.00000001642,
       0.00000000588,    -0.00000001487,
      -0.00000000468,    -0.00000000131,
       0.00000001571,    -0.00000034092,
      -0.00000036313,     0.00000002753,
       0.00000003927,     0.00000019199,
       0.00000006741,    -0.00000002249,
      -0.00000000826,    -0.00000001726,
      -0.00000010864,    -0.00000000189,
       0.00000001171,     0.00000011543,
       0.00000006095,    -0.00000001400,
      -0.00000000765,    -0.00000002142,
      -0.00000000594,     0.00000000310,
       0.00000011980,     0.00000008609,
       0.00000006644,    -0.00000012092,
      -0.00000005967,    -0.00000002433,
      -0.00000000528,     0.00000002018,
       0.00000000476,     0.00000000025,
       0.00000008556,    -0.00000011931,
      -0.00000012026,    -0.00000006598,
      -0.00000002376,     0.00000006005,
       0.00000002013,     0.00000000506,
       0.00000000094,    -0.00000000443,
      -0.00000005933,    -0.00000004262,
      -0.00000003268,     0.00000005990,
       0.00000002998,     0.00000001168,
       0.00000000239,    -0.00000000993,
      -0.00000002608,     0.00000003673,
       0.00000003716,     0.00000001985,
       0.00000000726,    -0.00000001861,
      -0.00000000571,    -0.00000000145,
       0.00000000091,    -0.00000000552,
      -0.00000000397,    -0.00000000006,
       0.00000000019,     0.00000000152,
      -0.00000000539,     0.00000000041,
       0.00000000109,     0.00000000351,
       0.00000000111,    -0.00000000054,
      -0.00000000154,     0.00000000049,
       0.00000000039,     0.00000000153,
       0.00000027997,     0.00000013974,
       0.00000010226,    -0.00000027256,
      -0.00000013265,    -0.00000003350,
      -0.00000000560,     0.00000004280,
       0.00000000985,     0.00000000038,
       0.00000005824,    -0.00000008068,
      -0.00000008182,    -0.00000004492,
      -0.00000001612,     0.00000004120,
       0.00000001348,     0.00000000312,
       0.00000000036,    -0.00000000341,
      -0.00000000202,     0.00000001762,
       0.00000001167,    -0.00000000061,
      -0.00000000071,    -0.00000000375,
       0.00000001819,     0.00000000113,
      -0.00000000147,    -0.00000001212,
      -0.00000000384,     0.00000000099,
       0.00000000378,    -0.00000000112,
      -0.00000000102,    -0.00000000198,
      -0.00000000100,    -0.00000000376,
      -0.00000000199,     0.00000000088,
       0.00000000893,    -0.00000002623,
      -0.00000001920,    -0.00000000395,
      -0.00000000069,     0.00000000713,
       0.00000000146,    -0.00000000369,
      -0.00000000246,    -0.00000000053,
       0.00000006023,    -0.00000018110,
      -0.00000012531,    -0.00000002380,
      -0.00000000210,     0.00000004344,
       0.00000000991,    -0.00000000098,
      -0.00000018456,    -0.00000005917,
      -0.00000002218,     0.00000012822,
       0.00000004449,     0.00000000080,
      -0.00000000164,    -0.00000001011,
       0.00000000072,    -0.00000000244,
      -0.00000000243,    -0.00000000053,
       0.00000093206,    -0.00000015025,
      -0.00000023642,    -0.00000072437,
      -0.00000027694,     0.00000014205,
       0.00000005098,     0.00000006871,
       0.00000001219,    -0.00000001288,
      -0.00000001588,    -0.00000003039,
      -0.00000002407,     0.00000001506,
       0.00000000713,     0.00000000930,
       0.00000000267,    -0.00000000242,
      -0.00000023606,     0.00000079058,
       0.00000054628,     0.00000008390,
       0.00000000095,    -0.00000018844,
      -0.00000004273,     0.00000000673,
       0.00000000228,     0.00000000708,
       0.00000079168,     0.00000023375,
       0.00000008161,    -0.00000054702,
      -0.00000018857,     0.00000000017,
       0.00000000730,     0.00000004272,
       0.00000000698,    -0.00000000255,
      -0.00000000172,     0.00000000578,
       0.00000000235,     0.00000000037,
       0.00000000585,     0.00000000147,
       0.00000000019,    -0.00000000240,
       0.00000000111,    -0.00000000030,
      -0.00000000037,    -0.00000000112,
      -0.00000000042,     0.00000000084,
       0.00000000085,     0.00000000041,
       0.00000030283,    -0.00000091967,
      -0.00000063823,    -0.00000011583,
      -0.00000000737,     0.00000022082,
       0.00000005019,    -0.00000000630,
      -0.00000000260,    -0.00000000835,
      -0.00000038825,    -0.00000009655,
      -0.00000003175,     0.00000026614,
       0.00000009121,    -0.00000000153,
      -0.00000000347,    -0.00000002059,
      -0.00000000344,     0.00000000115,
       0.00000028118,     0.00000033533,
       0.00000015915,    -0.00000017070,
      -0.00000005199,    -0.00000003676,
      -0.00000000526,     0.00000001044,
       0.00000032481,    -0.00000029481,
      -0.00000018076,    -0.00000014988,
      -0.00000003256,     0.00000005570,
       0.00000001127,     0.00000000402,
       0.00000000025,     0.00000000122,
       0.00000000024,    -0.00000000206,
      -0.00000184329,     0.00000656307,
       0.00000452820,     0.00000059701,
      -0.00000003189,    -0.00000155604,
      -0.00000035062,     0.00000006570,
       0.00000002203,     0.00000005772,
       0.00000000701,    -0.00000000438,
       0.00000092857,     0.00000024976,
       0.00000009089,    -0.00000063782,
      -0.00000021962,    -0.00000000268,
       0.00000000664,     0.00000004974,
       0.00000000841,    -0.00000000256,
      -0.00000122730,    -0.00000142701,
      -0.00000067241,     0.00000074771,
       0.00000022864,     0.00000015291,
       0.00000002110,    -0.00000004617,
      -0.00000000679,    -0.00000000176,
      -0.00000140809,     0.00000124782,
       0.00000076251,     0.00000065634,
       0.00000014594,    -0.00000023405,
      -0.00000004742,    -0.00000001898,
      -0.00000000126,     0.00000000702,
      -0.00000000080,    -0.00000000128,
       0.00000000038,    -0.00000000107,
      -0.00000000035,     0.00000000279,
       0.00000000920,    -0.00000000270,
      -0.00000000047,    -0.00000000136,
      -0.00000000045,     0.00000000163,
       0.00000000168,     0.00000000037,
       0.00000000111,     0.00000000054,
       0.00000000009,     0.00000000007,
       0.00000000020,     0.00000000034,
      -0.00000000034,     0.00000000020,
       0.00000000111,    -0.00000000165,
       0.00000000162,     0.00000000105,
       0.00000182571,     0.00000210090,
       0.00000098124,    -0.00000111760,
      -0.00000034220,    -0.00000022023,
      -0.00000002938,     0.00000006899,
       0.00000001020,     0.00000000213,
       0.00000074224,    -0.00000068295,
      -0.00000041331,    -0.00000034816,
      -0.00000007818,     0.00000012548,
       0.00000002521,     0.00000001047,
       0.00000000078,    -0.00000000366,
      -0.00000044589,     0.00000008348,
       0.00000004448,     0.00000017083,
       0.00000003388,    -0.00000001182,
      -0.00000000179,    -0.00000000451,
       0.00000014415,     0.00000044552,
       0.00000016711,    -0.00000008460,
      -0.00000002524,    -0.00000003130,
      -0.00000000344,     0.00000000506,
       0.00000000647,    -0.00000000308,
      -0.00000000145,    -0.00000000135,
       0.00000000172,    -0.00000000570,
      -0.00000000131,    -0.00000000050,
      -0.00000991504,    -0.00001121079,
      -0.00000522769,     0.00000604738,
       0.00000184700,     0.00000117076,
       0.00000015551,    -0.00000037183,
      -0.00000005511,    -0.00000001144,
      -0.00000174785,     0.00000161482,
       0.00000097503,     0.00000082021,
       0.00000018464,    -0.00000029583,
      -0.00000005926,    -0.00000002459,
      -0.00000000173,     0.00000000870,
       0.00000198113,    -0.00000040901,
      -0.00000022868,    -0.00000075855,
      -0.00000014978,     0.00000006389,
       0.00000001121,     0.00000001934,
       0.00000000186,    -0.00000000146,
      -0.00000056008,    -0.00000198262,
      -0.00000074829,     0.00000032881,
       0.00000009722,     0.00000014201,
       0.00000001636,    -0.00000001889,
      -0.00000000257,    -0.00000000123,
      -0.00000001890,     0.00000000436,
       0.00000000258,     0.00000000571,
       0.00000000140,    -0.00000000074,
      -0.00000000005,     0.00000001795,
       0.00000000553,    -0.00000000089,
      -0.00000000044,    -0.00000000139,
      -0.00000000170,     0.00000000035,
       0.00000000044,     0.00000000107,
       0.00000000210,     0.00000002486,
       0.00000001645,    -0.00000000438,
      -0.00000000259,    -0.00000000536,
      -0.00000000111,     0.00000000087,
      -0.00000002494,     0.00000000193,
       0.00000000431,     0.00000001649,
       0.00000000538,    -0.00000000258,
      -0.00000000086,    -0.00000000111,
       0.00000000264,     0.00000003114,
       0.00000002061,    -0.00000000549,
      -0.00000000325,    -0.00000000672,
      -0.00000000139,     0.00000000108,
      -0.00000003125,     0.00000000242,
       0.00000000540,     0.00000002066,
       0.00000000673,    -0.00000000323,
      -0.00000000108,    -0.00000000139,
      -0.00000000098,    -0.00000000024,
      -0.00000000315,     0.00000001354,
       0.00000000222,     0.00000000053,
       0.00000000095,     0.00000000035,
      -0.00000393088,     0.00000097503,
       0.00000056890,     0.00000149392,
       0.00000028823,    -0.00000016371,
      -0.00000003033,    -0.00000003499,
      -0.00000000284,     0.00000000401,
       0.00000023615,     0.00000128310,
       0.00000049194,    -0.00000015955,
      -0.00000004897,    -0.00000009560,
      -0.00000001177,     0.00000000936,
      -0.00000044077,     0.00000097207,
       0.00000029495,     0.00000005674,
      -0.00000000710,    -0.00000004644,
      -0.00000000483,     0.00000000315,
       0.00000083654,     0.00000059524,
       0.00000014642,    -0.00000023182,
      -0.00000003232,    -0.00000001938,
      -0.00000000211,     0.00000000292,
      -0.00000000312,     0.00000001208,
       0.00000000091,    -0.00000000036,
       0.00000000816,     0.00000000655,
       0.00001308262,    -0.00000420972,
      -0.00000220105,    -0.00000488838,
      -0.00000093517,     0.00000059642,
       0.00000010676,     0.00000011173,
       0.00000000884,    -0.00000001377,
      -0.00000052090,    -0.00000276800,
      -0.00000105815,     0.00000034836,
       0.00000010644,     0.00000020555,
       0.00000002510,    -0.00000002035,
      -0.00000000275,    -0.00000000207,
       0.00000210349,    -0.00000399985,
      -0.00000118131,    -0.00000036894,
      -0.00000000671,     0.00000018013,
       0.00000001798,    -0.00000000605,
      -0.00000379976,    -0.00000232136,
      -0.00000049689,     0.00000108789,
       0.00000015939,     0.00000004480,
       0.00000000145,    -0.00000001525,
       0.00000001341,    -0.00000002992,
      -0.00000000385,    -0.00000000127,
      -0.00000002750,    -0.00000001505,
      -0.00000000105,     0.00000000377,
       0.00000000203,     0.00000000274,
       0.00000000138,    -0.00000000122,
      -0.00000000295,     0.00000000186,
       0.00000000119,     0.00000000144,
       0.00000000254,     0.00000000344,
       0.00000000172,    -0.00000000153,
      -0.00000000370,     0.00000000234,
       0.00000000149,     0.00000000180,
       0.00000000004,     0.00000000002,
       0.00000000004,    -0.00000000006,
       0.00000000006,     0.00000000004,
      -0.00000405529,     0.00000669222,
       0.00000195827,     0.00000081411,
       0.00000005795,    -0.00000029326,
      -0.00000002886,     0.00000000194,
       0.00000221904,     0.00000111080,
       0.00000022351,    -0.00000063484,
      -0.00000009354,    -0.00000001403,
       0.00000000105,     0.00000000907,
      -0.00000146502,    -0.00000206789,
      -0.00000023235,     0.00000029037,
       0.00000003861,     0.00000000417,
      -0.00000000148,    -0.00000000360,
      -0.00000210366,     0.00000108014,
       0.00000014220,     0.00000027044,
       0.00000001737,    -0.00000000970,
       0.00000000252,    -0.00000000898,
       0.00000000141,    -0.00000000310,
      -0.00000000477,     0.00000001315,
      -0.00000000003,    -0.00000000279,
       0.00002368605,    -0.00004128321,
      -0.00001204601,    -0.00000461281,
      -0.00000028556,     0.00000179562,
       0.00000017559,    -0.00000002060,
      -0.00000000601,    -0.00000001224,
      -0.00000548165,    -0.00000272534,
      -0.00000054700,     0.00000156636,
       0.00000023088,     0.00000003396,
      -0.00000000268,    -0.00000002242,
       0.00000569437,     0.00000912471,
       0.00000109207,    -0.00000099797,
      -0.00000011655,    -0.00000004303,
       0.00000000183,     0.00000000920,
       0.00000912763,    -0.00000538643,
      -0.00000087981,    -0.00000111646,
      -0.00000005306,     0.00000009342,
       0.00000000640,     0.00000000016,
      -0.00000000144,     0.00000001019,
      -0.00000000767,     0.00000000864,
       0.00000000124,     0.00000000067,
       0.00000000542,    -0.00000001672,
       0.00000000386,     0.00000000970,
      -0.00000000240,    -0.00000000092,
      -0.00000000046,     0.00000000108,
       0.00000000131,    -0.00000000265,
      -0.00000000112,    -0.00000000050,
      -0.00000000301,    -0.00000000116,
      -0.00000000057,     0.00000000135,
       0.00000000165,    -0.00000000332,
      -0.00000000141,    -0.00000000063,
      -0.00000000064,     0.00000000039,
      -0.00000000003,    -0.00000000004,
      -0.00000000005,    -0.00000000014,
       0.00000000014,    -0.00000000005,
      -0.00000000033,     0.00000000020,
      -0.00000000060,     0.00000000046,
      -0.00000000045,    -0.00000000060,
      -0.00001124646,    -0.00001843469,
      -0.00000220192,     0.00000193947,
       0.00000021059,     0.00000009642,
      -0.00000000156,    -0.00000001521,
      -0.00000543342,     0.00000352541,
       0.00000057831,     0.00000066615,
       0.00000002927,    -0.00000006189,
      -0.00000000451,     0.00000000037,
       0.00000309460,     0.00000019851,
       0.00000009304,     0.00000003149,
       0.00000000254,    -0.00000000200,
      -0.00000051901,    -0.00000335726,
      -0.00000002635,     0.00000010559,
       0.00000002646,     0.00000000143,
      -0.00000001515,     0.00000001841,
       0.00000000080,     0.00000000169,
      -0.00000001872,     0.00000000163,
       0.00000000131,     0.00000000093,
       0.00006016365,     0.00009842434,
       0.00001175713,    -0.00001036537,
      -0.00000112442,    -0.00000051400,
       0.00000000830,     0.00000008138,
       0.00001299756,    -0.00000848154,
      -0.00000138786,    -0.00000159305,
      -0.00000006983,     0.00000014863,
       0.00000001084,    -0.00000000087,
      -0.00001413793,    -0.00000001596,
      -0.00000015650,    -0.00000005549,
      -0.00000000450,    -0.00000002869,
       0.00000000238,    -0.00000000028,
       0.00000124858,     0.00001458057,
       0.00000003602,    -0.00000019336,
      -0.00000007877,    -0.00000000026,
       0.00000000041,     0.00000000241,
       0.00000005733,    -0.00000001700,
      -0.00000000180,    -0.00000001205,
      -0.00000000189,     0.00000000074,
       0.00000002004,    -0.00000004127,
      -0.00000001040,    -0.00000000177,
      -0.00000000020,     0.00000000159,
      -0.00000000358,     0.00000000039,
       0.00000000013,     0.00000000113,
      -0.00000000032,    -0.00000000310,
      -0.00000000113,     0.00000000013,
      -0.00000000448,     0.00000000049,
       0.00000000016,     0.00000000141,
      -0.00000000040,    -0.00000000388,
      -0.00000000141,     0.00000000016,
      -0.00000000006,    -0.00000000002,
       0.00000000002,    -0.00000000006,
       0.00000000014,     0.00000000056,
      -0.00000000055,     0.00000000013,
       0.00003398126,    -0.00000197413,
      -0.00000012510,     0.00000003552,
       0.00000000745,     0.00000011807,
      -0.00000037686,    -0.00000969630,
       0.00000000667,     0.00000002552,
       0.00000003249,    -0.00000000130,
      -0.00000017724,    -0.00000126476,
      -0.00000006761,     0.00000016281,
       0.00000000917,     0.00000001096,
       0.00000070589,    -0.00000129419,
      -0.00000006224,    -0.00000018044,
       0.00000000015,     0.00000001365,
      -0.00000000867,    -0.00000002566,
       0.00000000292,    -0.00000000411,
       0.00000001097,    -0.00000001990,
       0.00000000245,     0.00000000519,
      -0.00015579911,     0.00000886082,
       0.00000058306,    -0.00000015976,
      -0.00000003395,    -0.00000053754,
       0.00000093656,     0.00002245159,
      -0.00000001626,    -0.00000006609,
      -0.00000007600,     0.00000000321,
      -0.00000120153,     0.00000263968,
      -0.00000004908,    -0.00000037413,
      -0.00000000790,    -0.00000003114,
       0.00000000164,     0.00000000119,
      -0.00000030574,     0.00000337825,
      -0.00000003843,     0.00000041851,
      -0.00000001517,    -0.00000002620,
      -0.00000000126,    -0.00000000143,
      -0.00000003204,     0.00000006508,
       0.00000000196,     0.00000000976,
       0.00000004279,     0.00000004660,
       0.00000000041,    -0.00000001295,
      -0.00000000283,    -0.00000000252,
       0.00000000254,    -0.00000000271,
      -0.00000000355,    -0.00000000315,
       0.00000000319,    -0.00000000339,
       0.00000000024,    -0.00000000021,
      -0.00000000006,    -0.00000000003,
       0.00000000090,    -0.00000000044,
      -0.00000000015,    -0.00000000014,
       0.00000000043,     0.00000000090,
       0.00000000014,    -0.00000000015,
       0.00000000002,    -0.00000000004,
       0.00000000005,     0.00000000002,
      -0.00000000040,     0.00000000017,
      -0.00000000050,    -0.00000000353,
      -0.00000000049,     0.00000000012,
       0.00000000354,    -0.00000000049,
      -0.00000000012,    -0.00000000049,
       0.00000666367,    -0.00000206187,
       0.00000053100,     0.00000052167,
      -0.00000002604,     0.00000005566,
      -0.00000059703,    -0.00000151650,
       0.00000012852,    -0.00000013494,
       0.00000001324,     0.00000000626,
       0.00000230021,    -0.00000022866,
       0.00000002361,     0.00000000371,
       0.00000000035,     0.00000001196,
      -0.00000002284,     0.00000228824,
      -0.00000000056,    -0.00000004579,
      -0.00000000185,     0.00000000046,
      -0.00000001594,    -0.00000000139,
       0.00000000073,     0.00000000253,
       0.00000000022,    -0.00000001613,
      -0.00000000269,     0.00000000054,
      -0.00002476497,     0.00000702481,
      -0.00000195672,    -0.00000185041,
       0.00000008926,    -0.00000020938,
       0.00000001308,     0.00000000038,
       0.00000142889,     0.00000359160,
      -0.00000030481,     0.00000032143,
      -0.00000003143,    -0.00000001477,
       0.00000000015,    -0.00000000204,
      -0.00000283839,     0.00000065615,
      -0.00000013375,    -0.00000000687,
      -0.00000000053,    -0.00000002960,
       0.00000039791,    -0.00000283278,
       0.00000000026,     0.00000016010,
      -0.00000001273,    -0.00000000069,
       0.00000003061,     0.00000000687,
      -0.00000000113,    -0.00000000757,
       0.00000001085,     0.00000003516,
       0.00000000872,    -0.00000000083,
       0.00000000077,     0.00000000051,
       0.00000000858,    -0.00000001017,
      -0.00000000200,    -0.00000000072,
       0.00000001018,     0.00000000860,
       0.00000000072,    -0.00000000199,
       0.00000001074,    -0.00000001273,
      -0.00000000250,    -0.00000000090,
       0.00000001275,     0.00000001077,
       0.00000000090,    -0.00000000249,
       0.00000000129,    -0.00000000014,
      -0.00000000014,    -0.00000000128,
       0.00000003124,    -0.00000000331,
      -0.00000000020,    -0.00000000167,
      -0.00000000438,     0.00000000068,
       0.00000000050,     0.00000000180,
       0.00000000067,     0.00000000429,
       0.00000000176,    -0.00000000049,
      -0.00000010404,     0.00000001632,
       0.00000001188,     0.00000004273,
       0.00000000098,     0.00000000613,
       0.00000000252,    -0.00000000071,
       0.00000000858,    -0.00000000201,
      -0.00000000120,    -0.00000000330,
      -0.00000000195,    -0.00000000814,
      -0.00000000311,     0.00000000116,
      -0.00000000471,     0.00000000111,
       0.00000000028,     0.00000000129,
       0.00000019633,    -0.00000004708,
      -0.00000002806,    -0.00000007516,
      -0.00000001456,     0.00000000781,
      -0.00000000302,    -0.00000001325,
      -0.00000000508,     0.00000000183,
       0.00000000052,     0.00000000098,
       0.00000000131,     0.00000000130,
       0.00000000102,    -0.00000000296,
      -0.00000000120,    -0.00000000016,
       0.00000002662,    -0.00000000460,
      -0.00000000019,    -0.00000000108,
       0.00000000004,    -0.00000000001,
       0.00000000010,     0.00000000006,
      -0.00000000006,     0.00000000010,
       0.00000000018,    -0.00000000066,
       0.00000000066,     0.00000000018,
       0.00000031747,    -0.00000106017,
       0.00000032080,     0.00000004825,
      -0.00000000367,     0.00000004732,
      -0.00000028574,    -0.00000001695,
       0.00000000371,    -0.00000008011,
       0.00000001182,     0.00000000013,
       0.00000051794,    -0.00000009721,
       0.00000002845,     0.00000005615,
      -0.00000000628,     0.00000000235,
       0.00000008879,     0.00000040155,
      -0.00000001124,     0.00000003289,
      -0.00000000412,     0.00000000258,
       0.00000001260,    -0.00000001896,
      -0.00000000274,    -0.00000000185,
       0.00000001878,     0.00000001322,
       0.00000000175,    -0.00000000280,
      -0.00000024135,     0.00000468697,
      -0.00000140522,    -0.00000005817,
       0.00000000818,    -0.00000020428,
       0.00000001777,     0.00000000134,
       0.00000069029,     0.00000004649,
      -0.00000000968,     0.00000019388,
      -0.00000002860,    -0.00000000030,
       0.00000000007,    -0.00000000243,
      -0.00000071926,     0.00000009028,
      -0.00000001731,    -0.00000013281,
       0.00000002174,     0.00000000111,
      -0.00000000059,     0.00000000265,
      -0.00000006692,    -0.00000019766,
      -0.00000006911,    -0.00000004154,
       0.00000000772,    -0.00000001821,
       0.00000000264,     0.00000000093,
      -0.00000003114,     0.00000004836,
       0.00000000631,     0.00000000426,
      -0.00000004617,    -0.00000003477,
      -0.00000000366,     0.00000000673,
       0.00000000269,    -0.00000000056,
       0.00000000056,     0.00000000270,
       0.00000000337,    -0.00000000070,
       0.00000000070,     0.00000000338,
       0.00000000002,     0.00000000005,
       0.00000000000,     0.00000000016,
      -0.00000000016,     0.00000000000,
       0.00000000047,     0.00000000004,
      -0.00000000004,     0.00000000047,
       0.00000063234,    -0.00000004047,
      -0.00000002225,     0.00000024052,
      -0.00000004706,    -0.00000000940,
       0.00000000176,    -0.00000000615,
      -0.00000000382,    -0.00000015969,
       0.00000006149,     0.00000000880,
      -0.00000000304,     0.00000001213,
      -0.00000000160,    -0.00000000053,
       0.00000003875,    -0.00000005989,
       0.00000002062,     0.00000001056,
      -0.00000000219,     0.00000000330,
       0.00000004993,     0.00000000833,
       0.00000000481,     0.00000001588,
      -0.00000000209,     0.00000000182,
       0.00000000258,    -0.00000000203,
       0.00000000180,     0.00000000237,
      -0.00000279457,     0.00000013547,
       0.00000012698,    -0.00000108538,
       0.00000021462,     0.00000004885,
      -0.00000000872,     0.00000002826,
       0.00000001234,     0.00000037382,
      -0.00000014357,    -0.00000001916,
       0.00000000680,    -0.00000002827,
       0.00000000372,     0.00000000120,
      -0.00000006888,     0.00000008911,
      -0.00000003249,    -0.00000003264,
       0.00000000859,    -0.00000000576,
       0.00000000074,     0.00000000152,
      -0.00000004724,     0.00000006027,
      -0.00000003264,    -0.00000001241,
       0.00000000056,    -0.00000000848,
      -0.00000000730,     0.00000000498,
      -0.00000000313,    -0.00000000586,
      -0.00000000108,     0.00000000028,
       0.00000000088,     0.00000000076,
      -0.00000000076,     0.00000000088,
       0.00000000109,     0.00000000095,
      -0.00000000095,     0.00000000110,
       0.00000000002,     0.00000000005,
       0.00000000002,     0.00000000015,
      -0.00000000015,     0.00000000002,
       0.00000000036,     0.00000000025,
      -0.00000000025,     0.00000000036,
       0.00000016914,    -0.00000003376,
       0.00000001587,     0.00000008250,
      -0.00000002101,     0.00000000417,
      -0.00000000937,    -0.00000003756,
       0.00000001870,    -0.00000000443,
       0.00000000115,     0.00000000484,
       0.00000004371,    -0.00000000367,
      -0.00000000026,     0.00000001703,
      -0.00000000347,    -0.00000000017,
      -0.00000000199,     0.00000004122,
      -0.00000001528,    -0.00000000400,
       0.00000000142,    -0.00000000284,
      -0.00000065491,     0.00000011866,
      -0.00000005554,    -0.00000031896,
       0.00000008098,    -0.00000001503,
       0.00000000292,     0.00000001399,
       0.00000002289,     0.00000008950,
      -0.00000004456,     0.00000001082,
      -0.00000000278,    -0.00000001153,
       0.00000000199,    -0.00000000050,
      -0.00000005559,     0.00000001346,
      -0.00000000511,    -0.00000002333,
       0.00000000519,    -0.00000000160,
       0.00000001129,    -0.00000004613,
       0.00000001657,     0.00000001092,
      -0.00000000376,     0.00000000268,
       0.00000000067,     0.00000000044,
       0.00000000077,     0.00000000068,
      -0.00000000068,     0.00000000077,
       0.00000000097,     0.00000000085,
      -0.00000000085,     0.00000000097,
       0.00000001745,    -0.00000002300,
       0.00000001557,     0.00000001102,
      -0.00000000366,     0.00000000534,
      -0.00000000592,    -0.00000000285,
       0.00000000196,    -0.00000000393,
       0.00000000132,     0.00000000069,
       0.00000001188,    -0.00000000158,
       0.00000000063,     0.00000000591,
      -0.00000000155,     0.00000000012,
       0.00000000118,     0.00000001043,
      -0.00000000481,     0.00000000050,
      -0.00000004662,     0.00000011239,
      -0.00000006942,    -0.00000003156,
       0.00000001173,    -0.00000002326,
       0.00000001308,     0.00000000745,
      -0.00000000510,     0.00000000939,
      -0.00000000325,    -0.00000000172,
      -0.00000001509,     0.00000000181,
      -0.00000000060,    -0.00000000815,
       0.00000000236,    -0.00000000004,
      -0.00000000040,    -0.00000000868,
       0.00000000322,    -0.00000000032,
      -0.00000000053,     0.00000000076,
      -0.00000000068,    -0.00000000059,
       0.00000000300,     0.00000000030,
       0.00000000029,    -0.00000000294,
       0.00000007494,     0.00000000759,
       0.00000000032,    -0.00000000371,
      -0.00000001052,    -0.00000000062,
      -0.00000000060,     0.00000001026,
       0.00000000006,    -0.00000000121,
      -0.00000026008,    -0.00000001538,
      -0.00000000068,     0.00000001370,
       0.00000002693,     0.00000000031,
       0.00000000023,    -0.00000002573,
      -0.00000001299,    -0.00000000013,
      -0.00000000003,     0.00000000340,
       0.00000000136,    -0.00000000004,
       0.00000065273,     0.00000000628,
       0.00000000029,    -0.00000003703,
      -0.00000003564,     0.00000000221,
       0.00000000238,     0.00000003117,
       0.00000001907,    -0.00000000136,
      -0.00000000028,    -0.00000000495,
      -0.00000000128,     0.00000000007,
       0.00000000008,    -0.00000000236,
      -0.00000082259,     0.00000005911,
       0.00000000316,     0.00000005156,
      -0.00000001488,    -0.00000000289,
      -0.00000000033,    -0.00000000092,
      -0.00000000294,     0.00000002723,
       0.00000001322,     0.00000000207,
       0.00000000044,    -0.00000000290,
       0.00000000328,    -0.00000000012,
       0.00000000022,     0.00000000301,
      -0.00000047776,    -0.00000007500,
      -0.00000000800,    -0.00000001779,
      -0.00000000468,     0.00000003137,
      -0.00000000115,     0.00000000050,
      -0.00000001383,    -0.00000000117,
      -0.00000000217,    -0.00000000402,
       0.00000000200,     0.00000000036,
      -0.00000000034,     0.00000000197,
      -0.00000010158,    -0.00000003935,
      -0.00000000258,     0.00000000714,
      -0.00000000524,    -0.00000000117,
       0.00000000045,    -0.00000000402,
      -0.00000000055,     0.00000000086,
      -0.00000000092,    -0.00000000044,
      -0.00000000113,    -0.00000000030,
      -0.00000000006,     0.00000000094,
      -0.00000000031,     0.00000000112,
       0.00000000095,     0.00000000006,
       0.00000000056,     0.00000000051,
      -0.00000000247,    -0.00000000110,
      -0.00000000115,     0.00000000244,
       0.00000000376,     0.00000000176,
       0.00000000048,    -0.00000000126,
      -0.00000000108,     0.00000000063,
       0.00000000062,     0.00000000107,
      -0.00000002268,    -0.00000001102,
      -0.00000000068,     0.00000001000,
      -0.00000000122,     0.00000000305,
       0.00000000129,     0.00000000003,
       0.00000000476,    -0.00000000270,
      -0.00000000120,    -0.00000000076,
      -0.00000000276,    -0.00000000471,
      -0.00000000073,     0.00000000124,
      -0.00000000832,     0.00000000504,
       0.00000000159,     0.00000000247,
       0.00000000025,    -0.00000000131,
      -0.00000000140,     0.00000000013,
       0.00000004439,    -0.00000002709,
      -0.00000001200,    -0.00000000677,
      -0.00000000379,    -0.00000000593,
      -0.00000000090,     0.00000000160,
      -0.00000000058,     0.00000000599,
       0.00000000618,    -0.00000000005,
      -0.00000000075,    -0.00000000056,
      -0.00000000126,    -0.00000001340,
      -0.00000000360,    -0.00000000009,
       0.00000000078,     0.00000005734,
       0.00000000491,    -0.00000000651,
       0.00000000834,     0.00000000024,
      -0.00000000089,    -0.00000000076,
       0.00000000204,     0.00000000031,
       0.00000000120,    -0.00000000089,
       0.00000004310,     0.00000003016,
       0.00000000723,    -0.00000001773,
       0.00000000199,    -0.00000000320,
      -0.00000000131,    -0.00000000044,
       0.00000000169,     0.00000000020,
       0.00000000099,     0.00000000192,
       0.00000000065,    -0.00000000039,
      -0.00000000038,     0.00000000083,
       0.00000000135,    -0.00000000012,
       0.00000000887,     0.00000000934,
       0.00000000096,    -0.00000000193,
      -0.00000000052,    -0.00000000189,
      -0.00000000368,     0.00000000059,
       0.00000000117,     0.00000000560,
       0.00000000149,    -0.00000000033,
      -0.00000000734,    -0.00000002759,
      -0.00000000358,     0.00000000080,
       0.00000000293,     0.00000000046,
       0.00000000014,    -0.00000000267,
      -0.00000000631,    -0.00000000053,
      -0.00000000020,     0.00000000168,
       0.00000002927,     0.00000000251,
      -0.00000000456,     0.00000000625,
       0.00000000044,    -0.00000000395,
       0.00000000083,     0.00000000063,
       0.00000000080,    -0.00000000072,
       0.00000000013,    -0.00000000110,
       0.00000000973,    -0.00000000198,
      -0.00000000035,    -0.00000000128,
      -0.00000001089,    -0.00000002972,
      -0.00000001887,     0.00000001264,
       0.00000000111,     0.00000000179,
       0.00000000129,    -0.00000000101,
       0.00000000077,     0.00000000946,
       0.00000000626,    -0.00000000165,
      -0.00000000098,    -0.00000000206,
       0.00000000948,    -0.00000000068,
      -0.00000000159,    -0.00000000628,
      -0.00000000206,     0.00000000098,
       0.00000000073,     0.00000000107,
       0.00000000107,    -0.00000000073,
      -0.00000000093,    -0.00000000041,
      -0.00000000041,     0.00000000093,
       0.00000000042,    -0.00000000059,
      -0.00000000129,     0.00000000058,
       0.00000000061,     0.00000000129,
      -0.00000000120,     0.00000000012,
       0.00000000013,     0.00000000122,
      -0.00000001163,     0.00000000507,
       0.00000000082,     0.00000000153,
       0.00000000002,    -0.00000000182,
      -0.00000000189,     0.00000000011,
      -0.00000000082,    -0.00000000093,
      -0.00000000092,     0.00000000084,
       0.00000000103,    -0.00000000003,
       0.00000000029,    -0.00000001534,
      -0.00000000241,     0.00000000013,
       0.00000000338,    -0.00000000335,
      -0.00000000064,    -0.00000000039,
      -0.00000000385,    -0.00000000274,
      -0.00000000030,    -0.00000000074,
       0.00000000063,    -0.00000000148,
      -0.00000000148,    -0.00000000064,
       0.00000000380,    -0.00000000890,
       0.00000000040,     0.00000000073,
       0.00000000071,    -0.00000000059,
       0.00000000074,    -0.00000000051,
      -0.00000000175,    -0.00000000028,
       0.00000000019,     0.00000000196,
      -0.00000001327,    -0.00000000047,
       0.00000000002,     0.00000000221,
       0.00000000166,    -0.00000000240,
      -0.00000000249,    -0.00000000160,
      -0.00000000250,     0.00000000385,
       0.00000000145,     0.00000000096,
       0.00000001142,    -0.00000001766,
      -0.00000000326,    -0.00000000214,
       0.00000001246,     0.00000003707,
      -0.00000000185,     0.00000000062,
       0.00000001637,    -0.00000000531,
       0.00000005465,     0.00000068568,
       0.00000045297,    -0.00000011915,
      -0.00000007097,    -0.00000014764,
      -0.00000003052,     0.00000002363,
      -0.00000003196,     0.00000000261,
       0.00000000560,     0.00000002114,
       0.00000000688,    -0.00000000332,
      -0.00000000112,    -0.00000000143,
       0.00000005222,     0.00000007670,
       0.00000003825,    -0.00000003272,
      -0.00000001016,    -0.00000000942,
      -0.00000000368,     0.00000000251,
       0.00000000156,     0.00000000183,
      -0.00000006813,    -0.00000003016,
      -0.00000001288,     0.00000002970,
       0.00000000144,    -0.00000000326,
      -0.00000000142,    -0.00000000062,
       0.00000000021,     0.00000000086,
       0.00000000088,    -0.00000000023,
      -0.00000008853,     0.00000000799,
       0.00000000340,     0.00000003016,
      -0.00000000046,    -0.00000000436,
      -0.00000000146,     0.00000000017,
       0.00000000196,     0.00000000787,
       0.00000000107,    -0.00000000029,
       0.00000000011,     0.00000000095,
      -0.00000006953,    -0.00000006516,
       0.00000000328,    -0.00000000379,
       0.00000000009,     0.00000000114,
       0.00000022485,    -0.00000026961,
      -0.00000005286,    -0.00000001919,
       0.00000001330,     0.00000001123,
       0.00000000095,    -0.00000000261,
       0.00000007130,    -0.00000001477,
       0.00000000073,     0.00000000352,
       0.00000002309,     0.00000002012,
      -0.00000000100,     0.00000000114,
       0.00000002039,     0.00000001807,
      -0.00000000089,     0.00000000101,
       0.00000000159,    -0.00000000002,
      -0.00000000085,     0.00000000029,
      -0.00000000053,     0.00000000098,
       0.00000000096,     0.00000000054,
      -0.00000000544,     0.00000000970,
       0.00000000123,     0.00000000062,
      -0.00000000124,    -0.00000000200,
      -0.00000000198,     0.00000000126,
       0.00000000219,     0.00000000351,
       0.00000000106,    -0.00000000071,
      -0.00000001288,    -0.00000002019,
      -0.00000000258,     0.00000000173,
       0.00000000274,    -0.00000000014,
      -0.00000000041,    -0.00000000282,
      -0.00000000566,     0.00000000061,
       0.00000000015,     0.00000000165,
       0.00000002822,    -0.00000000301,
      -0.00000000035,    -0.00000000389,
      -0.00000000072,    -0.00000000070,
      -0.00000000068,     0.00000000077,
      -0.00000000631,    -0.00000000587,
      -0.00000000084,     0.00000000095,
       0.00000000067,    -0.00000000034,
      -0.00000000039,    -0.00000000068,
      -0.00000000038,    -0.00000000086,
       0.00000000121,    -0.00000000220,
      -0.00000000211,    -0.00000000131,
      -0.00000005035,    -0.00000001338,
      -0.00000000037,     0.00000003180,
       0.00000000063,    -0.00000000232,
      -0.00000000147,    -0.00000000003,
      -0.00000000202,     0.00000000324,
       0.00000000114,     0.00000000058,
      -0.00000000083,    -0.00000000104,
      -0.00000000102,     0.00000000079,
       0.00000001361,    -0.00000002446,
      -0.00000000718,    -0.00000000265,
      -0.00000000298,    -0.00000000152,
       0.00000000350,     0.00000000444,
       0.00000000443,    -0.00000000345,
      -0.00000000591,    -0.00000000771,
      -0.00000000241,     0.00000000196,
       0.00000000159,    -0.00000000022,
      -0.00000000053,    -0.00000000168,
       0.00000003812,     0.00000005033,
       0.00000000515,    -0.00000000698,
       0.00000000597,    -0.00000000489,
      -0.00000000085,    -0.00000000063,
      -0.00000000709,     0.00000000126,
       0.00000000190,     0.00000000729,
       0.00000001421,    -0.00000000303,
      -0.00000000088,    -0.00000000426,
       0.00000000077,    -0.00000000064,
      -0.00000008110,     0.00000001672,
       0.00000000214,     0.00000001027,
      -0.00000000070,    -0.00000000132,
      -0.00000000256,     0.00000000186,
       0.00000000297,     0.00000000418,
       0.00000000121,    -0.00000000073,
      -0.00000001415,    -0.00000001966,
      -0.00000000280,     0.00000000178,
      -0.00000000135,     0.00000000023,
      -0.00000000027,    -0.00000000104,
      -0.00000000061,    -0.00000000047,
       0.00000002044,    -0.00000002394,
      -0.00000002915,    -0.00000001594,
      -0.00000000057,    -0.00000000049,
      -0.00000000041,     0.00000000075,
       0.00000002108,    -0.00000002495,
      -0.00000002962,    -0.00000001676,
      -0.00000000474,     0.00000001748,
       0.00000000587,    -0.00000000020,
       0.00000040869,     0.00000033538,
       0.00000025904,    -0.00000048438,
      -0.00000028471,    -0.00000006947,
       0.00000000293,     0.00000008715,
       0.00000001560,     0.00000004876,
       0.00000003725,    -0.00000002347,
      -0.00000001067,    -0.00000001077,
       0.00000004990,    -0.00000001470,
      -0.00000002285,    -0.00000003851,
      -0.00000001079,     0.00000001062,
      -0.00000024500,     0.00000027945,
       0.00000034013,     0.00000019014,
       0.00000005231,    -0.00000018590,
      -0.00000005564,    -0.00000000149,
      -0.00000100092,    -0.00000084343,
      -0.00000063850,     0.00000120038,
       0.00000064053,     0.00000016836,
       0.00000000499,    -0.00000018145,
      -0.00000005350,    -0.00000018630,
      -0.00000014874,     0.00000008055,
       0.00000004532,     0.00000005101,
      -0.00000018486,     0.00000005341,
       0.00000008227,     0.00000014580,
       0.00000004908,    -0.00000004991,
       0.00000000672,    -0.00000003259,
      -0.00000002720,    -0.00000000312,
       0.00000001992,     0.00000000397,
       0.00000000177,    -0.00000001538,
      -0.00000000231,    -0.00000000891,
      -0.00000000762,     0.00000000386,
       0.00000000251,     0.00000000296,
      -0.00000000913,     0.00000000304,
       0.00000000439,     0.00000000764,
       0.00000000286,    -0.00000000259,
       0.00000000834,     0.00000000291,
       0.00000000071,    -0.00000000694,
      -0.00000000227,     0.00000000041,
       0.00000000258,    -0.00000000854,
      -0.00000000700,    -0.00000000045,
       0.00000000036,     0.00000000225,
       0.00000001993,     0.00000007577,
       0.00000006525,    -0.00000003258,
      -0.00000002104,    -0.00000002574,
      -0.00000000544,     0.00000000668,
       0.00000002219,    -0.00000000693,
      -0.00000001060,    -0.00000001925,
      -0.00000000753,     0.00000000656,
       0.00000000214,     0.00000000154,
      -0.00000003563,    -0.00000001096,
      -0.00000000238,     0.00000002947,
       0.00000001088,    -0.00000000156,
      -0.00000000059,    -0.00000000160,
      -0.00000001166,     0.00000003486,
       0.00000002877,     0.00000000333,
      -0.00000000132,    -0.00000001074,
      -0.00000000151,     0.00000000067,
       0.00000002197,     0.00000000727,
       0.00000000181,    -0.00000001832,
      -0.00000000691,     0.00000000084,
       0.00000000376,    -0.00000001376,
      -0.00000001126,    -0.00000000068,
       0.00000000063,     0.00000000400,
      -0.00000000829,     0.00000000697,
       0.00000000577,     0.00000000423,
       0.00000000079,    -0.00000000201,
       0.00000000691,     0.00000000810,
       0.00000000420,    -0.00000000559,
      -0.00000000185,    -0.00000000086,
      -0.00000015011,    -0.00000004580,
      -0.00000000983,     0.00000012459,
       0.00000004710,    -0.00000000641,
      -0.00000000294,    -0.00000000900,
      -0.00000000889,     0.00000003210,
       0.00000002635,     0.00000000160,
      -0.00000000161,    -0.00000001011,
      -0.00000000207,     0.00000000073,
       0.00000003540,    -0.00000002891,
      -0.00000002395,    -0.00000001861,
      -0.00000000384,     0.00000000852,
      -0.00000002880,    -0.00000003504,
      -0.00000001825,     0.00000002398,
       0.00000000826,     0.00000000392,
      -0.00000003888,     0.00000003309,
       0.00000002715,     0.00000001996,
       0.00000000410,    -0.00000000925,
       0.00000001528,     0.00000001824,
       0.00000000940,    -0.00000001256,
      -0.00000000430,    -0.00000000196,
      -0.00000000032,     0.00000000202,
       0.00000000113,    -0.00000000010,
       0.00000000272,     0.00000000036,
      -0.00000000038,    -0.00000000143,
       0.00000016552,    -0.00000013643,
      -0.00000011205,    -0.00000008604,
      -0.00000001805,     0.00000003820,
      -0.00000003330,    -0.00000003944,
      -0.00000002024,     0.00000002735,
       0.00000000937,     0.00000000418,
      -0.00000000296,    -0.00000000855,
      -0.00000000392,     0.00000000334,
       0.00000000118,     0.00000000072,
      -0.00000000875,     0.00000000169,
       0.00000000251,     0.00000000422,
       0.00000000082,    -0.00000000105,
      -0.00000000099,     0.00000001122,
       0.00000000596,    -0.00000000129,
       0.00000000463,     0.00000000005,
      -0.00000000071,    -0.00000000240,
      -0.00000000317,    -0.00000000367,
      -0.00000000101,     0.00000000163,
      -0.00000000314,     0.00000000375,
       0.00000000190,     0.00000000056,
      -0.00000005283,    -0.00000006241,
      -0.00000002429,     0.00000004344,
       0.00000001412,     0.00000000419,
      -0.00000001263,    -0.00000000520,
      -0.00000000126,     0.00000000735,
       0.00000000203,    -0.00000000013,
       0.00000001548,     0.00000001427,
       0.00000000310,    -0.00000000794,
      -0.00000000185,    -0.00000000014,
       0.00000001511,    -0.00000001429,
      -0.00000000727,    -0.00000000385,
      -0.00000000028,     0.00000000170,
      -0.00000002224,    -0.00000002092,
      -0.00000000478,     0.00000001128,
      -0.00000000737,     0.00000000840,
       0.00000000414,     0.00000000166,
       0.00000000599,    -0.00000000024,
      -0.00000000073,    -0.00000000134,
      -0.00000000104,    -0.00000000582,
      -0.00000000107,     0.00000000107,
       0.00000012543,     0.00000011730,
       0.00000002674,    -0.00000006369,
      -0.00000001492,    -0.00000000120,
       0.00000001765,    -0.00000001990,
      -0.00000000981,    -0.00000000398,
      -0.00000000017,     0.00000000226,
      -0.00000002802,     0.00000000196,
       0.00000000369,     0.00000000659,
       0.00000000094,    -0.00000000092,
       0.00000000363,     0.00000002436,
       0.00000000420,    -0.00000000405,
      -0.00000000248,    -0.00000000018,
       0.00000000023,     0.00000000153,
       0.00000000018,    -0.00000000248,
      -0.00000000152,     0.00000000022,
      -0.00000000310,    -0.00000000023,
       0.00000000028,     0.00000000191,
       0.00000000024,    -0.00000000310,
      -0.00000000191,     0.00000000028,
      -0.00000012960,    -0.00000000850,
       0.00000001341,     0.00000007996,
       0.00000000027,    -0.00000000406,
      -0.00000000251,     0.00000000042,
       0.00000004764,    -0.00000000624,
      -0.00000000741,    -0.00000000964,
      -0.00000000176,    -0.00000001578,
      -0.00000000324,     0.00000000237,
       0.00000000006,    -0.00000000273,
      -0.00000000142,    -0.00000000205,
      -0.00000020657,     0.00000002691,
       0.00000003231,     0.00000004150,
       0.00000000412,    -0.00000000764,
       0.00000000395,     0.00000003567,
       0.00000000726,    -0.00000000538,
      -0.00000000127,    -0.00000000073,
      -0.00000000286,     0.00000000974,
       0.00000000163,    -0.00000000079,
       0.00000000777,     0.00000000576,
       0.00000000051,    -0.00000000122,
       0.00000000874,    -0.00000001523,
      -0.00000000477,    -0.00000000228,
       0.00000000307,     0.00000000108,
       0.00000000158,     0.00000000143,
      -0.00000004045,     0.00000008349,
       0.00000001353,    -0.00000000126,
       0.00000001164,     0.00000000548,
      -0.00000000020,    -0.00000000187,
      -0.00000000593,    -0.00000000560,
      -0.00000000611,     0.00000000108,
       0.00000000907,     0.00000001145,
       0.00000000359,    -0.00000000259,
       0.00000000103,    -0.00000000217,
      -0.00000004665,    -0.00000006169,
      -0.00000000856,     0.00000000621,
      -0.00000000325,     0.00000000761,
      -0.00000000026,    -0.00000000114,
       0.00000000527,     0.00000000241,
      -0.00000000108,     0.00000000061,
       0.00000000584,    -0.00000001243,
      -0.00000000360,    -0.00000000160,
      -0.00000000104,     0.00000000056,
      -0.00000002502,     0.00000005732,
      -0.00000000405,    -0.00000001002,
       0.00000000834,     0.00000000386,
      -0.00000000144,     0.00000000057,
      -0.00000000085,     0.00000000037,
       0.00000000241,    -0.00000000084,
       0.00000000070,     0.00000000032,
      -0.00000000206,     0.00000001227,
       0.00000000174,     0.00000000013,
      -0.00000000031,     0.00000000073,
      -0.00000000092,    -0.00000000040,
      -0.00000000144,     0.00000000004,
       0.00000000004,     0.00000000144,
      -0.00000005629,     0.00000000173,
       0.00000000006,     0.00000000211,
      -0.00000000085,    -0.00000000017,
      -0.00000000020,     0.00000000088,
      -0.00000003420,    -0.00000000767,
      -0.00000000029,     0.00000000129,
      -0.00000000112,    -0.00000000006,
      -0.00000000006,     0.00000000112,
       0.00000000090,     0.00000001115,
       0.00000000738,    -0.00000000195,
      -0.00000000117,    -0.00000000242,
       0.00000001116,    -0.00000000091,
      -0.00000000195,    -0.00000000738,
      -0.00000000242,     0.00000000117,
       0.00000000086,     0.00000000126,
       0.00000000063,    -0.00000000054,
       0.00000000126,    -0.00000000086,
      -0.00000000054,    -0.00000000063,
      -0.00000000111,    -0.00000000049,
      -0.00000000049,     0.00000000111,
      -0.00000000146,     0.00000000016,
       0.00000000016,     0.00000000146,
      -0.00000000121,    -0.00000000110,
      -0.00000000110,     0.00000000121,
      -0.00000000431,    -0.00000004999,
       0.00000000375,    -0.00000000437,
      -0.00000000085,    -0.00000000033,
      -0.00000000443,    -0.00000000369,
      -0.00000000030,     0.00000000087,
       0.00000000116,    -0.00000000024,
      -0.00000000024,    -0.00000000116,
       0.00000000156,     0.00000000011,
       0.00000000010,    -0.00000000149,
       0.00000003323,     0.00000000222,
       0.00000000012,    -0.00000000188,
      -0.00000000173,     0.00000000001,
       0.00000000004,     0.00000000149,
      -0.00000003441,     0.00000000060,
       0.00000000001,     0.00000000214,
      -0.00000000019,     0.00000000144,
      -0.00000002193,    -0.00000000364,
      -0.00000000023,     0.00000000148,
      -0.00000000054,     0.00000000065,
       0.00000000060,     0.00000000055,
       0.00000000071,     0.00000000066,
      -0.00000000028,    -0.00000000113,
      -0.00000000112,     0.00000000022,
      -0.00000000128,     0.00000000028,
       0.00000000123,    -0.00000000083,
      -0.00000000083,    -0.00000000123,
       0.00000000044,     0.00000000062,
       0.00000001023,    -0.00000000677,
      -0.00000000107,    -0.00000000148,
       0.00000000007,     0.00000000212,
       0.00000000209,    -0.00000000027,
      -0.00000000127,     0.00000000009,
       0.00000000167,     0.00000001721,
       0.00000000286,    -0.00000000023,
       0.00000002419,     0.00000007193,
      -0.00000000234,     0.00000000079,
       0.00000002474,    -0.00000001903,
       0.00000000049,     0.00000000064,
       0.00000003093,    -0.00000000909,
      -0.00000001057,    -0.00000002363,
       0.00000000028,     0.00000000095,
       0.00000000072,    -0.00000000033,
       0.00000010502,     0.00000130150,
       0.00000086087,    -0.00000022674,
      -0.00000013498,    -0.00000028016,
      -0.00000005784,     0.00000004491,
      -0.00000004080,     0.00000000331,
       0.00000000712,     0.00000002698,
       0.00000000879,    -0.00000000424,
      -0.00000000141,    -0.00000000181,
       0.00000009965,     0.00000014639,
       0.00000007240,    -0.00000006163,
      -0.00000001935,    -0.00000001794,
      -0.00000000468,     0.00000000320,
       0.00000000198,     0.00000000231,
       0.00000000057,    -0.00000000061,
      -0.00000012941,    -0.00000005814,
      -0.00000002471,     0.00000005689,
       0.00000000184,    -0.00000000415,
      -0.00000000181,    -0.00000000079,
      -0.00000017029,     0.00000001545,
       0.00000000574,     0.00000005773,
      -0.00000000058,    -0.00000000550,
      -0.00000000186,     0.00000000021,
      -0.00000013897,    -0.00000012519,
      -0.00000001067,     0.00000002559,
       0.00000000417,    -0.00000000457,
      -0.00000000084,    -0.00000000036,
       0.00000043374,    -0.00000052064,
      -0.00000010206,    -0.00000003699,
       0.00000001682,     0.00000001421,
       0.00000000119,    -0.00000000329,
       0.00000013693,    -0.00000002929,
       0.00000000092,     0.00000000445,
       0.00000004472,     0.00000003894,
      -0.00000000126,     0.00000000145,
       0.00000003952,     0.00000003482,
      -0.00000000113,     0.00000000128,
       0.00000000380,     0.00000000584,
       0.00000000064,    -0.00000000045,
      -0.00000000989,     0.00000000113,
       0.00000000014,     0.00000000118,
       0.00000000216,    -0.00000000016,
      -0.00000000015,    -0.00000000223,
       0.00000004715,    -0.00000000310,
      -0.00000000018,    -0.00000000290,
       0.00000000010,    -0.00000000112,
       0.00000001994,     0.00000000253,
       0.00000000015,    -0.00000000129,
       0.00000000020,     0.00000000091,
      -0.00000000053,    -0.00000000210,
       0.00000000199,    -0.00000000054,
       0.00000000055,    -0.00000000090,
       0.00000000087,     0.00000000070,
      -0.00000000029,    -0.00000000106,
       0.00000000111,    -0.00000000033,
       0.00000000032,    -0.00000000050,
       0.00000000051,     0.00000000033,
      -0.00000000550,    -0.00000000410,
       0.00000000047,    -0.00000000087,
       0.00000000663,     0.00000000347,
      -0.00000000101,    -0.00000000086,
      -0.00000000089,     0.00000000171,
       0.00000000155,     0.00000000032,
      -0.00000000185,    -0.00000000267,
      -0.00000000298,     0.00000000288,
       0.00000000280,     0.00000000179,
      -0.00000000250,     0.00000000192,
       0.00000000299,     0.00000000272,
       0.00000000149,    -0.00000000288,
      -0.00000000455,    -0.00000000507,
      -0.00000000568,     0.00000000694,
       0.00000000528,     0.00000000290,
      -0.00000001005,     0.00000000975,
       0.00000001443,     0.00000001109,
       0.00000000534,    -0.00000000966,
      -0.00000000561,    -0.00000000208,
       0.00000000521,    -0.00000000233,
      -0.00000000353,    -0.00000000596,
      -0.00000000328,     0.00000000298,
       0.00000000159,     0.00000000108,
      -0.00000000231,    -0.00000000535,
      -0.00000000613,     0.00000000358,
       0.00000000297,     0.00000000328,
       0.00000000108,    -0.00000000159,
       0.00000005728,     0.00000006204,
       0.00000006984,    -0.00000008730,
      -0.00000006481,    -0.00000003666,
      -0.00000001179,     0.00000003274,
       0.00000001102,     0.00000000118,
       0.00000002373,    -0.00000002359,
      -0.00000003591,    -0.00000002769,
      -0.00000001480,     0.00000002600,
       0.00000001316,     0.00000000479,
       0.00000000048,    -0.00000000450,
      -0.00000004111,     0.00000001917,
       0.00000002912,     0.00000004571,
       0.00000002530,    -0.00000002188,
      -0.00000001024,    -0.00000000828,
       0.00000001950,     0.00000004091,
       0.00000004546,    -0.00000002949,
      -0.00000002216,    -0.00000002466,
      -0.00000000810,     0.00000001066,
       0.00000000065,    -0.00000000056,
      -0.00000000056,    -0.00000000065,
       0.00000001228,    -0.00000001165,
      -0.00000000070,     0.00000000051,
      -0.00000000051,    -0.00000000070,
      -0.00000000088,     0.00000000064,
      -0.00000000064,    -0.00000000088,
       0.00000001085,    -0.00000000499,
      -0.00000000767,    -0.00000001209,
      -0.00000000658,     0.00000000562,
      -0.00000000452,    -0.00000000775,
      -0.00000000849,     0.00000000648,
       0.00000000459,     0.00000000459,
       0.00000000160,    -0.00000000202,
      -0.00000000152,     0.00000001152,
       0.00000001238,    -0.00000000002,
      -0.00000000095,    -0.00000000670,
      -0.00000000221,     0.00000000072,
       0.00000001151,     0.00000000133,
      -0.00000000019,    -0.00000001234,
      -0.00000000668,     0.00000000108,
       0.00000000066,     0.00000000224,
      -0.00000015870,     0.00000007781,
       0.00000011739,     0.00000017665,
       0.00000009648,    -0.00000008505,
      -0.00000004000,    -0.00000003437,
      -0.00000000775,     0.00000001296,
       0.00000000935,     0.00000001617,
       0.00000001793,    -0.00000001345,
      -0.00000000957,    -0.00000000998,
      -0.00000000389,     0.00000000453,
       0.00000000620,    -0.00000004980,
      -0.00000005356,     0.00000000016,
       0.00000000369,     0.00000002886,
       0.00000000968,    -0.00000000280,
      -0.00000000100,    -0.00000000206,
      -0.00000004990,    -0.00000000575,
       0.00000000085,     0.00000005371,
       0.00000002862,    -0.00000000438,
      -0.00000000318,    -0.00000000966,
      -0.00000000201,     0.00000000109,
      -0.00000000424,     0.00000003375,
       0.00000003637,    -0.00000000022,
      -0.00000000281,    -0.00000001946,
      -0.00000000685,     0.00000000204,
       0.00000001841,     0.00000000093,
      -0.00000000143,    -0.00000001966,
      -0.00000001069,     0.00000000211,
       0.00000000123,     0.00000000366,
      -0.00000001416,    -0.00000001122,
      -0.00000000874,     0.00000001437,
       0.00000000717,     0.00000000328,
       0.00000000070,    -0.00000000237,
      -0.00000001070,     0.00000001450,
       0.00000001471,     0.00000000812,
       0.00000000280,    -0.00000000749,
      -0.00000000239,    -0.00000000060,
       0.00000002776,    -0.00000024153,
      -0.00000025993,     0.00000000398,
       0.00000002085,     0.00000013887,
       0.00000004894,    -0.00000001450,
      -0.00000000566,    -0.00000001227,
      -0.00000004260,    -0.00000000223,
       0.00000000316,     0.00000004550,
       0.00000002413,    -0.00000000479,
      -0.00000000291,    -0.00000000870,
      -0.00000000249,     0.00000000125,
       0.00000005996,     0.00000004740,
       0.00000003706,    -0.00000006081,
      -0.00000003032,    -0.00000001390,
      -0.00000000319,     0.00000001015,
       0.00000000242,     0.00000000031,
       0.00000004592,    -0.00000006173,
      -0.00000006271,    -0.00000003506,
      -0.00000001238,     0.00000003166,
       0.00000001060,     0.00000000244,
       0.00000000037,    -0.00000000238,
      -0.00000000085,     0.00000000048,
       0.00000000070,     0.00000000094,
       0.00000000050,    -0.00000000050,
      -0.00000000047,    -0.00000000086,
      -0.00000000094,     0.00000000069,
      -0.00000000107,     0.00000000061,
       0.00000000088,     0.00000000117,
       0.00000000063,    -0.00000000063,
      -0.00000000059,    -0.00000000107,
      -0.00000000118,     0.00000000086,
       0.00000000061,     0.00000000063,
      -0.00000006226,    -0.00000004741,
      -0.00000003647,     0.00000006322,
       0.00000003186,     0.00000001311,
       0.00000000271,    -0.00000001059,
      -0.00000002031,     0.00000002689,
       0.00000002732,     0.00000001556,
       0.00000000570,    -0.00000001379,
      -0.00000000443,    -0.00000000119,
       0.00000000057,    -0.00000000355,
      -0.00000000235,    -0.00000000001,
      -0.00000000469,    -0.00000000002,
       0.00000000062,     0.00000000342,
       0.00000000122,    -0.00000000040,
      -0.00000000074,     0.00000000027,
       0.00000030248,     0.00000021122,
       0.00000016034,    -0.00000030384,
      -0.00000015202,    -0.00000005658,
      -0.00000001119,     0.00000005015,
       0.00000001183,     0.00000000100,
       0.00000004482,    -0.00000006045,
      -0.00000006135,    -0.00000003452,
      -0.00000001235,     0.00000003088,
       0.00000001035,     0.00000000253,
       0.00000000027,    -0.00000000265,
      -0.00000000435,     0.00000001601,
       0.00000001120,     0.00000000148,
       0.00000000009,    -0.00000000395,
       0.00000001852,     0.00000000376,
       0.00000000061,    -0.00000001373,
      -0.00000000509,     0.00000000039,
       0.00000000188,    -0.00000000058,
      -0.00000000053,    -0.00000000103,
      -0.00000000047,    -0.00000000188,
      -0.00000000105,     0.00000000044,
       0.00000002701,    -0.00000003291,
      -0.00000002662,    -0.00000001756,
      -0.00000000569,     0.00000001098,
      -0.00000000356,    -0.00000000948,
      -0.00000000675,     0.00000000377,
       0.00000000186,     0.00000000246,
       0.00000003677,    -0.00000011282,
      -0.00000007841,    -0.00000001371,
      -0.00000000057,     0.00000002722,
       0.00000000619,    -0.00000000087,
      -0.00000011147,    -0.00000003931,
      -0.00000001613,     0.00000007739,
       0.00000002683,     0.00000000158,
      -0.00000000073,    -0.00000000613,
       0.00000000035,    -0.00000000122,
      -0.00000000122,    -0.00000000026,
       0.00000043727,     0.00000001548,
      -0.00000004820,    -0.00000035085,
      -0.00000013788,     0.00000004373,
       0.00000001844,     0.00000003525,
       0.00000000651,    -0.00000000515,
       0.00000000174,     0.00000000078,
      -0.00000000061,    -0.00000000144,
      -0.00000016076,     0.00000048177,
       0.00000033462,     0.00000006231,
       0.00000000405,    -0.00000011605,
      -0.00000002651,     0.00000000345,
       0.00000000136,     0.00000000443,
       0.00000047967,     0.00000016295,
       0.00000006461,    -0.00000033290,
      -0.00000011539,    -0.00000000518,
       0.00000000325,     0.00000002634,
       0.00000000437,    -0.00000000143,
      -0.00000000085,     0.00000000292,
       0.00000000120,     0.00000000020,
       0.00000000297,     0.00000000070,
       0.00000000009,    -0.00000000123,
       0.00000023953,    -0.00000068163,
      -0.00000047447,    -0.00000009548,
      -0.00000000828,     0.00000016471,
       0.00000003756,    -0.00000000420,
      -0.00000000189,    -0.00000000626,
      -0.00000024652,    -0.00000006711,
      -0.00000002365,     0.00000016956,
       0.00000005831,    -0.00000000002,
      -0.00000000210,    -0.00000001321,
      -0.00000000223,     0.00000000074,
       0.00000017188,     0.00000019532,
       0.00000009140,    -0.00000010496,
      -0.00000003233,    -0.00000002042,
      -0.00000000272,     0.00000000655,
       0.00000019204,    -0.00000017006,
      -0.00000010339,    -0.00000008931,
      -0.00000001974,     0.00000003164,
       0.00000000637,     0.00000000253,
      -0.00000000013,    -0.00000000111,
      -0.00000145785,     0.00000432387,
       0.00000300328,     0.00000056530,
       0.00000003911,    -0.00000104049,
      -0.00000023698,     0.00000002964,
       0.00000001235,     0.00000003952,
       0.00000060650,     0.00000016336,
       0.00000005716,    -0.00000041702,
      -0.00000014351,     0.00000000028,
       0.00000000512,     0.00000003243,
       0.00000000541,    -0.00000000179,
      -0.00000073551,    -0.00000084707,
      -0.00000039759,     0.00000044755,
       0.00000013712,     0.00000008957,
       0.00000001204,    -0.00000002776,
      -0.00000000409,    -0.00000000096,
      -0.00000083649,     0.00000074589,
       0.00000045505,     0.00000038877,
       0.00000008581,    -0.00000013990,
      -0.00000002843,    -0.00000001086,
      -0.00000000064,     0.00000000421,
       0.00000000048,    -0.00000000094,
       0.00000000054,     0.00000000115,
      -0.00000000029,    -0.00000000083,
      -0.00000000024,     0.00000000088,
       0.00000000092,     0.00000000018,
       0.00000126121,     0.00000143462,
       0.00000066748,    -0.00000077118,
      -0.00000023661,    -0.00000014834,
      -0.00000001926,     0.00000004784,
       0.00000000708,     0.00000000125,
       0.00000045467,    -0.00000041689,
      -0.00000025248,    -0.00000021249,
      -0.00000004734,     0.00000007696,
       0.00000001553,     0.00000000621,
       0.00000000041,    -0.00000000224,
      -0.00000021606,     0.00000004043,
       0.00000001743,     0.00000008215,
       0.00000001642,    -0.00000000376,
      -0.00000000030,    -0.00000000226,
       0.00000009681,     0.00000022100,
       0.00000008154,    -0.00000005522,
      -0.00000001651,    -0.00000001475,
      -0.00000000143,     0.00000000338,
       0.00000000366,    -0.00000000194,
      -0.00000000090,    -0.00000000080,
       0.00000000122,    -0.00000000310,
      -0.00000659294,    -0.00000732392,
      -0.00000339482,     0.00000401152,
       0.00000122612,     0.00000075027,
       0.00000009630,    -0.00000024724,
      -0.00000003658,    -0.00000000623,
      -0.00000107276,     0.00000098848,
       0.00000059799,     0.00000050169,
       0.00000011205,    -0.00000018226,
      -0.00000003669,    -0.00000001458,
      -0.00000000092,     0.00000000538,
       0.00000098024,    -0.00000024361,
      -0.00000012411,    -0.00000037038,
      -0.00000007216,     0.00000003299,
       0.00000000554,     0.00000000915,
      -0.00000034351,    -0.00000099172,
      -0.00000037103,     0.00000019116,
       0.00000005564,     0.00000006969,
       0.00000000780,    -0.00000001087,
      -0.00000001101,     0.00000000256,
       0.00000000156,     0.00000000360,
      -0.00000000022,     0.00000001036,
       0.00000000344,    -0.00000000046,
      -0.00000000175,     0.00000000816,
       0.00000000123,     0.00000000028,
      -0.00000000013,     0.00000000006,
      -0.00000000006,    -0.00000000013,
       0.00000000036,     0.00000000050,
      -0.00000000050,     0.00000000036,
      -0.00000219904,     0.00000065019,
       0.00000035848,     0.00000082745,
       0.00000015766,    -0.00000010107,
      -0.00000001857,    -0.00000001863,
       0.00000014363,     0.00000065756,
       0.00000024994,    -0.00000009049,
      -0.00000002727,    -0.00000004816,
      -0.00000000580,     0.00000000520,
      -0.00000025005,     0.00000058911,
       0.00000018211,     0.00000002517,
      -0.00000000729,    -0.00000002942,
      -0.00000000311,     0.00000000254,
       0.00000047083,     0.00000037430,
       0.00000009853,    -0.00000012715,
      -0.00000001722,    -0.00000001470,
      -0.00000000185,     0.00000000152,
      -0.00000000147,     0.00000000608,
       0.00000000414,     0.00000000329,
       0.00000721315,    -0.00000282602,
      -0.00000140047,    -0.00000265479,
      -0.00000049925,     0.00000037012,
       0.00000006544,     0.00000005748,
       0.00000000413,    -0.00000000836,
      -0.00000032500,    -0.00000142178,
      -0.00000053875,     0.00000020118,
       0.00000006013,     0.00000010367,
       0.00000001240,    -0.00000001139,
      -0.00000000153,    -0.00000000096,
       0.00000125635,    -0.00000237963,
      -0.00000070792,    -0.00000021412,
      -0.00000000020,     0.00000010943,
       0.00000001104,    -0.00000000458,
      -0.00000221304,    -0.00000143172,
      -0.00000031811,     0.00000063106,
       0.00000009275,     0.00000003136,
       0.00000000156,    -0.00000000896,
       0.00000000684,    -0.00000001526,
      -0.00000000198,    -0.00000000082,
      -0.00000001436,    -0.00000000737,
      -0.00000000037,     0.00000000222,
      -0.00000000064,     0.00000000301,
       0.00000000195,     0.00000000019,
      -0.00000000309,    -0.00000000071,
      -0.00000000020,     0.00000000198,
      -0.00000000080,     0.00000000377,
       0.00000000244,     0.00000000024,
      -0.00000000387,    -0.00000000088,
      -0.00000000026,     0.00000000247,
      -0.00000000367,     0.00000000639,
       0.00000000078,     0.00000000041,
      -0.00000000075,    -0.00000000116,
      -0.00000000116,     0.00000000075,
      -0.00000265432,     0.00000424380,
       0.00000124440,     0.00000053977,
       0.00000003852,    -0.00000018820,
      -0.00000001875,     0.00000000142,
       0.00000130225,     0.00000068585,
       0.00000013948,    -0.00000037411,
      -0.00000005565,    -0.00000000877,
       0.00000000068,     0.00000000546,
      -0.00000081394,    -0.00000108268,
      -0.00000011667,     0.00000016565,
       0.00000002303,     0.00000000018,
      -0.00000000113,    -0.00000000220,
      -0.00000111745,     0.00000055843,
       0.00000006801,     0.00000014621,
       0.00000000995,    -0.00000000418,
       0.00000000269,    -0.00000000387,
       0.00000000065,    -0.00000000186,
      -0.00000000154,     0.00000000736,
       0.00000000028,    -0.00000000152,
       0.00001478446,    -0.00002504836,
      -0.00000732475,    -0.00000291101,
      -0.00000017852,     0.00000110177,
       0.00000010896,    -0.00000001427,
      -0.00000000407,    -0.00000000763,
      -0.00000321493,    -0.00000167788,
      -0.00000034021,     0.00000092294,
       0.00000013717,     0.00000002097,
      -0.00000000175,    -0.00000001348,
       0.00000310851,     0.00000483194,
       0.00000056943,    -0.00000055086,
      -0.00000006745,    -0.00000001872,
       0.00000000165,     0.00000000546,
       0.00000485181,    -0.00000289004,
      -0.00000046792,    -0.00000059260,
      -0.00000002701,     0.00000005138,
       0.00000000356,    -0.00000000003,
      -0.00000000336,     0.00000000215,
      -0.00000000427,     0.00000000516,
      -0.00000000057,    -0.00000000842,
       0.00000000166,     0.00000000560,
      -0.00000000057,     0.00000000073,
      -0.00000000055,    -0.00000000069,
       0.00000000063,    -0.00000000043,
      -0.00000000801,    -0.00000001223,
      -0.00000000152,     0.00000000104,
       0.00000000145,    -0.00000000010,
      -0.00000000029,    -0.00000000151,
      -0.00000643587,    -0.00001029858,
      -0.00000121865,     0.00000111266,
       0.00000012615,     0.00000004795,
      -0.00000000176,    -0.00000000932,
      -0.00000287260,     0.00000188284,
       0.00000031123,     0.00000034890,
       0.00000001390,    -0.00000003483,
      -0.00000000259,     0.00000000042,
       0.00000140957,     0.00000012434,
       0.00000006321,     0.00000002511,
       0.00000000219,    -0.00000000313,
      -0.00000036355,    -0.00000162105,
      -0.00000002138,     0.00000007358,
       0.00000001677,     0.00000000119,
      -0.00000000842,     0.00000001018,
       0.00000000056,     0.00000000089,
      -0.00000001063,     0.00000000078,
       0.00000000066,     0.00000000059,
       0.00003366253,     0.00005380075,
       0.00000636597,    -0.00000581522,
      -0.00000065868,    -0.00000024994,
       0.00000000918,     0.00000004878,
       0.00000688373,    -0.00000454025,
      -0.00000074893,    -0.00000083552,
      -0.00000003315,     0.00000008388,
       0.00000000628,    -0.00000000101,
      -0.00000661079,     0.00000006805,
      -0.00000009683,    -0.00000004219,
      -0.00000000398,    -0.00000001341,
       0.00000089250,     0.00000694532,
       0.00000002742,    -0.00000013390,
      -0.00000004689,     0.00000000003,
       0.00000000030,     0.00000000164,
       0.00000003194,    -0.00000000921,
      -0.00000000130,    -0.00000000664,
      -0.00000000105,     0.00000000049,
       0.00000001167,    -0.00000002279,
      -0.00000000562,    -0.00000000112,
      -0.00000000077,     0.00000000045,
      -0.00000000061,     0.00000000010,
       0.00000001557,    -0.00000000220,
      -0.00000000025,    -0.00000000207,
      -0.00000000004,    -0.00000000001,
       0.00000000001,    -0.00000000004,
       0.00001655539,    -0.00000145872,
      -0.00000009011,     0.00000002592,
       0.00000000633,     0.00000006930,
      -0.00000028645,    -0.00000450782,
       0.00000000485,     0.00000001842,
       0.00000001816,    -0.00000000115,
      -0.00000010729,    -0.00000075973,
      -0.00000004093,     0.00000009816,
       0.00000000620,     0.00000000641,
       0.00000043722,    -0.00000077000,
      -0.00000003997,    -0.00000010710,
       0.00000000055,     0.00000000856,
      -0.00000000424,    -0.00000001159,
       0.00000000163,    -0.00000000215,
       0.00000000453,    -0.00000000940,
       0.00000000135,     0.00000000276,
      -0.00007587776,     0.00000662445,
       0.00000043299,    -0.00000011533,
      -0.00000002918,    -0.00000031739,
       0.00000070407,     0.00001048568,
      -0.00000001198,    -0.00000004592,
      -0.00000004251,     0.00000000279,
      -0.00000074406,     0.00000155756,
      -0.00000003097,    -0.00000022261,
      -0.00000000587,    -0.00000001888,
      -0.00000019792,     0.00000202455,
      -0.00000002016,     0.00000024721,
      -0.00000001048,    -0.00000001577,
      -0.00000001555,     0.00000003012,
       0.00000000047,     0.00000000500,
       0.00000002285,     0.00000002113,
      -0.00000000025,    -0.00000000696,
      -0.00000000100,    -0.00000000118,
       0.00000000120,    -0.00000000095,
      -0.00000000126,    -0.00000000148,
       0.00000000150,    -0.00000000119,
       0.00000401868,    -0.00000131901,
       0.00000032648,     0.00000031411,
      -0.00000001468,     0.00000003636,
      -0.00000035967,    -0.00000090296,
       0.00000007584,    -0.00000008070,
       0.00000000844,     0.00000000345,
       0.00000118135,    -0.00000016295,
       0.00000001414,     0.00000000342,
       0.00000000035,     0.00000000754,
       0.00000000505,     0.00000116095,
       0.00000000003,    -0.00000003012,
      -0.00000000949,    -0.00000000123,
       0.00000000042,     0.00000000141,
       0.00000000031,    -0.00000000968,
      -0.00000000153,     0.00000000034,
      -0.00001539894,     0.00000469702,
      -0.00000124326,    -0.00000115387,
       0.00000005209,    -0.00000014071,
       0.00000086387,     0.00000214688,
      -0.00000018039,     0.00000019274,
      -0.00000002011,    -0.00000000817,
      -0.00000149942,     0.00000044199,
      -0.00000008349,    -0.00000000780,
      -0.00000000050,    -0.00000001901,
       0.00000022357,    -0.00000143807,
      -0.00000000316,     0.00000010455,
      -0.00000000807,    -0.00000000053,
       0.00000001802,     0.00000000519,
      -0.00000000063,    -0.00000000432,
       0.00000000633,     0.00000002105,
       0.00000000505,    -0.00000000052,
       0.00000000130,    -0.00000000046,
       0.00000000046,     0.00000000131,
       0.00000000163,    -0.00000000057,
       0.00000000058,     0.00000000164,
       0.00000000012,     0.00000000000,
       0.00000000000,     0.00000000012,
       0.00000027125,    -0.00000070186,
       0.00000021408,     0.00000003902,
      -0.00000000235,     0.00000003153,
      -0.00000018103,    -0.00000002811,
       0.00000000449,    -0.00000005158,
       0.00000000757,     0.00000000011,
       0.00000032081,    -0.00000007240,
       0.00000001986,     0.00000003345,
      -0.00000000348,     0.00000000199,
       0.00000006178,     0.00000025799,
      -0.00000000942,     0.00000002011,
      -0.00000000252,     0.00000000122,
       0.00000000599,    -0.00000000967,
      -0.00000000143,    -0.00000000093,
       0.00000000954,     0.00000000635,
       0.00000000087,    -0.00000000147,
      -0.00000049499,     0.00000302997,
      -0.00000091921,    -0.00000007912,
       0.00000000554,    -0.00000013386,
       0.00000001171,     0.00000000028,
       0.00000043902,     0.00000007160,
      -0.00000001125,     0.00000012510,
      -0.00000001835,    -0.00000000026,
      -0.00000044280,     0.00000008058,
      -0.00000001818,    -0.00000007556,
       0.00000001177,    -0.00000000092,
      -0.00000004289,    -0.00000016064,
      -0.00000003283,    -0.00000002235,
       0.00000000386,    -0.00000000952,
      -0.00000001495,     0.00000002495,
       0.00000000323,     0.00000000207,
      -0.00000002332,    -0.00000001709,
      -0.00000000169,     0.00000000350,
       0.00000000128,    -0.00000000063,
       0.00000000063,     0.00000000128,
       0.00000000160,    -0.00000000079,
       0.00000000079,     0.00000000161,
      -0.00000000148,     0.00000000025,
       0.00000000025,     0.00000000143,
      -0.00000003563,     0.00000000621,
       0.00000000036,     0.00000000205,
       0.00000000084,    -0.00000000025,
       0.00000000239,    -0.00000000065,
      -0.00000000037,    -0.00000000092,
      -0.00000000062,    -0.00000000219,
      -0.00000000083,     0.00000000035,
       0.00000005530,    -0.00000001555,
      -0.00000000880,    -0.00000002108,
      -0.00000000094,    -0.00000000356,
      -0.00000000136,     0.00000000055,
       0.00000000030,    -0.00000000119,
       0.00000001903,     0.00000000908,
       0.00000000058,    -0.00000000132,
       0.00000000000,     0.00000000006,
      -0.00000000006,     0.00000000000,
       0.00000035893,    -0.00000005314,
      -0.00000000106,     0.00000013450,
      -0.00000002599,    -0.00000000291,
      -0.00000000888,    -0.00000008655,
       0.00000003308,     0.00000000235,
      -0.00000000114,     0.00000000648,
       0.00000003085,    -0.00000004223,
       0.00000001448,     0.00000000756,
      -0.00000000142,     0.00000000232,
       0.00000003548,     0.00000001143,
       0.00000000222,     0.00000001130,
      -0.00000000151,     0.00000000111,
       0.00000000144,    -0.00000000154,
       0.00000000141,     0.00000000134,
      -0.00000157076,     0.00000019314,
       0.00000002671,    -0.00000060430,
       0.00000011858,     0.00000001802,
      -0.00000000352,     0.00000001557,
       0.00000002374,     0.00000020304,
      -0.00000007740,    -0.00000000444,
       0.00000000246,    -0.00000001512,
       0.00000000199,     0.00000000046,
      -0.00000005212,     0.00000006109,
      -0.00000002229,    -0.00000002208,
       0.00000000553,    -0.00000000397,
      -0.00000003265,     0.00000003123,
      -0.00000001983,    -0.00000000876,
       0.00000000045,    -0.00000000537,
      -0.00000000408,     0.00000000383,
      -0.00000000275,    -0.00000000336,
       0.00000011135,    -0.00000002587,
       0.00000001211,     0.00000005415,
      -0.00000001375,     0.00000000320,
      -0.00000000665,    -0.00000002433,
       0.00000001208,    -0.00000000313,
       0.00000000082,     0.00000000312,
       0.00000002625,    -0.00000000424,
       0.00000000075,     0.00000001011,
      -0.00000000206,     0.00000000011,
       0.00000000035,     0.00000002426,
      -0.00000000876,    -0.00000000182,
       0.00000000075,    -0.00000000159,
      -0.00000043812,     0.00000009354,
      -0.00000004364,    -0.00000021302,
       0.00000005402,    -0.00000001181,
       0.00000000230,     0.00000000931,
       0.00000001619,     0.00000005816,
      -0.00000002887,     0.00000000763,
      -0.00000000198,    -0.00000000743,
      -0.00000003436,     0.00000001063,
      -0.00000000453,    -0.00000001444,
       0.00000000324,    -0.00000000137,
       0.00000000593,    -0.00000002635,
       0.00000000899,     0.00000000646,
      -0.00000000232,     0.00000000134,
       0.00000001385,    -0.00000001656,
       0.00000001120,     0.00000000858,
      -0.00000000279,     0.00000000385,
      -0.00000000413,    -0.00000000235,
       0.00000000156,    -0.00000000274,
       0.00000000816,    -0.00000000133,
       0.00000000061,     0.00000000404,
       0.00000000093,     0.00000000726,
      -0.00000000335,     0.00000000041,
      -0.00000004358,     0.00000007026,
      -0.00000004791,    -0.00000002838,
       0.00000000968,    -0.00000001650,
       0.00000001004,     0.00000000578,
      -0.00000000381,     0.00000000668,
      -0.00000000227,    -0.00000000128,
      -0.00000001033,     0.00000000191,
      -0.00000000083,    -0.00000000551,
       0.00000000158,    -0.00000000018,
      -0.00000000033,    -0.00000000633,
       0.00000000248,    -0.00000000012,
       0.00000000540,    -0.00000000179,
      -0.00000000035,    -0.00000000114,
       0.00000000112,    -0.00000000096,
       0.00000000065,     0.00000000072,
       0.00000000086,     0.00000000079,
      -0.00000002217,     0.00000000741,
      -0.00000000279,    -0.00000001672,
       0.00000000081,     0.00000000273,
      -0.00000000207,     0.00000000035,
      -0.00000000153,     0.00000000127,
      -0.00000000088,    -0.00000000114,
      -0.00000000089,    -0.00000000012,
       0.00000002525,     0.00000000244,
       0.00000000010,    -0.00000000125,
      -0.00000000317,    -0.00000000016,
      -0.00000000015,     0.00000000307,
      -0.00000007875,    -0.00000000404,
      -0.00000000018,     0.00000000412,
       0.00000000714,    -0.00000000001,
      -0.00000000004,    -0.00000000671,
       0.00000017348,    -0.00000000072,
      -0.00000000004,    -0.00000000973,
      -0.00000000783,     0.00000000074,
       0.00000000080,     0.00000000640,
      -0.00000000010,    -0.00000000105,
      -0.00000017824,     0.00000001988,
       0.00000000106,     0.00000001096,
      -0.00000000427,    -0.00000000078,
      -0.00000000076,     0.00000000772,
      -0.00000013950,    -0.00000001997,
      -0.00000000123,     0.00000000886,
      -0.00000000382,    -0.00000000042,
      -0.00000003841,    -0.00000001435,
      -0.00000000091,     0.00000000257,
      -0.00000000172,    -0.00000000040,
       0.00000000006,    -0.00000000114,
      -0.00000000160,    -0.00000000072,
      -0.00000000077,     0.00000000156,
       0.00000000033,    -0.00000000082,
      -0.00000000064,     0.00000000038,
      -0.00000001537,    -0.00000000770,
      -0.00000000084,     0.00000000199,
       0.00000000284,    -0.00000000155,
      -0.00000000158,    -0.00000000283,
      -0.00000000533,     0.00000000310,
       0.00000000091,     0.00000000150,
       0.00000002787,    -0.00000001635,
      -0.00000000737,    -0.00000000434,
      -0.00000000218,    -0.00000000360,
      -0.00000000056,     0.00000000094,
      -0.00000000041,     0.00000000314,
       0.00000000329,    -0.00000000006,
       0.00000000010,    -0.00000000698,
      -0.00000000191,    -0.00000000008,
       0.00000000023,     0.00000003194,
       0.00000000444,     0.00000000020,
       0.00000000121,     0.00000000031,
       0.00000000090,    -0.00000000040,
       0.00000000527,     0.00000000716,
       0.00000000079,    -0.00000000110,
      -0.00000000028,    -0.00000000115,
      -0.00000000224,     0.00000000031,
       0.00000000091,    -0.00000000018,
      -0.00000000414,    -0.00000001720,
      -0.00000000219,     0.00000000043,
       0.00000000159,     0.00000000022,
      -0.00000000004,    -0.00000000141,
       0.00000001622,     0.00000000065,
       0.00000000015,    -0.00000000213,
      -0.00000000022,    -0.00000000083,
      -0.00000000032,     0.00000000113,
       0.00000000109,     0.00000000018,
      -0.00000000079,     0.00000000039,
       0.00000000040,     0.00000000079,
      -0.00000000744,     0.00000000356,
       0.00000000055,     0.00000000096,
      -0.00000000003,    -0.00000000103,
      -0.00000000109,     0.00000000013,
      -0.00000000044,    -0.00000000938,
      -0.00000000140,     0.00000000015,
      -0.00000000103,    -0.00000000009,
       0.00000000002,     0.00000000108,
       0.00000000001,     0.00000000133,
       0.00000000089,    -0.00000000129,
      -0.00000000136,    -0.00000000084,
       0.00000000080,     0.00000000052,
       0.00000000706,    -0.00000001123,
      -0.00000000183,    -0.00000000117,
      -0.00000002512,     0.00000001398,
       0.00000002049,     0.00000002753,
       0.00000001469,    -0.00000001453,
      -0.00000000059,    -0.00000000106,
      -0.00000000116,     0.00000000085,
       0.00000000063,     0.00000000064,
      -0.00000001810,     0.00000008191,
       0.00000005266,     0.00000000533,
      -0.00000000397,    -0.00000000087,
      -0.00000000025,     0.00000000256,
      -0.00000001346,     0.00000001360,
      -0.00000000066,    -0.00000000065,
      -0.00000001424,     0.00000000902,
      -0.00000002521,    -0.00000003064,
       0.00000000156,    -0.00000000128,
       0.00000003362,    -0.00000001198,
       0.00000000061,     0.00000000170,
       0.00000003376,    -0.00000001662,
       0.00000000083,     0.00000000168,
       0.00000001408,    -0.00000001674,
      -0.00000002037,    -0.00000001115,
       0.00000000073,     0.00000000061,
       0.00000000049,    -0.00000000089,
      -0.00000000051,     0.00000000059,
       0.00000000074,    -0.00000000136,
      -0.00000000128,    -0.00000000082,
       0.00000000069,     0.00000000037,
      -0.00000000050,    -0.00000000058,
      -0.00000000058,     0.00000000046,
       0.00000000900,    -0.00000001564,
      -0.00000000180,    -0.00000000096,
       0.00000000209,     0.00000000254,
       0.00000000255,    -0.00000000205,
      -0.00000000380,    -0.00000000476,
      -0.00000000138,     0.00000000116,
      -0.00000000035,    -0.00000000091,
       0.00000002350,     0.00000002975,
       0.00000000342,    -0.00000000289,
      -0.00000000376,     0.00000000078,
       0.00000000123,     0.00000000392,
       0.00000000800,    -0.00000000207,
      -0.00000000056,    -0.00000000226,
      -0.00000004438,     0.00000001117,
       0.00000000137,     0.00000000545,
      -0.00000000044,    -0.00000000058,
      -0.00000000140,     0.00000000121,
       0.00000000062,    -0.00000000048,
      -0.00000000952,    -0.00000001055,
      -0.00000000143,     0.00000000116,
      -0.00000000023,     0.00000000083,
       0.00000000159,    -0.00000000035,
      -0.00000000091,    -0.00000000160,
      -0.00000000891,    -0.00000001286,
      -0.00000001141,     0.00000001332,
       0.00000000875,     0.00000000426,
      -0.00000000403,     0.00000000045,
       0.00000000076,     0.00000000418,
       0.00000000225,    -0.00000000153,
       0.00000000464,    -0.00000000495,
      -0.00000000579,    -0.00000000377,
      -0.00000000141,     0.00000000354,
      -0.00000000507,    -0.00000000436,
      -0.00000000322,     0.00000000628,
       0.00000000397,     0.00000000062,
       0.00000000083,     0.00000000081,
       0.00000000063,    -0.00000000101,
      -0.00000000043,    -0.00000000180,
      -0.00000000148,     0.00000000073,
      -0.00000000179,     0.00000000047,
       0.00000000083,     0.00000000145,
       0.00000003995,    -0.00000004490,
      -0.00000005505,    -0.00000003176,
      -0.00000000964,     0.00000003098,
       0.00000000955,     0.00000000059,
       0.00000000164,     0.00000000764,
       0.00000000638,    -0.00000000287,
      -0.00000000179,    -0.00000000230,
       0.00000000769,    -0.00000000180,
      -0.00000000311,    -0.00000000641,
      -0.00000000218,     0.00000000202,
      -0.00000000125,    -0.00000000569,
      -0.00000000491,     0.00000000227,
      -0.00000000308,     0.00000000098,
       0.00000000144,     0.00000000260,
       0.00000000100,    -0.00000000087,
       0.00000000335,     0.00000000134,
       0.00000000045,    -0.00000000281,
       0.00000000104,    -0.00000000354,
      -0.00000000293,    -0.00000000012,
       0.00000000966,     0.00000004379,
       0.00000003789,    -0.00000001735,
      -0.00000001154,    -0.00000001496,
       0.00000000684,    -0.00000000203,
      -0.00000000315,    -0.00000000596,
      -0.00000000242,     0.00000000202,
      -0.00000001430,    -0.00000000566,
      -0.00000000204,     0.00000001199,
       0.00000000445,    -0.00000000002,
      -0.00000000460,     0.00000001506,
       0.00000001261,     0.00000000078,
      -0.00000000100,    -0.00000000470,
      -0.00000000077,    -0.00000000041,
       0.00000000041,    -0.00000000076,
       0.00000000064,    -0.00000000079,
      -0.00000000095,    -0.00000000051,
       0.00000000079,     0.00000000065,
       0.00000000051,    -0.00000000095,
       0.00000002783,    -0.00000003427,
      -0.00000004105,    -0.00000002196,
      -0.00000000650,     0.00000002420,
       0.00000000100,     0.00000000082,
       0.00000000065,    -0.00000000121,
       0.00000001658,     0.00000000579,
       0.00000000157,    -0.00000001387,
      -0.00000000527,     0.00000000056,
       0.00000000219,    -0.00000000736,
      -0.00000000607,    -0.00000000047,
       0.00000000030,     0.00000000222,
      -0.00000000408,     0.00000000328,
       0.00000000274,     0.00000000211,
       0.00000000348,     0.00000000388,
       0.00000000197,    -0.00000000283,
      -0.00000010164,    -0.00000003488,
      -0.00000000927,     0.00000008507,
       0.00000003233,    -0.00000000361,
      -0.00000000515,     0.00000001741,
       0.00000001441,     0.00000000109,
      -0.00000000075,    -0.00000000549,
       0.00000001756,    -0.00000001442,
      -0.00000001195,    -0.00000000921,
      -0.00000000196,     0.00000000418,
      -0.00000001493,    -0.00000001720,
      -0.00000000882,     0.00000001234,
       0.00000000420,     0.00000000183,
      -0.00000002658,     0.00000002295,
       0.00000001879,     0.00000001354,
       0.00000000274,    -0.00000000641,
       0.00000000846,     0.00000001009,
       0.00000000520,    -0.00000000697,
      -0.00000000239,    -0.00000000108,
       0.00000000009,     0.00000000154,
       0.00000000158,     0.00000000073,
       0.00000011441,    -0.00000010168,
      -0.00000008259,    -0.00000005790,
      -0.00000001149,     0.00000002807,
      -0.00000001914,    -0.00000002259,
      -0.00000001153,     0.00000001569,
       0.00000000534,     0.00000000232,
       0.00000000046,    -0.00000000658,
      -0.00000000348,     0.00000000092,
      -0.00000000639,    -0.00000000209,
      -0.00000000025,     0.00000000356,
      -0.00000000258,     0.00000000927,
       0.00000000321,     0.00000000051,
      -0.00000000025,    -0.00000000171,
      -0.00000000293,    -0.00000000126,
       0.00000000015,     0.00000000150,
      -0.00000000281,     0.00000000103,
       0.00000000054,     0.00000000108,
       0.00000001228,    -0.00000005531,
      -0.00000003037,     0.00000000208,
       0.00000000287,     0.00000000798,
      -0.00000000785,    -0.00000000126,
       0.00000000047,     0.00000000424,
       0.00000000110,    -0.00000000042,
       0.00000000934,     0.00000000805,
       0.00000000149,    -0.00000000480,
       0.00000000962,    -0.00000000747,
      -0.00000000384,    -0.00000000278,
      -0.00000001473,    -0.00000001455,
      -0.00000000337,     0.00000000754,
      -0.00000000461,     0.00000000493,
       0.00000000246,     0.00000000106,
       0.00000000313,     0.00000000003,
      -0.00000000074,    -0.00000000306,
       0.00000007600,     0.00000007772,
       0.00000001811,    -0.00000003864,
      -0.00000000922,    -0.00000000094,
       0.00000001104,    -0.00000001206,
      -0.00000000604,    -0.00000000251,
      -0.00000001385,     0.00000000092,
       0.00000000175,     0.00000000290,
       0.00000000217,     0.00000001386,
       0.00000000275,    -0.00000000239,
       0.00000002781,    -0.00000000385,
      -0.00000000440,    -0.00000000556,
      -0.00000000096,    -0.00000000845,
      -0.00000000173,     0.00000000128,
      -0.00000000002,    -0.00000000182,
      -0.00000000065,    -0.00000000149,
      -0.00000012071,     0.00000001667,
       0.00000001920,     0.00000002392,
       0.00000000218,     0.00000001925,
       0.00000000388,    -0.00000000293,
      -0.00000000208,     0.00000000607,
       0.00000000105,    -0.00000000044,
       0.00000000445,     0.00000000406,
       0.00000000664,    -0.00000000991,
      -0.00000000290,    -0.00000000166,
       0.00000000188,     0.00000000048,
       0.00000000085,     0.00000000118,
      -0.00000003008,     0.00000005184,
       0.00000000706,     0.00000000396,
      -0.00000000325,    -0.00000000277,
      -0.00000000317,    -0.00000000021,
       0.00000000443,     0.00000000611,
       0.00000000184,    -0.00000000117,
       0.00000000073,    -0.00000000130,
      -0.00000002205,    -0.00000003259,
      -0.00000000437,     0.00000000280,
      -0.00000000210,     0.00000000428,
       0.00000000271,     0.00000000130,
       0.00000000368,    -0.00000000711,
      -0.00000000197,    -0.00000000097,
      -0.00000001622,     0.00000003260,
       0.00000000458,     0.00000000230,
       0.00000000155,    -0.00000000041,
      -0.00000000147,     0.00000000887,
       0.00000000122,     0.00000000011,
      -0.00000000054,    -0.00000000074,
      -0.00000000053,    -0.00000000086,
      -0.00000000081,     0.00000000071,
       0.00000000058,     0.00000000089,
       0.00000000136,     0.00000000281,
      -0.00000000112,     0.00000000086,
       0.00000000091,     0.00000000110,
      -0.00000003199,     0.00000002435,
      -0.00000000179,    -0.00000000237,
      -0.00000000029,     0.00000000134,
       0.00000000133,     0.00000000028,
       0.00000000073,    -0.00000000052,
      -0.00000000047,    -0.00000000074,
       0.00000000662,    -0.00000000443,
      -0.00000000066,    -0.00000000091,
       0.00000000006,     0.00000000114,
       0.00000000115,    -0.00000000017,
       0.00000000118,     0.00000001019,
       0.00000000156,    -0.00000000016,
      -0.00000003798,     0.00000002764,
      -0.00000000081,    -0.00000000112,
      -0.00000004626,     0.00000002577,
       0.00000003775,     0.00000005069,
       0.00000002703,    -0.00000002673,
      -0.00000000077,    -0.00000000138,
      -0.00000000152,     0.00000000113,
       0.00000000079,     0.00000000080,
       0.00000001056,    -0.00000002401,
       0.00000000076,     0.00000000033,
      -0.00000003513,     0.00000015678,
       0.00000010109,     0.00000001018,
      -0.00000000133,    -0.00000003320,
      -0.00000000502,    -0.00000000111,
      -0.00000000032,     0.00000000323,
       0.00000000106,    -0.00000000004,
      -0.00000002588,     0.00000002621,
      -0.00000000084,    -0.00000000083,
      -0.00000002683,     0.00000001632,
      -0.00000000056,    -0.00000000088,
      -0.00000004872,    -0.00000005919,
       0.00000000197,    -0.00000000161,
       0.00000006459,    -0.00000002328,
       0.00000000076,     0.00000000217,
       0.00000006539,    -0.00000003224,
       0.00000000104,     0.00000000212,
      -0.00000000014,    -0.00000000071,
       0.00000000062,    -0.00000000014,
       0.00000000035,     0.00000000027,
       0.00000000379,     0.00000000302,
       0.00000000494,    -0.00000000738,
      -0.00000000902,    -0.00000000486,
      -0.00000000646,     0.00000000693,
       0.00000001223,     0.00000000914,
       0.00000000509,    -0.00000000919,
      -0.00000000867,    -0.00000000797,
      -0.00000001187,     0.00000001584,
       0.00000001143,     0.00000000646,
      -0.00000000094,    -0.00000000547,
      -0.00000000711,     0.00000000224,
      -0.00000000535,     0.00000000028,
       0.00000000102,     0.00000000682,
       0.00000000125,    -0.00000000038,
      -0.00000000072,    -0.00000000174,
      -0.00000000122,     0.00000000073,
      -0.00000000102,    -0.00000000014,
      -0.00000000010,     0.00000000131,
      -0.00000000028,     0.00000000088,
       0.00000000115,     0.00000000027,
       0.00000000124,    -0.00000000043,
      -0.00000000071,    -0.00000000144,
      -0.00000000043,    -0.00000000121,
      -0.00000000139,     0.00000000077,
       0.00000000481,     0.00000000561,
       0.00000000639,    -0.00000000750,
      -0.00000000511,     0.00000000185,
       0.00000000293,     0.00000000576,
       0.00000000325,    -0.00000000227,
      -0.00000000112,    -0.00000000105,
       0.00000000191,     0.00000000508,
       0.00000000568,    -0.00000000310,
      -0.00000000251,    -0.00000000307,
      -0.00000000095,     0.00000000120,
       0.00000000431,    -0.00000000138,
      -0.00000000239,    -0.00000000489,
      -0.00000000115,    -0.00000000192,
      -0.00000000211,     0.00000000166,
       0.00000000121,     0.00000000120,
      -0.00000000099,     0.00000000663,
       0.00000000713,     0.00000000010,
      -0.00000000047,    -0.00000000386,
       0.00000000666,     0.00000000084,
      -0.00000000007,    -0.00000000715,
      -0.00000000387,     0.00000000061,
      -0.00000005073,     0.00000002068,
       0.00000003307,     0.00000005712,
       0.00000003155,    -0.00000002486,
      -0.00000001202,    -0.00000001137,
       0.00000000320,     0.00000000565,
       0.00000000632,    -0.00000000467,
      -0.00000000331,    -0.00000000353,
      -0.00000000131,     0.00000000160,
       0.00000000363,    -0.00000002852,
      -0.00000003071,     0.00000000006,
       0.00000000219,     0.00000001648,
       0.00000000575,    -0.00000000158,
      -0.00000002858,    -0.00000000340,
       0.00000000050,     0.00000003077,
       0.00000001639,    -0.00000000250,
      -0.00000000177,    -0.00000000577,
      -0.00000000427,     0.00000003028,
       0.00000003268,     0.00000000022,
      -0.00000000232,    -0.00000001752,
      -0.00000000618,     0.00000000176,
       0.00000001255,     0.00000000092,
      -0.00000000071,    -0.00000001343,
      -0.00000000729,     0.00000000134,
       0.00000000080,     0.00000000251,
      -0.00000000918,    -0.00000000734,
      -0.00000000575,     0.00000000927,
       0.00000000465,     0.00000000214,
       0.00000000046,    -0.00000000155,
      -0.00000000686,     0.00000000960,
       0.00000000971,     0.00000000514,
       0.00000000172,    -0.00000000493,
      -0.00000000158,    -0.00000000035,
       0.00000002210,    -0.00000020211,
      -0.00000021724,     0.00000000542,
       0.00000001910,     0.00000011599,
       0.00000004071,    -0.00000001290,
      -0.00000000499,    -0.00000001023,
      -0.00000002990,    -0.00000000250,
       0.00000000130,     0.00000003204,
       0.00000001703,    -0.00000000288,
      -0.00000000191,    -0.00000000607,
      -0.00000000168,     0.00000000086,
       0.00000004007,     0.00000003112,
       0.00000002414,    -0.00000004059,
      -0.00000002036,    -0.00000000894,
      -0.00000000200,     0.00000000681,
       0.00000003036,    -0.00000004073,
      -0.00000004129,    -0.00000002317,
      -0.00000000823,     0.00000002088,
       0.00000000699,     0.00000000165,
      -0.00000005504,    -0.00000004191,
      -0.00000003214,     0.00000005586,
       0.00000002818,     0.00000001150,
       0.00000000233,    -0.00000000938,
      -0.00000001502,     0.00000001964,
       0.00000001996,     0.00000001153,
       0.00000000420,    -0.00000001011,
      -0.00000000329,    -0.00000000087,
      -0.00000000113,    -0.00000000191,
      -0.00000000088,     0.00000000113,
      -0.00000000368,     0.00000000146,
       0.00000000166,     0.00000000268,
       0.00000000098,    -0.00000000079,
       0.00000025762,     0.00000018060,
       0.00000013654,    -0.00000025873,
      -0.00000012951,    -0.00000004782,
      -0.00000000934,     0.00000004277,
       0.00000001039,     0.00000000072,
       0.00000003383,    -0.00000004485,
      -0.00000004554,    -0.00000002607,
      -0.00000000930,     0.00000002297,
       0.00000000767,     0.00000000187,
       0.00000000294,     0.00000001025,
       0.00000000612,    -0.00000000360,
      -0.00000000165,    -0.00000000163,
       0.00000001401,    -0.00000000336,
      -0.00000000446,    -0.00000000999,
      -0.00000000360,     0.00000000231,
       0.00000000109,    -0.00000000035,
      -0.00000000026,    -0.00000000109,
       0.00000001593,    -0.00000002771,
      -0.00000002169,    -0.00000000935,
      -0.00000000263,     0.00000000867,
      -0.00000000330,    -0.00000000477,
      -0.00000000316,     0.00000000299,
       0.00000000134,     0.00000000108,
       0.00000002334,    -0.00000007339,
      -0.00000005110,    -0.00000000820,
       0.00000000008,     0.00000001778,
       0.00000000405,    -0.00000000072,
      -0.00000007102,    -0.00000002688,
      -0.00000001173,     0.00000004929,
       0.00000001708,     0.00000000159,
      -0.00000000032,    -0.00000000390,
       0.00000028745,     0.00000004464,
      -0.00000000673,    -0.00000023481,
      -0.00000009368,     0.00000001975,
       0.00000001009,     0.00000002432,
       0.00000000396,    -0.00000000152,
      -0.00000000235,    -0.00000000272,
      -0.00000000091,     0.00000000128,
      -0.00000010618,     0.00000031329,
       0.00000021796,     0.00000004102,
       0.00000000238,    -0.00000007575,
      -0.00000001730,     0.00000000243,
       0.00000000097,     0.00000000289,
       0.00000031025,     0.00000010989,
       0.00000004484,    -0.00000021563,
      -0.00000007484,    -0.00000000435,
       0.00000000189,     0.00000001708,
       0.00000000282,    -0.00000000089,
      -0.00000000045,     0.00000000156,
       0.00000000160,     0.00000000034,
       0.00000000048,    -0.00000000071,
      -0.00000000076,    -0.00000000045,
       0.00000018019,    -0.00000049763,
      -0.00000034687,    -0.00000007294,
      -0.00000000681,     0.00000012065,
       0.00000002755,    -0.00000000303,
      -0.00000000140,    -0.00000000460,
      -0.00000016194,    -0.00000004681,
      -0.00000001712,     0.00000011167,
       0.00000003850,     0.00000000036,
      -0.00000000134,    -0.00000000874,
       0.00000010426,     0.00000011475,
       0.00000005325,    -0.00000006358,
      -0.00000001962,    -0.00000001167,
      -0.00000000150,     0.00000000400,
       0.00000011346,    -0.00000010166,
      -0.00000006154,    -0.00000005268,
      -0.00000001159,     0.00000001881,
       0.00000000378,     0.00000000145,
      -0.00000103755,     0.00000299788,
       0.00000208495,     0.00000040767,
       0.00000003034,    -0.00000072368,
      -0.00000016520,     0.00000002071,
       0.00000000882,     0.00000002755,
       0.00000039795,     0.00000011376,
       0.00000004133,    -0.00000027431,
      -0.00000009462,    -0.00000000071,
       0.00000000328,     0.00000002144,
       0.00000000359,    -0.00000000120,
      -0.00000044601,    -0.00000050417,
      -0.00000023567,     0.00000027076,
       0.00000008307,     0.00000005253,
       0.00000000686,    -0.00000001684,
      -0.00000000248,    -0.00000000050,
      -0.00000049828,     0.00000045225,
       0.00000027534,     0.00000023072,
       0.00000005040,    -0.00000008477,
      -0.00000001727,    -0.00000000617,
      -0.00000000030,     0.00000000256,
       0.00000000073,     0.00000000046,
       0.00000000411,    -0.00000000679,
      -0.00000000102,    -0.00000000063,
       0.00000084070,     0.00000093608,
       0.00000043333,    -0.00000051276,
      -0.00000015749,    -0.00000009512,
      -0.00000001193,     0.00000003190,
       0.00000000472,     0.00000000066,
       0.00000027532,    -0.00000025424,
      -0.00000015398,    -0.00000012816,
      -0.00000002826,     0.00000004707,
       0.00000000953,     0.00000000358,
      -0.00000009938,     0.00000001881,
       0.00000000498,     0.00000003708,
       0.00000000744,    -0.00000000020,
       0.00000006529,     0.00000010700,
       0.00000003870,    -0.00000003646,
      -0.00000001092,    -0.00000000668,
      -0.00000000050,     0.00000000225,
       0.00000000211,    -0.00000000121,
       0.00000000082,    -0.00000000174,
      -0.00000426858,    -0.00000465601,
      -0.00000215050,     0.00000259181,
       0.00000079368,     0.00000047064,
       0.00000005865,    -0.00000016061,
      -0.00000002382,    -0.00000000327,
      -0.00000064872,     0.00000060455,
       0.00000036551,     0.00000030169,
       0.00000006656,    -0.00000011167,
      -0.00000002253,    -0.00000000839,
      -0.00000000047,     0.00000000332,
       0.00000046573,    -0.00000014095,
      -0.00000006334,    -0.00000017303,
      -0.00000003338,     0.00000001528,
       0.00000000230,     0.00000000423,
      -0.00000022065,    -0.00000047922,
      -0.00000017600,     0.00000011726,
       0.00000003359,     0.00000003204,
       0.00000000326,    -0.00000000654,
      -0.00000000650,     0.00000000153,
       0.00000000094,     0.00000000223,
      -0.00000000020,     0.00000000608,
       0.00000000213,    -0.00000000025,
      -0.00000115903,     0.00000042626,
       0.00000022124,     0.00000042951,
       0.00000008022,    -0.00000006061,
      -0.00000001097,    -0.00000000907,
       0.00000008822,     0.00000032504,
       0.00000012206,    -0.00000005153,
      -0.00000001510,    -0.00000002319,
      -0.00000000271,     0.00000000285,
      -0.00000014224,     0.00000035840,
       0.00000011230,     0.00000001027,
      -0.00000000604,    -0.00000001848,
      -0.00000000196,     0.00000000187,
       0.00000026840,     0.00000023357,
       0.00000006469,    -0.00000007071,
      -0.00000000936,    -0.00000001040,
      -0.00000000140,     0.00000000081,
      -0.00000000073,     0.00000000305,
       0.00000000221,     0.00000000163,
       0.00000363166,    -0.00000186324,
      -0.00000086971,    -0.00000130112,
      -0.00000023688,     0.00000022152,
       0.00000003820,     0.00000002535,
      -0.00000020256,    -0.00000069999,
      -0.00000026153,     0.00000011585,
       0.00000003355,     0.00000004949,
       0.00000000571,    -0.00000000626,
       0.00000075112,    -0.00000142642,
      -0.00000042716,    -0.00000012473,
       0.00000000201,     0.00000006698,
       0.00000000687,    -0.00000000328,
      -0.00000130018,    -0.00000087739,
      -0.00000020026,     0.00000036865,
       0.00000005421,     0.00000002086,
       0.00000000127,    -0.00000000528,
       0.00000000350,    -0.00000000774,
      -0.00000000093,    -0.00000000050,
      -0.00000000763,    -0.00000000351,
      -0.00000000002,     0.00000000128,
      -0.00000169881,     0.00000266601,
       0.00000078262,     0.00000034809,
       0.00000002439,    -0.00000011947,
      -0.00000001203,     0.00000000111,
       0.00000077139,     0.00000042148,
       0.00000008625,    -0.00000022214,
      -0.00000003336,    -0.00000000534,
       0.00000000046,     0.00000000330,
      -0.00000045064,    -0.00000056856,
      -0.00000005857,     0.00000009255,
       0.00000001319,    -0.00000000098,
      -0.00000000077,    -0.00000000124,
      -0.00000059656,     0.00000029885,
       0.00000003517,     0.00000007917,
       0.00000000560,    -0.00000000226,
       0.00000000215,    -0.00000000167,
       0.00000000031,    -0.00000000114,
      -0.00000000035,     0.00000000423,
       0.00000912681,    -0.00001519491,
      -0.00000444836,    -0.00000180876,
      -0.00000010759,     0.00000067481,
       0.00000006742,    -0.00000001018,
      -0.00000190193,    -0.00000102885,
      -0.00000020985,     0.00000054759,
       0.00000008217,     0.00000001268,
      -0.00000000120,    -0.00000000814,
       0.00000170799,     0.00000256563,
       0.00000029717,    -0.00000030365,
      -0.00000003844,    -0.00000000762,
       0.00000000124,     0.00000000312,
       0.00000258813,    -0.00000157780,
      -0.00000025514,    -0.00000031513,
      -0.00000001348,     0.00000002925,
       0.00000000210,    -0.00000000013,
      -0.00000000311,    -0.00000000045,
      -0.00000000243,     0.00000000311,
      -0.00000000201,    -0.00000000442,
       0.00000000068,     0.00000000328,
      -0.00000367867,    -0.00000569715,
      -0.00000066650,     0.00000063707,
       0.00000007507,     0.00000002274,
      -0.00000000153,    -0.00000000564,
      -0.00000152404,     0.00000101863,
       0.00000016943,     0.00000018312,
       0.00000000642,    -0.00000001975,
       0.00000064085,     0.00000007837,
       0.00000004185,     0.00000001832,
       0.00000000165,    -0.00000000297,
      -0.00000024448,    -0.00000079630,
      -0.00000001575,     0.00000004950,
       0.00000001052,     0.00000000085,
      -0.00000000462,     0.00000000562,
      -0.00000000596,     0.00000000042,
       0.00001890078,     0.00002927573,
       0.00000342474,    -0.00000327084,
      -0.00000038502,    -0.00000011663,
       0.00000000788,     0.00000002901,
       0.00000365647,    -0.00000246121,
      -0.00000040887,    -0.00000043900,
      -0.00000001525,     0.00000004761,
       0.00000000361,    -0.00000000089,
      -0.00000311019,     0.00000007733,
      -0.00000005983,    -0.00000002975,
      -0.00000000305,    -0.00000000582,
       0.00000060522,     0.00000334569,
       0.00000001929,    -0.00000008913,
      -0.00000002772,     0.00000000026,
       0.00000001762,    -0.00000000503,
      -0.00000000084,    -0.00000000362,
       0.00000000656,    -0.00000001262,
      -0.00000000308,    -0.00000000063,
       0.00000807936,    -0.00000101265,
      -0.00000006157,     0.00000001818,
       0.00000000500,     0.00000003993,
      -0.00000020146,    -0.00000212225,
       0.00000000338,     0.00000001259,
       0.00000001007,    -0.00000000093,
      -0.00000006669,    -0.00000044426,
      -0.00000002366,     0.00000005795,
       0.00000000404,     0.00000000359,
       0.00000025924,    -0.00000044977,
      -0.00000002507,    -0.00000006181,
       0.00000000065,     0.00000000524,
      -0.00000000209,    -0.00000000513,
       0.00000000093,    -0.00000000114,
       0.00000000170,    -0.00000000445,
       0.00000000076,     0.00000000150,
      -0.00003702166,     0.00000459935,
       0.00000029432,    -0.00000008110,
      -0.00000002301,    -0.00000018282,
       0.00000049469,     0.00000495137,
      -0.00000000836,    -0.00000003129,
      -0.00000002368,     0.00000000225,
      -0.00000043803,     0.00000090251,
      -0.00000001931,    -0.00000012954,
      -0.00000000417,    -0.00000001116,
      -0.00000011674,     0.00000118405,
      -0.00000000930,     0.00000014274,
      -0.00000000701,    -0.00000000925,
      -0.00000000754,     0.00000001382,
      -0.00000000007,     0.00000000260,
       0.00000001242,     0.00000000944,
      -0.00000000039,    -0.00000000381,
       0.00000235517,    -0.00000081899,
       0.00000019506,     0.00000018305,
      -0.00000000788,     0.00000002301,
      -0.00000021233,    -0.00000052287,
       0.00000004343,    -0.00000004709,
       0.00000000522,     0.00000000182,
       0.00000060181,    -0.00000011155,
       0.00000000844,     0.00000000268,
       0.00000000030,     0.00000000463,
       0.00000001351,     0.00000058274,
       0.00000000015,    -0.00000001938,
      -0.00000000554,    -0.00000000086,
       0.00000000023,    -0.00000000568,
      -0.00000922700,     0.00000301218,
      -0.00000075904,    -0.00000068800,
       0.00000002861,    -0.00000009106,
       0.00000051111,     0.00000124634,
      -0.00000010355,     0.00000011279,
      -0.00000001248,    -0.00000000432,
      -0.00000078402,     0.00000028869,
      -0.00000005211,    -0.00000000672,
      -0.00000000048,    -0.00000001190,
       0.00000012687,    -0.00000071656,
      -0.00000000331,     0.00000006659,
      -0.00000000522,    -0.00000000037,
       0.00000001044,     0.00000000341,
      -0.00000000037,    -0.00000000244,
       0.00000000368,     0.00000001230,
       0.00000000290,    -0.00000000034,
       0.00000020388,    -0.00000045580,
       0.00000013961,     0.00000002823,
      -0.00000000134,     0.00000002057,
      -0.00000011347,    -0.00000002614,
       0.00000000381,    -0.00000003272,
       0.00000000478,     0.00000000006,
       0.00000019212,    -0.00000005032,
       0.00000001312,     0.00000001942,
      -0.00000000188,     0.00000000149,
       0.00000004056,     0.00000015831,
      -0.00000000665,     0.00000001181,
      -0.00000000149,     0.00000000058,
       0.00000000279,    -0.00000000495,
       0.00000000486,     0.00000000299,
      -0.00000048542,     0.00000193958,
      -0.00000059259,    -0.00000006985,
       0.00000000311,    -0.00000008648,
       0.00000027566,     0.00000006541,
      -0.00000000942,     0.00000007950,
      -0.00000001163,    -0.00000000012,
      -0.00000026509,     0.00000006319,
      -0.00000001499,    -0.00000004229,
       0.00000000626,    -0.00000000145,
      -0.00000002595,    -0.00000011301,
      -0.00000001543,    -0.00000001119,
       0.00000000181,    -0.00000000493,
      -0.00000000704,     0.00000001291,
       0.00000000165,     0.00000000101,
      -0.00000001176,    -0.00000000831,
      -0.00000000076,     0.00000000183,
      -0.00005984603,     0.00006623681,
       0.00000020246,    -0.00000005041,
       0.00000000724,     0.00000007437,
      -0.00000001412,    -0.00000000002,
      -0.00000000940,    -0.00000004682,
       0.00000001768,    -0.00000000039,
      -0.00000000028,     0.00000000342,
       0.00000002242,    -0.00000002894,
       0.00000000986,     0.00000000511,
      -0.00000000089,     0.00000000158,
       0.00000002440,     0.00000001034,
       0.00000000093,     0.00000000773,
      -0.00000000103,     0.00000000067,
       0.00000000079,    -0.00000000105,
       0.00000000097,     0.00000000074,
      -0.00000087561,     0.00000018956,
      -0.00000001592,    -0.00000033252,
       0.00000006448,     0.00000000370,
       0.00000002420,     0.00000011020,
      -0.00000004144,     0.00000000168,
       0.00000000050,    -0.00000000798,
      -0.00000003685,     0.00000004131,
      -0.00000001500,    -0.00000001432,
       0.00000000344,    -0.00000000269,
      -0.00000002219,     0.00000001537,
      -0.00000001179,    -0.00000000595,
       0.00000000032,    -0.00000000332,
      -0.00000000225,     0.00000000262,
      -0.00000000199,    -0.00000000190,
       0.00000000022,     0.00000000037,
       0.00000000044,    -0.00000000010,
       0.00000007108,    -0.00000001919,
       0.00000000894,     0.00000003441,
      -0.00000000870,     0.00000000237,
      -0.00000000463,    -0.00000001530,
       0.00000000757,    -0.00000000216,
       0.00000000056,     0.00000000194,
       0.00000001558,    -0.00000000394,
       0.00000000106,     0.00000000591,
       0.00000000131,     0.00000001408,
      -0.00000000493,    -0.00000000067,
      -0.00000028287,     0.00000007036,
      -0.00000003281,    -0.00000013720,
       0.00000003468,    -0.00000000887,
       0.00000001135,     0.00000003659,
      -0.00000001810,     0.00000000530,
      -0.00000000138,    -0.00000000465,
      -0.00000002076,     0.00000000837,
      -0.00000000370,    -0.00000000874,
       0.00000000198,    -0.00000000109,
       0.00000000287,    -0.00000001465,
       0.00000000471,     0.00000000374,
      -0.00000000141,     0.00000000061,
       0.00000001033,    -0.00000001170,
       0.00000000789,     0.00000000630,
      -0.00000000283,    -0.00000000179,
       0.00000000116,    -0.00000000188,
       0.00000000537,    -0.00000000112,
       0.00000000054,     0.00000000264,
       0.00000000077,     0.00000000481,
      -0.00000000222,     0.00000000033,
      -0.00000003419,     0.00000004904,
      -0.00000003332,    -0.00000002166,
       0.00000000717,    -0.00000001147,
       0.00000000690,     0.00000000441,
      -0.00000000283,     0.00000000458,
      -0.00000000156,    -0.00000000094,
      -0.00000000686,     0.00000000164,
      -0.00000000081,    -0.00000000362,
      -0.00000000024,    -0.00000000438,
       0.00000000175,    -0.00000000002,
      -0.00000000033,    -0.00000000071,
       0.00000000085,    -0.00000000071,
       0.00000000063,     0.00000000063,
      -0.00000001424,     0.00000000624,
      -0.00000000332,    -0.00000001072,
       0.00000000084,     0.00000000171,
      -0.00000000128,     0.00000000046,
      -0.00000000115,     0.00000000094,
      -0.00000000065,    -0.00000000083,
      -0.00000002336,    -0.00000000097,
      -0.00000000004,     0.00000000122,
       0.00000000187,    -0.00000000003,
      -0.00000000004,    -0.00000000172,
       0.00000004540,    -0.00000000098,
      -0.00000000005,    -0.00000000252,
      -0.00000000169,     0.00000000024,
       0.00000000025,     0.00000000125,
      -0.00000003731,     0.00000000633,
       0.00000000034,     0.00000000227,
      -0.00000000119,    -0.00000000020,
      -0.00000000018,     0.00000000213,
      -0.00000003928,    -0.00000000497,
      -0.00000000030,     0.00000000243,
      -0.00000000105,    -0.00000000014,
      -0.00000000104,    -0.00000000047,
      -0.00000000052,     0.00000000100,
      -0.00000001024,    -0.00000000528,
      -0.00000000058,     0.00000000130,
       0.00000000169,    -0.00000000088,
      -0.00000000090,    -0.00000000168,
       0.00000000052,     0.00000000090,
       0.00000001714,    -0.00000000972,
      -0.00000000125,    -0.00000000216,
      -0.00000000026,     0.00000000163,
       0.00000000175,    -0.00000000007,
      -0.00000000100,    -0.00000000006,
       0.00000000018,     0.00000001750,
       0.00000000235,     0.00000000012,
       0.00000000058,    -0.00000000060,
      -0.00000000133,     0.00000000017,
      -0.00000000232,    -0.00000001050,
      -0.00000000131,     0.00000000022,
       0.00000000001,    -0.00000000114,
      -0.00000000045,    -0.00000000068,
      -0.00000000067,     0.00000000045,
      -0.00000000496,    -0.00000000733,
      -0.00000000089,     0.00000000062,
      -0.00000000020,    -0.00000000081,
       0.00000000851,    -0.00000000156,
      -0.00000000018,    -0.00000000110,
       0.00000000046,    -0.00000000086,
      -0.00000000080,    -0.00000000053,
       0.00000000585,    -0.00000001003,
      -0.00000000112,    -0.00000000061,
       0.00000000126,     0.00000000146,
       0.00000000148,    -0.00000000123,
      -0.00000000243,    -0.00000000290,
      -0.00000000080,     0.00000000069,
       0.00000001455,     0.00000001762,
       0.00000000197,    -0.00000000173,
      -0.00000000199,     0.00000000049,
       0.00000000079,     0.00000000212,
       0.00000000446,    -0.00000000139,
      -0.00000000036,    -0.00000000120,
      -0.00000002428,     0.00000000735,
       0.00000000087,     0.00000000290,
      -0.00000000076,     0.00000000077,
      -0.00000000620,    -0.00000000553,
      -0.00000000072,     0.00000000074,
      -0.00000001533,    -0.00000000613,
      -0.00000000254,     0.00000001936,
       0.00000000437,    -0.00000001188,
      -0.00000001605,    -0.00000000154,
       0.00000002457,     0.00000000856,
       0.00000000295,    -0.00000003080,
      -0.00000000105,     0.00000000065,
       0.00000000112,     0.00000000123,
       0.00000000150,     0.00000000007,
      -0.00000000026,    -0.00000000169,
       0.00000000013,    -0.00000000145,
      -0.00000000161,     0.00000000013,
       0.00000000010,    -0.00000000135,
      -0.00000000172,     0.00000000019,
      -0.00000000070,     0.00000000101,
       0.00000000123,     0.00000000050,
       0.00000000095,     0.00000000073,
       0.00000000060,    -0.00000000111,
      -0.00000000016,    -0.00000000090,
      -0.00000000090,     0.00000000025,
       0.00000000083,     0.00000000067,
       0.00000000047,    -0.00000000094,
       0.00000000090,     0.00000000375,
       0.00000000315,    -0.00000000153,
      -0.00000000082,    -0.00000000119,
       0.00000000376,    -0.00000000092,
      -0.00000000160,    -0.00000000317,
      -0.00000000113,     0.00000000102,
      -0.00000000375,     0.00000000166,
      -0.00000000179,     0.00000000052,
       0.00000000080,     0.00000000153,
       0.00000000188,     0.00000000079,
       0.00000000028,    -0.00000000159,
       0.00000000059,    -0.00000000201,
      -0.00000000166,    -0.00000000008,
       0.00000000876,     0.00000003063,
       0.00000002593,    -0.00000001421,
      -0.00000000896,    -0.00000000988,
       0.00000000427,    -0.00000000101,
      -0.00000000174,    -0.00000000375,
      -0.00000000153,     0.00000000116,
      -0.00000000841,    -0.00000000309,
      -0.00000000095,     0.00000000707,
       0.00000000261,    -0.00000000016,
      -0.00000000299,     0.00000000843,
       0.00000000709,     0.00000000081,
      -0.00000000036,    -0.00000000266,
       0.00000001211,     0.00000000433,
       0.00000000121,    -0.00000001016,
       0.00000000141,    -0.00000000447,
      -0.00000000370,    -0.00000000034,
       0.00000000017,     0.00000000138,
      -0.00000000219,     0.00000000168,
       0.00000000139,     0.00000000114,
       0.00000000196,     0.00000000201,
       0.00000000098,    -0.00000000157,
      -0.00000006866,    -0.00000002415,
      -0.00000000662,     0.00000005759,
       0.00000002193,    -0.00000000230,
      -0.00000000334,     0.00000001063,
       0.00000000879,     0.00000000078,
      -0.00000000043,    -0.00000000334,
       0.00000000952,    -0.00000000767,
      -0.00000000633,    -0.00000000501,
      -0.00000000107,     0.00000000220,
      -0.00000000836,    -0.00000000907,
      -0.00000000453,     0.00000000681,
       0.00000000228,     0.00000000092,
      -0.00000001667,     0.00000001468,
       0.00000001198,     0.00000000842,
       0.00000000167,    -0.00000000409,
       0.00000000473,     0.00000000558,
       0.00000000287,    -0.00000000389,
      -0.00000000134,    -0.00000000060,
       0.00000000017,     0.00000000112,
       0.00000000101,     0.00000000071,
       0.00000007205,    -0.00000006568,
      -0.00000005313,    -0.00000003610,
      -0.00000000698,     0.00000001798,
      -0.00000001076,    -0.00000001255,
      -0.00000000639,     0.00000000883,
       0.00000000303,     0.00000000128,
       0.00000000037,    -0.00000000463,
      -0.00000000245,     0.00000000068,
      -0.00000000435,    -0.00000000184,
      -0.00000000047,     0.00000000245,
      -0.00000000243,     0.00000000729,
       0.00000000226,     0.00000000051,
      -0.00000000008,    -0.00000000122,
      -0.00000000194,    -0.00000000057,
       0.00000000024,     0.00000000099,
      -0.00000000186,     0.00000000038,
       0.00000001135,    -0.00000004068,
      -0.00000002267,     0.00000000019,
      -0.00000000554,    -0.00000000124,
       0.00000000014,     0.00000000306,
       0.00000000593,     0.00000000461,
       0.00000000068,    -0.00000000306,
       0.00000000610,    -0.00000000419,
      -0.00000000218,    -0.00000000188,
      -0.00000000957,    -0.00000000963,
      -0.00000000222,     0.00000000492,
      -0.00000000284,     0.00000000293,
       0.00000000148,     0.00000000066,
       0.00000000155,     0.00000000015,
      -0.00000000056,    -0.00000000168,
       0.00000005017,     0.00000005059,
       0.00000001170,    -0.00000002586,
       0.00000000678,    -0.00000000699,
      -0.00000000352,    -0.00000000156,
      -0.00000000721,     0.00000000037,
       0.00000000085,     0.00000000149,
       0.00000000141,     0.00000000735,
       0.00000000144,    -0.00000000139,
       0.00000001565,    -0.00000000240,
      -0.00000000255,    -0.00000000306,
      -0.00000000055,    -0.00000000448,
      -0.00000000091,     0.00000000069,
      -0.00000000004,    -0.00000000119,
      -0.00000000029,    -0.00000000103,
      -0.00000006800,     0.00000001028,
       0.00000001110,     0.00000001323,
       0.00000000126,     0.00000001022,
       0.00000000204,    -0.00000000159,
      -0.00000000131,     0.00000000381,
       0.00000000248,     0.00000000287,
       0.00000000473,    -0.00000000629,
      -0.00000000175,    -0.00000000115,
       0.00000000110,     0.00000000020,
       0.00000000046,     0.00000000080,
      -0.00000002109,     0.00000003185,
       0.00000000426,     0.00000000275,
      -0.00000000174,    -0.00000000133,
      -0.00000000163,    -0.00000000047,
       0.00000000213,     0.00000000316,
       0.00000000092,    -0.00000000052,
       0.00000000050,    -0.00000000078,
      -0.00000001023,    -0.00000001685,
      -0.00000000219,     0.00000000125,
      -0.00000000131,     0.00000000241,
       0.00000000137,     0.00000000067,
       0.00000000224,    -0.00000000401,
      -0.00000000107,    -0.00000000057,
      -0.00000000985,     0.00000001836,
       0.00000000250,     0.00000000135,
       0.00000000097,    -0.00000000019,
      -0.00000001080,     0.00000001305,
      -0.00000006852,    -0.00000005671,
      -0.00000000094,    -0.00000000030,
      -0.00000000026,     0.00000000095,
      -0.00000000033,     0.00000000101,
       0.00000000116,    -0.00000000088,
      -0.00000000087,    -0.00000000116,
       0.00000000046,     0.00000000061,
       0.00000000592,    -0.00000000414,
      -0.00000000098,    -0.00000000132,
      -0.00000002723,     0.00000000761,
      -0.00000000024,    -0.00000000085,
      -0.00000001706,     0.00000001477,
      -0.00000000048,    -0.00000000055,
      -0.00000001581,    -0.00000001861,
       0.00000000063,    -0.00000000053,
       0.00000002210,    -0.00000001034,
       0.00000000034,     0.00000000074,
      -0.00000000075,    -0.00000000078,
      -0.00000000055,     0.00000000100,
       0.00000000165,     0.00000000046,
      -0.00000000072,     0.00000000077,
       0.00000000012,     0.00000000106,
       0.00000000148,    -0.00000000037,
       0.00000000105,    -0.00000000006,
      -0.00000000028,    -0.00000000145,
       0.00000000058,    -0.00000000043,
      -0.00000000044,    -0.00000000057,
      -0.00000000034,    -0.00000000122,
      -0.00000000088,    -0.00000000030,
      -0.00000000027,     0.00000000121,
      -0.00000000032,     0.00000000084,
       0.00000000118,     0.00000000032,
      -0.00000000057,    -0.00000000074,
      -0.00000000036,    -0.00000000087,
      -0.00000000084,     0.00000000044,
      -0.00000000315,     0.00000000094,
       0.00000000166,     0.00000000360,
       0.00000000205,    -0.00000000127,
       0.00000000093,     0.00000000312,
       0.00000000359,    -0.00000000168,
      -0.00000000143,    -0.00000000195,
      -0.00000000062,    -0.00000000117,
      -0.00000000131,     0.00000000093,
      -0.00000000070,     0.00000000460,
       0.00000000493,     0.00000000011,
      -0.00000000028,    -0.00000000266,
       0.00000000460,     0.00000000064,
       0.00000000000,    -0.00000000494,
      -0.00000000266,     0.00000000041,
      -0.00000003517,     0.00000001226,
       0.00000002059,     0.00000003995,
       0.00000002230,    -0.00000001589,
      -0.00000000790,    -0.00000000814,
       0.00000000176,     0.00000000343,
       0.00000000386,    -0.00000000262,
      -0.00000000188,    -0.00000000218,
       0.00000000293,    -0.00000001969,
      -0.00000002126,    -0.00000000034,
       0.00000000133,     0.00000001143,
       0.00000000403,    -0.00000000106,
      -0.00000001970,    -0.00000000280,
      -0.00000000016,     0.00000002127,
       0.00000001141,    -0.00000000154,
      -0.00000000122,    -0.00000000404,
      -0.00000000401,     0.00000002592,
       0.00000002803,     0.00000000054,
      -0.00000000182,    -0.00000001503,
      -0.00000000531,     0.00000000147,
       0.00000000920,     0.00000000085,
      -0.00000000034,    -0.00000000987,
      -0.00000000532,     0.00000000090,
       0.00000000058,     0.00000000185,
      -0.00000000622,    -0.00000000479,
      -0.00000000372,     0.00000000626,
       0.00000000313,     0.00000000138,
      -0.00000000451,     0.00000000642,
       0.00000000647,     0.00000000335,
       0.00000000110,    -0.00000000328,
       0.00000002168,    -0.00000015635,
      -0.00000016864,    -0.00000000074,
       0.00000001219,     0.00000009041,
       0.00000003191,    -0.00000000910,
      -0.00000000385,    -0.00000000824,
      -0.00000002234,    -0.00000000202,
       0.00000000089,     0.00000002394,
       0.00000001274,    -0.00000000216,
      -0.00000000142,    -0.00000000449,
       0.00000002720,     0.00000002069,
       0.00000001602,    -0.00000002745,
      -0.00000001377,    -0.00000000587,
      -0.00000000129,     0.00000000460,
       0.00000002010,    -0.00000002779,
      -0.00000002807,    -0.00000001522,
      -0.00000000534,     0.00000001418,
       0.00000000475,     0.00000000104,
      -0.00000004416,    -0.00000003288,
      -0.00000002505,     0.00000004470,
       0.00000002251,     0.00000000884,
       0.00000000173,    -0.00000000748,
      -0.00000001058,     0.00000001393,
       0.00000001413,     0.00000000810,
       0.00000000292,    -0.00000000715,
      -0.00000000235,    -0.00000000061,
      -0.00000000225,    -0.00000000083,
       0.00000000004,     0.00000000189,
      -0.00000000275,     0.00000000241,
       0.00000000228,     0.00000000196,
       0.00000000069,    -0.00000000101,
       0.00000020246,     0.00000013744,
       0.00000010278,    -0.00000020258,
      -0.00000010124,    -0.00000003529,
      -0.00000000649,     0.00000003337,
       0.00000002398,    -0.00000003203,
      -0.00000003249,    -0.00000001835,
      -0.00000000650,     0.00000001640,
       0.00000000546,     0.00000000126,
       0.00000000833,     0.00000000627,
       0.00000000278,    -0.00000000739,
      -0.00000000305,    -0.00000000021,
       0.00000000944,    -0.00000000844,
      -0.00000000790,    -0.00000000612,
      -0.00000000193,     0.00000000351,
       0.00000000200,    -0.00000002045,
      -0.00000001490,     0.00000000088,
       0.00000000114,     0.00000000550,
      -0.00000000230,    -0.00000000030,
       0.00000000019,     0.00000000173,
       0.00000001475,    -0.00000004819,
      -0.00000003361,    -0.00000000480,
       0.00000000036,     0.00000001171,
       0.00000000268,    -0.00000000058,
      -0.00000004565,    -0.00000001830,
      -0.00000000837,     0.00000003167,
       0.00000001100,     0.00000000138,
      -0.00000000010,    -0.00000000251,
       0.00000022467,     0.00000004059,
      -0.00000000066,    -0.00000018418,
      -0.00000007378,     0.00000001366,
       0.00000000745,     0.00000001925,
       0.00000000322,    -0.00000000704,
      -0.00000000640,    -0.00000000153,
      -0.00000000017,     0.00000000277,
      -0.00000006972,     0.00000020532,
       0.00000014295,     0.00000002662,
       0.00000000124,    -0.00000004975,
      -0.00000001137,     0.00000000174,
       0.00000000066,     0.00000000186,
       0.00000020225,     0.00000007378,
       0.00000003077,    -0.00000014057,
      -0.00000004883,    -0.00000000332,
       0.00000000115,     0.00000001117,
       0.00000000182,    -0.00000000057,
      -0.00000000023,     0.00000000083,
       0.00000000088,     0.00000000015,
       0.00000013018,    -0.00000035383,
      -0.00000024679,    -0.00000005309,
      -0.00000000504,     0.00000008594,
       0.00000001965,    -0.00000000220,
      -0.00000010669,    -0.00000003220,
      -0.00000001206,     0.00000007370,
       0.00000002546,     0.00000000042,
      -0.00000000086,    -0.00000000579,
       0.00000006281,     0.00000006674,
       0.00000003070,    -0.00000003814,
      -0.00000001175,    -0.00000000661,
      -0.00000000081,     0.00000000239,
       0.00000006629,    -0.00000006113,
      -0.00000003687,    -0.00000003063,
      -0.00000000667,     0.00000001128,
       0.00000000228,     0.00000000081,
      -0.00000071617,     0.00000204667,
       0.00000142400,     0.00000028239,
       0.00000002073,    -0.00000049477,
      -0.00000011310,     0.00000001470,
       0.00000000628,     0.00000001885,
       0.00000026200,     0.00000007807,
       0.00000002900,    -0.00000018089,
      -0.00000006247,    -0.00000000086,
       0.00000000216,     0.00000001420,
       0.00000000236,    -0.00000000081,
      -0.00000026939,    -0.00000029602,
      -0.00000013759,     0.00000016277,
       0.00000004989,     0.00000003025,
       0.00000000382,    -0.00000001011,
      -0.00000029263,     0.00000027407,
       0.00000016632,     0.00000013460,
       0.00000002893,    -0.00000005123,
      -0.00000001046,    -0.00000000338,
       0.00000054498,     0.00000058793,
       0.00000027029,    -0.00000033114,
      -0.00000010168,    -0.00000005837,
      -0.00000000699,     0.00000002061,
       0.00000016389,    -0.00000015405,
      -0.00000009317,    -0.00000007588,
      -0.00000001650,     0.00000002853,
       0.00000000578,     0.00000000202,
      -0.00000004188,     0.00000000779,
      -0.00000000045,     0.00000001509,
       0.00000000304,     0.00000000118,
       0.00000004403,     0.00000004947,
       0.00000001731,    -0.00000002424,
      -0.00000000727,    -0.00000000272,
       0.00000000122,    -0.00000000074,
       0.00000000052,    -0.00000000099,
      -0.00000270770,    -0.00000285321,
      -0.00000130785,     0.00000163709,
       0.00000050112,     0.00000028129,
       0.00000003336,    -0.00000010146,
      -0.00000001504,    -0.00000000140,
      -0.00000038640,     0.00000036698,
       0.00000022156,     0.00000017865,
       0.00000003885,    -0.00000006780,
      -0.00000001370,    -0.00000000471,
       0.00000020700,    -0.00000008202,
      -0.00000003165,    -0.00000007462,
      -0.00000001414,     0.00000000646,
       0.00000000074,     0.00000000181,
      -0.00000014188,    -0.00000021930,
      -0.00000007808,     0.00000007241,
       0.00000002037,     0.00000001343,
       0.00000000112,    -0.00000000396,
      -0.00000000384,     0.00000000091,
       0.00000000057,     0.00000000136,
      -0.00000000013,     0.00000000358,
       0.00000000130,    -0.00000000015,
      -0.00000057452,     0.00000027571,
       0.00000013464,     0.00000020781,
       0.00000003752,    -0.00000003557,
      -0.00000000628,    -0.00000000392,
       0.00000005451,     0.00000015341,
       0.00000005650,    -0.00000002944,
      -0.00000000830,    -0.00000001048,
      -0.00000000117,     0.00000000154,
      -0.00000008053,     0.00000021709,
       0.00000006866,     0.00000000349,
      -0.00000000451,    -0.00000001144,
      -0.00000000121,     0.00000000129,
       0.00000015413,     0.00000014388,
       0.00000004141,    -0.00000003975,
      -0.00000000518,    -0.00000000699,
      -0.00000000038,     0.00000000150,
       0.00000000122,     0.00000000078,
       0.00000165047,    -0.00000120967,
      -0.00000053171,    -0.00000056243,
      -0.00000009577,     0.00000012938,
       0.00000002151,     0.00000000857,
      -0.00000012637,    -0.00000032590,
      -0.00000011896,     0.00000006667,
       0.00000001850,     0.00000002187,
       0.00000000235,    -0.00000000337,
       0.00000044563,    -0.00000085344,
      -0.00000025646,    -0.00000007196,
       0.00000000247,     0.00000004060,
       0.00000000418,    -0.00000000227,
      -0.00000076694,    -0.00000053236,
      -0.00000012399,     0.00000021666,
       0.00000003205,     0.00000001334,
       0.00000000090,    -0.00000000318,
       0.00000000178,    -0.00000000386,
      -0.00000000409,    -0.00000000158,
      -0.00000106381,     0.00000165706,
       0.00000048663,     0.00000021860,
       0.00000001472,    -0.00000007493,
      -0.00000000761,     0.00000000090,
       0.00000045729,     0.00000025596,
       0.00000005254,    -0.00000013199,
      -0.00000001999,    -0.00000000314,
       0.00000000032,     0.00000000200,
      -0.00000024784,    -0.00000029787,
      -0.00000002916,     0.00000005064,
       0.00000000728,    -0.00000000109,
      -0.00000031867,     0.00000016465,
       0.00000001969,     0.00000004278,
       0.00000000307,    -0.00000000151,
       0.00000000153,    -0.00000000071,
       0.00000000004,     0.00000000246,
       0.00000556035,    -0.00000918326,
      -0.00000268972,    -0.00000110478,
      -0.00000006232,     0.00000041137,
       0.00000004147,    -0.00000000733,
      -0.00000112713,    -0.00000062444,
      -0.00000012757,     0.00000032510,
       0.00000004918,     0.00000000738,
      -0.00000000083,    -0.00000000493,
       0.00000094083,     0.00000135979,
       0.00000015437,    -0.00000016674,
      -0.00000002161,    -0.00000000267,
       0.00000000087,     0.00000000174,
       0.00000137927,    -0.00000087191,
      -0.00000014177,    -0.00000016709,
      -0.00000000654,     0.00000001703,
      -0.00000000237,    -0.00000000105,
      -0.00000000140,     0.00000000188,
      -0.00000000196,    -0.00000000239,
       0.00000000026,     0.00000000193,
      -0.00000209969,    -0.00000312072,
      -0.00000035999,     0.00000036404,
       0.00000004436,     0.00000001002,
      -0.00000080843,     0.00000055608,
       0.00000009292,     0.00000009586,
       0.00000000283,    -0.00000001122,
       0.00000028837,     0.00000004952,
       0.00000002712,     0.00000001259,
       0.00000000113,    -0.00000000238,
      -0.00000015949,    -0.00000039584,
      -0.00000001086,     0.00000003240,
       0.00000000653,     0.00000000054,
      -0.00000000249,     0.00000000309,
      -0.00000000330,     0.00000000026,
       0.00001063211,     0.00001583193,
       0.00000182635,    -0.00000184200,
      -0.00000022423,    -0.00000005087,
       0.00000000604,     0.00000001709,
       0.00000194015,    -0.00000134525,
      -0.00000022465,    -0.00000022980,
      -0.00000000666,     0.00000002710,
       0.00000000208,    -0.00000000069,
      -0.00000146378,     0.00000006309,
      -0.00000003673,    -0.00000002004,
      -0.00000000218,    -0.00000000222,
       0.00000039602,     0.00000162199,
       0.00000001288,    -0.00000005781,
      -0.00000001627,     0.00000000036,
       0.00000000964,    -0.00000000277,
      -0.00000000052,    -0.00000000195,
       0.00000000358,    -0.00000000700,
      -0.00000000170,    -0.00000000031,
       0.00000393854,    -0.00000067281,
      -0.00000004052,     0.00000001241,
       0.00000000375,     0.00000002265,
      -0.00000013464,    -0.00000100525,
       0.00000000229,     0.00000000828,
       0.00000000555,    -0.00000000070,
      -0.00000004220,    -0.00000025473,
      -0.00000001322,     0.00000003374,
       0.00000000258,     0.00000000196,
       0.00000014904,    -0.00000025927,
      -0.00000001553,    -0.00000003493,
       0.00000000059,     0.00000000317,
      -0.00000000104,    -0.00000000220,
       0.00000000051,    -0.00000000210,
       0.00000000043,     0.00000000082,
      -0.00001804858,     0.00000305582,
       0.00000019304,    -0.00000005535,
      -0.00000001719,    -0.00000010369,
       0.00000033036,     0.00000235086,
      -0.00000000566,    -0.00000002052,
      -0.00000001307,     0.00000000171,
      -0.00000024846,     0.00000051556,
      -0.00000001202,    -0.00000007414,
      -0.00000000285,    -0.00000000646,
      -0.00000006423,     0.00000068015,
      -0.00000000351,     0.00000008097,
      -0.00000000456,    -0.00000000534,
      -0.00000000363,     0.00000000623,
      -0.00000000022,     0.00000000136,
       0.00000000683,     0.00000000411,
      -0.00000000036,    -0.00000000212,
       0.00000135074,    -0.00000049842,
       0.00000011416,     0.00000010404,
      -0.00000000403,     0.00000001421,
      -0.00000012377,    -0.00000029667,
       0.00000002432,    -0.00000002700,
       0.00000000316,     0.00000000092,
       0.00000030449,    -0.00000007380,
       0.00000000502,     0.00000000192,
       0.00000000025,     0.00000000280,
       0.00000001362,     0.00000028972,
       0.00000000011,    -0.00000001223,
      -0.00000000319,    -0.00000000053,
       0.00000000014,    -0.00000000328,
      -0.00000537947,     0.00000187500,
      -0.00000045163,    -0.00000039763,
       0.00000001482,    -0.00000005715,
       0.00000029879,     0.00000070908,
      -0.00000005808,     0.00000006481,
      -0.00000000758,    -0.00000000219,
      -0.00000040734,     0.00000018476,
      -0.00000003215,    -0.00000000501,
      -0.00000000042,    -0.00000000732,
       0.00000007166,    -0.00000035147,
      -0.00000000259,     0.00000004173,
      -0.00000000334,    -0.00000000021,
       0.00000000598,     0.00000000208,
      -0.00000000022,    -0.00000000137,
       0.00000000214,     0.00000000706,
       0.00000000164,    -0.00000000022,
       0.00000014209,    -0.00000029072,
       0.00000008919,     0.00000001902,
      -0.00000000067,     0.00000001317,
      -0.00000007028,    -0.00000002024,
       0.00000000279,    -0.00000002046,
       0.00000000299,     0.00000000002,
       0.00000011223,    -0.00000003347,
       0.00000000836,     0.00000001103,
      -0.00000000098,     0.00000000104,
       0.00000002561,     0.00000009407,
      -0.00000000431,     0.00000000671,
       0.00000000126,    -0.00000000254,
       0.00000000247,     0.00000000138,
      -0.00000038778,     0.00000122619,
      -0.00000037610,    -0.00000005236,
       0.00000000140,    -0.00000005506,
       0.00000017096,     0.00000005024,
      -0.00000000686,     0.00000004975,
      -0.00000000728,    -0.00000000002,
      -0.00000015499,     0.00000004572,
      -0.00000001117,    -0.00000002335,
       0.00000000327,    -0.00000000137,
      -0.00000001464,    -0.00000007327,
      -0.00000000712,    -0.00000000511,
       0.00000000077,    -0.00000000251,
      -0.00000000324,     0.00000000671,
       0.00000000085,     0.00000000048,
      -0.00000000593,    -0.00000000399,
      -0.00000000032,     0.00000000096,
       0.00000011357,    -0.00000004168,
       0.00000000920,     0.00000004068,
      -0.00000000754,     0.00000000103,
      -0.00000000808,    -0.00000002528,
       0.00000000938,    -0.00000000131,
       0.00000000007,     0.00000000178,
       0.00000001534,    -0.00000001932,
       0.00000000653,     0.00000000331,
      -0.00000000053,     0.00000000105,
       0.00000001629,     0.00000000801,
       0.00000000034,     0.00000000512,
       0.00000000043,    -0.00000000067,
       0.00000000062,     0.00000000040,
      -0.00000048548,     0.00000015971,
      -0.00000002941,    -0.00000018106,
       0.00000003448,    -0.00000000213,
       0.00000002048,     0.00000005966,
      -0.00000002202,     0.00000000362,
      -0.00000000027,    -0.00000000416,
      -0.00000002485,     0.00000002747,
      -0.00000000992,    -0.00000000900,
       0.00000000208,    -0.00000000179,
      -0.00000001468,     0.00000000708,
      -0.00000000687,    -0.00000000390,
       0.00000000022,    -0.00000000200,
      -0.00000000122,     0.00000000169,
      -0.00000000132,    -0.00000000107,
       0.00000004425,    -0.00000001389,
       0.00000000644,     0.00000002129,
      -0.00000000535,     0.00000000170,
      -0.00000000318,    -0.00000000939,
       0.00000000462,    -0.00000000148,
       0.00000000039,     0.00000000118,
       0.00000000915,    -0.00000000327,
       0.00000000104,     0.00000000341,
       0.00000000151,     0.00000000809,
      -0.00000000273,    -0.00000000012,
      -0.00000017752,     0.00000005155,
      -0.00000002393,    -0.00000008575,
       0.00000002160,    -0.00000000646,
       0.00000000783,     0.00000002248,
      -0.00000001106,     0.00000000363,
      -0.00000000095,    -0.00000000283,
      -0.00000001242,     0.00000000631,
      -0.00000000285,    -0.00000000522,
       0.00000000120,    -0.00000000081,
       0.00000000119,    -0.00000000802,
       0.00000000238,     0.00000000211,
       0.00000000735,    -0.00000000811,
       0.00000000544,     0.00000000443,
      -0.00000000191,    -0.00000000129,
       0.00000000082,    -0.00000000126,
       0.00000000344,    -0.00000000089,
       0.00000000043,     0.00000000168,
       0.00000000060,     0.00000000310,
      -0.00000000142,     0.00000000025,
      -0.00000002521,     0.00000003366,
      -0.00000002280,    -0.00000001566,
       0.00000000507,    -0.00000000784,
       0.00000000466,     0.00000000317,
      -0.00000000200,     0.00000000309,
      -0.00000000105,    -0.00000000065,
      -0.00000000442,     0.00000000135,
      -0.00000000070,    -0.00000000231,
      -0.00000000018,    -0.00000000289,
       0.00000000062,    -0.00000000051,
      -0.00000000904,     0.00000000536,
      -0.00000000323,    -0.00000000675,
       0.00000000071,     0.00000000107,
      -0.00000000080,     0.00000000044,
      -0.00000000082,     0.00000000069,
      -0.00000000672,    -0.00000000357,
      -0.00000000039,     0.00000000084,
       0.00000000099,    -0.00000000050,
      -0.00000000051,    -0.00000000099,
       0.00000001035,    -0.00000000572,
      -0.00000000071,    -0.00000000128,
       0.00000000022,     0.00000000945,
       0.00000000123,     0.00000000006,
       0.00000000374,    -0.00000000643,
      -0.00000000070,    -0.00000000039,
       0.00000000076,     0.00000000084,
       0.00000000086,    -0.00000000075,
       0.00000000896,     0.00000001029,
       0.00000000114,    -0.00000000105,
      -0.00000000105,     0.00000000030,
       0.00000000051,     0.00000000114,
      -0.00000001316,     0.00000000476,
       0.00000000055,     0.00000000155,
       0.00000000102,    -0.00000000022,
      -0.00000000026,    -0.00000000104,
      -0.00000000073,    -0.00000000028,
      -0.00000000024,     0.00000000098,
      -0.00000000027,     0.00000000073,
       0.00000000100,     0.00000000024,
      -0.00000000238,     0.00000000064,
       0.00000000119,     0.00000000274,
       0.00000000155,    -0.00000000095,
       0.00000000061,     0.00000000239,
       0.00000000274,    -0.00000000117,
      -0.00000000100,    -0.00000000152,
      -0.00000000041,    -0.00000000089,
      -0.00000000100,     0.00000000064,
      -0.00000000048,     0.00000000328,
       0.00000000351,     0.00000000005,
      -0.00000000022,    -0.00000000189,
       0.00000000327,     0.00000000049,
       0.00000000004,    -0.00000000350,
      -0.00000000189,     0.00000000028,
      -0.00000002745,     0.00000000850,
       0.00000001490,     0.00000003134,
       0.00000001759,    -0.00000001182,
      -0.00000000593,    -0.00000000639,
       0.00000000118,     0.00000000256,
       0.00000000290,    -0.00000000180,
      -0.00000000130,    -0.00000000163,
       0.00000000215,    -0.00000001416,
      -0.00000001529,    -0.00000000028,
       0.00000000096,     0.00000000820,
       0.00000000287,    -0.00000000075,
      -0.00000001415,    -0.00000000210,
      -0.00000000020,     0.00000001527,
       0.00000000819,    -0.00000000108,
      -0.00000000085,    -0.00000000287,
      -0.00000000340,     0.00000002125,
       0.00000002298,     0.00000000054,
      -0.00000000145,    -0.00000001232,
      -0.00000000435,     0.00000000120,
       0.00000000678,     0.00000000072,
      -0.00000000017,    -0.00000000729,
      -0.00000000393,     0.00000000063,
      -0.00000000412,    -0.00000000305,
      -0.00000000235,     0.00000000413,
       0.00000000205,     0.00000000086,
      -0.00000000284,     0.00000000429,
       0.00000000430,     0.00000000208,
       0.00000000067,    -0.00000000216,
       0.00000001744,    -0.00000012218,
      -0.00000013182,    -0.00000000099,
       0.00000000941,     0.00000007071,
       0.00000002491,    -0.00000000713,
      -0.00000001650,    -0.00000000168,
       0.00000000048,     0.00000001770,
       0.00000000945,    -0.00000000151,
      -0.00000000102,    -0.00000000331,
       0.00000001819,     0.00000001335,
       0.00000001024,    -0.00000001825,
      -0.00000000914,    -0.00000000370,
      -0.00000000080,     0.00000000303,
       0.00000001291,    -0.00000001871,
      -0.00000001883,    -0.00000000963,
      -0.00000000329,     0.00000000946,
       0.00000000314,     0.00000000059,
      -0.00000003309,    -0.00000002369,
      -0.00000001786,     0.00000003335,
       0.00000001675,     0.00000000618,
       0.00000000114,    -0.00000000554,
      -0.00000000709,     0.00000000956,
       0.00000000968,     0.00000000541,
       0.00000000193,    -0.00000000488,
      -0.00000000161,    -0.00000000039,
      -0.00000000265,    -0.00000000019,
       0.00000000052,     0.00000000213,
      -0.00000000200,     0.00000000268,
       0.00000000240,     0.00000000138,
       0.00000014841,     0.00000009637,
       0.00000007123,    -0.00000014777,
      -0.00000007361,    -0.00000002389,
      -0.00000000406,     0.00000002415,
       0.00000001603,    -0.00000002207,
      -0.00000002230,    -0.00000001217,
      -0.00000000425,     0.00000001123,
       0.00000000373,     0.00000000080,
       0.00000001049,     0.00000000340,
       0.00000000058,    -0.00000000878,
      -0.00000000349,     0.00000000068,
       0.00000000630,    -0.00000001047,
      -0.00000000913,    -0.00000000365,
      -0.00000000091,     0.00000000387,
      -0.00000000755,    -0.00000001401,
      -0.00000000919,     0.00000000772,
       0.00000000361,     0.00000000293,
      -0.00000000142,     0.00000000232,
       0.00000000211,     0.00000000077,
       0.00000000919,    -0.00000003148,
      -0.00000002196,    -0.00000000273,
       0.00000000045,     0.00000000765,
       0.00000000175,    -0.00000000045,
      -0.00000002927,    -0.00000001225,
      -0.00000000580,     0.00000002030,
       0.00000000705,     0.00000000107,
       0.00000018531,     0.00000002985,
      -0.00000000258,    -0.00000015146,
      -0.00000006066,     0.00000001186,
       0.00000000628,     0.00000001583,
       0.00000000203,    -0.00000001026,
      -0.00000000868,    -0.00000000027,
       0.00000000046,     0.00000000358,
      -0.00000004509,     0.00000013388,
       0.00000009321,     0.00000001694,
       0.00000000054,    -0.00000003245,
      -0.00000000742,     0.00000000124,
       0.00000013121,     0.00000004862,
       0.00000002055,    -0.00000009116,
      -0.00000003172,    -0.00000000234,
       0.00000000072,     0.00000000726,
       0.00000009057,    -0.00000024533,
      -0.00000017110,    -0.00000003694,
      -0.00000000340,     0.00000005962,
       0.00000001364,    -0.00000000161,
      -0.00000006979,    -0.00000002165,
      -0.00000000821,     0.00000004827,
       0.00000001669,     0.00000000033,
      -0.00000000057,    -0.00000000379,
       0.00000003736,     0.00000003821,
       0.00000001743,    -0.00000002250,
      -0.00000000689,    -0.00000000370,
       0.00000003810,    -0.00000003684,
      -0.00000002216,    -0.00000001744,
      -0.00000000372,     0.00000000679,
      -0.00000048045,     0.00000137351,
       0.00000095542,     0.00000018877,
       0.00000001281,    -0.00000033214,
      -0.00000007597,     0.00000001046,
       0.00000000441,     0.00000001264,
       0.00000017128,     0.00000005238,
       0.00000001968,    -0.00000011837,
      -0.00000004093,    -0.00000000066,
       0.00000000144,     0.00000000933,
      -0.00000016128,    -0.00000017089,
      -0.00000007884,     0.00000009680,
       0.00000002958,     0.00000001707,
       0.00000000208,    -0.00000000597,
      -0.00000016890,     0.00000016536,
       0.00000009998,     0.00000007695,
       0.00000001617,    -0.00000003080,
      -0.00000000629,    -0.00000000176,
       0.00000034545,     0.00000035742,
       0.00000016283,    -0.00000020886,
      -0.00000006405,    -0.00000003443,
      -0.00000000387,     0.00000001297,
       0.00000009572,    -0.00000009255,
      -0.00000005583,    -0.00000004399,
      -0.00000000941,     0.00000001710,
       0.00000000348,     0.00000000110,
      -0.00000001478,     0.00000000249,
      -0.00000000238,     0.00000000488,
       0.00000000101,     0.00000000152,
       0.00000002940,     0.00000002125,
       0.00000000700,    -0.00000001604,
      -0.00000000480,    -0.00000000088,
       0.00000000070,    -0.00000000045,
      -0.00000168627,    -0.00000169844,
      -0.00000077081,     0.00000101403,
       0.00000030990,     0.00000016206,
       0.00000001797,    -0.00000006270,
      -0.00000000927,    -0.00000000037,
      -0.00000022561,     0.00000022079,
       0.00000013295,     0.00000010348,
       0.00000002208,    -0.00000004066,
      -0.00000000823,    -0.00000000254,
       0.00000008179,    -0.00000004759,
      -0.00000001524,    -0.00000002769,
      -0.00000000503,     0.00000000224,
      -0.00000009073,    -0.00000009173,
      -0.00000003077,     0.00000004473,
       0.00000001233,     0.00000000465,
       0.00000000016,    -0.00000000239,
      -0.00000000224,     0.00000000054,
      -0.00000000007,     0.00000000210,
      -0.00000026349,     0.00000017603,
       0.00000008098,     0.00000009137,
       0.00000001547,    -0.00000002048,
       0.00000003371,     0.00000006765,
       0.00000002408,    -0.00000001684,
      -0.00000000453,    -0.00000000427,
      -0.00000004519,     0.00000013046,
       0.00000004149,     0.00000000060,
      -0.00000000318,    -0.00000000696,
       0.00000008898,     0.00000008739,
       0.00000002591,    -0.00000002257,
      -0.00000000295,    -0.00000000453,
       0.00000000070,     0.00000000035,
       0.00000063466,    -0.00000077237,
      -0.00000031972,    -0.00000019158,
      -0.00000002663,     0.00000007351,
       0.00000001159,     0.00000000078,
      -0.00000007877,    -0.00000013962,
      -0.00000004882,     0.00000003837,
       0.00000001010,     0.00000000846,
       0.00000000078,    -0.00000000177,
       0.00000026171,    -0.00000050840,
      -0.00000015301,    -0.00000004101,
       0.00000000223,     0.00000002438,
       0.00000000251,    -0.00000000152,
      -0.00000045233,    -0.00000031898,
      -0.00000007537,     0.00000012745,
       0.00000001900,     0.00000000822,
       0.00000000056,    -0.00000000191,
       0.00000000089,    -0.00000000189,
      -0.00000000220,    -0.00000000066,
       0.00000000550,     0.00000000596,
       0.00000000065,    -0.00000000063,
      -0.00000065278,     0.00000101917,
       0.00000029925,     0.00000013395,
       0.00000000847,    -0.00000004644,
      -0.00000000476,     0.00000000072,
       0.00000027056,     0.00000015340,
       0.00000003146,    -0.00000007821,
      -0.00000001194,    -0.00000000176,
      -0.00000013521,    -0.00000015519,
      -0.00000001431,     0.00000002712,
       0.00000000388,    -0.00000000089,
      -0.00000016991,     0.00000009294,
       0.00000001178,     0.00000002303,
       0.00000000165,    -0.00000000112,
       0.00000000102,    -0.00000000029,
       0.00000000014,     0.00000000144,
       0.00000333996,    -0.00000552030,
      -0.00000161679,    -0.00000066273,
      -0.00000003447,     0.00000024918,
       0.00000002529,    -0.00000000528,
      -0.00000066652,    -0.00000037373,
      -0.00000007625,     0.00000019250,
       0.00000002936,     0.00000000411,
      -0.00000000059,    -0.00000000297,
       0.00000051829,     0.00000071752,
       0.00000007959,    -0.00000009100,
      -0.00000001195,    -0.00000000062,
       0.00000073241,    -0.00000048632,
      -0.00000008003,    -0.00000008804,
      -0.00000000301,     0.00000001007,
      -0.00000000164,    -0.00000000097,
      -0.00000000080,     0.00000000113,
      -0.00000000151,    -0.00000000130,
       0.00000000010,     0.00000000115,
      -0.00000000709,     0.00000000304,
       0.00000000034,     0.00000000082,
      -0.00000119659,    -0.00000169260,
      -0.00000019190,     0.00000020762,
       0.00000002602,     0.00000000389,
      -0.00000042723,     0.00000030515,
       0.00000005121,     0.00000004989,
       0.00000000114,    -0.00000000638,
       0.00000012734,     0.00000003126,
       0.00000001726,     0.00000000827,
       0.00000000073,    -0.00000000175,
      -0.00000010157,    -0.00000019838,
      -0.00000000713,     0.00000002075,
       0.00000000400,     0.00000000030,
      -0.00000000132,     0.00000000169,
      -0.00000000181,     0.00000000018,
       0.00000598613,     0.00000850032,
       0.00000096391,    -0.00000103786,
      -0.00000013002,    -0.00000001964,
       0.00000000437,     0.00000000997,
       0.00000102622,    -0.00000073961,
      -0.00000012401,    -0.00000011955,
      -0.00000000265,     0.00000001543,
      -0.00000068609,     0.00000004494,
      -0.00000002242,    -0.00000001301,
      -0.00000000147,    -0.00000000062,
       0.00000025210,     0.00000078860,
       0.00000000821,    -0.00000003671,
      -0.00000000949,     0.00000000038,
       0.00000000521,    -0.00000000153,
      -0.00000000030,    -0.00000000104,
       0.00000000190,    -0.00000000389,
       0.00000191325,    -0.00000043240,
      -0.00000002590,     0.00000000831,
       0.00000000270,     0.00000001266,
      -0.00000008675,    -0.00000047724,
       0.00000000152,     0.00000000529,
       0.00000000303,    -0.00000000051,
      -0.00000002693,    -0.00000014376,
      -0.00000000716,     0.00000001944,
       0.00000000161,     0.00000000104,
       0.00000008357,    -0.00000014795,
      -0.00000000952,    -0.00000001939,
       0.00000000047,     0.00000000187,
      -0.00000000051,    -0.00000000089,
       0.00000000006,    -0.00000000098,
      -0.00000877091,     0.00000196384,
       0.00000012312,    -0.00000003703,
      -0.00000001234,    -0.00000005799,
       0.00000021292,     0.00000111829,
      -0.00000000375,    -0.00000001307,
      -0.00000000715,     0.00000000124,
      -0.00000013671,     0.00000029118,
      -0.00000000746,    -0.00000004191,
      -0.00000000190,    -0.00000000367,
      -0.00000003308,     0.00000038529,
      -0.00000000072,     0.00000004526,
      -0.00000000290,    -0.00000000304,
      -0.00000000174,     0.00000000274,
       0.00000000379,     0.00000000171,
      -0.00000000027,    -0.00000000119,
      -0.00000000040,    -0.00000000075,
       0.00000076129,    -0.00000029899,
       0.00000006575,     0.00000005793,
      -0.00000000195,     0.00000000861,
      -0.00000007153,    -0.00000016564,
       0.00000001338,    -0.00000001528,
       0.00000000189,     0.00000000045,
       0.00000015304,    -0.00000004749,
       0.00000000297,     0.00000000130,
       0.00000000020,     0.00000000166,
       0.00000001095,     0.00000014264,
       0.00000000004,    -0.00000000758,
      -0.00000000181,    -0.00000000030,
       0.00000000007,    -0.00000000187,
      -0.00000307080,     0.00000114430,
      -0.00000026340,    -0.00000022406,
       0.00000000719,    -0.00000003506,
       0.00000017306,     0.00000039635,
      -0.00000003196,     0.00000003673,
      -0.00000000452,    -0.00000000105,
      -0.00000021046,     0.00000011606,
      -0.00000001961,    -0.00000000343,
      -0.00000000035,    -0.00000000443,
       0.00000004036,    -0.00000016966,
      -0.00000000177,     0.00000002576,
      -0.00000000213,    -0.00000000012,
       0.00000000339,     0.00000000121,
       0.00000000125,     0.00000000400,
       0.00000009406,    -0.00000018244,
       0.00000005595,     0.00000001219,
      -0.00000000027,     0.00000000828,
      -0.00000004303,    -0.00000001426,
       0.00000000188,    -0.00000001261,
       0.00000000185,    -0.00000000001,
       0.00000006427,    -0.00000002160,
       0.00000000520,     0.00000000615,
       0.00000001573,     0.00000005456,
      -0.00000000265,     0.00000000370,
       0.00000000055,    -0.00000000130,
       0.00000000126,     0.00000000062,
      -0.00000027895,     0.00000076513,
      -0.00000023513,    -0.00000003585,
       0.00000000036,    -0.00000003455,
       0.00000010477,     0.00000003523,
      -0.00000000461,     0.00000003069,
      -0.00000000450,     0.00000000004,
      -0.00000008893,     0.00000003152,
      -0.00000000783,    -0.00000001271,
       0.00000000167,    -0.00000000109,
      -0.00000000780,    -0.00000004518,
      -0.00000000320,    -0.00000000201,
       0.00000000026,    -0.00000000127,
      -0.00000000145,     0.00000000349,
      -0.00000000298,    -0.00000000188,
       0.00000006337,    -0.00000003183,
       0.00000000841,     0.00000002201,
      -0.00000000627,    -0.00000001361,
       0.00000000493,    -0.00000000141,
       0.00000001006,    -0.00000001262,
       0.00000000424,     0.00000000206,
       0.00000001060,     0.00000000569,
       0.00000000009,     0.00000000330,
      -0.00000026785,     0.00000012349,
      -0.00000002958,    -0.00000009755,
       0.00000001810,    -0.00000000387,
       0.00000001577,     0.00000003220,
      -0.00000001159,     0.00000000368,
      -0.00000000049,    -0.00000000213,
      -0.00000001612,     0.00000001796,
      -0.00000000645,    -0.00000000550,
       0.00000000123,    -0.00000000117,
      -0.00000000944,     0.00000000300,
      -0.00000000394,    -0.00000000247,
      -0.00000000065,     0.00000000104,
      -0.00000000083,    -0.00000000059,
       0.00000002697,    -0.00000000984,
       0.00000000455,     0.00000001287,
      -0.00000000216,    -0.00000000564,
       0.00000000276,    -0.00000000100,
       0.00000000532,    -0.00000000252,
       0.00000000087,     0.00000000194,
       0.00000000137,     0.00000000460,
      -0.00000000148,     0.00000000011,
      -0.00000010887,     0.00000003688,
      -0.00000001705,    -0.00000005227,
       0.00000001308,    -0.00000000457,
       0.00000000533,     0.00000001352,
      -0.00000000661,     0.00000000245,
      -0.00000000065,    -0.00000000168,
      -0.00000000737,     0.00000000459,
      -0.00000000210,    -0.00000000308,
       0.00000000036,    -0.00000000433,
       0.00000000115,     0.00000000116,
       0.00000000503,    -0.00000000553,
       0.00000000369,     0.00000000300,
      -0.00000000127,    -0.00000000089,
       0.00000000056,    -0.00000000084,
       0.00000000215,    -0.00000000068,
       0.00000000033,     0.00000000104,
       0.00000000045,     0.00000000194,
      -0.00000001774,     0.00000002279,
      -0.00000001536,    -0.00000001086,
       0.00000000310,     0.00000000219,
      -0.00000000136,     0.00000000205,
      -0.00000000278,     0.00000000104,
      -0.00000000056,    -0.00000000144,
      -0.00000000013,    -0.00000000184,
      -0.00000000572,     0.00000000431,
       0.00000000057,     0.00000000066,
      -0.00000000057,     0.00000000049,
      -0.00000000191,     0.00000000051,
       0.00000000096,     0.00000000219,
       0.00000000124,    -0.00000000076,
       0.00000000044,     0.00000000192,
       0.00000000221,    -0.00000000089,
      -0.00000000077,    -0.00000000123,
      -0.00000000031,    -0.00000000073,
      -0.00000000083,     0.00000000049,
      -0.00000000032,     0.00000000233,
       0.00000000250,     0.00000000000,
      -0.00000000017,    -0.00000000135,
       0.00000000230,     0.00000000036,
       0.00000000005,    -0.00000000247,
      -0.00000000134,     0.00000000019,
      -0.00000002195,     0.00000000629,
       0.00000001136,     0.00000002511,
       0.00000001412,    -0.00000000914,
      -0.00000000461,    -0.00000000513,
       0.00000000085,     0.00000000206,
       0.00000000234,    -0.00000000134,
      -0.00000000099,    -0.00000000132,
       0.00000000152,    -0.00000001012,
      -0.00000001093,    -0.00000000017,
       0.00000000071,     0.00000000586,
       0.00000000203,    -0.00000000054,
      -0.00000001009,    -0.00000000152,
      -0.00000000016,     0.00000001090,
       0.00000000584,    -0.00000000077,
      -0.00000000060,    -0.00000000204,
      -0.00000000267,     0.00000001666,
       0.00000001801,     0.00000000042,
      -0.00000000115,    -0.00000000966,
       0.00000000493,     0.00000000056,
      -0.00000000009,    -0.00000000531,
      -0.00000000286,     0.00000000044,
      -0.00000000266,    -0.00000000187,
      -0.00000000143,     0.00000000264,
       0.00000000130,     0.00000000052,
      -0.00000000172,     0.00000000282,
       0.00000000282,     0.00000000121,
       0.00000000037,    -0.00000000140,
       0.00000001303,    -0.00000009224,
      -0.00000009950,    -0.00000000053,
       0.00000000727,     0.00000005332,
       0.00000001876,    -0.00000000548,
      -0.00000001199,    -0.00000000130,
       0.00000000027,     0.00000001289,
       0.00000000687,    -0.00000000107,
      -0.00000000075,    -0.00000000242,
       0.00000001186,     0.00000000825,
       0.00000000626,    -0.00000001181,
      -0.00000000589,    -0.00000000224,
      -0.00000000047,     0.00000000195,
       0.00000000791,    -0.00000001232,
      -0.00000001232,    -0.00000000578,
      -0.00000000190,     0.00000000618,
       0.00000000204,     0.00000000031,
      -0.00000002356,    -0.00000001591,
      -0.00000001181,     0.00000002360,
       0.00000001180,     0.00000000397,
      -0.00000000454,     0.00000000636,
       0.00000000641,     0.00000000342,
       0.00000000120,    -0.00000000321,
      -0.00000000255,     0.00000000014,
       0.00000000069,     0.00000000201,
      -0.00000000142,     0.00000000251,
       0.00000000218,     0.00000000096,
       0.00000010361,     0.00000006221,
       0.00000004495,    -0.00000010240,
      -0.00000005070,    -0.00000001440,
      -0.00000000213,     0.00000001656,
       0.00000001023,    -0.00000001472,
      -0.00000001480,    -0.00000000767,
      -0.00000000261,     0.00000000743,
       0.00000000246,     0.00000000046,
       0.00000001039,     0.00000000164,
      -0.00000000056,    -0.00000000848,
      -0.00000000332,     0.00000000102,
       0.00000000408,    -0.00000001030,
      -0.00000000871,    -0.00000000204,
      -0.00000000033,     0.00000000362,
      -0.00000001210,    -0.00000000911,
      -0.00000000514,     0.00000001075,
       0.00000000462,     0.00000000122,
      -0.00000000080,     0.00000000336,
       0.00000000284,     0.00000000019,
       0.00000000563,    -0.00000002035,
      -0.00000001418,    -0.00000000148,
       0.00000000041,     0.00000000494,
      -0.00000001865,    -0.00000000801,
      -0.00000000389,     0.00000001291,
       0.00000000448,     0.00000000076,
       0.00000015140,     0.00000001946,
      -0.00000000526,    -0.00000012314,
      -0.00000004918,     0.00000001075,
       0.00000000539,     0.00000001283,
       0.00000000110,    -0.00000001094,
      -0.00000000902,     0.00000000048,
       0.00000000076,     0.00000000365,
      -0.00000002859,     0.00000008639,
       0.00000006011,     0.00000001050,
       0.00000000012,    -0.00000002092,
      -0.00000000478,     0.00000000087,
       0.00000008437,     0.00000003137,
       0.00000001332,    -0.00000005860,
      -0.00000002039,    -0.00000000151,
       0.00000000045,     0.00000000467,
       0.00000006088,    -0.00000016626,
      -0.00000011587,    -0.00000002465,
      -0.00000000208,     0.00000004037,
       0.00000000923,    -0.00000000119,
      -0.00000004517,    -0.00000001420,
      -0.00000000539,     0.00000003125,
       0.00000001082,     0.00000000021,
      -0.00000000038,    -0.00000000246,
       0.00000002187,     0.00000002147,
       0.00000000970,    -0.00000001302,
      -0.00000000395,    -0.00000000202,
       0.00000002146,    -0.00000002221,
      -0.00000001335,    -0.00000000969,
      -0.00000000200,     0.00000000411,
      -0.00000031350,     0.00000090628,
       0.00000062991,     0.00000012185,
       0.00000000699,    -0.00000021902,
      -0.00000005008,     0.00000000747,
       0.00000000307,     0.00000000834,
       0.00000011077,     0.00000003427,
       0.00000001289,    -0.00000007658,
      -0.00000002652,    -0.00000000039,
       0.00000000097,     0.00000000604,
      -0.00000009555,    -0.00000009677,
      -0.00000004425,     0.00000005685,
       0.00000001730,     0.00000000941,
       0.00000000110,    -0.00000000348,
      -0.00000009556,     0.00000009921,
       0.00000005973,     0.00000004297,
       0.00000000876,    -0.00000001840,
      -0.00000000376,    -0.00000000086,
       0.00000021495,     0.00000021096,
       0.00000009497,    -0.00000012918,
      -0.00000003951,    -0.00000001953,
      -0.00000000201,     0.00000000799,
       0.00000005480,    -0.00000005511,
      -0.00000003311,    -0.00000002494,
      -0.00000000522,     0.00000001014,
       0.00000000206,     0.00000000058,
      -0.00000000295,     0.00000000014,
      -0.00000000269,     0.00000000055,
       0.00000000015,     0.00000000141,
       0.00000001937,     0.00000000796,
       0.00000000227,    -0.00000001051,
      -0.00000000313,    -0.00000000011,
      -0.00000103377,    -0.00000098348,
      -0.00000044055,     0.00000061763,
       0.00000018824,     0.00000008985,
       0.00000000903,    -0.00000003799,
      -0.00000012900,     0.00000013161,
       0.00000007893,     0.00000005857,
       0.00000001220,    -0.00000002411,
      -0.00000000488,    -0.00000000129,
       0.00000002468,    -0.00000002743,
      -0.00000000691,    -0.00000000681,
      -0.00000005752,    -0.00000003210,
      -0.00000000920,     0.00000002752,
       0.00000000745,     0.00000000083,
      -0.00000000130,     0.00000000032,
      -0.00000000003,     0.00000000122,
      -0.00000010690,     0.00000011091,
       0.00000004816,     0.00000003392,
       0.00000000487,    -0.00000001157,
       0.00000002076,     0.00000002656,
       0.00000000880,    -0.00000000962,
      -0.00000000244,    -0.00000000141,
      -0.00000002506,     0.00000007769,
       0.00000002475,    -0.00000000048,
      -0.00000000214,    -0.00000000415,
       0.00000005158,     0.00000005234,
       0.00000001588,    -0.00000001295,
      -0.00000000171,    -0.00000000284,
       0.00000015774,    -0.00000048585,
      -0.00000018965,    -0.00000002392,
       0.00000000320,     0.00000004074,
      -0.00000004879,    -0.00000005128,
      -0.00000001619,     0.00000002199,
       0.00000000544,     0.00000000236,
       0.00000015191,    -0.00000030110,
      -0.00000009060,    -0.00000002300,
       0.00000000178,     0.00000001450,
       0.00000000148,    -0.00000000100,
      -0.00000026633,    -0.00000018863,
      -0.00000004499,     0.00000007494,
       0.00000001128,     0.00000000492,
       0.00000000044,    -0.00000000090,
      -0.00000000119,    -0.00000000022,
      -0.00000039312,     0.00000062067,
       0.00000018213,     0.00000008022,
       0.00000000462,    -0.00000002847,
       0.00000015952,     0.00000009069,
       0.00000001852,    -0.00000004615,
      -0.00000000709,    -0.00000000094,
      -0.00000007310,    -0.00000008022,
      -0.00000000688,     0.00000001420,
       0.00000000199,    -0.00000000064,
      -0.00000009027,     0.00000005347,
       0.00000000738,     0.00000001232,
       0.00000000086,    -0.00000000084,
       0.00000197749,    -0.00000329857,
      -0.00000096557,    -0.00000039023,
      -0.00000001798,     0.00000014986,
       0.00000001531,    -0.00000000377,
      -0.00000039277,    -0.00000022064,
      -0.00000004480,     0.00000011355,
       0.00000001745,     0.00000000217,
      -0.00000000041,    -0.00000000177,
       0.00000028519,     0.00000037625,
       0.00000004060,    -0.00000004933,
      -0.00000000650,     0.00000000014,
       0.00000038691,    -0.00000027312,
      -0.00000004573,    -0.00000004599,
      -0.00000000126,     0.00000000602,
      -0.00000000108,    -0.00000000073,
      -0.00000000105,    -0.00000000072,
      -0.00000068078,    -0.00000090885,
      -0.00000010085,     0.00000011817,
       0.00000001516,     0.00000000110,
      -0.00000022467,     0.00000016812,
       0.00000002832,     0.00000002576,
       0.00000000038,    -0.00000000363,
       0.00000005459,     0.00000001968,
       0.00000001082,     0.00000000524,
       0.00000000046,    -0.00000000121,
      -0.00000006343,    -0.00000009994,
      -0.00000000449,     0.00000001305,
       0.00000000243,     0.00000000015,
      -0.00000000069,     0.00000000092,
      -0.00000000098,     0.00000000012,
       0.00000337092,     0.00000452768,
       0.00000050268,    -0.00000058472,
      -0.00000007502,    -0.00000000562,
       0.00000053963,    -0.00000040800,
      -0.00000006868,    -0.00000006167,
      -0.00000000084,     0.00000000876,
      -0.00000031896,     0.00000002959,
      -0.00000001358,    -0.00000000819,
       0.00000015698,     0.00000038343,
       0.00000000503,    -0.00000002289,
      -0.00000000550,     0.00000000034,
       0.00000000279,    -0.00000000085,
       0.00000000098,    -0.00000000216,
       0.00000092402,    -0.00000027063,
      -0.00000001617,     0.00000000548,
       0.00000000188,     0.00000000698,
      -0.00000005434,    -0.00000022631,
       0.00000000099,     0.00000000329,
       0.00000000164,    -0.00000000036,
      -0.00000001720,    -0.00000008003,
      -0.00000000377,     0.00000001111,
       0.00000000099,     0.00000000053,
       0.00000004585,    -0.00000008374,
      -0.00000000579,    -0.00000001059,
      -0.00000423917,     0.00000122916,
       0.00000007675,    -0.00000002441,
      -0.00000000859,    -0.00000003199,
       0.00000013325,     0.00000053061,
      -0.00000000245,    -0.00000000814,
      -0.00000000388,     0.00000000086,
      -0.00000007319,     0.00000016289,
      -0.00000000462,    -0.00000002343,
      -0.00000000124,    -0.00000000207,
      -0.00000001580,     0.00000021577,
       0.00000000042,     0.00000002496,
      -0.00000000181,    -0.00000000171,
      -0.00000000083,     0.00000000115,
       0.00000000212,     0.00000000066,
       0.00000042275,    -0.00000017738,
       0.00000003738,     0.00000003167,
      -0.00000000088,     0.00000000513,
      -0.00000004109,    -0.00000009123,
       0.00000000724,    -0.00000000856,
       0.00000007637,    -0.00000002986,
       0.00000000175,     0.00000000085,
       0.00000000790,     0.00000006948,
      -0.00000000001,    -0.00000000463,
      -0.00000000102,    -0.00000000016,
       0.00000000003,    -0.00000000105,
      -0.00000172265,     0.00000068817,
      -0.00000015121,    -0.00000012357,
       0.00000000319,    -0.00000002112,
       0.00000009958,     0.00000021845,
      -0.00000001729,     0.00000002058,
      -0.00000000267,    -0.00000000047,
      -0.00000010813,     0.00000007170,
      -0.00000001183,    -0.00000000223,
      -0.00000000028,    -0.00000000266,
       0.00000002270,    -0.00000008047,
      -0.00000000110,     0.00000001568,
       0.00000000191,     0.00000000067,
       0.00000000073,     0.00000000224,
       0.00000005993,    -0.00000011286,
       0.00000003454,     0.00000000750,
      -0.00000002606,    -0.00000000946,
       0.00000000119,    -0.00000000767,
       0.00000003621,    -0.00000001364,
       0.00000000317,     0.00000000337,
       0.00000000947,     0.00000003103,
      -0.00000000157,     0.00000000200,
      -0.00000018800,     0.00000047161,
      -0.00000014494,    -0.00000002305,
      -0.00000000018,    -0.00000002138,
       0.00000006351,     0.00000002329,
      -0.00000000292,     0.00000001868,
      -0.00000000274,     0.00000000007,
      -0.00000005022,     0.00000002101,
      -0.00000000527,    -0.00000000681,
       0.00000000082,    -0.00000000079,
      -0.00000000391,    -0.00000002691,
      -0.00000000139,    -0.00000000054,
      -0.00000000061,     0.00000000182,
      -0.00000000149,    -0.00000000087,
       0.00000003516,    -0.00000002305,
       0.00000000672,     0.00000001177,
      -0.00000000457,    -0.00000000730,
       0.00000000257,    -0.00000000120,
       0.00000000638,    -0.00000000809,
       0.00000000269,     0.00000000125,
       0.00000000675,     0.00000000382,
       0.00000000000,     0.00000000207,
      -0.00000014701,     0.00000009025,
      -0.00000002473,    -0.00000005196,
       0.00000000931,    -0.00000000379,
       0.00000001144,     0.00000001732,
      -0.00000000604,     0.00000000307,
      -0.00000000048,    -0.00000000106,
      -0.00000001014,     0.00000001157,
      -0.00000000414,    -0.00000000328,
      -0.00000000593,     0.00000000110,
      -0.00000000222,    -0.00000000151,
       0.00000001614,    -0.00000000685,
       0.00000000315,     0.00000000762,
      -0.00000000146,    -0.00000000333,
       0.00000000162,    -0.00000000067,
       0.00000000308,    -0.00000000185,
       0.00000000067,     0.00000000109,
       0.00000000110,     0.00000000260,
      -0.00000006546,     0.00000002587,
      -0.00000001190,    -0.00000003118,
       0.00000000774,    -0.00000000317,
       0.00000000359,     0.00000000799,
      -0.00000000387,     0.00000000165,
      -0.00000000433,     0.00000000324,
      -0.00000000149,    -0.00000000180,
      -0.00000000002,    -0.00000000231,
       0.00000000334,    -0.00000000371,
      -0.00000000083,    -0.00000000060,
       0.00000000132,    -0.00000000050,
       0.00000000033,     0.00000000119,
      -0.00000001203,     0.00000001522,
      -0.00000001019,    -0.00000000726,
       0.00000000204,     0.00000000146,
      -0.00000000090,     0.00000000134,
      -0.00000000171,     0.00000000078,
      -0.00000000043,    -0.00000000088,
      -0.00000000010,    -0.00000000114,
      -0.00000000154,     0.00000000042,
       0.00000000078,     0.00000000177,
       0.00000000099,    -0.00000000061,
       0.00000000035,     0.00000000155,
       0.00000000178,    -0.00000000069,
      -0.00000000060,    -0.00000000100,
      -0.00000000021,     0.00000000164,
       0.00000000176,    -0.00000000002,
       0.00000000160,     0.00000000025,
       0.00000000004,    -0.00000000173,
      -0.00000001752,     0.00000000483,
       0.00000000885,     0.00000002008,
       0.00000001128,    -0.00000000715,
       0.00000000064,     0.00000000168,
       0.00000000191,    -0.00000000103,
      -0.00000000078,    -0.00000000109,
       0.00000000103,    -0.00000000711,
      -0.00000000766,    -0.00000000008,
       0.00000000053,     0.00000000411,
      -0.00000000709,    -0.00000000105,
      -0.00000000009,     0.00000000765,
       0.00000000410,    -0.00000000056,
      -0.00000000196,     0.00000001257,
       0.00000001357,     0.00000000026,
      -0.00000000091,    -0.00000000727,
       0.00000000351,     0.00000000041,
      -0.00000000006,    -0.00000000378,
      -0.00000000203,     0.00000000031,
      -0.00000000167,    -0.00000000110,
      -0.00000000083,     0.00000000164,
      -0.00000000098,     0.00000000183,
       0.00000000181,     0.00000000066,
       0.00000000915,    -0.00000006749,
      -0.00000007271,     0.00000000004,
       0.00000000560,     0.00000003892,
       0.00000001367,    -0.00000000413,
      -0.00000000854,    -0.00000000095,
       0.00000000017,     0.00000000918,
       0.00000000490,    -0.00000000076,
      -0.00000000054,    -0.00000000173,
       0.00000000754,     0.00000000486,
       0.00000000363,    -0.00000000743,
      -0.00000000367,    -0.00000000126,
       0.00000000461,    -0.00000000793,
      -0.00000000788,    -0.00000000325,
      -0.00000000100,     0.00000000393,
      -0.00000001609,    -0.00000001002,
      -0.00000000728,     0.00000001600,
       0.00000000795,     0.00000000233,
      -0.00000000276,     0.00000000411,
       0.00000000412,     0.00000000205,
       0.00000000070,    -0.00000000206,
      -0.00000000219,     0.00000000028,
       0.00000000069,     0.00000000171,
      -0.00000000099,     0.00000000212,
       0.00000000182,     0.00000000066,
       0.00000006943,     0.00000003726,
       0.00000002596,    -0.00000006795,
      -0.00000003338,    -0.00000000767,
      -0.00000000081,     0.00000001081,
       0.00000000620,    -0.00000000953,
      -0.00000000953,    -0.00000000456,
      -0.00000000151,     0.00000000476,
       0.00000000906,     0.00000000063,
      -0.00000000101,    -0.00000000730,
      -0.00000000283,     0.00000000106,
       0.00000000257,    -0.00000000896,
      -0.00000000745,    -0.00000000106,
      -0.00000000002,     0.00000000306,
      -0.00000001302,    -0.00000000565,
      -0.00000000255,     0.00000001107,
      -0.00000000041,     0.00000000343,
       0.00000000282,    -0.00000000010,
       0.00000000337,    -0.00000001298,
      -0.00000000904,    -0.00000000077,
       0.00000000035,     0.00000000314,
      -0.00000001178,    -0.00000000512,
      -0.00000000252,     0.00000000815,
       0.00000000283,     0.00000000050,
       0.00000011997,     0.00000001146,
      -0.00000000681,    -0.00000009710,
      -0.00000003868,     0.00000000947,
       0.00000000452,     0.00000001005,
       0.00000000048,    -0.00000000999,
      -0.00000000813,     0.00000000082,
       0.00000000084,     0.00000000326,
      -0.00000001775,     0.00000005508,
       0.00000003827,     0.00000000633,
      -0.00000000009,    -0.00000001331,
      -0.00000000304,     0.00000000061,
       0.00000005371,     0.00000001978,
       0.00000000836,    -0.00000003726,
      -0.00000001297,    -0.00000000090,
       0.00000000031,     0.00000000298,
       0.00000003965,    -0.00000011039,
      -0.00000007684,    -0.00000001581,
      -0.00000000113,     0.00000002676,
       0.00000000611,    -0.00000000087,
      -0.00000002890,    -0.00000000908,
      -0.00000000342,     0.00000001998,
       0.00000000692,     0.00000000011,
       0.00000001259,     0.00000001182,
       0.00000000528,    -0.00000000737,
      -0.00000000220,    -0.00000000108,
       0.00000001183,    -0.00000001340,
      -0.00000000805,    -0.00000000522,
      -0.00000000102,     0.00000000249,
      -0.00000019913,     0.00000058849,
       0.00000040850,     0.00000007591,
       0.00000000309,    -0.00000014200,
      -0.00000003241,     0.00000000535,
       0.00000007079,     0.00000002185,
       0.00000000814,    -0.00000004892,
      -0.00000001694,    -0.00000000016,
       0.00000000066,     0.00000000386,
      -0.00000005598,    -0.00000005363,
      -0.00000002426,     0.00000003296,
       0.00000000995,     0.00000000505,
       0.00000000057,    -0.00000000197,
      -0.00000005289,     0.00000005914,
       0.00000003545,     0.00000002335,
       0.00000000455,    -0.00000001092,
      -0.00000000223,    -0.00000000037,
       0.00000013166,     0.00000012105,
       0.00000005366,    -0.00000007857,
      -0.00000002394,    -0.00000001063,
      -0.00000000095,     0.00000000482,
       0.00000003072,    -0.00000003253,
      -0.00000001944,    -0.00000001380,
      -0.00000000280,     0.00000000594,
       0.00000000155,    -0.00000000075,
      -0.00000000235,    -0.00000000100,
       0.00000001257,     0.00000000207,
       0.00000000027,    -0.00000000680,
      -0.00000000203,     0.00000000019,
      -0.00000062508,    -0.00000055403,
      -0.00000024390,     0.00000037061,
       0.00000011249,     0.00000004772,
       0.00000000409,    -0.00000002259,
      -0.00000007214,     0.00000007773,
       0.00000004638,     0.00000003231,
       0.00000000651,    -0.00000001414,
      -0.00000000286,    -0.00000000061,
       0.00000000103,    -0.00000001564,
      -0.00000000282,     0.00000000143,
      -0.00000003607,    -0.00000000626,
      -0.00000000027,     0.00000001683,
       0.00000000446,    -0.00000000062,
      -0.00000003323,     0.00000006891,
       0.00000002830,     0.00000000778,
       0.00000000029,    -0.00000000640,
       0.00000001269,     0.00000000800,
       0.00000000209,    -0.00000000547,
      -0.00000001370,     0.00000004581,
       0.00000001457,    -0.00000000076,
      -0.00000000140,    -0.00000000243,
       0.00000003001,     0.00000003092,
       0.00000000955,    -0.00000000752,
      -0.00000000102,    -0.00000000174,
      -0.00000003919,    -0.00000030102,
      -0.00000011091,     0.00000004000,
       0.00000001334,     0.00000002194,
      -0.00000002993,    -0.00000001228,
      -0.00000000228,     0.00000001252,
       0.00000000286,    -0.00000000011,
       0.00000008709,    -0.00000017719,
      -0.00000005322,    -0.00000001268,
       0.00000000132,     0.00000000853,
      -0.00000015643,    -0.00000011008,
      -0.00000002640,     0.00000004401,
       0.00000000670,     0.00000000286,
      -0.00000023266,     0.00000037456,
       0.00000010980,     0.00000004702,
       0.00000000237,    -0.00000001727,
       0.00000009365,     0.00000005288,
       0.00000001071,    -0.00000002710,
      -0.00000000419,    -0.00000000048,
      -0.00000003914,    -0.00000004106,
      -0.00000000321,     0.00000000724,
      -0.00000004773,     0.00000003121,
       0.00000000475,     0.00000000654,
       0.00000115426,    -0.00000195902,
      -0.00000057293,    -0.00000022553,
      -0.00000000865,     0.00000008950,
       0.00000000919,    -0.00000000267,
      -0.00000023048,    -0.00000012847,
      -0.00000002585,     0.00000006665,
       0.00000001031,     0.00000000107,
       0.00000015661,     0.00000019578,
       0.00000002044,    -0.00000002654,
      -0.00000000348,     0.00000000035,
       0.00000020306,    -0.00000015416,
      -0.00000002637,    -0.00000002377,
      -0.00000000045,     0.00000000362,
      -0.00000000069,    -0.00000000050,
      -0.00000000069,    -0.00000000039,
      -0.00000038660,    -0.00000048296,
      -0.00000005218,     0.00000006712,
       0.00000000877,    -0.00000000004,
      -0.00000011739,     0.00000009287,
       0.00000001569,     0.00000001316,
       0.00000000006,    -0.00000000205,
       0.00000002234,     0.00000001234,
       0.00000000669,     0.00000000322,
      -0.00000003897,    -0.00000005047,
      -0.00000000273,     0.00000000808,
       0.00000000147,     0.00000000007,
       0.00000189746,     0.00000239072,
       0.00000025855,    -0.00000032921,
      -0.00000004303,     0.00000000008,
       0.00000028186,    -0.00000022561,
      -0.00000003811,    -0.00000003146,
      -0.00000000010,     0.00000000496,
      -0.00000014644,     0.00000001845,
      -0.00000000819,    -0.00000000503,
       0.00000009597,     0.00000018597,
       0.00000000298,    -0.00000001406,
      -0.00000000316,     0.00000000029,
       0.00000000148,    -0.00000000047,
       0.00000000049,    -0.00000000121,
       0.00000044259,    -0.00000016570,
      -0.00000000989,     0.00000000357,
      -0.00000003327,    -0.00000010686,
       0.00000000064,     0.00000000202,
      -0.00000001095,    -0.00000004401,
      -0.00000000192,     0.00000000630,
       0.00000002465,    -0.00000004707,
      -0.00000000349,    -0.00000000569,
      -0.00000203291,     0.00000075270,
       0.00000004693,    -0.00000001590,
      -0.00000000582,    -0.00000001743,
       0.00000008151,     0.00000025050,
      -0.00000000158,    -0.00000000497,
      -0.00000000207,     0.00000000059,
      -0.00000003814,     0.00000009034,
      -0.00000000285,    -0.00000001298,
      -0.00000000079,    -0.00000000113,
      -0.00000000678,     0.00000011965,
       0.00000000077,     0.00000001360,
      -0.00000000112,    -0.00000000095,
       0.00000000119,     0.00000000021,
       0.00000023167,    -0.00000010430,
       0.00000002102,     0.00000001703,
      -0.00000002349,    -0.00000004964,
       0.00000000386,    -0.00000000475,
       0.00000003779,    -0.00000001841,
       0.00000000102,     0.00000000054,
       0.00000000533,     0.00000003343,
      -0.00000000004,    -0.00000000279,
      -0.00000095190,     0.00000040912,
      -0.00000008568,    -0.00000006684,
       0.00000000123,    -0.00000001251,
       0.00000005701,     0.00000011890,
      -0.00000000921,     0.00000001143,
      -0.00000000155,    -0.00000000019,
      -0.00000005521,     0.00000004364,
      -0.00000000707,    -0.00000000138,
      -0.00000000020,    -0.00000000157,
       0.00000001277,    -0.00000003737,
      -0.00000000063,     0.00000000943,
       0.00000000107,     0.00000000036,
       0.00000000042,     0.00000000124,
       0.00000003704,    -0.00000006893,
       0.00000002102,     0.00000000446,
      -0.00000001562,    -0.00000000601,
       0.00000000073,    -0.00000000461,
       0.00000002010,    -0.00000000846,
       0.00000000191,     0.00000000182,
       0.00000000561,     0.00000001736,
      -0.00000000091,     0.00000000105,
      -0.00000012098,     0.00000028739,
      -0.00000008819,    -0.00000001412,
      -0.00000000039,    -0.00000001305,
       0.00000003809,     0.00000001477,
      -0.00000000177,     0.00000001123,
      -0.00000000165,     0.00000000008,
      -0.00000002795,     0.00000001366,
      -0.00000000344,    -0.00000000360,
      -0.00000000181,    -0.00000001559,
      -0.00000000024,     0.00000000095,
      -0.00000000075,    -0.00000000039,
       0.00000001940,    -0.00000001605,
       0.00000000497,     0.00000000621,
      -0.00000000319,    -0.00000000391,
       0.00000000132,    -0.00000000092,
       0.00000000393,    -0.00000000510,
       0.00000000168,     0.00000000074,
       0.00000000422,     0.00000000246,
      -0.00000000003,     0.00000000128,
      -0.00000008024,     0.00000006332,
      -0.00000001879,    -0.00000002732,
       0.00000000796,     0.00000000928,
      -0.00000000310,     0.00000000232,
      -0.00000000623,     0.00000000735,
      -0.00000000262,    -0.00000000191,
      -0.00000000364,     0.00000000028,
      -0.00000000123,    -0.00000000091,
       0.00000000950,    -0.00000000469,
       0.00000000215,     0.00000000443,
      -0.00000000097,    -0.00000000194,
       0.00000000093,    -0.00000000044,
       0.00000000176,    -0.00000000131,
       0.00000000082,     0.00000000146,
      -0.00000003866,     0.00000001783,
      -0.00000000817,    -0.00000001823,
       0.00000000240,     0.00000000464,
      -0.00000000223,     0.00000000110,
      -0.00000000252,     0.00000000223,
      -0.00000000102,    -0.00000000104,
      -0.00000000015,    -0.00000000121,
       0.00000000079,    -0.00000000036,
      -0.00000000792,     0.00000001005,
      -0.00000000668,    -0.00000000472,
       0.00000000133,     0.00000000095,
      -0.00000000058,     0.00000000087,
      -0.00000000104,     0.00000000056,
      -0.00000000122,     0.00000000034,
       0.00000000063,     0.00000000139,
       0.00000000027,     0.00000000123,
       0.00000000141,    -0.00000000054,
      -0.00000000013,     0.00000000113,
       0.00000000121,    -0.00000000003,
       0.00000000110,     0.00000000017,
      -0.00000001379,     0.00000000377,
       0.00000000690,     0.00000001581,
       0.00000000889,    -0.00000000560,
       0.00000000048,     0.00000000135,
       0.00000000154,    -0.00000000079,
       0.00000000067,    -0.00000000490,
      -0.00000000527,    -0.00000000001,
       0.00000000039,     0.00000000282,
      -0.00000000489,    -0.00000000069,
      -0.00000000002,     0.00000000526,
       0.00000000282,    -0.00000000041,
      -0.00000000136,     0.00000000917,
       0.00000000988,     0.00000000011,
      -0.00000000071,    -0.00000000529,
       0.00000000245,     0.00000000028,
      -0.00000000004,    -0.00000000263,
      -0.00000000141,     0.00000000022,
      -0.00000000101,    -0.00000000061,
      -0.00000000045,     0.00000000097,
      -0.00000000052,     0.00000000116,
       0.00000000114,     0.00000000032,
       0.00000000605,    -0.00000004797,
      -0.00000005159,     0.00000000052,
       0.00000000427,     0.00000002756,
       0.00000000968,    -0.00000000307,
      -0.00000000595,    -0.00000000065,
       0.00000000014,     0.00000000639,
       0.00000000341,    -0.00000000055,
       0.00000000466,     0.00000000271,
       0.00000000198,    -0.00000000453,
      -0.00000000222,    -0.00000000067,
       0.00000000252,    -0.00000000501,
      -0.00000000493,    -0.00000000169,
      -0.00000000045,     0.00000000244,
      -0.00000001062,    -0.00000000592,
      -0.00000000415,     0.00000001046,
       0.00000000515,     0.00000000123,
      -0.00000000160,     0.00000000259,
       0.00000000258,     0.00000000116,
       0.00000000038,    -0.00000000128,
      -0.00000000175,     0.00000000031,
       0.00000000061,     0.00000000135,
      -0.00000000067,     0.00000000168,
       0.00000000143,     0.00000000044,
       0.00000004492,     0.00000002050,
       0.00000001341,    -0.00000004343,
      -0.00000002108,    -0.00000000336,
       0.00000000355,    -0.00000000600,
      -0.00000000595,    -0.00000000254,
      -0.00000000078,     0.00000000295,
       0.00000000730,     0.00000000010,
      -0.00000000109,    -0.00000000582,
      -0.00000000225,     0.00000000094,
       0.00000000157,    -0.00000000721,
      -0.00000000593,    -0.00000000048,
       0.00000000013,     0.00000000241,
      -0.00000001186,    -0.00000000335,
      -0.00000000103,     0.00000000986,
      -0.00000000017,     0.00000000301,
       0.00000000244,    -0.00000000023,
       0.00000000197,    -0.00000000817,
      -0.00000000567,    -0.00000000037,
       0.00000000026,     0.00000000196,
      -0.00000000739,    -0.00000000319,
      -0.00000000158,     0.00000000510,
       0.00000000177,     0.00000000031,
       0.00000009180,     0.00000000601,
      -0.00000000711,    -0.00000007395,
      -0.00000002939,     0.00000000795,
       0.00000000364,     0.00000000761,
       0.00000000012,    -0.00000000830,
      -0.00000000670,     0.00000000089,
       0.00000000077,     0.00000000267,
      -0.00000001078,     0.00000003469,
       0.00000002405,     0.00000000369,
      -0.00000000018,    -0.00000000835,
      -0.00000000190,     0.00000000042,
       0.00000003383,     0.00000001218,
       0.00000000508,    -0.00000002344,
      -0.00000000815,    -0.00000000049,
       0.00000000022,     0.00000000187,
       0.00000002506,    -0.00000007197,
      -0.00000005001,    -0.00000000975,
      -0.00000000050,     0.00000001739,
      -0.00000001826,    -0.00000000566,
      -0.00000000210,     0.00000001262,
       0.00000000437,     0.00000000003,
       0.00000000711,     0.00000000635,
       0.00000000281,    -0.00000000407,
      -0.00000000119,    -0.00000000057,
       0.00000000635,    -0.00000000808,
      -0.00000000486,    -0.00000000271,
      -0.00000000048,     0.00000000151,
      -0.00000012318,     0.00000037650,
       0.00000026088,     0.00000004558,
       0.00000000072,    -0.00000009061,
      -0.00000002065,     0.00000000382,
       0.00000004470,     0.00000001358,
       0.00000000496,    -0.00000003087,
      -0.00000001068,    -0.00000000001,
       0.00000000044,     0.00000000243,
      -0.00000003243,    -0.00000002903,
      -0.00000001295,     0.00000001885,
       0.00000000563,     0.00000000261,
      -0.00000002856,     0.00000003504,
       0.00000002091,     0.00000001229,
       0.00000000224,    -0.00000000644,
       0.00000007954,     0.00000006751,
       0.00000002931,    -0.00000004708,
      -0.00000001427,    -0.00000000551,
       0.00000001683,    -0.00000001904,
      -0.00000001131,    -0.00000000744,
      -0.00000000145,     0.00000000344,
       0.00000000275,    -0.00000000095,
      -0.00000000184,    -0.00000000131,
       0.00000000804,    -0.00000000026,
      -0.00000000045,    -0.00000000434,
      -0.00000000129,     0.00000000025,
      -0.00000037330,    -0.00000030321,
      -0.00000013038,     0.00000021940,
       0.00000006623,     0.00000002404,
       0.00000000153,    -0.00000001323,
      -0.00000003941,     0.00000004552,
       0.00000002697,     0.00000001732,
       0.00000000334,    -0.00000000819,
      -0.00000000702,    -0.00000000878,
      -0.00000000089,     0.00000000390,
      -0.00000002233,     0.00000000351,
       0.00000000277,     0.00000001020,
       0.00000000265,    -0.00000000101,
      -0.00000000185,     0.00000004219,
       0.00000001641,    -0.00000000265,
       0.00000000768,     0.00000000039,
      -0.00000000052,    -0.00000000308,
      -0.00000000737,     0.00000002676,
       0.00000000846,    -0.00000000071,
      -0.00000000090,    -0.00000000139,
       0.00000001751,     0.00000001803,
       0.00000000564,    -0.00000000441,
      -0.00000000062,    -0.00000000103,
      -0.00000010119,    -0.00000018360,
      -0.00000006383,     0.00000005540,
       0.00000001457,     0.00000001140,
      -0.00000001814,     0.00000000295,
       0.00000000277,     0.00000000706,
       0.00000000147,    -0.00000000092,
       0.00000004928,    -0.00000010358,
      -0.00000003100,    -0.00000000684,
       0.00000000093,     0.00000000496,
      -0.00000009161,    -0.00000006341,
      -0.00000001522,     0.00000002582,
       0.00000000398,     0.00000000161,
      -0.00000013544,     0.00000022415,
       0.00000006562,     0.00000002698,
       0.00000000111,    -0.00000001037,
       0.00000005473,     0.00000003042,
       0.00000000608,    -0.00000001583,
      -0.00000000246,    -0.00000000022,
      -0.00000002073,    -0.00000002077,
      -0.00000000144,     0.00000000359,
      -0.00000002508,     0.00000001840,
       0.00000000308,     0.00000000344,
       0.00000066432,    -0.00000115649,
      -0.00000033779,    -0.00000012789,
      -0.00000000364,     0.00000005305,
      -0.00000013463,    -0.00000007379,
      -0.00000001465,     0.00000003892,
       0.00000000606,     0.00000000048,
       0.00000008577,     0.00000010095,
       0.00000001013,    -0.00000001417,
      -0.00000000182,     0.00000000034,
       0.00000010574,    -0.00000008732,
      -0.00000001531,    -0.00000001213,
      -0.00000000010,     0.00000000218,
      -0.00000021909,    -0.00000025385,
      -0.00000002651,     0.00000003803,
       0.00000000503,    -0.00000000042,
      -0.00000006088,     0.00000005139,
       0.00000000870,     0.00000000663,
       0.00000000847,     0.00000000770,
       0.00000000409,     0.00000000193,
      -0.00000002360,    -0.00000002549,
      -0.00000000161,     0.00000000494,
       0.00000106710,     0.00000125035,
       0.00000013083,    -0.00000018515,
      -0.00000002452,     0.00000000196,
       0.00000014606,    -0.00000012493,
      -0.00000002115,    -0.00000001582,
       0.00000000015,     0.00000000279,
      -0.00000006605,     0.00000001103,
      -0.00000000490,    -0.00000000302,
       0.00000005776,     0.00000008975,
       0.00000000169,    -0.00000000853,
      -0.00000000181,     0.00000000022,
       0.00000000078,    -0.00000000026,
       0.00000020963,    -0.00000009957,
      -0.00000000595,     0.00000000231,
      -0.00000001999,    -0.00000005008,
       0.00000000041,     0.00000000121,
      -0.00000000693,    -0.00000002393,
      -0.00000000094,     0.00000000356,
       0.00000001299,    -0.00000002630,
      -0.00000000210,    -0.00000000301,
      -0.00000096463,     0.00000045241,
       0.00000002822,    -0.00000001026,
      -0.00000000385,    -0.00000000935,
       0.00000004891,     0.00000011727,
      -0.00000000102,    -0.00000000299,
      -0.00000001931,     0.00000004971,
      -0.00000000175,    -0.00000000714,
      -0.00000000238,     0.00000006577,
       0.00000000075,     0.00000000733,
       0.00000012541,    -0.00000006086,
       0.00000001170,     0.00000000901,
      -0.00000001337,    -0.00000002671,
       0.00000000203,    -0.00000000262,
       0.00000001851,    -0.00000001115,
       0.00000000343,     0.00000001584,
      -0.00000000005,    -0.00000000166,
      -0.00000051886,     0.00000024090,
      -0.00000004800,    -0.00000003550,
       0.00000003250,     0.00000006398,
      -0.00000000484,     0.00000000629,
      -0.00000002799,     0.00000002621,
      -0.00000000418,    -0.00000000083,
       0.00000000720,    -0.00000001692,
      -0.00000000034,     0.00000000560,
       0.00000002232,    -0.00000004162,
       0.00000001264,     0.00000000257,
      -0.00000000928,    -0.00000000370,
       0.00000000043,    -0.00000000274,
       0.00000001102,    -0.00000000518,
       0.00000000114,     0.00000000096,
       0.00000000328,     0.00000000957,
      -0.00000007513,     0.00000017329,
      -0.00000005303,    -0.00000000830,
       0.00000002263,     0.00000000906,
      -0.00000000103,     0.00000000668,
      -0.00000001535,     0.00000000870,
      -0.00000000220,    -0.00000000186,
      -0.00000000074,    -0.00000000885,
       0.00000001063,    -0.00000001085,
       0.00000000350,     0.00000000323,
      -0.00000000216,    -0.00000000208,
       0.00000000067,    -0.00000000066,
       0.00000000237,    -0.00000000317,
       0.00000000104,     0.00000000042,
       0.00000000259,     0.00000000154,
      -0.00000004352,     0.00000004305,
      -0.00000001347,    -0.00000001415,
       0.00000000537,     0.00000000494,
      -0.00000000157,     0.00000000165,
      -0.00000000373,     0.00000000462,
      -0.00000000163,    -0.00000000109,
      -0.00000000219,    -0.00000000003,
       0.00000000550,    -0.00000000316,
      -0.00000000064,    -0.00000000111,
       0.00000000100,    -0.00000000089,
       0.00000000059,     0.00000000081,
      -0.00000002246,     0.00000001210,
      -0.00000000553,    -0.00000001046,
       0.00000000158,     0.00000000266,
      -0.00000000126,     0.00000000073,
      -0.00000000146,     0.00000000150,
      -0.00000000069,    -0.00000000059,
      -0.00000000506,     0.00000000656,
       0.00000000085,     0.00000000060,
      -0.00000000062,     0.00000000040,
      -0.00000000095,     0.00000000027,
       0.00000000049,     0.00000000108,
       0.00000000022,     0.00000000095,
       0.00000000109,    -0.00000000042,
      -0.00000001063,     0.00000000292,
       0.00000000536,     0.00000001217,
       0.00000000684,    -0.00000000434,
       0.00000000037,     0.00000000106,
       0.00000000122,    -0.00000000061,
       0.00000000041,    -0.00000000331,
      -0.00000000356,     0.00000000003,
       0.00000000028,     0.00000000189,
      -0.00000000331,    -0.00000000043,
       0.00000000003,     0.00000000356,
       0.00000000190,    -0.00000000030,
      -0.00000000089,     0.00000000649,
       0.00000000699,     0.00000000000,
       0.00000000167,     0.00000000018,
      -0.00000000003,    -0.00000000180,
       0.00000000374,    -0.00000003324,
      -0.00000003567,     0.00000000085,
       0.00000000322,     0.00000001901,
      -0.00000000406,    -0.00000000042,
       0.00000000012,     0.00000000435,
       0.00000000232,    -0.00000000039,
       0.00000000281,     0.00000000141,
       0.00000000099,    -0.00000000268,
      -0.00000000129,    -0.00000000032,
       0.00000000126,    -0.00000000310,
      -0.00000000302,    -0.00000000075,
      -0.00000000015,     0.00000000146,
      -0.00000000680,    -0.00000000325,
      -0.00000000215,     0.00000000662,
      -0.00000000087,     0.00000000159,
       0.00000000157,     0.00000000062,
      -0.00000000132,     0.00000000029,
       0.00000000049,     0.00000000102,
      -0.00000000045,     0.00000000127,
       0.00000000106,     0.00000000028,
       0.00000002815,     0.00000001003,
       0.00000000577,    -0.00000002680,
      -0.00000001282,    -0.00000000087,
       0.00000000190,    -0.00000000368,
      -0.00000000361,    -0.00000000130,
      -0.00000000035,     0.00000000178,
       0.00000000556,    -0.00000000015,
      -0.00000000097,    -0.00000000440,
      -0.00000000169,     0.00000000076,
       0.00000000093,    -0.00000000549,
      -0.00000000448,    -0.00000000017,
       0.00000000018,     0.00000000181,
      -0.00000000982,    -0.00000000188,
      -0.00000000020,     0.00000000805,
      -0.00000000005,     0.00000000242,
       0.00000000194,    -0.00000000025,
       0.00000000113,    -0.00000000507,
      -0.00000000350,    -0.00000000016,
      -0.00000000459,    -0.00000000193,
      -0.00000000095,     0.00000000317,
       0.00000006795,     0.00000000260,
      -0.00000000657,    -0.00000005451,
      -0.00000002160,     0.00000000640,
      -0.00000000007,    -0.00000000647,
      -0.00000000519,     0.00000000081,
       0.00000000064,     0.00000000206,
      -0.00000000640,     0.00000002158,
       0.00000001493,     0.00000000208,
      -0.00000000020,    -0.00000000517,
       0.00000002109,     0.00000000732,
       0.00000000298,    -0.00000001459,
      -0.00000000508,    -0.00000000024,
       0.00000001540,    -0.00000004616,
      -0.00000003200,    -0.00000000578,
      -0.00000000013,     0.00000001111,
      -0.00000001140,    -0.00000000344,
      -0.00000000124,     0.00000000787,
       0.00000000272,    -0.00000000001,
       0.00000000394,     0.00000000332,
       0.00000000146,    -0.00000000219,
       0.00000000331,    -0.00000000486,
      -0.00000000294,    -0.00000000135,
      -0.00000007420,     0.00000023758,
       0.00000016427,     0.00000002628,
      -0.00000000054,    -0.00000005697,
      -0.00000001296,     0.00000000271,
       0.00000002790,     0.00000000823,
       0.00000000291,    -0.00000001924,
      -0.00000000666,     0.00000000008,
      -0.00000001858,    -0.00000001530,
      -0.00000000670,     0.00000001062,
       0.00000000313,     0.00000000130,
      -0.00000001498,     0.00000002064,
       0.00000001225,     0.00000000622,
       0.00000000102,    -0.00000000376,
       0.00000004746,     0.00000003653,
       0.00000001542,    -0.00000002784,
      -0.00000000838,    -0.00000000268,
       0.00000000899,    -0.00000001105,
      -0.00000000651,    -0.00000000388,
      -0.00000000071,     0.00000000197,
       0.00000000263,    -0.00000000088,
      -0.00000000135,    -0.00000000117,
       0.00000000506,    -0.00000000100,
      -0.00000000062,    -0.00000000273,
      -0.00000022040,    -0.00000016067,
      -0.00000006685,     0.00000012827,
       0.00000003843,     0.00000001125,
      -0.00000002095,     0.00000002641,
       0.00000001553,     0.00000000899,
       0.00000000162,    -0.00000000469,
      -0.00000000836,    -0.00000000482,
      -0.00000000006,     0.00000000398,
      -0.00000001365,     0.00000000610,
       0.00000000328,     0.00000000611,
       0.00000000156,    -0.00000000095,
       0.00000000920,     0.00000002544,
       0.00000000937,    -0.00000000574,
       0.00000000458,    -0.00000000218,
      -0.00000000127,    -0.00000000171,
      -0.00000000389,     0.00000001548,
       0.00000000485,    -0.00000000056,
       0.00000001025,     0.00000001038,
       0.00000000328,    -0.00000000262,
      -0.00000010459,    -0.00000011017,
      -0.00000003607,     0.00000005087,
       0.00000001242,     0.00000000562,
      -0.00000001085,     0.00000000744,
       0.00000000393,     0.00000000393,
       0.00000000073,    -0.00000000102,
       0.00000002752,    -0.00000006015,
      -0.00000001792,    -0.00000000359,
       0.00000000065,     0.00000000285,
      -0.00000005349,    -0.00000003606,
      -0.00000000863,     0.00000001511,
       0.00000000236,     0.00000000088,
      -0.00000007761,     0.00000013311,
       0.00000003890,     0.00000001516,
       0.00000000044,    -0.00000000617,
       0.00000003183,     0.00000001726,
       0.00000000339,    -0.00000000920,
      -0.00000000144,    -0.00000000008,
      -0.00000001085,    -0.00000001036,
      -0.00000000061,     0.00000000170,
      -0.00000001308,     0.00000001091,
       0.00000000200,     0.00000000178,
       0.00000037701,    -0.00000067876,
      -0.00000019794,    -0.00000007110,
      -0.00000000112,     0.00000003122,
      -0.00000007827,    -0.00000004181,
      -0.00000000815,     0.00000002261,
       0.00000000354,     0.00000000016,
       0.00000004683,     0.00000005150,
       0.00000000492,    -0.00000000749,
       0.00000005456,    -0.00000004957,
      -0.00000000892,    -0.00000000609,
      -0.00000012386,    -0.00000013185,
      -0.00000001319,     0.00000002149,
      -0.00000003129,     0.00000002845,
       0.00000000483,     0.00000000329,
       0.00000000277,     0.00000000478,
       0.00000000248,     0.00000000113,
      -0.00000001412,    -0.00000001284,
      -0.00000000092,     0.00000000298,
       0.00000059929,     0.00000064704,
       0.00000006492,    -0.00000010397,
      -0.00000001388,     0.00000000222,
       0.00000007499,    -0.00000006921,
      -0.00000001175,    -0.00000000782,
       0.00000000021,     0.00000000157,
      -0.00000002905,     0.00000000635,
      -0.00000000293,    -0.00000000178,
       0.00000003428,     0.00000004298,
       0.00000000093,    -0.00000000511,
       0.00000009783,    -0.00000005886,
      -0.00000001181,    -0.00000002321,
      -0.00000000434,    -0.00000001286,
      -0.00000000044,     0.00000000199,
       0.00000000671,    -0.00000001462,
      -0.00000000125,    -0.00000000156,
      -0.00000045133,     0.00000026750,
       0.00000001672,    -0.00000000656,
       0.00000002887,     0.00000005425,
      -0.00000000065,    -0.00000000176,
      -0.00000000946,     0.00000002715,
      -0.00000000107,    -0.00000000390,
      -0.00000000042,     0.00000003587,
       0.00000000062,     0.00000000389,
       0.00000006709,    -0.00000003526,
       0.00000000646,     0.00000000469,
      -0.00000000758,    -0.00000001422,
       0.00000000105,    -0.00000000143,
       0.00000000896,    -0.00000000666,
       0.00000000214,     0.00000000737,
      -0.00000027921,     0.00000014067,
      -0.00000002662,    -0.00000001852,
       0.00000001844,     0.00000003404,
      -0.00000000250,     0.00000000344,
      -0.00000001405,     0.00000001555,
      -0.00000000245,    -0.00000000048,
       0.00000000407,    -0.00000000740,
      -0.00000000017,     0.00000000330,
       0.00000001315,    -0.00000002487,
       0.00000000751,     0.00000000144,
      -0.00000000546,    -0.00000000221,
       0.00000000024,    -0.00000000161,
       0.00000000596,    -0.00000000313,
       0.00000000190,     0.00000000521,
      -0.00000004531,     0.00000010348,
      -0.00000003153,    -0.00000000469,
       0.00000001333,     0.00000000541,
      -0.00000000058,     0.00000000393,
      -0.00000000832,     0.00000000545,
      -0.00000000138,    -0.00000000095,
      -0.00000000024,    -0.00000000493,
       0.00000000578,    -0.00000000716,
      -0.00000000142,    -0.00000000110,
       0.00000000140,    -0.00000000194,
       0.00000000157,     0.00000000093,
      -0.00000002343,     0.00000002855,
      -0.00000000926,    -0.00000000719,
       0.00000000353,     0.00000000261,
      -0.00000000078,     0.00000000113,
      -0.00000000219,     0.00000000287,
      -0.00000000101,    -0.00000000061,
      -0.00000000130,    -0.00000000012,
       0.00000000314,    -0.00000000211,
      -0.00000000042,    -0.00000000062,
       0.00000000056,    -0.00000000060,
      -0.00000001285,     0.00000000809,
      -0.00000000369,    -0.00000000590,
       0.00000000104,     0.00000000150,
      -0.00000000083,     0.00000000099,
      -0.00000000800,     0.00000000226,
       0.00000000410,     0.00000000916,
       0.00000000028,     0.00000000081,
       0.00000000093,    -0.00000000046,
       0.00000000024,    -0.00000000220,
      -0.00000000235,     0.00000000006,
      -0.00000000220,    -0.00000000025,
       0.00000000006,     0.00000000236,
       0.00000000112,     0.00000000011,
      -0.00000000004,    -0.00000000121,
       0.00000000214,    -0.00000002251,
      -0.00000002409,     0.00000000100,
       0.00000000242,     0.00000001280,
      -0.00000000271,    -0.00000000025,
       0.00000000011,     0.00000000290,
       0.00000000154,    -0.00000000029,
       0.00000000164,     0.00000000066,
       0.00000000043,    -0.00000000153,
       0.00000000054,    -0.00000000187,
      -0.00000000180,    -0.00000000025,
      -0.00000000423,    -0.00000000161,
      -0.00000000044,     0.00000000096,
       0.00000000093,     0.00000000029,
      -0.00000000095,     0.00000000024,
      -0.00000000029,     0.00000000092,
       0.00000001712,     0.00000000395,
       0.00000000150,    -0.00000001599,
       0.00000000093,    -0.00000000220,
      -0.00000000213,    -0.00000000057,
       0.00000000405,    -0.00000000024,
      -0.00000000080,    -0.00000000319,
      -0.00000000123,     0.00000000059,
       0.00000000052,    -0.00000000401,
      -0.00000000326,    -0.00000000001,
      -0.00000000762,    -0.00000000098,
       0.00000000019,     0.00000000619,
       0.00000000002,     0.00000000183,
       0.00000000146,    -0.00000000023,
       0.00000000063,    -0.00000000310,
      -0.00000000213,    -0.00000000006,
      -0.00000000283,    -0.00000000115,
      -0.00000000055,     0.00000000195,
       0.00000004882,     0.00000000063,
      -0.00000000561,    -0.00000003901,
      -0.00000001541,     0.00000000495,
      -0.00000000016,    -0.00000000481,
      -0.00000000384,     0.00000000067,
       0.00000000052,     0.00000000152,
      -0.00000000371,     0.00000001327,
       0.00000000914,     0.00000000112,
      -0.00000000017,    -0.00000000316,
       0.00000001302,     0.00000000429,
       0.00000000168,    -0.00000000899,
      -0.00000000313,    -0.00000000007,
       0.00000000919,    -0.00000002917,
      -0.00000002016,    -0.00000000328,
       0.00000000007,     0.00000000698,
      -0.00000000704,    -0.00000000204,
      -0.00000000070,     0.00000000486,
       0.00000000168,    -0.00000000003,
       0.00000000214,     0.00000000169,
       0.00000000073,    -0.00000000114,
       0.00000000166,    -0.00000000292,
      -0.00000000178,    -0.00000000062,
      -0.00000004349,     0.00000014803,
       0.00000010207,     0.00000001446,
      -0.00000000109,    -0.00000003532,
      -0.00000000801,     0.00000000190,
       0.00000001722,     0.00000000486,
       0.00000000163,    -0.00000001185,
      -0.00000000409,     0.00000000012,
      -0.00000001052,    -0.00000000781,
      -0.00000000333,     0.00000000590,
       0.00000000170,     0.00000000061,
      -0.00000000758,     0.00000001208,
       0.00000000713,     0.00000000298,
       0.00000000039,    -0.00000000219,
       0.00000002799,     0.00000001909,
       0.00000000774,    -0.00000001625,
      -0.00000000485,    -0.00000000119,
       0.00000000467,    -0.00000000635,
      -0.00000000371,    -0.00000000195,
       0.00000000208,    -0.00000000070,
      -0.00000000095,    -0.00000000089,
       0.00000000315,    -0.00000000106,
      -0.00000000056,    -0.00000000169,
      -0.00000012873,    -0.00000008192,
      -0.00000003246,     0.00000007409,
       0.00000002197,     0.00000000466,
      -0.00000001080,     0.00000001520,
       0.00000000885,     0.00000000447,
       0.00000000073,    -0.00000000266,
      -0.00000000723,    -0.00000000258,
       0.00000000025,     0.00000000324,
      -0.00000000823,     0.00000000582,
       0.00000000284,     0.00000000362,
       0.00000000089,    -0.00000000075,
       0.00000001126,     0.00000001509,
       0.00000000525,    -0.00000000574,
       0.00000000270,    -0.00000000261,
      -0.00000000128,    -0.00000000094,
      -0.00000000200,     0.00000000887,
       0.00000000274,    -0.00000000041,
       0.00000000601,     0.00000000591,
       0.00000000188,    -0.00000000157,
      -0.00000008693,    -0.00000006498,
      -0.00000001994,     0.00000004017,
       0.00000000944,     0.00000000258,
      -0.00000000638,     0.00000000751,
       0.00000000357,     0.00000000215,
       0.00000001514,    -0.00000003471,
      -0.00000001027,    -0.00000000183,
       0.00000000044,     0.00000000162,
      -0.00000003114,    -0.00000002024,
      -0.00000000481,     0.00000000882,
       0.00000000139,     0.00000000046,
      -0.00000004378,     0.00000007848,
       0.00000002289,     0.00000000833,
       0.00000001843,     0.00000000966,
       0.00000000186,    -0.00000000532,
      -0.00000000560,    -0.00000000509,
      -0.00000000676,     0.00000000649,
       0.00000000129,     0.00000000091,
       0.00000021094,    -0.00000039614,
      -0.00000011531,    -0.00000003871,
       0.00000000003,     0.00000001824,
      -0.00000004530,    -0.00000002337,
      -0.00000000444,     0.00000001306,
       0.00000000205,     0.00000000003,
       0.00000002547,     0.00000002596,
       0.00000000233,    -0.00000000392,
       0.00000002786,    -0.00000002816,
      -0.00000000520,    -0.00000000300,
      -0.00000006984,    -0.00000006759,
      -0.00000000639,     0.00000001211,
      -0.00000001591,     0.00000001576,
       0.00000000268,     0.00000000159,
       0.00000000060,     0.00000000295,
       0.00000000148,     0.00000000064,
      -0.00000000836,    -0.00000000644,
      -0.00000000051,     0.00000000178,
       0.00000033595,     0.00000033086,
       0.00000003142,    -0.00000005826,
      -0.00000000781,     0.00000000189,
       0.00000003809,    -0.00000003835,
      -0.00000000652,    -0.00000000378,
      -0.00000001232,     0.00000000354,
      -0.00000000173,    -0.00000000103,
       0.00000002010,     0.00000002035,
       0.00000000048,    -0.00000000303,
       0.00000004475,    -0.00000003427,
      -0.00000000687,    -0.00000001059,
      -0.00000000270,    -0.00000000684,
      -0.00000000019,     0.00000000111,
       0.00000000338,    -0.00000000808,
      -0.00000000074,    -0.00000000080,
      -0.00000020724,     0.00000015585,
       0.00000000977,    -0.00000000417,
       0.00000001678,     0.00000002467,
      -0.00000000041,    -0.00000000103,
      -0.00000000446,     0.00000001471,
      -0.00000000066,    -0.00000000211,
       0.00000000033,     0.00000001941,
       0.00000000046,     0.00000000204,
       0.00000003547,    -0.00000002030,
       0.00000000354,     0.00000000240,
      -0.00000000428,    -0.00000000748,
       0.00000000054,    -0.00000000077,
       0.00000000427,    -0.00000000392,
       0.00000000129,     0.00000000335,
      -0.00000014838,     0.00000008152,
      -0.00000001462,    -0.00000000947,
       0.00000001042,     0.00000001791,
      -0.00000000127,     0.00000000186,
      -0.00000000698,     0.00000000913,
      -0.00000000142,    -0.00000000027,
       0.00000000231,    -0.00000000309,
      -0.00000000007,     0.00000000192,
       0.00000000760,    -0.00000001472,
      -0.00000000319,    -0.00000000129,
       0.00000000319,    -0.00000000187,
       0.00000000109,     0.00000000280,
      -0.00000002665,     0.00000006124,
      -0.00000001859,    -0.00000000255,
       0.00000000779,     0.00000000315,
      -0.00000000031,     0.00000000230,
      -0.00000000446,     0.00000000337,
      -0.00000000086,    -0.00000000047,
      -0.00000000002,    -0.00000000271,
       0.00000000311,    -0.00000000462,
      -0.00000000092,    -0.00000000057,
       0.00000000081,    -0.00000000118,
       0.00000000094,     0.00000000056,
      -0.00000001249,     0.00000001854,
      -0.00000000616,    -0.00000000357,
       0.00000000227,     0.00000000137,
      -0.00000000126,     0.00000000176,
      -0.00000000724,     0.00000000534,
       0.00000000067,     0.00000000083,
      -0.00000000047,     0.00000000064,
       0.00000000308,     0.00000000672,
       0.00000000013,    -0.00000000143,
      -0.00000000152,     0.00000000007,
      -0.00000000144,    -0.00000000013,
       0.00000000007,     0.00000000154,
       0.00000000109,    -0.00000001492,
      -0.00000001592,     0.00000000102,
       0.00000000181,     0.00000000842,
      -0.00000000178,    -0.00000000014,
       0.00000000010,     0.00000000190,
       0.00000000093,     0.00000000025,
       0.00000000016,    -0.00000000111,
       0.00000001008,     0.00000000073,
      -0.00000000062,    -0.00000000918,
       0.00000000038,    -0.00000000129,
      -0.00000000122,    -0.00000000018,
       0.00000000286,    -0.00000000025,
      -0.00000000061,    -0.00000000224,
       0.00000000028,    -0.00000000283,
      -0.00000000229,     0.00000000007,
      -0.00000000564,    -0.00000000045,
       0.00000000004,     0.00000000133,
       0.00000000106,    -0.00000000019,
       0.00000000034,    -0.00000000188,
      -0.00000000128,     0.00000000000,
      -0.00000000174,    -0.00000000066,
      -0.00000000031,     0.00000000120,
       0.00000003417,    -0.00000000038,
      -0.00000000455,    -0.00000002721,
      -0.00000001072,     0.00000000372,
      -0.00000000018,    -0.00000000344,
      -0.00000000274,     0.00000000054,
      -0.00000000209,     0.00000000806,
       0.00000000553,     0.00000000057,
      -0.00000000014,    -0.00000000190,
       0.00000000796,     0.00000000244,
       0.00000000090,    -0.00000000548,
      -0.00000000190,     0.00000000001,
       0.00000000533,    -0.00000001818,
      -0.00000001253,    -0.00000000176,
      -0.00000000430,    -0.00000000118,
      -0.00000000038,     0.00000000296,
       0.00000000114,     0.00000000082,
       0.00000000079,    -0.00000000175,
      -0.00000000107,    -0.00000000025,
      -0.00000002475,     0.00000009114,
       0.00000006265,     0.00000000749,
      -0.00000000123,    -0.00000002161,
       0.00000001051,     0.00000000279,
       0.00000000087,    -0.00000000722,
      -0.00000000248,     0.00000000013,
      -0.00000000588,    -0.00000000382,
      -0.00000000157,     0.00000000322,
      -0.00000000366,     0.00000000702,
       0.00000000413,     0.00000000132,
       0.00000001633,     0.00000000956,
       0.00000000365,    -0.00000000937,
       0.00000000234,    -0.00000000363,
      -0.00000000210,    -0.00000000093,
       0.00000000151,    -0.00000000053,
      -0.00000000065,    -0.00000000062,
       0.00000000192,    -0.00000000089,
      -0.00000000044,    -0.00000000104,
      -0.00000007440,    -0.00000003971,
      -0.00000001455,     0.00000004228,
       0.00000001240,     0.00000000146,
      -0.00000000535,     0.00000000867,
       0.00000000500,     0.00000000210,
       0.00000000028,    -0.00000000149,
      -0.00000000549,    -0.00000000133,
       0.00000000032,     0.00000000238,
      -0.00000000489,     0.00000000465,
       0.00000000217,     0.00000000211,
       0.00000000988,     0.00000000880,
       0.00000000288,    -0.00000000469,
       0.00000000156,    -0.00000000225,
      -0.00000000103,    -0.00000000050,
      -0.00000000101,     0.00000000503,
       0.00000000153,    -0.00000000029,
       0.00000000353,     0.00000000332,
       0.00000000106,    -0.00000000095,
      -0.00000006516,    -0.00000003764,
      -0.00000001072,     0.00000002925,
      -0.00000000369,     0.00000000612,
       0.00000000278,     0.00000000115,
       0.00000000821,    -0.00000001990,
      -0.00000000584,    -0.00000000090,
      -0.00000001807,    -0.00000001121,
      -0.00000000264,     0.00000000514,
      -0.00000002431,     0.00000004596,
       0.00000001337,     0.00000000447,
       0.00000001062,     0.00000000533,
       0.00000000099,    -0.00000000306,
      -0.00000000285,    -0.00000000245,
      -0.00000000346,     0.00000000386,
       0.00000000082,     0.00000000045,
       0.00000011631,    -0.00000022994,
      -0.00000006681,    -0.00000002058,
       0.00000000045,     0.00000001060,
      -0.00000002609,    -0.00000001287,
      -0.00000000237,     0.00000000751,
       0.00000001379,     0.00000001290,
       0.00000000107,    -0.00000000203,
       0.00000001405,    -0.00000001600,
      -0.00000000304,    -0.00000000143,
      -0.00000003926,    -0.00000003414,
      -0.00000000299,     0.00000000680,
      -0.00000000799,     0.00000000872,
       0.00000000148,     0.00000000075,
      -0.00000000012,     0.00000000181,
       0.00000000089,     0.00000000036,
      -0.00000000490,    -0.00000000320,
      -0.00000000028,     0.00000000106,
       0.00000018790,     0.00000016688,
       0.00000001471,    -0.00000003257,
       0.00000001910,    -0.00000002122,
      -0.00000000361,    -0.00000000178,
      -0.00000000494,     0.00000000191,
      -0.00000000102,    -0.00000000058,
       0.00000001165,     0.00000000949,
       0.00000000024,    -0.00000000178,
       0.00000001991,    -0.00000001969,
      -0.00000000394,    -0.00000000472,
      -0.00000000166,    -0.00000000359,
       0.00000000166,    -0.00000000445,
      -0.00000009275,     0.00000008957,
       0.00000000962,     0.00000001095,
      -0.00000000199,     0.00000000791,
      -0.00000000040,    -0.00000000114,
       0.00000000052,     0.00000001043,
       0.00000000032,     0.00000000106,
       0.00000001853,    -0.00000001161,
      -0.00000000241,    -0.00000000390,
       0.00000000199,    -0.00000000228,
       0.00000000076,     0.00000000147,
      -0.00000007787,     0.00000004690,
      -0.00000000796,    -0.00000000475,
       0.00000000587,     0.00000000932,
      -0.00000000064,     0.00000000100,
      -0.00000000342,     0.00000000530,
       0.00000000131,    -0.00000000119,
       0.00000000431,    -0.00000000864,
      -0.00000000185,    -0.00000000074,
       0.00000000168,    -0.00000000111,
       0.00000000062,     0.00000000149,
      -0.00000001532,     0.00000003594,
      -0.00000001084,    -0.00000000132,
       0.00000000452,     0.00000000180,
      -0.00000000016,     0.00000000133,
      -0.00000000236,     0.00000000206,
       0.00000000006,    -0.00000000147,
       0.00000000046,    -0.00000000071,
      -0.00000000658,     0.00000001183,
       0.00000000144,     0.00000000071,
      -0.00000000071,     0.00000000107,
       0.00000000044,    -0.00000000969,
      -0.00000001028,     0.00000000094,
      -0.00000000115,    -0.00000000007,
       0.00000000009,     0.00000000122,
       0.00000000196,    -0.00000000022,
      -0.00000000046,    -0.00000000153,
       0.00000000013,    -0.00000000195,
      -0.00000000157,     0.00000000010,
       0.00000000018,    -0.00000000112,
      -0.00000000106,    -0.00000000037,
       0.00000002336,    -0.00000000081,
      -0.00000000353,    -0.00000001854,
      -0.00000000727,     0.00000000272,
      -0.00000000017,    -0.00000000239,
      -0.00000000190,     0.00000000041,
      -0.00000000115,     0.00000000485,
       0.00000000330,     0.00000000027,
       0.00000000482,     0.00000000135,
       0.00000000045,    -0.00000000332,
       0.00000000299,    -0.00000001120,
      -0.00000000768,    -0.00000000089,
      -0.00000000260,    -0.00000000066,
      -0.00000000019,     0.00000000178,
       0.00000000034,    -0.00000000105,
      -0.00000001363,     0.00000005550,
       0.00000003801,     0.00000000355,
      -0.00000000113,    -0.00000001308,
       0.00000000635,     0.00000000156,
       0.00000000043,    -0.00000000435,
      -0.00000000149,     0.00000000011,
      -0.00000000324,    -0.00000000176,
      -0.00000000069,     0.00000000173,
      -0.00000000165,     0.00000000405,
       0.00000000237,     0.00000000050,
       0.00000000943,     0.00000000452,
       0.00000000155,    -0.00000000534,
       0.00000000111,    -0.00000000205,
      -0.00000000117,    -0.00000000041,
       0.00000000103,    -0.00000000038,
       0.00000000116,    -0.00000000067,
      -0.00000004255,    -0.00000001788,
      -0.00000000565,     0.00000002383,
      -0.00000000251,     0.00000000490,
       0.00000000279,     0.00000000090,
      -0.00000000387,    -0.00000000065,
       0.00000000029,     0.00000000165,
      -0.00000000286,     0.00000000339,
       0.00000000154,     0.00000000122,
       0.00000000759,     0.00000000504,
       0.00000000089,    -0.00000000171,
      -0.00000000049,     0.00000000283,
       0.00000000208,     0.00000000184,
      -0.00000004592,    -0.00000002138,
      -0.00000000557,     0.00000002022,
      -0.00000000209,     0.00000000450,
       0.00000000200,     0.00000000059,
       0.00000000438,    -0.00000001134,
      -0.00000000330,    -0.00000000041,
      -0.00000001046,    -0.00000000612,
      -0.00000000141,     0.00000000299,
      -0.00000001328,     0.00000002674,
       0.00000000776,     0.00000000233,
       0.00000000609,     0.00000000290,
       0.00000000052,    -0.00000000175,
      -0.00000000142,    -0.00000000115,
      -0.00000000174,     0.00000000230,
       0.00000006315,    -0.00000013279,
      -0.00000003848,    -0.00000001064,
      -0.00000001495,    -0.00000000698,
      -0.00000000123,     0.00000000430,
       0.00000000744,     0.00000000631,
       0.00000000047,    -0.00000000104,
       0.00000000698,    -0.00000000909,
      -0.00000000177,    -0.00000000066,
      -0.00000002199,    -0.00000001694,
      -0.00000000396,     0.00000000482,
      -0.00000000028,     0.00000000110,
      -0.00000000285,    -0.00000000157,
       0.00000010481,     0.00000008283,
       0.00000000659,    -0.00000001815,
       0.00000000943,    -0.00000001173,
      -0.00000000200,    -0.00000000081,
      -0.00000000180,     0.00000000099,
       0.00000000669,     0.00000000433,
       0.00000000851,    -0.00000001116,
      -0.00000000223,    -0.00000000204,
      -0.00000000101,    -0.00000000187,
       0.00000000078,    -0.00000000244,
      -0.00000004002,     0.00000005083,
       0.00000000545,     0.00000000470,
      -0.00000000081,     0.00000000422,
       0.00000000049,     0.00000000556,
       0.00000000956,    -0.00000000660,
      -0.00000000135,    -0.00000000201,
       0.00000000091,    -0.00000000131,
       0.00000000044,     0.00000000061,
      -0.00000004034,     0.00000002679,
       0.00000000329,     0.00000000479,
      -0.00000000164,     0.00000000305,
       0.00000000075,    -0.00000000040,
       0.00000000240,    -0.00000000503,
      -0.00000000107,    -0.00000000041,
       0.00000000088,    -0.00000000065,
       0.00000000035,     0.00000000078,
      -0.00000000862,     0.00000002093,
       0.00000000260,     0.00000000101,
      -0.00000000123,     0.00000000124,
      -0.00000000342,     0.00000000743,
       0.00000000090,     0.00000000036,
      -0.00000000040,     0.00000000065,
       0.00000000131,    -0.00000000018,
      -0.00000000033,    -0.00000000102,
       0.00000000005,    -0.00000000131,
       0.00000001566,    -0.00000000092,
      -0.00000000266,    -0.00000001238,
      -0.00000000015,    -0.00000000162,
      -0.00000000128,     0.00000000030,
      -0.00000000061,     0.00000000288,
       0.00000000195,     0.00000000011,
       0.00000000290,     0.00000000072,
       0.00000000021,    -0.00000000199,
       0.00000000162,    -0.00000000681,
      -0.00000000156,    -0.00000000036,
      -0.00000000721,     0.00000003344,
       0.00000002281,     0.00000000145,
       0.00000000380,     0.00000000084,
       0.00000000019,    -0.00000000259,
      -0.00000000176,    -0.00000000074,
      -0.00000000066,     0.00000000233,
       0.00000000135,     0.00000000012,
       0.00000000538,     0.00000000196,
       0.00000000049,    -0.00000000115,
       0.00000000069,    -0.00000000047,
      -0.00000002407,    -0.00000000709,
      -0.00000000152,     0.00000001326,
      -0.00000000108,     0.00000000274,
       0.00000000154,     0.00000000032,
      -0.00000000261,    -0.00000000029,
       0.00000000023,     0.00000000109,
      -0.00000000165,     0.00000000234,
       0.00000000105,     0.00000000069,
       0.00000000542,     0.00000000282,
       0.00000000049,    -0.00000000120,
      -0.00000000023,     0.00000000157,
       0.00000000122,     0.00000000101,
      -0.00000003104,    -0.00000001187,
      -0.00000000276,     0.00000001348,
      -0.00000000116,     0.00000000312,
       0.00000000136,     0.00000000029,
       0.00000000230,    -0.00000000642,
      -0.00000000185,    -0.00000000018,
      -0.00000000603,    -0.00000000330,
      -0.00000000074,     0.00000000173,
      -0.00000000713,     0.00000001547,
       0.00000000448,     0.00000000118,
       0.00000000348,     0.00000000155,
       0.00000000026,    -0.00000000100,
      -0.00000000069,    -0.00000000053,
      -0.00000000087,     0.00000000136,
       0.00000003372,    -0.00000007629,
      -0.00000002205,    -0.00000000532,
      -0.00000000853,    -0.00000000373,
      -0.00000000062,     0.00000000245,
       0.00000000399,     0.00000000302,
       0.00000000340,    -0.00000000515,
      -0.00000000103,    -0.00000000029,
      -0.00000001228,    -0.00000000823,
      -0.00000000192,     0.00000000265,
      -0.00000000164,    -0.00000000076,
       0.00000005828,     0.00000004031,
       0.00000000274,    -0.00000001007,
       0.00000000457,    -0.00000000646,
      -0.00000000110,    -0.00000000034,
      -0.00000000054,     0.00000000049,
       0.00000000380,     0.00000000192,
       0.00000000342,    -0.00000000625,
      -0.00000000125,    -0.00000000084,
      -0.00000000061,    -0.00000000095,
       0.00000000035,    -0.00000000133,
      -0.00000001633,     0.00000002849,
       0.00000000304,     0.00000000191,
      -0.00000000028,     0.00000000224,
       0.00000000039,     0.00000000295,
       0.00000000487,    -0.00000000373,
      -0.00000000075,    -0.00000000102,
       0.00000000040,    -0.00000000074,
      -0.00000002060,     0.00000001520,
       0.00000000183,     0.00000000243,
      -0.00000000077,     0.00000000174,
      -0.00000000475,     0.00000001210,
       0.00000000149,     0.00000000055,
      -0.00000000063,     0.00000000074,
       0.00000001030,    -0.00000000086,
      -0.00000000195,    -0.00000000812,
      -0.00000000012,    -0.00000000107,
      -0.00000000031,     0.00000000170,
       0.00000000173,     0.00000000037,
       0.00000000008,    -0.00000000118,
      -0.00000000093,    -0.00000000019,
      -0.00000000362,     0.00000001995,
       0.00000001354,     0.00000000039,
       0.00000000226,     0.00000000043,
       0.00000000007,    -0.00000000153,
      -0.00000000094,    -0.00000000026,
      -0.00000000020,     0.00000000132,
      -0.00000001346,    -0.00000000209,
      -0.00000000040,     0.00000000152,
      -0.00000000170,    -0.00000000011,
      -0.00000000093,     0.00000000156,
       0.00000000369,     0.00000000155,
       0.00000000027,    -0.00000000081,
       0.00000000072,     0.00000000054,
      -0.00000002034,    -0.00000000642,
      -0.00000000127,     0.00000000874,
      -0.00000000063,     0.00000000208,
       0.00000000118,    -0.00000000362,
      -0.00000000347,    -0.00000000175,
      -0.00000000038,     0.00000000100,
      -0.00000000376,     0.00000000890,
       0.00000000198,     0.00000000082,
      -0.00000000042,     0.00000000080,
       0.00000001767,    -0.00000004361,
      -0.00000001256,    -0.00000000254,
      -0.00000000485,    -0.00000000195,
      -0.00000000030,     0.00000000139,
       0.00000000213,     0.00000000141,
       0.00000000162,    -0.00000000291,
      -0.00000000682,    -0.00000000389,
      -0.00000000091,     0.00000000146,
      -0.00000000094,    -0.00000000036,
       0.00000003229,     0.00000001914,
       0.00000000216,    -0.00000000355,
       0.00000000214,     0.00000000081,
      -0.00000000069,    -0.00000000032,
      -0.00000000605,     0.00000001578,
       0.00000000168,     0.00000000071,
      -0.00000000006,     0.00000000117,
       0.00000000028,     0.00000000155,
      -0.00000001037,     0.00000000857,
       0.00000000102,     0.00000000121,
      -0.00000000035,     0.00000000098,
       0.00000000085,     0.00000000030,
      -0.00000000015,     0.00000000099,
       0.00000000102,     0.00000000017,
      -0.00000000169,     0.00000001177,
       0.00000000132,     0.00000000021,
      -0.00000000108,    -0.00000000003,
      -0.00000000052,     0.00000000101,
      -0.00000001302,    -0.00000000337,
      -0.00000000033,     0.00000000134,
       0.00000000059,    -0.00000000202,
      -0.00000000199,    -0.00000000091,
      -0.00000000194,     0.00000000509,
       0.00000000112,     0.00000000042,
       0.00000000906,    -0.00000002480,
      -0.00000000274,    -0.00000000100,
       0.00000000112,     0.00000000063,
       0.00000000075,    -0.00000000164,
      -0.00000000378,    -0.00000000177,
      -0.00000000042,     0.00000000080,
       0.00000001782,     0.00000000879,
       0.00000000099,    -0.00000000194,
       0.00000000120,     0.00000000031,
      -0.00000000182,     0.00000000864,
       0.00000000092,     0.00000000022,
      -0.00000000511,     0.00000000479,
       0.00000000056,     0.00000000060,
      -0.00000000817,    -0.00000000170,
      -0.00000000016,     0.00000000085,
       0.00000000029,    -0.00000000112,
      -0.00000000113,    -0.00000000046,
       0.00000000453,    -0.00000001403,
      -0.00000000154,    -0.00000000050,
       0.00000000033,    -0.00000000092,
       0.00000000978,     0.00000000385,
       0.00000000043,    -0.00000000106,
       0.00000000218,    -0.00000000790,
      -0.00000000086,    -0.00000000024,
       0.00000029130,     0.00000005148,
      -0.00000002624,     0.00000003510,
      -0.00000000157,    -0.00000000135,
      -0.00000008264,     0.00000001801,
       0.00000000035,     0.00000000741,
      -0.00000000016,    -0.00000000027,
      -0.00000002415,    -0.00000013546,
      -0.00000000020,     0.00000000244,
       0.00000000010,     0.00000000026,
      -0.00000000766,     0.00000000330,
       0.00000000136,    -0.00000000049,
      -0.00000000003,     0.00000000003,
       0.00000000605,    -0.00000000136,
      -0.00000000108,     0.00000000206,
      -0.00000000006,    -0.00000000007,
       0.00000098986,     0.00000004087,
      -0.00000006656,     0.00000009515,
      -0.00000000402,    -0.00000000336,
      -0.00000042660,    -0.00000000227,
       0.00000000034,     0.00000000062,
       0.00000000686,    -0.00000042743,
      -0.00000000150,     0.00000000046,
      -0.00000000759,     0.00000000182,
       0.00000000159,    -0.00000000058,
       0.00000000947,    -0.00000000051,
      -0.00000000038,     0.00000000153,
       0.00000155384,     0.00000006396,
      -0.00000010723,     0.00000015083,
      -0.00000000631,    -0.00000000538,
      -0.00000096198,    -0.00000000453,
       0.00000000007,     0.00000000250,
       0.00000000504,    -0.00000096974,
      -0.00000000066,     0.00000000211,
       0.00000002814,     0.00000000255,
      -0.00000000106,     0.00000000127,
       0.00000000839,     0.00000000560,
      -0.00000000194,     0.00000000086,
       0.00000382276,     0.00000148395,
      -0.00000031269,     0.00000036317,
      -0.00000001542,    -0.00000001475,
      -0.00000282332,     0.00000012554,
       0.00000002548,    -0.00000001904,
      -0.00000000009,     0.00000000165,
      -0.00000038237,    -0.00000275274,
       0.00000001318,    -0.00000002277,
       0.00000000277,     0.00000000014,
       0.00000000739,     0.00000001844,
       0.00000000432,     0.00000000634,
      -0.00000000042,     0.00000000027,
       0.00000005348,     0.00000001444,
      -0.00000000876,     0.00000000533,
      -0.00000000024,    -0.00000000038,
      -0.00000444403,     0.00000534205,
      -0.00000052633,    -0.00000085668,
       0.00000007700,    -0.00000001766,
      -0.00000000123,     0.00000000457,
       0.00003623667,     0.00003260972,
      -0.00000454932,     0.00000476899,
      -0.00000029051,    -0.00000035286,
       0.00000001948,    -0.00000000921,
      -0.00008213453,    -0.00000347631,
       0.00000130673,    -0.00000024245,
       0.00000000739,     0.00000020687,
      -0.00000001685,    -0.00000000183,
      -0.00000611597,    -0.00007890724,
      -0.00000020529,    -0.00000124402,
       0.00000016508,    -0.00000000707,
      -0.00000000235,     0.00000001663,
       0.00000017144,     0.00000049428,
       0.00000003200,     0.00000010522,
      -0.00000000422,    -0.00000000043,
       0.00000004184,     0.00000015616,
      -0.00000010263,     0.00000000799,
       0.00000000678,    -0.00000000698,
      -0.00095469142,     0.00107275612,
      -0.00014699792,    -0.00012653780,
       0.00000937735,    -0.00000925885,
       0.00000036215,     0.00000050264,
      -0.00000002069,     0.00000000726,
      -0.00004865221,    -0.00011157793,
       0.00001058356,    -0.00000952495,
       0.00000059311,     0.00000075347,
      -0.00000004056,     0.00000001759,
      -0.00000000035,    -0.00000000209,
       0.00126313586,     0.00001479346,
      -0.00000473401,     0.00000140936,
      -0.00000007195,    -0.00000116514,
       0.00000006526,     0.00000000703,
      -0.00000000118,     0.00000000410,
       0.00002060916,     0.00124520369,
       0.00000132005,     0.00000538921,
      -0.00000028746,     0.00000005863,
       0.00000000726,    -0.00000006648,
       0.00000000415,     0.00000000154,
      -0.00000048624,    -0.00000164484,
      -0.00000007072,    -0.00000028821,
       0.00000001182,     0.00000000623,
       0.00000000149,     0.00000000324,
       0.00000039253,    -0.00000036714,
       0.00000029311,    -0.00000001364,
      -0.00000002228,     0.00000001712,
      -0.00000000352,    -0.00000000171,
      -0.00000212209,     0.00000458183,
      -0.00000043359,    -0.00000040552,
       0.00000003130,    -0.00000002285,
       0.00001210439,     0.00003905553,
      -0.00000112657,     0.00000056240,
      -0.00000001819,    -0.00000004719,
       0.00000000189,    -0.00000000078,
       0.00061367652,     0.00000014898,
       0.00000010061,     0.00000005278,
      -0.00000000646,    -0.00000023409,
      -0.00000006889,     0.00061308823,
       0.00000007217,     0.00000009351,
       0.00000021656,     0.00000000735,
      -0.00000000057,    -0.00000000075,
      -0.00000029858,     0.00000049141,
       0.00000003465,    -0.00000000524,
       0.00000000021,    -0.00000000186,
      -0.00000010207,     0.00000005395,
       0.00000001878,     0.00000000148,
       0.00000000089,    -0.00000000082,
      -0.00000051497,     0.00000300862,
      -0.00000028385,    -0.00000022409,
       0.00000001783,    -0.00000001721,
       0.00000862896,    -0.00000674386,
       0.00000001863,     0.00000008824,
      -0.00000000767,    -0.00000000785,
       0.00076137557,    -0.00000015290,
       0.00000010287,     0.00000001957,
      -0.00000000767,    -0.00000028718,
       0.00000018322,     0.00076068362,
       0.00000006573,     0.00000018006,
       0.00000027033,     0.00000000350,
      -0.00000000042,    -0.00000000094,
      -0.00000000545,    -0.00000004045,
      -0.00000000383,    -0.00000000447,
      -0.00000000530,    -0.00000003574,
      -0.00000000249,    -0.00000000344,
       0.00000000011,    -0.00000000001,
      -0.00000000010,    -0.00000000091,
       0.00000000011,     0.00000000003,
       0.00000000007,    -0.00000000057,
       0.00000000006,     0.00000000004,
       0.00000000059,     0.00000000010,
      -0.00000000004,     0.00000000005,
       0.00000005462,    -0.00000007343,
      -0.00000000842,    -0.00000000622,
       0.00000000082,    -0.00000001089,
       0.00000000085,     0.00000000057,
       0.00000001099,     0.00000000068,
      -0.00000000056,     0.00000000086,
       0.00000001979,     0.00000001018,
      -0.00000001018,     0.00000001980,
       0.00000000005,    -0.00000000006,
       0.00000000006,    -0.00000000002,
      -0.00000001247,     0.00000000640,
      -0.00000000023,    -0.00000000124,
       0.00000001364,     0.00000000317,
       0.00000000039,    -0.00000000159,
       0.00000000222,    -0.00000000033,
       0.00000000036,     0.00000000223,
      -0.00000000116,     0.00000000439,
      -0.00000000440,    -0.00000000116,
       0.00000000017,    -0.00000000022,
      -0.00000000104,    -0.00000000004,
       0.00000000004,    -0.00000000104,
       0.00000000022,    -0.00000000506,
       0.00000000049,     0.00000000034,
       0.00000000507,     0.00000000024,
      -0.00000000034,     0.00000000048,
       0.00000003802,    -0.00000003724,
      -0.00000001477,     0.00000001487,
      -0.00000000652,    -0.00000000339,
       0.00000000058,     0.00000000125,
       0.00000000617,    -0.00000012073,
       0.00000001137,     0.00000000801,
       0.00000012063,     0.00000000596,
      -0.00000000797,     0.00000001133,
       0.00000006764,     0.00000003887,
       0.00000000249,    -0.00000000438,
      -0.00000003887,     0.00000006768,
       0.00000000437,     0.00000000248,
      -0.00000000089,     0.00000000022,
       0.00000000022,     0.00000000089,
      -0.00000000388,     0.00000000006,
       0.00000000007,     0.00000000389,
      -0.00000000133,     0.00000000002,
       0.00000000002,     0.00000000133,
      -0.00000000075,     0.00000000037,
       0.00000000037,     0.00000000075,
      -0.00000001259,     0.00000000033,
      -0.00000000003,    -0.00000000181,
      -0.00000005847,     0.00000000501,
      -0.00000000006,     0.00000000273,
      -0.00000000159,     0.00000000021,
       0.00000000053,     0.00000000244,
      -0.00000014170,     0.00000007130,
      -0.00000001079,    -0.00000001076,
      -0.00000000247,    -0.00000001369,
       0.00000000088,    -0.00000000076,
       0.00000000958,    -0.00000000023,
      -0.00000000053,     0.00000000878,
       0.00000000037,     0.00000000385,
       0.00000000387,    -0.00000000038,
       0.00000002093,     0.00000012795,
       0.00000000892,    -0.00000001218,
      -0.00000000719,     0.00000000089,
       0.00000000069,     0.00000000049,
       0.00000000084,    -0.00000000338,
      -0.00000000329,    -0.00000000087,
      -0.00000000055,    -0.00000000096,
      -0.00000004216,    -0.00000006316,
       0.00000000921,     0.00000000268,
      -0.00000000043,    -0.00000000075,
       0.00000050772,     0.00000000080,
      -0.00000000004,    -0.00000000527,
       0.00000000603,     0.00000000000,
      -0.00000000001,     0.00000000601,
       0.00000001388,    -0.00000002321,
       0.00000113851,    -0.00000000108,
       0.00000000000,    -0.00000000844,
       0.00000000854,    -0.00000000001,
       0.00000000000,     0.00000000855,
       0.00000000059,     0.00000000058,
      -0.00000000058,     0.00000000059,
       0.00000000000,     0.00000000084,
       0.00000000000,    -0.00000000053,
       0.00000000093,    -0.00000000002,
      -0.00000000066,     0.00000000001,
       0.00000000010,    -0.00000000068,
       0.00000000063,    -0.00000001292,
       0.00000000124,     0.00000000088,
       0.00000001284,     0.00000000054,
      -0.00000000088,     0.00000000125,
       0.00000000033,    -0.00000000075,
       0.00000000074,     0.00000000033,
       0.00000012831,     0.00000007109,
       0.00000000418,    -0.00000000764,
       0.00000000627,    -0.00000012570,
       0.00000001193,     0.00000000844,
       0.00000012541,     0.00000000614,
      -0.00000000839,     0.00000001187,
       0.00000003955,    -0.00000009022,
      -0.00000000883,    -0.00000000227,
       0.00000009027,     0.00000003955,
       0.00000000226,    -0.00000000882,
       0.00000000021,     0.00000000032,
      -0.00000000115,    -0.00000000179,
      -0.00000000326,     0.00000000094,
       0.00000000095,     0.00000000327,
      -0.00000000110,     0.00000000085,
      -0.00000001190,     0.00000000037,
       0.00000000038,     0.00000001193,
      -0.00000000152,     0.00000000002,
       0.00000000002,     0.00000000151,
       0.00000000008,     0.00000000230,
      -0.00000000203,     0.00000000085,
       0.00000000085,     0.00000000203,
      -0.00000001399,     0.00000000037,
      -0.00000000003,    -0.00000000207,
      -0.00000006233,     0.00000000536,
      -0.00000000011,     0.00000000332,
      -0.00000000176,     0.00000000023,
       0.00000000069,     0.00000000272,
       0.00000000096,     0.00000000000,
       0.00000000002,     0.00000000096,
      -0.00000015284,     0.00000007621,
      -0.00000001167,    -0.00000001150,
      -0.00000000279,    -0.00000001410,
       0.00000000100,    -0.00000000075,
       0.00000001034,     0.00000000004,
      -0.00000000030,     0.00000000945,
       0.00000000043,     0.00000000410,
       0.00000000413,    -0.00000000041,
       0.00000002225,     0.00000013590,
       0.00000000947,    -0.00000001293,
      -0.00000000760,     0.00000000094,
       0.00000000077,     0.00000000054,
       0.00000000083,    -0.00000000358,
      -0.00000000348,    -0.00000000087,
      -0.00000000058,    -0.00000000101,
      -0.00000004457,    -0.00000006678,
       0.00000000971,     0.00000000279,
      -0.00000000046,    -0.00000000079,
       0.00000053865,     0.00000000076,
      -0.00000000004,    -0.00000000509,
       0.00000000111,     0.00000000009,
       0.00000000631,     0.00000000000,
      -0.00000000001,     0.00000000630,
       0.00000001459,    -0.00000002437,
       0.00000120554,    -0.00000000026,
       0.00000000000,    -0.00000000898,
       0.00000000893,     0.00000000000,
       0.00000000000,     0.00000000894,
       0.00000000206,     0.00000000059,
      -0.00000000059,     0.00000000205,
       0.00000000134,     0.00000000100,
      -0.00000000100,     0.00000000134,
       0.00000000054,    -0.00000000195,
       0.00000000046,    -0.00000000035,
       0.00000000043,    -0.00000000215,
      -0.00000000119,    -0.00000000030,
      -0.00000000330,     0.00000001691,
       0.00000000888,     0.00000000241,
      -0.00000000369,    -0.00000001503,
       0.00000000284,    -0.00000000018,
      -0.00000000002,     0.00000000025,
      -0.00000003349,     0.00000000762,
      -0.00000000045,    -0.00000000629,
       0.00000000057,     0.00000000004,
       0.00000001853,    -0.00000005419,
       0.00000000515,     0.00000000437,
       0.00000005466,     0.00000001810,
      -0.00000000433,     0.00000000517,
       0.00000000390,    -0.00000000203,
      -0.00000000021,    -0.00000000035,
       0.00000000171,     0.00000000410,
       0.00000000034,    -0.00000000024,
      -0.00000000954,     0.00000001644,
       0.00000000155,     0.00000000137,
       0.00000000049,    -0.00000001046,
       0.00000000082,     0.00000000056,
       0.00000001042,     0.00000000049,
      -0.00000000055,     0.00000000081,
       0.00000000721,    -0.00000000286,
       0.00000000286,     0.00000000721,
      -0.00000000080,     0.00000000044,
      -0.00000000105,    -0.00000000273,
      -0.00000000192,    -0.00000000053,
       0.00000000069,    -0.00000000052,
      -0.00000000233,    -0.00000000287,
       0.00000000058,     0.00000000232,
       0.00000000231,    -0.00000000062,
       0.00000000405,    -0.00000000034,
       0.00000000002,    -0.00000000040,
      -0.00000000175,    -0.00000001363,
      -0.00000000132,     0.00000000007,
      -0.00000000768,     0.00000001045,
       0.00000001052,     0.00000000749,
       0.00000001331,     0.00000000661,
       0.00000000065,    -0.00000000104,
       0.00000001979,    -0.00000003733,
      -0.00000000298,    -0.00000000187,
      -0.00000005631,     0.00000000118,
       0.00000000180,     0.00000005580,
      -0.00000000208,    -0.00000000042,
      -0.00000000119,     0.00000000954,
       0.00000000030,     0.00000000045,
      -0.00000000751,     0.00000000608,
      -0.00000000016,    -0.00000000142,
       0.00000000718,     0.00000000825,
      -0.00000000143,     0.00000000016,
       0.00000000032,    -0.00000000056,
      -0.00000000111,    -0.00000000005,
       0.00000000397,     0.00000000010,
       0.00000000033,     0.00000000338,
      -0.00000001327,     0.00000000743,
      -0.00000000013,    -0.00000000118,
       0.00000000147,     0.00000000001,
      -0.00000000001,     0.00000000147,
       0.00000004450,    -0.00000000004,
      -0.00000000023,    -0.00000000102,
       0.00000000103,    -0.00000000024,
       0.00000000017,     0.00000000081,
      -0.00000000082,     0.00000000017,
      -0.00000000036,    -0.00000000094,
       0.00000000094,    -0.00000000036,
       0.00000000052,    -0.00000000053,
       0.00000000054,     0.00000000052,
      -0.00000000051,     0.00000000064,
       0.00000000066,     0.00000000052,
      -0.00000000082,     0.00000000102,
       0.00000000097,     0.00000000062,
      -0.00000000088,     0.00000000011,
       0.00000000005,     0.00000000094,
      -0.00000000155,     0.00000000020,
       0.00000000014,     0.00000000130,
       0.00000000010,     0.00000000052,
       0.00000000010,     0.00000000070,
       0.00000000093,    -0.00000000035,
      -0.00000000184,     0.00000000249,
       0.00000000258,     0.00000000188,
      -0.00000000221,     0.00000000281,
       0.00000000317,     0.00000000195,
      -0.00000000329,     0.00000000019,
       0.00000000014,     0.00000000345,
      -0.00000000474,     0.00000000051,
       0.00000000032,     0.00000000458,
      -0.00000000031,     0.00000000043,
      -0.00000000062,     0.00000000046,
       0.00000000025,     0.00000000034,
      -0.00000000053,     0.00000000007,
      -0.00000000012,    -0.00000000001,
      -0.00000000011,     0.00000000001,
       0.00000000000,     0.00000000012,
      -0.00000000001,     0.00000000009,
       0.00000000019,     0.00000000075,
      -0.00000000168,     0.00000000048,
       0.00000000091,     0.00000000316,
      -0.00000000165,     0.00000000000,
       0.00000000003,     0.00000000164,
      -0.00000000152,     0.00000000000,
       0.00000000001,     0.00000000239,
      -0.00000000484,     0.00000000620,
       0.00000000925,     0.00000000522,
      -0.00000000025,    -0.00000000025,
      -0.00000001170,     0.00000000037,
       0.00000000059,     0.00000001233,
      -0.00000001428,     0.00000000096,
       0.00000000085,     0.00000001533,
      -0.00000000064,     0.00000000107,
      -0.00000000158,     0.00000000118,
       0.00000000086,     0.00000000110,
      -0.00000000144,     0.00000000011,
      -0.00000000005,    -0.00000000142,
       0.00000000002,    -0.00000000008,
      -0.00000000071,    -0.00000000013,
      -0.00000000049,     0.00000000002,
       0.00000000001,     0.00000000052,
      -0.00000000012,     0.00000000000,
       0.00000000002,     0.00000000058,
       0.00000000000,    -0.00000000008,
       0.00000000000,    -0.00000000002,
      -0.00000000006,     0.00000000001,
      -0.00000000173,     0.00000000003,
       0.00000000007,     0.00000000434,
      -0.00000000669,     0.00000000014,
       0.00000000021,     0.00000001245,
      -0.00000000094,     0.00000000002,
      -0.00000000001,    -0.00000000094,
      -0.00000004096,     0.00000000114,
       0.00000000140,     0.00000005366,
       0.00000000057,     0.00000000192,
       0.00000000062,     0.00000000293,
      -0.00000000317,     0.00000000238,
      -0.00000000021,    -0.00000000029,
       0.00000000204,     0.00000000239,
      -0.00000000381,     0.00000000007,
       0.00000000000,    -0.00000000380,
      -0.00000000035,     0.00000000026,
      -0.00000000023,    -0.00000000035,
      -0.00000000011,     0.00000000004,
      -0.00000000007,     0.00000000003,
       0.00000000003,     0.00000000007,
      -0.00000000002,     0.00000000001,
       0.00000000005,     0.00000000008,
       0.00000000033,     0.00000000057,
      -0.00000000120,     0.00000000042,
       0.00000000076,     0.00000000214,
      -0.00000000012,    -0.00000000071,
      -0.00000000068,     0.00000000011,
      -0.00000000010,    -0.00000000064,
      -0.00000000093,     0.00000000015,
      -0.00000000677,     0.00000000493,
      -0.00000000008,    -0.00000000118,
       0.00000000692,     0.00000000854,
      -0.00000000150,     0.00000000019,
      -0.00000000080,    -0.00000000449,
       0.00000000094,    -0.00000000001,
      -0.00000000436,     0.00000000187,
      -0.00000000011,    -0.00000000096,
      -0.00000000191,    -0.00000000541,
       0.00000000119,    -0.00000000017,
      -0.00000000566,     0.00000000193,
      -0.00000000013,    -0.00000000125,
      -0.00000000019,     0.00000000066,
      -0.00000000088,    -0.00000000156,
       0.00000000079,    -0.00000000036,
      -0.00000000020,    -0.00000000075,
      -0.00000000017,    -0.00000000062,
       0.00000000053,    -0.00000000032,
       0.00000001231,    -0.00000000096,
      -0.00000000060,     0.00000000931,
      -0.00000000010,    -0.00000000289,
       0.00000000220,     0.00000000028,
       0.00000000159,    -0.00000000156,
       0.00000000109,     0.00000000107,
       0.00000000143,     0.00000000092,
      -0.00000000046,     0.00000000101,
      -0.00000005280,     0.00000000271,
       0.00000000414,    -0.00000004056,
       0.00000001557,     0.00000000385,
       0.00000000034,     0.00000000680,
      -0.00000000518,    -0.00000000056,
       0.00000000050,    -0.00000000199,
      -0.00000000212,     0.00000000208,
      -0.00000000147,    -0.00000000183,
      -0.00000000154,     0.00000000073,
      -0.00000000074,    -0.00000000117,
      -0.00000000995,     0.00000000111,
       0.00000000022,     0.00000000129,
      -0.00000000116,     0.00000000012,
      -0.00000000123,     0.00000000010,
       0.00000000253,     0.00000002702,
      -0.00000000185,     0.00000000016,
      -0.00000000259,    -0.00000002847,
       0.00000000121,    -0.00000000010,
       0.00000000114,     0.00000000041,
      -0.00000000074,    -0.00000000026,
       0.00000001065,    -0.00000001052,
       0.00000000083,     0.00000000059,
      -0.00000000523,    -0.00000002322,
       0.00000000134,    -0.00000000040,
       0.00000000395,     0.00000003992,
      -0.00000000280,     0.00000000027,
      -0.00000000407,    -0.00000004205,
       0.00000000182,    -0.00000000017,
       0.00000000587,    -0.00000002605,
       0.00000000194,     0.00000000066,
       0.00000000027,     0.00000000093,
      -0.00000000144,    -0.00000000006,
      -0.00000000835,     0.00000002869,
      -0.00000000126,    -0.00000000044,
      -0.00000000019,    -0.00000000110,
       0.00000000136,    -0.00000000003,
       0.00000001681,    -0.00000001553,
       0.00000000128,     0.00000000092,
       0.00000000033,    -0.00000000090,
      -0.00000001638,     0.00000001988,
      -0.00000000086,    -0.00000000056,
      -0.00000000032,     0.00000000090,
      -0.00000000081,    -0.00000000024,
      -0.00000000045,     0.00000000059,
       0.00000000436,     0.00000006706,
      -0.00000000422,     0.00000000044,
      -0.00000000015,    -0.00000000201,
       0.00000000108,    -0.00000000034,
      -0.00000000640,    -0.00000006157,
       0.00000000274,    -0.00000000028,
       0.00000000941,    -0.00000004264,
       0.00000000329,     0.00000000107,
       0.00000000040,     0.00000000140,
      -0.00000000215,    -0.00000000007,
      -0.00000001347,     0.00000004751,
      -0.00000000215,    -0.00000000072,
      -0.00000000030,    -0.00000000164,
       0.00000000201,    -0.00000000006,
       0.00000002356,    -0.00000002270,
       0.00000000198,     0.00000000147,
       0.00000000051,    -0.00000000149,
       0.00000000126,     0.00000000031,
      -0.00000002545,     0.00000002974,
      -0.00000000132,    -0.00000000088,
      -0.00000000052,     0.00000000151,
      -0.00000000137,    -0.00000000040,
       0.00000000073,    -0.00000000091,
       0.00000000085,     0.00000000068,
      -0.00000000071,     0.00000000090,
      -0.00000000087,    -0.00000000068,
       0.00000000013,     0.00000000137,
       0.00000000133,    -0.00000000013,
       0.00000000956,     0.00000008451,
      -0.00000000632,     0.00000000071,
       0.00000000076,     0.00000000055,
       0.00000000030,    -0.00000000134,
      -0.00000000996,    -0.00000008921,
       0.00000000408,    -0.00000000044,
       0.00000001786,    -0.00000007080,
       0.00000000558,     0.00000000173,
       0.00000000060,     0.00000000208,
      -0.00000000315,    -0.00000000005,
      -0.00000002302,     0.00000007798,
      -0.00000000366,    -0.00000000118,
      -0.00000000047,    -0.00000000244,
       0.00000000297,    -0.00000000012,
       0.00000003654,    -0.00000003299,
       0.00000000305,     0.00000000239,
       0.00000000083,    -0.00000000250,
       0.00000000218,     0.00000000055,
      -0.00000003938,     0.00000004445,
      -0.00000000204,    -0.00000000140,
      -0.00000000083,     0.00000000251,
      -0.00000000234,    -0.00000000067,
       0.00000001495,    -0.00000000657,
       0.00000000078,     0.00000000106,
       0.00000000116,    -0.00000000139,
       0.00000000131,     0.00000000108,
      -0.00000000053,    -0.00000000061,
      -0.00000000114,     0.00000000139,
      -0.00000000134,    -0.00000000108,
       0.00000000760,    -0.00000001480,
      -0.00000000380,     0.00000000660,
       0.00000000019,     0.00000000189,
       0.00000000185,    -0.00000000019,
       0.00000001467,     0.00000012066,
       0.00000001036,    -0.00000001449,
      -0.00000000940,     0.00000000116,
       0.00000000115,     0.00000000080,
       0.00000000045,    -0.00000000195,
      -0.00000000120,    -0.00000000055,
      -0.00000001537,    -0.00000012749,
       0.00000000605,    -0.00000000072,
      -0.00000000073,    -0.00000000052,
      -0.00000000019,     0.00000000105,
       0.00000002513,    -0.00000011049,
       0.00000000951,     0.00000000282,
      -0.00000000045,    -0.00000000107,
       0.00000000090,     0.00000000311,
      -0.00000000462,    -0.00000000001,
      -0.00000003679,     0.00000012755,
      -0.00000000621,    -0.00000000195,
      -0.00000000072,    -0.00000000359,
       0.00000000436,    -0.00000000023,
       0.00000005648,    -0.00000004788,
       0.00000000477,     0.00000000402,
       0.00000000135,    -0.00000000420,
       0.00000000375,     0.00000000095,
      -0.00000006146,     0.00000006660,
      -0.00000000320,    -0.00000000227,
      -0.00000000134,     0.00000000422,
      -0.00000000398,    -0.00000000111,
       0.00000002263,    -0.00000000915,
       0.00000000118,     0.00000000178,
       0.00000000189,    -0.00000000218,
       0.00000000206,     0.00000000179,
      -0.00000002728,     0.00000001600,
      -0.00000000082,    -0.00000000100,
      -0.00000000185,     0.00000000217,
      -0.00000000210,    -0.00000000177,
       0.00000000091,    -0.00000000058,
       0.00000000058,     0.00000000091,
      -0.00000000085,     0.00000000059,
      -0.00000000057,    -0.00000000084,
       0.00000000950,    -0.00000005580,
      -0.00000001190,    -0.00000000400,
      -0.00000000720,     0.00000001140,
       0.00000000029,     0.00000000252,
       0.00000000249,    -0.00000000028,
      -0.00000000011,    -0.00000000098,
      -0.00000000094,     0.00000000010,
       0.00000002228,     0.00000016924,
       0.00000001436,    -0.00000002060,
      -0.00000001388,     0.00000000188,
       0.00000000173,     0.00000000118,
       0.00000000069,    -0.00000000278,
      -0.00000000199,    -0.00000000080,
      -0.00000002349,    -0.00000017904,
      -0.00000001505,     0.00000002197,
       0.00000000891,    -0.00000000116,
      -0.00000000109,    -0.00000000075,
      -0.00000000030,     0.00000000143,
       0.00000000075,     0.00000000039,
       0.00000003905,    -0.00000017622,
      -0.00000001874,     0.00000000946,
       0.00000001639,     0.00000000464,
      -0.00000000083,    -0.00000000179,
       0.00000000134,     0.00000000459,
      -0.00000000673,     0.00000000008,
       0.00000000076,     0.00000000061,
      -0.00000005866,     0.00000020817,
       0.00000002226,    -0.00000001047,
      -0.00000001067,    -0.00000000323,
       0.00000000053,     0.00000000111,
      -0.00000000111,    -0.00000000524,
      -0.00000000042,     0.00000000067,
       0.00000000635,    -0.00000000043,
      -0.00000000075,    -0.00000000054,
       0.00000009134,    -0.00000006990,
       0.00000000758,     0.00000000551,
      -0.00000000043,    -0.00000000079,
       0.00000000219,    -0.00000000714,
      -0.00000000076,     0.00000000035,
       0.00000000653,     0.00000000166,
      -0.00000009631,     0.00000010055,
      -0.00000000515,    -0.00000000385,
      -0.00000000219,     0.00000000718,
       0.00000000077,    -0.00000000036,
      -0.00000000685,    -0.00000000187,
       0.00000000034,     0.00000000072,
       0.00000003451,    -0.00000001313,
       0.00000000188,     0.00000000337,
       0.00000000325,    -0.00000000352,
       0.00000000335,     0.00000000309,
      -0.00000004210,     0.00000002415,
      -0.00000000132,    -0.00000000179,
      -0.00000000313,     0.00000000350,
      -0.00000000340,    -0.00000000302,
       0.00000000038,     0.00000000127,
       0.00000000163,    -0.00000000098,
       0.00000000097,     0.00000000162,
      -0.00000000151,     0.00000000096,
      -0.00000000095,    -0.00000000149,
       0.00000000039,     0.00000000322,
       0.00000000318,    -0.00000000039,
      -0.00000000015,    -0.00000000122,
      -0.00000000119,     0.00000000014,
       0.00000003370,     0.00000022931,
       0.00000001944,    -0.00000002870,
      -0.00000002029,     0.00000000300,
       0.00000000256,     0.00000000170,
       0.00000000101,    -0.00000000387,
      -0.00000000316,    -0.00000000114,
      -0.00000003539,    -0.00000024525,
      -0.00000002039,     0.00000003069,
       0.00000001297,    -0.00000000188,
      -0.00000000162,    -0.00000000108,
      -0.00000000046,     0.00000000190,
       0.00000000136,     0.00000000054,
       0.00000006047,    -0.00000028197,
      -0.00000002982,     0.00000001582,
       0.00000002839,     0.00000000836,
      -0.00000000153,    -0.00000000313,
       0.00000000199,     0.00000000671,
       0.00000000051,    -0.00000000095,
      -0.00000000970,     0.00000000032,
       0.00000000113,     0.00000000085,
      -0.00000009416,     0.00000034050,
       0.00000003620,    -0.00000001807,
      -0.00000001869,    -0.00000000546,
       0.00000000098,     0.00000000192,
      -0.00000000170,    -0.00000000758,
      -0.00000000061,     0.00000000099,
       0.00000000915,    -0.00000000074,
      -0.00000000110,    -0.00000000077,
       0.00000014812,    -0.00000010290,
       0.00000001267,     0.00000000882,
      -0.00000000074,    -0.00000000132,
       0.00000000367,    -0.00000001251,
      -0.00000000133,     0.00000000066,
       0.00000001165,     0.00000000296,
      -0.00000000060,    -0.00000000122,
      -0.00000015421,     0.00000015459,
      -0.00000000865,    -0.00000000704,
       0.00000000045,     0.00000000084,
      -0.00000000370,     0.00000001257,
       0.00000000132,    -0.00000000067,
      -0.00000001208,    -0.00000000324,
       0.00000000063,     0.00000000129,
       0.00000005360,    -0.00000001890,
       0.00000000375,     0.00000000355,
       0.00000000472,    -0.00000000616,
       0.00000000576,     0.00000000441,
      -0.00000006878,     0.00000003926,
      -0.00000000238,    -0.00000000387,
      -0.00000000576,     0.00000000605,
      -0.00000000588,    -0.00000000559,
       0.00000001731,    -0.00000000438,
       0.00000000123,     0.00000000292,
       0.00000000371,    -0.00000000210,
       0.00000000208,     0.00000000370,
      -0.00000002470,     0.00000001068,
      -0.00000000091,    -0.00000000156,
      -0.00000000338,     0.00000000197,
      -0.00000000196,    -0.00000000334,
       0.00000004945,     0.00000000007,
       0.00000000039,     0.00000007796,
      -0.00000000580,    -0.00000000400,
       0.00000000880,    -0.00000000321,
       0.00000000317,     0.00000000885,
      -0.00000003220,     0.00000001016,
      -0.00000001007,    -0.00000001987,
      -0.00000000788,     0.00000000355,
      -0.00000000351,    -0.00000000789,
       0.00000000171,     0.00000000074,
      -0.00000000057,    -0.00000000044,
       0.00000000075,    -0.00000000171,
      -0.00000000045,     0.00000000058,
       0.00000000003,    -0.00000000388,
      -0.00000000031,     0.00000000075,
      -0.00000000390,    -0.00000000001,
       0.00000000078,     0.00000000027,
       0.00000000083,     0.00000000304,
       0.00000000300,    -0.00000000082,
      -0.00000000019,    -0.00000000139,
      -0.00000000138,     0.00000000018,
       0.00000005383,     0.00000028421,
       0.00000002351,    -0.00000003544,
      -0.00000002929,     0.00000000482,
       0.00000000376,     0.00000000241,
      -0.00000000194,    -0.00000000212,
      -0.00000000166,     0.00000000183,
      -0.00000005090,    -0.00000032512,
      -0.00000002650,     0.00000004140,
       0.00000001864,    -0.00000000300,
      -0.00000000239,    -0.00000000152,
      -0.00000000069,     0.00000000245,
       0.00000000222,     0.00000000071,
       0.00000009106,    -0.00000044904,
      -0.00000004767,     0.00000002645,
       0.00000005189,     0.00000001453,
      -0.00000000295,    -0.00000000568,
       0.00000000382,     0.00000001040,
       0.00000000078,    -0.00000000139,
      -0.00000001325,    -0.00000000019,
       0.00000000167,     0.00000000120,
      -0.00000015200,     0.00000056216,
       0.00000005944,    -0.00000003126,
      -0.00000003412,    -0.00000000957,
       0.00000000188,     0.00000000349,
      -0.00000000260,    -0.00000001088,
      -0.00000000086,     0.00000000145,
       0.00000001308,    -0.00000000130,
      -0.00000000161,    -0.00000000108,
       0.00000024339,    -0.00000014590,
      -0.00000001434,     0.00000000741,
       0.00000002315,     0.00000001660,
      -0.00000000142,    -0.00000000250,
       0.00000000951,    -0.00000002392,
      -0.00000000245,     0.00000000126,
       0.00000002173,     0.00000000264,
      -0.00000000119,    -0.00000000232,
      -0.00000025999,     0.00000025285,
       0.00000002568,    -0.00000001428,
      -0.00000001622,    -0.00000001147,
       0.00000000087,     0.00000000158,
      -0.00000000662,     0.00000002341,
       0.00000000247,    -0.00000000130,
      -0.00000002268,    -0.00000000598,
       0.00000000123,     0.00000000240,
       0.00000009696,    -0.00000003362,
       0.00000000939,     0.00000001032,
      -0.00000000081,    -0.00000000069,
       0.00000001123,    -0.00000001300,
      -0.00000000128,     0.00000000071,
       0.00000001239,     0.00000000921,
      -0.00000000081,    -0.00000000141,
      -0.00000013098,     0.00000008049,
      -0.00000000678,    -0.00000000766,
       0.00000000063,     0.00000000075,
      -0.00000001028,     0.00000001269,
       0.00000000133,    -0.00000000073,
      -0.00000001239,    -0.00000000997,
       0.00000000072,     0.00000000132,
      -0.00000000420,    -0.00000001110,
       0.00000014929,    -0.00000007833,
      -0.00000000863,     0.00000001385,
       0.00000009859,     0.00000016234,
      -0.00000001578,    -0.00000000984,
       0.00000002276,    -0.00000002211,
       0.00000001949,     0.00000002072,
      -0.00000016346,     0.00000009932,
       0.00000001465,    -0.00000002259,
      -0.00000004247,    -0.00000005896,
       0.00000000838,     0.00000000543,
      -0.00000002179,     0.00000001872,
       0.00000000236,    -0.00000000178,
      -0.00000001860,    -0.00000002182,
       0.00000000178,     0.00000000237,
      -0.00000000110,     0.00000000038,
      -0.00000000038,    -0.00000000110,
       0.00000000095,    -0.00000000042,
       0.00000000041,     0.00000000095,
       0.00000000175,     0.00000000293,
      -0.00000000037,    -0.00000000096,
       0.00000000290,    -0.00000000177,
      -0.00000000096,     0.00000000041,
       0.00000005584,     0.00000001300,
      -0.00000001969,    -0.00000001312,
      -0.00000000063,    -0.00000000122,
      -0.00000001608,    -0.00000001757,
       0.00000000346,     0.00000000691,
       0.00000000021,    -0.00000000132,
      -0.00000001763,     0.00000001608,
       0.00000000703,    -0.00000000350,
      -0.00000000135,    -0.00000000017,
      -0.00000000171,     0.00000000069,
       0.00000000067,     0.00000000172,
       0.00000000340,    -0.00000000384,
      -0.00000000387,    -0.00000000334,
       0.00000000044,     0.00000000060,
      -0.00000000745,    -0.00000014447,
      -0.00000000979,     0.00000002857,
      -0.00000000175,     0.00000000097,
      -0.00000001743,     0.00000003343,
       0.00000000447,    -0.00000000290,
       0.00000003346,     0.00000001762,
      -0.00000000287,    -0.00000000453,
       0.00000000090,     0.00000000086,
       0.00000000090,    -0.00000000086,
      -0.00000000017,    -0.00000000140,
      -0.00000000160,     0.00000000012,
       0.00000000089,    -0.00000000059,
      -0.00000000067,    -0.00000000088,
       0.00000007429,     0.00000033963,
       0.00000002818,    -0.00000004415,
      -0.00000004332,     0.00000000779,
       0.00000000544,     0.00000000330,
      -0.00000000294,    -0.00000000126,
      -0.00000000070,     0.00000000031,
      -0.00000000142,     0.00000000288,
       0.00000000042,     0.00000000083,
      -0.00000007066,    -0.00000041131,
      -0.00000003338,     0.00000005329,
       0.00000002603,    -0.00000000485,
      -0.00000000346,    -0.00000000212,
      -0.00000000097,     0.00000000294,
       0.00000000343,     0.00000000090,
       0.00000000100,     0.00000000022,
      -0.00000000556,     0.00000000056,
       0.00000000071,     0.00000000561,
       0.00000004060,    -0.00000058281,
      -0.00000005873,     0.00000003048,
       0.00000009954,     0.00000002970,
      -0.00000000620,    -0.00000001136,
      -0.00000000071,     0.00000000047,
       0.00000003021,     0.00000000887,
       0.00000000003,    -0.00000000187,
      -0.00000002461,    -0.00000002459,
       0.00000000266,     0.00000000284,
      -0.00000025797,     0.00000097043,
       0.00000010198,    -0.00000005625,
      -0.00000006815,    -0.00000001831,
       0.00000000392,     0.00000000695,
      -0.00000000406,    -0.00000001543,
      -0.00000000120,     0.00000000211,
       0.00000001881,    -0.00000000219,
      -0.00000000234,    -0.00000000152,
       0.00000000151,    -0.00000000024,
       0.00000045218,    -0.00000023252,
      -0.00000002497,     0.00000001414,
       0.00000005675,     0.00000004654,
      -0.00000000376,    -0.00000000635,
       0.00000002092,    -0.00000005227,
      -0.00000000534,     0.00000000286,
       0.00000004975,     0.00000000645,
      -0.00000000278,    -0.00000000517,
      -0.00000050865,     0.00000049868,
       0.00000005019,    -0.00000002902,
      -0.00000004123,    -0.00000003087,
       0.00000000235,     0.00000000405,
      -0.00000001412,     0.00000005181,
       0.00000000543,    -0.00000000297,
      -0.00000005056,    -0.00000001311,
       0.00000000287,     0.00000000532,
      -0.00000007694,     0.00000004301,
      -0.00000003801,    -0.00000001466,
       0.00000007430,    -0.00000019258,
       0.00000000650,     0.00000000910,
       0.00000002961,     0.00000000296,
       0.00000000382,    -0.00000003256,
       0.00000080801,    -0.00000053410,
      -0.00000003071,     0.00000004020,
       0.00000050739,     0.00000060270,
      -0.00000003550,    -0.00000002711,
       0.00000005108,    -0.00000010690,
      -0.00000000928,     0.00000000499,
       0.00000007027,     0.00000008942,
      -0.00000000576,    -0.00000001019,
      -0.00000067941,     0.00000053476,
       0.00000005667,    -0.00000004593,
      -0.00000017861,    -0.00000020250,
       0.00000001476,     0.00000001862,
      -0.00000007018,     0.00000009036,
       0.00000000908,    -0.00000000550,
      -0.00000008846,    -0.00000006877,
       0.00000000551,     0.00000000911,
       0.00000000913,     0.00000001575,
       0.00000000121,     0.00000000392,
      -0.00000000334,     0.00000000343,
      -0.00000000356,    -0.00000000314,
       0.00000001178,    -0.00000001848,
       0.00000000321,    -0.00000000307,
       0.00000000306,     0.00000000323,
       0.00000000061,    -0.00000000128,
      -0.00000000039,     0.00000000141,
       0.00000002193,     0.00000001716,
      -0.00000000374,    -0.00000000624,
      -0.00000000070,    -0.00000000215,
      -0.00000000024,     0.00000000107,
      -0.00000000052,    -0.00000004929,
      -0.00000000177,     0.00000001071,
       0.00000000105,    -0.00000000094,
      -0.00000005059,     0.00000000013,
       0.00000001077,     0.00000000175,
      -0.00000000100,    -0.00000000099,
      -0.00000000435,    -0.00000000013,
      -0.00000000023,     0.00000000439,
      -0.00000064819,    -0.00000077216,
       0.00000011852,     0.00000026770,
       0.00000000597,    -0.00000004720,
       0.00000000424,     0.00000001245,
       0.00000000258,    -0.00000000482,
      -0.00000000091,     0.00000000038,
       0.00000007111,     0.00000036666,
       0.00000000713,    -0.00000008209,
      -0.00000000763,     0.00000000771,
       0.00000000107,     0.00000000002,
       0.00000036762,    -0.00000007051,
      -0.00000008228,    -0.00000000746,
       0.00000000767,     0.00000000780,
      -0.00000000001,    -0.00000000105,
       0.00000002006,     0.00000000118,
      -0.00000000260,     0.00000000110,
       0.00000000151,    -0.00000002008,
       0.00000000102,     0.00000000258,
       0.00000000000,    -0.00000000247,
      -0.00000000247,     0.00000000000,
      -0.00000000281,     0.00000000082,
      -0.00000000069,     0.00000000096,
       0.00000000099,     0.00000000064,
      -0.00000000264,    -0.00000011968,
      -0.00000000127,    -0.00000000022,
       0.00000000145,    -0.00000004790,
      -0.00000000540,     0.00000000487,
       0.00000000697,    -0.00000000030,
      -0.00000000054,    -0.00000000080,
      -0.00000005321,     0.00000001407,
       0.00000000298,     0.00000000127,
       0.00000001423,     0.00000005355,
       0.00000000125,    -0.00000000306,
      -0.00000000001,     0.00000000108,
       0.00000000117,     0.00000000018,
      -0.00000050993,     0.00000199496,
       0.00000022447,    -0.00000021434,
      -0.00000002842,    -0.00000000905,
      -0.00000008750,     0.00000001935,
       0.00000001049,     0.00000000749,
       0.00000027228,    -0.00000012130,
      -0.00000002248,    -0.00000000779,
      -0.00000000044,     0.00000000279,
      -0.00000012135,    -0.00000027426,
      -0.00000000816,     0.00000002294,
       0.00000000292,     0.00000000037,
      -0.00000000039,    -0.00000000557,
      -0.00000000565,    -0.00000000014,
      -0.00000009135,    -0.00000044983,
      -0.00000003401,     0.00000006062,
       0.00000003498,    -0.00000000757,
      -0.00000000482,    -0.00000000273,
      -0.00000000185,     0.00000000229,
       0.00000000699,     0.00000000156,
      -0.00000000052,    -0.00000000087,
       0.00000000242,    -0.00000001134,
       0.00000000563,     0.00000000077,
      -0.00000000065,    -0.00000000099,
      -0.00000001161,     0.00000000090,
       0.00000000114,     0.00000001176,
       0.00000008014,    -0.00000094882,
      -0.00000010157,     0.00000005258,
       0.00000024168,     0.00000007108,
      -0.00000001383,    -0.00000002633,
      -0.00000000178,     0.00000000127,
       0.00000005079,     0.00000001177,
       0.00000000080,    -0.00000000276,
      -0.00000003794,    -0.00000004070,
       0.00000000452,     0.00000000349,
       0.00000000089,     0.00000000022,
      -0.00000054192,     0.00000196411,
       0.00000020726,    -0.00000011234,
      -0.00000017589,    -0.00000004663,
       0.00000001052,     0.00000001801,
       0.00000000126,    -0.00000000089,
      -0.00000000629,    -0.00000002250,
      -0.00000000167,     0.00000000322,
       0.00000002698,    -0.00000000474,
      -0.00000000355,    -0.00000000212,
      -0.00000000076,    -0.00000000027,
       0.00000000176,     0.00000000368,
      -0.00000006956,     0.00000003320,
      -0.00000001163,    -0.00000000150,
      -0.00000011929,     0.00000006856,
       0.00000001540,    -0.00000000541,
      -0.00000026734,    -0.00000045682,
       0.00000002237,     0.00000006370,
       0.00000013437,     0.00000004227,
       0.00000004182,    -0.00000013877,
       0.00000473105,    -0.00000411308,
      -0.00000045146,     0.00000018910,
       0.00000304350,     0.00000255203,
      -0.00000014085,    -0.00000032014,
      -0.00000000570,     0.00000000274,
      -0.00000009430,    -0.00000059036,
      -0.00000005168,     0.00000002844,
       0.00000005272,     0.00000026993,
      -0.00000000577,    -0.00000001408,
      -0.00000224633,     0.00000256081,
       0.00000025637,    -0.00000015793,
      -0.00000072594,    -0.00000056338,
       0.00000004417,     0.00000007166,
       0.00000000297,    -0.00000000238,
      -0.00000008881,     0.00000032833,
       0.00000003398,    -0.00000001950,
      -0.00000000130,    -0.00000000174,
      -0.00000032325,    -0.00000008183,
       0.00000001912,     0.00000003373,
       0.00000000174,    -0.00000000129,
      -0.00000000135,    -0.00000000445,
       0.00000010208,     0.00000009963,
       0.00000000977,     0.00000002919,
      -0.00000000162,    -0.00000000098,
      -0.00000001460,     0.00000002033,
       0.00000000206,    -0.00000000115,
      -0.00000001919,    -0.00000001241,
       0.00000000133,     0.00000000227,
       0.00000003729,    -0.00000011464,
      -0.00000000231,    -0.00000000142,
       0.00000001352,    -0.00000001907,
      -0.00000000195,     0.00000000127,
       0.00000001869,     0.00000001367,
      -0.00000000129,    -0.00000000194,
      -0.00000000079,     0.00000000099,
      -0.00000000106,    -0.00000000026,
      -0.00000000019,     0.00000000096,
      -0.00000019338,    -0.00000000757,
       0.00000000575,    -0.00000000878,
      -0.00000000753,    -0.00000000028,
      -0.00000000371,     0.00000000690,
      -0.00000000076,    -0.00000000033,
      -0.00000000008,     0.00000000099,
       0.00000000083,    -0.00000000139,
       0.00000000136,     0.00000000092,
       0.00000000025,    -0.00000000104,
      -0.00000001979,    -0.00000000603,
       0.00000000167,     0.00000000141,
      -0.00000000093,     0.00000000008,
      -0.00000028083,    -0.00000055975,
      -0.00000001101,     0.00000011734,
       0.00000000944,    -0.00000000906,
       0.00000010009,    -0.00000002258,
      -0.00000001264,    -0.00000000928,
      -0.00000000055,     0.00000000147,
      -0.00000042449,     0.00000074178,
       0.00000007251,    -0.00000009383,
      -0.00000001111,     0.00000000048,
       0.00000000048,     0.00000000063,
       0.00000074453,     0.00000042951,
      -0.00000009396,    -0.00000007360,
       0.00000000056,     0.00000001129,
       0.00000000049,    -0.00000000062,
       0.00000004306,     0.00000002204,
      -0.00000000419,     0.00000000138,
       0.00000002420,    -0.00000004232,
       0.00000000092,     0.00000000403,
       0.00000510226,     0.00001870737,
       0.00000048530,    -0.00000394260,
      -0.00000033328,     0.00000033680,
       0.00000004446,     0.00000000270,
      -0.00000043314,     0.00000009553,
       0.00000004753,     0.00000004667,
       0.00000000182,    -0.00000000549,
       0.00000123291,    -0.00000609111,
      -0.00000054184,     0.00000078635,
       0.00000008477,    -0.00000000620,
      -0.00000000345,    -0.00000000513,
      -0.00000609523,    -0.00000125398,
       0.00000078458,     0.00000054775,
      -0.00000000520,    -0.00000008535,
      -0.00000000525,     0.00000000354,
      -0.00000020017,    -0.00000010408,
       0.00000001982,    -0.00000000531,
      -0.00000010882,     0.00000019705,
      -0.00000000422,    -0.00000001912,
       0.00000000136,    -0.00000000012,
       0.00000000437,     0.00000002596,
       0.00000000222,    -0.00000000338,
       0.00000002616,    -0.00000000439,
      -0.00000000338,    -0.00000000224,
       0.00000000317,    -0.00000000030,
      -0.00000000405,     0.00000000201,
       0.00000000227,     0.00000000398,
       0.00000001230,     0.00000009373,
       0.00000000563,    -0.00000001744,
      -0.00000000851,     0.00000001074,
       0.00000000164,    -0.00000000066,
       0.00000001394,    -0.00000001098,
      -0.00000000102,     0.00000000021,
      -0.00000001332,    -0.00000001409,
       0.00000000103,     0.00000000202,
       0.00000019521,     0.00000154700,
       0.00000013187,    -0.00000018604,
       0.00000002380,    -0.00000001062,
      -0.00000000418,    -0.00000000154,
      -0.00000000046,    -0.00000000084,
       0.00000000157,    -0.00000000020,
      -0.00000001155,    -0.00000000458,
      -0.00000000741,     0.00000001027,
      -0.00000085704,     0.00000141695,
       0.00000019815,    -0.00000001019,
      -0.00000000242,    -0.00000002157,
      -0.00000471706,    -0.00000267186,
       0.00000004280,     0.00000065628,
       0.00000007536,    -0.00000000847,
       0.00000143776,     0.00000076974,
       0.00000007488,    -0.00000006092,
      -0.00000000498,    -0.00000000360,
       0.00000076073,    -0.00000144639,
      -0.00000006198,    -0.00000007160,
       0.00000002054,     0.00000002145,
       0.00000002498,    -0.00000002887,
       0.00003041452,    -0.00004331066,
      -0.00000592373,     0.00000045041,
       0.00000006813,     0.00000057480,
       0.00000003454,    -0.00000000227,
       0.00002526051,     0.00001317573,
      -0.00000040413,    -0.00000336982,
      -0.00000031861,     0.00000004305,
       0.00000000138,     0.00000002103,
      -0.00000354257,    -0.00000651266,
      -0.00000052744,     0.00000049954,
       0.00000003049,     0.00000002586,
      -0.00000672016,     0.00000366056,
       0.00000053600,     0.00000054147,
       0.00000002545,    -0.00000003063,
      -0.00000009894,    -0.00000009161,
       0.00000001228,    -0.00000001068,
      -0.00000009932,     0.00000012494,
      -0.00000001012,    -0.00000001409,
      -0.00000212226,     0.00000831309,
       0.00000087404,    -0.00000052737,
      -0.00000004383,    -0.00000005760,
      -0.00000219163,    -0.00000054390,
       0.00000013853,     0.00000022763,
       0.00000001456,    -0.00000001127,
      -0.00000001087,     0.00000002783,
       0.00000000342,    -0.00000000193,
      -0.00000001781,    -0.00000000352,
       0.00000000096,     0.00000000232,
      -0.00000000194,     0.00000000056,
      -0.00000000412,    -0.00000000132,
       0.00000002830,    -0.00000001354,
      -0.00000001808,    -0.00000004356,
       0.00000000107,     0.00000000187,
      -0.00000001334,    -0.00000000132,
      -0.00000000128,     0.00000001471,
       0.00000083688,     0.00000054517,
       0.00000005195,    -0.00000003374,
       0.00000009736,     0.00000018885,
      -0.00000000652,    -0.00000001005,
      -0.00000003388,     0.00000009373,
       0.00000000944,    -0.00000000726,
      -0.00000008795,    -0.00000001781,
       0.00000000379,     0.00000000921,
       0.00000008453,    -0.00000058474,
      -0.00000005696,     0.00000003743,
      -0.00000000723,    -0.00000000250,
       0.00000000074,     0.00000000099,
       0.00000002365,    -0.00000008779,
      -0.00000000921,     0.00000000552,
       0.00000008697,     0.00000002245,
      -0.00000000549,    -0.00000000916,
       0.00000000064,    -0.00000000064,
      -0.00000002131,    -0.00000002681,
      -0.00000001951,     0.00000001898,
       0.00000000048,     0.00000000082,
      -0.00000002285,    -0.00000000488,
      -0.00000000453,     0.00000002088,
      -0.00000228793,     0.00000028259,
       0.00000003228,    -0.00000006289,
       0.00000008757,    -0.00000010048,
      -0.00000000171,    -0.00000000534,
      -0.00000014618,    -0.00000002731,
       0.00000000275,    -0.00000000475,
      -0.00000004195,     0.00000015429,
      -0.00000000355,    -0.00000000069,
      -0.00000011992,    -0.00000004974,
       0.00000000064,     0.00000000224,
       0.00000000244,    -0.00000000769,
      -0.00000000075,     0.00000000053,
       0.00000000735,     0.00000000277,
      -0.00000000054,    -0.00000000076,
      -0.00000008920,    -0.00000001232,
       0.00000000123,    -0.00000000514,
       0.00000000132,     0.00000000105,
      -0.00000000113,     0.00000000174,
      -0.00000004711,    -0.00000001473,
      -0.00000000038,     0.00000000165,
       0.00000000066,    -0.00000000056,
       0.00000000101,    -0.00000000155,
      -0.00000031314,    -0.00000007694,
       0.00000000596,    -0.00000001346,
       0.00000000221,    -0.00000001058,
      -0.00000000151,    -0.00000000027,
       0.00000000019,    -0.00000000117,
       0.00000006187,     0.00000001295,
      -0.00000000022,     0.00000000093,
       0.00000000025,    -0.00000000117,
       0.00000003243,     0.00000000922,
       0.00000000072,    -0.00000000020,
      -0.00000000007,    -0.00000000005,
       0.00000000002,     0.00000000023,
       0.00000000027,     0.00000000014,
      -0.00000000003,    -0.00000000004,
      -0.00000000004,     0.00000000003,
       0.00000000253,    -0.00000000075,
      -0.00000000037,     0.00000000021,
       0.00000000021,     0.00000000037,
       0.00000000324,    -0.00000000106,
      -0.00000000043,    -0.00000000018,
      -0.00000000084,     0.00000000036,
       0.00000000036,     0.00000000084,
       0.00000000695,    -0.00000000434,
      -0.00000000100,    -0.00000000026,
      -0.00000000308,     0.00000000116,
       0.00000000123,     0.00000000327,
      -0.00000000022,    -0.00000000034,
      -0.00000000030,     0.00000000025,
      -0.00000001935,    -0.00000000249,
       0.00000011087,    -0.00000037540,
      -0.00000007156,    -0.00000000468,
      -0.00000013157,     0.00000013921,
       0.00000002975,    -0.00000002035,
       0.00000007364,     0.00000015203,
      -0.00000000797,    -0.00000001021,
       0.00000701267,     0.00004523080,
       0.00000365354,    -0.00000585153,
      -0.00000057636,     0.00000003668,
       0.00000002346,     0.00000003495,
      -0.00012817835,     0.00002135490,
       0.00001654981,     0.00001034959,
      -0.00000008552,    -0.00000161143,
      -0.00000009773,     0.00000006211,
       0.00004481372,    -0.00002589942,
      -0.00000325904,     0.00000139705,
       0.00000017265,     0.00000018152,
       0.00000000862,    -0.00000000682,
      -0.00002575038,    -0.00004512028,
       0.00000133669,     0.00000325343,
       0.00000018130,    -0.00000016543,
      -0.00000000805,    -0.00000000652,
      -0.00000087560,    -0.00000106562,
       0.00000008495,    -0.00000004804,
       0.00000000248,     0.00000000395,
      -0.00000118090,     0.00000075836,
      -0.00000003252,    -0.00000006603,
       0.00000000398,    -0.00000000400,
      -0.00011384026,    -0.00123813777,
      -0.00010153043,     0.00015932042,
       0.00001594594,    -0.00000098275,
      -0.00000063837,    -0.00000089546,
      -0.00000003181,     0.00000006432,
       0.00058969484,    -0.00009452568,
      -0.00007564109,    -0.00004755891,
       0.00000040672,     0.00000742477,
       0.00000041494,    -0.00000029694,
      -0.00000002819,    -0.00000001476,
      -0.00008712241,     0.00024449197,
       0.00002527179,    -0.00001567875,
      -0.00000135254,    -0.00000149543,
      -0.00000006654,     0.00000006421,
       0.00024408073,     0.00008861164,
      -0.00001542498,    -0.00002546948,
      -0.00000154115,     0.00000134721,
       0.00000006703,     0.00000007212,
       0.00000415515,     0.00000518194,
      -0.00000038696,     0.00000015708,
      -0.00000001544,    -0.00000001528,
       0.00000532796,    -0.00000381911,
       0.00000012421,     0.00000033673,
      -0.00000001562,     0.00000001801,
      -0.00000007913,     0.00000023389,
       0.00000003165,    -0.00000002101,
      -0.00000011352,    -0.00000002140,
       0.00000000956,     0.00000001249,
       0.00000000316,    -0.00000003414,
      -0.00000000305,     0.00000000221,
      -0.00000003890,    -0.00000000405,
       0.00000000236,     0.00000000340,
       0.00000000067,     0.00000000171,
       0.00000000150,    -0.00000000060,
      -0.00000000068,    -0.00000000148,
       0.00000000102,     0.00000003918,
      -0.00000000124,    -0.00000000015,
       0.00000000029,    -0.00000000340,
      -0.00000000336,    -0.00000000034,
       0.00000000647,     0.00000000160,
       0.00000000711,    -0.00000002897,
      -0.00000004109,    -0.00000001124,
       0.00000000241,     0.00000000344,
      -0.00000002804,    -0.00000000022,
      -0.00000000009,     0.00000002838,
       0.00000016756,     0.00000076536,
       0.00000007408,    -0.00000005249,
       0.00000011492,     0.00000003253,
      -0.00000000721,    -0.00000001175,
      -0.00000002453,     0.00000000937,
       0.00000000079,    -0.00000000064,
      -0.00000001048,     0.00000002037,
       0.00000000051,     0.00000000094,
      -0.00000000141,    -0.00000000057,
      -0.00000014078,    -0.00000128332,
      -0.00000012728,     0.00000008758,
       0.00000005587,     0.00000000784,
      -0.00000000334,    -0.00000000482,
      -0.00000000045,    -0.00000000820,
      -0.00000000093,     0.00000000077,
       0.00000001115,     0.00000000303,
      -0.00000000086,    -0.00000000102,
       0.00000000142,     0.00000000039,
      -0.00000043688,    -0.00000072083,
      -0.00000005603,     0.00000005615,
      -0.00000052987,     0.00000037605,
       0.00000004135,     0.00000004091,
       0.00000000139,    -0.00000000165,
      -0.00000068737,     0.00000004141,
       0.00000000244,    -0.00000001061,
       0.00000003759,     0.00000067832,
      -0.00000001042,    -0.00000000225,
       0.00000000073,    -0.00000000227,
       0.00000000137,    -0.00000000083,
      -0.00003019934,     0.00001433534,
       0.00000080782,    -0.00000104035,
      -0.00000005304,    -0.00000003879,
       0.00000233549,    -0.00000185579,
      -0.00000018165,    -0.00000018119,
      -0.00000000971,     0.00000001079,
      -0.00000436553,     0.00000027174,
       0.00000002503,    -0.00000006973,
      -0.00000000062,     0.00000000325,
       0.00000019381,     0.00000442684,
      -0.00000006592,    -0.00000001821,
       0.00000000351,     0.00000000049,
      -0.00000000125,     0.00000000386,
      -0.00000000010,     0.00000000141,
      -0.00000000867,     0.00000000589,
       0.00000000131,    -0.00000000028,
      -0.00000066916,    -0.00000009755,
       0.00000000590,     0.00000001289,
      -0.00000000227,    -0.00000002531,
      -0.00000000284,     0.00000000182,
       0.00000002302,     0.00000000026,
      -0.00000000184,    -0.00000000287,
      -0.00000000105,     0.00000000367,
      -0.00000000104,     0.00000000004,
       0.00000000002,    -0.00000000098,
      -0.00000077792,    -0.00000004860,
       0.00000000419,    -0.00000003885,
       0.00000000110,     0.00000000042,
       0.00000001097,     0.00000000321,
      -0.00000000456,     0.00000001360,
      -0.00000035076,    -0.00000007401,
      -0.00000000053,     0.00000000309,
       0.00000000669,    -0.00000000123,
       0.00000000375,    -0.00000001554,
      -0.00000000007,     0.00000000474,
       0.00000000127,     0.00000000034,
      -0.00000000045,     0.00000000146,
      -0.00000601716,    -0.00000114464,
       0.00000006381,    -0.00000017710,
       0.00000003588,    -0.00000020941,
       0.00000000595,     0.00000000151,
      -0.00000002506,    -0.00000000204,
       0.00000000015,    -0.00000001416,
       0.00000000086,    -0.00000000074,
      -0.00000000088,    -0.00000000100,
       0.00000066267,     0.00000008943,
      -0.00000000115,     0.00000000688,
       0.00000000665,     0.00000000122,
       0.00000000236,    -0.00000001137,
       0.00000048419,     0.00000010268,
      -0.00000000249,     0.00000001183,
       0.00000000156,     0.00000000024,
      -0.00000000033,     0.00000000163,
      -0.00000000009,    -0.00000000006,
      -0.00000000100,    -0.00000000039,
       0.00000000120,     0.00000000521,
      -0.00000000273,     0.00000000470,
       0.00000000029,     0.00000000090,
       0.00000002244,    -0.00000000210,
       0.00000000412,    -0.00000001821,
      -0.00000878283,     0.00004828990,
       0.00000483598,    -0.00000313955,
      -0.00000026033,    -0.00000029120,
      -0.00000001332,     0.00000001250,
       0.00003603706,     0.00000822492,
      -0.00000233268,    -0.00000364214,
      -0.00000021957,     0.00000019723,
       0.00000000945,     0.00000000985,
       0.00004685959,     0.00000041639,
       0.00000004706,    -0.00000001506,
       0.00000000070,    -0.00000002207,
       0.00000024899,    -0.00004719596,
      -0.00000003953,     0.00000001355,
      -0.00000001680,     0.00000000058,
      -0.00000005074,    -0.00000005087,
       0.00000000507,     0.00000000815,
      -0.00000007941,    -0.00000011719,
       0.00000001277,     0.00000001100,
      -0.00106084003,    -0.00103228516,
      -0.00008389331,     0.00007735745,
       0.00000466477,     0.00000546657,
       0.00000023803,    -0.00000021851,
      -0.00015947479,    -0.00004109744,
       0.00001028120,     0.00001619835,
       0.00000096580,    -0.00000086692,
      -0.00000004153,    -0.00000004398,
      -0.00000000123,     0.00000000154,
       0.00028383116,    -0.00000191408,
       0.00000015466,     0.00000008756,
      -0.00000001127,    -0.00000009383,
       0.00000000058,     0.00000000043,
      -0.00000261202,    -0.00028213098,
       0.00000028783,    -0.00000020074,
      -0.00000010778,    -0.00000000276,
      -0.00000000095,     0.00000000159,
       0.00000031304,     0.00000039120,
      -0.00000000574,    -0.00000008341,
      -0.00000000389,    -0.00000000101,
       0.00000029053,     0.00000012370,
      -0.00000009628,    -0.00000003752,
      -0.00000000128,     0.00000000321,
      -0.00000035348,     0.00000006399,
       0.00000003504,     0.00000000711,
      -0.00000000253,    -0.00000000307,
      -0.00000000329,    -0.00000003546,
      -0.00000000209,     0.00000000333,
       0.00000001797,    -0.00000000521,
      -0.00000000308,    -0.00000000156,
       0.00000004524,     0.00000000003,
      -0.00000000046,     0.00000000147,
       0.00000000143,     0.00000000049,
       0.00000001130,     0.00000003146,
       0.00000000112,    -0.00000001082,
      -0.00000000106,     0.00000000136,
      -0.00000001096,    -0.00000000108,
       0.00000000147,     0.00000000100,
       0.00000000094,     0.00000000234,
       0.00000000045,    -0.00000000062,
       0.00000000246,    -0.00000000096,
      -0.00000000062,    -0.00000000045,
      -0.00000000050,    -0.00000000363,
      -0.00000000364,     0.00000000051,
      -0.00000000820,     0.00000000002,
      -0.00000000579,     0.00000000125,
       0.00000000181,     0.00000000599,
       0.00000001870,    -0.00000011075,
      -0.00000001133,     0.00000001538,
      -0.00000006531,     0.00000001098,
       0.00000000922,     0.00000000727,
      -0.00000000057,    -0.00000003771,
      -0.00000000263,     0.00000000164,
      -0.00000004197,     0.00000000186,
       0.00000000203,     0.00000000350,
       0.00000000080,    -0.00000000098,
      -0.00000022262,    -0.00000119598,
      -0.00000009592,     0.00000015721,
       0.00000009411,    -0.00000000927,
      -0.00000001074,    -0.00000000780,
      -0.00000000004,     0.00000000106,
      -0.00000000563,     0.00000002312,
       0.00000000204,    -0.00000000137,
       0.00000002461,     0.00000000555,
      -0.00000000155,    -0.00000000208,
      -0.00000000707,     0.00000000474,
       0.00000000278,     0.00000000627,
       0.00000000558,     0.00000000074,
       0.00000000060,    -0.00000000557,
      -0.00000175766,    -0.00000002224,
      -0.00000000318,    -0.00000006589,
       0.00000000180,     0.00000000611,
      -0.00000000662,    -0.00000000037,
      -0.00000176763,    -0.00000030256,
      -0.00000000662,     0.00000000163,
       0.00000000060,     0.00000000088,
       0.00000011358,    -0.00000000268,
      -0.00000000051,     0.00000000208,
       0.00000000694,    -0.00000011424,
       0.00000000203,     0.00000000057,
      -0.00000023364,     0.00000003551,
       0.00000002008,     0.00000013833,
      -0.00000000303,     0.00000000043,
       0.00000002037,     0.00000000490,
       0.00000000121,     0.00000001964,
      -0.00000000020,    -0.00000000106,
      -0.00000000026,    -0.00000000083,
      -0.00015906521,     0.00000868636,
       0.00000071899,    -0.00000247224,
      -0.00000000703,     0.00000011286,
      -0.00000036463,    -0.00000565402,
       0.00000009167,     0.00000002043,
      -0.00000000346,     0.00000000013,
      -0.00000044240,     0.00000002075,
       0.00000000363,    -0.00000000458,
      -0.00000000586,    -0.00000013921,
      -0.00000000295,    -0.00000000251,
       0.00000001945,    -0.00000002345,
      -0.00000000178,    -0.00000000048,
      -0.00000002662,    -0.00000001537,
      -0.00000000112,     0.00000000122,
       0.00000739292,     0.00000037870,
      -0.00000002419,     0.00000012061,
      -0.00000000736,     0.00000006523,
      -0.00000000099,     0.00000000002,
       0.00000014125,     0.00000001876,
      -0.00000000157,     0.00000000474,
       0.00000002148,    -0.00000017541,
       0.00000000470,     0.00000000153,
      -0.00000001697,    -0.00000000436,
       0.00000846319,     0.00000121736,
      -0.00000008288,     0.00000025227,
      -0.00000003013,     0.00000021062,
      -0.00000000596,    -0.00000000191,
       0.00000002130,     0.00000000094,
      -0.00000000233,     0.00000002255,
      -0.00000000416,     0.00000000022,
      -0.00000000018,    -0.00000000418,
      -0.00000000065,     0.00000000069,
      -0.00000000059,    -0.00000000057,
       0.00000000414,     0.00000000020,
      -0.00000000016,     0.00000000409,
       0.00000000000,     0.00000000014,
       0.00000000000,     0.00000000055,
      -0.00000000001,     0.00000000157,
      -0.00000000106,     0.00000001426,
       0.00000000025,    -0.00000000114,
       0.00000000135,     0.00000000017,
       0.00003519431,     0.00000096953,
      -0.00000006460,    -0.00000008069,
       0.00000000292,    -0.00000001671,
       0.00000043524,    -0.00002115025,
      -0.00000003612,     0.00000004444,
      -0.00000000941,    -0.00000000210,
       0.00000085898,    -0.00000061031,
       0.00000000373,     0.00000004045,
      -0.00000000067,     0.00000000326,
      -0.00000017518,     0.00000008751,
       0.00000001856,    -0.00000006218,
       0.00000000035,     0.00000000393,
       0.00000001124,     0.00000018622,
      -0.00000000119,     0.00000000784,
      -0.00000001502,     0.00000015196,
       0.00000000028,    -0.00000001437,
       0.02068375002,    -0.00000729670,
       0.00000503019,     0.00000059205,
      -0.00000006393,    -0.00000773574,
      -0.00000193927,     0.00075571302,
       0.00000016156,    -0.00000008498,
       0.00000029016,     0.00000000938,
      -0.00000000087,    -0.00000000047,
       0.00001202183,    -0.00000161566,
      -0.00000044239,     0.00000032365,
       0.00000002314,    -0.00000002125,
       0.00000000136,     0.00000000094,
       0.00000264818,    -0.00002754257,
      -0.00000008045,     0.00000028085,
      -0.00000003773,    -0.00000001327,
       0.00000000157,    -0.00000000049,
      -0.00000285083,     0.00000280254,
       0.00000024986,     0.00000015678,
       0.00000001292,    -0.00000001123,
      -0.00000000054,    -0.00000000056,
       0.00000361163,     0.00000218593,
       0.00000018721,    -0.00000018606,
      -0.00000001357,    -0.00000001055,
       0.00003328275,    -0.00000408906,
      -0.00000005686,     0.00000011216,
      -0.00000005966,     0.00000021211,
       0.00000000605,     0.00000000625,
       0.00000346343,    -0.00000016950,
      -0.00000001225,     0.00000005209,
       0.00000000007,    -0.00000000228,
      -0.00000022535,    -0.00000347756,
       0.00000005620,     0.00000001782,
      -0.00000000226,    -0.00000000009,
       0.00000000020,     0.00000000080,
       0.00000000090,    -0.00000000015,
       0.00000000028,     0.00000000089,
       0.00000000097,    -0.00000000030,
      -0.00000000024,    -0.00000000095,
      -0.00000000100,     0.00000000025,
       0.00000001136,     0.00000004434,
       0.00000000452,    -0.00000000733,
      -0.00000009753,     0.00000002551,
       0.00000001747,     0.00000001078,
       0.00000004029,    -0.00000002843,
      -0.00000002869,    -0.00000003983,
      -0.00000018794,    -0.00000109830,
      -0.00000011116,     0.00000016389,
       0.00000002073,    -0.00000000424,
       0.00000041197,    -0.00000010631,
      -0.00000006098,    -0.00000003911,
       0.00000000176,     0.00000000861,
      -0.00000006340,     0.00000025977,
       0.00000003191,    -0.00000001931,
       0.00000025967,     0.00000006522,
      -0.00000001997,    -0.00000003194,
       0.00000000445,     0.00000000431,
       0.00000000454,    -0.00000000392,
      -0.00000003607,    -0.00000011972,
       0.00000001002,    -0.00000000348,
      -0.00000000155,    -0.00000000092,
      -0.00000000587,    -0.00000002527,
      -0.00000000214,     0.00000000337,
      -0.00000002514,     0.00000000590,
       0.00000000346,     0.00000000219,
      -0.00000005109,    -0.00000025697,
      -0.00000002379,     0.00000003445,
      -0.00000008219,     0.00000001618,
       0.00000001118,     0.00000000797,
       0.00000000013,    -0.00000000122,
      -0.00000001442,    -0.00000008053,
      -0.00000000580,     0.00000000871,
      -0.00000008027,     0.00000001459,
       0.00000000877,     0.00000000575,
       0.00000012107,     0.00000047338,
       0.00000003951,    -0.00000006680,
       0.00000006452,    -0.00000001433,
      -0.00000000889,    -0.00000000533,
       0.00000001334,     0.00000006769,
       0.00000000507,    -0.00000000940,
       0.00000006738,    -0.00000001328,
      -0.00000000932,    -0.00000000503,
       0.00000000024,     0.00000002300,
       0.00000001247,     0.00000000122,
      -0.00000000032,    -0.00000000084,
       0.00000002321,     0.00000000572,
       0.00000000066,    -0.00000000062,
       0.00000000566,    -0.00000002340,
      -0.00000000062,    -0.00000000066,
      -0.00000033959,    -0.00000017732,
      -0.00000001321,     0.00000001701,
      -0.00000002451,    -0.00000001141,
       0.00000000194,     0.00000000248,
       0.00000005140,     0.00000000202,
      -0.00000000268,    -0.00000005027,
      -0.00000004003,     0.00000052236,
       0.00000008997,    -0.00000009249,
      -0.00000005767,    -0.00000000656,
       0.00000000301,     0.00000000670,
       0.00000039670,    -0.00000013384,
      -0.00000000498,     0.00000000623,
      -0.00000013473,    -0.00000039624,
       0.00000000624,     0.00000000496,
       0.00000002252,    -0.00000000187,
      -0.00000000115,    -0.00000001151,
      -0.00000000104,    -0.00000000029,
      -0.00000000015,    -0.00000000091,
       0.00000008446,     0.00000001086,
      -0.00000000173,     0.00000000629,
      -0.00000000179,     0.00000000025,
      -0.00000000022,    -0.00000000193,
       0.00018941904,    -0.00001064185,
      -0.00000081693,     0.00000296176,
       0.00000000440,    -0.00000012451,
       0.00000027157,     0.00000479837,
      -0.00000007517,    -0.00000002057,
       0.00000000305,     0.00000000010,
       0.00000009974,    -0.00000001964,
      -0.00000000615,     0.00000012976,
       0.00000000363,     0.00000001229,
       0.00000000110,     0.00000000080,
       0.00000001241,    -0.00000000359,
       0.00000000080,    -0.00000000109,
      -0.00000000133,     0.00000000028,
      -0.00000000336,    -0.00000000100,
       0.00000000062,    -0.00000000251,
       0.00000000735,    -0.00000000717,
      -0.00000000047,     0.00000000079,
      -0.00000005309,     0.00000001127,
      -0.00000000045,    -0.00000000132,
      -0.00000001054,    -0.00000005303,
       0.00000000124,    -0.00000000037,
      -0.00000000774,     0.00000000981,
       0.00000000059,     0.00000000047,
      -0.00000000984,    -0.00000000756,
      -0.00000000048,     0.00000000059,
       0.00000005532,     0.00000000627,
      -0.00000000066,    -0.00000000151,
       0.00000005105,    -0.00000000458,
      -0.00000000020,     0.00000000083,
       0.00000000451,     0.00000005104,
      -0.00000000082,    -0.00000000021,
      -0.00000000044,    -0.00000000186,
       0.00000000184,    -0.00000000043,
      -0.00000005361,     0.00000000216,
      -0.00000000010,     0.00000000280,
       0.00000000334,     0.00000000450,
      -0.00000000449,     0.00000000336,
      -0.00000000174,    -0.00000000862,
      -0.00000000075,    -0.00000000055,
       0.00000000862,    -0.00000000174,
       0.00000000055,    -0.00000000075,
       0.00000000085,    -0.00000000050,
       0.00000000053,     0.00000000084,
       0.00000000129,    -0.00000000050,
      -0.00000005346,     0.00000024730,
      -0.00000002248,    -0.00000004202,
       0.00000008976,     0.00000001346,
      -0.00000001585,     0.00000000984,
       0.00000000074,    -0.00000000179,
       0.00000022733,     0.00000015755,
      -0.00000001834,    -0.00000000839,
       0.00000000111,    -0.00000000114,
      -0.00000015602,     0.00000022404,
       0.00000000863,    -0.00000001865,
       0.00000000108,     0.00000000116,
      -0.00000003754,     0.00000004856,
       0.00000000356,     0.00000000209,
      -0.00000004983,    -0.00000003650,
      -0.00000000191,     0.00000000339,
       0.00000247802,     0.00000482779,
      -0.00000036189,     0.00000045303,
      -0.00000005323,    -0.00000013612,
      -0.00000004583,    -0.00000032680,
       0.00000012018,    -0.00000002919,
      -0.00000001506,     0.00000001448,
      -0.00000000063,    -0.00000000103,
       0.00000237965,    -0.00000156510,
       0.00000011766,     0.00000012175,
      -0.00000000800,     0.00000000644,
       0.00000156894,     0.00000241119,
      -0.00000012850,     0.00000011387,
      -0.00000000524,    -0.00000000844,
      -0.00000121636,     0.00000153662,
       0.00000010483,     0.00000008430,
       0.00000000571,    -0.00000000599,
      -0.00000153005,    -0.00000122117,
      -0.00000008506,     0.00000010577,
       0.00000000599,     0.00000000564,
       0.00000000204,     0.00000000131,
      -0.00000000132,     0.00000000201,
      -0.00000000316,     0.00000000100,
       0.00000000068,     0.00000000041,
      -0.00000000137,    -0.00000000382,
      -0.00000000419,     0.00000000151,
       0.00000000281,    -0.00000000111,
       0.00000000128,     0.00000000396,
       0.00000000039,    -0.00000000067,
       0.00000000396,    -0.00000000130,
      -0.00000000067,    -0.00000000039,
       0.00000000136,    -0.00000000002,
       0.00000000207,    -0.00000000022,
      -0.00000000051,    -0.00000000215,
      -0.00000006240,    -0.00000007197,
      -0.00000000770,     0.00000001258,
      -0.00000001448,     0.00000000025,
       0.00000000180,     0.00000000189,
       0.00000000851,    -0.00000001077,
      -0.00000000100,     0.00000000044,
      -0.00000001070,    -0.00000000845,
       0.00000000030,     0.00000000086,
       0.00000001370,     0.00000005260,
       0.00000000393,    -0.00000000128,
      -0.00000000102,     0.00000001519,
       0.00000000216,    -0.00000000025,
       0.00000001509,     0.00000000104,
      -0.00000000024,    -0.00000000215,
      -0.00000000164,    -0.00000000138,
      -0.00000000304,    -0.00000000659,
      -0.00000000102,     0.00000000069,
      -0.00000000659,     0.00000000304,
       0.00000000069,     0.00000000102,
      -0.00000005196,    -0.00000013779,
      -0.00000001105,     0.00000001949,
      -0.00000001757,     0.00000000532,
       0.00000000253,     0.00000000151,
       0.00000000073,    -0.00000001675,
      -0.00000000131,     0.00000000217,
      -0.00000001619,    -0.00000000085,
       0.00000000205,     0.00000000121,
       0.00000000604,     0.00000033655,
       0.00000004171,    -0.00000001247,
       0.00000001674,     0.00000000114,
      -0.00000000062,    -0.00000000258,
      -0.00000005141,     0.00000015550,
       0.00000001188,    -0.00000000665,
      -0.00000000057,    -0.00000000061,
       0.00000015516,     0.00000005148,
      -0.00000000656,    -0.00000001187,
      -0.00000000061,     0.00000000057,
       0.00000001363,    -0.00000006890,
      -0.00000000804,     0.00000000412,
      -0.00000003077,    -0.00000000819,
       0.00000000213,     0.00000000351,
      -0.00000006385,    -0.00000000427,
       0.00000000035,     0.00000000109,
      -0.00000000374,     0.00000006467,
       0.00000000101,    -0.00000000021,
       0.00000647490,    -0.00000135604,
      -0.00000009633,    -0.00000014204,
      -0.00000000699,     0.00000026327,
       0.00000001407,     0.00000000061,
       0.00000008232,    -0.00000000284,
      -0.00000000074,    -0.00000000120,
       0.00000000073,    -0.00000008831,
      -0.00000000224,    -0.00000000037,
      -0.00000000197,     0.00000000048,
       0.00000000078,     0.00000000162,
      -0.00000246673,     0.00000441285,
       0.00000023527,    -0.00000009136,
       0.00000009852,    -0.00000002241,
      -0.00000000993,    -0.00000001191,
      -0.00000000067,     0.00000000059,
      -0.00000346272,    -0.00000019137,
       0.00000001519,     0.00000005419,
      -0.00000019315,     0.00000346254,
       0.00000005457,    -0.00000001443,
      -0.00000002858,    -0.00000000317,
      -0.00000000138,     0.00000001462,
      -0.00000000146,     0.00000000034,
       0.00000000025,     0.00000000170,
       0.00000087663,     0.00000010293,
       0.00000001416,    -0.00000002743,
       0.00000001653,     0.00000004011,
      -0.00000000097,    -0.00000000127,
      -0.00000000801,    -0.00000000448,
       0.00000000368,     0.00000000534,
      -0.00017640980,    -0.00000997773,
       0.00000074843,     0.00000277874,
       0.00000025109,    -0.00000465826,
      -0.00000007248,     0.00000002084,
      -0.00000026450,     0.00000002465,
       0.00000000774,     0.00000025605,
      -0.00000000200,    -0.00000001205,
      -0.00000000139,    -0.00000000084,
      -0.00000001227,     0.00000000195,
      -0.00000000081,     0.00000000132,
       0.00000005059,    -0.00000000441,
      -0.00000000066,    -0.00000000139,
       0.00000000367,    -0.00000000002,
       0.00000000005,     0.00000000366,
       0.00000000598,     0.00000009622,
      -0.00000000258,     0.00000000160,
      -0.00000002244,    -0.00000000435,
       0.00000000437,    -0.00000002253,
       0.00000000097,     0.00000000938,
       0.00000000096,     0.00000000062,
      -0.00000000936,     0.00000000096,
      -0.00000000062,     0.00000000096,
       0.00000000005,     0.00000000230,
      -0.00000000233,     0.00000000007,
       0.00000000580,     0.00000000009,
       0.00000000005,    -0.00000000372,
       0.00000000146,     0.00000000029,
      -0.00000004566,    -0.00000000633,
       0.00000000352,    -0.00000001680,
      -0.00000000225,     0.00000000705,
      -0.00000000243,    -0.00000000068,
       0.00000000177,    -0.00000001493,
       0.00000000117,     0.00000000357,
       0.00000001261,     0.00000000145,
      -0.00000000389,     0.00000000026,
       0.00000000885,    -0.00000000259,
      -0.00000000083,    -0.00000000027,
       0.00000000270,     0.00000000885,
       0.00000000025,    -0.00000000082,
      -0.00000049254,    -0.00000023074,
       0.00000010890,    -0.00000005666,
       0.00000000630,    -0.00000000058,
      -0.00000000455,     0.00000000335,
       0.00000000059,    -0.00000000060,
      -0.00000000296,     0.00000015013,
      -0.00000001548,    -0.00000001620,
       0.00000000186,    -0.00000000017,
      -0.00000015068,    -0.00000000265,
       0.00000001630,    -0.00000001569,
       0.00000000016,     0.00000000190,
      -0.00000002808,    -0.00000000839,
       0.00000000274,     0.00000000070,
       0.00000000778,    -0.00000002813,
      -0.00000000055,     0.00000000269,
       0.00000000138,    -0.00000000040,
       0.00000000038,     0.00000000149,
       0.00000000028,     0.00000000082,
       0.00000000081,    -0.00000000028,
      -0.00000000098,     0.00000000036,
       0.00000000057,    -0.00000000186,
      -0.00000000195,    -0.00000000053,
       0.00000000055,     0.00000000152,
       0.00000000151,    -0.00000000055,
      -0.00000000062,    -0.00000000076,
      -0.00000000067,     0.00000000074,
      -0.00000001059,    -0.00000002288,
      -0.00000000317,     0.00000000103,
      -0.00000000137,    -0.00000000463,
      -0.00000000038,     0.00000000067,
      -0.00000000447,     0.00000000136,
       0.00000000073,     0.00000000041,
       0.00000001338,     0.00000004014,
       0.00000000209,    -0.00000000073,
       0.00000000101,     0.00000000992,
       0.00000000037,    -0.00000000069,
       0.00000000984,    -0.00000000100,
      -0.00000000069,    -0.00000000037,
      -0.00000000108,     0.00000000035,
      -0.00000000253,     0.00000000045,
       0.00000000045,     0.00000000253,
       0.00000032644,    -0.00000017580,
      -0.00000000770,     0.00000001743,
       0.00000000233,     0.00000000065,
       0.00000000523,    -0.00000001104,
      -0.00000000078,     0.00000000135,
      -0.00000001028,    -0.00000000576,
       0.00000000123,     0.00000000069,
      -0.00000000159,     0.00000028946,
       0.00000001966,    -0.00000002168,
       0.00000000949,    -0.00000000343,
      -0.00000000102,    -0.00000000055,
      -0.00000012158,     0.00000003673,
       0.00000000210,     0.00000000223,
       0.00000003638,     0.00000012167,
       0.00000000229,    -0.00000000211,
       0.00000006183,    -0.00000028479,
      -0.00000002138,     0.00000003251,
      -0.00000002330,     0.00000000973,
       0.00000000337,     0.00000000188,
      -0.00000000258,     0.00000000390,
       0.00000000614,     0.00000000189,
      -0.00000000057,    -0.00000000063,
      -0.00000642694,     0.00000149456,
       0.00000008802,     0.00000015413,
      -0.00000001617,    -0.00000017832,
      -0.00000000762,     0.00000000068,
      -0.00000008555,     0.00000000373,
       0.00000000033,     0.00000000134,
       0.00000000198,     0.00000008572,
       0.00000000185,     0.00000000025,
       0.00000000002,     0.00000000246,
       0.00000027437,     0.00000012519,
       0.00000001364,    -0.00000001670,
       0.00000002455,     0.00000013529,
       0.00000000111,    -0.00000000137,
      -0.00000000493,    -0.00000000424,
       0.00000000189,    -0.00000000469,
      -0.00000000059,     0.00000000082,
       0.00000000072,     0.00000000047,
      -0.00000130923,    -0.00000003463,
      -0.00000000905,     0.00000003119,
      -0.00000002365,    -0.00000020764,
      -0.00000000262,     0.00000000133,
       0.00000000367,     0.00000000133,
      -0.00000000063,     0.00000000442,
       0.00000104875,    -0.00000001627,
      -0.00000000110,    -0.00000003172,
       0.00000000359,     0.00000000228,
      -0.00000000226,     0.00000000248,
      -0.00000244642,     0.00000001264,
       0.00000000068,     0.00000002518,
      -0.00000001980,     0.00000000111,
      -0.00000000048,    -0.00000001984,
       0.00000000086,     0.00000000076,
      -0.00000000063,     0.00000000072,
       0.00000006135,    -0.00000001827,
       0.00000000049,     0.00000000174,
      -0.00000000089,     0.00000000016,
      -0.00000000023,    -0.00000000088,
      -0.00000000011,     0.00000000001,
      -0.00000000001,    -0.00000000011,
       0.00000000014,     0.00000000039,
      -0.00000000038,     0.00000000014,
       0.00000000366,     0.00000000425,
       0.00000000263,    -0.00000000241,
       0.00000000068,     0.00000000121,
       0.00000000052,     0.00000000186,
      -0.00000000254,     0.00000000015,
      -0.00000000851,    -0.00000003038,
       0.00000001254,    -0.00000000112,
      -0.00000000639,     0.00000000621,
      -0.00000000176,    -0.00000000313,
      -0.00000000952,    -0.00000000633,
       0.00000000206,    -0.00000000147,
       0.00000001025,    -0.00000000823,
       0.00000000011,     0.00000000371,
      -0.00000000101,    -0.00000000019,
       0.00000000019,    -0.00000000101,
       0.00000000086,    -0.00000000107,
      -0.00000000159,    -0.00000000568,
       0.00000000134,     0.00000000146,
       0.00000000574,    -0.00000000157,
      -0.00000000150,     0.00000000134,
       0.00000000461,     0.00000000037,
      -0.00000000034,     0.00000000462,
      -0.00000000541,    -0.00000000104,
       0.00000000113,    -0.00000000089,
       0.00000000102,    -0.00000000541,
       0.00000000089,     0.00000000113,
      -0.00000000035,    -0.00000000088,
      -0.00000000093,     0.00000000037,
       0.00000000035,     0.00000000090,
       0.00000000090,    -0.00000000035,
       0.00000001575,    -0.00000002364,
      -0.00000000167,     0.00000000187,
      -0.00000000092,    -0.00000000298,
      -0.00000000285,     0.00000000090,
       0.00000000902,     0.00000003265,
       0.00000000132,    -0.00000000049,
      -0.00000000521,     0.00000000474,
       0.00000000468,     0.00000000523,
      -0.00000000737,    -0.00000007375,
      -0.00000000695,     0.00000000201,
       0.00000000106,     0.00000000058,
      -0.00000000137,    -0.00000000602,
      -0.00000000048,     0.00000000086,
      -0.00000000554,     0.00000000122,
       0.00000000076,     0.00000000042,
      -0.00000029275,     0.00000016982,
       0.00000000243,    -0.00000001025,
      -0.00000000068,    -0.00000000035,
      -0.00000000575,     0.00000000752,
       0.00000000045,    -0.00000000086,
       0.00000000720,     0.00000000584,
      -0.00000000083,    -0.00000000043,
      -0.00000000015,    -0.00000018210,
      -0.00000001314,     0.00000002399,
      -0.00000001310,     0.00000001403,
       0.00000000209,     0.00000000112,
      -0.00000000061,     0.00000000288,
       0.00000000469,     0.00000000016,
      -0.00000012057,     0.00000021105,
       0.00000001491,    -0.00000002029,
       0.00000000639,    -0.00000001949,
      -0.00000000129,    -0.00000000069,
       0.00000000035,    -0.00000000320,
      -0.00000000410,    -0.00000000012,
       0.00000013329,     0.00000010552,
       0.00000001159,    -0.00000001231,
       0.00000000838,    -0.00000000361,
      -0.00000000068,    -0.00000000084,
       0.00000000066,    -0.00000000260,
       0.00000000190,    -0.00000000167,
      -0.00000040784,    -0.00000006495,
      -0.00000000524,     0.00000000364,
      -0.00000000463,     0.00000000158,
      -0.00000000057,     0.00000000200,
       0.00000052475,    -0.00000001210,
      -0.00000000218,    -0.00000000435,
       0.00000000723,     0.00000000180,
      -0.00000000159,     0.00000000690,
      -0.00000000022,    -0.00000000083,
       0.00000000074,    -0.00000000031,
      -0.00000118959,     0.00000001683,
       0.00000000070,     0.00000000503,
       0.00000000107,     0.00000000045,
       0.00000000058,    -0.00000000095,
       0.00000000117,    -0.00000000019,
       0.00000000021,     0.00000000115,
       0.00000003067,     0.00000001615,
       0.00000000085,    -0.00000000168,
      -0.00000000037,    -0.00000000108,
      -0.00000000112,    -0.00000000059,
      -0.00000023677,    -0.00000005271,
       0.00000005094,    -0.00000003849,
       0.00000000131,    -0.00000000687,
       0.00000000111,     0.00000000142,
      -0.00000000132,     0.00000000049,
      -0.00000000064,    -0.00000000198,
      -0.00000000187,     0.00000000062,
      -0.00000001208,     0.00000002414,
       0.00000000088,    -0.00000000034,
       0.00000000064,     0.00000000201,
       0.00000000196,    -0.00000000063,
      -0.00000001190,    -0.00000005173,
      -0.00000000434,     0.00000000244,
       0.00000000071,     0.00000000039,
      -0.00000000090,    -0.00000000375,
      -0.00000000336,     0.00000000077,
       0.00000000040,     0.00000006741,
       0.00000000266,    -0.00000000244,
       0.00000000089,     0.00000000386,
       0.00000000365,    -0.00000000082,
      -0.00000000667,    -0.00000011742,
      -0.00000000840,     0.00000001575,
      -0.00000000935,     0.00000000109,
       0.00000000130,     0.00000000067,
      -0.00000000047,     0.00000000209,
       0.00000000349,     0.00000000009,
      -0.00000002863,     0.00000012918,
       0.00000000593,    -0.00000000044,
      -0.00000000081,    -0.00000000042,
       0.00000000026,    -0.00000000238,
      -0.00000000306,    -0.00000000007,
       0.00000007145,     0.00000008319,
       0.00000000632,    -0.00000000007,
      -0.00000000052,    -0.00000000062,
       0.00000000035,    -0.00000000189,
       0.00000000124,    -0.00000000106,
      -0.00000020603,    -0.00000006009,
      -0.00000000411,     0.00000000027,
      -0.00000000100,     0.00000000093,
      -0.00000000032,     0.00000000125,
       0.00000029915,    -0.00000001481,
      -0.00000000068,    -0.00000000316,
       0.00000000001,     0.00000000151,
      -0.00000000150,     0.00000000001,
      -0.00000062551,     0.00000001095,
       0.00000000045,     0.00000000163,
       0.00000000137,     0.00000000021,
      -0.00000000091,     0.00000000036,
      -0.00000000044,    -0.00000000134,
      -0.00000000125,     0.00000000042,
       0.00000000045,     0.00000000135,
       0.00000000131,    -0.00000000043,
      -0.00000001132,    -0.00000003597,
      -0.00000000307,     0.00000000098,
      -0.00000000060,    -0.00000000235,
      -0.00000000204,     0.00000000048,
       0.00000000757,     0.00000004545,
       0.00000000202,    -0.00000000064,
       0.00000000058,     0.00000000242,
       0.00000000224,    -0.00000000052,
      -0.00000000664,    -0.00000007528,
      -0.00000000560,     0.00000000130,
       0.00000000081,     0.00000000041,
      -0.00000000034,     0.00000000149,
       0.00000000252,     0.00000000004,
      -0.00000001118,     0.00000008185,
       0.00000000353,    -0.00000000083,
       0.00000000019,    -0.00000000172,
      -0.00000000222,    -0.00000000004,
       0.00000004381,     0.00000006157,
       0.00000000457,     0.00000000025,
       0.00000000018,    -0.00000000133,
       0.00000000075,    -0.00000000068,
      -0.00000011672,    -0.00000004846,
      -0.00000000298,    -0.00000000016,
      -0.00000000060,     0.00000000069,
       0.00000016807,    -0.00000001162,
      -0.00000000065,    -0.00000000131,
      -0.00000034614,     0.00000000996,
       0.00000000043,     0.00000000064,
      -0.00000000031,    -0.00000000091,
       0.00000000031,     0.00000000091,
       0.00000000088,    -0.00000000030,
      -0.00000000812,    -0.00000002523,
      -0.00000000205,     0.00000000068,
      -0.00000000039,    -0.00000000146,
      -0.00000000122,     0.00000000030,
       0.00000000663,     0.00000003122,
       0.00000000135,    -0.00000000044,
       0.00000000038,     0.00000000150,
       0.00000000137,    -0.00000000034,
      -0.00000000430,    -0.00000004766,
      -0.00000000343,     0.00000000084,
      -0.00000000024,     0.00000000103,
       0.00000000180,     0.00000000001,
      -0.00000000573,     0.00000005171,
       0.00000000215,    -0.00000000054,
       0.00000000013,    -0.00000000121,
      -0.00000000157,    -0.00000000002,
       0.00000002672,     0.00000004463,
       0.00000000321,     0.00000000018,
      -0.00000006965,    -0.00000003660,
      -0.00000000209,    -0.00000000012,
       0.00000009670,    -0.00000000953,
      -0.00000019610,     0.00000000825,
      -0.00000000574,    -0.00000001751,
      -0.00000000137,     0.00000000047,
      -0.00000000026,    -0.00000000092,
       0.00000000090,    -0.00000000030,
       0.00000000025,     0.00000000093,
       0.00000000083,    -0.00000000021,
      -0.00000000285,    -0.00000002994,
      -0.00000000208,     0.00000000054,
       0.00000000126,    -0.00000000001,
      -0.00000000332,     0.00000003244,
       0.00000000131,    -0.00000000034,
      -0.00000000110,    -0.00000000001,
       0.00000001687,     0.00000003155,
       0.00000000221,     0.00000000012,
      -0.00000004258,    -0.00000002678,
      -0.00000000145,    -0.00000000008,
       0.00000005624,    -0.00000000752,
      -0.00000011327,     0.00000000635,
      -0.00000000324,    -0.00000001823,
      -0.00000000125,     0.00000000034,
       0.00000000079,    -0.00000000022,
       0.00000001029,     0.00000002200,
       0.00000000151,     0.00000000008,
      -0.00000002648,    -0.00000001905,
      -0.00000000100,    -0.00000000006,
       0.00000003272,    -0.00000000589,
      -0.00000006606,     0.00000000492,
       0.00000000669,     0.00000001515,
       0.00000001941,    -0.00000000436,
      -0.00000003883,     0.00000000374,
      -0.00000001135,    -0.00000005340,
       0.00000001622,    -0.00000000539,
      -0.00000000178,     0.00000001721,
       0.00000001885,     0.00000000111,
       0.00000000171,    -0.00000000065,
      -0.00000000140,    -0.00000000465,
      -0.00000000458,     0.00000000139,
      -0.00000001195,    -0.00000005904,
      -0.00000002129,     0.00000000428,
       0.00000000347,     0.00000000221,
      -0.00000000664,    -0.00000002057,
      -0.00000000160,     0.00000000278,
      -0.00000002031,     0.00000000671,
       0.00000000311,     0.00000000180,
       0.00000003466,     0.00000013842,
       0.00000002346,    -0.00000002465,
       0.00000001780,    -0.00000000464,
      -0.00000000287,    -0.00000000240,
       0.00000002269,     0.00000002132,
       0.00000000151,    -0.00000001292,
      -0.00000000235,     0.00000000034,
       0.00000002113,    -0.00000002267,
      -0.00000001294,    -0.00000000151,
       0.00000000034,     0.00000000235,
      -0.00000001000,     0.00000003894,
       0.00000002844,     0.00000001065,
       0.00000000175,    -0.00000000251,
       0.00000002002,     0.00000001993,
       0.00000000792,     0.00000000260,
       0.00000000036,    -0.00000000117,
       0.00000001973,    -0.00000002028,
       0.00000000288,    -0.00000000771,
      -0.00000000117,    -0.00000000036,
      -0.00000000099,     0.00000000044,
       0.00000000044,     0.00000000099,
      -0.00000086910,    -0.00000088066,
      -0.00000009265,     0.00000012257,
      -0.00000016407,    -0.00000001342,
       0.00000001362,     0.00000001940,
       0.00000000117,    -0.00000000151,
       0.00000018497,    -0.00000006055,
      -0.00000000477,    -0.00000000181,
      -0.00000006098,    -0.00000018341,
      -0.00000000139,     0.00000000465,
       0.00000028144,     0.00000024844,
       0.00000001423,    -0.00000017577,
      -0.00000003520,     0.00000000609,
       0.00000002566,    -0.00000002509,
      -0.00000001345,    -0.00000000224,
       0.00000000031,     0.00000000239,
       0.00000013554,    -0.00000031359,
      -0.00000006879,    -0.00000003868,
      -0.00000000355,     0.00000001096,
       0.00000000103,     0.00000000020,
      -0.00000031411,    -0.00000013549,
      -0.00000003865,     0.00000006892,
       0.00000001089,     0.00000000360,
       0.00000000020,    -0.00000000103,
       0.00000000309,    -0.00000000141,
      -0.00000000141,    -0.00000000309,
      -0.00000000117,     0.00000000023,
       0.00000000055,     0.00000000144,
       0.00000000152,    -0.00000000049,
      -0.00000008937,     0.00000021461,
       0.00000005245,     0.00000002435,
       0.00000000263,    -0.00000000813,
       0.00000011800,     0.00000007601,
       0.00000001750,    -0.00000003173,
      -0.00000000493,    -0.00000000180,
       0.00000021471,     0.00000039776,
       0.00000005839,    -0.00000004350,
      -0.00000000337,    -0.00000000351,
       0.00000039845,    -0.00000021536,
      -0.00000004374,    -0.00000005856,
      -0.00000000354,     0.00000000326,
      -0.00000000206,     0.00000001057,
      -0.00000000102,     0.00000000072,
       0.00000001092,     0.00000000193,
       0.00000000070,     0.00000000098,
      -0.00000021754,    -0.00000061714,
      -0.00000004784,     0.00000008180,
      -0.00000007082,     0.00000001179,
       0.00000000927,     0.00000000558,
       0.00000002857,     0.00000000254,
       0.00000000508,    -0.00000002861,
      -0.00000000070,    -0.00000000086,
       0.00000249168,    -0.00000978759,
      -0.00000154652,    -0.00000085360,
      -0.00000008876,     0.00000024041,
      -0.00000031359,    -0.00000024643,
      -0.00000006025,     0.00000009565,
       0.00000001552,     0.00000000566,
       0.00000000031,    -0.00000000148,
       0.00000310297,    -0.00000900007,
      -0.00000049299,     0.00000030827,
       0.00000002511,     0.00000002860,
       0.00000000137,    -0.00000000133,
      -0.00000900288,    -0.00000310036,
       0.00000030985,     0.00000049390,
       0.00000002862,    -0.00000002532,
      -0.00000000134,    -0.00000000139,
       0.00000000689,    -0.00000003306,
       0.00000000333,    -0.00000000243,
      -0.00000003323,    -0.00000000679,
      -0.00000000240,    -0.00000000329,
      -0.00000000207,     0.00000000111,
       0.00000074162,     0.00000637662,
       0.00000061667,    -0.00000044470,
      -0.00000003397,    -0.00000003776,
       0.00000281037,    -0.00000040095,
      -0.00000019798,    -0.00000027235,
      -0.00000001665,     0.00000001502,
       0.00000000073,     0.00000000077,
       0.00000612665,    -0.00000000859,
       0.00000000420,     0.00000000220,
       0.00000000013,    -0.00000000256,
      -0.00000000968,    -0.00000614736,
      -0.00000000319,    -0.00000000031,
      -0.00000000190,     0.00000000023,
      -0.00000000241,     0.00000000194,
       0.00000000075,    -0.00000000286,
       0.00015137690,     0.00000821615,
      -0.00000066637,    -0.00000235679,
      -0.00000034304,     0.00000570985,
       0.00000009124,    -0.00000002605,
       0.00000105629,    -0.00000000976,
      -0.00000000541,     0.00000000150,
       0.00000003594,    -0.00000117265,
      -0.00000000504,     0.00000000255,
      -0.00000002261,     0.00000001818,
       0.00000000224,     0.00000000152,
       0.00000002539,     0.00000001669,
       0.00000000160,    -0.00000000178,
       0.00023963836,    -0.00040288164,
      -0.00001178350,     0.00000536940,
       0.00000051580,     0.00000061793,
       0.00000003024,    -0.00000003057,
      -0.00000652989,     0.00000210595,
       0.00000058943,     0.00000080249,
       0.00000004938,    -0.00000004469,
      -0.00000000212,    -0.00000000215,
       0.00034074985,    -0.00000006735,
       0.00000006304,    -0.00000001374,
      -0.00000000137,    -0.00000012360,
      -0.00000002789,    -0.00034070791,
      -0.00000000493,    -0.00000007396,
      -0.00000012531,     0.00000000058,
       0.00000001870,     0.00000000255,
       0.00000000138,    -0.00000000317,
       0.00000000312,    -0.00000001543,
      -0.00000000294,    -0.00000000127,
       0.00000010199,    -0.00000000021,
       0.00000010339,    -0.00000000007,
      -0.00000000075,    -0.00000000111,
      -0.00000000110,     0.00000000075,
       0.00000000010,    -0.00000000195,
      -0.00000000195,    -0.00000000005,
      -0.00000002233,    -0.00000003092,
       0.00000000192,     0.00000001413,
       0.00000000081,    -0.00000000346,
       0.00000001415,    -0.00000000186,
      -0.00000000352,    -0.00000000075,
      -0.00000000696,    -0.00000001774,
       0.00000000313,    -0.00000000003,
      -0.00000001400,     0.00000002156,
       0.00000000286,    -0.00000000257,
       0.00000002139,     0.00000001390,
      -0.00000000277,    -0.00000000277,
       0.00000000128,     0.00000000070,
       0.00000000073,    -0.00000000128,
       0.00000011817,     0.00000056739,
       0.00000003014,    -0.00000012169,
      -0.00000001268,     0.00000000991,
      -0.00000001193,     0.00000000187,
       0.00000000124,     0.00000000151,
       0.00000004814,    -0.00000018003,
      -0.00000002065,     0.00000002282,
       0.00000000311,     0.00000000014,
      -0.00000018001,    -0.00000004876,
       0.00000002271,     0.00000002087,
      -0.00000000002,    -0.00000000318,
      -0.00000000583,    -0.00000000345,
      -0.00000000354,     0.00000000573,
       0.00000007888,     0.00000091309,
       0.00000009432,    -0.00000012018,
      -0.00000001490,    -0.00000000093,
      -0.00000225300,     0.00000021681,
       0.00000029492,     0.00000023390,
       0.00000000249,    -0.00000003972,
       0.00000092944,    -0.00000046141,
      -0.00000007724,     0.00000001024,
       0.00000000329,     0.00000000535,
      -0.00000045978,    -0.00000093602,
       0.00000001130,     0.00000007845,
       0.00000000399,    -0.00000000486,
      -0.00000001653,    -0.00000002311,
      -0.00000002487,     0.00000001426,
      -0.00000054654,    -0.00002456833,
      -0.00000259347,     0.00000318708,
       0.00000039427,     0.00000001439,
      -0.00000001622,    -0.00000002294,
       0.00001035083,    -0.00000099734,
      -0.00000134376,    -0.00000106952,
      -0.00000000540,     0.00000016096,
       0.00000001066,    -0.00000000753,
      -0.00000213065,     0.00000482903,
       0.00000061764,    -0.00000027925,
      -0.00000003202,    -0.00000003956,
       0.00000481891,     0.00000215937,
      -0.00000027305,    -0.00000062145,
      -0.00000004057,     0.00000003243,
       0.00000007811,     0.00000011105,
      -0.00000000683,     0.00000000169,
       0.00000011362,    -0.00000007115,
       0.00000000099,     0.00000000566,
      -0.00000006516,    -0.00000040311,
      -0.00000003672,     0.00000005498,
       0.00000002837,    -0.00000000417,
      -0.00000000426,    -0.00000000307,
      -0.00000000533,    -0.00000005341,
      -0.00000000432,     0.00000000613,
      -0.00000005337,     0.00000000541,
       0.00000000616,     0.00000000444,
      -0.00000000071,     0.00000000124,
      -0.00000018773,     0.00000074327,
       0.00000009147,    -0.00000004546,
      -0.00000000514,    -0.00000000575,
       0.00000056282,     0.00000016975,
      -0.00000003372,    -0.00000006983,
      -0.00000000414,     0.00000000376,
       0.00000072653,     0.00000005742,
      -0.00000000163,    -0.00000001184,
       0.00000005613,    -0.00000073276,
      -0.00000001288,     0.00000000240,
      -0.00000000062,    -0.00000000062,
      -0.00000000113,    -0.00000000213,
      -0.00001462476,    -0.00001756200,
      -0.00000162615,     0.00000151322,
       0.00000010734,     0.00000009980,
      -0.00000261651,    -0.00000080452,
       0.00000016291,     0.00000032149,
       0.00000001965,    -0.00000001864,
      -0.00000000111,    -0.00000000100,
       0.00000410690,     0.00000013382,
      -0.00000002228,    -0.00000005891,
       0.00000000058,     0.00000000045,
       0.00000011870,    -0.00000408031,
      -0.00000005398,     0.00000002297,
       0.00000000003,    -0.00000000127,
       0.00000000396,     0.00000000597,
       0.00000000009,    -0.00000000153,
       0.00000000391,     0.00000000340,
      -0.00000000161,    -0.00000000080,
       0.00000017261,     0.00000081710,
       0.00000005772,    -0.00000011852,
       0.00000007560,    -0.00000001427,
      -0.00000001042,    -0.00000000558,
       0.00000000005,     0.00000000098,
       0.00000000177,    -0.00000001399,
      -0.00000000133,    -0.00000000064,
      -0.00000001427,    -0.00000000167,
      -0.00000000065,     0.00000000145,
      -0.00000003591,    -0.00000000134,
      -0.00000000096,     0.00000003551,
       0.00000000050,    -0.00000000313,
       0.00000221845,     0.00000001099,
       0.00000000831,     0.00000181473,
      -0.00000000041,     0.00000000171,
      -0.00000005637,    -0.00000001779,
       0.00000000089,     0.00000000174,
      -0.00000000259,    -0.00000004776,
      -0.00000000052,    -0.00000000128,
      -0.00000000708,     0.00000001080,
       0.00000000060,     0.00000000048,
       0.00000000889,     0.00000000855,
       0.00000000051,    -0.00000000076,
       0.00000413021,     0.00000001504,
      -0.00000000384,    -0.00000068600,
       0.00000005104,     0.00000000440,
      -0.00000000030,    -0.00000000087,
      -0.00000000176,     0.00000004937,
      -0.00000000031,    -0.00000000188,
      -0.00000000193,     0.00000000047,
       0.00000000688,     0.00000001243,
       0.00000000516,    -0.00000000361,
       0.00000000958,    -0.00000000124,
       0.00000000119,     0.00000000966,
      -0.00000022231,     0.00000003405,
      -0.00000002401,     0.00000003495,
       0.00000000039,    -0.00000000276,
       0.00000009308,     0.00000001071,
      -0.00000000024,    -0.00000000095,
      -0.00000000908,     0.00000009436,
       0.00000001016,    -0.00000001210,
      -0.00000000104,    -0.00000000070,
       0.00000001216,     0.00000000975,
       0.00000000071,    -0.00000000101,
      -0.00000028241,     0.00000005394,
       0.00000000131,     0.00000000124,
       0.00000000181,    -0.00000000039,
      -0.00000000143,    -0.00000000075,
       0.00000000075,    -0.00000000144,
      -0.00000002033,    -0.00000001253,
       0.00000000179,    -0.00000000127,
       0.00000000233,     0.00000000544,
      -0.00000000553,     0.00000000238,
       0.00000001158,    -0.00000003239,
      -0.00000000190,    -0.00000000032,
      -0.00000000202,    -0.00000000035,
      -0.00000000047,     0.00000000227,
      -0.00000000033,    -0.00000000099,
       0.00000000101,    -0.00000000033,
       0.00000000000,     0.00000000006,
       0.00000000001,     0.00000000241,
       0.00000485310,    -0.00000002746,
      -0.00000001079,    -0.00000255646,
      -0.00000000049,     0.00000000239,
      -0.00000002287,    -0.00000006673,
       0.00000000061,     0.00000000453,
      -0.00000003354,    -0.00000000190,
       0.00000000008,    -0.00000000643,
      -0.00000000024,     0.00000001311,
      -0.00000000105,    -0.00000000261,
      -0.00000000009,    -0.00000000127,
       0.00000151607,    -0.00000000245,
       0.00000000027,     0.00000006989,
      -0.00000000012,    -0.00000000138,
       0.03597274261,    -0.00000759497,
       0.00000675220,    -0.00000106510,
      -0.00000009899,    -0.00001315093,
       0.00000017273,     0.00092919547,
       0.00000002052,     0.00000016630,
       0.00000033921,    -0.00000000194,
      -0.00000183579,    -0.00000405558,
      -0.00000006858,    -0.00000004554,
       0.00000000273,     0.00000000042,
      -0.00000130912,     0.00000360914,
      -0.00000007043,     0.00000000260,
      -0.00000000291,    -0.00000000009,
       0.00000046989,     0.00000245652,
       0.00000023907,     0.00000016701,
       0.00000001326,    -0.00000001552,
      -0.00000000055,    -0.00000000050,
       0.00000249374,    -0.00000046163,
       0.00000016976,    -0.00000023522,
      -0.00000001573,    -0.00000001311,
      -0.00000000050,     0.00000000055,
      -0.00000001010,    -0.00000001285,
      -0.00000000300,     0.00000000408,
      -0.00000000062,    -0.00000000107,
      -0.00000000909,     0.00000002255,
      -0.00000000289,    -0.00000000217,
      -0.00000002235,    -0.00000000996,
       0.00000000213,    -0.00000000288,
       0.00000000094,     0.00000000497,
      -0.00000000501,     0.00000000094,
       0.00001128747,     0.00000010680,
      -0.00000122599,     0.00000137935,
      -0.00000018087,    -0.00000027785,
       0.00000004983,    -0.00000055892,
       0.00000006180,     0.00000005207,
      -0.00000001314,     0.00000000835,
      -0.00000000024,    -0.00000000097,
      -0.00000060896,    -0.00000089979,
       0.00000001926,     0.00000000619,
      -0.00000000137,     0.00000000136,
       0.00000090114,    -0.00000060961,
      -0.00000000616,     0.00000001940,
      -0.00000000140,    -0.00000000138,
       0.00000032412,     0.00000172474,
       0.00000016525,     0.00000011734,
       0.00000000919,    -0.00000001089,
      -0.00000172452,     0.00000032459,
      -0.00000011730,     0.00000016529,
       0.00000001089,     0.00000000919,
      -0.00000000012,     0.00000000141,
       0.00000000497,     0.00000000187,
      -0.00000000050,     0.00000000051,
      -0.00000000188,     0.00000000496,
      -0.00000000051,    -0.00000000050,
       0.00000002028,     0.00000004012,
      -0.00000000026,    -0.00000000235,
       0.00000000100,    -0.00000000026,
       0.00000001649,    -0.00000001354,
       0.00000000116,     0.00000000136,
       0.00000001359,     0.00000001655,
      -0.00000000139,     0.00000000117,
      -0.00000000855,     0.00000001176,
       0.00000000065,     0.00000000036,
      -0.00000001171,    -0.00000000859,
      -0.00000000037,     0.00000000067,
       0.00000004869,     0.00000008307,
       0.00000000073,    -0.00000000158,
       0.00000007042,    -0.00000000423,
      -0.00000000652,     0.00000000845,
      -0.00000000111,    -0.00000000156,
       0.00000000424,     0.00000007042,
      -0.00000000844,    -0.00000000652,
       0.00000000156,    -0.00000000111,
      -0.00000000548,     0.00000000489,
      -0.00000000489,    -0.00000000548,
       0.00000000040,     0.00000000110,
      -0.00000000110,     0.00000000040,
       0.00000012608,     0.00000024713,
      -0.00000001663,     0.00000000483,
       0.00000000007,    -0.00000000038,
       0.00000018865,     0.00000084021,
      -0.00000002292,     0.00000000684,
      -0.00000000008,    -0.00000000027,
      -0.00000084115,     0.00000018531,
      -0.00000000708,    -0.00000002226,
       0.00000000023,    -0.00000000012,
       0.00000004113,     0.00000002389,
       0.00000000003,    -0.00000000183,
      -0.00000000004,     0.00000000004,
      -0.00000002442,     0.00000005157,
       0.00000000155,    -0.00000000023,
      -0.00000000004,    -0.00000000006,
       0.00000000082,     0.00000003757,
      -0.00000000813,    -0.00000000601,
       0.00000000061,    -0.00000000042,
      -0.00000004695,     0.00000002034,
       0.00000000681,    -0.00000000149,
      -0.00000002025,    -0.00000004672,
       0.00000000148,     0.00000000674,
       0.00000003460,     0.00000001959,
       0.00000000135,    -0.00000000217,
      -0.00000001986,     0.00000003502,
       0.00000000215,     0.00000000126,
      -0.00000001506,     0.00000007863,
      -0.00000001048,    -0.00000000777,
       0.00000000118,    -0.00000000092,
      -0.00000006165,     0.00000021535,
      -0.00000000719,    -0.00000000544,
      -0.00000021502,    -0.00000006163,
       0.00000000534,    -0.00000000707,
       0.00000003181,    -0.00000007192,
      -0.00000000725,    -0.00000000197,
       0.00000007287,     0.00000003243,
       0.00000000184,    -0.00000000725,
      -0.00000049115,    -0.00000014837,
       0.00000001224,    -0.00000009882,
       0.00000000921,     0.00000000170,
       0.00000295654,    -0.00000154545,
       0.00000015518,     0.00000023132,
      -0.00000000949,     0.00000000648,
       0.00000154649,     0.00000294532,
      -0.00000022774,     0.00000015570,
      -0.00000000666,    -0.00000000907,
       0.00000013874,    -0.00000005894,
      -0.00000000474,    -0.00000001215,
      -0.00000000058,     0.00000000017,
       0.00000005091,     0.00000014027,
       0.00000001192,    -0.00000000555,
      -0.00000000013,    -0.00000000058,
      -0.00000493672,    -0.00001170122,
       0.00000333816,    -0.00000073012,
       0.00000002051,     0.00000047755,
      -0.00000004330,    -0.00000000734,
      -0.00000920109,     0.00000290379,
      -0.00000050786,    -0.00000262047,
       0.00000038174,    -0.00000002926,
      -0.00000000320,     0.00000003398,
      -0.00000000232,    -0.00000000058,
       0.00001714088,    -0.00000341766,
       0.00000072784,     0.00000220000,
      -0.00000028235,    -0.00000002870,
       0.00000001336,    -0.00000003093,
       0.00000000253,     0.00000000191,
       0.00000435947,     0.00001173131,
      -0.00000011945,     0.00000137005,
      -0.00000015169,     0.00000012764,
      -0.00000002173,    -0.00000001780,
       0.00000000193,    -0.00000000236,
       0.00000050141,    -0.00000073776,
      -0.00000010252,    -0.00000008040,
      -0.00000000347,     0.00000000420,
       0.00000071005,     0.00000054174,
       0.00000007417,    -0.00000010502,
      -0.00000000403,    -0.00000000403,
       0.00001262191,    -0.00002925294,
       0.00000488126,     0.00000081324,
       0.00000019548,     0.00000063930,
      -0.00000005098,     0.00000004038,
       0.00002506838,    -0.00000569250,
       0.00000093944,     0.00000664344,
      -0.00000094911,     0.00000004491,
       0.00000000904,    -0.00000008342,
       0.00000000576,     0.00000000141,
      -0.00006821034,     0.00006554213,
      -0.00000759255,    -0.00001230089,
       0.00000148627,    -0.00000014587,
      -0.00000004934,     0.00000014629,
      -0.00000001174,    -0.00000000780,
      -0.00007006332,    -0.00004526143,
       0.00000346146,    -0.00001049077,
       0.00000094008,    -0.00000024951,
       0.00000007577,     0.00000008604,
      -0.00000000835,     0.00000000887,
      -0.00000122910,     0.00000189091,
       0.00000024948,     0.00000020302,
       0.00000000788,    -0.00000001126,
      -0.00000000081,    -0.00000000109,
      -0.00000183560,    -0.00000137267,
      -0.00000018359,     0.00000025679,
       0.00000001104,     0.00000000948,
      -0.00000004862,    -0.00000060566,
       0.00000007294,     0.00000004560,
      -0.00000238763,     0.00000011904,
       0.00000015268,    -0.00000030510,
       0.00000001752,     0.00000001847,
      -0.00000000118,     0.00000000068,
      -0.00000273833,     0.00003357897,
      -0.00000324658,    -0.00000231868,
       0.00000021432,    -0.00000019546,
       0.00000000623,     0.00000001202,
      -0.00003355903,    -0.00000271848,
       0.00000231444,    -0.00000324246,
       0.00000019502,     0.00000021295,
      -0.00000001187,     0.00000000624,
       0.00000014840,    -0.00000023298,
      -0.00000001329,    -0.00000001004,
      -0.00000000078,     0.00000000082,
       0.00000021905,     0.00000016900,
       0.00000000996,    -0.00000001520,
      -0.00000000075,    -0.00000000081,
      -0.00000015623,    -0.00000002675,
       0.00000002246,    -0.00000002721,
       0.00000034614,     0.00000045731,
      -0.00000006850,     0.00000001175,
      -0.00000000003,    -0.00000000508,
      -0.00000327516,     0.00004224843,
      -0.00000408710,    -0.00000291475,
       0.00000026970,    -0.00000024619,
       0.00000000783,     0.00000001505,
      -0.00004222259,    -0.00000327209,
       0.00000290950,    -0.00000408259,
       0.00000024566,     0.00000026826,
      -0.00000001489,     0.00000000781,
      -0.00000000164,    -0.00000002038,
      -0.00000000187,    -0.00000000134,
       0.00000002247,    -0.00000000409,
       0.00000000147,    -0.00000000189,
       0.00000000003,     0.00000000002,
       0.00000000015,    -0.00000000018,
       0.00000000046,    -0.00000000028,
       0.00000000027,     0.00000000045,
       0.00000000004,    -0.00000000015,
       0.00000000015,     0.00000000004,
       0.00000000156,    -0.00000000187,
       0.00000000188,     0.00000000156,
      -0.00000000005,     0.00000000001,
       0.00000000010,     0.00000000002,
       0.00000000033,     0.00000000011,
       0.00000000258,    -0.00000000080,
      -0.00000000053,     0.00000000058,
       0.00000000080,     0.00000000258,
      -0.00000000058,    -0.00000000053,
      -0.00000000115,     0.00000000178,
      -0.00000000178,    -0.00000000115,
       0.00000000960,    -0.00000002404,
       0.00000000110,     0.00000000117,
      -0.00000000442,     0.00000005872,
       0.00000000014,    -0.00000000696,
      -0.00000000010,     0.00000000480,
       0.00000000120,     0.00000000028,
      -0.00000000028,     0.00000000120,
      -0.00000000610,     0.00000000501,
       0.00000000488,     0.00000000194,
      -0.00000000193,     0.00000000487,
       0.00000000235,     0.00000000119,
      -0.00000000119,     0.00000000236,
      -0.00000004456,    -0.00000006656,
       0.00000000517,    -0.00000000370,
      -0.00000000121,     0.00000000000,
      -0.00000000001,    -0.00000000121,
      -0.00000000454,     0.00000000306,
      -0.00000000026,    -0.00000000035,
       0.00000036370,     0.00000054216,
      -0.00000004188,     0.00000003052,
       0.00000000958,     0.00000000033,
       0.00000000060,     0.00000000971,
       0.00000001059,    -0.00000002512,
       0.00000000101,     0.00000000120,
      -0.00000000470,     0.00000006218,
       0.00000000002,    -0.00000000183,
       0.00000000171,    -0.00000000001,
       0.00000000008,     0.00000001462,
       0.00000000027,    -0.00000000053,
      -0.00000000014,    -0.00000000098,
      -0.00000000100,     0.00000000175,
      -0.00000000044,     0.00000000028,
      -0.00000000009,    -0.00000000044,
      -0.00000000012,    -0.00000000133,
      -0.00000000053,    -0.00000000001,
       0.00000000016,    -0.00000000038,
       0.00000000107,     0.00000000070,
      -0.00000000019,     0.00000000032,
       0.00000000141,    -0.00000000256,
       0.00000000071,     0.00000000038,
      -0.00000000001,     0.00000000805,
      -0.00000000177,     0.00000000025,
      -0.00000000805,     0.00000000004,
      -0.00000000026,    -0.00000000177,
      -0.00000000089,     0.00000000181,
      -0.00000000181,    -0.00000000092,
       0.00000000080,     0.00000000059,
      -0.00000000149,     0.00000000015,
      -0.00000000016,    -0.00000000148,
       0.00000000024,    -0.00000000033,
      -0.00000000033,    -0.00000000026,
       0.00000000034,    -0.00000000045,
      -0.00000000046,    -0.00000000030,
      -0.00000000011,    -0.00000000048,
      -0.00000000013,    -0.00000000054,
      -0.00000000055,     0.00000000013,
      -0.00000000014,    -0.00000000072,
      -0.00000000066,     0.00000000029,
       0.00000000061,    -0.00000000080,
      -0.00000000081,    -0.00000000064,
       0.00000000089,    -0.00000000112,
      -0.00000000113,    -0.00000000074,
       0.00000000100,    -0.00000000028,
      -0.00000000018,    -0.00000000110,
       0.00000000157,    -0.00000000044,
      -0.00000000035,    -0.00000000142,
       0.00000000034,    -0.00000000036,
       0.00000000047,    -0.00000000030,
      -0.00000000023,    -0.00000000037,
       0.00000000044,    -0.00000000016,
      -0.00000000017,    -0.00000000044,
      -0.00000000020,    -0.00000000057,
       0.00000000054,    -0.00000000010,
      -0.00000000009,    -0.00000000050,
       0.00000000050,    -0.00000000009,
      -0.00000000012,    -0.00000000072,
      -0.00000000165,     0.00000000228,
       0.00000000015,     0.00000000038,
       0.00000012795,     0.00000009195,
       0.00000002104,    -0.00000000818,
      -0.00000000565,    -0.00000000138,
       0.00000000036,     0.00000000020,
      -0.00000000130,     0.00000000575,
       0.00000000023,    -0.00000000037,
       0.00000000049,    -0.00000000003,
      -0.00000000036,    -0.00000000123,
      -0.00000000125,     0.00000000036,
      -0.00000000030,    -0.00000000127,
      -0.00000000133,     0.00000000070,
       0.00000000173,    -0.00000000214,
      -0.00000000218,    -0.00000000177,
       0.00000000234,    -0.00000000314,
      -0.00000000324,    -0.00000000192,
       0.00000000308,    -0.00000000062,
      -0.00000000045,    -0.00000000330,
       0.00000000490,    -0.00000000103,
      -0.00000000083,    -0.00000000437,
       0.00000000082,    -0.00000000078,
       0.00000000013,    -0.00000000050,
       0.00000000125,    -0.00000000070,
      -0.00000000053,    -0.00000000098,
       0.00000000071,    -0.00000000023,
       0.00000000005,     0.00000000062,
       0.00000000004,     0.00000000001,
       0.00000000006,     0.00000000001,
       0.00000000002,    -0.00000000006,
       0.00000000002,    -0.00000000006,
       0.00000000001,    -0.00000000004,
       0.00000000058,    -0.00000000030,
      -0.00000000030,    -0.00000000058,
      -0.00000000031,    -0.00000000062,
       0.00000000057,    -0.00000000014,
      -0.00000000014,    -0.00000000056,
       0.00000000034,    -0.00000000008,
      -0.00000000018,    -0.00000000074,
       0.00000000177,    -0.00000000064,
      -0.00000000064,    -0.00000000177,
       0.00000000100,    -0.00000000037,
      -0.00000000068,    -0.00000000201,
       0.00000000197,    -0.00000000035,
      -0.00000000037,    -0.00000000196,
       0.00000000156,    -0.00000000032,
      -0.00000000049,    -0.00000000251,
       0.00000000141,    -0.00000000008,
       0.00000000000,    -0.00000000114,
       0.00000000112,    -0.00000000002,
      -0.00000000004,    -0.00000000160,
      -0.00000000275,     0.00000000167,
       0.00000000004,     0.00000000037,
       0.00000001157,     0.00000001738,
       0.00000000226,    -0.00000000036,
      -0.00000000275,    -0.00000000766,
      -0.00000000764,     0.00000000291,
      -0.00000000025,     0.00000000028,
       0.00000000030,     0.00000000026,
      -0.00000000054,    -0.00000000259,
      -0.00000000336,     0.00000000211,
       0.00000000547,    -0.00000000694,
      -0.00000000714,    -0.00000000563,
       0.00000000628,    -0.00000000767,
      -0.00000000907,    -0.00000000570,
       0.00000001059,    -0.00000000135,
      -0.00000000117,    -0.00000001110,
       0.00000001528,    -0.00000000239,
      -0.00000000200,    -0.00000001461,
       0.00000000205,    -0.00000000172,
       0.00000000027,    -0.00000000159,
       0.00000000365,    -0.00000000174,
       0.00000000025,     0.00000000025,
      -0.00000000130,    -0.00000000291,
       0.00000000188,    -0.00000000047,
       0.00000000009,     0.00000000169,
       0.00000000041,    -0.00000000054,
      -0.00000000033,    -0.00000000025,
       0.00000000059,    -0.00000000033,
       0.00000000029,     0.00000000058,
      -0.00000000001,     0.00000000006,
       0.00000000053,     0.00000000009,
       0.00000000035,     0.00000000004,
       0.00000000005,    -0.00000000038,
       0.00000000008,     0.00000000001,
       0.00000000007,    -0.00000000038,
       0.00000000010,     0.00000000005,
      -0.00000000001,    -0.00000000012,
       0.00000000004,     0.00000000002,
       0.00000000005,    -0.00000000015,
       0.00000000003,     0.00000000003,
       0.00000000320,    -0.00000000085,
      -0.00000000085,    -0.00000000320,
       0.00000000139,    -0.00000000034,
      -0.00000000089,    -0.00000000351,
       0.00000000154,    -0.00000000006,
      -0.00000000008,    -0.00000000153,
       0.00000000104,    -0.00000000004,
      -0.00000000009,    -0.00000000209,
       0.00000000042,    -0.00000000010,
       0.00000000026,     0.00000000034,
       0.00000000138,    -0.00000000118,
       0.00000001068,    -0.00000000192,
       0.00000000025,    -0.00000000041,
      -0.00000000193,    -0.00000001070,
      -0.00000000041,    -0.00000000027,
       0.00000000556,    -0.00000000103,
       0.00000000013,    -0.00000000022,
      -0.00000000182,    -0.00000001077,
      -0.00000000042,    -0.00000000025,
       0.00000000620,    -0.00000000017,
      -0.00000000029,    -0.00000000615,
       0.00000000512,    -0.00000000015,
      -0.00000000026,    -0.00000000806,
       0.00000000116,    -0.00000000015,
       0.00000000014,     0.00000000062,
       0.00000000099,    -0.00000000001,
       0.00000000002,     0.00000000101,
      -0.00000001487,     0.00000000131,
      -0.00000000028,     0.00000000142,
       0.00000000732,     0.00000005918,
       0.00000000557,     0.00000000077,
       0.00000002604,    -0.00000003723,
      -0.00000000040,     0.00000000031,
      -0.00000003751,    -0.00000002549,
       0.00000000043,     0.00000000046,
      -0.00000000087,     0.00000000002,
       0.00000000006,     0.00000000102,
       0.00000001395,    -0.00000001820,
      -0.00000002811,    -0.00000001488,
       0.00000000054,     0.00000000049,
       0.00000003889,    -0.00000000303,
      -0.00000000378,    -0.00000004104,
       0.00000004633,    -0.00000000489,
       0.00000000037,     0.00000000004,
      -0.00000000495,    -0.00000005015,
       0.00000000555,    -0.00000000383,
       0.00000000024,     0.00000000042,
       0.00000000004,    -0.00000000515,
       0.00000000029,    -0.00000000036,
       0.00000001012,    -0.00000000408,
       0.00000000059,     0.00000000067,
      -0.00000000326,    -0.00000000865,
       0.00000000057,    -0.00000000051,
       0.00000000514,    -0.00000000090,
       0.00000000025,     0.00000000461,
       0.00000000093,    -0.00000000110,
      -0.00000000068,    -0.00000000057,
       0.00000000133,    -0.00000000063,
       0.00000000056,     0.00000000131,
      -0.00000000005,     0.00000000027,
       0.00000000234,     0.00000000043,
       0.00000000162,    -0.00000000005,
      -0.00000000001,    -0.00000000173,
       0.00000000039,    -0.00000000001,
      -0.00000000002,    -0.00000000185,
       0.00000000002,     0.00000000026,
      -0.00000000006,    -0.00000000009,
       0.00000000003,     0.00000000008,
       0.00000000022,    -0.00000000010,
       0.00000000013,     0.00000000001,
       0.00000000001,     0.00000000015,
      -0.00000000001,     0.00000000001,
       0.00000000002,     0.00000000003,
       0.00000000001,     0.00000000004,
      -0.00000000004,     0.00000000002,
       0.00000000040,    -0.00000000114,
       0.00000001308,    -0.00000000051,
      -0.00000000053,    -0.00000001311,
       0.00000000495,    -0.00000000115,
       0.00000000281,    -0.00000000012,
      -0.00000000012,    -0.00000000280,
       0.00000000575,    -0.00000000023,
      -0.00000000057,    -0.00000001448,
       0.00000000050,    -0.00000000007,
      -0.00000000008,    -0.00000000050,
      -0.00000000007,    -0.00000000051,
       0.00000000126,    -0.00000000005,
       0.00000000005,     0.00000000127,
       0.00000000081,     0.00000000116,
       0.00000000433,    -0.00000000370,
      -0.00000000032,    -0.00000000032,
       0.00000004050,    -0.00000000149,
      -0.00000000155,    -0.00000004063,
       0.00000000566,    -0.00000000135,
       0.00000000305,    -0.00000000014,
      -0.00000000014,    -0.00000000304,
       0.00000002268,    -0.00000000087,
      -0.00000000160,    -0.00000004283,
       0.00000000184,    -0.00000000004,
      -0.00000000042,    -0.00000000180,
       0.00000000147,     0.00000000007,
      -0.00000000013,    -0.00000000235,
       0.00000000346,    -0.00000000010,
       0.00000000013,     0.00000000350,
       0.00000000027,     0.00000000033,
      -0.00000005091,    -0.00000002327,
      -0.00000000240,     0.00000000373,
      -0.00000007644,     0.00000015903,
       0.00000001190,     0.00000000759,
       0.00000020707,    -0.00000001127,
       0.00000000104,    -0.00000000004,
      -0.00000001361,    -0.00000020517,
       0.00000000021,    -0.00000000069,
       0.00000013554,    -0.00000000808,
       0.00000000052,     0.00000000005,
      -0.00000001132,    -0.00000018160,
       0.00000000002,    -0.00000000065,
       0.00000001065,    -0.00000000647,
       0.00000000064,     0.00000000145,
      -0.00000000246,    -0.00000002256,
       0.00000000062,    -0.00000000113,
      -0.00000000003,    -0.00000000049,
       0.00000002554,    -0.00000000826,
       0.00000000136,     0.00000000166,
      -0.00000000733,    -0.00000002451,
       0.00000000158,    -0.00000000135,
       0.00000001234,    -0.00000000121,
       0.00000000006,     0.00000001148,
       0.00000000190,    -0.00000000203,
       0.00000000037,     0.00000000023,
      -0.00000000126,    -0.00000000116,
       0.00000000309,    -0.00000000113,
       0.00000000103,     0.00000000307,
       0.00000000032,    -0.00000000034,
       0.00000000034,     0.00000000032,
       0.00000000001,     0.00000000009,
       0.00000000080,    -0.00000000016,
       0.00000000003,     0.00000000005,
       0.00000000054,    -0.00000000014,
       0.00000000003,     0.00000000003,
      -0.00000000012,    -0.00000000063,
       0.00000000003,    -0.00000000003,
       0.00000000013,    -0.00000000003,
      -0.00000000014,    -0.00000000064,
       0.00000000003,    -0.00000000003,
       0.00000000001,     0.00000000009,
       0.00000000025,     0.00000000010,
       0.00000000001,     0.00000000003,
       0.00000000007,    -0.00000000003,
       0.00000000004,    -0.00000000001,
       0.00000000001,     0.00000000005,
       0.00000000043,    -0.00000000001,
      -0.00000000005,    -0.00000000256,
       0.00000000403,    -0.00000000067,
       0.00000000023,     0.00000000027,
      -0.00000000088,    -0.00000000436,
       0.00000000027,    -0.00000000023,
       0.00000004581,    -0.00000000121,
       0.00000000199,    -0.00000000036,
      -0.00000000034,    -0.00000000194,
      -0.00000002610,    -0.00000000052,
       0.00000000000,     0.00000000116,
       0.00000000189,    -0.00000000030,
       0.00000000010,     0.00000000012,
      -0.00000000079,    -0.00000000467,
       0.00000000030,    -0.00000000026,
       0.00000000213,     0.00000000025,
       0.00000000083,    -0.00000001026,
       0.00000001239,    -0.00000000254,
       0.00000000070,     0.00000000080,
      -0.00000000282,    -0.00000001369,
       0.00000000078,    -0.00000000073,
       0.00000005115,    -0.00000000135,
       0.00000000225,    -0.00000000040,
      -0.00000000038,    -0.00000000218,
      -0.00000002800,    -0.00000000071,
      -0.00000000001,     0.00000000142,
       0.00000000773,    -0.00000000137,
       0.00000000043,     0.00000000047,
      -0.00000000261,    -0.00000001429,
       0.00000000087,    -0.00000000080,
       0.00000000114,    -0.00000000021,
       0.00000000019,     0.00000000112,
       0.00000000437,    -0.00000000432,
      -0.00000000078,    -0.00000000003,
      -0.00000001258,    -0.00000002982,
      -0.00000000006,     0.00000000204,
       0.00000005775,    -0.00000000899,
       0.00000000317,     0.00000000353,
      -0.00000001179,    -0.00000006001,
       0.00000000354,    -0.00000000310,
       0.00000000000,    -0.00000000067,
       0.00000004623,    -0.00000000837,
       0.00000000245,     0.00000000276,
      -0.00000001048,    -0.00000005910,
       0.00000000353,    -0.00000000314,
      -0.00000000307,    -0.00000000126,
      -0.00000000053,    -0.00000000412,
       0.00000000245,    -0.00000000235,
       0.00000000044,     0.00000000029,
      -0.00000000183,    -0.00000000163,
       0.00000000021,    -0.00000000033,
       0.00000000413,    -0.00000000081,
       0.00000000024,     0.00000000028,
       0.00000000077,     0.00000000417,
      -0.00000000028,     0.00000000023,
       0.00000000033,    -0.00000000028,
       0.00000000028,     0.00000000033,
       0.00000000005,    -0.00000000002,
      -0.00000000022,    -0.00000000045,
       0.00000000009,     0.00000000002,
      -0.00000000024,    -0.00000000030,
       0.00000000008,    -0.00000000001,
      -0.00000000033,    -0.00000000007,
      -0.00000000001,    -0.00000000009,
      -0.00000000052,    -0.00000000152,
      -0.00000000153,     0.00000000013,
      -0.00000000002,    -0.00000000008,
      -0.00000000035,     0.00000000009,
       0.00000000000,    -0.00000000007,
       0.00000000007,    -0.00000000001,
       0.00000000002,    -0.00000000001,
      -0.00000000002,    -0.00000000007,
      -0.00000000166,    -0.00000000266,
       0.00000000056,     0.00000000002,
      -0.00000000264,    -0.00000000056,
       0.00000000002,    -0.00000000056,
       0.00000001899,    -0.00000000349,
       0.00000000059,     0.00000000371,
       0.00000000146,    -0.00000000002,
       0.00000000000,     0.00000000158,
      -0.00000002496,    -0.00000001601,
      -0.00000000088,    -0.00000000118,
      -0.00000000618,    -0.00000001851,
       0.00000000397,     0.00000000008,
      -0.00000001870,     0.00000000180,
       0.00000000008,    -0.00000000398,
      -0.00000000026,    -0.00000000125,
       0.00000000026,     0.00000000001,
      -0.00000000301,     0.00000000061,
       0.00000000002,    -0.00000000064,
       0.00000000055,    -0.00000000023,
      -0.00000000088,    -0.00000000261,
      -0.00000000490,    -0.00000000728,
       0.00000000153,     0.00000000004,
      -0.00000000733,    -0.00000000161,
       0.00000000004,    -0.00000000153,
       0.00000002076,    -0.00000000382,
       0.00000000062,     0.00000000404,
       0.00000000161,    -0.00000000002,
       0.00000000000,     0.00000000163,
      -0.00000002762,    -0.00000001692,
       0.00000000557,    -0.00000000472,
      -0.00000000088,    -0.00000000120,
      -0.00000000687,    -0.00000002040,
       0.00000000435,     0.00000000009,
      -0.00000002055,     0.00000000198,
       0.00000000009,    -0.00000000435,
      -0.00000000097,    -0.00000000461,
       0.00000000098,     0.00000000003,
      -0.00000000823,     0.00000000172,
       0.00000000005,    -0.00000000175,
       0.00000000098,     0.00000000015,
       0.00000000016,    -0.00000000093,
       0.00000000093,     0.00000000015,
      -0.00000000008,     0.00000000027,
       0.00000000022,    -0.00000000140,
       0.00000000040,     0.00000000011,
      -0.00000000015,    -0.00000000068,
       0.00000000069,    -0.00000000014,
      -0.00000000051,     0.00000001077,
      -0.00000000107,    -0.00000000057,
       0.00000002569,    -0.00000000326,
      -0.00000000109,     0.00000000229,
      -0.00000002383,    -0.00000003163,
       0.00000000670,    -0.00000000054,
      -0.00000003200,    -0.00000000276,
      -0.00000000057,    -0.00000000672,
      -0.00000000034,    -0.00000000169,
      -0.00000000169,     0.00000000034,
      -0.00000000901,    -0.00000002833,
       0.00000000597,    -0.00000000044,
      -0.00000003462,     0.00000001152,
      -0.00000000067,    -0.00000000735,
       0.00000000442,     0.00000000169,
      -0.00000000040,     0.00000000136,
      -0.00000000108,    -0.00000000574,
       0.00000000149,     0.00000000008,
       0.00000000554,    -0.00000000072,
       0.00000000000,     0.00000000167,
      -0.00000000016,    -0.00000000618,
       0.00000000180,     0.00000000015,
      -0.00000000059,    -0.00000000304,
       0.00000000068,     0.00000000007,
       0.00000000270,    -0.00000000130,
       0.00000000021,     0.00000000058,
       0.00000000084,    -0.00000000046,
       0.00000000013,     0.00000000030,
      -0.00000000028,    -0.00000000060,
       0.00000000056,    -0.00000000017,
       0.00000000013,     0.00000000052,
      -0.00000001324,     0.00000000090,
      -0.00000000002,    -0.00000001165,
       0.00000000023,     0.00000000177,
      -0.00000000158,     0.00000000011,
      -0.00000000107,     0.00000000022,
      -0.00000000236,     0.00000000083,
       0.00000000100,    -0.00000000104,
       0.00000000093,     0.00000000041,
      -0.00000000066,    -0.00000000035,
       0.00000003107,     0.00000000895,
      -0.00000000355,     0.00000001233,
      -0.00000003028,    -0.00000000641,
       0.00000000083,    -0.00000000391,
      -0.00000000188,    -0.00000000010,
       0.00000000010,    -0.00000000175,
       0.00000000130,     0.00000000014,
       0.00000000776,    -0.00000001389,
       0.00000000170,     0.00000000075,
      -0.00000000125,    -0.00000000065,
      -0.00000001963,     0.00000000422,
      -0.00000000061,    -0.00000000307,
       0.00000000062,    -0.00000000080,
       0.00000000012,     0.00000000136,
      -0.00000000059,     0.00000000084,
      -0.00000000091,    -0.00000000028,
       0.00000008949,     0.00000000332,
      -0.00000000152,     0.00000003644,
      -0.00000000037,    -0.00000000201,
      -0.00000008182,    -0.00000000477,
       0.00000000068,    -0.00000001302,
       0.00000000136,    -0.00000000018,
       0.00000000011,     0.00000000188,
      -0.00000002968,    -0.00000000751,
      -0.00000000199,     0.00000000786,
      -0.00000000168,    -0.00000000011,
       0.00000000019,    -0.00000000168,
       0.00000000033,    -0.00000000112,
       0.00000000167,     0.00000000012,
      -0.00000000012,     0.00000000160,
       0.00000000036,     0.00000000104,
       0.00000000104,    -0.00000000036,
       0.00000000245,     0.00000000023,
       0.00000000110,    -0.00000000170,
      -0.00000000175,    -0.00000000107,
      -0.00000000144,    -0.00000000013,
       0.00000001658,    -0.00000002608,
       0.00000000449,     0.00000000246,
      -0.00000000118,    -0.00000000065,
       0.00000000129,     0.00000000129,
      -0.00000001820,     0.00000003633,
      -0.00000000290,    -0.00000000149,
      -0.00000000096,    -0.00000000022,
      -0.00000007602,     0.00000000922,
      -0.00000000444,    -0.00000002980,
       0.00000000368,    -0.00000000056,
       0.00000000555,    -0.00000000029,
       0.00000006927,    -0.00000001121,
       0.00000000176,     0.00000001036,
      -0.00000000220,     0.00000000300,
      -0.00000000326,    -0.00000000091,
       0.00000000049,    -0.00000000472,
      -0.00000011795,    -0.00000000820,
      -0.00000000165,     0.00000002821,
       0.00000000159,    -0.00000000026,
       0.00000000048,     0.00000000197,
       0.00000008632,     0.00000000592,
       0.00000000033,    -0.00000000430,
      -0.00000000143,     0.00000000033,
      -0.00000000027,    -0.00000000191,
      -0.00000000065,    -0.00000000040,
      -0.00000000040,     0.00000000065,
       0.00000000485,     0.00000000317,
      -0.00000000164,    -0.00000000149,
       0.00000000319,    -0.00000000488,
      -0.00000000147,     0.00000000168,
      -0.00000000047,     0.00000000169,
       0.00000000175,     0.00000000040,
       0.00000001557,     0.00000004777,
       0.00000000334,    -0.00000001339,
      -0.00000000132,     0.00000000237,
      -0.00000001340,    -0.00000000340,
       0.00000000244,     0.00000000130,
       0.00000000359,     0.00000000050,
      -0.00000000109,    -0.00000000029,
       0.00000000191,    -0.00000000369,
      -0.00000000376,    -0.00000000185,
      -0.00000000399,     0.00000003941,
      -0.00000000321,    -0.00000000028,
       0.00000000079,     0.00000000022,
      -0.00000000562,     0.00000000915,
      -0.00000001599,    -0.00000000981,
       0.00000000566,    -0.00000000309,
      -0.00000000372,    -0.00000000567,
       0.00000018305,    -0.00000030341,
      -0.00000002834,     0.00000004216,
       0.00000008820,     0.00000005294,
      -0.00000001100,    -0.00000000749,
      -0.00000004758,     0.00000002439,
       0.00000000361,    -0.00000000186,
       0.00000002998,     0.00000004788,
      -0.00000000251,    -0.00000000367,
      -0.00000005392,     0.00000010950,
      -0.00000001635,    -0.00000000801,
       0.00000000260,     0.00000000136,
       0.00000000041,     0.00000000251,
      -0.00000000299,    -0.00000000080,
       0.00000000160,     0.00000000711,
       0.00000013271,    -0.00000003655,
      -0.00000000919,    -0.00000003804,
      -0.00000000678,    -0.00000000135,
      -0.00000000718,     0.00000000366,
      -0.00000007864,     0.00000001966,
       0.00000000101,     0.00000000441,
       0.00000000227,    -0.00000000294,
       0.00000000320,     0.00000000078,
      -0.00000003102,    -0.00000002702,
      -0.00000000335,     0.00000000156,
      -0.00000000676,     0.00000000010,
       0.00000000045,     0.00000000828,
       0.00000004049,     0.00000000350,
      -0.00000000019,    -0.00000000085,
      -0.00000000678,    -0.00000000492,
       0.00000000166,    -0.00000000055,
       0.00000000537,     0.00000001101,
      -0.00000000109,    -0.00000000337,
       0.00000001101,    -0.00000000545,
      -0.00000000332,     0.00000000114,
       0.00000000091,    -0.00000000041,
      -0.00000000041,    -0.00000000092,
       0.00000022013,     0.00000014003,
      -0.00000007053,    -0.00000006604,
       0.00000000679,     0.00000001747,
      -0.00000000619,     0.00000000174,
       0.00000000178,    -0.00000000017,
      -0.00000004373,    -0.00000008786,
       0.00000000845,     0.00000002691,
       0.00000000089,    -0.00000000444,
      -0.00000008808,     0.00000004352,
       0.00000002722,    -0.00000000837,
      -0.00000000447,    -0.00000000089,
      -0.00000000423,     0.00000000178,
       0.00000000067,    -0.00000000057,
       0.00000000171,     0.00000000427,
      -0.00000000054,    -0.00000000068,
       0.00000000006,    -0.00000000115,
      -0.00000000115,    -0.00000000006,
      -0.00000001375,     0.00000006119,
       0.00000000573,    -0.00000001104,
      -0.00000008894,    -0.00000002316,
       0.00000001789,     0.00000000929,
       0.00000003533,    -0.00000003342,
      -0.00000000574,     0.00000000216,
      -0.00000003345,    -0.00000003564,
       0.00000000316,     0.00000000561,
       0.00000037127,    -0.00000171851,
      -0.00000015189,     0.00000030230,
       0.00000003886,    -0.00000001406,
       0.00000041950,     0.00000010046,
      -0.00000007170,    -0.00000003697,
       0.00000000323,     0.00000000865,
      -0.00000029320,     0.00000028249,
       0.00000004472,    -0.00000001817,
      -0.00000000416,    -0.00000000318,
       0.00000028256,     0.00000029559,
      -0.00000001776,    -0.00000004521,
      -0.00000000296,     0.00000000445,
       0.00000000504,     0.00000000579,
       0.00000000611,    -0.00000000463,
      -0.00000000857,     0.00000009312,
      -0.00000001422,    -0.00000000105,
       0.00000000342,     0.00000000084,
       0.00000001389,     0.00000000795,
       0.00000000727,    -0.00000001313,
      -0.00000002502,    -0.00000001386,
      -0.00000000715,     0.00000000438,
       0.00000000518,     0.00000000710,
      -0.00000020127,     0.00000036355,
       0.00000002965,    -0.00000004865,
       0.00000012969,     0.00000007077,
      -0.00000001874,    -0.00000001124,
       0.00000006087,    -0.00000003636,
      -0.00000000451,     0.00000000245,
      -0.00000004241,    -0.00000006126,
       0.00000000324,     0.00000000455,
       0.00000005273,    -0.00000010582,
      -0.00000000560,    -0.00000000283,
      -0.00000000054,    -0.00000000236,
       0.00000000297,     0.00000000093,
      -0.00000001876,     0.00000001150,
       0.00000000883,     0.00000001330,
      -0.00000000061,    -0.00000000102,
       0.00000000757,    -0.00000000030,
      -0.00000000029,    -0.00000000768,
       0.00000044674,    -0.00000027678,
      -0.00000003976,     0.00000001968,
      -0.00000004099,    -0.00000006371,
       0.00000000302,     0.00000000590,
      -0.00000006556,     0.00000000490,
       0.00000000121,    -0.00000000001,
       0.00000000194,     0.00000006465,
       0.00000000018,    -0.00000000127,
      -0.00000003164,     0.00000001270,
       0.00000000080,    -0.00000000096,
       0.00000000105,     0.00000000020,
      -0.00000005534,    -0.00000002134,
      -0.00000000245,     0.00000000299,
      -0.00000000602,    -0.00000000057,
      -0.00000000038,     0.00000000632,
       0.00000003137,     0.00000000321,
      -0.00000000003,    -0.00000000174,
      -0.00000009426,     0.00000000801,
       0.00000000058,     0.00000000755,
       0.00000000302,     0.00000000102,
       0.00000000135,    -0.00000000325,
       0.00000018078,     0.00000035814,
      -0.00000003227,    -0.00000010799,
      -0.00000000371,     0.00000001732,
      -0.00000049237,     0.00000022037,
       0.00000014393,    -0.00000003651,
      -0.00000002264,    -0.00000000575,
       0.00000000233,     0.00000000271,
       0.00000002844,    -0.00000046551,
      -0.00000001762,     0.00000009497,
       0.00000000940,    -0.00000000768,
      -0.00000046683,    -0.00000003090,
       0.00000009526,     0.00000001832,
      -0.00000000838,    -0.00000000879,
      -0.00000002763,    -0.00000000050,
       0.00000000348,    -0.00000000145,
      -0.00000000167,     0.00000002785,
      -0.00000000133,    -0.00000000356,
      -0.00000496536,    -0.00001005431,
       0.00000093888,     0.00000297833,
       0.00000010506,    -0.00000046760,
      -0.00000004625,     0.00000003236,
       0.00000227803,    -0.00000097788,
      -0.00000065090,     0.00000017452,
       0.00000009878,     0.00000002612,
      -0.00000000599,    -0.00000000974,
      -0.00000017740,     0.00000394934,
       0.00000015474,    -0.00000079745,
      -0.00000007302,     0.00000006542,
       0.00000000910,     0.00000000057,
       0.00000395596,     0.00000019084,
      -0.00000079744,    -0.00000016042,
       0.00000006465,     0.00000007446,
       0.00000000024,    -0.00000000923,
       0.00000012815,     0.00000000524,
      -0.00000001661,     0.00000000692,
       0.00000000823,    -0.00000012780,
       0.00000000600,     0.00000001651,
      -0.00000000136,    -0.00000000002,
       0.00000000020,    -0.00000000103,
      -0.00000000114,    -0.00000000013,
      -0.00000000219,    -0.00000000482,
       0.00000000049,     0.00000000156,
      -0.00000000484,     0.00000000222,
       0.00000000156,    -0.00000000049,
      -0.00000000996,    -0.00000000101,
      -0.00000002158,     0.00000011760,
       0.00000000394,    -0.00000002155,
       0.00000004237,     0.00000000767,
      -0.00000000782,    -0.00000000143,
       0.00000001804,    -0.00000002508,
      -0.00000000193,     0.00000000252,
      -0.00000002681,    -0.00000001756,
       0.00000000252,     0.00000000193,
       0.00000000633,    -0.00000004266,
      -0.00000000131,    -0.00000000036,
      -0.00000011891,     0.00000012415,
       0.00000001828,    -0.00000000954,
       0.00000009939,     0.00000008764,
      -0.00000000811,    -0.00000001449,
      -0.00000000066,     0.00000000122,
       0.00000008361,    -0.00000001212,
      -0.00000000391,    -0.00000000338,
      -0.00000001143,    -0.00000008444,
      -0.00000000358,     0.00000000401,
       0.00000000028,    -0.00000000115,
      -0.00000000126,    -0.00000000065,
       0.00000273285,    -0.00000288681,
      -0.00000042114,     0.00000021956,
       0.00000004083,     0.00000002562,
      -0.00000046706,    -0.00000041196,
       0.00000003675,     0.00000006738,
       0.00000000419,    -0.00000000692,
      -0.00000070877,     0.00000011302,
       0.00000003421,     0.00000002939,
       0.00000000177,    -0.00000000462,
       0.00000010886,     0.00000071201,
       0.00000003108,    -0.00000003429,
      -0.00000000479,    -0.00000000177,
      -0.00000000110,     0.00000000593,
       0.00000000596,     0.00000000200,
       0.00000001253,    -0.00000002512,
       0.00000000088,     0.00000000043,
      -0.00000001444,     0.00000000886,
       0.00000000472,     0.00000000698,
       0.00000000524,    -0.00000000045,
      -0.00000000042,    -0.00000000532,
       0.00000031268,    -0.00000018635,
      -0.00000002319,     0.00000001358,
      -0.00000002150,    -0.00000003150,
       0.00000000163,     0.00000000267,
      -0.00000004477,     0.00000000387,
       0.00000000328,     0.00000004487,
      -0.00000002049,     0.00000001094,
       0.00000004158,    -0.00000000391,
      -0.00000000189,    -0.00000001582,
      -0.00000000391,    -0.00000000179,
      -0.00000000206,     0.00000000394,
      -0.00000080550,     0.00000009291,
       0.00000002108,    -0.00000000361,
       0.00000000697,     0.00000007158,
      -0.00000000022,    -0.00000000161,
       0.00000003298,     0.00000001722,
       0.00000000141,    -0.00000000005,
       0.00000001883,    -0.00000003211,
      -0.00000000027,    -0.00000000150,
       0.00000002568,     0.00000000584,
       0.00000000011,    -0.00000000172,
      -0.00000009452,     0.00000000288,
       0.00000000035,     0.00000000843,
       0.00000000135,     0.00000000041,
       0.00000000058,    -0.00000000179,
      -0.00000000888,    -0.00000000190,
      -0.00000000047,     0.00000000288,
       0.00000000130,    -0.00000000157,
      -0.00000000174,    -0.00000000121,
       0.00000005690,     0.00000002535,
       0.00000000207,    -0.00000000542,
       0.00000000216,     0.00000000069,
       0.00000000032,    -0.00000000268,
       0.00000000138,     0.00000000115,
      -0.00000012660,     0.00000174343,
       0.00000007709,    -0.00000034510,
      -0.00000003237,     0.00000002709,
       0.00000139613,    -0.00000000118,
      -0.00000029179,    -0.00000004803,
       0.00000002467,     0.00000002589,
       0.00000000026,    -0.00000000244,
       0.00000081436,    -0.00000124020,
      -0.00000012306,     0.00000015085,
       0.00000001774,    -0.00000000039,
      -0.00000000047,    -0.00000000098,
      -0.00000123936,    -0.00000082590,
       0.00000014942,     0.00000012604,
      -0.00000000020,    -0.00000001821,
      -0.00000000076,     0.00000000081,
      -0.00000005146,    -0.00000002513,
       0.00000000506,    -0.00000000167,
      -0.00000002927,     0.00000004908,
      -0.00000000077,    -0.00000000464,
       0.00000268258,    -0.00004117141,
      -0.00000181439,     0.00000809453,
       0.00000075102,    -0.00000062403,
      -0.00000008719,    -0.00000001228,
      -0.00000654138,     0.00000000736,
       0.00000134867,     0.00000022073,
      -0.00000011286,    -0.00000012068,
      -0.00000000145,     0.00000001496,
       0.00000000077,    -0.00000000044,
      -0.00000660475,     0.00001096436,
       0.00000108806,    -0.00000134356,
      -0.00000015472,     0.00000000504,
       0.00000000515,     0.00000000858,
       0.00001095572,     0.00000665698,
      -0.00000133481,    -0.00000110108,
       0.00000000205,     0.00000015552,
       0.00000000894,    -0.00000000510,
       0.00000024011,     0.00000012301,
      -0.00000002394,     0.00000000540,
       0.00000013095,    -0.00000023340,
       0.00000000355,     0.00000002269,
       0.00000000699,     0.00000001610,
      -0.00000002071,     0.00000000760,
       0.00000000750,    -0.00000000221,
       0.00000000247,    -0.00000002027,
      -0.00000000113,     0.00000000444,
      -0.00000002026,    -0.00000000215,
       0.00000000469,     0.00000000087,
      -0.00000018445,    -0.00000044459,
       0.00000002916,     0.00000013442,
       0.00000000638,    -0.00000002332,
       0.00000009339,    -0.00000003265,
      -0.00000002662,     0.00000000520,
       0.00000000576,     0.00000000158,
      -0.00000001725,     0.00000016733,
       0.00000001059,    -0.00000003440,
      -0.00000000339,     0.00000000318,
       0.00000016773,     0.00000001808,
      -0.00000003459,    -0.00000001048,
       0.00000000318,     0.00000000339,
       0.00000000545,     0.00000000033,
       0.00000000049,    -0.00000000542,
       0.00000000344,     0.00000001757,
       0.00000000703,    -0.00000000138,
       0.00000000193,    -0.00000000707,
      -0.00000000801,    -0.00000000171,
       0.00000000194,    -0.00000000060,
      -0.00000000023,     0.00000000121,
       0.00000000131,     0.00000000019,
       0.00000000226,     0.00000000149,
       0.00000000246,    -0.00000000108,
      -0.00000000128,    -0.00000000243,
       0.00000007276,    -0.00000009078,
      -0.00000000940,     0.00000001001,
      -0.00000001070,    -0.00000000759,
       0.00000000135,     0.00000000121,
      -0.00000002125,     0.00000000999,
       0.00000000121,    -0.00000000045,
       0.00000000992,     0.00000002141,
      -0.00000000053,    -0.00000000123,
       0.00000040192,    -0.00000009987,
      -0.00000002091,    -0.00000000956,
      -0.00000004465,    -0.00000016148,
      -0.00000000435,     0.00000000940,
      -0.00000006378,    -0.00000005810,
      -0.00000000629,     0.00000000466,
      -0.00000006073,     0.00000006499,
       0.00000000504,     0.00000000636,
      -0.00000000108,    -0.00000000110,
      -0.00000000153,     0.00000000172,
      -0.00000761359,     0.00000205622,
       0.00000042320,     0.00000018543,
       0.00000000768,    -0.00000004688,
       0.00000019822,     0.00000073945,
       0.00000001996,    -0.00000004136,
      -0.00000000495,    -0.00000000090,
       0.00000054354,     0.00000052280,
       0.00000005299,    -0.00000004226,
      -0.00000000261,    -0.00000000253,
       0.00000053628,    -0.00000054308,
      -0.00000004411,    -0.00000005453,
      -0.00000000255,     0.00000000278,
       0.00000000602,     0.00000000451,
      -0.00000000080,     0.00000000055,
       0.00000000554,    -0.00000000756,
       0.00000000050,     0.00000000090,
      -0.00000001389,     0.00000001844,
       0.00000003889,    -0.00000000266,
      -0.00000000098,    -0.00000001264,
      -0.00000000203,    -0.00000000004,
      -0.00000000021,     0.00000000203,
      -0.00000064638,     0.00000003444,
       0.00000000464,     0.00000005440,
      -0.00000000042,    -0.00000000066,
       0.00000001718,     0.00000000239,
       0.00000000373,    -0.00000001638,
       0.00000002764,    -0.00000000323,
      -0.00000000006,     0.00000000096,
      -0.00000000085,     0.00000000091,
       0.00000000073,     0.00000000071,
      -0.00000005595,    -0.00000001848,
      -0.00000000588,     0.00000001706,
      -0.00000000299,    -0.00000000504,
      -0.00000000515,     0.00000000229,
       0.00000065997,     0.00000028151,
       0.00000002432,    -0.00000006206,
       0.00000000019,    -0.00000000154,
       0.00000005018,     0.00000001181,
      -0.00000000020,     0.00000000158,
       0.00000001057,    -0.00000005377,
       0.00000000167,     0.00000000047,
       0.00000001908,     0.00000001560,
       0.00000000074,    -0.00000000218,
       0.00000003055,    -0.00000000260,
       0.00000000085,    -0.00000000291,
       0.00000000140,     0.00000000245,
       0.00000000124,    -0.00000000002,
       0.00000000004,    -0.00000000128,
       0.00000360993,    -0.00000558926,
      -0.00000056325,     0.00000067580,
       0.00000007920,    -0.00000000113,
      -0.00000245249,    -0.00000125725,
       0.00000031540,     0.00000023073,
      -0.00000000287,    -0.00000003568,
      -0.00000000202,     0.00000000140,
      -0.00000330444,     0.00000179846,
       0.00000022073,    -0.00000009269,
      -0.00000001065,    -0.00000001100,
      -0.00000000064,     0.00000000043,
       0.00000177307,     0.00000334253,
      -0.00000008428,    -0.00000022385,
      -0.00000001219,     0.00000001035,
       0.00000000051,     0.00000000054,
       0.00000004734,     0.00000005384,
      -0.00000000460,     0.00000000244,
       0.00000006004,    -0.00000003394,
       0.00000000111,     0.00000000268,
      -0.00006613342,     0.00011190742,
       0.00001116643,    -0.00001356811,
      -0.00000156405,     0.00000002096,
       0.00000005262,     0.00000008863,
       0.00001105786,     0.00000580568,
      -0.00000142465,    -0.00000104527,
       0.00000001202,     0.00000016095,
       0.00000000904,    -0.00000000585,
      -0.00000000067,    -0.00000000035,
       0.00002801011,    -0.00001542936,
      -0.00000193972,     0.00000078774,
       0.00000009410,     0.00000009763,
       0.00000000536,    -0.00000000398,
      -0.00001531258,    -0.00002813527,
       0.00000074968,     0.00000195357,
       0.00000010296,    -0.00000009173,
      -0.00000000420,    -0.00000000597,
      -0.00000021711,    -0.00000027406,
       0.00000001939,    -0.00000000524,
       0.00000000095,     0.00000000063,
      -0.00000028313,     0.00000018829,
      -0.00000000282,    -0.00000001525,
       0.00000000065,    -0.00000000112,
      -0.00000000026,     0.00000000077,
       0.00000002001,    -0.00000000946,
       0.00000002103,    -0.00000001004,
      -0.00000001639,    -0.00000002651,
       0.00000000482,    -0.00000000272,
      -0.00000000163,     0.00000000099,
       0.00000000057,     0.00000001018,
       0.00000001020,    -0.00000000058,
      -0.00000000890,     0.00000006849,
       0.00000000493,    -0.00000001367,
       0.00000005590,     0.00000000281,
      -0.00000001162,    -0.00000000332,
       0.00000000083,     0.00000000109,
       0.00000003474,    -0.00000004648,
      -0.00000000639,     0.00000000545,
      -0.00000004638,    -0.00000003493,
       0.00000000571,     0.00000000626,
      -0.00000000196,    -0.00000000110,
      -0.00000000117,     0.00000000197,
       0.00000019147,    -0.00000158796,
      -0.00000011404,     0.00000031622,
       0.00000003619,    -0.00000002432,
      -0.00000026076,    -0.00000001504,
       0.00000005389,     0.00000001619,
      -0.00000000463,    -0.00000000625,
      -0.00000028167,     0.00000040683,
       0.00000005318,    -0.00000004689,
      -0.00000000700,    -0.00000000056,
       0.00000040616,     0.00000028355,
      -0.00000004642,    -0.00000005399,
      -0.00000000047,     0.00000000707,
       0.00000000903,     0.00000000524,
       0.00000000546,    -0.00000000876,
       0.00000000805,    -0.00000002028,
      -0.00000000216,    -0.00000000074,
      -0.00000000355,     0.00000000305,
       0.00000000312,     0.00000000356,
      -0.00000000196,     0.00000003324,
       0.00000001530,    -0.00000000677,
      -0.00000000246,    -0.00000000485,
      -0.00000000401,     0.00000000000,
      -0.00000000003,     0.00000000408,
      -0.00000024418,     0.00000010667,
       0.00000001130,     0.00000002230,
      -0.00000000055,    -0.00000000155,
       0.00000003456,     0.00000000057,
       0.00000000107,    -0.00000003455,
      -0.00000002664,     0.00000005912,
      -0.00000000164,    -0.00000000158,
       0.00000000104,    -0.00000000191,
      -0.00000000115,     0.00000000063,
       0.00000000074,     0.00000000130,
      -0.00000075644,    -0.00000038328,
      -0.00000003862,     0.00000003108,
      -0.00000013220,     0.00000023169,
       0.00000001065,     0.00000001295,
       0.00000000065,    -0.00000000063,
      -0.00000019998,     0.00000000582,
       0.00000000084,    -0.00000000228,
       0.00000000550,     0.00000019611,
      -0.00000000285,    -0.00000000120,
       0.00001066613,     0.00000578123,
       0.00000055979,    -0.00000047406,
      -0.00000002780,    -0.00000003167,
       0.00000056769,    -0.00000099847,
      -0.00000004633,    -0.00000005702,
      -0.00000000315,     0.00000000292,
       0.00000175102,    -0.00000010789,
      -0.00000000067,     0.00000002836,
      -0.00000000024,    -0.00000000192,
      -0.00000013241,    -0.00000178474,
       0.00000002777,     0.00000000454,
      -0.00000000158,     0.00000000023,
      -0.00000000265,     0.00000000245,
      -0.00000000061,     0.00000000320,
      -0.00000000068,     0.00000000176,
      -0.00000000183,    -0.00000000092,
      -0.00000001975,    -0.00000000211,
      -0.00000000061,     0.00000000518,
       0.00000020285,     0.00000009589,
       0.00000000319,    -0.00000002217,
      -0.00000000100,    -0.00000000133,
      -0.00000000172,    -0.00000000211,
       0.00000001113,    -0.00000001956,
      -0.00000006397,    -0.00000001364,
      -0.00000000327,     0.00000001515,
       0.00000061176,     0.00000001008,
       0.00000000722,     0.00000002643,
       0.00000001459,    -0.00000006180,
       0.00000000173,     0.00000000065,
       0.00000000530,    -0.00000000655,
       0.00000000092,     0.00000000085,
      -0.00000000914,    -0.00000000082,
       0.00000000091,    -0.00000000101,
      -0.00000000019,    -0.00000000032,
      -0.00000000003,    -0.00000000004,
      -0.00000000048,    -0.00000000085,
      -0.00000000078,    -0.00000000139,
      -0.00000000161,    -0.00000000246,
       0.00000001140,    -0.00000000946,
      -0.00000003214,    -0.00000004468,
       0.00000000301,     0.00000000431,
       0.00000000510,     0.00000001339,
      -0.00000000305,    -0.00000000207,
       0.00000002182,    -0.00000000246,
      -0.00000000246,     0.00000000223,
      -0.00002823220,     0.00001487527,
       0.00000190316,    -0.00000075796,
      -0.00000009186,    -0.00000009506,
      -0.00000000568,     0.00000000418,
       0.00000529708,     0.00000848524,
      -0.00000028198,    -0.00000064928,
      -0.00000003351,     0.00000003094,
       0.00000000137,     0.00000000187,
       0.00000896294,    -0.00000005887,
       0.00000003401,     0.00000001399,
      -0.00000000128,    -0.00000000749,
      -0.00000001413,    -0.00000907315,
      -0.00000000068,    -0.00000002799,
      -0.00000000643,     0.00000000223,
      -0.00000002262,    -0.00000001931,
       0.00000000178,     0.00000000092,
      -0.00000000674,    -0.00000003334,
       0.00000000253,     0.00000000367,
       0.00040548190,    -0.00021584156,
      -0.00002767438,     0.00001079346,
       0.00000134474,     0.00000140301,
       0.00000007882,    -0.00000005807,
      -0.00002283994,    -0.00003685880,
       0.00000119857,     0.00000279698,
       0.00000014611,    -0.00000013507,
      -0.00000000602,    -0.00000000822,
      -0.00007826281,    -0.00000160202,
       0.00000002273,     0.00000010045,
      -0.00000001202,     0.00000006841,
      -0.00000247799,     0.00007833033,
       0.00000023924,     0.00000003233,
       0.00000006453,    -0.00000000145,
       0.00000003113,     0.00000015432,
       0.00000000588,    -0.00000001367,
      -0.00000000107,    -0.00000000056,
       0.00000012399,     0.00000007482,
      -0.00000001644,    -0.00000001662,
      -0.00000000069,     0.00000000092,
       0.00000000826,     0.00000004848,
       0.00000000390,    -0.00000000641,
      -0.00000004935,     0.00000000978,
       0.00000000647,     0.00000000375,
      -0.00000005478,    -0.00000000295,
      -0.00000005719,    -0.00000000308,
       0.00000000301,    -0.00000000018,
       0.00000000165,    -0.00000000340,
      -0.00000000353,    -0.00000000157,
      -0.00000000143,    -0.00000010140,
      -0.00000000461,     0.00000002097,
      -0.00000001666,     0.00000000095,
       0.00000000406,     0.00000000072,
      -0.00000001412,     0.00000002743,
       0.00000000311,    -0.00000000317,
       0.00000002750,     0.00000001415,
      -0.00000000338,    -0.00000000296,
      -0.00000000052,     0.00000000102,
       0.00000000102,     0.00000000052,
       0.00000011977,    -0.00000016401,
      -0.00000002185,     0.00000001879,
      -0.00000007331,    -0.00000004323,
       0.00000000907,     0.00000000908,
      -0.00000000002,    -0.00000000128,
      -0.00000010440,     0.00000004815,
       0.00000000819,    -0.00000000091,
       0.00000004742,     0.00000010543,
      -0.00000000075,    -0.00000000842,
       0.00000000131,     0.00000000173,
       0.00000000192,    -0.00000000094,
      -0.00000222357,     0.00000325073,
       0.00000042576,    -0.00000037035,
      -0.00000005700,    -0.00000000645,
       0.00000033142,     0.00000020036,
      -0.00000004007,    -0.00000004103,
      -0.00000000044,     0.00000000612,
       0.00000088150,    -0.00000041340,
      -0.00000007155,     0.00000000784,
       0.00000000337,     0.00000000423,
      -0.00000040931,    -0.00000088518,
       0.00000000641,     0.00000007183,
       0.00000000443,    -0.00000000332,
      -0.00000000601,    -0.00000000875,
      -0.00000000893,     0.00000000514,
       0.00000003029,     0.00000000604,
       0.00000000414,     0.00000002213,
       0.00000000172,    -0.00000000293,
       0.00000002213,    -0.00000000414,
      -0.00000000293,    -0.00000000172,
      -0.00000000068,    -0.00000000068,
      -0.00000000096,     0.00000000023,
       0.00000000033,     0.00000000097,
      -0.00000003844,     0.00000003037,
       0.00000000307,     0.00000000337,
       0.00000000883,    -0.00000000260,
      -0.00000000262,    -0.00000000887,
       0.00000041978,     0.00000234175,
       0.00000018375,    -0.00000030785,
      -0.00000002887,     0.00000000220,
      -0.00000005942,     0.00000001094,
       0.00000000784,     0.00000000467,
      -0.00000004008,    -0.00000000048,
      -0.00000000012,     0.00000001004,
       0.00000042218,     0.00000001930,
       0.00000000056,    -0.00000004200,
      -0.00000000030,    -0.00000000263,
      -0.00000000352,    -0.00000000049,
      -0.00000000087,     0.00000000100,
      -0.00000197008,     0.00000011006,
       0.00000000626,    -0.00000003021,
       0.00000002661,     0.00000046548,
      -0.00000000723,    -0.00000000153,
      -0.00000000109,    -0.00000001072,
       0.00000000367,     0.00000000622,
      -0.00000000033,    -0.00000000108,
      -0.00000000062,    -0.00000000125,
       0.00000000049,    -0.00000000108,
       0.00001813772,    -0.00000269874,
       0.00000010976,     0.00000038127,
      -0.00000009088,    -0.00000189970,
       0.00000002833,     0.00000000784,
      -0.00000000210,     0.00000000002,
       0.00000006743,    -0.00000022515,
       0.00000002955,     0.00000001589,
      -0.00000000193,     0.00000000160,
      -0.00000026644,     0.00000000516,
       0.00000001853,    -0.00000002549,
       0.00000000182,     0.00000000180,
      -0.00000000178,     0.00000000600,
       0.00000000410,     0.00000000452,
       0.00000001563,    -0.00000004237,
      -0.00000000105,    -0.00000000043,
      -0.00000000122,    -0.00000000009,
      -0.00000000038,     0.00000000080,
      -0.00000000707,    -0.00000000618,
       0.00000000053,    -0.00000000067,
       0.00000000818,    -0.00000003987,
       0.00000000155,     0.00000000002,
       0.00000000041,     0.00000000080,
      -0.00000000089,     0.00000000042,
      -0.00000002550,    -0.00000002296,
      -0.00000000049,     0.00000000129,
       0.00000000096,     0.00000000053,
       0.00000008077,    -0.00000029302,
       0.00000003344,     0.00000002987,
       0.00000001091,     0.00000000234,
      -0.00000000106,     0.00000000125,
       0.00000000132,    -0.00000000067,
       0.00000000017,     0.00000000124,
      -0.00000001257,     0.00000003030,
      -0.00000000665,     0.00000002483,
       0.00000000000,    -0.00000000019,
       0.00000000000,    -0.00000000066,
       0.00000000001,    -0.00000000107,
       0.00000000001,    -0.00000000149,
      -0.00000001299,    -0.00000000025,
       0.00000000077,    -0.00000000435,
      -0.00000000152,     0.00000000348,
      -0.00000000191,     0.00000000316,
       0.00013277458,     0.00000087755,
       0.00000012105,    -0.00000007107,
      -0.00000000116,    -0.00000010266,
       0.00000022628,    -0.00003154048,
      -0.00000001877,    -0.00000002160,
      -0.00000002451,     0.00000000013,
       0.00000036504,     0.00000069957,
       0.00000001624,    -0.00000004579,
      -0.00000000057,    -0.00000000474,
      -0.00000021790,    -0.00000010745,
       0.00000001654,     0.00000006833,
       0.00000000105,    -0.00000000377,
       0.00000003752,     0.00000009663,
      -0.00000000333,     0.00000000430,
      -0.00000004809,     0.00000007740,
      -0.00000000189,    -0.00000000700,
      -0.00120280851,    -0.00006128399,
       0.00000277240,     0.00000514770,
      -0.00000031791,     0.00000110978,
      -0.00000055457,     0.00012930358,
       0.00000004434,     0.00000013461,
       0.00000009902,    -0.00000000355,
      -0.00000198126,     0.00001468792,
      -0.00000166170,    -0.00000103652,
       0.00000010814,    -0.00000008211,
       0.00000000325,     0.00000000529,
       0.00001728050,     0.00000400958,
      -0.00000119061,     0.00000137999,
      -0.00000009810,    -0.00000009439,
       0.00000000591,    -0.00000000242,
       0.00000016797,    -0.00000043952,
      -0.00000001408,    -0.00000002105,
      -0.00000000089,     0.00000000138,
      -0.00000020046,    -0.00000036412,
      -0.00000001453,     0.00000003269,
       0.00000000088,     0.00000000172,
       0.00000006554,    -0.00000005173,
      -0.00000000184,    -0.00000000285,
      -0.00000000680,    -0.00000001711,
      -0.00000000152,     0.00000000093,
       0.00000001296,     0.00000001416,
      -0.00000000101,    -0.00000000152,
      -0.00000000232,     0.00000002519,
      -0.00000000236,     0.00000002612,
      -0.00000000058,     0.00000000192,
       0.00000000192,     0.00000000058,
       0.00000000549,    -0.00000001051,
      -0.00000000456,    -0.00000000180,
      -0.00000000551,     0.00000000363,
       0.00000000362,     0.00000000554,
      -0.00000010175,     0.00000019898,
       0.00000002472,    -0.00000002218,
       0.00000002025,     0.00000000882,
      -0.00000000257,    -0.00000000255,
       0.00000004642,    -0.00000002997,
      -0.00000000441,     0.00000000036,
      -0.00000002972,    -0.00000004676,
       0.00000000030,     0.00000000445,
      -0.00000000692,     0.00000001468,
       0.00000000101,     0.00000000050,
       0.00000000280,    -0.00000000169,
      -0.00000000177,    -0.00000000278,
      -0.00000066739,     0.00000030057,
       0.00000005292,    -0.00000000575,
       0.00000010900,     0.00000020224,
      -0.00000000273,    -0.00000001793,
      -0.00000000104,     0.00000000083,
       0.00000020289,     0.00000001416,
      -0.00000000015,    -0.00000000300,
       0.00000001611,    -0.00000020562,
      -0.00000000393,    -0.00000000003,
       0.00000954209,    -0.00000433979,
      -0.00000076285,     0.00000007457,
       0.00000003667,     0.00000004592,
      -0.00000047234,    -0.00000088190,
       0.00000001146,     0.00000007791,
       0.00000000492,    -0.00000000370,
      -0.00000176653,    -0.00000016389,
       0.00000000499,     0.00000003219,
      -0.00000000017,     0.00000000115,
      -0.00000018362,     0.00000176688,
       0.00000003530,    -0.00000000327,
       0.00000000098,    -0.00000000026,
       0.00000000030,     0.00000000331,
       0.00000000245,     0.00000000221,
       0.00000000179,    -0.00000000540,
      -0.00000000712,    -0.00000000160,
       0.00000000051,     0.00000000074,
      -0.00000000816,     0.00000000200,
       0.00000000066,     0.00000000205,
       0.00000000120,     0.00000000004,
      -0.00000000003,    -0.00000000127,
       0.00000010182,    -0.00000002616,
      -0.00000000297,    -0.00000000889,
      -0.00000001193,    -0.00000000057,
      -0.00000000078,     0.00000001186,
       0.00000019083,    -0.00000069501,
      -0.00000007040,     0.00000004361,
       0.00000001892,     0.00000000439,
      -0.00000000125,    -0.00000000192,
       0.00000000135,    -0.00000000112,
      -0.00000000112,    -0.00000000141,
       0.00000002567,     0.00000000531,
       0.00000000064,    -0.00000000491,
       0.00000000419,    -0.00000008999,
       0.00000000245,     0.00000000069,
      -0.00000029806,    -0.00000014020,
      -0.00000000390,     0.00000001409,
      -0.00000000708,     0.00000000448,
      -0.00000000054,    -0.00000000059,
       0.00000000493,     0.00000000725,
      -0.00000000063,     0.00000000054,
       0.00000000030,     0.00000001146,
       0.00000000289,    -0.00000000054,
       0.00000000327,    -0.00000000035,
       0.00000000024,     0.00000000324,
       0.00000022393,    -0.00000790435,
       0.00000091017,     0.00000054640,
      -0.00000006224,     0.00000005715,
       0.00000030156,    -0.00000000157,
      -0.00000002051,     0.00000003459,
      -0.00000000217,    -0.00000000231,
       0.00000003314,    -0.00000001818,
       0.00000000173,     0.00000000145,
       0.00000000498,     0.00000003481,
       0.00000000030,     0.00000000182,
       0.00000000027,    -0.00000000248,
      -0.00000016571,     0.00000034431,
      -0.00000003772,    -0.00000003166,
      -0.00000000500,     0.00000000633,
      -0.00000000198,     0.00000000649,
      -0.00000000082,    -0.00000000065,
       0.00000000890,     0.00000000076,
      -0.00000000079,     0.00000000103,
      -0.00000009686,     0.00000043201,
      -0.00000005070,    -0.00000004163,
      -0.00000001171,    -0.00000000235,
       0.00000000108,    -0.00000000134,
      -0.00000000024,     0.00000000101,
      -0.00000000107,    -0.00000000032,
      -0.00000000096,    -0.00000000083,
       0.00000000192,     0.00000000003,
      -0.00000000006,     0.00000000193,
       0.00000213197,    -0.00000145731,
       0.00000011564,     0.00000011651,
      -0.00000000837,     0.00000000496,
      -0.00000037855,    -0.00000035695,
       0.00000002805,    -0.00000003230,
       0.00000000173,     0.00000000246,
      -0.00000052173,     0.00000001637,
      -0.00000000192,    -0.00000000139,
      -0.00000000561,    -0.00000052355,
       0.00000000033,     0.00000000280,
      -0.00000000027,    -0.00000001326,
      -0.00000000109,     0.00000000214,
       0.00000001469,    -0.00000000553,
      -0.00000000230,    -0.00000000060,
      -0.00008684129,     0.00102387544,
      -0.00009891712,    -0.00007065114,
       0.00000653010,    -0.00000594799,
       0.00000019148,     0.00000037011,
      -0.00003933500,    -0.00000192643,
       0.00000270887,    -0.00000382019,
       0.00000023106,     0.00000025075,
      -0.00000001413,     0.00000000736,
      -0.00000438774,     0.00000058928,
      -0.00000009081,    -0.00000002735,
       0.00000000454,    -0.00000000288,
       0.00000119113,    -0.00000445738,
      -0.00000007814,     0.00000011241,
      -0.00000001145,    -0.00000000735,
      -0.00000016203,    -0.00000009280,
       0.00000001986,    -0.00000000528,
      -0.00000022435,     0.00000019579,
       0.00000001432,    -0.00000001258,
       0.00000000114,    -0.00000000104,
      -0.00000265212,     0.00000182580,
      -0.00000016163,    -0.00000014127,
      -0.00000000320,     0.00000015934,
      -0.00000000188,    -0.00000000166,
      -0.00000002102,     0.00000016914,
      -0.00000001947,    -0.00000001122,
       0.00000000134,    -0.00000000125,
       0.00000017245,    -0.00000001492,
      -0.00000001152,     0.00000001935,
      -0.00000000125,    -0.00000000132,
       0.00000000151,     0.00000000053,
       0.00000000288,    -0.00000000217,
      -0.00000000227,    -0.00000000289,
      -0.00000003308,     0.00000002166,
       0.00000000735,     0.00000000980,
      -0.00000000016,    -0.00000000101,
       0.00000001065,    -0.00000000082,
      -0.00000000071,    -0.00000001080,
       0.00000046602,    -0.00000030114,
      -0.00000004426,     0.00000000183,
      -0.00000003186,    -0.00000004265,
       0.00000000045,     0.00000000459,
      -0.00000009292,     0.00000000516,
       0.00000000055,     0.00000000305,
       0.00000000406,     0.00000009310,
       0.00000000332,    -0.00000000046,
       0.00000002921,    -0.00000001822,
      -0.00000000184,    -0.00000000239,
      -0.00000000615,     0.00000000026,
       0.00000000028,     0.00000000620,
      -0.00000000095,    -0.00000000390,
      -0.00000000391,     0.00000000096,
       0.00000202329,     0.00000015700,
      -0.00000000355,    -0.00000003447,
       0.00000003859,    -0.00000048361,
      -0.00000000824,     0.00000000120,
       0.00000002130,     0.00000001281,
       0.00000000050,    -0.00000000090,
      -0.00000000215,    -0.00000001833,
       0.00000000023,     0.00000000114,
       0.00000000049,     0.00000000155,
      -0.00000000081,     0.00000000120,
      -0.00001820632,    -0.00000220651,
       0.00000007902,     0.00000038728,
      -0.00000014614,     0.00000198575,
       0.00000003337,    -0.00000000427,
       0.00000000109,     0.00000000000,
      -0.00000012979,     0.00000021534,
      -0.00000002036,    -0.00000001687,
       0.00000000146,    -0.00000000061,
       0.00000025304,     0.00000016306,
      -0.00000001864,     0.00000001512,
      -0.00000000095,    -0.00000000116,
       0.00000000291,    -0.00000000674,
      -0.00000000272,    -0.00000000598,
      -0.00000003919,    -0.00000011714,
      -0.00000002210,     0.00000000801,
      -0.00000000554,     0.00000000086,
       0.00000000028,     0.00000000099,
      -0.00000004509,    -0.00000001338,
      -0.00000000173,     0.00000000159,
      -0.00000001289,     0.00000004529,
       0.00000000158,     0.00000000167,
       0.00000009936,     0.00000002318,
       0.00000000583,    -0.00000002018,
      -0.00000000064,    -0.00000000067,
       0.00000001546,    -0.00000000045,
      -0.00000000098,    -0.00000001550,
      -0.00000017456,    -0.00000001772,
       0.00000000024,     0.00000000679,
      -0.00000000015,     0.00000000232,
       0.00000000248,     0.00000000054,
      -0.00000186605,    -0.00000031831,
      -0.00000004082,     0.00000003971,
      -0.00000001677,     0.00000003176,
       0.00000000184,     0.00000000207,
      -0.00000014082,     0.00000002772,
      -0.00000000162,    -0.00000000323,
       0.00000002870,     0.00000014077,
      -0.00000000326,     0.00000000146,
      -0.00000020837,     0.00000981036,
      -0.00000113207,    -0.00000067614,
       0.00000007667,    -0.00000007149,
      -0.00000027791,    -0.00000000613,
       0.00000001895,    -0.00000003173,
       0.00000000197,     0.00000000210,
      -0.00000000314,     0.00000000537,
      -0.00000000682,    -0.00000000468,
       0.00000000055,    -0.00000000062,
      -0.00000000021,    -0.00000000354,
       0.00000000355,    -0.00000000021,
       0.00000000010,     0.00000000283,
      -0.00000000283,     0.00000000010,
       0.00000000061,    -0.00000000097,
      -0.00000000156,     0.00000000189,
      -0.00000000195,    -0.00000000159,
       0.00000000410,    -0.00000000289,
       0.00000000291,     0.00000000410,
      -0.00000026178,     0.00000007872,
      -0.00000000842,    -0.00000003747,
       0.00000000546,     0.00000001626,
      -0.00000000308,     0.00000000232,
      -0.00000000020,    -0.00000000087,
       0.00000001621,     0.00000021584,
      -0.00000002970,    -0.00000001334,
       0.00000000174,    -0.00000000157,
      -0.00000021751,     0.00000001582,
       0.00000001386,    -0.00000003010,
       0.00000000157,     0.00000000188,
      -0.00000013654,    -0.00000009278,
       0.00000001243,     0.00000000360,
       0.00000009266,    -0.00000013642,
      -0.00000000357,     0.00000001240,
      -0.00000000069,     0.00000000095,
       0.00000000095,     0.00000000069,
      -0.00000000023,    -0.00000000141,
      -0.00000000081,    -0.00000000167,
      -0.00000000167,     0.00000000081,
      -0.00000005354,     0.00000000815,
       0.00000000083,     0.00000000589,
      -0.00000000071,     0.00000000072,
       0.00000000092,     0.00000000077,
      -0.00000001681,     0.00000002205,
       0.00000000112,     0.00000000106,
       0.00000000713,     0.00000001510,
       0.00000000243,    -0.00000000166,
       0.00000001511,    -0.00000000715,
      -0.00000000165,    -0.00000000244,
      -0.00000001096,    -0.00000002404,
      -0.00000000663,     0.00000000259,
       0.00000000068,     0.00000000102,
      -0.00000001680,    -0.00000000104,
      -0.00000000097,     0.00000001693,
       0.00000016588,     0.00000036102,
       0.00000005811,    -0.00000003969,
       0.00000001979,    -0.00000000750,
      -0.00000000201,    -0.00000000301,
       0.00000015131,     0.00000000629,
      -0.00000000011,    -0.00000000247,
       0.00000000612,    -0.00000015153,
      -0.00000000252,     0.00000000017,
      -0.00000031240,    -0.00000001978,
      -0.00000000416,     0.00000006425,
       0.00000000105,    -0.00000000024,
      -0.00000000044,    -0.00000000212,
       0.00000000056,     0.00000000051,
      -0.00000001396,     0.00000032897,
      -0.00000002467,    -0.00000001931,
      -0.00000001273,     0.00000000657,
       0.00000000088,    -0.00000000096,
      -0.00000000212,     0.00000000366,
       0.00000000356,    -0.00000000097,
       0.00000277801,     0.00000003096,
       0.00000000772,    -0.00000004787,
       0.00000001165,    -0.00000017869,
      -0.00000000281,     0.00000000071,
       0.00000002732,    -0.00000016728,
       0.00000001296,     0.00000001273,
      -0.00000000098,     0.00000000063,
      -0.00000017006,    -0.00000002676,
       0.00000001296,    -0.00000001259,
       0.00000000063,     0.00000000098,
      -0.00000000890,     0.00000000131,
       0.00000000030,     0.00000000167,
      -0.00000003386,     0.00000004604,
      -0.00000000206,     0.00000000044,
       0.00000124451,    -0.00000901475,
       0.00000069546,     0.00000070271,
      -0.00000005205,     0.00000003270,
       0.00000026753,     0.00000003241,
      -0.00000002071,     0.00000002058,
      -0.00000000098,    -0.00000000154,
       0.00000000435,    -0.00000001310,
       0.00000000135,     0.00000000074,
      -0.00000001320,     0.00000000479,
       0.00000000091,    -0.00000000130,
       0.00000000051,     0.00000000050,
       0.00000000007,    -0.00000000202,
       0.00000000349,     0.00000000126,
       0.00000000104,     0.00000000273,
      -0.00000001703,     0.00000007857,
      -0.00000011560,    -0.00000000423,
       0.00000000765,    -0.00000001111,
       0.00000000071,     0.00000000079,
      -0.00000000403,    -0.00000000334,
       0.00000000349,    -0.00000000439,
      -0.00000000033,    -0.00000000073,
      -0.00000001758,     0.00000021654,
      -0.00000002057,    -0.00000001458,
       0.00000003809,     0.00000000278,
      -0.00000000264,     0.00000000378,
       0.00000000051,     0.00000000292,
      -0.00000000274,     0.00000000066,
       0.00000000032,    -0.00000000201,
       0.00000000201,     0.00000000032,
      -0.00000000112,    -0.00000000015,
       0.00000001649,    -0.00000004116,
       0.00000000142,     0.00000000014,
      -0.00000000330,    -0.00000000036,
       0.00000000033,    -0.00000000332,
       0.00000000408,     0.00000000102,
      -0.00000000101,     0.00000000408,
      -0.00000000776,    -0.00000000854,
      -0.00000000025,    -0.00000000087,
       0.00000000062,    -0.00000000042,
       0.00000000525,    -0.00000000109,
      -0.00000000111,    -0.00000000527,
      -0.00000001131,     0.00000000199,
       0.00000000041,     0.00000000248,
      -0.00000000071,     0.00000000061,
       0.00000010412,    -0.00000002104,
      -0.00000000110,    -0.00000000669,
       0.00000000056,    -0.00000000614,
      -0.00000000624,    -0.00000000051,
      -0.00000001646,    -0.00000033669,
       0.00000002587,     0.00000001887,
       0.00000001035,    -0.00000000261,
      -0.00000000063,     0.00000000082,
       0.00000000039,    -0.00000000442,
      -0.00000000436,    -0.00000000001,
      -0.00000001934,     0.00000001234,
      -0.00000000850,     0.00000000106,
       0.00000000056,    -0.00000000079,
       0.00000001312,    -0.00000007007,
       0.00000001145,    -0.00000000060,
      -0.00000000071,     0.00000000097,
      -0.00000000354,     0.00000005157,
       0.00000000220,     0.00000000002,
       0.00000001395,    -0.00000012295,
      -0.00000000135,    -0.00000000014,
      -0.00000000001,    -0.00000000123,
       0.00000000125,     0.00000000002,
       0.00000000353,    -0.00000001742,
       0.00000000047,     0.00000000929,
      -0.00000000084,    -0.00000000063,
      -0.00000000939,     0.00000000048,
       0.00000000063,    -0.00000000088,
       0.00000000034,     0.00000000161,
      -0.00000000161,     0.00000000035,
      -0.00000000126,     0.00000000876,
       0.00000000110,     0.00000000028,
       0.00000000147,    -0.00000000024,
       0.00000000018,     0.00000000137,
       0.00000000096,    -0.00000000012,
      -0.00000000079,     0.00000002596,
       0.00000001516,    -0.00000005781,
       0.00000000050,     0.00000000055,
       0.00000002736,    -0.00000002123,
      -0.00000000231,    -0.00000000270,
      -0.00000000569,     0.00000000078,
       0.00000000071,     0.00000000569,
       0.00000000204,     0.00000000024,
      -0.00000000006,    -0.00000000097,
       0.00000000024,    -0.00000000204,
      -0.00000000097,     0.00000000006,
      -0.00000000024,     0.00000000101,
      -0.00000000265,     0.00000000413,
       0.00000000143,     0.00000000066,
       0.00000000413,     0.00000000265,
       0.00000000065,    -0.00000000143,
       0.00000006445,     0.00000000235,
      -0.00000000239,    -0.00000003033,
       0.00000000083,    -0.00000000325,
      -0.00000000153,    -0.00000000005,
       0.00000002405,    -0.00000003666,
      -0.00000001259,    -0.00000000566,
      -0.00000000063,     0.00000000197,
      -0.00000003670,    -0.00000002405,
      -0.00000000566,     0.00000001260,
       0.00000000196,     0.00000000063,
       0.00000000144,    -0.00000003308,
       0.00000006213,     0.00000005057,
       0.00000001972,     0.00000000560,
       0.00000001562,    -0.00000000908,
       0.00000000252,    -0.00000000607,
      -0.00000000098,    -0.00000000036,
       0.00000004422,     0.00000007634,
       0.00000001122,    -0.00000000830,
      -0.00000000063,    -0.00000000070,
       0.00000007591,    -0.00000004420,
      -0.00000000843,    -0.00000001127,
      -0.00000000070,     0.00000000059,
      -0.00000000028,     0.00000000138,
       0.00000000146,     0.00000000024,
      -0.00000089834,     0.00000004532,
       0.00000000757,     0.00000003230,
       0.00000000901,     0.00000009844,
       0.00000000324,    -0.00000000059,
      -0.00000000809,     0.00000001090,
      -0.00000000101,    -0.00000000051,
       0.00000001301,     0.00000000937,
      -0.00000000065,     0.00000000066,
       0.00000072721,    -0.00000119150,
      -0.00000040271,    -0.00000017416,
      -0.00000001987,     0.00000006262,
      -0.00000005547,    -0.00000004597,
      -0.00000001028,     0.00000001953,
       0.00000000306,     0.00000000109,
      -0.00000037952,    -0.00000069257,
      -0.00000010180,     0.00000007591,
       0.00000000539,     0.00000000648,
      -0.00000069290,     0.00000038031,
       0.00000007629,     0.00000010187,
       0.00000000647,    -0.00000000547,
       0.00000000093,    -0.00000000433,
      -0.00000000436,    -0.00000000090,
       0.00000086132,     0.00000149643,
       0.00000022274,    -0.00000016578,
      -0.00000001158,    -0.00000001379,
       0.00000040162,    -0.00000018743,
      -0.00000004121,    -0.00000005562,
      -0.00000000350,     0.00000000295,
       0.00000105403,     0.00000000394,
       0.00000000208,     0.00000000082,
      -0.00000000323,    -0.00000106165,
      -0.00000000094,     0.00000000020,
      -0.00000005051,    -0.00000000627,
       0.00000000045,     0.00000000546,
      -0.00000000051,     0.00000000131,
       0.00000000150,     0.00000000059,
      -0.00001186597,    -0.00002205697,
      -0.00000323224,     0.00000241785,
       0.00000017069,     0.00000020432,
      -0.00000119909,     0.00000054690,
       0.00000012138,     0.00000016418,
       0.00000001042,    -0.00000000871,
      -0.00000951523,     0.00000018190,
      -0.00000003462,     0.00000000621,
      -0.00000000017,     0.00000000540,
       0.00000019296,     0.00000952822,
       0.00000000893,     0.00000003055,
       0.00000000476,    -0.00000000009,
       0.00000000182,    -0.00000000128,
      -0.00000000114,    -0.00000000115,
       0.00000000005,     0.00000000179,
       0.00003017389,    -0.00000003572,
       0.00000002608,    -0.00000000168,
      -0.00000000039,    -0.00000002175,
      -0.00000000956,    -0.00000615737,
      -0.00000000050,    -0.00000000416,
      -0.00000000446,     0.00000000003,
      -0.00000000948,     0.00000019635,
       0.00000000411,    -0.00000001372,
      -0.00000000013,    -0.00000000119,
      -0.00000004787,     0.00000000906,
       0.00000000279,     0.00000001938,
       0.00000000026,    -0.00000000108,
      -0.00000000126,     0.00000000878,
      -0.00000000063,     0.00000000060,
      -0.00000000654,    -0.00000000168,
      -0.00000000045,    -0.00000000084,
      -0.00000145435,     0.00000747397,
      -0.00000057907,    -0.00000058234,
       0.00000004309,    -0.00000002679,
      -0.00000028942,     0.00000002808,
       0.00000002256,    -0.00000002203,
       0.00000000107,     0.00000000165,
      -0.00000003876,     0.00000004075,
      -0.00000000455,    -0.00000000232,
       0.00000005414,    -0.00000002804,
      -0.00000000413,     0.00000000436,
      -0.00000000103,    -0.00000000038,
      -0.00000000193,     0.00000000161,
      -0.00026023236,     0.00001330930,
      -0.00000216764,     0.00000052101,
      -0.00000000853,     0.00000005402,
      -0.00000009379,     0.00001659107,
      -0.00000000274,    -0.00000000356,
       0.00000001308,    -0.00000000033,
      -0.00000128114,     0.00001655418,
      -0.00000162353,    -0.00000113544,
       0.00000010671,    -0.00000009594,
       0.00000000312,     0.00000000585,
       0.00001680671,     0.00000121202,
      -0.00000115719,     0.00000159364,
      -0.00000009781,    -0.00000010517,
       0.00000000593,    -0.00000000304,
      -0.00000000429,    -0.00000002743,
      -0.00000000202,    -0.00000000194,
      -0.00000002069,     0.00000000571,
      -0.00000000141,     0.00000000273,
      -0.00000005836,    -0.00000023022,
       0.00000001401,     0.00000000925,
      -0.00000004986,     0.00000001180,
       0.00000000216,    -0.00000000325,
      -0.00000006956,    -0.00000000057,
       0.00000000025,    -0.00000007003,
      -0.00000000720,     0.00000007538,
      -0.00000000392,    -0.00000000026,
      -0.00000000279,     0.00000000000,
       0.00000000006,    -0.00000000279,
      -0.00014305493,     0.00185114405,
      -0.00017899114,    -0.00012760130,
       0.00001178514,    -0.00001077750,
       0.00000034853,     0.00000066830,
      -0.00000002742,     0.00000000298,
      -0.00005360810,    -0.00000420098,
       0.00000370000,    -0.00000518929,
       0.00000031238,     0.00000034164,
      -0.00000001902,     0.00000000994,
      -0.00000063497,    -0.00000007635,
      -0.00000000321,     0.00000002534,
      -0.00000000186,     0.00000000016,
      -0.00000017695,    -0.00000096204,
       0.00000002144,    -0.00000001306,
       0.00000000000,     0.00000000178,
      -0.00000013093,     0.00000001596,
      -0.00000000311,    -0.00000000249,
       0.00000000102,    -0.00000000031,
       0.00000001750,     0.00000013411,
      -0.00000000158,     0.00000000294,
      -0.00000000032,    -0.00000000112,
      -0.00000004058,     0.00000047498,
      -0.00000008347,    -0.00000009365,
       0.00000002575,    -0.00000000966,
       0.00000002084,     0.00000000376,
      -0.00000000361,     0.00000000325,
      -0.00000000050,    -0.00000000084,
       0.00000004625,    -0.00000003539,
       0.00000000120,     0.00000000689,
       0.00000003541,     0.00000004630,
      -0.00000000690,     0.00000000119,
      -0.00000011057,     0.00000001370,
      -0.00000000253,    -0.00000000174,
       0.00000000090,    -0.00000000027,
      -0.00000001373,    -0.00000011056,
       0.00000000174,    -0.00000000253,
       0.00000000027,     0.00000000090,
       0.00000000025,     0.00000000159,
      -0.00000000161,     0.00000000026,
      -0.00000000109,    -0.00000000063,
       0.00000000063,    -0.00000000109,
       0.00000000030,     0.00000000382,
      -0.00000000076,    -0.00000000050,
      -0.00000000383,     0.00000000030,
       0.00000000050,    -0.00000000076,
      -0.00000000842,    -0.00000002576,
       0.00000000260,    -0.00000000083,
      -0.00000000137,    -0.00000000034,
      -0.00000000077,    -0.00000000234,
       0.00000000236,    -0.00000000074,
      -0.00000000021,     0.00000000115,
      -0.00000000115,    -0.00000000039,
      -0.00000000013,     0.00000000257,
       0.00000000001,    -0.00000000135,
      -0.00000003044,     0.00000001199,
       0.00000000006,    -0.00000000588,
       0.00000000040,     0.00000000008,
      -0.00000010309,     0.00000001321,
       0.00000000462,    -0.00000001462,
       0.00000000080,     0.00000000035,
      -0.00000001286,    -0.00000010330,
       0.00000001461,     0.00000000469,
      -0.00000000035,     0.00000000080,
      -0.00000000323,     0.00000000440,
      -0.00000000007,    -0.00000000062,
      -0.00000000565,    -0.00000000342,
       0.00000000074,    -0.00000000021,
       0.00000000002,     0.00000000003,
      -0.00000000170,    -0.00000000123,
       0.00000000052,    -0.00000000063,
      -0.00000000413,    -0.00000000689,
       0.00000000115,     0.00000000074,
       0.00000000688,    -0.00000000413,
      -0.00000000074,     0.00000000114,
      -0.00000000278,     0.00000000435,
      -0.00000000437,    -0.00000000280,
      -0.00000000526,    -0.00000000447,
       0.00000000117,    -0.00000000097,
      -0.00000003088,    -0.00000000924,
       0.00000000397,    -0.00000000381,
       0.00000000019,     0.00000000033,
       0.00000000918,    -0.00000003096,
       0.00000000381,     0.00000000396,
      -0.00000000033,     0.00000000019,
       0.00000000898,     0.00000000448,
      -0.00000000072,    -0.00000000029,
      -0.00000000452,     0.00000000904,
       0.00000000029,    -0.00000000074,
       0.00000004770,    -0.00000005803,
       0.00000001337,     0.00000001366,
      -0.00000000182,     0.00000000162,
       0.00000011820,     0.00000045641,
      -0.00000008962,     0.00000000335,
       0.00000000034,    -0.00000000741,
      -0.00000045570,     0.00000011874,
      -0.00000000351,    -0.00000008932,
       0.00000000737,     0.00000000031,
       0.00000000336,     0.00000002218,
      -0.00000000061,    -0.00000000092,
      -0.00000002196,     0.00000000231,
       0.00000000111,    -0.00000000054,
       0.00001293452,     0.00000225212,
      -0.00000154680,     0.00000500502,
      -0.00000098353,    -0.00000043970,
       0.00000007700,    -0.00000012572,
       0.00000001162,     0.00000000961,
       0.00000121479,    -0.00000613703,
       0.00000237571,     0.00000077423,
      -0.00000021554,     0.00000046879,
      -0.00000005976,    -0.00000003737,
       0.00000000469,    -0.00000000553,
      -0.00000050312,     0.00000182863,
      -0.00000036081,    -0.00000008974,
       0.00000000820,    -0.00000004472,
       0.00000000346,     0.00000000072,
      -0.00000187152,    -0.00000055666,
       0.00000012094,    -0.00000038357,
       0.00000005025,     0.00000001716,
      -0.00000000245,     0.00000000446,
       0.00000006461,     0.00000011720,
       0.00000000472,    -0.00000000395,
       0.00000000040,     0.00000000174,
      -0.00000012378,     0.00000005379,
       0.00000000697,     0.00000000331,
      -0.00000000162,     0.00000000082,
      -0.00005011592,    -0.00001137825,
       0.00000690730,    -0.00001951953,
       0.00000381166,     0.00000187663,
      -0.00000032061,     0.00000048533,
      -0.00000004455,    -0.00000003922,
      -0.00000274120,     0.00001374199,
      -0.00000528545,    -0.00000173907,
       0.00000048106,    -0.00000104093,
       0.00000013289,     0.00000008293,
      -0.00000001039,     0.00000001230,
      -0.00000051669,    -0.00000621253,
       0.00000148189,     0.00000008652,
      -0.00000005792,     0.00000018814,
      -0.00000001290,    -0.00000001238,
       0.00000718568,    -0.00000227928,
       0.00000064177,     0.00000194790,
      -0.00000029969,     0.00000009913,
      -0.00000001033,    -0.00000003064,
       0.00000000268,    -0.00000000084,
      -0.00000017911,    -0.00000030054,
      -0.00000000630,     0.00000000181,
       0.00000000074,    -0.00000000237,
       0.00000034141,    -0.00000012727,
      -0.00000001725,     0.00000000221,
       0.00000000151,    -0.00000000163,
       0.00000003937,    -0.00000001339,
      -0.00000000199,    -0.00000014542,
       0.00000003186,     0.00000001917,
      -0.00000000391,     0.00000000303,
      -0.00000204898,    -0.00000026806,
       0.00000030683,    -0.00000037567,
       0.00000003115,     0.00000005415,
      -0.00000000553,     0.00000000133,
       0.00000026672,    -0.00000204738,
       0.00000037532,     0.00000030640,
      -0.00000005408,     0.00000003110,
      -0.00000000133,    -0.00000000551,
      -0.00000000018,     0.00000000815,
      -0.00000000947,    -0.00000000118,
       0.00000001275,    -0.00000003932,
      -0.00000002889,     0.00000001968,
      -0.00000000082,    -0.00000000828,
       0.00000000105,     0.00000000025,
      -0.00000256999,    -0.00000032943,
       0.00000038582,    -0.00000047134,
       0.00000003905,     0.00000006785,
      -0.00000000692,     0.00000000167,
       0.00000032916,    -0.00000256797,
       0.00000047071,     0.00000038540,
      -0.00000006772,     0.00000003895,
      -0.00000000167,    -0.00000000691,
       0.00000000003,     0.00000000002,
       0.00000000004,     0.00000000008,
      -0.00000000008,     0.00000000004,
       0.00000000020,     0.00000000092,
      -0.00000000002,     0.00000000005,
       0.00000001007,     0.00000000061,
      -0.00000000056,     0.00000000081,
      -0.00000000677,    -0.00000000034,
       0.00000000039,    -0.00000000057,
       0.00000000189,     0.00000000049,
      -0.00000000019,     0.00000000074,
      -0.00000001503,    -0.00000000393,
       0.00000000086,    -0.00000000127,
      -0.00000000053,    -0.00000000027,
      -0.00000109100,    -0.00000036826,
      -0.00000006341,     0.00000019862,
       0.00000001990,     0.00000000600,
       0.00000000042,    -0.00000000116,
      -0.00000000038,    -0.00000000013,
       0.00000517349,     0.00000187970,
       0.00000044899,    -0.00000125917,
      -0.00000015320,    -0.00000005360,
      -0.00000000430,     0.00000001240,
      -0.00000000055,    -0.00000000172,
      -0.00000000056,     0.00000000005,
      -0.00000000144,     0.00000000068,
       0.00000000003,     0.00000000055,
      -0.00000000132,    -0.00000000080,
       0.00000000020,    -0.00000000029,
       0.00000000079,    -0.00000000132,
       0.00000000029,     0.00000000020,
      -0.00000000027,    -0.00000000024,
       0.00000000025,    -0.00000000027,
       0.00000000057,     0.00000000095,
       0.00000000024,    -0.00000000031,
      -0.00000000032,    -0.00000000025,
      -0.00000000040,    -0.00000000026,
       0.00000000034,    -0.00000000027,
      -0.00000000025,    -0.00000000037,
       0.00000000030,    -0.00000000052,
      -0.00000000044,    -0.00000000024,
       0.00000000036,    -0.00000000058,
      -0.00000000041,    -0.00000000022,
       0.00000000038,    -0.00000000013,
      -0.00000000013,    -0.00000000038,
      -0.00000000009,    -0.00000000051,
      -0.00000000046,     0.00000000047,
       0.00000000607,     0.00000000618,
       0.00000000116,    -0.00000000005,
      -0.00000000053,    -0.00000000110,
      -0.00000000110,     0.00000000056,
      -0.00000000071,     0.00000000045,
       0.00000000058,    -0.00000000086,
      -0.00000000068,    -0.00000000057,
       0.00000000093,    -0.00000000082,
      -0.00000000095,    -0.00000000078,
       0.00000000071,    -0.00000000093,
      -0.00000000093,    -0.00000000076,
       0.00000000096,    -0.00000000143,
      -0.00000000124,    -0.00000000088,
       0.00000000060,    -0.00000000109,
      -0.00000000091,    -0.00000000055,
       0.00000000095,    -0.00000000167,
      -0.00000000148,    -0.00000000087,
       0.00000000009,     0.00000000001,
       0.00000000005,     0.00000000000,
       0.00000000001,    -0.00000000005,
       0.00000000044,    -0.00000000011,
      -0.00000000011,    -0.00000000044,
      -0.00000000014,    -0.00000000058,
       0.00000000145,    -0.00000000082,
      -0.00000000082,    -0.00000000145,
       0.00000000053,    -0.00000000011,
      -0.00000000020,    -0.00000000105,
       0.00000000079,    -0.00000000057,
      -0.00000000058,    -0.00000000077,
      -0.00000000063,     0.00000000100,
       0.00000000298,    -0.00000000560,
       0.00000000064,    -0.00000000046,
      -0.00000000074,    -0.00000000102,
      -0.00000000039,    -0.00000000027,
      -0.00000000189,     0.00000000087,
       0.00000000459,     0.00000000950,
       0.00000000105,     0.00000000030,
       0.00000000156,    -0.00000000481,
      -0.00000000489,    -0.00000000164,
       0.00000000076,    -0.00000000193,
      -0.00000000270,    -0.00000000066,
       0.00000000277,    -0.00000000289,
      -0.00000000303,    -0.00000000297,
       0.00000000329,    -0.00000000345,
       0.00000000042,    -0.00000000001,
      -0.00000000351,    -0.00000000322,
       0.00000000211,    -0.00000000363,
       0.00000000034,     0.00000000029,
      -0.00000000329,    -0.00000000206,
       0.00000000026,    -0.00000000034,
       0.00000000303,    -0.00000000484,
       0.00000000055,     0.00000000037,
      -0.00000000467,    -0.00000000298,
       0.00000000037,    -0.00000000048,
       0.00000000026,    -0.00000000059,
       0.00000000008,    -0.00000000075,
       0.00000000035,    -0.00000000059,
       0.00000000049,     0.00000000029,
       0.00000000001,     0.00000000004,
       0.00000000032,    -0.00000000008,
       0.00000000021,    -0.00000000009,
      -0.00000000009,    -0.00000000023,
       0.00000000005,    -0.00000000002,
      -0.00000000010,    -0.00000000025,
      -0.00000000004,    -0.00000000003,
       0.00000000153,    -0.00000000099,
      -0.00000000099,    -0.00000000153,
       0.00000000065,    -0.00000000042,
      -0.00000000108,    -0.00000000167,
       0.00000000026,    -0.00000000039,
      -0.00000000039,    -0.00000000026,
       0.00000000016,    -0.00000000024,
      -0.00000000051,    -0.00000000034,
       0.00000000065,    -0.00000000057,
       0.00000000509,    -0.00000000327,
      -0.00000000328,    -0.00000000511,
       0.00000000267,    -0.00000000163,
      -0.00000000311,    -0.00000000516,
       0.00000000116,    -0.00000000170,
      -0.00000000171,    -0.00000000111,
       0.00000000094,    -0.00000000143,
       0.00000000014,     0.00000000012,
      -0.00000000223,    -0.00000000149,
       0.00000000022,     0.00000000039,
      -0.00000000740,     0.00000000122,
      -0.00000000031,     0.00000000041,
       0.00000000399,     0.00000002603,
       0.00000000144,     0.00000000111,
       0.00000001759,    -0.00000001302,
       0.00000000153,    -0.00000000019,
      -0.00000001317,    -0.00000001723,
      -0.00000000013,    -0.00000000148,
       0.00000001045,    -0.00000000757,
       0.00000000086,    -0.00000000004,
      -0.00000000957,    -0.00000001301,
      -0.00000000011,    -0.00000000117,
       0.00000000699,    -0.00000001139,
       0.00000000113,     0.00000000098,
      -0.00000001169,    -0.00000000831,
       0.00000000095,    -0.00000000124,
       0.00000000918,    -0.00000001411,
       0.00000000148,     0.00000000115,
      -0.00000001507,    -0.00000000987,
       0.00000000127,    -0.00000000156,
       0.00000000044,    -0.00000000079,
       0.00000000012,     0.00000000103,
      -0.00000000008,    -0.00000000156,
      -0.00000000086,     0.00000000014,
       0.00000000104,    -0.00000000166,
       0.00000000164,     0.00000000120,
       0.00000000007,     0.00000000007,
       0.00000000056,    -0.00000000059,
       0.00000000005,     0.00000000006,
       0.00000000028,    -0.00000000046,
       0.00000000004,     0.00000000003,
      -0.00000000048,    -0.00000000032,
       0.00000000004,    -0.00000000005,
       0.00000000007,    -0.00000000011,
      -0.00000000053,    -0.00000000033,
       0.00000000004,    -0.00000000004,
       0.00000000007,     0.00000000005,
       0.00000000002,     0.00000000001,
       0.00000000003,    -0.00000000007,
       0.00000000003,    -0.00000000004,
       0.00000000005,     0.00000000003,
       0.00000000242,    -0.00000000369,
       0.00000000038,     0.00000000031,
      -0.00000000371,    -0.00000000242,
       0.00000000031,    -0.00000000038,
       0.00000000729,    -0.00000000144,
       0.00000000068,    -0.00000000102,
      -0.00000000102,    -0.00000000068,
      -0.00000001200,    -0.00000000039,
       0.00000000106,    -0.00000000165,
       0.00000000017,     0.00000000014,
      -0.00000000412,    -0.00000000270,
       0.00000000034,    -0.00000000042,
       0.00000000022,    -0.00000000033,
       0.00000000033,     0.00000000022,
       0.00000000068,     0.00000000000,
      -0.00000000018,    -0.00000000294,
       0.00000000682,    -0.00000001087,
       0.00000000109,     0.00000000091,
      -0.00000001091,    -0.00000000714,
       0.00000000091,    -0.00000000109,
       0.00000000809,    -0.00000000164,
       0.00000000078,    -0.00000000117,
      -0.00000000116,    -0.00000000078,
      -0.00000001310,    -0.00000000037,
       0.00000000410,    -0.00000000625,
       0.00000000063,     0.00000000052,
      -0.00000001169,    -0.00000000765,
       0.00000000098,    -0.00000000117,
       0.00000000065,    -0.00000000099,
       0.00000000099,     0.00000000065,
      -0.00000001294,     0.00000000856,
      -0.00000000015,    -0.00000000052,
       0.00000002418,     0.00000003663,
      -0.00000000157,     0.00000000033,
       0.00000003614,    -0.00000005508,
       0.00000000547,     0.00000000459,
      -0.00000005540,    -0.00000003566,
       0.00000000462,    -0.00000000551,
       0.00000002512,    -0.00000003893,
       0.00000000394,     0.00000000334,
      -0.00000005164,    -0.00000003346,
       0.00000000440,    -0.00000000524,
      -0.00000000270,     0.00000000089,
      -0.00000000157,    -0.00000000139,
      -0.00000000049,    -0.00000000254,
       0.00000000050,     0.00000000001,
      -0.00000000152,     0.00000000065,
       0.00000000238,    -0.00000000363,
       0.00000000036,     0.00000000031,
       0.00000000358,     0.00000000243,
      -0.00000000031,     0.00000000036,
      -0.00000000001,    -0.00000000001,
      -0.00000000012,     0.00000000004,
      -0.00000000004,     0.00000000004,
       0.00000000004,     0.00000000002,
      -0.00000000049,     0.00000000025,
       0.00000000673,    -0.00000001017,
       0.00000000112,     0.00000000076,
       0.00000000778,    -0.00000001245,
       0.00000000029,     0.00000000022,
       0.00000000029,    -0.00000000048,
      -0.00000000253,    -0.00000000157,
      -0.00000000100,     0.00000000075,
       0.00000000063,     0.00000000009,
       0.00000000754,    -0.00000001135,
       0.00000000118,     0.00000000081,
       0.00000000831,    -0.00000001324,
      -0.00000000102,     0.00000000001,
       0.00000000002,    -0.00000000109,
      -0.00000000033,     0.00000000042,
       0.00000000077,     0.00000000060,
       0.00000000496,     0.00000000633,
       0.00000000440,    -0.00000000567,
       0.00000000057,     0.00000000051,
      -0.00000001936,    -0.00000001443,
       0.00000000171,    -0.00000000197,
      -0.00000000383,     0.00000000487,
      -0.00000000076,    -0.00000000033,
       0.00000000372,     0.00000000159,
      -0.00000000031,     0.00000000062,
      -0.00000000218,     0.00000000256,
      -0.00000000056,    -0.00000000026,
       0.00000000323,     0.00000000271,
      -0.00000000030,     0.00000000072,
      -0.00000000004,     0.00000000008,
      -0.00000000002,    -0.00000000001,
       0.00000000069,     0.00000000034,
      -0.00000000011,     0.00000000020,
       0.00000000049,     0.00000000012,
      -0.00000000004,     0.00000000015,
       0.00000000004,    -0.00000000059,
       0.00000000016,     0.00000000005,
      -0.00000000907,    -0.00000003455,
       0.00000000131,    -0.00000000035,
       0.00000000556,     0.00000000079,
      -0.00000000044,     0.00000000159,
       0.00000000096,    -0.00000000547,
       0.00000000159,     0.00000000044,
       0.00000000014,     0.00000000002,
      -0.00000000001,     0.00000000004,
       0.00000000009,    -0.00000000061,
       0.00000000018,     0.00000000005,
      -0.00000000002,     0.00000000008,
       0.00000000642,    -0.00000000329,
       0.00000000007,     0.00000000103,
       0.00000000102,    -0.00000000004,
       0.00000000000,     0.00000000003,
       0.00000000007,    -0.00000000001,
      -0.00000000001,     0.00000000005,
       0.00000000379,     0.00000000088,
      -0.00000000033,     0.00000000107,
       0.00000000032,    -0.00000000414,
       0.00000000112,     0.00000000027,
      -0.00000010191,    -0.00000039347,
       0.00000008389,    -0.00000000399,
       0.00000001506,    -0.00000000403,
       0.00000000020,     0.00000000322,
       0.00000006273,     0.00000000929,
      -0.00000000499,     0.00000001794,
       0.00000001068,    -0.00000006206,
       0.00000001785,     0.00000000518,
       0.00000000189,     0.00000000029,
      -0.00000000014,     0.00000000052,
       0.00000000069,    -0.00000000458,
       0.00000000126,     0.00000000035,
       0.00000000051,     0.00000000005,
       0.00000000992,     0.00000000257,
      -0.00000000093,     0.00000000288,
       0.00000000072,    -0.00000001095,
       0.00000000301,     0.00000000074,
      -0.00000010999,    -0.00000042557,
       0.00000009069,    -0.00000000423,
       0.00000001549,    -0.00000000410,
       0.00000000023,     0.00000000338,
       0.00000006794,     0.00000001007,
      -0.00000000540,     0.00000001942,
       0.00000001165,    -0.00000006709,
       0.00000001930,     0.00000000562,
       0.00000000672,     0.00000000103,
      -0.00000000054,     0.00000000189,
       0.00000000184,    -0.00000001204,
       0.00000000337,     0.00000000096,
      -0.00000000038,     0.00000000095,
       0.00000000095,     0.00000000038,
       0.00000000104,     0.00000000016,
      -0.00000000016,     0.00000000104,
      -0.00000000990,    -0.00000000678,
       0.00000000172,    -0.00000000196,
      -0.00000001683,     0.00000002300,
      -0.00000000451,    -0.00000000420,
       0.00000003685,     0.00000001023,
      -0.00000000354,     0.00000001079,
       0.00000000265,    -0.00000004092,
       0.00000001124,     0.00000000277,
      -0.00000000972,    -0.00000003813,
       0.00000000108,    -0.00000000028,
       0.00000000615,     0.00000000094,
      -0.00000000049,     0.00000000176,
       0.00000000094,    -0.00000000616,
       0.00000000176,     0.00000000049,
       0.00000003622,     0.00000000610,
      -0.00000000294,     0.00000001030,
       0.00000000732,    -0.00000004324,
       0.00000001228,     0.00000000351,
      -0.00000000022,    -0.00000000119,
       0.00000000035,     0.00000000014,
       0.00000000020,    -0.00000000251,
       0.00000000070,     0.00000000008,
       0.00000000338,    -0.00000000114,
       0.00000000032,     0.00000000122,
      -0.00000000090,    -0.00000000247,
       0.00000000089,    -0.00000000024,
       0.00000000393,     0.00000000056,
      -0.00000000030,     0.00000000107,
      -0.00000000066,     0.00000000385,
      -0.00000000105,    -0.00000000033,
       0.00000000049,    -0.00000000016,
       0.00000000015,     0.00000000049,
       0.00000000915,     0.00000000168,
       0.00000000969,     0.00000000180,
       0.00000000019,    -0.00000000135,
      -0.00000001078,     0.00000001444,
       0.00000000889,     0.00000000664,
      -0.00000000123,    -0.00000000127,
      -0.00000000035,    -0.00000000103,
       0.00000000100,    -0.00000000055,
      -0.00000000149,    -0.00000000015,
      -0.00000000016,     0.00000000149,
      -0.00000001987,    -0.00000002084,
       0.00000000380,    -0.00000000370,
      -0.00000000134,     0.00000000109,
       0.00000000178,     0.00000000770,
      -0.00000000025,    -0.00000000180,
       0.00000000771,    -0.00000000179,
      -0.00000000180,     0.00000000025,
       0.00000001347,     0.00000002484,
       0.00000002231,    -0.00000001210,
       0.00000000408,    -0.00000002581,
      -0.00000000337,    -0.00000000053,
      -0.00000000292,     0.00000000373,
       0.00000000390,     0.00000000283,
      -0.00000000052,     0.00000000462,
       0.00000000044,    -0.00000000211,
      -0.00000000422,    -0.00000000312,
       0.00000000147,     0.00000000115,
      -0.00000000312,     0.00000000422,
       0.00000000115,    -0.00000000147,
      -0.00000011662,    -0.00000001392,
       0.00000004887,     0.00000001455,
       0.00000000329,    -0.00000001575,
      -0.00000000226,     0.00000000649,
       0.00000000141,    -0.00000000177,
       0.00000003350,     0.00000002509,
      -0.00000001084,    -0.00000001052,
       0.00000000123,     0.00000000273,
       0.00000002525,    -0.00000003354,
      -0.00000001069,     0.00000001091,
       0.00000000273,    -0.00000000123,
       0.00000000118,    -0.00000000141,
      -0.00000000141,    -0.00000000118,
       0.00000000940,    -0.00000001030,
       0.00000000468,    -0.00000000091,
      -0.00000000126,     0.00000000006,
       0.00000000154,    -0.00000000424,
      -0.00000000445,    -0.00000000133,
      -0.00000003039,    -0.00000014529,
      -0.00000000007,     0.00000003748,
      -0.00000002549,     0.00000000630,
       0.00000000692,    -0.00000000026,
      -0.00000000071,    -0.00000000044,
      -0.00000001215,     0.00000003395,
       0.00000000381,    -0.00000000549,
       0.00000003394,     0.00000001237,
      -0.00000000564,    -0.00000000378,
      -0.00000000174,     0.00000000007,
      -0.00000000160,     0.00000000279,
       0.00000000289,     0.00000000154,
      -0.00000000085,     0.00000000033,
       0.00000000045,     0.00000000086,
      -0.00000003170,     0.00000004219,
       0.00000000388,     0.00000000263,
       0.00000000720,    -0.00000000334,
      -0.00000000322,    -0.00000000724,
      -0.00000000140,    -0.00000000013,
      -0.00000000014,     0.00000000151,
       0.00000002695,     0.00000002244,
      -0.00000000808,    -0.00000000884,
       0.00000001950,    -0.00000002729,
      -0.00000000814,     0.00000000854,
       0.00000000197,    -0.00000000069,
      -0.00000001285,    -0.00000003175,
       0.00000000248,     0.00000000916,
      -0.00000003194,     0.00000001289,
       0.00000000917,    -0.00000000279,
      -0.00000000190,     0.00000000085,
       0.00000000085,     0.00000000190,
      -0.00000066901,    -0.00000053830,
       0.00000020666,     0.00000021930,
      -0.00000001700,    -0.00000005358,
      -0.00000009064,     0.00000012807,
       0.00000003864,    -0.00000004042,
      -0.00000000993,     0.00000000354,
       0.00000000117,     0.00000000029,
       0.00000010913,     0.00000026578,
      -0.00000002090,    -0.00000007558,
      -0.00000000297,     0.00000001162,
       0.00000000104,    -0.00000000087,
       0.00000026656,    -0.00000010810,
      -0.00000007618,     0.00000002035,
       0.00000001165,     0.00000000302,
      -0.00000000087,    -0.00000000104,
       0.00000000893,    -0.00000000378,
      -0.00000000144,     0.00000000125,
      -0.00000000358,    -0.00000000901,
       0.00000000112,     0.00000000143,
       0.00000000026,    -0.00000000154,
       0.00000000229,     0.00000000180,
       0.00000000181,    -0.00000000229,
      -0.00000000014,     0.00000000112,
       0.00000000112,     0.00000000014,
       0.00000000505,    -0.00000001526,
      -0.00000000671,    -0.00000000153,
       0.00000000072,     0.00000000046,
      -0.00000000660,     0.00000000520,
       0.00000000096,    -0.00000000016,
       0.00000000516,     0.00000000674,
      -0.00000000032,    -0.00000000091,
      -0.00000011029,     0.00000033744,
       0.00000003479,    -0.00000005442,
       0.00000003161,     0.00000000771,
      -0.00000000540,    -0.00000000311,
       0.00000005607,    -0.00000004385,
      -0.00000000794,     0.00000000188,
       0.00000000064,     0.00000000071,
      -0.00000004347,    -0.00000005658,
       0.00000000173,     0.00000000805,
       0.00000000066,    -0.00000000068,
      -0.00000000058,    -0.00000000081,
      -0.00000000086,     0.00000000050,
      -0.00000000076,     0.00000000042,
       0.00000000052,     0.00000000075,
      -0.00000002491,     0.00000004158,
       0.00000000345,     0.00000000172,
       0.00000000638,    -0.00000000411,
      -0.00000000405,    -0.00000000640,
      -0.00000000655,     0.00000000330,
       0.00000000109,     0.00000000186,
       0.00000000163,     0.00000000003,
       0.00000000007,    -0.00000000165,
       0.00000012391,    -0.00000006293,
      -0.00000000517,    -0.00000000918,
       0.00000000030,     0.00000000074,
      -0.00000001399,    -0.00000000021,
      -0.00000000060,     0.00000001392,
      -0.00000000180,    -0.00000000018,
      -0.00000000012,     0.00000000192,
      -0.00000003137,    -0.00000000249,
      -0.00000000018,     0.00000000212,
       0.00000000039,    -0.00000000088,
      -0.00000004399,    -0.00000010967,
       0.00000000811,     0.00000003101,
      -0.00000004568,     0.00000002515,
       0.00000001389,    -0.00000000527,
      -0.00000000225,    -0.00000000036,
      -0.00000001042,     0.00000009255,
       0.00000000393,    -0.00000001798,
      -0.00000000179,     0.00000000146,
       0.00000009270,     0.00000001125,
      -0.00000001798,    -0.00000000421,
       0.00000000156,     0.00000000170,
       0.00000000413,     0.00000000004,
       0.00000000031,    -0.00000000414,
       0.00000099644,     0.00000246338,
      -0.00000018795,    -0.00000069236,
      -0.00000002812,     0.00000010269,
       0.00000021839,    -0.00000012011,
      -0.00000006570,     0.00000002454,
       0.00000001061,     0.00000000192,
      -0.00000000080,    -0.00000000103,
       0.00000008417,    -0.00000078448,
      -0.00000003298,     0.00000015308,
       0.00000001391,    -0.00000001187,
      -0.00000000169,    -0.00000000016,
      -0.00000078497,    -0.00000008815,
       0.00000015270,     0.00000003449,
      -0.00000001163,    -0.00000001423,
      -0.00000000017,     0.00000000171,
      -0.00000001935,    -0.00000000074,
       0.00000000251,    -0.00000000103,
      -0.00000000135,     0.00000001921,
      -0.00000000085,    -0.00000000248,
       0.00000000126,    -0.00000000159,
      -0.00000000086,    -0.00000000210,
      -0.00000000210,     0.00000000086,
      -0.00000004086,    -0.00000003837,
       0.00000001362,     0.00000001616,
      -0.00000000689,     0.00000000791,
       0.00000000316,    -0.00000000289,
       0.00000000610,     0.00000001707,
      -0.00000000118,    -0.00000000522,
       0.00000001714,    -0.00000000610,
      -0.00000000528,     0.00000000136,
       0.00000000125,     0.00000000015,
       0.00000000191,    -0.00000000251,
      -0.00000000254,    -0.00000000193,
      -0.00000004528,     0.00000003903,
       0.00000001343,     0.00000001349,
      -0.00000000090,    -0.00000000219,
       0.00000001956,    -0.00000000074,
      -0.00000000074,    -0.00000000116,
      -0.00000000030,    -0.00000001975,
      -0.00000000125,     0.00000000074,
       0.00000086584,    -0.00000075706,
      -0.00000012201,     0.00000004521,
       0.00000001016,     0.00000000907,
      -0.00000006300,    -0.00000006376,
       0.00000000408,     0.00000001013,
       0.00000000080,    -0.00000000092,
      -0.00000016734,     0.00000000537,
       0.00000000605,     0.00000000980,
       0.00000000069,    -0.00000000112,
       0.00000000303,     0.00000016794,
       0.00000001041,    -0.00000000595,
      -0.00000000117,    -0.00000000073,
      -0.00000000051,     0.00000000107,
       0.00000000106,     0.00000000075,
      -0.00000000668,     0.00000000411,
       0.00000000120,     0.00000000174,
       0.00000000160,    -0.00000000018,
      -0.00000000013,    -0.00000000163,
       0.00000011825,    -0.00000007222,
      -0.00000000583,    -0.00000000837,
       0.00000000041,     0.00000000064,
      -0.00000001408,     0.00000000160,
       0.00000000126,     0.00000001405,
       0.00000001880,    -0.00000000064,
      -0.00000000026,    -0.00000000458,
      -0.00000000090,    -0.00000000051,
      -0.00000000069,     0.00000000080,
      -0.00000029694,     0.00000001164,
       0.00000000102,     0.00000002143,
       0.00000000761,     0.00000000492,
       0.00000000569,    -0.00000000684,
      -0.00000003882,    -0.00000000415,
      -0.00000000023,     0.00000000279,
       0.00000002057,     0.00000000640,
       0.00000000058,    -0.00000000165,
       0.00000000032,    -0.00000000105,
      -0.00000008006,     0.00000062045,
       0.00000002812,    -0.00000011937,
      -0.00000001128,     0.00000000924,
       0.00000020319,    -0.00000000116,
      -0.00000004219,    -0.00000000562,
       0.00000000368,     0.00000000366,
       0.00000020685,    -0.00000030264,
      -0.00000002968,     0.00000003600,
       0.00000000409,    -0.00000000011,
      -0.00000030143,    -0.00000021076,
       0.00000003505,     0.00000003077,
       0.00000000006,    -0.00000000420,
      -0.00000000967,    -0.00000000444,
       0.00000000096,    -0.00000000033,
      -0.00000000555,     0.00000000883,
       0.00000139703,    -0.00001205129,
      -0.00000053011,     0.00000232612,
       0.00000021312,    -0.00000017580,
      -0.00000002389,    -0.00000000399,
      -0.00000095208,     0.00000000477,
       0.00000019785,     0.00000002686,
      -0.00000001672,    -0.00000001704,
      -0.00000000020,     0.00000000216,
      -0.00000176534,     0.00000259404,
       0.00000025790,    -0.00000030925,
      -0.00000003497,     0.00000000086,
       0.00000000108,     0.00000000199,
       0.00000258399,     0.00000178338,
      -0.00000030457,    -0.00000026215,
      -0.00000000023,     0.00000003513,
       0.00000000212,    -0.00000000102,
       0.00000004519,     0.00000002267,
      -0.00000000455,     0.00000000083,
       0.00000002463,    -0.00000004325,
       0.00000000038,     0.00000000419,
       0.00000000066,     0.00000000080,
       0.00000000080,    -0.00000000066,
      -0.00000000258,     0.00000000131,
      -0.00000000090,     0.00000000496,
       0.00000000031,    -0.00000000104,
       0.00000000500,     0.00000000083,
      -0.00000000112,    -0.00000000024,
       0.00000004578,     0.00000013551,
      -0.00000000573,    -0.00000003940,
       0.00000001233,    -0.00000000581,
      -0.00000000389,     0.00000000090,
       0.00000000694,    -0.00000004165,
      -0.00000000294,     0.00000000830,
       0.00000000089,    -0.00000000068,
      -0.00000004168,    -0.00000000720,
       0.00000000827,     0.00000000295,
      -0.00000000075,    -0.00000000088,
      -0.00000000103,    -0.00000000006,
      -0.00000000010,     0.00000000102,
       0.00000000066,     0.00000000045,
       0.00000000088,    -0.00000000036,
      -0.00000000042,    -0.00000000087,
       0.00000003247,    -0.00000003790,
      -0.00000000311,    -0.00000000220,
      -0.00000000779,     0.00000000362,
       0.00000000355,     0.00000000785,
       0.00000020451,    -0.00000002793,
      -0.00000000833,    -0.00000000800,
      -0.00000000805,    -0.00000005173,
      -0.00000000233,     0.00000000256,
      -0.00000001436,    -0.00000002446,
      -0.00000000272,     0.00000000191,
      -0.00000002634,     0.00000001361,
       0.00000000217,     0.00000000293,
      -0.00000000054,     0.00000000059,
      -0.00000315493,     0.00000043680,
       0.00000013133,     0.00000012690,
       0.00000000808,    -0.00000001806,
       0.00000003718,     0.00000023857,
       0.00000001072,    -0.00000001144,
      -0.00000000159,    -0.00000000071,
       0.00000012308,     0.00000021950,
       0.00000002409,    -0.00000001762,
      -0.00000000106,    -0.00000000108,
       0.00000022801,    -0.00000011897,
      -0.00000001851,    -0.00000002523,
      -0.00000000113,     0.00000000118,
       0.00000000210,     0.00000000133,
       0.00000000181,    -0.00000000265,
       0.00000001943,    -0.00000000184,
      -0.00000000045,    -0.00000000441,
      -0.00000027852,     0.00000002489,
       0.00000000207,     0.00000002002,
       0.00000000715,    -0.00000000012,
       0.00000000069,    -0.00000000641,
      -0.00000001822,    -0.00000000787,
      -0.00000000195,     0.00000000382,
      -0.00000000204,    -0.00000000095,
      -0.00000000076,     0.00000000268,
       0.00000021758,     0.00000009646,
       0.00000000861,    -0.00000001680,
       0.00000001970,     0.00000000761,
       0.00000000715,    -0.00000002233,
       0.00000000070,     0.00000000049,
       0.00000000432,    -0.00000000853,
      -0.00000000177,    -0.00000000127,
       0.00000000096,     0.00000000266,
       0.00000000265,    -0.00000000096,
       0.00000000046,    -0.00000000130,
       0.00000206059,    -0.00000290292,
      -0.00000029344,     0.00000034072,
       0.00000003867,    -0.00000000009,
      -0.00000079662,    -0.00000044426,
       0.00000010041,     0.00000007501,
      -0.00000000080,    -0.00000001122,
      -0.00000000064,     0.00000000041,
      -0.00000116593,     0.00000062846,
       0.00000007411,    -0.00000003262,
      -0.00000000333,    -0.00000000329,
       0.00000060688,     0.00000117925,
      -0.00000002681,    -0.00000007549,
      -0.00000000403,     0.00000000296,
       0.00000001383,     0.00000001420,
      -0.00000000135,     0.00000000069,
       0.00000001629,    -0.00000000764,
      -0.00003171630,     0.00004564379,
       0.00000459356,    -0.00000536709,
      -0.00000061178,     0.00000000201,
       0.00000001793,     0.00000003468,
       0.00000368649,     0.00000204796,
      -0.00000046393,    -0.00000034672,
       0.00000000333,     0.00000005127,
       0.00000000291,    -0.00000000168,
       0.00001015205,    -0.00000545756,
      -0.00000065737,     0.00000027566,
       0.00000002922,     0.00000003034,
       0.00000000201,    -0.00000000110,
      -0.00000535813,    -0.00001020021,
       0.00000024998,     0.00000066118,
       0.00000003310,    -0.00000002752,
      -0.00000000117,    -0.00000000235,
      -0.00000006013,    -0.00000007583,
       0.00000000502,    -0.00000000071,
      -0.00000007877,     0.00000004835,
       0.00000000018,    -0.00000000335,
       0.00000000088,    -0.00000000054,
       0.00000000009,    -0.00000000295,
      -0.00000000294,    -0.00000000004,
      -0.00000000530,     0.00000002887,
       0.00000000938,     0.00000000040,
      -0.00000000198,    -0.00000000049,
       0.00000001030,    -0.00000001318,
      -0.00000000178,     0.00000000151,
      -0.00000001314,    -0.00000001046,
       0.00000000154,     0.00000000178,
       0.00000009518,    -0.00000054556,
      -0.00000003985,     0.00000010623,
       0.00000001234,    -0.00000000759,
      -0.00000004380,    -0.00000000234,
       0.00000000917,     0.00000000253,
      -0.00000000078,    -0.00000000107,
      -0.00000008747,     0.00000011291,
       0.00000001492,    -0.00000001249,
      -0.00000000194,    -0.00000000019,
       0.00000011236,     0.00000008830,
      -0.00000001223,    -0.00000001519,
      -0.00000000020,     0.00000000196,
       0.00000000200,     0.00000000115,
       0.00000000121,    -0.00000000191,
      -0.00000000143,     0.00000000124,
       0.00000000127,     0.00000000144,
       0.00000000934,    -0.00000000416,
      -0.00000000105,    -0.00000000209,
      -0.00000000163,     0.00000000003,
      -0.00000000002,     0.00000000165,
      -0.00000012828,     0.00000005613,
       0.00000000483,     0.00000000954,
       0.00000001443,    -0.00000000013,
       0.00000000024,    -0.00000001434,
      -0.00000029553,    -0.00000025953,
      -0.00000002930,     0.00000002096,
      -0.00000006359,     0.00000006414,
       0.00000000517,     0.00000000679,
      -0.00000008888,     0.00000000119,
      -0.00000000014,    -0.00000000136,
       0.00000000412,     0.00000009615,
      -0.00000000112,    -0.00000000055,
       0.00000360221,     0.00000322756,
       0.00000036758,    -0.00000026171,
      -0.00000001678,    -0.00000001783,
       0.00000028153,    -0.00000028292,
      -0.00000002266,    -0.00000003028,
      -0.00000000154,     0.00000000140,
       0.00000080708,    -0.00000002188,
      -0.00000000104,     0.00000001183,
      -0.00000000004,    -0.00000000137,
      -0.00000003226,    -0.00000083506,
       0.00000001083,     0.00000000313,
      -0.00000000114,     0.00000000009,
      -0.00000000139,     0.00000000075,
      -0.00000000049,     0.00000000135,
      -0.00000001079,    -0.00000000001,
      -0.00000000010,     0.00000000219,
       0.00000010147,    -0.00000000443,
       0.00000000025,    -0.00000000943,
       0.00000000143,     0.00000000100,
       0.00000000066,    -0.00000000385,
       0.00000000127,     0.00000000114,
      -0.00000000040,    -0.00000000152,
      -0.00000000147,     0.00000000034,
      -0.00000002207,    -0.00000002556,
      -0.00000000530,     0.00000000409,
       0.00000000159,     0.00000000640,
       0.00000000638,    -0.00000000140,
       0.00000032968,     0.00000010304,
       0.00000000879,    -0.00000002894,
       0.00000000076,     0.00000000034,
       0.00000000207,     0.00000000191,
       0.00000000063,     0.00000000099,
       0.00000000023,     0.00000000088,
      -0.00001525836,     0.00000795306,
       0.00000097146,    -0.00000039098,
      -0.00000004231,    -0.00000004570,
       0.00000195313,     0.00000325772,
      -0.00000009972,    -0.00000023429,
      -0.00000001146,     0.00000001032,
       0.00000000041,     0.00000000077,
       0.00000305217,    -0.00000008555,
       0.00000001151,     0.00000001246,
       0.00000000046,    -0.00000000373,
       0.00000006126,    -0.00000303984,
      -0.00000001024,    -0.00000002165,
      -0.00000000315,     0.00000000180,
      -0.00000001113,    -0.00000000832,
      -0.00000000087,    -0.00000001470,
       0.00000000086,     0.00000000167,
       0.00018954553,    -0.00009916653,
      -0.00001202864,     0.00000488280,
       0.00000052659,     0.00000057083,
       0.00000003827,    -0.00000002127,
      -0.00000865820,    -0.00001437562,
       0.00000044123,     0.00000103795,
       0.00000005074,    -0.00000004544,
      -0.00000000180,    -0.00000000334,
      -0.00002711219,     0.00000030863,
      -0.00000009994,    -0.00000004389,
       0.00000000107,     0.00000003166,
      -0.00000023146,     0.00002701059,
       0.00000003446,     0.00000013997,
       0.00000003005,    -0.00000000893,
       0.00000001279,     0.00000006687,
       0.00000000310,    -0.00000000465,
       0.00000005259,     0.00000003133,
      -0.00000000561,    -0.00000000763,
      -0.00000000051,     0.00000000167,
      -0.00000000138,    -0.00000000130,
      -0.00000002098,     0.00000000017,
      -0.00000002173,     0.00000000018,
       0.00000000057,    -0.00000000095,
      -0.00000000103,    -0.00000000053,
       0.00000000224,    -0.00000003617,
      -0.00000000287,     0.00000000011,
      -0.00000000473,     0.00000000795,
       0.00000000096,    -0.00000000088,
       0.00000000789,     0.00000000477,
      -0.00000000094,    -0.00000000095,
       0.00000008468,    -0.00000010548,
      -0.00000001391,     0.00000001136,
      -0.00000002951,    -0.00000001877,
       0.00000000354,     0.00000000371,
      -0.00000004547,     0.00000002067,
       0.00000000341,    -0.00000000039,
       0.00000001975,     0.00000004602,
      -0.00000000015,    -0.00000000347,
       0.00000000047,     0.00000000056,
      -0.00000130478,     0.00000165326,
       0.00000021998,    -0.00000017973,
      -0.00000002786,    -0.00000000376,
       0.00000013554,     0.00000008667,
      -0.00000001605,    -0.00000001693,
      -0.00000000019,     0.00000000234,
       0.00000039556,    -0.00000017917,
      -0.00000003024,     0.00000000303,
       0.00000000133,     0.00000000171,
      -0.00000017526,    -0.00000039716,
       0.00000000197,     0.00000003026,
       0.00000000183,    -0.00000000126,
      -0.00000000207,    -0.00000000303,
      -0.00000000310,     0.00000000162,
      -0.00000000038,     0.00000000063,
       0.00000000068,     0.00000000039,
      -0.00000002257,     0.00000001854,
       0.00000000150,     0.00000000158,
       0.00000000426,    -0.00000000139,
      -0.00000000136,    -0.00000000426,
      -0.00000004079,     0.00000006868,
      -0.00000000193,    -0.00000000114,
      -0.00000002176,     0.00000000010,
      -0.00000000002,     0.00000000446,
       0.00000022715,    -0.00000000121,
       0.00000000001,    -0.00000001900,
       0.00000005097,     0.00000001832,
      -0.00000000048,     0.00000000163,
       0.00000000099,     0.00000000192,
      -0.00000000007,    -0.00000000212,
      -0.00000000202,    -0.00000000039,
      -0.00000120668,     0.00000002538,
       0.00000000538,    -0.00000001404,
       0.00000000533,     0.00000023251,
      -0.00000000330,    -0.00000000086,
      -0.00000001049,    -0.00000000203,
      -0.00000000199,    -0.00000000022,
       0.00000001006,    -0.00000000207,
      -0.00000000038,     0.00000000089,
      -0.00000000039,    -0.00000000073,
       0.00001137089,    -0.00000032613,
      -0.00000004138,     0.00000015425,
      -0.00000003083,    -0.00000099435,
       0.00000001370,     0.00000000350,
      -0.00000000153,    -0.00000000002,
       0.00000005348,     0.00000005229,
      -0.00000000288,    -0.00000000286,
       0.00000001622,    -0.00000000350,
      -0.00000000006,     0.00000000661,
      -0.00000000127,     0.00000000342,
       0.00000000246,     0.00000000289,
       0.00000000094,     0.00000000051,
      -0.00000000051,     0.00000000094,
       0.00000007950,     0.00000002331,
      -0.00000000082,     0.00000000346,
      -0.00000000082,    -0.00000000024,
       0.00000000022,    -0.00000000082,
       0.00000000763,     0.00000001726,
       0.00000000141,    -0.00000000018,
       0.00000000188,     0.00000000059,
      -0.00000000027,     0.00000000109,
      -0.00000000010,     0.00000000004,
      -0.00000000004,    -0.00000000010,
       0.00000000025,     0.00000000035,
      -0.00000000035,     0.00000000025,
       0.00000001545,     0.00000003735,
      -0.00000000542,    -0.00000000551,
       0.00000002670,    -0.00000001200,
      -0.00000000336,     0.00000000404,
      -0.00000003486,     0.00000000019,
       0.00000000532,    -0.00000000603,
      -0.00000001107,     0.00000002243,
      -0.00000000536,    -0.00000000433,
       0.00000000102,    -0.00000000053,
       0.00005415478,    -0.00000052784,
       0.00000023858,     0.00000005939,
      -0.00000001307,    -0.00000005982,
      -0.00000008463,    -0.00001054303,
       0.00000000997,    -0.00000004555,
      -0.00000001149,     0.00000000256,
       0.00000002834,     0.00000047375,
      -0.00000001573,    -0.00000000271,
      -0.00000000204,    -0.00000000719,
       0.00000000947,    -0.00000049608,
       0.00000002379,     0.00000007077,
      -0.00000000407,    -0.00000000047,
       0.00000001525,     0.00000003745,
      -0.00000000135,     0.00000000164,
      -0.00000002002,     0.00000003042,
      -0.00000000073,    -0.00000000263,
      -0.00050468036,     0.00000406292,
      -0.00000203161,    -0.00000036940,
       0.00000008569,     0.00000055983,
       0.00000008907,     0.00004429219,
      -0.00000000842,     0.00000016371,
       0.00000004981,    -0.00000000633,
      -0.00000013427,    -0.00000232507,
       0.00000006777,     0.00000014919,
      -0.00000000756,     0.00000001420,
      -0.00000078627,     0.00000193899,
       0.00000006417,    -0.00000023890,
       0.00000000515,     0.00000001509,
      -0.00000000055,     0.00000000048,
       0.00000006442,    -0.00000017378,
      -0.00000000536,    -0.00000000775,
      -0.00000008239,    -0.00000013983,
      -0.00000000656,     0.00000001209,
       0.00000000035,     0.00000000080,
       0.00000003230,    -0.00000002088,
      -0.00000000173,     0.00000000099,
       0.00000002074,     0.00000003221,
      -0.00000000098,    -0.00000000173,
       0.00000000421,    -0.00000000706,
      -0.00000000192,    -0.00000000087,
      -0.00000000255,     0.00000000159,
       0.00000000155,     0.00000000260,
      -0.00000006560,     0.00000010917,
       0.00000001371,    -0.00000001152,
       0.00000000875,     0.00000000421,
      -0.00000000103,    -0.00000000110,
       0.00000002216,    -0.00000001356,
      -0.00000000201,     0.00000000006,
      -0.00000001329,    -0.00000002233,
       0.00000000001,     0.00000000202,
       0.00000000149,    -0.00000000091,
      -0.00000000095,    -0.00000000150,
      -0.00000000025,    -0.00000000163,
      -0.00000000163,     0.00000000025,
      -0.00000047566,     0.00000021429,
       0.00000003591,    -0.00000000376,
       0.00000005344,     0.00000010254,
      -0.00000000114,    -0.00000000869,
       0.00000009083,     0.00000000453,
       0.00000000008,    -0.00000000111,
       0.00000000920,    -0.00000009025,
      -0.00000000190,    -0.00000000048,
       0.00000590822,    -0.00000261084,
      -0.00000044253,     0.00000003908,
       0.00000001928,     0.00000002527,
      -0.00000023270,    -0.00000045319,
       0.00000000482,     0.00000003818,
       0.00000000226,    -0.00000000163,
      -0.00000080611,    -0.00000005350,
      -0.00000000082,     0.00000001235,
      -0.00000006979,     0.00000080199,
       0.00000001465,     0.00000000248,
       0.00000000080,    -0.00000000035,
       0.00000000015,     0.00000000194,
       0.00000000139,     0.00000000120,
      -0.00000000551,    -0.00000005590,
      -0.00000000222,     0.00000000050,
       0.00000001274,    -0.00000000818,
      -0.00000000070,     0.00000000040,
      -0.00000000825,    -0.00000001276,
       0.00000000041,     0.00000000071,
       0.00000000038,     0.00000000110,
       0.00000006151,    -0.00000001980,
      -0.00000000172,    -0.00000000479,
      -0.00000000532,     0.00000000017,
      -0.00000000001,     0.00000000522,
       0.00000143116,    -0.00000092727,
      -0.00000007702,     0.00000004413,
       0.00000002645,     0.00000004095,
      -0.00000000126,    -0.00000000220,
      -0.00000000147,     0.00000000010,
       0.00000000010,     0.00000000147,
      -0.00000002491,    -0.00000007639,
       0.00000000926,     0.00000000450,
      -0.00000000912,     0.00000000447,
       0.00000000054,    -0.00000000116,
      -0.00000002491,     0.00000000118,
      -0.00000000141,    -0.00000002511,
       0.00000025307,     0.00000046077,
      -0.00000006873,    -0.00000000967,
       0.00000003840,    -0.00000001819,
      -0.00000000245,     0.00000000477,
       0.00000006040,     0.00000000081,
      -0.00000000226,     0.00000000326,
       0.00000000243,     0.00000003348,
       0.00000000207,     0.00000000184,
      -0.00000000088,     0.00000000036,
      -0.00000000038,    -0.00000000073,
      -0.00000009415,     0.00000000254,
      -0.00000000011,    -0.00000000259,
       0.00000001707,     0.00000000533,
      -0.00000000337,     0.00000000556,
       0.00000000498,     0.00000000355,
      -0.00000000387,    -0.00000000052,
       0.00000000079,    -0.00000000376,
       0.00000144162,     0.00000518079,
      -0.00000051852,    -0.00000034067,
       0.00000003515,    -0.00000003341,
       0.00000061652,    -0.00000027425,
      -0.00000003923,     0.00000006216,
      -0.00000000416,    -0.00000000415,
       0.00000170799,     0.00000002466,
       0.00000000391,    -0.00000000322,
       0.00000000016,    -0.00000000181,
      -0.00000001293,     0.00000171948,
      -0.00000000073,     0.00000000536,
       0.00000000209,    -0.00000000027,
      -0.00000000148,    -0.00000000611,
      -0.00000000074,     0.00000000067,
       0.00000000645,    -0.00000000260,
      -0.00000000070,    -0.00000000064,
      -0.00000608118,    -0.00002962126,
       0.00000251191,     0.00000241523,
      -0.00000022963,     0.00000011879,
      -0.00000267993,     0.00000118656,
       0.00000016680,    -0.00000027702,
       0.00000001867,     0.00000001728,
      -0.00000000113,     0.00000000059,
      -0.00000402310,    -0.00000015005,
       0.00000012504,    -0.00000015230,
       0.00000001309,     0.00000002598,
      -0.00000000227,     0.00000000053,
      -0.00000018706,    -0.00000230437,
      -0.00000014917,    -0.00000015193,
       0.00000001901,    -0.00000001213,
       0.00000000050,     0.00000000234,
       0.00000001078,     0.00000003708,
       0.00000000221,    -0.00000000442,
      -0.00000003473,     0.00000000156,
       0.00000000400,     0.00000000304,
      -0.00000010814,    -0.00000000577,
      -0.00000000034,     0.00000000415,
       0.00000002211,     0.00000000035,
      -0.00000000136,     0.00000002644,
       0.00000000143,    -0.00000000101,
      -0.00000000105,    -0.00000000144,
      -0.00000002491,     0.00000001548,
       0.00000000373,     0.00000000537,
       0.00000000500,    -0.00000000040,
      -0.00000000015,    -0.00000000503,
       0.00000030961,    -0.00000019095,
      -0.00000002758,     0.00000000067,
      -0.00000001643,    -0.00000002364,
       0.00000000012,     0.00000000236,
      -0.00000004461,     0.00000000283,
       0.00000000008,     0.00000000141,
       0.00000000182,     0.00000004449,
       0.00000000161,     0.00000000002,
       0.00000002077,    -0.00000001386,
      -0.00000000111,    -0.00000000152,
      -0.00000000324,     0.00000000034,
       0.00000000032,     0.00000000325,
       0.00000002287,    -0.00000001242,
       0.00000000038,     0.00000000072,
       0.00000122040,     0.00000008034,
       0.00000000190,    -0.00000001903,
       0.00000001762,    -0.00000023862,
      -0.00000000397,    -0.00000000018,
       0.00000000585,     0.00000001040,
       0.00000000046,    -0.00000000067,
       0.00000000029,    -0.00000001749,
      -0.00000000013,     0.00000000127,
       0.00000000030,     0.00000000090,
      -0.00000000052,     0.00000000067,
      -0.00001133103,    -0.00000074894,
      -0.00000001230,     0.00000018044,
      -0.00000007541,     0.00000100102,
       0.00000001671,     0.00000000123,
       0.00000000102,    -0.00000000012,
      -0.00000005851,    -0.00000005202,
       0.00000000015,     0.00000000365,
      -0.00000001973,     0.00000010183,
       0.00000000251,    -0.00000000494,
       0.00000000171,    -0.00000000390,
      -0.00000000163,    -0.00000000340,
       0.00000000988,    -0.00000000050,
       0.00000000027,    -0.00000000849,
      -0.00000000002,    -0.00000000143,
      -0.00000006535,     0.00000000301,
       0.00000000013,     0.00000000582,
       0.00000106350,     0.00000002746,
      -0.00000000170,     0.00000003043,
      -0.00000000329,    -0.00000000051,
      -0.00000000052,     0.00000000332,
       0.00000000022,    -0.00000000090,
      -0.00000004539,    -0.00000001368,
      -0.00000000067,     0.00000000239,
       0.00000045613,     0.00000003762,
      -0.00000006720,     0.00000009324,
      -0.00000000870,    -0.00000001305,
      -0.00000000045,     0.00000001969,
      -0.00000000410,    -0.00000000263,
       0.00000000144,     0.00000000139,
      -0.00000000141,     0.00000000062,
      -0.00000000047,     0.00000000070,
      -0.00000002547,    -0.00000000708,
      -0.00000000019,    -0.00000000082,
       0.00000003628,     0.00000011659,
      -0.00000003130,    -0.00000000358,
       0.00000001278,    -0.00000000509,
      -0.00000000048,     0.00000000413,
       0.00000004503,    -0.00000002344,
       0.00000000152,     0.00000000153,
       0.00000002384,     0.00000004580,
      -0.00000000170,     0.00000000142,
      -0.00000001632,     0.00000002109,
       0.00000000142,     0.00000000107,
      -0.00000002112,    -0.00000001623,
      -0.00000000107,     0.00000000141,
      -0.00005886240,    -0.00000800081,
       0.00000892421,    -0.00001074835,
       0.00000088693,     0.00000155472,
      -0.00000015907,     0.00000003421,
       0.00000028210,    -0.00000254836,
       0.00000047342,     0.00000036742,
      -0.00000006745,     0.00000003977,
      -0.00000000171,    -0.00000000685,
      -0.00000007621,    -0.00000018789,
       0.00000002253,     0.00000000333,
      -0.00000000034,     0.00000000220,
       0.00000018520,     0.00000002738,
      -0.00000002605,     0.00000000938,
       0.00000000066,    -0.00000000261,
       0.00000008444,    -0.00000010640,
      -0.00000000657,    -0.00000000249,
       0.00000008751,     0.00000006516,
       0.00000000733,    -0.00000000659,
      -0.00000009407,    -0.00000007494,
       0.00000002131,    -0.00000001485,
      -0.00000000452,    -0.00000000098,
      -0.00000001341,    -0.00000000225,
       0.00000000160,    -0.00000000199,
       0.00000000133,     0.00000000665,
      -0.00000000217,    -0.00000000122,
       0.00000001915,    -0.00000001369,
      -0.00000000122,    -0.00000000153,
      -0.00000000283,     0.00000000034,
       0.00000000026,     0.00000000281,
       0.00000006339,    -0.00000000512,
      -0.00000000088,    -0.00000001249,
       0.00000000065,     0.00000000057,
      -0.00000000013,    -0.00000000121,
      -0.00000058829,     0.00000004794,
       0.00000000111,     0.00000001916,
       0.00000000361,     0.00000005223,
       0.00000000175,    -0.00000000006,
      -0.00000000629,    -0.00000000236,
      -0.00000000033,     0.00000000828,
      -0.00000000025,    -0.00000000139,
      -0.00000000135,     0.00000000029,
       0.00000000132,     0.00000000051,
       0.00000000052,    -0.00000000131,
      -0.00000002907,     0.00000002600,
      -0.00000000063,     0.00000000400,
      -0.00000000210,    -0.00000004436,
      -0.00000000193,     0.00000000061,
      -0.00000001216,    -0.00000000478,
      -0.00000000062,     0.00000000056,
      -0.00000000477,     0.00000001219,
       0.00000000057,     0.00000000062,
       0.00000013574,     0.00000008904,
      -0.00000000457,    -0.00000000728,
       0.00000001171,    -0.00000002374,
      -0.00000000078,     0.00000000071,
       0.00000003014,     0.00000000293,
      -0.00000000194,     0.00000002017,
      -0.00000058477,    -0.00000044686,
       0.00000002531,     0.00000004145,
      -0.00000004069,     0.00000008238,
       0.00000000319,    -0.00000000309,
      -0.00000006227,    -0.00000001043,
       0.00000000248,    -0.00000000094,
      -0.00000000333,    -0.00000003369,
      -0.00000000251,    -0.00000000266,
      -0.00000040984,    -0.00000012954,
      -0.00000001699,     0.00000001516,
      -0.00000000634,     0.00000001583,
       0.00000000070,     0.00000000077,
      -0.00000004529,     0.00000000075,
       0.00000000113,     0.00000004599,
       0.00000011561,    -0.00000000259,
      -0.00000000074,    -0.00000001980,
       0.00000000119,     0.00000000110,
      -0.00000000002,     0.00000000125,
      -0.00000112126,    -0.00000004719,
       0.00000000439,     0.00000005540,
      -0.00000000095,    -0.00000000005,
      -0.00000000235,    -0.00000000476,
      -0.00000000352,    -0.00000000054,
      -0.00000058225,    -0.00000004172,
       0.00000008459,    -0.00000011960,
       0.00000000084,    -0.00000001738,
       0.00000000358,     0.00000000243,
      -0.00000001235,    -0.00000002039,
      -0.00000001403,     0.00000000104,
       0.00000000190,    -0.00000000312,
      -0.00000000099,    -0.00000001415,
       0.00000000314,     0.00000000194,
       0.00000000755,    -0.00000000813,
       0.00000000005,     0.00000000192,
       0.00000000812,     0.00000000756,
      -0.00000000192,     0.00000000005,
       0.00000000267,     0.00000000541,
       0.00000000088,    -0.00000000060,
       0.00000000541,    -0.00000000268,
      -0.00000000060,    -0.00000000088,
      -0.00000000482,    -0.00000001102,
      -0.00000000241,     0.00000000101,
      -0.00000000523,    -0.00000000036,
      -0.00000000031,     0.00000000531,
       0.00000007449,     0.00000014569,
       0.00000002408,    -0.00000001635,
       0.00000000723,    -0.00000000299,
      -0.00000000075,    -0.00000000113,
       0.00000004908,     0.00000000311,
       0.00000000302,    -0.00000004919,
      -0.00000012545,    -0.00000000786,
      -0.00000000134,     0.00000002155,
      -0.00000000028,    -0.00000000121,
       0.00000123042,     0.00000007560,
       0.00000000379,    -0.00000006039,
      -0.00000000097,     0.00000000014,
       0.00000000147,     0.00000000498,
       0.00000000357,    -0.00000000105,
      -0.00000000226,    -0.00000001445,
      -0.00000000179,     0.00000000052,
      -0.00000000487,    -0.00000000031,
       0.00000000027,    -0.00000000492,
      -0.00000002935,    -0.00000000239,
       0.00000001352,     0.00000007057,
       0.00000000498,    -0.00000000146,
       0.00000001392,     0.00000000201,
      -0.00000000159,     0.00000000152,
       0.00000000150,    -0.00000000575,
       0.00000000165,     0.00000000157,
       0.00000053026,     0.00000010383,
      -0.00000008760,     0.00000008450,
      -0.00000000277,     0.00000001697,
      -0.00000000273,    -0.00000000270,
       0.00000000086,     0.00000000014,
       0.00000000134,    -0.00000000043,
       0.00000000059,    -0.00000000696,
       0.00000000128,     0.00000000103,
      -0.00000000024,     0.00000000243,
       0.00000000159,    -0.00000000033,
      -0.00000000033,    -0.00000000159,
      -0.00000000017,    -0.00000000088,
      -0.00000000100,     0.00000000146,
       0.00000000189,     0.00000000058,
       0.00000004296,    -0.00000000787,
      -0.00000000038,    -0.00000000212,
      -0.00000000083,     0.00000000136,
       0.00000000135,     0.00000000083,
      -0.00000003589,     0.00000000499,
       0.00000000041,     0.00000000322,
       0.00000000023,    -0.00000000113,
       0.00000000748,    -0.00000001214,
      -0.00000000416,    -0.00000000182,
      -0.00000001216,    -0.00000000748,
      -0.00000000182,     0.00000000417,
      -0.00000001427,     0.00000002545,
       0.00000000872,     0.00000000369,
       0.00000000583,     0.00000000449,
       0.00000000102,    -0.00000000203,
       0.00000001233,     0.00000002143,
       0.00000000320,    -0.00000000234,
       0.00000002150,    -0.00000001251,
      -0.00000000242,    -0.00000000323,
       0.00000024494,    -0.00000042022,
      -0.00000014324,    -0.00000006048,
      -0.00000001811,    -0.00000001396,
      -0.00000000320,     0.00000000633,
       0.00000000099,     0.00000000035,
      -0.00000011362,    -0.00000019850,
      -0.00000002963,     0.00000002186,
       0.00000000148,     0.00000000194,
      -0.00000019862,     0.00000011396,
       0.00000002202,     0.00000002965,
       0.00000000195,    -0.00000000151,
       0.00000000022,    -0.00000000092,
      -0.00000000093,    -0.00000000021,
       0.00000031060,     0.00000051206,
       0.00000007655,    -0.00000005729,
       0.00000011116,    -0.00000005610,
      -0.00000001149,    -0.00000001560,
      -0.00000000101,     0.00000000077,
       0.00000024811,     0.00000000221,
       0.00000000099,     0.00000000037,
      -0.00000000066,    -0.00000024883,
      -0.00000005770,    -0.00000001841,
      -0.00000000143,     0.00000000699,
      -0.00000000335,    -0.00000000020,
      -0.00000000063,    -0.00000000139,
      -0.00000395653,    -0.00000669160,
      -0.00000101215,     0.00000074822,
       0.00000005090,     0.00000006625,
      -0.00000033115,     0.00000016131,
       0.00000003418,     0.00000004649,
       0.00000000303,    -0.00000000232,
      -0.00000232777,    -0.00000000770,
      -0.00000000401,    -0.00000000056,
       0.00000000003,     0.00000000274,
      -0.00000000313,     0.00000233239,
       0.00000000043,     0.00000000237,
       0.00000000249,    -0.00000000013,
       0.00000000062,    -0.00000000039,
       0.00000789748,     0.00000002351,
       0.00000001127,     0.00000000040,
       0.00000000223,    -0.00000135386,
      -0.00000000001,    -0.00000000182,
      -0.00000000151,     0.00000000004,
       0.00000000898,     0.00000007463,
       0.00000000111,    -0.00000000607,
      -0.00000000342,     0.00000001717,
       0.00000000083,     0.00000000847,
      -0.00000000034,     0.00000000225,
      -0.00000000171,    -0.00000000043,
      -0.00007782765,    -0.00000019313,
      -0.00000012066,    -0.00000000385,
       0.00000000289,     0.00000008773,
      -0.00000000600,     0.00000377588,
       0.00000000001,     0.00000000559,
       0.00000000424,    -0.00000000013,
       0.00000000860,    -0.00000031776,
       0.00000002266,     0.00000002126,
      -0.00000000130,     0.00000000214,
      -0.00000023311,    -0.00000003983,
       0.00000001320,    -0.00000003375,
       0.00000000143,     0.00000000184,
      -0.00000000097,    -0.00000000695,
      -0.00000000539,     0.00000000133,
      -0.00000000037,     0.00000000067,
       0.00000000069,     0.00000000045,
      -0.00000012810,     0.00000145100,
      -0.00000014394,    -0.00000010119,
       0.00000000970,    -0.00000000853,
       0.00000017078,     0.00000001605,
      -0.00000001200,     0.00000001699,
      -0.00000000102,    -0.00000000117,
       0.00000046350,    -0.00000000091,
       0.00000000136,    -0.00000000040,
       0.00000000012,     0.00000046757,
      -0.00000000031,     0.00000000071,
      -0.00000043596,    -0.00000011405,
       0.00000007408,    -0.00000006632,
      -0.00000000199,    -0.00000001756,
       0.00000000343,     0.00000000247,
      -0.00000000416,    -0.00000000187,
       0.00000000110,     0.00000000158,
      -0.00000000069,    -0.00000000033,
       0.00000000075,    -0.00000000090,
       0.00000000078,     0.00000000056,
      -0.00000000937,    -0.00000656725,
       0.00000068908,     0.00000028041,
      -0.00000002018,     0.00000004876,
      -0.00000048329,    -0.00000005768,
       0.00000003565,    -0.00000004786,
       0.00000000274,     0.00000000347,
      -0.00000133807,    -0.00000012221,
       0.00000014318,    -0.00000017617,
       0.00000001515,     0.00000002553,
      -0.00000000262,     0.00000000063,
      -0.00000012111,     0.00000058275,
      -0.00000017618,    -0.00000014466,
       0.00000002468,    -0.00000001511,
       0.00000000063,     0.00000000262,
       0.00000000124,    -0.00000000036,
       0.00000000808,    -0.00000000159,
      -0.00000000027,    -0.00000000103,
      -0.00000000062,    -0.00000000497,
       0.00000000498,    -0.00000000057,
       0.00000000032,     0.00000000174,
      -0.00000000173,     0.00000000032,
      -0.00010946495,    -0.00001401724,
       0.00001643012,    -0.00002007375,
       0.00000166160,     0.00000288769,
      -0.00000029542,     0.00000006473,
       0.00000042622,    -0.00000331690,
       0.00000060799,     0.00000049805,
      -0.00000008756,     0.00000005038,
      -0.00000000216,    -0.00000000893,
       0.00000000616,    -0.00000000873,
      -0.00000000095,     0.00000000089,
       0.00000002905,    -0.00000000856,
      -0.00000000027,     0.00000000488,
      -0.00000000164,    -0.00000001325,
      -0.00000000251,    -0.00000000036,
       0.00000000063,    -0.00000000146,
      -0.00000003911,    -0.00000000471,
      -0.00000000011,     0.00000000100,
       0.00000000220,     0.00000000304,
      -0.00000000304,     0.00000000220,
      -0.00000000056,    -0.00000000699,
       0.00000000084,     0.00000000022,
       0.00000000699,    -0.00000000056,
      -0.00000000022,     0.00000000084,
       0.00000006286,     0.00000020881,
       0.00000000762,    -0.00000000235,
       0.00000000034,     0.00000000083,
      -0.00000000090,     0.00000000024,
      -0.00000002657,    -0.00000001555,
      -0.00000000189,    -0.00000000282,
       0.00000000165,     0.00000000569,
      -0.00000000565,     0.00000000177,
      -0.00000000650,    -0.00000000680,
       0.00000000637,    -0.00000000175,
       0.00000000173,     0.00000000740,
       0.00000000065,     0.00000000222,
      -0.00000000081,    -0.00000000292,
       0.00000000090,     0.00000000017,
      -0.00000000003,     0.00000000006,
      -0.00000000006,    -0.00000000967,
       0.00000000234,     0.00000000119,
      -0.00000000016,     0.00000000019,
       0.00000000969,    -0.00000000004,
      -0.00000000119,     0.00000000234,
      -0.00000000019,    -0.00000000016,
      -0.00000000035,    -0.00000000031,
       0.00000000009,    -0.00000000001,
       0.00000000038,    -0.00000000045,
       0.00000000003,     0.00000000011,
       0.00000000072,    -0.00000000046,
       0.00000000045,     0.00000000072,
      -0.00000000042,    -0.00000000030,
       0.00000000030,    -0.00000000043,
       0.00000000022,    -0.00000000037,
       0.00000000127,    -0.00000000294,
       0.00000000061,     0.00000000073,
       0.00000000294,     0.00000000127,
      -0.00000000073,     0.00000000062,
      -0.00000000049,     0.00000000087,
      -0.00000000087,    -0.00000000049,
       0.00000000411,     0.00000000675,
      -0.00000000276,     0.00000000102,
      -0.00000000012,    -0.00000000045,
      -0.00000005203,     0.00000000121,
       0.00000000436,    -0.00000001534,
       0.00000000173,     0.00000000081,
      -0.00000000125,    -0.00000005201,
       0.00000001535,     0.00000000435,
      -0.00000000080,     0.00000000173,
      -0.00000000258,    -0.00000000026,
       0.00000000027,    -0.00000000031,
       0.00000000036,    -0.00000000252,
       0.00000000028,     0.00000000030,
      -0.00000012766,    -0.00000006248,
       0.00000003282,    -0.00000005454,
       0.00000001225,     0.00000000864,
      -0.00000001317,     0.00000008401,
      -0.00000003768,    -0.00000000766,
       0.00000000200,    -0.00000000924,
       0.00000000157,     0.00000000041,
      -0.00000021759,    -0.00000014918,
       0.00000006323,    -0.00000006887,
       0.00000001232,     0.00000001440,
      -0.00000000241,     0.00000000138,
       0.00000011134,    -0.00000020835,
       0.00000006674,     0.00000003808,
      -0.00000000592,     0.00000001279,
      -0.00000000175,    -0.00000000058,
      -0.00000002158,    -0.00000000054,
       0.00000000095,    -0.00000000279,
       0.00000000127,    -0.00000002117,
       0.00000000262,     0.00000000115,
       0.00001713353,     0.00000764877,
      -0.00000361817,     0.00000729690,
      -0.00000162448,    -0.00000087156,
       0.00000014029,    -0.00000024576,
       0.00000002863,     0.00000001718,
      -0.00000057041,     0.00000111678,
      -0.00000046717,    -0.00000026462,
       0.00000006263,    -0.00000010132,
       0.00000001462,     0.00000000971,
      -0.00000000112,     0.00000000158,
       0.00000037751,     0.00000018789,
      -0.00000009647,     0.00000013198,
      -0.00000002653,    -0.00000002709,
       0.00000000540,    -0.00000000305,
      -0.00000010430,     0.00000030042,
      -0.00000009622,    -0.00000002644,
       0.00000000030,    -0.00000001995,
       0.00000000279,    -0.00000000105,
       0.00000005772,    -0.00000000111,
      -0.00000000161,     0.00000000836,
      -0.00000000600,     0.00000004745,
      -0.00000000500,    -0.00000000354,
       0.00000006084,     0.00000003551,
      -0.00000001646,     0.00000002460,
       0.00000001087,    -0.00000000064,
      -0.00000000197,     0.00000000291,
       0.00000001415,    -0.00000014557,
       0.00000003982,     0.00000002882,
      -0.00000000754,     0.00000000440,
      -0.00000000009,    -0.00000000116,
       0.00000014464,     0.00000001381,
      -0.00000002856,     0.00000003940,
      -0.00000000432,    -0.00000000748,
       0.00000000115,    -0.00000000009,
       0.00000013692,     0.00000005917,
      -0.00000002826,     0.00000005820,
      -0.00000000130,    -0.00000000188,
       0.00000001718,    -0.00000018222,
       0.00000004965,     0.00000003597,
      -0.00000000942,     0.00000000546,
      -0.00000000011,    -0.00000000145,
       0.00000018216,     0.00000001716,
      -0.00000003594,     0.00000004963,
      -0.00000000546,    -0.00000000940,
       0.00000000144,    -0.00000000011,
      -0.00000000051,     0.00000000409,
      -0.00000000064,    -0.00000000049,
       0.00000000037,    -0.00000000288,
       0.00000000043,     0.00000000032,
      -0.00000000110,     0.00000000194,
      -0.00000000034,    -0.00000000039,
       0.00000001009,    -0.00000001565,
       0.00000000227,     0.00000000288,
       0.00000000113,    -0.00000000006,
       0.00000000379,     0.00000008423,
       0.00000000928,     0.00000000301,
       0.00000000005,     0.00000000302,
       0.00000000305,     0.00000000009,
       0.00000000037,    -0.00000000027,
      -0.00000000028,    -0.00000000039,
      -0.00000000045,    -0.00000000028,
       0.00000000038,     0.00000000032,
       0.00000000034,    -0.00000000040,
       0.00000000049,     0.00000000031,
       0.00000000037,    -0.00000000055,
      -0.00000000017,     0.00000000035,
       0.00000000036,     0.00000000020,
       0.00000000037,     0.00000000018,
       0.00000000052,    -0.00000000009,
      -0.00000000016,    -0.00000000070,
      -0.00000000013,    -0.00000000081,
       0.00000000029,    -0.00000000031,
      -0.00000000084,    -0.00000000034,
      -0.00000000262,     0.00000000673,
       0.00000000050,     0.00000000026,
       0.00000000281,    -0.00000000175,
      -0.00000000178,    -0.00000000278,
       0.00000000049,    -0.00000000045,
      -0.00000000080,    -0.00000000085,
       0.00000000144,     0.00000000076,
       0.00000000076,    -0.00000000153,
       0.00000000159,     0.00000000080,
       0.00000000086,    -0.00000000188,
      -0.00000000040,     0.00000000072,
       0.00000000075,     0.00000000044,
      -0.00000000021,     0.00000000096,
       0.00000000098,     0.00000000045,
       0.00000000047,     0.00000000019,
       0.00000000001,     0.00000000006,
       0.00000000017,     0.00000000038,
       0.00000000038,    -0.00000000017,
       0.00000000022,    -0.00000000040,
       0.00000000063,     0.00000000090,
       0.00000000090,    -0.00000000063,
       0.00000000032,     0.00000000046,
       0.00000000089,    -0.00000000061,
      -0.00000000030,     0.00000000047,
       0.00000000047,     0.00000000028,
      -0.00000000024,     0.00000000039,
       0.00000000061,     0.00000000038,
      -0.00000000120,    -0.00000000210,
      -0.00000000927,     0.00000000544,
       0.00000000052,     0.00000000020,
       0.00000000927,     0.00000000360,
       0.00000000347,    -0.00000000927,
       0.00000000007,     0.00000000044,
       0.00000000437,     0.00000000172,
       0.00000000227,    -0.00000000666,
      -0.00000000163,     0.00000000235,
       0.00000000290,     0.00000000097,
      -0.00000000118,     0.00000000314,
      -0.00000000037,    -0.00000000011,
       0.00000000347,     0.00000000143,
      -0.00000000018,     0.00000000037,
       0.00000000068,     0.00000000075,
       0.00000000039,     0.00000000014,
       0.00000000033,     0.00000000102,
       0.00000000091,    -0.00000000027,
      -0.00000000023,     0.00000000043,
      -0.00000000041,    -0.00000000024,
      -0.00000000001,    -0.00000000001,
      -0.00000000012,     0.00000000014,
      -0.00000000006,     0.00000000012,
       0.00000000012,     0.00000000006,
      -0.00000000002,     0.00000000003,
       0.00000000013,     0.00000000007,
      -0.00000000051,     0.00000000084,
       0.00000000084,     0.00000000051,
      -0.00000000022,     0.00000000036,
       0.00000000093,     0.00000000056,
       0.00000000007,     0.00000000046,
      -0.00000000139,     0.00000000291,
      -0.00000000033,    -0.00000000017,
       0.00000000292,     0.00000000139,
      -0.00000000017,     0.00000000033,
      -0.00000000072,     0.00000000161,
      -0.00000000018,    -0.00000000010,
       0.00000000305,     0.00000000145,
      -0.00000000018,     0.00000000035,
       0.00000000430,    -0.00000000275,
      -0.00000000867,    -0.00000001496,
       0.00000000033,    -0.00000000034,
      -0.00000000735,     0.00000001496,
      -0.00000000161,    -0.00000000079,
       0.00000001473,     0.00000000704,
      -0.00000000076,     0.00000000160,
      -0.00000000458,     0.00000000947,
      -0.00000000103,    -0.00000000052,
       0.00000001266,     0.00000000646,
      -0.00000000070,     0.00000000136,
       0.00000000111,     0.00000000188,
      -0.00000000050,     0.00000000017,
       0.00000000269,    -0.00000000081,
       0.00000000015,     0.00000000055,
       0.00000000113,     0.00000000319,
      -0.00000000072,     0.00000000019,
       0.00000000331,    -0.00000000120,
       0.00000000018,     0.00000000072,
      -0.00000000037,     0.00000000085,
      -0.00000000082,    -0.00000000052,
       0.00000000003,     0.00000000009,
       0.00000000003,     0.00000000007,
       0.00000000007,    -0.00000000003,
       0.00000000008,    -0.00000000004,
       0.00000000016,     0.00000000041,
       0.00000000041,    -0.00000000016,
      -0.00000000199,     0.00000000378,
       0.00000000378,    -0.00000000600,
       0.00000000046,    -0.00000000018,
       0.00000000089,     0.00000000053,
       0.00000000059,     0.00000000183,
      -0.00000000043,     0.00000000014,
       0.00000000198,    -0.00000000110,
       0.00000000013,     0.00000000044,
      -0.00000000221,     0.00000000422,
       0.00000000402,    -0.00000000638,
       0.00000000038,     0.00000000092,
      -0.00000000026,     0.00000000008,
       0.00000000168,    -0.00000000068,
       0.00000000014,     0.00000000048,
      -0.00000000087,    -0.00000000118,
      -0.00000000191,     0.00000000304,
      -0.00000000023,    -0.00000000040,
       0.00000000384,     0.00000000862,
      -0.00000000199,     0.00000000068,
       0.00000000892,    -0.00000000487,
       0.00000000068,     0.00000000204,
       0.00000000329,     0.00000000671,
      -0.00000000152,     0.00000000050,
       0.00000000868,    -0.00000000426,
       0.00000000065,     0.00000000198,
       0.00000000017,    -0.00000000033,
       0.00000000016,     0.00000000041,
      -0.00000000041,     0.00000000015,
       0.00000000025,    -0.00000000041,
       0.00000000040,     0.00000000025,
       0.00000000174,    -0.00000000093,
       0.00000000018,     0.00000000044,
       0.00000000078,    -0.00000000103,
       0.00000000117,     0.00000000058,
       0.00000010915,     0.00000001673,
      -0.00000000870,     0.00000003115,
      -0.00000000169,     0.00000001105,
      -0.00000000317,    -0.00000000088,
       0.00000000148,     0.00000002234,
      -0.00000000209,     0.00000000013,
       0.00000122062,     0.00000018952,
      -0.00000009751,     0.00000034872,
      -0.00000001869,     0.00000012112,
      -0.00000003470,    -0.00000000968,
      -0.00000000074,    -0.00000000603,
       0.00000000166,     0.00000000009,
       0.00000000140,    -0.00000000180,
      -0.00000000055,    -0.00000000048,
      -0.00000000636,     0.00000000526,
      -0.00000000048,    -0.00000000058,
       0.00000000012,    -0.00000000077,
       0.00000000019,     0.00000000043,
      -0.00000000065,     0.00000000015,
       0.00000130481,     0.00000020379,
      -0.00000010460,     0.00000037280,
      -0.00000001969,     0.00000012715,
      -0.00000003646,    -0.00000001015,
      -0.00000000084,    -0.00000000690,
       0.00000000178,     0.00000000009,
       0.00000000150,    -0.00000000192,
      -0.00000000805,     0.00000002010,
      -0.00000000643,    -0.00000000390,
      -0.00000000182,    -0.00000000073,
       0.00000000191,     0.00000000074,
      -0.00000000025,     0.00000000048,
       0.00000000179,    -0.00000000530,
       0.00000000141,     0.00000000065,
       0.00000000046,     0.00000000107,
      -0.00000000203,     0.00000000055,
       0.00000011450,     0.00000001786,
      -0.00000000916,     0.00000003298,
      -0.00000000162,     0.00000001062,
      -0.00000000303,    -0.00000000084,
      -0.00000000045,    -0.00000000010,
      -0.00000000010,     0.00000000044,
      -0.00000000050,    -0.00000000002,
       0.00000001017,    -0.00000000277,
       0.00000000024,     0.00000000090,
      -0.00000000011,     0.00000000049,
       0.00000000122,    -0.00000000035,
      -0.00000000467,    -0.00000001615,
       0.00000000189,    -0.00000000054,
      -0.00000000116,     0.00000000033,
      -0.00000000584,    -0.00000002015,
       0.00000000297,    -0.00000000086,
      -0.00000000180,     0.00000000053,
       0.00000000111,    -0.00000000030,
       0.00000000106,    -0.00000000035,
      -0.00000000104,     0.00000000033,
       0.00000000042,     0.00000000137,
      -0.00000000138,     0.00000000041,
      -0.00000001295,     0.00000000666,
       0.00000000137,    -0.00000000038,
      -0.00000001338,    -0.00000004650,
      -0.00000000163,     0.00000000050,
      -0.00000000460,    -0.00000000640,
       0.00000001881,     0.00000007964,
       0.00000033776,    -0.00000007975,
      -0.00000000760,    -0.00000000750,
      -0.00000003145,    -0.00000012413,
      -0.00000017134,     0.00000004341,
      -0.00000001597,    -0.00000006070,
       0.00000022130,    -0.00000005824,
       0.00000002042,     0.00000009614,
      -0.00000011409,     0.00000002424,
       0.00000000144,    -0.00000000036,
       0.00000001341,     0.00000004637,
      -0.00000000169,     0.00000000045,
       0.00000000660,     0.00000000580,
       0.00000000108,    -0.00000000031,
       0.00000000109,    -0.00000000033,
      -0.00000000105,     0.00000000028,
      -0.00000000040,    -0.00000000138,
      -0.00000000140,     0.00000000041,
       0.00000000564,     0.00000002041,
       0.00000000301,    -0.00000000085,
      -0.00000000182,     0.00000000052,
       0.00000000458,     0.00000001646,
       0.00000000191,    -0.00000000054,
      -0.00000000117,     0.00000000034,
       0.00000000124,    -0.00000000035,
      -0.00000001882,     0.00000000523,
       0.00000000093,     0.00000000359,
      -0.00000000042,    -0.00000000203,
       0.00000000665,     0.00000000132,
      -0.00000000278,    -0.00000000065,
       0.00000000133,    -0.00000000665,
      -0.00000000065,     0.00000000278,
       0.00000000141,     0.00000000128,
       0.00000000125,    -0.00000000148,
       0.00000004759,     0.00000001063,
      -0.00000001910,    -0.00000000704,
       0.00000000045,    -0.00000000463,
      -0.00000000051,     0.00000000195,
      -0.00000001177,    -0.00000001021,
       0.00000000367,     0.00000000396,
      -0.00000000038,    -0.00000000093,
      -0.00000001027,     0.00000001170,
       0.00000000399,    -0.00000000368,
      -0.00000000093,     0.00000000038,
      -0.00000000058,     0.00000000098,
       0.00000000100,     0.00000000057,
       0.00000000827,     0.00000000753,
       0.00000000199,    -0.00000000314,
      -0.00000000085,     0.00000000101,
      -0.00000000286,    -0.00000000773,
       0.00000000055,     0.00000000214,
      -0.00000000779,     0.00000000280,
       0.00000000215,    -0.00000000058,
      -0.00000017867,    -0.00000015750,
       0.00000005426,     0.00000006028,
      -0.00000000978,     0.00000001501,
       0.00000000420,    -0.00000000477,
      -0.00000000110,     0.00000000043,
       0.00000002424,     0.00000006526,
      -0.00000000463,    -0.00000001787,
      -0.00000000074,     0.00000000268,
       0.00000006547,    -0.00000002388,
      -0.00000001795,     0.00000000445,
       0.00000000268,     0.00000000076,
       0.00000000171,    -0.00000000076,
      -0.00000000071,    -0.00000000172,
      -0.00000001229,     0.00000001927,
       0.00000000133,     0.00000000065,
       0.00000000279,    -0.00000000181,
      -0.00000000177,    -0.00000000280,
      -0.00000002215,    -0.00000006376,
       0.00000000401,     0.00000001702,
      -0.00000001590,     0.00000000854,
       0.00000000476,    -0.00000000180,
      -0.00000000521,     0.00000003880,
       0.00000000166,    -0.00000000733,
      -0.00000000072,     0.00000000057,
       0.00000003877,     0.00000000575,
      -0.00000000729,    -0.00000000186,
       0.00000000058,     0.00000000069,
       0.00000000139,    -0.00000000001,
       0.00000000012,    -0.00000000137,
       0.00000043290,     0.00000119817,
      -0.00000008012,    -0.00000032406,
      -0.00000001381,     0.00000004635,
       0.00000007742,    -0.00000004107,
      -0.00000002270,     0.00000000836,
       0.00000000356,     0.00000000065,
       0.00000004437,    -0.00000033213,
      -0.00000001451,     0.00000006295,
       0.00000000569,    -0.00000000471,
      -0.00000033185,    -0.00000004672,
       0.00000006266,     0.00000001534,
      -0.00000000458,    -0.00000000580,
      -0.00000000654,    -0.00000000020,
       0.00000000085,    -0.00000000034,
      -0.00000000046,     0.00000000646,
      -0.00000000027,    -0.00000000083,
       0.00000000097,    -0.00000000122,
      -0.00000000123,    -0.00000000098,
       0.00000000055,     0.00000000079,
       0.00000006068,    -0.00000003746,
      -0.00000000262,    -0.00000000373,
      -0.00000000625,     0.00000000087,
       0.00000000065,     0.00000000620,
      -0.00000000010,     0.00000000119,
      -0.00000006193,     0.00000039438,
       0.00000001857,    -0.00000007366,
      -0.00000000672,     0.00000000533,
       0.00000009255,     0.00000000177,
      -0.00000001881,    -0.00000000277,
       0.00000000156,     0.00000000161,
       0.00000009707,    -0.00000013934,
      -0.00000001334,     0.00000001627,
       0.00000000177,    -0.00000000008,
      -0.00000013738,    -0.00000009964,
       0.00000001549,     0.00000001401,
       0.00000000007,    -0.00000000181,
      -0.00000000366,    -0.00000000156,
      -0.00000000211,     0.00000000316,
       0.00000093190,    -0.00000650184,
      -0.00000029718,     0.00000122111,
       0.00000011080,    -0.00000008868,
      -0.00000043974,    -0.00000000897,
       0.00000008868,     0.00000001326,
      -0.00000000721,    -0.00000000756,
      -0.00000000011,     0.00000000092,
      -0.00000084438,     0.00000120660,
       0.00000011820,    -0.00000014062,
      -0.00000001537,     0.00000000034,
       0.00000000041,     0.00000000088,
       0.00000119801,     0.00000085497,
      -0.00000013734,    -0.00000012063,
      -0.00000000035,     0.00000001541,
       0.00000000095,    -0.00000000037,
       0.00000001712,     0.00000000845,
      -0.00000000171,     0.00000000024,
       0.00000000938,    -0.00000001609,
       0.00000000001,     0.00000000153,
       0.00000001908,    -0.00000002266,
      -0.00000000160,    -0.00000000114,
      -0.00000000391,     0.00000000183,
       0.00000000179,     0.00000000392,
       0.00000001034,    -0.00000000122,
      -0.00000000024,    -0.00000000196,
      -0.00000013531,     0.00000001520,
       0.00000000113,     0.00000000884,
       0.00000000333,    -0.00000000033,
       0.00000000019,    -0.00000000279,
       0.00000000078,    -0.00000000048,
       0.00000124595,    -0.00000171031,
      -0.00000017078,     0.00000019637,
       0.00000002167,     0.00000000006,
      -0.00000037829,    -0.00000022181,
       0.00000004640,     0.00000003547,
      -0.00000000029,    -0.00000000506,
      -0.00000051414,     0.00000028042,
       0.00000003103,    -0.00000001487,
      -0.00000000132,    -0.00000000120,
       0.00000026299,     0.00000052112,
      -0.00000001046,    -0.00000003147,
      -0.00000000171,     0.00000000098,
       0.00000000553,     0.00000000502,
       0.00000000593,    -0.00000000212,
      -0.00001730996,     0.00002418994,
       0.00000240404,    -0.00000278013,
      -0.00000030711,    -0.00000000056,
       0.00000000769,     0.00000001749,
       0.00000174745,     0.00000102082,
      -0.00000021426,    -0.00000016380,
       0.00000000120,     0.00000002301,
       0.00000000131,    -0.00000000066,
       0.00000455306,    -0.00000243937,
      -0.00000028003,     0.00000012161,
       0.00000001123,     0.00000001177,
       0.00000000095,    -0.00000000037,
      -0.00000236732,    -0.00000457450,
       0.00000010472,     0.00000028097,
       0.00000001342,    -0.00000001002,
      -0.00000000038,    -0.00000000115,
      -0.00000002307,    -0.00000002888,
       0.00000000174,     0.00000000003,
      -0.00000003014,     0.00000001704,
       0.00000000045,    -0.00000000089,
       0.00000000009,    -0.00000000148,
      -0.00000000148,    -0.00000000011,
      -0.00000000080,     0.00000000070,
       0.00000000071,     0.00000000080,
      -0.00000000054,    -0.00000000110,
      -0.00000007292,     0.00000003231,
       0.00000000251,     0.00000000496,
       0.00000000684,    -0.00000000019,
       0.00000000008,    -0.00000000672,
       0.00000000001,     0.00000000100,
       0.00000005547,    -0.00000000531,
      -0.00000000017,    -0.00000000435,
      -0.00000001797,    -0.00000000706,
      -0.00000000119,     0.00000000310,
       0.00000016807,     0.00000006539,
       0.00000000504,    -0.00000001318,
       0.00000000087,     0.00000000141,
      -0.00000000139,     0.00000000050,
       0.00000000443,     0.00000001343,
       0.00000000303,    -0.00000000108,
      -0.00000801548,     0.00000418160,
       0.00000048376,    -0.00000020092,
      -0.00000001877,    -0.00000002131,
       0.00000085866,     0.00000146630,
      -0.00000004239,    -0.00000009957,
      -0.00000000458,     0.00000000388,
       0.00000119461,    -0.00000007560,
       0.00000000778,     0.00000000912,
       0.00000000023,    -0.00000000185,
       0.00000003280,    -0.00000116244,
      -0.00000000557,    -0.00000001585,
      -0.00000000157,     0.00000000126,
      -0.00000000574,    -0.00000000397,
       0.00000000017,    -0.00000000700,
       0.00000000034,     0.00000000081,
       0.00009323801,    -0.00004862037,
      -0.00000563715,     0.00000233325,
       0.00000021882,     0.00000024835,
       0.00000002006,    -0.00000000790,
      -0.00000382906,    -0.00000650214,
       0.00000018833,     0.00000044155,
       0.00000002038,    -0.00000001705,
      -0.00000000058,    -0.00000000159,
      -0.00001065652,     0.00000032095,
      -0.00000007829,    -0.00000003574,
       0.00000000164,     0.00000001591,
      -0.00000002413,     0.00001054447,
       0.00000001114,     0.00000010583,
       0.00000001523,    -0.00000000636,
       0.00000000611,     0.00000003206,
       0.00000000164,    -0.00000000186,
       0.00000002485,     0.00000001471,
      -0.00000000221,    -0.00000000378,
       0.00000000041,    -0.00000000061,
       0.00000000060,     0.00000000052,
       0.00000000228,    -0.00000002366,
      -0.00000000158,     0.00000000001,
      -0.00000000282,     0.00000000443,
       0.00000000054,    -0.00000000047,
       0.00000000440,     0.00000000285,
      -0.00000000049,    -0.00000000054,
      -0.00000001425,     0.00000001202,
       0.00000000087,     0.00000000091,
       0.00000000228,    -0.00000000079,
      -0.00000000077,    -0.00000000229,
       0.00000002104,    -0.00000002688,
       0.00000000076,     0.00000000059,
      -0.00000001193,     0.00000000024,
       0.00000000002,     0.00000000213,
       0.00000012076,    -0.00000000261,
      -0.00000000012,    -0.00000000915,
       0.00002317657,    -0.00000054127,
       0.00000019171,     0.00000004442,
      -0.00000000844,    -0.00000003292,
      -0.00000004884,    -0.00000399816,
       0.00000000412,    -0.00000002877,
      -0.00000000587,     0.00000000120,
      -0.00000013231,     0.00000030694,
       0.00000000336,    -0.00000001712,
       0.00000000028,    -0.00000000201,
      -0.00000002102,    -0.00000029351,
       0.00000000023,     0.00000003275,
      -0.00000000017,    -0.00000000166,
       0.00000000710,     0.00000001590,
      -0.00000000056,     0.00000000068,
      -0.00000000872,     0.00000001286,
      -0.00000000030,    -0.00000000104,
      -0.00021554080,     0.00000501615,
      -0.00000177997,    -0.00000040504,
       0.00000007803,     0.00000030597,
       0.00000021936,     0.00001697680,
      -0.00000001875,     0.00000013160,
       0.00000002484,    -0.00000000548,
      -0.00000009335,    -0.00000154298,
       0.00000006930,     0.00000008855,
      -0.00000000593,     0.00000001091,
      -0.00000069898,     0.00000123202,
       0.00000004424,    -0.00000016262,
       0.00000000629,     0.00000000959,
       0.00000002622,    -0.00000007376,
      -0.00000000215,    -0.00000000313,
      -0.00000003450,    -0.00000005848,
      -0.00000000312,     0.00000000477,
      -0.00000000120,     0.00000000048,
      -0.00000000038,    -0.00000000114,
      -0.00000000113,    -0.00000000058,
      -0.00000000143,     0.00000000085,
       0.00000000083,     0.00000000146,
      -0.00000004637,     0.00000007237,
       0.00000000518,     0.00000000271,
      -0.00000000058,    -0.00000000066,
       0.00000001263,    -0.00000000743,
      -0.00000000109,     0.00000000000,
      -0.00000000719,    -0.00000001272,
      -0.00000000004,     0.00000000110,
       0.00000000087,    -0.00000000053,
      -0.00000000055,    -0.00000000088,
       0.00000003771,    -0.00000001291,
      -0.00000000101,    -0.00000000270,
      -0.00000000273,     0.00000000017,
       0.00000000004,     0.00000000262,
      -0.00000005577,     0.00000001723,
      -0.00000000057,    -0.00000000152,
       0.00000000006,     0.00000000006,
      -0.00000000006,     0.00000000006,
       0.00000000022,    -0.00000000035,
       0.00000000034,     0.00000000021,
      -0.00000013407,    -0.00000002433,
       0.00000002250,    -0.00000002678,
       0.00000000258,     0.00000000455,
      -0.00000001168,     0.00000006316,
      -0.00000001257,    -0.00000001068,
       0.00000000213,    -0.00000000123,
       0.00000000033,    -0.00000000142,
       0.00000000300,     0.00000000267,
       0.00000168977,     0.00000353647,
      -0.00000044278,    -0.00000009749,
       0.00000001262,    -0.00000004372,
       0.00000044873,    -0.00000032219,
       0.00000000172,     0.00000006643,
      -0.00000000783,    -0.00000000026,
       0.00000082587,     0.00000000049,
       0.00000000681,    -0.00000000207,
       0.00000000003,    -0.00000000107,
      -0.00000000165,     0.00000084980,
      -0.00000000196,     0.00000000250,
       0.00000000141,    -0.00000000042,
      -0.00000000102,    -0.00000000306,
       0.00000000308,    -0.00000000108,
      -0.00000780965,    -0.00002372193,
       0.00000239666,     0.00000144737,
      -0.00000016262,     0.00000016704,
      -0.00000184469,     0.00000080586,
       0.00000011308,    -0.00000018749,
       0.00000001331,     0.00000001265,
      -0.00000000084,     0.00000000038,
      -0.00000162736,    -0.00000001612,
      -0.00000001308,     0.00000001174,
      -0.00000000084,     0.00000000116,
      -0.00000007825,    -0.00000172020,
       0.00000001692,    -0.00000001275,
      -0.00000000362,     0.00000000188,
       0.00000000761,     0.00000002051,
       0.00000000112,    -0.00000000203,
      -0.00000001790,    -0.00000000314,
       0.00000000170,     0.00000000206,
       0.00000000410,     0.00000000426,
      -0.00000000430,     0.00000000431,
       0.00000000084,    -0.00000000057,
      -0.00000000059,    -0.00000000085,
      -0.00000001743,     0.00000001037,
       0.00000000209,     0.00000000321,
       0.00000000254,    -0.00000000025,
       0.00000000001,    -0.00000000251,
       0.00000020258,    -0.00000011978,
      -0.00000000931,    -0.00000001421,
       0.00000000001,     0.00000000133,
      -0.00000002284,     0.00000000138,
       0.00000000053,     0.00000002262,
       0.00000001412,    -0.00000000844,
      -0.00000000069,    -0.00000000096,
      -0.00000000175,     0.00000000019,
       0.00000000017,     0.00000000175,
      -0.00000004055,     0.00000000227,
       0.00000000014,     0.00000000319,
       0.00000016931,     0.00000018528,
      -0.00000002831,    -0.00000001347,
      -0.00000000552,     0.00000000535,
       0.00000000036,    -0.00000000087,
       0.00000000002,     0.00000000008,
      -0.00000000008,     0.00000000002,
       0.00000000121,    -0.00000000044,
       0.00000000040,     0.00000000115,
      -0.00000025124,     0.00000008163,
       0.00000002028,    -0.00000005839,
       0.00000000555,     0.00000000621,
       0.00000001182,     0.00000001988,
      -0.00000000523,    -0.00000000029,
       0.00000000054,    -0.00000000053,
       0.00000002193,     0.00000012834,
      -0.00000001268,    -0.00000000867,
       0.00000000086,    -0.00000000081,
      -0.00000012844,     0.00000002292,
       0.00000000853,    -0.00000001292,
       0.00000000083,     0.00000000085,
      -0.00000000734,     0.00000000955,
       0.00000000063,     0.00000000050,
      -0.00000000954,    -0.00000000731,
      -0.00000000050,     0.00000000062,
       0.00000102769,    -0.00000027386,
      -0.00000011266,     0.00000021170,
      -0.00000001435,    -0.00000003123,
      -0.00000005281,    -0.00000009209,
       0.00000002437,     0.00000000135,
      -0.00000000227,     0.00000000255,
      -0.00000003254,    -0.00000024025,
       0.00000003342,     0.00000002321,
      -0.00000000449,     0.00000000262,
       0.00000012391,    -0.00000004883,
       0.00000000080,     0.00000000303,
       0.00000000013,     0.00000000196,
       0.00000003344,    -0.00000004376,
      -0.00000000289,    -0.00000000234,
       0.00000004430,     0.00000003340,
       0.00000000228,    -0.00000000282,
      -0.00000000029,     0.00000000090,
      -0.00000000031,     0.00000000283,
      -0.00000000305,    -0.00000000038,
       0.00000001305,    -0.00000000894,
      -0.00000000071,    -0.00000000095,
      -0.00000000149,     0.00000000016,
       0.00000000009,     0.00000000147,
       0.00000003780,    -0.00000000276,
      -0.00000000036,    -0.00000000658,
      -0.00000035048,     0.00000002541,
       0.00000000151,     0.00000002782,
       0.00000000093,     0.00000000014,
      -0.00000000403,    -0.00000000238,
      -0.00000000079,     0.00000000589,
      -0.00000002498,     0.00000000201,
       0.00000000027,     0.00000000209,
      -0.00000004429,     0.00000012056,
      -0.00000000401,     0.00000000117,
      -0.00000000836,     0.00000000238,
       0.00000000094,     0.00000000272,
      -0.00000000277,     0.00000000096,
      -0.00000000044,    -0.00000000109,
       0.00000000109,    -0.00000000044,
       0.00000041092,    -0.00000406879,
       0.00000110317,     0.00000080886,
      -0.00000021048,     0.00000012110,
      -0.00000000482,    -0.00000003007,
       0.00000018299,     0.00000001492,
      -0.00000003546,     0.00000005081,
      -0.00000000562,    -0.00000000946,
       0.00000000149,    -0.00000000013,
       0.00000000646,    -0.00000000587,
      -0.00000000019,     0.00000000190,
      -0.00000000095,     0.00000000629,
      -0.00000000053,    -0.00000000228,
       0.00000000261,     0.00000000569,
      -0.00000000442,     0.00000000145,
       0.00000000025,    -0.00000000082,
      -0.00000002207,     0.00000000277,
       0.00000000019,     0.00000000177,
       0.00000000876,     0.00000000478,
       0.00000000039,    -0.00000000125,
       0.00000000100,    -0.00000000002,
       0.00000000013,     0.00000000095,
      -0.00000005463,    -0.00000002426,
      -0.00000000183,     0.00000000541,
      -0.00000000220,    -0.00000000028,
      -0.00000000058,    -0.00000000176,
      -0.00000000403,    -0.00000000179,
      -0.00000000179,     0.00000000404,
       0.00000001075,     0.00000000456,
       0.00000000080,    -0.00000000178,
       0.00000000159,     0.00000000003,
      -0.00000000009,    -0.00000000158,
      -0.00000014024,    -0.00000005122,
      -0.00000000243,     0.00000000523,
      -0.00000001501,     0.00000000020,
       0.00000000040,     0.00000001500,
       0.00000000100,     0.00000000005,
       0.00000000204,    -0.00000000005,
       0.00000000005,     0.00000000208,
       0.00000001067,    -0.00000006262,
      -0.00000000298,     0.00000000006,
      -0.00000000159,    -0.00000000002,
      -0.00000000007,    -0.00000000191,
       0.00000000167,    -0.00000004062,
       0.00000000114,     0.00000000011,
       0.00000000049,     0.00000000053,
      -0.00000000053,     0.00000000049,
       0.00000000102,     0.00000000202,
       0.00000000202,    -0.00000000102,
      -0.00000000090,     0.00000000040,
      -0.00000000170,    -0.00000000012,
      -0.00000000011,     0.00000000172,
       0.00000003062,     0.00000005901,
       0.00000000272,    -0.00000000117,
       0.00000001608,     0.00000000111,
       0.00000000106,    -0.00000001612,
      -0.00000004589,    -0.00000000310,
      -0.00000000047,     0.00000000712,
       0.00000045101,     0.00000003019,
       0.00000000136,    -0.00000002028,
       0.00000000037,     0.00000000289,
       0.00000000224,    -0.00000000014,
      -0.00000000066,    -0.00000000920,
      -0.00000000107,     0.00000000014,
      -0.00000000220,    -0.00000000015,
       0.00000000013,    -0.00000000226,
       0.00000000180,     0.00000006772,
       0.00000000315,    -0.00000000041,
       0.00000000177,     0.00000000020,
      -0.00000000006,     0.00000000211,
      -0.00000000622,     0.00000003710,
      -0.00000000110,    -0.00000000025,
       0.00000000258,    -0.00000000437,
      -0.00000000149,    -0.00000000064,
      -0.00000000437,    -0.00000000258,
      -0.00000000064,     0.00000000150,
      -0.00000000083,    -0.00000000035,
      -0.00000000637,     0.00000001166,
       0.00000000204,     0.00000000086,
      -0.00000000153,    -0.00000000249,
      -0.00000000246,     0.00000000164,
      -0.00000000528,     0.00000001075,
       0.00000000206,     0.00000000149,
       0.00000000035,    -0.00000000071,
       0.00000000383,     0.00000000659,
       0.00000000099,    -0.00000000072,
       0.00000000662,    -0.00000000392,
      -0.00000000074,    -0.00000000100,
       0.00000008836,    -0.00000015663,
      -0.00000005334,    -0.00000002224,
      -0.00000000642,    -0.00000000467,
      -0.00000000110,     0.00000000223,
      -0.00000003592,    -0.00000006174,
      -0.00000000932,     0.00000000679,
       0.00000000043,     0.00000000063,
      -0.00000006178,     0.00000003607,
       0.00000000685,     0.00000000932,
       0.00000000063,    -0.00000000045,
       0.00000010517,     0.00000017576,
       0.00000002682,    -0.00000001962,
       0.00000003320,    -0.00000001694,
      -0.00000000347,    -0.00000000475,
       0.00000006380,     0.00000000115,
      -0.00000000023,    -0.00000006506,
      -0.00000129715,    -0.00000216563,
      -0.00000033056,     0.00000024177,
       0.00000001552,     0.00000002230,
      -0.00000009961,     0.00000005050,
       0.00000001036,     0.00000001424,
       0.00000000095,    -0.00000000066,
      -0.00000061036,    -0.00000000466,
      -0.00000000184,    -0.00000000028,
       0.00000000002,     0.00000000099,
      -0.00000000278,     0.00000061203,
       0.00000000009,     0.00000000118,
       0.00000217153,     0.00000001597,
       0.00000000161,    -0.00000033614,
       0.00000000388,     0.00000002821,
       0.00000000031,    -0.00000000233,
       0.00000000105,     0.00000000703,
       0.00000000025,     0.00000000327,
      -0.00002141501,    -0.00000015083,
      -0.00000005483,    -0.00000000406,
       0.00000000128,     0.00000003303,
      -0.00000000501,     0.00000095563,
      -0.00000000011,     0.00000000255,
       0.00000000146,    -0.00000000006,
       0.00000000717,    -0.00000013751,
       0.00000001042,     0.00000001017,
      -0.00000000080,     0.00000000088,
      -0.00000010772,    -0.00000001970,
       0.00000000728,    -0.00000001441,
       0.00000000061,     0.00000000099,
      -0.00000000026,    -0.00000000188,
      -0.00000000148,     0.00000000036,
      -0.00000005045,     0.00000058114,
      -0.00000005782,    -0.00000004054,
       0.00000006736,     0.00000000625,
      -0.00000000478,     0.00000000681,
       0.00000013915,     0.00000000115,
      -0.00000000004,     0.00000014130,
       0.00000036869,    -0.00000431401,
       0.00000042687,     0.00000029760,
      -0.00000003075,     0.00000002577,
      -0.00000019975,    -0.00000001957,
       0.00000001400,    -0.00000001999,
       0.00000000119,     0.00000000144,
      -0.00000011097,     0.00000000085,
      -0.00000000179,     0.00000000190,
       0.00000000197,    -0.00000013118,
       0.00000000193,     0.00000000104,
      -0.00000007578,    -0.00000001236,
       0.00000001145,    -0.00000001373,
      -0.00000000101,     0.00000000601,
      -0.00000000109,    -0.00000000094,
      -0.00000000272,     0.00000003581,
      -0.00000000350,    -0.00000000244,
      -0.00000003596,    -0.00000000276,
       0.00000000246,    -0.00000000353,
       0.00000000647,    -0.00000003047,
       0.00000012176,    -0.00000000802,
      -0.00000000627,     0.00000002550,
       0.00000000352,    -0.00000001710,
       0.00000000305,     0.00000000286,
       0.00000000787,    -0.00000008851,
       0.00000002024,     0.00000001456,
      -0.00000000384,     0.00000000184,
      -0.00000004395,    -0.00000000449,
       0.00000001153,    -0.00000001588,
       0.00000000156,     0.00000000352,
      -0.00000000028,    -0.00000000149,
       0.00000000154,    -0.00000000028,
      -0.00000000096,    -0.00000000081,
      -0.00000000081,     0.00000000096,
      -0.00000001173,    -0.00000001197,
      -0.00000000078,     0.00000000098,
       0.00000000146,     0.00000000462,
      -0.00000000024,    -0.00000000135,
       0.00000000463,    -0.00000000146,
      -0.00000000134,     0.00000000028,
      -0.00000000101,     0.00000000049,
      -0.00000000047,     0.00000000234,
       0.00000000235,     0.00000000046,
       0.00000002215,     0.00000007446,
      -0.00000000260,    -0.00000002090,
       0.00000000486,    -0.00000000222,
      -0.00000000149,     0.00000000033,
       0.00000000387,    -0.00000001998,
      -0.00000000146,     0.00000000386,
      -0.00000001996,    -0.00000000403,
       0.00000000383,     0.00000000148,
      -0.00000000445,     0.00000002069,
       0.00000000496,     0.00000000034,
      -0.00000000101,    -0.00000000026,
       0.00000000558,    -0.00000000700,
      -0.00000000093,     0.00000000077,
      -0.00000000691,    -0.00000000572,
       0.00000000078,     0.00000000094,
       0.00000006993,    -0.00000034070,
      -0.00000002554,     0.00000006428,
      -0.00000002332,    -0.00000000187,
       0.00000000470,     0.00000000138,
      -0.00000004856,     0.00000006071,
       0.00000000796,    -0.00000000651,
      -0.00000000097,    -0.00000000010,
       0.00000006020,     0.00000004910,
      -0.00000000630,    -0.00000000812,
      -0.00000000012,     0.00000000100,
       0.00000000088,     0.00000000049,
       0.00000000054,    -0.00000000081,
       0.00000006107,    -0.00000007376,
      -0.00000000975,     0.00000000775,
      -0.00000001660,    -0.00000001116,
       0.00000000193,     0.00000000209,
      -0.00000002395,     0.00000001098,
       0.00000000172,    -0.00000000023,
       0.00000001006,     0.00000002425,
      -0.00000000001,    -0.00000000174,
      -0.00000085063,     0.00000104395,
       0.00000013796,    -0.00000010990,
      -0.00000001681,    -0.00000000254,
       0.00000007629,     0.00000005141,
      -0.00000000870,    -0.00000000953,
      -0.00000000013,     0.00000000126,
       0.00000021174,    -0.00000009486,
      -0.00000001544,     0.00000000147,
       0.00000000063,     0.00000000082,
      -0.00000009150,    -0.00000021249,
       0.00000000062,     0.00000001539,
       0.00000000090,    -0.00000000056,
      -0.00000000093,    -0.00000000138,
      -0.00000000141,     0.00000000066,
      -0.00000031136,     0.00000013613,
       0.00000002224,    -0.00000000176,
       0.00000002841,     0.00000005738,
      -0.00000000051,    -0.00000000460,
       0.00000004392,     0.00000000093,
       0.00000000495,    -0.00000004259,
      -0.00000000098,    -0.00000000057,
       0.00000361666,    -0.00000157917,
      -0.00000025893,     0.00000002017,
       0.00000001001,     0.00000001412,
      -0.00000012646,    -0.00000025456,
       0.00000000217,     0.00000002024,
       0.00000000113,    -0.00000000076,
      -0.00000039180,    -0.00000002076,
      -0.00000000195,     0.00000000530,
      -0.00000003358,     0.00000038671,
       0.00000000701,     0.00000000335,
       0.00000000009,     0.00000000117,
       0.00000000082,     0.00000000068,
       0.00000068838,     0.00000003783,
       0.00000000393,    -0.00000001013,
       0.00000000830,    -0.00000011882,
      -0.00000000190,    -0.00000000060,
      -0.00000000031,     0.00000000891,
      -0.00000000007,    -0.00000001307,
      -0.00000000018,     0.00000000101,
      -0.00000637856,    -0.00000036133,
      -0.00000003654,     0.00000009423,
      -0.00000003577,     0.00000050391,
       0.00000000801,     0.00000000298,
      -0.00000003469,    -0.00000004611,
       0.00000000089,     0.00000000298,
      -0.00000002300,     0.00000007086,
       0.00000000232,    -0.00000000438,
       0.00000000094,    -0.00000000216,
      -0.00000000089,    -0.00000000187,
       0.00000000156,     0.00000000154,
       0.00000000168,    -0.00000000167,
       0.00000009001,     0.00000007996,
      -0.00000000751,    -0.00000000479,
       0.00000001012,    -0.00000001492,
      -0.00000000077,     0.00000000074,
       0.00000001875,     0.00000000199,
      -0.00000000075,     0.00000001871,
      -0.00000058706,    -0.00000054015,
       0.00000004119,     0.00000004073,
      -0.00000004260,     0.00000005890,
       0.00000000325,    -0.00000000318,
      -0.00000003802,    -0.00000000699,
       0.00000000011,     0.00000000111,
      -0.00000000215,    -0.00000003737,
      -0.00000000104,     0.00000000105,
       0.00000000082,     0.00000000121,
       0.00000004036,     0.00000000864,
       0.00000000100,    -0.00000000659,
       0.00000000246,     0.00000000245,
      -0.00000000187,     0.00000000199,
       0.00000000571,    -0.00000003900,
      -0.00000000545,    -0.00000000007,
       0.00000000062,    -0.00000000043,
      -0.00000000254,    -0.00000000375,
       0.00000000190,    -0.00000000329,
       0.00000000059,    -0.00000000068,
       0.00000000070,     0.00000000055,
      -0.00000042045,     0.00000000333,
       0.00000000035,     0.00000001862,
      -0.00000000039,    -0.00000000266,
      -0.00000000218,    -0.00000000083,
       0.00000073252,    -0.00000765182,
       0.00000208417,     0.00000151474,
      -0.00000039643,     0.00000022905,
      -0.00000000908,    -0.00000005687,
       0.00000023717,     0.00000002230,
      -0.00000004678,     0.00000006461,
      -0.00000000710,    -0.00000001225,
       0.00000000187,    -0.00000000014,
      -0.00000000083,     0.00000000032,
       0.00000000049,     0.00000000071,
       0.00000000074,    -0.00000000071,
      -0.00000000036,    -0.00000000257,
      -0.00000000253,     0.00000000032,
       0.00000000042,    -0.00000000097,
      -0.00000000103,    -0.00000000036,
      -0.00000000490,    -0.00000003797,
      -0.00000000292,     0.00000000069,
      -0.00000000349,     0.00000000791,
       0.00000000100,    -0.00000000119,
       0.00000000789,     0.00000000357,
      -0.00000000122,    -0.00000000100,
       0.00000000330,    -0.00000000802,
      -0.00000000221,    -0.00000000063,
      -0.00000000277,     0.00000000189,
       0.00000000185,     0.00000000285,
      -0.00000005977,     0.00000015555,
       0.00000001736,    -0.00000002267,
       0.00000001057,     0.00000000321,
      -0.00000000170,    -0.00000000115,
       0.00000002391,    -0.00000001586,
      -0.00000000324,     0.00000000024,
      -0.00000001556,    -0.00000002414,
       0.00000000015,     0.00000000328,
      -0.00000002882,     0.00000002145,
       0.00000000536,     0.00000000633,
      -0.00000000025,    -0.00000000096,
       0.00000000833,     0.00000000064,
       0.00000000100,    -0.00000000839,
       0.00000047095,    -0.00000035513,
      -0.00000006308,     0.00000001333,
      -0.00000002540,    -0.00000003001,
       0.00000000108,     0.00000000451,
      -0.00000007239,    -0.00000000655,
       0.00000000187,     0.00000000572,
      -0.00000000818,     0.00000007249,
       0.00000000608,    -0.00000000175,
       0.00000011263,    -0.00000000328,
      -0.00000000354,    -0.00000000645,
      -0.00000000087,    -0.00000002336,
      -0.00000000151,     0.00000000093,
      -0.00000000268,    -0.00000001245,
      -0.00000000141,     0.00000000099,
      -0.00000001381,     0.00000000175,
       0.00000000111,     0.00000000163,
      -0.00000156647,     0.00000004587,
       0.00000004997,     0.00000009137,
       0.00000000400,     0.00000010754,
       0.00000000689,    -0.00000000410,
      -0.00000000074,    -0.00000000051,
       0.00000002152,     0.00000011452,
       0.00000001294,    -0.00000000905,
      -0.00000000052,    -0.00000000054,
       0.00000011973,    -0.00000001772,
      -0.00000000948,    -0.00000001375,
      -0.00000000059,     0.00000000059,
       0.00000000095,     0.00000000051,
       0.00000000078,    -0.00000000119,
      -0.00000009521,    -0.00000016506,
      -0.00000001957,     0.00000001318,
      -0.00000003442,     0.00000001782,
       0.00000000276,     0.00000000384,
      -0.00000004236,     0.00000000034,
       0.00000000155,     0.00000004815,
       0.00000107562,     0.00000192901,
       0.00000022813,    -0.00000015423,
      -0.00000000942,    -0.00000001029,
       0.00000015317,    -0.00000007844,
      -0.00000001207,    -0.00000001723,
      -0.00000000082,     0.00000000070,
       0.00000039709,    -0.00000000805,
       0.00000000046,     0.00000000595,
      -0.00000001191,    -0.00000041629,
       0.00000000492,     0.00000000065,
      -0.00000070373,     0.00000001267,
       0.00000000139,    -0.00000000926,
       0.00000000268,     0.00000012126,
      -0.00000000160,    -0.00000000022,
      -0.00000000683,    -0.00000000939,
       0.00000000093,    -0.00000000173,
      -0.00000000016,    -0.00000000108,
       0.00000655723,    -0.00000012167,
      -0.00000001251,     0.00000008693,
      -0.00000001003,    -0.00000051615,
       0.00000000674,     0.00000000079,
      -0.00000000102,     0.00000000005,
       0.00000003248,     0.00000004623,
      -0.00000000279,    -0.00000000284,
       0.00000002006,    -0.00000000972,
      -0.00000000082,     0.00000000565,
      -0.00000000073,     0.00000000190,
       0.00000000141,     0.00000000169,
       0.00000000083,    -0.00000000995,
      -0.00000000468,    -0.00000000041,
      -0.00000003162,    -0.00000003299,
       0.00000000093,     0.00000000534,
       0.00000000034,     0.00000000103,
      -0.00000001875,     0.00000000010,
      -0.00000000115,    -0.00000001977,
       0.00000019022,     0.00000052782,
      -0.00000006390,    -0.00000003245,
       0.00000004078,    -0.00000001930,
      -0.00000000241,     0.00000000504,
       0.00000003659,     0.00000000118,
       0.00000000167,     0.00000003867,
      -0.00000000046,    -0.00000000182,
       0.00000000182,    -0.00000000046,
      -0.00000001662,     0.00000001540,
       0.00000000089,     0.00000000132,
       0.00000000056,     0.00000000354,
      -0.00000000185,     0.00000000079,
      -0.00000000044,     0.00000000060,
      -0.00000000061,    -0.00000000044,
      -0.00000000169,     0.00000003149,
      -0.00000000133,    -0.00000000010,
      -0.00000000071,     0.00000000083,
       0.00000000085,     0.00000000071,
      -0.00000001475,     0.00000001684,
       0.00000000116,     0.00000000086,
       0.00000000280,    -0.00000000114,
      -0.00000000106,    -0.00000000282,
       0.00000000043,     0.00000000078,
       0.00000006106,    -0.00000002756,
      -0.00000000189,    -0.00000000381,
      -0.00000000547,    -0.00000000037,
      -0.00000000060,     0.00000000539,
       0.00000000947,     0.00000000012,
       0.00000000001,    -0.00000000186,
      -0.00000013404,    -0.00000000121,
      -0.00000000009,     0.00000000868,
       0.00000000164,     0.00000000161,
       0.00000000206,    -0.00000000106,
      -0.00000000076,     0.00000000103,
      -0.00000000075,    -0.00000000048,
      -0.00000000041,     0.00000000123,
       0.00000006420,     0.00000003977,
       0.00000000336,    -0.00000000455,
       0.00000000809,     0.00000000403,
       0.00000000386,    -0.00000000974,
      -0.00000000171,    -0.00000000709,
      -0.00000000333,     0.00000000079,
       0.00000000679,     0.00000002938,
       0.00000000731,    -0.00000000175,
       0.00000000135,    -0.00000000016,
       0.00000000104,     0.00000000041,
      -0.00000000035,     0.00000000112,
       0.00000000044,     0.00000000130,
       0.00000000130,    -0.00000000044,
       0.00000001460,     0.00000002788,
      -0.00000000441,    -0.00000001321,
      -0.00000001310,     0.00000000453,
      -0.00000000027,    -0.00000000079,
      -0.00000000081,     0.00000000027,
       0.00000073036,     0.00000040506,
       0.00000004100,     0.00000022730,
      -0.00001945340,     0.00000649910,
       0.00000050430,     0.00000088300,
       0.00000003260,    -0.00000001370,
       0.00000003810,    -0.00000004460,
       0.00000076180,     0.00000067310,
      -0.00000000564,    -0.00000003057,
       0.00000000473,     0.00000001306,
      -0.00000001324,     0.00000000442,
       0.00000000026,     0.00000000079,
      -0.00000000081,     0.00000000027,
      -0.00000000044,    -0.00000000131,
       0.00000000131,    -0.00000000044,
       0.00000000445,     0.00000001179,
      -0.00000000419,     0.00000000102,
       0.00000000336,     0.00000004063,
      -0.00000001301,    -0.00000000040,
      -0.00000004074,     0.00000000315,
       0.00000000044,    -0.00000001304,
       0.00000000145,     0.00000000171,
      -0.00000000052,     0.00000000031,
      -0.00000000188,     0.00000000187,
      -0.00000000041,    -0.00000000058,
       0.00000000105,    -0.00000000118,
       0.00000000046,     0.00000000033,
      -0.00000000518,     0.00000000122,
      -0.00000000003,    -0.00000000182,
      -0.00000000121,    -0.00000000519,
       0.00000000183,    -0.00000000002,
       0.00000000064,     0.00000000203,
      -0.00000000045,     0.00000000013,
      -0.00000000201,     0.00000000064,
      -0.00000000013,    -0.00000000044,
       0.00000000083,    -0.00000000195,
       0.00000000078,     0.00000000024,
      -0.00000001081,     0.00000001121,
      -0.00000000356,    -0.00000000438,
      -0.00000001123,    -0.00000001084,
       0.00000000439,    -0.00000000357,
       0.00000000061,    -0.00000000200,
       0.00000000041,     0.00000000029,
       0.00000000198,     0.00000000060,
      -0.00000000028,     0.00000000040,
      -0.00000001315,    -0.00000001367,
       0.00000000628,    -0.00000000576,
       0.00000013886,    -0.00000005594,
       0.00000001601,     0.00000005308,
      -0.00000000024,     0.00000000025,
       0.00000005582,     0.00000013903,
      -0.00000005314,     0.00000001598,
      -0.00000000026,    -0.00000000024,
       0.00000000478,     0.00000000077,
      -0.00000000024,     0.00000000109,
      -0.00000000089,     0.00000000469,
      -0.00000000107,    -0.00000000028,
      -0.00000002643,    -0.00000009562,
       0.00000006504,    -0.00000000739,
      -0.00000000083,     0.00000002212,
      -0.00000000495,    -0.00000000095,
      -0.00000003699,     0.00000000886,
      -0.00000000228,    -0.00000002474,
       0.00000000839,     0.00000000033,
      -0.00000000034,     0.00000000191,
       0.00000002470,    -0.00000000422,
      -0.00000000128,     0.00000001397,
      -0.00000000417,    -0.00000000130,
       0.00000000921,     0.00000000693,
      -0.00000000050,     0.00000000459,
      -0.00000000182,     0.00000000088,
       0.00000000247,    -0.00000000320,
       0.00000000067,     0.00000000083,
       0.00000000313,     0.00000000258,
      -0.00000000087,     0.00000000062,
       0.00000014611,     0.00000047425,
      -0.00000033190,     0.00000004595,
       0.00000000244,    -0.00000011398,
       0.00000002562,     0.00000000468,
       0.00000013002,    -0.00000000153,
      -0.00000000845,     0.00000007970,
      -0.00000002529,    -0.00000000556,
       0.00000000180,    -0.00000000542,
       0.00000041280,     0.00000023545,
      -0.00000010142,     0.00000015855,
      -0.00000002932,    -0.00000002070,
       0.00000000212,    -0.00000000304,
      -0.00000025827,     0.00000049152,
      -0.00000021817,    -0.00000012776,
       0.00000003438,    -0.00000005186,
       0.00000000866,     0.00000000660,
      -0.00000000533,     0.00000000841,
      -0.00000000179,    -0.00000000195,
      -0.00000000848,    -0.00000000568,
       0.00000000213,    -0.00000000174,
      -0.00000001503,     0.00000005108,
      -0.00000002252,    -0.00000000801,
       0.00000000212,    -0.00000000515,
       0.00000000080,     0.00000000037,
       0.00000071284,     0.00000030658,
      -0.00000014814,     0.00000030194,
      -0.00000006664,    -0.00000003608,
       0.00000000566,    -0.00000000998,
       0.00000000116,     0.00000000070,
      -0.00000030676,     0.00000071303,
      -0.00000030206,    -0.00000014822,
       0.00000003609,    -0.00000006668,
       0.00000000998,     0.00000000566,
      -0.00000000070,     0.00000000117,
       0.00000001229,    -0.00000000411,
       0.00000000166,     0.00000000543,
      -0.00000000125,     0.00000000035,
       0.00000089238,     0.00000038369,
      -0.00000018538,     0.00000037804,
      -0.00000008344,    -0.00000004514,
       0.00000000709,    -0.00000001250,
       0.00000000146,     0.00000000087,
      -0.00000038404,     0.00000089274,
      -0.00000037821,    -0.00000018555,
       0.00000004518,    -0.00000008349,
       0.00000001251,     0.00000000709,
      -0.00000000087,     0.00000000146,
       0.00000000005,    -0.00000000002,
       0.00000000002,     0.00000000005,
       0.00000000025,    -0.00000000148,
      -0.00000000017,     0.00000000102,
      -0.00000000040,     0.00000000274,
       0.00000000064,    -0.00000000082,
      -0.00000000482,     0.00000000681,
       0.00000000101,    -0.00000000059,
      -0.00000000080,    -0.00000000104,
      -0.00000000035,     0.00000000024,
      -0.00000000024,    -0.00000000035,
       0.00000000027,    -0.00000000023,
       0.00000000039,     0.00000000045,
       0.00000000032,    -0.00000000036,
      -0.00000000011,     0.00000000040,
       0.00000000043,     0.00000000016,
      -0.00000000023,     0.00000000075,
       0.00000000059,     0.00000000016,
       0.00000000012,     0.00000000040,
      -0.00000000028,    -0.00000000023,
       0.00000000079,    -0.00000000012,
      -0.00000000026,    -0.00000000042,
       0.00000000066,     0.00000000068,
       0.00000000070,    -0.00000000068,
       0.00000000096,     0.00000000130,
      -0.00000000035,     0.00000000015,
       0.00000000089,    -0.00000000102,
      -0.00000000041,     0.00000000130,
      -0.00000000032,    -0.00000000019,
       0.00000000139,     0.00000000051,
      -0.00000000022,     0.00000000033,
      -0.00000000067,     0.00000000219,
      -0.00000000053,    -0.00000000030,
       0.00000000192,     0.00000000051,
      -0.00000000025,     0.00000000047,
      -0.00000000063,     0.00000000017,
      -0.00000000006,     0.00000000092,
      -0.00000000145,     0.00000000028,
      -0.00000000006,    -0.00000000045,
       0.00000000016,     0.00000000120,
      -0.00000000019,     0.00000000037,
       0.00000000041,     0.00000000019,
       0.00000000018,     0.00000000035,
      -0.00000000139,    -0.00000000063,
      -0.00000001236,     0.00000002826,
       0.00000000248,    -0.00000000324,
       0.00000000038,     0.00000000025,
      -0.00000000332,    -0.00000000262,
       0.00000000025,    -0.00000000037,
      -0.00000000025,    -0.00000000072,
       0.00000000142,    -0.00000000013,
      -0.00000000014,    -0.00000000146,
       0.00000000156,    -0.00000000009,
      -0.00000000036,    -0.00000000169,
       0.00000000182,     0.00000000229,
      -0.00000000059,     0.00000000035,
       0.00000000219,    -0.00000000194,
       0.00000000031,     0.00000000061,
       0.00000000264,     0.00000000288,
      -0.00000000082,     0.00000000042,
       0.00000000252,    -0.00000000296,
       0.00000000051,     0.00000000077,
      -0.00000000119,     0.00000000401,
      -0.00000000096,    -0.00000000057,
       0.00000000423,     0.00000000135,
      -0.00000000061,     0.00000000100,
      -0.00000000180,     0.00000000630,
      -0.00000000149,    -0.00000000083,
       0.00000000589,     0.00000000138,
      -0.00000000071,     0.00000000143,
      -0.00000000244,     0.00000000024,
      -0.00000000002,    -0.00000000077,
      -0.00000000023,     0.00000000328,
      -0.00000000095,    -0.00000000017,
      -0.00000000460,     0.00000000047,
      -0.00000000003,    -0.00000000136,
       0.00000000020,     0.00000000394,
      -0.00000000116,    -0.00000000003,
      -0.00000000055,     0.00000000109,
      -0.00000000028,    -0.00000000025,
      -0.00000000058,     0.00000000030,
      -0.00000000101,     0.00000000059,
      -0.00000000019,    -0.00000000038,
       0.00000000035,     0.00000000067,
      -0.00000000060,     0.00000000011,
      -0.00000000001,    -0.00000000054,
       0.00000000045,     0.00000000010,
       0.00000000044,     0.00000000138,
       0.00000000138,    -0.00000000045,
       0.00000000024,     0.00000000055,
       0.00000000106,    -0.00000000050,
      -0.00000000038,     0.00000000131,
       0.00000000130,     0.00000000037,
      -0.00000000035,     0.00000000111,
      -0.00000000024,    -0.00000000014,
       0.00000000175,     0.00000000059,
      -0.00000000023,     0.00000000037,
      -0.00000000110,     0.00000000081,
      -0.00000000016,    -0.00000000033,
       0.00000000064,     0.00000000111,
      -0.00000000107,     0.00000000069,
      -0.00000000014,    -0.00000000031,
       0.00000000098,     0.00000000152,
      -0.00000000045,     0.00000000019,
      -0.00000000069,     0.00000000005,
      -0.00000000032,     0.00000000057,
      -0.00000000069,    -0.00000000015,
       0.00000000007,    -0.00000000020,
      -0.00000000018,     0.00000000085,
      -0.00000000166,    -0.00000000245,
      -0.00000001357,     0.00000001009,
      -0.00000000037,    -0.00000000115,
       0.00000000933,    -0.00000000187,
      -0.00000000001,     0.00000000143,
      -0.00000000201,    -0.00000000932,
       0.00000000143,     0.00000000001,
       0.00000000344,    -0.00000000042,
      -0.00000000005,     0.00000000050,
      -0.00000000161,    -0.00000000486,
       0.00000000072,    -0.00000000013,
       0.00000000610,     0.00000000564,
      -0.00000000165,     0.00000000104,
       0.00000000578,    -0.00000000625,
       0.00000000106,     0.00000000168,
       0.00000000678,     0.00000000642,
      -0.00000000186,     0.00000000112,
       0.00000000587,    -0.00000000866,
       0.00000000152,     0.00000000191,
      -0.00000000296,     0.00000001267,
      -0.00000000293,    -0.00000000157,
       0.00000001317,     0.00000000359,
      -0.00000000167,     0.00000000302,
      -0.00000000428,     0.00000001815,
      -0.00000000420,    -0.00000000216,
       0.00000001819,     0.00000000329,
      -0.00000000193,     0.00000000422,
      -0.00000000848,    -0.00000000020,
       0.00000000027,    -0.00000000256,
      -0.00000000080,     0.00000001050,
      -0.00000000292,    -0.00000000058,
      -0.00000001429,     0.00000000031,
       0.00000000031,    -0.00000000422,
      -0.00000000027,     0.00000001296,
      -0.00000000376,    -0.00000000044,
      -0.00000000130,     0.00000000288,
      -0.00000000072,    -0.00000000061,
      -0.00000000175,     0.00000000093,
      -0.00000000024,    -0.00000000037,
      -0.00000000263,     0.00000000133,
      -0.00000000041,    -0.00000000099,
       0.00000000092,     0.00000000184,
      -0.00000000071,     0.00000000029,
      -0.00000000187,     0.00000000023,
      -0.00000000003,    -0.00000000058,
      -0.00000000001,    -0.00000000172,
       0.00000000052,     0.00000000007,
      -0.00000000041,     0.00000000020,
      -0.00000000017,    -0.00000000041,
      -0.00000000004,    -0.00000000002,
      -0.00000000016,     0.00000000032,
      -0.00000000005,    -0.00000000005,
      -0.00000000007,     0.00000000019,
      -0.00000000003,    -0.00000000003,
       0.00000000020,     0.00000000008,
      -0.00000000003,     0.00000000003,
      -0.00000000002,     0.00000000005,
       0.00000000021,     0.00000000010,
      -0.00000000004,     0.00000000003,
      -0.00000000009,     0.00000000003,
       0.00000000005,     0.00000000005,
      -0.00000000003,     0.00000000001,
       0.00000000005,     0.00000000010,
      -0.00000000001,    -0.00000000001,
      -0.00000000002,     0.00000000005,
      -0.00000000036,     0.00000000261,
      -0.00000000049,    -0.00000000014,
       0.00000000261,     0.00000000036,
      -0.00000000014,     0.00000000049,
      -0.00000000019,     0.00000000117,
      -0.00000000022,    -0.00000000007,
       0.00000000291,     0.00000000045,
      -0.00000000017,     0.00000000055,
      -0.00000000110,     0.00000000107,
      -0.00000000022,    -0.00000000034,
       0.00000000106,     0.00000000109,
      -0.00000000034,     0.00000000022,
      -0.00000000068,     0.00000000065,
      -0.00000000014,    -0.00000000021,
       0.00000000139,     0.00000000145,
      -0.00000000045,     0.00000000029,
      -0.00000000060,     0.00000000010,
      -0.00000000032,     0.00000000057,
      -0.00000000045,    -0.00000000008,
      -0.00000000014,     0.00000000079,
       0.00000000036,     0.00000000081,
      -0.00000000134,     0.00000000755,
      -0.00000000146,    -0.00000000048,
       0.00000000756,     0.00000000133,
      -0.00000000048,     0.00000000147,
      -0.00000000069,     0.00000000400,
      -0.00000000077,    -0.00000000027,
       0.00000000762,     0.00000000144,
      -0.00000000053,     0.00000000145,
      -0.00000000440,     0.00000000391,
      -0.00000000081,    -0.00000000133,
       0.00000000391,     0.00000000431,
      -0.00000000131,     0.00000000079,
      -0.00000000364,     0.00000000324,
      -0.00000000067,    -0.00000000109,
       0.00000000503,     0.00000000572,
      -0.00000000171,     0.00000000104,
      -0.00000000357,    -0.00000000010,
       0.00000000019,    -0.00000000104,
      -0.00000000118,     0.00000000335,
      -0.00000000090,    -0.00000000041,
      -0.00000000333,    -0.00000000066,
       0.00000000030,    -0.00000000093,
      -0.00000000090,     0.00000000472,
      -0.00000000133,    -0.00000000042,
      -0.00000000052,     0.00000000050,
      -0.00000000058,    -0.00000000066,
      -0.00000000058,    -0.00000000012,
       0.00000000012,    -0.00000000058,
       0.00000000135,    -0.00000001140,
       0.00000000152,     0.00000000046,
      -0.00000004277,    -0.00000000275,
       0.00000000150,    -0.00000000589,
       0.00000003067,     0.00000002374,
      -0.00000000697,     0.00000000544,
       0.00000002307,    -0.00000003124,
       0.00000000547,     0.00000000694,
       0.00000000015,    -0.00000000067,
      -0.00000000076,    -0.00000000012,
       0.00000001583,     0.00000001350,
      -0.00000000384,     0.00000000271,
       0.00000001409,    -0.00000002592,
       0.00000000471,     0.00000000481,
      -0.00000000724,     0.00000004265,
      -0.00000000974,    -0.00000000445,
       0.00000004463,     0.00000000699,
      -0.00000000452,     0.00000001008,
      -0.00000000819,     0.00000005041,
      -0.00000001150,    -0.00000000515,
       0.00000005710,     0.00000000566,
      -0.00000000494,     0.00000001310,
      -0.00000002945,    -0.00000000293,
       0.00000000175,    -0.00000000869,
      -0.00000000358,     0.00000003330,
      -0.00000000945,    -0.00000000219,
      -0.00000004389,    -0.00000000199,
       0.00000000193,    -0.00000001280,
      -0.00000000325,     0.00000004229,
      -0.00000001227,    -0.00000000229,
      -0.00000000260,     0.00000000725,
      -0.00000000176,    -0.00000000135,
      -0.00000000474,     0.00000000263,
      -0.00000000063,    -0.00000000102,
      -0.00000000651,     0.00000000297,
      -0.00000000088,    -0.00000000244,
       0.00000000211,     0.00000000459,
      -0.00000000174,     0.00000000063,
      -0.00000000540,     0.00000000015,
       0.00000000009,    -0.00000000159,
       0.00000000026,    -0.00000000511,
       0.00000000146,     0.00000000025,
      -0.00000000057,     0.00000000071,
      -0.00000000031,    -0.00000000027,
       0.00000000036,     0.00000000028,
      -0.00000000103,     0.00000000053,
      -0.00000000016,    -0.00000000040,
      -0.00000000048,    -0.00000000103,
       0.00000000040,    -0.00000000013,
      -0.00000000012,    -0.00000000015,
       0.00000000004,    -0.00000000002,
      -0.00000000120,     0.00000000106,
      -0.00000000018,    -0.00000000033,
      -0.00000000056,     0.00000000074,
      -0.00000000013,    -0.00000000018,
       0.00000000079,     0.00000000063,
      -0.00000000020,     0.00000000015,
      -0.00000000015,     0.00000000018,
      -0.00000000003,    -0.00000000005,
       0.00000000076,     0.00000000063,
      -0.00000000020,     0.00000000013,
      -0.00000000042,    -0.00000000015,
       0.00000000007,    -0.00000000011,
       0.00000000000,     0.00000000029,
      -0.00000000010,    -0.00000000001,
      -0.00000000015,    -0.00000000004,
       0.00000000002,    -0.00000000004,
      -0.00000000013,     0.00000000055,
      -0.00000000014,    -0.00000000006,
      -0.00000000010,    -0.00000000004,
       0.00000000000,    -0.00000000004,
      -0.00000000010,     0.00000000000,
      -0.00000000006,    -0.00000000002,
       0.00000000002,    -0.00000000007,
       0.00000000044,     0.00000000043,
      -0.00000000611,     0.00000000893,
      -0.00000000191,    -0.00000000201,
       0.00000000896,     0.00000000609,
      -0.00000000200,     0.00000000192,
      -0.00000000126,     0.00000000205,
       0.00000000205,     0.00000000126,
      -0.00000000268,     0.00000000394,
      -0.00000000083,    -0.00000000089,
       0.00000000990,     0.00000000679,
      -0.00000000223,     0.00000000211,
      -0.00000000452,    -0.00000000073,
       0.00000000038,    -0.00000000129,
      -0.00000000068,     0.00000000449,
      -0.00000000127,    -0.00000000037,
      -0.00000000055,    -0.00000000029,
      -0.00000000280,    -0.00000000045,
       0.00000000024,    -0.00000000080,
      -0.00000000094,     0.00000000595,
      -0.00000000169,    -0.00000000049,
      -0.00000000049,     0.00000000083,
      -0.00000000083,    -0.00000000059,
      -0.00000000065,    -0.00000000011,
       0.00000000011,    -0.00000000065,
      -0.00000000095,    -0.00000000008,
       0.00000000009,    -0.00000000017,
       0.00000000008,     0.00000000353,
      -0.00000000063,    -0.00000000023,
      -0.00000001781,     0.00000002584,
      -0.00000000546,    -0.00000000588,
       0.00000002592,     0.00000001777,
      -0.00000000588,     0.00000000548,
      -0.00000000144,     0.00000000236,
      -0.00000000047,    -0.00000000062,
       0.00000000235,     0.00000000144,
      -0.00000000062,     0.00000000047,
      -0.00000001027,     0.00000001463,
      -0.00000000312,    -0.00000000334,
       0.00000002704,     0.00000001932,
      -0.00000000625,     0.00000000575,
      -0.00000001764,    -0.00000000313,
       0.00000000151,    -0.00000000504,
      -0.00000000280,     0.00000001740,
      -0.00000000495,    -0.00000000141,
      -0.00000001485,    -0.00000000264,
       0.00000000130,    -0.00000000424,
      -0.00000000401,     0.00000002309,
      -0.00000000658,    -0.00000000195,
      -0.00000000136,     0.00000000202,
      -0.00000000041,    -0.00000000043,
      -0.00000000228,    -0.00000000187,
       0.00000000062,    -0.00000000049,
       0.00000000031,    -0.00000000027,
      -0.00000000025,    -0.00000000035,
      -0.00000000268,    -0.00000000047,
       0.00000000023,    -0.00000000076,
       0.00000000048,    -0.00000000268,
       0.00000000076,     0.00000000024,
       0.00000003338,    -0.00000003491,
       0.00000000390,     0.00000000751,
      -0.00000011079,    -0.00000010044,
       0.00000002287,    -0.00000001238,
      -0.00000001127,     0.00000019895,
      -0.00000004488,    -0.00000001463,
       0.00000019821,     0.00000000860,
      -0.00000001433,     0.00000004491,
       0.00000000139,     0.00000000019,
      -0.00000000014,     0.00000000050,
       0.00000000015,    -0.00000000165,
       0.00000000060,     0.00000000008,
      -0.00000000948,     0.00000013387,
      -0.00000003015,    -0.00000001042,
       0.00000018787,    -0.00000000024,
      -0.00000001131,     0.00000004262,
      -0.00000010572,    -0.00000002014,
       0.00000000880,    -0.00000003045,
      -0.00000001048,     0.00000011174,
      -0.00000003195,    -0.00000000750,
      -0.00000013032,    -0.00000001302,
       0.00000000806,    -0.00000003768,
      -0.00000001671,     0.00000013925,
      -0.00000004011,    -0.00000000952,
      -0.00000000359,     0.00000001636,
      -0.00000000390,    -0.00000000242,
      -0.00000001132,     0.00000000694,
      -0.00000000157,    -0.00000000247,
      -0.00000001433,     0.00000000596,
      -0.00000000174,    -0.00000000541,
       0.00000000444,     0.00000001027,
      -0.00000000389,     0.00000000128,
      -0.00000001457,    -0.00000000076,
       0.00000000065,    -0.00000000426,
       0.00000000136,    -0.00000001414,
       0.00000000405,     0.00000000090,
      -0.00000000109,     0.00000000126,
      -0.00000000053,    -0.00000000050,
       0.00000000061,     0.00000000050,
      -0.00000000026,     0.00000000029,
      -0.00000000230,     0.00000000096,
      -0.00000000029,    -0.00000000087,
      -0.00000000090,    -0.00000000229,
       0.00000000086,    -0.00000000027,
       0.00000000029,    -0.00000000073,
       0.00000000019,     0.00000000011,
      -0.00000000626,    -0.00000000249,
       0.00000000088,    -0.00000000162,
      -0.00000000461,    -0.00000000067,
       0.00000000034,    -0.00000000132,
      -0.00000000082,     0.00000000490,
      -0.00000000140,    -0.00000000040,
      -0.00000000006,     0.00000000003,
      -0.00000000156,    -0.00000000025,
      -0.00000000025,     0.00000000156,
      -0.00000000114,    -0.00000000018,
       0.00000000009,    -0.00000000033,
      -0.00000000082,     0.00000000530,
      -0.00000000152,    -0.00000000042,
       0.00000000016,    -0.00000000076,
       0.00000000022,     0.00000000007,
       0.00000000012,     0.00000000017,
      -0.00000000006,     0.00000000003,
       0.00000000003,    -0.00000000002,
       0.00000000002,     0.00000000003,
      -0.00000000001,    -0.00000000024,
       0.00000000008,     0.00000000000,
      -0.00000000068,     0.00000000003,
       0.00000000001,    -0.00000000020,
      -0.00000000035,    -0.00000000007,
       0.00000000003,    -0.00000000011,
       0.00000000005,    -0.00000000041,
       0.00000000013,     0.00000000003,
       0.00000000005,    -0.00000000001,
      -0.00000000002,    -0.00000000008,
      -0.00000000001,    -0.00000000012,
       0.00000000013,    -0.00000000001,
       0.00000000012,     0.00000000140,
      -0.00000000034,    -0.00000000019,
      -0.00000003565,    -0.00000000573,
       0.00000000291,    -0.00000001017,
      -0.00000000559,     0.00000003565,
      -0.00000001016,    -0.00000000285,
      -0.00000000846,     0.00000002782,
      -0.00000000608,    -0.00000000353,
      -0.00000000168,    -0.00000000039,
      -0.00000002008,    -0.00000000318,
       0.00000000159,    -0.00000000573,
      -0.00000000322,     0.00000002008,
      -0.00000000573,    -0.00000000159,
      -0.00000001621,    -0.00000000258,
       0.00000000131,    -0.00000000463,
      -0.00000000641,     0.00000004031,
      -0.00000001151,    -0.00000000325,
      -0.00000000367,    -0.00000000056,
       0.00000000029,    -0.00000000104,
       0.00000000056,    -0.00000000367,
       0.00000000104,     0.00000000029,
      -0.00000000063,    -0.00000000217,
       0.00000000055,    -0.00000000002,
      -0.00000000608,     0.00000000278,
      -0.00000000028,    -0.00000000162,
      -0.00000010238,    -0.00000001670,
       0.00000000841,    -0.00000002923,
      -0.00000001605,     0.00000010224,
      -0.00000002925,    -0.00000000820,
      -0.00000000947,     0.00000003114,
      -0.00000000681,    -0.00000000395,
      -0.00000000151,    -0.00000000043,
      -0.00000002244,    -0.00000000356,
       0.00000000178,    -0.00000000641,
      -0.00000000360,     0.00000002244,
      -0.00000000640,    -0.00000000178,
      -0.00000006173,    -0.00000000984,
       0.00000000501,    -0.00000001762,
      -0.00000001803,     0.00000011323,
      -0.00000003231,    -0.00000000915,
       0.00000000068,    -0.00000000121,
       0.00000000031,     0.00000000029,
       0.00000000063,    -0.00000000063,
       0.00000000081,    -0.00000000075,
       0.00000000023,     0.00000000032,
      -0.00000000076,    -0.00000000108,
       0.00000000041,    -0.00000000022,
      -0.00000000979,    -0.00000000148,
       0.00000000079,    -0.00000000280,
       0.00000000149,    -0.00000000979,
       0.00000000280,     0.00000000079,
       0.00000010491,     0.00000008537,
      -0.00000002151,     0.00000002048,
       0.00000023926,    -0.00000027551,
       0.00000005292,     0.00000005982,
      -0.00000045234,    -0.00000007797,
       0.00000003798,    -0.00000012893,
       0.00000001810,     0.00000000770,
      -0.00000007086,     0.00000044739,
      -0.00000012824,    -0.00000003671,
       0.00000000770,    -0.00000001800,
      -0.00000000067,    -0.00000000018,
       0.00000000061,    -0.00000000021,
      -0.00000000196,    -0.00000000031,
      -0.00000000031,     0.00000000196,
      -0.00000036315,    -0.00000005984,
       0.00000003011,    -0.00000010372,
      -0.00000007642,     0.00000046158,
      -0.00000013182,    -0.00000003833,
       0.00000000644,     0.00000001174,
      -0.00000000320,     0.00000000001,
      -0.00000000396,     0.00000002129,
      -0.00000000577,    -0.00000000095,
       0.00000000071,     0.00000000003,
      -0.00000000008,     0.00000000091,
      -0.00000002654,     0.00000000943,
      -0.00000000260,    -0.00000000996,
       0.00000000752,     0.00000001902,
      -0.00000000717,     0.00000000210,
      -0.00000003511,    -0.00000000554,
       0.00000000281,    -0.00000001003,
       0.00000000611,    -0.00000003466,
       0.00000000984,     0.00000000304,
      -0.00000000171,     0.00000000173,
      -0.00000000072,    -0.00000000078,
       0.00000000088,     0.00000000077,
      -0.00000000037,     0.00000000036,
      -0.00000000430,     0.00000000144,
      -0.00000000040,    -0.00000000159,
      -0.00000000139,    -0.00000000429,
       0.00000000159,    -0.00000000038,
      -0.00000000032,     0.00000000032,
      -0.00000000032,    -0.00000000032,
      -0.00000000002,     0.00000000018,
      -0.00000000008,     0.00000000000,
       0.00000000003,    -0.00000000006,
       0.00000000012,    -0.00000000006,
       0.00000000001,     0.00000000004,
      -0.00000001776,    -0.00000000275,
       0.00000000141,    -0.00000000508,
       0.00000000047,    -0.00000000299,
       0.00000000094,     0.00000000026,
       0.00000000001,     0.00000000001,
       0.00000000004,    -0.00000000006,
      -0.00000000058,    -0.00000000008,
       0.00000000004,    -0.00000000016,
      -0.00000000053,     0.00000000367,
      -0.00000000103,    -0.00000000028,
       0.00000000022,    -0.00000000072,
       0.00000000101,    -0.00000000046,
      -0.00000020221,    -0.00000003151,
       0.00000001622,    -0.00000005792,
       0.00000000544,    -0.00000003565,
       0.00000001016,     0.00000000282,
       0.00000000063,     0.00000000084,
       0.00000000017,    -0.00000000039,
      -0.00000000297,    -0.00000000041,
       0.00000000021,    -0.00000000083,
      -0.00000000188,     0.00000001367,
      -0.00000000388,    -0.00000000102,
       0.00000000084,    -0.00000000159,
       0.00000000025,     0.00000000036,
       0.00000000256,    -0.00000000115,
       0.00000000024,     0.00000000062,
      -0.00000021999,    -0.00000003437,
       0.00000001768,    -0.00000006305,
       0.00000000565,    -0.00000003701,
       0.00000001054,     0.00000000293,
       0.00000000069,     0.00000000092,
       0.00000000064,     0.00000000030,
      -0.00000000011,     0.00000000020,
       0.00000000054,    -0.00000000113,
       0.00000000031,     0.00000000020,
      -0.00000000082,     0.00000000044,
      -0.00000002961,    -0.00000000689,
       0.00000000298,    -0.00000000836,
      -0.00000001972,     0.00000009332,
      -0.00000002641,    -0.00000000887,
       0.00000000362,    -0.00000000874,
       0.00000000178,     0.00000000184,
       0.00000001078,    -0.00000000790,
       0.00000000190,     0.00000000260,
      -0.00000000010,     0.00000000065,
      -0.00000001983,    -0.00000000303,
       0.00000000158,    -0.00000000567,
       0.00000000044,    -0.00000000279,
       0.00000000087,     0.00000000024,
       0.00000000478,     0.00000000111,
      -0.00000000040,     0.00000000152,
       0.00000000115,    -0.00000000605,
       0.00000000195,     0.00000000043,
       0.00000000221,     0.00000000004,
      -0.00000000009,     0.00000000062,
      -0.00000000009,     0.00000000196,
      -0.00000000055,    -0.00000000010,
       0.00000000050,     0.00000000010,
      -0.00000000011,     0.00000000049,
       0.00000000004,     0.00000000001,
       0.00000000013,    -0.00000000003,
      -0.00000000058,     0.00000000000,
      -0.00000000033,     0.00000000018,
       0.00000000028,     0.00000000052,
      -0.00000000055,    -0.00000000033,
       0.00000000022,    -0.00000000059,
      -0.00000001272,    -0.00000000275,
       0.00000000029,    -0.00000000132,
      -0.00000001360,    -0.00000000295,
       0.00000000030,    -0.00000000138,
      -0.00000000061,    -0.00000000021,
       0.00000000020,    -0.00000000060,
       0.00000000042,    -0.00000000057,
      -0.00000000027,     0.00000000073,
      -0.00000000043,    -0.00000000148,
       0.00000000047,    -0.00000000004,
       0.00000000149,    -0.00000000044,
       0.00000000004,     0.00000000047,
       0.00000000136,     0.00000000032,
       0.00000000032,    -0.00000000136,
       0.00000000061,     0.00000000059,
       0.00000000058,    -0.00000000065,
       0.00000002330,     0.00000000634,
       0.00000000016,    -0.00000000157,
      -0.00000000517,    -0.00000000483,
       0.00000000158,     0.00000000179,
      -0.00000000486,     0.00000000513,
       0.00000000180,    -0.00000000158,
       0.00000000529,     0.00000000520,
       0.00000000101,    -0.00000000151,
      -0.00000000142,    -0.00000000413,
       0.00000000028,     0.00000000111,
      -0.00000000417,     0.00000000137,
       0.00000000111,    -0.00000000028,
      -0.00000010184,    -0.00000009722,
       0.00000003013,     0.00000003528,
      -0.00000000499,     0.00000000722,
       0.00000000203,    -0.00000000224,
       0.00000001209,     0.00000003538,
      -0.00000000229,    -0.00000000939,
      -0.00000000040,     0.00000000136,
       0.00000003548,    -0.00000001183,
      -0.00000000942,     0.00000000216,
       0.00000000136,     0.00000000042,
       0.00000000076,    -0.00000000034,
      -0.00000000031,    -0.00000000076,
      -0.00000001374,    -0.00000004367,
       0.00000000249,     0.00000001141,
      -0.00000000884,     0.00000000444,
       0.00000000256,    -0.00000000092,
      -0.00000000306,     0.00000002100,
       0.00000000088,    -0.00000000389,
       0.00000002092,     0.00000000348,
      -0.00000000383,    -0.00000000102,
       0.00000024554,     0.00000074130,
      -0.00000004477,    -0.00000019498,
      -0.00000000858,     0.00000002707,
       0.00000004297,    -0.00000002128,
      -0.00000001222,     0.00000000427,
       0.00000000185,     0.00000000037,
       0.00000002711,    -0.00000018173,
      -0.00000000802,     0.00000003369,
       0.00000000298,    -0.00000000245,
      -0.00000018135,    -0.00000002873,
       0.00000003341,     0.00000000859,
      -0.00000000234,    -0.00000000306,
      -0.00000000299,    -0.00000000006,
      -0.00000000020,     0.00000000294,
      -0.00000004396,     0.00000025669,
       0.00000001214,    -0.00000004680,
       0.00000005117,     0.00000000216,
      -0.00000001013,    -0.00000000163,
       0.00000000081,     0.00000000086,
       0.00000005038,    -0.00000007226,
      -0.00000000666,     0.00000000833,
      -0.00000007050,    -0.00000005215,
       0.00000000769,     0.00000000710,
      -0.00000000163,    -0.00000000063,
      -0.00000000094,     0.00000000131,
       0.00000061876,    -0.00000387432,
      -0.00000017975,     0.00000071153,
       0.00000006337,    -0.00000004981,
      -0.00000024271,    -0.00000001050,
       0.00000004768,     0.00000000777,
      -0.00000000373,    -0.00000000401,
      -0.00000044569,     0.00000063012,
       0.00000006014,    -0.00000007191,
      -0.00000000752,     0.00000000017,
       0.00000062313,     0.00000045240,
      -0.00000006952,    -0.00000006164,
      -0.00000000029,     0.00000000752,
       0.00000000763,     0.00000000370,
       0.00000000421,    -0.00000000701,
       0.00000074934,    -0.00000102040,
      -0.00000009905,     0.00000011433,
       0.00000001212,     0.00000000008,
      -0.00000019678,    -0.00000011936,
       0.00000002350,     0.00000001820,
      -0.00000000010,    -0.00000000246,
      -0.00000024234,     0.00000013659,
       0.00000001371,    -0.00000000740,
      -0.00000000056,    -0.00000000044,
       0.00000012232,     0.00000024538,
      -0.00000000424,    -0.00000001378,
      -0.00000000078,     0.00000000030,
       0.00000000250,     0.00000000198,
       0.00000000241,    -0.00000000055,
      -0.00000972708,     0.00001344124,
       0.00000130291,    -0.00000151066,
      -0.00000016018,    -0.00000000087,
       0.00000090795,     0.00000054849,
      -0.00000010825,    -0.00000008380,
       0.00000000045,     0.00000001124,
       0.00000216696,    -0.00000117588,
      -0.00000012507,     0.00000005797,
       0.00000000436,     0.00000000483,
      -0.00000112563,    -0.00000217650,
       0.00000004688,     0.00000012501,
       0.00000000584,    -0.00000000353,
      -0.00000000997,    -0.00000001235,
      -0.00000001294,     0.00000000669,
      -0.00000418484,     0.00000220680,
       0.00000023732,    -0.00000010316,
      -0.00000000777,    -0.00000000987,
       0.00000040243,     0.00000069331,
      -0.00000001919,    -0.00000004415,
      -0.00000000193,     0.00000000147,
       0.00000049702,    -0.00000005402,
       0.00000000499,     0.00000000611,
       0.00000001637,    -0.00000046981,
      -0.00000000311,    -0.00000001063,
      -0.00000000078,     0.00000000083,
      -0.00000000299,    -0.00000000195,
       0.00000000028,    -0.00000000345,
       0.00004673369,    -0.00002466779,
      -0.00000265224,     0.00000115166,
       0.00000008700,     0.00000011044,
      -0.00000180282,    -0.00000308764,
       0.00000008540,     0.00000019567,
       0.00000000860,    -0.00000000638,
      -0.00000442746,     0.00000025407,
      -0.00000005442,    -0.00000002554,
       0.00000000144,     0.00000000792,
       0.00000003902,     0.00000433925,
       0.00000000229,     0.00000007216,
       0.00000000763,    -0.00000000418,
       0.00000000313,     0.00000001588,
       0.00000000084,    -0.00000000080,
       0.00000001214,     0.00000000714,
      -0.00000000090,    -0.00000000188,
       0.00001009819,    -0.00000044908,
       0.00000014035,     0.00000003544,
      -0.00000000635,    -0.00000001727,
      -0.00000004595,    -0.00000160770,
       0.00000000373,    -0.00000001993,
      -0.00000000287,     0.00000000087,
      -0.00000008441,     0.00000017572,
       0.00000000106,    -0.00000000932,
       0.00000000027,    -0.00000000115,
      -0.00000000674,    -0.00000017367,
      -0.00000000010,     0.00000001835,
      -0.00000000022,    -0.00000000089,
       0.00000000327,     0.00000000693,
      -0.00000000388,     0.00000000565,
      -0.00009390810,     0.00000416781,
      -0.00000130557,    -0.00000031789,
       0.00000005733,     0.00000015941,
       0.00000021278,     0.00000688760,
      -0.00000001654,     0.00000009204,
       0.00000001219,    -0.00000000394,
      -0.00000004805,    -0.00000092935,
       0.00000004861,     0.00000004809,
      -0.00000000399,     0.00000000678,
      -0.00000046771,     0.00000071818,
       0.00000002504,    -0.00000009758,
       0.00000000448,     0.00000000561,
       0.00000001063,    -0.00000003254,
      -0.00000000087,    -0.00000000130,
      -0.00000001466,    -0.00000002520,
      -0.00000000152,     0.00000000191,
       0.00000000064,    -0.00000000094,
      -0.00000000090,    -0.00000000039,
      -0.00000000037,     0.00000000088,
       0.00000071801,     0.00000205889,
      -0.00000021015,    -0.00000011874,
       0.00000001419,    -0.00000001564,
       0.00000025149,    -0.00000010835,
      -0.00000001472,     0.00000002557,
      -0.00000000192,    -0.00000000175,
       0.00000039251,    -0.00000000702,
       0.00000000498,    -0.00000000085,
       0.00000001828,     0.00000040471,
      -0.00000000283,     0.00000000486,
       0.00000000086,    -0.00000000037,
      -0.00000000063,    -0.00000000153,
       0.00000000149,    -0.00000000047,
      -0.00000551712,    -0.00001535252,
       0.00000159195,     0.00000082215,
      -0.00000009635,     0.00000012349,
      -0.00000111445,     0.00000049361,
       0.00000006108,    -0.00000011493,
       0.00000000856,     0.00000000765,
      -0.00000074051,    -0.00000000165,
      -0.00000001747,     0.00000002388,
      -0.00000000318,    -0.00000000114,
      -0.00000005763,    -0.00000090065,
       0.00000002821,    -0.00000000338,
      -0.00000000399,     0.00000000403,
       0.00000000472,     0.00000001121,
       0.00000000056,    -0.00000000095,
      -0.00000000920,    -0.00000000306,
       0.00000000072,     0.00000000124,
      -0.00000001489,    -0.00000000184,
       0.00000000173,    -0.00000000580,
      -0.00000000087,     0.00000000590,
      -0.00000000160,    -0.00000000088,
       0.00000000389,    -0.00000000097,
       0.00000000034,     0.00000000162,
      -0.00000000198,    -0.00000000161,
       0.00000000122,    -0.00000000093,
      -0.00000032651,     0.00000000955,
       0.00000004602,    -0.00000006711,
       0.00000000737,     0.00000001057,
       0.00000001422,     0.00000000246,
      -0.00000000048,     0.00000000158,
       0.00000001949,     0.00000008938,
      -0.00000001200,    -0.00000000699,
       0.00000000058,    -0.00000000057,
      -0.00000007347,     0.00000002411,
       0.00000000233,    -0.00000000460,
       0.00000000060,     0.00000000058,
      -0.00000000330,     0.00000000436,
      -0.00000000436,    -0.00000000329,
       0.00000181378,    -0.00000023004,
      -0.00000020380,     0.00000051116,
      -0.00000006917,    -0.00000005644,
      -0.00000003050,    -0.00000012931,
       0.00000003795,     0.00000001162,
      -0.00000000391,     0.00000000509,
      -0.00000014393,    -0.00000015643,
       0.00000002803,    -0.00000003592,
       0.00000000791,     0.00000000517,
      -0.00000000076,     0.00000000114,
       0.00000011005,     0.00000007446,
      -0.00000005089,    -0.00000000082,
       0.00000000313,    -0.00000000967,
       0.00000000113,     0.00000000075,
       0.00000001490,    -0.00000001985,
      -0.00000000125,    -0.00000000114,
       0.00000002042,     0.00000001490,
       0.00000000105,    -0.00000000119,
       0.00000000018,     0.00000000116,
       0.00000004491,     0.00000000669,
      -0.00000000502,     0.00000001742,
       0.00000000244,    -0.00000001637,
       0.00000000634,     0.00000000183,
      -0.00000000055,     0.00000000124,
       0.00000000044,    -0.00000000075,
      -0.00000000603,    -0.00000001150,
      -0.00000000060,     0.00000000060,
      -0.00000000814,     0.00000000154,
       0.00000000071,    -0.00000000177,
      -0.00000000160,    -0.00000000812,
       0.00000000178,     0.00000000070,
      -0.00000000048,    -0.00000000059,
       0.00000000058,    -0.00000000048,
      -0.00000182593,    -0.00000062082,
       0.00000033176,    -0.00000073327,
       0.00000015081,     0.00000008653,
      -0.00000001456,     0.00000002013,
      -0.00000001178,     0.00000013527,
      -0.00000005327,    -0.00000001172,
       0.00000000400,    -0.00000000985,
       0.00000000124,     0.00000000075,
       0.00000027942,     0.00000011443,
      -0.00000005762,     0.00000011627,
      -0.00000002503,    -0.00000001334,
       0.00000000215,    -0.00000000377,
       0.00000011794,    -0.00000025355,
       0.00000010955,     0.00000005422,
      -0.00000001303,     0.00000002458,
      -0.00000000371,    -0.00000000209,
       0.00000000220,     0.00000000271,
      -0.00000000271,     0.00000000223,
       0.00001954167,     0.00000842184,
      -0.00000406697,     0.00000827800,
      -0.00000182723,    -0.00000099005,
       0.00000016026,    -0.00000027355,
       0.00000003198,     0.00000001922,
      -0.00000039392,     0.00000091744,
      -0.00000038867,    -0.00000019044,
       0.00000004641,    -0.00000008579,
       0.00000001285,     0.00000000729,
      -0.00000000091,     0.00000000151,
       0.00000001300,     0.00000000233,
      -0.00000000145,     0.00000000593,
      -0.00000000139,    -0.00000000043,
       0.00000000845,    -0.00000002343,
       0.00000001014,     0.00000000428,
      -0.00000000109,     0.00000000228,
      -0.00000000150,     0.00000000076,
      -0.00000000033,    -0.00000000071,
       0.00000000074,     0.00000000144,
      -0.00000000071,     0.00000000033,
       0.00000004955,     0.00000001298,
       0.00000000327,     0.00000000121,
      -0.00000000062,     0.00000000144,
       0.00000000121,    -0.00000000335,
       0.00000000147,     0.00000000063,
      -0.00000000135,    -0.00000000067,
      -0.00000000067,     0.00000000136,
       0.00000001496,     0.00000000015,
      -0.00000000001,    -0.00000000217,
      -0.00000014818,    -0.00000000178,
       0.00000000000,     0.00000000622,
      -0.00000000016,    -0.00000000130,
      -0.00000000105,    -0.00000000032,
       0.00000000530,     0.00000000156,
      -0.00000000083,     0.00000000235,
      -0.00000000156,     0.00000000531,
      -0.00000000235,    -0.00000000083,
      -0.00000000129,     0.00000000065,
      -0.00000000065,    -0.00000000129,
       0.00000000038,     0.00000000075,
       0.00000000075,    -0.00000000038,
       0.00000001207,     0.00000002318,
       0.00000000102,    -0.00000000045,
       0.00000000528,     0.00000000040,
       0.00000000038,    -0.00000000530,
      -0.00000001614,    -0.00000000120,
      -0.00000000017,     0.00000000234,
       0.00000015877,     0.00000001168,
       0.00000000049,    -0.00000000675,
       0.00000000009,     0.00000000138,
       0.00000000111,     0.00000000002,
      -0.00000000027,     0.00000003607,
       0.00000000156,    -0.00000000008,
      -0.00000017680,    -0.00000009203,
       0.00000004279,    -0.00000007127,
       0.00000000296,    -0.00000000567,
       0.00000000230,     0.00000000136,
       0.00000000091,    -0.00000000159,
      -0.00000000159,    -0.00000000091,
       0.00000000075,     0.00000000051,
       0.00000000122,     0.00000000209,
       0.00000000210,    -0.00000000125,
       0.00000003215,    -0.00000005827,
      -0.00000000231,    -0.00000000160,
      -0.00000000039,     0.00000000080,
      -0.00000001156,    -0.00000001971,
      -0.00000000300,     0.00000000215,
      -0.00000001972,     0.00000001161,
       0.00000000218,     0.00000000300,
       0.00000000252,     0.00000000383,
       0.00000000146,    -0.00000000094,
      -0.00000001157,    -0.00000001753,
      -0.00000000326,     0.00000000212,
       0.00000003604,     0.00000005998,
       0.00000000923,    -0.00000000664,
       0.00000001029,    -0.00000000539,
      -0.00000000107,    -0.00000000149,
       0.00000001728,     0.00000000051,
      -0.00000000006,    -0.00000001776,
      -0.00000042662,    -0.00000070950,
      -0.00000010908,     0.00000007855,
      -0.00000003100,     0.00000001612,
       0.00000000323,     0.00000000450,
      -0.00000016712,    -0.00000000229,
      -0.00000000155,     0.00000016772,
       0.00000061153,     0.00000000772,
       0.00000000084,    -0.00000008861,
       0.00000000149,     0.00000001015,
       0.00000000076,     0.00000000264,
       0.00000000007,     0.00000000119,
      -0.00000603260,    -0.00000007645,
      -0.00000000259,     0.00000025505,
      -0.00000000007,     0.00000000107,
       0.00000000281,    -0.00000005258,
       0.00000000409,     0.00000000383,
      -0.00000004246,    -0.00000000739,
       0.00000000284,    -0.00000000548,
      -0.00000002017,     0.00000021831,
      -0.00000002165,    -0.00000001506,
       0.00000002509,     0.00000000247,
      -0.00000000175,     0.00000000251,
       0.00000004086,     0.00000000023,
      -0.00000000053,     0.00000004243,
       0.00000015654,    -0.00000172722,
       0.00000017093,     0.00000011886,
      -0.00000007438,    -0.00000000738,
       0.00000000516,    -0.00000000743,
       0.00000000044,     0.00000000056,
      -0.00000003015,     0.00000000225,
      -0.00000000117,     0.00000000114,
       0.00000000295,    -0.00000004397,
       0.00000000116,     0.00000000082,
      -0.00000000181,     0.00000000274,
       0.00000000270,     0.00000000184,
      -0.00000000071,    -0.00000000040,
      -0.00000000083,     0.00000000048,
       0.00000000046,     0.00000000085,
      -0.00000003259,     0.00000004879,
       0.00000000327,     0.00000000183,
       0.00000000740,    -0.00000000426,
      -0.00000000406,    -0.00000000745,
      -0.00000001157,     0.00000000671,
       0.00000000120,     0.00000000193,
       0.00000000133,    -0.00000000016,
       0.00000000005,    -0.00000000127,
       0.00000012907,    -0.00000007435,
      -0.00000000537,    -0.00000000857,
      -0.00000001182,     0.00000000080,
       0.00000000013,     0.00000001159,
       0.00000002161,    -0.00000000091,
      -0.00000000015,    -0.00000000344,
      -0.00000019900,     0.00000001411,
       0.00000000070,     0.00000001466,
      -0.00000000240,    -0.00000000193,
      -0.00000000080,     0.00000000387,
       0.00000000036,    -0.00000000083,
      -0.00000004097,    -0.00000002286,
      -0.00000000164,     0.00000000364,
      -0.00000000148,    -0.00000000029,
      -0.00000000048,    -0.00000000123,
      -0.00000004810,    -0.00000002076,
      -0.00000000090,     0.00000000176,
      -0.00000000494,    -0.00000000003,
       0.00000000004,     0.00000000493,
      -0.00000000311,    -0.00000000159,
       0.00000000074,    -0.00000000127,
      -0.00000000159,     0.00000000311,
      -0.00000000127,    -0.00000000074,
      -0.00000003515,    -0.00000000049,
       0.00000000441,    -0.00000000660,
       0.00000000049,     0.00000000275,
      -0.00000000060,    -0.00000000051,
      -0.00000000118,     0.00000001451,
      -0.00000000141,    -0.00000000097,
      -0.00000001464,    -0.00000000121,
       0.00000000101,    -0.00000000145,
       0.00000028225,     0.00000005980,
      -0.00000004085,     0.00000005985,
       0.00000000377,    -0.00000001165,
       0.00000000223,     0.00000000220,
      -0.00000001199,    -0.00000001429,
       0.00000000360,    -0.00000000415,
       0.00000000088,     0.00000000075,
       0.00000000603,     0.00000001364,
      -0.00000000553,    -0.00000000160,
       0.00000000075,    -0.00000000088,
       0.00000001266,     0.00000000189,
      -0.00000000141,     0.00000000489,
       0.00000000066,    -0.00000000441,
       0.00000000171,     0.00000000050,
      -0.00000000240,    -0.00000000034,
       0.00000000035,    -0.00000000240,
       0.00000013855,     0.00000006724,
      -0.00000003196,     0.00000005603,
      -0.00000001174,    -0.00000000750,
      -0.00000000321,     0.00000000671,
      -0.00000000273,    -0.00000000152,
       0.00000000093,     0.00000000040,
       0.00000000042,    -0.00000000101,
       0.00000184019,     0.00000007853,
      -0.00000010836,     0.00000072949,
      -0.00000014739,    -0.00000003734,
      -0.00000000033,     0.00000001742,
      -0.00000000693,    -0.00000000087,
       0.00000000015,    -0.00000000102,
       0.00000032026,     0.00000013760,
      -0.00000006654,     0.00000013535,
      -0.00000002966,    -0.00000001570,
       0.00000000253,    -0.00000000447,
       0.00000013688,    -0.00000031742,
       0.00000013477,     0.00000006607,
      -0.00000001569,     0.00000002964,
      -0.00000000447,    -0.00000000253,
       0.00000000079,     0.00000000275,
       0.00000000276,    -0.00000000078,
      -0.00000000031,     0.00000000142,
       0.00000000142,     0.00000000031,
       0.00000001381,     0.00000005152,
       0.00000000302,    -0.00000000128,
      -0.00000000090,     0.00000000018,
       0.00000000259,    -0.00000001222,
      -0.00000000091,     0.00000000228,
      -0.00000001219,    -0.00000000271,
       0.00000000227,     0.00000000093,
      -0.00000000357,     0.00000001534,
       0.00000000312,     0.00000000030,
       0.00000000330,    -0.00000000412,
      -0.00000000402,    -0.00000000340,
       0.00000005148,    -0.00000023105,
      -0.00000001752,     0.00000004254,
      -0.00000001461,    -0.00000000155,
       0.00000000286,     0.00000000090,
      -0.00000002921,     0.00000003598,
       0.00000000463,    -0.00000000376,
       0.00000003552,     0.00000002960,
      -0.00000000359,    -0.00000000473,
       0.00000004280,    -0.00000005101,
      -0.00000000663,     0.00000000520,
      -0.00000000998,    -0.00000000697,
       0.00000000112,     0.00000000125,
      -0.00000001311,     0.00000000617,
       0.00000000090,    -0.00000000014,
       0.00000000536,     0.00000001325,
      -0.00000055692,     0.00000067204,
       0.00000008734,    -0.00000006871,
       0.00000004586,     0.00000003208,
      -0.00000000505,    -0.00000000568,
       0.00000011711,    -0.00000005281,
      -0.00000000809,     0.00000000079,
      -0.00000005010,    -0.00000011744,
       0.00000000014,     0.00000000801,
      -0.00000000046,    -0.00000000068,
      -0.00000019419,     0.00000008523,
       0.00000001314,    -0.00000000097,
       0.00000001576,     0.00000003240,
      -0.00000000024,    -0.00000000246,
       0.00000002176,    -0.00000000045,
       0.00000000269,    -0.00000002034,
       0.00000216686,    -0.00000095102,
      -0.00000014690,     0.00000001067,
      -0.00000007042,    -0.00000014423,
       0.00000000100,     0.00000001078,
      -0.00000019348,    -0.00000000608,
      -0.00000000196,     0.00000000212,
      -0.00000001557,     0.00000018884,
       0.00000000330,     0.00000000302,
       0.00000037005,     0.00000001514,
       0.00000000353,    -0.00000005902,
      -0.00000000086,    -0.00000000063,
      -0.00000000132,     0.00000000630,
       0.00000000016,    -0.00000000885,
      -0.00000343791,    -0.00000014161,
      -0.00000003988,     0.00000004495,
      -0.00000001511,     0.00000025251,
       0.00000000366,     0.00000000300,
      -0.00000001869,    -0.00000003423,
       0.00000000084,     0.00000000213,
      -0.00000001870,     0.00000004511,
       0.00000000176,    -0.00000000312,
       0.00000000049,    -0.00000000119,
      -0.00000000046,    -0.00000000100,
       0.00000005636,     0.00000006237,
       0.00000000769,    -0.00000000852,
      -0.00000000059,     0.00000000058,
       0.00000001173,     0.00000000122,
      -0.00000000018,     0.00000001182,
      -0.00000042571,    -0.00000046392,
       0.00000003646,     0.00000003164,
      -0.00000003381,     0.00000003783,
       0.00000000247,    -0.00000000252,
      -0.00000002304,    -0.00000000491,
      -0.00000000232,    -0.00000002571,
      -0.00000000367,     0.00000000811,
       0.00000000106,     0.00000000014,
       0.00000000152,     0.00000000188,
      -0.00000000184,     0.00000000159,
       0.00000003091,    -0.00000004275,
      -0.00000000420,    -0.00000000179,
      -0.00000000469,    -0.00000000382,
       0.00000000038,    -0.00000000091,
       0.00000000232,    -0.00000000034,
      -0.00000000089,    -0.00000000037,
      -0.00000002439,    -0.00000001100,
      -0.00000000029,     0.00000000194,
       0.00000000367,     0.00000000189,
      -0.00000000088,     0.00000000151,
       0.00000000182,    -0.00000000355,
       0.00000000145,     0.00000000085,
       0.00000000211,    -0.00000003411,
      -0.00000000146,    -0.00000000009,
      -0.00000000098,     0.00000000229,
      -0.00000000097,    -0.00000000047,
       0.00000000035,    -0.00000000081,
       0.00003717030,     0.00001599170,
      -0.00000772603,     0.00001574625,
      -0.00000347565,    -0.00000188105,
       0.00000030443,    -0.00000052047,
       0.00000006029,     0.00000003755,
      -0.00000050255,     0.00000116775,
      -0.00000049467,    -0.00000024278,
       0.00000005910,    -0.00000010918,
       0.00000001636,     0.00000000927,
      -0.00000000114,     0.00000000190,
       0.00000001378,    -0.00000000269,
       0.00000000055,     0.00000000538,
      -0.00000000108,     0.00000000001,
      -0.00000000575,     0.00000001654,
      -0.00000000649,    -0.00000000319,
       0.00000000083,    -0.00000000130,
       0.00000000257,     0.00000000098,
      -0.00000000053,     0.00000000097,
       0.00000000098,    -0.00000000257,
       0.00000000097,     0.00000000053,
       0.00000000126,    -0.00000000046,
      -0.00000000044,    -0.00000000125,
       0.00000002240,    -0.00000000814,
      -0.00000000059,    -0.00000000151,
      -0.00000000140,     0.00000000015,
       0.00000000004,     0.00000000131,
      -0.00000002273,     0.00000000213,
       0.00000000012,     0.00000000167,
      -0.00000001492,     0.00000002128,
      -0.00000000063,    -0.00000000056,
      -0.00000000041,    -0.00000000130,
       0.00000000130,    -0.00000000041,
       0.00000000239,     0.00000000091,
      -0.00000000050,     0.00000000090,
      -0.00000000091,     0.00000000239,
      -0.00000000090,    -0.00000000050,
       0.00000000056,    -0.00000000070,
      -0.00000000070,    -0.00000000057,
       0.00000001194,    -0.00000001380,
      -0.00000000092,    -0.00000000067,
      -0.00000000211,     0.00000000101,
       0.00000000097,     0.00000000211,
      -0.00000004196,     0.00000001913,
       0.00000000137,     0.00000000268,
       0.00000000334,    -0.00000000019,
       0.00000000000,    -0.00000000324,
       0.00000000003,     0.00000000103,
       0.00000006291,    -0.00000000281,
      -0.00000000014,    -0.00000000448,
       0.00000020085,     0.00000007333,
      -0.00000003840,     0.00000008824,
      -0.00000000222,     0.00000000603,
      -0.00000000266,    -0.00000000114,
      -0.00000000008,    -0.00000000113,
      -0.00000000112,     0.00000000010,
      -0.00000000195,    -0.00000002279,
      -0.00000000138,     0.00000000027,
      -0.00000000206,     0.00000000394,
       0.00000000054,    -0.00000000054,
       0.00000000391,     0.00000000211,
      -0.00000000055,    -0.00000000054,
      -0.00000000114,    -0.00000000041,
      -0.00000000149,     0.00000000086,
       0.00000000082,     0.00000000154,
      -0.00000004109,     0.00000008652,
       0.00000001100,    -0.00000001201,
       0.00000000544,     0.00000000204,
      -0.00000000081,    -0.00000000066,
       0.00000001298,    -0.00000000724,
      -0.00000000169,    -0.00000000015,
      -0.00000000700,    -0.00000001311,
      -0.00000000021,     0.00000000171,
      -0.00000001868,     0.00000001195,
       0.00000000254,     0.00000000354,
       0.00000000392,     0.00000000082,
       0.00000000109,    -0.00000000392,
       0.00000027860,    -0.00000017993,
      -0.00000003565,     0.00000000173,
      -0.00000001199,    -0.00000001670,
       0.00000000018,     0.00000000238,
      -0.00000003450,    -0.00000000799,
       0.00000000047,     0.00000000354,
      -0.00000000914,     0.00000003439,
       0.00000000377,    -0.00000000036,
       0.00000006153,     0.00000000551,
       0.00000000105,    -0.00000001136,
      -0.00000000098,     0.00000000032,
       0.00000000042,    -0.00000000664,
      -0.00000000075,     0.00000000051,
      -0.00000000754,    -0.00000000125,
       0.00000000059,     0.00000000091,
      -0.00000080060,    -0.00000007443,
       0.00000001651,     0.00000006347,
      -0.00000000481,     0.00000005225,
       0.00000000447,    -0.00000000140,
      -0.00000000589,     0.00000006220,
       0.00000000713,    -0.00000000484,
       0.00000006534,     0.00000000890,
      -0.00000000502,    -0.00000000770,
      -0.00000001793,    -0.00000010152,
      -0.00000001235,     0.00000000793,
      -0.00000001897,     0.00000000313,
       0.00000000149,     0.00000000218,
      -0.00000002057,     0.00000000028,
       0.00000000053,     0.00000002447,
       0.00000018600,     0.00000113819,
       0.00000013804,    -0.00000008903,
       0.00000008459,    -0.00000001362,
      -0.00000000650,    -0.00000000979,
       0.00000019797,    -0.00000000392,
       0.00000000098,     0.00000000314,
      -0.00000000466,    -0.00000020996,
       0.00000000230,    -0.00000000045,
      -0.00000038729,     0.00000000557,
       0.00000000098,     0.00000006165,
      -0.00000000337,    -0.00000000669,
       0.00000361211,    -0.00000005376,
       0.00000000491,     0.00000004864,
      -0.00000000348,    -0.00000026491,
       0.00000000341,    -0.00000000049,
       0.00000001895,     0.00000003471,
      -0.00000000221,    -0.00000000209,
       0.00000001711,    -0.00000001067,
      -0.00000000076,     0.00000000422,
      -0.00000000040,     0.00000000102,
       0.00000000078,     0.00000000095,
      -0.00000002247,    -0.00000006135,
       0.00000000736,     0.00000000371,
      -0.00000000749,     0.00000000343,
       0.00000000045,    -0.00000000092,
      -0.00000001192,    -0.00000000010,
      -0.00000000025,    -0.00000001227,
       0.00000016127,     0.00000045457,
      -0.00000005379,    -0.00000002978,
       0.00000003294,    -0.00000001519,
      -0.00000000191,     0.00000000402,
       0.00000002355,     0.00000000137,
       0.00000000153,     0.00000002618,
       0.00000000936,    -0.00000001454,
       0.00000000095,     0.00000000019,
      -0.00000000049,    -0.00000000182,
       0.00000000182,    -0.00000000051,
      -0.00000001748,     0.00000001290,
       0.00000000130,     0.00000000176,
       0.00000000233,     0.00000000293,
      -0.00000000295,    -0.00000000080,
      -0.00000001682,    -0.00000001143,
       0.00000000039,    -0.00000000265,
       0.00000000108,     0.00000000030,
      -0.00000000408,    -0.00000000154,
       0.00000000078,    -0.00000000181,
      -0.00000000153,     0.00000000418,
      -0.00000000187,    -0.00000000081,
      -0.00000015438,    -0.00000005653,
       0.00000002947,    -0.00000006765,
       0.00000001527,     0.00000000747,
       0.00000000259,    -0.00000000702,
       0.00000000308,     0.00000000133,
      -0.00000000034,     0.00000000069,
       0.00000000143,    -0.00000000094,
      -0.00000000091,    -0.00000000144,
       0.00000003332,    -0.00000002088,
      -0.00000000136,    -0.00000000193,
      -0.00000000302,     0.00000000051,
       0.00000000037,     0.00000000297,
      -0.00000000014,    -0.00000000094,
      -0.00000006816,     0.00000000957,
       0.00000000065,     0.00000000423,
       0.00000000161,    -0.00000000031,
       0.00000000001,    -0.00000000124,
       0.00000002785,    -0.00000000533,
      -0.00000000024,    -0.00000000204,
       0.00000000136,    -0.00000000049,
      -0.00000000044,    -0.00000000137,
       0.00000003313,    -0.00000001337,
      -0.00000000084,    -0.00000000191,
      -0.00000000234,    -0.00000000028,
      -0.00000000041,     0.00000000227,
      -0.00000006248,    -0.00000000356,
      -0.00000000027,     0.00000000378,
       0.00000001304,     0.00000001416,
       0.00000000121,    -0.00000000087,
       0.00000000325,     0.00000000208,
       0.00000000205,    -0.00000000418,
      -0.00000000851,    -0.00000000406,
      -0.00000000064,     0.00000000135,
       0.00000007947,     0.00000003755,
       0.00000000273,    -0.00000000581,
       0.00000000014,    -0.00000000956,
      -0.00000003017,     0.00000002167,
      -0.00000001130,    -0.00000001574,
       0.00000000615,    -0.00000000244,
      -0.00000000275,    -0.00000000655,
      -0.00000000028,    -0.00000000087,
      -0.00000000087,     0.00000000028,
      -0.00000001363,     0.00000002756,
       0.00000000004,     0.00000000346,
      -0.00000000178,    -0.00000000569,
      -0.00000000570,     0.00000000176,
      -0.00000006311,    -0.00000020875,
       0.00000000762,    -0.00000000226,
      -0.00000000001,     0.00000000006,
      -0.00000000006,    -0.00000000001,
       0.00000000045,     0.00000000020,
      -0.00000000020,     0.00000000045,
       0.00000005374,     0.00000001389,
      -0.00000001685,     0.00000004094,
      -0.00000001554,    -0.00000000895,
       0.00000000294,    -0.00000000386,
       0.00000000512,    -0.00000001780,
       0.00000001349,     0.00000000595,
      -0.00000000308,     0.00000000511,
      -0.00000000126,    -0.00000000098,
      -0.00000000050,    -0.00000000134,
       0.00000000113,    -0.00000000010,
       0.00000000117,    -0.00000000159,
       0.00000000107,     0.00000000103,
      -0.00000024608,    -0.00000006078,
       0.00000007532,    -0.00000018789,
       0.00000007134,     0.00000004036,
      -0.00000001341,     0.00000001776,
      -0.00000001173,     0.00000004061,
      -0.00000003079,    -0.00000001358,
       0.00000000704,    -0.00000001170,
       0.00000000290,     0.00000000229,
      -0.00000000100,     0.00000000098,
      -0.00000000096,    -0.00000000122,
      -0.00000000030,     0.00000000304,
      -0.00000000247,    -0.00000000055,
       0.00000000000,    -0.00000000057,
      -0.00000000065,     0.00000000033,
       0.00000000117,     0.00000000020,
       0.00000000023,    -0.00000000124,
      -0.00000000308,    -0.00000000304,
       0.00000000092,     0.00000000108,
      -0.00000000306,     0.00000000305,
       0.00000000109,    -0.00000000092,
       0.00000000064,    -0.00000000090,
      -0.00000000082,    -0.00000000250,
      -0.00000000253,     0.00000000078,
      -0.00000006720,    -0.00000006796,
       0.00000001981,     0.00000002461,
      -0.00000000320,     0.00000000435,
       0.00000000124,    -0.00000000132,
       0.00000000701,     0.00000002166,
      -0.00000000131,    -0.00000000563,
      -0.00000000024,     0.00000000080,
       0.00000002170,    -0.00000000680,
      -0.00000000563,     0.00000000121,
       0.00000000080,     0.00000000025,
       0.00000000077,    -0.00000000052,
      -0.00000000050,    -0.00000000078,
       0.00000001791,    -0.00000000642,
      -0.00000000038,    -0.00000000100,
      -0.00000000100,    -0.00000000018,
      -0.00000000026,     0.00000000094,
      -0.00000000910,    -0.00000003068,
      -0.00000000547,     0.00000000257,
       0.00000000154,    -0.00000000052,
      -0.00000000182,     0.00000001211,
       0.00000000049,    -0.00000000220,
       0.00000001201,     0.00000000214,
      -0.00000000215,    -0.00000000059,
       0.00000015045,     0.00000048157,
      -0.00000002706,    -0.00000012376,
      -0.00000000549,     0.00000001676,
       0.00000002649,    -0.00000001232,
      -0.00000000733,     0.00000000242,
       0.00000000107,     0.00000000023,
       0.00000001661,    -0.00000010557,
      -0.00000000460,     0.00000001918,
       0.00000000165,    -0.00000000136,
      -0.00000010515,    -0.00000001775,
       0.00000001895,     0.00000000499,
      -0.00000000128,    -0.00000000169,
      -0.00000000149,    -0.00000000001,
      -0.00000000010,     0.00000000146,
       0.00000001860,    -0.00000001201,
      -0.00000000074,    -0.00000000104,
      -0.00000000150,     0.00000000031,
       0.00000000021,     0.00000000145,
      -0.00000002818,    -0.00000000296,
      -0.00000000021,     0.00000000164,
      -0.00000002981,     0.00000016629,
       0.00000000778,    -0.00000002976,
       0.00000002979,     0.00000000178,
      -0.00000000575,    -0.00000000098,
       0.00000002684,    -0.00000003893,
      -0.00000000341,     0.00000000442,
      -0.00000003747,    -0.00000002804,
       0.00000000393,     0.00000000370,
      -0.00000000078,    -0.00000000027,
      -0.00000000045,     0.00000000058,
       0.00000039492,    -0.00000234932,
      -0.00000010791,     0.00000042293,
       0.00000003685,    -0.00000002880,
      -0.00000014089,    -0.00000000856,
       0.00000002702,     0.00000000467,
      -0.00000000203,    -0.00000000223,
      -0.00000024056,     0.00000034058,
       0.00000003126,    -0.00000003805,
      -0.00000000376,     0.00000000009,
       0.00000033518,     0.00000024487,
      -0.00000003634,    -0.00000003220,
      -0.00000000021,     0.00000000375,
       0.00000000363,     0.00000000173,
       0.00000000202,    -0.00000000326,
      -0.00000000115,     0.00000000057,
       0.00000000053,     0.00000000115,
      -0.00000003488,     0.00000000600,
       0.00000000037,     0.00000000208,
       0.00000000123,     0.00000000105,
       0.00000000106,    -0.00000000173,
       0.00000044057,    -0.00000060134,
      -0.00000005631,     0.00000006589,
      -0.00000010550,    -0.00000006533,
       0.00000001226,     0.00000000953,
      -0.00000000004,    -0.00000000124,
      -0.00000011682,     0.00000006877,
       0.00000000610,    -0.00000000384,
       0.00000005850,     0.00000011820,
      -0.00000000167,    -0.00000000604,
       0.00000000121,     0.00000000082,
       0.00000000103,    -0.00000000009,
      -0.00000544746,     0.00000752789,
       0.00000070440,    -0.00000082714,
      -0.00000008336,    -0.00000000079,
       0.00000048658,     0.00000030000,
      -0.00000005638,    -0.00000004377,
       0.00000000016,     0.00000000561,
       0.00000105263,    -0.00000058562,
      -0.00000005612,     0.00000002858,
       0.00000000159,     0.00000000200,
      -0.00000055158,    -0.00000105668,
       0.00000002142,     0.00000005579,
       0.00000000261,    -0.00000000106,
      -0.00000000460,    -0.00000000562,
      -0.00000000591,     0.00000000278,
      -0.00000002389,     0.00000001129,
       0.00000000076,     0.00000000148,
       0.00000000165,    -0.00000000015,
      -0.00000000003,    -0.00000000157,
       0.00000003595,     0.00000002202,
       0.00000000140,    -0.00000000246,
      -0.00000216287,     0.00000116634,
       0.00000011315,    -0.00000005293,
       0.00000019515,     0.00000033489,
      -0.00000000893,    -0.00000001969,
      -0.00000000083,     0.00000000052,
       0.00000021409,    -0.00000003587,
       0.00000000309,     0.00000000382,
       0.00000000856,    -0.00000019414,
      -0.00000000174,    -0.00000000670,
      -0.00000000156,    -0.00000000097,
       0.00000000021,    -0.00000000173,
       0.00002346261,    -0.00001266224,
      -0.00000122831,     0.00000057392,
       0.00000003086,     0.00000004923,
      -0.00000087602,    -0.00000149427,
       0.00000003982,     0.00000008717,
       0.00000000372,    -0.00000000219,
      -0.00000189223,     0.00000017604,
      -0.00000003538,    -0.00000001654,
       0.00000000106,     0.00000000385,
       0.00000004613,     0.00000183192,
      -0.00000000041,     0.00000004611,
       0.00000000371,    -0.00000000259,
       0.00000000167,     0.00000000796,
       0.00000000600,     0.00000000352,
      -0.00000000037,    -0.00000000094,
       0.00000003245,    -0.00000000239,
      -0.00000000013,    -0.00000000221,
       0.00000444591,    -0.00000032291,
       0.00000009391,     0.00000002453,
      -0.00000000430,    -0.00000000869,
      -0.00000003442,    -0.00000066847,
       0.00000000268,    -0.00000001282,
      -0.00000000138,     0.00000000057,
      -0.00000004938,     0.00000009712,
       0.00000000011,    -0.00000000483,
      -0.00000000265,    -0.00000009806,
      -0.00000000002,     0.00000000979,
       0.00000000154,     0.00000000308,
      -0.00000000175,     0.00000000255,
      -0.00004134119,     0.00000299696,
      -0.00000087499,    -0.00000022369,
       0.00000003938,     0.00000008035,
       0.00000015850,     0.00000288212,
      -0.00000001224,     0.00000005886,
       0.00000000588,    -0.00000000258,
      -0.00000002439,    -0.00000052352,
       0.00000002932,     0.00000002589,
      -0.00000000241,     0.00000000404,
      -0.00000027482,     0.00000039779,
       0.00000001400,    -0.00000005386,
       0.00000000293,     0.00000000306,
       0.00000000422,    -0.00000001472,
      -0.00000000626,    -0.00000001104,
      -0.00000000076,     0.00000000075,
       0.00000043144,     0.00000116437,
      -0.00000011968,    -0.00000006179,
       0.00000000798,    -0.00000000965,
       0.00000014062,    -0.00000006149,
      -0.00000000767,     0.00000001434,
      -0.00000000117,    -0.00000000098,
       0.00000018076,    -0.00000000615,
       0.00000000364,    -0.00000000055,
       0.00000001445,     0.00000018979,
      -0.00000000215,     0.00000000345,
      -0.00000000036,    -0.00000000077,
      -0.00000337222,    -0.00000900678,
       0.00000093176,     0.00000046928,
      -0.00000006066,     0.00000007549,
      -0.00000062315,     0.00000027820,
       0.00000003289,    -0.00000006395,
       0.00000000521,     0.00000000429,
      -0.00000035014,    -0.00000000839,
      -0.00000000939,     0.00000001009,
      -0.00000005403,    -0.00000042914,
       0.00000001355,    -0.00000000572,
      -0.00000000162,     0.00000000143,
       0.00000000276,     0.00000000599,
      -0.00000000472,    -0.00000000212,
       0.00000000031,     0.00000000069,
      -0.00000014323,     0.00000007745,
       0.00000000867,    -0.00000003580,
       0.00000000919,     0.00000001291,
      -0.00000000350,    -0.00000000051,
       0.00000001380,     0.00000004858,
      -0.00000000481,    -0.00000000276,
      -0.00000004908,     0.00000001451,
       0.00000000268,    -0.00000000511,
      -0.00000000148,     0.00000000200,
      -0.00000000200,    -0.00000000148,
       0.00000096248,    -0.00000044883,
      -0.00000007065,     0.00000024890,
      -0.00000002605,    -0.00000002278,
      -0.00000004438,    -0.00000005522,
       0.00000001479,     0.00000000125,
      -0.00000000143,     0.00000000166,
      -0.00000002171,    -0.00000008431,
       0.00000000751,     0.00000000489,
       0.00000008491,    -0.00000002652,
      -0.00000000440,     0.00000000970,
      -0.00000000085,    -0.00000000068,
       0.00000000660,    -0.00000000903,
      -0.00000000054,    -0.00000000056,
       0.00000000951,     0.00000000661,
       0.00000000651,    -0.00000003158,
       0.00000000099,     0.00000000020,
      -0.00000000003,    -0.00000000010,
       0.00000000000,    -0.00000000035,
       0.00000000035,     0.00000000000,
      -0.00000000120,     0.00000000036,
      -0.00000000036,    -0.00000000120,
       0.00000000258,    -0.00000001186,
       0.00000000565,     0.00000000323,
      -0.00000000434,    -0.00000000079,
       0.00000000107,    -0.00000000212,
       0.00000000161,     0.00000000180,
      -0.00000000126,     0.00000000079,
       0.00000000148,    -0.00000000188,
       0.00000000098,     0.00000000138,
      -0.00000002147,     0.00000005709,
      -0.00000002805,    -0.00000001797,
       0.00000000566,    -0.00000000491,
       0.00000000897,     0.00000000266,
      -0.00000000257,     0.00000000491,
      -0.00000000088,    -0.00000000087,
      -0.00000001702,    -0.00000000698,
       0.00000000716,    -0.00000000577,
       0.00000000085,     0.00000000212,
      -0.00000000753,    -0.00000000187,
      -0.00000000191,    -0.00000000445,
       0.00000000179,    -0.00000000068,
       0.00000003368,     0.00000004181,
      -0.00000001908,    -0.00000000282,
       0.00000000736,    -0.00000000454,
       0.00000000055,     0.00000000294,
       0.00000000265,    -0.00000000613,
       0.00000000043,     0.00000000087,
       0.00000000639,     0.00000000272,
      -0.00000000094,     0.00000000049,
       0.00000000161,     0.00000000142,
      -0.00000000142,     0.00000000165,
      -0.00000000671,    -0.00000000320,
       0.00000000144,    -0.00000000124,
       0.00000000320,    -0.00000000671,
       0.00000000124,     0.00000000144,
      -0.00000028194,    -0.00000013500,
       0.00000006010,    -0.00000005183,
       0.00000000419,    -0.00000000878,
       0.00000000162,     0.00000000188,
      -0.00000000460,     0.00000001689,
      -0.00000000162,    -0.00000000005,
      -0.00000000002,     0.00000000162,
       0.00000000173,     0.00000000014,
       0.00000000014,    -0.00000000174,
       0.00000005450,     0.00000000446,
       0.00000000018,    -0.00000000223,
       0.00000000039,     0.00000000067,
       0.00000000067,    -0.00000000040,
       0.00000001165,    -0.00000002145,
      -0.00000000083,    -0.00000000056,
      -0.00000000373,    -0.00000000635,
      -0.00000000097,     0.00000000068,
      -0.00000000636,     0.00000000375,
       0.00000000070,     0.00000000097,
       0.00000001215,     0.00000002024,
       0.00000000324,    -0.00000000172,
       0.00000000480,     0.00000000021,
      -0.00000000002,    -0.00000000499,
      -0.00000013983,    -0.00000023277,
      -0.00000003602,     0.00000002544,
      -0.00000000979,     0.00000000518,
       0.00000000101,     0.00000000144,
      -0.00000004687,    -0.00000000101,
      -0.00000000073,     0.00000004709,
       0.00000017491,     0.00000000348,
       0.00000000038,    -0.00000002418,
       0.00000000054,     0.00000000355,
       0.00000000035,     0.00000000095,
      -0.00000172516,    -0.00000003445,
      -0.00000000118,     0.00000007023,
       0.00000000106,    -0.00000001908,
       0.00000000151,     0.00000000137,
      -0.00000001568,    -0.00000000267,
       0.00000000104,    -0.00000000199,
      -0.00000000749,     0.00000007810,
      -0.00000000775,    -0.00000000534,
       0.00000000888,     0.00000000089,
      -0.00000000061,     0.00000000089,
       0.00000001195,     0.00000000013,
      -0.00000000026,     0.00000001268,
       0.00000006086,    -0.00000064260,
       0.00000006371,     0.00000004391,
      -0.00000002648,    -0.00000000268,
       0.00000000182,    -0.00000000265,
      -0.00000000814,     0.00000000040,
      -0.00000000056,     0.00000000055,
       0.00000000076,    -0.00000001419,
      -0.00000001703,    -0.00000000283,
      -0.00000000027,     0.00000000153,
      -0.00000000046,     0.00000000542,
      -0.00000000550,    -0.00000000047,
       0.00000011237,     0.00000001840,
      -0.00000001735,     0.00000002076,
       0.00000000084,    -0.00000000474,
       0.00000000087,     0.00000000074,
       0.00000000026,    -0.00000000374,
       0.00000000438,     0.00000000044,
      -0.00000000117,     0.00000000172,
       0.00000000169,     0.00000000119,
      -0.00000002225,     0.00000003241,
       0.00000000209,     0.00000000123,
       0.00000000432,    -0.00000000247,
      -0.00000000230,    -0.00000000435,
      -0.00000000736,     0.00000000420,
       0.00000000070,     0.00000000115,
       0.00000007979,    -0.00000004537,
      -0.00000000311,    -0.00000000510,
      -0.00000000609,     0.00000000052,
       0.00000000002,     0.00000000589,
       0.00000001175,    -0.00000000093,
      -0.00000000008,    -0.00000000178,
      -0.00000010899,     0.00000000851,
       0.00000000037,     0.00000000762,
      -0.00000000138,    -0.00000000140,
      -0.00000000065,     0.00000000242,
      -0.00000002781,    -0.00000001810,
      -0.00000000124,     0.00000000229,
      -0.00000005081,    -0.00000000173,
      -0.00000000004,     0.00000000206,
      -0.00000000116,    -0.00000000018,
       0.00000000018,    -0.00000000116,
       0.00000000049,     0.00000000185,
       0.00000000185,    -0.00000000048,
      -0.00000000019,     0.00000000090,
       0.00000000090,     0.00000000020,
       0.00000000921,     0.00000003708,
       0.00000000206,    -0.00000000080,
       0.00000000174,    -0.00000000785,
      -0.00000000058,     0.00000000144,
      -0.00000000781,    -0.00000000184,
       0.00000000142,     0.00000000060,
      -0.00000000270,     0.00000001113,
       0.00000000203,     0.00000000024,
       0.00000000197,    -0.00000000250,
      -0.00000000241,    -0.00000000205,
       0.00000003668,    -0.00000015722,
      -0.00000001188,     0.00000002830,
      -0.00000000950,    -0.00000000119,
       0.00000000181,     0.00000000060,
      -0.00000001770,     0.00000002176,
       0.00000000272,    -0.00000000222,
       0.00000002136,     0.00000001800,
      -0.00000000208,    -0.00000000278,
       0.00000002873,    -0.00000003417,
      -0.00000000607,    -0.00000000435,
       0.00000000066,     0.00000000075,
      -0.00000000720,     0.00000000354,
       0.00000000289,     0.00000000726,
      -0.00000035612,     0.00000042784,
       0.00000005416,    -0.00000004254,
       0.00000002788,     0.00000001998,
      -0.00000000297,    -0.00000000339,
       0.00000006480,    -0.00000002984,
      -0.00000000419,     0.00000000045,
      -0.00000002774,    -0.00000006491,
      -0.00000000003,     0.00000000412,
      -0.00000011665,     0.00000005207,
       0.00000000880,     0.00000001817,
      -0.00000000010,    -0.00000000129,
       0.00000001086,    -0.00000000082,
       0.00000000146,    -0.00000000966,
       0.00000126526,    -0.00000056480,
      -0.00000008026,     0.00000000575,
      -0.00000003944,    -0.00000008104,
       0.00000000046,     0.00000000564,
      -0.00000009563,    -0.00000000010,
      -0.00000000158,     0.00000000074,
      -0.00000000673,     0.00000009199,
       0.00000000151,     0.00000000234,
       0.00000019356,     0.00000000356,
       0.00000000126,    -0.00000002911,
      -0.00000000130,     0.00000000413,
       0.00000000016,    -0.00000000564,
      -0.00000179855,    -0.00000003407,
      -0.00000003446,     0.00000001936,
      -0.00000000527,     0.00000012536,
       0.00000000157,     0.00000000244,
      -0.00000000987,    -0.00000002303,
       0.00000000065,     0.00000000135,
      -0.00000001314,     0.00000002743,
       0.00000000116,    -0.00000000203,
       0.00000003543,     0.00000004375,
       0.00000000534,    -0.00000000512,
       0.00000000673,     0.00000000067,
       0.00000000014,     0.00000000688,
      -0.00000027650,    -0.00000033789,
       0.00000002584,     0.00000002289,
      -0.00000002358,     0.00000002286,
       0.00000000162,    -0.00000000175,
      -0.00000001375,    -0.00000000372,
      -0.00000000274,    -0.00000001489,
      -0.00000000305,     0.00000000710,
       0.00000000085,     0.00000000020,
       0.00000000106,     0.00000000149,
      -0.00000000147,     0.00000000113,
       0.00000001988,    -0.00000004394,
      -0.00000000386,    -0.00000000095,
      -0.00000000160,    -0.00000000265,
       0.00000000250,    -0.00000000211,
      -0.00000000241,    -0.00000000115,
      -0.00000000115,     0.00000000241,
      -0.00000000857,     0.00000003220,
      -0.00000000102,    -0.00000000027,
      -0.00000000132,     0.00000000221,
       0.00000000219,     0.00000000136,
      -0.00000000084,     0.00000000041,
       0.00000000037,     0.00000000088,
      -0.00000002831,     0.00000005192,
       0.00000000310,     0.00000000138,
       0.00000000741,    -0.00000000340,
      -0.00000000093,    -0.00000000025,
      -0.00000000321,    -0.00000000748,
      -0.00000000029,     0.00000000094,
      -0.00000001192,     0.00000000647,
       0.00000000124,     0.00000000206,
       0.00000000190,     0.00000000068,
       0.00000000089,    -0.00000000187,
       0.00000016671,    -0.00000009113,
      -0.00000002044,    -0.00000000250,
      -0.00000000582,    -0.00000000969,
      -0.00000000013,     0.00000000132,
      -0.00000001674,    -0.00000000669,
      -0.00000000001,     0.00000000220,
      -0.00000000749,     0.00000001656,
       0.00000000234,     0.00000000010,
       0.00000003277,     0.00000000758,
       0.00000000136,    -0.00000000560,
       0.00000000100,    -0.00000000359,
      -0.00000000415,    -0.00000000164,
      -0.00000040584,    -0.00000009625,
       0.00000000284,     0.00000004220,
      -0.00000000619,     0.00000002575,
       0.00000000284,    -0.00000000035,
      -0.00000001074,     0.00000003407,
       0.00000000390,    -0.00000000259,
       0.00000003589,     0.00000001290,
      -0.00000000266,    -0.00000000426,
       0.00000000738,    -0.00000006052,
      -0.00000000742,     0.00000000461,
      -0.00000001045,    -0.00000000108,
       0.00000000080,     0.00000000123,
      -0.00000001002,     0.00000000032,
       0.00000000015,     0.00000001241,
      -0.00000008698,     0.00000065934,
       0.00000008066,    -0.00000005020,
       0.00000004667,     0.00000000499,
      -0.00000000347,    -0.00000000548,
       0.00000009872,    -0.00000000260,
       0.00000000100,     0.00000000173,
      -0.00000000208,    -0.00000010572,
       0.00000000109,    -0.00000000078,
      -0.00000020698,     0.00000000347,
       0.00000000047,     0.00000003110,
      -0.00000000148,    -0.00000000440,
       0.00000192837,    -0.00000003214,
       0.00000001233,     0.00000002679,
      -0.00000000167,    -0.00000013458,
       0.00000000177,    -0.00000000090,
       0.00000001056,     0.00000002359,
      -0.00000000155,    -0.00000000137,
       0.00000001249,    -0.00000000882,
      -0.00000000057,     0.00000000285,
      -0.00000001638,    -0.00000004381,
      -0.00000000528,     0.00000000238,
      -0.00000000696,    -0.00000000014,
      -0.00000000020,    -0.00000000728,
       0.00000012744,     0.00000033620,
      -0.00000004050,    -0.00000001977,
       0.00000002339,    -0.00000001061,
      -0.00000000132,     0.00000000282,
       0.00000001363,     0.00000000129,
       0.00000000138,     0.00000001621,
      -0.00000000042,    -0.00000000147,
       0.00000000148,    -0.00000000044,
      -0.00000002670,     0.00000002336,
       0.00000000134,     0.00000000175,
       0.00000000064,     0.00000000249,
      -0.00000000257,     0.00000000082,
      -0.00000000087,    -0.00000000077,
      -0.00000000074,     0.00000000093,
      -0.00000000093,     0.00000000106,
       0.00000000074,     0.00000000234,
       0.00000000235,    -0.00000000072,
       0.00000000602,     0.00000002464,
       0.00000000262,    -0.00000000075,
       0.00000000162,     0.00000000027,
      -0.00000000035,     0.00000000127,
      -0.00000000050,     0.00000000163,
      -0.00000000124,    -0.00000000062,
      -0.00000000907,    -0.00000000294,
       0.00000000286,    -0.00000000739,
      -0.00000000014,     0.00000000163,
      -0.00000000138,    -0.00000000017,
      -0.00000000189,     0.00000000019,
      -0.00000000005,    -0.00000000150,
       0.00000000088,    -0.00000000193,
       0.00000000139,     0.00000000107,
      -0.00000000198,    -0.00000000203,
       0.00000000060,     0.00000000070,
      -0.00000000204,     0.00000000196,
       0.00000000070,    -0.00000000058,
       0.00000000043,    -0.00000000058,
      -0.00000000050,    -0.00000000157,
      -0.00000000158,     0.00000000047,
      -0.00000004447,    -0.00000004649,
       0.00000001278,     0.00000001589,
      -0.00000000217,     0.00000000279,
       0.00000000080,    -0.00000000084,
       0.00000000427,     0.00000001364,
      -0.00000000080,    -0.00000000348,
       0.00000001366,    -0.00000000411,
      -0.00000000347,     0.00000000073,
      -0.00000000608,    -0.00000002120,
      -0.00000000347,     0.00000000155,
       0.00000000095,    -0.00000000030,
      -0.00000000106,     0.00000000710,
       0.00000000027,    -0.00000000128,
       0.00000000700,     0.00000000130,
      -0.00000000123,    -0.00000000034,
       0.00000009434,     0.00000031338,
      -0.00000001679,    -0.00000007905,
       0.00000001676,    -0.00000000740,
      -0.00000000452,     0.00000000143,
       0.00000000993,    -0.00000006219,
      -0.00000000261,     0.00000001109,
       0.00000000093,    -0.00000000077,
      -0.00000006182,    -0.00000001076,
       0.00000001090,     0.00000000289,
      -0.00000000072,    -0.00000000095,
      -0.00000001919,     0.00000010587,
       0.00000000481,    -0.00000001858,
       0.00000001756,     0.00000000128,
      -0.00000000331,    -0.00000000059,
       0.00000001433,    -0.00000002125,
      -0.00000000174,     0.00000000238,
      -0.00000002011,    -0.00000001514,
       0.00000000202,     0.00000000193,
       0.00000024353,    -0.00000142186,
      -0.00000006361,     0.00000025106,
       0.00000002126,    -0.00000001660,
      -0.00000008283,    -0.00000000604,
       0.00000001551,     0.00000000277,
      -0.00000000112,    -0.00000000126,
      -0.00000012998,     0.00000018596,
       0.00000001623,    -0.00000002033,
      -0.00000000188,     0.00000000005,
       0.00000018200,     0.00000013274,
      -0.00000001914,    -0.00000001681,
      -0.00000000015,     0.00000000186,
       0.00000000180,     0.00000000084,
       0.00000000100,    -0.00000000157,
      -0.00000001766,     0.00000000370,
       0.00000000022,     0.00000000103,
       0.00000000052,    -0.00000000068,
       0.00000025299,    -0.00000034885,
      -0.00000003116,     0.00000003734,
      -0.00000005709,    -0.00000003569,
       0.00000000645,     0.00000000498,
      -0.00000005679,     0.00000003535,
       0.00000000268,    -0.00000000204,
       0.00000002831,     0.00000005735,
      -0.00000000060,    -0.00000000259,
      -0.00000301796,     0.00000420508,
       0.00000037539,    -0.00000045093,
      -0.00000004269,    -0.00000000071,
       0.00000026310,     0.00000016370,
      -0.00000002961,    -0.00000002282,
       0.00000000004,     0.00000000280,
       0.00000051462,    -0.00000029662,
      -0.00000002476,     0.00000001431,
       0.00000000049,     0.00000000084,
      -0.00000027433,    -0.00000051622,
       0.00000000980,     0.00000002443,
       0.00000000119,    -0.00000000015,
      -0.00000000221,    -0.00000000265,
      -0.00000000279,     0.00000000120,
       0.00000001487,     0.00000001047,
       0.00000000070,    -0.00000000100,
      -0.00000110482,     0.00000061457,
       0.00000005207,    -0.00000002697,
       0.00000009625,     0.00000016277,
      -0.00000000423,    -0.00000000868,
       0.00000009411,    -0.00000002255,
       0.00000000187,     0.00000000228,
       0.00000000453,    -0.00000008089,
      -0.00000000098,    -0.00000000404,
      -0.00000000081,    -0.00000000048,
       0.00000000013,    -0.00000000087,
       0.00001174001,    -0.00000653284,
      -0.00000055368,     0.00000028681,
       0.00000000827,     0.00000002201,
      -0.00000043305,    -0.00000072746,
       0.00000001883,     0.00000003832,
       0.00000000163,    -0.00000000058,
      -0.00000082218,     0.00000011332,
      -0.00000002194,    -0.00000001003,
       0.00000000070,     0.00000000181,
       0.00000003672,     0.00000078402,
      -0.00000000087,     0.00000002815,
       0.00000000176,    -0.00000000153,
       0.00000000091,     0.00000000400,
       0.00000000297,     0.00000000174,
       0.00000196678,    -0.00000021389,
       0.00000005951,     0.00000001544,
      -0.00000002315,    -0.00000028330,
       0.00000000172,    -0.00000000786,
      -0.00000002723,     0.00000005242,
      -0.00000000022,    -0.00000000240,
      -0.00000000140,    -0.00000005351,
       0.00000000007,     0.00000000504,
       0.00000000073,     0.00000000138,
      -0.00000000080,     0.00000000117,
      -0.00001829184,     0.00000198170,
      -0.00000055254,    -0.00000014328,
       0.00000002516,     0.00000003903,
       0.00000010628,     0.00000122805,
      -0.00000000782,     0.00000003605,
       0.00000000276,    -0.00000000157,
      -0.00000001268,    -0.00000028512,
       0.00000001669,     0.00000001325,
      -0.00000000136,     0.00000000235,
      -0.00000015247,     0.00000021377,
       0.00000000719,    -0.00000002848,
       0.00000000182,     0.00000000156,
       0.00000000162,    -0.00000000678,
      -0.00000000268,    -0.00000000488,
       0.00000024991,     0.00000063486,
      -0.00000006603,    -0.00000003014,
       0.00000000426,    -0.00000000578,
       0.00000007583,    -0.00000003398,
      -0.00000000375,     0.00000000777,
      -0.00000000069,    -0.00000000052,
       0.00000008230,    -0.00000000467,
       0.00000000243,    -0.00000000035,
       0.00000001031,     0.00000008838,
      -0.00000000147,     0.00000000229,
      -0.00000199005,    -0.00000504619,
       0.00000052478,     0.00000023959,
      -0.00000003377,     0.00000004589,
      -0.00000033698,     0.00000015286,
       0.00000001636,    -0.00000003457,
       0.00000000309,     0.00000000228,
      -0.00000015748,    -0.00000000690,
      -0.00000000564,     0.00000000613,
      -0.00000003983,    -0.00000020614,
       0.00000000854,    -0.00000000450,
      -0.00000000097,     0.00000000100,
       0.00000000156,     0.00000000315,
      -0.00000000242,    -0.00000000129,
      -0.00000009361,     0.00000005635,
       0.00000000520,    -0.00000002405,
       0.00000000657,     0.00000000872,
      -0.00000000241,    -0.00000000034,
       0.00000000855,     0.00000002748,
      -0.00000000273,    -0.00000000143,
      -0.00000002791,     0.00000000915,
       0.00000000137,    -0.00000000298,
      -0.00000000067,     0.00000000092,
      -0.00000000093,    -0.00000000066,
       0.00000062164,    -0.00000037794,
      -0.00000003415,     0.00000016016,
      -0.00000001663,    -0.00000001458,
      -0.00000003021,    -0.00000003996,
       0.00000001102,     0.00000000148,
      -0.00000000095,     0.00000000117,
      -0.00000001337,    -0.00000004849,
       0.00000000418,     0.00000000268,
       0.00000005167,    -0.00000001823,
      -0.00000000230,     0.00000000612,
      -0.00000000060,    -0.00000000043,
       0.00000000290,    -0.00000000410,
       0.00000000446,     0.00000000290,
      -0.00000000649,    -0.00000000918,
      -0.00000000061,     0.00000000063,
      -0.00000000420,     0.00000000191,
       0.00000000032,    -0.00000000099,
      -0.00000000195,    -0.00000000422,
       0.00000000101,     0.00000000031,
       0.00000003523,     0.00000004535,
      -0.00000001999,    -0.00000000171,
       0.00000000295,    -0.00000000301,
       0.00000000012,     0.00000000148,
       0.00000000615,    -0.00000000283,
      -0.00000000052,     0.00000000143,
       0.00000000317,     0.00000000627,
      -0.00000000155,    -0.00000000045,
       0.00000000103,     0.00000000073,
      -0.00000000073,     0.00000000105,
       0.00000001426,     0.00000001297,
      -0.00000001049,     0.00000000726,
       0.00000000434,    -0.00000000466,
       0.00000000235,     0.00000000350,
      -0.00000000115,     0.00000000047,
      -0.00000006511,    -0.00000006011,
       0.00000004830,    -0.00000003312,
       0.00000000671,     0.00000001604,
      -0.00000000991,     0.00000001066,
      -0.00000000542,    -0.00000000800,
       0.00000000264,    -0.00000000107,
      -0.00000000079,     0.00000000058,
      -0.00000000096,     0.00000000073,
      -0.00000000120,    -0.00000000205,
      -0.00000000206,     0.00000000121,
       0.00000000404,     0.00000000677,
       0.00000000103,    -0.00000000055,
       0.00000000136,     0.00000000008,
       0.00000000000,    -0.00000000143,
      -0.00000004560,    -0.00000007624,
      -0.00000000312,     0.00000000166,
      -0.00000001334,    -0.00000000042,
      -0.00000000031,     0.00000001342,
       0.00000005050,     0.00000000146,
       0.00000000016,    -0.00000000674,
       0.00000000019,     0.00000000121,
      -0.00000049804,    -0.00000001444,
      -0.00000000049,     0.00000001971,
       0.00000000040,    -0.00000000671,
       0.00000000054,     0.00000000048,
      -0.00000000557,    -0.00000000095,
       0.00000000037,    -0.00000000069,
      -0.00000000252,     0.00000002672,
       0.00000000306,     0.00000000032,
       0.00000000349,     0.00000000006,
      -0.00000000012,     0.00000000380,
       0.00000002275,    -0.00000022992,
       0.00000002279,     0.00000001555,
      -0.00000000918,    -0.00000000096,
       0.00000000062,    -0.00000000092,
      -0.00000000201,     0.00000000017,
       0.00000000034,    -0.00000000468,
      -0.00000000017,     0.00000000193,
      -0.00000000198,    -0.00000000018,
       0.00000005029,     0.00000000849,
       0.00000000037,    -0.00000000203,
       0.00000000009,    -0.00000000134,
       0.00000000171,     0.00000000019,
      -0.00000000074,     0.00000000107,
       0.00000000105,     0.00000000075,
      -0.00000001472,     0.00000002113,
       0.00000000132,     0.00000000080,
       0.00000000250,    -0.00000000143,
      -0.00000000130,    -0.00000000251,
       0.00000000040,     0.00000000067,
       0.00000004800,    -0.00000002725,
      -0.00000000179,    -0.00000000298,
      -0.00000000311,     0.00000000036,
       0.00000000001,     0.00000000296,
      -0.00000005812,     0.00000000550,
       0.00000000023,     0.00000000390,
      -0.00000000078,    -0.00000000094,
      -0.00000000046,     0.00000000146,
      -0.00000001782,    -0.00000001294,
      -0.00000000086,     0.00000000139,
       0.00000000035,     0.00000000108,
       0.00000000109,    -0.00000000033,
       0.00000000134,    -0.00000000041,
      -0.00000000112,    -0.00000000051,
      -0.00000000131,    -0.00000000136,
      -0.00000000138,     0.00000000129,
      -0.00000000032,    -0.00000000098,
      -0.00000000099,     0.00000000028,
      -0.00000002995,    -0.00000003218,
       0.00000000853,     0.00000001075,
      -0.00000000148,     0.00000000183,
       0.00000000053,    -0.00000000054,
       0.00000000265,     0.00000000860,
      -0.00000000050,    -0.00000000215,
       0.00000000861,    -0.00000000252,
      -0.00000000214,     0.00000000045,
      -0.00000000404,    -0.00000001433,
      -0.00000000221,     0.00000000094,
      -0.00000000059,     0.00000000416,
       0.00000000408,     0.00000000077,
       0.00000005953,     0.00000020181,
      -0.00000001057,    -0.00000004997,
       0.00000001062,    -0.00000000450,
      -0.00000000279,     0.00000000086,
       0.00000000578,    -0.00000003660,
      -0.00000000146,     0.00000000641,
      -0.00000003630,    -0.00000000637,
       0.00000000627,     0.00000000165,
      -0.00000001184,     0.00000006602,
       0.00000000287,    -0.00000001137,
       0.00000001033,     0.00000000084,
      -0.00000000191,    -0.00000000035,
       0.00000000761,    -0.00000001163,
      -0.00000000088,     0.00000000129,
      -0.00000001080,    -0.00000000814,
       0.00000000104,     0.00000000100,
       0.00000014526,    -0.00000085206,
      -0.00000003650,     0.00000014748,
       0.00000001212,    -0.00000000947,
      -0.00000004863,    -0.00000000392,
       0.00000000890,     0.00000000161,
      -0.00000000063,    -0.00000000071,
      -0.00000006975,     0.00000010164,
       0.00000000833,    -0.00000001085,
       0.00000009886,     0.00000007151,
      -0.00000001005,    -0.00000000869,
      -0.00000000011,     0.00000000093,
       0.00000000090,     0.00000000041,
       0.00000000051,    -0.00000000077,
       0.00000014225,    -0.00000019956,
      -0.00000001678,     0.00000002079,
      -0.00000003093,    -0.00000001933,
       0.00000000339,     0.00000000258,
      -0.00000002765,     0.00000001839,
       0.00000000114,    -0.00000000110,
       0.00000001375,     0.00000002786,
      -0.00000000017,    -0.00000000107,
      -0.00000164966,     0.00000233497,
       0.00000019635,    -0.00000024373,
      -0.00000002143,    -0.00000000061,
       0.00000014247,     0.00000008861,
      -0.00000001554,    -0.00000001176,
      -0.00000000001,     0.00000000137,
       0.00000025162,    -0.00000015168,
      -0.00000001058,     0.00000000721,
      -0.00000013750,    -0.00000025217,
       0.00000000444,     0.00000001033,
      -0.00000000109,    -0.00000000127,
      -0.00000000135,     0.00000000053,
      -0.00000055827,     0.00000032301,
       0.00000002289,    -0.00000001362,
       0.00000004799,     0.00000007920,
      -0.00000000201,    -0.00000000374,
       0.00000004189,    -0.00000001363,
       0.00000000111,     0.00000000132,
       0.00000000243,    -0.00000003367,
      -0.00000000056,    -0.00000000236,
       0.00000583782,    -0.00000337877,
      -0.00000023951,     0.00000014253,
      -0.00000021636,    -0.00000035436,
       0.00000000894,     0.00000001638,
      -0.00000035976,     0.00000006965,
      -0.00000001315,    -0.00000000578,
       0.00000000044,     0.00000000082,
       0.00000002536,     0.00000033681,
      -0.00000000068,     0.00000001661,
       0.00000000080,    -0.00000000087,
       0.00000000051,     0.00000000201,
       0.00000000147,     0.00000000086,
       0.00000087047,    -0.00000013394,
       0.00000003619,     0.00000000907,
      -0.00000001459,    -0.00000012132,
       0.00000000102,    -0.00000000466,
      -0.00000001442,     0.00000002783,
      -0.00000000028,    -0.00000000115,
      -0.00000000093,    -0.00000002852,
       0.00000000010,     0.00000000251,
      -0.00000809423,     0.00000124135,
      -0.00000033611,    -0.00000008417,
       0.00000001507,     0.00000001801,
       0.00000006689,     0.00000052757,
      -0.00000000462,     0.00000002134,
       0.00000000126,    -0.00000000092,
      -0.00000000836,    -0.00000015194,
       0.00000000943,     0.00000000620,
      -0.00000000074,     0.00000000135,
      -0.00000008164,     0.00000011407,
       0.00000000315,    -0.00000001485,
       0.00000000111,     0.00000000075,
       0.00000000058,    -0.00000000316,
      -0.00000000114,    -0.00000000216,
       0.00000014135,     0.00000033711,
      -0.00000003555,    -0.00000001372,
       0.00000003989,    -0.00000001850,
      -0.00000000172,     0.00000000412,
       0.00000003703,    -0.00000000318,
       0.00000000153,    -0.00000000022,
       0.00000000682,     0.00000004089,
      -0.00000000093,     0.00000000145,
      -0.00000114899,    -0.00000273785,
       0.00000028851,     0.00000011139,
      -0.00000001758,     0.00000002750,
      -0.00000017774,     0.00000008345,
       0.00000000748,    -0.00000001838,
       0.00000000179,     0.00000000115,
      -0.00000006803,    -0.00000000425,
      -0.00000000357,     0.00000000400,
      -0.00000002622,    -0.00000009960,
       0.00000000550,    -0.00000000287,
      -0.00000000056,     0.00000000067,
       0.00000000086,     0.00000000162,
      -0.00000000124,    -0.00000000073,
      -0.00000005817,     0.00000003798,
       0.00000000286,    -0.00000001528,
       0.00000000437,     0.00000000553,
      -0.00000000155,    -0.00000000020,
       0.00000000506,     0.00000001499,
      -0.00000000148,    -0.00000000069,
      -0.00000001532,     0.00000000554,
       0.00000000065,    -0.00000000168,
       0.00000040443,    -0.00000026672,
      -0.00000001954,     0.00000010635,
      -0.00000001154,    -0.00000000927,
      -0.00000002013,    -0.00000002531,
       0.00000000708,     0.00000000086,
      -0.00000000058,     0.00000000078,
      -0.00000000788,    -0.00000002700,
       0.00000000219,     0.00000000138,
       0.00000002954,    -0.00000001172,
      -0.00000000109,     0.00000000372,
       0.00000000126,    -0.00000000186,
       0.00000000211,     0.00000000126,
      -0.00000000533,    -0.00000000662,
      -0.00000000269,     0.00000000137,
      -0.00000000141,    -0.00000000271,
       0.00000003099,     0.00000003696,
       0.00000000226,    -0.00000000248,
       0.00000000012,     0.00000000116,
       0.00000000419,    -0.00000000210,
      -0.00000000033,     0.00000000099,
       0.00000000242,     0.00000000432,
      -0.00000000111,    -0.00000000026,
       0.00000000069,     0.00000000058,
      -0.00000000066,     0.00000000065,
      -0.00000000040,     0.00000000087,
      -0.00000000090,    -0.00000000053,
       0.00000000092,    -0.00000000071,
      -0.00000000039,    -0.00000000066,
      -0.00000000067,     0.00000000039,
      -0.00000001478,    -0.00000002490,
      -0.00000000099,     0.00000000053,
      -0.00000000384,    -0.00000000017,
      -0.00000000013,     0.00000000386,
       0.00000001467,     0.00000000058,
       0.00000000006,    -0.00000000191,
      -0.00000014464,    -0.00000000576,
      -0.00000000020,     0.00000000560,
       0.00000000015,    -0.00000000231,
      -0.00000000193,    -0.00000000033,
      -0.00000000098,     0.00000000928,
       0.00000000104,     0.00000000011,
       0.00000000102,     0.00000000002,
      -0.00000000004,     0.00000000114,
       0.00000000833,    -0.00000008012,
      -0.00000000312,    -0.00000000034,
       0.00000000014,    -0.00000000155,
      -0.00000000945,     0.00000001351,
       0.00000000082,     0.00000000052,
       0.00000000142,    -0.00000000083,
      -0.00000000074,    -0.00000000142,
       0.00000002817,    -0.00000001614,
      -0.00000000103,    -0.00000000171,
      -0.00000000157,     0.00000000025,
       0.00000000000,     0.00000000146,
      -0.00000003025,     0.00000000368,
       0.00000000016,     0.00000000196,
      -0.00000000056,     0.00000000081,
      -0.00000000087,    -0.00000000091,
      -0.00000000092,     0.00000000085,
      -0.00000002007,    -0.00000002189,
      -0.00000000100,     0.00000000119,
       0.00000000165,     0.00000000538,
      -0.00000000031,    -0.00000000132,
       0.00000000538,    -0.00000000156,
      -0.00000000131,     0.00000000028,
      -0.00000000265,    -0.00000000945,
      -0.00000000139,     0.00000000058,
      -0.00000000032,     0.00000000243,
       0.00000000236,     0.00000000045,
       0.00000003751,     0.00000012807,
      -0.00000000667,    -0.00000003118,
       0.00000000666,    -0.00000000274,
      -0.00000000171,     0.00000000052,
       0.00000000327,    -0.00000002140,
      -0.00000000080,     0.00000000368,
      -0.00000002116,    -0.00000000369,
       0.00000000358,     0.00000000093,
      -0.00000000704,     0.00000004037,
       0.00000000604,     0.00000000051,
      -0.00000000109,    -0.00000000019,
       0.00000000400,    -0.00000000636,
      -0.00000000044,     0.00000000070,
      -0.00000000578,    -0.00000000434,
       0.00000000053,     0.00000000051,
       0.00000008398,    -0.00000050452,
      -0.00000002040,     0.00000008561,
      -0.00000002836,    -0.00000000239,
       0.00000000506,     0.00000000090,
      -0.00000003707,     0.00000005541,
       0.00000000420,    -0.00000000577,
       0.00000005352,     0.00000003817,
      -0.00000000524,    -0.00000000443,
       0.00000007844,    -0.00000011265,
      -0.00000000880,     0.00000001139,
      -0.00000001671,    -0.00000001036,
       0.00000000178,     0.00000000132,
      -0.00000001344,     0.00000000965,
       0.00000000046,    -0.00000000059,
       0.00000000668,     0.00000001351,
      -0.00000088943,     0.00000128711,
       0.00000010057,    -0.00000013029,
       0.00000007695,     0.00000004744,
      -0.00000000811,    -0.00000000597,
       0.00000012265,    -0.00000007801,
      -0.00000000429,     0.00000000364,
      -0.00000006922,    -0.00000012280,
       0.00000000197,     0.00000000413,
      -0.00000000055,    -0.00000000063,
      -0.00000027908,     0.00000016926,
       0.00000000945,    -0.00000000680,
       0.00000002409,     0.00000003845,
      -0.00000000095,    -0.00000000154,
       0.00000001877,    -0.00000000800,
       0.00000000065,     0.00000000073,
       0.00000000132,    -0.00000001390,
      -0.00000000031,    -0.00000000134,
       0.00000288268,    -0.00000174891,
      -0.00000009754,     0.00000007019,
      -0.00000010883,    -0.00000017216,
       0.00000000423,     0.00000000669,
      -0.00000015786,     0.00000004132,
      -0.00000000768,    -0.00000000320,
       0.00000001612,     0.00000014456,
      -0.00000000040,     0.00000000955,
       0.00000000028,     0.00000000101,
       0.00000000073,     0.00000000043,
       0.00000038408,    -0.00000008043,
       0.00000002133,     0.00000000503,
      -0.00000000879,    -0.00000005219,
       0.00000000057,    -0.00000000269,
      -0.00000000741,     0.00000001459,
      -0.00000000068,    -0.00000001493,
       0.00000000010,     0.00000000122,
      -0.00000357223,     0.00000074621,
      -0.00000019813,    -0.00000004673,
       0.00000004027,     0.00000022742,
      -0.00000000257,     0.00000001231,
       0.00000000054,    -0.00000000051,
      -0.00000000564,    -0.00000007949,
       0.00000000522,     0.00000000271,
      -0.00000000038,     0.00000000076,
      -0.00000004245,     0.00000006012,
       0.00000000119,    -0.00000000758,
       0.00000000018,    -0.00000000148,
      -0.00000000048,    -0.00000000096,
       0.00000000079,    -0.00000000048,
      -0.00000000041,    -0.00000000079,
       0.00000007871,     0.00000017557,
      -0.00000001878,    -0.00000000573,
       0.00000002059,    -0.00000000997,
      -0.00000000075,     0.00000000215,
       0.00000001646,    -0.00000000202,
       0.00000000093,    -0.00000000014,
       0.00000000428,     0.00000001880,
      -0.00000000056,     0.00000000088,
      -0.00000064946,    -0.00000144857,
       0.00000015477,     0.00000004719,
      -0.00000000870,     0.00000001612,
      -0.00000009192,     0.00000004508,
       0.00000000316,    -0.00000000960,
       0.00000000102,     0.00000000055,
      -0.00000002847,    -0.00000000261,
      -0.00000000215,     0.00000000249,
      -0.00000001646,    -0.00000004795,
       0.00000000338,    -0.00000000179,
       0.00000000047,     0.00000000083,
      -0.00000000063,    -0.00000000040,
       0.00000001616,    -0.00000000947,
      -0.00000000059,    -0.00000000096,
      -0.00000003475,     0.00000002436,
       0.00000000144,    -0.00000000931,
       0.00000000278,     0.00000000335,
      -0.00000000095,    -0.00000000011,
       0.00000000291,     0.00000000796,
      -0.00000000078,    -0.00000000031,
      -0.00000000820,     0.00000000326,
       0.00000000028,    -0.00000000092,
       0.00000025061,    -0.00000017698,
      -0.00000001021,     0.00000006717,
      -0.00000001277,    -0.00000001534,
       0.00000000434,     0.00000000044,
      -0.00000000443,    -0.00000001451,
       0.00000000107,     0.00000000067,
       0.00000001636,    -0.00000000728,
      -0.00000000045,     0.00000000220,
       0.00000000054,    -0.00000000084,
       0.00000000100,     0.00000000053,
      -0.00000000164,     0.00000000092,
      -0.00000000095,    -0.00000000166,
       0.00000002470,     0.00000002707,
       0.00000000159,    -0.00000000188,
       0.00000000267,    -0.00000000144,
       0.00000000173,     0.00000000279,
      -0.00000000111,    -0.00000000006,
      -0.00000000005,     0.00000000112,
      -0.00000004216,    -0.00000000222,
      -0.00000000008,     0.00000000160,
       0.00000000300,    -0.00000002738,
      -0.00000000105,    -0.00000000012,
      -0.00000001328,    -0.00000001455,
      -0.00000000066,     0.00000000077,
       0.00000000103,     0.00000000333,
      -0.00000000019,    -0.00000000081,
       0.00000000332,    -0.00000000097,
      -0.00000000087,     0.00000000035,
      -0.00000000017,     0.00000000140,
       0.00000000135,     0.00000000025,
       0.00000002352,     0.00000008000,
      -0.00000000422,    -0.00000001915,
       0.00000000412,    -0.00000000167,
      -0.00000000104,     0.00000000032,
       0.00000000180,    -0.00000001240,
      -0.00000000042,     0.00000000210,
      -0.00000001223,    -0.00000000208,
       0.00000000202,     0.00000000050,
      -0.00000000404,     0.00000002424,
       0.00000000350,     0.00000000030,
       0.00000000207,    -0.00000000347,
      -0.00000000307,    -0.00000000230,
       0.00000004712,    -0.00000029515,
      -0.00000001106,     0.00000004903,
      -0.00000001640,    -0.00000000139,
       0.00000000285,     0.00000000049,
      -0.00000001948,     0.00000003008,
       0.00000000207,    -0.00000000304,
       0.00000002883,     0.00000002017,
      -0.00000000270,    -0.00000000221,
       0.00000004250,    -0.00000006283,
      -0.00000000448,     0.00000000614,
      -0.00000000900,    -0.00000000549,
       0.00000000092,     0.00000000066,
      -0.00000000650,     0.00000000508,
       0.00000000324,     0.00000000651,
      -0.00000047305,     0.00000070418,
       0.00000005035,    -0.00000006891,
       0.00000004140,     0.00000002510,
      -0.00000000421,    -0.00000000298,
       0.00000005948,    -0.00000004025,
      -0.00000000159,     0.00000000183,
      -0.00000003492,    -0.00000005950,
       0.00000000084,     0.00000000149,
      -0.00000013815,     0.00000008850,
       0.00000001215,     0.00000001859,
      -0.00000000045,    -0.00000000059,
       0.00000000844,    -0.00000000459,
       0.00000000073,    -0.00000000564,
       0.00000141238,    -0.00000090506,
      -0.00000003597,     0.00000003414,
      -0.00000005497,    -0.00000008325,
       0.00000000198,     0.00000000253,
      -0.00000006920,     0.00000002385,
      -0.00000000440,    -0.00000000171,
       0.00000000971,     0.00000006170,
      -0.00000000019,     0.00000000537,
       0.00000016857,    -0.00000004685,
       0.00000001227,     0.00000000265,
      -0.00000000513,    -0.00000002245,
       0.00000000030,    -0.00000000152,
      -0.00000000370,     0.00000000757,
      -0.00000000049,    -0.00000000771,
      -0.00000156762,     0.00000043468,
      -0.00000011388,    -0.00000002459,
       0.00000002345,     0.00000009790,
      -0.00000000136,     0.00000000695,
      -0.00000000375,    -0.00000004099,
       0.00000000284,     0.00000000108,
      -0.00000002155,     0.00000003139,
       0.00000000033,    -0.00000000380,
       0.00000004331,     0.00000008993,
      -0.00000000977,    -0.00000000208,
       0.00000001046,    -0.00000000533,
      -0.00000000029,     0.00000000110,
       0.00000000722,    -0.00000000122,
       0.00000000258,     0.00000000859,
      -0.00000036188,    -0.00000075130,
       0.00000008153,     0.00000001734,
      -0.00000004679,     0.00000002418,
       0.00000000118,    -0.00000000494,
      -0.00000001145,    -0.00000000160,
      -0.00000000124,     0.00000000151,
      -0.00000000998,    -0.00000002301,
       0.00000000199,    -0.00000000110,
      -0.00000002014,     0.00000001507,
       0.00000000170,     0.00000000196,
       0.00000000164,     0.00000000414,
      -0.00000000430,     0.00000000188,
       0.00000014942,    -0.00000011243,
      -0.00000000479,     0.00000004074,
      -0.00000000782,    -0.00000000897,
       0.00000000257,     0.00000000019,
      -0.00000000241,    -0.00000000759,
       0.00000000885,    -0.00000000441,
      -0.00000000015,     0.00000000127,
      -0.00000000097,     0.00000000059,
      -0.00000000062,    -0.00000000098,
       0.00000001866,     0.00000001862,
       0.00000000106,    -0.00000000134,
       0.00000000163,    -0.00000000094,
       0.00000000116,     0.00000000172,
       0.00000000064,     0.00000000203,
       0.00000000203,    -0.00000000059,
       0.00000001466,     0.00000004926,
       0.00000000253,    -0.00000000101,
       0.00000000096,    -0.00000000712,
      -0.00000000021,     0.00000000118,
      -0.00000000700,    -0.00000000114,
       0.00000000113,     0.00000000027,
      -0.00000000224,     0.00000001433,
       0.00000000201,     0.00000000017,
       0.00000000106,    -0.00000000189,
      -0.00000000162,    -0.00000000121,
       0.00000002568,    -0.00000017062,
      -0.00000000578,     0.00000002769,
      -0.00000000940,    -0.00000000077,
       0.00000000159,     0.00000000026,
      -0.00000001011,     0.00000001625,
       0.00000000099,    -0.00000000159,
       0.00000001545,     0.00000001054,
      -0.00000000138,    -0.00000000108,
       0.00000002264,    -0.00000003468,
      -0.00000000482,    -0.00000000287,
      -0.00000000312,     0.00000000268,
       0.00000000156,     0.00000000312,
      -0.00000024825,     0.00000038247,
       0.00000002454,    -0.00000003599,
       0.00000002217,     0.00000001312,
      -0.00000000216,    -0.00000000145,
       0.00000002865,    -0.00000002080,
      -0.00000000049,     0.00000000091,
      -0.00000001762,    -0.00000002864,
      -0.00000006770,     0.00000004615,
       0.00000000615,     0.00000000894,
       0.00000000379,    -0.00000000259,
       0.00000000041,    -0.00000000223,
       0.00000068623,    -0.00000046796,
      -0.00000001082,     0.00000001632,
      -0.00000002784,    -0.00000004002,
       0.00000000092,     0.00000000082,
      -0.00000003019,     0.00000001347,
      -0.00000000247,    -0.00000000088,
       0.00000000562,     0.00000002608,
      -0.00000000007,     0.00000000297,
       0.00000007334,    -0.00000002661,
      -0.00000000291,    -0.00000000961,
      -0.00000000181,     0.00000000390,
      -0.00000000035,    -0.00000000393,
      -0.00000068196,     0.00000024691,
      -0.00000006414,    -0.00000001227,
       0.00000001330,     0.00000004194,
      -0.00000000068,     0.00000000385,
      -0.00000000244,    -0.00000002088,
       0.00000000152,     0.00000000037,
      -0.00000001071,     0.00000001625,
       0.00000000000,    -0.00000000186,
       0.00000002361,     0.00000004540,
       0.00000000525,    -0.00000000284,
       0.00000000311,    -0.00000000071,
       0.00000000152,     0.00000000390,
      -0.00000019928,    -0.00000038313,
       0.00000004223,     0.00000000458,
      -0.00000002349,     0.00000001289,
       0.00000000034,    -0.00000000251,
      -0.00000000435,    -0.00000000097,
      -0.00000000068,     0.00000000089,
      -0.00000000589,    -0.00000001101,
       0.00000000115,    -0.00000000066,
      -0.00000001138,     0.00000000907,
       0.00000000101,     0.00000000112,
       0.00000000091,     0.00000000212,
      -0.00000000222,     0.00000000107,
       0.00000008638,    -0.00000006910,
      -0.00000000187,     0.00000002397,
      -0.00000000466,    -0.00000000511,
       0.00000000148,     0.00000000006,
      -0.00000000127,    -0.00000000388,
       0.00000000470,    -0.00000000261,
       0.00000001308,     0.00000001205,
       0.00000000067,    -0.00000000091,
       0.00000000096,    -0.00000000058,
       0.00000000076,     0.00000000102,
       0.00000000040,     0.00000000123,
       0.00000000122,    -0.00000000036,
       0.00000000912,     0.00000002987,
       0.00000000152,    -0.00000000061,
       0.00000000049,    -0.00000000405,
      -0.00000000397,    -0.00000000061,
       0.00000000114,     0.00000000009,
       0.00000000053,    -0.00000000102,
      -0.00000000086,    -0.00000000062,
       0.00000001357,    -0.00000009756,
      -0.00000000533,    -0.00000000041,
      -0.00000000518,     0.00000000872,
       0.00000000046,    -0.00000000083,
       0.00000000822,     0.00000000544,
      -0.00000000069,    -0.00000000052,
       0.00000001187,    -0.00000001896,
      -0.00000000257,    -0.00000000149,
      -0.00000000149,     0.00000000142,
       0.00000000075,     0.00000000148,
      -0.00000012853,     0.00000020629,
       0.00000001160,    -0.00000001858,
       0.00000001181,     0.00000000678,
      -0.00000000110,    -0.00000000068,
       0.00000001370,    -0.00000001076,
      -0.00000000889,    -0.00000001369,
      -0.00000003283,     0.00000002399,
       0.00000000311,     0.00000000426,
       0.00000000170,    -0.00000000144,
       0.00000000023,    -0.00000000085,
       0.00000033041,    -0.00000024157,
      -0.00000001412,    -0.00000001910,
      -0.00000001307,     0.00000000746,
      -0.00000000137,    -0.00000000043,
       0.00000000315,     0.00000001085,
      -0.00000000001,     0.00000000161,
       0.00000003152,    -0.00000001480,
      -0.00000000161,    -0.00000000408,
      -0.00000000087,     0.00000000199,
      -0.00000000023,    -0.00000000199,
      -0.00000029307,     0.00000013733,
      -0.00000003549,    -0.00000000573,
       0.00000000739,     0.00000001779,
      -0.00000000031,     0.00000000210,
      -0.00000000154,    -0.00000001052,
      -0.00000000521,     0.00000000836,
       0.00000001277,     0.00000002261,
       0.00000000260,    -0.00000000151,
       0.00000000132,    -0.00000000041,
       0.00000000087,     0.00000000176,
      -0.00000010869,    -0.00000019246,
       0.00000002154,    -0.00000000020,
      -0.00000001165,     0.00000000684,
       0.00000000002,    -0.00000000126,
      -0.00000000150,    -0.00000000059,
      -0.00000000340,    -0.00000000525,
       0.00000000064,    -0.00000000039,
      -0.00000000630,     0.00000000534,
       0.00000000059,     0.00000000063,
       0.00000000050,     0.00000000107,
      -0.00000000113,     0.00000000060,
       0.00000004866,    -0.00000004137,
      -0.00000000272,    -0.00000000285,
      -0.00000000066,    -0.00000000194,
       0.00000000245,    -0.00000000153,
       0.00000000564,     0.00000001786,
       0.00000000091,    -0.00000000037,
       0.00000000024,    -0.00000000229,
      -0.00000000223,    -0.00000000032,
       0.00000000694,    -0.00000005522,
      -0.00000000300,    -0.00000000020,
      -0.00000000262,     0.00000000467,
       0.00000000436,     0.00000000278,
       0.00000000613,    -0.00000001027,
      -0.00000000137,    -0.00000000076,
      -0.00000000070,     0.00000000075,
       0.00000000036,     0.00000000070,
      -0.00000006566,     0.00000011047,
       0.00000000626,     0.00000000345,
       0.00000000649,    -0.00000000556,
      -0.00000000448,    -0.00000000648,
      -0.00000001574,     0.00000001243,
       0.00000000158,     0.00000000202,
       0.00000000075,    -0.00000000079,
       0.00000015752,    -0.00000012445,
      -0.00000000716,    -0.00000000904,
      -0.00000000559,     0.00000000407,
       0.00000000172,     0.00000000442,
       0.00000001331,    -0.00000000808,
      -0.00000000088,    -0.00000000171,
      -0.00000000041,     0.00000000101,
      -0.00000000015,    -0.00000000100,
      -0.00000012380,     0.00000007501,
      -0.00000001934,    -0.00000000246,
       0.00000000403,     0.00000000744,
      -0.00000000014,     0.00000000113,
      -0.00000000095,    -0.00000000525,
      -0.00000000249,     0.00000000428,
       0.00000000685,     0.00000001112,
       0.00000000128,    -0.00000000079,
       0.00000000049,     0.00000000079,
      -0.00000005873,    -0.00000009534,
      -0.00000000570,     0.00000000361,
      -0.00000000193,    -0.00000000249,
       0.00000002686,    -0.00000002428,
      -0.00000000156,    -0.00000000155,
      -0.00000000033,    -0.00000000095,
       0.00000000126,    -0.00000000088,
       0.00000000011,    -0.00000000128,
      -0.00000000124,    -0.00000000016,
       0.00000000339,    -0.00000003096,
      -0.00000000167,    -0.00000000009,
      -0.00000000131,     0.00000000249,
       0.00000000230,     0.00000000141,
      -0.00000000072,    -0.00000000038,
      -0.00000003309,     0.00000005883,
       0.00000000331,     0.00000000174,
       0.00000000304,    -0.00000000288,
      -0.00000000226,    -0.00000000304,
      -0.00000000746,     0.00000000642,
       0.00000000080,     0.00000000095,
       0.00000007427,    -0.00000006394,
      -0.00000000363,    -0.00000000424,
      -0.00000000235,     0.00000000219,
       0.00000000093,     0.00000000174,
       0.00000000549,    -0.00000000434,
      -0.00000000047,    -0.00000000071,
      -0.00000005106,     0.00000004033,
       0.00000000217,     0.00000000305,
      -0.00000000058,    -0.00000000259,
      -0.00000000117,     0.00000000218,
       0.00000000062,    -0.00000000042,
      -0.00000003153,    -0.00000004657,
      -0.00000000276,     0.00000000190,
      -0.00000000107,    -0.00000000117,
       0.00000001452,    -0.00000001400,
      -0.00000000088,    -0.00000000083,
      -0.00000000064,     0.00000000132,
       0.00000000120,     0.00000000070,
      -0.00000001643,     0.00000003114,
       0.00000000174,     0.00000000086,
       0.00000000141,    -0.00000000148,
      -0.00000000113,    -0.00000000141,
       0.00000003460,    -0.00000003276,
      -0.00000000184,    -0.00000000197,
      -0.00000000096,     0.00000000117,
      -0.00000002034,     0.00000002135,
       0.00000000114,     0.00000000121,
      -0.00000000034,    -0.00000000127,
      -0.00000000053,     0.00000000110,
      -0.00000001680,    -0.00000002245,
      -0.00000000132,     0.00000000100,
      -0.00000000805,     0.00000001641,
       0.00000000091,     0.00000000042,
       0.00000000065,    -0.00000000076,
      -0.00000000057,    -0.00000000065,
       0.00000001589,    -0.00000001673,
      -0.00000000093,    -0.00000000090,
      -0.00000000013,     0.00000001919,
      -0.00000000031,    -0.00000000013,
      -0.00000000187,     0.00000000499,
      -0.00000000006,     0.00000000000,
      -0.00000000243,    -0.00000000270,
       0.00000000002,     0.00000000004,
       0.00000000004,    -0.00000000037,
      -0.00000000019,     0.00000000061,
       0.00000000509,     0.00000004376,
      -0.00000000063,    -0.00000000015,
      -0.00000001135,     0.00000000030,
      -0.00000000010,    -0.00000001123,
      -0.00000000007,    -0.00000000060,
       0.00000000654,     0.00000007139,
      -0.00000000099,    -0.00000000039,
      -0.00000002550,     0.00000000157,
      -0.00000000070,    -0.00000002553,
       0.00000003170,     0.00000017163,
      -0.00000000465,     0.00000000090,
      -0.00000006452,    -0.00000000688,
       0.00000000036,     0.00000000115,
       0.00000000120,    -0.00000007393,
      -0.00000000033,    -0.00000000026,
      -0.00000000017,    -0.00000000089,
       0.00000000029,     0.00000000122,
       0.00000000182,     0.00000144088,
      -0.00000003708,     0.00000000740,
      -0.00000000089,    -0.00000000100,
      -0.00000164696,     0.00000002045,
       0.00000000039,     0.00000000306,
      -0.00000001353,    -0.00000170383,
       0.00000000060,    -0.00000000737,
       0.00000000103,    -0.00000000482,
      -0.00000005538,    -0.00000005405,
      -0.00000299774,     0.00000350274,
       0.00000017372,    -0.00000035185,
       0.00000002128,     0.00000002052,
      -0.00000000117,     0.00000000082,
      -0.00000922504,     0.00000024495,
       0.00000000456,    -0.00000002762,
       0.00000000062,    -0.00000000086,
      -0.00000007478,    -0.00000921707,
      -0.00000000096,     0.00000002737,
       0.00000000089,    -0.00000000250,
       0.00000002104,    -0.00000007004,
      -0.00000000268,     0.00000000974,
       0.00000006578,    -0.00000007365,
      -0.00000001262,     0.00000000661,
      -0.00310599366,    -0.00044142123,
       0.00004478070,    -0.00003576112,
      -0.00000143624,     0.00000225497,
      -0.00000008400,    -0.00000009963,
       0.00003249139,    -0.00002787003,
      -0.00000064759,     0.00000205338,
      -0.00000010733,    -0.00000010027,
       0.00000000507,    -0.00000000374,
       0.00136013837,    -0.00000082022,
      -0.00000005201,    -0.00000000220,
       0.00000000557,     0.00000001715,
       0.00000118074,     0.00136132786,
      -0.00000005861,    -0.00000012902,
      -0.00000000567,     0.00000000394,
      -0.00000002790,     0.00000058592,
       0.00000000666,    -0.00000008321,
      -0.00000000007,    -0.00000000142,
      -0.00000050942,     0.00000034356,
       0.00000009363,    -0.00000002681,
       0.00000000170,    -0.00000000081,
      -0.00002715287,     0.00000118540,
       0.00000018031,    -0.00000041839,
       0.00000104536,     0.00000525064,
      -0.00000008541,    -0.00000002423,
       0.00000000169,     0.00000000038,
       0.00008693527,    -0.00000002088,
       0.00000000734,    -0.00000000297,
       0.00000007522,     0.00008735586,
      -0.00000000827,     0.00000000329,
      -0.00000006003,     0.00000004559,
       0.00000000072,     0.00000000203,
      -0.00000002102,     0.00000001290,
      -0.00000000235,     0.00000000028,
      -0.00000000603,     0.00000000154,
       0.00000000159,     0.00000000605,
       0.00000000263,    -0.00000000026,
      -0.00000000027,    -0.00000000261,
      -0.00000002040,     0.00000000154,
      -0.00000000013,     0.00000000673,
      -0.00000004130,     0.00000000529,
       0.00000000070,     0.00000000175,
       0.00000000530,     0.00000004136,
       0.00000000178,    -0.00000000072,
       0.00000000481,     0.00000000112,
       0.00000000107,    -0.00000000466,
       0.00000011142,    -0.00000001182,
       0.00000000012,    -0.00000000459,
       0.00000004038,    -0.00000000672,
      -0.00000000068,    -0.00000000168,
      -0.00000000675,    -0.00000004035,
      -0.00000000166,     0.00000000069,
      -0.00000000319,    -0.00000000073,
      -0.00000000070,     0.00000000311,
      -0.00000000005,     0.00000000007,
      -0.00000000008,     0.00000000026,
       0.00000000026,     0.00000000007,
      -0.00000000087,    -0.00000000046,
      -0.00000000046,     0.00000000087,
      -0.00000000351,     0.00000000072,
       0.00000000072,     0.00000000351,
      -0.00000000105,    -0.00000000025,
      -0.00000000330,     0.00000001100,
       0.00000000112,    -0.00000000074,
       0.00000001123,     0.00000000324,
      -0.00000000074,    -0.00000000112,
      -0.00000143103,     0.00000033965,
       0.00000003832,     0.00000005412,
       0.00000001186,     0.00000009541,
       0.00000000318,    -0.00000000294,
      -0.00000018551,     0.00000001328,
       0.00000000159,     0.00000000605,
       0.00000001320,     0.00000018646,
       0.00000000573,    -0.00000000161,
       0.00000005162,     0.00000001542,
       0.00000001467,    -0.00000004944,
       0.00000196018,    -0.00000031320,
      -0.00000005557,    -0.00000009035,
      -0.00000000092,    -0.00000006448,
      -0.00000000290,     0.00000000145,
       0.00000066591,    -0.00000008649,
      -0.00000001210,    -0.00000001801,
       0.00000000059,     0.00000000060,
      -0.00000008671,    -0.00000066597,
      -0.00000001801,     0.00000001212,
       0.00000000060,    -0.00000000059,
      -0.00000003318,    -0.00000001150,
      -0.00000001115,     0.00000003200,
       0.00000000316,     0.00000000020,
       0.00000000033,    -0.00000000316,
       0.00000000426,    -0.00000000106,
       0.00000001848,     0.00000000155,
       0.00000000139,    -0.00000001819,
      -0.00001260277,     0.00000092099,
       0.00000011281,     0.00000036864,
      -0.00000010105,     0.00000048595,
       0.00000001429,    -0.00000000500,
      -0.00001167789,     0.00000084910,
       0.00000003811,     0.00000018541,
      -0.00000000001,    -0.00000000208,
       0.00000084837,     0.00001170545,
       0.00000018575,    -0.00000003863,
      -0.00000000222,     0.00000000002,
       0.00000049949,     0.00000021353,
      -0.00000000473,     0.00000000594,
       0.00000019927,    -0.00000046166,
       0.00000000677,     0.00000000448,
       0.00004343211,    -0.00000476313,
      -0.00000056770,    -0.00000124393,
      -0.00000000242,     0.00000003014,
      -0.00000002669,    -0.00000044856,
      -0.00000001051,     0.00000000935,
       0.00000379851,    -0.00000061074,
      -0.00000000635,    -0.00000006273,
      -0.00000000029,     0.00000000113,
      -0.00000061617,    -0.00000379677,
      -0.00000006224,     0.00000000642,
      -0.00000032930,    -0.00000014370,
       0.00000000308,    -0.00000000424,
      -0.00000013772,     0.00000031172,
      -0.00000000461,    -0.00000000298,
       0.00000000000,    -0.00000000014,
       0.00000000000,    -0.00000000055,
       0.00000000000,    -0.00000000105,
       0.00000000023,    -0.00000000380,
      -0.00000000306,    -0.00000000024,
       0.00000000009,    -0.00000000291,
      -0.00000000326,    -0.00000001305,
      -0.00000001422,    -0.00000000167,
       0.00000000033,    -0.00000001423,
       0.00036971774,    -0.00000197571,
      -0.00000008074,     0.00000000569,
      -0.00000049914,    -0.00013047816,
       0.00000000176,     0.00000002394,
       0.00000000101,    -0.00000000117,
       0.00000170333,    -0.00000022439,
      -0.00000001035,     0.00000009479,
       0.00000000024,    -0.00000000121,
      -0.00000056788,    -0.00000600698,
       0.00000002760,     0.00000002471,
      -0.00000000068,     0.00000000037,
      -0.00000005105,    -0.00000047508,
       0.00000000028,    -0.00000001436,
       0.00000002697,    -0.00000109406,
       0.00000000122,     0.00000001079,
       0.00462318438,     0.00000088525,
       0.00000032127,     0.00000007602,
      -0.00000001247,     0.00000004831,
       0.00000042493,     0.00016297410,
      -0.00000000369,    -0.00000001318,
      -0.00000000141,     0.00000000085,
       0.00000269520,    -0.00000021026,
      -0.00000001412,    -0.00000017570,
       0.00000000023,     0.00000000147,
      -0.00000028987,     0.00001134985,
       0.00000004139,    -0.00000004596,
       0.00000045476,     0.00000011365,
      -0.00000000447,     0.00000000379,
      -0.00000019229,     0.00000023814,
      -0.00000000577,    -0.00000000216,
       0.00000000005,     0.00000000164,
      -0.00000000164,     0.00000000005,
       0.00000234738,    -0.00000002461,
      -0.00000002327,     0.00000006555,
      -0.00000002937,    -0.00000045627,
       0.00000001192,     0.00000000520,
      -0.00000084992,    -0.00000004033,
       0.00000000301,    -0.00000001312,
       0.00000003711,    -0.00000086314,
       0.00000001341,     0.00000000324,
       0.00000019256,    -0.00000008160,
      -0.00000000190,    -0.00000000235,
       0.00000007865,     0.00000018609,
       0.00000000249,    -0.00000000184,
      -0.00000474828,     0.00000079535,
       0.00000003574,    -0.00000015413,
       0.00000002384,     0.00000026359,
      -0.00000000591,    -0.00000000357,
       0.00000056382,    -0.00000005439,
      -0.00000000064,     0.00000000971,
       0.00000005516,     0.00000057237,
      -0.00000000997,    -0.00000000077,
       0.00000003219,    -0.00000001211,
       0.00000001449,     0.00000003747,
      -0.00000000043,    -0.00000000170,
      -0.00000000488,     0.00000000069,
      -0.00000000068,    -0.00000000491,
       0.00000000221,    -0.00000000061,
       0.00000000060,     0.00000000217,
       0.00000000036,     0.00000000154,
       0.00000000494,    -0.00000000085,
       0.00000000083,     0.00000000498,
      -0.00000000158,     0.00000000046,
      -0.00000000046,    -0.00000000156,
       0.00000000610,     0.00000000220,
      -0.00000000220,     0.00000000615,
       0.00000000715,     0.00000001274,
      -0.00000000072,     0.00000000034,
       0.00000001051,     0.00000004548,
      -0.00000000124,     0.00000000032,
      -0.00000004566,     0.00000001032,
      -0.00000000033,    -0.00000000125,
       0.00000000169,     0.00000000189,
       0.00000000004,    -0.00000000007,
      -0.00000000206,     0.00000000220,
       0.00000000005,     0.00000000004,
      -0.00000000025,    -0.00000000117,
      -0.00000000189,     0.00000000291,
      -0.00000000292,    -0.00000000190,
       0.00000000060,     0.00000000268,
      -0.00000000266,     0.00000000059,
      -0.00000000059,    -0.00000000200,
      -0.00000000188,     0.00000001551,
      -0.00000000079,    -0.00000000057,
      -0.00000001554,    -0.00000000187,
       0.00000000057,    -0.00000000079,
      -0.00000000036,    -0.00000000127,
       0.00000000126,    -0.00000000036,
      -0.00000002019,    -0.00000002320,
       0.00000000246,    -0.00000000321,
       0.00000016101,    -0.00000006953,
       0.00000000676,     0.00000001134,
      -0.00000000045,     0.00000000028,
       0.00000006983,     0.00000016099,
      -0.00000001137,     0.00000000681,
      -0.00000000029,    -0.00000000045,
       0.00000000332,     0.00000000219,
      -0.00000000222,     0.00000000309,
       0.00000000037,     0.00000000011,
      -0.00000000367,    -0.00000000699,
       0.00000000817,    -0.00000010579,
       0.00000000703,    -0.00000000493,
       0.00000069812,     0.00000013124,
       0.00000001353,     0.00000002953,
      -0.00000000129,     0.00000000121,
      -0.00000013138,     0.00000069596,
      -0.00000002866,     0.00000001344,
      -0.00000000126,    -0.00000000140,
      -0.00000001058,     0.00000001299,
      -0.00000001264,    -0.00000001005,
       0.00000004919,     0.00000008293,
      -0.00000002412,     0.00000000764,
       0.00000020860,     0.00000007301,
      -0.00000005437,     0.00000000930,
       0.00000000455,    -0.00000000590,
       0.00000056876,     0.00000195528,
      -0.00000019922,    -0.00000012915,
       0.00000001150,    -0.00000001217,
      -0.00000195849,     0.00000058895,
       0.00000012320,    -0.00000020147,
       0.00000001203,     0.00000001172,
      -0.00000010532,     0.00000013441,
       0.00000000985,     0.00000000622,
      -0.00000013902,    -0.00000010190,
      -0.00000000537,     0.00000000913,
       0.00018202708,     0.00004570729,
      -0.00000332266,     0.00000118260,
       0.00000048457,     0.00000025425,
      -0.00000004528,     0.00000003185,
      -0.00000232135,     0.00000202550,
       0.00000018373,    -0.00000007899,
      -0.00000001996,     0.00000002638,
      -0.00000000183,    -0.00000000198,
      -0.00008518996,    -0.00001246035,
       0.00000129531,    -0.00000093411,
      -0.00000004323,     0.00000006565,
      -0.00000000240,    -0.00000000285,
       0.00001248825,    -0.00008522812,
       0.00000092798,     0.00000130777,
      -0.00000006688,    -0.00000004465,
       0.00000000296,    -0.00000000242,
       0.00000078185,    -0.00000085651,
      -0.00000004147,    -0.00000002406,
      -0.00000000134,     0.00000000153,
       0.00000086076,     0.00000075836,
       0.00000002234,    -0.00000003902,
      -0.00000000159,    -0.00000000151,
      -0.00000042365,    -0.00000011540,
      -0.00000001543,     0.00000014213,
      -0.00000000264,    -0.00000000241,
      -0.00000367939,     0.00000045263,
       0.00000001595,    -0.00000005905,
      -0.00000045380,    -0.00000368294,
       0.00000005916,     0.00000001603,
      -0.00000011533,     0.00000004827,
       0.00000000117,     0.00000000137,
      -0.00000004724,    -0.00000011372,
      -0.00000000141,     0.00000000115,
      -0.00000000508,     0.00000000095,
       0.00000000098,     0.00000000504,
       0.00000000516,    -0.00000000091,
      -0.00000000092,    -0.00000000513,
      -0.00000014383,     0.00000003693,
       0.00000000095,     0.00000000896,
      -0.00000005256,     0.00000000853,
       0.00000000105,     0.00000000268,
       0.00000000846,     0.00000005229,
       0.00000000261,    -0.00000000105,
       0.00000000661,     0.00000000142,
       0.00000000134,    -0.00000000633,
       0.00000027495,    -0.00000005532,
      -0.00000000003,    -0.00000000640,
       0.00000007754,    -0.00000001422,
      -0.00000000138,    -0.00000000339,
      -0.00000001425,    -0.00000007735,
      -0.00000000326,     0.00000000136,
      -0.00000000437,    -0.00000000093,
      -0.00000000090,     0.00000000425,
      -0.00000207837,     0.00000036390,
       0.00000003594,     0.00000008920,
      -0.00000002863,    -0.00000001357,
      -0.00000000057,     0.00000000048,
      -0.00000135824,     0.00000020069,
       0.00000001318,     0.00000004164,
       0.00000000034,    -0.00000000067,
       0.00000020033,     0.00000135675,
       0.00000004134,    -0.00000001319,
      -0.00000000080,    -0.00000000044,
       0.00000008300,     0.00000002418,
       0.00000002276,    -0.00000007816,
       0.00000554887,    -0.00000101486,
      -0.00000010531,    -0.00000023778,
       0.00000001359,    -0.00000000543,
       0.00000107985,    -0.00000015047,
      -0.00000001041,    -0.00000003255,
      -0.00000000035,     0.00000000078,
      -0.00000015055,    -0.00000107597,
      -0.00000003220,     0.00000001035,
       0.00000000070,     0.00000000032,
      -0.00000005492,    -0.00000001637,
      -0.00000001575,     0.00000005259,
       0.00000000003,     0.00000000025,
       0.00000000006,     0.00000000044,
       0.00000000009,     0.00000000073,
       0.00000000016,     0.00000000155,
       0.00000000070,     0.00000001160,
       0.00000002845,     0.00000004969,
       0.00000004054,    -0.00000000192,
      -0.00000000681,    -0.00000004525,
      -0.00032521522,     0.00005092972,
       0.00000322021,     0.00000971227,
       0.00000008873,    -0.00000018005,
      -0.00009943822,    -0.00061134075,
      -0.00001821485,     0.00000624499,
       0.00000033359,     0.00000016509,
      -0.00014954670,     0.00001413413,
       0.00000047250,     0.00000235334,
       0.00000000217,    -0.00000002810,
       0.00001409337,     0.00014922415,
       0.00000232679,    -0.00000046943,
      -0.00000002879,    -0.00000000093,
       0.00000484331,     0.00000207845,
      -0.00000004412,     0.00000005662,
       0.00000189284,    -0.00000428422,
       0.00000006851,     0.00000004084,
       0.00068742597,    -0.00010694991,
      -0.00000676744,    -0.00002053292,
      -0.00000014711,     0.00000033999,
       0.00000001767,    -0.00000001263,
       0.00006709364,     0.00041160478,
       0.00001225615,    -0.00000418269,
      -0.00000021283,    -0.00000007836,
       0.00000000806,     0.00000001127,
       0.00015842816,    -0.00001271116,
      -0.00000053084,    -0.00000249738,
      -0.00000000134,     0.00000004008,
      -0.00001269983,    -0.00015783636,
      -0.00000246554,     0.00000051963,
       0.00000003619,    -0.00000000061,
      -0.00000313198,    -0.00000140607,
       0.00000002757,    -0.00000004189,
      -0.00000133468,     0.00000289857,
      -0.00000004724,    -0.00000002627,
       0.00000000104,    -0.00000000143,
       0.00000000187,     0.00000000016,
       0.00000000029,    -0.00000000204,
       0.00028612073,    -0.00002302619,
      -0.00000089771,    -0.00000450604,
      -0.00000000500,     0.00000006209,
      -0.00000496883,    -0.00007258706,
      -0.00000114679,     0.00000023238,
       0.00000001682,     0.00000000120,
      -0.00000000052,    -0.00000000070,
       0.00006933510,    -0.00000019318,
      -0.00000000177,    -0.00000000642,
      -0.00000020389,    -0.00006885311,
       0.00000001029,     0.00000000016,
      -0.00000012789,    -0.00000011939,
       0.00000000020,    -0.00000000335,
      -0.00000000015,    -0.00000018751,
      -0.00000000491,     0.00000000187,
      -0.00050680018,     0.00003205964,
       0.00000169258,     0.00000797830,
       0.00000000337,    -0.00000011918,
       0.00000330674,     0.00004668638,
       0.00000073701,    -0.00000014796,
      -0.00000001026,    -0.00000000064,
      -0.00006853376,     0.00000021463,
       0.00000000071,    -0.00000000370,
      -0.00000000009,    -0.00000000112,
       0.00000029874,     0.00006868515,
       0.00000000020,    -0.00000000337,
      -0.00000000814,     0.00000008459,
       0.00000000121,     0.00000000373,
       0.00000003874,     0.00000013219,
       0.00000000416,    -0.00000000186,
       0.00000000105,    -0.00000000003,
      -0.00000000007,     0.00000000103,
       0.00002909422,    -0.00000317939,
      -0.00000013846,     0.00000045831,
      -0.00000036333,    -0.00000400330,
       0.00000006307,     0.00000001798,
      -0.00000000128,    -0.00000000005,
      -0.00000264469,    -0.00000001591,
      -0.00000000151,     0.00000000055,
      -0.00000000968,    -0.00000273769,
       0.00000000045,     0.00000000060,
       0.00000004497,    -0.00000003559,
      -0.00000000064,    -0.00000000193,
       0.00000001907,     0.00000000846,
       0.00000000218,    -0.00000000032,
      -0.00021649129,     0.00002492036,
       0.00000093727,    -0.00000347577,
      -0.00000000869,     0.00000005841,
      -0.00000029444,    -0.00000183530,
       0.00000003018,     0.00000000750,
      -0.00000136693,     0.00000002373,
       0.00000001314,    -0.00000131461,
      -0.00000005387,     0.00000003499,
       0.00000000071,     0.00000000098,
      -0.00000000067,     0.00000001192,
      -0.00000000145,     0.00000000014,
       0.00000001965,    -0.00000001276,
      -0.00000000315,    -0.00000000988,
      -0.00000000175,     0.00000000307,
      -0.00000000325,    -0.00000000218,
       0.00000001448,    -0.00000000448,
       0.00000000436,     0.00000001426,
       0.00000028391,    -0.00000005581,
       0.00000000210,     0.00000001149,
       0.00000001727,    -0.00000000132,
       0.00000000139,     0.00000001766,
      -0.00000002051,     0.00000000910,
      -0.00000000906,    -0.00000002039,
      -0.00000000087,    -0.00000000137,
       0.00000000009,    -0.00000000005,
      -0.00000000144,    -0.00000000492,
       0.00000000020,    -0.00000000003,
       0.00000000493,    -0.00000000142,
       0.00000000003,     0.00000000020,
      -0.00000000019,    -0.00000000020,
       0.00000000021,    -0.00000000025,
       0.00000000020,    -0.00000000035,
       0.00000000035,     0.00000000020,
       0.00000000008,    -0.00000000170,
       0.00000000170,     0.00000000008,
       0.00000000242,     0.00000000220,
      -0.00000000026,     0.00000000045,
      -0.00000001733,     0.00000001086,
      -0.00000000115,    -0.00000000157,
      -0.00000001082,    -0.00000001750,
       0.00000000158,    -0.00000000116,
      -0.00000000044,    -0.00000000020,
       0.00000000021,    -0.00000000042,
       0.00000000244,    -0.00000000726,
      -0.00000001166,     0.00000000941,
       0.00000000034,    -0.00000000222,
      -0.00000008052,    -0.00000001696,
      -0.00000000196,    -0.00000000389,
       0.00000001992,    -0.00000008301,
       0.00000000455,    -0.00000000066,
       0.00000000138,    -0.00000000186,
       0.00000000182,     0.00000000136,
      -0.00000008699,    -0.00000001788,
       0.00000000994,    -0.00000001925,
       0.00000002062,    -0.00000001335,
       0.00000000159,     0.00000000916,
      -0.00000000187,    -0.00000000090,
       0.00000000355,    -0.00000013797,
       0.00000001384,     0.00000001704,
      -0.00000000201,     0.00000000045,
       0.00000012544,     0.00000001190,
      -0.00000001941,     0.00000000832,
       0.00000000038,    -0.00000000306,
       0.00000001990,    -0.00000001073,
      -0.00000000191,    -0.00000000063,
       0.00000001123,     0.00000001985,
       0.00000000053,    -0.00000000188,
      -0.00001151749,    -0.00000207412,
      -0.00000012845,    -0.00000019289,
      -0.00000001585,    -0.00000009354,
       0.00000009264,    -0.00000016283,
      -0.00000000977,    -0.00000001179,
       0.00000000491,    -0.00000000073,
       0.00000547112,     0.00000126648,
      -0.00000009027,     0.00000004546,
       0.00000001364,     0.00000000829,
      -0.00000000097,     0.00000000126,
      -0.00000126824,     0.00000546998,
      -0.00000004452,    -0.00000009101,
      -0.00000000824,     0.00000001379,
      -0.00000000128,    -0.00000000102,
      -0.00000010321,     0.00000006948,
       0.00000000778,     0.00000000201,
      -0.00000007074,    -0.00000010240,
      -0.00000000169,     0.00000000765,
       0.00000000338,     0.00000000179,
       0.00000018988,    -0.00000000192,
      -0.00000000495,     0.00000000049,
       0.00000000229,     0.00000019054,
      -0.00000000053,    -0.00000000494,
      -0.00000000390,     0.00000000091,
      -0.00000000086,    -0.00000000382,
       0.00000000016,     0.00000000073,
       0.00000000009,    -0.00000000092,
      -0.00000000041,     0.00000000061,
       0.00000000065,     0.00000000048,
       0.00000000046,     0.00000000239,
       0.00000000205,    -0.00000000088,
       0.00000000240,    -0.00000000043,
      -0.00000000087,    -0.00000000206,
      -0.00000000117,     0.00000000030,
       0.00000000049,     0.00000000100,
       0.00000000117,     0.00000000045,
       0.00000000039,    -0.00000000120,
       0.00000000382,     0.00000002127,
       0.00000001859,    -0.00000000768,
      -0.00000000533,    -0.00000000747,
       0.00000000283,    -0.00000000069,
      -0.00000000115,    -0.00000000248,
      -0.00000000100,     0.00000000076,
      -0.00000000515,    -0.00000000189,
      -0.00000000057,     0.00000000432,
       0.00000000158,    -0.00000000010,
      -0.00000000186,     0.00000000513,
       0.00000000431,     0.00000000052,
      -0.00000000021,    -0.00000000161,
       0.00000000847,     0.00000000305,
       0.00000000086,    -0.00000000711,
       0.00000000090,    -0.00000000278,
      -0.00000000231,    -0.00000000023,
      -0.00000000117,     0.00000000087,
       0.00000000073,     0.00000000062,
       0.00000000113,     0.00000000105,
       0.00000000048,    -0.00000000089,
      -0.00000004599,    -0.00000001617,
      -0.00000000437,     0.00000003855,
       0.00000001467,    -0.00000000157,
      -0.00000000215,     0.00000000664,
       0.00000000551,     0.00000000052,
      -0.00000000026,    -0.00000000209,
       0.00000000509,    -0.00000000422,
      -0.00000000343,    -0.00000000266,
      -0.00000000056,     0.00000000118,
      -0.00000000469,    -0.00000000488,
      -0.00000000240,     0.00000000381,
       0.00000000127,     0.00000000045,
      -0.00000000987,     0.00000000902,
       0.00000000730,     0.00000000491,
       0.00000000264,     0.00000000304,
       0.00000000155,    -0.00000000216,
       0.00000000065,     0.00000000058,
       0.00000004247,    -0.00000004029,
      -0.00000003234,    -0.00000002093,
      -0.00000000387,     0.00000001090,
      -0.00000000607,    -0.00000000689,
      -0.00000000346,     0.00000000495,
       0.00000000170,     0.00000000065,
       0.00000000029,    -0.00000000320,
      -0.00000000170,     0.00000000047,
      -0.00000000293,    -0.00000000146,
      -0.00000000046,     0.00000000167,
      -0.00000000200,     0.00000000548,
       0.00000000158,     0.00000000043,
      -0.00000000122,    -0.00000000025,
      -0.00000000117,     0.00000000014,
       0.00000000931,    -0.00000002915,
      -0.00000001640,    -0.00000000056,
      -0.00000000386,    -0.00000000103,
       0.00000000000,     0.00000000215,
       0.00000000361,     0.00000000264,
       0.00000000031,    -0.00000000185,
       0.00000000373,    -0.00000000238,
      -0.00000000125,    -0.00000000119,
      -0.00000000606,    -0.00000000612,
      -0.00000000172,     0.00000000173,
       0.00000000088,     0.00000000040,
      -0.00000000041,    -0.00000000091,
       0.00000003106,     0.00000003153,
       0.00000000722,    -0.00000001604,
       0.00000000409,    -0.00000000415,
      -0.00000000210,    -0.00000000094,
      -0.00000000369,     0.00000000013,
       0.00000000095,     0.00000000387,
       0.00000000073,    -0.00000000083,
       0.00000000854,    -0.00000000151,
      -0.00000000032,    -0.00000000234,
      -0.00000003690,     0.00000000657,
       0.00000000637,     0.00000000689,
       0.00000000074,     0.00000000535,
       0.00000000104,    -0.00000000087,
      -0.00000000086,     0.00000000232,
       0.00000000141,     0.00000000190,
       0.00000000321,    -0.00000000391,
      -0.00000000104,    -0.00000000077,
      -0.00000001421,     0.00000001931,
       0.00000000254,     0.00000000182,
      -0.00000000090,    -0.00000000062,
      -0.00000000082,    -0.00000000041,
      -0.00000000462,    -0.00000000847,
      -0.00000000107,     0.00000000054,
      -0.00000000080,     0.00000000135,
       0.00000000069,     0.00000000033,
      -0.00000000580,     0.00000001024,
       0.00000000136,     0.00000000078,
       0.00000000029,     0.00000000160,
       0.00000000138,    -0.00000000058,
       0.00000000160,    -0.00000000024,
      -0.00000000054,    -0.00000000138,
       0.00000000242,     0.00000001518,
       0.00000001333,    -0.00000000526,
       0.00000000193,    -0.00000000041,
      -0.00000000074,    -0.00000000170,
      -0.00000000318,    -0.00000000115,
      -0.00000000033,     0.00000000266,
      -0.00000000115,     0.00000000315,
       0.00000000264,     0.00000000033,
       0.00000000572,     0.00000000205,
       0.00000000056,    -0.00000000480,
       0.00000000058,    -0.00000000172,
      -0.00000000144,    -0.00000000015,
      -0.00000000062,     0.00000000045,
       0.00000000066,     0.00000000054,
      -0.00000003025,    -0.00000001064,
      -0.00000000284,     0.00000002535,
       0.00000000965,    -0.00000000107,
      -0.00000000135,     0.00000000413,
       0.00000000345,     0.00000000034,
       0.00000000271,    -0.00000000230,
      -0.00000000185,    -0.00000000141,
      -0.00000000268,    -0.00000000257,
      -0.00000000121,     0.00000000215,
      -0.00000000559,     0.00000000540,
       0.00000000433,     0.00000000272,
       0.00000000148,     0.00000000163,
       0.00000000082,    -0.00000000120,
       0.00000002382,    -0.00000002411,
      -0.00000001911,    -0.00000001144,
      -0.00000000341,    -0.00000000368,
      -0.00000000182,     0.00000000276,
       0.00000000021,    -0.00000000217,
      -0.00000000115,     0.00000000031,
      -0.00000000194,    -0.00000000107,
      -0.00000000037,     0.00000000112,
      -0.00000000151,     0.00000000394,
       0.00000000107,     0.00000000032,
       0.00000000699,    -0.00000002030,
      -0.00000001147,    -0.00000000068,
      -0.00000000263,    -0.00000000079,
      -0.00000000004,     0.00000000148,
       0.00000000216,     0.00000000150,
       0.00000000014,    -0.00000000110,
       0.00000000223,    -0.00000000138,
      -0.00000000072,    -0.00000000071,
      -0.00000000377,    -0.00000000375,
      -0.00000000101,     0.00000000102,
       0.00000001896,     0.00000001909,
       0.00000000430,    -0.00000000977,
       0.00000000242,    -0.00000000245,
      -0.00000000124,    -0.00000000055,
      -0.00000000185,     0.00000000002,
       0.00000000065,     0.00000000202,
       0.00000000454,    -0.00000000096,
      -0.00000000019,    -0.00000000120,
      -0.00000001952,     0.00000000419,
       0.00000000045,     0.00000000275,
      -0.00000000054,     0.00000000140,
       0.00000000079,     0.00000000121,
      -0.00000000062,    -0.00000000049,
      -0.00000000923,     0.00000001156,
       0.00000000149,     0.00000000117,
      -0.00000000048,     0.00000000076,
       0.00000000074,     0.00000000044,
       0.00000000019,     0.00000000107,
       0.00000000092,    -0.00000000039,
       0.00000000107,    -0.00000000014,
      -0.00000000034,    -0.00000000093,
       0.00000000157,     0.00000001066,
       0.00000000941,    -0.00000000358,
       0.00000000132,    -0.00000000025,
      -0.00000000049,    -0.00000000117,
      -0.00000000195,    -0.00000000069,
      -0.00000000018,     0.00000000162,
      -0.00000000070,     0.00000000193,
       0.00000000161,     0.00000000019,
       0.00000000036,    -0.00000000107,
      -0.00000001940,    -0.00000000666,
      -0.00000000169,     0.00000001625,
      -0.00000000084,     0.00000000255,
       0.00000000214,     0.00000000020,
       0.00000000141,    -0.00000000125,
      -0.00000000098,    -0.00000000073,
      -0.00000000154,    -0.00000000131,
      -0.00000000059,     0.00000000122,
      -0.00000000303,     0.00000000318,
       0.00000000082,     0.00000000085,
       0.00000001276,    -0.00000001417,
      -0.00000001104,    -0.00000000588,
      -0.00000000191,    -0.00000000192,
      -0.00000000092,     0.00000000153,
       0.00000000015,    -0.00000000144,
      -0.00000000127,    -0.00000000075,
       0.00000000494,    -0.00000001378,
      -0.00000000175,    -0.00000000056,
       0.00000000127,     0.00000000084,
       0.00000000130,    -0.00000000080,
      -0.00000000059,     0.00000000060,
       0.00000001145,     0.00000001126,
       0.00000000140,    -0.00000000143,
       0.00000000044,     0.00000000103,
      -0.00000001003,     0.00000000267,
       0.00000000028,     0.00000000138,
      -0.00000000034,     0.00000000083,
       0.00000000044,     0.00000000075,
      -0.00000000584,     0.00000000685,
       0.00000000087,     0.00000000073,
       0.00000000089,    -0.00000000016,
      -0.00000000118,    -0.00000000040,
      -0.00000000042,     0.00000000118,
      -0.00000001219,    -0.00000000403,
      -0.00000000092,     0.00000001019,
      -0.00000000050,     0.00000000156,
       0.00000000131,     0.00000000011,
       0.00000000070,    -0.00000000067,
      -0.00000000089,    -0.00000000065,
       0.00000000650,    -0.00000000821,
      -0.00000000107,    -0.00000000097,
      -0.00000000044,     0.00000000084,
       0.00000000010,    -0.00000000093,
      -0.00000000082,    -0.00000000050,
       0.00000000334,    -0.00000000912,
      -0.00000000114,    -0.00000000038,
       0.00000000073,     0.00000000046,
       0.00000000074,    -0.00000000047,
       0.00000000683,     0.00000000646,
       0.00000000079,    -0.00000000084,
      -0.00000000754,    -0.00000000235,
      -0.00000000029,     0.00000000094,
      -0.00000000059,    -0.00000000047,
      -0.00000000740,     0.00000000170,
       0.00000000174,     0.00000000730,
       0.00000003461,    -0.00000000889,
       0.00000000887,    -0.00000000206,
      -0.00000000207,    -0.00000000882,
      -0.00000026624,     0.00000006341,
       0.00000000759,     0.00000001555,
      -0.00000000424,    -0.00000000441,
      -0.00000015287,     0.00000003302,
       0.00000000283,     0.00000000681,
       0.00000003291,     0.00000015227,
       0.00000000676,    -0.00000000283,
       0.00000001087,     0.00000000218,
       0.00000000206,    -0.00000001031,
       0.00000065382,    -0.00000016085,
      -0.00000001828,    -0.00000003706,
       0.00000000272,     0.00000000173,
       0.00000013687,    -0.00000002823,
      -0.00000000281,    -0.00000000671,
      -0.00000002820,    -0.00000013635,
      -0.00000000660,     0.00000000280,
      -0.00000000716,    -0.00000000143,
      -0.00000000139,     0.00000000693,
      -0.00002667222,     0.00000583922,
       0.00000049230,     0.00000116003,
       0.00000001891,    -0.00000003244,
      -0.00000744600,    -0.00003263478,
      -0.00000141194,     0.00000061715,
       0.00000004241,     0.00000002471,
      -0.00001224740,     0.00000214091,
       0.00000012035,     0.00000037550,
       0.00000000272,    -0.00000000547,
       0.00000212483,     0.00001219352,
       0.00000036959,    -0.00000011936,
      -0.00000000693,    -0.00000000399,
       0.00000061944,     0.00000017664,
       0.00000016496,    -0.00000057310,
       0.00005640517,    -0.00001231537,
      -0.00000104182,    -0.00000246430,
      -0.00000003839,     0.00000005991,
       0.00000499387,     0.00002184017,
       0.00000095324,    -0.00000041506,
      -0.00000002549,    -0.00000001633,
       0.00001235858,    -0.00000211285,
      -0.00000012232,    -0.00000037438,
      -0.00000000326,     0.00000000709,
      -0.00000210429,    -0.00001229061,
      -0.00000036886,     0.00000012021,
       0.00000000637,     0.00000000295,
      -0.00000040621,    -0.00000012039,
      -0.00000011566,     0.00000038603,
       0.00003391475,    -0.00000587002,
      -0.00000032420,    -0.00000102814,
      -0.00000000941,     0.00000001847,
      -0.00000136392,    -0.00000900722,
      -0.00000027217,     0.00000008491,
       0.00000000497,     0.00000000242,
       0.00001142445,    -0.00000122457,
      -0.00000003374,    -0.00000018218,
      -0.00000000025,     0.00000000226,
      -0.00000121023,    -0.00001133678,
      -0.00000017730,     0.00000003235,
       0.00000000222,     0.00000000013,
      -0.00000030429,    -0.00000013205,
       0.00000000266,    -0.00000000347,
      -0.00000011824,     0.00000025883,
      -0.00000000446,    -0.00000000240,
      -0.00006285926,     0.00001071836,
       0.00000060074,     0.00000190658,
       0.00000001583,    -0.00000003103,
       0.00000091317,     0.00000607406,
       0.00000018358,    -0.00000005707,
      -0.00000000296,    -0.00000000149,
      -0.00001140441,     0.00000119702,
       0.00000003371,     0.00000018152,
       0.00000000026,    -0.00000000284,
       0.00000119584,     0.00001135798,
       0.00000017758,    -0.00000003244,
      -0.00000000252,    -0.00000000008,
       0.00000019307,     0.00000009017,
      -0.00000000158,     0.00000000270,
       0.00000008498,    -0.00000017503,
       0.00000000310,     0.00000000149,
       0.00000000001,     0.00000000007,
       0.00000000004,     0.00000000051,
       0.00000000025,     0.00000000282,
       0.00000000166,     0.00000001073,
       0.00000000263,    -0.00000000028,
       0.00000000048,     0.00000000131,
       0.00012614193,    -0.00001365551,
      -0.00000037199,    -0.00000199172,
      -0.00000000402,     0.00000002739,
      -0.00000191824,    -0.00002111650,
      -0.00000033404,     0.00000006255,
       0.00000000505,     0.00000000054,
       0.00002469606,    -0.00000002984,
       0.00000000286,     0.00000000562,
      -0.00000012878,    -0.00002503799,
      -0.00000000395,     0.00000000221,
      -0.00000004671,    -0.00000004487,
      -0.00000000006,    -0.00000000113,
       0.00000000153,    -0.00000008315,
      -0.00000000173,     0.00000000086,
      -0.00019786256,     0.00002003084,
       0.00000058297,     0.00000311781,
       0.00000000469,    -0.00000004153,
       0.00000123891,     0.00001398676,
       0.00000022086,    -0.00000004175,
      -0.00000000303,    -0.00000000031,
      -0.00002537301,     0.00000008763,
       0.00000000104,     0.00000000487,
       0.00000006891,     0.00002507613,
      -0.00000000603,     0.00000000189,
      -0.00000000659,     0.00000003218,
       0.00000000056,     0.00000000133,
       0.00000001466,     0.00000005522,
       0.00000000147,    -0.00000000079,
       0.00000000014,    -0.00000000273,
       0.00053293916,    -0.00000092395,
       0.00000001503,     0.00000000782,
      -0.00000014861,    -0.00006111859,
       0.00000000108,    -0.00000000070,
       0.00000000167,    -0.00000000004,
       0.00000097944,    -0.00000005736,
      -0.00000000475,    -0.00000002471,
      -0.00000007312,     0.00000144946,
       0.00000001643,    -0.00000000419,
      -0.00000007877,    -0.00000011205,
       0.00000000067,    -0.00000000341,
       0.00000003641,    -0.00000025591,
       0.00000000128,     0.00000000249,
      -0.00056976611,     0.00000100204,
      -0.00000004953,    -0.00000000927,
       0.00000010344,     0.00003802335,
      -0.00000000087,    -0.00000000057,
      -0.00000000103,     0.00000000004,
      -0.00000057527,     0.00000008280,
       0.00000000297,     0.00000001098,
       0.00000006081,    -0.00000067433,
      -0.00000000887,     0.00000000122,
      -0.00000006059,     0.00000007610,
       0.00000000073,     0.00000000229,
       0.00000002246,     0.00000017307,
       0.00000000050,    -0.00000000167,
      -0.00000000009,    -0.00000000025,
      -0.00000000011,    -0.00000000067,
       0.00000000067,    -0.00000000010,
      -0.00000000002,    -0.00000000003,
       0.00000000003,    -0.00000000003,
      -0.00000000249,     0.00000000091,
      -0.00000000091,    -0.00000000249,
      -0.00000000039,    -0.00000000332,
      -0.00000000477,    -0.00000000262,
       0.00000000263,    -0.00000000451,
      -0.00000000655,    -0.00000000429,
       0.00000000761,    -0.00000001276,
       0.00000000022,     0.00000000140,
       0.00000002634,    -0.00000003093,
       0.00000000240,     0.00000000240,
       0.00000003030,     0.00000002664,
      -0.00000000231,     0.00000000224,
       0.00000000157,    -0.00000000165,
       0.00000000168,     0.00000000193,
       0.00000041508,     0.00000201081,
      -0.00000019544,    -0.00000003615,
       0.00000000300,    -0.00000001902,
      -0.00000014599,     0.00000021099,
      -0.00000000339,    -0.00000000740,
       0.00000000113,     0.00000000037,
       0.00000278300,     0.00000024911,
      -0.00000002270,    -0.00000000821,
       0.00000000127,    -0.00000000085,
      -0.00000026547,     0.00000272016,
       0.00000000823,    -0.00000002108,
       0.00000000083,     0.00000000122,
       0.00000002246,    -0.00000001062,
       0.00000000264,     0.00000000510,
       0.00000261736,     0.00000005143,
      -0.00000000353,     0.00000004647,
       0.00000004359,    -0.00000003343,
       0.00000000109,    -0.00000000121,
      -0.00000170156,    -0.00000000054,
      -0.00000000247,     0.00000000045,
       0.00000000183,    -0.00000184810,
       0.00000000045,     0.00000000109,
      -0.00000002065,     0.00000001323,
       0.00000000037,     0.00000000067,
      -0.00000000848,    -0.00000001076,
      -0.00000000073,     0.00000000028,
      -0.00000011346,     0.00000000146,
      -0.00000000990,    -0.00000001573,
      -0.00000001418,    -0.00000004357,
      -0.00000000028,     0.00000000149,
       0.00000016608,    -0.00000002752,
      -0.00000000059,     0.00000000221,
       0.00000002591,     0.00000016326,
      -0.00000000219,    -0.00000000055,
       0.00000010288,    -0.00000004299,
      -0.00000000104,    -0.00000000122,
       0.00000004237,     0.00000010171,
       0.00000000125,    -0.00000000103,
       0.00000869594,    -0.00000004620,
      -0.00000022074,     0.00000002836,
       0.00000002638,    -0.00000000548,
       0.00000000879,     0.00000025428,
      -0.00000000034,    -0.00000000686,
       0.00000004425,    -0.00000000192,
       0.00000000178,     0.00000004841,
      -0.00000000083,    -0.00000000018,
      -0.00000006899,     0.00000002857,
       0.00000000071,     0.00000000082,
      -0.00000002939,    -0.00000006977,
      -0.00000000084,     0.00000000070,
       0.00000002392,    -0.00000000013,
      -0.00000001094,    -0.00000000154,
      -0.00000000016,    -0.00000000092,
       0.00000000159,    -0.00000001091,
       0.00000000092,    -0.00000000016,
       0.00000000008,     0.00000000011,
       0.00000000015,     0.00000000041,
      -0.00000000041,     0.00000000015,
       0.00000000144,    -0.00000000124,
       0.00000000124,     0.00000000144,
       0.00000000826,     0.00000000070,
      -0.00000000071,     0.00000000825,
      -0.00000000051,     0.00000000147,
      -0.00000000777,     0.00000001543,
      -0.00000000029,    -0.00000000352,
      -0.00000001543,    -0.00000000777,
       0.00000000355,    -0.00000000025,
      -0.00000000228,     0.00000000009,
      -0.00000000011,    -0.00000000230,
       0.00000080723,     0.00000003558,
       0.00000004584,     0.00000003628,
      -0.00000000169,     0.00000001331,
      -0.00000000042,     0.00000000219,
      -0.00000036248,    -0.00000007199,
      -0.00000000395,    -0.00000000434,
      -0.00000000066,    -0.00000000323,
       0.00000007206,    -0.00000036280,
       0.00000000424,    -0.00000000395,
       0.00000000326,    -0.00000000070,
       0.00000001240,    -0.00000000287,
      -0.00000000141,    -0.00000000059,
       0.00000000295,     0.00000001240,
       0.00000000056,    -0.00000000140,
      -0.00000000995,     0.00000000381,
      -0.00000000382,    -0.00000000996,
      -0.00000000100,     0.00000000031,
       0.00000000031,     0.00000000100,
       0.00000000099,    -0.00000000030,
      -0.00000000030,    -0.00000000099,
      -0.00000003039,     0.00000000941,
      -0.00000001629,     0.00000000458,
       0.00000000057,     0.00000000103,
       0.00000000461,     0.00000001602,
       0.00000000111,    -0.00000000062,
       0.00000000122,     0.00000000011,
       0.00000000011,    -0.00000000121,
       0.00000007339,    -0.00000002284,
       0.00000001658,    -0.00000000456,
      -0.00000000052,    -0.00000000101,
      -0.00000000455,    -0.00000001646,
      -0.00000000102,     0.00000000052,
      -0.00000231624,     0.00000066081,
       0.00000006945,     0.00000013584,
      -0.00000061563,    -0.00000206525,
      -0.00000011943,     0.00000006156,
      -0.00000105969,     0.00000026427,
       0.00000002014,     0.00000004751,
       0.00000026237,     0.00000105215,
       0.00000004675,    -0.00000001994,
       0.00000006623,     0.00000001236,
       0.00000001154,    -0.00000006191,
       0.00000492218,    -0.00000140798,
      -0.00000014815,    -0.00000028475,
       0.00000041625,     0.00000138541,
       0.00000008128,    -0.00000004320,
      -0.00000000322,    -0.00000000217,
       0.00000105947,    -0.00000025721,
      -0.00000002006,    -0.00000004704,
      -0.00000025567,    -0.00000105232,
      -0.00000004620,     0.00000001991,
      -0.00000004351,    -0.00000000824,
      -0.00000000794,     0.00000004168,
       0.00000400347,    -0.00000098958,
      -0.00000007495,    -0.00000017700,
      -0.00000024363,    -0.00000110323,
      -0.00000004832,     0.00000002001,
       0.00000000093,     0.00000000051,
       0.00000151631,    -0.00000029553,
      -0.00000001483,    -0.00000004703,
      -0.00000029231,    -0.00000150301,
      -0.00000004580,     0.00000001437,
       0.00000000082,     0.00000000046,
      -0.00000006692,    -0.00000001876,
      -0.00000001743,     0.00000006082,
      -0.00000772410,     0.00000188949,
       0.00000014531,     0.00000034539,
       0.00000016637,     0.00000075498,
       0.00000003312,    -0.00000001364,
      -0.00000151637,     0.00000028963,
       0.00000001470,     0.00000004644,
       0.00000000042,    -0.00000000089,
       0.00000028791,     0.00000150625,
       0.00000004556,    -0.00000001436,
      -0.00000000079,    -0.00000000038,
       0.00000004349,     0.00000001288,
       0.00000001237,    -0.00000004102,
       0.00001545306,    -0.00000303788,
      -0.00000014898,    -0.00000047180,
      -0.00000044572,    -0.00000266218,
      -0.00000008099,     0.00000002511,
       0.00000000149,     0.00000000073,
       0.00000448039,    -0.00000052319,
      -0.00000001300,    -0.00000007203,
      -0.00000000014,     0.00000000093,
      -0.00000052163,    -0.00000446381,
      -0.00000006912,     0.00000001207,
      -0.00000010158,    -0.00000004443,
       0.00000000085,    -0.00000000113,
      -0.00000003904,     0.00000008232,
      -0.00000000155,    -0.00000000075,
      -0.00002727522,     0.00000517638,
       0.00000025971,     0.00000083023,
       0.00000030231,     0.00000181893,
       0.00000005526,    -0.00000001708,
      -0.00000000087,    -0.00000000045,
      -0.00000453564,     0.00000052542,
       0.00000001301,     0.00000007237,
       0.00000000013,    -0.00000000111,
       0.00000051925,     0.00000449322,
       0.00000007007,    -0.00000001223,
      -0.00000000097,    -0.00000000005,
       0.00000006301,     0.00000003047,
      -0.00000000048,     0.00000000093,
       0.00000002850,    -0.00000005572,
       0.00000000109,     0.00000000044,
       0.00007932240,    -0.00000955234,
      -0.00000021536,    -0.00000125016,
      -0.00000000276,     0.00000001654,
      -0.00000100472,    -0.00000992720,
      -0.00000015719,     0.00000002823,
       0.00000000243,     0.00000000029,
       0.00001183409,    -0.00000005564,
      -0.00000000072,    -0.00000000485,
      -0.00000001929,    -0.00001151874,
       0.00000000563,    -0.00000000187,
      -0.00000002284,    -0.00000002239,
       0.00000000160,    -0.00000004684,
      -0.00000000081,     0.00000000049,
      -0.00011090487,     0.00001238144,
       0.00000031109,     0.00000174724,
       0.00000065332,     0.00000658719,
       0.00000010403,    -0.00000001892,
      -0.00000000144,    -0.00000000017,
      -0.00001205319,     0.00000004982,
       0.00000000064,     0.00000000299,
       0.00000002874,     0.00001184072,
      -0.00000000366,     0.00000000117,
      -0.00000000601,     0.00000001656,
       0.00000000035,     0.00000000067,
       0.00000000774,     0.00000003089,
       0.00000000074,    -0.00000000046,
       0.00000000044,    -0.00000000142,
       0.00000000143,    -0.00000000143,
       0.00000000121,     0.00000000159,
       0.00029696851,    -0.00000041695,
      -0.00000008197,    -0.00002947362,
       0.00000000090,    -0.00000000201,
       0.00000000107,     0.00000000003,
       0.00000034464,    -0.00000004843,
      -0.00000000176,    -0.00000001849,
      -0.00000001464,     0.00000111130,
       0.00000000564,    -0.00000000302,
      -0.00000004305,    -0.00000005975,
       0.00000000032,    -0.00000000183,
       0.00000002070,    -0.00000013473,
       0.00000000075,     0.00000000129,
      -0.00031426224,     0.00000061823,
       0.00000005090,     0.00001859325,
      -0.00000000044,     0.00000000079,
      -0.00000017094,     0.00000005603,
       0.00000000109,     0.00000001227,
       0.00000001215,    -0.00000074958,
      -0.00000000244,     0.00000000177,
      -0.00000003410,     0.00000004048,
       0.00000000043,     0.00000000123,
       0.00000001217,     0.00000009079,
       0.00000000025,    -0.00000000086,
      -0.00001470036,     0.00000031644,
       0.00000004311,    -0.00000024242,
      -0.00000002340,     0.00000067500,
      -0.00000001171,    -0.00000000103,
       0.00000337435,    -0.00000001035,
      -0.00000000279,     0.00000332927,
       0.00000000023,     0.00000000093,
       0.00000002142,    -0.00000001112,
      -0.00000000037,    -0.00000000064,
       0.00000000798,     0.00000001482,
       0.00000000070,    -0.00000000031,
       0.00000911489,    -0.00000017997,
      -0.00000002358,     0.00000015702,
       0.00000001790,    -0.00000042409,
       0.00000000728,     0.00000000064,
      -0.00000125741,     0.00000000475,
       0.00000000163,    -0.00000122439,
      -0.00000001260,     0.00000000755,
      -0.00000000618,    -0.00000001008,
       0.00000003808,    -0.00000001531,
      -0.00000000153,    -0.00000000285,
      -0.00000001011,     0.00000000264,
      -0.00000000270,    -0.00000001022,
       0.00000000195,    -0.00000000014,
       0.00000000011,     0.00000000191,
      -0.00000048367,     0.00000018490,
      -0.00000000382,    -0.00000001069,
       0.00000000137,    -0.00000000120,
       0.00000000130,     0.00000000141,
      -0.00000000132,     0.00000000010,
      -0.00000000002,    -0.00000000120,
       0.00000000092,    -0.00000000081,
       0.00000000082,     0.00000000094,
      -0.00000004748,     0.00000000712,
       0.00000000004,    -0.00000000155,
       0.00000002711,     0.00000000139,
       0.00000000164,     0.00000000115,
      -0.00000000144,     0.00000002507,
      -0.00000000083,     0.00000000160,
      -0.00000000088,    -0.00000000033,
       0.00000000032,    -0.00000000088,
      -0.00000000180,     0.00000000065,
       0.00000000065,     0.00000000180,
       0.00000000179,    -0.00000000064,
      -0.00000000062,    -0.00000000174,
      -0.00000020400,     0.00000007303,
       0.00000001053,     0.00000001664,
      -0.00000005310,    -0.00000014121,
      -0.00000001305,     0.00000000826,
      -0.00000009334,     0.00000002984,
       0.00000000328,     0.00000000584,
       0.00000002984,     0.00000009141,
       0.00000000628,    -0.00000000357,
       0.00000000628,     0.00000000047,
       0.00000000047,    -0.00000000628,
       0.00000044458,    -0.00000015872,
      -0.00000002014,    -0.00000003210,
       0.00000003672,     0.00000009766,
       0.00000000819,    -0.00000000514,
       0.00000009481,    -0.00000002981,
      -0.00000000300,    -0.00000000579,
      -0.00000002961,    -0.00000009386,
      -0.00000000580,     0.00000000301,
      -0.00000000423,    -0.00000000032,
      -0.00000000032,     0.00000000423,
       0.00000046144,    -0.00000014751,
      -0.00000001457,    -0.00000002759,
      -0.00000003796,    -0.00000013078,
      -0.00000000784,     0.00000000402,
       0.00000018546,    -0.00000005119,
      -0.00000000363,    -0.00000000839,
      -0.00000005061,    -0.00000018354,
      -0.00000000813,     0.00000000354,
      -0.00000001054,    -0.00000000185,
      -0.00000000172,     0.00000000975,
      -0.00000091877,     0.00000029254,
       0.00000002831,     0.00000005355,
       0.00000002627,     0.00000009068,
       0.00000000517,    -0.00000000270,
      -0.00000018486,     0.00000004997,
       0.00000000364,     0.00000000839,
       0.00000004956,     0.00000018333,
       0.00000000821,    -0.00000000359,
       0.00000000692,     0.00000000126,
       0.00000000124,    -0.00000000659,
       0.00000178403,    -0.00000049595,
      -0.00000003444,    -0.00000007948,
      -0.00000007549,    -0.00000031383,
      -0.00000001397,     0.00000000588,
       0.00000061054,    -0.00000012989,
      -0.00000000600,    -0.00000001905,
      -0.00000012819,    -0.00000060431,
      -0.00000001841,     0.00000000574,
      -0.00000002384,    -0.00000000659,
      -0.00000000610,     0.00000002128,
      -0.00000335366,     0.00000091243,
       0.00000006353,     0.00000014834,
       0.00000005238,     0.00000021846,
       0.00000000950,    -0.00000000394,
      -0.00000061489,     0.00000012902,
       0.00000000600,     0.00000001894,
       0.00000012773,     0.00000060902,
       0.00000001850,    -0.00000000582,
       0.00000001535,     0.00000000457,
       0.00000000439,    -0.00000001435,
       0.00000977892,    -0.00000211654,
      -0.00000009346,    -0.00000029773,
      -0.00000023758,    -0.00000128880,
      -0.00000003943,     0.00000001218,
       0.00000000072,     0.00000000035,
       0.00000233507,    -0.00000029249,
      -0.00000000660,    -0.00000003754,
      -0.00000028398,    -0.00000229615,
      -0.00000003554,     0.00000000589,
      -0.00000004598,    -0.00000002030,
      -0.00000001751,     0.00000003526,
      -0.00000000073,    -0.00000000031,
      -0.00001566266,     0.00000325027,
       0.00000014968,     0.00000047944,
       0.00000016163,     0.00000088313,
       0.00000002691,    -0.00000000829,
      -0.00000236608,     0.00000029386,
       0.00000000662,     0.00000003780,
       0.00000028864,     0.00000233738,
       0.00000003638,    -0.00000000610,
       0.00000002785,     0.00000001407,
       0.00000001306,    -0.00000002394,
       0.00004952724,    -0.00000637757,
      -0.00000012974,    -0.00000078079,
      -0.00000058835,    -0.00000535198,
      -0.00000008482,     0.00000001466,
       0.00000000134,     0.00000000018,
       0.00000614985,    -0.00000003092,
      -0.00000000041,    -0.00000000353,
      -0.00000000269,    -0.00000593777,
       0.00000000393,    -0.00000000152,
      -0.00000001231,    -0.00000001234,
       0.00000000127,    -0.00000002831,
      -0.00006627370,     0.00000794818,
       0.00000017950,     0.00000104359,
       0.00000038411,     0.00000356279,
       0.00000005622,    -0.00000000984,
      -0.00000630441,     0.00000002792,
       0.00000000025,     0.00000000214,
       0.00000000892,     0.00000615668,
      -0.00000000249,     0.00000000103,
      -0.00000000460,     0.00000000939,
       0.00000000447,     0.00000001859,
       0.00016880085,    -0.00000015912,
       0.00000001730,     0.00000000291,
      -0.00000003628,    -0.00001526533,
       0.00000000032,    -0.00000000126,
       0.00000027140,    -0.00000005865,
      -0.00000000102,    -0.00000001593,
       0.00000002777,     0.00000099086,
       0.00000000401,    -0.00000000257,
      -0.00000002408,    -0.00000003332,
       0.00000000017,    -0.00000000103,
       0.00000001211,    -0.00000007397,
       0.00000000045,     0.00000000069,
      -0.00017938320,     0.00000023855,
       0.00000002427,     0.00000969651,
      -0.00000000023,     0.00000000085,
      -0.00000000669,     0.00000003815,
       0.00000000030,     0.00000001016,
      -0.00000000377,    -0.00000062248,
       0.00000000010,     0.00000000157,
      -0.00000001980,     0.00000002258,
       0.00000000675,     0.00000004985,
      -0.00001324129,     0.00000028532,
       0.00000003988,    -0.00000021701,
      -0.00000000824,     0.00000073118,
      -0.00000001235,    -0.00000000166,
       0.00000219982,    -0.00000000679,
      -0.00000000256,     0.00000215816,
       0.00000000033,     0.00000000070,
       0.00000001445,    -0.00000000649,
       0.00000000522,     0.00000001188,
       0.00000940272,    -0.00000018669,
      -0.00000002363,     0.00000015716,
       0.00000000667,    -0.00000048150,
       0.00000000803,     0.00000000106,
      -0.00000084921,     0.00000000346,
       0.00000000210,    -0.00000081642,
      -0.00000000778,     0.00000000441,
      -0.00000000434,    -0.00000000801,
       0.00000084037,    -0.00000016499,
      -0.00000000925,     0.00000001928,
      -0.00000001254,    -0.00000004369,
       0.00000000069,     0.00000000076,
      -0.00000019407,     0.00000000792,
       0.00000000061,    -0.00000000316,
      -0.00000000815,    -0.00000019389,
       0.00000000311,     0.00000000063,
       0.00000003668,    -0.00000001531,
       0.00000001517,     0.00000003647,
      -0.00000047428,     0.00000009537,
       0.00000000831,     0.00000002846,
       0.00000005190,    -0.00000000262,
      -0.00000000015,     0.00000000087,
       0.00000000300,     0.00000005201,
      -0.00000000087,    -0.00000000016,
      -0.00000002482,     0.00000001036,
      -0.00000001022,    -0.00000002458,
       0.00000002759,    -0.00000001266,
       0.00000000040,     0.00000000075,
       0.00000004050,    -0.00000000081,
      -0.00000000001,     0.00000000168,
      -0.00000000232,     0.00000000027,
      -0.00000000028,    -0.00000000237,
      -0.00000001991,     0.00000000895,
      -0.00000000585,    -0.00000001301,
      -0.00000000870,     0.00000000341,
       0.00000000358,     0.00000000904,
       0.00000004111,    -0.00000001800,
       0.00000000383,     0.00000000874,
       0.00000000863,    -0.00000000338,
      -0.00000000330,    -0.00000000841,
       0.00000005155,    -0.00000002057,
      -0.00000000565,    -0.00000001539,
      -0.00000000077,     0.00000000052,
       0.00000002150,    -0.00000000755,
      -0.00000000079,    -0.00000000143,
      -0.00000000748,    -0.00000002093,
      -0.00000000153,     0.00000000086,
      -0.00000000132,    -0.00000000010,
      -0.00000000010,     0.00000000132,
      -0.00000010737,     0.00000004240,
       0.00000000390,     0.00000001072,
      -0.00000002180,     0.00000000757,
       0.00000000074,     0.00000000137,
       0.00000000749,     0.00000002154,
       0.00000000136,    -0.00000000074,
       0.00000019890,    -0.00000007090,
      -0.00000001112,    -0.00000003561,
      -0.00000000210,     0.00000000110,
       0.00000007446,    -0.00000002241,
      -0.00000000150,    -0.00000000338,
      -0.00000002208,    -0.00000007350,
      -0.00000000328,     0.00000000146,
      -0.00000000385,    -0.00000000063,
      -0.00000000059,     0.00000000352,
      -0.00000038645,     0.00000013608,
       0.00000001235,     0.00000002269,
       0.00000000775,     0.00000002487,
       0.00000000142,    -0.00000000076,
      -0.00000007454,     0.00000002208,
       0.00000000149,     0.00000000339,
       0.00000002185,     0.00000007380,
       0.00000000331,    -0.00000000146,
       0.00000000252,     0.00000000044,
       0.00000000043,    -0.00000000237,
       0.00000111782,    -0.00000034052,
      -0.00000002180,    -0.00000004946,
      -0.00000004011,    -0.00000015282,
      -0.00000000684,     0.00000000292,
       0.00000033187,    -0.00000007565,
      -0.00000000328,    -0.00000001037,
      -0.00000007418,    -0.00000032702,
      -0.00000000996,     0.00000000310,
      -0.00000001161,    -0.00000000318,
      -0.00000000294,     0.00000001016,
      -0.00000194090,     0.00000057520,
       0.00000003767,     0.00000008656,
       0.00000002791,     0.00000010672,
       0.00000000466,    -0.00000000196,
      -0.00000033496,     0.00000007555,
       0.00000000327,     0.00000001038,
       0.00000007461,     0.00000033132,
       0.00000001010,    -0.00000000316,
       0.00000000741,     0.00000000223,
       0.00000000214,    -0.00000000688,
       0.00000637251,    -0.00000148244,
      -0.00000006116,    -0.00000019509,
      -0.00000014565,    -0.00000072838,
      -0.00000002241,     0.00000000691,
       0.00000133197,    -0.00000017521,
      -0.00000000369,    -0.00000002149,
      -0.00000016852,    -0.00000130426,
      -0.00000002007,     0.00000000317,
      -0.00000002336,    -0.00000001043,
      -0.00000000883,     0.00000001682,
      -0.00000975781,     0.00000217185,
       0.00000009353,     0.00000029980,
       0.00000009927,     0.00000050093,
       0.00000001528,    -0.00000000469,
      -0.00000135624,     0.00000017729,
       0.00000000368,     0.00000002171,
       0.00000017309,     0.00000133641,
       0.00000002075,    -0.00000000333,
       0.00000001378,     0.00000000732,
       0.00000000674,    -0.00000001146,
       0.00003148381,    -0.00000425997,
      -0.00000007871,    -0.00000049577,
      -0.00000036004,    -0.00000307917,
      -0.00000004881,     0.00000000815,
       0.00000337261,    -0.00000002040,
      -0.00000000026,    -0.00000000261,
       0.00000000802,    -0.00000319932,
       0.00000000281,    -0.00000000117,
      -0.00000000698,    -0.00000000715,
       0.00000000093,    -0.00000001761,
      -0.00004100531,     0.00000517386,
       0.00000010692,     0.00000064600,
       0.00000023559,     0.00000205270,
       0.00000003238,    -0.00000000546,
      -0.00000345683,     0.00000001528,
       0.00000000008,     0.00000000154,
      -0.00000000015,     0.00000335484,
      -0.00000000170,     0.00000000082,
      -0.00000000331,     0.00000000557,
       0.00000000268,     0.00000001154,
       0.00009781311,     0.00000001099,
      -0.00000001480,    -0.00000823783,
       0.00000000015,    -0.00000000109,
       0.00000014700,    -0.00000004435,
      -0.00000000051,    -0.00000001169,
       0.00000002926,     0.00000072605,
       0.00000000215,    -0.00000000186,
      -0.00000001365,    -0.00000001909,
       0.00000000722,    -0.00000004160,
      -0.00010396542,     0.00000004853,
       0.00000000965,     0.00000526117,
       0.00000004887,     0.00000002585,
       0.00000000000,     0.00000000778,
      -0.00000000886,    -0.00000047778,
       0.00000000094,     0.00000000122,
      -0.00000001167,     0.00000001293,
       0.00000000377,     0.00000002805,
      -0.00001051604,     0.00000023310,
       0.00000003012,    -0.00000017283,
      -0.00000000287,     0.00000061974,
      -0.00000001034,    -0.00000000153,
       0.00000136959,    -0.00000000345,
      -0.00000000326,     0.00000133141,
       0.00000000969,    -0.00000000384,
       0.00000000341,     0.00000000891,
       0.00000832323,    -0.00000016597,
      -0.00000002214,     0.00000013815,
       0.00000000209,    -0.00000040835,
       0.00000000676,     0.00000000098,
      -0.00000055127,     0.00000000235,
       0.00000000197,    -0.00000052030,
      -0.00000000481,     0.00000000260,
      -0.00000000301,    -0.00000000600,
       0.00000078349,    -0.00000013611,
      -0.00000000757,     0.00000001992,
      -0.00000000912,    -0.00000003920,
       0.00000000080,     0.00000000056,
      -0.00000016148,     0.00000000630,
       0.00000000050,    -0.00000000265,
      -0.00000000629,    -0.00000016061,
       0.00000000260,     0.00000000050,
       0.00000002195,    -0.00000000917,
       0.00000000909,     0.00000002185,
      -0.00000045655,     0.00000008148,
       0.00000000627,     0.00000002686,
       0.00000004751,    -0.00000000222,
       0.00000000232,     0.00000004693,
      -0.00000001490,     0.00000000621,
      -0.00000000611,    -0.00000001474,
      -0.00000003006,     0.00000000582,
       0.00000001012,    -0.00000000170,
       0.00000000168,     0.00000001005,
      -0.00000000172,    -0.00000000001,
       0.00000000003,    -0.00000000164,
      -0.00000000065,    -0.00000000132,
       0.00000000248,    -0.00000000111,
      -0.00000000117,    -0.00000000259,
       0.00000000041,     0.00000000084,
      -0.00000000252,     0.00000000109,
       0.00000000106,     0.00000000246,
       0.00000002186,    -0.00000000964,
      -0.00000000158,    -0.00000000404,
       0.00000000847,    -0.00000000325,
      -0.00000000321,    -0.00000000822,
      -0.00000004351,     0.00000001888,
       0.00000000106,     0.00000000274,
      -0.00000000860,     0.00000000326,
       0.00000000323,     0.00000000849,
       0.00000012096,    -0.00000004719,
      -0.00000000584,    -0.00000001722,
      -0.00000000100,     0.00000000053,
       0.00000004125,    -0.00000001331,
      -0.00000000087,    -0.00000000193,
      -0.00000001308,    -0.00000004064,
      -0.00000000188,     0.00000000085,
      -0.00000000196,    -0.00000000031,
      -0.00000000029,     0.00000000178,
      -0.00000022339,     0.00000008555,
       0.00000000407,     0.00000001205,
       0.00000000069,    -0.00000000037,
      -0.00000004143,     0.00000001321,
       0.00000000084,     0.00000000190,
       0.00000001306,     0.00000004097,
       0.00000000185,    -0.00000000082,
       0.00000000128,     0.00000000022,
       0.00000000021,    -0.00000000120,
       0.00000073898,    -0.00000024268,
      -0.00000001476,    -0.00000003303,
      -0.00000002506,    -0.00000008838,
      -0.00000000398,     0.00000000172,
       0.00000019925,    -0.00000004795,
      -0.00000000198,    -0.00000000626,
      -0.00000004683,    -0.00000019583,
      -0.00000000597,     0.00000000185,
      -0.00000000633,    -0.00000000172,
      -0.00000000159,     0.00000000542,
      -0.00000123510,     0.00000039276,
       0.00000002443,     0.00000005540,
       0.00000001746,     0.00000006179,
       0.00000000271,    -0.00000000116,
      -0.00000020183,     0.00000004819,
       0.00000000197,     0.00000000628,
       0.00000004747,     0.00000019937,
       0.00000000609,    -0.00000000190,
       0.00000000401,     0.00000000122,
       0.00000000118,    -0.00000000368,
       0.00000427006,    -0.00000105324,
      -0.00000004109,    -0.00000013145,
      -0.00000009450,    -0.00000044237,
      -0.00000001365,     0.00000000419,
       0.00000079538,    -0.00000010880,
      -0.00000000216,    -0.00000001288,
      -0.00000010327,    -0.00000077445,
      -0.00000001186,     0.00000000178,
      -0.00000001262,    -0.00000000570,
      -0.00000000473,     0.00000000844,
      -0.00000632794,     0.00000148776,
       0.00000006068,     0.00000019493,
       0.00000006434,     0.00000030407,
       0.00000000930,    -0.00000000285,
      -0.00000081189,     0.00000011035,
       0.00000000215,     0.00000001303,
       0.00000010707,     0.00000079809,
       0.00000001237,    -0.00000000190,
       0.00000000722,     0.00000000405,
       0.00000000370,    -0.00000000578,
       0.00002024032,    -0.00000283463,
      -0.00000004876,    -0.00000031890,
      -0.00000022490,    -0.00000183373,
      -0.00000002908,     0.00000000469,
       0.00000190074,    -0.00000001249,
      -0.00000000017,    -0.00000000190,
       0.00000001018,    -0.00000177693,
       0.00000000200,    -0.00000000089,
      -0.00000000407,    -0.00000000427,
       0.00000000066,    -0.00000001112,
      -0.00002576469,     0.00000338276,
       0.00000006476,     0.00000040617,
       0.00000014741,     0.00000122365,
       0.00000001931,    -0.00000000314,
      -0.00000194902,     0.00000000808,
       0.00000000000,     0.00000000110,
      -0.00000000406,     0.00000187776,
      -0.00000000116,     0.00000000063,
      -0.00000000231,     0.00000000340,
       0.00000000164,     0.00000000727,
       0.00005725808,     0.00000009179,
      -0.00000000375,    -0.00000456593,
       0.00000007782,    -0.00000003203,
      -0.00000000023,    -0.00000000828,
       0.00000002463,     0.00000051190,
       0.00000000114,    -0.00000000129,
      -0.00000000779,    -0.00000001115,
       0.00000000437,    -0.00000002376,
      -0.00006083347,    -0.00000005654,
       0.00000000228,     0.00000292925,
       0.00000005998,     0.00000001742,
      -0.00000000011,     0.00000000567,
      -0.00000000906,    -0.00000034848,
       0.00000000107,     0.00000000089,
      -0.00000000696,     0.00000000753,
       0.00000000211,     0.00000001602,
      -0.00000792502,     0.00000016769,
       0.00000002273,    -0.00000013017,
      -0.00000000025,     0.00000047534,
      -0.00000000790,    -0.00000000122,
       0.00000085367,    -0.00000000148,
      -0.00000000332,     0.00000082020,
       0.00000000646,    -0.00000000229,
       0.00000000222,     0.00000000642,
       0.00000664712,    -0.00000012993,
      -0.00000001805,     0.00000010994,
       0.00000000029,    -0.00000031346,
       0.00000000518,     0.00000000079,
      -0.00000035095,     0.00000000141,
       0.00000000196,    -0.00000032242,
      -0.00000000299,     0.00000000154,
      -0.00000000206,    -0.00000000433,
       0.00000067895,    -0.00000010795,
      -0.00000000604,     0.00000001858,
      -0.00000000654,    -0.00000003440,
       0.00000000080,     0.00000000041,
      -0.00000012268,     0.00000000504,
       0.00000000039,    -0.00000000207,
      -0.00000000488,    -0.00000012139,
       0.00000000200,     0.00000000038,
       0.00000001316,    -0.00000000550,
       0.00000000546,     0.00000001310,
      -0.00000044643,     0.00000007112,
       0.00000000449,     0.00000002355,
       0.00000003969,    -0.00000000173,
       0.00000000170,     0.00000003877,
      -0.00000000896,     0.00000000373,
      -0.00000000365,    -0.00000000884,
      -0.00000003304,     0.00000000744,
       0.00000000981,    -0.00000000155,
       0.00000000152,     0.00000000974,
      -0.00000000065,     0.00000000044,
      -0.00000000044,    -0.00000000066,
      -0.00000000219,     0.00000000018,
      -0.00000000016,    -0.00000000210,
       0.00000000095,    -0.00000000047,
      -0.00000000047,    -0.00000000096,
      -0.00000000097,     0.00000000045,
       0.00000000044,     0.00000000094,
      -0.00000000083,    -0.00000000194,
       0.00000000473,    -0.00000000195,
      -0.00000000192,    -0.00000000458,
      -0.00000002497,     0.00000001174,
       0.00000000055,     0.00000000130,
      -0.00000000480,     0.00000000197,
       0.00000000194,     0.00000000474,
       0.00000008043,    -0.00000003387,
      -0.00000000367,    -0.00000001004,
       0.00000002547,    -0.00000000871,
      -0.00000000054,    -0.00000000120,
      -0.00000000853,    -0.00000002504,
      -0.00000000115,     0.00000000052,
      -0.00000000112,    -0.00000000017,
      -0.00000000016,     0.00000000100,
      -0.00000014367,     0.00000005907,
       0.00000000257,     0.00000000704,
      -0.00000002567,     0.00000000870,
       0.00000000052,     0.00000000116,
       0.00000000858,     0.00000002537,
       0.00000000114,    -0.00000000051,
       0.00000050816,    -0.00000017741,
      -0.00000001035,    -0.00000002287,
      -0.00000001676,    -0.00000005528,
      -0.00000000249,     0.00000000109,
       0.00000012532,    -0.00000003152,
      -0.00000000125,    -0.00000000395,
      -0.00000003064,    -0.00000012285,
      -0.00000000372,     0.00000000114,
      -0.00000000365,    -0.00000000099,
      -0.00000000091,     0.00000000306,
      -0.00000082346,     0.00000027753,
       0.00000001654,     0.00000003712,
       0.00000001167,     0.00000003865,
       0.00000000171,    -0.00000000074,
      -0.00000012732,     0.00000003182,
       0.00000000125,     0.00000000398,
       0.00000003127,     0.00000012559,
       0.00000000384,    -0.00000000119,
       0.00000000229,     0.00000000071,
       0.00000000069,    -0.00000000208,
       0.00000290207,    -0.00000074857,
      -0.00000002794,    -0.00000008965,
      -0.00000006285,    -0.00000027873,
      -0.00000000863,     0.00000000264,
       0.00000048630,    -0.00000006860,
      -0.00000000131,    -0.00000000793,
      -0.00000006423,    -0.00000047096,
      -0.00000000717,     0.00000000102,
      -0.00000000706,    -0.00000000323,
      -0.00000000263,     0.00000000434,
      -0.00000417747,     0.00000102787,
       0.00000004007,     0.00000012905,
       0.00000004276,     0.00000019144,
       0.00000000588,    -0.00000000179,
      -0.00000049753,     0.00000006970,
       0.00000000128,     0.00000000801,
       0.00000006714,     0.00000048773,
       0.00000000754,    -0.00000000110,
       0.00000000391,     0.00000000233,
       0.00000000211,    -0.00000000299,
       0.00001305921,    -0.00000187552,
      -0.00000003031,    -0.00000020587,
      -0.00000014186,    -0.00000111516,
      -0.00000001768,     0.00000000274,
       0.00000109198,    -0.00000000765,
      -0.00000000011,    -0.00000000138,
       0.00000000993,    -0.00000100364,
       0.00000000142,    -0.00000000065,
      -0.00000000242,    -0.00000000261,
       0.00000000046,    -0.00000000707,
      -0.00001631888,     0.00000220622,
       0.00000003951,     0.00000025741,
       0.00000009313,     0.00000074449,
       0.00000001175,    -0.00000000185,
      -0.00000111868,     0.00000000401,
      -0.00000000523,     0.00000106909,
      -0.00000000080,     0.00000000046,
      -0.00000000158,     0.00000000210,
       0.00000000101,     0.00000000461,
       0.00003374431,     0.00000011920,
       0.00000000149,    -0.00000257800,
       0.00000004244,    -0.00000002269,
      -0.00000000009,    -0.00000000575,
       0.00000001905,     0.00000035506,
       0.00000000059,    -0.00000000088,
      -0.00000000447,    -0.00000000659,
       0.00000000267,    -0.00000001370,
      -0.00003583895,    -0.00000009938,
      -0.00000000116,     0.00000165958,
       0.00000005424,     0.00000001168,
      -0.00000000013,     0.00000000400,
      -0.00000000757,    -0.00000024633,
       0.00000000095,     0.00000000062,
      -0.00000000418,     0.00000000444,
       0.00000000118,     0.00000000924,
      -0.00000574681,     0.00000011457,
       0.00000001641,    -0.00000009433,
       0.00000000061,     0.00000034576,
      -0.00000000573,    -0.00000000089,
       0.00000052890,    -0.00000000026,
      -0.00000000315,     0.00000050081,
       0.00000000428,    -0.00000000137,
       0.00000000144,     0.00000000451,
       0.00000500612,    -0.00000009377,
      -0.00000001367,     0.00000008264,
      -0.00000000032,    -0.00000022847,
       0.00000000376,     0.00000000058,
      -0.00000022144,     0.00000000077,
       0.00000000179,    -0.00000019661,
      -0.00000000186,     0.00000000092,
      -0.00000000140,    -0.00000000304,
       0.00000056881,    -0.00000008129,
      -0.00000000470,     0.00000001617,
      -0.00000000463,    -0.00000002904,
       0.00000000074,     0.00000000031,
      -0.00000009193,     0.00000000350,
       0.00000000029,    -0.00000000155,
      -0.00000000327,    -0.00000009048,
       0.00000000148,     0.00000000028,
       0.00000000789,    -0.00000000331,
       0.00000000328,     0.00000000786,
      -0.00000040405,     0.00000005775,
       0.00000000317,     0.00000001984,
       0.00000003114,    -0.00000000126,
       0.00000000114,     0.00000003000,
      -0.00000000541,     0.00000000225,
      -0.00000000218,    -0.00000000530,
      -0.00000003466,     0.00000000781,
       0.00000000023,     0.00000000125,
       0.00000000875,    -0.00000000129,
       0.00000000126,     0.00000000869,
      -0.00000000018,    -0.00000000091,
      -0.00000000224,     0.00000000024,
      -0.00000000022,    -0.00000000217,
      -0.00000000053,    -0.00000000115,
       0.00000000298,    -0.00000000131,
      -0.00000000127,    -0.00000000287,
       0.00000000034,     0.00000000076,
      -0.00000000302,     0.00000000132,
       0.00000000130,     0.00000000297,
       0.00000005606,    -0.00000002519,
      -0.00000000249,    -0.00000000637,
       0.00000001654,    -0.00000000594,
      -0.00000000036,    -0.00000000079,
      -0.00000000580,    -0.00000001622,
      -0.00000000076,     0.00000000035,
      -0.00000009726,     0.00000004250,
       0.00000000174,     0.00000000447,
      -0.00000001672,     0.00000000596,
       0.00000000034,     0.00000000076,
       0.00000000587,     0.00000001651,
       0.00000000074,    -0.00000000034,
       0.00000035658,    -0.00000013096,
      -0.00000000735,    -0.00000001611,
      -0.00000001156,    -0.00000003601,
      -0.00000000164,     0.00000000072,
       0.00000008076,    -0.00000002107,
      -0.00000000082,    -0.00000000258,
      -0.00000002037,    -0.00000007894,
      -0.00000000240,     0.00000000073,
      -0.00000000217,    -0.00000000059,
      -0.00000000054,     0.00000000177,
      -0.00000056120,     0.00000019857,
       0.00000001141,     0.00000002536,
       0.00000000804,     0.00000002517,
       0.00000000111,    -0.00000000049,
      -0.00000008225,     0.00000002134,
       0.00000000081,     0.00000000258,
       0.00000002091,     0.00000008101,
       0.00000000248,    -0.00000000076,
       0.00000000135,     0.00000000043,
       0.00000000041,    -0.00000000121,
       0.00000198095,    -0.00000052973,
      -0.00000001906,    -0.00000006141,
      -0.00000004228,    -0.00000017929,
      -0.00000000556,     0.00000000170,
       0.00000030161,    -0.00000004362,
      -0.00000000080,    -0.00000000495,
      -0.00000004020,    -0.00000029037,
      -0.00000000438,     0.00000000057,
      -0.00000000405,    -0.00000000188,
      -0.00000000150,     0.00000000225,
      -0.00000278312,     0.00000071037,
       0.00000002666,     0.00000008621,
       0.00000002874,     0.00000012304,
       0.00000000379,    -0.00000000115,
      -0.00000030907,     0.00000004433,
       0.00000000078,     0.00000000499,
       0.00000004237,     0.00000030209,
       0.00000000465,    -0.00000000065,
       0.00000000217,     0.00000000137,
       0.00000000124,    -0.00000000157,
       0.00000843406,    -0.00000123362,
      -0.00000001886,    -0.00000013306,
      -0.00000008990,    -0.00000068695,
      -0.00000001090,     0.00000000162,
       0.00000063570,    -0.00000000467,
      -0.00000000007,    -0.00000000099,
       0.00000000862,    -0.00000057316,
       0.00000000099,    -0.00000000047,
      -0.00000000144,    -0.00000000161,
       0.00000000032,    -0.00000000452,
      -0.00001038143,     0.00000143484,
       0.00000002420,     0.00000016388,
       0.00000005909,     0.00000045879,
       0.00000000724,    -0.00000000110,
      -0.00000064992,     0.00000000178,
      -0.00000000510,     0.00000061556,
      -0.00000000107,     0.00000000132,
       0.00000000063,     0.00000000294,
       0.00001998180,     0.00000011829,
       0.00000000356,    -0.00000147459,
       0.00000002307,    -0.00000001576,
      -0.00000000003,    -0.00000000392,
       0.00000001395,     0.00000024242,
      -0.00000000257,    -0.00000000393,
       0.00000000165,    -0.00000000795,
      -0.00002121482,    -0.00000010755,
      -0.00000000250,     0.00000095185,
       0.00000004355,     0.00000000780,
      -0.00000000011,     0.00000000277,
      -0.00000000576,    -0.00000017054,
       0.00000000075,     0.00000000043,
      -0.00000000253,     0.00000000265,
       0.00000000065,     0.00000000536,
      -0.00000405937,     0.00000007542,
       0.00000001153,    -0.00000006658,
       0.00000000073,     0.00000024377,
      -0.00000000404,    -0.00000000063,
       0.00000032645,     0.00000000037,
      -0.00000000277,     0.00000030373,
       0.00000000283,    -0.00000000083,
       0.00000000093,     0.00000000311,
       0.00000363301,    -0.00000006406,
      -0.00000000993,     0.00000005989,
      -0.00000000043,    -0.00000016133,
       0.00000000265,     0.00000000041,
      -0.00000013914,     0.00000000037,
       0.00000000153,    -0.00000011851,
      -0.00000000116,     0.00000000055,
      -0.00000000094,    -0.00000000209,
       0.00000046022,    -0.00000005942,
      -0.00000000325,    -0.00000002364,
      -0.00000006675,     0.00000000236,
       0.00000000021,    -0.00000000112,
      -0.00000000212,    -0.00000006530,
       0.00000000107,     0.00000000020,
       0.00000000473,    -0.00000000199,
       0.00000000197,     0.00000000471,
      -0.00000034480,     0.00000004460,
       0.00000000221,     0.00000001612,
       0.00000002344,    -0.00000000090,
       0.00000000074,     0.00000002224,
      -0.00000000327,     0.00000000135,
      -0.00000000130,    -0.00000000318,
      -0.00000003326,     0.00000000726,
       0.00000000026,     0.00000000131,
       0.00000000738,    -0.00000000101,
       0.00000000098,     0.00000000730,
      -0.00000000019,    -0.00000000094,
      -0.00000000205,     0.00000000023,
      -0.00000000021,    -0.00000000199,
       0.00000000198,    -0.00000000092,
      -0.00000000089,    -0.00000000191,
      -0.00000000200,     0.00000000092,
       0.00000000091,     0.00000000197,
       0.00000004001,    -0.00000001900,
      -0.00000000176,    -0.00000000424,
       0.00000001104,    -0.00000000413,
      -0.00000000402,    -0.00000001081,
      -0.00000006758,     0.00000003108,
       0.00000000123,     0.00000000297,
      -0.00000001118,     0.00000000416,
       0.00000000409,     0.00000001103,
       0.00000025228,    -0.00000009663,
      -0.00000000810,    -0.00000002400,
      -0.00000000110,     0.00000000048,
       0.00000005277,    -0.00000001419,
      -0.00000000053,    -0.00000000169,
      -0.00000001364,    -0.00000005142,
      -0.00000000156,     0.00000000048,
      -0.00000000131,    -0.00000000036,
      -0.00000000033,     0.00000000104,
      -0.00000038684,     0.00000014262,
       0.00000000562,     0.00000001677,
       0.00000000075,    -0.00000000033,
      -0.00000005384,     0.00000001441,
       0.00000000053,     0.00000000170,
       0.00000001408,     0.00000005294,
       0.00000000162,    -0.00000000050,
       0.00000000081,     0.00000000026,
       0.00000135427,    -0.00000037288,
      -0.00000001299,    -0.00000004204,
      -0.00000002858,    -0.00000011674,
      -0.00000000363,     0.00000000110,
       0.00000018876,    -0.00000002786,
      -0.00000000049,    -0.00000000312,
      -0.00000002521,    -0.00000018054,
      -0.00000000271,     0.00000000032,
      -0.00000000236,    -0.00000000111,
      -0.00000000087,     0.00000000117,
      -0.00000186289,     0.00000049012,
       0.00000001778,     0.00000005783,
       0.00000001941,     0.00000008004,
       0.00000000247,    -0.00000000075,
      -0.00000019360,     0.00000002829,
       0.00000000048,     0.00000000314,
       0.00000002680,     0.00000018863,
       0.00000000289,    -0.00000000038,
       0.00000000121,     0.00000000082,
       0.00000000074,    -0.00000000082,
       0.00000544344,    -0.00000080657,
      -0.00000001169,    -0.00000008600,
      -0.00000005709,    -0.00000042671,
      -0.00000000677,     0.00000000097,
       0.00000037361,    -0.00000000284,
       0.00000000701,    -0.00000032966,
       0.00000000069,    -0.00000000034,
      -0.00000000087,    -0.00000000100,
       0.00000000022,    -0.00000000289,
      -0.00000662007,     0.00000093002,
       0.00000001483,     0.00000010458,
       0.00000003756,     0.00000028506,
       0.00000000450,    -0.00000000066,
      -0.00000038083,     0.00000000061,
      -0.00000000441,     0.00000035712,
      -0.00000000072,     0.00000000083,
       0.00000000039,     0.00000000188,
       0.00001187216,     0.00000010449,
       0.00000000401,    -0.00000085139,
       0.00000001246,    -0.00000001079,
       0.00000000000,    -0.00000000265,
       0.00000000984,     0.00000016357,
      -0.00000000148,    -0.00000000238,
       0.00000000102,    -0.00000000462,
      -0.00001259894,    -0.00000009890,
      -0.00000000276,     0.00000055079,
       0.00000003283,     0.00000000520,
      -0.00000000010,     0.00000000189,
      -0.00000000415,    -0.00000011632,
      -0.00000000154,     0.00000000159,
       0.00000000036,     0.00000000313,
      -0.00000281326,     0.00000004943,
       0.00000000792,    -0.00000004614,
       0.00000000059,     0.00000016830,
      -0.00000000279,    -0.00000000044,
       0.00000020108,     0.00000000062,
      -0.00000000230,     0.00000018324,
       0.00000000186,    -0.00000000050,
       0.00000000060,     0.00000000212,
       0.00000257068,    -0.00000004195,
      -0.00000000698,     0.00000004235,
      -0.00000000035,    -0.00000011151,
       0.00000000183,     0.00000000028,
      -0.00000008725,     0.00000000013,
       0.00000000125,    -0.00000007073,
      -0.00000000073,     0.00000000033,
      -0.00000000062,    -0.00000000143,
       0.00000036091,    -0.00000004255,
      -0.00000000225,    -0.00000001865,
      -0.00000004739,     0.00000000156,
      -0.00000000134,    -0.00000004604,
       0.00000000284,    -0.00000000120,
       0.00000000118,     0.00000000282,
      -0.00000028178,     0.00000003330,
       0.00000000153,     0.00000001270,
       0.00000001714,    -0.00000000062,
       0.00000000046,     0.00000001599,
      -0.00000000198,     0.00000000082,
      -0.00000000078,    -0.00000000191,
      -0.00000003008,     0.00000000632,
       0.00000000025,     0.00000000126,
       0.00000000595,    -0.00000000077,
       0.00000000074,     0.00000000587,
      -0.00000000018,    -0.00000000089,
      -0.00000000177,     0.00000000020,
      -0.00000000018,    -0.00000000170,
       0.00000000136,    -0.00000000066,
      -0.00000000064,    -0.00000000131,
      -0.00000000136,     0.00000000066,
       0.00000000065,     0.00000000134,
       0.00000002891,    -0.00000001439,
      -0.00000000126,    -0.00000000290,
       0.00000000748,    -0.00000000290,
      -0.00000000281,    -0.00000000730,
      -0.00000004765,     0.00000002291,
       0.00000000088,     0.00000000202,
      -0.00000000759,     0.00000000293,
       0.00000000287,     0.00000000748,
       0.00000017895,    -0.00000007102,
      -0.00000000571,    -0.00000001622,
      -0.00000000074,     0.00000000033,
       0.00000003474,    -0.00000000959,
      -0.00000000035,    -0.00000000112,
      -0.00000000916,    -0.00000003375,
      -0.00000000103,     0.00000000031,
      -0.00000000081,    -0.00000000023,
      -0.00000026821,     0.00000010243,
       0.00000000396,     0.00000001132,
      -0.00000003550,     0.00000000976,
       0.00000000035,     0.00000000112,
       0.00000000950,     0.00000003485,
       0.00000000107,    -0.00000000033,
       0.00000092480,    -0.00000026081,
      -0.00000000885,    -0.00000002877,
      -0.00000001933,    -0.00000007655,
      -0.00000000239,     0.00000000072,
       0.00000011880,    -0.00000001783,
      -0.00000000031,    -0.00000000198,
      -0.00000001581,    -0.00000011283,
      -0.00000000167,     0.00000000018,
      -0.00000000139,    -0.00000000066,
      -0.00000000051,     0.00000000060,
      -0.00000124914,     0.00000033705,
       0.00000001187,     0.00000003883,
       0.00000001312,     0.00000005245,
       0.00000000162,    -0.00000000049,
      -0.00000012190,     0.00000001807,
       0.00000000030,     0.00000000199,
       0.00000001695,     0.00000011836,
       0.00000000181,    -0.00000000023,
       0.00000000069,     0.00000000051,
       0.00000351210,    -0.00000052485,
      -0.00000000723,    -0.00000005552,
      -0.00000003625,    -0.00000026652,
      -0.00000000424,     0.00000000059,
       0.00000022112,    -0.00000000172,
       0.00000000546,    -0.00000019044,
      -0.00000000053,    -0.00000000063,
       0.00000000015,    -0.00000000185,
      -0.00000422523,     0.00000060049,
       0.00000000907,     0.00000006681,
       0.00000002387,     0.00000017808,
       0.00000000281,    -0.00000000040,
      -0.00000022453,     0.00000000004,
      -0.00000000358,     0.00000020824,
      -0.00000000048,     0.00000000052,
       0.00000000024,     0.00000000120,
       0.00000706705,     0.00000008616,
       0.00000000371,    -0.00000049494,
       0.00000000667,    -0.00000000731,
       0.00000000001,    -0.00000000177,
       0.00000000677,     0.00000010937,
      -0.00000000085,    -0.00000000144,
       0.00000000065,    -0.00000000270,
      -0.00000749825,    -0.00000008354,
      -0.00000000253,     0.00000032076,
       0.00000002381,     0.00000000346,
      -0.00000000007,     0.00000000127,
      -0.00000000289,    -0.00000007845,
      -0.00000000094,     0.00000000096,
       0.00000000019,     0.00000000182,
      -0.00000192250,     0.00000002961,
       0.00000000537,    -0.00000003153,
       0.00000000039,     0.00000011446,
      -0.00000000189,    -0.00000000029,
       0.00000012370,     0.00000000068,
      -0.00000000184,     0.00000011005,
       0.00000000122,    -0.00000000030,
       0.00000000039,     0.00000000143,
       0.00000178533,    -0.00000002638,
      -0.00000000482,     0.00000002937,
      -0.00000000023,    -0.00000007591,
       0.00000000124,     0.00000000019,
      -0.00000005467,     0.00000000000,
       0.00000000098,    -0.00000004182,
      -0.00000000041,    -0.00000000096,
       0.00000027511,    -0.00000003538,
      -0.00000000155,    -0.00000001434,
      -0.00000003306,     0.00000000101,
      -0.00000000082,    -0.00000003188,
       0.00000000170,    -0.00000000072,
       0.00000000071,     0.00000000168,
      -0.00000022270,     0.00000002424,
       0.00000000105,     0.00000000976,
       0.00000001226,    -0.00000000042,
       0.00000000028,     0.00000001123,
      -0.00000000120,     0.00000000049,
      -0.00000000046,    -0.00000000114,
      -0.00000002600,     0.00000000522,
       0.00000000021,     0.00000000113,
       0.00000000465,    -0.00000000056,
       0.00000000053,     0.00000000458,
      -0.00000000145,     0.00000000016,
      -0.00000000014,    -0.00000000139,
       0.00000000096,    -0.00000000048,
      -0.00000000046,    -0.00000000092,
      -0.00000000095,     0.00000000047,
       0.00000000047,     0.00000000093,
       0.00000002100,    -0.00000001088,
      -0.00000000092,    -0.00000000201,
       0.00000000510,    -0.00000000204,
      -0.00000000197,    -0.00000000497,
      -0.00000003385,     0.00000001693,
       0.00000000063,     0.00000000139,
      -0.00000000519,     0.00000000206,
       0.00000000203,     0.00000000511,
       0.00000012689,    -0.00000005192,
      -0.00000000403,    -0.00000001105,
       0.00000002298,    -0.00000000648,
      -0.00000000615,    -0.00000002224,
      -0.00000018638,     0.00000007337,
       0.00000000279,     0.00000000769,
      -0.00000002351,     0.00000000660,
       0.00000000641,     0.00000002304,
       0.00000063023,    -0.00000018127,
      -0.00000000600,    -0.00000001964,
      -0.00000001307,    -0.00000005040,
      -0.00000000158,     0.00000000048,
       0.00000007504,    -0.00000001141,
      -0.00000000020,    -0.00000000127,
      -0.00000000990,    -0.00000007072,
      -0.00000000103,     0.00000000010,
      -0.00000000082,    -0.00000000040,
      -0.00000083781,     0.00000023085,
       0.00000000793,     0.00000002610,
       0.00000000887,     0.00000003450,
       0.00000000107,    -0.00000000032,
      -0.00000007701,     0.00000001154,
       0.00000000018,     0.00000000127,
       0.00000001070,     0.00000007450,
       0.00000000113,    -0.00000000013,
       0.00000226310,    -0.00000033980,
      -0.00000000446,    -0.00000003580,
      -0.00000002300,    -0.00000016706,
      -0.00000000266,     0.00000000035,
       0.00000013157,    -0.00000000103,
       0.00000000413,    -0.00000011028,
       0.00000000010,    -0.00000000118,
      -0.00000269773,     0.00000038628,
       0.00000000552,     0.00000004270,
       0.00000001515,     0.00000011164,
       0.00000000177,    -0.00000000024,
      -0.00000013298,    -0.00000000021,
      -0.00000000278,     0.00000012183,
       0.00000421403,     0.00000006797,
       0.00000000312,    -0.00000028917,
       0.00000000353,    -0.00000000491,
       0.00000000002,    -0.00000000119,
       0.00000000456,     0.00000007259,
      -0.00000000049,    -0.00000000087,
       0.00000000041,    -0.00000000158,
      -0.00000446928,    -0.00000006693,
      -0.00000000212,     0.00000018766,
       0.00000001682,     0.00000000230,
      -0.00000000195,    -0.00000005243,
      -0.00000000058,     0.00000000058,
       0.00000000010,     0.00000000107,
      -0.00000129970,     0.00000001752,
       0.00000000359,    -0.00000002129,
       0.00000000022,     0.00000007697,
      -0.00000000127,    -0.00000000020,
       0.00000007605,     0.00000000062,
      -0.00000000142,     0.00000006581,
       0.00000000025,     0.00000000096,
       0.00000122232,    -0.00000001590,
      -0.00000000013,    -0.00000005108,
      -0.00000003426,    -0.00000000005,
       0.00000000075,    -0.00000002450,
       0.00000020698,    -0.00000002083,
      -0.00000000106,    -0.00000001079,
      -0.00000002276,     0.00000000065,
      -0.00000000049,    -0.00000002177,
       0.00000000102,    -0.00000000044,
       0.00000000043,     0.00000000100,
      -0.00000017140,     0.00000001729,
       0.00000000072,     0.00000000733,
       0.00000000863,    -0.00000000028,
       0.00000000015,     0.00000000774,
      -0.00000000072,     0.00000000030,
      -0.00000002417,     0.00000001293,
       0.00000000355,    -0.00000000041,
       0.00000000038,     0.00000000348,
      -0.00000000115,     0.00000000012,
      -0.00000000011,    -0.00000000109,
      -0.00000000066,     0.00000000034,
       0.00000001525,    -0.00000000818,
      -0.00000000067,    -0.00000000141,
       0.00000000350,    -0.00000000144,
      -0.00000000138,    -0.00000000340,
      -0.00000002409,     0.00000001248,
       0.00000000046,     0.00000000097,
      -0.00000000357,     0.00000000145,
       0.00000000143,     0.00000000350,
       0.00000008983,    -0.00000003771,
      -0.00000000284,    -0.00000000756,
       0.00000001523,    -0.00000000438,
      -0.00000000412,    -0.00000001468,
      -0.00000012957,     0.00000005236,
       0.00000000197,     0.00000000525,
      -0.00000001560,     0.00000000446,
       0.00000000431,     0.00000001525,
       0.00000042834,    -0.00000012522,
      -0.00000000882,    -0.00000003326,
      -0.00000000104,     0.00000000031,
       0.00000004752,    -0.00000000731,
      -0.00000000618,    -0.00000004440,
      -0.00000056146,     0.00000015748,
       0.00000000598,     0.00000002275,
      -0.00000004875,     0.00000000736,
       0.00000000675,     0.00000004697,
       0.00000145605,    -0.00000021894,
      -0.00000000273,    -0.00000002304,
      -0.00000001457,    -0.00000010496,
      -0.00000000167,     0.00000000021,
       0.00000007860,    -0.00000000062,
       0.00000000305,    -0.00000006392,
      -0.00000172183,     0.00000024751,
       0.00000000334,     0.00000002727,
       0.00000000960,     0.00000007015,
       0.00000000110,    -0.00000000014,
      -0.00000007901,    -0.00000000029,
      -0.00000000210,     0.00000007141,
       0.00000251453,     0.00000005191,
       0.00000000248,    -0.00000016954,
       0.00000000183,    -0.00000000327,
       0.00000000303,     0.00000004787,
       0.00000000026,    -0.00000000092,
      -0.00000266569,    -0.00000005169,
      -0.00000000168,     0.00000011015,
       0.00000001165,     0.00000000152,
      -0.00000000129,    -0.00000003478,
      -0.00000087072,     0.00000000989,
       0.00000000010,     0.00000005129,
       0.00000004673,     0.00000000053,
      -0.00000000107,     0.00000003918,
       0.00000082741,    -0.00000000911,
      -0.00000000005,    -0.00000003406,
      -0.00000002148,    -0.00000000008,
       0.00000000056,    -0.00000001420,
       0.00000015256,    -0.00000001430,
      -0.00000000071,    -0.00000000799,
      -0.00000001551,     0.00000000040,
      -0.00000000028,    -0.00000001470,
      -0.00000012911,     0.00000001213,
       0.00000000048,     0.00000000541,
       0.00000000600,    -0.00000000018,
       0.00000000008,     0.00000000525,
      -0.00000001763,     0.00000000324,
       0.00000000266,    -0.00000000029,
       0.00000000027,     0.00000000260,
      -0.00000000049,    -0.00000000099,
       0.00000000241,    -0.00000000101,
      -0.00000000096,    -0.00000000233,
       0.00000000034,     0.00000000068,
      -0.00000000246,     0.00000000102,
       0.00000000100,     0.00000000241,
       0.00000006345,    -0.00000002722,
      -0.00000000200,    -0.00000000518,
       0.00000001011,    -0.00000000295,
      -0.00000000275,    -0.00000000969,
      -0.00000009002,     0.00000003721,
       0.00000000138,     0.00000000359,
      -0.00000001036,     0.00000000301,
       0.00000000290,     0.00000001011,
       0.00000029028,    -0.00000008601,
      -0.00000000594,    -0.00000002197,
       0.00000003014,    -0.00000000467,
      -0.00000000385,    -0.00000002791,
      -0.00000037579,     0.00000010697,
       0.00000000402,     0.00000001502,
      -0.00000003089,     0.00000000469,
       0.00000000424,     0.00000002963,
       0.00000093530,    -0.00000014044,
      -0.00000000921,    -0.00000006604,
      -0.00000000105,     0.00000000012,
       0.00000004711,    -0.00000000036,
       0.00000000221,    -0.00000003703,
      -0.00000109826,     0.00000015798,
       0.00000000607,     0.00000004415,
      -0.00000004705,    -0.00000000029,
      -0.00000000155,     0.00000004189,
       0.00000150063,     0.00000003870,
       0.00000000190,    -0.00000009965,
       0.00000000092,    -0.00000000217,
       0.00000000198,     0.00000003141,
      -0.00000159008,    -0.00000003884,
      -0.00000000128,     0.00000006480,
       0.00000000796,     0.00000000101,
      -0.00000000084,    -0.00000002293,
      -0.00000057894,     0.00000000522,
       0.00000000003,     0.00000003393,
       0.00000002870,     0.00000000043,
      -0.00000000080,     0.00000002322,
       0.00000055493,    -0.00000000488,
      -0.00000000001,    -0.00000002254,
      -0.00000001347,    -0.00000000008,
       0.00000000041,    -0.00000000814,
       0.00000011086,    -0.00000000971,
      -0.00000000048,    -0.00000000582,
      -0.00000001047,     0.00000000025,
      -0.00000000015,    -0.00000000983,
      -0.00000009555,     0.00000000838,
       0.00000000032,     0.00000000394,
       0.00000000413,    -0.00000000012,
       0.00000000004,     0.00000000352,
       0.00000000197,    -0.00000000020,
       0.00000000018,     0.00000000191,
       0.00000000166,    -0.00000000071,
      -0.00000000067,    -0.00000000160,
      -0.00000000169,     0.00000000072,
       0.00000000070,     0.00000000166,
       0.00000004467,    -0.00000001954,
      -0.00000000141,    -0.00000000355,
       0.00000000671,    -0.00000000199,
      -0.00000000184,    -0.00000000640,
      -0.00000006245,     0.00000002633,
       0.00000000097,     0.00000000246,
      -0.00000000688,     0.00000000203,
       0.00000000194,     0.00000000669,
       0.00000019613,    -0.00000005875,
      -0.00000000399,    -0.00000001452,
       0.00000001912,    -0.00000000299,
      -0.00000000238,    -0.00000001753,
      -0.00000025112,     0.00000007237,
       0.00000000270,     0.00000000991,
      -0.00000001958,     0.00000000298,
       0.00000000266,     0.00000001870,
       0.00000059975,    -0.00000008967,
      -0.00000000581,    -0.00000004159,
       0.00000002830,    -0.00000000021,
       0.00000000158,    -0.00000002141,
      -0.00000069990,     0.00000010044,
       0.00000000383,     0.00000002780,
      -0.00000002807,    -0.00000000024,
      -0.00000000112,     0.00000002457,
       0.00000089522,     0.00000002831,
       0.00000000142,    -0.00000005866,
       0.00000000044,    -0.00000000143,
       0.00000000128,     0.00000002051,
      -0.00000094813,    -0.00000002859,
      -0.00000000096,     0.00000003817,
       0.00000000537,     0.00000000067,
      -0.00000000053,    -0.00000001504,
      -0.00000038249,     0.00000000247,
      -0.00000000002,     0.00000002231,
       0.00000001763,     0.00000000033,
      -0.00000000058,     0.00000001369,
       0.00000036932,    -0.00000000233,
       0.00000000002,    -0.00000001483,
      -0.00000000846,    -0.00000000007,
       0.00000000030,    -0.00000000460,
       0.00000007958,    -0.00000000652,
      -0.00000000032,    -0.00000000419,
      -0.00000000702,     0.00000000015,
      -0.00000000007,    -0.00000000651,
      -0.00000006967,     0.00000000573,
       0.00000000021,     0.00000000283,
       0.00000000282,    -0.00000000008,
       0.00000000001,     0.00000000234,
       0.00000000143,    -0.00000000014,
       0.00000000012,     0.00000000138,
       0.00000000115,    -0.00000000049,
      -0.00000000047,    -0.00000000109,
      -0.00000000117,     0.00000000051,
       0.00000000049,     0.00000000114,
       0.00000003134,    -0.00000001393,
      -0.00000000099,    -0.00000000243,
       0.00000000445,    -0.00000000134,
      -0.00000000122,    -0.00000000423,
      -0.00000004323,     0.00000001855,
       0.00000000068,     0.00000000168,
      -0.00000000456,     0.00000000136,
       0.00000000129,     0.00000000443,
       0.00000013212,    -0.00000003992,
      -0.00000000267,    -0.00000000959,
       0.00000001214,    -0.00000000191,
      -0.00000000146,    -0.00000001100,
      -0.00000016746,     0.00000004875,
       0.00000000180,     0.00000000654,
      -0.00000001242,     0.00000000189,
       0.00000000166,     0.00000001179,
       0.00000038393,    -0.00000005700,
      -0.00000000365,    -0.00000002620,
       0.00000001703,    -0.00000000012,
       0.00000000112,    -0.00000001235,
      -0.00000044554,     0.00000006361,
       0.00000000241,     0.00000001751,
      -0.00000001675,    -0.00000000020,
      -0.00000000080,     0.00000001440,
       0.00000053363,     0.00000002038,
       0.00000000103,    -0.00000003456,
       0.00000000082,     0.00000001334,
      -0.00000056488,    -0.00000002069,
      -0.00000000070,     0.00000002250,
       0.00000000359,     0.00000000044,
      -0.00000000033,    -0.00000000982,
      -0.00000025130,     0.00000000092,
      -0.00000000004,     0.00000001459,
       0.00000001082,     0.00000000025,
      -0.00000000042,     0.00000000802,
       0.00000024418,    -0.00000000088,
       0.00000000003,    -0.00000000970,
      -0.00000000531,    -0.00000000006,
       0.00000000021,    -0.00000000255,
       0.00000005653,    -0.00000000434,
      -0.00000000021,    -0.00000000298,
      -0.00000000468,     0.00000000009,
      -0.00000000003,    -0.00000000429,
      -0.00000005017,     0.00000000386,
       0.00000000014,     0.00000000201,
       0.00000000191,    -0.00000000005,
       0.00000000000,     0.00000000153,
      -0.00000000080,     0.00000000035,
       0.00000000034,     0.00000000078,
       0.00000002190,    -0.00000000988,
      -0.00000000069,    -0.00000000167,
       0.00000000295,    -0.00000000090,
      -0.00000000081,    -0.00000000278,
      -0.00000002986,     0.00000001300,
       0.00000000047,     0.00000000115,
      -0.00000000302,     0.00000000091,
       0.00000000086,     0.00000000293,
       0.00000008871,    -0.00000002699,
      -0.00000000178,    -0.00000000633,
       0.00000000770,    -0.00000000122,
      -0.00000000089,    -0.00000000689,
      -0.00000011145,     0.00000003271,
       0.00000000120,     0.00000000431,
      -0.00000000787,     0.00000000120,
       0.00000000103,     0.00000000744,
       0.00000024533,    -0.00000003608,
      -0.00000000229,    -0.00000001651,
       0.00000001026,    -0.00000000007,
       0.00000000078,    -0.00000000709,
      -0.00000028330,     0.00000004012,
       0.00000000151,     0.00000001103,
      -0.00000001001,    -0.00000000015,
      -0.00000000057,     0.00000000842,
       0.00000031771,     0.00000001449,
       0.00000000074,    -0.00000002037,
       0.00000000052,     0.00000000864,
      -0.00000033613,    -0.00000001476,
      -0.00000000050,     0.00000001326,
       0.00000000239,     0.00000000029,
      -0.00000000021,    -0.00000000638,
      -0.00000016432,     0.00000000010,
      -0.00000000005,     0.00000000950,
       0.00000000664,     0.00000000019,
      -0.00000000029,     0.00000000467,
       0.00000016051,    -0.00000000010,
       0.00000000004,    -0.00000000632,
      -0.00000000334,    -0.00000000005,
       0.00000000015,    -0.00000000139,
       0.00000003979,    -0.00000000286,
      -0.00000000014,    -0.00000000211,
      -0.00000000310,     0.00000000005,
      -0.00000000001,    -0.00000000280,
      -0.00000003573,     0.00000000257,
       0.00000000009,     0.00000000142,
       0.00000000129,    -0.00000000003,
      -0.00000000001,     0.00000000100,
       0.00000001526,    -0.00000000697,
      -0.00000000048,    -0.00000000115,
       0.00000000196,    -0.00000000060,
      -0.00000000053,    -0.00000000183,
       0.00000000033,     0.00000000078,
      -0.00000000200,     0.00000000060,
       0.00000000057,     0.00000000193,
       0.00000005940,    -0.00000001817,
      -0.00000000119,    -0.00000000417,
       0.00000000489,    -0.00000000077,
      -0.00000000054,    -0.00000000432,
      -0.00000007402,     0.00000002187,
       0.00000000080,     0.00000000284,
      -0.00000000499,     0.00000000076,
       0.00000000064,     0.00000000469,
       0.00000015648,    -0.00000002274,
      -0.00000000143,    -0.00000001039,
       0.00000000619,    -0.00000000004,
       0.00000000054,    -0.00000000405,
      -0.00000017989,     0.00000002521,
       0.00000000094,     0.00000000694,
      -0.00000000598,    -0.00000000012,
      -0.00000000039,     0.00000000491,
       0.00000018883,     0.00000001018,
       0.00000000052,    -0.00000001199,
       0.00000000032,     0.00000000558,
      -0.00000019968,    -0.00000001041,
      -0.00000000035,     0.00000000781,
       0.00000000157,     0.00000000020,
      -0.00000000012,    -0.00000000413,
      -0.00000010694,    -0.00000000029,
      -0.00000000005,     0.00000000616,
       0.00000000407,     0.00000000014,
      -0.00000000021,     0.00000000269,
       0.00000010498,     0.00000000028,
       0.00000000003,    -0.00000000409,
      -0.00000000210,    -0.00000000004,
       0.00000002777,    -0.00000000186,
      -0.00000000009,    -0.00000000147,
      -0.00000000205,     0.00000000003,
       0.00000000000,    -0.00000000182,
       0.00000000006,     0.00000000099,
       0.00000000130,    -0.00000000040,
      -0.00000000035,    -0.00000000120,
      -0.00000000132,     0.00000000040,
       0.00000000038,     0.00000000127,
       0.00000003967,    -0.00000001217,
      -0.00000000079,    -0.00000000275,
       0.00000000310,    -0.00000000049,
      -0.00000000033,    -0.00000000271,
      -0.00000004907,     0.00000001457,
       0.00000000053,     0.00000000187,
      -0.00000000316,     0.00000000048,
       0.00000000039,     0.00000000294,
       0.00000009962,    -0.00000001427,
      -0.00000000089,    -0.00000000654,
       0.00000000374,    -0.00000000002,
       0.00000000037,    -0.00000000230,
      -0.00000011406,     0.00000001578,
       0.00000000059,     0.00000000437,
      -0.00000000357,    -0.00000000008,
      -0.00000000027,     0.00000000286,
       0.00000011202,     0.00000000708,
       0.00000000037,    -0.00000000706,
       0.00000000020,     0.00000000359,
      -0.00000011841,    -0.00000000726,
      -0.00000000025,     0.00000000460,
       0.00000000103,     0.00000000013,
      -0.00000000007,    -0.00000000267,
      -0.00000006932,    -0.00000000044,
      -0.00000000004,     0.00000000398,
       0.00000000250,     0.00000000010,
      -0.00000000014,     0.00000000154,
       0.00000006834,     0.00000000041,
       0.00000000003,    -0.00000000264,
      -0.00000000133,    -0.00000000003,
      -0.00000000135,     0.00000000001,
      -0.00000000088,     0.00000000027,
       0.00000000025,     0.00000000084,
       0.00000002641,    -0.00000000813,
      -0.00000000052,    -0.00000000181,
       0.00000000197,    -0.00000000031,
      -0.00000000020,    -0.00000000169,
      -0.00000003246,     0.00000000967,
       0.00000000035,     0.00000000123,
      -0.00000000200,     0.00000000030,
       0.00000000024,     0.00000000185,
       0.00000006331,    -0.00000000892,
      -0.00000000055,    -0.00000000412,
       0.00000000226,    -0.00000000001,
       0.00000000025,    -0.00000000129,
      -0.00000007222,     0.00000000983,
       0.00000000037,     0.00000000274,
      -0.00000000213,    -0.00000000006,
      -0.00000000019,     0.00000000166,
       0.00000006630,     0.00000000489,
       0.00000000026,    -0.00000000415,
       0.00000000012,     0.00000000230,
      -0.00000007005,    -0.00000000502,
      -0.00000000017,     0.00000000270,
      -0.00000000004,    -0.00000000172,
      -0.00000004477,    -0.00000000046,
      -0.00000000003,     0.00000000256,
       0.00000000153,     0.00000000007,
       0.00000004430,     0.00000000043,
       0.00000000002,    -0.00000000170,
       0.00000001754,    -0.00000000540,
      -0.00000000034,    -0.00000000119,
       0.00000000125,    -0.00000000020,
       0.00000000023,     0.00000000080,
      -0.00000000126,     0.00000000019,
       0.00000000015,     0.00000000116,
       0.00000004015,    -0.00000000553,
      -0.00000000034,    -0.00000000258,
       0.00000000137,    -0.00000000001,
      -0.00000004565,     0.00000000610,
       0.00000000023,     0.00000000173,
      -0.00000000127,    -0.00000000004,
      -0.00000000013,     0.00000000096,
       0.00000003912,     0.00000000335,
       0.00000000018,    -0.00000000243,
       0.00000000008,     0.00000000148,
      -0.00000004130,    -0.00000000344,
      -0.00000000012,     0.00000000158,
      -0.00000000002,    -0.00000000110,
      -0.00000002882,    -0.00000000041,
      -0.00000000003,     0.00000000165,
       0.00000000002,    -0.00000000109,
       0.00000002541,    -0.00000000342,
      -0.00000000021,    -0.00000000163,
      -0.00000002882,     0.00000000376,
       0.00000000014,     0.00000000108,
       0.00000002300,     0.00000000228,
       0.00000000012,    -0.00000000142,
      -0.00000000008,     0.00000000092,
      -0.00000000237,    -0.00000000115,
      -0.00000000115,     0.00000000236,
       0.00000000032,     0.00000000126,
       0.00000000126,    -0.00000000031,
       0.00000000622,     0.00000002652,
       0.00000000143,    -0.00000000052,
       0.00000000114,    -0.00000000509,
      -0.00000000036,     0.00000000092,
      -0.00000000504,    -0.00000000122,
       0.00000000090,     0.00000000038,
      -0.00000000194,     0.00000000785,
       0.00000000131,     0.00000000018,
       0.00000000117,    -0.00000000151,
      -0.00000000143,    -0.00000000123,
       0.00000002511,    -0.00000010538,
      -0.00000000783,     0.00000001858,
      -0.00000000617,    -0.00000000086,
       0.00000000115,     0.00000000039,
      -0.00000001062,     0.00000001315,
       0.00000000159,    -0.00000000132,
       0.00000001283,     0.00000001083,
      -0.00000000121,    -0.00000000162,
       0.00000001856,    -0.00000002220,
      -0.00000000366,    -0.00000000267,
      -0.00000000393,     0.00000000205,
       0.00000000155,     0.00000000395,
      -0.00000022186,     0.00000026763,
       0.00000003268,    -0.00000002586,
       0.00000001686,     0.00000001225,
      -0.00000000173,    -0.00000000200,
       0.00000003554,    -0.00000001690,
      -0.00000000212,     0.00000000027,
      -0.00000001537,    -0.00000003555,
      -0.00000000007,     0.00000000206,
      -0.00000006790,     0.00000003114,
       0.00000000491,     0.00000001006,
       0.00000000543,    -0.00000000079,
       0.00000000079,    -0.00000000453,
       0.00000072136,    -0.00000033090,
      -0.00000004207,     0.00000000310,
      -0.00000002207,    -0.00000004492,
       0.00000000022,     0.00000000287,
      -0.00000004704,     0.00000000185,
      -0.00000000115,     0.00000000016,
      -0.00000000259,     0.00000004444,
       0.00000000068,     0.00000000165,
       0.00000009890,    -0.00000000119,
       0.00000000027,    -0.00000001424,
      -0.00000000098,     0.00000000258,
       0.00000000011,    -0.00000000344,
      -0.00000091907,     0.00000001051,
      -0.00000002624,     0.00000000730,
      -0.00000000103,     0.00000006159,
       0.00000000061,     0.00000000178,
      -0.00000000518,    -0.00000001455,
       0.00000000047,     0.00000000079,
      -0.00000000846,     0.00000001617,
       0.00000000068,    -0.00000000124,
       0.00000002146,     0.00000002843,
       0.00000000343,    -0.00000000298,
       0.00000000365,     0.00000000033,
       0.00000000024,     0.00000000381,
      -0.00000017135,    -0.00000022577,
       0.00000001735,     0.00000001413,
      -0.00000001520,     0.00000001334,
       0.00000000097,    -0.00000000113,
      -0.00000000734,    -0.00000000240,
      -0.00000000228,    -0.00000000855,
       0.00000000068,     0.00000000105,
      -0.00000000103,     0.00000000074,
       0.00000001695,    -0.00000003591,
      -0.00000000287,    -0.00000000092,
      -0.00000000104,    -0.00000000195,
       0.00000000185,    -0.00000000150,
       0.00000000427,     0.00000001857,
       0.00000000098,    -0.00000000034,
       0.00000000073,    -0.00000000325,
      -0.00000000321,    -0.00000000078,
       0.00000000068,    -0.00000000091,
      -0.00000000084,    -0.00000000072,
       0.00000001661,    -0.00000006927,
      -0.00000000397,    -0.00000000059,
      -0.00000000627,     0.00000000787,
       0.00000000090,    -0.00000000077,
       0.00000000763,     0.00000000642,
      -0.00000000070,    -0.00000000093,
       0.00000001160,    -0.00000001405,
      -0.00000000219,    -0.00000000161,
      -0.00000000212,     0.00000000119,
       0.00000000082,     0.00000000213,
      -0.00000013482,     0.00000016443,
       0.00000001919,    -0.00000001542,
       0.00000001009,     0.00000000736,
      -0.00000000099,    -0.00000000115,
       0.00000001927,    -0.00000000957,
      -0.00000000104,     0.00000000016,
      -0.00000000849,    -0.00000001924,
      -0.00000000007,     0.00000000100,
      -0.00000003849,     0.00000001832,
       0.00000000274,     0.00000000549,
       0.00000000270,    -0.00000000062,
       0.00000000044,    -0.00000000208,
       0.00000040258,    -0.00000019155,
      -0.00000002109,     0.00000000169,
      -0.00000001230,    -0.00000002454,
       0.00000000010,     0.00000000141,
      -0.00000002297,     0.00000000207,
      -0.00000000079,     0.00000002121,
       0.00000000029,     0.00000000110,
       0.00000004953,    -0.00000000255,
      -0.00000000009,    -0.00000000691,
      -0.00000000065,     0.00000000156,
       0.00000000005,    -0.00000000202,
      -0.00000046031,     0.00000002340,
      -0.00000001846,     0.00000000214,
       0.00000000052,     0.00000002991,
       0.00000000021,     0.00000000121,
      -0.00000000273,    -0.00000000879,
      -0.00000000514,     0.00000000931,
       0.00000000037,    -0.00000000073,
       0.00000001267,     0.00000001750,
       0.00000000208,    -0.00000000170,
       0.00000000190,     0.00000000014,
       0.00000000024,     0.00000000203,
      -0.00000010314,    -0.00000014194,
      -0.00000000928,     0.00000000762,
       0.00000000053,    -0.00000000068,
      -0.00000000372,    -0.00000000148,
      -0.00000000172,    -0.00000000475,
       0.00000000042,     0.00000000069,
      -0.00000000068,     0.00000000046,
       0.00000001372,    -0.00000002635,
      -0.00000000200,    -0.00000000076,
      -0.00000000063,    -0.00000000130,
       0.00000000125,    -0.00000000101,
       0.00000000045,    -0.00000000205,
      -0.00000000202,    -0.00000000049,
       0.00000001057,    -0.00000004462,
      -0.00000000252,    -0.00000000039,
      -0.00000000364,     0.00000000467,
       0.00000000448,     0.00000000374,
       0.00000000705,    -0.00000000869,
      -0.00000000130,    -0.00000000095,
      -0.00000000114,     0.00000000069,
       0.00000000043,     0.00000000112,
      -0.00000008004,     0.00000009929,
       0.00000001096,    -0.00000000901,
       0.00000000597,     0.00000000433,
      -0.00000000056,    -0.00000000065,
       0.00000001032,    -0.00000000541,
      -0.00000000467,    -0.00000001028,
      -0.00000002134,     0.00000001063,
       0.00000000152,     0.00000000296,
       0.00000000135,    -0.00000000044,
       0.00000000023,    -0.00000000093,
       0.00000022041,    -0.00000010974,
      -0.00000000683,    -0.00000001322,
      -0.00000001112,     0.00000000171,
      -0.00000000010,     0.00000000998,
       0.00000002435,    -0.00000000247,
      -0.00000000019,    -0.00000000331,
      -0.00000000040,     0.00000000091,
       0.00000000002,    -0.00000000116,
      -0.00000022628,     0.00000002282,
       0.00000000088,     0.00000001435,
      -0.00000000146,    -0.00000000514,
      -0.00000000299,     0.00000000526,
       0.00000000734,     0.00000001034,
       0.00000000122,    -0.00000000096,
       0.00000000096,     0.00000000005,
       0.00000000020,     0.00000000106,
      -0.00000006063,    -0.00000008524,
      -0.00000000545,     0.00000000429,
      -0.00000000179,    -0.00000000089,
      -0.00000000121,    -0.00000000258,
       0.00000001016,    -0.00000001825,
      -0.00000000133,    -0.00000000056,
      -0.00000000036,    -0.00000000082,
       0.00000000080,    -0.00000000066,
       0.00000000027,    -0.00000000128,
      -0.00000000126,    -0.00000000030,
       0.00000000653,    -0.00000002823,
      -0.00000000157,    -0.00000000025,
      -0.00000000208,     0.00000000274,
       0.00000000261,     0.00000000214,
      -0.00000000076,    -0.00000000055,
      -0.00000004654,     0.00000005910,
       0.00000000349,     0.00000000250,
       0.00000000545,    -0.00000000305,
      -0.00000000256,    -0.00000000542,
      -0.00000001160,     0.00000000610,
       0.00000000084,     0.00000000157,
       0.00000011857,    -0.00000006230,
      -0.00000000378,    -0.00000000702,
      -0.00000000532,     0.00000000124,
       0.00000000012,     0.00000000462,
       0.00000001175,    -0.00000000193,
      -0.00000000017,    -0.00000000156,
      -0.00000010916,     0.00000001787,
       0.00000000080,     0.00000000679,
      -0.00000000079,    -0.00000000292,
      -0.00000000167,     0.00000000293,
       0.00000000419,     0.00000000592,
       0.00000000070,    -0.00000000053,
      -0.00000003507,    -0.00000004938,
      -0.00000000310,     0.00000000239,
      -0.00000000082,    -0.00000000052,
      -0.00000000081,    -0.00000000138,
       0.00000000390,    -0.00000001756,
      -0.00000000117,     0.00000000159,
       0.00000000150,     0.00000000120,
      -0.00000002653,     0.00000003469,
       0.00000000202,     0.00000000142,
       0.00000000285,    -0.00000000171,
      -0.00000000139,    -0.00000000283,
      -0.00000000619,     0.00000000347,
       0.00000000046,     0.00000000082,
       0.00000006282,    -0.00000003511,
      -0.00000000209,    -0.00000000368,
      -0.00000000252,     0.00000000084,
       0.00000000016,     0.00000000210,
      -0.00000005171,     0.00000001256,
       0.00000000060,     0.00000000316,
      -0.00000000043,    -0.00000000163,
      -0.00000000091,     0.00000000161,
      -0.00000002004,    -0.00000002781,
      -0.00000000171,     0.00000000133,
      -0.00000000052,    -0.00000000072,
      -0.00000000065,     0.00000000091,
       0.00000000085,     0.00000000067,
      -0.00000001484,     0.00000002010,
       0.00000000115,     0.00000000079,
       0.00000000147,    -0.00000000096,
      -0.00000000075,    -0.00000000145,
       0.00000003275,    -0.00000001965,
      -0.00000000114,    -0.00000000190,
      -0.00000000117,     0.00000000053,
      -0.00000002397,     0.00000000825,
       0.00000000041,     0.00000000144,
      -0.00000000048,     0.00000000088,
      -0.00000001134,    -0.00000001527,
      -0.00000000093,     0.00000000073,
       0.00000000075,    -0.00000000053,
       0.00000001683,    -0.00000001093,
      -0.00000000063,    -0.00000000097,
      -0.00000000015,     0.00000000416,
       0.00000000020,     0.00000000082,
      -0.00000000075,     0.00000001100,
       0.00000000072,     0.00000000000,
      -0.00000000001,    -0.00000000072,
      -0.00000000110,     0.00000001807,
       0.00000000156,    -0.00000000001,
      -0.00000000001,    -0.00000000162,
      -0.00000000246,     0.00000003716,
       0.00000000095,     0.00000000001,
       0.00000000442,    -0.00000000087,
      -0.00000000002,    -0.00000000687,
       0.00000000214,     0.00000030809,
       0.00000000722,     0.00000000029,
       0.00000019530,     0.00000000434,
       0.00000000052,    -0.00000022319,
      -0.00000000600,     0.00000000750,
       0.00000028781,     0.00000106554,
      -0.00000001189,    -0.00000005009,
       0.00000135687,     0.00000006810,
      -0.00000000786,    -0.00000137511,
      -0.00000000890,    -0.00000002710,
      -0.00000001001,    -0.00000002557,
      -0.00391148786,    -0.00022897364,
       0.00001631180,     0.00006221660,
      -0.00000015724,    -0.00000116572,
      -0.00000009088,     0.00000006384,
       0.00083470982,    -0.01501623169,
      -0.00023442555,     0.00006951474,
       0.00000516751,    -0.00000114987,
      -0.00000025031,    -0.00000035635,
      -0.00209611201,     0.00000366856,
       0.00000004315,    -0.00000009749,
       0.00000000020,    -0.00000001180,
       0.00000557048,     0.00210406689,
       0.00000002423,    -0.00000022613,
      -0.00000001180,     0.00000000020,
       0.00000347182,     0.00000360181,
      -0.00000001588,     0.00000011261,
       0.00000031788,     0.00000423425,
       0.00000015493,    -0.00000003938,
       0.00831185648,     0.00043929062,
      -0.00003413324,    -0.00013212968,
       0.00000021174,     0.00000223182,
       0.00000008216,    -0.00000006538,
      -0.00056728847,     0.01016045634,
       0.00015927049,    -0.00004568863,
      -0.00000291669,     0.00000042810,
       0.00000007712,     0.00000009754,
       0.00059960178,    -0.00000300906,
      -0.00000012691,    -0.00000002848,
      -0.00000411250,    -0.00059734670,
       0.00000007372,     0.00000011034,
      -0.00000001114,    -0.00000000597,
      -0.00000066501,    -0.00000218483,
      -0.00000001272,    -0.00000010061,
      -0.00000086217,    -0.00000281996,
      -0.00000011466,     0.00000003333,
       0.00000000001,     0.00000000067,
       0.00000000003,     0.00000000007,
      -0.00000000001,     0.00000000168,
      -0.00000000001,     0.00000000275,
      -0.00000000014,     0.00000000501,
       0.00000000064,    -0.00000000008,
       0.00000000066,     0.00000003697,
       0.00000001217,     0.00000000090,
      -0.00000000061,     0.00000000956,
       0.00000002378,     0.00000015728,
      -0.00000000190,     0.00000000134,
       0.00000006630,     0.00000001337,
      -0.00000000043,    -0.00000000114,
      -0.00000000308,     0.00000005856,
       0.00000000095,     0.00000000087,
      -0.00000000032,    -0.00000000183,
      -0.00000000028,    -0.00000000142,
       0.00090077875,     0.00000266396,
      -0.00000039147,    -0.00000011282,
       0.00000001683,     0.00000001140,
      -0.00000058824,    -0.00014437447,
       0.00000001227,    -0.00000000533,
       0.00000000328,     0.00000000056,
      -0.00000888702,    -0.00000049534,
       0.00000003771,     0.00000015208,
      -0.00000000054,    -0.00000000261,
       0.00000023341,    -0.00000969872,
      -0.00000013918,     0.00000004431,
       0.00000000367,    -0.00000000041,
      -0.00000014851,    -0.00000021939,
       0.00000000120,    -0.00000000583,
       0.00000007077,    -0.00000051857,
       0.00000000149,     0.00000000447,
      -0.00105686700,     0.00000430137,
       0.00000001281,    -0.00000002796,
       0.00000021452,     0.00008886007,
      -0.00000000159,    -0.00000000198,
      -0.00000000159,     0.00000000022,
      -0.00000445175,     0.00000036765,
       0.00000002016,    -0.00000003746,
       0.00000054105,     0.00000234792,
      -0.00000007129,    -0.00000001064,
       0.00000000145,    -0.00000000011,
      -0.00000011086,     0.00000015484,
       0.00000000128,     0.00000000463,
       0.00000004281,     0.00000035591,
       0.00000000101,    -0.00000000347,
      -0.00000000003,    -0.00000000011,
       0.00000000011,    -0.00000000003,
      -0.00000000038,     0.00000000012,
      -0.00000000012,    -0.00000000038,
      -0.00000000192,    -0.00000000027,
       0.00000000027,    -0.00000000190,
      -0.00000000093,    -0.00000000092,
      -0.00000000374,    -0.00000000303,
       0.00000000317,    -0.00000000345,
      -0.00000032332,    -0.00000042916,
      -0.00000000532,    -0.00000006695,
       0.00000000007,     0.00000000201,
       0.00000012363,     0.00000004390,
      -0.00000000458,     0.00000000204,
      -0.00000004484,     0.00000012289,
      -0.00000000205,    -0.00000000460,
       0.00000006054,    -0.00000002284,
      -0.00000000062,    -0.00000000073,
       0.00000002252,     0.00000006017,
       0.00000000073,    -0.00000000062,
      -0.00000035313,     0.00000005981,
       0.00000001163,     0.00000003407,
      -0.00000000040,    -0.00000000070,
       0.00000006030,    -0.00000000063,
       0.00000000486,     0.00000007051,
      -0.00000000131,    -0.00000000017,
      -0.00000004146,     0.00000001732,
      -0.00000001709,    -0.00000004104,
       0.00000000015,     0.00000000138,
      -0.00000000140,     0.00000000014,
       0.00000008143,     0.00000002193,
      -0.00000000136,     0.00000000087,
      -0.00000002432,    -0.00000000924,
       0.00000000090,    -0.00000000022,
       0.00000000922,    -0.00000002444,
       0.00000000021,     0.00000000089,
      -0.00000000159,    -0.00000000028,
       0.00000000006,    -0.00000000197,
       0.00000000301,     0.00000000073,
      -0.00000000076,     0.00000000278,
       0.00000000049,    -0.00000000118,
       0.00000000002,    -0.00000000004,
      -0.00000000093,    -0.00000000029,
       0.00000000010,    -0.00000000106,
      -0.00000000004,     0.00000000005,
      -0.00000000002,    -0.00000000009,
       0.00000000373,    -0.00000000356,
      -0.00000000429,    -0.00000000002,
       0.00000000005,    -0.00000000432,
       0.00000000512,    -0.00000000590,
      -0.00000000970,    -0.00000000015,
       0.00000000008,    -0.00000000978,
       0.00000001790,     0.00000000248,
      -0.00000000085,     0.00000000059,
      -0.00000002848,     0.00000000103,
      -0.00000000175,    -0.00000002786,
       0.00000003255,     0.00000008964,
      -0.00000000661,     0.00000000366,
      -0.00000062065,     0.00000000152,
      -0.00000000382,    -0.00000062135,
      -0.00000000696,     0.00000001278,
      -0.00000075068,    -0.00000025116,
       0.00000007044,    -0.00000009470,
       0.00000000559,     0.00000000556,
      -0.00000303661,    -0.00000001753,
       0.00000000334,    -0.00000000055,
       0.00000000665,    -0.00000302627,
       0.00000000238,     0.00000000283,
      -0.00000000247,    -0.00000000601,
       0.00000515190,    -0.00000020184,
      -0.00000002317,     0.00000007876,
       0.00000025106,    -0.00001262146,
       0.00000019831,     0.00000004634,
      -0.00000000345,     0.00000000034,
      -0.00002576283,    -0.00000008964,
      -0.00000000076,     0.00000000940,
      -0.00000000006,    -0.00000000153,
       0.00000005643,    -0.00002643488,
       0.00000000816,     0.00000000748,
       0.00000000124,    -0.00000000001,
       0.00000011972,    -0.00000012860,
      -0.00000000070,    -0.00000000551,
       0.00000003596,    -0.00000010393,
       0.00000000680,     0.00000000098,
       0.00002942760,    -0.00013878423,
       0.00002907329,     0.00001002937,
      -0.00000115164,     0.00000131617,
      -0.00000004917,    -0.00000006381,
       0.00001687037,    -0.00000408348,
      -0.00000024931,     0.00000063377,
      -0.00000002952,    -0.00000002759,
       0.00000000177,    -0.00000000131,
       0.00136232980,    -0.00000002359,
      -0.00000000153,    -0.00000001135,
       0.00000000122,     0.00000000100,
      -0.00000007335,     0.00136278888,
      -0.00000001611,    -0.00000003119,
       0.00000000369,    -0.00000000064,
      -0.00000006803,     0.00000003295,
       0.00000000045,    -0.00000002052,
       0.00000001495,     0.00000007735,
       0.00000001894,     0.00000000015,
       0.00000000440,     0.00000000800,
      -0.00000000042,     0.00000000018,
       0.00000000620,     0.00000002774,
      -0.00000000076,     0.00000000019,
      -0.00000002781,     0.00000000606,
      -0.00000000019,    -0.00000000076,
       0.00000000137,     0.00000000102,
       0.00000000002,    -0.00000000004,
      -0.00000000107,     0.00000000173,
       0.00000000004,     0.00000000002,
       0.00000000032,     0.00000000026,
      -0.00000000166,     0.00000000178,
      -0.00000000178,    -0.00000000166,
       0.00000000123,     0.00000000125,
      -0.00000000125,     0.00000000123,
      -0.00000000020,     0.00000000075,
      -0.00000000228,     0.00000000945,
      -0.00000000048,    -0.00000000037,
      -0.00000000945,    -0.00000000228,
       0.00000000036,    -0.00000000048,
       0.00000000121,    -0.00000000135,
       0.00000000135,     0.00000000121,
      -0.00000001052,    -0.00000000905,
       0.00000000121,    -0.00000000186,
       0.00000009381,    -0.00000004224,
       0.00000000411,     0.00000000684,
       0.00000004212,     0.00000009402,
      -0.00000000684,     0.00000000411,
       0.00000000475,     0.00000000010,
      -0.00000000025,     0.00000000460,
       0.00000000782,    -0.00000000998,
       0.00000000236,    -0.00000000060,
       0.00000032014,     0.00000007978,
       0.00000000810,     0.00000001548,
      -0.00000000077,     0.00000000071,
      -0.00000007965,     0.00000031874,
      -0.00000001549,     0.00000000810,
      -0.00000000071,    -0.00000000077,
       0.00000001219,    -0.00000000053,
       0.00000000049,     0.00000001212,
       0.00000000868,     0.00000001519,
      -0.00000003059,     0.00000011504,
      -0.00000001476,    -0.00000001631,
       0.00000000242,    -0.00000000124,
      -0.00000016431,     0.00000103281,
      -0.00000010102,    -0.00000007329,
       0.00000000616,    -0.00000000611,
      -0.00000103283,    -0.00000016131,
       0.00000007317,    -0.00000010051,
       0.00000000612,     0.00000000598,
       0.00000000672,     0.00000003531,
       0.00000000343,     0.00000000241,
      -0.00000003507,     0.00000000606,
      -0.00000000237,     0.00000000333,
      -0.00002199855,    -0.00000190566,
       0.00000047047,    -0.00000058581,
      -0.00000001115,     0.00000003112,
       0.00000020303,    -0.00000111224,
       0.00000002249,     0.00000003063,
      -0.00000000134,    -0.00000000096,
       0.00000600780,    -0.00000042780,
      -0.00000003184,     0.00000009162,
       0.00000000060,    -0.00000000244,
       0.00000042548,     0.00000599273,
      -0.00000009126,    -0.00000003235,
       0.00000000351,     0.00000000057,
       0.00000033997,    -0.00000013940,
      -0.00000000336,    -0.00000000410,
       0.00000013396,     0.00000032899,
       0.00000000515,    -0.00000000353,
       0.00000472437,    -0.00000819565,
       0.00000205118,    -0.00000059592,
       0.00000008836,     0.00000032066,
      -0.00000002984,     0.00000000342,
       0.00000019596,     0.00000053614,
       0.00000001100,     0.00000004981,
      -0.00000000899,     0.00000000294,
       0.00000092628,    -0.00000246678,
       0.00000050500,     0.00000018796,
      -0.00000002029,     0.00000002293,
      -0.00000000089,    -0.00000000116,
       0.00000246799,     0.00000094815,
      -0.00000018963,     0.00000050528,
      -0.00000002316,    -0.00000002050,
       0.00000000114,    -0.00000000092,
      -0.00000035388,    -0.00000000877,
      -0.00000000454,    -0.00000000155,
      -0.00000000052,     0.00000000066,
       0.00000001236,    -0.00000034472,
       0.00000000133,    -0.00000000453,
      -0.00000000066,    -0.00000000052,
      -0.00000000087,     0.00000000128,
       0.00000000127,     0.00000000089,
      -0.00000001925,     0.00000003118,
       0.00000000182,     0.00000000094,
       0.00000000427,    -0.00000000153,
      -0.00000000139,    -0.00000000431,
      -0.00000000743,     0.00000000332,
       0.00000000059,     0.00000000121,
       0.00000000092,     0.00000000050,
       0.00000000065,    -0.00000000088,
       0.00000009892,    -0.00000004443,
      -0.00000000277,    -0.00000000566,
      -0.00000000806,    -0.00000000491,
      -0.00000000015,     0.00000000134,
      -0.00000000545,     0.00000000790,
       0.00000000143,     0.00000000021,
       0.00000001692,     0.00000000685,
       0.00000000115,    -0.00000000274,
       0.00000000086,    -0.00000000196,
      -0.00000000227,    -0.00000000132,
      -0.00000020207,    -0.00000008335,
      -0.00000000189,     0.00000002707,
      -0.00000000525,     0.00000001263,
       0.00000000175,     0.00000000003,
      -0.00000000913,     0.00000001865,
       0.00000000208,    -0.00000000139,
       0.00000001965,     0.00000001058,
      -0.00000000139,    -0.00000000231,
       0.00000001239,    -0.00000003522,
      -0.00000000573,    -0.00000000180,
       0.00000000042,     0.00000000067,
      -0.00000000487,     0.00000000032,
       0.00000000000,     0.00000000625,
      -0.00000013486,     0.00000037582,
       0.00000004544,    -0.00000002773,
       0.00000002562,     0.00000000813,
      -0.00000000184,    -0.00000000299,
       0.00000004899,    -0.00000000200,
       0.00000000083,     0.00000000096,
      -0.00000000113,    -0.00000005288,
       0.00000000053,    -0.00000000075,
      -0.00000010811,     0.00000000274,
       0.00000000031,     0.00000001555,
      -0.00000000058,    -0.00000000275,
       0.00000100668,    -0.00000002634,
       0.00000001302,     0.00000001509,
      -0.00000000119,    -0.00000006768,
       0.00000000095,    -0.00000000089,
       0.00000000576,     0.00000001505,
      -0.00000000102,    -0.00000000083,
       0.00000000833,    -0.00000000637,
      -0.00000000036,     0.00000000181,
      -0.00000001108,    -0.00000002894,
      -0.00000000344,     0.00000000154,
      -0.00000000385,    -0.00000000010,
      -0.00000000017,    -0.00000000412,
       0.00000008831,     0.00000023076,
      -0.00000002745,    -0.00000001266,
       0.00000001535,    -0.00000000692,
      -0.00000000081,     0.00000000183,
       0.00000000736,     0.00000000102,
       0.00000000117,     0.00000000954,
      -0.00000000031,    -0.00000000105,
       0.00000000106,    -0.00000000033,
      -0.00000002316,     0.00000001474,
       0.00000000119,     0.00000000149,
       0.00000000048,     0.00000000184,
      -0.00000000194,     0.00000000067,
      -0.00000000056,     0.00000000074,
       0.00000000073,     0.00000000059,
      -0.00000001285,     0.00000001845,
       0.00000000105,     0.00000000064,
       0.00000000246,    -0.00000000063,
      -0.00000000053,    -0.00000000247,
       0.00000005787,    -0.00000002030,
      -0.00000000125,    -0.00000000329,
      -0.00000000383,    -0.00000000335,
      -0.00000000370,     0.00000000369,
       0.00000000086,     0.00000000020,
       0.00000000848,     0.00000000529,
       0.00000000084,    -0.00000000132,
       0.00000000060,    -0.00000000106,
      -0.00000000124,    -0.00000000091,
      -0.00000009835,    -0.00000006230,
      -0.00000000289,     0.00000001681,
      -0.00000000385,     0.00000000611,
       0.00000000106,     0.00000000013,
      -0.00000000637,     0.00000001018,
       0.00000000108,    -0.00000000074,
       0.00000001071,     0.00000000730,
      -0.00000000071,    -0.00000000122,
       0.00000001067,    -0.00000002013,
      -0.00000000312,    -0.00000000151,
      -0.00000000236,     0.00000000027,
      -0.00000000003,     0.00000000312,
      -0.00000011315,     0.00000021133,
       0.00000002475,    -0.00000001503,
       0.00000001397,     0.00000000679,
      -0.00000000096,    -0.00000000159,
       0.00000002414,    -0.00000000154,
       0.00000000062,     0.00000000055,
      -0.00000000073,    -0.00000002621,
      -0.00000005543,     0.00000000225,
       0.00000000024,     0.00000000772,
      -0.00000000018,    -0.00000000165,
       0.00000051598,    -0.00000002134,
       0.00000001097,     0.00000000854,
      -0.00000000098,    -0.00000003367,
       0.00000000051,    -0.00000000072,
       0.00000000311,     0.00000000918,
      -0.00000000064,    -0.00000000048,
       0.00000000524,    -0.00000000425,
      -0.00000000021,     0.00000000109,
      -0.00000000714,    -0.00000001812,
      -0.00000000214,     0.00000000096,
      -0.00000000204,    -0.00000000006,
      -0.00000000014,    -0.00000000225,
       0.00000005808,     0.00000014746,
      -0.00000001751,    -0.00000000743,
       0.00000000955,    -0.00000000431,
      -0.00000000047,     0.00000000113,
       0.00000000377,     0.00000000074,
       0.00000000092,     0.00000000544,
      -0.00000001826,     0.00000001228,
       0.00000000093,     0.00000000114,
       0.00000000033,     0.00000000124,
      -0.00000000134,     0.00000000051,
       0.00000000140,    -0.00000000022,
      -0.00000000015,    -0.00000000140,
       0.00000003336,    -0.00000000835,
      -0.00000000052,    -0.00000000189,
      -0.00000000178,    -0.00000000218,
      -0.00000000239,     0.00000000167,
       0.00000000058,    -0.00000000063,
      -0.00000000067,    -0.00000000057,
      -0.00000004661,    -0.00000004300,
      -0.00000000261,     0.00000000289,
      -0.00000000404,     0.00000000554,
       0.00000000580,     0.00000000463,
       0.00000000759,    -0.00000001134,
      -0.00000000169,    -0.00000000104,
      -0.00000000114,     0.00000000021,
      -0.00000000003,     0.00000000153,
      -0.00000007903,     0.00000011752,
       0.00000001303,    -0.00000000801,
       0.00000000757,     0.00000000470,
      -0.00000000049,    -0.00000000082,
       0.00000001180,    -0.00000000114,
      -0.00000000050,    -0.00000001287,
      -0.00000002796,     0.00000000177,
       0.00000000018,     0.00000000379,
       0.00000026018,    -0.00000001661,
      -0.00000000078,    -0.00000001658,
       0.00000000167,     0.00000000541,
       0.00000000315,    -0.00000000270,
      -0.00000000444,    -0.00000001090,
      -0.00000000128,     0.00000000059,
      -0.00000000105,    -0.00000000003,
      -0.00000000011,    -0.00000000120,
       0.00000003673,     0.00000009006,
       0.00000000571,    -0.00000000261,
       0.00000000184,     0.00000000050,
       0.00000000068,     0.00000000303,
      -0.00000001331,     0.00000000940,
       0.00000000069,     0.00000000081,
      -0.00000000088,     0.00000000036,
       0.00000001898,    -0.00000000276,
      -0.00000000018,    -0.00000000107,
      -0.00000000080,    -0.00000000137,
      -0.00000000150,     0.00000000072,
      -0.00000002140,    -0.00000002815,
      -0.00000000168,     0.00000000133,
      -0.00000000242,     0.00000000301,
       0.00000000312,     0.00000000277,
       0.00000000490,    -0.00000000632,
      -0.00000000091,    -0.00000000066,
      -0.00000005034,     0.00000006482,
       0.00000000408,     0.00000000296,
       0.00000000571,    -0.00000000080,
      -0.00000000035,    -0.00000000624,
      -0.00000001388,     0.00000000131,
       0.00000000013,     0.00000000184,
       0.00000012918,    -0.00000001222,
      -0.00000000059,    -0.00000000807,
       0.00000000090,     0.00000000311,
       0.00000000183,    -0.00000000164,
      -0.00000000074,     0.00000000034,
       0.00000002257,     0.00000005317,
       0.00000000331,    -0.00000000155,
       0.00000000085,     0.00000000032,
       0.00000000048,     0.00000000165,
      -0.00000000035,    -0.00000000084,
      -0.00000000091,     0.00000000029,
      -0.00000000947,    -0.00000001776,
      -0.00000000105,     0.00000000060,
      -0.00000000139,     0.00000000163,
       0.00000000167,     0.00000000160,
      -0.00000003027,     0.00000003550,
       0.00000000219,     0.00000000176,
       0.00000000274,    -0.00000000054,
      -0.00000000024,    -0.00000000299,
       0.00000006323,    -0.00000000853,
      -0.00000000042,    -0.00000000389,
       0.00000000049,     0.00000000175,
       0.00000000103,    -0.00000000097,
       0.00000001357,     0.00000003050,
       0.00000000187,    -0.00000000090,
       0.00000000033,     0.00000000090,
      -0.00000000078,     0.00000000088,
       0.00000000089,     0.00000000090,
      -0.00000001745,     0.00000001932,
       0.00000000117,     0.00000000101,
       0.00000000129,    -0.00000000035,
      -0.00000000016,    -0.00000000142,
       0.00000003046,    -0.00000000568,
      -0.00000000029,    -0.00000000185,
       0.00000000027,     0.00000000096,
       0.00000000803,     0.00000001708,
       0.00000000103,    -0.00000000052,
       0.00000000011,     0.00000000017,
       0.00000000019,     0.00000000063,
      -0.00000000064,     0.00000000019,
       0.00000000003,     0.00000000003,
      -0.00000000003,     0.00000000003,
      -0.00000000030,    -0.00000000031,
       0.00000000232,    -0.00000000116,
       0.00000000117,     0.00000000233,
       0.00000000015,    -0.00000000162,
       0.00000001045,     0.00000000138,
      -0.00000000136,     0.00000001057,
       0.00000000324,     0.00000000123,
       0.00000001107,     0.00000002928,
      -0.00000000365,    -0.00000000173,
      -0.00000002938,     0.00000001111,
       0.00000000183,    -0.00000000369,
      -0.00000000147,     0.00000000216,
      -0.00000000223,    -0.00000000145,
       0.00000281568,     0.00000048287,
      -0.00000006981,     0.00000006542,
      -0.00000003591,     0.00000000965,
       0.00000000298,    -0.00000000149,
      -0.00000134030,    -0.00000011124,
       0.00000002840,    -0.00000003553,
      -0.00000000066,     0.00000000183,
       0.00000011134,    -0.00000134169,
       0.00000003547,     0.00000002862,
      -0.00000000179,    -0.00000000067,
       0.00000003406,    -0.00000002120,
       0.00000002105,     0.00000003319,
       0.00000008942,    -0.00000003007,
       0.00000000243,     0.00000001832,
      -0.00000000299,     0.00000000634,
      -0.00000000637,    -0.00000000288,
      -0.00000001686,     0.00000000577,
      -0.00000000563,    -0.00000001652,
      -0.00000000004,    -0.00000000010,
       0.00000000010,    -0.00000000004,
      -0.00000000035,     0.00000000025,
      -0.00000000025,    -0.00000000035,
      -0.00000000187,    -0.00000000016,
       0.00000000016,    -0.00000000189,
      -0.00000000002,    -0.00000000378,
       0.00000000054,     0.00000000058,
       0.00000000345,     0.00000000018,
      -0.00000000065,     0.00000000051,
      -0.00000025572,    -0.00000003322,
       0.00000000168,    -0.00000000453,
       0.00000012215,     0.00000002193,
      -0.00000000300,     0.00000000281,
      -0.00000002193,     0.00000012215,
      -0.00000000281,    -0.00000000302,
      -0.00000000095,     0.00000000091,
      -0.00000000094,    -0.00000000093,
       0.00000000592,    -0.00000000197,
       0.00000000199,     0.00000000597,
       0.00000000116,    -0.00000000016,
      -0.00000000004,     0.00000000015,
      -0.00000000025,     0.00000000043,
      -0.00000000043,    -0.00000000025,
       0.00000000147,     0.00000000140,
      -0.00000000140,     0.00000000147,
      -0.00000000041,     0.00000000439,
      -0.00000000439,    -0.00000000041,
       0.00000000029,     0.00000000233,
       0.00000001043,     0.00000001780,
      -0.00000000445,     0.00000000236,
      -0.00000000027,    -0.00000000074,
      -0.00000001771,     0.00000001043,
      -0.00000000236,    -0.00000000445,
       0.00000000074,    -0.00000000027,
       0.00000000089,     0.00000000044,
      -0.00000000044,     0.00000000090,
       0.00000009814,     0.00000001263,
      -0.00000000366,    -0.00000005848,
       0.00000000233,     0.00000000094,
      -0.00000028372,    -0.00000001424,
       0.00000000263,    -0.00000000812,
       0.00000001399,    -0.00000028517,
       0.00000000814,     0.00000000268,
       0.00000003782,    -0.00000001158,
       0.00000001114,     0.00000003677,
       0.00000027227,     0.00000006436,
      -0.00000001333,    -0.00000004263,
      -0.00000000163,     0.00000003232,
       0.00000000061,    -0.00000000114,
       0.00000016575,    -0.00000015200,
       0.00000003980,    -0.00000000983,
       0.00000000179,     0.00000000646,
       0.00000015202,     0.00000016614,
       0.00000000988,     0.00000003979,
      -0.00000000643,     0.00000000180,
      -0.00000002747,     0.00000000508,
      -0.00000000484,    -0.00000002678,
       0.00000000026,     0.00000000171,
      -0.00000000156,     0.00000000027,
       0.00000012882,     0.00000004727,
      -0.00000000227,    -0.00000000207,
      -0.00000007788,    -0.00000001741,
       0.00000000215,    -0.00000000297,
       0.00000001744,    -0.00000007790,
       0.00000000292,     0.00000000216,
       0.00000000349,    -0.00000000089,
       0.00000000085,     0.00000000341,
       0.00000004421,     0.00000000269,
       0.00000000028,     0.00000000268,
       0.00000001123,     0.00000000123,
      -0.00000000010,    -0.00000000128,
      -0.00000000125,     0.00000001124,
       0.00000000128,    -0.00000000010,
      -0.00000000232,     0.00000000057,
      -0.00000000055,    -0.00000000227,
       0.00000000124,    -0.00000000002,
       0.00000000002,     0.00000000124,
      -0.00000000949,     0.00000000000,
       0.00000000000,     0.00000001474,
       0.00000000096,     0.00000000000,
       0.00000000000,    -0.00000000088,
       0.00000000339,     0.00000000000,
       0.00000000000,    -0.00000000571,
       0.00000000000,     0.00000000052,
       0.00000000061,     0.00000000000,
       0.00000000000,    -0.00000000188,
};

static
const unsigned char vsop87_instructions[] = {
  0,0,49,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  13,
  14,
  19,
  20,
  25,
  26,
  34,
  38,
  44,
  50,
  56,
  65,
  66,
  71,
  77,
  98,
  99,
  104,
  110,
  116,
  122,
  133,
  134,
  139,
  144,
  149,
  163,
  164,
  169,
  175,
  186,
  197,
  198,
  203,
  208,
  213,
  227,
  233,
  239,
  245,
  249,
  0,49,2,
  7,
  13,
  0xFE,
  0,47,6,
  3,
  7,
  8,
  13,
  14,
  25,
  0,102,1,
  3,
  0xFE,
  0xFE,
  0,45,9,
  0,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  0,98,1,
  3,
  0xFE,
  0xFE,
  0,43,16,
  0,
  1,
  3,
  4,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  20,
  25,
  26,
  38,
  44,
  0,90,2,
  7,
  13,
  0xFE,
  0,92,2,
  7,
  13,
  0xFE,
  0,94,3,
  3,
  7,
  13,
  0xFE,
  0,96,1,
  3,
  0xFE,
  0xFE,
  0,41,19,
  0,
  1,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  20,
  25,
  26,
  31,
  34,
  38,
  44,
  0,88,2,
  7,
  13,
  0xFE,
  0,90,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,92,9,
  0,
  3,
  4,
  7,
  13,
  25,
  34,
  35,
  65,
  0xFE,
  0xFE,
  0,39,25,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  20,
  25,
  26,
  31,
  32,
  34,
  38,
  39,
  44,
  45,
  50,
  56,
  0,88,5,
  0,
  3,
  4,
  7,
  13,
  0xFE,
  0,90,5,
  0,
  3,
  4,
  7,
  13,
  0xFE,
  0,144,1,
  3,
  0xFE,
  0xFE,
  0,37,29,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  20,
  21,
  25,
  26,
  27,
  31,
  32,
  33,
  34,
  35,
  38,
  39,
  44,
  45,
  50,
  56,
  0,88,2,
  0,
  3,
  0xFE,
  0,140,1,
  3,
  0xFE,
  0,144,1,
  3,
  0xFE,
  0xFE,
  0,35,27,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  20,
  21,
  25,
  26,
  27,
  31,
  32,
  34,
  35,
  38,
  44,
  45,
  50,
  56,
  0xFE,
  0,33,22,
  0,
  1,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  19,
  20,
  21,
  25,
  26,
  31,
  34,
  38,
  39,
  44,
  50,
  56,
  0,91,3,
  3,
  7,
  13,
  0xFE,
  0,89,3,
  3,
  7,
  13,
  0xFE,
  0,87,3,
  3,
  7,
  13,
  0xFE,
  0,145,1,
  3,
  0xFE,
  0xFE,
  0,34,15,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  20,
  25,
  26,
  34,
  38,
  44,
  0xFE,
  0,36,12,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  20,
  25,
  26,
  0,103,1,
  3,
  0xFE,
  0xFE,
  0,38,9,
  0,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  0xFE,
  0,40,4,
  3,
  7,
  13,
  25,
  0,111,3,
  3,
  7,
  13,
  0xFE,
  0xFE,
  0,42,0,
  0,111,3,
  3,
  7,
  13,
  0xFE,
  0xFE,
  0,46,0,
  0,119,1,
  3,
  0xFE,
  0xFE,
  0,105,0,
  0,160,9,
  0,
  3,
  4,
  7,
  13,
  65,
  66,
  98,
  99,
  0xFE,
  0xFE,
  0,103,0,
  0,156,1,
  3,
  0xFE,
  0xFE,
  0,101,0,
  0,152,1,
  3,
  0xFE,
  0xFE,
  0,99,3,
  3,
  7,
  13,
  0xFE,
  0,97,8,
  0,
  3,
  7,
  8,
  13,
  14,
  19,
  25,
  0xFE,
  0,95,13,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  26,
  71,
  77,
  0xFE,
  0,93,27,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  20,
  21,
  25,
  26,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  89,
  0xFE,
  0,91,26,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  20,
  25,
  26,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  89,
  0,143,1,
  3,
  0xFE,
  0xFE,
  0,89,23,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  19,
  20,
  25,
  26,
  62,
  63,
  65,
  66,
  71,
  77,
  83,
  89,
  0,147,2,
  3,
  4,
  0xFE,
  0xFE,
  0,87,20,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  20,
  25,
  26,
  62,
  65,
  71,
  77,
  83,
  84,
  89,
  90,
  0,35,4,
  3,
  4,
  7,
  13,
  0xFE,
  0,151,1,
  3,
  0xFE,
  0xFE,
  0,88,11,
  0,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  71,
  77,
  0xFE,
  0,90,6,
  0,
  3,
  7,
  13,
  19,
  25,
  0xFE,
  0,92,3,
  3,
  7,
  13,
  0xFE,
  0,110,0,
  0,53,3,
  3,
  7,
  13,
  0xFE,
  0,51,3,
  3,
  7,
  13,
  0xFE,
  0xFE,
  0,155,7,
  3,
  4,
  5,
  7,
  8,
  13,
  14,
  0xFE,
  0,153,6,
  3,
  4,
  7,
  8,
  13,
  14,
  0xFE,
  0,151,9,
  0,
  3,
  4,
  7,
  8,
  13,
  14,
  104,
  110,
  0xFE,
  0,149,14,
  0,
  1,
  3,
  4,
  5,
  7,
  8,
  13,
  14,
  25,
  95,
  98,
  104,
  110,
  0xFE,
  0,147,17,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  95,
  96,
  98,
  104,
  105,
  110,
  111,
  0xFE,
  0,145,19,
  0,
  1,
  3,
  4,
  5,
  7,
  8,
  13,
  14,
  19,
  25,
  95,
  96,
  97,
  98,
  104,
  105,
  110,
  111,
  0xFE,
  0,143,18,
  0,
  1,
  3,
  4,
  7,
  13,
  14,
  19,
  25,
  95,
  96,
  97,
  98,
  99,
  104,
  110,
  116,
  122,
  0xFE,
  0,141,15,
  0,
  3,
  7,
  13,
  14,
  19,
  25,
  95,
  98,
  104,
  105,
  110,
  111,
  116,
  122,
  0xFE,
  0,140,15,
  0,
  3,
  4,
  7,
  13,
  19,
  25,
  95,
  98,
  104,
  105,
  110,
  111,
  116,
  122,
  0xFE,
  0,142,7,
  3,
  7,
  13,
  19,
  25,
  104,
  110,
  0,95,3,
  3,
  7,
  13,
  0xFE,
  0,93,1,
  3,
  0xFE,
  0xFE,
  0,144,2,
  7,
  13,
  0xFE,
  0,215,4,
  0,
  3,
  7,
  13,
  1,10,2,
  3,
  13,
  0xFE,
  0xFE,
  0,213,10,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  0xFE,
  0,211,16,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  25,
  139,
  144,
  0xFE,
  0,209,21,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  20,
  25,
  26,
  127,
  139,
  140,
  144,
  145,
  0,154,0,
  0,101,3,
  3,
  7,
  13,
  0xFE,
  0,97,1,
  3,
  0xFE,
  0xFE,
  0xFE,
  0,207,23,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  20,
  25,
  26,
  127,
  128,
  133,
  134,
  139,
  144,
  221,
  0,93,2,
  3,
  4,
  0xFE,
  1,1,6,
  0,
  3,
  4,
  7,
  13,
  14,
  0xFE,
  0,251,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,249,3,
  3,
  7,
  13,
  0xFE,
  0xFE,
  0,205,24,
  0,
  1,
  3,
  4,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  25,
  26,
  127,
  133,
  139,
  140,
  144,
  145,
  149,
  153,
  157,
  191,
  221,
  0,35,1,
  3,
  0xFE,
  0,33,1,
  3,
  0xFE,
  0,93,1,
  3,
  0xFE,
  1,1,8,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  0xFE,
  0,251,7,
  0,
  3,
  4,
  7,
  8,
  13,
  14,
  0xFE,
  0,249,4,
  0,
  3,
  7,
  13,
  0xFE,
  1,45,2,
  203,
  208,
  0xFE,
  1,44,2,
  203,
  208,
  0xFE,
  1,93,2,
  233,
  239,
  0xFE,
  1,92,2,
  233,
  239,
  0xFE,
  0xFE,
  1,3,0,
  1,48,2,
  3,
  13,
  0xFE,
  0xFE,
  1,1,11,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  25,
  139,
  144,
  0xFE,
  0,255,10,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  169,
  175,
  0xFE,
  0,253,19,
  0,
  1,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  25,
  157,
  158,
  169,
  170,
  175,
  176,
  0xFE,
  0,251,16,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  157,
  158,
  163,
  164,
  169,
  175,
  0xFE,
  0,249,19,
  0,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  157,
  163,
  169,
  170,
  175,
  176,
  181,
  186,
  191,
  221,
  0xFE,
  1,51,2,
  203,
  208,
  0xFE,
  1,49,7,
  0,
  3,
  7,
  13,
  191,
  203,
  208,
  0,211,0,
  1,2,2,
  3,
  13,
  0xFE,
  0xFE,
  1,94,4,
  191,
  197,
  203,
  208,
  0xFE,
  0xFE,
  1,47,11,
  0,
  3,
  7,
  13,
  191,
  192,
  197,
  203,
  208,
  213,
  217,
  1,94,2,
  191,
  197,
  0xFE,
  0xFE,
  1,45,12,
  3,
  191,
  197,
  203,
  208,
  213,
  214,
  217,
  218,
  221,
  233,
  239,
  1,95,3,
  191,
  203,
  208,
  0xFE,
  1,92,2,
  221,
  227,
  0xFE,
  0xFE,
  1,44,8,
  191,
  197,
  203,
  208,
  213,
  214,
  217,
  218,
  1,99,2,
  233,
  239,
  0xFE,
  1,97,2,
  221,
  227,
  0xFE,
  1,95,2,
  197,
  227,
  0xFE,
  0xFE,
  1,46,4,
  203,
  208,
  213,
  217,
  0xFE,
  1,97,2,
  233,
  239,
  0xFE,
  1,95,12,
  0,
  3,
  7,
  13,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,93,8,
  221,
  227,
  233,
  239,
  245,
  246,
  249,
  250,
  0xFE,
  1,92,10,
  221,
  222,
  227,
  228,
  233,
  239,
  245,
  246,
  249,
  250,
  0xFE,
  1,94,2,
  233,
  239,
  0xFE,
  0xFE,
  0,2,55,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  20,
  25,
  26,
  34,
  38,
  39,
  44,
  45,
  50,
  56,
  65,
  71,
  72,
  77,
  78,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  133,
  134,
  139,
  140,
  144,
  145,
  163,
  164,
  169,
  170,
  175,
  176,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  227,
  233,
  234,
  239,
  240,
  0,53,0,
  0,104,1,
  3,
  0xFE,
  0xFE,
  0,51,3,
  3,
  7,
  13,
  0,100,6,
  0,
  3,
  4,
  7,
  13,
  34,
  0xFE,
  0xFE,
  0,49,9,
  0,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  0,94,1,
  3,
  0xFE,
  0,96,3,
  3,
  7,
  13,
  0xFE,
  0xFE,
  0,47,11,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  26,
  0,90,2,
  7,
  13,
  0xFE,
  0,92,3,
  3,
  7,
  13,
  0xFE,
  0,94,5,
  0,
  3,
  4,
  7,
  13,
  0,207,1,
  3,
  0xFE,
  0xFE,
  0,98,0,
  0,147,1,
  3,
  0xFE,
  0xFE,
  0xFE,
  0,45,17,
  0,
  1,
  3,
  4,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  20,
  25,
  26,
  31,
  38,
  44,
  0,88,3,
  3,
  7,
  13,
  0xFE,
  0,90,13,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  34,
  35,
  65,
  0xFE,
  0,92,3,
  3,
  7,
  13,
  0xFE,
  0,94,1,
  3,
  0xFE,
  0,150,1,
  3,
  0xFE,
  0xFE,
  0,43,24,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  20,
  25,
  26,
  31,
  32,
  34,
  38,
  39,
  44,
  45,
  56,
  0,88,5,
  0,
  3,
  4,
  7,
  13,
  0xFE,
  0,90,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,92,1,
  3,
  0xFE,
  0,142,1,
  3,
  0xFE,
  0,144,5,
  0,
  3,
  4,
  7,
  13,
  0xFE,
  0xFE,
  0,41,32,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  20,
  21,
  25,
  26,
  27,
  31,
  32,
  33,
  34,
  35,
  36,
  38,
  39,
  40,
  44,
  45,
  46,
  50,
  56,
  0,88,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,90,2,
  0,
  3,
  0xFE,
  0,207,5,
  3,
  4,
  7,
  13,
  34,
  0xFE,
  0,205,5,
  0,
  3,
  4,
  7,
  13,
  0xFE,
  1,1,1,
  3,
  0xFE,
  0,249,1,
  3,
  0xFE,
  0xFE,
  0,39,27,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  20,
  25,
  26,
  27,
  31,
  32,
  34,
  35,
  38,
  39,
  44,
  45,
  50,
  56,
  0,89,4,
  3,
  4,
  7,
  13,
  0xFE,
  0,87,3,
  3,
  7,
  13,
  0xFE,
  0,88,2,
  0,
  3,
  0xFE,
  0,145,2,
  3,
  4,
  0xFE,
  0xFE,
  0,37,24,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  20,
  25,
  26,
  31,
  32,
  34,
  35,
  38,
  44,
  50,
  56,
  0,91,3,
  3,
  7,
  13,
  0xFE,
  0,89,1,
  3,
  0xFE,
  0,87,1,
  3,
  0xFE,
  0xFE,
  0,35,22,
  0,
  1,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  19,
  20,
  25,
  26,
  31,
  34,
  38,
  39,
  44,
  45,
  50,
  56,
  0,99,0,
  0,146,1,
  3,
  0xFE,
  0xFE,
  0,95,5,
  0,
  3,
  4,
  7,
  13,
  0xFE,
  0,91,1,
  3,
  0xFE,
  0,87,1,
  3,
  0xFE,
  0xFE,
  0,33,20,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  20,
  25,
  26,
  31,
  32,
  34,
  35,
  38,
  44,
  50,
  56,
  0,91,1,
  3,
  0xFE,
  0xFE,
  0,34,6,
  0,
  3,
  7,
  13,
  19,
  25,
  0xFE,
  0,36,5,
  3,
  7,
  13,
  19,
  25,
  0xFE,
  0,125,0,
  0,182,2,
  98,
  99,
  0xFE,
  0xFE,
  0,111,0,
  0,36,4,
  3,
  4,
  7,
  13,
  0xFE,
  0,38,1,
  3,
  0xFE,
  0,40,1,
  3,
  0xFE,
  0xFE,
  0,103,3,
  3,
  7,
  13,
  0xFE,
  0,101,9,
  0,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  0xFE,
  0,99,9,
  0,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  0xFE,
  0,97,12,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  71,
  77,
  0xFE,
  0,95,14,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  62,
  65,
  71,
  77,
  0xFE,
  0,93,17,
  0,
  1,
  3,
  4,
  5,
  7,
  8,
  13,
  14,
  19,
  20,
  25,
  26,
  62,
  65,
  71,
  77,
  0xFE,
  0,91,17,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  13,
  14,
  19,
  25,
  26,
  62,
  65,
  71,
  77,
  0xFE,
  0,89,18,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  20,
  25,
  26,
  62,
  65,
  71,
  72,
  77,
  78,
  0xFE,
  0,87,19,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  20,
  25,
  26,
  62,
  63,
  65,
  66,
  77,
  83,
  89,
  0xFE,
  0,88,5,
  3,
  7,
  13,
  19,
  25,
  0xFE,
  0,110,0,
  0,57,4,
  3,
  4,
  7,
  13,
  0xFE,
  0,55,1,
  3,
  0xFE,
  0,53,1,
  3,
  0xFE,
  0xFE,
  0,153,3,
  3,
  7,
  13,
  0xFE,
  0,151,7,
  0,
  3,
  4,
  7,
  8,
  13,
  14,
  0xFE,
  0,149,10,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  104,
  110,
  0xFE,
  0,147,12,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  95,
  98,
  104,
  110,
  0xFE,
  0,145,15,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  95,
  96,
  98,
  104,
  110,
  0xFE,
  0,143,14,
  0,
  3,
  7,
  13,
  14,
  19,
  25,
  95,
  96,
  98,
  104,
  105,
  110,
  111,
  0xFE,
  0,141,16,
  0,
  3,
  4,
  7,
  13,
  25,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  0xFE,
  0,140,6,
  95,
  98,
  104,
  110,
  116,
  122,
  0xFE,
  0,215,4,
  0,
  3,
  7,
  13,
  1,10,1,
  3,
  0xFE,
  0xFE,
  0,213,8,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  0xFE,
  0,211,12,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  13,
  14,
  19,
  25,
  0xFE,
  0,209,16,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  19,
  20,
  25,
  139,
  144,
  0xFE,
  0,207,20,
  0,
  1,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  19,
  20,
  25,
  26,
  127,
  133,
  139,
  140,
  144,
  145,
  1,1,5,
  0,
  3,
  4,
  7,
  13,
  0xFE,
  0,251,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,249,2,
  0,
  3,
  0xFE,
  0xFE,
  0,205,24,
  0,
  1,
  3,
  4,
  5,
  7,
  8,
  13,
  14,
  19,
  20,
  25,
  26,
  127,
  128,
  133,
  134,
  139,
  144,
  149,
  153,
  157,
  191,
  221,
  1,1,8,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  0xFE,
  0,251,5,
  0,
  3,
  4,
  7,
  13,
  0xFE,
  0,249,2,
  0,
  3,
  0xFE,
  0xFE,
  1,3,0,
  1,48,1,
  3,
  0xFE,
  0xFE,
  1,1,10,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  0xFE,
  0,255,7,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  0xFE,
  0,253,14,
  0,
  1,
  3,
  4,
  5,
  7,
  8,
  13,
  14,
  19,
  25,
  157,
  169,
  175,
  0xFE,
  0,251,15,
  0,
  3,
  7,
  8,
  13,
  14,
  19,
  25,
  157,
  158,
  163,
  169,
  170,
  175,
  176,
  0xFE,
  0,249,16,
  0,
  1,
  3,
  4,
  7,
  13,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  181,
  186,
  221,
  0xFE,
  1,49,5,
  0,
  3,
  191,
  203,
  208,
  0,211,0,
  1,2,1,
  3,
  0xFE,
  0xFE,
  1,94,2,
  203,
  208,
  0xFE,
  0xFE,
  1,47,10,
  0,
  3,
  7,
  13,
  191,
  197,
  203,
  204,
  208,
  209,
  1,94,2,
  191,
  197,
  0xFE,
  0xFE,
  1,45,10,
  191,
  192,
  197,
  198,
  203,
  208,
  213,
  217,
  233,
  239,
  1,95,2,
  203,
  208,
  0xFE,
  1,92,2,
  221,
  227,
  0xFE,
  0xFE,
  1,44,6,
  191,
  197,
  203,
  208,
  213,
  217,
  1,99,2,
  233,
  239,
  0xFE,
  1,97,2,
  221,
  227,
  0xFE,
  1,95,1,
  197,
  0xFE,
  0xFE,
  1,97,2,
  233,
  239,
  0xFE,
  1,95,10,
  0,
  3,
  191,
  197,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,93,8,
  221,
  222,
  227,
  228,
  233,
  239,
  245,
  249,
  0xFE,
  1,92,4,
  221,
  227,
  245,
  249,
  0xFE,
  0xFE,
  0,4,45,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  38,
  39,
  44,
  45,
  65,
  71,
  72,
  77,
  78,
  98,
  99,
  104,
  105,
  110,
  111,
  133,
  139,
  140,
  144,
  145,
  163,
  169,
  170,
  175,
  176,
  197,
  203,
  204,
  208,
  209,
  227,
  233,
  234,
  239,
  240,
  0,61,0,
  0,108,1,
  3,
  0xFE,
  0xFE,
  0,57,0,
  0,100,1,
  3,
  0xFE,
  0xFE,
  0,55,2,
  7,
  13,
  0,98,2,
  3,
  4,
  0xFE,
  0,100,1,
  3,
  0xFE,
  0xFE,
  0,53,6,
  0,
  3,
  7,
  8,
  13,
  14,
  0,94,1,
  3,
  0xFE,
  0xFE,
  0,51,10,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  0,90,3,
  3,
  7,
  13,
  0xFE,
  0,92,5,
  0,
  3,
  4,
  7,
  13,
  0xFE,
  0,94,1,
  3,
  0xFE,
  0xFE,
  0,49,15,
  0,
  1,
  3,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  20,
  25,
  26,
  38,
  44,
  0,88,5,
  0,
  3,
  4,
  7,
  13,
  0xFE,
  0,90,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,92,1,
  3,
  0xFE,
  0,146,1,
  3,
  0xFE,
  0xFE,
  0,47,25,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  20,
  25,
  26,
  31,
  32,
  34,
  35,
  38,
  39,
  44,
  45,
  56,
  0,88,3,
  3,
  7,
  13,
  0xFE,
  0,90,4,
  0,
  3,
  7,
  13,
  0xFE,
  0xFE,
  0,45,26,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  20,
  25,
  26,
  31,
  32,
  34,
  35,
  38,
  39,
  44,
  45,
  50,
  56,
  0,87,4,
  3,
  4,
  7,
  13,
  0xFE,
  0,88,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,90,3,
  0,
  3,
  13,
  0xFE,
  0,143,2,
  3,
  4,
  0xFE,
  0xFE,
  0,43,25,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  20,
  25,
  26,
  31,
  32,
  34,
  35,
  38,
  39,
  44,
  45,
  56,
  0,91,1,
  3,
  0xFE,
  0,88,3,
  0,
  3,
  13,
  0xFE,
  0,90,2,
  0,
  3,
  0xFE,
  0xFE,
  0,41,25,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  20,
  25,
  26,
  31,
  32,
  34,
  35,
  38,
  39,
  44,
  45,
  56,
  0,88,1,
  3,
  0xFE,
  0xFE,
  0,39,23,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  20,
  25,
  26,
  31,
  32,
  34,
  35,
  38,
  44,
  56,
  0,97,1,
  3,
  0xFE,
  0xFE,
  0,37,18,
  0,
  1,
  3,
  4,
  7,
  8,
  9,
  13,
  14,
  19,
  20,
  25,
  26,
  31,
  34,
  38,
  44,
  56,
  0xFE,
  0,35,18,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  13,
  14,
  19,
  20,
  25,
  26,
  31,
  34,
  38,
  44,
  0xFE,
  0,33,18,
  0,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  20,
  25,
  26,
  31,
  32,
  34,
  35,
  38,
  44,
  50,
  0xFE,
  0,105,1,
  13,
  0xFE,
  0,103,3,
  3,
  7,
  13,
  0xFE,
  0,101,7,
  0,
  3,
  4,
  7,
  13,
  14,
  25,
  0xFE,
  0,99,10,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  0xFE,
  0,97,12,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  71,
  77,
  0xFE,
  0,95,14,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  62,
  65,
  71,
  77,
  0xFE,
  0,93,15,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  26,
  62,
  65,
  71,
  77,
  0xFE,
  0,91,16,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  20,
  25,
  26,
  62,
  65,
  71,
  77,
  0xFE,
  0,89,18,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  20,
  25,
  26,
  62,
  65,
  71,
  72,
  77,
  78,
  0xFE,
  0,87,17,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  20,
  25,
  26,
  62,
  63,
  65,
  66,
  83,
  0xFE,
  0,153,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,151,5,
  0,
  3,
  4,
  7,
  13,
  0xFE,
  0,149,6,
  0,
  1,
  3,
  4,
  7,
  13,
  0xFE,
  0,147,10,
  0,
  1,
  3,
  4,
  7,
  13,
  95,
  98,
  104,
  110,
  0xFE,
  0,145,9,
  0,
  3,
  7,
  13,
  14,
  95,
  98,
  104,
  110,
  0xFE,
  0,143,11,
  0,
  3,
  4,
  7,
  13,
  95,
  98,
  104,
  105,
  110,
  111,
  0xFE,
  0,141,10,
  0,
  3,
  7,
  13,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0,215,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,213,7,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  0xFE,
  0,211,10,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  0xFE,
  0,209,13,
  0,
  1,
  3,
  4,
  5,
  7,
  8,
  13,
  14,
  19,
  25,
  139,
  144,
  0xFE,
  0,207,16,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  20,
  25,
  26,
  127,
  133,
  139,
  144,
  1,1,3,
  0,
  3,
  7,
  0xFE,
  0,251,2,
  0,
  3,
  0xFE,
  0xFE,
  0,205,16,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  127,
  128,
  133,
  134,
  144,
  221,
  1,1,7,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  0xFE,
  0,251,3,
  0,
  3,
  13,
  0xFE,
  0xFE,
  1,1,8,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  0xFE,
  0,255,5,
  0,
  1,
  3,
  4,
  13,
  0xFE,
  0,253,10,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  169,
  175,
  0xFE,
  0,251,14,
  0,
  1,
  3,
  4,
  7,
  13,
  19,
  25,
  157,
  163,
  169,
  170,
  175,
  176,
  0xFE,
  0,249,12,
  0,
  3,
  7,
  13,
  157,
  158,
  163,
  164,
  169,
  175,
  181,
  186,
  0xFE,
  1,49,2,
  203,
  208,
  0xFE,
  1,47,10,
  0,
  3,
  7,
  13,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,45,8,
  191,
  192,
  197,
  198,
  203,
  208,
  213,
  217,
  0xFE,
  1,44,2,
  213,
  217,
  0xFE,
  1,95,5,
  221,
  233,
  234,
  239,
  240,
  0xFE,
  1,93,6,
  221,
  222,
  227,
  228,
  245,
  249,
  0xFE,
  1,92,2,
  245,
  249,
  0xFE,
  0xFE,
  0,6,36,
  0,
  1,
  3,
  7,
  8,
  13,
  14,
  19,
  25,
  38,
  44,
  71,
  77,
  98,
  104,
  105,
  110,
  111,
  139,
  140,
  144,
  145,
  163,
  169,
  170,
  175,
  176,
  197,
  203,
  204,
  208,
  209,
  233,
  234,
  239,
  240,
  0,65,0,
  0,106,2,
  3,
  4,
  0xFE,
  0xFE,
  0,61,0,
  0,100,1,
  3,
  0xFE,
  0xFE,
  0,59,0,
  0,96,2,
  3,
  4,
  0xFE,
  0xFE,
  0,57,3,
  3,
  7,
  13,
  0,94,1,
  3,
  0xFE,
  0xFE,
  0,55,9,
  0,
  1,
  3,
  7,
  8,
  13,
  14,
  19,
  25,
  0xFE,
  0,53,17,
  0,
  1,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  20,
  25,
  26,
  38,
  44,
  0xFE,
  0,51,24,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  20,
  25,
  26,
  31,
  32,
  34,
  35,
  38,
  39,
  44,
  45,
  0xFE,
  0,49,22,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  20,
  25,
  26,
  31,
  34,
  35,
  38,
  44,
  45,
  0,89,4,
  3,
  4,
  7,
  13,
  0xFE,
  0xFE,
  0,47,24,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  20,
  25,
  26,
  31,
  32,
  34,
  35,
  38,
  39,
  44,
  45,
  0xFE,
  0,45,24,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  20,
  25,
  26,
  31,
  32,
  34,
  35,
  38,
  39,
  44,
  45,
  0xFE,
  0,43,21,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  13,
  14,
  15,
  19,
  20,
  25,
  26,
  31,
  32,
  34,
  35,
  38,
  44,
  0xFE,
  0,41,22,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  20,
  25,
  26,
  31,
  32,
  34,
  35,
  38,
  44,
  0xFE,
  0,39,16,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  20,
  25,
  26,
  31,
  34,
  38,
  44,
  0xFE,
  0,37,18,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  13,
  14,
  19,
  20,
  25,
  26,
  31,
  34,
  38,
  44,
  0xFE,
  0,35,16,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  20,
  25,
  26,
  31,
  34,
  38,
  44,
  0xFE,
  0,33,10,
  0,
  3,
  7,
  13,
  19,
  25,
  31,
  32,
  34,
  35,
  0xFE,
  0,105,3,
  3,
  7,
  13,
  0xFE,
  0,103,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,101,8,
  0,
  3,
  4,
  7,
  8,
  13,
  14,
  25,
  0xFE,
  0,99,10,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  0xFE,
  0,97,12,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  71,
  77,
  0xFE,
  0,95,14,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  62,
  65,
  71,
  77,
  0xFE,
  0,93,14,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  62,
  65,
  71,
  77,
  0xFE,
  0,91,12,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  62,
  65,
  0xFE,
  0,89,12,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  71,
  77,
  0xFE,
  0,87,10,
  0,
  3,
  7,
  13,
  19,
  25,
  62,
  63,
  65,
  66,
  0xFE,
  0,90,0,
  0,55,5,
  0,
  3,
  4,
  7,
  13,
  0xFE,
  0,53,1,
  3,
  0xFE,
  0,49,1,
  3,
  0xFE,
  0,47,1,
  3,
  0xFE,
  0xFE,
  0,153,2,
  0,
  3,
  0xFE,
  0,151,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,149,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,147,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,145,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,143,7,
  0,
  3,
  7,
  13,
  95,
  104,
  110,
  0xFE,
  0,141,4,
  95,
  96,
  98,
  99,
  0xFE,
  0,215,3,
  0,
  3,
  7,
  0xFE,
  0,213,5,
  0,
  3,
  4,
  7,
  13,
  0xFE,
  0,211,8,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  0xFE,
  0,209,10,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  0xFE,
  0,207,12,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  139,
  144,
  0xFE,
  0,205,13,
  0,
  1,
  3,
  4,
  7,
  13,
  14,
  19,
  25,
  127,
  128,
  133,
  134,
  0xFE,
  1,1,7,
  0,
  1,
  3,
  7,
  8,
  13,
  14,
  0,207,1,
  3,
  0xFE,
  0,205,4,
  0,
  3,
  7,
  13,
  0xFE,
  0xFE,
  0,255,3,
  0,
  3,
  13,
  0xFE,
  0,253,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,251,9,
  0,
  1,
  3,
  4,
  7,
  13,
  157,
  169,
  175,
  0xFE,
  0,249,4,
  157,
  158,
  163,
  164,
  0xFE,
  1,47,3,
  191,
  203,
  208,
  0xFE,
  1,45,4,
  191,
  192,
  197,
  198,
  0xFE,
  1,95,2,
  233,
  239,
  0xFE,
  1,93,4,
  221,
  222,
  227,
  228,
  0xFE,
  0xFE,
  0,8,23,
  0,
  1,
  3,
  7,
  8,
  13,
  14,
  19,
  25,
  38,
  44,
  71,
  77,
  104,
  110,
  139,
  144,
  169,
  175,
  203,
  208,
  233,
  239,
  0,75,0,
  0,114,3,
  3,
  34,
  65,
  0xFE,
  0xFE,
  0,61,2,
  7,
  13,
  0xFE,
  0,59,6,
  0,
  3,
  7,
  8,
  13,
  14,
  0xFE,
  0,57,19,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  20,
  25,
  26,
  34,
  38,
  44,
  0xFE,
  0,55,19,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  20,
  25,
  26,
  34,
  38,
  44,
  0xFE,
  0,53,18,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  25,
  26,
  34,
  38,
  44,
  0,91,1,
  3,
  0xFE,
  0xFE,
  0,51,19,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  25,
  26,
  31,
  34,
  38,
  44,
  0xFE,
  0,49,19,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  25,
  26,
  31,
  34,
  38,
  44,
  0xFE,
  0,47,19,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  13,
  14,
  15,
  19,
  25,
  26,
  31,
  32,
  34,
  38,
  44,
  0xFE,
  0,45,20,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  13,
  14,
  15,
  19,
  20,
  25,
  26,
  31,
  32,
  34,
  38,
  44,
  0xFE,
  0,43,15,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  26,
  31,
  34,
  38,
  44,
  0xFE,
  0,41,16,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  20,
  25,
  26,
  31,
  34,
  38,
  44,
  0xFE,
  0,39,15,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  20,
  25,
  26,
  31,
  34,
  44,
  0xFE,
  0,37,13,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  31,
  34,
  44,
  0xFE,
  0,35,11,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  38,
  0xFE,
  0,33,8,
  0,
  3,
  7,
  13,
  19,
  25,
  31,
  34,
  0xFE,
  0,105,3,
  3,
  7,
  13,
  0xFE,
  0,103,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,101,9,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  25,
  0xFE,
  0,99,8,
  0,
  1,
  3,
  4,
  7,
  13,
  14,
  25,
  0xFE,
  0,97,10,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  0xFE,
  0,95,10,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  0xFE,
  0,93,10,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  0xFE,
  0,91,10,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  0xFE,
  0,89,12,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  71,
  77,
  0xFE,
  0,87,8,
  0,
  3,
  7,
  13,
  19,
  25,
  62,
  65,
  0xFE,
  0,153,1,
  3,
  0xFE,
  0,151,2,
  0,
  3,
  0xFE,
  0,149,3,
  0,
  3,
  7,
  0xFE,
  0,147,3,
  0,
  3,
  13,
  0xFE,
  0,145,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,143,6,
  0,
  3,
  7,
  13,
  104,
  110,
  0xFE,
  0,141,3,
  95,
  96,
  98,
  0xFE,
  0,215,3,
  0,
  3,
  7,
  0xFE,
  0,213,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,211,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,209,8,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  0xFE,
  0,207,10,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  139,
  144,
  0xFE,
  0,205,6,
  0,
  3,
  7,
  13,
  127,
  133,
  0xFE,
  0,253,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,251,6,
  0,
  3,
  7,
  13,
  169,
  175,
  0xFE,
  0,249,2,
  157,
  163,
  0xFE,
  1,47,2,
  203,
  208,
  0xFE,
  1,45,2,
  191,
  197,
  0xFE,
  1,95,2,
  233,
  239,
  0xFE,
  1,93,4,
  221,
  222,
  227,
  228,
  0xFE,
  0xFE,
  0,10,9,
  0,
  7,
  13,
  169,
  175,
  203,
  208,
  233,
  239,
  0,85,0,
  0,122,1,
  3,
  0xFE,
  0xFE,
  0,69,0,
  0,96,1,
  3,
  0xFE,
  0xFE,
  0,63,7,
  0,
  3,
  4,
  7,
  8,
  13,
  14,
  0xFE,
  0,61,13,
  0,
  1,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  25,
  0xFE,
  0,59,11,
  0,
  1,
  3,
  4,
  5,
  7,
  8,
  13,
  14,
  19,
  25,
  0xFE,
  0,57,11,
  0,
  1,
  3,
  4,
  5,
  7,
  8,
  13,
  14,
  19,
  25,
  0xFE,
  0,55,11,
  0,
  1,
  3,
  4,
  5,
  7,
  8,
  13,
  14,
  19,
  25,
  0xFE,
  0,53,16,
  0,
  1,
  3,
  4,
  5,
  7,
  8,
  13,
  14,
  15,
  19,
  25,
  26,
  34,
  38,
  44,
  0xFE,
  0,51,18,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  13,
  14,
  15,
  19,
  25,
  26,
  31,
  34,
  38,
  44,
  0xFE,
  0,49,18,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  13,
  14,
  15,
  19,
  25,
  26,
  31,
  34,
  38,
  44,
  0xFE,
  0,47,17,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  13,
  14,
  19,
  25,
  26,
  31,
  34,
  38,
  44,
  0xFE,
  0,45,15,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  26,
  31,
  34,
  38,
  44,
  0xFE,
  0,43,14,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  31,
  34,
  38,
  44,
  0xFE,
  0,41,13,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  31,
  34,
  44,
  0xFE,
  0,39,12,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  31,
  34,
  0xFE,
  0,37,10,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  0xFE,
  0,35,10,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  0xFE,
  0,107,3,
  3,
  7,
  13,
  0xFE,
  0,105,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,103,5,
  0,
  3,
  4,
  7,
  13,
  0xFE,
  0,101,6,
  0,
  1,
  3,
  4,
  7,
  13,
  0xFE,
  0,99,7,
  0,
  1,
  3,
  4,
  7,
  13,
  19,
  0xFE,
  0,97,6,
  0,
  3,
  4,
  7,
  13,
  14,
  0xFE,
  0,95,9,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  19,
  25,
  0xFE,
  0,93,7,
  0,
  1,
  3,
  4,
  7,
  13,
  14,
  0xFE,
  0,91,10,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  0xFE,
  0,89,7,
  0,
  1,
  3,
  7,
  8,
  13,
  14,
  0xFE,
  0,87,1,
  62,
  0xFE,
  0,141,2,
  95,
  98,
  0xFE,
  0,213,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,211,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,209,7,
  0,
  1,
  3,
  7,
  8,
  13,
  14,
  0xFE,
  0,207,8,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  0xFE,
  0,205,2,
  127,
  133,
  0xFE,
  0,251,3,
  0,
  7,
  13,
  0xFE,
  0,249,2,
  157,
  163,
  0xFE,
  1,45,2,
  191,
  197,
  0xFE,
  1,93,2,
  221,
  227,
  0xFE,
  0xFE,
  0,12,2,
  7,
  13,
  0,67,9,
  0,
  3,
  4,
  5,
  7,
  8,
  13,
  14,
  25,
  0xFE,
  0,65,7,
  0,
  3,
  4,
  7,
  8,
  13,
  14,
  0xFE,
  0,63,7,
  0,
  3,
  4,
  7,
  8,
  13,
  14,
  0xFE,
  0,61,8,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  0xFE,
  0,59,10,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  0xFE,
  0,57,10,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  0xFE,
  0,55,12,
  0,
  1,
  3,
  4,
  5,
  7,
  8,
  13,
  14,
  19,
  25,
  44,
  0xFE,
  0,53,16,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  13,
  14,
  19,
  25,
  31,
  34,
  38,
  44,
  0xFE,
  0,51,16,
  0,
  1,
  2,
  3,
  4,
  5,
  7,
  8,
  13,
  14,
  19,
  25,
  31,
  34,
  38,
  44,
  0xFE,
  0,49,14,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  31,
  34,
  38,
  44,
  0xFE,
  0,47,12,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  31,
  34,
  0xFE,
  0,45,11,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  38,
  0xFE,
  0,43,12,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  31,
  34,
  0xFE,
  0,41,11,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  34,
  0xFE,
  0,39,10,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  0xFE,
  0,37,10,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  0xFE,
  0,35,6,
  0,
  3,
  7,
  8,
  13,
  14,
  0xFE,
  0,107,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,105,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,103,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,101,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,99,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,97,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,95,6,
  0,
  1,
  3,
  4,
  7,
  13,
  0xFE,
  0,93,7,
  0,
  1,
  3,
  4,
  7,
  13,
  14,
  0xFE,
  0,91,6,
  0,
  1,
  3,
  7,
  13,
  14,
  0xFE,
  0,89,3,
  0,
  7,
  13,
  0xFE,
  0,211,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,209,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,207,6,
  0,
  3,
  7,
  8,
  13,
  14,
  0xFE,
  0xFE,
  0,14,0,
  0,71,2,
  3,
  4,
  0xFE,
  0,69,3,
  3,
  7,
  13,
  0xFE,
  0,67,3,
  3,
  7,
  13,
  0xFE,
  0,65,7,
  0,
  3,
  4,
  7,
  8,
  13,
  14,
  0xFE,
  0,63,7,
  0,
  3,
  4,
  7,
  8,
  13,
  14,
  0xFE,
  0,61,8,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  0xFE,
  0,59,9,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  25,
  0xFE,
  0,57,9,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  25,
  0xFE,
  0,55,10,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  25,
  44,
  0xFE,
  0,53,13,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  31,
  34,
  44,
  0xFE,
  0,51,12,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  31,
  34,
  0xFE,
  0,49,10,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  0xFE,
  0,47,9,
  0,
  1,
  3,
  4,
  7,
  13,
  14,
  19,
  25,
  0xFE,
  0,45,10,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  0xFE,
  0,43,9,
  0,
  1,
  3,
  4,
  7,
  13,
  14,
  19,
  25,
  0xFE,
  0,41,10,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  0xFE,
  0,39,8,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  0xFE,
  0,37,6,
  0,
  3,
  7,
  8,
  13,
  14,
  0xFE,
  0,35,3,
  0,
  7,
  13,
  0xFE,
  0,107,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,105,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,103,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,101,3,
  3,
  7,
  13,
  0xFE,
  0,99,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,97,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,95,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,93,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,91,3,
  0,
  7,
  13,
  0xFE,
  0,89,2,
  7,
  13,
  0xFE,
  0,209,3,
  0,
  7,
  13,
  0xFE,
  0,207,3,
  0,
  7,
  13,
  0xFE,
  0xFE,
  0,16,0,
  0,77,9,
  0,
  3,
  4,
  5,
  7,
  8,
  13,
  14,
  34,
  0xFE,
  0,69,3,
  3,
  7,
  13,
  0xFE,
  0,67,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,65,7,
  0,
  3,
  4,
  7,
  8,
  13,
  14,
  0xFE,
  0,63,8,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  0xFE,
  0,61,9,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  25,
  0xFE,
  0,59,9,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  25,
  0xFE,
  0,57,9,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  25,
  0xFE,
  0,55,10,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  0xFE,
  0,53,10,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  0xFE,
  0,51,10,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  0xFE,
  0,49,7,
  0,
  3,
  7,
  13,
  14,
  19,
  25,
  0xFE,
  0,47,9,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  0xFE,
  0,45,7,
  0,
  1,
  3,
  4,
  7,
  13,
  14,
  0xFE,
  0,43,8,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  0xFE,
  0,41,8,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  0xFE,
  0,39,5,
  0,
  3,
  7,
  13,
  14,
  0xFE,
  0,37,3,
  0,
  7,
  13,
  0xFE,
  0,109,1,
  3,
  0xFE,
  0,107,2,
  0,
  3,
  0xFE,
  0,105,3,
  0,
  3,
  7,
  0xFE,
  0,103,2,
  3,
  7,
  0xFE,
  0,101,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,99,3,
  0,
  3,
  13,
  0xFE,
  0,97,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,95,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,93,3,
  0,
  7,
  13,
  0xFE,
  0xFE,
  0,18,0,
  0,69,3,
  3,
  7,
  13,
  0xFE,
  0,67,5,
  0,
  3,
  4,
  7,
  13,
  0xFE,
  0,65,6,
  0,
  3,
  4,
  7,
  13,
  14,
  0xFE,
  0,63,8,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  0xFE,
  0,61,9,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  25,
  0xFE,
  0,59,9,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  25,
  0xFE,
  0,57,8,
  0,
  1,
  3,
  4,
  7,
  13,
  14,
  25,
  0xFE,
  0,55,9,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  25,
  0xFE,
  0,53,6,
  0,
  3,
  7,
  8,
  13,
  25,
  0xFE,
  0,51,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,49,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,47,7,
  0,
  1,
  3,
  4,
  7,
  13,
  14,
  0xFE,
  0,45,7,
  0,
  1,
  3,
  4,
  7,
  13,
  14,
  0xFE,
  0,43,7,
  0,
  1,
  3,
  4,
  7,
  13,
  14,
  0xFE,
  0,41,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,39,3,
  0,
  7,
  13,
  0xFE,
  0,103,2,
  0,
  3,
  0xFE,
  0,101,3,
  0,
  3,
  13,
  0xFE,
  0,99,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,97,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,95,2,
  7,
  13,
  0xFE,
  0xFE,
  0,20,0,
  0,71,3,
  3,
  7,
  13,
  0xFE,
  0,69,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,67,5,
  0,
  3,
  4,
  7,
  13,
  0xFE,
  0,65,6,
  0,
  3,
  4,
  7,
  13,
  14,
  0xFE,
  0,63,8,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  0xFE,
  0,61,7,
  0,
  1,
  3,
  4,
  7,
  13,
  14,
  0xFE,
  0,59,7,
  0,
  1,
  3,
  4,
  7,
  13,
  14,
  0xFE,
  0,57,6,
  0,
  3,
  4,
  7,
  8,
  13,
  0xFE,
  0,55,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,53,3,
  0,
  3,
  7,
  0xFE,
  0,51,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,49,6,
  0,
  1,
  3,
  4,
  7,
  13,
  0xFE,
  0,47,6,
  0,
  1,
  3,
  4,
  7,
  13,
  0xFE,
  0,45,6,
  0,
  1,
  3,
  4,
  7,
  13,
  0xFE,
  0,43,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,41,3,
  0,
  7,
  13,
  0xFE,
  0,103,1,
  0,
  0xFE,
  0,101,2,
  0,
  3,
  0xFE,
  0xFE,
  0,22,0,
  0,71,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,69,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,67,5,
  0,
  3,
  4,
  7,
  13,
  0xFE,
  0,65,5,
  0,
  3,
  4,
  7,
  13,
  0xFE,
  0,63,6,
  0,
  1,
  3,
  4,
  7,
  13,
  0xFE,
  0,61,6,
  0,
  1,
  3,
  4,
  7,
  13,
  0xFE,
  0,59,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,57,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,55,3,
  0,
  3,
  7,
  0xFE,
  0,53,3,
  0,
  7,
  13,
  0xFE,
  0,51,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,49,5,
  0,
  1,
  3,
  7,
  13,
  0xFE,
  0,47,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,45,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,43,3,
  0,
  7,
  13,
  0xFE,
  0xFE,
  0,24,0,
  0,73,2,
  3,
  13,
  0xFE,
  0,71,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,69,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,67,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,65,5,
  0,
  3,
  4,
  7,
  13,
  0xFE,
  0,63,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,61,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,59,3,
  0,
  3,
  7,
  0xFE,
  0,57,3,
  0,
  3,
  7,
  0xFE,
  0,55,3,
  0,
  3,
  7,
  0xFE,
  0,53,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,51,3,
  0,
  3,
  13,
  0xFE,
  0,49,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,47,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,45,1,
  13,
  0xFE,
  0xFE,
  0,26,0,
  0,73,1,
  3,
  0xFE,
  0,71,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,69,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,67,3,
  0,
  3,
  13,
  0xFE,
  0,65,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,63,1,
  7,
  0xFE,
  0,61,2,
  0,
  3,
  0xFE,
  0,59,3,
  0,
  3,
  7,
  0xFE,
  0,57,2,
  0,
  3,
  0xFE,
  0,55,3,
  0,
  3,
  7,
  0xFE,
  0,53,3,
  0,
  3,
  13,
  0xFE,
  0,51,3,
  0,
  3,
  13,
  0xFE,
  0,49,4,
  0,
  3,
  7,
  13,
  0xFE,
  0xFE,
  0,28,0,
  0,73,1,
  3,
  0xFE,
  0,71,3,
  0,
  3,
  13,
  0xFE,
  0,69,2,
  0,
  3,
  0xFE,
  0,67,1,
  3,
  0xFE,
  0,63,1,
  3,
  0xFE,
  0,55,2,
  0,
  3,
  0xFE,
  0,53,2,
  0,
  3,
  0xFE,
  0,51,1,
  0,
  0xFE,
  0xFE,
  0,30,0,
  0,71,1,
  3,
  0xFE,
  0xFE,
  0,32,87,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  19,
  20,
  21,
  25,
  26,
  27,
  31,
  32,
  34,
  35,
  36,
  38,
  39,
  40,
  44,
  45,
  46,
  50,
  51,
  56,
  57,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  84,
  89,
  90,
  98,
  99,
  100,
  104,
  105,
  110,
  111,
  116,
  117,
  122,
  123,
  133,
  134,
  135,
  139,
  140,
  144,
  149,
  153,
  157,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  182,
  186,
  187,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  214,
  217,
  227,
  228,
  233,
  234,
  239,
  240,
  245,
  249,
  0,0,20,
  0,
  1,
  3,
  4,
  5,
  7,
  8,
  13,
  14,
  19,
  20,
  25,
  26,
  31,
  34,
  35,
  38,
  44,
  50,
  56,
  0,97,1,
  3,
  0xFE,
  0,95,1,
  3,
  0xFE,
  0xFE,
  0,2,14,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  20,
  25,
  26,
  50,
  56,
  0xFE,
  0,4,4,
  3,
  13,
  19,
  25,
  0xFE,
  0,101,0,
  0,162,11,
  34,
  35,
  65,
  66,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  0xFE,
  0xFE,
  0,99,0,
  0,158,3,
  65,
  98,
  99,
  0xFE,
  0xFE,
  0,97,0,
  0,154,1,
  98,
  0xFE,
  0xFE,
  0,95,8,
  34,
  38,
  44,
  50,
  56,
  71,
  77,
  89,
  0,150,1,
  98,
  0xFE,
  0xFE,
  0,93,18,
  31,
  34,
  35,
  38,
  39,
  44,
  45,
  50,
  56,
  62,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  89,
  0,146,1,
  98,
  0xFE,
  0xFE,
  0,91,33,
  31,
  32,
  33,
  34,
  35,
  36,
  38,
  39,
  40,
  44,
  45,
  46,
  50,
  51,
  56,
  57,
  58,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  84,
  89,
  90,
  0,142,3,
  34,
  95,
  98,
  0xFE,
  0,144,3,
  34,
  65,
  98,
  0xFE,
  0xFE,
  0,89,34,
  3,
  7,
  13,
  31,
  32,
  33,
  34,
  35,
  36,
  38,
  39,
  40,
  44,
  45,
  46,
  50,
  51,
  56,
  57,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  77,
  78,
  83,
  84,
  89,
  90,
  98,
  0,140,3,
  98,
  104,
  110,
  0xFE,
  0xFE,
  0,87,29,
  31,
  32,
  34,
  35,
  38,
  39,
  40,
  44,
  45,
  50,
  51,
  52,
  56,
  57,
  62,
  63,
  65,
  66,
  71,
  72,
  73,
  77,
  78,
  83,
  84,
  85,
  89,
  90,
  221,
  0,141,2,
  98,
  104,
  0xFE,
  0xFE,
  0,88,16,
  34,
  38,
  39,
  44,
  45,
  50,
  56,
  62,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  89,
  0,153,4,
  98,
  99,
  104,
  110,
  0xFE,
  0,151,1,
  98,
  0xFE,
  0xFE,
  0,90,6,
  38,
  44,
  50,
  56,
  71,
  77,
  0,157,1,
  98,
  0xFE,
  0,155,1,
  98,
  0xFE,
  0xFE,
  0,102,0,
  0,179,4,
  65,
  66,
  98,
  99,
  0xFE,
  0xFE,
  0,151,6,
  95,
  96,
  104,
  105,
  110,
  111,
  0xFE,
  0,149,16,
  38,
  39,
  44,
  45,
  95,
  96,
  97,
  98,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  0xFE,
  0,147,21,
  31,
  34,
  38,
  39,
  44,
  45,
  50,
  56,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  0,218,0,
  1,1,1,
  98,
  0xFE,
  0xFE,
  0xFE,
  0,145,37,
  3,
  7,
  13,
  31,
  32,
  33,
  34,
  35,
  36,
  38,
  39,
  40,
  44,
  45,
  46,
  50,
  51,
  56,
  57,
  95,
  96,
  97,
  98,
  99,
  100,
  101,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  118,
  122,
  123,
  0xFE,
  0,143,20,
  31,
  32,
  34,
  35,
  38,
  39,
  44,
  50,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  110,
  111,
  116,
  122,
  0xFE,
  0,141,20,
  31,
  34,
  38,
  39,
  44,
  45,
  50,
  95,
  96,
  98,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  0xFE,
  0,140,15,
  34,
  38,
  44,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  117,
  122,
  123,
  0xFE,
  0,142,5,
  38,
  44,
  98,
  104,
  110,
  0xFE,
  0,215,2,
  38,
  44,
  0xFE,
  0,213,7,
  34,
  38,
  39,
  44,
  45,
  139,
  144,
  0xFE,
  0,211,17,
  31,
  32,
  34,
  35,
  38,
  39,
  40,
  44,
  45,
  46,
  50,
  56,
  127,
  139,
  140,
  144,
  145,
  1,8,1,
  163,
  0xFE,
  0xFE,
  0,209,23,
  31,
  32,
  34,
  35,
  38,
  39,
  40,
  44,
  45,
  46,
  50,
  51,
  56,
  57,
  95,
  98,
  127,
  128,
  133,
  139,
  140,
  144,
  145,
  0,97,0,
  0,154,2,
  38,
  44,
  0xFE,
  0xFE,
  0,93,0,
  0,148,6,
  34,
  38,
  44,
  98,
  104,
  110,
  0xFE,
  0,150,2,
  104,
  110,
  0xFE,
  0,152,2,
  95,
  98,
  0xFE,
  0xFE,
  0,141,2,
  104,
  110,
  0xFE,
  0xFE,
  0,207,28,
  31,
  32,
  33,
  34,
  35,
  36,
  38,
  39,
  40,
  44,
  45,
  46,
  50,
  56,
  127,
  128,
  129,
  133,
  134,
  139,
  140,
  144,
  145,
  149,
  153,
  157,
  191,
  221,
  0,143,3,
  98,
  104,
  110,
  0xFE,
  0,141,3,
  95,
  96,
  98,
  0xFE,
  0,140,1,
  95,
  0xFE,
  1,1,4,
  38,
  39,
  44,
  45,
  0xFE,
  0,251,2,
  38,
  44,
  0xFE,
  0,249,2,
  38,
  44,
  0xFE,
  1,2,0,
  1,53,2,
  203,
  208,
  0xFE,
  1,51,2,
  191,
  197,
  0xFE,
  0xFE,
  1,93,2,
  233,
  239,
  0xFE,
  1,92,3,
  221,
  233,
  239,
  0xFE,
  0xFE,
  0,205,31,
  31,
  34,
  38,
  39,
  44,
  45,
  50,
  56,
  57,
  127,
  128,
  133,
  139,
  140,
  144,
  145,
  149,
  150,
  153,
  154,
  157,
  163,
  169,
  170,
  175,
  176,
  191,
  197,
  203,
  221,
  227,
  0,145,1,
  98,
  0xFE,
  1,1,7,
  34,
  38,
  39,
  44,
  45,
  139,
  144,
  0xFE,
  0,251,4,
  38,
  44,
  139,
  144,
  0xFE,
  0,249,7,
  38,
  44,
  139,
  144,
  157,
  169,
  175,
  0xFE,
  1,45,3,
  191,
  203,
  208,
  0xFE,
  1,44,3,
  191,
  203,
  208,
  0xFE,
  1,93,3,
  221,
  233,
  239,
  0xFE,
  1,92,3,
  221,
  233,
  239,
  0xFE,
  0xFE,
  1,1,16,
  31,
  34,
  35,
  38,
  39,
  44,
  45,
  127,
  128,
  133,
  139,
  140,
  144,
  145,
  169,
  175,
  0xFE,
  0,255,10,
  38,
  39,
  44,
  45,
  157,
  158,
  169,
  170,
  175,
  176,
  0xFE,
  0,253,19,
  31,
  34,
  38,
  39,
  44,
  45,
  139,
  144,
  157,
  158,
  159,
  163,
  164,
  169,
  170,
  171,
  175,
  176,
  177,
  1,48,1,
  157,
  0xFE,
  0xFE,
  0,251,24,
  31,
  32,
  34,
  35,
  38,
  44,
  127,
  133,
  139,
  144,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  181,
  186,
  191,
  221,
  0,103,0,
  0,166,1,
  98,
  0xFE,
  0xFE,
  1,48,2,
  157,
  163,
  0xFE,
  0xFE,
  0,249,27,
  31,
  34,
  38,
  39,
  44,
  45,
  127,
  133,
  157,
  158,
  163,
  164,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  182,
  186,
  187,
  191,
  197,
  203,
  208,
  221,
  0,145,4,
  34,
  35,
  98,
  99,
  0xFE,
  1,93,2,
  233,
  239,
  0xFE,
  1,92,2,
  233,
  239,
  0xFE,
  0xFE,
  1,53,2,
  203,
  208,
  1,94,2,
  203,
  208,
  0xFE,
  1,98,2,
  203,
  208,
  0xFE,
  0xFE,
  1,51,3,
  191,
  203,
  208,
  1,94,3,
  191,
  203,
  208,
  0xFE,
  1,96,1,
  191,
  0xFE,
  1,98,4,
  191,
  197,
  203,
  208,
  0xFE,
  0xFE,
  1,49,11,
  38,
  44,
  157,
  163,
  191,
  192,
  197,
  203,
  204,
  208,
  209,
  1,94,6,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,98,2,
  191,
  197,
  0xFE,
  0xFE,
  1,47,10,
  38,
  44,
  191,
  192,
  197,
  198,
  203,
  208,
  213,
  217,
  1,95,3,
  191,
  203,
  208,
  0xFE,
  1,92,5,
  191,
  197,
  221,
  233,
  239,
  0xFE,
  1,94,12,
  191,
  192,
  197,
  198,
  203,
  208,
  213,
  217,
  221,
  227,
  233,
  239,
  0xFE,
  1,96,2,
  221,
  227,
  0xFE,
  0xFE,
  1,45,16,
  191,
  197,
  203,
  204,
  208,
  209,
  213,
  214,
  217,
  218,
  221,
  227,
  233,
  234,
  239,
  240,
  0,249,3,
  191,
  203,
  208,
  0xFE,
  1,95,6,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,93,2,
  233,
  239,
  0xFE,
  1,92,8,
  221,
  222,
  227,
  228,
  233,
  239,
  245,
  249,
  0xFE,
  1,94,4,
  203,
  208,
  233,
  239,
  0xFE,
  1,96,2,
  245,
  249,
  0xFE,
  0xFE,
  1,44,12,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  214,
  217,
  218,
  0,251,1,
  191,
  0xFE,
  0,249,5,
  191,
  192,
  197,
  203,
  208,
  0xFE,
  1,101,2,
  233,
  239,
  0xFE,
  1,99,10,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,97,10,
  191,
  197,
  221,
  222,
  227,
  228,
  233,
  239,
  245,
  249,
  0xFE,
  1,95,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,93,4,
  233,
  239,
  245,
  249,
  0xFE,
  0xFE,
  1,46,2,
  213,
  217,
  0,249,4,
  191,
  192,
  197,
  198,
  0xFE,
  1,103,2,
  233,
  239,
  0xFE,
  1,101,4,
  221,
  227,
  233,
  239,
  0xFE,
  1,99,3,
  197,
  221,
  227,
  0xFE,
  1,97,2,
  191,
  221,
  0xFE,
  1,95,4,
  203,
  208,
  213,
  217,
  0xFE,
  0xFE,
  1,99,4,
  203,
  208,
  233,
  239,
  1,48,1,
  221,
  0xFE,
  0xFE,
  1,97,3,
  221,
  233,
  239,
  0xFE,
  1,95,15,
  38,
  44,
  191,
  192,
  197,
  198,
  203,
  208,
  213,
  217,
  221,
  222,
  227,
  233,
  239,
  0xFE,
  1,93,13,
  191,
  197,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  245,
  246,
  249,
  250,
  0xFE,
  1,92,10,
  221,
  222,
  227,
  228,
  233,
  239,
  245,
  246,
  249,
  250,
  0xFE,
  1,94,6,
  203,
  208,
  213,
  217,
  233,
  239,
  0xFE,
  0xFE,
  0,34,68,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  20,
  25,
  26,
  27,
  31,
  32,
  33,
  34,
  35,
  36,
  38,
  39,
  44,
  45,
  50,
  51,
  56,
  57,
  65,
  66,
  67,
  71,
  72,
  77,
  78,
  83,
  84,
  89,
  90,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  133,
  139,
  140,
  144,
  145,
  163,
  164,
  169,
  170,
  175,
  176,
  197,
  203,
  204,
  208,
  209,
  227,
  233,
  234,
  239,
  240,
  0,117,0,
  0,186,2,
  98,
  99,
  0xFE,
  0xFE,
  0,105,0,
  0,164,1,
  98,
  0xFE,
  0xFE,
  0,103,0,
  0,160,3,
  34,
  98,
  99,
  0xFE,
  0xFE,
  0,101,0,
  0,154,1,
  98,
  0xFE,
  0,156,10,
  34,
  35,
  65,
  66,
  95,
  96,
  98,
  99,
  104,
  110,
  0,249,1,
  98,
  0xFE,
  0xFE,
  0xFE,
  0,99,0,
  0,150,1,
  98,
  0xFE,
  0,152,12,
  34,
  35,
  65,
  66,
  71,
  77,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0xFE,
  0,97,9,
  34,
  38,
  44,
  50,
  56,
  71,
  77,
  83,
  89,
  0,148,7,
  34,
  65,
  95,
  98,
  99,
  104,
  110,
  0xFE,
  0xFE,
  0,95,20,
  31,
  34,
  35,
  38,
  39,
  40,
  44,
  45,
  50,
  56,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  89,
  0,144,4,
  34,
  65,
  95,
  98,
  0xFE,
  0,146,2,
  34,
  95,
  0xFE,
  0xFE,
  0,93,33,
  31,
  32,
  33,
  34,
  35,
  38,
  39,
  40,
  44,
  45,
  46,
  50,
  51,
  52,
  56,
  57,
  58,
  62,
  63,
  64,
  65,
  66,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  84,
  89,
  90,
  91,
  0xFE,
  0,91,39,
  3,
  7,
  13,
  31,
  32,
  33,
  34,
  35,
  36,
  38,
  39,
  40,
  44,
  45,
  46,
  50,
  51,
  52,
  56,
  57,
  58,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  84,
  85,
  89,
  90,
  91,
  0xFE,
  0,89,29,
  3,
  31,
  32,
  33,
  34,
  35,
  36,
  38,
  39,
  44,
  45,
  50,
  51,
  52,
  56,
  57,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  84,
  85,
  89,
  90,
  0,143,2,
  38,
  44,
  0xFE,
  0xFE,
  0,87,24,
  31,
  32,
  34,
  35,
  38,
  39,
  44,
  45,
  50,
  51,
  56,
  57,
  62,
  63,
  64,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  84,
  89,
  0xFE,
  0,155,5,
  95,
  104,
  105,
  110,
  111,
  0xFE,
  0,153,16,
  34,
  38,
  39,
  44,
  45,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  106,
  110,
  111,
  112,
  0xFE,
  0,151,31,
  31,
  32,
  33,
  34,
  35,
  36,
  38,
  39,
  40,
  44,
  45,
  46,
  50,
  56,
  95,
  96,
  97,
  98,
  99,
  100,
  101,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  0,251,1,
  98,
  0xFE,
  0xFE,
  0,149,25,
  31,
  32,
  33,
  34,
  35,
  36,
  38,
  39,
  44,
  45,
  50,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  0xFE,
  0,147,24,
  31,
  32,
  33,
  34,
  35,
  36,
  38,
  39,
  44,
  45,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  0xFE,
  0,145,19,
  31,
  32,
  33,
  34,
  35,
  38,
  44,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  110,
  111,
  116,
  122,
  0xFE,
  0,143,14,
  31,
  34,
  38,
  44,
  50,
  95,
  96,
  98,
  104,
  105,
  110,
  111,
  116,
  122,
  0xFE,
  0,141,14,
  31,
  32,
  34,
  38,
  44,
  95,
  96,
  97,
  98,
  99,
  104,
  110,
  116,
  122,
  0xFE,
  0,215,4,
  31,
  34,
  38,
  44,
  1,10,1,
  34,
  0xFE,
  0xFE,
  0,213,9,
  31,
  32,
  33,
  34,
  35,
  38,
  39,
  44,
  45,
  0xFE,
  0,211,13,
  31,
  32,
  33,
  34,
  35,
  36,
  38,
  39,
  44,
  45,
  50,
  139,
  144,
  0xFE,
  0,209,15,
  31,
  32,
  33,
  34,
  35,
  36,
  38,
  39,
  44,
  50,
  56,
  127,
  133,
  139,
  144,
  0,91,4,
  34,
  35,
  65,
  66,
  0xFE,
  0,154,0,
  0,99,6,
  34,
  38,
  44,
  65,
  71,
  77,
  0xFE,
  0,97,2,
  34,
  65,
  0xFE,
  0xFE,
  0xFE,
  0,207,20,
  31,
  32,
  33,
  34,
  35,
  36,
  38,
  39,
  44,
  45,
  50,
  51,
  56,
  57,
  127,
  133,
  139,
  140,
  144,
  145,
  0,105,0,
  0,164,1,
  98,
  0xFE,
  0xFE,
  1,1,8,
  31,
  32,
  34,
  35,
  38,
  39,
  44,
  45,
  0xFE,
  0,251,4,
  31,
  34,
  38,
  44,
  0xFE,
  0,249,1,
  34,
  0xFE,
  0xFE,
  0,205,17,
  31,
  32,
  33,
  34,
  35,
  38,
  44,
  50,
  56,
  127,
  128,
  133,
  134,
  139,
  144,
  191,
  221,
  0,91,2,
  34,
  65,
  0xFE,
  0,151,1,
  98,
  0xFE,
  1,1,8,
  31,
  32,
  34,
  35,
  38,
  39,
  44,
  45,
  0xFE,
  0,251,2,
  31,
  34,
  0xFE,
  0,249,2,
  31,
  34,
  0xFE,
  0xFE,
  1,1,6,
  31,
  32,
  34,
  35,
  38,
  44,
  0xFE,
  0,255,9,
  31,
  32,
  33,
  34,
  35,
  38,
  39,
  44,
  45,
  0xFE,
  0,253,9,
  31,
  32,
  33,
  34,
  35,
  38,
  44,
  169,
  175,
  0xFE,
  0,251,13,
  31,
  34,
  38,
  44,
  50,
  56,
  157,
  158,
  163,
  169,
  170,
  175,
  176,
  0xFE,
  0,249,12,
  31,
  32,
  34,
  35,
  38,
  44,
  157,
  158,
  163,
  164,
  169,
  175,
  0xFE,
  1,49,4,
  31,
  34,
  203,
  208,
  0xFE,
  1,47,8,
  31,
  34,
  191,
  197,
  203,
  204,
  208,
  209,
  1,94,2,
  191,
  197,
  0xFE,
  0xFE,
  1,45,8,
  191,
  192,
  197,
  198,
  203,
  208,
  213,
  217,
  1,92,2,
  221,
  227,
  0xFE,
  0xFE,
  1,44,4,
  203,
  208,
  213,
  217,
  1,97,2,
  221,
  227,
  0xFE,
  0xFE,
  1,95,9,
  31,
  34,
  191,
  197,
  221,
  233,
  234,
  239,
  240,
  0xFE,
  1,93,6,
  221,
  222,
  227,
  228,
  245,
  249,
  0xFE,
  1,92,2,
  245,
  249,
  0xFE,
  0xFE,
  0,36,32,
  0,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  20,
  25,
  26,
  31,
  32,
  34,
  38,
  39,
  44,
  45,
  50,
  56,
  65,
  71,
  72,
  77,
  78,
  83,
  89,
  233,
  239,
  0,103,0,
  0,154,1,
  98,
  0xFE,
  0xFE,
  0,101,0,
  0,150,4,
  34,
  65,
  98,
  104,
  0,251,1,
  98,
  0xFE,
  0xFE,
  0xFE,
  0,99,10,
  34,
  38,
  44,
  50,
  56,
  65,
  71,
  77,
  83,
  89,
  0,146,21,
  31,
  34,
  35,
  38,
  39,
  44,
  45,
  62,
  65,
  66,
  71,
  72,
  77,
  78,
  95,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0xFE,
  0,97,23,
  31,
  32,
  34,
  38,
  39,
  40,
  44,
  45,
  46,
  50,
  56,
  62,
  63,
  65,
  66,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  89,
  0xFE,
  0,95,37,
  3,
  31,
  32,
  33,
  34,
  35,
  36,
  38,
  39,
  40,
  44,
  45,
  46,
  50,
  51,
  52,
  56,
  57,
  58,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  84,
  85,
  89,
  90,
  91,
  0xFE,
  0,93,29,
  31,
  32,
  33,
  34,
  35,
  36,
  38,
  39,
  40,
  44,
  45,
  50,
  51,
  56,
  57,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  77,
  78,
  83,
  84,
  89,
  90,
  0xFE,
  0,91,26,
  3,
  31,
  32,
  33,
  34,
  35,
  38,
  39,
  44,
  45,
  50,
  51,
  56,
  57,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  84,
  89,
  90,
  0xFE,
  0,89,22,
  31,
  32,
  34,
  35,
  38,
  39,
  44,
  45,
  50,
  51,
  56,
  57,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  89,
  0xFE,
  0,87,25,
  31,
  32,
  33,
  34,
  35,
  36,
  38,
  39,
  44,
  45,
  50,
  51,
  56,
  57,
  62,
  63,
  65,
  66,
  67,
  71,
  77,
  83,
  84,
  89,
  90,
  0xFE,
  0,159,4,
  104,
  105,
  110,
  111,
  0xFE,
  0,157,21,
  31,
  32,
  34,
  35,
  36,
  38,
  39,
  44,
  45,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  0xFE,
  0,155,18,
  34,
  35,
  38,
  39,
  44,
  45,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  0xFE,
  0,153,20,
  31,
  32,
  34,
  35,
  38,
  39,
  44,
  45,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  0xFE,
  0,151,22,
  31,
  32,
  34,
  35,
  38,
  39,
  44,
  45,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  0xFE,
  0,149,22,
  31,
  32,
  34,
  35,
  38,
  39,
  44,
  45,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  0xFE,
  0,147,18,
  31,
  32,
  34,
  35,
  38,
  44,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  110,
  111,
  116,
  122,
  0xFE,
  0,145,12,
  31,
  34,
  38,
  44,
  95,
  98,
  104,
  105,
  110,
  111,
  116,
  122,
  0xFE,
  0,143,11,
  31,
  32,
  34,
  38,
  44,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0,141,7,
  38,
  44,
  98,
  104,
  110,
  116,
  122,
  0xFE,
  0,217,3,
  34,
  38,
  44,
  0xFE,
  0,215,6,
  31,
  32,
  34,
  35,
  38,
  44,
  0xFE,
  0,213,9,
  31,
  32,
  33,
  34,
  35,
  38,
  39,
  44,
  45,
  0xFE,
  0,211,12,
  31,
  32,
  33,
  34,
  35,
  36,
  38,
  39,
  44,
  45,
  50,
  56,
  0xFE,
  0,209,10,
  31,
  32,
  34,
  35,
  38,
  39,
  44,
  45,
  50,
  56,
  0,154,0,
  0,103,2,
  34,
  65,
  0xFE,
  0,99,1,
  34,
  0xFE,
  0xFE,
  1,1,1,
  34,
  0xFE,
  0xFE,
  0,207,6,
  31,
  32,
  34,
  35,
  38,
  44,
  0,121,0,
  0,188,1,
  98,
  0xFE,
  0xFE,
  1,1,5,
  31,
  32,
  34,
  35,
  38,
  0xFE,
  0,251,2,
  31,
  34,
  0xFE,
  0xFE,
  0,205,7,
  34,
  38,
  39,
  44,
  45,
  50,
  56,
  0,157,2,
  98,
  99,
  0xFE,
  1,1,6,
  31,
  32,
  34,
  35,
  38,
  44,
  0xFE,
  0xFE,
  1,1,7,
  31,
  32,
  34,
  38,
  39,
  44,
  45,
  0xFE,
  0,255,5,
  31,
  32,
  34,
  35,
  44,
  0xFE,
  0,253,6,
  31,
  34,
  38,
  39,
  44,
  45,
  0xFE,
  0,251,4,
  31,
  34,
  38,
  44,
  0xFE,
  0,249,4,
  38,
  44,
  157,
  163,
  0xFE,
  1,47,2,
  38,
  44,
  0xFE,
  1,45,2,
  191,
  197,
  0xFE,
  1,93,2,
  221,
  227,
  0xFE,
  0xFE,
  0,38,14,
  3,
  4,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  20,
  25,
  26,
  38,
  44,
  0,115,0,
  0,170,4,
  34,
  65,
  98,
  99,
  0xFE,
  0xFE,
  0,101,11,
  31,
  34,
  38,
  44,
  50,
  56,
  62,
  71,
  77,
  83,
  89,
  0xFE,
  0,99,25,
  31,
  32,
  34,
  35,
  38,
  39,
  40,
  44,
  45,
  46,
  50,
  56,
  57,
  62,
  63,
  65,
  66,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  89,
  0xFE,
  0,97,34,
  31,
  32,
  33,
  34,
  35,
  36,
  38,
  39,
  40,
  44,
  45,
  46,
  50,
  51,
  52,
  56,
  57,
  58,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  84,
  89,
  90,
  0xFE,
  0,95,28,
  31,
  32,
  33,
  34,
  35,
  36,
  38,
  39,
  44,
  45,
  50,
  51,
  56,
  57,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  77,
  78,
  83,
  84,
  89,
  90,
  0xFE,
  0,93,24,
  31,
  32,
  34,
  35,
  38,
  39,
  44,
  45,
  50,
  51,
  56,
  57,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  84,
  89,
  90,
  0xFE,
  0,91,22,
  31,
  32,
  34,
  35,
  38,
  39,
  44,
  45,
  50,
  51,
  56,
  57,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  89,
  0xFE,
  0,89,23,
  31,
  32,
  33,
  34,
  35,
  36,
  38,
  44,
  45,
  50,
  51,
  56,
  57,
  62,
  63,
  65,
  66,
  71,
  77,
  83,
  84,
  89,
  90,
  0xFE,
  0,87,10,
  34,
  38,
  44,
  50,
  56,
  65,
  71,
  77,
  83,
  89,
  0xFE,
  0,163,8,
  34,
  98,
  99,
  100,
  104,
  105,
  110,
  111,
  0xFE,
  0,161,5,
  98,
  104,
  105,
  110,
  111,
  0xFE,
  0,159,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,157,12,
  34,
  38,
  44,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,155,13,
  31,
  34,
  38,
  44,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,153,20,
  31,
  32,
  34,
  35,
  38,
  39,
  44,
  45,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  0xFE,
  0,151,17,
  31,
  32,
  34,
  35,
  38,
  39,
  44,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  110,
  111,
  0xFE,
  0,149,16,
  31,
  32,
  34,
  35,
  38,
  44,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  0xFE,
  0,147,11,
  31,
  34,
  38,
  44,
  95,
  98,
  104,
  105,
  110,
  111,
  116,
  0xFE,
  0,145,10,
  31,
  34,
  38,
  44,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0,143,5,
  38,
  44,
  98,
  104,
  110,
  0xFE,
  0,217,4,
  31,
  34,
  38,
  44,
  0xFE,
  0,215,6,
  31,
  32,
  34,
  35,
  38,
  44,
  0xFE,
  0,213,6,
  31,
  32,
  34,
  35,
  38,
  44,
  0xFE,
  0,211,6,
  31,
  34,
  38,
  39,
  44,
  45,
  0xFE,
  0,209,6,
  31,
  32,
  34,
  35,
  38,
  44,
  0xFE,
  0,207,4,
  38,
  39,
  44,
  45,
  1,1,2,
  31,
  34,
  0xFE,
  0xFE,
  0,205,0,
  1,1,2,
  38,
  44,
  0xFE,
  0xFE,
  0,255,4,
  31,
  34,
  38,
  44,
  0xFE,
  0,253,2,
  38,
  44,
  0xFE,
  0xFE,
  0,40,9,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  26,
  0,113,0,
  0,160,3,
  34,
  65,
  98,
  0xFE,
  0xFE,
  0,105,4,
  38,
  44,
  71,
  77,
  0xFE,
  0,103,14,
  31,
  38,
  39,
  44,
  45,
  50,
  56,
  62,
  71,
  72,
  77,
  78,
  83,
  89,
  0xFE,
  0,101,32,
  31,
  32,
  33,
  34,
  35,
  36,
  38,
  39,
  40,
  44,
  45,
  46,
  50,
  51,
  56,
  57,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  84,
  89,
  90,
  0,205,1,
  34,
  0xFE,
  0xFE,
  0,99,32,
  31,
  32,
  33,
  34,
  35,
  36,
  38,
  39,
  40,
  44,
  45,
  46,
  50,
  51,
  56,
  57,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  84,
  89,
  90,
  0xFE,
  0,97,27,
  31,
  32,
  33,
  34,
  35,
  36,
  38,
  44,
  45,
  50,
  51,
  56,
  57,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  77,
  78,
  83,
  84,
  89,
  90,
  0,147,12,
  34,
  35,
  38,
  44,
  65,
  66,
  71,
  77,
  98,
  99,
  104,
  110,
  0xFE,
  0xFE,
  0,95,24,
  31,
  32,
  34,
  35,
  38,
  39,
  44,
  45,
  50,
  51,
  56,
  57,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  84,
  89,
  90,
  0xFE,
  0,93,21,
  31,
  32,
  33,
  34,
  35,
  38,
  39,
  44,
  45,
  50,
  56,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  89,
  0xFE,
  0,91,22,
  31,
  32,
  33,
  34,
  35,
  38,
  44,
  45,
  50,
  51,
  56,
  57,
  62,
  63,
  65,
  66,
  71,
  77,
  83,
  84,
  89,
  90,
  0xFE,
  0,89,4,
  34,
  65,
  71,
  77,
  0xFE,
  0,161,6,
  95,
  98,
  104,
  105,
  110,
  111,
  0xFE,
  0,159,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,157,12,
  34,
  38,
  44,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,155,14,
  31,
  32,
  34,
  38,
  44,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,153,15,
  31,
  32,
  34,
  35,
  38,
  44,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,151,13,
  31,
  32,
  34,
  35,
  38,
  44,
  95,
  96,
  98,
  99,
  104,
  110,
  111,
  0xFE,
  0,149,9,
  31,
  34,
  38,
  44,
  95,
  98,
  104,
  105,
  110,
  0xFE,
  0,147,10,
  31,
  34,
  38,
  44,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0,145,3,
  98,
  104,
  110,
  0xFE,
  0,217,3,
  31,
  34,
  38,
  0xFE,
  0,215,3,
  31,
  34,
  38,
  0xFE,
  0,213,4,
  31,
  34,
  38,
  44,
  0xFE,
  0,211,4,
  31,
  34,
  38,
  44,
  0xFE,
  0,209,2,
  38,
  44,
  0xFE,
  0xFE,
  0,42,8,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  0,113,0,
  0,154,4,
  34,
  65,
  98,
  99,
  0xFE,
  0xFE,
  0,107,4,
  38,
  44,
  71,
  77,
  0xFE,
  0,105,23,
  31,
  32,
  34,
  35,
  36,
  38,
  39,
  44,
  45,
  50,
  51,
  56,
  57,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  89,
  0xFE,
  0,103,24,
  31,
  32,
  34,
  35,
  38,
  39,
  40,
  44,
  45,
  46,
  50,
  56,
  62,
  63,
  65,
  66,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  89,
  0xFE,
  0,101,32,
  31,
  32,
  33,
  34,
  35,
  36,
  38,
  39,
  40,
  44,
  45,
  46,
  50,
  51,
  56,
  57,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  84,
  89,
  90,
  0xFE,
  0,99,23,
  31,
  32,
  33,
  34,
  35,
  36,
  38,
  44,
  45,
  50,
  51,
  56,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  77,
  83,
  84,
  89,
  0xFE,
  0,97,24,
  31,
  32,
  34,
  35,
  38,
  39,
  44,
  45,
  50,
  51,
  56,
  57,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  84,
  89,
  90,
  0xFE,
  0,95,20,
  31,
  32,
  34,
  35,
  38,
  39,
  44,
  45,
  50,
  56,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  89,
  0xFE,
  0,93,21,
  31,
  32,
  33,
  34,
  35,
  38,
  44,
  50,
  51,
  56,
  57,
  62,
  63,
  65,
  66,
  71,
  77,
  83,
  84,
  89,
  90,
  0xFE,
  0,91,5,
  34,
  38,
  44,
  71,
  77,
  0xFE,
  0,163,2,
  104,
  110,
  0xFE,
  0,161,7,
  95,
  96,
  98,
  104,
  105,
  110,
  111,
  0xFE,
  0,159,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,157,13,
  31,
  34,
  38,
  44,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,155,14,
  31,
  32,
  34,
  38,
  44,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,153,10,
  31,
  34,
  38,
  44,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0,151,8,
  31,
  34,
  38,
  44,
  95,
  98,
  104,
  110,
  0xFE,
  0,149,8,
  31,
  34,
  38,
  44,
  95,
  98,
  104,
  110,
  0xFE,
  0,147,2,
  104,
  110,
  0xFE,
  0,217,1,
  34,
  0xFE,
  0,215,4,
  31,
  34,
  38,
  44,
  0xFE,
  0,213,2,
  38,
  44,
  0xFE,
  0xFE,
  0,44,3,
  3,
  7,
  13,
  0,109,4,
  38,
  44,
  71,
  77,
  0xFE,
  0,107,18,
  31,
  34,
  35,
  38,
  39,
  44,
  45,
  50,
  56,
  62,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  89,
  0xFE,
  0,105,22,
  31,
  32,
  34,
  35,
  38,
  39,
  40,
  44,
  45,
  46,
  50,
  56,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  89,
  0xFE,
  0,103,32,
  31,
  32,
  33,
  34,
  35,
  36,
  38,
  39,
  40,
  44,
  45,
  46,
  50,
  51,
  56,
  57,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  84,
  89,
  90,
  0xFE,
  0,101,21,
  31,
  32,
  33,
  34,
  35,
  36,
  38,
  44,
  45,
  50,
  56,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  77,
  83,
  89,
  0xFE,
  0,99,23,
  31,
  32,
  34,
  35,
  38,
  39,
  44,
  45,
  50,
  51,
  56,
  57,
  62,
  63,
  65,
  71,
  72,
  77,
  78,
  83,
  84,
  89,
  90,
  0xFE,
  0,97,20,
  31,
  32,
  34,
  35,
  38,
  39,
  44,
  45,
  50,
  56,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  89,
  0xFE,
  0,95,21,
  31,
  32,
  33,
  34,
  35,
  38,
  44,
  50,
  51,
  56,
  57,
  62,
  63,
  65,
  66,
  71,
  77,
  83,
  84,
  89,
  90,
  0xFE,
  0,93,2,
  38,
  44,
  0xFE,
  0,165,2,
  104,
  110,
  0xFE,
  0,163,5,
  95,
  96,
  98,
  104,
  110,
  0xFE,
  0,161,7,
  95,
  96,
  98,
  104,
  105,
  110,
  111,
  0xFE,
  0,159,11,
  34,
  38,
  44,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,157,11,
  31,
  34,
  38,
  44,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  0xFE,
  0,155,9,
  31,
  34,
  38,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0,153,8,
  31,
  34,
  38,
  44,
  95,
  98,
  104,
  110,
  0xFE,
  0,151,5,
  34,
  95,
  98,
  104,
  110,
  0xFE,
  0xFE,
  0,46,0,
  0,127,0,
  0,168,2,
  34,
  98,
  0xFE,
  0xFE,
  0,119,0,
  0,154,0,
  0,209,2,
  34,
  65,
  0xFE,
  0xFE,
  0xFE,
  0,111,37,
  3,
  7,
  13,
  31,
  32,
  33,
  34,
  35,
  36,
  38,
  39,
  40,
  44,
  45,
  46,
  50,
  51,
  52,
  56,
  57,
  58,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  84,
  89,
  90,
  0xFE,
  0,109,14,
  31,
  34,
  38,
  39,
  44,
  45,
  50,
  56,
  62,
  65,
  71,
  77,
  83,
  89,
  0xFE,
  0,107,20,
  31,
  32,
  34,
  35,
  38,
  39,
  44,
  45,
  50,
  56,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  89,
  0xFE,
  0,105,27,
  31,
  32,
  33,
  34,
  35,
  36,
  38,
  39,
  44,
  45,
  50,
  51,
  56,
  57,
  62,
  63,
  64,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  84,
  89,
  90,
  0xFE,
  0,103,21,
  31,
  32,
  33,
  34,
  35,
  36,
  38,
  44,
  45,
  50,
  56,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  77,
  83,
  89,
  0xFE,
  0,101,22,
  31,
  32,
  34,
  35,
  38,
  39,
  44,
  45,
  50,
  51,
  56,
  57,
  62,
  63,
  65,
  71,
  72,
  77,
  78,
  83,
  89,
  90,
  0xFE,
  0,99,20,
  31,
  32,
  34,
  35,
  38,
  39,
  44,
  45,
  50,
  56,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  89,
  0xFE,
  0,97,18,
  31,
  32,
  34,
  35,
  38,
  44,
  50,
  51,
  56,
  57,
  62,
  63,
  65,
  66,
  71,
  77,
  83,
  89,
  0xFE,
  0,95,2,
  38,
  44,
  0xFE,
  0,165,2,
  104,
  110,
  0xFE,
  0,163,5,
  95,
  96,
  98,
  104,
  110,
  0xFE,
  0,161,5,
  95,
  96,
  98,
  104,
  110,
  0xFE,
  0,159,8,
  31,
  34,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0,157,7,
  31,
  34,
  95,
  96,
  98,
  104,
  110,
  0xFE,
  0,155,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,153,2,
  95,
  98,
  0xFE,
  0xFE,
  0,48,0,
  0,127,0,
  0,162,1,
  65,
  0xFE,
  0xFE,
  0,113,6,
  34,
  38,
  44,
  65,
  71,
  77,
  0xFE,
  0,111,10,
  31,
  34,
  38,
  44,
  50,
  56,
  62,
  65,
  71,
  77,
  0xFE,
  0,109,19,
  31,
  32,
  34,
  35,
  38,
  39,
  44,
  45,
  50,
  56,
  62,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  89,
  0xFE,
  0,107,23,
  31,
  32,
  33,
  34,
  35,
  38,
  39,
  44,
  45,
  50,
  56,
  57,
  62,
  63,
  64,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  89,
  0xFE,
  0,105,19,
  31,
  32,
  33,
  34,
  35,
  36,
  38,
  44,
  50,
  56,
  62,
  63,
  64,
  65,
  66,
  71,
  77,
  83,
  89,
  0xFE,
  0,103,19,
  31,
  32,
  34,
  38,
  39,
  44,
  45,
  50,
  56,
  57,
  62,
  65,
  71,
  72,
  77,
  78,
  83,
  89,
  90,
  0xFE,
  0,101,20,
  31,
  32,
  34,
  35,
  38,
  39,
  44,
  45,
  50,
  56,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  89,
  0xFE,
  0,99,16,
  31,
  32,
  34,
  35,
  38,
  44,
  50,
  56,
  62,
  63,
  65,
  66,
  71,
  77,
  83,
  89,
  0xFE,
  0,97,2,
  38,
  44,
  0xFE,
  0,165,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,163,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,161,5,
  95,
  96,
  98,
  104,
  110,
  0xFE,
  0,159,3,
  95,
  98,
  104,
  0xFE,
  0,157,3,
  95,
  98,
  110,
  0xFE,
  0,155,1,
  98,
  0xFE,
  0xFE,
  0,50,0,
  0,117,1,
  65,
  0xFE,
  0,115,4,
  38,
  44,
  71,
  77,
  0xFE,
  0,113,6,
  34,
  38,
  44,
  65,
  71,
  77,
  0xFE,
  0,111,16,
  31,
  34,
  38,
  39,
  44,
  45,
  50,
  56,
  62,
  65,
  71,
  72,
  77,
  78,
  83,
  89,
  0xFE,
  0,109,22,
  31,
  32,
  33,
  34,
  35,
  38,
  39,
  44,
  45,
  50,
  56,
  62,
  63,
  64,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  89,
  0xFE,
  0,107,18,
  31,
  32,
  33,
  34,
  35,
  38,
  44,
  50,
  56,
  62,
  63,
  64,
  65,
  66,
  71,
  77,
  83,
  89,
  0xFE,
  0,105,16,
  31,
  32,
  34,
  38,
  44,
  45,
  50,
  56,
  57,
  62,
  65,
  71,
  77,
  78,
  83,
  89,
  0xFE,
  0,103,17,
  31,
  32,
  34,
  35,
  38,
  39,
  44,
  45,
  50,
  56,
  62,
  63,
  65,
  66,
  71,
  77,
  83,
  0xFE,
  0,101,16,
  31,
  32,
  34,
  35,
  38,
  44,
  50,
  56,
  62,
  63,
  65,
  66,
  71,
  77,
  83,
  89,
  0xFE,
  0,99,2,
  38,
  44,
  0xFE,
  0,91,0,
  1,145,2,
  65,
  66,
  0xFE,
  0xFE,
  0,165,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,163,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,161,2,
  95,
  98,
  0xFE,
  0xFE,
  0,52,0,
  0,115,6,
  34,
  38,
  44,
  65,
  71,
  77,
  0xFE,
  0,113,16,
  31,
  34,
  38,
  39,
  44,
  45,
  50,
  56,
  62,
  65,
  71,
  72,
  77,
  78,
  83,
  89,
  0xFE,
  0,111,21,
  31,
  32,
  33,
  34,
  35,
  38,
  39,
  44,
  45,
  50,
  56,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  89,
  0xFE,
  0,109,16,
  31,
  32,
  33,
  34,
  35,
  38,
  44,
  50,
  62,
  63,
  64,
  65,
  66,
  71,
  77,
  83,
  0xFE,
  0,107,14,
  31,
  34,
  38,
  44,
  45,
  50,
  56,
  62,
  65,
  71,
  77,
  78,
  83,
  89,
  0xFE,
  0,105,14,
  31,
  32,
  34,
  35,
  38,
  44,
  50,
  56,
  62,
  63,
  65,
  66,
  71,
  77,
  0xFE,
  0,103,16,
  31,
  32,
  34,
  35,
  38,
  44,
  50,
  56,
  62,
  63,
  65,
  66,
  71,
  77,
  83,
  89,
  0xFE,
  0,101,2,
  38,
  44,
  0xFE,
  0xFE,
  0,54,0,
  0,117,5,
  34,
  38,
  44,
  71,
  77,
  0xFE,
  0,115,14,
  31,
  34,
  38,
  39,
  44,
  45,
  50,
  56,
  62,
  65,
  71,
  72,
  77,
  78,
  0xFE,
  0,113,20,
  31,
  32,
  34,
  35,
  38,
  39,
  44,
  45,
  50,
  56,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  89,
  0xFE,
  0,111,15,
  31,
  32,
  33,
  34,
  35,
  38,
  44,
  50,
  62,
  63,
  65,
  66,
  71,
  77,
  83,
  0xFE,
  0,109,14,
  31,
  34,
  38,
  44,
  45,
  50,
  56,
  62,
  65,
  71,
  77,
  78,
  83,
  89,
  0xFE,
  0,107,12,
  31,
  32,
  34,
  35,
  38,
  44,
  62,
  63,
  65,
  66,
  71,
  77,
  0xFE,
  0,105,14,
  31,
  32,
  34,
  35,
  38,
  44,
  50,
  56,
  62,
  65,
  71,
  77,
  83,
  89,
  0xFE,
  0,103,1,
  44,
  0xFE,
  0xFE,
  0,56,0,
  0,119,4,
  38,
  44,
  71,
  77,
  0xFE,
  0,117,10,
  31,
  34,
  38,
  39,
  44,
  45,
  62,
  65,
  71,
  77,
  0xFE,
  0,115,20,
  31,
  32,
  34,
  35,
  38,
  39,
  44,
  45,
  50,
  56,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  89,
  0xFE,
  0,113,13,
  31,
  32,
  34,
  35,
  38,
  44,
  50,
  62,
  63,
  65,
  66,
  71,
  77,
  0xFE,
  0,111,14,
  31,
  34,
  38,
  44,
  45,
  50,
  56,
  62,
  65,
  71,
  77,
  78,
  83,
  89,
  0xFE,
  0,109,12,
  31,
  32,
  34,
  35,
  38,
  44,
  62,
  63,
  65,
  66,
  71,
  77,
  0xFE,
  0,107,11,
  31,
  32,
  34,
  38,
  44,
  50,
  56,
  62,
  65,
  83,
  89,
  0xFE,
  0xFE,
  0,58,0,
  0,121,4,
  38,
  44,
  71,
  77,
  0xFE,
  0,119,8,
  31,
  34,
  38,
  44,
  62,
  65,
  71,
  77,
  0xFE,
  0,117,17,
  31,
  32,
  34,
  35,
  38,
  39,
  44,
  45,
  50,
  56,
  62,
  63,
  65,
  71,
  72,
  77,
  78,
  0xFE,
  0,115,12,
  31,
  32,
  34,
  35,
  38,
  44,
  62,
  63,
  65,
  66,
  71,
  77,
  0xFE,
  0,113,12,
  31,
  34,
  38,
  44,
  45,
  50,
  56,
  62,
  65,
  71,
  77,
  83,
  0xFE,
  0,111,11,
  31,
  32,
  34,
  35,
  38,
  44,
  62,
  63,
  65,
  71,
  77,
  0xFE,
  0,109,10,
  31,
  34,
  38,
  44,
  50,
  56,
  62,
  65,
  83,
  89,
  0xFE,
  0xFE,
  0,60,0,
  0,121,7,
  31,
  34,
  38,
  44,
  65,
  71,
  77,
  0xFE,
  0,119,16,
  31,
  32,
  34,
  35,
  38,
  39,
  44,
  45,
  56,
  62,
  63,
  65,
  71,
  72,
  77,
  78,
  0xFE,
  0,117,12,
  31,
  32,
  34,
  35,
  38,
  44,
  62,
  63,
  65,
  66,
  71,
  77,
  0xFE,
  0,115,11,
  31,
  34,
  38,
  44,
  45,
  50,
  62,
  65,
  71,
  77,
  83,
  0xFE,
  0,113,9,
  31,
  32,
  34,
  38,
  44,
  62,
  65,
  71,
  77,
  0xFE,
  0,111,8,
  31,
  34,
  38,
  44,
  50,
  56,
  62,
  65,
  0xFE,
  0xFE,
  0,62,0,
  0,123,6,
  34,
  38,
  44,
  65,
  71,
  77,
  0xFE,
  0,121,13,
  31,
  32,
  34,
  38,
  39,
  44,
  45,
  62,
  65,
  71,
  72,
  77,
  78,
  0xFE,
  0,119,12,
  31,
  32,
  34,
  35,
  38,
  44,
  62,
  63,
  65,
  66,
  71,
  77,
  0xFE,
  0,117,8,
  31,
  34,
  38,
  44,
  50,
  62,
  65,
  71,
  0xFE,
  0,115,8,
  31,
  34,
  38,
  44,
  62,
  65,
  71,
  77,
  0xFE,
  0,113,4,
  31,
  34,
  62,
  65,
  0xFE,
  0xFE,
  0,64,0,
  0,125,5,
  34,
  38,
  44,
  71,
  77,
  0xFE,
  0,123,8,
  31,
  34,
  38,
  44,
  62,
  65,
  71,
  77,
  0xFE,
  0,121,12,
  31,
  32,
  34,
  35,
  38,
  44,
  62,
  63,
  65,
  66,
  71,
  77,
  0xFE,
  0,119,7,
  31,
  34,
  38,
  44,
  62,
  65,
  71,
  0xFE,
  0,117,8,
  31,
  34,
  38,
  44,
  62,
  65,
  71,
  77,
  0xFE,
  0,115,3,
  31,
  34,
  65,
  0xFE,
  0xFE,
  0,66,0,
  0,127,4,
  38,
  44,
  71,
  77,
  0xFE,
  0,125,8,
  31,
  34,
  38,
  44,
  62,
  65,
  71,
  77,
  0xFE,
  0,123,11,
  31,
  32,
  34,
  35,
  38,
  44,
  62,
  63,
  65,
  71,
  77,
  0xFE,
  0,121,7,
  31,
  34,
  38,
  44,
  62,
  65,
  71,
  0xFE,
  0,119,8,
  31,
  34,
  38,
  44,
  62,
  65,
  71,
  77,
  0xFE,
  0,117,2,
  31,
  34,
  0xFE,
  0xFE,
  0,68,0,
  0,129,4,
  38,
  44,
  71,
  77,
  0xFE,
  0,127,8,
  31,
  34,
  38,
  44,
  62,
  65,
  71,
  77,
  0xFE,
  0,125,10,
  31,
  32,
  34,
  38,
  44,
  62,
  63,
  65,
  71,
  77,
  0xFE,
  0,123,7,
  31,
  34,
  38,
  44,
  62,
  65,
  71,
  0xFE,
  0,121,7,
  31,
  34,
  38,
  44,
  65,
  71,
  77,
  0xFE,
  0xFE,
  0,70,0,
  0,129,8,
  31,
  34,
  38,
  44,
  62,
  65,
  71,
  77,
  0xFE,
  0,127,9,
  31,
  32,
  34,
  38,
  44,
  62,
  65,
  71,
  77,
  0xFE,
  0,125,6,
  31,
  34,
  38,
  62,
  65,
  71,
  0xFE,
  0,123,5,
  34,
  38,
  44,
  71,
  77,
  0xFE,
  0xFE,
  0,72,0,
  0,131,6,
  31,
  34,
  38,
  44,
  71,
  77,
  0xFE,
  0,129,8,
  31,
  34,
  38,
  44,
  62,
  65,
  71,
  77,
  0xFE,
  0,127,5,
  31,
  34,
  38,
  62,
  65,
  0xFE,
  0,125,4,
  38,
  44,
  71,
  77,
  0xFE,
  0xFE,
  0,74,0,
  0,133,5,
  34,
  38,
  44,
  71,
  77,
  0xFE,
  0,131,8,
  31,
  34,
  38,
  44,
  62,
  65,
  71,
  77,
  0xFE,
  0,129,5,
  31,
  34,
  38,
  62,
  65,
  0xFE,
  0,127,2,
  38,
  44,
  0xFE,
  0xFE,
  0,76,0,
  0,135,4,
  38,
  44,
  71,
  77,
  0xFE,
  0,133,8,
  31,
  34,
  38,
  44,
  62,
  65,
  71,
  77,
  0xFE,
  0,131,4,
  31,
  34,
  62,
  65,
  0xFE,
  0,129,2,
  38,
  44,
  0xFE,
  0xFE,
  0,78,0,
  0,135,8,
  31,
  34,
  38,
  44,
  62,
  65,
  71,
  77,
  0xFE,
  0,133,4,
  31,
  34,
  62,
  65,
  0xFE,
  0xFE,
  0,80,0,
  0,137,6,
  31,
  34,
  38,
  44,
  71,
  77,
  0xFE,
  0,135,4,
  31,
  34,
  62,
  65,
  0xFE,
  0xFE,
  0,82,0,
  0,137,3,
  31,
  34,
  65,
  0xFE,
  0xFE,
  0,84,0,
  0,139,2,
  31,
  34,
  0xFE,
  0xFE,
  0,86,86,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  20,
  25,
  26,
  34,
  35,
  36,
  38,
  39,
  40,
  44,
  45,
  46,
  50,
  51,
  56,
  57,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  77,
  78,
  83,
  84,
  89,
  90,
  98,
  99,
  100,
  104,
  105,
  110,
  111,
  116,
  117,
  122,
  123,
  133,
  134,
  135,
  139,
  144,
  149,
  153,
  157,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  181,
  182,
  186,
  187,
  191,
  197,
  198,
  203,
  208,
  209,
  213,
  214,
  227,
  228,
  233,
  234,
  239,
  240,
  245,
  249,
  0,0,17,
  0,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  20,
  25,
  26,
  62,
  65,
  71,
  77,
  83,
  89,
  0,37,5,
  0,
  3,
  4,
  7,
  13,
  0xFE,
  0xFE,
  0,2,10,
  0,
  3,
  7,
  13,
  19,
  25,
  62,
  65,
  83,
  89,
  0,43,3,
  3,
  7,
  13,
  0xFE,
  0,41,3,
  3,
  7,
  13,
  0xFE,
  0,39,1,
  3,
  0xFE,
  0,37,1,
  3,
  0xFE,
  0xFE,
  0,4,0,
  0,47,1,
  3,
  0xFE,
  0xFE,
  0,6,0,
  0,53,1,
  3,
  0xFE,
  0xFE,
  0,32,27,
  31,
  32,
  33,
  34,
  35,
  36,
  38,
  39,
  44,
  45,
  50,
  51,
  56,
  57,
  62,
  63,
  65,
  66,
  67,
  71,
  72,
  77,
  78,
  83,
  84,
  89,
  90,
  0,149,7,
  34,
  65,
  95,
  98,
  99,
  104,
  110,
  0xFE,
  1,95,1,
  221,
  0xFE,
  0xFE,
  0,34,15,
  31,
  34,
  38,
  39,
  44,
  45,
  50,
  56,
  65,
  71,
  72,
  77,
  78,
  83,
  89,
  0,155,14,
  34,
  35,
  65,
  66,
  71,
  77,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0xFE,
  0,36,0,
  0,161,1,
  98,
  0xFE,
  0xFE,
  0,153,4,
  104,
  105,
  110,
  111,
  0xFE,
  0,151,15,
  71,
  77,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  0xFE,
  0,149,19,
  65,
  71,
  72,
  77,
  78,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  0xFE,
  0,147,24,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  89,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  0xFE,
  0,145,31,
  62,
  63,
  64,
  65,
  66,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  89,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  118,
  122,
  123,
  124,
  0xFE,
  0,143,35,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  84,
  89,
  90,
  95,
  96,
  97,
  98,
  99,
  100,
  101,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  118,
  122,
  123,
  124,
  0xFE,
  0,141,27,
  62,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  84,
  89,
  90,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  118,
  122,
  123,
  0xFE,
  0,140,21,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  89,
  95,
  96,
  98,
  99,
  100,
  104,
  105,
  110,
  111,
  116,
  122,
  0xFE,
  0,142,17,
  65,
  71,
  72,
  77,
  78,
  83,
  89,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  0xFE,
  0,144,9,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  0xFE,
  0,146,2,
  104,
  110,
  0xFE,
  0,217,2,
  71,
  77,
  1,10,2,
  71,
  77,
  0xFE,
  0xFE,
  0,215,5,
  65,
  71,
  72,
  77,
  78,
  1,2,2,
  71,
  77,
  0xFE,
  1,4,2,
  71,
  77,
  0xFE,
  1,6,2,
  71,
  77,
  0xFE,
  1,10,2,
  71,
  77,
  0xFE,
  0xFE,
  0,213,14,
  62,
  63,
  65,
  66,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  89,
  139,
  144,
  0xFE,
  0,211,19,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  89,
  127,
  139,
  140,
  144,
  145,
  0,142,2,
  95,
  98,
  0xFE,
  1,2,0,
  1,49,2,
  71,
  77,
  0xFE,
  0xFE,
  1,8,1,
  163,
  0xFE,
  0xFE,
  0,209,26,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  84,
  89,
  90,
  98,
  104,
  110,
  127,
  128,
  133,
  139,
  140,
  144,
  145,
  0,141,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,140,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,142,3,
  95,
  98,
  104,
  0xFE,
  0,144,1,
  98,
  0xFE,
  1,1,2,
  71,
  77,
  0xFE,
  0,251,2,
  71,
  77,
  0xFE,
  0,249,2,
  71,
  77,
  0xFE,
  0xFE,
  0,207,34,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  89,
  95,
  96,
  98,
  99,
  104,
  110,
  127,
  128,
  129,
  133,
  134,
  139,
  140,
  144,
  145,
  149,
  153,
  157,
  191,
  221,
  0,143,0,
  1,4,2,
  65,
  98,
  0xFE,
  0xFE,
  0,141,5,
  95,
  98,
  99,
  104,
  110,
  0xFE,
  0,140,4,
  95,
  98,
  104,
  110,
  0xFE,
  1,1,5,
  65,
  71,
  72,
  77,
  78,
  0xFE,
  0,253,2,
  71,
  77,
  0xFE,
  0,251,5,
  65,
  71,
  72,
  77,
  78,
  0xFE,
  0,249,2,
  71,
  77,
  0xFE,
  1,2,0,
  1,53,2,
  203,
  208,
  0xFE,
  1,51,2,
  191,
  197,
  0xFE,
  0xFE,
  1,93,2,
  233,
  239,
  0xFE,
  1,92,3,
  221,
  233,
  239,
  0xFE,
  0xFE,
  0,205,38,
  3,
  62,
  63,
  65,
  66,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  84,
  89,
  90,
  98,
  127,
  128,
  133,
  139,
  140,
  144,
  145,
  149,
  150,
  153,
  154,
  157,
  163,
  169,
  170,
  175,
  176,
  191,
  197,
  203,
  221,
  227,
  0,143,1,
  98,
  0xFE,
  0,141,2,
  95,
  98,
  0xFE,
  1,1,12,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  139,
  140,
  144,
  145,
  0xFE,
  0,255,2,
  71,
  77,
  0xFE,
  0,253,2,
  71,
  77,
  0xFE,
  0,251,8,
  62,
  65,
  71,
  72,
  77,
  78,
  139,
  144,
  0xFE,
  0,249,9,
  65,
  71,
  77,
  139,
  144,
  157,
  163,
  169,
  175,
  0xFE,
  1,45,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,44,3,
  191,
  203,
  208,
  0xFE,
  1,93,3,
  221,
  233,
  239,
  0xFE,
  1,92,3,
  221,
  233,
  239,
  0xFE,
  0xFE,
  1,1,17,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  127,
  128,
  133,
  139,
  140,
  144,
  145,
  169,
  175,
  0xFE,
  0,255,15,
  62,
  65,
  66,
  71,
  72,
  73,
  77,
  78,
  79,
  157,
  158,
  169,
  170,
  175,
  176,
  0xFE,
  0,253,25,
  62,
  63,
  65,
  66,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  89,
  139,
  144,
  157,
  158,
  159,
  163,
  164,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  0,251,26,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  127,
  133,
  139,
  144,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  181,
  186,
  191,
  221,
  0xFE,
  0,249,29,
  62,
  65,
  71,
  72,
  77,
  78,
  83,
  89,
  127,
  133,
  157,
  158,
  163,
  164,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  182,
  186,
  187,
  191,
  197,
  203,
  208,
  221,
  1,93,2,
  233,
  239,
  0xFE,
  1,92,2,
  233,
  239,
  0xFE,
  0xFE,
  1,53,2,
  203,
  208,
  1,94,2,
  203,
  208,
  0xFE,
  1,98,3,
  191,
  203,
  208,
  0xFE,
  0xFE,
  1,51,5,
  191,
  203,
  204,
  208,
  209,
  1,94,3,
  191,
  203,
  208,
  0xFE,
  1,96,1,
  191,
  0xFE,
  1,98,4,
  191,
  197,
  203,
  208,
  0xFE,
  0xFE,
  1,49,11,
  71,
  77,
  157,
  163,
  191,
  192,
  197,
  203,
  204,
  208,
  209,
  1,94,6,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,98,2,
  191,
  197,
  0xFE,
  0xFE,
  1,47,12,
  62,
  65,
  71,
  77,
  191,
  192,
  197,
  198,
  203,
  208,
  213,
  217,
  1,95,3,
  191,
  203,
  208,
  0xFE,
  1,92,5,
  191,
  197,
  221,
  233,
  239,
  0xFE,
  1,94,10,
  191,
  192,
  197,
  198,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,96,2,
  221,
  227,
  0xFE,
  0xFE,
  1,45,18,
  65,
  71,
  191,
  197,
  203,
  204,
  208,
  209,
  213,
  214,
  217,
  218,
  221,
  227,
  233,
  234,
  239,
  240,
  0,249,3,
  191,
  203,
  208,
  0xFE,
  1,95,6,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,93,2,
  233,
  239,
  0xFE,
  1,92,6,
  221,
  222,
  227,
  228,
  233,
  239,
  0xFE,
  1,94,4,
  203,
  208,
  233,
  239,
  0xFE,
  0xFE,
  1,44,10,
  191,
  197,
  203,
  204,
  208,
  209,
  213,
  214,
  217,
  218,
  0,251,1,
  191,
  0xFE,
  0,249,5,
  191,
  192,
  197,
  203,
  208,
  0xFE,
  1,101,2,
  233,
  239,
  0xFE,
  1,99,10,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,97,8,
  191,
  197,
  221,
  222,
  227,
  228,
  233,
  239,
  0xFE,
  1,95,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,93,2,
  233,
  239,
  0xFE,
  0xFE,
  1,46,4,
  203,
  208,
  213,
  217,
  0,143,1,
  98,
  0xFE,
  0,249,4,
  191,
  192,
  197,
  198,
  0xFE,
  1,103,2,
  233,
  239,
  0xFE,
  1,101,4,
  221,
  227,
  233,
  239,
  0xFE,
  1,99,3,
  197,
  221,
  227,
  0xFE,
  1,97,2,
  191,
  221,
  0xFE,
  1,95,2,
  203,
  208,
  0xFE,
  0xFE,
  1,48,0,
  1,3,2,
  71,
  77,
  0xFE,
  0,253,1,
  157,
  0xFE,
  0,251,2,
  157,
  163,
  0xFE,
  1,99,1,
  221,
  0xFE,
  0xFE,
  1,99,4,
  203,
  208,
  233,
  239,
  0xFE,
  1,97,3,
  221,
  233,
  239,
  0xFE,
  1,95,14,
  65,
  71,
  77,
  191,
  192,
  197,
  198,
  203,
  208,
  221,
  222,
  227,
  233,
  239,
  0xFE,
  1,93,13,
  191,
  197,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  245,
  246,
  249,
  250,
  0xFE,
  1,92,10,
  221,
  222,
  227,
  228,
  233,
  239,
  245,
  246,
  249,
  250,
  0xFE,
  1,94,4,
  203,
  208,
  233,
  239,
  0xFE,
  1,141,4,
  62,
  65,
  71,
  77,
  1,144,2,
  71,
  77,
  0xFE,
  1,146,2,
  71,
  77,
  0xFE,
  0xFE,
  1,139,0,
  1,142,2,
  83,
  89,
  0xFE,
  0xFE,
  1,138,0,
  1,143,4,
  83,
  85,
  89,
  91,
  0xFE,
  1,149,3,
  65,
  84,
  90,
  0xFE,
  0xFE,
  1,140,2,
  71,
  77,
  0xFE,
  1,145,2,
  71,
  77,
  0xFE,
  0xFE,
  0,88,75,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  20,
  25,
  26,
  34,
  35,
  36,
  38,
  39,
  44,
  45,
  50,
  51,
  56,
  57,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  77,
  78,
  83,
  84,
  89,
  90,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  117,
  122,
  123,
  133,
  134,
  139,
  140,
  144,
  145,
  163,
  164,
  169,
  170,
  175,
  176,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  227,
  233,
  234,
  235,
  239,
  240,
  241,
  0,6,0,
  0,53,1,
  3,
  0xFE,
  0,47,1,
  3,
  0xFE,
  0,45,1,
  3,
  0xFE,
  0xFE,
  0,8,0,
  0,59,14,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  34,
  35,
  65,
  66,
  0xFE,
  0xFE,
  0,34,1,
  71,
  0,159,1,
  98,
  0xFE,
  0xFE,
  0,157,2,
  104,
  110,
  0xFE,
  0,155,6,
  95,
  96,
  104,
  105,
  110,
  111,
  0xFE,
  0,153,17,
  71,
  72,
  77,
  78,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  0xFE,
  0,151,24,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  89,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  0xFE,
  0,149,27,
  62,
  63,
  64,
  65,
  66,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  89,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  0xFE,
  0,147,35,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  84,
  89,
  90,
  95,
  96,
  97,
  98,
  99,
  100,
  101,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  118,
  122,
  123,
  124,
  0xFE,
  0,145,31,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  77,
  78,
  83,
  84,
  89,
  90,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  110,
  111,
  112,
  116,
  117,
  118,
  122,
  123,
  124,
  0,213,1,
  98,
  0xFE,
  0xFE,
  0,143,21,
  62,
  65,
  71,
  72,
  77,
  78,
  83,
  89,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  117,
  122,
  123,
  0xFE,
  0,141,21,
  62,
  63,
  65,
  66,
  71,
  77,
  83,
  89,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  117,
  122,
  123,
  0xFE,
  0,217,3,
  65,
  71,
  77,
  1,2,2,
  62,
  65,
  0xFE,
  1,10,1,
  65,
  0xFE,
  0xFE,
  0,215,8,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  1,2,1,
  65,
  0xFE,
  1,4,2,
  62,
  65,
  0xFE,
  1,6,1,
  65,
  0xFE,
  1,10,2,
  62,
  65,
  0xFE,
  0xFE,
  0,213,14,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  89,
  0xFE,
  0,211,16,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  89,
  139,
  144,
  0xFE,
  0,209,18,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  77,
  78,
  83,
  89,
  98,
  104,
  110,
  127,
  133,
  139,
  144,
  0,145,3,
  98,
  104,
  110,
  0xFE,
  0,143,1,
  98,
  0xFE,
  0,141,3,
  98,
  104,
  110,
  0xFE,
  1,1,6,
  62,
  65,
  71,
  72,
  77,
  78,
  0xFE,
  0,251,3,
  65,
  71,
  77,
  0xFE,
  0xFE,
  0,207,22,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  84,
  89,
  90,
  127,
  133,
  139,
  140,
  144,
  145,
  0,145,3,
  98,
  104,
  110,
  0xFE,
  0,143,6,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0,141,3,
  95,
  98,
  110,
  0xFE,
  1,1,8,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  0xFE,
  0,253,2,
  62,
  65,
  0xFE,
  0,251,6,
  62,
  63,
  65,
  66,
  71,
  77,
  0xFE,
  0,249,4,
  62,
  65,
  71,
  77,
  0xFE,
  0xFE,
  0,205,20,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  77,
  78,
  83,
  89,
  127,
  128,
  133,
  134,
  139,
  144,
  157,
  191,
  221,
  0,145,1,
  98,
  0xFE,
  1,1,8,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  0xFE,
  0,253,3,
  65,
  71,
  77,
  0xFE,
  0,251,6,
  62,
  63,
  65,
  66,
  71,
  77,
  0xFE,
  0,249,2,
  62,
  65,
  0xFE,
  0xFE,
  1,3,0,
  1,48,2,
  62,
  65,
  0xFE,
  0xFE,
  1,1,8,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  0xFE,
  0,255,10,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  77,
  78,
  0xFE,
  0,253,12,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  77,
  83,
  89,
  169,
  175,
  0xFE,
  0,251,17,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  89,
  157,
  158,
  163,
  169,
  170,
  175,
  176,
  0xFE,
  0,249,19,
  62,
  63,
  65,
  66,
  71,
  77,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  181,
  186,
  221,
  0xFE,
  1,49,5,
  62,
  65,
  191,
  203,
  208,
  0,211,0,
  1,2,2,
  62,
  65,
  0xFE,
  0xFE,
  1,94,2,
  203,
  208,
  0xFE,
  0xFE,
  1,47,9,
  62,
  65,
  191,
  192,
  197,
  203,
  204,
  208,
  209,
  1,94,2,
  191,
  197,
  0xFE,
  0xFE,
  1,45,10,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  208,
  233,
  239,
  1,95,2,
  203,
  208,
  0xFE,
  1,92,2,
  221,
  227,
  0xFE,
  0xFE,
  1,44,2,
  203,
  208,
  1,99,2,
  233,
  239,
  0xFE,
  1,97,2,
  221,
  227,
  0xFE,
  0xFE,
  1,97,2,
  233,
  239,
  0xFE,
  1,95,10,
  62,
  65,
  191,
  197,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,93,10,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  239,
  245,
  249,
  0xFE,
  1,92,2,
  245,
  249,
  0xFE,
  1,141,2,
  71,
  77,
  0xFE,
  0xFE,
  0,90,45,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  20,
  25,
  26,
  34,
  38,
  39,
  44,
  45,
  50,
  56,
  62,
  65,
  71,
  72,
  77,
  78,
  83,
  89,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  139,
  144,
  169,
  175,
  203,
  204,
  208,
  209,
  233,
  234,
  239,
  240,
  0,159,5,
  96,
  104,
  105,
  110,
  111,
  0xFE,
  0,157,15,
  71,
  77,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  0xFE,
  0,155,19,
  62,
  63,
  71,
  72,
  77,
  78,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  0xFE,
  0,153,29,
  62,
  63,
  64,
  65,
  66,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  89,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  0xFE,
  0,151,33,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  89,
  95,
  96,
  97,
  98,
  99,
  100,
  101,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  118,
  122,
  123,
  124,
  0xFE,
  0,149,32,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  89,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  118,
  122,
  123,
  124,
  0,211,6,
  65,
  66,
  98,
  99,
  104,
  110,
  0xFE,
  0xFE,
  0,147,28,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  77,
  78,
  83,
  89,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  0xFE,
  0,145,22,
  62,
  65,
  71,
  72,
  77,
  78,
  83,
  89,
  95,
  96,
  98,
  99,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  0xFE,
  0,143,22,
  62,
  63,
  65,
  66,
  71,
  77,
  83,
  89,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  110,
  111,
  116,
  117,
  122,
  123,
  0xFE,
  0,141,17,
  65,
  71,
  77,
  83,
  89,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  117,
  122,
  123,
  0xFE,
  0,219,2,
  71,
  77,
  0xFE,
  0,217,4,
  62,
  65,
  71,
  77,
  0xFE,
  0,215,9,
  62,
  63,
  64,
  65,
  66,
  71,
  72,
  77,
  78,
  0xFE,
  0,213,11,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  77,
  78,
  83,
  0xFE,
  0,211,15,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  77,
  78,
  83,
  89,
  133,
  139,
  144,
  0xFE,
  0,209,18,
  62,
  63,
  64,
  65,
  66,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  84,
  89,
  90,
  98,
  127,
  133,
  0,149,1,
  98,
  0xFE,
  1,1,8,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  0xFE,
  0,251,3,
  62,
  65,
  71,
  0xFE,
  0,249,1,
  65,
  0xFE,
  0xFE,
  0,207,13,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  77,
  78,
  83,
  89,
  139,
  144,
  0,149,3,
  98,
  104,
  110,
  1,1,1,
  98,
  0xFE,
  0xFE,
  0,145,2,
  95,
  98,
  0xFE,
  1,1,8,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  0xFE,
  0,251,2,
  62,
  65,
  0xFE,
  0,249,2,
  62,
  65,
  0xFE,
  0xFE,
  0,205,14,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  84,
  89,
  90,
  127,
  133,
  0,149,1,
  98,
  0xFE,
  1,1,8,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  0xFE,
  0,253,2,
  62,
  65,
  0xFE,
  0,251,4,
  62,
  65,
  71,
  77,
  0xFE,
  0xFE,
  1,1,8,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  0xFE,
  0,255,8,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  0xFE,
  0,253,6,
  62,
  65,
  71,
  72,
  77,
  78,
  0xFE,
  0,251,7,
  62,
  63,
  65,
  71,
  77,
  169,
  175,
  0xFE,
  0,249,6,
  71,
  77,
  157,
  158,
  163,
  164,
  0xFE,
  1,49,1,
  65,
  0xFE,
  1,47,4,
  71,
  77,
  203,
  208,
  0xFE,
  1,45,4,
  191,
  192,
  197,
  198,
  0xFE,
  1,93,4,
  221,
  222,
  227,
  228,
  0xFE,
  0xFE,
  0,92,15,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  34,
  38,
  44,
  50,
  56,
  71,
  77,
  0,163,2,
  104,
  110,
  0xFE,
  0,161,7,
  95,
  96,
  97,
  104,
  105,
  110,
  111,
  0xFE,
  0,159,17,
  71,
  72,
  77,
  78,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  0xFE,
  0,157,26,
  62,
  63,
  64,
  65,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  0xFE,
  0,155,30,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  89,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  0xFE,
  0,153,33,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  89,
  95,
  96,
  97,
  98,
  99,
  100,
  101,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  118,
  122,
  123,
  124,
  0xFE,
  0,151,29,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  73,
  77,
  78,
  83,
  89,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  0xFE,
  0,149,28,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  77,
  78,
  83,
  89,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  0xFE,
  0,147,21,
  62,
  65,
  71,
  72,
  77,
  78,
  83,
  89,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  0xFE,
  0,145,22,
  62,
  63,
  65,
  66,
  71,
  77,
  83,
  89,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  110,
  111,
  116,
  117,
  122,
  123,
  0xFE,
  0,143,18,
  62,
  65,
  71,
  77,
  83,
  89,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  117,
  122,
  123,
  0xFE,
  0,141,4,
  104,
  110,
  116,
  122,
  0xFE,
  0,219,4,
  62,
  65,
  71,
  77,
  0xFE,
  0,217,7,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  0xFE,
  0,215,9,
  62,
  63,
  64,
  65,
  66,
  71,
  72,
  77,
  78,
  0xFE,
  0,213,9,
  62,
  63,
  64,
  65,
  66,
  71,
  72,
  77,
  78,
  0xFE,
  0,211,10,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  89,
  1,1,1,
  65,
  0xFE,
  0xFE,
  0,209,7,
  62,
  63,
  64,
  65,
  66,
  71,
  77,
  0,153,6,
  65,
  95,
  98,
  99,
  104,
  110,
  0xFE,
  1,1,5,
  62,
  63,
  65,
  66,
  71,
  0xFE,
  0,251,1,
  65,
  0xFE,
  0xFE,
  0,207,9,
  62,
  63,
  65,
  71,
  72,
  77,
  78,
  83,
  89,
  0,153,2,
  98,
  99,
  0xFE,
  0,149,2,
  104,
  110,
  0xFE,
  0,147,1,
  98,
  0xFE,
  0,145,2,
  95,
  98,
  0xFE,
  1,1,6,
  62,
  63,
  65,
  66,
  71,
  77,
  0xFE,
  0xFE,
  0,205,2,
  71,
  77,
  0,153,3,
  98,
  104,
  110,
  0xFE,
  0,151,2,
  104,
  110,
  0xFE,
  0,149,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,147,2,
  95,
  98,
  0xFE,
  1,1,6,
  62,
  65,
  71,
  72,
  77,
  78,
  0xFE,
  0xFE,
  1,1,3,
  62,
  65,
  77,
  0xFE,
  0,255,4,
  62,
  65,
  71,
  77,
  0xFE,
  0,253,4,
  62,
  65,
  71,
  77,
  0xFE,
  0,251,2,
  71,
  77,
  0xFE,
  0xFE,
  0,94,5,
  3,
  7,
  13,
  19,
  25,
  0,165,5,
  96,
  104,
  105,
  110,
  111,
  0xFE,
  0,163,12,
  71,
  77,
  95,
  96,
  97,
  98,
  104,
  105,
  106,
  110,
  111,
  112,
  0xFE,
  0,161,19,
  62,
  63,
  71,
  72,
  77,
  78,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  0xFE,
  0,159,30,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  89,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  0xFE,
  0,157,31,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  89,
  95,
  96,
  97,
  98,
  99,
  100,
  101,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  0,209,2,
  98,
  99,
  0xFE,
  0,207,5,
  65,
  98,
  99,
  104,
  110,
  0xFE,
  0,205,1,
  98,
  0xFE,
  0xFE,
  0,155,30,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  89,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  0xFE,
  0,153,28,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  77,
  78,
  83,
  89,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  0xFE,
  0,151,28,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  77,
  78,
  83,
  89,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  0xFE,
  0,149,20,
  62,
  65,
  71,
  72,
  77,
  78,
  83,
  89,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  112,
  116,
  117,
  122,
  0xFE,
  0,147,20,
  62,
  63,
  65,
  66,
  71,
  77,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  110,
  111,
  116,
  117,
  122,
  123,
  0xFE,
  0,145,17,
  62,
  65,
  71,
  77,
  83,
  89,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  123,
  0xFE,
  0,219,4,
  62,
  65,
  71,
  77,
  0xFE,
  0,217,6,
  62,
  63,
  65,
  66,
  71,
  77,
  0xFE,
  0,215,6,
  62,
  63,
  65,
  66,
  71,
  77,
  0xFE,
  0,213,6,
  62,
  65,
  71,
  72,
  77,
  78,
  0xFE,
  0,211,6,
  62,
  63,
  65,
  66,
  71,
  77,
  0xFE,
  0,209,6,
  62,
  65,
  71,
  72,
  77,
  78,
  0xFE,
  0,207,2,
  71,
  77,
  0xFE,
  1,1,1,
  65,
  0,211,1,
  65,
  0xFE,
  0,209,2,
  62,
  65,
  0xFE,
  0,207,2,
  71,
  77,
  0xFE,
  0xFE,
  0,255,2,
  71,
  77,
  0xFE,
  0xFE,
  0,96,2,
  7,
  13,
  0,167,7,
  95,
  96,
  97,
  104,
  105,
  110,
  111,
  0xFE,
  0,165,11,
  71,
  77,
  95,
  96,
  97,
  104,
  105,
  106,
  110,
  111,
  112,
  0xFE,
  0,163,28,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  73,
  77,
  78,
  79,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  0xFE,
  0,161,32,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  89,
  95,
  96,
  97,
  98,
  99,
  100,
  101,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  118,
  122,
  123,
  0,207,6,
  65,
  66,
  98,
  99,
  104,
  110,
  0xFE,
  0,205,2,
  98,
  99,
  0,251,1,
  98,
  0xFE,
  0xFE,
  0xFE,
  0,159,30,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  89,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  0xFE,
  0,157,30,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  89,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  0xFE,
  0,155,28,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  77,
  78,
  83,
  89,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  0xFE,
  0,153,27,
  62,
  63,
  64,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  89,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  0xFE,
  0,151,19,
  62,
  65,
  71,
  72,
  77,
  78,
  83,
  89,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  117,
  122,
  0xFE,
  0,149,17,
  62,
  63,
  65,
  66,
  71,
  77,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  0xFE,
  0,147,14,
  62,
  65,
  71,
  77,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  0xFE,
  0,219,4,
  62,
  65,
  71,
  77,
  0xFE,
  0,217,5,
  62,
  63,
  65,
  71,
  77,
  0xFE,
  0,215,4,
  62,
  65,
  71,
  77,
  0xFE,
  0,213,4,
  62,
  65,
  71,
  77,
  0xFE,
  0,211,2,
  71,
  77,
  0xFE,
  0xFE,
  0,98,0,
  0,171,4,
  104,
  105,
  110,
  111,
  0xFE,
  0,169,9,
  95,
  96,
  97,
  104,
  105,
  106,
  110,
  111,
  112,
  0xFE,
  0,167,23,
  62,
  63,
  64,
  65,
  66,
  71,
  72,
  77,
  78,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  0xFE,
  0,165,36,
  62,
  63,
  64,
  65,
  66,
  67,
  68,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  89,
  95,
  96,
  97,
  98,
  99,
  100,
  101,
  102,
  103,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  118,
  122,
  123,
  124,
  0,205,10,
  65,
  66,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0xFE,
  0,163,29,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  0xFE,
  0,161,28,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  73,
  77,
  78,
  83,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  0xFE,
  0,159,26,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  77,
  78,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  0xFE,
  0,157,26,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  77,
  78,
  83,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  0xFE,
  0,155,23,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  110,
  111,
  112,
  116,
  117,
  122,
  0xFE,
  0,153,16,
  62,
  65,
  71,
  72,
  77,
  78,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  0xFE,
  0,151,17,
  62,
  63,
  65,
  66,
  71,
  77,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  0xFE,
  0,149,14,
  62,
  65,
  71,
  77,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  0xFE,
  0,219,2,
  62,
  65,
  0xFE,
  0,217,4,
  62,
  65,
  71,
  77,
  0xFE,
  0,215,4,
  62,
  65,
  71,
  77,
  0xFE,
  0,213,2,
  71,
  77,
  0xFE,
  0xFE,
  0,100,0,
  0,173,7,
  95,
  96,
  97,
  104,
  105,
  110,
  111,
  0xFE,
  0,171,17,
  65,
  71,
  72,
  77,
  78,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  0xFE,
  0,169,42,
  62,
  63,
  64,
  65,
  66,
  67,
  68,
  69,
  70,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  84,
  89,
  90,
  95,
  96,
  97,
  98,
  99,
  100,
  101,
  102,
  103,
  104,
  105,
  106,
  107,
  110,
  111,
  112,
  113,
  116,
  117,
  118,
  122,
  123,
  124,
  0,249,2,
  98,
  99,
  0xFE,
  1,47,2,
  98,
  99,
  0xFE,
  0xFE,
  0,167,28,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  73,
  77,
  78,
  79,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  0xFE,
  0,165,27,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  73,
  77,
  78,
  79,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  0,207,2,
  96,
  98,
  0xFE,
  0xFE,
  0,163,24,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  77,
  78,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  0xFE,
  0,161,24,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  77,
  78,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  0xFE,
  0,159,24,
  62,
  63,
  64,
  65,
  66,
  71,
  72,
  77,
  78,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  0xFE,
  0,157,20,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  110,
  111,
  116,
  122,
  0xFE,
  0,155,15,
  62,
  65,
  71,
  72,
  77,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  0xFE,
  0,153,17,
  62,
  63,
  65,
  66,
  71,
  77,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  0xFE,
  0,151,13,
  62,
  65,
  71,
  77,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  116,
  122,
  0xFE,
  0,149,2,
  104,
  110,
  0xFE,
  0xFE,
  0,102,0,
  0,175,11,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  106,
  110,
  111,
  112,
  0xFE,
  0,173,35,
  62,
  63,
  64,
  65,
  66,
  67,
  68,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  89,
  95,
  96,
  97,
  98,
  99,
  100,
  101,
  102,
  103,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  118,
  122,
  123,
  0xFE,
  0,171,24,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  77,
  78,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  0xFE,
  0,169,23,
  62,
  63,
  64,
  65,
  66,
  71,
  72,
  77,
  78,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  0xFE,
  0,167,24,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  77,
  78,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  0xFE,
  0,165,24,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  77,
  78,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  0xFE,
  0,163,23,
  62,
  63,
  64,
  65,
  66,
  71,
  72,
  77,
  78,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  0xFE,
  0,161,22,
  62,
  63,
  64,
  65,
  66,
  71,
  72,
  77,
  78,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  116,
  122,
  0xFE,
  0,159,20,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  110,
  111,
  116,
  122,
  0xFE,
  0,157,13,
  62,
  65,
  71,
  77,
  95,
  96,
  98,
  104,
  105,
  110,
  111,
  116,
  122,
  0xFE,
  0,155,15,
  62,
  63,
  65,
  71,
  77,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  0xFE,
  0,153,12,
  62,
  65,
  71,
  77,
  95,
  96,
  98,
  99,
  104,
  110,
  116,
  122,
  0xFE,
  0,151,2,
  104,
  110,
  0xFE,
  0xFE,
  0,104,0,
  0,179,6,
  96,
  97,
  104,
  105,
  110,
  111,
  0xFE,
  0,177,24,
  62,
  63,
  64,
  65,
  66,
  67,
  71,
  72,
  77,
  78,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  0xFE,
  0,175,22,
  62,
  63,
  65,
  66,
  67,
  71,
  72,
  77,
  78,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  0xFE,
  0,173,21,
  62,
  63,
  64,
  65,
  66,
  71,
  72,
  77,
  78,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  0xFE,
  0,171,23,
  62,
  63,
  64,
  65,
  66,
  71,
  72,
  77,
  78,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  0xFE,
  0,169,23,
  62,
  63,
  64,
  65,
  66,
  71,
  72,
  77,
  78,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  0xFE,
  0,167,23,
  62,
  63,
  64,
  65,
  66,
  71,
  72,
  77,
  78,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  0xFE,
  0,165,23,
  62,
  63,
  64,
  65,
  66,
  71,
  72,
  77,
  78,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  0xFE,
  0,163,20,
  62,
  63,
  64,
  65,
  66,
  71,
  72,
  77,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  110,
  111,
  116,
  122,
  0xFE,
  0,161,17,
  62,
  63,
  65,
  66,
  71,
  77,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  0xFE,
  0,159,12,
  62,
  65,
  71,
  77,
  95,
  98,
  104,
  105,
  110,
  111,
  116,
  122,
  0xFE,
  0,157,13,
  62,
  65,
  71,
  77,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  122,
  0xFE,
  0,155,9,
  65,
  71,
  77,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0xFE,
  0,106,0,
  0,181,14,
  65,
  66,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  0xFE,
  0,179,16,
  65,
  66,
  71,
  77,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  0xFE,
  0,177,16,
  65,
  66,
  71,
  77,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  0xFE,
  0,175,18,
  62,
  63,
  65,
  66,
  71,
  77,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  0xFE,
  0,173,20,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  0xFE,
  0,171,22,
  62,
  63,
  64,
  65,
  66,
  71,
  72,
  77,
  78,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  0xFE,
  0,169,23,
  62,
  63,
  64,
  65,
  66,
  71,
  72,
  77,
  78,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  0xFE,
  0,167,23,
  62,
  63,
  64,
  65,
  66,
  71,
  72,
  77,
  78,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  0xFE,
  0,165,19,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  110,
  111,
  116,
  122,
  0xFE,
  0,163,16,
  62,
  63,
  65,
  66,
  71,
  77,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  0xFE,
  0,161,10,
  62,
  65,
  71,
  77,
  95,
  98,
  104,
  105,
  110,
  111,
  0xFE,
  0,159,10,
  62,
  65,
  71,
  77,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0,157,7,
  71,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0xFE,
  0,108,0,
  0,185,10,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  110,
  111,
  0xFE,
  0,183,14,
  65,
  66,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  0xFE,
  0,181,12,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  0xFE,
  0,179,15,
  65,
  71,
  77,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  0xFE,
  0,177,17,
  63,
  65,
  66,
  71,
  77,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  0xFE,
  0,175,20,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  0xFE,
  0,173,21,
  62,
  63,
  64,
  65,
  66,
  71,
  72,
  77,
  78,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  0xFE,
  0,171,21,
  62,
  63,
  64,
  65,
  66,
  71,
  72,
  77,
  78,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  0xFE,
  0,169,18,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  110,
  111,
  116,
  0xFE,
  0,167,17,
  62,
  63,
  65,
  66,
  71,
  77,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  110,
  111,
  116,
  0xFE,
  0,165,13,
  62,
  63,
  65,
  71,
  77,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,163,9,
  62,
  65,
  71,
  77,
  95,
  98,
  104,
  105,
  110,
  0xFE,
  0,161,8,
  62,
  65,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0,159,5,
  95,
  96,
  98,
  104,
  110,
  0xFE,
  0xFE,
  0,110,0,
  0,187,12,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  0xFE,
  0,185,9,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,183,11,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  106,
  110,
  111,
  112,
  0xFE,
  0,181,12,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  0xFE,
  0,179,15,
  65,
  71,
  77,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  0xFE,
  0,177,18,
  62,
  63,
  65,
  71,
  72,
  77,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  0xFE,
  0,175,20,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  0xFE,
  0,173,19,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  0xFE,
  0,171,16,
  62,
  63,
  65,
  66,
  71,
  77,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  110,
  111,
  0xFE,
  0,169,15,
  62,
  63,
  65,
  66,
  71,
  77,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,167,12,
  62,
  65,
  71,
  77,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,165,6,
  71,
  77,
  95,
  98,
  104,
  110,
  0xFE,
  0,163,7,
  62,
  65,
  95,
  96,
  98,
  104,
  110,
  0xFE,
  0,161,4,
  95,
  98,
  104,
  110,
  0xFE,
  0xFE,
  0,112,0,
  0,191,9,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  110,
  111,
  0xFE,
  0,189,8,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,187,8,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,185,9,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,183,12,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  0xFE,
  0,181,13,
  65,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  0xFE,
  0,179,16,
  62,
  65,
  71,
  77,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  0xFE,
  0,177,18,
  62,
  63,
  65,
  66,
  71,
  77,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  0xFE,
  0,175,17,
  62,
  63,
  65,
  66,
  71,
  77,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  0xFE,
  0,173,16,
  62,
  63,
  65,
  66,
  71,
  77,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  110,
  111,
  0xFE,
  0,171,14,
  62,
  63,
  65,
  71,
  77,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,169,12,
  62,
  65,
  71,
  77,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,167,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,165,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,163,3,
  95,
  98,
  104,
  0xFE,
  0xFE,
  0,114,0,
  0,195,4,
  97,
  98,
  99,
  100,
  0xFE,
  0,189,5,
  96,
  97,
  98,
  104,
  110,
  0xFE,
  0,187,9,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,185,9,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,183,10,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  110,
  111,
  0xFE,
  0,181,15,
  65,
  71,
  77,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  0xFE,
  0,179,17,
  62,
  63,
  65,
  66,
  71,
  77,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  0xFE,
  0,177,16,
  62,
  63,
  65,
  66,
  71,
  77,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  110,
  111,
  0xFE,
  0,175,14,
  62,
  63,
  65,
  71,
  77,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,173,12,
  62,
  65,
  71,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,171,7,
  62,
  65,
  95,
  96,
  98,
  104,
  110,
  0xFE,
  0,169,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,167,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,165,2,
  95,
  98,
  0xFE,
  0xFE,
  0,116,0,
  0,199,5,
  66,
  97,
  98,
  99,
  100,
  0xFE,
  0,191,3,
  96,
  104,
  110,
  0xFE,
  0,189,8,
  95,
  96,
  97,
  98,
  104,
  105,
  110,
  111,
  0xFE,
  0,187,9,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,185,9,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,183,10,
  71,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,181,15,
  62,
  63,
  65,
  71,
  77,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  110,
  111,
  0xFE,
  0,179,15,
  62,
  63,
  65,
  71,
  77,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  110,
  111,
  0xFE,
  0,177,12,
  62,
  65,
  71,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,175,8,
  62,
  65,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0,173,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,171,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,169,3,
  95,
  98,
  110,
  0xFE,
  0,167,1,
  98,
  0xFE,
  0xFE,
  0,118,0,
  0,203,12,
  65,
  66,
  67,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  110,
  111,
  0xFE,
  0,191,7,
  95,
  96,
  98,
  104,
  105,
  110,
  111,
  0xFE,
  0,189,9,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,187,9,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,185,9,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,183,12,
  62,
  65,
  71,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,181,11,
  62,
  65,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,179,10,
  62,
  65,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  0xFE,
  0,177,8,
  62,
  65,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0,175,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,173,3,
  95,
  98,
  110,
  0xFE,
  0,171,2,
  95,
  98,
  0xFE,
  0xFE,
  0,120,0,
  0,193,5,
  96,
  104,
  105,
  110,
  111,
  0xFE,
  0,191,9,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,189,9,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,187,9,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,185,10,
  65,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,183,10,
  65,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,181,6,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0,179,5,
  95,
  96,
  98,
  104,
  110,
  0xFE,
  0,177,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,175,1,
  98,
  0xFE,
  0xFE,
  0,122,0,
  0,195,2,
  104,
  110,
  0xFE,
  0,193,9,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,191,9,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,189,9,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,187,9,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,185,8,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  0xFE,
  0,183,6,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0,181,3,
  95,
  98,
  104,
  0xFE,
  0,179,1,
  110,
  0xFE,
  0xFE,
  0,124,0,
  0,197,2,
  104,
  110,
  0xFE,
  0,195,7,
  95,
  96,
  97,
  98,
  104,
  105,
  110,
  0xFE,
  0,193,8,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  0xFE,
  0,191,9,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,189,8,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  0xFE,
  0,187,6,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0,185,5,
  95,
  96,
  98,
  104,
  110,
  0xFE,
  0,183,2,
  95,
  98,
  0xFE,
  0xFE,
  0,126,0,
  0,197,5,
  95,
  96,
  98,
  104,
  110,
  0xFE,
  0,195,6,
  95,
  96,
  97,
  98,
  104,
  110,
  0xFE,
  0,193,5,
  95,
  96,
  98,
  104,
  110,
  0xFE,
  0,191,6,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0,189,6,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0,187,2,
  95,
  98,
  0xFE,
  0,185,2,
  95,
  98,
  0xFE,
  0xFE,
  0,128,0,
  0,199,2,
  96,
  104,
  0xFE,
  0,197,5,
  95,
  96,
  98,
  104,
  110,
  0xFE,
  0,195,5,
  95,
  96,
  98,
  104,
  110,
  0xFE,
  0,193,5,
  95,
  96,
  98,
  104,
  110,
  0xFE,
  0,191,4,
  95,
  96,
  98,
  104,
  0xFE,
  0,189,2,
  95,
  98,
  0xFE,
  0xFE,
  0,130,0,
  0,199,1,
  96,
  0xFE,
  0,197,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,195,4,
  95,
  96,
  98,
  104,
  0xFE,
  0,193,2,
  95,
  98,
  0xFE,
  0xFE,
  0,132,0,
  0,197,2,
  95,
  98,
  0xFE,
  0xFE,
  0,140,59,
  3,
  4,
  5,
  7,
  8,
  13,
  14,
  19,
  20,
  25,
  34,
  35,
  38,
  44,
  50,
  65,
  66,
  67,
  71,
  72,
  77,
  78,
  83,
  89,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  118,
  122,
  123,
  124,
  133,
  134,
  139,
  144,
  163,
  164,
  169,
  175,
  186,
  197,
  198,
  203,
  208,
  213,
  227,
  233,
  239,
  0,4,0,
  0,47,2,
  3,
  4,
  0xFE,
  0xFE,
  0,34,4,
  104,
  110,
  116,
  122,
  0xFE,
  0,38,0,
  0,99,1,
  34,
  0xFE,
  0xFE,
  0,88,12,
  71,
  77,
  83,
  89,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  0xFE,
  0,231,0,
  1,28,1,
  98,
  0xFE,
  0xFE,
  0,227,0,
  1,12,1,
  110,
  0xFE,
  1,18,3,
  98,
  99,
  104,
  0xFE,
  0xFE,
  0,225,0,
  1,10,1,
  98,
  0xFE,
  1,12,1,
  98,
  0xFE,
  1,14,2,
  98,
  99,
  1,45,1,
  98,
  0xFE,
  0xFE,
  0xFE,
  0,223,0,
  1,6,3,
  98,
  104,
  110,
  0xFE,
  1,8,9,
  95,
  98,
  99,
  104,
  105,
  110,
  111,
  133,
  163,
  0xFE,
  1,10,4,
  98,
  99,
  104,
  110,
  0xFE,
  0xFE,
  0,221,0,
  1,2,4,
  98,
  99,
  104,
  110,
  1,46,1,
  98,
  0xFE,
  0xFE,
  1,4,5,
  95,
  98,
  99,
  104,
  110,
  1,45,1,
  98,
  0xFE,
  0xFE,
  1,6,3,
  98,
  104,
  110,
  0xFE,
  1,8,1,
  98,
  0xFE,
  0xFE,
  0,219,0,
  1,2,5,
  95,
  98,
  99,
  104,
  110,
  0xFE,
  1,4,3,
  98,
  104,
  110,
  0xFE,
  1,6,1,
  98,
  0xFE,
  1,8,3,
  98,
  104,
  110,
  0xFE,
  1,10,3,
  98,
  104,
  110,
  0xFE,
  0xFE,
  0,217,4,
  98,
  99,
  104,
  110,
  1,2,3,
  98,
  104,
  110,
  0xFE,
  1,4,3,
  98,
  104,
  110,
  0xFE,
  1,6,1,
  98,
  0xFE,
  1,8,3,
  98,
  104,
  110,
  0xFE,
  1,10,4,
  95,
  98,
  104,
  110,
  0xFE,
  1,50,0,
  1,94,1,
  98,
  0xFE,
  0xFE,
  1,52,2,
  98,
  99,
  0xFE,
  1,54,0,
  1,95,1,
  98,
  0xFE,
  0xFE,
  0xFE,
  0,215,12,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  110,
  111,
  116,
  122,
  0,249,1,
  98,
  0xFE,
  1,2,5,
  98,
  104,
  105,
  110,
  111,
  0xFE,
  1,4,7,
  95,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  1,6,3,
  98,
  104,
  110,
  0xFE,
  1,10,4,
  95,
  98,
  104,
  110,
  0xFE,
  0xFE,
  0,213,15,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  116,
  117,
  122,
  123,
  1,1,0,
  1,52,1,
  98,
  0xFE,
  0xFE,
  0,253,1,
  98,
  0xFE,
  1,2,5,
  98,
  104,
  105,
  110,
  111,
  1,49,3,
  98,
  104,
  110,
  0xFE,
  0xFE,
  1,4,3,
  98,
  104,
  110,
  0xFE,
  1,6,1,
  98,
  0xFE,
  1,10,1,
  98,
  0xFE,
  0xFE,
  0,211,18,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  139,
  144,
  1,1,4,
  98,
  99,
  104,
  110,
  0xFE,
  0,253,1,
  98,
  0xFE,
  0,249,3,
  98,
  104,
  110,
  0xFE,
  1,2,0,
  1,49,4,
  95,
  98,
  104,
  110,
  0xFE,
  0xFE,
  1,46,3,
  98,
  104,
  110,
  0xFE,
  0xFE,
  0,209,25,
  65,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  118,
  122,
  123,
  124,
  127,
  133,
  139,
  140,
  144,
  145,
  0,88,2,
  95,
  98,
  0xFE,
  1,1,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,253,3,
  98,
  104,
  110,
  0xFE,
  0,251,3,
  98,
  104,
  110,
  0xFE,
  0,249,3,
  98,
  104,
  110,
  0xFE,
  1,2,0,
  1,49,1,
  98,
  0xFE,
  0xFE,
  1,44,2,
  104,
  110,
  0xFE,
  0xFE,
  0,207,25,
  95,
  96,
  97,
  98,
  99,
  100,
  101,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  118,
  122,
  123,
  124,
  127,
  128,
  133,
  139,
  144,
  221,
  1,1,10,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  0xFE,
  0,253,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,251,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,249,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0xFE,
  0,205,29,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  118,
  122,
  123,
  124,
  127,
  133,
  139,
  140,
  144,
  145,
  149,
  153,
  157,
  191,
  221,
  1,3,0,
  1,48,3,
  98,
  104,
  110,
  0xFE,
  0xFE,
  1,1,10,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  0xFE,
  0,255,3,
  98,
  104,
  110,
  0xFE,
  0,253,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,251,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,249,5,
  95,
  98,
  99,
  104,
  110,
  0xFE,
  1,45,2,
  203,
  208,
  0xFE,
  1,44,2,
  203,
  208,
  0xFE,
  1,93,2,
  233,
  239,
  0xFE,
  1,92,2,
  233,
  239,
  0xFE,
  0xFE,
  1,3,6,
  98,
  99,
  104,
  105,
  110,
  111,
  1,48,4,
  95,
  98,
  104,
  110,
  0xFE,
  0xFE,
  1,1,16,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  139,
  144,
  0xFE,
  0,255,16,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  169,
  175,
  1,46,2,
  104,
  110,
  0xFE,
  1,48,2,
  104,
  110,
  0xFE,
  0xFE,
  0,253,21,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  123,
  157,
  158,
  169,
  170,
  175,
  176,
  0xFE,
  0,251,20,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  110,
  111,
  116,
  117,
  122,
  123,
  157,
  158,
  163,
  164,
  169,
  175,
  0xFE,
  0,249,23,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  157,
  163,
  169,
  170,
  175,
  176,
  181,
  186,
  191,
  221,
  0xFE,
  1,51,5,
  98,
  104,
  110,
  203,
  208,
  0xFE,
  1,49,8,
  95,
  98,
  99,
  104,
  110,
  191,
  203,
  208,
  0xFE,
  1,47,11,
  95,
  96,
  98,
  99,
  104,
  110,
  191,
  192,
  197,
  203,
  208,
  0xFE,
  1,45,12,
  95,
  98,
  104,
  191,
  197,
  203,
  208,
  213,
  217,
  221,
  233,
  239,
  1,95,6,
  98,
  104,
  110,
  191,
  203,
  208,
  0xFE,
  1,92,2,
  221,
  227,
  0xFE,
  0xFE,
  1,44,9,
  98,
  104,
  110,
  191,
  197,
  203,
  208,
  213,
  217,
  1,99,2,
  233,
  239,
  0xFE,
  1,97,2,
  221,
  227,
  0xFE,
  1,95,2,
  197,
  227,
  0xFE,
  0xFE,
  1,97,4,
  104,
  110,
  233,
  239,
  0xFE,
  1,95,14,
  95,
  96,
  98,
  99,
  104,
  110,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,93,7,
  98,
  221,
  227,
  233,
  239,
  245,
  249,
  0xFE,
  1,92,6,
  221,
  227,
  233,
  239,
  245,
  249,
  0xFE,
  1,94,2,
  233,
  239,
  1,49,7,
  98,
  104,
  110,
  191,
  197,
  203,
  208,
  0xFE,
  1,47,2,
  191,
  197,
  0xFE,
  0xFE,
  0xFE,
  0,142,64,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  34,
  35,
  38,
  39,
  44,
  45,
  50,
  56,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  89,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  118,
  122,
  123,
  124,
  133,
  139,
  140,
  144,
  145,
  163,
  164,
  169,
  170,
  175,
  176,
  197,
  198,
  203,
  204,
  208,
  209,
  227,
  233,
  234,
  239,
  240,
  0,34,0,
  0,93,2,
  34,
  65,
  0xFE,
  0xFE,
  0,36,0,
  0,97,11,
  34,
  35,
  38,
  44,
  65,
  66,
  71,
  77,
  95,
  98,
  99,
  0xFE,
  0xFE,
  0,88,4,
  104,
  110,
  116,
  122,
  0xFE,
  0,245,0,
  1,32,2,
  98,
  99,
  0xFE,
  0xFE,
  0,241,0,
  1,22,2,
  98,
  99,
  0xFE,
  0xFE,
  0,237,0,
  1,12,2,
  98,
  99,
  0xFE,
  0xFE,
  0,223,0,
  1,8,1,
  98,
  0xFE,
  0xFE,
  0,221,0,
  1,4,1,
  98,
  0xFE,
  0xFE,
  0,219,0,
  1,4,1,
  98,
  0xFE,
  1,6,2,
  95,
  98,
  0xFE,
  1,10,3,
  98,
  104,
  110,
  0xFE,
  0xFE,
  0,217,3,
  98,
  104,
  110,
  1,4,4,
  95,
  98,
  104,
  110,
  0xFE,
  1,6,3,
  98,
  104,
  110,
  0xFE,
  1,8,1,
  98,
  0xFE,
  1,10,4,
  95,
  98,
  104,
  110,
  0xFE,
  0xFE,
  0,215,9,
  95,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  1,4,3,
  95,
  98,
  110,
  0xFE,
  1,6,2,
  95,
  98,
  0xFE,
  1,10,2,
  95,
  98,
  0xFE,
  0xFE,
  0,213,12,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  117,
  122,
  123,
  1,4,2,
  95,
  98,
  0xFE,
  0xFE,
  0,211,18,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  139,
  144,
  1,46,2,
  95,
  98,
  0xFE,
  0xFE,
  0,209,20,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  127,
  133,
  139,
  144,
  0,34,0,
  0,93,1,
  98,
  0xFE,
  0xFE,
  1,1,7,
  95,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,253,3,
  98,
  104,
  110,
  0xFE,
  0,251,3,
  98,
  104,
  110,
  0xFE,
  0,249,4,
  95,
  98,
  104,
  110,
  0xFE,
  0xFE,
  0,207,24,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  118,
  122,
  123,
  124,
  127,
  133,
  139,
  140,
  144,
  145,
  1,1,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,253,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,251,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,249,6,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0xFE,
  0,205,24,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  127,
  128,
  133,
  134,
  139,
  144,
  191,
  221,
  1,3,0,
  1,48,4,
  95,
  98,
  104,
  110,
  0xFE,
  0xFE,
  1,1,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,255,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,253,6,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0,251,6,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0,249,4,
  95,
  98,
  104,
  110,
  0xFE,
  0xFE,
  1,5,2,
  104,
  110,
  0xFE,
  1,3,9,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  1,48,2,
  95,
  98,
  0xFE,
  0xFE,
  1,1,14,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  0xFE,
  0,255,13,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  122,
  0xFE,
  0,253,15,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  110,
  111,
  112,
  116,
  122,
  169,
  175,
  0xFE,
  0,251,19,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  157,
  163,
  169,
  170,
  175,
  176,
  0xFE,
  0,249,19,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  110,
  116,
  122,
  157,
  158,
  163,
  164,
  169,
  175,
  181,
  186,
  0xFE,
  1,2,0,
  0,221,1,
  98,
  0xFE,
  0,219,2,
  95,
  98,
  0xFE,
  0,217,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,215,2,
  95,
  98,
  0xFE,
  0,213,2,
  96,
  99,
  1,49,3,
  98,
  104,
  110,
  0xFE,
  0xFE,
  0,211,0,
  1,49,2,
  95,
  98,
  0xFE,
  0xFE,
  0xFE,
  1,51,4,
  95,
  98,
  104,
  110,
  0xFE,
  1,49,6,
  95,
  98,
  104,
  110,
  203,
  208,
  1,94,2,
  95,
  98,
  0xFE,
  0xFE,
  1,47,10,
  95,
  98,
  104,
  110,
  191,
  197,
  203,
  204,
  208,
  209,
  1,94,2,
  191,
  197,
  0xFE,
  0xFE,
  1,45,8,
  95,
  98,
  191,
  192,
  197,
  198,
  203,
  208,
  1,95,2,
  95,
  98,
  0xFE,
  1,92,2,
  221,
  227,
  0xFE,
  0xFE,
  1,44,2,
  203,
  208,
  1,97,2,
  221,
  227,
  0xFE,
  0xFE,
  1,97,4,
  98,
  110,
  233,
  239,
  0xFE,
  1,95,12,
  95,
  98,
  104,
  110,
  191,
  197,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,93,9,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  245,
  249,
  0xFE,
  1,92,2,
  245,
  249,
  0xFE,
  0xFE,
  0,144,53,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  34,
  38,
  44,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  89,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  139,
  140,
  144,
  145,
  163,
  169,
  170,
  175,
  176,
  197,
  203,
  204,
  208,
  209,
  233,
  234,
  239,
  240,
  0,38,0,
  0,101,2,
  34,
  98,
  0xFE,
  0xFE,
  0,44,0,
  0,111,1,
  98,
  0xFE,
  0xFE,
  0,219,2,
  104,
  110,
  1,4,1,
  98,
  0xFE,
  1,10,3,
  95,
  98,
  110,
  0xFE,
  0xFE,
  0,217,4,
  95,
  98,
  104,
  110,
  1,4,2,
  95,
  98,
  0xFE,
  1,6,2,
  95,
  98,
  0xFE,
  1,10,2,
  95,
  98,
  0xFE,
  0xFE,
  0,215,10,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  1,4,1,
  98,
  0xFE,
  0xFE,
  0,213,15,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  0xFE,
  0,211,16,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  139,
  144,
  1,1,4,
  104,
  105,
  110,
  111,
  0xFE,
  0,251,2,
  104,
  110,
  0xFE,
  0,249,2,
  104,
  110,
  0xFE,
  0xFE,
  0,209,18,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  127,
  133,
  1,1,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,253,2,
  95,
  98,
  0xFE,
  0,251,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,249,4,
  95,
  98,
  104,
  110,
  0xFE,
  0xFE,
  0,207,18,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  133,
  139,
  144,
  1,1,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,255,1,
  110,
  0xFE,
  0,253,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,251,6,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0,249,3,
  95,
  98,
  104,
  0xFE,
  0xFE,
  0,205,20,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  127,
  128,
  133,
  134,
  1,3,0,
  1,48,2,
  95,
  98,
  0xFE,
  0xFE,
  1,1,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,255,2,
  95,
  98,
  0xFE,
  0,253,5,
  95,
  96,
  98,
  104,
  110,
  0xFE,
  0,251,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,249,3,
  98,
  104,
  110,
  0xFE,
  0xFE,
  1,5,4,
  95,
  98,
  104,
  110,
  0xFE,
  1,3,9,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  1,1,10,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  110,
  111,
  0xFE,
  0,255,10,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  110,
  111,
  0xFE,
  0,253,11,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  112,
  116,
  122,
  0xFE,
  0,251,15,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  110,
  111,
  116,
  122,
  157,
  169,
  175,
  0xFE,
  0,249,14,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  157,
  158,
  163,
  164,
  0xFE,
  1,51,2,
  95,
  98,
  0xFE,
  1,49,4,
  95,
  98,
  104,
  110,
  0xFE,
  1,47,10,
  95,
  96,
  98,
  104,
  110,
  191,
  203,
  204,
  208,
  209,
  0xFE,
  1,45,4,
  191,
  192,
  197,
  198,
  0xFE,
  1,97,1,
  98,
  0xFE,
  1,95,8,
  95,
  98,
  104,
  110,
  233,
  234,
  239,
  240,
  0xFE,
  1,93,4,
  221,
  222,
  227,
  228,
  0xFE,
  0xFE,
  0,146,31,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  71,
  77,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  139,
  144,
  169,
  175,
  203,
  208,
  233,
  239,
  0,219,4,
  95,
  98,
  104,
  110,
  1,10,1,
  98,
  0xFE,
  0xFE,
  0,217,7,
  95,
  98,
  104,
  105,
  110,
  111,
  122,
  0xFE,
  0,215,11,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  0xFE,
  0,213,12,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  110,
  111,
  116,
  122,
  1,1,2,
  104,
  110,
  0xFE,
  0xFE,
  0,211,16,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  1,1,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,253,1,
  98,
  0xFE,
  0,251,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,249,4,
  95,
  98,
  104,
  110,
  0xFE,
  0xFE,
  0,209,16,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  1,1,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,253,1,
  110,
  0xFE,
  0,251,5,
  95,
  96,
  98,
  104,
  110,
  0xFE,
  0,249,3,
  95,
  98,
  104,
  0xFE,
  0xFE,
  0,207,16,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  1,1,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,255,1,
  98,
  0xFE,
  0,253,2,
  95,
  98,
  0xFE,
  0,251,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,249,2,
  104,
  110,
  0xFE,
  0xFE,
  0,205,13,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  127,
  133,
  1,1,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,255,1,
  98,
  0xFE,
  0,253,3,
  95,
  98,
  104,
  0xFE,
  0,251,4,
  95,
  98,
  104,
  110,
  0xFE,
  0xFE,
  1,5,5,
  95,
  96,
  98,
  104,
  110,
  0xFE,
  1,3,6,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  1,1,6,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0,255,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,253,9,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,251,9,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,249,5,
  95,
  104,
  110,
  157,
  163,
  0xFE,
  1,47,2,
  104,
  110,
  0xFE,
  1,45,2,
  191,
  197,
  0xFE,
  1,93,2,
  221,
  227,
  0xFE,
  0xFE,
  0,148,12,
  3,
  7,
  13,
  95,
  96,
  98,
  104,
  105,
  110,
  111,
  116,
  122,
  0,221,3,
  98,
  104,
  110,
  0xFE,
  0,219,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,217,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,215,11,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  0xFE,
  0,213,12,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  112,
  116,
  122,
  0xFE,
  0,211,12,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  110,
  111,
  116,
  122,
  0,249,1,
  98,
  0xFE,
  0xFE,
  0,209,14,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  0,253,1,
  98,
  0xFE,
  0xFE,
  0,207,13,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  0xFE,
  0,205,7,
  95,
  96,
  98,
  104,
  105,
  110,
  111,
  0xFE,
  1,5,3,
  95,
  98,
  110,
  0xFE,
  1,3,4,
  95,
  96,
  98,
  110,
  0xFE,
  1,1,4,
  95,
  98,
  104,
  110,
  0,213,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,211,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,209,7,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  0xFE,
  0,207,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,205,6,
  95,
  98,
  104,
  105,
  110,
  111,
  0xFE,
  0xFE,
  0,255,6,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0,253,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,251,7,
  95,
  96,
  98,
  104,
  105,
  110,
  111,
  0,213,1,
  98,
  0xFE,
  0,211,3,
  95,
  98,
  104,
  0xFE,
  0,209,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,207,3,
  95,
  104,
  110,
  0xFE,
  0xFE,
  0xFE,
  0,150,2,
  104,
  110,
  0,223,2,
  104,
  110,
  0xFE,
  0,221,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,219,5,
  95,
  96,
  98,
  104,
  110,
  0xFE,
  0,217,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,215,10,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  0xFE,
  0,213,10,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  110,
  111,
  0xFE,
  0,211,13,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  116,
  122,
  0xFE,
  0,209,13,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  0xFE,
  0,207,8,
  95,
  96,
  97,
  98,
  104,
  105,
  110,
  111,
  0xFE,
  1,5,1,
  98,
  0xFE,
  1,3,4,
  95,
  98,
  104,
  110,
  0xFE,
  1,1,4,
  95,
  98,
  104,
  110,
  0,213,6,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0,211,5,
  95,
  96,
  98,
  99,
  104,
  0xFE,
  0,209,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,207,4,
  95,
  98,
  104,
  110,
  0xFE,
  0xFE,
  0,255,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,253,2,
  104,
  110,
  0xFE,
  0,251,2,
  104,
  110,
  0,213,1,
  98,
  0xFE,
  0,211,2,
  95,
  98,
  0xFE,
  0xFE,
  0xFE,
  0,152,0,
  0,223,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,221,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,219,5,
  95,
  96,
  98,
  104,
  110,
  0xFE,
  0,217,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,215,8,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  0xFE,
  0,213,9,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,211,9,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,209,7,
  95,
  96,
  98,
  104,
  105,
  110,
  111,
  0xFE,
  0,207,2,
  104,
  110,
  0xFE,
  1,1,0,
  0,215,2,
  95,
  98,
  0xFE,
  0,213,5,
  95,
  96,
  98,
  99,
  104,
  0xFE,
  0,211,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,209,2,
  104,
  110,
  0xFE,
  0xFE,
  0xFE,
  0,154,0,
  0,223,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,221,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,219,5,
  95,
  96,
  98,
  104,
  110,
  0xFE,
  0,217,7,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  0xFE,
  0,215,9,
  95,
  96,
  97,
  98,
  99,
  104,
  105,
  110,
  111,
  1,1,1,
  98,
  0xFE,
  0xFE,
  0,213,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,211,7,
  95,
  96,
  98,
  104,
  105,
  110,
  111,
  0xFE,
  0,209,2,
  104,
  110,
  0xFE,
  0xFE,
  0,156,0,
  0,225,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,223,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,221,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,219,6,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0,217,7,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  0xFE,
  0,215,5,
  95,
  96,
  98,
  104,
  110,
  0xFE,
  0,213,2,
  104,
  110,
  0xFE,
  0xFE,
  0,158,0,
  0,225,2,
  95,
  98,
  0xFE,
  0,223,3,
  95,
  98,
  104,
  0xFE,
  0,221,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,219,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,217,2,
  104,
  110,
  0xFE,
  0xFE,
  0,160,0,
  0,225,2,
  95,
  98,
  0xFE,
  0,223,2,
  95,
  98,
  0xFE,
  0,221,1,
  104,
  0xFE,
  0xFE,
  0,204,136,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  20,
  21,
  25,
  26,
  27,
  34,
  35,
  36,
  38,
  39,
  44,
  45,
  50,
  51,
  52,
  56,
  57,
  65,
  66,
  67,
  71,
  72,
  77,
  78,
  79,
  83,
  84,
  85,
  89,
  90,
  95,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  118,
  122,
  123,
  124,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  149,
  150,
  151,
  153,
  154,
  155,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  183,
  184,
  186,
  187,
  188,
  189,
  191,
  192,
  193,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  215,
  217,
  218,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  246,
  249,
  250,
  0,0,17,
  0,
  1,
  3,
  4,
  5,
  7,
  8,
  13,
  14,
  19,
  25,
  127,
  133,
  139,
  144,
  149,
  153,
  0,37,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,35,1,
  3,
  0,87,1,
  3,
  0xFE,
  0xFE,
  0,93,4,
  3,
  4,
  7,
  13,
  0xFE,
  0,91,3,
  3,
  7,
  13,
  0xFE,
  0,155,1,
  3,
  0xFE,
  1,93,2,
  233,
  239,
  0xFE,
  0xFE,
  0,2,15,
  0,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  127,
  133,
  139,
  144,
  149,
  153,
  0,45,0,
  0,90,1,
  3,
  0xFE,
  0xFE,
  0,41,5,
  0,
  3,
  4,
  7,
  13,
  0xFE,
  0,39,1,
  3,
  0xFE,
  0,37,1,
  3,
  0xFE,
  0xFE,
  0,4,6,
  0,
  3,
  7,
  13,
  19,
  25,
  0,47,1,
  3,
  0xFE,
  0,45,1,
  3,
  0xFE,
  0xFE,
  0,6,0,
  0,51,1,
  3,
  0xFE,
  0xFE,
  0,32,26,
  31,
  32,
  34,
  35,
  38,
  39,
  44,
  45,
  50,
  51,
  56,
  57,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  149,
  150,
  153,
  154,
  203,
  208,
  0,89,1,
  34,
  0xFE,
  0,147,1,
  98,
  0xFE,
  0,145,3,
  98,
  104,
  110,
  0xFE,
  0,143,1,
  95,
  0xFE,
  1,44,0,
  1,99,1,
  221,
  0xFE,
  0xFE,
  1,97,1,
  221,
  0xFE,
  1,93,2,
  233,
  239,
  0xFE,
  0xFE,
  0,34,2,
  50,
  56,
  0,97,0,
  0,148,1,
  98,
  0xFE,
  0xFE,
  0,93,4,
  38,
  44,
  71,
  77,
  0xFE,
  0,91,4,
  31,
  34,
  62,
  65,
  0xFE,
  0,151,1,
  98,
  0xFE,
  0xFE,
  0,36,0,
  0,113,0,
  0,172,1,
  98,
  0xFE,
  0xFE,
  0,95,10,
  31,
  34,
  35,
  38,
  44,
  62,
  65,
  66,
  71,
  77,
  0xFE,
  0,93,3,
  34,
  62,
  65,
  0xFE,
  0xFE,
  0,38,0,
  0,97,4,
  38,
  44,
  71,
  77,
  0xFE,
  0,95,1,
  34,
  0xFE,
  0xFE,
  0,40,0,
  0,101,1,
  34,
  0xFE,
  0xFE,
  0,86,26,
  62,
  63,
  65,
  66,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  84,
  89,
  90,
  127,
  133,
  139,
  140,
  144,
  145,
  149,
  150,
  153,
  154,
  203,
  208,
  0,145,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,143,5,
  65,
  95,
  98,
  104,
  110,
  0xFE,
  1,97,1,
  221,
  0xFE,
  1,93,2,
  233,
  239,
  0xFE,
  0xFE,
  0,88,6,
  71,
  77,
  83,
  89,
  139,
  144,
  0,149,2,
  104,
  110,
  0xFE,
  0,147,11,
  62,
  65,
  66,
  71,
  77,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0,145,2,
  95,
  98,
  0xFE,
  0xFE,
  0,90,0,
  0,151,11,
  65,
  66,
  71,
  77,
  95,
  96,
  98,
  99,
  104,
  110,
  111,
  0xFE,
  0,149,5,
  65,
  95,
  98,
  104,
  110,
  0xFE,
  0,147,2,
  95,
  98,
  0xFE,
  0xFE,
  0,92,0,
  0,155,7,
  65,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0,153,5,
  95,
  98,
  99,
  104,
  110,
  0xFE,
  0,151,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,149,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,147,2,
  104,
  110,
  0xFE,
  0xFE,
  0,94,0,
  0,159,3,
  98,
  104,
  110,
  0xFE,
  0,157,5,
  95,
  98,
  99,
  104,
  110,
  0xFE,
  0,155,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,153,2,
  95,
  98,
  0xFE,
  0xFE,
  0,96,0,
  0,161,5,
  95,
  98,
  99,
  104,
  110,
  0xFE,
  0,159,2,
  104,
  110,
  0xFE,
  0,155,2,
  65,
  98,
  0xFE,
  0xFE,
  0,98,0,
  0,165,4,
  98,
  99,
  104,
  110,
  0xFE,
  0xFE,
  0,100,0,
  0,169,2,
  98,
  99,
  0xFE,
  0xFE,
  0,102,0,
  0,173,2,
  98,
  99,
  0xFE,
  0xFE,
  0,104,0,
  0,177,4,
  65,
  66,
  98,
  99,
  0xFE,
  0xFE,
  0,140,23,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  124,
  127,
  133,
  139,
  144,
  149,
  153,
  1,93,2,
  233,
  239,
  0xFE,
  0xFE,
  0,142,12,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  117,
  122,
  123,
  0xFE,
  0,144,4,
  104,
  110,
  116,
  122,
  0xFE,
  1,11,2,
  7,
  13,
  0,140,3,
  98,
  104,
  110,
  0xFE,
  0,142,3,
  98,
  104,
  110,
  0xFE,
  0,144,2,
  95,
  98,
  0xFE,
  1,54,1,
  191,
  0xFE,
  1,56,3,
  191,
  197,
  208,
  0xFE,
  1,58,7,
  157,
  163,
  191,
  197,
  198,
  203,
  208,
  0xFE,
  1,60,5,
  163,
  191,
  197,
  203,
  208,
  0xFE,
  1,62,4,
  191,
  197,
  203,
  208,
  0xFE,
  0xFE,
  1,7,3,
  144,
  175,
  176,
  0xFE,
  1,5,28,
  7,
  13,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  181,
  203,
  208,
  233,
  239,
  0,140,2,
  104,
  110,
  0xFE,
  0,142,1,
  98,
  0xFE,
  1,47,0,
  1,104,1,
  157,
  0xFE,
  0xFE,
  1,44,2,
  191,
  197,
  0xFE,
  1,46,2,
  191,
  192,
  0xFE,
  1,48,25,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  191,
  192,
  197,
  203,
  208,
  0xFE,
  1,50,11,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  208,
  209,
  0xFE,
  1,52,12,
  157,
  163,
  164,
  169,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,54,12,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,56,11,
  157,
  163,
  169,
  175,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,58,7,
  163,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,60,2,
  203,
  208,
  0xFE,
  1,92,4,
  221,
  222,
  227,
  228,
  0xFE,
  1,100,2,
  221,
  227,
  0xFE,
  1,102,2,
  233,
  239,
  0xFE,
  0xFE,
  1,3,35,
  127,
  128,
  129,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  153,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  186,
  187,
  1,47,0,
  1,104,4,
  127,
  133,
  157,
  163,
  0xFE,
  0xFE,
  1,44,4,
  203,
  204,
  208,
  209,
  0xFE,
  1,46,11,
  163,
  169,
  175,
  191,
  192,
  193,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,48,39,
  3,
  7,
  104,
  110,
  127,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  163,
  164,
  169,
  170,
  171,
  175,
  176,
  177,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  227,
  233,
  234,
  235,
  239,
  240,
  241,
  0,140,1,
  98,
  0xFE,
  1,92,6,
  221,
  222,
  223,
  227,
  228,
  229,
  0xFE,
  0xFE,
  1,50,22,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  208,
  209,
  210,
  233,
  239,
  1,93,4,
  221,
  222,
  227,
  228,
  0xFE,
  0xFE,
  1,52,20,
  157,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  1,99,6,
  169,
  175,
  203,
  208,
  233,
  239,
  0xFE,
  1,97,2,
  221,
  227,
  0xFE,
  0xFE,
  1,54,12,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  1,103,6,
  169,
  175,
  203,
  208,
  233,
  239,
  0xFE,
  1,101,2,
  221,
  227,
  0xFE,
  1,99,4,
  191,
  197,
  233,
  239,
  0xFE,
  0xFE,
  1,56,11,
  157,
  163,
  169,
  175,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  1,107,6,
  169,
  175,
  203,
  208,
  233,
  239,
  0xFE,
  1,105,2,
  221,
  227,
  0xFE,
  1,103,4,
  191,
  197,
  233,
  239,
  0xFE,
  1,99,3,
  191,
  203,
  208,
  0xFE,
  0xFE,
  1,58,7,
  163,
  191,
  197,
  203,
  204,
  208,
  209,
  1,107,2,
  191,
  197,
  0xFE,
  1,103,2,
  203,
  208,
  0xFE,
  1,99,2,
  203,
  208,
  0xFE,
  0xFE,
  1,60,2,
  203,
  208,
  0xFE,
  1,62,7,
  163,
  164,
  191,
  197,
  198,
  203,
  208,
  0xFE,
  1,74,0,
  1,109,1,
  163,
  0xFE,
  0xFE,
  1,92,4,
  233,
  234,
  239,
  240,
  0xFE,
  1,96,5,
  133,
  157,
  163,
  221,
  227,
  0xFE,
  1,98,6,
  157,
  163,
  221,
  227,
  233,
  239,
  0xFE,
  1,100,4,
  221,
  227,
  233,
  239,
  0xFE,
  1,102,2,
  233,
  239,
  0xFE,
  0xFE,
  1,1,106,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  26,
  34,
  35,
  38,
  39,
  44,
  45,
  65,
  66,
  71,
  72,
  77,
  78,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  127,
  128,
  129,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  173,
  175,
  176,
  177,
  178,
  179,
  181,
  182,
  186,
  187,
  191,
  197,
  198,
  199,
  200,
  201,
  203,
  204,
  205,
  206,
  207,
  208,
  209,
  210,
  211,
  212,
  213,
  214,
  221,
  222,
  227,
  228,
  229,
  230,
  233,
  234,
  235,
  236,
  237,
  238,
  239,
  240,
  241,
  242,
  243,
  244,
  0,0,10,
  0,
  3,
  4,
  7,
  8,
  13,
  127,
  128,
  133,
  134,
  0xFE,
  0,2,2,
  0,
  3,
  0xFE,
  0,32,15,
  31,
  32,
  34,
  35,
  38,
  39,
  44,
  127,
  128,
  133,
  134,
  139,
  144,
  169,
  175,
  0xFE,
  0,34,4,
  31,
  32,
  34,
  35,
  0xFE,
  0,86,16,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  127,
  128,
  133,
  134,
  139,
  144,
  169,
  175,
  0xFE,
  0,88,4,
  62,
  63,
  65,
  66,
  0xFE,
  0,90,2,
  71,
  77,
  0xFE,
  0,98,0,
  0,165,1,
  98,
  0xFE,
  0xFE,
  0,140,13,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  122,
  127,
  128,
  133,
  134,
  0xFE,
  0,142,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,144,6,
  95,
  98,
  104,
  105,
  110,
  111,
  0xFE,
  1,47,2,
  203,
  208,
  1,96,1,
  221,
  0xFE,
  1,98,3,
  197,
  227,
  228,
  0xFE,
  1,100,6,
  221,
  222,
  227,
  228,
  233,
  239,
  0xFE,
  1,102,2,
  233,
  239,
  0xFE,
  1,104,7,
  133,
  139,
  144,
  157,
  163,
  169,
  175,
  0xFE,
  0xFE,
  1,45,16,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  215,
  217,
  218,
  219,
  1,96,10,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  239,
  240,
  0xFE,
  1,98,11,
  191,
  192,
  197,
  198,
  221,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  1,100,2,
  233,
  239,
  0xFE,
  0xFE,
  1,44,41,
  133,
  134,
  169,
  175,
  191,
  192,
  193,
  194,
  195,
  196,
  197,
  198,
  199,
  200,
  201,
  202,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  214,
  215,
  217,
  218,
  219,
  221,
  222,
  223,
  227,
  228,
  233,
  234,
  235,
  239,
  240,
  241,
  1,93,10,
  221,
  222,
  223,
  224,
  227,
  228,
  229,
  230,
  233,
  239,
  0xFE,
  1,92,4,
  233,
  234,
  239,
  240,
  0xFE,
  0xFE,
  1,46,31,
  127,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  163,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  207,
  208,
  209,
  210,
  211,
  212,
  1,97,4,
  221,
  222,
  227,
  228,
  0xFE,
  1,93,3,
  191,
  233,
  239,
  0xFE,
  0xFE,
  1,48,35,
  127,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  181,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  217,
  0xFE,
  1,50,20,
  139,
  144,
  157,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,52,15,
  139,
  144,
  157,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,54,16,
  133,
  139,
  144,
  157,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  1,103,2,
  157,
  163,
  0xFE,
  0xFE,
  1,56,15,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  1,107,2,
  157,
  163,
  0xFE,
  0xFE,
  1,58,3,
  198,
  203,
  208,
  0xFE,
  1,74,0,
  1,109,2,
  157,
  163,
  0xFE,
  0xFE,
  1,99,0,
  1,48,4,
  191,
  192,
  197,
  198,
  0xFE,
  1,50,5,
  191,
  203,
  204,
  208,
  209,
  0xFE,
  1,52,3,
  133,
  157,
  163,
  0xFE,
  0xFE,
  1,93,16,
  221,
  222,
  227,
  228,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  246,
  247,
  249,
  250,
  251,
  0xFE,
  1,92,24,
  221,
  222,
  223,
  224,
  225,
  226,
  227,
  228,
  229,
  230,
  231,
  232,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  246,
  247,
  249,
  250,
  251,
  0xFE,
  1,96,13,
  157,
  163,
  169,
  175,
  221,
  227,
  228,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  1,98,14,
  157,
  163,
  169,
  175,
  203,
  204,
  208,
  209,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,100,2,
  233,
  239,
  0xFE,
  0xFE,
  0,255,92,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  34,
  65,
  71,
  77,
  98,
  104,
  105,
  110,
  111,
  122,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  137,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  149,
  150,
  151,
  153,
  154,
  155,
  157,
  158,
  159,
  160,
  161,
  162,
  163,
  164,
  165,
  166,
  167,
  168,
  169,
  170,
  171,
  172,
  173,
  175,
  176,
  177,
  178,
  179,
  181,
  182,
  183,
  186,
  187,
  188,
  191,
  192,
  193,
  197,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  221,
  222,
  223,
  224,
  227,
  233,
  234,
  235,
  239,
  240,
  241,
  0,0,3,
  3,
  7,
  13,
  0xFE,
  0,2,1,
  3,
  0xFE,
  0,32,8,
  38,
  44,
  157,
  163,
  169,
  170,
  175,
  176,
  0xFE,
  0,34,1,
  34,
  0xFE,
  0,86,8,
  71,
  77,
  157,
  163,
  169,
  170,
  175,
  176,
  0xFE,
  0,88,2,
  62,
  65,
  0xFE,
  0,140,7,
  95,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,142,4,
  95,
  96,
  98,
  99,
  0xFE,
  1,49,2,
  169,
  175,
  0xFE,
  1,47,6,
  191,
  192,
  193,
  197,
  198,
  199,
  1,104,7,
  133,
  139,
  144,
  157,
  163,
  169,
  175,
  0xFE,
  0xFE,
  1,45,11,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  210,
  0xFE,
  1,44,21,
  139,
  144,
  157,
  163,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  217,
  233,
  1,93,1,
  227,
  0xFE,
  1,96,1,
  191,
  0xFE,
  0xFE,
  1,46,25,
  127,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  1,92,2,
  221,
  227,
  0xFE,
  1,98,1,
  197,
  0xFE,
  0xFE,
  1,48,38,
  127,
  128,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  186,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,50,27,
  127,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  1,97,2,
  191,
  197,
  0xFE,
  0xFE,
  1,52,16,
  133,
  139,
  157,
  163,
  164,
  169,
  175,
  176,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  1,99,8,
  133,
  139,
  144,
  157,
  163,
  164,
  169,
  175,
  0xFE,
  1,97,2,
  203,
  208,
  0xFE,
  1,93,1,
  197,
  0xFE,
  0xFE,
  1,54,6,
  133,
  157,
  163,
  169,
  203,
  208,
  1,103,7,
  133,
  139,
  144,
  157,
  163,
  169,
  175,
  0xFE,
  1,99,1,
  197,
  0xFE,
  1,97,3,
  191,
  197,
  227,
  0xFE,
  0xFE,
  1,74,0,
  1,113,2,
  233,
  239,
  0xFE,
  1,111,2,
  221,
  227,
  0xFE,
  0xFE,
  1,109,0,
  1,68,2,
  203,
  208,
  0xFE,
  1,70,6,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,72,5,
  191,
  192,
  197,
  198,
  208,
  0xFE,
  0xFE,
  1,93,7,
  221,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  1,92,16,
  139,
  144,
  169,
  175,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  1,96,10,
  157,
  163,
  169,
  175,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,98,6,
  163,
  169,
  175,
  221,
  233,
  239,
  0xFE,
  1,100,2,
  233,
  239,
  0xFE,
  0xFE,
  0,253,107,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  34,
  35,
  38,
  39,
  44,
  45,
  50,
  56,
  65,
  66,
  71,
  72,
  77,
  78,
  89,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  127,
  128,
  129,
  130,
  131,
  133,
  134,
  135,
  136,
  137,
  139,
  140,
  141,
  142,
  143,
  144,
  145,
  146,
  147,
  148,
  149,
  150,
  151,
  152,
  153,
  154,
  155,
  156,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  172,
  173,
  175,
  176,
  177,
  178,
  179,
  181,
  182,
  183,
  184,
  186,
  187,
  188,
  189,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  221,
  222,
  223,
  227,
  233,
  234,
  235,
  239,
  240,
  0,0,3,
  3,
  157,
  163,
  0xFE,
  0,32,7,
  127,
  133,
  157,
  163,
  164,
  169,
  175,
  0xFE,
  0,86,10,
  65,
  71,
  77,
  127,
  133,
  157,
  163,
  164,
  169,
  175,
  0xFE,
  0,88,2,
  62,
  65,
  0xFE,
  0,140,7,
  95,
  98,
  99,
  104,
  110,
  157,
  163,
  0xFE,
  0,142,2,
  95,
  98,
  0xFE,
  1,49,9,
  139,
  144,
  169,
  170,
  175,
  176,
  191,
  203,
  208,
  0xFE,
  1,47,11,
  163,
  164,
  169,
  175,
  191,
  192,
  198,
  203,
  204,
  208,
  209,
  1,104,8,
  127,
  133,
  157,
  158,
  163,
  164,
  169,
  175,
  0xFE,
  0xFE,
  1,45,13,
  157,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  1,96,2,
  221,
  227,
  0xFE,
  1,98,6,
  163,
  197,
  221,
  227,
  233,
  239,
  0xFE,
  1,100,4,
  163,
  227,
  233,
  239,
  0xFE,
  0xFE,
  1,44,29,
  127,
  128,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  165,
  169,
  175,
  191,
  192,
  193,
  197,
  198,
  199,
  200,
  203,
  204,
  208,
  209,
  213,
  217,
  227,
  233,
  239,
  1,93,2,
  221,
  227,
  0xFE,
  1,92,8,
  133,
  157,
  163,
  197,
  221,
  227,
  233,
  239,
  0xFE,
  0xFE,
  1,46,32,
  127,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  181,
  186,
  191,
  192,
  193,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  227,
  1,97,2,
  203,
  208,
  0xFE,
  1,93,5,
  133,
  163,
  197,
  227,
  233,
  0xFE,
  0xFE,
  1,48,31,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  186,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  227,
  1,99,1,
  197,
  0xFE,
  1,97,6,
  133,
  163,
  191,
  197,
  221,
  227,
  0xFE,
  0xFE,
  1,50,13,
  133,
  157,
  158,
  163,
  164,
  169,
  175,
  176,
  191,
  192,
  197,
  203,
  208,
  1,99,2,
  163,
  197,
  0xFE,
  1,97,2,
  203,
  208,
  0xFE,
  0xFE,
  1,52,8,
  157,
  163,
  164,
  169,
  175,
  191,
  197,
  204,
  1,99,6,
  139,
  144,
  157,
  163,
  169,
  175,
  0xFE,
  0xFE,
  1,54,3,
  157,
  191,
  208,
  1,103,3,
  144,
  169,
  175,
  0xFE,
  0xFE,
  1,93,10,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,92,23,
  127,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  175,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,96,20,
  127,
  133,
  134,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  197,
  221,
  222,
  227,
  228,
  229,
  233,
  234,
  239,
  240,
  0xFE,
  1,98,7,
  163,
  169,
  175,
  233,
  234,
  239,
  240,
  0xFE,
  0xFE,
  0,251,108,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  26,
  34,
  35,
  38,
  44,
  50,
  65,
  66,
  71,
  77,
  83,
  89,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  137,
  139,
  140,
  141,
  142,
  143,
  144,
  145,
  146,
  147,
  148,
  149,
  150,
  151,
  152,
  153,
  154,
  155,
  156,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  172,
  173,
  175,
  176,
  177,
  178,
  179,
  181,
  182,
  183,
  184,
  186,
  187,
  188,
  189,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  221,
  222,
  223,
  227,
  228,
  233,
  234,
  235,
  239,
  240,
  241,
  0,0,1,
  157,
  0xFE,
  0,2,1,
  157,
  0xFE,
  0,32,7,
  34,
  127,
  133,
  157,
  163,
  169,
  175,
  0xFE,
  0,86,7,
  65,
  127,
  133,
  157,
  163,
  169,
  175,
  0xFE,
  0,88,1,
  157,
  0xFE,
  0,140,5,
  95,
  98,
  104,
  110,
  157,
  0xFE,
  0,142,2,
  104,
  110,
  0xFE,
  1,59,0,
  1,116,1,
  197,
  0xFE,
  0xFE,
  1,57,0,
  1,112,1,
  197,
  0xFE,
  0xFE,
  1,51,0,
  1,100,1,
  227,
  0xFE,
  0xFE,
  1,49,23,
  127,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  169,
  170,
  171,
  175,
  176,
  181,
  186,
  191,
  198,
  203,
  208,
  0xFE,
  1,47,27,
  127,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  191,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  1,92,1,
  227,
  0xFE,
  0xFE,
  1,45,25,
  127,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  200,
  203,
  204,
  208,
  209,
  213,
  217,
  0xFE,
  1,44,15,
  127,
  133,
  139,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  203,
  204,
  205,
  208,
  209,
  1,97,2,
  203,
  208,
  0xFE,
  1,93,2,
  221,
  227,
  0xFE,
  0xFE,
  1,46,21,
  127,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  203,
  208,
  209,
  1,97,2,
  191,
  197,
  0xFE,
  1,93,2,
  221,
  227,
  0xFE,
  0xFE,
  1,48,24,
  127,
  133,
  134,
  139,
  140,
  144,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,50,9,
  191,
  192,
  193,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,52,6,
  191,
  192,
  197,
  198,
  203,
  208,
  1,99,1,
  157,
  0xFE,
  0xFE,
  1,97,8,
  157,
  163,
  227,
  228,
  233,
  234,
  239,
  240,
  0xFE,
  1,93,19,
  133,
  139,
  144,
  157,
  163,
  169,
  175,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,92,10,
  133,
  157,
  163,
  169,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,96,4,
  169,
  175,
  233,
  239,
  1,49,3,
  191,
  197,
  227,
  0xFE,
  1,45,4,
  191,
  221,
  227,
  233,
  0xFE,
  0xFE,
  0xFE,
  0,249,88,
  3,
  7,
  13,
  19,
  25,
  34,
  38,
  44,
  56,
  65,
  71,
  77,
  98,
  104,
  105,
  110,
  111,
  122,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  149,
  150,
  151,
  153,
  154,
  155,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  183,
  184,
  186,
  187,
  188,
  189,
  191,
  192,
  193,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  249,
  0,0,3,
  3,
  169,
  175,
  0xFE,
  0,32,3,
  163,
  169,
  175,
  0xFE,
  0,86,3,
  163,
  169,
  175,
  0,143,1,
  98,
  0xFE,
  0xFE,
  0,140,6,
  95,
  98,
  104,
  110,
  169,
  175,
  0xFE,
  1,61,0,
  1,108,1,
  197,
  0xFE,
  0xFE,
  1,55,0,
  1,102,2,
  169,
  175,
  0xFE,
  0xFE,
  1,53,3,
  163,
  203,
  208,
  1,98,3,
  163,
  169,
  175,
  0xFE,
  0xFE,
  1,51,17,
  133,
  157,
  158,
  163,
  164,
  165,
  169,
  175,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  0xFE,
  1,49,29,
  127,
  128,
  133,
  134,
  135,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,47,26,
  127,
  133,
  139,
  144,
  145,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  0xFE,
  1,45,25,
  133,
  139,
  144,
  157,
  158,
  163,
  169,
  175,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  213,
  214,
  217,
  218,
  221,
  233,
  239,
  1,92,4,
  221,
  227,
  233,
  239,
  0xFE,
  1,96,2,
  233,
  239,
  0xFE,
  0xFE,
  1,44,22,
  133,
  139,
  144,
  157,
  163,
  164,
  169,
  175,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  0xFE,
  1,46,26,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  208,
  209,
  210,
  213,
  217,
  0xFE,
  1,48,22,
  127,
  133,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  203,
  204,
  208,
  209,
  213,
  217,
  233,
  239,
  0xFE,
  1,50,6,
  203,
  204,
  208,
  209,
  213,
  217,
  1,103,2,
  203,
  208,
  0xFE,
  0xFE,
  1,52,0,
  1,103,4,
  191,
  197,
  233,
  239,
  0xFE,
  1,101,2,
  221,
  227,
  0xFE,
  0xFE,
  1,54,0,
  1,105,1,
  221,
  0xFE,
  0xFE,
  1,99,0,
  1,44,2,
  233,
  239,
  0xFE,
  1,46,2,
  203,
  208,
  0xFE,
  1,48,3,
  191,
  203,
  208,
  0xFE,
  1,50,5,
  191,
  192,
  197,
  233,
  239,
  0xFE,
  1,52,2,
  197,
  227,
  0xFE,
  0xFE,
  1,97,6,
  169,
  175,
  233,
  234,
  239,
  240,
  1,44,5,
  221,
  222,
  227,
  233,
  239,
  0xFE,
  1,46,1,
  221,
  0xFE,
  1,50,4,
  221,
  222,
  227,
  228,
  0xFE,
  0xFE,
  1,93,14,
  157,
  163,
  169,
  175,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  239,
  245,
  249,
  1,44,4,
  221,
  227,
  233,
  239,
  0xFE,
  1,48,2,
  221,
  227,
  0xFE,
  0xFE,
  1,92,11,
  157,
  163,
  169,
  175,
  221,
  222,
  227,
  233,
  239,
  245,
  249,
  0xFE,
  0xFE,
  0,248,81,
  3,
  4,
  7,
  13,
  19,
  25,
  34,
  65,
  71,
  77,
  98,
  104,
  110,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  149,
  150,
  151,
  153,
  154,
  155,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  172,
  173,
  175,
  176,
  177,
  178,
  179,
  181,
  182,
  183,
  186,
  187,
  188,
  191,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  227,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  249,
  0,36,0,
  0,95,2,
  34,
  65,
  0xFE,
  0xFE,
  0,88,0,
  0,147,1,
  98,
  0xFE,
  0xFE,
  1,71,0,
  1,106,1,
  197,
  0xFE,
  0xFE,
  1,63,1,
  197,
  0xFE,
  1,57,5,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,55,13,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,53,18,
  133,
  139,
  144,
  157,
  163,
  169,
  175,
  191,
  192,
  193,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  1,96,2,
  221,
  227,
  0xFE,
  0xFE,
  1,51,23,
  157,
  163,
  164,
  169,
  175,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  233,
  239,
  1,92,2,
  221,
  227,
  0xFE,
  0xFE,
  1,49,27,
  127,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  213,
  217,
  1,93,1,
  221,
  0xFE,
  0xFE,
  1,47,19,
  127,
  133,
  157,
  163,
  169,
  175,
  191,
  192,
  193,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  1,92,1,
  221,
  0xFE,
  0xFE,
  1,45,17,
  157,
  169,
  175,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  208,
  213,
  217,
  221,
  222,
  227,
  233,
  1,93,1,
  221,
  0xFE,
  1,92,4,
  221,
  227,
  233,
  239,
  0xFE,
  0xFE,
  1,44,1,
  163,
  1,101,1,
  221,
  0xFE,
  1,97,2,
  233,
  239,
  0xFE,
  0xFE,
  1,46,0,
  1,103,1,
  221,
  0xFE,
  0xFE,
  1,48,1,
  163,
  0xFE,
  1,99,4,
  191,
  221,
  233,
  239,
  1,49,2,
  203,
  208,
  0xFE,
  1,47,2,
  191,
  197,
  0xFE,
  1,45,2,
  191,
  197,
  0xFE,
  1,44,3,
  221,
  222,
  227,
  0xFE,
  0xFE,
  1,97,8,
  133,
  157,
  163,
  221,
  222,
  227,
  233,
  239,
  0xFE,
  1,93,8,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  239,
  0xFE,
  1,98,0,
  1,59,2,
  203,
  208,
  0xFE,
  1,57,3,
  191,
  203,
  208,
  0xFE,
  1,55,2,
  191,
  197,
  0xFE,
  1,53,3,
  169,
  175,
  197,
  0xFE,
  1,51,1,
  191,
  0xFE,
  0xFE,
  1,102,0,
  1,69,1,
  197,
  0xFE,
  1,59,2,
  203,
  208,
  0xFE,
  1,57,2,
  191,
  197,
  0xFE,
  1,55,2,
  169,
  175,
  0xFE,
  0xFE,
  0xFE,
  0,250,81,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  38,
  44,
  71,
  77,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  127,
  128,
  129,
  130,
  131,
  133,
  134,
  135,
  136,
  137,
  139,
  140,
  141,
  144,
  145,
  146,
  147,
  149,
  150,
  153,
  154,
  155,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  183,
  186,
  187,
  188,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  227,
  233,
  234,
  235,
  239,
  240,
  241,
  0,88,0,
  0,147,1,
  98,
  0xFE,
  0xFE,
  1,91,0,
  1,134,1,
  197,
  0xFE,
  0xFE,
  1,75,0,
  1,108,1,
  163,
  0xFE,
  0xFE,
  1,59,4,
  203,
  204,
  208,
  209,
  0xFE,
  1,57,8,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,55,12,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  1,102,3,
  163,
  169,
  175,
  0xFE,
  0xFE,
  1,53,12,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  1,98,5,
  133,
  157,
  163,
  169,
  175,
  0xFE,
  0xFE,
  1,51,10,
  157,
  163,
  164,
  169,
  191,
  192,
  197,
  198,
  203,
  208,
  0xFE,
  1,49,26,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  163,
  164,
  169,
  170,
  171,
  175,
  176,
  177,
  186,
  191,
  192,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,47,18,
  127,
  133,
  139,
  144,
  157,
  163,
  169,
  191,
  192,
  193,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,45,11,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  221,
  1,93,1,
  221,
  0xFE,
  0xFE,
  1,44,0,
  1,101,1,
  221,
  0xFE,
  1,99,1,
  221,
  0xFE,
  0xFE,
  1,46,0,
  1,105,2,
  169,
  175,
  0xFE,
  0xFE,
  1,101,2,
  233,
  239,
  0xFE,
  1,99,5,
  221,
  222,
  227,
  233,
  239,
  0xFE,
  1,97,5,
  157,
  221,
  227,
  233,
  239,
  0xFE,
  1,93,10,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  239,
  240,
  0xFE,
  0xFE,
  0,252,58,
  3,
  7,
  13,
  104,
  110,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  186,
  187,
  197,
  203,
  204,
  205,
  208,
  209,
  210,
  233,
  234,
  235,
  239,
  240,
  241,
  0,90,0,
  0,151,1,
  98,
  0xFE,
  0xFE,
  1,75,0,
  1,108,2,
  157,
  163,
  0xFE,
  0xFE,
  1,59,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,57,10,
  163,
  169,
  175,
  191,
  192,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,55,12,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  1,102,3,
  133,
  157,
  163,
  0xFE,
  0xFE,
  1,53,9,
  157,
  163,
  169,
  191,
  192,
  197,
  198,
  203,
  208,
  1,98,4,
  127,
  133,
  157,
  163,
  0xFE,
  0xFE,
  1,51,9,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  203,
  208,
  0xFE,
  1,49,23,
  127,
  128,
  133,
  134,
  135,
  139,
  144,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  203,
  208,
  0xFE,
  1,47,9,
  191,
  192,
  197,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,45,8,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  208,
  0xFE,
  1,46,0,
  1,105,4,
  157,
  163,
  169,
  175,
  0xFE,
  0xFE,
  1,101,3,
  221,
  233,
  239,
  0xFE,
  1,99,2,
  157,
  221,
  0xFE,
  1,97,2,
  221,
  227,
  0xFE,
  1,93,8,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  239,
  0xFE,
  0xFE,
  0,254,43,
  7,
  13,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  186,
  187,
  203,
  208,
  233,
  239,
  1,75,0,
  1,104,2,
  233,
  239,
  0xFE,
  1,106,2,
  221,
  227,
  0xFE,
  0xFE,
  1,61,3,
  191,
  203,
  208,
  0xFE,
  1,59,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,57,9,
  157,
  163,
  169,
  191,
  192,
  197,
  198,
  203,
  208,
  0xFE,
  1,55,8,
  157,
  163,
  191,
  192,
  197,
  198,
  203,
  208,
  0xFE,
  1,53,7,
  157,
  163,
  175,
  191,
  197,
  203,
  208,
  0xFE,
  1,51,5,
  157,
  191,
  192,
  203,
  208,
  0xFE,
  1,49,14,
  127,
  133,
  139,
  140,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  197,
  0xFE,
  1,45,4,
  191,
  192,
  197,
  198,
  0xFE,
  1,46,0,
  1,105,3,
  133,
  157,
  163,
  0xFE,
  0xFE,
  1,101,1,
  221,
  0xFE,
  1,99,1,
  221,
  0xFE,
  1,93,4,
  221,
  222,
  227,
  228,
  0xFE,
  1,108,0,
  1,81,2,
  203,
  208,
  0xFE,
  1,79,6,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,77,5,
  191,
  192,
  197,
  198,
  208,
  0xFE,
  0xFE,
  0xFE,
  1,0,68,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  20,
  25,
  26,
  38,
  44,
  71,
  72,
  77,
  78,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  177,
  197,
  198,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  227,
  233,
  234,
  235,
  236,
  239,
  240,
  241,
  242,
  0,42,0,
  0,105,2,
  34,
  65,
  0xFE,
  0xFE,
  0,92,0,
  0,155,5,
  65,
  95,
  98,
  104,
  110,
  0xFE,
  0xFE,
  1,63,2,
  203,
  208,
  0xFE,
  1,61,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,59,6,
  157,
  163,
  191,
  197,
  203,
  208,
  0xFE,
  1,57,6,
  157,
  163,
  191,
  197,
  203,
  208,
  0xFE,
  1,55,7,
  157,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,53,10,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,51,14,
  127,
  133,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,49,26,
  127,
  128,
  133,
  134,
  139,
  144,
  145,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  210,
  0xFE,
  1,47,13,
  133,
  139,
  144,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,45,11,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  208,
  0xFE,
  1,103,1,
  221,
  0xFE,
  1,99,4,
  233,
  234,
  239,
  240,
  0xFE,
  1,97,4,
  221,
  222,
  227,
  228,
  0xFE,
  1,93,12,
  221,
  222,
  223,
  224,
  227,
  228,
  229,
  230,
  233,
  234,
  239,
  240,
  0xFE,
  0xFE,
  1,2,14,
  127,
  133,
  139,
  140,
  144,
  145,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,77,0,
  1,128,1,
  197,
  0xFE,
  0xFE,
  1,75,0,
  1,126,1,
  197,
  0xFE,
  0xFE,
  1,73,0,
  1,120,3,
  197,
  233,
  239,
  0xFE,
  0xFE,
  1,71,0,
  1,118,3,
  197,
  203,
  208,
  0xFE,
  0xFE,
  1,69,0,
  1,112,1,
  197,
  0xFE,
  1,114,1,
  233,
  0xFE,
  0xFE,
  1,67,0,
  1,8,1,
  191,
  0xFE,
  1,110,3,
  197,
  198,
  227,
  0xFE,
  1,112,1,
  191,
  0xFE,
  1,114,2,
  233,
  239,
  0xFE,
  0xFE,
  1,65,0,
  1,2,1,
  203,
  0xFE,
  1,4,3,
  191,
  203,
  208,
  0xFE,
  1,6,1,
  191,
  0xFE,
  1,8,1,
  191,
  0xFE,
  1,104,1,
  197,
  0xFE,
  1,106,1,
  197,
  0xFE,
  1,110,2,
  191,
  221,
  0xFE,
  1,112,1,
  191,
  0xFE,
  0xFE,
  1,63,0,
  1,2,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,4,3,
  191,
  197,
  203,
  0xFE,
  1,6,1,
  191,
  0xFE,
  1,8,1,
  191,
  0xFE,
  1,102,2,
  197,
  227,
  0xFE,
  1,106,1,
  191,
  0xFE,
  1,108,3,
  191,
  197,
  221,
  0xFE,
  1,110,2,
  191,
  221,
  0xFE,
  0xFE,
  1,61,0,
  1,2,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,4,2,
  191,
  197,
  0xFE,
  1,6,1,
  191,
  0xFE,
  1,104,1,
  191,
  0xFE,
  1,106,4,
  191,
  197,
  221,
  227,
  0xFE,
  1,108,4,
  191,
  203,
  221,
  227,
  0xFE,
  1,110,4,
  197,
  203,
  233,
  239,
  0xFE,
  0xFE,
  1,59,2,
  203,
  208,
  1,2,4,
  157,
  191,
  197,
  203,
  0xFE,
  1,4,1,
  191,
  0xFE,
  1,6,1,
  191,
  0xFE,
  1,98,1,
  203,
  0xFE,
  1,102,1,
  191,
  0xFE,
  1,104,4,
  191,
  197,
  221,
  227,
  0xFE,
  1,106,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,108,6,
  191,
  197,
  203,
  208,
  233,
  239,
  0xFE,
  0xFE,
  1,57,14,
  133,
  134,
  139,
  144,
  157,
  163,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  1,2,4,
  157,
  191,
  197,
  208,
  0xFE,
  1,4,1,
  191,
  0xFE,
  1,98,3,
  197,
  203,
  208,
  0xFE,
  1,100,5,
  191,
  203,
  208,
  221,
  227,
  0xFE,
  1,102,6,
  191,
  197,
  203,
  208,
  221,
  227,
  0xFE,
  1,104,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,106,6,
  191,
  197,
  208,
  221,
  233,
  239,
  0xFE,
  0xFE,
  1,55,16,
  133,
  139,
  144,
  164,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  1,2,2,
  191,
  197,
  0xFE,
  1,96,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,98,6,
  191,
  197,
  203,
  208,
  221,
  227,
  0xFE,
  1,100,6,
  191,
  197,
  203,
  208,
  221,
  227,
  0xFE,
  1,102,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,104,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,106,4,
  191,
  197,
  203,
  208,
  0xFE,
  0xFE,
  1,53,24,
  127,
  133,
  134,
  139,
  140,
  144,
  145,
  163,
  169,
  175,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  1,92,2,
  203,
  208,
  0xFE,
  1,96,6,
  191,
  197,
  203,
  208,
  221,
  227,
  0xFE,
  1,98,12,
  157,
  163,
  169,
  175,
  191,
  197,
  203,
  204,
  208,
  209,
  221,
  227,
  0xFE,
  1,100,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,102,11,
  191,
  192,
  197,
  203,
  204,
  208,
  209,
  221,
  227,
  233,
  239,
  0xFE,
  1,104,6,
  191,
  197,
  203,
  208,
  221,
  239,
  0xFE,
  1,106,2,
  191,
  197,
  0xFE,
  0xFE,
  1,51,31,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  149,
  153,
  157,
  163,
  169,
  170,
  175,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  217,
  1,92,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,96,9,
  191,
  197,
  203,
  204,
  208,
  209,
  221,
  227,
  239,
  0xFE,
  1,98,14,
  139,
  144,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  221,
  227,
  233,
  239,
  0xFE,
  1,100,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,102,9,
  191,
  192,
  197,
  198,
  203,
  208,
  221,
  227,
  239,
  0xFE,
  1,104,2,
  221,
  227,
  0xFE,
  0xFE,
  1,49,44,
  127,
  128,
  133,
  134,
  135,
  139,
  140,
  144,
  145,
  149,
  153,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  214,
  217,
  218,
  221,
  233,
  239,
  0,32,2,
  191,
  197,
  0xFE,
  0,86,2,
  191,
  197,
  0xFE,
  1,2,13,
  7,
  13,
  127,
  128,
  133,
  134,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  1,93,2,
  221,
  227,
  0xFE,
  0xFE,
  1,4,3,
  163,
  169,
  175,
  0xFE,
  1,93,2,
  203,
  208,
  0xFE,
  1,92,9,
  191,
  192,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,96,12,
  191,
  192,
  197,
  198,
  203,
  208,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,98,17,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  213,
  217,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  0xFE,
  1,100,8,
  203,
  208,
  221,
  222,
  227,
  228,
  233,
  239,
  0xFE,
  1,102,4,
  203,
  208,
  233,
  239,
  0xFE,
  0xFE,
  1,47,41,
  3,
  98,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  144,
  145,
  149,
  153,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  217,
  218,
  221,
  227,
  233,
  239,
  0,0,2,
  191,
  197,
  0xFE,
  0,2,1,
  191,
  0xFE,
  0,32,2,
  191,
  197,
  0xFE,
  0,86,2,
  191,
  197,
  0xFE,
  0,88,1,
  191,
  0xFE,
  0,140,2,
  191,
  197,
  0xFE,
  0,142,1,
  191,
  0xFE,
  1,99,3,
  197,
  203,
  208,
  0xFE,
  1,93,5,
  191,
  203,
  208,
  233,
  239,
  0xFE,
  1,92,12,
  191,
  197,
  203,
  208,
  213,
  217,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,96,13,
  191,
  203,
  208,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,98,10,
  203,
  208,
  213,
  217,
  221,
  227,
  233,
  239,
  245,
  249,
  0xFE,
  1,100,4,
  233,
  239,
  245,
  249,
  0xFE,
  0xFE,
  1,45,50,
  3,
  34,
  65,
  98,
  127,
  133,
  134,
  139,
  140,
  144,
  145,
  149,
  153,
  157,
  163,
  169,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  214,
  215,
  216,
  217,
  218,
  219,
  220,
  221,
  222,
  227,
  228,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  249,
  0,0,2,
  203,
  208,
  0xFE,
  0,32,4,
  197,
  203,
  208,
  221,
  0xFE,
  0,86,4,
  197,
  203,
  208,
  221,
  0xFE,
  0,140,2,
  203,
  208,
  0xFE,
  1,99,6,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,97,3,
  191,
  203,
  208,
  0xFE,
  1,93,13,
  139,
  144,
  191,
  192,
  197,
  203,
  208,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,92,16,
  191,
  197,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  239,
  240,
  245,
  246,
  249,
  250,
  0xFE,
  1,96,8,
  221,
  227,
  233,
  239,
  245,
  246,
  249,
  250,
  0xFE,
  1,98,6,
  203,
  208,
  213,
  217,
  233,
  239,
  0xFE,
  0xFE,
  1,44,34,
  127,
  133,
  134,
  139,
  140,
  144,
  145,
  149,
  153,
  163,
  169,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  215,
  217,
  218,
  219,
  221,
  233,
  239,
  245,
  249,
  1,103,8,
  191,
  197,
  203,
  204,
  208,
  209,
  233,
  239,
  0xFE,
  1,101,8,
  191,
  203,
  208,
  221,
  233,
  234,
  239,
  240,
  0xFE,
  1,99,25,
  133,
  139,
  144,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  222,
  227,
  228,
  233,
  234,
  235,
  239,
  240,
  241,
  0,86,1,
  221,
  0xFE,
  0xFE,
  1,97,21,
  133,
  191,
  192,
  197,
  198,
  203,
  208,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  239,
  240,
  245,
  246,
  249,
  250,
  0xFE,
  1,93,12,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  245,
  246,
  249,
  250,
  0xFE,
  0xFE,
  1,46,21,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  217,
  218,
  1,107,2,
  203,
  208,
  0xFE,
  1,105,4,
  203,
  208,
  233,
  239,
  0xFE,
  1,103,14,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,101,14,
  191,
  197,
  203,
  208,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,99,16,
  191,
  192,
  197,
  198,
  203,
  208,
  213,
  217,
  221,
  222,
  227,
  228,
  233,
  239,
  245,
  249,
  0xFE,
  1,97,12,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,93,4,
  203,
  208,
  233,
  239,
  0xFE,
  0xFE,
  1,48,16,
  127,
  133,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  197,
  203,
  204,
  208,
  209,
  213,
  217,
  1,9,6,
  139,
  144,
  163,
  169,
  175,
  176,
  0xFE,
  1,7,14,
  127,
  128,
  129,
  133,
  134,
  135,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  175,
  0xFE,
  1,107,8,
  191,
  197,
  203,
  208,
  233,
  234,
  239,
  240,
  0xFE,
  1,105,10,
  191,
  197,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  0xFE,
  1,103,9,
  191,
  197,
  198,
  203,
  208,
  221,
  227,
  228,
  233,
  0xFE,
  1,101,9,
  191,
  197,
  203,
  208,
  221,
  222,
  227,
  233,
  239,
  0xFE,
  1,99,9,
  191,
  197,
  203,
  208,
  213,
  217,
  221,
  227,
  239,
  0xFE,
  1,97,6,
  191,
  197,
  203,
  208,
  233,
  239,
  0xFE,
  0xFE,
  1,50,0,
  1,7,3,
  191,
  192,
  208,
  0xFE,
  1,111,2,
  233,
  239,
  0xFE,
  1,109,4,
  221,
  227,
  233,
  239,
  0xFE,
  1,107,4,
  191,
  197,
  221,
  227,
  0xFE,
  1,105,5,
  191,
  197,
  221,
  227,
  233,
  0xFE,
  1,103,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,101,5,
  191,
  197,
  208,
  221,
  227,
  0xFE,
  1,99,4,
  191,
  221,
  233,
  239,
  0xFE,
  0xFE,
  1,52,0,
  1,9,1,
  191,
  0xFE,
  1,7,10,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  208,
  0xFE,
  1,109,1,
  221,
  0xFE,
  1,107,4,
  191,
  221,
  233,
  239,
  0xFE,
  1,105,6,
  191,
  197,
  221,
  227,
  233,
  239,
  0xFE,
  1,103,4,
  191,
  197,
  221,
  227,
  0xFE,
  1,101,6,
  191,
  203,
  208,
  221,
  233,
  239,
  0xFE,
  0xFE,
  1,54,0,
  1,9,8,
  157,
  163,
  175,
  191,
  192,
  197,
  203,
  208,
  0xFE,
  1,7,8,
  157,
  163,
  191,
  192,
  197,
  198,
  203,
  208,
  0xFE,
  1,109,2,
  191,
  221,
  0xFE,
  1,107,5,
  191,
  221,
  227,
  233,
  239,
  0xFE,
  1,105,3,
  191,
  197,
  221,
  0xFE,
  1,103,2,
  191,
  221,
  0xFE,
  0xFE,
  1,56,0,
  1,9,8,
  157,
  163,
  191,
  192,
  197,
  198,
  203,
  208,
  0xFE,
  1,7,12,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,111,2,
  191,
  221,
  0xFE,
  1,109,2,
  191,
  221,
  0xFE,
  1,107,3,
  191,
  221,
  227,
  0xFE,
  1,105,2,
  191,
  221,
  0xFE,
  0xFE,
  1,58,0,
  1,13,2,
  191,
  197,
  0xFE,
  1,9,10,
  157,
  163,
  169,
  175,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,7,9,
  163,
  169,
  175,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,113,1,
  221,
  0xFE,
  1,111,2,
  191,
  221,
  0xFE,
  1,109,2,
  191,
  221,
  0xFE,
  1,107,2,
  191,
  221,
  0xFE,
  0xFE,
  1,60,0,
  1,15,1,
  191,
  0xFE,
  1,13,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,9,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,7,3,
  191,
  203,
  208,
  0xFE,
  1,113,2,
  191,
  221,
  0xFE,
  1,111,2,
  191,
  221,
  0xFE,
  0xFE,
  1,62,0,
  1,15,2,
  191,
  197,
  0xFE,
  1,13,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,9,2,
  203,
  208,
  0xFE,
  1,115,2,
  191,
  221,
  0xFE,
  1,113,2,
  191,
  221,
  0xFE,
  0xFE,
  1,64,0,
  1,15,3,
  191,
  197,
  203,
  0xFE,
  1,13,2,
  203,
  208,
  0xFE,
  1,117,1,
  221,
  0xFE,
  1,115,2,
  191,
  221,
  0xFE,
  0xFE,
  1,66,0,
  1,117,1,
  221,
  0xFE,
  0xFE,
  1,103,2,
  203,
  208,
  0xFE,
  1,101,4,
  203,
  208,
  233,
  239,
  0xFE,
  1,99,16,
  133,
  139,
  144,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  221,
  233,
  234,
  239,
  240,
  0xFE,
  1,97,28,
  127,
  133,
  139,
  140,
  144,
  145,
  157,
  163,
  169,
  175,
  191,
  197,
  203,
  204,
  208,
  209,
  221,
  222,
  227,
  228,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  249,
  0xFE,
  1,95,41,
  127,
  128,
  133,
  134,
  135,
  139,
  144,
  157,
  163,
  175,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  213,
  214,
  217,
  218,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  246,
  249,
  250,
  0,0,2,
  221,
  227,
  0xFE,
  0,2,1,
  221,
  0xFE,
  0,32,2,
  221,
  227,
  0xFE,
  0,34,1,
  221,
  0xFE,
  0,86,2,
  221,
  227,
  0xFE,
  0,88,1,
  221,
  0xFE,
  0,140,2,
  221,
  227,
  0xFE,
  0,142,1,
  221,
  0xFE,
  1,7,0,
  1,56,1,
  197,
  0xFE,
  0xFE,
  1,5,0,
  1,54,2,
  191,
  197,
  0xFE,
  1,56,4,
  191,
  197,
  203,
  208,
  0xFE,
  0xFE,
  1,3,0,
  1,50,7,
  169,
  175,
  197,
  203,
  208,
  233,
  239,
  0xFE,
  1,52,9,
  191,
  192,
  193,
  197,
  198,
  203,
  208,
  233,
  239,
  0xFE,
  1,54,6,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,56,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,64,1,
  197,
  0xFE,
  0xFE,
  1,1,2,
  233,
  239,
  1,46,17,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  208,
  221,
  222,
  227,
  233,
  239,
  240,
  0xFE,
  1,48,11,
  191,
  192,
  193,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,50,14,
  133,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,52,5,
  163,
  191,
  197,
  203,
  208,
  0xFE,
  1,54,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,56,3,
  197,
  203,
  208,
  0xFE,
  1,58,5,
  163,
  191,
  197,
  203,
  208,
  0xFE,
  0xFE,
  0,255,0,
  1,48,6,
  163,
  169,
  175,
  191,
  197,
  198,
  0xFE,
  1,50,20,
  127,
  133,
  139,
  140,
  144,
  145,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,52,11,
  133,
  157,
  163,
  164,
  169,
  175,
  191,
  197,
  198,
  203,
  208,
  0xFE,
  1,54,2,
  203,
  208,
  0xFE,
  0xFE,
  0,253,0,
  1,44,1,
  175,
  0xFE,
  1,46,4,
  163,
  191,
  197,
  198,
  0xFE,
  1,48,13,
  133,
  157,
  163,
  164,
  169,
  175,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,50,10,
  133,
  139,
  144,
  157,
  163,
  169,
  175,
  176,
  203,
  208,
  0xFE,
  1,52,2,
  163,
  197,
  0xFE,
  0xFE,
  0,251,18,
  133,
  157,
  158,
  163,
  164,
  165,
  169,
  175,
  198,
  221,
  227,
  228,
  233,
  234,
  235,
  239,
  240,
  241,
  1,45,1,
  197,
  0xFE,
  1,44,3,
  163,
  169,
  175,
  0xFE,
  1,46,3,
  163,
  191,
  197,
  0xFE,
  1,48,7,
  157,
  163,
  169,
  175,
  191,
  203,
  208,
  0xFE,
  1,50,8,
  157,
  158,
  163,
  169,
  175,
  191,
  203,
  208,
  0xFE,
  1,52,3,
  191,
  197,
  208,
  0xFE,
  1,54,1,
  191,
  0xFE,
  0xFE,
  0,249,21,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  203,
  208,
  221,
  222,
  227,
  228,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  249,
  1,49,1,
  197,
  0xFE,
  1,47,7,
  133,
  139,
  144,
  157,
  163,
  170,
  176,
  0xFE,
  1,45,3,
  197,
  203,
  208,
  0xFE,
  1,44,10,
  157,
  163,
  169,
  175,
  191,
  197,
  203,
  208,
  221,
  227,
  0xFE,
  1,46,7,
  191,
  192,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,48,6,
  191,
  192,
  197,
  198,
  203,
  208,
  0xFE,
  1,50,2,
  203,
  208,
  0xFE,
  0xFE,
  0,248,18,
  157,
  163,
  169,
  175,
  191,
  197,
  203,
  208,
  221,
  222,
  223,
  227,
  233,
  234,
  239,
  240,
  245,
  249,
  1,53,2,
  203,
  208,
  0xFE,
  1,51,6,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,49,8,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,47,6,
  157,
  163,
  191,
  192,
  197,
  198,
  0xFE,
  1,45,7,
  157,
  163,
  191,
  192,
  197,
  203,
  208,
  0xFE,
  1,44,3,
  163,
  203,
  208,
  0xFE,
  0xFE,
  0,250,11,
  221,
  222,
  223,
  227,
  228,
  233,
  234,
  235,
  239,
  240,
  241,
  1,53,2,
  191,
  197,
  0xFE,
  1,51,3,
  191,
  197,
  203,
  0xFE,
  1,49,2,
  191,
  197,
  0xFE,
  1,47,1,
  191,
  0xFE,
  1,45,1,
  191,
  0xFE,
  1,44,1,
  163,
  0xFE,
  0xFE,
  0,252,9,
  221,
  222,
  227,
  233,
  234,
  235,
  239,
  240,
  241,
  1,53,1,
  191,
  0xFE,
  1,51,1,
  191,
  0xFE,
  1,49,1,
  191,
  0xFE,
  1,47,1,
  163,
  0xFE,
  0xFE,
  0,254,1,
  221,
  1,55,1,
  191,
  0xFE,
  1,53,1,
  191,
  0xFE,
  1,51,1,
  191,
  0xFE,
  0xFE,
  1,0,10,
  221,
  222,
  223,
  227,
  233,
  234,
  235,
  239,
  240,
  241,
  1,55,1,
  191,
  0xFE,
  1,53,1,
  191,
  0xFE,
  0xFE,
  1,55,5,
  191,
  197,
  198,
  203,
  208,
  0xFE,
  1,53,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,51,8,
  139,
  144,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,49,10,
  133,
  139,
  144,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,47,20,
  127,
  133,
  139,
  144,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  0xFE,
  1,45,22,
  127,
  133,
  134,
  139,
  144,
  157,
  191,
  192,
  193,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  221,
  233,
  239,
  0,32,1,
  191,
  0xFE,
  0,86,1,
  191,
  0xFE,
  0xFE,
  1,44,22,
  133,
  163,
  169,
  191,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  227,
  228,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,46,16,
  139,
  144,
  191,
  197,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  227,
  233,
  239,
  245,
  249,
  0xFE,
  1,48,7,
  191,
  197,
  203,
  208,
  213,
  217,
  221,
  0xFE,
  0xFE,
  1,93,35,
  98,
  127,
  133,
  139,
  144,
  157,
  163,
  191,
  197,
  203,
  208,
  213,
  217,
  221,
  222,
  223,
  224,
  227,
  228,
  229,
  230,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  246,
  247,
  248,
  249,
  250,
  251,
  252,
  0xFE,
  1,92,22,
  139,
  144,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  246,
  247,
  248,
  249,
  250,
  251,
  252,
  0xFE,
  1,94,25,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  214,
  217,
  218,
  221,
  222,
  227,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  246,
  249,
  250,
  1,7,0,
  1,52,1,
  197,
  0xFE,
  0xFE,
  1,5,2,
  221,
  222,
  1,50,2,
  191,
  197,
  0xFE,
  1,52,3,
  197,
  203,
  208,
  0xFE,
  0xFE,
  1,3,5,
  221,
  222,
  223,
  227,
  228,
  1,46,1,
  197,
  0xFE,
  1,48,11,
  191,
  192,
  197,
  198,
  221,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  1,50,6,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,52,2,
  203,
  208,
  0xFE,
  1,60,4,
  163,
  197,
  203,
  208,
  0xFE,
  0xFE,
  1,1,37,
  133,
  139,
  144,
  163,
  169,
  175,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  208,
  209,
  210,
  221,
  222,
  223,
  224,
  227,
  228,
  229,
  233,
  234,
  235,
  236,
  237,
  239,
  240,
  241,
  242,
  243,
  1,47,4,
  203,
  204,
  208,
  209,
  0xFE,
  1,45,12,
  191,
  197,
  198,
  199,
  203,
  208,
  221,
  227,
  228,
  229,
  233,
  239,
  0xFE,
  1,44,10,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,46,9,
  157,
  158,
  163,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,48,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,50,5,
  163,
  191,
  197,
  203,
  208,
  0xFE,
  1,52,3,
  163,
  203,
  208,
  0xFE,
  1,54,4,
  191,
  197,
  203,
  208,
  0xFE,
  0xFE,
  0,255,17,
  139,
  144,
  157,
  163,
  164,
  169,
  175,
  221,
  222,
  227,
  228,
  233,
  234,
  235,
  239,
  240,
  241,
  1,45,5,
  157,
  158,
  163,
  169,
  175,
  0xFE,
  1,44,6,
  163,
  169,
  175,
  191,
  197,
  198,
  0xFE,
  1,46,14,
  133,
  139,
  144,
  157,
  163,
  164,
  169,
  175,
  176,
  191,
  197,
  203,
  208,
  209,
  0xFE,
  1,48,11,
  157,
  163,
  164,
  169,
  191,
  197,
  198,
  203,
  208,
  233,
  239,
  0xFE,
  1,50,3,
  197,
  203,
  208,
  0xFE,
  0xFE,
  0,253,23,
  127,
  133,
  139,
  144,
  157,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  197,
  203,
  208,
  221,
  227,
  228,
  229,
  233,
  234,
  239,
  240,
  1,45,9,
  133,
  139,
  144,
  157,
  163,
  169,
  170,
  175,
  176,
  0xFE,
  1,44,7,
  133,
  163,
  169,
  191,
  197,
  203,
  208,
  0xFE,
  1,46,11,
  133,
  139,
  144,
  157,
  163,
  164,
  169,
  170,
  175,
  203,
  208,
  0xFE,
  0xFE,
  0,251,15,
  133,
  139,
  144,
  157,
  163,
  164,
  169,
  175,
  191,
  197,
  221,
  227,
  228,
  233,
  239,
  1,51,4,
  139,
  144,
  169,
  175,
  0xFE,
  1,49,1,
  203,
  0xFE,
  1,47,4,
  197,
  203,
  208,
  227,
  0xFE,
  1,45,7,
  133,
  139,
  157,
  163,
  169,
  175,
  197,
  0xFE,
  1,44,8,
  157,
  163,
  169,
  175,
  191,
  197,
  203,
  208,
  0xFE,
  1,46,6,
  157,
  158,
  163,
  191,
  203,
  208,
  0xFE,
  0xFE,
  0,249,15,
  163,
  191,
  192,
  197,
  198,
  203,
  208,
  221,
  227,
  233,
  234,
  239,
  240,
  245,
  249,
  1,51,12,
  133,
  139,
  144,
  157,
  163,
  164,
  169,
  175,
  176,
  197,
  203,
  208,
  0xFE,
  1,49,6,
  169,
  175,
  191,
  197,
  203,
  208,
  0xFE,
  1,47,6,
  191,
  192,
  197,
  203,
  208,
  221,
  0xFE,
  1,45,9,
  157,
  163,
  169,
  191,
  192,
  197,
  203,
  208,
  221,
  0xFE,
  1,44,6,
  191,
  192,
  197,
  198,
  203,
  208,
  0xFE,
  0xFE,
  0,248,0,
  1,57,3,
  191,
  203,
  208,
  0xFE,
  1,55,7,
  191,
  192,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,53,11,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  221,
  233,
  239,
  0xFE,
  1,51,10,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  208,
  0xFE,
  1,49,7,
  157,
  163,
  191,
  192,
  197,
  203,
  208,
  0xFE,
  1,47,4,
  191,
  203,
  208,
  221,
  0xFE,
  0xFE,
  0,250,0,
  1,59,3,
  191,
  203,
  208,
  0xFE,
  1,57,2,
  191,
  197,
  0xFE,
  1,55,2,
  191,
  197,
  0xFE,
  1,53,1,
  191,
  0xFE,
  1,51,9,
  133,
  139,
  144,
  157,
  163,
  169,
  175,
  191,
  197,
  0xFE,
  1,49,1,
  191,
  0xFE,
  1,47,1,
  191,
  0xFE,
  0xFE,
  0,252,0,
  1,57,1,
  191,
  0xFE,
  1,55,1,
  191,
  0xFE,
  1,53,1,
  191,
  0xFE,
  1,51,5,
  127,
  133,
  157,
  163,
  164,
  0xFE,
  0xFE,
  1,59,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,57,4,
  191,
  197,
  203,
  208,
  1,0,1,
  191,
  0xFE,
  0xFE,
  1,55,7,
  139,
  144,
  191,
  197,
  198,
  203,
  208,
  0,254,1,
  191,
  0xFE,
  0xFE,
  1,53,11,
  133,
  139,
  144,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,51,21,
  127,
  133,
  139,
  144,
  157,
  158,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  221,
  227,
  0xFE,
  1,49,22,
  127,
  133,
  139,
  144,
  157,
  191,
  192,
  193,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  221,
  227,
  233,
  239,
  0,32,1,
  191,
  0xFE,
  0,86,1,
  191,
  0xFE,
  0xFE,
  1,47,28,
  127,
  133,
  139,
  144,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  213,
  214,
  217,
  218,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,45,15,
  197,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  227,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,44,4,
  203,
  208,
  213,
  217,
  0xFE,
  0xFE,
  1,96,2,
  233,
  239,
  0xFE,
  0xFE,
  0,206,143,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  20,
  21,
  25,
  26,
  27,
  34,
  35,
  36,
  38,
  39,
  40,
  44,
  45,
  46,
  50,
  51,
  52,
  56,
  57,
  58,
  65,
  66,
  67,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  84,
  85,
  89,
  90,
  91,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  118,
  122,
  123,
  124,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  149,
  150,
  151,
  153,
  154,
  155,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  173,
  175,
  176,
  177,
  178,
  179,
  181,
  182,
  183,
  186,
  187,
  188,
  191,
  192,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  214,
  217,
  218,
  221,
  222,
  227,
  228,
  229,
  233,
  234,
  235,
  236,
  237,
  239,
  240,
  241,
  242,
  243,
  245,
  246,
  249,
  250,
  0,0,12,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  139,
  144,
  0,37,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,35,2,
  0,
  3,
  0xFE,
  0,95,2,
  7,
  13,
  0xFE,
  0,93,5,
  0,
  3,
  4,
  7,
  13,
  0xFE,
  0,91,1,
  3,
  0xFE,
  0,155,1,
  3,
  0xFE,
  0xFE,
  0,2,6,
  0,
  3,
  7,
  13,
  19,
  25,
  0,43,2,
  7,
  13,
  0xFE,
  0,41,5,
  0,
  3,
  4,
  7,
  13,
  0xFE,
  0,39,1,
  3,
  0,89,1,
  3,
  0xFE,
  0xFE,
  0xFE,
  0,4,0,
  0,47,3,
  3,
  4,
  13,
  0xFE,
  0,45,3,
  3,
  7,
  13,
  0xFE,
  0,43,1,
  3,
  0xFE,
  0xFE,
  0,6,0,
  0,51,1,
  3,
  0xFE,
  0xFE,
  0,8,0,
  0,57,1,
  3,
  0xFE,
  0xFE,
  0,12,0,
  0,67,2,
  3,
  4,
  0xFE,
  0xFE,
  0,32,11,
  34,
  38,
  39,
  44,
  45,
  50,
  56,
  139,
  144,
  149,
  153,
  0,91,2,
  34,
  65,
  0xFE,
  0,89,6,
  34,
  38,
  44,
  65,
  71,
  77,
  0xFE,
  0,90,0,
  0,157,1,
  98,
  0xFE,
  0xFE,
  0,149,4,
  95,
  96,
  98,
  99,
  0xFE,
  0,147,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,145,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,143,1,
  98,
  0xFE,
  1,47,2,
  191,
  197,
  0xFE,
  1,45,2,
  203,
  208,
  0xFE,
  1,93,2,
  233,
  239,
  0xFE,
  0xFE,
  0,34,0,
  0,93,3,
  34,
  62,
  65,
  0xFE,
  0,91,6,
  34,
  38,
  44,
  65,
  71,
  77,
  0xFE,
  0,151,4,
  34,
  98,
  104,
  110,
  0xFE,
  0xFE,
  0,36,0,
  0,97,4,
  38,
  44,
  71,
  77,
  0xFE,
  0,95,10,
  31,
  34,
  35,
  38,
  44,
  50,
  62,
  65,
  71,
  77,
  0xFE,
  0xFE,
  0,38,0,
  0,99,5,
  38,
  44,
  65,
  71,
  77,
  0xFE,
  0,97,2,
  34,
  65,
  0xFE,
  0xFE,
  0,40,0,
  0,101,6,
  34,
  38,
  44,
  65,
  71,
  77,
  0xFE,
  0,99,2,
  34,
  65,
  0xFE,
  0xFE,
  0,42,0,
  0,105,6,
  34,
  38,
  44,
  65,
  71,
  77,
  0xFE,
  0,103,2,
  38,
  44,
  0xFE,
  0xFE,
  0,86,13,
  62,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  89,
  139,
  144,
  149,
  153,
  0,147,2,
  95,
  98,
  0xFE,
  0,145,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,143,8,
  65,
  71,
  77,
  95,
  98,
  99,
  104,
  110,
  0xFE,
  1,47,2,
  191,
  197,
  0xFE,
  1,45,2,
  203,
  208,
  0xFE,
  1,93,2,
  233,
  239,
  0xFE,
  0xFE,
  0,88,0,
  0,4,0,
  0,49,1,
  3,
  0xFE,
  0xFE,
  0,151,1,
  98,
  0xFE,
  0,149,5,
  95,
  98,
  99,
  104,
  110,
  0xFE,
  0,147,11,
  62,
  65,
  71,
  77,
  95,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,145,4,
  95,
  98,
  104,
  110,
  0xFE,
  0xFE,
  0,90,0,
  0,155,2,
  104,
  110,
  0xFE,
  0,153,3,
  98,
  104,
  110,
  0xFE,
  0,151,16,
  62,
  63,
  65,
  66,
  71,
  77,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  0xFE,
  0,149,5,
  65,
  95,
  98,
  104,
  110,
  0xFE,
  0,147,2,
  95,
  98,
  0xFE,
  0xFE,
  0,92,0,
  0,157,2,
  104,
  110,
  0xFE,
  0,155,12,
  65,
  66,
  71,
  77,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,153,6,
  65,
  95,
  98,
  99,
  104,
  110,
  0xFE,
  0,151,2,
  95,
  98,
  0xFE,
  0xFE,
  0,94,0,
  0,159,8,
  65,
  95,
  96,
  98,
  99,
  104,
  110,
  111,
  0xFE,
  0,157,6,
  65,
  95,
  98,
  99,
  104,
  110,
  0xFE,
  0,155,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,153,2,
  95,
  98,
  0xFE,
  0xFE,
  0,96,0,
  0,163,7,
  65,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0,161,5,
  95,
  98,
  99,
  104,
  110,
  0xFE,
  0,159,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,157,2,
  95,
  98,
  0xFE,
  0,155,2,
  95,
  98,
  0xFE,
  0xFE,
  0,98,0,
  0,167,3,
  98,
  104,
  110,
  0xFE,
  0,165,5,
  95,
  98,
  99,
  104,
  110,
  0xFE,
  0,163,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,161,3,
  98,
  104,
  110,
  0xFE,
  0,159,2,
  95,
  98,
  0xFE,
  0xFE,
  0,100,0,
  0,173,3,
  96,
  98,
  99,
  0xFE,
  0,169,4,
  98,
  99,
  104,
  110,
  0xFE,
  0,163,1,
  98,
  0xFE,
  0xFE,
  0,102,0,
  0,173,2,
  98,
  99,
  0xFE,
  0xFE,
  0,104,0,
  0,177,5,
  65,
  98,
  99,
  104,
  110,
  0xFE,
  0xFE,
  0,106,0,
  0,181,11,
  65,
  66,
  71,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0xFE,
  0,140,14,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  0xFE,
  0,142,6,
  104,
  105,
  110,
  111,
  116,
  122,
  0xFE,
  1,27,0,
  1,70,2,
  191,
  197,
  0xFE,
  0xFE,
  1,25,0,
  1,64,2,
  191,
  197,
  0xFE,
  1,66,4,
  191,
  197,
  198,
  208,
  0xFE,
  1,68,5,
  191,
  197,
  198,
  203,
  208,
  0xFE,
  1,70,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,72,2,
  203,
  208,
  0xFE,
  0xFE,
  1,23,0,
  1,62,4,
  191,
  197,
  198,
  208,
  0xFE,
  1,64,7,
  191,
  192,
  197,
  198,
  203,
  208,
  209,
  0xFE,
  1,66,9,
  163,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,68,8,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,70,3,
  191,
  203,
  208,
  0xFE,
  0xFE,
  1,21,0,
  1,60,7,
  163,
  191,
  192,
  197,
  198,
  208,
  209,
  0xFE,
  1,62,10,
  157,
  163,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,64,10,
  157,
  163,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,66,9,
  163,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,68,6,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,70,2,
  203,
  208,
  0xFE,
  0xFE,
  1,19,0,
  1,56,3,
  191,
  197,
  208,
  0xFE,
  1,58,11,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  208,
  209,
  0xFE,
  1,60,16,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  0xFE,
  1,62,15,
  157,
  163,
  164,
  169,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  0xFE,
  1,64,12,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,66,8,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,68,2,
  203,
  208,
  0xFE,
  0xFE,
  1,17,0,
  1,54,7,
  169,
  175,
  191,
  197,
  198,
  203,
  208,
  0xFE,
  1,56,17,
  157,
  158,
  163,
  164,
  169,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  0xFE,
  1,58,19,
  157,
  158,
  163,
  164,
  169,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,60,19,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,62,15,
  157,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,64,10,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,66,5,
  191,
  203,
  204,
  208,
  209,
  0xFE,
  0xFE,
  1,15,0,
  1,52,10,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,54,19,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  210,
  0xFE,
  1,56,22,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,58,21,
  157,
  158,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,60,20,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,62,13,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,64,6,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,66,2,
  203,
  208,
  0xFE,
  1,70,0,
  1,93,2,
  221,
  227,
  0xFE,
  0xFE,
  1,104,1,
  227,
  0xFE,
  0xFE,
  1,13,0,
  1,48,12,
  169,
  170,
  175,
  176,
  203,
  204,
  208,
  209,
  233,
  234,
  239,
  240,
  1,92,6,
  221,
  222,
  223,
  227,
  228,
  229,
  0xFE,
  0xFE,
  1,50,12,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,52,21,
  157,
  158,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,54,24,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,56,26,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  0xFE,
  1,58,20,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,60,14,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,62,6,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,64,2,
  203,
  208,
  0xFE,
  1,100,2,
  221,
  227,
  0xFE,
  1,102,5,
  163,
  221,
  222,
  227,
  228,
  0xFE,
  1,104,2,
  233,
  239,
  0xFE,
  0xFE,
  1,11,21,
  3,
  7,
  13,
  104,
  110,
  169,
  170,
  175,
  176,
  203,
  204,
  205,
  208,
  209,
  210,
  233,
  234,
  235,
  239,
  240,
  241,
  0,0,3,
  3,
  7,
  13,
  0xFE,
  0,2,1,
  3,
  0xFE,
  0,86,2,
  71,
  77,
  0xFE,
  0,88,1,
  65,
  0xFE,
  0,140,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,142,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,144,6,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0,146,2,
  95,
  98,
  0xFE,
  1,44,7,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  0xFE,
  1,46,4,
  203,
  204,
  208,
  209,
  0xFE,
  1,48,19,
  157,
  158,
  159,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  0xFE,
  1,50,25,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  191,
  192,
  193,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,52,26,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  0xFE,
  1,54,28,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  0xFE,
  1,56,24,
  157,
  158,
  163,
  164,
  169,
  170,
  171,
  175,
  176,
  177,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  0xFE,
  1,58,17,
  157,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,60,11,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,62,5,
  191,
  197,
  203,
  204,
  208,
  0xFE,
  1,64,2,
  203,
  208,
  0xFE,
  1,92,8,
  221,
  222,
  223,
  224,
  227,
  228,
  229,
  230,
  0xFE,
  1,98,9,
  163,
  169,
  175,
  221,
  222,
  227,
  228,
  233,
  239,
  0xFE,
  1,100,12,
  157,
  158,
  163,
  164,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  0xFE,
  1,102,4,
  233,
  234,
  239,
  240,
  0xFE,
  0xFE,
  1,9,40,
  3,
  7,
  8,
  13,
  14,
  104,
  105,
  110,
  111,
  139,
  140,
  141,
  144,
  145,
  146,
  157,
  158,
  159,
  163,
  164,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  186,
  203,
  204,
  205,
  208,
  209,
  210,
  233,
  234,
  235,
  239,
  240,
  241,
  0,86,4,
  71,
  77,
  169,
  175,
  0xFE,
  0,88,1,
  65,
  0xFE,
  0,90,1,
  65,
  0xFE,
  0,140,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,142,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,144,7,
  95,
  96,
  98,
  99,
  104,
  110,
  111,
  0xFE,
  0,146,4,
  95,
  96,
  98,
  110,
  0xFE,
  1,44,6,
  191,
  192,
  193,
  197,
  198,
  199,
  0xFE,
  1,46,15,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  213,
  217,
  0xFE,
  1,48,33,
  127,
  128,
  129,
  133,
  134,
  135,
  144,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  208,
  209,
  210,
  1,92,2,
  221,
  227,
  0xFE,
  0xFE,
  1,50,32,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  177,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  213,
  214,
  217,
  0xFE,
  1,52,30,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  217,
  0xFE,
  1,54,26,
  157,
  158,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  0xFE,
  1,56,19,
  157,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,58,14,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,60,8,
  169,
  175,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,62,3,
  197,
  203,
  208,
  0xFE,
  1,64,3,
  197,
  203,
  208,
  0xFE,
  1,74,0,
  1,109,3,
  133,
  163,
  164,
  0xFE,
  0xFE,
  1,92,6,
  221,
  222,
  223,
  227,
  228,
  229,
  0xFE,
  1,96,17,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  221,
  222,
  227,
  228,
  229,
  233,
  234,
  239,
  240,
  0xFE,
  1,98,15,
  157,
  158,
  163,
  164,
  165,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  239,
  240,
  0xFE,
  1,100,10,
  169,
  175,
  221,
  227,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  1,102,2,
  233,
  239,
  0xFE,
  0xFE,
  1,7,67,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  26,
  71,
  77,
  98,
  99,
  104,
  105,
  110,
  111,
  127,
  128,
  133,
  134,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  153,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  186,
  187,
  197,
  198,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  227,
  233,
  234,
  235,
  236,
  239,
  240,
  241,
  242,
  0,0,7,
  0,
  3,
  4,
  7,
  8,
  13,
  14,
  0xFE,
  0,2,3,
  0,
  3,
  4,
  0xFE,
  0,4,1,
  3,
  0xFE,
  0,32,9,
  38,
  39,
  44,
  45,
  157,
  169,
  170,
  175,
  176,
  0xFE,
  0,34,4,
  31,
  32,
  34,
  35,
  0xFE,
  0,86,9,
  71,
  72,
  77,
  78,
  157,
  169,
  170,
  175,
  176,
  0xFE,
  0,88,6,
  62,
  63,
  65,
  66,
  71,
  77,
  0xFE,
  0,90,4,
  62,
  63,
  65,
  66,
  0xFE,
  0,140,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,142,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,144,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,146,4,
  95,
  98,
  104,
  110,
  0xFE,
  1,47,0,
  1,104,6,
  169,
  175,
  203,
  208,
  233,
  239,
  0xFE,
  1,106,2,
  221,
  227,
  0xFE,
  1,108,2,
  233,
  239,
  0xFE,
  0xFE,
  1,45,0,
  1,100,2,
  221,
  227,
  0xFE,
  1,102,2,
  233,
  239,
  0xFE,
  1,104,2,
  191,
  197,
  0xFE,
  0xFE,
  1,44,17,
  169,
  175,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  208,
  209,
  213,
  214,
  217,
  1,92,1,
  227,
  0xFE,
  1,96,2,
  191,
  197,
  0xFE,
  0xFE,
  1,46,29,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  213,
  217,
  0xFE,
  1,48,54,
  3,
  7,
  13,
  104,
  110,
  127,
  128,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  177,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  201,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  213,
  214,
  217,
  218,
  233,
  234,
  239,
  240,
  0,140,2,
  104,
  110,
  0xFE,
  0,142,1,
  98,
  0xFE,
  1,92,4,
  221,
  222,
  227,
  228,
  0xFE,
  1,96,2,
  233,
  239,
  0xFE,
  0xFE,
  1,50,32,
  157,
  158,
  159,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  217,
  0xFE,
  1,52,31,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  214,
  217,
  1,99,1,
  169,
  0xFE,
  0xFE,
  1,54,25,
  157,
  163,
  164,
  169,
  170,
  171,
  175,
  176,
  177,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  217,
  0xFE,
  1,56,18,
  133,
  157,
  163,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  1,99,1,
  197,
  0xFE,
  0xFE,
  1,58,13,
  133,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  1,103,1,
  197,
  0xFE,
  1,99,3,
  197,
  203,
  208,
  0xFE,
  0xFE,
  1,60,9,
  163,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,92,8,
  221,
  222,
  223,
  224,
  227,
  228,
  229,
  230,
  0xFE,
  1,96,22,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  203,
  208,
  221,
  222,
  223,
  227,
  228,
  229,
  230,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  1,98,16,
  157,
  163,
  169,
  170,
  175,
  176,
  221,
  222,
  227,
  228,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  1,100,6,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,102,2,
  233,
  239,
  0xFE,
  1,104,2,
  203,
  208,
  0xFE,
  0xFE,
  1,5,100,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  20,
  25,
  26,
  34,
  38,
  44,
  50,
  56,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  89,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  127,
  128,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  147,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  169,
  170,
  171,
  172,
  173,
  175,
  176,
  177,
  178,
  179,
  181,
  182,
  186,
  187,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  207,
  208,
  209,
  210,
  211,
  212,
  213,
  214,
  217,
  218,
  227,
  228,
  229,
  233,
  234,
  235,
  236,
  237,
  239,
  240,
  241,
  242,
  243,
  0,0,13,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  157,
  169,
  170,
  175,
  176,
  0xFE,
  0,2,7,
  0,
  1,
  3,
  4,
  7,
  169,
  175,
  0xFE,
  0,4,3,
  0,
  3,
  13,
  0xFE,
  0,32,12,
  38,
  39,
  44,
  45,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  0,34,4,
  31,
  32,
  34,
  35,
  0xFE,
  0,86,13,
  65,
  71,
  72,
  77,
  78,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  0,88,6,
  62,
  63,
  65,
  66,
  169,
  175,
  0xFE,
  0,90,2,
  71,
  77,
  0xFE,
  0,140,13,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  157,
  169,
  170,
  175,
  176,
  0xFE,
  0,142,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,144,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,146,2,
  104,
  110,
  0xFE,
  1,47,0,
  1,104,3,
  133,
  157,
  163,
  0xFE,
  0xFE,
  1,45,8,
  203,
  204,
  208,
  209,
  213,
  214,
  217,
  218,
  1,96,4,
  221,
  222,
  227,
  228,
  0xFE,
  1,98,8,
  163,
  221,
  227,
  228,
  233,
  234,
  239,
  240,
  0xFE,
  1,100,4,
  233,
  234,
  239,
  240,
  0xFE,
  0xFE,
  1,44,36,
  157,
  158,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  194,
  195,
  197,
  198,
  199,
  200,
  201,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  217,
  218,
  221,
  227,
  228,
  233,
  234,
  239,
  240,
  1,93,4,
  221,
  222,
  227,
  228,
  0xFE,
  1,92,2,
  233,
  239,
  0xFE,
  1,98,2,
  197,
  203,
  0xFE,
  0xFE,
  1,46,40,
  139,
  144,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  175,
  176,
  181,
  186,
  191,
  192,
  193,
  194,
  195,
  197,
  198,
  199,
  200,
  201,
  203,
  204,
  205,
  206,
  207,
  208,
  209,
  210,
  211,
  212,
  213,
  214,
  217,
  218,
  1,98,2,
  203,
  208,
  0xFE,
  0xFE,
  1,48,56,
  127,
  128,
  133,
  134,
  139,
  140,
  141,
  144,
  145,
  146,
  157,
  158,
  159,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  201,
  203,
  204,
  205,
  206,
  207,
  208,
  209,
  210,
  211,
  212,
  213,
  214,
  217,
  218,
  227,
  233,
  234,
  235,
  239,
  240,
  241,
  0,32,2,
  157,
  163,
  0xFE,
  0,86,2,
  157,
  163,
  0xFE,
  1,92,6,
  221,
  222,
  223,
  227,
  228,
  229,
  0xFE,
  1,96,2,
  233,
  239,
  0xFE,
  0xFE,
  1,50,39,
  133,
  139,
  144,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  178,
  181,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  214,
  217,
  218,
  233,
  239,
  1,93,4,
  221,
  222,
  227,
  228,
  0xFE,
  0xFE,
  1,52,36,
  127,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  191,
  192,
  193,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  214,
  217,
  218,
  1,99,6,
  169,
  175,
  203,
  208,
  233,
  239,
  0xFE,
  1,97,2,
  221,
  227,
  0xFE,
  0xFE,
  1,54,24,
  127,
  133,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  1,103,6,
  169,
  175,
  203,
  208,
  233,
  239,
  0xFE,
  1,101,1,
  221,
  0xFE,
  1,99,2,
  191,
  197,
  0xFE,
  1,97,1,
  197,
  0xFE,
  0xFE,
  1,56,12,
  133,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  203,
  204,
  208,
  209,
  1,103,2,
  191,
  197,
  0xFE,
  1,99,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,97,1,
  197,
  0xFE,
  0xFE,
  1,58,2,
  203,
  208,
  1,103,3,
  197,
  203,
  208,
  0xFE,
  1,99,5,
  191,
  197,
  203,
  208,
  227,
  0xFE,
  0xFE,
  1,93,4,
  233,
  239,
  245,
  249,
  0xFE,
  1,92,19,
  169,
  175,
  221,
  222,
  223,
  224,
  225,
  227,
  228,
  229,
  230,
  231,
  233,
  234,
  239,
  240,
  241,
  245,
  249,
  0xFE,
  1,96,23,
  157,
  163,
  164,
  169,
  170,
  171,
  175,
  176,
  177,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  235,
  236,
  239,
  240,
  241,
  242,
  0xFE,
  1,98,14,
  157,
  163,
  169,
  175,
  197,
  221,
  227,
  228,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  1,100,2,
  233,
  239,
  0xFE,
  0xFE,
  1,3,105,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  34,
  35,
  38,
  39,
  44,
  45,
  65,
  66,
  71,
  72,
  77,
  78,
  98,
  99,
  104,
  105,
  110,
  111,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  149,
  150,
  151,
  153,
  154,
  155,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  179,
  181,
  182,
  183,
  186,
  187,
  188,
  191,
  192,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  207,
  208,
  209,
  210,
  211,
  212,
  213,
  214,
  217,
  227,
  228,
  229,
  230,
  233,
  234,
  235,
  236,
  237,
  239,
  240,
  241,
  242,
  243,
  0,0,6,
  157,
  158,
  163,
  164,
  169,
  175,
  0xFE,
  0,2,2,
  157,
  163,
  0xFE,
  0,32,8,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  0,34,2,
  157,
  163,
  0xFE,
  0,86,10,
  65,
  71,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  0,88,3,
  65,
  157,
  163,
  0xFE,
  0,140,13,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  157,
  158,
  163,
  164,
  175,
  0xFE,
  0,142,6,
  95,
  96,
  98,
  99,
  157,
  163,
  0xFE,
  1,53,0,
  1,106,3,
  191,
  197,
  227,
  0xFE,
  0xFE,
  1,51,0,
  1,104,4,
  191,
  197,
  221,
  227,
  0xFE,
  1,106,3,
  203,
  233,
  239,
  0xFE,
  0xFE,
  1,49,0,
  1,102,5,
  191,
  197,
  198,
  221,
  227,
  0xFE,
  1,104,6,
  197,
  203,
  208,
  227,
  233,
  239,
  0xFE,
  0xFE,
  1,47,2,
  203,
  208,
  1,98,1,
  227,
  0xFE,
  1,100,8,
  191,
  192,
  197,
  198,
  221,
  222,
  227,
  228,
  0xFE,
  1,102,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,104,11,
  133,
  139,
  144,
  157,
  163,
  164,
  169,
  175,
  208,
  233,
  239,
  0xFE,
  0xFE,
  1,45,20,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  215,
  217,
  218,
  219,
  1,96,9,
  203,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  239,
  0xFE,
  1,98,14,
  191,
  192,
  197,
  198,
  221,
  222,
  227,
  228,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  1,100,13,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,102,5,
  197,
  203,
  208,
  233,
  239,
  0xFE,
  0xFE,
  1,44,47,
  133,
  139,
  144,
  157,
  158,
  159,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  181,
  186,
  191,
  192,
  193,
  194,
  195,
  197,
  198,
  199,
  200,
  201,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  214,
  215,
  217,
  218,
  219,
  221,
  222,
  233,
  234,
  235,
  239,
  240,
  241,
  0,32,2,
  203,
  208,
  0xFE,
  0,86,2,
  203,
  208,
  0xFE,
  1,93,7,
  221,
  222,
  223,
  227,
  228,
  229,
  239,
  0xFE,
  1,92,10,
  191,
  197,
  198,
  221,
  222,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,96,14,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  221,
  227,
  228,
  233,
  234,
  239,
  240,
  0xFE,
  1,98,15,
  133,
  163,
  191,
  197,
  203,
  204,
  208,
  209,
  221,
  227,
  228,
  233,
  234,
  239,
  240,
  0xFE,
  0xFE,
  1,46,55,
  127,
  128,
  133,
  134,
  135,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  186,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  201,
  203,
  204,
  205,
  206,
  207,
  208,
  209,
  210,
  211,
  212,
  213,
  214,
  217,
  218,
  221,
  227,
  233,
  234,
  239,
  240,
  0,32,2,
  191,
  197,
  0xFE,
  0,86,2,
  191,
  197,
  0xFE,
  1,97,3,
  221,
  222,
  227,
  0xFE,
  1,93,5,
  197,
  221,
  227,
  233,
  239,
  0xFE,
  1,92,13,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  221,
  222,
  227,
  228,
  233,
  239,
  0xFE,
  1,96,1,
  233,
  0xFE,
  0xFE,
  1,48,66,
  3,
  7,
  13,
  34,
  38,
  44,
  65,
  71,
  77,
  98,
  104,
  110,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  144,
  145,
  149,
  153,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  186,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  214,
  217,
  218,
  227,
  228,
  233,
  239,
  1,99,3,
  191,
  192,
  197,
  0xFE,
  1,93,8,
  197,
  198,
  203,
  208,
  221,
  227,
  228,
  239,
  0xFE,
  1,92,1,
  221,
  0xFE,
  0xFE,
  1,50,39,
  127,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  186,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  214,
  217,
  218,
  233,
  234,
  239,
  240,
  1,99,7,
  191,
  203,
  204,
  208,
  209,
  233,
  239,
  0xFE,
  1,97,9,
  203,
  208,
  209,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,93,8,
  191,
  197,
  198,
  204,
  221,
  222,
  227,
  228,
  0xFE,
  0xFE,
  1,52,20,
  127,
  133,
  139,
  144,
  157,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  203,
  204,
  205,
  208,
  209,
  210,
  1,103,2,
  203,
  208,
  0xFE,
  1,101,2,
  203,
  208,
  0xFE,
  1,99,18,
  127,
  133,
  139,
  144,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,97,16,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  0xFE,
  1,93,1,
  221,
  0xFE,
  0xFE,
  1,54,10,
  133,
  157,
  163,
  169,
  175,
  191,
  203,
  204,
  208,
  209,
  1,105,2,
  233,
  239,
  0xFE,
  1,103,15,
  133,
  157,
  163,
  169,
  175,
  191,
  197,
  203,
  208,
  221,
  222,
  227,
  228,
  233,
  239,
  0xFE,
  1,101,8,
  197,
  198,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,99,13,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,97,8,
  191,
  192,
  197,
  198,
  221,
  222,
  227,
  228,
  0xFE,
  0xFE,
  1,56,2,
  203,
  208,
  1,107,6,
  133,
  163,
  191,
  197,
  221,
  227,
  0xFE,
  1,105,1,
  227,
  0xFE,
  1,103,3,
  203,
  208,
  233,
  0xFE,
  1,101,10,
  197,
  203,
  204,
  208,
  209,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,99,8,
  191,
  192,
  197,
  198,
  221,
  222,
  227,
  228,
  0xFE,
  0xFE,
  1,58,0,
  1,107,2,
  197,
  227,
  0xFE,
  1,103,5,
  197,
  203,
  208,
  233,
  239,
  0xFE,
  1,101,7,
  191,
  192,
  197,
  198,
  221,
  227,
  228,
  0xFE,
  0xFE,
  1,60,0,
  1,105,4,
  203,
  208,
  233,
  239,
  0xFE,
  1,103,5,
  191,
  197,
  198,
  221,
  227,
  0xFE,
  0xFE,
  1,62,0,
  1,107,2,
  233,
  239,
  0xFE,
  1,105,4,
  191,
  197,
  221,
  227,
  0xFE,
  0xFE,
  1,64,0,
  1,107,3,
  191,
  197,
  227,
  0xFE,
  0xFE,
  1,93,12,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  245,
  246,
  249,
  250,
  0xFE,
  1,92,30,
  157,
  158,
  163,
  164,
  165,
  169,
  175,
  221,
  222,
  223,
  224,
  225,
  226,
  227,
  228,
  229,
  230,
  231,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  246,
  247,
  249,
  250,
  251,
  0xFE,
  1,96,24,
  133,
  139,
  144,
  157,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  221,
  222,
  227,
  228,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  1,98,19,
  133,
  157,
  163,
  169,
  175,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  221,
  227,
  228,
  233,
  234,
  239,
  240,
  0xFE,
  1,100,12,
  163,
  197,
  203,
  204,
  208,
  209,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,102,3,
  227,
  233,
  239,
  0xFE,
  0xFE,
  1,1,131,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  20,
  25,
  26,
  34,
  35,
  38,
  39,
  44,
  45,
  50,
  51,
  56,
  57,
  65,
  66,
  67,
  71,
  72,
  77,
  78,
  83,
  84,
  89,
  90,
  98,
  99,
  100,
  101,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  127,
  128,
  129,
  130,
  131,
  132,
  133,
  134,
  135,
  136,
  137,
  138,
  139,
  140,
  141,
  142,
  143,
  144,
  145,
  146,
  147,
  148,
  149,
  150,
  151,
  153,
  154,
  155,
  157,
  158,
  159,
  160,
  161,
  162,
  163,
  164,
  165,
  166,
  167,
  168,
  169,
  170,
  171,
  172,
  173,
  175,
  176,
  177,
  178,
  179,
  181,
  182,
  183,
  186,
  187,
  188,
  191,
  192,
  193,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  217,
  218,
  227,
  228,
  229,
  230,
  233,
  234,
  235,
  236,
  239,
  240,
  241,
  242,
  245,
  0,0,1,
  163,
  0xFE,
  0,2,0,
  0,41,1,
  3,
  0xFE,
  0xFE,
  0,32,5,
  133,
  134,
  157,
  163,
  164,
  0xFE,
  0,86,6,
  65,
  133,
  134,
  157,
  163,
  164,
  0xFE,
  0,140,7,
  95,
  96,
  98,
  99,
  104,
  110,
  163,
  0xFE,
  0,142,2,
  95,
  98,
  0xFE,
  1,49,4,
  169,
  170,
  175,
  176,
  1,102,1,
  163,
  0xFE,
  0xFE,
  1,47,8,
  169,
  175,
  191,
  192,
  203,
  204,
  208,
  209,
  1,96,1,
  227,
  0xFE,
  1,98,6,
  133,
  157,
  163,
  191,
  197,
  227,
  0xFE,
  1,104,7,
  133,
  139,
  144,
  157,
  163,
  169,
  175,
  0xFE,
  0xFE,
  1,45,23,
  163,
  164,
  169,
  170,
  175,
  176,
  181,
  186,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  217,
  218,
  1,96,2,
  221,
  227,
  0xFE,
  1,98,3,
  163,
  233,
  239,
  0xFE,
  0xFE,
  1,44,39,
  127,
  128,
  133,
  134,
  135,
  139,
  140,
  144,
  145,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  181,
  186,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  217,
  218,
  233,
  239,
  1,93,1,
  221,
  0xFE,
  1,92,2,
  169,
  175,
  0xFE,
  0xFE,
  1,46,37,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  217,
  239,
  1,99,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,92,2,
  221,
  227,
  0xFE,
  0xFE,
  1,48,40,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  227,
  233,
  234,
  235,
  239,
  240,
  241,
  1,99,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,97,2,
  233,
  239,
  0xFE,
  1,93,6,
  221,
  222,
  223,
  227,
  228,
  229,
  0xFE,
  0xFE,
  1,50,16,
  127,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  197,
  203,
  204,
  208,
  209,
  1,103,3,
  197,
  203,
  208,
  0xFE,
  1,99,4,
  191,
  197,
  233,
  239,
  0xFE,
  1,97,4,
  221,
  222,
  227,
  228,
  0xFE,
  0xFE,
  1,52,10,
  127,
  133,
  157,
  163,
  164,
  169,
  175,
  191,
  203,
  208,
  1,103,2,
  191,
  197,
  0xFE,
  1,101,2,
  221,
  227,
  0xFE,
  1,99,7,
  163,
  169,
  175,
  203,
  208,
  233,
  239,
  0xFE,
  0xFE,
  1,54,3,
  133,
  157,
  163,
  1,105,1,
  221,
  0xFE,
  1,103,6,
  169,
  175,
  203,
  208,
  233,
  239,
  0xFE,
  0xFE,
  1,93,10,
  163,
  221,
  222,
  227,
  228,
  233,
  234,
  235,
  239,
  240,
  0xFE,
  1,92,28,
  127,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  221,
  222,
  223,
  224,
  227,
  228,
  229,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  249,
  0xFE,
  1,96,12,
  133,
  157,
  163,
  169,
  170,
  175,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,98,3,
  163,
  233,
  239,
  0xFE,
  0xFE,
  0,255,119,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  20,
  25,
  26,
  34,
  35,
  38,
  39,
  44,
  45,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  127,
  128,
  129,
  130,
  131,
  133,
  134,
  135,
  136,
  137,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  149,
  150,
  151,
  153,
  154,
  155,
  157,
  158,
  159,
  160,
  161,
  162,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  172,
  173,
  175,
  176,
  177,
  178,
  179,
  181,
  182,
  183,
  184,
  186,
  187,
  188,
  189,
  191,
  192,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  207,
  208,
  209,
  210,
  211,
  212,
  213,
  214,
  217,
  218,
  221,
  227,
  228,
  229,
  230,
  233,
  234,
  235,
  236,
  237,
  239,
  240,
  241,
  242,
  243,
  0,32,6,
  157,
  163,
  169,
  170,
  175,
  176,
  0xFE,
  0,86,6,
  157,
  163,
  169,
  170,
  175,
  176,
  0xFE,
  0,140,4,
  95,
  98,
  104,
  110,
  0xFE,
  1,55,0,
  1,102,2,
  197,
  227,
  0xFE,
  0xFE,
  1,53,4,
  203,
  204,
  208,
  209,
  1,96,1,
  227,
  0xFE,
  1,98,4,
  191,
  197,
  198,
  227,
  0xFE,
  0xFE,
  1,51,21,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  217,
  1,96,2,
  197,
  198,
  0xFE,
  1,98,4,
  191,
  197,
  208,
  209,
  0xFE,
  0xFE,
  1,49,52,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  186,
  191,
  192,
  193,
  194,
  195,
  197,
  198,
  199,
  200,
  201,
  203,
  204,
  205,
  206,
  207,
  208,
  209,
  210,
  211,
  212,
  213,
  214,
  217,
  218,
  239,
  1,92,2,
  197,
  227,
  0xFE,
  1,98,2,
  191,
  197,
  0xFE,
  0xFE,
  1,47,47,
  133,
  134,
  139,
  144,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  186,
  191,
  192,
  193,
  194,
  195,
  197,
  198,
  199,
  200,
  201,
  203,
  204,
  205,
  206,
  207,
  208,
  209,
  210,
  211,
  212,
  213,
  214,
  217,
  218,
  1,93,1,
  227,
  0xFE,
  1,92,1,
  239,
  0xFE,
  1,96,2,
  221,
  227,
  0xFE,
  0xFE,
  1,45,49,
  133,
  139,
  144,
  157,
  158,
  159,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  181,
  186,
  191,
  192,
  193,
  194,
  195,
  197,
  198,
  199,
  200,
  201,
  203,
  204,
  205,
  208,
  209,
  210,
  211,
  213,
  214,
  215,
  217,
  218,
  219,
  221,
  222,
  227,
  233,
  234,
  235,
  239,
  240,
  241,
  1,93,4,
  233,
  234,
  239,
  240,
  0xFE,
  1,92,7,
  221,
  222,
  223,
  227,
  228,
  229,
  239,
  0xFE,
  0xFE,
  1,44,31,
  127,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  215,
  217,
  218,
  219,
  1,101,2,
  233,
  239,
  0xFE,
  1,99,13,
  191,
  192,
  197,
  198,
  203,
  208,
  221,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  1,97,8,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  239,
  0xFE,
  1,92,1,
  157,
  0xFE,
  0xFE,
  1,46,25,
  127,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  1,103,2,
  233,
  239,
  0xFE,
  1,101,2,
  221,
  227,
  0xFE,
  1,99,4,
  197,
  203,
  208,
  227,
  0xFE,
  0xFE,
  1,48,29,
  3,
  7,
  13,
  104,
  110,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  197,
  203,
  208,
  233,
  234,
  239,
  240,
  1,103,2,
  203,
  208,
  0xFE,
  1,99,2,
  191,
  197,
  0xFE,
  1,97,4,
  157,
  163,
  233,
  239,
  0xFE,
  1,93,4,
  221,
  222,
  227,
  228,
  0xFE,
  0xFE,
  1,50,9,
  127,
  133,
  157,
  163,
  164,
  175,
  191,
  203,
  208,
  1,103,1,
  197,
  0xFE,
  0xFE,
  1,52,1,
  163,
  1,99,1,
  175,
  0xFE,
  0xFE,
  1,109,0,
  1,70,2,
  203,
  208,
  0xFE,
  1,72,2,
  191,
  197,
  0xFE,
  0xFE,
  1,99,8,
  197,
  203,
  204,
  209,
  233,
  234,
  239,
  240,
  0xFE,
  1,97,16,
  169,
  170,
  175,
  176,
  221,
  222,
  227,
  228,
  233,
  234,
  235,
  236,
  239,
  240,
  241,
  242,
  0xFE,
  1,93,30,
  157,
  158,
  163,
  164,
  165,
  169,
  175,
  221,
  222,
  223,
  224,
  225,
  226,
  227,
  228,
  229,
  230,
  231,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  246,
  247,
  249,
  250,
  251,
  0xFE,
  1,92,27,
  127,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  175,
  221,
  222,
  227,
  228,
  229,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  246,
  247,
  249,
  250,
  251,
  0xFE,
  1,96,5,
  133,
  157,
  158,
  233,
  239,
  0xFE,
  0xFE,
  0,253,114,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  20,
  25,
  26,
  34,
  35,
  38,
  44,
  50,
  56,
  65,
  66,
  71,
  72,
  77,
  83,
  84,
  89,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  117,
  122,
  123,
  127,
  128,
  129,
  130,
  131,
  133,
  134,
  135,
  136,
  137,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  149,
  150,
  151,
  153,
  154,
  155,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  183,
  184,
  186,
  187,
  188,
  189,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  214,
  217,
  218,
  221,
  222,
  227,
  228,
  229,
  233,
  234,
  235,
  236,
  239,
  240,
  241,
  242,
  0,0,3,
  3,
  7,
  13,
  0xFE,
  0,32,3,
  127,
  157,
  163,
  0xFE,
  0,86,5,
  71,
  77,
  127,
  157,
  163,
  0,143,2,
  65,
  98,
  0xFE,
  0xFE,
  0,140,4,
  98,
  104,
  110,
  157,
  0xFE,
  1,57,2,
  203,
  208,
  0xFE,
  1,55,18,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,53,31,
  157,
  158,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  214,
  217,
  218,
  0xFE,
  1,51,36,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  186,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  207,
  208,
  209,
  210,
  211,
  212,
  213,
  214,
  217,
  218,
  0xFE,
  1,49,48,
  127,
  128,
  133,
  134,
  135,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  201,
  203,
  204,
  205,
  206,
  207,
  208,
  209,
  210,
  211,
  212,
  213,
  214,
  217,
  218,
  0xFE,
  1,47,40,
  133,
  139,
  144,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  181,
  186,
  191,
  192,
  193,
  194,
  195,
  197,
  198,
  199,
  200,
  201,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  214,
  217,
  218,
  0xFE,
  1,45,39,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  208,
  209,
  210,
  211,
  213,
  214,
  217,
  218,
  219,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  1,99,2,
  203,
  208,
  0xFE,
  1,93,4,
  233,
  234,
  239,
  240,
  0xFE,
  1,92,4,
  221,
  222,
  227,
  228,
  0xFE,
  1,96,2,
  191,
  197,
  0xFE,
  0xFE,
  1,44,17,
  127,
  133,
  157,
  163,
  169,
  175,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  214,
  217,
  218,
  1,101,4,
  233,
  234,
  239,
  240,
  0xFE,
  1,99,9,
  163,
  191,
  221,
  227,
  228,
  233,
  234,
  239,
  240,
  0xFE,
  1,97,6,
  203,
  208,
  221,
  222,
  227,
  228,
  0xFE,
  0xFE,
  1,46,18,
  127,
  128,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  197,
  203,
  208,
  233,
  239,
  1,99,1,
  197,
  0xFE,
  1,97,2,
  191,
  197,
  0xFE,
  1,93,2,
  221,
  227,
  0xFE,
  0xFE,
  1,48,16,
  127,
  128,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  197,
  203,
  208,
  1,101,2,
  233,
  239,
  0xFE,
  1,99,2,
  221,
  227,
  0xFE,
  0xFE,
  1,99,10,
  169,
  175,
  197,
  227,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  1,97,25,
  157,
  158,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  235,
  236,
  239,
  240,
  241,
  242,
  0xFE,
  1,93,22,
  157,
  163,
  169,
  175,
  221,
  222,
  223,
  224,
  225,
  227,
  228,
  229,
  230,
  233,
  234,
  239,
  240,
  241,
  245,
  246,
  249,
  250,
  0xFE,
  1,92,13,
  127,
  133,
  157,
  163,
  221,
  227,
  228,
  233,
  239,
  245,
  246,
  249,
  250,
  0xFE,
  1,96,2,
  233,
  239,
  0xFE,
  1,98,0,
  1,57,1,
  197,
  0xFE,
  1,55,1,
  197,
  0xFE,
  1,53,1,
  197,
  0xFE,
  0xFE,
  0xFE,
  0,251,87,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  20,
  25,
  26,
  34,
  38,
  44,
  50,
  56,
  65,
  71,
  72,
  77,
  78,
  83,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  150,
  151,
  153,
  154,
  155,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  183,
  186,
  187,
  188,
  191,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  221,
  222,
  227,
  228,
  233,
  234,
  235,
  239,
  240,
  241,
  0,0,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,32,3,
  38,
  44,
  157,
  0xFE,
  0,36,0,
  0,95,1,
  34,
  0xFE,
  0xFE,
  0,86,3,
  71,
  77,
  157,
  0xFE,
  0,88,0,
  0,147,1,
  98,
  0xFE,
  0xFE,
  0,104,0,
  0,177,1,
  98,
  0xFE,
  0xFE,
  0,140,5,
  95,
  98,
  99,
  104,
  110,
  0xFE,
  1,67,0,
  1,104,1,
  197,
  0xFE,
  0xFE,
  1,65,0,
  1,100,1,
  197,
  0xFE,
  0xFE,
  1,63,0,
  1,96,3,
  133,
  197,
  227,
  0xFE,
  0xFE,
  1,59,8,
  191,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  0xFE,
  1,57,16,
  163,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,55,26,
  157,
  158,
  163,
  164,
  169,
  170,
  171,
  175,
  176,
  177,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  217,
  0xFE,
  1,53,32,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  214,
  217,
  218,
  0xFE,
  1,51,33,
  157,
  158,
  159,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  217,
  0xFE,
  1,49,45,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  181,
  186,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  201,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  213,
  214,
  217,
  218,
  233,
  239,
  1,93,2,
  221,
  227,
  0xFE,
  0xFE,
  1,47,31,
  127,
  133,
  134,
  139,
  144,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  213,
  1,99,2,
  203,
  208,
  0xFE,
  0xFE,
  1,45,27,
  127,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  213,
  214,
  217,
  218,
  221,
  233,
  239,
  1,99,1,
  197,
  0xFE,
  1,93,1,
  227,
  0xFE,
  1,92,2,
  221,
  227,
  0xFE,
  0xFE,
  1,44,12,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  208,
  213,
  217,
  1,103,2,
  233,
  239,
  0xFE,
  1,101,1,
  227,
  0xFE,
  1,99,2,
  233,
  239,
  0xFE,
  1,97,3,
  221,
  227,
  228,
  0xFE,
  0xFE,
  1,46,8,
  133,
  157,
  163,
  175,
  191,
  197,
  203,
  208,
  0xFE,
  1,48,9,
  133,
  157,
  158,
  163,
  164,
  175,
  191,
  203,
  208,
  0xFE,
  1,50,1,
  191,
  0xFE,
  1,101,6,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  1,99,20,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  1,97,20,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  221,
  222,
  223,
  227,
  228,
  229,
  230,
  233,
  234,
  239,
  240,
  0xFE,
  1,93,15,
  157,
  163,
  169,
  175,
  221,
  222,
  223,
  224,
  227,
  228,
  229,
  233,
  239,
  245,
  249,
  0xFE,
  1,92,8,
  221,
  222,
  227,
  228,
  233,
  239,
  245,
  249,
  0xFE,
  0xFE,
  0,249,74,
  3,
  4,
  7,
  8,
  13,
  14,
  34,
  38,
  44,
  65,
  71,
  77,
  98,
  104,
  105,
  110,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  149,
  150,
  151,
  153,
  154,
  155,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  183,
  186,
  187,
  188,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  221,
  222,
  223,
  227,
  233,
  234,
  239,
  240,
  0,36,0,
  0,95,1,
  34,
  0xFE,
  0xFE,
  0,88,0,
  0,147,1,
  98,
  0xFE,
  0xFE,
  1,75,0,
  1,108,4,
  133,
  163,
  164,
  165,
  0xFE,
  0xFE,
  1,61,4,
  203,
  204,
  208,
  209,
  0xFE,
  1,59,12,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,57,21,
  157,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  0xFE,
  1,55,29,
  157,
  158,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  217,
  0xFE,
  1,53,30,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  217,
  0xFE,
  1,51,31,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  217,
  218,
  0xFE,
  1,49,31,
  127,
  128,
  133,
  134,
  139,
  144,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  0xFE,
  1,47,19,
  127,
  133,
  157,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  0xFE,
  1,45,20,
  127,
  133,
  139,
  144,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  233,
  239,
  1,92,2,
  221,
  227,
  0xFE,
  0xFE,
  1,44,12,
  133,
  191,
  192,
  193,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  1,99,2,
  233,
  239,
  0xFE,
  1,97,2,
  221,
  227,
  0xFE,
  0xFE,
  1,46,9,
  157,
  163,
  191,
  192,
  193,
  197,
  198,
  203,
  208,
  0xFE,
  1,48,6,
  133,
  157,
  163,
  175,
  203,
  208,
  0xFE,
  1,103,4,
  233,
  234,
  239,
  240,
  0xFE,
  1,101,13,
  163,
  169,
  175,
  221,
  222,
  227,
  228,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  1,99,16,
  157,
  158,
  163,
  164,
  165,
  169,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  239,
  240,
  0xFE,
  1,97,16,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  0xFE,
  1,93,11,
  157,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,92,8,
  221,
  227,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  0xFE,
  0,248,57,
  3,
  7,
  13,
  104,
  110,
  127,
  128,
  129,
  130,
  131,
  133,
  134,
  135,
  136,
  137,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  186,
  187,
  188,
  203,
  204,
  208,
  209,
  233,
  234,
  239,
  240,
  1,65,2,
  203,
  208,
  0xFE,
  1,63,6,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,61,14,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,59,20,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,57,25,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  0xFE,
  1,55,24,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,53,22,
  127,
  133,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  1,98,4,
  157,
  163,
  169,
  175,
  0xFE,
  0xFE,
  1,51,13,
  157,
  163,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  0xFE,
  1,49,24,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  163,
  164,
  169,
  170,
  171,
  175,
  176,
  177,
  191,
  192,
  197,
  198,
  203,
  208,
  1,93,2,
  221,
  227,
  0xFE,
  0xFE,
  1,47,10,
  157,
  163,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,45,9,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  221,
  0xFE,
  1,105,2,
  233,
  239,
  0xFE,
  1,103,7,
  163,
  221,
  222,
  227,
  228,
  233,
  239,
  0xFE,
  1,101,2,
  221,
  227,
  0xFE,
  1,99,1,
  221,
  1,44,1,
  221,
  0xFE,
  0xFE,
  1,97,2,
  221,
  227,
  0xFE,
  1,93,8,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  0xFE,
  0xFE,
  0,250,57,
  3,
  7,
  8,
  13,
  14,
  19,
  25,
  104,
  110,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  186,
  187,
  203,
  204,
  208,
  209,
  233,
  234,
  239,
  240,
  1,65,6,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,63,12,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,61,20,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,59,21,
  157,
  158,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,57,21,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  0xFE,
  1,55,19,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  210,
  1,102,1,
  163,
  0xFE,
  0xFE,
  1,53,11,
  157,
  163,
  169,
  175,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  1,98,3,
  133,
  157,
  163,
  0xFE,
  0xFE,
  1,51,5,
  157,
  191,
  197,
  203,
  208,
  0xFE,
  1,49,19,
  127,
  128,
  133,
  134,
  139,
  144,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  191,
  197,
  203,
  0xFE,
  1,47,5,
  191,
  192,
  197,
  203,
  208,
  0xFE,
  1,45,6,
  191,
  192,
  197,
  198,
  203,
  208,
  0xFE,
  1,105,1,
  227,
  1,46,3,
  163,
  169,
  175,
  0xFE,
  0xFE,
  1,97,1,
  221,
  0xFE,
  1,93,6,
  221,
  222,
  227,
  228,
  233,
  239,
  0xFE,
  0xFE,
  0,252,47,
  3,
  7,
  13,
  104,
  110,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  153,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  186,
  187,
  203,
  204,
  208,
  209,
  233,
  234,
  239,
  240,
  1,67,4,
  203,
  204,
  208,
  209,
  0xFE,
  1,65,10,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,63,18,
  157,
  163,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,61,19,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,59,16,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  0xFE,
  1,57,17,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  0xFE,
  1,55,10,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  208,
  209,
  0xFE,
  1,53,3,
  191,
  203,
  208,
  0xFE,
  1,51,2,
  191,
  197,
  0xFE,
  1,49,12,
  127,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,47,2,
  203,
  208,
  0xFE,
  1,45,4,
  191,
  192,
  197,
  198,
  0xFE,
  1,46,0,
  1,105,2,
  157,
  163,
  0xFE,
  0xFE,
  1,99,1,
  221,
  0xFE,
  1,93,4,
  221,
  222,
  227,
  228,
  0xFE,
  0xFE,
  0,254,24,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  177,
  1,77,0,
  1,108,2,
  191,
  197,
  0xFE,
  0xFE,
  1,69,2,
  203,
  208,
  0xFE,
  1,67,8,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,65,14,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,63,17,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  0xFE,
  1,61,15,
  157,
  158,
  163,
  164,
  169,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  0xFE,
  1,59,11,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  208,
  209,
  0xFE,
  1,57,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,55,1,
  191,
  0xFE,
  1,53,1,
  191,
  0xFE,
  1,49,11,
  127,
  133,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  0xFE,
  1,0,34,
  3,
  4,
  7,
  8,
  13,
  14,
  25,
  104,
  105,
  110,
  111,
  127,
  133,
  139,
  140,
  144,
  145,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  203,
  204,
  208,
  209,
  233,
  234,
  235,
  239,
  240,
  241,
  1,69,6,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,67,8,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,65,11,
  157,
  163,
  169,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,63,10,
  157,
  163,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,61,7,
  157,
  163,
  191,
  192,
  197,
  198,
  208,
  0xFE,
  1,59,3,
  191,
  197,
  208,
  0xFE,
  1,57,3,
  191,
  203,
  208,
  0xFE,
  1,55,6,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,53,8,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,51,11,
  133,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  0xFE,
  1,49,15,
  127,
  128,
  133,
  134,
  157,
  158,
  163,
  164,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  0xFE,
  1,47,3,
  203,
  208,
  209,
  0xFE,
  1,45,6,
  191,
  192,
  193,
  197,
  198,
  199,
  0xFE,
  1,99,1,
  227,
  0xFE,
  1,93,6,
  221,
  222,
  223,
  227,
  228,
  229,
  0xFE,
  0xFE,
  1,2,0,
  1,71,2,
  203,
  208,
  0xFE,
  1,69,6,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,67,9,
  163,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,65,7,
  163,
  191,
  192,
  197,
  198,
  203,
  208,
  0xFE,
  1,63,4,
  191,
  197,
  198,
  208,
  0xFE,
  0xFE,
  1,4,0,
  1,71,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,69,5,
  191,
  197,
  198,
  203,
  208,
  0xFE,
  1,67,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,65,2,
  191,
  197,
  0xFE,
  0xFE,
  1,6,0,
  1,71,2,
  191,
  197,
  0xFE,
  1,69,2,
  191,
  197,
  0xFE,
  0xFE,
  1,63,2,
  203,
  208,
  0xFE,
  1,61,10,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,59,19,
  157,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  1,98,2,
  203,
  208,
  0xFE,
  0xFE,
  1,57,26,
  157,
  158,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  191,
  192,
  193,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  1,96,2,
  203,
  208,
  0xFE,
  1,98,4,
  191,
  197,
  203,
  208,
  0xFE,
  0xFE,
  1,55,32,
  127,
  133,
  139,
  144,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  1,96,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,98,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,100,1,
  191,
  0xFE,
  1,102,1,
  191,
  0xFE,
  0xFE,
  1,53,33,
  127,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  1,92,2,
  203,
  208,
  0xFE,
  1,96,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,98,7,
  169,
  175,
  191,
  197,
  203,
  208,
  221,
  0xFE,
  1,100,2,
  191,
  208,
  0xFE,
  1,102,4,
  191,
  197,
  203,
  208,
  0xFE,
  0xFE,
  1,51,33,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  1,92,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,96,6,
  191,
  197,
  203,
  208,
  221,
  227,
  0xFE,
  1,98,10,
  133,
  191,
  192,
  197,
  203,
  204,
  208,
  209,
  221,
  233,
  0xFE,
  1,102,2,
  191,
  197,
  0xFE,
  0xFE,
  1,49,30,
  127,
  128,
  133,
  134,
  139,
  144,
  145,
  149,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  1,99,2,
  203,
  208,
  0xFE,
  1,93,4,
  203,
  208,
  221,
  227,
  0xFE,
  1,92,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,96,7,
  191,
  197,
  203,
  208,
  221,
  227,
  239,
  0xFE,
  1,98,10,
  191,
  192,
  197,
  198,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,100,2,
  221,
  227,
  0xFE,
  0xFE,
  1,47,35,
  127,
  133,
  134,
  139,
  140,
  144,
  145,
  149,
  153,
  157,
  163,
  169,
  175,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  214,
  217,
  218,
  221,
  233,
  239,
  1,99,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,93,6,
  191,
  197,
  203,
  208,
  221,
  227,
  0xFE,
  1,92,9,
  191,
  192,
  197,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,96,6,
  221,
  222,
  227,
  228,
  233,
  239,
  0xFE,
  1,98,4,
  203,
  208,
  233,
  239,
  0xFE,
  0xFE,
  1,45,40,
  127,
  133,
  139,
  144,
  157,
  163,
  169,
  175,
  191,
  192,
  193,
  194,
  195,
  197,
  198,
  199,
  200,
  201,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  215,
  217,
  218,
  219,
  221,
  222,
  227,
  228,
  233,
  234,
  235,
  239,
  240,
  241,
  1,99,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,97,2,
  203,
  208,
  0xFE,
  1,93,14,
  127,
  133,
  139,
  144,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,92,13,
  197,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,96,4,
  233,
  239,
  245,
  249,
  0xFE,
  0xFE,
  1,44,16,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  215,
  217,
  218,
  219,
  1,103,4,
  203,
  208,
  233,
  239,
  0xFE,
  1,101,8,
  139,
  144,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,99,17,
  127,
  133,
  191,
  192,
  197,
  198,
  203,
  208,
  221,
  222,
  227,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  1,97,16,
  191,
  197,
  203,
  208,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,93,8,
  203,
  208,
  221,
  227,
  233,
  239,
  245,
  249,
  0xFE,
  0xFE,
  1,46,7,
  191,
  203,
  204,
  208,
  209,
  213,
  217,
  1,105,2,
  233,
  239,
  0xFE,
  1,103,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,101,8,
  191,
  197,
  221,
  222,
  227,
  228,
  233,
  239,
  0xFE,
  1,99,6,
  191,
  197,
  198,
  221,
  227,
  228,
  0xFE,
  1,97,7,
  191,
  197,
  221,
  222,
  227,
  233,
  239,
  0xFE,
  0xFE,
  1,48,1,
  163,
  1,107,2,
  233,
  239,
  0xFE,
  1,105,4,
  221,
  227,
  233,
  239,
  0xFE,
  1,103,3,
  197,
  221,
  227,
  0xFE,
  1,101,4,
  191,
  197,
  221,
  227,
  0xFE,
  1,99,6,
  191,
  197,
  203,
  208,
  221,
  227,
  0xFE,
  0xFE,
  1,105,2,
  233,
  239,
  1,52,1,
  221,
  0xFE,
  1,54,2,
  191,
  221,
  0xFE,
  0xFE,
  1,103,8,
  203,
  208,
  221,
  227,
  233,
  234,
  239,
  240,
  1,50,1,
  221,
  0xFE,
  1,52,2,
  191,
  221,
  0xFE,
  0xFE,
  1,101,11,
  157,
  158,
  163,
  164,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  1,50,2,
  191,
  221,
  0xFE,
  0xFE,
  1,99,22,
  127,
  133,
  139,
  144,
  157,
  163,
  169,
  175,
  191,
  197,
  203,
  204,
  208,
  209,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  0xFE,
  1,97,22,
  127,
  133,
  134,
  139,
  144,
  157,
  163,
  169,
  175,
  191,
  203,
  208,
  221,
  222,
  227,
  228,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  1,95,39,
  127,
  133,
  139,
  140,
  144,
  153,
  157,
  169,
  175,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  235,
  236,
  239,
  240,
  241,
  242,
  245,
  246,
  249,
  250,
  0,32,2,
  221,
  227,
  0xFE,
  0,86,2,
  221,
  227,
  0xFE,
  1,17,0,
  1,60,2,
  191,
  197,
  0xFE,
  1,62,2,
  203,
  208,
  0xFE,
  0xFE,
  1,15,0,
  1,58,4,
  191,
  192,
  197,
  198,
  0xFE,
  1,60,5,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,62,1,
  208,
  0xFE,
  1,72,2,
  191,
  197,
  0xFE,
  1,74,2,
  203,
  208,
  0xFE,
  0xFE,
  1,13,0,
  1,54,1,
  197,
  0xFE,
  1,56,5,
  163,
  191,
  192,
  197,
  198,
  0xFE,
  1,58,6,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,60,2,
  203,
  208,
  0xFE,
  1,70,2,
  157,
  163,
  0xFE,
  0xFE,
  1,11,0,
  1,52,1,
  197,
  0xFE,
  1,54,10,
  157,
  163,
  164,
  191,
  192,
  197,
  198,
  199,
  203,
  208,
  0xFE,
  1,56,6,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,58,2,
  203,
  208,
  0xFE,
  0xFE,
  1,9,0,
  1,50,2,
  163,
  197,
  0xFE,
  1,52,12,
  157,
  158,
  163,
  164,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  208,
  0xFE,
  1,54,8,
  169,
  175,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,56,4,
  191,
  197,
  203,
  208,
  0xFE,
  0xFE,
  1,7,0,
  1,48,1,
  203,
  0xFE,
  1,50,13,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  0xFE,
  1,52,10,
  169,
  175,
  191,
  197,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,54,6,
  191,
  192,
  197,
  198,
  203,
  208,
  0xFE,
  1,56,3,
  191,
  203,
  208,
  0xFE,
  1,58,1,
  197,
  0xFE,
  1,60,3,
  197,
  203,
  208,
  0xFE,
  0xFE,
  1,5,0,
  1,46,5,
  191,
  192,
  193,
  197,
  198,
  0xFE,
  1,48,14,
  157,
  158,
  163,
  164,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  0xFE,
  1,50,19,
  163,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  233,
  234,
  239,
  240,
  0xFE,
  1,52,9,
  163,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,54,10,
  163,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,56,12,
  133,
  157,
  163,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  227,
  0xFE,
  0xFE,
  1,3,0,
  1,46,13,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  203,
  204,
  208,
  221,
  233,
  239,
  0xFE,
  1,48,25,
  133,
  157,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  227,
  233,
  239,
  0xFE,
  1,50,30,
  127,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  239,
  0xFE,
  1,52,15,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  204,
  209,
  221,
  222,
  227,
  228,
  233,
  239,
  0xFE,
  1,54,2,
  203,
  208,
  0xFE,
  0xFE,
  1,1,0,
  1,44,6,
  133,
  163,
  191,
  197,
  203,
  208,
  0xFE,
  1,46,9,
  163,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,48,20,
  133,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  221,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  1,50,18,
  133,
  139,
  144,
  157,
  163,
  164,
  169,
  170,
  175,
  176,
  203,
  204,
  208,
  209,
  233,
  234,
  239,
  240,
  0xFE,
  0xFE,
  0,255,39,
  157,
  158,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  221,
  222,
  223,
  224,
  227,
  228,
  229,
  230,
  233,
  234,
  235,
  236,
  237,
  239,
  240,
  241,
  242,
  243,
  1,49,2,
  203,
  208,
  0xFE,
  1,47,16,
  163,
  164,
  169,
  175,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,45,20,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  0xFE,
  1,44,9,
  191,
  197,
  198,
  199,
  203,
  208,
  221,
  227,
  228,
  0xFE,
  1,46,2,
  191,
  197,
  0xFE,
  1,48,7,
  133,
  157,
  163,
  221,
  227,
  233,
  239,
  0xFE,
  1,50,7,
  127,
  133,
  144,
  157,
  163,
  169,
  175,
  0xFE,
  0xFE,
  0,253,32,
  157,
  158,
  159,
  163,
  164,
  165,
  166,
  169,
  170,
  175,
  191,
  192,
  197,
  198,
  203,
  208,
  221,
  222,
  223,
  224,
  227,
  228,
  229,
  230,
  233,
  234,
  235,
  236,
  239,
  240,
  241,
  242,
  1,53,2,
  203,
  208,
  0xFE,
  1,51,8,
  163,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,49,16,
  163,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,47,19,
  157,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,45,14,
  157,
  158,
  163,
  164,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  0xFE,
  1,44,3,
  197,
  198,
  227,
  0xFE,
  1,48,1,
  163,
  0xFE,
  1,50,3,
  133,
  157,
  163,
  0xFE,
  0xFE,
  0,251,27,
  157,
  158,
  163,
  164,
  169,
  170,
  171,
  175,
  176,
  177,
  191,
  192,
  197,
  198,
  203,
  208,
  221,
  222,
  227,
  228,
  229,
  233,
  234,
  235,
  239,
  240,
  241,
  1,53,4,
  203,
  204,
  208,
  209,
  0xFE,
  1,51,13,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,49,16,
  157,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,47,15,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  0xFE,
  1,45,6,
  191,
  197,
  203,
  208,
  233,
  239,
  0xFE,
  1,44,2,
  197,
  227,
  0xFE,
  0xFE,
  0,249,12,
  157,
  163,
  169,
  175,
  191,
  197,
  221,
  227,
  233,
  234,
  239,
  240,
  1,55,4,
  203,
  204,
  208,
  209,
  0xFE,
  1,53,10,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,51,16,
  157,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,49,13,
  157,
  158,
  163,
  164,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  0xFE,
  1,47,7,
  157,
  163,
  191,
  197,
  198,
  203,
  208,
  0xFE,
  1,45,2,
  203,
  208,
  0xFE,
  1,44,3,
  197,
  203,
  208,
  0xFE,
  1,46,3,
  191,
  203,
  208,
  0xFE,
  1,48,2,
  191,
  197,
  0xFE,
  0xFE,
  0,248,9,
  221,
  222,
  223,
  227,
  228,
  233,
  234,
  239,
  240,
  1,59,2,
  203,
  208,
  0xFE,
  1,57,6,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,55,9,
  163,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,53,8,
  157,
  163,
  191,
  192,
  197,
  198,
  203,
  208,
  0xFE,
  1,51,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,49,2,
  191,
  197,
  0xFE,
  1,47,1,
  197,
  0xFE,
  1,45,1,
  191,
  0xFE,
  0xFE,
  0,250,5,
  221,
  222,
  227,
  233,
  239,
  1,61,1,
  203,
  0xFE,
  1,59,6,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,57,8,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,55,6,
  191,
  192,
  197,
  198,
  203,
  208,
  0xFE,
  1,53,2,
  191,
  197,
  0xFE,
  0xFE,
  0,252,2,
  233,
  239,
  1,61,2,
  203,
  208,
  0xFE,
  1,59,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,57,2,
  191,
  197,
  0xFE,
  0xFE,
  0,254,0,
  1,61,2,
  191,
  197,
  0xFE,
  1,59,1,
  197,
  0xFE,
  0xFE,
  1,0,5,
  221,
  233,
  234,
  239,
  240,
  0xFE,
  1,57,2,
  203,
  208,
  0xFE,
  1,55,6,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,53,10,
  157,
  163,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,51,13,
  139,
  144,
  157,
  163,
  164,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,49,12,
  127,
  133,
  139,
  144,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,47,14,
  127,
  133,
  139,
  144,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  233,
  239,
  0xFE,
  1,45,16,
  127,
  133,
  139,
  144,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  233,
  239,
  0xFE,
  1,44,15,
  191,
  192,
  197,
  198,
  199,
  203,
  208,
  213,
  217,
  221,
  222,
  227,
  228,
  233,
  239,
  0xFE,
  1,46,9,
  191,
  197,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  0xFE,
  1,48,2,
  203,
  208,
  0xFE,
  0xFE,
  1,93,33,
  127,
  133,
  139,
  144,
  169,
  175,
  191,
  197,
  198,
  203,
  208,
  221,
  222,
  223,
  224,
  225,
  227,
  228,
  229,
  230,
  231,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  246,
  247,
  249,
  250,
  251,
  0xFE,
  1,92,16,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  239,
  240,
  245,
  246,
  247,
  249,
  250,
  251,
  0xFE,
  1,94,10,
  203,
  208,
  213,
  217,
  233,
  234,
  239,
  240,
  245,
  249,
  1,17,0,
  1,56,2,
  191,
  197,
  0xFE,
  1,58,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,60,2,
  203,
  208,
  0xFE,
  0xFE,
  1,15,0,
  1,52,1,
  197,
  0xFE,
  1,54,5,
  191,
  192,
  197,
  198,
  208,
  0xFE,
  1,56,6,
  191,
  197,
  198,
  203,
  204,
  208,
  0xFE,
  1,58,4,
  191,
  197,
  203,
  208,
  0xFE,
  0xFE,
  1,13,0,
  1,48,4,
  233,
  234,
  239,
  240,
  0xFE,
  1,50,2,
  191,
  197,
  0xFE,
  1,52,6,
  163,
  191,
  192,
  197,
  198,
  208,
  0xFE,
  1,54,8,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,56,6,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,58,2,
  203,
  208,
  0xFE,
  0xFE,
  1,11,6,
  233,
  234,
  235,
  239,
  240,
  241,
  1,48,2,
  191,
  197,
  0xFE,
  1,50,9,
  157,
  163,
  191,
  192,
  197,
  198,
  203,
  208,
  209,
  0xFE,
  1,52,9,
  163,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,54,7,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,56,2,
  203,
  208,
  0xFE,
  0xFE,
  1,9,6,
  169,
  175,
  233,
  234,
  239,
  240,
  1,46,4,
  191,
  197,
  198,
  203,
  0xFE,
  1,48,11,
  157,
  163,
  164,
  191,
  192,
  197,
  198,
  199,
  203,
  208,
  209,
  0xFE,
  1,50,14,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,52,10,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,54,6,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,56,2,
  203,
  208,
  0xFE,
  0xFE,
  1,7,23,
  157,
  158,
  163,
  164,
  169,
  170,
  171,
  175,
  176,
  177,
  221,
  222,
  227,
  228,
  229,
  233,
  234,
  235,
  236,
  239,
  240,
  241,
  242,
  1,44,5,
  191,
  197,
  203,
  233,
  239,
  0xFE,
  1,46,13,
  157,
  158,
  163,
  164,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  208,
  209,
  0xFE,
  1,48,22,
  157,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,50,13,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,52,6,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,54,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,56,3,
  197,
  203,
  208,
  0xFE,
  0xFE,
  1,5,34,
  157,
  158,
  159,
  163,
  164,
  165,
  166,
  169,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  203,
  221,
  222,
  223,
  224,
  227,
  228,
  229,
  230,
  233,
  234,
  235,
  236,
  237,
  239,
  240,
  241,
  242,
  243,
  1,45,2,
  197,
  198,
  0xFE,
  1,44,14,
  157,
  158,
  163,
  164,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  0xFE,
  1,46,19,
  157,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,48,24,
  163,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  221,
  227,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  1,50,11,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,52,11,
  133,
  157,
  163,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,54,2,
  203,
  208,
  0xFE,
  0xFE,
  1,3,41,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  221,
  222,
  223,
  224,
  227,
  228,
  229,
  233,
  234,
  235,
  236,
  237,
  239,
  240,
  241,
  242,
  243,
  1,45,6,
  191,
  197,
  198,
  199,
  227,
  228,
  0xFE,
  1,44,25,
  133,
  157,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  221,
  222,
  227,
  228,
  233,
  239,
  0xFE,
  1,46,34,
  127,
  128,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  0xFE,
  1,48,10,
  133,
  157,
  163,
  169,
  175,
  191,
  203,
  204,
  208,
  209,
  0xFE,
  1,50,2,
  203,
  208,
  0xFE,
  0xFE,
  1,1,31,
  127,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  163,
  164,
  169,
  170,
  171,
  175,
  176,
  177,
  191,
  192,
  197,
  198,
  221,
  222,
  227,
  228,
  233,
  234,
  235,
  239,
  240,
  241,
  1,45,19,
  127,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  197,
  198,
  199,
  200,
  203,
  221,
  227,
  228,
  0xFE,
  1,44,11,
  133,
  157,
  163,
  169,
  175,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,46,10,
  133,
  139,
  144,
  157,
  163,
  164,
  169,
  175,
  203,
  208,
  0xFE,
  0xFE,
  0,255,14,
  127,
  133,
  139,
  144,
  157,
  163,
  169,
  175,
  221,
  227,
  233,
  234,
  239,
  240,
  1,53,4,
  203,
  204,
  208,
  209,
  0xFE,
  1,51,21,
  157,
  163,
  164,
  169,
  170,
  175,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  221,
  227,
  228,
  0xFE,
  1,49,21,
  157,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  0xFE,
  1,47,12,
  163,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  221,
  239,
  0xFE,
  1,45,9,
  133,
  139,
  144,
  157,
  163,
  164,
  169,
  175,
  203,
  0xFE,
  1,44,3,
  133,
  157,
  163,
  0xFE,
  1,46,6,
  127,
  133,
  157,
  163,
  169,
  175,
  0xFE,
  0xFE,
  0,253,10,
  127,
  133,
  157,
  163,
  164,
  169,
  221,
  227,
  233,
  239,
  1,57,4,
  197,
  198,
  203,
  208,
  0xFE,
  1,55,15,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,53,15,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,51,18,
  157,
  158,
  163,
  164,
  170,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,49,14,
  157,
  158,
  163,
  164,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  0xFE,
  1,47,4,
  191,
  192,
  197,
  198,
  0xFE,
  1,45,3,
  133,
  157,
  163,
  0xFE,
  1,46,1,
  163,
  0xFE,
  0xFE,
  0,251,2,
  233,
  239,
  1,59,2,
  203,
  208,
  0xFE,
  1,57,7,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,55,11,
  163,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,53,17,
  157,
  163,
  164,
  169,
  175,
  176,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,51,17,
  133,
  139,
  144,
  157,
  163,
  164,
  169,
  175,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  0xFE,
  1,49,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,47,3,
  191,
  192,
  197,
  0xFE,
  0xFE,
  0,249,0,
  1,61,2,
  203,
  208,
  0xFE,
  1,59,6,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,57,8,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,55,13,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,53,12,
  157,
  158,
  163,
  164,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  208,
  0xFE,
  1,51,7,
  127,
  133,
  157,
  163,
  191,
  197,
  208,
  0xFE,
  1,49,3,
  191,
  203,
  208,
  0xFE,
  1,47,2,
  191,
  197,
  0xFE,
  0xFE,
  0,248,0,
  1,61,2,
  203,
  208,
  0xFE,
  1,59,7,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,57,8,
  157,
  163,
  191,
  192,
  197,
  198,
  203,
  208,
  0xFE,
  1,55,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,53,2,
  191,
  197,
  0xFE,
  1,51,6,
  133,
  157,
  163,
  169,
  175,
  197,
  0xFE,
  1,49,1,
  191,
  0xFE,
  0xFE,
  0,250,0,
  1,63,2,
  203,
  208,
  0xFE,
  1,61,6,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,59,5,
  191,
  192,
  197,
  198,
  203,
  0xFE,
  1,57,2,
  191,
  197,
  0xFE,
  1,51,3,
  133,
  157,
  163,
  0xFE,
  0xFE,
  0,252,0,
  1,63,2,
  203,
  208,
  0xFE,
  1,61,2,
  191,
  197,
  0xFE,
  0xFE,
  0,254,0,
  1,63,1,
  197,
  0xFE,
  0xFE,
  1,61,2,
  203,
  208,
  0xFE,
  1,59,6,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,57,9,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,55,12,
  133,
  139,
  144,
  157,
  163,
  164,
  191,
  192,
  197,
  198,
  203,
  208,
  0xFE,
  1,53,9,
  127,
  133,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,51,14,
  127,
  133,
  139,
  144,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,49,16,
  127,
  133,
  144,
  191,
  192,
  193,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  221,
  227,
  0xFE,
  1,47,19,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  213,
  217,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  0xFE,
  1,45,9,
  197,
  203,
  204,
  208,
  209,
  213,
  217,
  233,
  239,
  0xFE,
  0xFE,
  0xFE,
  0,208,118,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  20,
  25,
  26,
  27,
  34,
  35,
  38,
  39,
  44,
  45,
  50,
  51,
  56,
  57,
  65,
  66,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  84,
  89,
  90,
  95,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  186,
  187,
  191,
  197,
  198,
  199,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  217,
  221,
  227,
  228,
  229,
  233,
  234,
  235,
  236,
  239,
  240,
  241,
  242,
  245,
  249,
  0,0,9,
  0,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  0,37,1,
  3,
  0xFE,
  0,93,3,
  3,
  7,
  13,
  0xFE,
  0xFE,
  0,2,2,
  7,
  13,
  0,41,1,
  3,
  0xFE,
  0xFE,
  0,4,0,
  0,47,1,
  3,
  0xFE,
  0xFE,
  0,32,2,
  38,
  44,
  0,92,0,
  0,161,7,
  34,
  38,
  44,
  95,
  98,
  104,
  110,
  0xFE,
  0,159,2,
  104,
  110,
  0xFE,
  0,157,2,
  95,
  98,
  0xFE,
  0xFE,
  0,151,2,
  95,
  98,
  0xFE,
  0,149,2,
  104,
  110,
  0xFE,
  0,145,1,
  98,
  0xFE,
  0xFE,
  0,34,0,
  0,93,2,
  34,
  65,
  0xFE,
  0xFE,
  0,36,0,
  0,95,3,
  34,
  35,
  65,
  0xFE,
  0xFE,
  0,38,0,
  0,99,4,
  34,
  38,
  44,
  65,
  0xFE,
  0,97,1,
  34,
  0xFE,
  0xFE,
  0,40,0,
  0,101,1,
  65,
  0xFE,
  0xFE,
  0,42,0,
  0,105,10,
  31,
  34,
  35,
  38,
  44,
  62,
  65,
  66,
  71,
  77,
  0xFE,
  0xFE,
  0,86,3,
  66,
  71,
  77,
  0,149,2,
  95,
  98,
  0xFE,
  0,147,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,145,5,
  65,
  95,
  98,
  104,
  110,
  0xFE,
  0,143,4,
  95,
  98,
  104,
  110,
  0xFE,
  0xFE,
  0,88,0,
  0,153,1,
  98,
  0xFE,
  0,151,2,
  95,
  98,
  0xFE,
  0,149,5,
  65,
  95,
  98,
  104,
  110,
  0xFE,
  0,147,7,
  65,
  71,
  77,
  95,
  98,
  104,
  110,
  0xFE,
  0,145,2,
  104,
  110,
  0xFE,
  0xFE,
  0,90,0,
  0,153,2,
  95,
  98,
  0xFE,
  0,151,8,
  65,
  71,
  77,
  95,
  98,
  99,
  104,
  110,
  0xFE,
  0,149,3,
  98,
  104,
  110,
  0xFE,
  0xFE,
  0,92,0,
  0,34,0,
  0,167,3,
  98,
  104,
  110,
  0xFE,
  0xFE,
  0,159,0,
  0,250,2,
  95,
  98,
  0xFE,
  0xFE,
  0,157,3,
  98,
  104,
  110,
  0xFE,
  0,153,4,
  95,
  98,
  104,
  110,
  0xFE,
  0xFE,
  0,94,0,
  0,161,2,
  104,
  110,
  0xFE,
  0,159,8,
  65,
  71,
  77,
  95,
  98,
  99,
  104,
  110,
  0xFE,
  0,157,2,
  95,
  98,
  0xFE,
  0xFE,
  0,96,0,
  0,165,2,
  104,
  110,
  0xFE,
  0,163,7,
  65,
  71,
  77,
  95,
  98,
  104,
  110,
  0xFE,
  0,161,5,
  65,
  95,
  98,
  104,
  110,
  0xFE,
  0,159,2,
  95,
  98,
  0xFE,
  0xFE,
  0,98,0,
  0,167,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,165,5,
  65,
  95,
  98,
  104,
  110,
  0xFE,
  0,163,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,161,2,
  95,
  98,
  0xFE,
  0xFE,
  0,100,0,
  0,171,3,
  98,
  104,
  110,
  0xFE,
  0,169,5,
  65,
  95,
  98,
  104,
  110,
  0xFE,
  0,167,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,165,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,163,2,
  95,
  98,
  0xFE,
  0xFE,
  0,102,0,
  0,173,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,171,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,169,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,167,2,
  95,
  98,
  0xFE,
  0,165,1,
  98,
  0xFE,
  0xFE,
  0,104,0,
  0,177,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,175,3,
  98,
  104,
  110,
  0xFE,
  0,173,3,
  98,
  104,
  110,
  0xFE,
  0,171,3,
  95,
  98,
  104,
  0xFE,
  0,169,1,
  98,
  0xFE,
  0xFE,
  0,106,0,
  0,181,4,
  65,
  98,
  104,
  110,
  0xFE,
  0,173,1,
  98,
  0xFE,
  0xFE,
  0,108,0,
  0,185,5,
  65,
  98,
  99,
  104,
  110,
  0xFE,
  0xFE,
  0,110,0,
  0,189,1,
  98,
  0xFE,
  0xFE,
  0,155,0,
  0,0,2,
  3,
  13,
  0xFE,
  0,34,0,
  0,86,6,
  34,
  38,
  44,
  65,
  71,
  77,
  0xFE,
  0xFE,
  0,36,0,
  0,87,1,
  34,
  0xFE,
  0,86,1,
  34,
  0xFE,
  0xFE,
  0,46,0,
  0,103,2,
  34,
  65,
  0xFE,
  0xFE,
  0,88,0,
  0,0,1,
  3,
  0xFE,
  0,32,2,
  38,
  44,
  0xFE,
  0,34,2,
  34,
  65,
  0xFE,
  0xFE,
  0,92,27,
  34,
  62,
  63,
  65,
  66,
  67,
  68,
  69,
  71,
  72,
  77,
  78,
  83,
  95,
  96,
  98,
  99,
  100,
  101,
  102,
  104,
  105,
  110,
  111,
  116,
  122,
  133,
  0xFE,
  0xFE,
  0,140,10,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  0xFE,
  1,15,0,
  1,52,3,
  191,
  197,
  198,
  0xFE,
  1,54,3,
  197,
  203,
  208,
  0xFE,
  1,68,6,
  133,
  134,
  163,
  164,
  197,
  198,
  0xFE,
  0xFE,
  1,13,10,
  139,
  140,
  144,
  145,
  169,
  170,
  171,
  175,
  176,
  177,
  1,48,21,
  163,
  169,
  170,
  171,
  175,
  176,
  177,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  233,
  234,
  235,
  239,
  240,
  241,
  1,92,6,
  221,
  222,
  223,
  227,
  228,
  229,
  0xFE,
  0xFE,
  1,50,8,
  163,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  1,93,3,
  221,
  222,
  227,
  0xFE,
  0xFE,
  1,52,7,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,54,2,
  203,
  208,
  0xFE,
  1,56,1,
  197,
  0xFE,
  1,58,2,
  191,
  197,
  0xFE,
  1,60,5,
  163,
  191,
  197,
  203,
  208,
  0xFE,
  1,62,6,
  157,
  163,
  191,
  197,
  203,
  208,
  0xFE,
  0xFE,
  1,11,63,
  3,
  7,
  13,
  104,
  110,
  133,
  134,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  149,
  153,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  169,
  170,
  171,
  172,
  173,
  175,
  176,
  177,
  178,
  179,
  181,
  182,
  186,
  187,
  197,
  198,
  199,
  203,
  204,
  205,
  206,
  207,
  208,
  209,
  210,
  211,
  212,
  227,
  228,
  233,
  234,
  235,
  236,
  237,
  239,
  240,
  241,
  242,
  243,
  0,140,1,
  98,
  0xFE,
  1,44,15,
  191,
  192,
  193,
  194,
  195,
  197,
  198,
  199,
  200,
  201,
  203,
  204,
  208,
  209,
  210,
  1,98,2,
  157,
  163,
  0xFE,
  0xFE,
  1,46,13,
  191,
  192,
  193,
  197,
  198,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  0xFE,
  1,48,31,
  127,
  128,
  133,
  134,
  135,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  0xFE,
  1,50,8,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,52,8,
  163,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  1,99,1,
  157,
  0xFE,
  0xFE,
  1,54,6,
  163,
  191,
  197,
  198,
  203,
  208,
  0xFE,
  1,56,9,
  133,
  157,
  163,
  191,
  192,
  197,
  198,
  203,
  208,
  0xFE,
  1,58,7,
  133,
  157,
  163,
  191,
  197,
  203,
  208,
  0xFE,
  1,60,2,
  203,
  208,
  0xFE,
  1,92,10,
  221,
  222,
  223,
  224,
  225,
  227,
  228,
  229,
  230,
  231,
  0xFE,
  0xFE,
  1,9,52,
  7,
  13,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  172,
  173,
  175,
  176,
  177,
  178,
  179,
  181,
  182,
  186,
  187,
  203,
  204,
  208,
  209,
  233,
  234,
  239,
  240,
  1,47,0,
  1,104,6,
  169,
  175,
  203,
  208,
  233,
  239,
  0xFE,
  1,106,2,
  221,
  227,
  0xFE,
  1,108,2,
  233,
  239,
  0xFE,
  0xFE,
  1,45,0,
  1,104,2,
  191,
  197,
  0xFE,
  0xFE,
  1,44,4,
  191,
  192,
  197,
  198,
  1,96,1,
  191,
  0xFE,
  1,98,2,
  163,
  175,
  0xFE,
  0xFE,
  1,46,10,
  157,
  163,
  164,
  191,
  192,
  197,
  198,
  199,
  204,
  208,
  0xFE,
  1,48,38,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  157,
  158,
  159,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  186,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,50,20,
  127,
  133,
  134,
  139,
  157,
  158,
  163,
  164,
  165,
  169,
  175,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  0xFE,
  1,52,16,
  133,
  134,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  1,99,7,
  133,
  139,
  144,
  157,
  163,
  169,
  175,
  0xFE,
  0xFE,
  1,54,8,
  133,
  157,
  163,
  169,
  175,
  191,
  203,
  208,
  1,103,2,
  169,
  175,
  0xFE,
  0xFE,
  1,56,4,
  163,
  191,
  203,
  208,
  0xFE,
  1,92,5,
  221,
  222,
  223,
  227,
  228,
  0xFE,
  1,96,4,
  169,
  175,
  233,
  239,
  0xFE,
  1,98,2,
  221,
  227,
  0xFE,
  1,102,2,
  169,
  175,
  0xFE,
  0xFE,
  1,7,68,
  3,
  7,
  13,
  104,
  110,
  127,
  128,
  129,
  130,
  131,
  133,
  134,
  135,
  136,
  137,
  139,
  140,
  141,
  142,
  143,
  144,
  145,
  146,
  147,
  148,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  160,
  161,
  162,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  172,
  173,
  175,
  176,
  177,
  178,
  179,
  181,
  182,
  183,
  186,
  187,
  188,
  197,
  198,
  203,
  204,
  208,
  209,
  233,
  234,
  235,
  239,
  240,
  241,
  1,49,0,
  1,104,2,
  203,
  208,
  0xFE,
  0xFE,
  1,47,0,
  1,102,2,
  233,
  239,
  0xFE,
  1,104,6,
  133,
  157,
  163,
  197,
  221,
  227,
  0xFE,
  1,106,2,
  233,
  239,
  0xFE,
  0xFE,
  1,45,0,
  1,100,3,
  133,
  163,
  227,
  0xFE,
  1,104,2,
  203,
  208,
  0xFE,
  0xFE,
  1,44,14,
  133,
  134,
  157,
  163,
  164,
  169,
  175,
  191,
  192,
  193,
  197,
  198,
  199,
  200,
  0xFE,
  1,46,31,
  127,
  128,
  133,
  134,
  135,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  166,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,48,33,
  127,
  128,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  157,
  158,
  159,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  186,
  191,
  192,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,50,12,
  133,
  139,
  144,
  157,
  163,
  164,
  169,
  175,
  191,
  197,
  203,
  208,
  0xFE,
  1,52,9,
  133,
  157,
  163,
  169,
  175,
  191,
  197,
  203,
  208,
  1,99,8,
  133,
  139,
  144,
  157,
  163,
  164,
  169,
  175,
  0xFE,
  0xFE,
  1,54,3,
  157,
  203,
  208,
  1,103,7,
  133,
  139,
  144,
  157,
  163,
  169,
  175,
  0xFE,
  0xFE,
  1,92,7,
  163,
  221,
  222,
  223,
  227,
  228,
  229,
  0xFE,
  1,96,19,
  127,
  133,
  134,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  197,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  0xFE,
  1,98,6,
  133,
  157,
  163,
  221,
  233,
  239,
  0xFE,
  1,100,1,
  163,
  0xFE,
  0xFE,
  1,5,87,
  3,
  4,
  7,
  13,
  34,
  65,
  66,
  71,
  77,
  98,
  99,
  104,
  110,
  127,
  128,
  129,
  130,
  131,
  133,
  134,
  135,
  136,
  137,
  139,
  140,
  141,
  142,
  143,
  144,
  145,
  146,
  147,
  148,
  149,
  150,
  151,
  153,
  154,
  155,
  157,
  158,
  159,
  160,
  161,
  162,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  172,
  173,
  174,
  175,
  176,
  177,
  178,
  179,
  180,
  181,
  182,
  183,
  186,
  187,
  188,
  191,
  192,
  197,
  199,
  200,
  203,
  204,
  205,
  208,
  209,
  210,
  221,
  222,
  223,
  233,
  234,
  235,
  239,
  240,
  0,32,1,
  157,
  0xFE,
  0,86,1,
  157,
  0xFE,
  1,49,2,
  169,
  175,
  1,104,2,
  191,
  197,
  0xFE,
  0xFE,
  1,47,0,
  1,100,2,
  233,
  239,
  0xFE,
  1,102,2,
  221,
  227,
  0xFE,
  1,104,6,
  169,
  175,
  203,
  208,
  233,
  239,
  0xFE,
  0xFE,
  1,45,8,
  191,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  1,96,1,
  191,
  0xFE,
  0xFE,
  1,44,19,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  209,
  0xFE,
  1,46,24,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,48,22,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  203,
  208,
  209,
  0xFE,
  1,50,11,
  127,
  133,
  139,
  157,
  163,
  169,
  175,
  191,
  197,
  203,
  208,
  0xFE,
  1,52,9,
  133,
  157,
  163,
  169,
  175,
  191,
  197,
  203,
  208,
  1,99,4,
  133,
  157,
  163,
  169,
  0xFE,
  0xFE,
  1,54,1,
  175,
  1,103,2,
  157,
  163,
  0xFE,
  0xFE,
  1,56,0,
  1,107,2,
  157,
  163,
  0xFE,
  0xFE,
  1,93,2,
  221,
  241,
  0xFE,
  1,92,10,
  157,
  163,
  169,
  175,
  221,
  222,
  227,
  228,
  229,
  233,
  0xFE,
  1,96,5,
  163,
  169,
  175,
  233,
  239,
  0xFE,
  0xFE,
  1,3,83,
  3,
  4,
  7,
  13,
  65,
  66,
  98,
  99,
  104,
  110,
  127,
  128,
  129,
  130,
  131,
  133,
  134,
  135,
  136,
  137,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  148,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  160,
  161,
  162,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  172,
  173,
  175,
  176,
  177,
  178,
  179,
  181,
  182,
  183,
  184,
  186,
  187,
  188,
  189,
  191,
  192,
  193,
  194,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  221,
  222,
  223,
  224,
  227,
  233,
  234,
  239,
  240,
  0,32,1,
  157,
  0xFE,
  0,86,1,
  157,
  0xFE,
  1,49,12,
  133,
  134,
  157,
  158,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  177,
  0xFE,
  1,47,6,
  169,
  175,
  203,
  204,
  208,
  209,
  0xFE,
  1,45,13,
  169,
  175,
  191,
  192,
  193,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,44,26,
  127,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  0xFE,
  1,46,26,
  127,
  128,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  208,
  209,
  1,99,2,
  203,
  208,
  0xFE,
  1,92,2,
  221,
  227,
  0xFE,
  0xFE,
  1,48,38,
  3,
  65,
  98,
  104,
  110,
  127,
  128,
  133,
  134,
  139,
  144,
  145,
  157,
  158,
  163,
  164,
  169,
  170,
  171,
  175,
  176,
  177,
  191,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  227,
  233,
  234,
  235,
  239,
  240,
  241,
  1,99,3,
  191,
  203,
  208,
  0xFE,
  1,93,6,
  221,
  222,
  223,
  227,
  228,
  229,
  0xFE,
  0xFE,
  1,50,9,
  127,
  133,
  144,
  157,
  163,
  169,
  175,
  203,
  208,
  1,103,3,
  191,
  203,
  208,
  0xFE,
  1,99,4,
  191,
  197,
  233,
  239,
  0xFE,
  1,97,2,
  221,
  227,
  0xFE,
  0xFE,
  1,52,3,
  133,
  157,
  163,
  1,103,4,
  191,
  197,
  233,
  239,
  0xFE,
  1,101,2,
  221,
  227,
  0xFE,
  1,99,6,
  169,
  175,
  203,
  208,
  233,
  239,
  0xFE,
  0xFE,
  1,54,1,
  163,
  1,107,4,
  191,
  197,
  233,
  239,
  0xFE,
  1,105,2,
  221,
  227,
  0xFE,
  1,103,6,
  169,
  175,
  203,
  208,
  233,
  239,
  0xFE,
  0xFE,
  1,56,0,
  1,111,2,
  233,
  239,
  0xFE,
  1,109,2,
  221,
  227,
  0xFE,
  1,107,6,
  169,
  175,
  203,
  208,
  233,
  239,
  0xFE,
  0xFE,
  1,93,12,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  1,92,17,
  133,
  139,
  144,
  157,
  163,
  169,
  175,
  221,
  222,
  227,
  228,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  1,96,3,
  163,
  169,
  175,
  0xFE,
  0xFE,
  1,1,118,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  20,
  25,
  26,
  34,
  35,
  38,
  39,
  44,
  45,
  50,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  117,
  122,
  123,
  127,
  128,
  129,
  130,
  131,
  133,
  134,
  135,
  136,
  137,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  172,
  173,
  175,
  176,
  177,
  178,
  179,
  181,
  182,
  183,
  186,
  187,
  188,
  191,
  192,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  207,
  208,
  209,
  210,
  211,
  212,
  213,
  214,
  221,
  222,
  227,
  228,
  229,
  230,
  233,
  234,
  235,
  236,
  237,
  238,
  239,
  240,
  241,
  242,
  243,
  244,
  0,0,3,
  3,
  7,
  13,
  0xFE,
  0,32,3,
  38,
  44,
  157,
  0xFE,
  0,86,3,
  71,
  77,
  157,
  0,143,1,
  98,
  0xFE,
  0xFE,
  0,88,0,
  0,147,1,
  98,
  0xFE,
  0xFE,
  0,140,7,
  95,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  1,57,5,
  133,
  163,
  164,
  197,
  198,
  0xFE,
  1,53,2,
  203,
  208,
  0xFE,
  1,51,9,
  191,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  1,98,3,
  191,
  203,
  208,
  0xFE,
  0xFE,
  1,49,29,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  1,98,4,
  191,
  192,
  197,
  198,
  0xFE,
  0xFE,
  1,47,28,
  127,
  133,
  134,
  139,
  140,
  144,
  145,
  169,
  175,
  191,
  192,
  193,
  194,
  195,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  207,
  208,
  209,
  210,
  211,
  212,
  1,92,3,
  191,
  233,
  239,
  0xFE,
  1,96,4,
  221,
  222,
  227,
  228,
  0xFE,
  0xFE,
  1,45,42,
  133,
  134,
  163,
  169,
  175,
  191,
  192,
  193,
  194,
  195,
  196,
  197,
  198,
  199,
  200,
  201,
  202,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  214,
  215,
  217,
  218,
  219,
  221,
  222,
  223,
  227,
  228,
  233,
  234,
  235,
  239,
  240,
  241,
  1,93,4,
  233,
  234,
  239,
  240,
  0xFE,
  1,92,9,
  221,
  222,
  223,
  224,
  227,
  228,
  229,
  233,
  239,
  0xFE,
  0xFE,
  1,44,27,
  127,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  215,
  217,
  218,
  219,
  1,101,2,
  233,
  239,
  0xFE,
  1,99,11,
  191,
  192,
  197,
  198,
  221,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  1,97,12,
  221,
  222,
  223,
  224,
  227,
  228,
  229,
  230,
  233,
  234,
  239,
  240,
  0xFE,
  0xFE,
  1,46,16,
  127,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  197,
  203,
  208,
  1,103,2,
  233,
  239,
  0xFE,
  1,101,6,
  221,
  222,
  227,
  228,
  233,
  239,
  0xFE,
  1,99,4,
  197,
  198,
  227,
  228,
  0xFE,
  1,97,1,
  221,
  0xFE,
  0xFE,
  1,48,18,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  203,
  208,
  0xFE,
  1,50,3,
  133,
  157,
  163,
  0xFE,
  1,99,4,
  203,
  204,
  208,
  209,
  0xFE,
  1,97,7,
  221,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  1,95,35,
  139,
  140,
  144,
  145,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  208,
  209,
  210,
  221,
  222,
  223,
  224,
  227,
  228,
  229,
  233,
  234,
  235,
  236,
  237,
  239,
  240,
  241,
  242,
  243,
  1,47,4,
  203,
  204,
  208,
  209,
  0xFE,
  1,45,11,
  191,
  192,
  193,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,44,8,
  191,
  197,
  198,
  199,
  221,
  227,
  228,
  229,
  0xFE,
  1,46,4,
  203,
  204,
  208,
  209,
  0xFE,
  1,48,1,
  163,
  0xFE,
  1,50,5,
  133,
  157,
  163,
  169,
  175,
  0xFE,
  0xFE,
  1,93,24,
  221,
  222,
  223,
  224,
  225,
  226,
  227,
  228,
  229,
  230,
  231,
  232,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  246,
  247,
  249,
  250,
  251,
  0xFE,
  1,92,23,
  133,
  139,
  144,
  157,
  163,
  169,
  175,
  221,
  222,
  227,
  228,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  246,
  247,
  249,
  250,
  251,
  0xFE,
  1,94,10,
  127,
  133,
  157,
  163,
  169,
  175,
  233,
  234,
  239,
  240,
  1,59,1,
  197,
  0xFE,
  1,51,6,
  191,
  192,
  203,
  204,
  208,
  209,
  0xFE,
  1,49,11,
  191,
  192,
  193,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,47,15,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  221,
  222,
  227,
  233,
  239,
  0xFE,
  1,45,9,
  133,
  157,
  163,
  169,
  175,
  203,
  208,
  233,
  239,
  0xFE,
  1,44,2,
  157,
  163,
  0xFE,
  1,46,3,
  133,
  157,
  163,
  0xFE,
  0xFE,
  0xFE,
  0,255,81,
  3,
  4,
  7,
  8,
  13,
  19,
  25,
  34,
  35,
  65,
  66,
  71,
  77,
  98,
  99,
  104,
  105,
  110,
  111,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  149,
  150,
  151,
  153,
  154,
  155,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  183,
  186,
  187,
  188,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  221,
  227,
  233,
  234,
  235,
  239,
  240,
  241,
  0,32,1,
  157,
  0xFE,
  0,86,1,
  157,
  0xFE,
  0,140,2,
  104,
  110,
  0xFE,
  1,51,4,
  203,
  204,
  208,
  209,
  0xFE,
  1,49,24,
  127,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,47,21,
  133,
  139,
  144,
  157,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,45,17,
  133,
  139,
  144,
  157,
  163,
  169,
  175,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  1,92,2,
  221,
  227,
  0xFE,
  0xFE,
  1,44,7,
  127,
  133,
  157,
  163,
  175,
  191,
  197,
  1,97,2,
  221,
  227,
  0xFE,
  0xFE,
  1,46,13,
  127,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  175,
  203,
  208,
  0xFE,
  1,48,13,
  127,
  128,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  175,
  176,
  1,93,2,
  221,
  227,
  0xFE,
  0xFE,
  1,97,2,
  233,
  239,
  0xFE,
  1,93,12,
  169,
  175,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  239,
  240,
  0xFE,
  1,92,6,
  127,
  133,
  157,
  163,
  221,
  227,
  0xFE,
  0xFE,
  0,253,79,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  34,
  38,
  44,
  65,
  71,
  72,
  77,
  78,
  98,
  99,
  104,
  105,
  110,
  111,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  149,
  150,
  151,
  153,
  154,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  183,
  186,
  187,
  188,
  191,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  221,
  227,
  228,
  233,
  234,
  235,
  239,
  240,
  241,
  0,140,2,
  104,
  110,
  0xFE,
  1,53,1,
  191,
  0xFE,
  1,51,2,
  191,
  192,
  0xFE,
  1,49,23,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,47,22,
  127,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,45,22,
  133,
  139,
  144,
  157,
  163,
  169,
  175,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  233,
  239,
  1,92,2,
  221,
  227,
  0xFE,
  0xFE,
  1,44,9,
  163,
  169,
  175,
  191,
  197,
  203,
  208,
  213,
  217,
  1,99,2,
  233,
  239,
  0xFE,
  1,97,3,
  221,
  222,
  227,
  0xFE,
  0xFE,
  1,46,4,
  157,
  163,
  169,
  175,
  0xFE,
  1,48,8,
  127,
  133,
  139,
  144,
  157,
  163,
  169,
  175,
  0xFE,
  1,97,2,
  233,
  239,
  0xFE,
  1,93,11,
  163,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  239,
  245,
  249,
  0xFE,
  1,92,6,
  221,
  227,
  233,
  239,
  245,
  249,
  0xFE,
  0xFE,
  0,251,82,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  34,
  38,
  44,
  65,
  66,
  71,
  77,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  150,
  151,
  153,
  154,
  155,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  183,
  186,
  187,
  188,
  191,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  221,
  227,
  233,
  234,
  235,
  239,
  240,
  241,
  0,86,2,
  71,
  77,
  0xFE,
  0,151,0,
  0,90,2,
  65,
  98,
  0xFE,
  0,92,2,
  95,
  98,
  0xFE,
  0xFE,
  0,140,2,
  104,
  110,
  0xFE,
  1,53,2,
  191,
  192,
  0xFE,
  1,51,4,
  191,
  192,
  203,
  208,
  0xFE,
  1,49,18,
  127,
  128,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,47,15,
  127,
  133,
  157,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,45,17,
  127,
  133,
  157,
  163,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  208,
  213,
  217,
  221,
  233,
  239,
  1,92,2,
  221,
  227,
  0xFE,
  0xFE,
  1,44,6,
  191,
  197,
  203,
  208,
  213,
  217,
  1,99,2,
  233,
  239,
  0xFE,
  1,97,2,
  221,
  227,
  0xFE,
  0xFE,
  1,46,2,
  157,
  163,
  0xFE,
  1,48,3,
  157,
  163,
  169,
  0xFE,
  1,97,2,
  233,
  239,
  0xFE,
  1,93,10,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  239,
  245,
  249,
  0xFE,
  1,92,6,
  221,
  227,
  233,
  239,
  245,
  249,
  0xFE,
  0xFE,
  0,249,73,
  3,
  7,
  13,
  19,
  25,
  71,
  77,
  98,
  104,
  105,
  110,
  111,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  149,
  150,
  151,
  153,
  154,
  155,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  173,
  175,
  176,
  177,
  178,
  179,
  181,
  182,
  183,
  186,
  187,
  188,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  221,
  222,
  227,
  233,
  234,
  239,
  240,
  0,90,0,
  0,151,1,
  98,
  0xFE,
  0xFE,
  1,55,1,
  191,
  0xFE,
  1,53,1,
  191,
  0xFE,
  1,51,3,
  163,
  203,
  208,
  0xFE,
  1,49,14,
  127,
  133,
  139,
  144,
  157,
  163,
  169,
  170,
  175,
  176,
  191,
  197,
  203,
  208,
  0xFE,
  1,47,10,
  127,
  133,
  157,
  163,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,45,7,
  133,
  191,
  192,
  197,
  198,
  203,
  208,
  0xFE,
  1,44,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,46,2,
  191,
  197,
  0xFE,
  1,93,6,
  221,
  222,
  227,
  228,
  233,
  239,
  0xFE,
  1,92,2,
  233,
  239,
  0xFE,
  0xFE,
  0,248,42,
  7,
  13,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  149,
  153,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  186,
  203,
  208,
  233,
  239,
  1,59,3,
  191,
  203,
  208,
  0xFE,
  1,57,6,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,55,6,
  191,
  192,
  197,
  198,
  203,
  208,
  0xFE,
  1,53,4,
  191,
  197,
  203,
  208,
  1,98,2,
  157,
  163,
  0xFE,
  0xFE,
  1,51,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,49,15,
  127,
  128,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  191,
  197,
  0xFE,
  1,47,4,
  191,
  192,
  197,
  203,
  0xFE,
  1,45,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,97,1,
  221,
  0xFE,
  1,93,6,
  221,
  222,
  227,
  228,
  233,
  239,
  0xFE,
  0xFE,
  0,250,35,
  7,
  13,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  153,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  181,
  182,
  186,
  187,
  1,53,2,
  203,
  208,
  0xFE,
  1,51,2,
  191,
  197,
  0xFE,
  0xFE,
  0,252,24,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  0,254,13,
  133,
  139,
  140,
  144,
  145,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,0,2,
  7,
  13,
  1,57,2,
  203,
  208,
  0xFE,
  1,55,6,
  191,
  197,
  198,
  203,
  204,
  208,
  0xFE,
  1,53,5,
  191,
  192,
  197,
  198,
  208,
  0xFE,
  1,51,2,
  191,
  197,
  0xFE,
  0xFE,
  1,61,2,
  203,
  208,
  0xFE,
  1,59,6,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,57,11,
  133,
  139,
  144,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  1,96,2,
  203,
  208,
  0xFE,
  0xFE,
  1,55,12,
  127,
  133,
  139,
  144,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  1,96,2,
  191,
  197,
  0xFE,
  0xFE,
  1,53,16,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  0xFE,
  1,51,20,
  127,
  128,
  133,
  134,
  139,
  144,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  0xFE,
  1,49,32,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  233,
  239,
  0,250,11,
  127,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  0xFE,
  0,252,8,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  0,254,5,
  133,
  157,
  158,
  163,
  164,
  0xFE,
  1,99,2,
  203,
  208,
  0xFE,
  1,93,2,
  221,
  227,
  0xFE,
  1,98,2,
  191,
  197,
  0xFE,
  0xFE,
  1,47,25,
  127,
  128,
  133,
  134,
  139,
  144,
  169,
  175,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  217,
  233,
  239,
  1,99,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,93,2,
  221,
  227,
  0xFE,
  1,92,3,
  191,
  233,
  239,
  0xFE,
  1,96,2,
  221,
  227,
  0xFE,
  0xFE,
  1,45,29,
  127,
  133,
  139,
  144,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  217,
  218,
  221,
  222,
  227,
  233,
  234,
  239,
  240,
  1,99,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,97,2,
  233,
  239,
  0xFE,
  1,93,6,
  127,
  133,
  221,
  227,
  233,
  239,
  0xFE,
  1,92,6,
  221,
  222,
  227,
  228,
  233,
  239,
  0xFE,
  0xFE,
  1,44,10,
  191,
  197,
  203,
  204,
  208,
  209,
  213,
  214,
  217,
  218,
  1,103,2,
  233,
  239,
  0xFE,
  1,101,6,
  127,
  133,
  221,
  227,
  233,
  239,
  0xFE,
  1,99,8,
  191,
  197,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,97,6,
  221,
  222,
  227,
  228,
  233,
  239,
  0xFE,
  0xFE,
  1,46,2,
  203,
  208,
  1,103,2,
  233,
  239,
  0xFE,
  1,101,4,
  221,
  227,
  233,
  239,
  0xFE,
  1,99,2,
  197,
  227,
  0xFE,
  1,97,1,
  221,
  0xFE,
  0xFE,
  1,54,0,
  1,19,1,
  163,
  0xFE,
  1,17,4,
  169,
  175,
  191,
  197,
  0xFE,
  0xFE,
  1,101,4,
  233,
  234,
  239,
  240,
  0xFE,
  1,99,13,
  127,
  133,
  139,
  144,
  197,
  203,
  208,
  221,
  222,
  227,
  228,
  233,
  239,
  0xFE,
  1,97,12,
  127,
  133,
  144,
  221,
  222,
  227,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,95,29,
  127,
  128,
  133,
  134,
  139,
  144,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  221,
  222,
  223,
  227,
  228,
  233,
  234,
  235,
  236,
  239,
  240,
  241,
  242,
  245,
  249,
  1,17,0,
  1,72,1,
  191,
  0xFE,
  0xFE,
  1,15,0,
  1,68,2,
  203,
  208,
  0xFE,
  1,70,4,
  133,
  163,
  191,
  197,
  0xFE,
  1,72,2,
  203,
  208,
  0xFE,
  0xFE,
  1,13,0,
  1,52,1,
  191,
  0xFE,
  1,68,1,
  191,
  0xFE,
  0xFE,
  1,11,0,
  1,58,1,
  197,
  0xFE,
  0xFE,
  1,9,0,
  1,52,1,
  197,
  0xFE,
  0xFE,
  1,7,0,
  1,48,1,
  163,
  0xFE,
  0xFE,
  1,5,2,
  233,
  239,
  1,48,1,
  169,
  0xFE,
  0xFE,
  1,3,4,
  233,
  234,
  239,
  240,
  1,44,2,
  203,
  208,
  0xFE,
  1,46,5,
  191,
  203,
  204,
  208,
  209,
  0xFE,
  1,48,16,
  163,
  169,
  175,
  191,
  192,
  193,
  197,
  198,
  199,
  221,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  0xFE,
  0,255,13,
  191,
  197,
  221,
  222,
  223,
  227,
  228,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  0,253,13,
  191,
  192,
  197,
  198,
  221,
  222,
  227,
  233,
  234,
  235,
  239,
  240,
  241,
  1,45,2,
  203,
  208,
  0xFE,
  1,44,1,
  197,
  0xFE,
  0xFE,
  0,251,11,
  191,
  197,
  203,
  208,
  221,
  222,
  227,
  233,
  234,
  239,
  240,
  1,47,2,
  203,
  208,
  0xFE,
  1,45,3,
  191,
  203,
  208,
  0xFE,
  1,44,2,
  197,
  227,
  0xFE,
  0xFE,
  0,249,5,
  221,
  233,
  234,
  239,
  240,
  0xFE,
  0,248,5,
  221,
  222,
  227,
  233,
  239,
  0xFE,
  0,250,1,
  221,
  0xFE,
  1,55,2,
  203,
  208,
  0xFE,
  1,53,6,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,51,11,
  133,
  139,
  144,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,49,12,
  127,
  133,
  139,
  144,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,47,12,
  127,
  133,
  144,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  233,
  239,
  0xFE,
  1,45,14,
  127,
  133,
  139,
  144,
  191,
  197,
  203,
  204,
  208,
  209,
  221,
  227,
  233,
  239,
  0xFE,
  1,44,6,
  191,
  197,
  198,
  221,
  227,
  228,
  0xFE,
  1,46,2,
  203,
  208,
  0xFE,
  1,50,0,
  1,13,2,
  233,
  239,
  0xFE,
  1,11,4,
  157,
  158,
  163,
  164,
  0xFE,
  1,9,11,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,7,14,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,5,6,
  133,
  157,
  158,
  163,
  164,
  175,
  0xFE,
  1,3,8,
  157,
  163,
  169,
  175,
  203,
  208,
  233,
  239,
  0xFE,
  0,255,3,
  133,
  157,
  163,
  0xFE,
  0xFE,
  0xFE,
  1,93,18,
  221,
  222,
  223,
  224,
  227,
  228,
  229,
  230,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  246,
  249,
  250,
  0xFE,
  1,92,10,
  221,
  227,
  233,
  234,
  239,
  240,
  245,
  246,
  249,
  250,
  1,53,2,
  203,
  208,
  0xFE,
  1,51,2,
  191,
  197,
  0xFE,
  0xFE,
  1,94,0,
  1,13,0,
  1,48,6,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  0xFE,
  1,11,10,
  221,
  222,
  233,
  234,
  235,
  236,
  239,
  240,
  241,
  242,
  1,54,1,
  197,
  0xFE,
  0xFE,
  1,9,4,
  221,
  227,
  233,
  239,
  1,48,1,
  197,
  0xFE,
  1,50,1,
  197,
  0xFE,
  0xFE,
  1,7,16,
  133,
  157,
  163,
  164,
  165,
  169,
  175,
  221,
  227,
  228,
  233,
  234,
  235,
  239,
  240,
  241,
  1,44,1,
  163,
  0xFE,
  0xFE,
  1,5,12,
  133,
  139,
  144,
  157,
  163,
  164,
  169,
  170,
  175,
  176,
  221,
  227,
  1,45,6,
  157,
  163,
  169,
  170,
  175,
  176,
  0xFE,
  1,44,2,
  169,
  175,
  0xFE,
  0xFE,
  1,3,12,
  127,
  133,
  139,
  144,
  157,
  163,
  164,
  169,
  175,
  191,
  227,
  233,
  1,45,5,
  157,
  158,
  163,
  169,
  175,
  0xFE,
  1,44,4,
  163,
  169,
  175,
  191,
  0xFE,
  0xFE,
  0,255,5,
  133,
  157,
  163,
  233,
  239,
  1,45,3,
  133,
  157,
  163,
  0xFE,
  0xFE,
  1,59,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,57,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,55,5,
  133,
  191,
  197,
  203,
  208,
  0xFE,
  1,53,9,
  133,
  139,
  144,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,51,13,
  127,
  133,
  139,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  208,
  0,255,2,
  169,
  175,
  0xFE,
  0,253,3,
  163,
  169,
  175,
  0xFE,
  0,251,3,
  133,
  157,
  163,
  0xFE,
  0,249,2,
  169,
  175,
  0xFE,
  0,248,2,
  157,
  163,
  0xFE,
  0xFE,
  1,49,12,
  127,
  133,
  139,
  144,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0,253,3,
  191,
  203,
  208,
  0xFE,
  0,251,3,
  191,
  203,
  208,
  0xFE,
  0xFE,
  1,47,11,
  191,
  192,
  193,
  197,
  198,
  203,
  208,
  221,
  227,
  233,
  239,
  0,255,2,
  191,
  197,
  0xFE,
  0,253,3,
  191,
  192,
  197,
  0xFE,
  0,251,2,
  191,
  197,
  0xFE,
  0xFE,
  1,46,0,
  1,9,1,
  169,
  0xFE,
  1,7,6,
  133,
  163,
  169,
  170,
  175,
  176,
  0xFE,
  1,5,5,
  157,
  158,
  163,
  169,
  175,
  0xFE,
  1,3,4,
  157,
  163,
  169,
  175,
  0xFE,
  0,255,1,
  163,
  0xFE,
  0xFE,
  0xFE,
  1,98,0,
  1,59,1,
  203,
  0xFE,
  1,57,2,
  191,
  197,
  0xFE,
  1,55,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,53,5,
  163,
  169,
  175,
  191,
  197,
  0xFE,
  1,51,3,
  191,
  203,
  208,
  0xFE,
  0xFE,
  0xFE,
  0,210,90,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  20,
  25,
  26,
  34,
  38,
  39,
  44,
  45,
  50,
  56,
  66,
  71,
  72,
  77,
  78,
  83,
  89,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  153,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  177,
  178,
  181,
  186,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  227,
  228,
  233,
  234,
  235,
  236,
  239,
  240,
  241,
  242,
  0,0,5,
  3,
  7,
  13,
  19,
  25,
  0xFE,
  0,32,0,
  0,89,1,
  65,
  0xFE,
  0xFE,
  0,34,0,
  0,153,1,
  98,
  0xFE,
  0xFE,
  0,36,0,
  0,95,2,
  34,
  65,
  0xFE,
  0xFE,
  0,38,0,
  0,99,2,
  34,
  65,
  0xFE,
  0xFE,
  0,42,0,
  0,105,2,
  34,
  65,
  0xFE,
  0xFE,
  0,86,0,
  0,151,1,
  98,
  0xFE,
  0,145,2,
  104,
  110,
  0xFE,
  0xFE,
  0,90,0,
  0,151,1,
  98,
  0xFE,
  0xFE,
  0,92,0,
  0,159,1,
  98,
  0xFE,
  0,157,2,
  95,
  98,
  0xFE,
  0,155,1,
  98,
  0xFE,
  0xFE,
  0,94,0,
  0,159,5,
  65,
  95,
  98,
  99,
  104,
  0xFE,
  0xFE,
  0,96,0,
  0,163,3,
  65,
  98,
  99,
  0xFE,
  0xFE,
  0,98,0,
  0,167,1,
  98,
  0xFE,
  0xFE,
  0,100,0,
  0,171,0,
  1,0,2,
  65,
  98,
  0xFE,
  0xFE,
  0xFE,
  0,140,2,
  104,
  110,
  0xFE,
  1,27,0,
  1,58,2,
  197,
  198,
  0xFE,
  1,60,3,
  191,
  197,
  198,
  0xFE,
  0xFE,
  1,25,0,
  1,56,4,
  191,
  192,
  197,
  198,
  0xFE,
  1,58,8,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,60,4,
  203,
  204,
  208,
  209,
  0xFE,
  0xFE,
  1,23,0,
  1,54,8,
  157,
  163,
  191,
  192,
  193,
  197,
  198,
  199,
  0xFE,
  1,56,11,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,58,4,
  203,
  204,
  208,
  209,
  0xFE,
  0xFE,
  1,21,0,
  1,50,2,
  197,
  198,
  0xFE,
  1,52,12,
  157,
  158,
  163,
  164,
  191,
  192,
  193,
  197,
  198,
  199,
  208,
  209,
  0xFE,
  1,54,16,
  133,
  157,
  163,
  164,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,56,6,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  0xFE,
  1,19,0,
  1,48,2,
  197,
  198,
  0xFE,
  1,50,14,
  157,
  158,
  163,
  164,
  165,
  191,
  192,
  193,
  197,
  198,
  199,
  200,
  208,
  209,
  0xFE,
  1,52,16,
  157,
  158,
  163,
  164,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,54,11,
  163,
  169,
  175,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  0xFE,
  1,17,8,
  203,
  204,
  208,
  209,
  233,
  234,
  239,
  240,
  1,46,4,
  197,
  198,
  203,
  204,
  0xFE,
  1,48,31,
  139,
  144,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  208,
  209,
  210,
  233,
  234,
  239,
  240,
  1,92,6,
  221,
  222,
  223,
  227,
  228,
  229,
  0xFE,
  0xFE,
  1,50,25,
  157,
  158,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  0xFE,
  1,52,13,
  163,
  169,
  170,
  175,
  176,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,54,4,
  175,
  197,
  203,
  208,
  0xFE,
  1,58,2,
  163,
  197,
  0xFE,
  1,60,4,
  157,
  163,
  191,
  197,
  0xFE,
  1,92,6,
  221,
  222,
  223,
  227,
  228,
  229,
  0xFE,
  1,96,4,
  221,
  222,
  227,
  228,
  0xFE,
  1,98,3,
  233,
  234,
  239,
  0xFE,
  0xFE,
  1,15,34,
  7,
  13,
  104,
  110,
  139,
  140,
  141,
  144,
  145,
  146,
  157,
  158,
  169,
  170,
  171,
  175,
  176,
  177,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  233,
  234,
  235,
  236,
  239,
  240,
  241,
  242,
  0,86,2,
  169,
  175,
  0xFE,
  0,140,2,
  104,
  110,
  0xFE,
  0,142,1,
  98,
  0xFE,
  1,44,10,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  208,
  209,
  0xFE,
  1,46,20,
  157,
  158,
  159,
  163,
  164,
  165,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,48,37,
  133,
  139,
  144,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  233,
  234,
  235,
  239,
  240,
  241,
  1,92,8,
  221,
  222,
  223,
  224,
  227,
  228,
  229,
  230,
  0xFE,
  0xFE,
  1,50,21,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  0xFE,
  1,52,13,
  163,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,54,14,
  133,
  134,
  157,
  158,
  163,
  164,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,56,2,
  163,
  191,
  0xFE,
  1,70,0,
  1,97,2,
  221,
  227,
  0xFE,
  0xFE,
  1,92,8,
  221,
  222,
  223,
  224,
  227,
  228,
  229,
  230,
  0xFE,
  1,96,7,
  221,
  227,
  228,
  233,
  234,
  239,
  240,
  0xFE,
  1,98,2,
  233,
  239,
  0xFE,
  0xFE,
  1,13,68,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  173,
  175,
  176,
  177,
  178,
  179,
  181,
  182,
  186,
  187,
  197,
  198,
  199,
  203,
  204,
  205,
  206,
  207,
  208,
  209,
  210,
  211,
  212,
  227,
  228,
  233,
  234,
  235,
  236,
  237,
  239,
  240,
  241,
  242,
  243,
  0,32,2,
  157,
  163,
  0xFE,
  0,86,2,
  157,
  163,
  0xFE,
  1,47,0,
  1,104,7,
  133,
  139,
  144,
  157,
  163,
  169,
  175,
  0xFE,
  0xFE,
  1,44,17,
  163,
  164,
  191,
  192,
  193,
  194,
  195,
  197,
  198,
  199,
  200,
  201,
  203,
  204,
  205,
  208,
  209,
  1,98,3,
  133,
  163,
  227,
  0xFE,
  0xFE,
  1,46,33,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  201,
  203,
  204,
  205,
  206,
  207,
  208,
  209,
  210,
  211,
  212,
  1,92,2,
  221,
  227,
  0xFE,
  0xFE,
  1,48,44,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  186,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  1,93,1,
  203,
  0xFE,
  1,92,2,
  203,
  208,
  0xFE,
  0xFE,
  1,50,14,
  133,
  157,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  204,
  209,
  1,93,4,
  197,
  221,
  227,
  228,
  0xFE,
  0xFE,
  1,52,2,
  203,
  208,
  1,99,9,
  127,
  133,
  157,
  158,
  163,
  164,
  169,
  175,
  227,
  0xFE,
  1,97,4,
  197,
  198,
  227,
  228,
  0xFE,
  0xFE,
  1,54,0,
  1,103,4,
  157,
  163,
  169,
  175,
  0xFE,
  1,97,1,
  197,
  0xFE,
  0xFE,
  1,92,10,
  221,
  222,
  223,
  224,
  225,
  227,
  228,
  229,
  230,
  231,
  0xFE,
  1,96,8,
  163,
  169,
  175,
  197,
  233,
  234,
  239,
  240,
  0xFE,
  1,102,7,
  133,
  157,
  163,
  164,
  221,
  227,
  228,
  0xFE,
  0xFE,
  1,11,93,
  3,
  4,
  7,
  13,
  19,
  25,
  34,
  35,
  38,
  44,
  65,
  66,
  71,
  77,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  127,
  128,
  129,
  130,
  131,
  132,
  133,
  134,
  135,
  136,
  137,
  138,
  139,
  140,
  141,
  142,
  143,
  144,
  145,
  146,
  147,
  148,
  149,
  150,
  151,
  153,
  154,
  155,
  157,
  158,
  159,
  160,
  161,
  162,
  163,
  164,
  165,
  166,
  167,
  168,
  169,
  170,
  171,
  172,
  173,
  174,
  175,
  176,
  177,
  178,
  179,
  180,
  181,
  182,
  183,
  186,
  187,
  188,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  227,
  228,
  233,
  239,
  1,47,0,
  1,104,8,
  133,
  139,
  144,
  157,
  163,
  164,
  169,
  175,
  0xFE,
  0xFE,
  1,46,25,
  133,
  134,
  139,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,48,40,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  157,
  158,
  159,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  191,
  197,
  203,
  204,
  205,
  208,
  209,
  210,
  233,
  234,
  235,
  239,
  240,
  241,
  1,93,7,
  221,
  222,
  223,
  224,
  227,
  228,
  229,
  0xFE,
  0xFE,
  1,50,2,
  169,
  175,
  0xFE,
  1,52,1,
  163,
  1,99,7,
  133,
  139,
  144,
  157,
  163,
  169,
  175,
  0xFE,
  0xFE,
  1,54,0,
  1,103,3,
  163,
  169,
  175,
  0xFE,
  0xFE,
  1,92,5,
  221,
  222,
  223,
  227,
  228,
  0xFE,
  1,96,10,
  133,
  139,
  144,
  157,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  0xFE,
  1,9,72,
  3,
  98,
  127,
  128,
  129,
  130,
  131,
  133,
  134,
  135,
  136,
  137,
  139,
  140,
  141,
  142,
  143,
  144,
  145,
  146,
  147,
  148,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  160,
  161,
  162,
  163,
  164,
  165,
  166,
  167,
  168,
  169,
  170,
  171,
  172,
  173,
  175,
  176,
  177,
  178,
  179,
  181,
  182,
  186,
  187,
  191,
  197,
  198,
  199,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  233,
  234,
  235,
  236,
  239,
  240,
  241,
  242,
  1,49,18,
  133,
  134,
  157,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,47,20,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  208,
  209,
  210,
  1,104,3,
  133,
  157,
  163,
  0xFE,
  0xFE,
  1,45,15,
  163,
  191,
  192,
  193,
  194,
  195,
  197,
  198,
  199,
  200,
  203,
  204,
  208,
  209,
  210,
  0xFE,
  1,44,12,
  157,
  163,
  164,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  0xFE,
  1,46,13,
  133,
  139,
  144,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  0xFE,
  1,48,31,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  203,
  204,
  208,
  209,
  233,
  234,
  239,
  240,
  1,93,4,
  221,
  222,
  227,
  228,
  0xFE,
  0xFE,
  1,50,4,
  157,
  163,
  169,
  175,
  0xFE,
  1,52,0,
  1,99,7,
  133,
  139,
  144,
  157,
  163,
  169,
  175,
  0xFE,
  0xFE,
  1,54,0,
  1,103,4,
  157,
  163,
  169,
  175,
  0xFE,
  0xFE,
  1,93,9,
  221,
  222,
  223,
  224,
  225,
  227,
  228,
  229,
  230,
  0xFE,
  1,92,3,
  221,
  227,
  228,
  0xFE,
  1,96,3,
  157,
  158,
  163,
  0xFE,
  0xFE,
  1,7,73,
  3,
  34,
  65,
  98,
  127,
  128,
  129,
  130,
  131,
  133,
  134,
  135,
  136,
  137,
  139,
  140,
  141,
  142,
  143,
  144,
  145,
  146,
  147,
  148,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  160,
  161,
  162,
  163,
  164,
  165,
  166,
  167,
  168,
  169,
  170,
  171,
  172,
  173,
  175,
  176,
  177,
  178,
  179,
  181,
  182,
  183,
  186,
  187,
  188,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  233,
  234,
  235,
  239,
  240,
  241,
  1,51,4,
  203,
  204,
  208,
  209,
  1,104,3,
  191,
  203,
  208,
  0xFE,
  0xFE,
  1,49,15,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  1,104,2,
  191,
  197,
  0xFE,
  0xFE,
  1,47,13,
  163,
  164,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  1,100,2,
  233,
  239,
  0xFE,
  1,102,2,
  221,
  227,
  0xFE,
  1,104,6,
  169,
  175,
  203,
  208,
  233,
  239,
  0xFE,
  0xFE,
  1,45,10,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  1,96,2,
  191,
  197,
  0xFE,
  0xFE,
  1,44,5,
  139,
  144,
  169,
  175,
  197,
  0xFE,
  1,46,13,
  127,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,48,19,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  0xFE,
  1,50,5,
  133,
  157,
  163,
  169,
  175,
  0xFE,
  1,52,3,
  157,
  163,
  169,
  1,99,4,
  127,
  133,
  157,
  163,
  0xFE,
  0xFE,
  1,54,0,
  1,103,3,
  133,
  157,
  163,
  0xFE,
  0xFE,
  1,56,0,
  1,107,1,
  163,
  0xFE,
  0xFE,
  1,93,6,
  221,
  222,
  223,
  227,
  228,
  229,
  0xFE,
  1,92,3,
  169,
  175,
  227,
  0xFE,
  1,96,5,
  175,
  203,
  208,
  233,
  239,
  0xFE,
  0xFE,
  1,5,67,
  3,
  4,
  34,
  65,
  98,
  99,
  104,
  127,
  128,
  129,
  130,
  131,
  133,
  134,
  135,
  136,
  137,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  172,
  173,
  175,
  176,
  177,
  178,
  179,
  181,
  182,
  183,
  186,
  187,
  188,
  203,
  204,
  205,
  208,
  209,
  210,
  233,
  234,
  235,
  239,
  240,
  241,
  1,49,4,
  163,
  191,
  197,
  198,
  0xFE,
  1,47,6,
  191,
  192,
  203,
  204,
  208,
  209,
  0xFE,
  1,45,6,
  191,
  192,
  193,
  197,
  198,
  199,
  0xFE,
  1,46,11,
  127,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  175,
  0xFE,
  1,48,13,
  127,
  128,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  0xFE,
  1,50,4,
  157,
  163,
  169,
  175,
  0xFE,
  1,52,2,
  157,
  163,
  1,99,1,
  157,
  0xFE,
  0xFE,
  1,93,6,
  221,
  222,
  223,
  227,
  228,
  229,
  0xFE,
  1,92,5,
  163,
  169,
  175,
  221,
  227,
  0xFE,
  0xFE,
  1,3,67,
  3,
  4,
  7,
  34,
  65,
  98,
  110,
  127,
  128,
  129,
  130,
  131,
  133,
  134,
  135,
  136,
  137,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  183,
  186,
  187,
  188,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  221,
  222,
  233,
  234,
  239,
  240,
  0,32,1,
  157,
  0xFE,
  0,86,1,
  157,
  0xFE,
  1,53,7,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,51,5,
  191,
  192,
  197,
  198,
  199,
  0xFE,
  1,49,7,
  139,
  144,
  169,
  170,
  175,
  176,
  203,
  0xFE,
  1,47,5,
  169,
  175,
  197,
  203,
  208,
  0xFE,
  1,45,11,
  169,
  175,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  0xFE,
  1,44,17,
  127,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,46,13,
  127,
  133,
  139,
  144,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  208,
  0xFE,
  1,48,25,
  3,
  7,
  13,
  104,
  110,
  127,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  203,
  204,
  208,
  209,
  233,
  234,
  239,
  240,
  1,93,4,
  221,
  222,
  227,
  228,
  0xFE,
  0xFE,
  1,50,3,
  133,
  157,
  163,
  1,97,2,
  221,
  227,
  0xFE,
  0xFE,
  1,52,0,
  1,101,2,
  221,
  227,
  0xFE,
  0xFE,
  1,93,6,
  221,
  222,
  223,
  227,
  228,
  229,
  0xFE,
  1,92,7,
  133,
  157,
  163,
  169,
  175,
  221,
  227,
  0xFE,
  0xFE,
  1,1,107,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  20,
  25,
  26,
  34,
  35,
  38,
  39,
  44,
  45,
  50,
  51,
  56,
  57,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  84,
  89,
  90,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  117,
  122,
  123,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  183,
  186,
  187,
  188,
  191,
  197,
  198,
  199,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  221,
  227,
  228,
  229,
  233,
  234,
  235,
  236,
  239,
  240,
  241,
  242,
  0,0,2,
  7,
  13,
  0xFE,
  0,32,1,
  157,
  0xFE,
  0,36,0,
  0,95,2,
  34,
  65,
  0xFE,
  0xFE,
  0,86,1,
  157,
  0xFE,
  0,88,0,
  0,147,1,
  98,
  0xFE,
  0xFE,
  0,90,0,
  0,151,2,
  65,
  98,
  0xFE,
  0xFE,
  0,106,0,
  0,181,2,
  65,
  98,
  0xFE,
  0xFE,
  0,140,2,
  104,
  110,
  0xFE,
  1,55,8,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,53,8,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,51,12,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,49,30,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  0xFE,
  1,47,30,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  144,
  157,
  163,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  0xFE,
  1,45,26,
  157,
  163,
  169,
  175,
  191,
  192,
  193,
  194,
  195,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  221,
  233,
  234,
  239,
  240,
  1,92,4,
  221,
  222,
  227,
  228,
  0xFE,
  0xFE,
  1,44,13,
  127,
  133,
  157,
  163,
  175,
  191,
  197,
  203,
  204,
  208,
  209,
  213,
  217,
  1,99,4,
  233,
  234,
  239,
  240,
  0xFE,
  1,97,4,
  221,
  222,
  227,
  228,
  0xFE,
  0xFE,
  1,46,10,
  127,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  175,
  0xFE,
  1,48,12,
  127,
  128,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  175,
  0xFE,
  1,97,7,
  221,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  1,93,20,
  221,
  222,
  223,
  224,
  225,
  227,
  228,
  229,
  230,
  231,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  246,
  249,
  250,
  0xFE,
  1,92,11,
  133,
  157,
  163,
  221,
  227,
  233,
  239,
  245,
  246,
  249,
  250,
  0xFE,
  0xFE,
  0,255,67,
  3,
  7,
  8,
  13,
  14,
  34,
  38,
  44,
  65,
  71,
  77,
  98,
  104,
  110,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  183,
  186,
  187,
  188,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  227,
  233,
  234,
  235,
  239,
  240,
  241,
  0,90,0,
  0,151,1,
  98,
  0xFE,
  0xFE,
  1,57,5,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,55,4,
  191,
  192,
  197,
  198,
  0xFE,
  1,49,16,
  127,
  128,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  203,
  208,
  0xFE,
  1,47,17,
  127,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,45,13,
  133,
  157,
  163,
  169,
  175,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  208,
  1,92,2,
  221,
  227,
  0xFE,
  0xFE,
  1,46,2,
  169,
  175,
  0xFE,
  1,48,8,
  127,
  133,
  139,
  144,
  157,
  163,
  169,
  175,
  0xFE,
  1,93,7,
  221,
  222,
  223,
  227,
  228,
  229,
  239,
  0xFE,
  0xFE,
  0,253,69,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  34,
  38,
  44,
  65,
  71,
  77,
  98,
  104,
  110,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  150,
  151,
  153,
  154,
  155,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  183,
  186,
  187,
  188,
  197,
  203,
  204,
  208,
  209,
  227,
  233,
  234,
  239,
  240,
  0,92,0,
  0,155,1,
  98,
  0xFE,
  0xFE,
  1,57,3,
  191,
  197,
  198,
  0xFE,
  1,51,2,
  203,
  208,
  0xFE,
  1,49,16,
  127,
  128,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  197,
  203,
  208,
  0xFE,
  1,47,12,
  127,
  133,
  157,
  163,
  164,
  175,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,45,9,
  133,
  157,
  163,
  191,
  192,
  197,
  198,
  203,
  208,
  1,92,2,
  221,
  227,
  0xFE,
  0xFE,
  1,97,2,
  233,
  239,
  0xFE,
  1,93,5,
  221,
  222,
  227,
  228,
  239,
  0xFE,
  0xFE,
  0,251,80,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  34,
  38,
  44,
  50,
  56,
  65,
  71,
  77,
  83,
  89,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  183,
  186,
  187,
  188,
  197,
  203,
  204,
  208,
  209,
  227,
  233,
  234,
  235,
  239,
  240,
  241,
  0,92,0,
  0,155,2,
  65,
  98,
  0xFE,
  0xFE,
  1,51,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,49,13,
  127,
  133,
  144,
  157,
  163,
  169,
  170,
  175,
  176,
  191,
  197,
  203,
  208,
  0xFE,
  1,47,7,
  163,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,45,8,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  208,
  1,92,2,
  221,
  227,
  0xFE,
  0xFE,
  1,93,8,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  239,
  0xFE,
  0xFE,
  0,249,49,
  3,
  7,
  13,
  104,
  110,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  186,
  187,
  203,
  208,
  221,
  233,
  239,
  1,51,3,
  197,
  203,
  208,
  0xFE,
  1,49,16,
  127,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  197,
  203,
  208,
  0xFE,
  1,47,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,45,2,
  191,
  197,
  0xFE,
  1,93,2,
  221,
  227,
  0xFE,
  0xFE,
  0,248,29,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  153,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  181,
  186,
  1,59,2,
  203,
  208,
  0xFE,
  1,57,2,
  191,
  197,
  0xFE,
  1,53,1,
  191,
  0xFE,
  1,49,8,
  127,
  133,
  139,
  144,
  157,
  163,
  169,
  175,
  0xFE,
  0xFE,
  0,250,21,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  0xFE,
  0,252,11,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,61,2,
  203,
  208,
  0xFE,
  1,59,6,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,57,12,
  127,
  133,
  139,
  144,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,55,12,
  127,
  133,
  139,
  144,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  1,98,1,
  197,
  0xFE,
  0xFE,
  1,53,13,
  127,
  133,
  134,
  139,
  144,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,51,14,
  127,
  133,
  139,
  144,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  0xFE,
  1,49,23,
  127,
  128,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  217,
  0,250,3,
  157,
  163,
  175,
  0xFE,
  0,252,2,
  157,
  163,
  0xFE,
  1,99,2,
  203,
  208,
  0xFE,
  0xFE,
  1,47,18,
  127,
  133,
  139,
  140,
  144,
  145,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  1,99,2,
  191,
  197,
  0xFE,
  0xFE,
  1,45,16,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  208,
  209,
  217,
  221,
  233,
  239,
  1,97,2,
  233,
  239,
  0xFE,
  1,92,2,
  221,
  227,
  0xFE,
  0xFE,
  1,44,2,
  203,
  208,
  1,17,5,
  191,
  192,
  193,
  197,
  198,
  0xFE,
  1,11,22,
  133,
  163,
  164,
  169,
  170,
  171,
  175,
  191,
  192,
  193,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  0xFE,
  1,5,12,
  133,
  139,
  144,
  157,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  197,
  0xFE,
  0,255,3,
  163,
  169,
  175,
  0xFE,
  0,253,2,
  203,
  208,
  0xFE,
  0,251,2,
  203,
  208,
  0xFE,
  1,105,2,
  233,
  239,
  0xFE,
  1,103,2,
  221,
  227,
  0xFE,
  1,99,3,
  221,
  233,
  239,
  0xFE,
  1,97,2,
  221,
  227,
  0,255,2,
  221,
  227,
  0xFE,
  0,253,2,
  221,
  227,
  0xFE,
  0,251,2,
  221,
  227,
  0xFE,
  0xFE,
  0xFE,
  1,103,2,
  233,
  239,
  0xFE,
  1,101,6,
  139,
  144,
  221,
  227,
  233,
  239,
  0xFE,
  1,99,9,
  127,
  133,
  144,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,97,10,
  127,
  133,
  139,
  144,
  221,
  222,
  227,
  228,
  233,
  239,
  0xFE,
  1,95,16,
  127,
  139,
  144,
  191,
  197,
  221,
  222,
  223,
  227,
  228,
  233,
  234,
  235,
  239,
  240,
  241,
  1,17,0,
  1,70,2,
  157,
  163,
  0xFE,
  0xFE,
  1,15,0,
  1,52,1,
  191,
  0xFE,
  1,66,2,
  203,
  208,
  0xFE,
  1,68,2,
  191,
  197,
  0xFE,
  0xFE,
  1,13,0,
  1,52,4,
  197,
  198,
  227,
  228,
  0xFE,
  0xFE,
  1,9,8,
  233,
  234,
  235,
  236,
  239,
  240,
  241,
  242,
  0xFE,
  1,7,4,
  233,
  234,
  239,
  240,
  0xFE,
  1,5,4,
  233,
  234,
  239,
  240,
  0xFE,
  1,3,4,
  233,
  234,
  239,
  240,
  0xFE,
  1,1,24,
  127,
  128,
  133,
  134,
  191,
  192,
  197,
  198,
  203,
  208,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  235,
  236,
  239,
  240,
  241,
  242,
  1,47,2,
  203,
  208,
  0xFE,
  1,45,7,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,44,3,
  197,
  198,
  227,
  0xFE,
  0xFE,
  0,255,8,
  191,
  221,
  222,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  0,253,9,
  191,
  197,
  221,
  222,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  0,251,8,
  191,
  197,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  0,249,2,
  233,
  239,
  0xFE,
  0,248,1,
  221,
  0xFE,
  1,57,2,
  203,
  208,
  0xFE,
  1,55,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,53,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,51,6,
  127,
  133,
  191,
  197,
  203,
  208,
  0xFE,
  1,49,6,
  127,
  133,
  191,
  197,
  203,
  208,
  0xFE,
  1,47,7,
  133,
  139,
  144,
  191,
  197,
  203,
  208,
  0xFE,
  1,45,5,
  197,
  203,
  208,
  221,
  227,
  0xFE,
  1,44,1,
  197,
  0xFE,
  1,48,0,
  1,11,6,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  1,9,5,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,3,2,
  233,
  239,
  0xFE,
  0xFE,
  1,50,0,
  1,13,21,
  127,
  128,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  197,
  198,
  221,
  222,
  227,
  228,
  0xFE,
  1,11,10,
  133,
  139,
  144,
  157,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,9,13,
  127,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,7,6,
  127,
  133,
  157,
  163,
  164,
  169,
  0xFE,
  1,5,2,
  157,
  163,
  0xFE,
  1,3,2,
  163,
  175,
  0xFE,
  1,1,2,
  157,
  163,
  0xFE,
  0xFE,
  0xFE,
  1,93,14,
  221,
  222,
  223,
  224,
  227,
  228,
  229,
  230,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,92,2,
  245,
  249,
  0xFE,
  1,94,0,
  1,17,0,
  1,48,5,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  0xFE,
  1,15,13,
  163,
  164,
  221,
  222,
  227,
  228,
  229,
  233,
  234,
  235,
  239,
  240,
  241,
  1,48,6,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  0xFE,
  1,13,15,
  163,
  164,
  221,
  222,
  227,
  228,
  229,
  233,
  234,
  235,
  236,
  239,
  240,
  241,
  242,
  1,44,4,
  191,
  197,
  198,
  227,
  0xFE,
  0xFE,
  1,11,9,
  133,
  157,
  163,
  164,
  233,
  234,
  239,
  240,
  241,
  0xFE,
  1,9,2,
  169,
  175,
  0xFE,
  1,7,6,
  157,
  163,
  169,
  175,
  221,
  227,
  0xFE,
  1,5,5,
  133,
  157,
  163,
  169,
  175,
  0xFE,
  1,3,8,
  127,
  133,
  157,
  163,
  169,
  175,
  233,
  239,
  0xFE,
  1,1,4,
  157,
  163,
  233,
  239,
  1,51,4,
  203,
  204,
  208,
  209,
  0xFE,
  1,49,7,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,47,4,
  191,
  192,
  197,
  198,
  0xFE,
  1,45,1,
  163,
  0xFE,
  0xFE,
  1,61,2,
  203,
  208,
  0xFE,
  1,59,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,57,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,55,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,53,6,
  127,
  133,
  191,
  197,
  203,
  208,
  0xFE,
  1,51,8,
  127,
  139,
  144,
  163,
  191,
  197,
  203,
  208,
  0,255,2,
  169,
  175,
  0xFE,
  0,253,2,
  157,
  163,
  0xFE,
  0xFE,
  1,49,3,
  191,
  203,
  208,
  0xFE,
  1,47,2,
  191,
  197,
  0,255,2,
  191,
  197,
  0xFE,
  0,253,2,
  191,
  197,
  0xFE,
  0,251,2,
  191,
  197,
  0xFE,
  0xFE,
  1,45,0,
  1,9,3,
  163,
  169,
  175,
  0xFE,
  1,7,3,
  163,
  169,
  175,
  0xFE,
  1,5,1,
  163,
  0xFE,
  1,3,1,
  163,
  0xFE,
  0xFE,
  1,46,0,
  1,13,6,
  133,
  163,
  164,
  197,
  221,
  227,
  0xFE,
  1,11,2,
  169,
  175,
  0xFE,
  1,9,3,
  163,
  169,
  175,
  0xFE,
  1,7,1,
  163,
  0xFE,
  1,3,1,
  163,
  0xFE,
  0xFE,
  0xFE,
  0xFE,
  0,212,58,
  3,
  4,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  25,
  38,
  44,
  71,
  72,
  77,
  78,
  98,
  104,
  105,
  110,
  111,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  181,
  186,
  197,
  203,
  204,
  208,
  209,
  233,
  234,
  235,
  239,
  240,
  241,
  0,32,0,
  0,147,1,
  98,
  0xFE,
  0xFE,
  0,34,0,
  0,153,1,
  98,
  0xFE,
  0xFE,
  0,38,0,
  0,99,1,
  34,
  0xFE,
  0,165,1,
  98,
  0xFE,
  0xFE,
  0,88,0,
  0,155,1,
  65,
  0xFE,
  0xFE,
  0,92,0,
  0,161,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,159,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,157,4,
  95,
  98,
  104,
  110,
  0xFE,
  0xFE,
  0,96,0,
  0,163,1,
  98,
  0xFE,
  0xFE,
  0,98,0,
  0,167,1,
  98,
  0xFE,
  0xFE,
  0,114,0,
  0,197,3,
  65,
  98,
  99,
  0xFE,
  0xFE,
  0,159,0,
  0,36,2,
  34,
  98,
  0xFE,
  0,94,0,
  0,253,2,
  65,
  98,
  0xFE,
  0xFE,
  0xFE,
  1,25,0,
  1,56,1,
  191,
  0xFE,
  1,58,2,
  203,
  208,
  0xFE,
  1,68,1,
  197,
  0xFE,
  0xFE,
  1,21,8,
  169,
  170,
  171,
  175,
  176,
  177,
  234,
  240,
  1,56,2,
  197,
  198,
  0xFE,
  1,92,5,
  221,
  222,
  223,
  227,
  228,
  0xFE,
  0xFE,
  1,19,19,
  139,
  140,
  141,
  144,
  145,
  146,
  157,
  158,
  159,
  163,
  164,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  0xFE,
  1,17,34,
  127,
  128,
  129,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  186,
  1,102,2,
  169,
  175,
  0xFE,
  0xFE,
  1,15,52,
  127,
  128,
  129,
  130,
  131,
  133,
  134,
  135,
  136,
  137,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  160,
  161,
  162,
  163,
  164,
  165,
  166,
  167,
  168,
  169,
  170,
  171,
  172,
  173,
  174,
  175,
  176,
  177,
  178,
  179,
  180,
  181,
  182,
  183,
  186,
  187,
  188,
  0xFE,
  1,13,51,
  127,
  128,
  129,
  130,
  131,
  133,
  134,
  135,
  136,
  137,
  139,
  140,
  141,
  142,
  143,
  144,
  145,
  146,
  147,
  148,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  160,
  161,
  162,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  172,
  173,
  175,
  176,
  177,
  178,
  179,
  181,
  182,
  186,
  187,
  221,
  222,
  0xFE,
  1,11,80,
  3,
  13,
  34,
  65,
  98,
  104,
  110,
  127,
  128,
  129,
  130,
  131,
  133,
  134,
  135,
  136,
  137,
  139,
  140,
  141,
  142,
  143,
  144,
  145,
  146,
  147,
  148,
  149,
  153,
  154,
  157,
  158,
  159,
  160,
  161,
  162,
  163,
  164,
  165,
  166,
  167,
  168,
  169,
  170,
  171,
  172,
  173,
  175,
  176,
  177,
  178,
  179,
  181,
  186,
  187,
  197,
  198,
  199,
  203,
  204,
  205,
  206,
  207,
  208,
  209,
  210,
  211,
  212,
  227,
  228,
  233,
  234,
  235,
  236,
  237,
  239,
  240,
  241,
  242,
  243,
  0xFE,
  1,9,60,
  3,
  98,
  127,
  128,
  129,
  130,
  131,
  133,
  134,
  135,
  136,
  137,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  172,
  173,
  175,
  176,
  177,
  178,
  179,
  181,
  182,
  183,
  186,
  187,
  188,
  203,
  204,
  208,
  209,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  1,7,63,
  104,
  110,
  127,
  128,
  129,
  130,
  131,
  133,
  134,
  135,
  136,
  137,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  172,
  173,
  175,
  176,
  177,
  178,
  179,
  181,
  182,
  183,
  186,
  187,
  188,
  197,
  203,
  204,
  205,
  208,
  209,
  210,
  233,
  234,
  235,
  239,
  240,
  241,
  1,92,2,
  169,
  175,
  0xFE,
  0xFE,
  1,5,63,
  3,
  7,
  13,
  65,
  104,
  110,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  183,
  186,
  187,
  188,
  197,
  203,
  204,
  205,
  208,
  209,
  210,
  233,
  234,
  235,
  239,
  240,
  241,
  1,92,6,
  157,
  163,
  169,
  175,
  221,
  227,
  0xFE,
  0xFE,
  1,3,58,
  3,
  7,
  13,
  98,
  104,
  110,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  183,
  186,
  187,
  188,
  197,
  203,
  204,
  208,
  209,
  233,
  234,
  239,
  240,
  0,32,1,
  157,
  0xFE,
  0,86,1,
  157,
  0xFE,
  1,92,4,
  157,
  163,
  221,
  227,
  0xFE,
  0xFE,
  1,1,77,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  20,
  25,
  26,
  34,
  38,
  44,
  50,
  56,
  65,
  71,
  72,
  77,
  78,
  98,
  104,
  105,
  110,
  111,
  116,
  122,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  186,
  187,
  188,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  227,
  233,
  234,
  235,
  239,
  240,
  241,
  0,42,0,
  0,105,2,
  34,
  65,
  0xFE,
  0xFE,
  0,92,0,
  0,155,5,
  65,
  95,
  98,
  104,
  110,
  0xFE,
  0xFE,
  1,55,4,
  203,
  204,
  208,
  209,
  0xFE,
  1,53,10,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,51,14,
  127,
  133,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  1,94,1,
  197,
  0xFE,
  0xFE,
  1,49,24,
  127,
  128,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  0xFE,
  1,47,19,
  133,
  139,
  140,
  144,
  145,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  1,94,1,
  191,
  0xFE,
  0xFE,
  1,45,13,
  157,
  163,
  169,
  175,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  1,92,2,
  221,
  227,
  0xFE,
  0xFE,
  1,46,1,
  169,
  0xFE,
  1,48,7,
  133,
  139,
  144,
  157,
  163,
  169,
  175,
  0xFE,
  1,99,4,
  233,
  234,
  239,
  240,
  0xFE,
  1,97,4,
  221,
  222,
  227,
  228,
  1,44,2,
  221,
  227,
  0xFE,
  0xFE,
  1,93,12,
  221,
  222,
  223,
  224,
  227,
  228,
  229,
  230,
  233,
  234,
  239,
  240,
  0xFE,
  0xFE,
  0,255,57,
  3,
  7,
  8,
  13,
  14,
  19,
  25,
  104,
  110,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  183,
  186,
  187,
  188,
  197,
  203,
  204,
  208,
  209,
  233,
  234,
  239,
  240,
  0xFE,
  0,253,57,
  3,
  7,
  8,
  13,
  14,
  25,
  71,
  77,
  98,
  104,
  110,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  186,
  187,
  197,
  203,
  204,
  208,
  209,
  233,
  234,
  239,
  240,
  0xFE,
  0,251,55,
  3,
  7,
  8,
  13,
  14,
  19,
  25,
  71,
  77,
  98,
  104,
  110,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  153,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  186,
  187,
  203,
  204,
  208,
  209,
  233,
  234,
  239,
  240,
  1,51,2,
  203,
  208,
  0xFE,
  0xFE,
  0,249,29,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  186,
  0xFE,
  1,63,2,
  203,
  208,
  0xFE,
  1,61,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,59,6,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,57,12,
  127,
  133,
  139,
  144,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,55,12,
  127,
  133,
  139,
  144,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,53,12,
  127,
  133,
  139,
  144,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,51,12,
  127,
  133,
  139,
  144,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,49,15,
  127,
  133,
  139,
  144,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  1,11,6,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,9,2,
  169,
  175,
  0xFE,
  1,7,1,
  163,
  1,104,1,
  191,
  0xFE,
  0xFE,
  1,5,6,
  139,
  144,
  169,
  170,
  175,
  176,
  0xFE,
  1,3,12,
  127,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  0,255,10,
  127,
  133,
  134,
  139,
  157,
  158,
  163,
  164,
  169,
  175,
  0xFE,
  0,253,10,
  127,
  133,
  144,
  157,
  163,
  169,
  175,
  176,
  191,
  197,
  0xFE,
  0,251,13,
  127,
  133,
  139,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  197,
  0xFE,
  0,249,9,
  127,
  133,
  157,
  158,
  163,
  164,
  169,
  175,
  197,
  0xFE,
  0xFE,
  1,47,7,
  139,
  144,
  191,
  203,
  204,
  208,
  209,
  1,11,15,
  191,
  192,
  193,
  197,
  198,
  203,
  204,
  205,
  206,
  207,
  208,
  209,
  210,
  211,
  212,
  1,94,2,
  191,
  192,
  0xFE,
  0xFE,
  1,9,4,
  203,
  204,
  208,
  209,
  0xFE,
  1,7,7,
  191,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,5,7,
  169,
  175,
  191,
  203,
  204,
  208,
  209,
  0xFE,
  1,3,9,
  157,
  163,
  169,
  175,
  191,
  203,
  204,
  208,
  209,
  0xFE,
  0,255,8,
  127,
  133,
  157,
  163,
  175,
  191,
  203,
  208,
  0xFE,
  0,253,3,
  163,
  203,
  208,
  0xFE,
  0,251,2,
  203,
  208,
  0xFE,
  1,104,0,
  1,17,6,
  127,
  133,
  157,
  163,
  169,
  175,
  0xFE,
  1,15,7,
  133,
  139,
  144,
  157,
  163,
  169,
  175,
  0xFE,
  1,13,7,
  133,
  139,
  144,
  157,
  163,
  169,
  175,
  0xFE,
  1,11,4,
  127,
  133,
  157,
  163,
  0xFE,
  1,9,1,
  157,
  0xFE,
  0xFE,
  0xFE,
  1,45,7,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  1,11,15,
  191,
  192,
  193,
  194,
  195,
  197,
  198,
  199,
  200,
  201,
  203,
  204,
  208,
  209,
  210,
  1,92,1,
  221,
  0xFE,
  0xFE,
  1,9,6,
  191,
  192,
  193,
  197,
  198,
  199,
  0xFE,
  1,7,6,
  191,
  192,
  193,
  197,
  198,
  199,
  0xFE,
  1,5,6,
  191,
  192,
  193,
  197,
  198,
  199,
  0xFE,
  1,3,7,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  0xFE,
  0,255,6,
  157,
  163,
  191,
  192,
  197,
  198,
  0xFE,
  0,253,4,
  191,
  192,
  197,
  198,
  0xFE,
  0,251,4,
  191,
  192,
  197,
  198,
  0xFE,
  0xFE,
  1,44,0,
  1,21,3,
  191,
  192,
  198,
  0xFE,
  1,17,2,
  197,
  198,
  1,98,2,
  163,
  175,
  0xFE,
  0xFE,
  1,15,0,
  1,98,2,
  157,
  163,
  0xFE,
  0xFE,
  1,13,6,
  191,
  192,
  203,
  204,
  208,
  209,
  0xFE,
  1,9,2,
  169,
  175,
  0xFE,
  1,7,7,
  133,
  139,
  144,
  157,
  163,
  169,
  175,
  0xFE,
  1,5,8,
  127,
  133,
  139,
  144,
  157,
  163,
  169,
  175,
  0xFE,
  1,3,5,
  127,
  133,
  157,
  163,
  175,
  0xFE,
  0xFE,
  1,46,0,
  1,17,13,
  133,
  134,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  199,
  0xFE,
  1,15,7,
  157,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,13,13,
  133,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  192,
  193,
  197,
  198,
  199,
  0xFE,
  1,11,8,
  139,
  144,
  157,
  163,
  169,
  170,
  175,
  176,
  0xFE,
  1,9,12,
  127,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,7,10,
  127,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  175,
  0xFE,
  1,5,7,
  127,
  133,
  144,
  157,
  163,
  169,
  175,
  0xFE,
  1,3,8,
  127,
  133,
  139,
  144,
  157,
  163,
  169,
  175,
  0xFE,
  0xFE,
  1,48,0,
  1,23,6,
  169,
  175,
  233,
  234,
  239,
  240,
  1,92,4,
  221,
  222,
  227,
  228,
  0xFE,
  0xFE,
  1,21,13,
  139,
  144,
  157,
  158,
  159,
  163,
  164,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,19,23,
  127,
  128,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,17,27,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  0xFE,
  1,15,20,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,13,35,
  127,
  128,
  133,
  134,
  135,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  233,
  234,
  235,
  239,
  240,
  241,
  1,93,6,
  221,
  222,
  223,
  227,
  228,
  229,
  0xFE,
  0xFE,
  1,11,21,
  127,
  128,
  133,
  134,
  135,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,9,17,
  127,
  128,
  133,
  134,
  135,
  139,
  144,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  0xFE,
  1,7,11,
  127,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  175,
  0xFE,
  1,5,10,
  127,
  133,
  139,
  144,
  157,
  163,
  169,
  170,
  175,
  176,
  0xFE,
  1,3,13,
  7,
  13,
  127,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  175,
  1,93,2,
  221,
  227,
  0xFE,
  0xFE,
  0xFE,
  1,50,0,
  1,23,3,
  191,
  192,
  198,
  0xFE,
  1,19,4,
  163,
  191,
  197,
  198,
  0xFE,
  1,13,0,
  1,97,3,
  221,
  222,
  227,
  0xFE,
  0xFE,
  1,11,2,
  169,
  175,
  0xFE,
  1,9,2,
  157,
  163,
  0xFE,
  1,7,2,
  163,
  175,
  0xFE,
  1,5,1,
  163,
  0xFE,
  0xFE,
  1,52,0,
  1,19,4,
  163,
  191,
  197,
  198,
  0xFE,
  1,99,0,
  1,17,3,
  163,
  169,
  175,
  0xFE,
  1,15,3,
  157,
  169,
  175,
  0xFE,
  1,13,3,
  163,
  169,
  175,
  0xFE,
  1,11,5,
  133,
  157,
  163,
  169,
  175,
  0xFE,
  1,9,4,
  127,
  133,
  157,
  163,
  0xFE,
  0xFE,
  0xFE,
  1,54,0,
  1,25,4,
  203,
  208,
  233,
  239,
  1,92,2,
  221,
  227,
  0xFE,
  0xFE,
  1,23,2,
  157,
  163,
  0xFE,
  1,21,6,
  133,
  139,
  144,
  163,
  169,
  175,
  0xFE,
  1,19,1,
  175,
  0xFE,
  1,103,0,
  1,13,1,
  175,
  0xFE,
  1,11,3,
  163,
  169,
  175,
  0xFE,
  1,9,3,
  133,
  157,
  163,
  0xFE,
  0xFE,
  0xFE,
  1,103,2,
  233,
  239,
  0xFE,
  1,101,6,
  127,
  133,
  221,
  227,
  233,
  239,
  0xFE,
  1,99,8,
  127,
  133,
  139,
  144,
  221,
  227,
  233,
  239,
  0xFE,
  1,97,6,
  139,
  144,
  221,
  227,
  233,
  239,
  0xFE,
  1,95,7,
  221,
  222,
  227,
  233,
  234,
  239,
  240,
  1,13,0,
  1,48,6,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  0xFE,
  1,11,14,
  191,
  192,
  221,
  222,
  233,
  234,
  235,
  236,
  237,
  239,
  240,
  241,
  242,
  243,
  0xFE,
  1,9,4,
  233,
  234,
  239,
  240,
  0xFE,
  1,7,6,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  1,5,6,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  1,3,4,
  233,
  234,
  239,
  240,
  0xFE,
  1,1,11,
  191,
  197,
  221,
  222,
  227,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  0,255,3,
  221,
  233,
  239,
  0xFE,
  0,253,3,
  221,
  233,
  239,
  0xFE,
  0,251,3,
  221,
  233,
  239,
  0xFE,
  1,57,2,
  203,
  208,
  0xFE,
  1,55,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,53,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,51,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,49,3,
  191,
  197,
  203,
  0xFE,
  1,50,0,
  1,17,7,
  133,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,15,6,
  157,
  158,
  163,
  164,
  169,
  175,
  0xFE,
  1,13,7,
  157,
  163,
  169,
  175,
  176,
  233,
  239,
  0xFE,
  1,11,8,
  133,
  144,
  157,
  158,
  163,
  164,
  169,
  175,
  0xFE,
  1,9,4,
  127,
  133,
  157,
  163,
  0xFE,
  0xFE,
  0xFE,
  1,93,8,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  239,
  1,11,10,
  221,
  222,
  223,
  224,
  225,
  227,
  228,
  229,
  230,
  231,
  0xFE,
  1,9,6,
  221,
  222,
  223,
  227,
  228,
  229,
  0xFE,
  1,7,6,
  221,
  222,
  223,
  227,
  228,
  229,
  0xFE,
  1,5,6,
  221,
  222,
  223,
  227,
  228,
  229,
  0xFE,
  1,3,4,
  221,
  222,
  227,
  228,
  0xFE,
  0,255,4,
  221,
  222,
  227,
  228,
  0xFE,
  0,253,4,
  221,
  222,
  227,
  228,
  0xFE,
  0,251,4,
  221,
  222,
  227,
  228,
  0xFE,
  0xFE,
  1,94,0,
  1,9,2,
  169,
  175,
  0xFE,
  1,7,4,
  157,
  163,
  169,
  175,
  0xFE,
  1,5,3,
  133,
  157,
  163,
  0xFE,
  1,59,2,
  203,
  208,
  0xFE,
  1,57,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,55,5,
  133,
  191,
  197,
  203,
  208,
  0xFE,
  1,53,3,
  191,
  197,
  203,
  0xFE,
  0xFE,
  1,96,0,
  1,17,3,
  133,
  157,
  163,
  0xFE,
  1,15,3,
  163,
  169,
  175,
  0xFE,
  1,13,3,
  163,
  169,
  175,
  0xFE,
  1,11,3,
  133,
  157,
  163,
  0xFE,
  0xFE,
  0xFE,
  0,214,28,
  3,
  7,
  8,
  13,
  14,
  104,
  110,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  169,
  170,
  175,
  176,
  203,
  208,
  233,
  239,
  0,32,0,
  0,147,1,
  98,
  0xFE,
  0xFE,
  0,34,0,
  0,153,1,
  98,
  0,249,1,
  98,
  0xFE,
  0xFE,
  0xFE,
  0,38,0,
  0,99,4,
  34,
  35,
  65,
  66,
  0xFE,
  0xFE,
  0,92,0,
  0,163,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,161,3,
  95,
  98,
  104,
  0xFE,
  0,159,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,157,3,
  98,
  104,
  110,
  0xFE,
  0xFE,
  0,98,0,
  0,167,1,
  98,
  0xFE,
  0xFE,
  0,100,0,
  0,171,4,
  65,
  66,
  98,
  99,
  0xFE,
  0xFE,
  1,27,0,
  1,54,2,
  169,
  175,
  1,92,2,
  221,
  227,
  0xFE,
  0xFE,
  1,56,2,
  191,
  197,
  0xFE,
  1,58,2,
  203,
  208,
  0xFE,
  0xFE,
  1,25,0,
  1,50,2,
  191,
  192,
  0xFE,
  1,54,12,
  133,
  134,
  157,
  163,
  164,
  169,
  175,
  191,
  197,
  198,
  203,
  208,
  0xFE,
  0xFE,
  1,23,26,
  139,
  140,
  141,
  144,
  145,
  146,
  157,
  158,
  159,
  163,
  164,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  204,
  209,
  233,
  234,
  235,
  239,
  240,
  1,44,4,
  191,
  192,
  193,
  198,
  0xFE,
  1,46,5,
  197,
  198,
  199,
  204,
  209,
  0xFE,
  1,50,1,
  208,
  0xFE,
  1,52,2,
  191,
  197,
  0xFE,
  1,54,2,
  169,
  175,
  0xFE,
  1,92,6,
  221,
  222,
  223,
  227,
  228,
  229,
  0xFE,
  1,102,2,
  133,
  163,
  0xFE,
  0xFE,
  1,21,38,
  127,
  128,
  129,
  130,
  131,
  133,
  134,
  135,
  136,
  137,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  172,
  173,
  175,
  176,
  177,
  178,
  179,
  1,92,2,
  228,
  229,
  0xFE,
  0xFE,
  1,19,37,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  137,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  172,
  173,
  175,
  176,
  177,
  178,
  179,
  0xFE,
  1,17,36,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  153,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  182,
  186,
  187,
  0xFE,
  1,15,38,
  127,
  128,
  129,
  133,
  134,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  149,
  153,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  173,
  175,
  176,
  177,
  178,
  179,
  181,
  182,
  186,
  187,
  1,52,0,
  1,99,2,
  169,
  175,
  0xFE,
  0xFE,
  1,96,2,
  169,
  175,
  0xFE,
  0xFE,
  1,13,44,
  127,
  128,
  129,
  130,
  131,
  133,
  134,
  135,
  136,
  137,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  149,
  153,
  157,
  158,
  159,
  160,
  161,
  162,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  172,
  173,
  175,
  176,
  177,
  178,
  179,
  181,
  182,
  186,
  1,52,0,
  1,99,4,
  157,
  163,
  169,
  175,
  0xFE,
  0xFE,
  1,96,1,
  163,
  0xFE,
  0xFE,
  1,11,71,
  3,
  7,
  13,
  104,
  110,
  127,
  128,
  129,
  130,
  131,
  133,
  134,
  135,
  136,
  137,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  172,
  173,
  175,
  176,
  177,
  178,
  179,
  181,
  182,
  183,
  186,
  187,
  188,
  197,
  198,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  233,
  234,
  235,
  236,
  239,
  240,
  241,
  242,
  1,52,0,
  1,99,3,
  133,
  157,
  163,
  0xFE,
  0xFE,
  1,54,0,
  1,103,1,
  163,
  0xFE,
  0xFE,
  0xFE,
  1,9,50,
  127,
  128,
  129,
  130,
  131,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  149,
  150,
  153,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  186,
  187,
  203,
  204,
  208,
  209,
  233,
  234,
  239,
  240,
  0xFE,
  1,7,52,
  3,
  7,
  13,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  183,
  186,
  187,
  203,
  204,
  208,
  209,
  233,
  234,
  239,
  240,
  1,92,1,
  163,
  0xFE,
  0xFE,
  1,5,54,
  3,
  7,
  13,
  65,
  104,
  110,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  183,
  186,
  187,
  188,
  203,
  204,
  208,
  209,
  233,
  234,
  239,
  240,
  0xFE,
  1,3,42,
  7,
  13,
  104,
  110,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  186,
  187,
  203,
  208,
  233,
  239,
  0xFE,
  1,1,56,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  98,
  104,
  105,
  110,
  111,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  182,
  186,
  187,
  197,
  203,
  204,
  208,
  209,
  233,
  234,
  239,
  240,
  1,57,2,
  203,
  208,
  0xFE,
  1,55,6,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,53,8,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,51,11,
  127,
  133,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  208,
  209,
  0xFE,
  1,99,2,
  221,
  227,
  0xFE,
  0xFE,
  0,255,45,
  3,
  7,
  13,
  104,
  110,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  186,
  187,
  203,
  208,
  233,
  239,
  0xFE,
  0,253,43,
  3,
  7,
  13,
  104,
  110,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  153,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  181,
  186,
  187,
  203,
  208,
  233,
  239,
  0xFE,
  0,251,36,
  3,
  7,
  13,
  104,
  110,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  203,
  208,
  233,
  239,
  0xFE,
  0,249,24,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  181,
  186,
  0xFE,
  1,63,2,
  203,
  208,
  0xFE,
  1,61,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,59,5,
  133,
  191,
  197,
  203,
  208,
  0xFE,
  1,57,7,
  127,
  133,
  139,
  191,
  197,
  203,
  208,
  0xFE,
  1,55,9,
  127,
  133,
  139,
  144,
  191,
  197,
  203,
  204,
  208,
  0xFE,
  1,53,10,
  127,
  133,
  139,
  144,
  191,
  192,
  197,
  198,
  203,
  208,
  0xFE,
  1,51,6,
  139,
  144,
  191,
  197,
  203,
  208,
  0xFE,
  1,49,2,
  197,
  203,
  1,11,4,
  203,
  204,
  208,
  209,
  0xFE,
  1,9,1,
  163,
  0xFE,
  1,7,2,
  169,
  175,
  0xFE,
  1,5,12,
  127,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,3,11,
  127,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  175,
  176,
  0xFE,
  1,1,15,
  127,
  133,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  208,
  209,
  0xFE,
  0,255,6,
  133,
  144,
  157,
  163,
  169,
  175,
  0xFE,
  0,253,6,
  127,
  133,
  157,
  163,
  169,
  175,
  0xFE,
  0,251,8,
  127,
  133,
  157,
  158,
  163,
  164,
  169,
  175,
  0xFE,
  0,249,2,
  157,
  163,
  0xFE,
  0xFE,
  1,47,0,
  1,11,10,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,9,2,
  203,
  208,
  0xFE,
  1,7,2,
  203,
  208,
  0xFE,
  1,5,5,
  163,
  169,
  175,
  203,
  208,
  0xFE,
  1,3,6,
  157,
  163,
  169,
  175,
  203,
  208,
  0xFE,
  1,1,6,
  133,
  157,
  163,
  203,
  204,
  208,
  0xFE,
  0,255,1,
  163,
  0xFE,
  1,104,0,
  1,21,2,
  169,
  175,
  0xFE,
  1,19,3,
  157,
  169,
  175,
  0xFE,
  1,17,3,
  163,
  169,
  175,
  0xFE,
  1,15,4,
  157,
  163,
  169,
  175,
  0xFE,
  1,13,4,
  127,
  133,
  157,
  163,
  0xFE,
  0xFE,
  0xFE,
  1,45,2,
  191,
  197,
  1,13,1,
  191,
  0xFE,
  1,11,8,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  0xFE,
  1,9,5,
  191,
  192,
  193,
  197,
  198,
  0xFE,
  1,7,4,
  191,
  192,
  197,
  198,
  0xFE,
  1,5,6,
  169,
  175,
  191,
  192,
  197,
  198,
  0xFE,
  1,3,5,
  157,
  163,
  175,
  191,
  197,
  0xFE,
  1,1,5,
  163,
  191,
  192,
  197,
  198,
  0xFE,
  0,255,2,
  191,
  197,
  0xFE,
  0,253,2,
  191,
  197,
  0xFE,
  0,251,2,
  191,
  197,
  0xFE,
  0xFE,
  1,44,0,
  1,21,2,
  197,
  199,
  0xFE,
  1,11,2,
  169,
  175,
  0xFE,
  1,9,4,
  157,
  163,
  169,
  175,
  0xFE,
  1,7,6,
  127,
  133,
  157,
  163,
  169,
  175,
  0xFE,
  1,5,4,
  127,
  133,
  157,
  163,
  0xFE,
  0xFE,
  1,46,0,
  1,17,2,
  169,
  175,
  0xFE,
  1,15,4,
  157,
  163,
  169,
  175,
  0xFE,
  1,13,7,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  0xFE,
  1,11,9,
  127,
  133,
  139,
  144,
  157,
  158,
  163,
  169,
  175,
  0xFE,
  1,9,10,
  127,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  175,
  0xFE,
  1,7,7,
  127,
  133,
  144,
  157,
  163,
  169,
  175,
  0xFE,
  1,5,6,
  127,
  133,
  157,
  163,
  169,
  175,
  0xFE,
  0xFE,
  1,48,0,
  1,25,1,
  204,
  1,92,4,
  221,
  222,
  227,
  228,
  0xFE,
  0xFE,
  1,23,32,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  144,
  145,
  146,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  177,
  191,
  192,
  197,
  198,
  0xFE,
  1,21,28,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  0xFE,
  1,19,26,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,17,22,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,15,20,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,13,27,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  203,
  204,
  208,
  209,
  233,
  234,
  239,
  240,
  1,93,4,
  221,
  222,
  227,
  228,
  0xFE,
  0xFE,
  1,11,16,
  127,
  128,
  133,
  134,
  139,
  144,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  0xFE,
  1,9,10,
  127,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  175,
  0xFE,
  1,7,9,
  127,
  133,
  139,
  157,
  163,
  169,
  170,
  175,
  176,
  0xFE,
  1,5,9,
  127,
  133,
  144,
  157,
  158,
  163,
  164,
  169,
  175,
  0xFE,
  1,3,5,
  144,
  157,
  163,
  169,
  175,
  0xFE,
  0xFE,
  1,105,2,
  233,
  239,
  0xFE,
  1,103,4,
  221,
  227,
  233,
  239,
  0xFE,
  1,101,2,
  221,
  227,
  0xFE,
  1,95,0,
  1,13,0,
  1,48,2,
  233,
  239,
  0xFE,
  0xFE,
  1,11,6,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  1,9,4,
  233,
  234,
  239,
  240,
  0xFE,
  1,7,2,
  233,
  239,
  0xFE,
  1,5,2,
  233,
  239,
  0xFE,
  1,3,2,
  233,
  239,
  0xFE,
  1,1,5,
  221,
  233,
  234,
  239,
  240,
  0xFE,
  0,255,2,
  233,
  239,
  0xFE,
  1,57,2,
  203,
  208,
  0xFE,
  1,55,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,53,3,
  191,
  197,
  208,
  0xFE,
  1,51,2,
  191,
  197,
  0xFE,
  1,50,0,
  1,21,2,
  169,
  175,
  0xFE,
  1,19,3,
  163,
  169,
  175,
  0xFE,
  1,17,1,
  163,
  0xFE,
  1,15,2,
  169,
  175,
  0xFE,
  1,13,5,
  133,
  157,
  163,
  169,
  175,
  0xFE,
  1,11,3,
  133,
  157,
  163,
  0xFE,
  0xFE,
  0xFE,
  1,93,2,
  221,
  227,
  1,23,0,
  1,54,2,
  221,
  227,
  0xFE,
  0xFE,
  1,13,4,
  221,
  222,
  227,
  228,
  0xFE,
  1,11,8,
  221,
  222,
  223,
  224,
  227,
  228,
  229,
  230,
  0xFE,
  1,9,6,
  221,
  222,
  223,
  227,
  228,
  229,
  0xFE,
  1,7,4,
  221,
  222,
  227,
  228,
  0xFE,
  1,5,4,
  221,
  222,
  227,
  228,
  0xFE,
  1,3,4,
  221,
  222,
  227,
  228,
  0xFE,
  1,1,4,
  221,
  222,
  227,
  228,
  0xFE,
  0,255,2,
  221,
  227,
  0xFE,
  0,253,2,
  221,
  227,
  0xFE,
  0,251,2,
  221,
  227,
  0xFE,
  0xFE,
  1,94,0,
  1,9,2,
  157,
  163,
  0xFE,
  1,7,2,
  157,
  163,
  0xFE,
  1,59,2,
  203,
  208,
  0xFE,
  1,57,3,
  191,
  197,
  208,
  0xFE,
  1,55,2,
  191,
  197,
  0xFE,
  0xFE,
  0xFE,
  0,216,1,
  163,
  0,92,0,
  0,163,2,
  95,
  98,
  0xFE,
  0,161,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,159,2,
  104,
  110,
  0xFE,
  0xFE,
  1,37,0,
  1,74,1,
  163,
  0xFE,
  0xFE,
  1,27,11,
  134,
  158,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,25,33,
  127,
  128,
  129,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  1,50,2,
  203,
  208,
  0xFE,
  1,52,1,
  191,
  0xFE,
  1,54,4,
  203,
  208,
  233,
  239,
  0xFE,
  0xFE,
  1,23,34,
  127,
  128,
  129,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  173,
  175,
  176,
  177,
  178,
  179,
  0xFE,
  1,21,47,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  172,
  173,
  175,
  176,
  177,
  178,
  179,
  203,
  204,
  205,
  208,
  209,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  1,19,34,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  144,
  145,
  146,
  149,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  182,
  186,
  187,
  0xFE,
  1,17,37,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  149,
  153,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  169,
  170,
  171,
  172,
  173,
  175,
  176,
  177,
  178,
  179,
  181,
  182,
  186,
  187,
  0xFE,
  1,15,41,
  127,
  128,
  129,
  130,
  131,
  133,
  134,
  135,
  136,
  137,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  153,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  172,
  173,
  175,
  176,
  177,
  178,
  179,
  182,
  186,
  1,52,0,
  1,99,3,
  163,
  169,
  175,
  0xFE,
  0xFE,
  0xFE,
  1,13,43,
  127,
  128,
  129,
  130,
  131,
  133,
  134,
  135,
  136,
  137,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  149,
  150,
  153,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  186,
  187,
  1,52,0,
  1,99,2,
  157,
  163,
  0xFE,
  0xFE,
  0xFE,
  1,11,52,
  7,
  13,
  104,
  110,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  149,
  153,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  186,
  187,
  203,
  204,
  208,
  209,
  233,
  234,
  239,
  240,
  1,44,4,
  157,
  163,
  169,
  175,
  0xFE,
  0xFE,
  1,9,42,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  186,
  187,
  233,
  239,
  1,44,5,
  133,
  157,
  163,
  169,
  175,
  0xFE,
  0xFE,
  1,7,39,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  177,
  178,
  181,
  182,
  183,
  186,
  187,
  233,
  0,94,0,
  0,159,1,
  98,
  0xFE,
  0xFE,
  1,44,2,
  157,
  163,
  0xFE,
  0xFE,
  1,5,38,
  7,
  13,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  182,
  186,
  187,
  203,
  208,
  233,
  239,
  0xFE,
  1,3,35,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  182,
  186,
  187,
  203,
  208,
  233,
  239,
  0xFE,
  1,1,39,
  7,
  13,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  182,
  186,
  187,
  203,
  208,
  233,
  239,
  1,57,2,
  203,
  208,
  0xFE,
  1,55,7,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,53,5,
  191,
  192,
  197,
  198,
  208,
  0xFE,
  1,51,2,
  191,
  197,
  0xFE,
  0xFE,
  0,255,32,
  7,
  13,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  153,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  186,
  0xFE,
  0,253,25,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  0,251,20,
  127,
  128,
  129,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  0xFE,
  0,249,12,
  127,
  133,
  139,
  140,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  0xFE,
  1,63,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,61,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,59,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,57,5,
  133,
  191,
  197,
  203,
  208,
  0xFE,
  1,55,3,
  191,
  197,
  203,
  0xFE,
  1,53,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,49,0,
  1,9,2,
  169,
  175,
  0xFE,
  1,7,6,
  133,
  139,
  157,
  163,
  169,
  175,
  0xFE,
  1,5,8,
  127,
  133,
  157,
  158,
  163,
  164,
  169,
  175,
  0xFE,
  1,3,6,
  127,
  133,
  157,
  163,
  169,
  175,
  0xFE,
  1,1,4,
  157,
  163,
  169,
  175,
  0xFE,
  0,255,5,
  133,
  157,
  163,
  169,
  175,
  0xFE,
  0,253,2,
  157,
  163,
  0xFE,
  0,251,1,
  163,
  0xFE,
  0xFE,
  1,47,0,
  1,11,2,
  203,
  208,
  0xFE,
  1,7,2,
  169,
  175,
  0xFE,
  1,5,4,
  157,
  163,
  169,
  175,
  0xFE,
  1,3,2,
  157,
  163,
  0xFE,
  1,104,0,
  1,19,2,
  169,
  175,
  0xFE,
  1,17,3,
  163,
  169,
  175,
  0xFE,
  1,15,3,
  133,
  157,
  163,
  0xFE,
  0xFE,
  0xFE,
  1,45,0,
  1,21,6,
  191,
  192,
  193,
  197,
  198,
  199,
  0xFE,
  1,11,4,
  191,
  192,
  197,
  198,
  0xFE,
  1,9,2,
  191,
  197,
  0xFE,
  1,7,1,
  191,
  0xFE,
  1,5,2,
  191,
  197,
  0xFE,
  1,3,2,
  191,
  197,
  0xFE,
  1,1,2,
  191,
  197,
  0xFE,
  0xFE,
  1,46,0,
  1,23,4,
  191,
  192,
  197,
  198,
  0xFE,
  1,17,1,
  163,
  0xFE,
  1,15,2,
  169,
  175,
  0xFE,
  1,13,7,
  133,
  139,
  144,
  157,
  163,
  169,
  175,
  0xFE,
  1,11,7,
  127,
  133,
  144,
  157,
  163,
  169,
  175,
  0xFE,
  1,9,6,
  127,
  133,
  157,
  163,
  169,
  175,
  0xFE,
  1,7,3,
  133,
  157,
  163,
  0xFE,
  0xFE,
  1,48,0,
  1,29,1,
  169,
  0xFE,
  1,27,14,
  133,
  134,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,25,16,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  160,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,23,27,
  127,
  128,
  133,
  134,
  135,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  204,
  209,
  233,
  234,
  239,
  240,
  1,93,5,
  221,
  222,
  223,
  227,
  228,
  0xFE,
  0xFE,
  1,21,22,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,19,18,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  144,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  0xFE,
  1,17,17,
  127,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  0xFE,
  1,15,15,
  127,
  128,
  133,
  134,
  139,
  144,
  157,
  158,
  159,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,13,13,
  127,
  128,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  0xFE,
  1,11,8,
  127,
  133,
  139,
  144,
  157,
  163,
  169,
  175,
  0xFE,
  1,9,7,
  127,
  133,
  139,
  157,
  163,
  169,
  175,
  0xFE,
  1,7,5,
  127,
  133,
  157,
  163,
  175,
  0xFE,
  1,5,2,
  157,
  163,
  0xFE,
  0xFE,
  1,54,0,
  1,31,1,
  175,
  0xFE,
  1,29,6,
  133,
  139,
  144,
  163,
  169,
  175,
  0xFE,
  1,27,2,
  157,
  163,
  0xFE,
  0xFE,
  1,95,0,
  1,11,2,
  233,
  239,
  0xFE,
  1,50,0,
  1,17,2,
  169,
  175,
  0xFE,
  1,15,4,
  157,
  163,
  169,
  175,
  0xFE,
  1,13,2,
  157,
  163,
  0xFE,
  0xFE,
  0xFE,
  1,93,0,
  1,25,0,
  1,54,2,
  221,
  227,
  0xFE,
  0xFE,
  1,21,6,
  221,
  222,
  223,
  227,
  228,
  229,
  0xFE,
  1,13,2,
  221,
  227,
  0xFE,
  1,11,6,
  221,
  222,
  223,
  227,
  228,
  229,
  0xFE,
  1,9,2,
  221,
  227,
  0xFE,
  1,7,2,
  221,
  227,
  0xFE,
  1,5,2,
  221,
  227,
  0xFE,
  1,3,2,
  221,
  227,
  0xFE,
  1,1,2,
  221,
  227,
  0xFE,
  0xFE,
  0xFE,
  0,218,0,
  0,92,0,
  0,165,2,
  95,
  98,
  0xFE,
  0,163,1,
  98,
  0xFE,
  0xFE,
  1,31,13,
  133,
  134,
  135,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  0xFE,
  1,29,7,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  0xFE,
  1,27,14,
  139,
  140,
  144,
  145,
  157,
  158,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  0xFE,
  1,25,27,
  127,
  128,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  0xFE,
  1,23,34,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  0xFE,
  1,21,30,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  144,
  145,
  146,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  186,
  0xFE,
  1,19,33,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  169,
  170,
  171,
  172,
  173,
  175,
  176,
  177,
  178,
  179,
  181,
  186,
  1,47,0,
  1,104,2,
  169,
  175,
  0xFE,
  0xFE,
  0xFE,
  1,17,38,
  127,
  128,
  129,
  130,
  131,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  172,
  173,
  175,
  176,
  177,
  178,
  179,
  186,
  1,47,0,
  1,104,2,
  157,
  163,
  0xFE,
  0xFE,
  0xFE,
  1,15,40,
  127,
  128,
  129,
  130,
  131,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  144,
  145,
  146,
  147,
  149,
  153,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  186,
  187,
  0xFE,
  1,13,38,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  149,
  153,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  186,
  187,
  0xFE,
  1,11,38,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  149,
  150,
  153,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  186,
  187,
  0xFE,
  1,9,33,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  153,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  182,
  186,
  187,
  0xFE,
  1,7,32,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  182,
  186,
  187,
  0xFE,
  1,5,29,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  149,
  153,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  182,
  186,
  187,
  0xFE,
  1,3,30,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  153,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  182,
  186,
  187,
  0,34,0,
  0,153,1,
  98,
  0xFE,
  0xFE,
  0xFE,
  1,1,28,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  153,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  186,
  0,38,0,
  0,99,1,
  34,
  0xFE,
  0xFE,
  0,100,0,
  0,171,1,
  98,
  0xFE,
  0xFE,
  0xFE,
  0,255,22,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  0,253,18,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  0xFE,
  0,251,9,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  0xFE,
  1,63,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,61,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,59,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,57,3,
  191,
  197,
  203,
  0xFE,
  1,49,0,
  1,11,2,
  169,
  175,
  0xFE,
  1,9,4,
  157,
  163,
  169,
  175,
  0xFE,
  1,7,5,
  133,
  157,
  163,
  169,
  175,
  0xFE,
  1,5,3,
  133,
  157,
  163,
  0xFE,
  1,3,2,
  163,
  175,
  0xFE,
  1,1,2,
  157,
  163,
  0xFE,
  0,255,1,
  163,
  0xFE,
  0xFE,
  1,44,0,
  1,13,2,
  169,
  175,
  0xFE,
  1,11,3,
  157,
  163,
  175,
  0xFE,
  1,9,2,
  157,
  163,
  0xFE,
  0xFE,
  1,46,0,
  1,17,2,
  169,
  175,
  0xFE,
  1,15,4,
  157,
  163,
  169,
  175,
  0xFE,
  1,13,6,
  127,
  133,
  157,
  163,
  169,
  175,
  0xFE,
  1,11,5,
  127,
  133,
  157,
  163,
  175,
  0xFE,
  1,9,2,
  157,
  163,
  0xFE,
  0xFE,
  1,48,0,
  1,33,6,
  133,
  134,
  157,
  158,
  163,
  164,
  0xFE,
  1,31,6,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,29,6,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,27,10,
  139,
  144,
  163,
  164,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,25,19,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,23,21,
  127,
  128,
  133,
  134,
  135,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,21,18,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  144,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  0xFE,
  1,19,14,
  127,
  133,
  134,
  139,
  140,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,17,14,
  127,
  128,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,15,12,
  127,
  128,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  175,
  0xFE,
  1,13,8,
  127,
  133,
  139,
  144,
  157,
  163,
  169,
  175,
  0xFE,
  1,11,7,
  127,
  133,
  139,
  157,
  163,
  169,
  175,
  0xFE,
  1,9,5,
  127,
  133,
  157,
  163,
  175,
  0xFE,
  0xFE,
  1,54,0,
  1,35,2,
  157,
  163,
  0xFE,
  1,33,2,
  169,
  175,
  0xFE,
  1,31,3,
  163,
  169,
  175,
  0xFE,
  1,29,3,
  133,
  157,
  163,
  0xFE,
  0xFE,
  1,93,0,
  1,21,2,
  222,
  228,
  0xFE,
  1,11,2,
  221,
  227,
  0xFE,
  1,7,2,
  221,
  227,
  0xFE,
  1,5,2,
  221,
  227,
  0xFE,
  0xFE,
  0xFE,
  0,220,0,
  0,108,0,
  0,187,1,
  98,
  0xFE,
  0xFE,
  1,35,2,
  163,
  164,
  0xFE,
  1,33,5,
  164,
  169,
  170,
  175,
  176,
  1,54,2,
  169,
  175,
  0xFE,
  0xFE,
  1,31,5,
  164,
  169,
  170,
  175,
  176,
  1,54,1,
  163,
  0xFE,
  0xFE,
  1,29,8,
  140,
  145,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,27,23,
  133,
  134,
  139,
  140,
  141,
  144,
  145,
  146,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  0xFE,
  1,25,31,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  144,
  145,
  146,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  0xFE,
  1,23,27,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  144,
  145,
  146,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,21,32,
  127,
  128,
  129,
  133,
  134,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  169,
  170,
  171,
  172,
  173,
  175,
  176,
  177,
  178,
  179,
  181,
  186,
  0xFE,
  1,19,34,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  0xFE,
  1,17,37,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  153,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  186,
  187,
  0xFE,
  1,15,35,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  149,
  153,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  186,
  0xFE,
  1,13,35,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  153,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  181,
  182,
  186,
  187,
  0xFE,
  1,11,32,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  153,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  182,
  186,
  187,
  0xFE,
  1,9,31,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  153,
  154,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  182,
  186,
  187,
  0xFE,
  1,7,29,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  149,
  153,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  182,
  186,
  187,
  0xFE,
  1,5,29,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  149,
  153,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  182,
  186,
  187,
  0xFE,
  1,3,27,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  149,
  153,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  186,
  0xFE,
  1,1,21,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  0xFE,
  0,255,17,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  0xFE,
  0,253,6,
  139,
  157,
  158,
  163,
  164,
  169,
  0xFE,
  1,49,0,
  1,11,4,
  157,
  163,
  169,
  175,
  0xFE,
  1,9,2,
  157,
  163,
  0xFE,
  1,7,2,
  157,
  163,
  0xFE,
  0xFE,
  1,46,0,
  1,17,4,
  157,
  163,
  169,
  175,
  0xFE,
  1,15,4,
  157,
  163,
  169,
  175,
  0xFE,
  1,13,2,
  157,
  163,
  0xFE,
  0xFE,
  0xFE,
  0,222,0,
  1,35,1,
  170,
  0xFE,
  1,33,5,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,31,6,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,29,23,
  133,
  134,
  139,
  140,
  141,
  144,
  145,
  146,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  0xFE,
  1,27,29,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  0xFE,
  1,25,26,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,23,27,
  127,
  128,
  129,
  133,
  134,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  0xFE,
  1,21,31,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  144,
  145,
  146,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  0xFE,
  1,19,33,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  177,
  178,
  181,
  186,
  0xFE,
  1,17,31,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  153,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  186,
  1,46,2,
  157,
  163,
  0xFE,
  0xFE,
  1,15,33,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  153,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  182,
  186,
  187,
  1,46,2,
  157,
  163,
  0xFE,
  0xFE,
  1,13,30,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  153,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  182,
  186,
  0xFE,
  1,11,29,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  153,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  186,
  187,
  1,49,1,
  163,
  0xFE,
  0xFE,
  1,9,27,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  149,
  153,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  186,
  0xFE,
  1,7,26,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  144,
  145,
  149,
  153,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  186,
  0xFE,
  1,5,21,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  181,
  186,
  0xFE,
  1,3,18,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  0xFE,
  1,1,14,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  0xFE,
  0,255,3,
  157,
  163,
  169,
  0xFE,
  0xFE,
  0,224,0,
  1,33,6,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,31,23,
  133,
  134,
  139,
  140,
  141,
  144,
  145,
  146,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  0xFE,
  1,29,28,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  0xFE,
  1,27,24,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,25,26,
  127,
  128,
  129,
  133,
  134,
  139,
  140,
  141,
  144,
  145,
  146,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  0xFE,
  1,23,31,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  144,
  145,
  146,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  0xFE,
  1,21,29,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  144,
  145,
  146,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  175,
  176,
  177,
  181,
  186,
  0xFE,
  1,19,27,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  181,
  186,
  0xFE,
  1,17,28,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  186,
  0xFE,
  1,15,27,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  153,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  177,
  181,
  186,
  0xFE,
  1,13,25,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  186,
  0xFE,
  1,11,24,
  127,
  128,
  129,
  133,
  134,
  139,
  140,
  144,
  145,
  149,
  153,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  181,
  186,
  0xFE,
  1,9,21,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  181,
  186,
  0xFE,
  1,7,18,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  0xFE,
  1,5,15,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  157,
  158,
  159,
  163,
  164,
  169,
  170,
  175,
  0xFE,
  1,3,11,
  127,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  0xFE,
  0xFE,
  0,226,0,
  1,35,6,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,33,16,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,31,24,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,29,23,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,27,25,
  127,
  128,
  133,
  134,
  139,
  140,
  141,
  144,
  145,
  146,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  0xFE,
  1,25,29,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  0xFE,
  1,23,25,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  139,
  140,
  144,
  145,
  146,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  175,
  176,
  177,
  0xFE,
  1,21,23,
  127,
  128,
  129,
  133,
  134,
  139,
  140,
  141,
  144,
  145,
  146,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,19,25,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  186,
  0xFE,
  1,17,24,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  144,
  145,
  146,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  177,
  181,
  186,
  0xFE,
  1,15,21,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  186,
  0xFE,
  1,13,21,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  181,
  186,
  0xFE,
  1,11,19,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  181,
  186,
  0xFE,
  1,9,16,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,7,13,
  127,
  128,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  0xFE,
  1,5,6,
  139,
  157,
  158,
  163,
  164,
  169,
  0xFE,
  0xFE,
  0,228,0,
  1,37,4,
  169,
  170,
  175,
  176,
  0xFE,
  1,35,14,
  139,
  140,
  144,
  157,
  158,
  159,
  163,
  164,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,33,23,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,31,22,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,29,25,
  127,
  128,
  133,
  134,
  139,
  140,
  141,
  144,
  145,
  146,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  0xFE,
  1,27,26,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  141,
  144,
  145,
  146,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,25,23,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  144,
  145,
  146,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  177,
  0xFE,
  1,23,22,
  127,
  128,
  133,
  134,
  139,
  140,
  141,
  144,
  145,
  146,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,21,22,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,19,21,
  127,
  128,
  129,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  177,
  181,
  0xFE,
  1,17,17,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,15,15,
  127,
  128,
  133,
  134,
  139,
  144,
  145,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,13,14,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  0xFE,
  1,11,14,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  0xFE,
  1,9,10,
  127,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  0xFE,
  1,7,3,
  157,
  163,
  169,
  0xFE,
  0xFE,
  0,230,0,
  1,39,4,
  169,
  170,
  175,
  176,
  0xFE,
  1,37,11,
  157,
  158,
  159,
  163,
  164,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,35,16,
  127,
  128,
  133,
  134,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,33,20,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,31,20,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,29,23,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,27,21,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  177,
  0xFE,
  1,25,20,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,23,20,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  0xFE,
  1,21,18,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  0xFE,
  1,19,14,
  127,
  133,
  139,
  140,
  144,
  145,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,17,14,
  127,
  128,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,15,13,
  127,
  128,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  0xFE,
  1,13,11,
  127,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  0xFE,
  1,11,7,
  127,
  133,
  139,
  157,
  158,
  163,
  169,
  0xFE,
  0xFE,
  0,232,0,
  1,39,8,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,37,12,
  133,
  134,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  0xFE,
  1,35,14,
  127,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  171,
  175,
  176,
  0xFE,
  1,33,20,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,31,23,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,29,20,
  127,
  128,
  129,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  177,
  0xFE,
  1,27,19,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,25,17,
  127,
  128,
  133,
  134,
  139,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  0xFE,
  1,23,17,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,21,14,
  127,
  133,
  139,
  140,
  144,
  145,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,19,14,
  127,
  128,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,17,11,
  127,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  0xFE,
  1,15,8,
  127,
  133,
  139,
  144,
  157,
  158,
  163,
  169,
  0xFE,
  1,13,3,
  157,
  163,
  169,
  0xFE,
  0xFE,
  0,234,0,
  1,41,5,
  158,
  169,
  170,
  175,
  176,
  0xFE,
  1,39,7,
  157,
  158,
  159,
  163,
  164,
  169,
  175,
  0xFE,
  1,37,8,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,35,18,
  127,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,33,17,
  127,
  128,
  133,
  134,
  139,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  0xFE,
  1,31,18,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  0xFE,
  1,29,16,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,27,16,
  127,
  128,
  133,
  134,
  139,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,25,16,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,23,10,
  127,
  133,
  139,
  144,
  157,
  163,
  169,
  170,
  175,
  176,
  0xFE,
  1,21,10,
  127,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  175,
  0xFE,
  1,19,10,
  127,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  175,
  0xFE,
  1,17,4,
  139,
  157,
  163,
  169,
  0xFE,
  0xFE,
  0,236,0,
  1,41,4,
  157,
  158,
  163,
  164,
  0xFE,
  1,39,4,
  169,
  170,
  175,
  176,
  0xFE,
  1,37,11,
  139,
  144,
  157,
  158,
  159,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,35,15,
  127,
  128,
  133,
  134,
  144,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  0xFE,
  1,33,14,
  127,
  128,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,31,12,
  127,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,29,14,
  127,
  128,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,27,10,
  127,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  175,
  0xFE,
  1,25,8,
  127,
  133,
  139,
  144,
  157,
  163,
  169,
  175,
  0xFE,
  1,23,8,
  127,
  133,
  139,
  144,
  157,
  163,
  169,
  175,
  0xFE,
  1,21,7,
  127,
  133,
  139,
  144,
  157,
  163,
  169,
  0xFE,
  1,19,3,
  157,
  163,
  169,
  0xFE,
  0xFE,
  0,238,0,
  1,41,3,
  169,
  170,
  175,
  0xFE,
  1,39,8,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,37,9,
  127,
  133,
  157,
  158,
  163,
  164,
  169,
  175,
  176,
  0xFE,
  1,35,11,
  127,
  133,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,33,12,
  127,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,31,13,
  127,
  128,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  175,
  176,
  0xFE,
  1,29,9,
  127,
  133,
  144,
  157,
  158,
  163,
  164,
  169,
  175,
  0xFE,
  1,27,8,
  127,
  133,
  139,
  144,
  157,
  163,
  169,
  175,
  0xFE,
  1,25,6,
  127,
  133,
  139,
  157,
  163,
  169,
  0xFE,
  1,23,3,
  157,
  163,
  169,
  0xFE,
  0xFE,
  0,240,0,
  1,41,6,
  157,
  158,
  163,
  169,
  175,
  176,
  0xFE,
  1,39,7,
  133,
  157,
  158,
  163,
  164,
  169,
  175,
  0xFE,
  1,37,4,
  157,
  163,
  169,
  175,
  0xFE,
  1,35,9,
  127,
  133,
  144,
  157,
  158,
  163,
  169,
  175,
  176,
  0xFE,
  1,33,9,
  127,
  133,
  144,
  157,
  158,
  163,
  164,
  169,
  175,
  0xFE,
  1,31,6,
  127,
  133,
  144,
  157,
  163,
  175,
  0xFE,
  1,29,5,
  133,
  157,
  163,
  169,
  175,
  0xFE,
  1,27,3,
  157,
  163,
  169,
  0xFE,
  1,25,1,
  163,
  0xFE,
  0xFE,
  0,242,0,
  1,43,2,
  169,
  175,
  0xFE,
  1,41,2,
  157,
  163,
  0xFE,
  1,39,2,
  169,
  175,
  0xFE,
  1,37,4,
  157,
  163,
  169,
  175,
  0xFE,
  1,35,6,
  127,
  133,
  157,
  163,
  169,
  175,
  0xFE,
  1,33,5,
  127,
  133,
  157,
  163,
  175,
  0xFE,
  1,31,2,
  157,
  163,
  0xFE,
  1,29,2,
  157,
  163,
  0xFE,
  0xFE,
  0,244,0,
  1,39,4,
  157,
  163,
  169,
  175,
  0xFE,
  1,37,3,
  157,
  163,
  175,
  0xFE,
  1,35,2,
  157,
  163,
  0xFE,
  0xFE,
  0,246,0,
  1,39,2,
  157,
  163,
  0xFE,
  0xFE,
  0,248,139,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  20,
  21,
  25,
  26,
  27,
  34,
  35,
  36,
  38,
  39,
  44,
  45,
  50,
  51,
  56,
  57,
  65,
  66,
  67,
  71,
  72,
  77,
  78,
  83,
  84,
  89,
  90,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  118,
  122,
  123,
  124,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  149,
  150,
  151,
  153,
  154,
  155,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  172,
  173,
  175,
  176,
  177,
  178,
  179,
  181,
  182,
  183,
  184,
  186,
  187,
  188,
  189,
  191,
  192,
  193,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  208,
  209,
  210,
  211,
  213,
  214,
  215,
  217,
  218,
  219,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  235,
  239,
  240,
  241,
  242,
  245,
  246,
  249,
  250,
  0,0,15,
  0,
  3,
  4,
  7,
  8,
  13,
  14,
  157,
  163,
  169,
  170,
  175,
  176,
  181,
  186,
  0,93,1,
  3,
  0xFE,
  0,205,3,
  3,
  157,
  163,
  0xFE,
  0xFE,
  0,2,6,
  157,
  163,
  169,
  175,
  181,
  186,
  0,41,1,
  3,
  0xFE,
  0xFE,
  0,4,2,
  181,
  186,
  0xFE,
  0,32,16,
  38,
  39,
  44,
  45,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  181,
  182,
  186,
  187,
  0,211,2,
  38,
  44,
  0xFE,
  0,209,4,
  38,
  44,
  139,
  144,
  0xFE,
  0,207,8,
  38,
  44,
  127,
  133,
  157,
  163,
  169,
  175,
  0xFE,
  0,205,6,
  157,
  158,
  163,
  164,
  169,
  175,
  0xFE,
  1,53,2,
  203,
  208,
  0xFE,
  1,51,6,
  191,
  192,
  197,
  198,
  203,
  208,
  0xFE,
  1,49,4,
  163,
  191,
  197,
  198,
  0xFE,
  1,47,2,
  191,
  197,
  0xFE,
  1,45,2,
  203,
  208,
  0xFE,
  1,97,1,
  221,
  0xFE,
  1,95,2,
  221,
  227,
  0xFE,
  1,93,2,
  233,
  239,
  0xFE,
  0xFE,
  0,34,2,
  169,
  175,
  0,91,2,
  34,
  65,
  0xFE,
  0xFE,
  0,36,0,
  0,95,2,
  34,
  65,
  0xFE,
  0xFE,
  0,86,17,
  65,
  71,
  72,
  77,
  78,
  83,
  89,
  157,
  163,
  169,
  170,
  175,
  176,
  181,
  182,
  186,
  187,
  0,143,2,
  65,
  98,
  0xFE,
  0,211,2,
  71,
  77,
  0xFE,
  0,209,5,
  65,
  71,
  77,
  139,
  144,
  0xFE,
  0,207,9,
  65,
  71,
  77,
  127,
  133,
  157,
  163,
  169,
  175,
  0xFE,
  0,205,8,
  71,
  77,
  157,
  158,
  163,
  164,
  169,
  175,
  0xFE,
  1,53,2,
  203,
  208,
  0xFE,
  1,51,6,
  191,
  192,
  197,
  198,
  203,
  208,
  0xFE,
  1,49,4,
  163,
  191,
  197,
  198,
  0xFE,
  1,47,3,
  191,
  197,
  208,
  0xFE,
  1,45,2,
  203,
  208,
  0xFE,
  1,97,1,
  221,
  0xFE,
  1,95,2,
  221,
  227,
  0xFE,
  1,93,2,
  233,
  239,
  0xFE,
  0xFE,
  0,88,4,
  169,
  175,
  181,
  186,
  0,147,1,
  98,
  0xFE,
  0xFE,
  0,90,0,
  0,151,1,
  98,
  0xFE,
  0xFE,
  0,102,0,
  0,173,4,
  65,
  66,
  98,
  99,
  0xFE,
  0xFE,
  0,140,22,
  95,
  96,
  97,
  98,
  99,
  100,
  104,
  105,
  110,
  111,
  116,
  117,
  122,
  123,
  157,
  163,
  169,
  170,
  175,
  176,
  181,
  186,
  0,217,1,
  98,
  1,46,1,
  98,
  0xFE,
  0xFE,
  0,215,1,
  98,
  0xFE,
  0,213,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,211,6,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0,209,6,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0,207,7,
  95,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,205,8,
  95,
  98,
  104,
  110,
  157,
  163,
  169,
  175,
  0xFE,
  1,47,1,
  191,
  0xFE,
  0xFE,
  0,142,8,
  104,
  110,
  116,
  122,
  169,
  175,
  181,
  186,
  0xFE,
  0,215,0,
  0,144,2,
  104,
  110,
  0xFE,
  0,146,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,148,2,
  95,
  98,
  0xFE,
  0xFE,
  0,213,0,
  0,90,1,
  65,
  0xFE,
  0,142,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,144,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,146,3,
  95,
  98,
  104,
  0xFE,
  0,148,3,
  95,
  98,
  104,
  0xFE,
  0xFE,
  0,211,0,
  0,0,3,
  3,
  7,
  13,
  0xFE,
  0,2,1,
  3,
  0xFE,
  0,34,1,
  34,
  0xFE,
  0,88,4,
  62,
  65,
  71,
  77,
  0xFE,
  0,90,2,
  62,
  65,
  0xFE,
  0,142,5,
  95,
  98,
  99,
  104,
  110,
  0xFE,
  0,144,3,
  95,
  98,
  104,
  0xFE,
  0,146,4,
  95,
  98,
  104,
  110,
  0xFE,
  0xFE,
  0,209,0,
  0,0,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,2,3,
  0,
  3,
  7,
  0xFE,
  0,4,2,
  0,
  3,
  0xFE,
  0,34,2,
  31,
  34,
  0xFE,
  0,88,2,
  62,
  65,
  0xFE,
  0,90,2,
  71,
  77,
  0xFE,
  0,142,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,144,5,
  95,
  96,
  98,
  104,
  110,
  0xFE,
  0,146,2,
  104,
  110,
  0xFE,
  0xFE,
  0,207,0,
  0,0,3,
  3,
  7,
  13,
  0xFE,
  0,2,2,
  0,
  3,
  0xFE,
  0,34,1,
  34,
  0xFE,
  0,88,4,
  62,
  65,
  71,
  77,
  0xFE,
  0,90,2,
  62,
  65,
  0xFE,
  0,142,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,144,6,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0,146,4,
  95,
  98,
  104,
  110,
  0xFE,
  0xFE,
  0,212,18,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  0xFE,
  0,214,3,
  157,
  163,
  164,
  0xFE,
  1,73,0,
  1,114,1,
  197,
  0xFE,
  0xFE,
  1,71,0,
  1,112,4,
  191,
  197,
  221,
  227,
  0xFE,
  1,114,2,
  197,
  227,
  0xFE,
  1,116,2,
  191,
  197,
  0xFE,
  0xFE,
  1,69,0,
  1,108,2,
  191,
  197,
  0xFE,
  1,110,4,
  191,
  197,
  221,
  227,
  0xFE,
  1,112,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,114,7,
  191,
  197,
  203,
  221,
  227,
  233,
  239,
  0xFE,
  1,116,1,
  233,
  0xFE,
  0xFE,
  1,67,0,
  1,108,6,
  191,
  197,
  203,
  208,
  221,
  227,
  0xFE,
  1,110,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,112,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,114,4,
  203,
  208,
  233,
  239,
  0xFE,
  0xFE,
  1,65,0,
  1,104,2,
  203,
  208,
  0xFE,
  1,106,6,
  191,
  197,
  198,
  203,
  221,
  227,
  0xFE,
  1,108,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,110,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,112,7,
  191,
  197,
  203,
  208,
  227,
  233,
  239,
  0xFE,
  1,122,2,
  197,
  227,
  0xFE,
  0xFE,
  1,63,0,
  1,102,2,
  203,
  208,
  0xFE,
  1,104,10,
  191,
  192,
  197,
  198,
  203,
  208,
  221,
  227,
  228,
  233,
  0xFE,
  1,106,9,
  191,
  197,
  198,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,108,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,110,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,112,4,
  203,
  208,
  233,
  239,
  0xFE,
  1,118,3,
  197,
  198,
  227,
  0xFE,
  0xFE,
  1,61,0,
  1,100,4,
  203,
  208,
  233,
  239,
  0xFE,
  1,102,12,
  191,
  192,
  197,
  198,
  203,
  208,
  221,
  222,
  227,
  228,
  233,
  239,
  0xFE,
  1,104,15,
  191,
  192,
  197,
  198,
  203,
  208,
  209,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  0xFE,
  1,106,12,
  191,
  197,
  198,
  203,
  204,
  208,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,108,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,110,5,
  197,
  203,
  208,
  233,
  239,
  0xFE,
  0xFE,
  1,59,0,
  1,98,4,
  203,
  208,
  233,
  239,
  0xFE,
  1,100,12,
  191,
  192,
  197,
  198,
  203,
  208,
  221,
  222,
  227,
  228,
  233,
  239,
  0xFE,
  1,102,16,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  0xFE,
  1,104,14,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  221,
  227,
  228,
  233,
  234,
  239,
  240,
  0xFE,
  1,106,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,108,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,110,9,
  191,
  197,
  198,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  0xFE,
  1,57,4,
  203,
  204,
  208,
  209,
  1,94,4,
  203,
  204,
  208,
  209,
  0xFE,
  1,96,4,
  203,
  208,
  233,
  239,
  0xFE,
  1,98,12,
  191,
  192,
  197,
  198,
  203,
  208,
  221,
  222,
  227,
  228,
  233,
  239,
  0xFE,
  1,100,16,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  0xFE,
  1,102,16,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  0xFE,
  1,104,14,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  221,
  227,
  228,
  233,
  234,
  239,
  240,
  0xFE,
  1,106,15,
  163,
  191,
  192,
  197,
  198,
  203,
  208,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  0xFE,
  1,108,4,
  203,
  208,
  233,
  239,
  0xFE,
  0xFE,
  1,55,15,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  1,94,14,
  169,
  175,
  176,
  191,
  192,
  197,
  203,
  204,
  208,
  209,
  213,
  217,
  233,
  239,
  0xFE,
  1,96,16,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  221,
  222,
  227,
  228,
  233,
  239,
  0xFE,
  1,98,20,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  0xFE,
  1,100,17,
  163,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  0xFE,
  1,102,22,
  133,
  157,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  0xFE,
  1,104,7,
  191,
  197,
  203,
  208,
  221,
  233,
  239,
  0xFE,
  1,106,2,
  197,
  208,
  0xFE,
  0xFE,
  1,53,32,
  157,
  158,
  163,
  164,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  186,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  214,
  217,
  218,
  233,
  239,
  1,92,5,
  197,
  203,
  208,
  221,
  227,
  0xFE,
  1,94,30,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  181,
  186,
  191,
  192,
  193,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  221,
  222,
  227,
  228,
  233,
  239,
  240,
  0xFE,
  1,96,25,
  157,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  217,
  221,
  222,
  227,
  228,
  229,
  233,
  234,
  239,
  240,
  249,
  0xFE,
  1,98,29,
  127,
  133,
  139,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  221,
  222,
  227,
  228,
  229,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  1,100,14,
  163,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,102,11,
  163,
  169,
  175,
  191,
  197,
  203,
  208,
  209,
  227,
  233,
  239,
  0xFE,
  1,104,1,
  197,
  0xFE,
  1,106,1,
  191,
  0xFE,
  0xFE,
  1,51,44,
  133,
  144,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  186,
  187,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  214,
  217,
  218,
  227,
  233,
  234,
  239,
  240,
  1,92,17,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  221,
  222,
  227,
  228,
  233,
  239,
  0xFE,
  1,94,45,
  139,
  144,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  181,
  186,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  217,
  218,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,96,21,
  157,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  0xFE,
  1,98,20,
  133,
  157,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,100,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,102,8,
  191,
  192,
  197,
  203,
  208,
  221,
  227,
  239,
  0xFE,
  1,104,1,
  221,
  0xFE,
  0xFE,
  1,49,80,
  3,
  4,
  7,
  13,
  19,
  25,
  34,
  38,
  44,
  65,
  66,
  71,
  77,
  98,
  99,
  104,
  110,
  116,
  122,
  127,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  183,
  186,
  187,
  188,
  191,
  192,
  193,
  194,
  195,
  197,
  198,
  199,
  200,
  201,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  214,
  215,
  217,
  218,
  219,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  1,93,7,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,92,24,
  133,
  157,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  217,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  249,
  0xFE,
  1,94,33,
  157,
  158,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  181,
  186,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  217,
  218,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,96,12,
  163,
  169,
  175,
  191,
  197,
  198,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,98,15,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  208,
  213,
  217,
  221,
  227,
  233,
  239,
  0xFE,
  1,100,4,
  221,
  227,
  233,
  239,
  0xFE,
  0xFE,
  1,47,55,
  3,
  98,
  127,
  133,
  134,
  139,
  144,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  182,
  186,
  187,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  201,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  214,
  215,
  217,
  218,
  219,
  221,
  227,
  228,
  233,
  234,
  239,
  240,
  0,0,1,
  191,
  0xFE,
  1,99,7,
  169,
  175,
  191,
  203,
  204,
  208,
  209,
  0xFE,
  1,97,6,
  203,
  204,
  208,
  209,
  233,
  239,
  0xFE,
  1,93,21,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  210,
  221,
  222,
  227,
  228,
  229,
  233,
  234,
  239,
  240,
  0xFE,
  1,92,15,
  157,
  163,
  169,
  175,
  191,
  197,
  203,
  208,
  221,
  227,
  228,
  233,
  234,
  239,
  240,
  0xFE,
  1,94,29,
  157,
  163,
  164,
  169,
  175,
  181,
  186,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  213,
  214,
  217,
  218,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  0xFE,
  1,96,7,
  191,
  221,
  222,
  227,
  228,
  233,
  239,
  0xFE,
  1,98,6,
  203,
  208,
  213,
  217,
  233,
  239,
  0xFE,
  0xFE,
  1,45,56,
  3,
  34,
  98,
  133,
  139,
  144,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  182,
  186,
  187,
  191,
  192,
  193,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  214,
  215,
  216,
  217,
  218,
  219,
  221,
  222,
  227,
  228,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  249,
  1,101,4,
  203,
  208,
  233,
  239,
  0xFE,
  1,99,25,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  227,
  228,
  233,
  234,
  239,
  240,
  0xFE,
  1,97,17,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  0xFE,
  1,93,21,
  157,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  203,
  208,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  0xFE,
  1,92,20,
  157,
  163,
  169,
  175,
  191,
  197,
  203,
  208,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  239,
  245,
  246,
  249,
  250,
  0xFE,
  1,94,21,
  163,
  169,
  175,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  214,
  217,
  218,
  221,
  227,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,96,8,
  221,
  227,
  233,
  239,
  245,
  246,
  249,
  250,
  0xFE,
  1,98,2,
  203,
  208,
  0xFE,
  0xFE,
  1,44,37,
  133,
  157,
  158,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  182,
  186,
  187,
  191,
  192,
  193,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  215,
  217,
  218,
  219,
  233,
  239,
  1,105,9,
  197,
  198,
  203,
  208,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,103,17,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,101,23,
  163,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  221,
  222,
  227,
  228,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  1,99,27,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  222,
  227,
  228,
  229,
  233,
  234,
  239,
  240,
  0xFE,
  1,97,20,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  208,
  221,
  222,
  227,
  228,
  233,
  239,
  245,
  246,
  249,
  250,
  0xFE,
  1,93,12,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  245,
  246,
  249,
  250,
  0xFE,
  1,94,2,
  203,
  208,
  0xFE,
  0xFE,
  1,46,32,
  127,
  133,
  134,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  181,
  182,
  186,
  187,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  217,
  218,
  233,
  239,
  1,109,2,
  233,
  239,
  0xFE,
  1,107,5,
  197,
  203,
  208,
  233,
  239,
  0xFE,
  1,105,14,
  169,
  175,
  191,
  197,
  203,
  204,
  208,
  209,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,103,18,
  163,
  169,
  175,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  0xFE,
  1,101,15,
  191,
  192,
  197,
  198,
  203,
  208,
  209,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  0xFE,
  1,99,15,
  163,
  191,
  192,
  197,
  198,
  203,
  208,
  213,
  217,
  221,
  222,
  227,
  228,
  233,
  239,
  0xFE,
  1,97,10,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  245,
  249,
  0xFE,
  1,93,4,
  221,
  227,
  233,
  239,
  0xFE,
  0xFE,
  1,48,30,
  7,
  13,
  104,
  110,
  127,
  133,
  134,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  181,
  186,
  197,
  203,
  204,
  208,
  209,
  213,
  217,
  233,
  234,
  239,
  240,
  1,109,4,
  203,
  208,
  233,
  239,
  0xFE,
  1,107,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,105,14,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  221,
  227,
  228,
  233,
  234,
  239,
  240,
  0xFE,
  1,103,12,
  191,
  192,
  197,
  198,
  203,
  208,
  221,
  222,
  227,
  228,
  233,
  239,
  0xFE,
  1,101,10,
  191,
  192,
  197,
  198,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,99,10,
  191,
  197,
  203,
  208,
  213,
  217,
  221,
  227,
  233,
  239,
  0xFE,
  1,97,6,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,93,4,
  221,
  222,
  227,
  228,
  0xFE,
  0xFE,
  1,50,0,
  1,109,7,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,107,9,
  191,
  197,
  198,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,105,11,
  191,
  192,
  197,
  198,
  203,
  208,
  221,
  227,
  228,
  233,
  239,
  0xFE,
  1,103,9,
  191,
  197,
  198,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,101,6,
  191,
  197,
  203,
  208,
  221,
  227,
  0xFE,
  1,99,1,
  203,
  0xFE,
  0xFE,
  1,52,0,
  1,111,5,
  197,
  203,
  208,
  233,
  239,
  0xFE,
  1,109,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,107,9,
  191,
  197,
  198,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,105,7,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  0xFE,
  1,103,4,
  191,
  197,
  221,
  227,
  0xFE,
  0xFE,
  1,54,0,
  1,113,3,
  203,
  233,
  239,
  0xFE,
  1,111,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,109,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,107,4,
  191,
  197,
  221,
  227,
  0xFE,
  1,105,2,
  191,
  221,
  0xFE,
  0xFE,
  1,56,0,
  1,113,6,
  191,
  197,
  203,
  227,
  233,
  239,
  0xFE,
  1,111,6,
  191,
  197,
  208,
  221,
  227,
  239,
  0xFE,
  1,109,4,
  191,
  197,
  221,
  227,
  0xFE,
  1,107,2,
  191,
  221,
  0xFE,
  0xFE,
  1,58,0,
  1,113,3,
  191,
  197,
  227,
  0xFE,
  1,111,4,
  191,
  197,
  221,
  227,
  0xFE,
  1,109,2,
  191,
  221,
  0xFE,
  0xFE,
  1,60,0,
  1,113,1,
  191,
  0xFE,
  1,111,2,
  191,
  221,
  0xFE,
  0xFE,
  1,103,7,
  191,
  197,
  203,
  208,
  227,
  233,
  239,
  0xFE,
  1,101,17,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  221,
  222,
  227,
  228,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  1,99,36,
  157,
  163,
  164,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  186,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  235,
  236,
  239,
  240,
  241,
  242,
  245,
  249,
  0xFE,
  1,97,45,
  133,
  139,
  144,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  182,
  186,
  187,
  191,
  192,
  197,
  198,
  203,
  208,
  209,
  221,
  222,
  223,
  227,
  228,
  229,
  230,
  233,
  234,
  235,
  236,
  239,
  240,
  241,
  242,
  245,
  246,
  249,
  250,
  0xFE,
  1,95,46,
  133,
  157,
  158,
  159,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  186,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  214,
  217,
  218,
  221,
  222,
  223,
  224,
  227,
  228,
  229,
  230,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  246,
  249,
  250,
  0,0,1,
  221,
  0xFE,
  0,140,1,
  221,
  0xFE,
  1,53,2,
  203,
  208,
  0xFE,
  1,51,7,
  169,
  175,
  191,
  203,
  204,
  208,
  209,
  0xFE,
  1,49,21,
  157,
  163,
  169,
  170,
  175,
  176,
  181,
  186,
  191,
  192,
  193,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  0xFE,
  1,47,41,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  186,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  217,
  218,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  245,
  0xFE,
  1,45,38,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  181,
  186,
  191,
  192,
  193,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  217,
  218,
  221,
  222,
  227,
  228,
  229,
  233,
  234,
  239,
  240,
  0xFE,
  1,44,21,
  157,
  163,
  175,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  214,
  217,
  218,
  221,
  227,
  233,
  234,
  239,
  245,
  249,
  0xFE,
  1,46,17,
  157,
  163,
  169,
  175,
  191,
  197,
  198,
  203,
  204,
  208,
  213,
  214,
  217,
  218,
  221,
  227,
  233,
  0xFE,
  1,48,10,
  169,
  175,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,50,2,
  203,
  208,
  0xFE,
  0xFE,
  1,93,35,
  3,
  133,
  157,
  163,
  164,
  169,
  170,
  175,
  176,
  181,
  186,
  187,
  191,
  197,
  208,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  246,
  247,
  248,
  249,
  250,
  251,
  252,
  0xFE,
  1,92,28,
  157,
  163,
  164,
  169,
  170,
  175,
  176,
  181,
  186,
  221,
  222,
  223,
  227,
  228,
  229,
  230,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  246,
  247,
  249,
  250,
  251,
  0xFE,
  1,94,26,
  163,
  169,
  170,
  175,
  176,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  214,
  217,
  218,
  221,
  227,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  249,
  0xFE,
  1,96,2,
  233,
  239,
  0xFE,
  0xFE,
  0,250,140,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  20,
  21,
  25,
  26,
  27,
  34,
  35,
  36,
  38,
  39,
  44,
  45,
  50,
  51,
  56,
  57,
  65,
  66,
  67,
  71,
  72,
  77,
  78,
  83,
  84,
  89,
  90,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  118,
  122,
  123,
  124,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  137,
  139,
  140,
  141,
  142,
  143,
  144,
  145,
  146,
  147,
  148,
  149,
  150,
  151,
  153,
  154,
  155,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  172,
  173,
  175,
  176,
  177,
  178,
  179,
  181,
  182,
  183,
  184,
  186,
  187,
  188,
  191,
  192,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  214,
  215,
  217,
  218,
  219,
  221,
  222,
  227,
  228,
  229,
  233,
  234,
  235,
  236,
  239,
  240,
  241,
  242,
  245,
  246,
  249,
  250,
  0,0,8,
  0,
  3,
  7,
  13,
  19,
  25,
  169,
  175,
  0,35,1,
  3,
  0xFE,
  0,93,1,
  3,
  0xFE,
  0xFE,
  0,2,0,
  0,41,1,
  3,
  0xFE,
  0xFE,
  0,32,6,
  169,
  170,
  175,
  176,
  181,
  186,
  0,205,2,
  157,
  163,
  0xFE,
  1,95,1,
  221,
  0xFE,
  0xFE,
  0,36,0,
  0,95,2,
  34,
  65,
  0xFE,
  0xFE,
  0,86,7,
  71,
  77,
  157,
  169,
  175,
  181,
  186,
  0,143,10,
  65,
  66,
  71,
  77,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0,205,2,
  157,
  163,
  0xFE,
  1,95,1,
  221,
  0xFE,
  0xFE,
  0,88,0,
  0,147,5,
  65,
  95,
  98,
  104,
  110,
  0xFE,
  0xFE,
  0,90,0,
  0,151,1,
  98,
  0xFE,
  0xFE,
  0,92,0,
  0,155,1,
  98,
  0xFE,
  0xFE,
  0,100,0,
  0,169,0,
  0,205,1,
  98,
  0xFE,
  0xFE,
  0xFE,
  0,102,0,
  0,173,2,
  98,
  99,
  0xFE,
  0xFE,
  0,104,0,
  0,177,1,
  98,
  0xFE,
  0xFE,
  0,140,10,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  0,205,3,
  98,
  104,
  110,
  0xFE,
  0xFE,
  0,221,0,
  0,148,2,
  104,
  110,
  0xFE,
  0,150,2,
  95,
  98,
  0xFE,
  0,152,1,
  98,
  0xFE,
  0xFE,
  0,219,0,
  0,144,2,
  104,
  110,
  0xFE,
  0,146,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,148,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,150,3,
  95,
  98,
  104,
  0xFE,
  0,152,2,
  95,
  98,
  0xFE,
  0xFE,
  0,217,0,
  0,88,2,
  71,
  77,
  0xFE,
  0,90,3,
  65,
  71,
  77,
  0xFE,
  0,92,2,
  62,
  65,
  0xFE,
  0,140,9,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  122,
  0xFE,
  0,142,2,
  104,
  110,
  0xFE,
  0,144,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,146,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,148,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,150,4,
  95,
  98,
  104,
  110,
  0xFE,
  0xFE,
  0,215,0,
  0,0,3,
  3,
  7,
  13,
  0xFE,
  0,2,1,
  3,
  0xFE,
  0,4,1,
  3,
  0xFE,
  0,32,2,
  38,
  44,
  0xFE,
  0,34,3,
  34,
  38,
  44,
  0xFE,
  0,36,2,
  31,
  34,
  0xFE,
  0,86,2,
  71,
  77,
  0xFE,
  0,88,4,
  62,
  65,
  71,
  77,
  0xFE,
  0,90,4,
  62,
  65,
  71,
  77,
  0xFE,
  0,92,2,
  62,
  65,
  0xFE,
  0,140,8,
  95,
  96,
  98,
  99,
  104,
  110,
  116,
  122,
  0xFE,
  0,142,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,144,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,146,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,148,5,
  95,
  96,
  98,
  104,
  110,
  0xFE,
  0,150,4,
  95,
  98,
  104,
  110,
  0xFE,
  0xFE,
  0,213,0,
  0,0,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,2,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,4,3,
  0,
  3,
  7,
  0xFE,
  0,32,2,
  38,
  44,
  0xFE,
  0,34,4,
  31,
  34,
  38,
  44,
  0xFE,
  0,36,3,
  31,
  34,
  38,
  0xFE,
  0,86,6,
  62,
  65,
  71,
  72,
  77,
  78,
  0xFE,
  0,88,6,
  62,
  63,
  65,
  66,
  71,
  77,
  0xFE,
  0,90,4,
  62,
  65,
  71,
  77,
  0xFE,
  0,92,2,
  71,
  77,
  0xFE,
  0,140,10,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  0xFE,
  0,142,5,
  95,
  98,
  99,
  104,
  110,
  0xFE,
  0,144,7,
  95,
  96,
  98,
  104,
  105,
  110,
  111,
  0xFE,
  0,146,6,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0,148,4,
  95,
  98,
  104,
  110,
  0xFE,
  0xFE,
  0,211,0,
  0,0,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,2,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,4,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,6,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,32,6,
  34,
  38,
  44,
  127,
  139,
  144,
  0xFE,
  0,34,4,
  31,
  34,
  38,
  44,
  0xFE,
  0,36,3,
  34,
  38,
  44,
  0xFE,
  0,86,8,
  62,
  65,
  66,
  71,
  77,
  127,
  139,
  144,
  0xFE,
  0,88,4,
  62,
  65,
  71,
  77,
  0xFE,
  0,90,4,
  62,
  65,
  71,
  77,
  0xFE,
  0,92,0,
  0,153,1,
  98,
  0xFE,
  0xFE,
  0,140,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,142,9,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  122,
  0xFE,
  0,144,6,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0,146,5,
  95,
  96,
  98,
  104,
  110,
  0xFE,
  0,148,2,
  104,
  110,
  0xFE,
  0xFE,
  0,209,0,
  0,0,7,
  0,
  3,
  4,
  7,
  8,
  13,
  14,
  0xFE,
  0,2,4,
  0,
  3,
  4,
  7,
  0,41,1,
  3,
  0xFE,
  0xFE,
  0,4,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,32,11,
  31,
  34,
  38,
  39,
  44,
  45,
  127,
  139,
  144,
  157,
  163,
  0xFE,
  0,34,4,
  31,
  32,
  34,
  35,
  0xFE,
  0,86,11,
  62,
  65,
  71,
  72,
  77,
  78,
  127,
  139,
  144,
  157,
  163,
  0xFE,
  0,88,4,
  62,
  63,
  65,
  66,
  0xFE,
  0,90,2,
  71,
  77,
  0xFE,
  0,140,9,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  122,
  0xFE,
  0,142,6,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0,144,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,146,2,
  104,
  110,
  0xFE,
  0xFE,
  0,207,0,
  0,0,9,
  0,
  3,
  4,
  7,
  8,
  13,
  14,
  169,
  175,
  0xFE,
  0,2,4,
  0,
  3,
  4,
  7,
  0xFE,
  0,4,2,
  0,
  3,
  0xFE,
  0,32,14,
  38,
  39,
  44,
  45,
  127,
  133,
  139,
  144,
  157,
  163,
  169,
  170,
  175,
  176,
  0xFE,
  0,34,4,
  31,
  32,
  34,
  35,
  0xFE,
  0,86,17,
  62,
  65,
  71,
  72,
  77,
  78,
  127,
  133,
  139,
  144,
  157,
  158,
  163,
  169,
  170,
  175,
  176,
  0xFE,
  0,88,6,
  62,
  63,
  65,
  66,
  71,
  77,
  0xFE,
  0,90,6,
  62,
  63,
  65,
  66,
  71,
  77,
  0xFE,
  0,140,10,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  169,
  175,
  0xFE,
  0,142,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,144,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,146,5,
  95,
  96,
  98,
  104,
  110,
  0xFE,
  0xFE,
  0,212,5,
  157,
  158,
  163,
  164,
  169,
  0xFE,
  1,67,0,
  1,114,2,
  197,
  227,
  0xFE,
  0xFE,
  1,65,0,
  1,104,2,
  197,
  227,
  0xFE,
  1,110,6,
  191,
  197,
  221,
  227,
  233,
  239,
  0xFE,
  0xFE,
  1,63,0,
  1,100,1,
  197,
  0xFE,
  1,102,3,
  191,
  197,
  227,
  0xFE,
  1,104,6,
  191,
  197,
  203,
  227,
  233,
  239,
  0xFE,
  1,106,7,
  191,
  197,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,108,7,
  191,
  197,
  203,
  208,
  227,
  233,
  239,
  0xFE,
  0xFE,
  1,61,0,
  1,94,2,
  203,
  208,
  0xFE,
  1,98,4,
  197,
  203,
  208,
  227,
  0xFE,
  1,100,7,
  191,
  197,
  203,
  208,
  221,
  227,
  239,
  0xFE,
  1,102,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,104,9,
  163,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  0xFE,
  1,59,6,
  169,
  175,
  203,
  204,
  208,
  209,
  1,94,7,
  169,
  175,
  191,
  203,
  204,
  208,
  209,
  0xFE,
  1,96,7,
  197,
  198,
  203,
  208,
  221,
  227,
  228,
  0xFE,
  1,98,17,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  221,
  227,
  228,
  233,
  239,
  0xFE,
  1,100,9,
  163,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,102,6,
  191,
  197,
  203,
  208,
  221,
  239,
  0xFE,
  0xFE,
  1,57,23,
  157,
  163,
  169,
  170,
  175,
  176,
  181,
  186,
  191,
  192,
  193,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  217,
  218,
  1,92,2,
  203,
  208,
  0xFE,
  1,94,25,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  221,
  227,
  228,
  0xFE,
  1,96,17,
  133,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  221,
  227,
  233,
  239,
  0xFE,
  1,98,16,
  157,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  221,
  233,
  239,
  0xFE,
  1,100,5,
  191,
  197,
  203,
  208,
  221,
  0xFE,
  1,102,5,
  163,
  191,
  197,
  203,
  208,
  0xFE,
  0xFE,
  1,55,40,
  157,
  158,
  159,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  182,
  186,
  187,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  214,
  217,
  218,
  227,
  233,
  239,
  1,95,4,
  203,
  204,
  208,
  209,
  0xFE,
  1,92,11,
  163,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  221,
  228,
  0xFE,
  1,94,27,
  157,
  158,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  181,
  186,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  221,
  227,
  0xFE,
  1,96,11,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  208,
  221,
  0xFE,
  1,98,13,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  221,
  0xFE,
  1,100,5,
  163,
  191,
  197,
  203,
  208,
  0xFE,
  1,102,8,
  157,
  163,
  169,
  175,
  191,
  197,
  203,
  208,
  0xFE,
  0xFE,
  1,53,41,
  133,
  157,
  158,
  159,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  186,
  187,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  201,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  214,
  217,
  218,
  1,95,21,
  157,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  0xFE,
  1,93,7,
  191,
  197,
  203,
  208,
  227,
  233,
  239,
  0xFE,
  1,92,11,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,94,27,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  181,
  186,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  217,
  218,
  221,
  0xFE,
  1,96,11,
  157,
  163,
  169,
  175,
  191,
  197,
  198,
  203,
  208,
  221,
  227,
  0xFE,
  1,98,13,
  139,
  144,
  157,
  163,
  169,
  175,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,100,2,
  191,
  197,
  0xFE,
  1,102,2,
  191,
  197,
  0xFE,
  0xFE,
  1,51,46,
  133,
  139,
  144,
  157,
  158,
  159,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  186,
  187,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  201,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  214,
  215,
  217,
  218,
  219,
  233,
  0,32,1,
  191,
  0xFE,
  0,86,1,
  191,
  0xFE,
  1,99,5,
  191,
  197,
  198,
  203,
  208,
  0xFE,
  1,95,25,
  157,
  158,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  181,
  186,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  0xFE,
  1,93,5,
  191,
  197,
  203,
  208,
  221,
  0xFE,
  1,92,11,
  157,
  163,
  169,
  175,
  191,
  197,
  198,
  203,
  208,
  221,
  227,
  0xFE,
  1,94,28,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  233,
  239,
  0xFE,
  1,96,7,
  157,
  163,
  191,
  197,
  203,
  208,
  221,
  0xFE,
  1,98,10,
  157,
  163,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  0xFE,
  1,49,54,
  3,
  7,
  34,
  65,
  98,
  127,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  182,
  186,
  187,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  215,
  217,
  218,
  219,
  233,
  234,
  239,
  240,
  0,32,1,
  191,
  0xFE,
  0,86,1,
  191,
  0xFE,
  1,99,11,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,97,2,
  203,
  208,
  0xFE,
  1,95,30,
  157,
  158,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  181,
  186,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  221,
  233,
  234,
  239,
  240,
  0xFE,
  1,93,12,
  163,
  169,
  175,
  191,
  197,
  203,
  208,
  221,
  222,
  223,
  227,
  228,
  0xFE,
  1,92,7,
  157,
  163,
  191,
  197,
  203,
  208,
  227,
  0xFE,
  1,94,26,
  157,
  158,
  163,
  164,
  169,
  175,
  176,
  181,
  186,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  221,
  227,
  233,
  239,
  0xFE,
  1,96,9,
  157,
  163,
  191,
  197,
  203,
  208,
  221,
  227,
  239,
  0xFE,
  1,98,7,
  191,
  192,
  197,
  198,
  203,
  208,
  221,
  0xFE,
  1,100,1,
  221,
  0xFE,
  0xFE,
  1,47,50,
  3,
  34,
  65,
  98,
  127,
  133,
  139,
  144,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  182,
  186,
  187,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  214,
  215,
  217,
  218,
  219,
  221,
  227,
  233,
  234,
  239,
  240,
  0,32,1,
  191,
  0xFE,
  0,86,1,
  191,
  0xFE,
  1,103,2,
  203,
  208,
  0xFE,
  1,99,12,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,97,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,95,26,
  157,
  158,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  233,
  239,
  240,
  0xFE,
  1,93,14,
  157,
  163,
  169,
  175,
  191,
  197,
  203,
  208,
  221,
  222,
  227,
  228,
  233,
  239,
  0xFE,
  1,92,10,
  157,
  163,
  191,
  197,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,94,22,
  157,
  163,
  169,
  175,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  213,
  217,
  221,
  222,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,96,6,
  221,
  222,
  227,
  228,
  233,
  239,
  0xFE,
  0xFE,
  1,45,46,
  133,
  139,
  144,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  181,
  182,
  186,
  187,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  217,
  218,
  219,
  221,
  222,
  227,
  228,
  233,
  234,
  235,
  239,
  240,
  241,
  1,103,3,
  197,
  203,
  208,
  0xFE,
  1,99,13,
  157,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,97,6,
  191,
  197,
  203,
  208,
  233,
  239,
  0xFE,
  1,95,31,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  181,
  186,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  0xFE,
  1,93,17,
  157,
  163,
  164,
  169,
  175,
  191,
  197,
  203,
  208,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  0xFE,
  1,92,9,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  239,
  240,
  0xFE,
  1,94,4,
  203,
  208,
  233,
  239,
  0xFE,
  0xFE,
  1,44,20,
  163,
  169,
  175,
  181,
  186,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  217,
  218,
  1,105,2,
  233,
  239,
  0xFE,
  1,103,13,
  163,
  169,
  175,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,101,13,
  157,
  163,
  164,
  169,
  175,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  0xFE,
  1,99,21,
  157,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  208,
  221,
  222,
  227,
  228,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  1,97,14,
  157,
  163,
  191,
  197,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  239,
  240,
  0xFE,
  1,95,17,
  133,
  163,
  169,
  175,
  191,
  197,
  198,
  199,
  203,
  208,
  213,
  221,
  222,
  227,
  228,
  233,
  239,
  0xFE,
  1,93,2,
  233,
  239,
  0xFE,
  0xFE,
  1,46,7,
  163,
  191,
  197,
  203,
  208,
  213,
  217,
  1,105,5,
  157,
  175,
  227,
  233,
  239,
  0xFE,
  1,103,7,
  157,
  163,
  197,
  221,
  227,
  233,
  239,
  0xFE,
  1,101,6,
  221,
  222,
  227,
  228,
  233,
  239,
  0xFE,
  1,99,6,
  191,
  197,
  198,
  221,
  227,
  228,
  0xFE,
  1,97,6,
  191,
  197,
  221,
  227,
  233,
  239,
  0xFE,
  1,95,7,
  191,
  203,
  204,
  208,
  209,
  213,
  217,
  0xFE,
  0xFE,
  1,48,4,
  157,
  163,
  169,
  175,
  1,103,1,
  227,
  0xFE,
  1,99,2,
  191,
  221,
  0xFE,
  0xFE,
  1,103,9,
  163,
  191,
  197,
  203,
  208,
  233,
  234,
  239,
  240,
  0xFE,
  1,101,15,
  163,
  169,
  170,
  175,
  176,
  221,
  222,
  227,
  228,
  233,
  234,
  235,
  239,
  240,
  241,
  1,50,1,
  221,
  0xFE,
  0xFE,
  1,99,35,
  157,
  158,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  186,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  249,
  0xFE,
  1,97,28,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  191,
  197,
  203,
  208,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  249,
  0xFE,
  1,95,47,
  133,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  182,
  186,
  187,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  235,
  236,
  239,
  240,
  241,
  242,
  245,
  246,
  249,
  250,
  0xFE,
  1,93,31,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  197,
  203,
  208,
  221,
  222,
  223,
  224,
  225,
  227,
  228,
  229,
  230,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  246,
  247,
  249,
  250,
  251,
  0xFE,
  1,92,16,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  239,
  240,
  245,
  246,
  247,
  249,
  250,
  251,
  0xFE,
  1,94,8,
  203,
  208,
  213,
  217,
  233,
  234,
  239,
  240,
  0xFE,
  1,108,0,
  1,71,2,
  191,
  197,
  0xFE,
  1,69,3,
  197,
  203,
  208,
  0xFE,
  1,67,2,
  203,
  208,
  0xFE,
  1,65,2,
  197,
  227,
  0xFE,
  0xFE,
  0xFE,
  0,252,122,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  20,
  25,
  26,
  27,
  34,
  35,
  38,
  39,
  44,
  45,
  50,
  56,
  65,
  66,
  71,
  72,
  73,
  77,
  78,
  79,
  83,
  84,
  89,
  90,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  127,
  128,
  129,
  130,
  131,
  133,
  134,
  135,
  136,
  137,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  149,
  150,
  151,
  153,
  154,
  155,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  179,
  181,
  182,
  183,
  186,
  187,
  188,
  191,
  197,
  198,
  199,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  217,
  221,
  227,
  228,
  229,
  233,
  234,
  235,
  236,
  239,
  240,
  241,
  242,
  0,0,3,
  3,
  7,
  13,
  0,155,1,
  98,
  0xFE,
  0xFE,
  0,2,0,
  0,41,1,
  3,
  0xFE,
  0xFE,
  0,36,0,
  0,95,4,
  34,
  35,
  65,
  66,
  0xFE,
  0xFE,
  0,88,0,
  0,147,4,
  65,
  95,
  98,
  99,
  0xFE,
  0xFE,
  0,90,0,
  0,151,1,
  98,
  0xFE,
  0xFE,
  0,104,0,
  0,177,13,
  65,
  66,
  67,
  95,
  96,
  98,
  99,
  100,
  101,
  104,
  105,
  110,
  111,
  0xFE,
  0xFE,
  0,140,6,
  104,
  105,
  110,
  111,
  116,
  122,
  0xFE,
  0,219,0,
  0,140,1,
  98,
  0xFE,
  0,144,2,
  104,
  110,
  0xFE,
  0,146,3,
  98,
  104,
  110,
  0xFE,
  0,148,2,
  95,
  98,
  0xFE,
  0,150,2,
  95,
  98,
  0xFE,
  0xFE,
  0,217,0,
  0,86,2,
  71,
  77,
  0xFE,
  0,90,1,
  65,
  0xFE,
  0,140,5,
  95,
  98,
  99,
  104,
  110,
  0xFE,
  0,142,3,
  98,
  104,
  110,
  0xFE,
  0,144,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,146,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,148,2,
  95,
  98,
  0xFE,
  0xFE,
  0,215,0,
  0,0,2,
  3,
  7,
  0xFE,
  0,2,1,
  3,
  0xFE,
  0,32,2,
  38,
  44,
  0xFE,
  0,34,1,
  34,
  0xFE,
  0,86,2,
  71,
  77,
  0xFE,
  0,88,4,
  62,
  65,
  71,
  77,
  0,145,2,
  65,
  98,
  0xFE,
  0xFE,
  0,90,2,
  62,
  65,
  0xFE,
  0,92,1,
  65,
  0xFE,
  0,140,5,
  95,
  98,
  99,
  104,
  110,
  0xFE,
  0,142,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,144,7,
  95,
  96,
  98,
  104,
  105,
  110,
  111,
  0xFE,
  0,146,5,
  95,
  96,
  98,
  99,
  104,
  0xFE,
  0,148,3,
  95,
  104,
  110,
  0xFE,
  0xFE,
  0,213,0,
  0,0,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,2,3,
  0,
  3,
  13,
  0xFE,
  0,32,2,
  38,
  44,
  0xFE,
  0,34,4,
  31,
  34,
  38,
  44,
  0xFE,
  0,36,2,
  31,
  34,
  0xFE,
  0,86,3,
  65,
  71,
  77,
  0xFE,
  0,88,4,
  62,
  65,
  71,
  77,
  0xFE,
  0,90,4,
  62,
  63,
  65,
  77,
  0xFE,
  0,140,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,142,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,144,6,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0,146,4,
  95,
  98,
  104,
  110,
  0xFE,
  0xFE,
  0,211,0,
  0,0,7,
  0,
  3,
  4,
  7,
  8,
  13,
  14,
  0xFE,
  0,2,4,
  0,
  3,
  4,
  7,
  0xFE,
  0,4,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,32,8,
  38,
  39,
  44,
  45,
  127,
  139,
  144,
  157,
  0xFE,
  0,34,4,
  31,
  32,
  34,
  35,
  0xFE,
  0,36,2,
  38,
  44,
  0xFE,
  0,86,10,
  62,
  65,
  71,
  72,
  77,
  78,
  127,
  139,
  144,
  157,
  0xFE,
  0,88,4,
  62,
  63,
  65,
  66,
  0xFE,
  0,90,2,
  71,
  77,
  0xFE,
  0,140,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,142,6,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0,144,7,
  95,
  96,
  98,
  104,
  105,
  110,
  111,
  0xFE,
  0xFE,
  0,209,0,
  0,0,3,
  0,
  3,
  7,
  0xFE,
  0,2,1,
  3,
  0xFE,
  0,32,4,
  38,
  127,
  133,
  157,
  0xFE,
  0,34,1,
  34,
  0xFE,
  0,86,9,
  62,
  65,
  71,
  77,
  127,
  133,
  157,
  169,
  175,
  0xFE,
  0,88,2,
  62,
  65,
  0xFE,
  0,96,0,
  0,161,1,
  98,
  0xFE,
  0xFE,
  0,140,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,142,4,
  95,
  96,
  98,
  99,
  0xFE,
  0xFE,
  0,207,0,
  0,0,14,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  157,
  163,
  169,
  170,
  175,
  176,
  0xFE,
  0,2,8,
  0,
  1,
  3,
  4,
  7,
  157,
  169,
  175,
  0xFE,
  0,4,3,
  0,
  3,
  13,
  0xFE,
  0,32,12,
  38,
  39,
  44,
  45,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  0,34,4,
  31,
  32,
  34,
  35,
  0xFE,
  0,86,13,
  65,
  71,
  72,
  77,
  78,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  0,88,6,
  62,
  63,
  65,
  66,
  169,
  175,
  0xFE,
  0,90,2,
  71,
  77,
  0xFE,
  0,140,14,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  157,
  163,
  169,
  170,
  175,
  176,
  0xFE,
  0,142,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,144,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,146,2,
  104,
  110,
  0xFE,
  0xFE,
  0,205,0,
  0,32,1,
  157,
  0xFE,
  0,86,1,
  157,
  0xFE,
  0xFE,
  1,71,0,
  1,112,1,
  227,
  0xFE,
  0xFE,
  1,67,0,
  1,104,3,
  191,
  197,
  227,
  0xFE,
  1,106,1,
  197,
  0xFE,
  0xFE,
  1,65,0,
  1,98,1,
  197,
  0xFE,
  0xFE,
  1,63,2,
  203,
  208,
  1,94,7,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,96,2,
  191,
  197,
  0xFE,
  1,98,4,
  191,
  197,
  203,
  208,
  0xFE,
  0xFE,
  1,61,19,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  1,92,1,
  197,
  0xFE,
  1,94,13,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  0xFE,
  1,96,3,
  197,
  203,
  208,
  0xFE,
  1,98,6,
  169,
  175,
  191,
  197,
  203,
  208,
  0xFE,
  1,102,2,
  203,
  208,
  0xFE,
  0xFE,
  1,59,30,
  157,
  158,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  181,
  186,
  191,
  192,
  193,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  214,
  217,
  218,
  1,95,3,
  197,
  203,
  208,
  0xFE,
  1,92,2,
  203,
  208,
  0xFE,
  1,94,19,
  157,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  0xFE,
  1,96,6,
  169,
  175,
  191,
  197,
  203,
  208,
  0xFE,
  1,98,9,
  157,
  163,
  169,
  175,
  191,
  197,
  198,
  203,
  208,
  0xFE,
  1,100,2,
  203,
  208,
  0xFE,
  1,102,3,
  191,
  197,
  208,
  0xFE,
  0xFE,
  1,57,32,
  157,
  158,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  186,
  191,
  192,
  193,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  214,
  217,
  218,
  1,95,11,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,92,3,
  197,
  203,
  208,
  0xFE,
  1,94,21,
  157,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  0xFE,
  1,96,9,
  157,
  163,
  169,
  175,
  191,
  197,
  198,
  203,
  208,
  0xFE,
  1,98,8,
  157,
  163,
  169,
  175,
  191,
  197,
  203,
  208,
  0xFE,
  1,100,2,
  191,
  197,
  0xFE,
  1,102,3,
  191,
  197,
  208,
  0xFE,
  0xFE,
  1,55,35,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  182,
  186,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  214,
  217,
  218,
  1,99,2,
  203,
  208,
  0xFE,
  1,95,17,
  163,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  0xFE,
  1,92,7,
  163,
  169,
  175,
  191,
  197,
  203,
  208,
  0xFE,
  1,94,23,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  186,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  0xFE,
  1,96,6,
  157,
  163,
  191,
  197,
  203,
  208,
  0xFE,
  1,98,8,
  157,
  163,
  169,
  175,
  191,
  197,
  203,
  208,
  0xFE,
  1,102,6,
  133,
  157,
  163,
  169,
  175,
  197,
  0xFE,
  0xFE,
  1,53,36,
  157,
  158,
  159,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  182,
  186,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  214,
  217,
  218,
  1,99,3,
  197,
  203,
  208,
  0xFE,
  1,95,21,
  157,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  0xFE,
  1,93,2,
  203,
  208,
  0xFE,
  1,92,8,
  157,
  163,
  169,
  175,
  191,
  197,
  203,
  208,
  0xFE,
  1,94,22,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  0xFE,
  1,96,6,
  157,
  163,
  191,
  197,
  203,
  208,
  0xFE,
  1,98,12,
  133,
  139,
  144,
  157,
  163,
  169,
  175,
  191,
  197,
  198,
  203,
  208,
  0xFE,
  0xFE,
  1,51,37,
  133,
  157,
  158,
  159,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  186,
  187,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  217,
  218,
  233,
  239,
  0,32,1,
  191,
  0xFE,
  0,86,1,
  191,
  0xFE,
  1,99,8,
  169,
  175,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,95,25,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  181,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  233,
  239,
  0xFE,
  1,93,6,
  191,
  197,
  203,
  208,
  221,
  227,
  0xFE,
  1,92,6,
  157,
  163,
  191,
  197,
  221,
  227,
  0xFE,
  1,94,25,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  181,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  213,
  217,
  233,
  239,
  0xFE,
  1,96,2,
  191,
  197,
  0xFE,
  1,98,6,
  169,
  175,
  191,
  197,
  203,
  208,
  0xFE,
  0xFE,
  1,49,51,
  7,
  13,
  104,
  110,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  146,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  182,
  186,
  187,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  211,
  213,
  214,
  217,
  218,
  219,
  233,
  234,
  239,
  240,
  1,99,12,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,97,4,
  203,
  208,
  233,
  239,
  0xFE,
  1,95,27,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  210,
  213,
  217,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,93,9,
  163,
  191,
  197,
  203,
  208,
  221,
  222,
  227,
  228,
  0xFE,
  1,92,2,
  233,
  239,
  0xFE,
  1,94,18,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  227,
  0xFE,
  1,98,1,
  191,
  0xFE,
  0xFE,
  1,47,43,
  127,
  133,
  139,
  144,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  182,
  186,
  187,
  191,
  192,
  193,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  214,
  217,
  218,
  221,
  227,
  233,
  239,
  1,103,2,
  203,
  208,
  0xFE,
  1,99,12,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,97,6,
  191,
  197,
  203,
  208,
  221,
  227,
  0xFE,
  1,95,23,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  181,
  186,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  213,
  217,
  233,
  239,
  0xFE,
  1,93,7,
  163,
  191,
  197,
  221,
  227,
  233,
  239,
  0xFE,
  1,92,3,
  163,
  221,
  227,
  0xFE,
  1,94,8,
  191,
  192,
  193,
  197,
  198,
  203,
  208,
  213,
  0xFE,
  1,96,1,
  221,
  0xFE,
  0xFE,
  1,45,38,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  181,
  182,
  186,
  187,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  217,
  218,
  221,
  222,
  227,
  233,
  234,
  239,
  240,
  1,103,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,99,12,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  208,
  233,
  239,
  0xFE,
  1,97,10,
  169,
  175,
  191,
  197,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,95,20,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  221,
  222,
  227,
  228,
  233,
  239,
  0xFE,
  1,93,9,
  157,
  163,
  169,
  175,
  221,
  227,
  228,
  233,
  239,
  0xFE,
  1,92,4,
  221,
  222,
  227,
  228,
  0xFE,
  0xFE,
  1,44,9,
  191,
  203,
  204,
  208,
  209,
  213,
  214,
  217,
  218,
  1,105,4,
  233,
  234,
  239,
  240,
  0xFE,
  1,103,10,
  157,
  163,
  169,
  175,
  221,
  222,
  227,
  228,
  233,
  239,
  0xFE,
  1,101,8,
  157,
  163,
  169,
  221,
  227,
  228,
  233,
  239,
  0xFE,
  1,99,11,
  157,
  163,
  169,
  175,
  191,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,97,5,
  221,
  222,
  227,
  228,
  233,
  0xFE,
  1,95,4,
  163,
  197,
  198,
  227,
  0xFE,
  0xFE,
  1,46,0,
  1,107,2,
  233,
  239,
  0xFE,
  1,105,5,
  163,
  169,
  175,
  221,
  227,
  0xFE,
  0xFE,
  1,103,9,
  169,
  175,
  191,
  197,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,101,21,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  221,
  222,
  227,
  228,
  229,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  249,
  0xFE,
  1,99,25,
  157,
  158,
  163,
  164,
  165,
  169,
  170,
  175,
  191,
  197,
  203,
  208,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  0xFE,
  1,97,26,
  157,
  158,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  181,
  186,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  249,
  250,
  0xFE,
  1,95,35,
  133,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  203,
  208,
  209,
  213,
  217,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  235,
  236,
  239,
  240,
  241,
  242,
  245,
  0xFE,
  1,93,21,
  157,
  163,
  169,
  175,
  181,
  186,
  221,
  222,
  223,
  224,
  227,
  228,
  229,
  230,
  233,
  234,
  239,
  240,
  245,
  249,
  250,
  0xFE,
  1,92,8,
  221,
  227,
  233,
  239,
  245,
  246,
  249,
  250,
  0xFE,
  1,108,0,
  1,73,2,
  191,
  197,
  0xFE,
  1,71,6,
  169,
  175,
  191,
  197,
  203,
  208,
  0xFE,
  1,69,4,
  197,
  203,
  208,
  227,
  0xFE,
  0xFE,
  0xFE,
  0,254,101,
  3,
  4,
  5,
  7,
  8,
  9,
  13,
  14,
  15,
  19,
  20,
  25,
  26,
  38,
  44,
  50,
  56,
  65,
  71,
  72,
  77,
  78,
  83,
  89,
  98,
  99,
  100,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  127,
  128,
  129,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  142,
  144,
  145,
  146,
  147,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  160,
  161,
  163,
  164,
  165,
  166,
  167,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  186,
  187,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  221,
  222,
  227,
  233,
  234,
  235,
  236,
  239,
  240,
  241,
  242,
  0,36,0,
  0,95,4,
  34,
  35,
  65,
  66,
  0xFE,
  0xFE,
  0,88,0,
  0,147,4,
  65,
  66,
  98,
  99,
  0xFE,
  0xFE,
  0,219,0,
  0,140,5,
  95,
  98,
  99,
  104,
  110,
  0xFE,
  0,142,2,
  104,
  110,
  0xFE,
  0,144,3,
  98,
  104,
  110,
  0xFE,
  0,146,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,148,1,
  98,
  0xFE,
  0xFE,
  0,217,0,
  0,86,2,
  71,
  77,
  0xFE,
  0,88,1,
  65,
  0xFE,
  0,90,1,
  65,
  0xFE,
  0,140,5,
  95,
  98,
  99,
  104,
  110,
  0xFE,
  0,142,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,144,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,146,3,
  95,
  98,
  104,
  0xFE,
  0xFE,
  0,215,0,
  0,0,1,
  3,
  0xFE,
  0,32,2,
  38,
  44,
  0xFE,
  0,34,1,
  34,
  0xFE,
  0,86,2,
  71,
  77,
  0xFE,
  0,88,4,
  62,
  65,
  71,
  77,
  0xFE,
  0,90,2,
  62,
  65,
  0xFE,
  0,140,6,
  95,
  98,
  99,
  104,
  110,
  111,
  0xFE,
  0,142,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,144,6,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0,146,4,
  95,
  98,
  104,
  110,
  0xFE,
  0xFE,
  0,213,0,
  0,0,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,2,2,
  0,
  3,
  0xFE,
  0,32,2,
  38,
  44,
  0xFE,
  0,34,2,
  31,
  34,
  0xFE,
  0,86,5,
  65,
  71,
  72,
  77,
  78,
  0xFE,
  0,88,4,
  62,
  63,
  65,
  66,
  0xFE,
  0,140,7,
  95,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,142,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,144,6,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0xFE,
  0,211,0,
  0,0,3,
  3,
  7,
  13,
  0xFE,
  0,2,1,
  3,
  0xFE,
  0,32,4,
  38,
  44,
  127,
  157,
  0xFE,
  0,34,1,
  34,
  0xFE,
  0,86,7,
  65,
  71,
  72,
  77,
  78,
  127,
  157,
  0xFE,
  0,88,4,
  62,
  63,
  65,
  66,
  0xFE,
  0,140,7,
  95,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,142,5,
  95,
  96,
  98,
  99,
  104,
  0xFE,
  0,144,2,
  104,
  110,
  0xFE,
  0xFE,
  0,209,0,
  0,86,2,
  71,
  77,
  0xFE,
  0,140,4,
  98,
  99,
  104,
  110,
  0xFE,
  0xFE,
  0,207,0,
  0,0,4,
  157,
  158,
  163,
  164,
  0xFE,
  0,2,2,
  157,
  163,
  0xFE,
  0,32,7,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  0,145,1,
  98,
  0xFE,
  0xFE,
  0,34,2,
  157,
  163,
  0xFE,
  0,86,10,
  65,
  71,
  77,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  0xFE,
  0,88,3,
  157,
  158,
  163,
  0xFE,
  0,140,10,
  95,
  96,
  98,
  99,
  104,
  110,
  157,
  158,
  163,
  164,
  1,49,2,
  104,
  110,
  0xFE,
  0xFE,
  0,142,3,
  98,
  157,
  163,
  1,49,1,
  98,
  0xFE,
  0xFE,
  0xFE,
  1,83,0,
  1,116,1,
  197,
  0xFE,
  0xFE,
  1,81,0,
  1,114,3,
  191,
  197,
  227,
  0xFE,
  0xFE,
  1,79,0,
  1,112,3,
  191,
  197,
  227,
  0xFE,
  1,114,5,
  197,
  208,
  227,
  233,
  239,
  0xFE,
  1,116,1,
  227,
  0xFE,
  0xFE,
  1,77,0,
  1,110,3,
  197,
  221,
  227,
  0xFE,
  1,112,6,
  163,
  191,
  197,
  208,
  221,
  227,
  0xFE,
  0xFE,
  1,73,0,
  1,104,4,
  191,
  197,
  221,
  227,
  0xFE,
  1,106,3,
  197,
  221,
  227,
  0xFE,
  0xFE,
  1,71,0,
  1,100,1,
  197,
  0xFE,
  1,102,5,
  197,
  208,
  227,
  233,
  239,
  0xFE,
  1,104,3,
  191,
  197,
  227,
  0xFE,
  0xFE,
  1,69,0,
  1,102,3,
  191,
  197,
  227,
  0xFE,
  0xFE,
  1,67,0,
  1,100,1,
  197,
  0xFE,
  0xFE,
  1,65,7,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,63,12,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  1,96,2,
  203,
  208,
  0xFE,
  0xFE,
  1,61,20,
  157,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  1,96,4,
  191,
  197,
  203,
  208,
  0xFE,
  0xFE,
  1,59,28,
  157,
  158,
  163,
  164,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  186,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  217,
  218,
  1,92,2,
  203,
  208,
  0xFE,
  1,96,5,
  163,
  191,
  197,
  203,
  208,
  0xFE,
  1,100,1,
  197,
  0xFE,
  0xFE,
  1,57,33,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  186,
  191,
  192,
  193,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  214,
  217,
  218,
  1,92,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,96,6,
  157,
  163,
  191,
  197,
  203,
  208,
  0xFE,
  1,106,1,
  163,
  0xFE,
  0xFE,
  1,55,34,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  186,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  214,
  217,
  218,
  1,99,2,
  203,
  208,
  0xFE,
  1,93,2,
  203,
  208,
  0xFE,
  1,92,5,
  163,
  191,
  197,
  203,
  208,
  0xFE,
  1,96,3,
  163,
  191,
  197,
  0xFE,
  1,98,5,
  157,
  163,
  191,
  197,
  203,
  0xFE,
  1,102,3,
  133,
  157,
  163,
  0xFE,
  0xFE,
  1,53,35,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  186,
  187,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  217,
  218,
  233,
  239,
  1,99,6,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,93,6,
  191,
  197,
  203,
  208,
  221,
  227,
  0xFE,
  1,92,4,
  157,
  163,
  191,
  197,
  0xFE,
  0xFE,
  1,51,34,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  182,
  186,
  187,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  217,
  218,
  233,
  239,
  1,99,10,
  163,
  169,
  175,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,97,2,
  203,
  208,
  0xFE,
  1,93,6,
  191,
  197,
  203,
  208,
  221,
  227,
  0xFE,
  0xFE,
  1,49,45,
  7,
  13,
  104,
  110,
  127,
  128,
  129,
  133,
  134,
  135,
  139,
  144,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  186,
  191,
  192,
  193,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  217,
  218,
  233,
  239,
  1,103,2,
  203,
  208,
  0xFE,
  1,99,9,
  157,
  163,
  169,
  175,
  191,
  197,
  198,
  203,
  208,
  0xFE,
  1,97,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,93,6,
  191,
  197,
  221,
  222,
  227,
  228,
  0xFE,
  0xFE,
  1,47,39,
  7,
  13,
  139,
  144,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  182,
  186,
  187,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  217,
  218,
  227,
  233,
  239,
  1,103,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,99,7,
  157,
  163,
  191,
  197,
  198,
  203,
  208,
  0xFE,
  1,97,2,
  191,
  197,
  0xFE,
  1,93,2,
  221,
  227,
  0xFE,
  0xFE,
  1,45,20,
  157,
  169,
  175,
  181,
  186,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  208,
  209,
  213,
  217,
  233,
  1,103,2,
  191,
  197,
  0xFE,
  1,99,10,
  157,
  163,
  169,
  175,
  191,
  197,
  203,
  208,
  233,
  239,
  0xFE,
  1,97,8,
  157,
  163,
  169,
  175,
  221,
  227,
  233,
  239,
  0xFE,
  1,93,7,
  163,
  169,
  175,
  221,
  227,
  233,
  239,
  0xFE,
  1,92,2,
  221,
  227,
  0xFE,
  0xFE,
  1,44,2,
  213,
  217,
  1,107,2,
  233,
  239,
  0xFE,
  1,105,7,
  163,
  169,
  175,
  221,
  227,
  233,
  239,
  0xFE,
  1,103,6,
  157,
  163,
  221,
  227,
  233,
  239,
  0xFE,
  1,101,8,
  157,
  163,
  169,
  175,
  221,
  227,
  233,
  239,
  0xFE,
  1,97,2,
  221,
  227,
  0xFE,
  0xFE,
  1,105,4,
  233,
  234,
  239,
  240,
  1,46,5,
  133,
  157,
  163,
  169,
  175,
  0xFE,
  0xFE,
  1,103,17,
  157,
  163,
  164,
  169,
  170,
  175,
  176,
  221,
  222,
  227,
  228,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  1,101,17,
  157,
  158,
  163,
  164,
  169,
  175,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  235,
  239,
  240,
  0xFE,
  1,99,20,
  157,
  163,
  169,
  170,
  175,
  176,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  249,
  0xFE,
  1,97,16,
  157,
  158,
  163,
  164,
  169,
  175,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  239,
  240,
  0xFE,
  1,95,21,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  221,
  222,
  227,
  228,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  249,
  1,61,2,
  203,
  208,
  0xFE,
  1,59,6,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,57,11,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,55,19,
  157,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  0xFE,
  1,53,22,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  0xFE,
  1,51,19,
  157,
  158,
  163,
  164,
  169,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  213,
  217,
  0xFE,
  1,49,19,
  157,
  158,
  163,
  164,
  169,
  175,
  176,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  233,
  239,
  0xFE,
  1,47,15,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  233,
  239,
  0xFE,
  1,45,15,
  157,
  163,
  169,
  175,
  191,
  197,
  198,
  203,
  208,
  213,
  217,
  221,
  227,
  233,
  239,
  0xFE,
  0xFE,
  1,93,10,
  221,
  222,
  223,
  224,
  227,
  228,
  229,
  230,
  233,
  239,
  0xFE,
  1,94,0,
  1,65,3,
  197,
  203,
  208,
  0xFE,
  1,63,8,
  169,
  175,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,61,12,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,59,13,
  157,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,57,21,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  0xFE,
  1,55,16,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  0xFE,
  1,53,18,
  157,
  158,
  163,
  164,
  169,
  175,
  176,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  213,
  217,
  0xFE,
  1,51,13,
  127,
  133,
  157,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  208,
  0xFE,
  1,49,8,
  169,
  175,
  191,
  197,
  203,
  208,
  213,
  217,
  0xFE,
  1,47,2,
  191,
  197,
  0xFE,
  0xFE,
  1,98,0,
  1,65,2,
  203,
  208,
  0xFE,
  1,63,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,61,5,
  163,
  191,
  197,
  203,
  208,
  0xFE,
  1,59,6,
  157,
  163,
  191,
  197,
  203,
  208,
  0xFE,
  1,57,7,
  163,
  169,
  175,
  191,
  197,
  203,
  208,
  0xFE,
  1,53,7,
  127,
  133,
  157,
  163,
  197,
  203,
  208,
  0xFE,
  0xFE,
  1,108,0,
  1,79,2,
  203,
  208,
  0xFE,
  1,77,5,
  191,
  203,
  208,
  233,
  239,
  0xFE,
  1,75,7,
  163,
  191,
  197,
  198,
  203,
  221,
  227,
  0xFE,
  1,73,5,
  191,
  203,
  208,
  233,
  239,
  0xFE,
  1,71,2,
  203,
  208,
  0xFE,
  0xFE,
  0xFE,
  1,0,107,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  20,
  25,
  26,
  34,
  35,
  38,
  39,
  44,
  45,
  50,
  51,
  56,
  57,
  65,
  66,
  71,
  72,
  77,
  78,
  83,
  84,
  89,
  90,
  98,
  99,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  117,
  122,
  123,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  150,
  153,
  154,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  172,
  175,
  176,
  177,
  178,
  181,
  182,
  186,
  187,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  207,
  208,
  209,
  210,
  211,
  212,
  227,
  228,
  229,
  233,
  234,
  235,
  236,
  237,
  239,
  240,
  241,
  242,
  243,
  0,0,2,
  7,
  13,
  0xFE,
  0,36,0,
  0,95,2,
  34,
  65,
  0xFE,
  0xFE,
  0,88,0,
  0,147,1,
  98,
  0xFE,
  0xFE,
  0,90,0,
  0,151,2,
  65,
  98,
  0xFE,
  0xFE,
  0,106,0,
  0,181,1,
  98,
  0xFE,
  0xFE,
  0,108,0,
  0,187,0,
  0,216,1,
  98,
  0xFE,
  0xFE,
  0xFE,
  0,140,2,
  104,
  110,
  0xFE,
  0,223,0,
  0,148,1,
  110,
  0xFE,
  0,150,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,152,2,
  95,
  98,
  0xFE,
  0xFE,
  0,221,0,
  0,140,1,
  98,
  0xFE,
  0,142,1,
  110,
  0xFE,
  0,146,4,
  95,
  98,
  104,
  110,
  0xFE,
  0,148,7,
  95,
  96,
  98,
  104,
  105,
  110,
  111,
  0xFE,
  0,150,6,
  95,
  96,
  98,
  99,
  104,
  110,
  0xFE,
  0,152,4,
  95,
  96,
  98,
  104,
  0xFE,
  0xFE,
  0,219,0,
  0,92,1,
  65,
  0xFE,
  0,94,1,
  65,
  0xFE,
  0,140,6,
  95,
  98,
  104,
  105,
  110,
  111,
  0xFE,
  0,142,3,
  98,
  104,
  110,
  0xFE,
  0,144,6,
  95,
  98,
  104,
  105,
  110,
  111,
  0xFE,
  0,146,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,148,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,150,7,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  0xFE,
  0,152,5,
  95,
  96,
  98,
  104,
  110,
  0xFE,
  0xFE,
  0,217,0,
  0,36,1,
  34,
  0xFE,
  0,86,2,
  71,
  77,
  0xFE,
  0,88,4,
  62,
  65,
  71,
  77,
  0xFE,
  0,90,7,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  0xFE,
  0,92,6,
  62,
  63,
  65,
  66,
  71,
  77,
  0xFE,
  0,94,3,
  62,
  63,
  65,
  0xFE,
  0,140,7,
  95,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,142,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,144,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,146,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,148,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,150,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,152,2,
  104,
  110,
  0xFE,
  0xFE,
  0,215,0,
  0,0,7,
  0,
  3,
  4,
  7,
  8,
  13,
  14,
  0xFE,
  0,2,5,
  0,
  3,
  4,
  7,
  13,
  0xFE,
  0,4,2,
  0,
  3,
  0xFE,
  0,6,2,
  0,
  3,
  0xFE,
  0,32,4,
  38,
  39,
  44,
  45,
  0xFE,
  0,34,8,
  31,
  32,
  34,
  35,
  38,
  39,
  44,
  45,
  0xFE,
  0,36,6,
  31,
  32,
  34,
  35,
  38,
  44,
  0xFE,
  0,38,2,
  31,
  34,
  0xFE,
  0,86,5,
  65,
  71,
  72,
  77,
  78,
  0xFE,
  0,88,8,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  0xFE,
  0,90,8,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  0xFE,
  0,92,6,
  62,
  63,
  65,
  66,
  71,
  77,
  0xFE,
  0,94,2,
  71,
  77,
  0xFE,
  0,140,10,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  0xFE,
  0,142,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,144,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,146,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,148,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,150,7,
  95,
  96,
  98,
  104,
  105,
  110,
  111,
  0xFE,
  0xFE,
  0,213,0,
  0,0,8,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  0xFE,
  0,2,8,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  0xFE,
  0,4,5,
  0,
  1,
  3,
  4,
  13,
  0xFE,
  0,6,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,32,7,
  34,
  38,
  39,
  44,
  45,
  139,
  144,
  0xFE,
  0,34,8,
  31,
  32,
  34,
  35,
  38,
  39,
  44,
  45,
  0,91,1,
  65,
  0xFE,
  0xFE,
  0,36,6,
  31,
  32,
  34,
  35,
  38,
  44,
  0xFE,
  0,38,2,
  38,
  44,
  0xFE,
  0,86,12,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  139,
  140,
  144,
  145,
  0xFE,
  0,88,8,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  0xFE,
  0,90,8,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  0xFE,
  0,92,6,
  62,
  65,
  71,
  72,
  77,
  78,
  0xFE,
  0,140,12,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  117,
  122,
  123,
  0xFE,
  0,142,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,144,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,146,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,148,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0xFE,
  0,211,0,
  0,0,11,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  25,
  139,
  144,
  0xFE,
  0,2,10,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  0xFE,
  0,4,8,
  0,
  1,
  3,
  4,
  7,
  8,
  13,
  14,
  0xFE,
  0,6,4,
  0,
  3,
  7,
  13,
  0xFE,
  0,32,13,
  34,
  35,
  38,
  39,
  44,
  45,
  127,
  128,
  133,
  139,
  140,
  144,
  145,
  0xFE,
  0,34,4,
  31,
  32,
  34,
  35,
  0xFE,
  0,36,4,
  38,
  39,
  44,
  45,
  0xFE,
  0,86,15,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  127,
  128,
  133,
  139,
  140,
  144,
  145,
  0xFE,
  0,88,7,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  0xFE,
  0,90,8,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  0xFE,
  0,140,14,
  95,
  96,
  98,
  99,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  139,
  144,
  0xFE,
  0,142,10,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  0xFE,
  0,144,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,146,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,148,2,
  104,
  110,
  0xFE,
  0xFE,
  0,209,0,
  0,0,9,
  3,
  4,
  7,
  13,
  14,
  127,
  128,
  133,
  134,
  0xFE,
  0,2,2,
  0,
  3,
  0xFE,
  0,32,11,
  31,
  32,
  34,
  35,
  38,
  44,
  127,
  128,
  133,
  134,
  139,
  0xFE,
  0,34,1,
  34,
  0xFE,
  0,86,13,
  62,
  63,
  65,
  66,
  71,
  72,
  77,
  78,
  127,
  128,
  133,
  134,
  139,
  0xFE,
  0,88,4,
  62,
  63,
  65,
  66,
  0xFE,
  0,98,0,
  0,165,1,
  98,
  0xFE,
  0xFE,
  0,140,13,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  122,
  127,
  128,
  133,
  134,
  0xFE,
  0,142,8,
  95,
  96,
  98,
  99,
  104,
  105,
  110,
  111,
  0xFE,
  0,144,2,
  104,
  110,
  0xFE,
  0xFE,
  0,207,0,
  0,0,0,
  0,93,1,
  3,
  0xFE,
  0xFE,
  0,2,0,
  0,41,1,
  3,
  0xFE,
  0xFE,
  0,46,0,
  0,111,1,
  34,
  0xFE,
  0xFE,
  0,100,0,
  0,169,1,
  98,
  0xFE,
  0xFE,
  0,140,3,
  98,
  104,
  110,
  0xFE,
  1,49,0,
  0,32,2,
  157,
  163,
  0xFE,
  0,86,2,
  157,
  163,
  0xFE,
  0xFE,
  0xFE,
  0,205,0,
  0,86,2,
  71,
  77,
  0xFE,
  0,88,0,
  0,147,1,
  98,
  0xFE,
  0xFE,
  0,140,5,
  98,
  104,
  105,
  110,
  111,
  0xFE,
  0xFE,
  1,67,2,
  203,
  208,
  0xFE,
  1,65,8,
  169,
  175,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,63,18,
  157,
  163,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  0xFE,
  1,61,22,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  0xFE,
  1,59,25,
  157,
  158,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  181,
  186,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  0xFE,
  1,57,28,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  186,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  0xFE,
  1,55,29,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  177,
  181,
  186,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  217,
  218,
  0xFE,
  1,53,30,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  186,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  217,
  218,
  0xFE,
  1,51,32,
  133,
  139,
  144,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  177,
  181,
  186,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  217,
  218,
  0xFE,
  1,49,39,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  186,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  214,
  217,
  218,
  1,97,1,
  197,
  0xFE,
  0xFE,
  1,47,32,
  127,
  128,
  133,
  134,
  135,
  144,
  157,
  163,
  169,
  170,
  175,
  176,
  181,
  186,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  217,
  0xFE,
  1,45,25,
  191,
  192,
  193,
  194,
  195,
  197,
  198,
  199,
  200,
  201,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  217,
  218,
  221,
  233,
  234,
  239,
  240,
  1,101,2,
  233,
  239,
  0xFE,
  1,97,6,
  157,
  163,
  221,
  227,
  233,
  239,
  0xFE,
  0xFE,
  1,44,6,
  203,
  204,
  208,
  209,
  213,
  217,
  1,109,2,
  233,
  239,
  0xFE,
  1,107,4,
  221,
  227,
  233,
  239,
  0xFE,
  1,105,6,
  157,
  163,
  221,
  227,
  233,
  239,
  0xFE,
  1,103,2,
  221,
  227,
  0xFE,
  1,97,4,
  221,
  222,
  227,
  228,
  0xFE,
  0xFE,
  1,107,2,
  233,
  239,
  0xFE,
  1,105,10,
  163,
  169,
  175,
  221,
  227,
  228,
  233,
  234,
  239,
  240,
  1,46,4,
  127,
  133,
  157,
  163,
  0xFE,
  0xFE,
  1,103,12,
  157,
  158,
  163,
  164,
  169,
  175,
  221,
  222,
  227,
  228,
  233,
  239,
  0xFE,
  1,101,12,
  157,
  163,
  169,
  175,
  176,
  221,
  227,
  228,
  233,
  234,
  239,
  240,
  0xFE,
  1,99,15,
  157,
  158,
  163,
  164,
  169,
  175,
  221,
  222,
  227,
  228,
  229,
  233,
  234,
  239,
  240,
  1,55,2,
  203,
  208,
  0xFE,
  1,53,7,
  163,
  169,
  175,
  191,
  197,
  203,
  208,
  0xFE,
  1,51,8,
  157,
  163,
  169,
  175,
  191,
  197,
  203,
  208,
  0xFE,
  1,49,6,
  157,
  163,
  191,
  197,
  203,
  208,
  0xFE,
  1,47,5,
  163,
  191,
  197,
  203,
  208,
  0xFE,
  1,45,4,
  221,
  227,
  233,
  239,
  0xFE,
  1,44,4,
  233,
  234,
  239,
  240,
  0xFE,
  0xFE,
  1,97,18,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  221,
  222,
  227,
  228,
  233,
  234,
  235,
  239,
  240,
  241,
  0xFE,
  1,95,27,
  127,
  128,
  133,
  134,
  169,
  175,
  191,
  192,
  193,
  197,
  198,
  203,
  208,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  235,
  236,
  239,
  240,
  241,
  242,
  1,61,2,
  203,
  208,
  0xFE,
  1,59,9,
  169,
  175,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,57,12,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,55,16,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  0xFE,
  1,53,14,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,51,15,
  157,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  0xFE,
  1,49,11,
  157,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  208,
  0xFE,
  1,47,10,
  157,
  163,
  169,
  175,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,45,8,
  191,
  197,
  203,
  204,
  208,
  209,
  221,
  227,
  0xFE,
  1,44,3,
  197,
  198,
  227,
  0xFE,
  0xFE,
  1,93,20,
  221,
  222,
  223,
  224,
  225,
  227,
  228,
  229,
  230,
  231,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  246,
  249,
  250,
  1,55,2,
  203,
  208,
  0xFE,
  1,53,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,51,2,
  191,
  197,
  0xFE,
  1,49,2,
  221,
  227,
  0xFE,
  0xFE,
  1,92,6,
  233,
  239,
  245,
  246,
  249,
  250,
  1,61,2,
  203,
  208,
  0xFE,
  1,59,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,57,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,55,3,
  163,
  191,
  197,
  0xFE,
  1,45,4,
  221,
  222,
  227,
  228,
  0xFE,
  0xFE,
  1,94,0,
  1,67,2,
  203,
  208,
  0xFE,
  1,65,6,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,63,11,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,61,12,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,59,15,
  157,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,57,14,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,55,13,
  157,
  163,
  169,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  0xFE,
  1,53,12,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  208,
  0xFE,
  1,51,8,
  157,
  163,
  169,
  175,
  191,
  197,
  203,
  208,
  0xFE,
  1,49,7,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,47,6,
  191,
  192,
  193,
  197,
  198,
  199,
  0xFE,
  0xFE,
  1,96,0,
  1,63,2,
  203,
  208,
  0xFE,
  1,61,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,59,5,
  163,
  191,
  197,
  203,
  208,
  0xFE,
  1,57,2,
  191,
  197,
  0xFE,
  0xFE,
  1,98,0,
  1,65,2,
  203,
  208,
  0xFE,
  1,63,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,61,2,
  191,
  197,
  0xFE,
  1,59,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,57,4,
  157,
  163,
  191,
  197,
  0xFE,
  1,53,1,
  157,
  0xFE,
  0xFE,
  1,108,0,
  1,81,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,79,6,
  169,
  175,
  191,
  197,
  203,
  208,
  0xFE,
  1,77,2,
  191,
  197,
  0xFE,
  0xFE,
  0xFE,
  1,2,28,
  7,
  13,
  104,
  110,
  127,
  128,
  129,
  130,
  133,
  134,
  135,
  136,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  160,
  163,
  164,
  165,
  166,
  169,
  170,
  175,
  176,
  0,144,0,
  0,221,1,
  98,
  0xFE,
  0,217,1,
  98,
  0xFE,
  0,213,0,
  1,49,2,
  95,
  98,
  0xFE,
  0xFE,
  0xFE,
  1,67,4,
  203,
  204,
  208,
  209,
  0xFE,
  1,65,13,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  1,96,2,
  203,
  208,
  0xFE,
  1,98,4,
  191,
  197,
  203,
  208,
  0xFE,
  0xFE,
  1,63,21,
  157,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  1,96,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,98,2,
  191,
  197,
  0xFE,
  0xFE,
  1,61,24,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  181,
  186,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  1,92,2,
  203,
  208,
  0xFE,
  1,96,2,
  191,
  197,
  0xFE,
  1,98,2,
  203,
  208,
  0xFE,
  0xFE,
  1,59,25,
  157,
  158,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  181,
  186,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  1,92,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,98,2,
  191,
  197,
  0xFE,
  0xFE,
  1,57,26,
  157,
  158,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  181,
  186,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  218,
  1,92,2,
  191,
  197,
  0xFE,
  0xFE,
  1,55,27,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  181,
  186,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  218,
  0xFE,
  1,53,25,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  181,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  210,
  213,
  217,
  218,
  0xFE,
  1,51,24,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  181,
  186,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  210,
  213,
  214,
  217,
  1,92,2,
  221,
  227,
  0xFE,
  0xFE,
  1,49,39,
  3,
  7,
  13,
  104,
  110,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  233,
  234,
  239,
  240,
  1,93,4,
  221,
  222,
  227,
  228,
  0xFE,
  0xFE,
  1,45,1,
  191,
  1,101,2,
  233,
  239,
  0xFE,
  0xFE,
  1,44,0,
  1,109,2,
  233,
  239,
  0xFE,
  1,107,2,
  221,
  227,
  0xFE,
  0xFE,
  1,107,8,
  169,
  175,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,105,10,
  157,
  163,
  169,
  175,
  221,
  222,
  227,
  228,
  233,
  239,
  0xFE,
  1,103,10,
  157,
  163,
  169,
  175,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,101,13,
  157,
  158,
  163,
  164,
  169,
  175,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  0xFE,
  1,99,10,
  157,
  163,
  169,
  175,
  221,
  222,
  227,
  228,
  233,
  239,
  1,57,2,
  203,
  208,
  0xFE,
  1,55,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,53,6,
  157,
  163,
  191,
  197,
  203,
  208,
  0xFE,
  1,51,6,
  157,
  163,
  191,
  197,
  203,
  208,
  0xFE,
  1,49,2,
  191,
  197,
  0xFE,
  1,45,2,
  221,
  227,
  0xFE,
  0xFE,
  1,97,2,
  169,
  175,
  0xFE,
  1,95,0,
  1,63,2,
  203,
  208,
  0xFE,
  1,61,8,
  169,
  175,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,59,12,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,57,13,
  157,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,55,13,
  157,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,53,12,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,51,10,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  208,
  0xFE,
  1,49,10,
  157,
  163,
  169,
  175,
  191,
  197,
  203,
  208,
  233,
  239,
  0xFE,
  0xFE,
  1,93,2,
  221,
  227,
  0xFE,
  1,94,0,
  1,67,2,
  203,
  208,
  0xFE,
  1,65,8,
  169,
  175,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,63,12,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,61,12,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,59,14,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,57,11,
  157,
  163,
  169,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,55,10,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  208,
  0xFE,
  1,53,6,
  169,
  175,
  191,
  197,
  203,
  208,
  0xFE,
  0xFE,
  1,108,0,
  1,83,2,
  203,
  208,
  0xFE,
  1,81,3,
  197,
  203,
  208,
  0xFE,
  1,79,2,
  191,
  197,
  0xFE,
  0xFE,
  0xFE,
  1,4,12,
  139,
  140,
  144,
  145,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  1,69,4,
  203,
  204,
  208,
  209,
  0xFE,
  1,67,11,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,65,17,
  157,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  0xFE,
  1,63,22,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  1,96,2,
  191,
  197,
  0xFE,
  1,98,1,
  208,
  0xFE,
  0xFE,
  1,61,22,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  1,98,2,
  191,
  197,
  0xFE,
  0xFE,
  1,59,24,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  181,
  186,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  0xFE,
  1,57,24,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  181,
  186,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  0xFE,
  1,55,23,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  210,
  213,
  217,
  0xFE,
  1,53,23,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  181,
  186,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  210,
  213,
  217,
  0xFE,
  1,51,16,
  157,
  163,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  0xFE,
  1,49,13,
  127,
  128,
  133,
  134,
  135,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  175,
  0xFE,
  1,109,2,
  233,
  239,
  0xFE,
  1,107,8,
  157,
  163,
  169,
  175,
  221,
  227,
  233,
  239,
  0xFE,
  1,105,9,
  157,
  163,
  169,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,103,10,
  157,
  163,
  169,
  175,
  221,
  222,
  227,
  228,
  233,
  239,
  0xFE,
  1,101,6,
  169,
  175,
  221,
  227,
  233,
  239,
  0xFE,
  1,99,0,
  1,59,2,
  203,
  208,
  0xFE,
  1,57,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,55,5,
  163,
  191,
  197,
  203,
  208,
  0xFE,
  1,53,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,51,2,
  191,
  197,
  0xFE,
  0xFE,
  1,108,0,
  1,83,2,
  203,
  208,
  0xFE,
  1,81,1,
  197,
  0xFE,
  0xFE,
  0xFE,
  1,6,0,
  0,104,0,
  0,177,0,
  0,207,1,
  98,
  0xFE,
  0xFE,
  0xFE,
  1,71,2,
  203,
  208,
  0xFE,
  1,69,10,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,67,12,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,65,21,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  210,
  213,
  217,
  0xFE,
  1,63,20,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  213,
  217,
  0xFE,
  1,61,20,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  0xFE,
  1,59,22,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  0xFE,
  1,57,21,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  210,
  213,
  217,
  0xFE,
  1,55,20,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  213,
  217,
  0xFE,
  1,53,10,
  157,
  169,
  175,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,49,5,
  139,
  144,
  163,
  169,
  175,
  0xFE,
  1,111,2,
  233,
  239,
  0xFE,
  1,109,4,
  221,
  227,
  233,
  239,
  0xFE,
  1,107,6,
  157,
  163,
  221,
  227,
  233,
  239,
  0xFE,
  1,105,7,
  157,
  163,
  169,
  175,
  221,
  227,
  239,
  0xFE,
  1,99,0,
  1,59,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,57,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,55,2,
  191,
  197,
  0xFE,
  1,53,1,
  197,
  0xFE,
  0xFE,
  0xFE,
  1,8,0,
  1,73,2,
  203,
  208,
  0xFE,
  1,71,6,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,69,12,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,67,15,
  157,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,65,16,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  0xFE,
  1,63,18,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  0xFE,
  1,61,19,
  157,
  158,
  163,
  164,
  169,
  175,
  176,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  213,
  217,
  1,99,2,
  203,
  208,
  0xFE,
  0xFE,
  1,59,20,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  213,
  217,
  1,99,2,
  191,
  197,
  0xFE,
  0xFE,
  1,57,18,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  0xFE,
  1,55,6,
  169,
  175,
  191,
  197,
  203,
  208,
  0xFE,
  1,111,2,
  233,
  239,
  0xFE,
  1,109,2,
  221,
  227,
  0xFE,
  1,107,2,
  221,
  227,
  0xFE,
  0xFE,
  1,10,0,
  1,73,5,
  191,
  197,
  203,
  204,
  208,
  0xFE,
  1,71,11,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,69,12,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,67,14,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,65,13,
  157,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,63,14,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,61,14,
  157,
  158,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,59,11,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  208,
  209,
  0xFE,
  1,57,6,
  169,
  175,
  191,
  197,
  203,
  208,
  0xFE,
  0xFE,
  1,12,0,
  1,75,2,
  203,
  208,
  0xFE,
  1,73,6,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,71,12,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,69,11,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  0xFE,
  1,67,12,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,65,12,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,63,12,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,61,10,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  208,
  0xFE,
  1,59,4,
  191,
  197,
  203,
  208,
  0xFE,
  0xFE,
  1,14,0,
  1,77,2,
  203,
  208,
  0xFE,
  1,75,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,73,9,
  163,
  169,
  175,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,71,10,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  208,
  0xFE,
  1,69,10,
  157,
  163,
  169,
  175,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,67,10,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  208,
  0xFE,
  1,65,11,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  208,
  209,
  0xFE,
  1,63,8,
  157,
  163,
  169,
  175,
  191,
  197,
  203,
  208,
  0xFE,
  0xFE,
  1,16,0,
  1,77,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,75,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,73,8,
  157,
  163,
  169,
  175,
  191,
  197,
  203,
  208,
  0xFE,
  1,71,7,
  157,
  163,
  169,
  191,
  197,
  203,
  208,
  0xFE,
  1,69,10,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  208,
  0xFE,
  1,67,6,
  157,
  163,
  175,
  191,
  197,
  208,
  0xFE,
  1,65,4,
  191,
  197,
  203,
  208,
  0xFE,
  0xFE,
  1,18,0,
  1,79,2,
  203,
  208,
  0xFE,
  1,77,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,75,5,
  163,
  191,
  197,
  203,
  208,
  0xFE,
  1,73,6,
  157,
  163,
  191,
  197,
  203,
  208,
  0xFE,
  1,71,6,
  157,
  163,
  191,
  197,
  203,
  208,
  0xFE,
  1,69,4,
  163,
  191,
  197,
  208,
  0xFE,
  1,67,2,
  191,
  197,
  0xFE,
  0xFE,
  1,20,0,
  1,79,2,
  203,
  208,
  0xFE,
  1,77,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,75,3,
  191,
  197,
  203,
  0xFE,
  1,73,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,71,2,
  191,
  197,
  0xFE,
  0xFE,
  1,22,0,
  1,79,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,77,2,
  191,
  197,
  0xFE,
  0xFE,
  1,44,81,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  34,
  35,
  38,
  44,
  50,
  65,
  66,
  71,
  77,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  133,
  134,
  135,
  139,
  140,
  144,
  149,
  153,
  157,
  163,
  164,
  165,
  166,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  182,
  186,
  187,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  215,
  217,
  218,
  219,
  221,
  222,
  227,
  228,
  229,
  233,
  234,
  239,
  240,
  245,
  246,
  249,
  250,
  1,103,4,
  203,
  208,
  233,
  239,
  0xFE,
  1,101,16,
  191,
  197,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  227,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,99,34,
  3,
  7,
  13,
  104,
  110,
  139,
  144,
  163,
  169,
  170,
  175,
  176,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  222,
  227,
  228,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  249,
  0xFE,
  1,97,33,
  139,
  144,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  217,
  218,
  221,
  222,
  223,
  227,
  228,
  233,
  234,
  235,
  239,
  240,
  245,
  246,
  249,
  250,
  0xFE,
  1,93,39,
  3,
  65,
  98,
  133,
  139,
  144,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  217,
  218,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  235,
  239,
  240,
  245,
  246,
  249,
  250,
  0xFE,
  1,92,24,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  214,
  217,
  218,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,96,10,
  197,
  203,
  208,
  213,
  217,
  227,
  233,
  239,
  245,
  249,
  0xFE,
  1,98,2,
  203,
  208,
  0xFE,
  0xFE,
  1,46,92,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  20,
  25,
  26,
  34,
  38,
  44,
  50,
  56,
  65,
  71,
  72,
  77,
  78,
  83,
  89,
  98,
  99,
  104,
  105,
  106,
  110,
  111,
  112,
  116,
  122,
  123,
  127,
  133,
  134,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  153,
  157,
  158,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  182,
  186,
  187,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  214,
  215,
  217,
  218,
  219,
  221,
  227,
  228,
  233,
  234,
  239,
  240,
  245,
  246,
  249,
  250,
  1,105,4,
  203,
  208,
  233,
  239,
  0xFE,
  1,103,16,
  191,
  197,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  227,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,101,23,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  221,
  222,
  227,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  249,
  0xFE,
  1,99,42,
  3,
  34,
  65,
  98,
  133,
  163,
  169,
  175,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  217,
  218,
  221,
  222,
  223,
  224,
  227,
  228,
  229,
  230,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  246,
  249,
  250,
  0xFE,
  1,97,33,
  163,
  169,
  175,
  191,
  192,
  193,
  197,
  198,
  199,
  200,
  203,
  204,
  208,
  209,
  213,
  214,
  217,
  218,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  235,
  239,
  240,
  245,
  246,
  249,
  250,
  0xFE,
  1,93,26,
  169,
  175,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  213,
  214,
  217,
  218,
  221,
  222,
  223,
  227,
  228,
  233,
  239,
  245,
  246,
  249,
  250,
  0xFE,
  1,92,12,
  191,
  197,
  203,
  208,
  213,
  217,
  221,
  227,
  233,
  239,
  245,
  249,
  0xFE,
  0xFE,
  1,48,69,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  25,
  38,
  44,
  71,
  77,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  127,
  133,
  134,
  139,
  140,
  144,
  145,
  149,
  153,
  157,
  158,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  182,
  186,
  187,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  206,
  208,
  209,
  210,
  211,
  213,
  214,
  217,
  218,
  227,
  233,
  234,
  239,
  240,
  245,
  249,
  0,102,0,
  0,173,1,
  98,
  0xFE,
  0xFE,
  0,220,0,
  1,31,2,
  163,
  169,
  0xFE,
  1,29,5,
  163,
  169,
  170,
  175,
  176,
  0xFE,
  1,27,16,
  133,
  134,
  139,
  144,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,25,20,
  127,
  128,
  133,
  134,
  139,
  140,
  144,
  145,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  0xFE,
  1,23,15,
  127,
  128,
  133,
  134,
  144,
  157,
  158,
  159,
  163,
  164,
  165,
  169,
  170,
  175,
  176,
  0xFE,
  1,21,12,
  127,
  133,
  139,
  144,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,19,11,
  127,
  133,
  134,
  144,
  157,
  158,
  163,
  164,
  169,
  175,
  176,
  0xFE,
  1,17,8,
  127,
  133,
  157,
  158,
  163,
  164,
  169,
  175,
  0xFE,
  1,15,6,
  127,
  133,
  157,
  163,
  169,
  175,
  0xFE,
  1,13,4,
  157,
  163,
  169,
  175,
  0xFE,
  1,11,2,
  157,
  163,
  0xFE,
  0xFE,
  0,222,0,
  1,31,4,
  169,
  170,
  175,
  176,
  0xFE,
  1,29,8,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,27,15,
  127,
  128,
  133,
  134,
  139,
  144,
  157,
  158,
  159,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,25,12,
  127,
  128,
  133,
  134,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,23,10,
  127,
  133,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,21,8,
  127,
  133,
  157,
  158,
  163,
  164,
  169,
  175,
  0xFE,
  1,19,6,
  127,
  133,
  157,
  163,
  169,
  175,
  0xFE,
  1,17,3,
  133,
  157,
  163,
  0xFE,
  1,15,1,
  169,
  0xFE,
  1,13,1,
  163,
  0xFE,
  0xFE,
  0,224,0,
  1,33,4,
  169,
  170,
  175,
  176,
  0xFE,
  1,31,8,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,29,9,
  133,
  157,
  158,
  163,
  164,
  169,
  170,
  175,
  176,
  0xFE,
  1,27,8,
  127,
  133,
  157,
  158,
  163,
  164,
  169,
  175,
  0xFE,
  1,25,4,
  157,
  163,
  169,
  175,
  0xFE,
  1,23,4,
  133,
  157,
  163,
  175,
  0xFE,
  1,21,4,
  157,
  163,
  169,
  175,
  0xFE,
  1,19,2,
  157,
  163,
  0xFE,
  0xFE,
  0,226,0,
  1,33,3,
  163,
  169,
  175,
  0xFE,
  1,31,6,
  157,
  158,
  163,
  164,
  169,
  175,
  0xFE,
  1,29,5,
  157,
  163,
  164,
  169,
  175,
  0xFE,
  1,27,4,
  157,
  163,
  169,
  175,
  0xFE,
  1,25,2,
  157,
  163,
  0xFE,
  0xFE,
  0,228,0,
  1,33,2,
  157,
  163,
  0xFE,
  1,31,1,
  163,
  0xFE,
  0xFE,
  1,107,5,
  203,
  208,
  221,
  233,
  239,
  0xFE,
  1,105,18,
  191,
  192,
  197,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  222,
  227,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,103,28,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  249,
  0xFE,
  1,101,32,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  217,
  218,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  246,
  249,
  250,
  0xFE,
  1,99,34,
  3,
  98,
  133,
  163,
  169,
  175,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  213,
  214,
  217,
  218,
  221,
  222,
  223,
  227,
  228,
  229,
  233,
  234,
  239,
  240,
  245,
  246,
  249,
  250,
  0xFE,
  1,97,27,
  163,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  213,
  214,
  217,
  218,
  221,
  222,
  227,
  228,
  229,
  233,
  234,
  239,
  240,
  245,
  246,
  249,
  250,
  0xFE,
  1,95,45,
  3,
  7,
  13,
  19,
  25,
  104,
  110,
  133,
  139,
  144,
  157,
  163,
  164,
  169,
  170,
  175,
  176,
  181,
  186,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  245,
  246,
  249,
  250,
  0xFE,
  1,93,24,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  214,
  217,
  218,
  221,
  222,
  223,
  227,
  228,
  233,
  239,
  240,
  245,
  246,
  249,
  250,
  0xFE,
  1,94,5,
  191,
  203,
  204,
  208,
  209,
  0xFE,
  0xFE,
  1,50,30,
  3,
  7,
  13,
  104,
  110,
  139,
  144,
  163,
  169,
  170,
  175,
  176,
  181,
  186,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  217,
  218,
  233,
  239,
  1,111,4,
  203,
  208,
  233,
  239,
  0xFE,
  1,109,12,
  191,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  233,
  234,
  239,
  240,
  0xFE,
  1,107,21,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  222,
  227,
  228,
  229,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,105,24,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  210,
  213,
  217,
  221,
  222,
  227,
  228,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  249,
  0xFE,
  1,103,29,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  213,
  214,
  217,
  218,
  221,
  222,
  227,
  228,
  229,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  246,
  249,
  250,
  0xFE,
  1,101,26,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  213,
  217,
  218,
  221,
  222,
  227,
  228,
  229,
  233,
  234,
  239,
  240,
  245,
  246,
  249,
  250,
  0xFE,
  1,99,26,
  163,
  169,
  175,
  191,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  213,
  214,
  217,
  218,
  221,
  227,
  228,
  233,
  234,
  239,
  240,
  245,
  246,
  249,
  250,
  0xFE,
  1,97,19,
  191,
  192,
  197,
  198,
  203,
  208,
  209,
  213,
  214,
  217,
  218,
  221,
  222,
  227,
  228,
  233,
  239,
  245,
  249,
  0xFE,
  1,93,12,
  191,
  197,
  203,
  208,
  213,
  217,
  221,
  227,
  233,
  239,
  245,
  249,
  0xFE,
  0xFE,
  1,52,10,
  169,
  175,
  191,
  197,
  203,
  204,
  208,
  209,
  213,
  217,
  1,113,4,
  203,
  208,
  233,
  239,
  0xFE,
  1,111,20,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,109,20,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,107,20,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,105,22,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  213,
  217,
  218,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,103,21,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,101,21,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  214,
  217,
  218,
  221,
  227,
  228,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,99,17,
  191,
  192,
  197,
  198,
  203,
  208,
  209,
  213,
  217,
  221,
  222,
  227,
  228,
  233,
  239,
  245,
  249,
  0xFE,
  1,97,18,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  227,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,93,2,
  221,
  227,
  0xFE,
  0xFE,
  1,54,4,
  191,
  197,
  203,
  208,
  1,115,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,113,15,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,111,19,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,109,20,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,107,20,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,105,20,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,103,17,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  227,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,101,16,
  191,
  192,
  197,
  198,
  203,
  208,
  213,
  217,
  221,
  222,
  227,
  228,
  233,
  239,
  245,
  249,
  0xFE,
  1,99,16,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  227,
  233,
  239,
  245,
  249,
  0xFE,
  1,97,5,
  191,
  203,
  208,
  233,
  239,
  0xFE,
  0xFE,
  1,56,0,
  1,117,6,
  197,
  203,
  208,
  227,
  233,
  239,
  0xFE,
  1,115,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,113,18,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  227,
  228,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,111,20,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,109,20,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,107,20,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,105,16,
  191,
  197,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  227,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,103,16,
  191,
  192,
  197,
  198,
  203,
  208,
  213,
  217,
  221,
  222,
  227,
  228,
  233,
  239,
  245,
  249,
  0xFE,
  1,101,16,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  227,
  233,
  239,
  245,
  249,
  0xFE,
  1,99,7,
  191,
  203,
  208,
  213,
  217,
  233,
  239,
  0xFE,
  0xFE,
  1,58,0,
  1,119,4,
  203,
  208,
  233,
  239,
  0xFE,
  1,117,7,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,115,14,
  191,
  197,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,113,20,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,111,20,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,109,19,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  222,
  227,
  228,
  233,
  239,
  240,
  245,
  249,
  0xFE,
  1,107,16,
  191,
  197,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  227,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,105,16,
  191,
  192,
  197,
  198,
  203,
  208,
  213,
  217,
  221,
  222,
  227,
  228,
  233,
  239,
  245,
  249,
  0xFE,
  1,103,16,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  227,
  233,
  239,
  245,
  249,
  0xFE,
  1,101,7,
  191,
  197,
  203,
  208,
  227,
  233,
  239,
  0xFE,
  0xFE,
  1,60,0,
  1,119,6,
  197,
  203,
  208,
  227,
  233,
  239,
  0xFE,
  1,117,12,
  191,
  197,
  203,
  204,
  208,
  209,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,115,20,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,113,20,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,111,18,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  222,
  227,
  228,
  233,
  239,
  245,
  249,
  0xFE,
  1,109,15,
  191,
  197,
  203,
  204,
  208,
  213,
  217,
  221,
  227,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,107,16,
  191,
  192,
  197,
  198,
  203,
  208,
  213,
  217,
  221,
  222,
  227,
  228,
  233,
  239,
  245,
  249,
  0xFE,
  1,105,14,
  191,
  197,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  227,
  233,
  239,
  245,
  249,
  0xFE,
  1,103,7,
  191,
  197,
  203,
  208,
  227,
  233,
  239,
  0xFE,
  0xFE,
  1,62,0,
  1,121,4,
  203,
  208,
  233,
  239,
  0xFE,
  1,119,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,117,17,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  227,
  228,
  233,
  234,
  239,
  240,
  245,
  0xFE,
  1,115,20,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,113,17,
  191,
  192,
  197,
  198,
  203,
  208,
  209,
  213,
  217,
  221,
  222,
  227,
  228,
  233,
  239,
  245,
  249,
  0xFE,
  1,111,14,
  191,
  197,
  203,
  204,
  208,
  213,
  217,
  221,
  227,
  233,
  234,
  239,
  245,
  249,
  0xFE,
  1,109,16,
  191,
  192,
  197,
  198,
  203,
  208,
  213,
  217,
  221,
  222,
  227,
  228,
  233,
  239,
  245,
  249,
  0xFE,
  1,107,12,
  191,
  197,
  203,
  208,
  213,
  217,
  221,
  227,
  233,
  239,
  245,
  249,
  0xFE,
  1,105,7,
  191,
  197,
  203,
  208,
  227,
  233,
  239,
  0xFE,
  0xFE,
  1,64,0,
  1,123,4,
  203,
  208,
  233,
  239,
  0xFE,
  1,121,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,119,14,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  221,
  227,
  233,
  234,
  239,
  240,
  0xFE,
  1,117,19,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  245,
  0xFE,
  1,115,16,
  191,
  192,
  197,
  198,
  203,
  208,
  213,
  217,
  221,
  222,
  227,
  228,
  233,
  239,
  245,
  249,
  0xFE,
  1,113,14,
  191,
  197,
  203,
  204,
  208,
  213,
  217,
  221,
  227,
  233,
  234,
  239,
  245,
  249,
  0xFE,
  1,111,15,
  191,
  192,
  197,
  198,
  203,
  208,
  213,
  217,
  221,
  222,
  227,
  228,
  233,
  239,
  249,
  0xFE,
  1,109,12,
  191,
  197,
  203,
  208,
  213,
  217,
  221,
  227,
  233,
  239,
  245,
  249,
  0xFE,
  1,107,6,
  191,
  197,
  203,
  208,
  233,
  239,
  0xFE,
  0xFE,
  1,66,0,
  1,125,4,
  203,
  208,
  233,
  239,
  0xFE,
  1,123,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,121,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,119,17,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  221,
  222,
  227,
  228,
  233,
  234,
  239,
  240,
  0xFE,
  1,117,13,
  191,
  192,
  197,
  198,
  203,
  208,
  217,
  221,
  222,
  227,
  228,
  233,
  239,
  0xFE,
  1,115,13,
  191,
  197,
  203,
  204,
  208,
  213,
  217,
  221,
  227,
  233,
  239,
  245,
  249,
  0xFE,
  1,113,11,
  191,
  192,
  197,
  198,
  203,
  208,
  217,
  221,
  227,
  233,
  239,
  0xFE,
  1,111,11,
  191,
  197,
  203,
  208,
  213,
  217,
  221,
  227,
  233,
  239,
  245,
  0xFE,
  1,109,5,
  191,
  203,
  208,
  233,
  239,
  0xFE,
  0xFE,
  1,68,0,
  1,127,1,
  233,
  0xFE,
  1,125,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,123,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,121,9,
  191,
  197,
  198,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,119,12,
  191,
  192,
  197,
  198,
  203,
  208,
  221,
  222,
  227,
  228,
  233,
  239,
  0xFE,
  1,117,9,
  191,
  197,
  203,
  208,
  217,
  221,
  227,
  233,
  239,
  0xFE,
  1,115,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,113,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,111,3,
  191,
  203,
  208,
  0xFE,
  0xFE,
  1,70,0,
  1,127,6,
  197,
  203,
  208,
  227,
  233,
  239,
  0xFE,
  1,125,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,123,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,121,9,
  191,
  197,
  198,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,119,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,117,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,115,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,113,2,
  203,
  208,
  0xFE,
  0xFE,
  1,72,0,
  1,129,4,
  203,
  208,
  233,
  239,
  0xFE,
  1,127,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,125,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,123,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,121,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,119,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,117,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,115,2,
  203,
  208,
  0xFE,
  0xFE,
  1,74,0,
  1,131,4,
  203,
  208,
  233,
  239,
  0xFE,
  1,129,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,127,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,125,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,123,7,
  191,
  197,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,121,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,119,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  0xFE,
  1,76,0,
  1,133,2,
  233,
  239,
  0xFE,
  1,131,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,129,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,127,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,125,7,
  191,
  197,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,123,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,121,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  0xFE,
  1,78,0,
  1,133,7,
  191,
  197,
  203,
  208,
  227,
  233,
  239,
  0xFE,
  1,131,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,129,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,127,7,
  191,
  197,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,125,7,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  0xFE,
  1,123,5,
  191,
  197,
  203,
  208,
  227,
  0xFE,
  0xFE,
  1,80,0,
  1,135,4,
  203,
  208,
  233,
  239,
  0xFE,
  1,133,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,131,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,129,7,
  191,
  197,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,127,7,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  0xFE,
  1,125,1,
  203,
  0xFE,
  0xFE,
  1,82,0,
  1,137,2,
  233,
  239,
  0xFE,
  1,135,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,133,8,
  191,
  197,
  203,
  208,
  221,
  227,
  233,
  239,
  0xFE,
  1,131,6,
  191,
  197,
  208,
  221,
  227,
  239,
  0xFE,
  1,129,5,
  191,
  197,
  203,
  221,
  227,
  0xFE,
  0xFE,
  1,84,0,
  1,137,6,
  191,
  197,
  203,
  227,
  233,
  239,
  0xFE,
  1,135,7,
  191,
  197,
  203,
  221,
  227,
  233,
  239,
  0xFE,
  1,133,6,
  191,
  197,
  208,
  221,
  227,
  239,
  0xFE,
  1,131,3,
  191,
  197,
  227,
  0xFE,
  0xFE,
  1,86,0,
  1,137,4,
  191,
  197,
  221,
  227,
  0xFE,
  1,135,3,
  191,
  197,
  227,
  0xFE,
  0xFE,
  1,95,0,
  0,90,2,
  233,
  239,
  0xFE,
  1,4,0,
  1,65,2,
  203,
  208,
  0xFE,
  1,63,6,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,61,12,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,59,12,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,57,12,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,55,12,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,53,10,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  208,
  0xFE,
  1,51,6,
  169,
  175,
  191,
  197,
  203,
  208,
  0xFE,
  0xFE,
  1,6,0,
  1,65,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,63,8,
  169,
  175,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,61,12,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,59,11,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  208,
  209,
  0xFE,
  1,57,11,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  208,
  209,
  0xFE,
  1,55,9,
  157,
  163,
  169,
  175,
  191,
  197,
  198,
  203,
  208,
  0xFE,
  1,53,6,
  169,
  175,
  191,
  197,
  203,
  208,
  0xFE,
  0xFE,
  1,8,0,
  1,67,2,
  203,
  208,
  0xFE,
  1,65,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,63,10,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  208,
  0xFE,
  1,61,8,
  157,
  163,
  169,
  175,
  191,
  197,
  203,
  208,
  0xFE,
  1,59,8,
  157,
  163,
  169,
  175,
  191,
  197,
  203,
  208,
  0xFE,
  1,57,8,
  157,
  163,
  169,
  175,
  191,
  197,
  203,
  208,
  0xFE,
  1,55,4,
  191,
  197,
  203,
  208,
  0xFE,
  0xFE,
  1,10,0,
  1,69,2,
  203,
  208,
  0xFE,
  1,67,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,65,5,
  163,
  191,
  197,
  203,
  208,
  0xFE,
  1,63,6,
  157,
  163,
  191,
  197,
  203,
  208,
  0xFE,
  1,61,6,
  157,
  163,
  191,
  197,
  203,
  208,
  0xFE,
  1,59,6,
  157,
  163,
  191,
  197,
  203,
  208,
  0xFE,
  0xFE,
  1,12,0,
  1,69,3,
  191,
  203,
  208,
  0xFE,
  1,67,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,65,6,
  157,
  163,
  191,
  197,
  203,
  208,
  0xFE,
  1,63,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,61,3,
  191,
  197,
  208,
  0xFE,
  0xFE,
  1,14,0,
  1,71,2,
  203,
  208,
  0xFE,
  1,69,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,67,3,
  191,
  197,
  203,
  0xFE,
  1,65,3,
  191,
  197,
  208,
  0xFE,
  1,63,2,
  191,
  197,
  0xFE,
  0xFE,
  1,16,0,
  1,71,1,
  203,
  0xFE,
  1,69,2,
  191,
  197,
  0xFE,
  0xFE,
  1,44,58,
  3,
  7,
  34,
  38,
  44,
  65,
  71,
  77,
  98,
  99,
  104,
  110,
  133,
  134,
  139,
  144,
  157,
  163,
  164,
  169,
  175,
  181,
  186,
  191,
  192,
  193,
  194,
  197,
  198,
  199,
  200,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  217,
  218,
  221,
  222,
  223,
  224,
  227,
  228,
  229,
  230,
  233,
  234,
  239,
  240,
  241,
  245,
  246,
  249,
  250,
  0xFE,
  1,46,47,
  3,
  7,
  34,
  65,
  98,
  104,
  133,
  139,
  144,
  163,
  164,
  169,
  170,
  175,
  176,
  181,
  186,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  217,
  218,
  221,
  222,
  227,
  228,
  229,
  233,
  234,
  239,
  240,
  241,
  245,
  246,
  249,
  250,
  0xFE,
  1,50,28,
  7,
  13,
  104,
  110,
  139,
  144,
  163,
  169,
  175,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  214,
  217,
  218,
  221,
  227,
  228,
  233,
  239,
  240,
  245,
  249,
  0xFE,
  1,52,10,
  169,
  175,
  191,
  197,
  203,
  204,
  208,
  209,
  233,
  239,
  0xFE,
  1,54,2,
  203,
  208,
  0xFE,
  0xFE,
  1,92,62,
  3,
  4,
  7,
  13,
  19,
  25,
  34,
  38,
  44,
  65,
  71,
  77,
  98,
  99,
  104,
  110,
  133,
  134,
  139,
  144,
  157,
  163,
  164,
  165,
  169,
  170,
  175,
  181,
  186,
  191,
  192,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  217,
  218,
  221,
  222,
  223,
  224,
  227,
  228,
  229,
  230,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  246,
  249,
  250,
  0xFE,
  1,94,91,
  3,
  4,
  7,
  8,
  13,
  14,
  19,
  20,
  25,
  26,
  34,
  38,
  44,
  50,
  56,
  65,
  71,
  72,
  77,
  78,
  83,
  89,
  98,
  99,
  104,
  105,
  110,
  111,
  116,
  122,
  133,
  134,
  139,
  140,
  141,
  144,
  145,
  146,
  149,
  153,
  157,
  163,
  164,
  165,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  182,
  186,
  187,
  191,
  192,
  193,
  197,
  198,
  199,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  214,
  217,
  218,
  221,
  222,
  223,
  224,
  227,
  228,
  229,
  233,
  234,
  235,
  236,
  239,
  240,
  241,
  242,
  245,
  246,
  247,
  249,
  250,
  251,
  1,4,0,
  1,69,2,
  203,
  208,
  0xFE,
  1,67,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,65,10,
  157,
  163,
  169,
  175,
  191,
  197,
  203,
  204,
  208,
  209,
  0xFE,
  1,63,12,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,61,13,
  157,
  163,
  164,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,59,11,
  157,
  163,
  169,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,57,10,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  208,
  0xFE,
  1,55,6,
  169,
  175,
  191,
  197,
  203,
  208,
  0xFE,
  0xFE,
  1,6,0,
  1,71,2,
  203,
  208,
  0xFE,
  1,69,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,67,5,
  191,
  197,
  203,
  208,
  209,
  0xFE,
  1,65,12,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,63,11,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  0xFE,
  1,61,11,
  157,
  163,
  169,
  191,
  192,
  197,
  198,
  203,
  204,
  208,
  209,
  0xFE,
  1,59,10,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  208,
  0xFE,
  1,57,4,
  191,
  197,
  203,
  208,
  0xFE,
  0xFE,
  1,8,0,
  1,71,2,
  203,
  208,
  0xFE,
  1,69,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,67,6,
  163,
  175,
  191,
  197,
  203,
  208,
  0xFE,
  1,65,10,
  157,
  163,
  169,
  175,
  191,
  192,
  197,
  198,
  203,
  208,
  0xFE,
  1,63,6,
  157,
  163,
  191,
  197,
  203,
  208,
  0xFE,
  1,61,8,
  157,
  163,
  169,
  175,
  191,
  197,
  203,
  208,
  0xFE,
  1,59,3,
  191,
  197,
  208,
  0xFE,
  0xFE,
  1,10,0,
  1,71,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,69,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,67,6,
  157,
  163,
  191,
  197,
  203,
  208,
  0xFE,
  1,65,6,
  157,
  163,
  191,
  197,
  203,
  208,
  0xFE,
  1,63,5,
  163,
  191,
  197,
  203,
  208,
  0xFE,
  0xFE,
  1,12,0,
  1,73,2,
  203,
  208,
  0xFE,
  1,71,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,69,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,67,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,65,3,
  191,
  197,
  208,
  0xFE,
  0xFE,
  1,14,0,
  1,73,2,
  203,
  208,
  0xFE,
  1,71,4,
  191,
  197,
  203,
  208,
  0xFE,
  1,69,3,
  191,
  197,
  203,
  0xFE,
  1,67,2,
  191,
  197,
  0xFE,
  0xFE,
  1,44,36,
  3,
  7,
  13,
  19,
  25,
  98,
  104,
  110,
  133,
  139,
  144,
  163,
  169,
  170,
  175,
  176,
  181,
  186,
  191,
  192,
  197,
  198,
  203,
  204,
  205,
  208,
  209,
  210,
  213,
  217,
  221,
  227,
  233,
  239,
  245,
  249,
  0xFE,
  1,46,20,
  7,
  13,
  104,
  110,
  139,
  144,
  169,
  170,
  175,
  176,
  191,
  197,
  203,
  204,
  208,
  209,
  213,
  217,
  233,
  239,
  0xFE,
  1,50,1,
  203,
  0xFE,
  0xFE,
  1,96,37,
  3,
  7,
  13,
  104,
  110,
  139,
  144,
  163,
  169,
  170,
  171,
  175,
  176,
  177,
  181,
  186,
  191,
  197,
  198,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  222,
  227,
  228,
  233,
  234,
  235,
  239,
  240,
  241,
  245,
  249,
  0xFE,
  1,98,18,
  169,
  175,
  191,
  197,
  203,
  204,
  208,
  209,
  213,
  217,
  221,
  227,
  233,
  234,
  239,
  240,
  245,
  249,
  0xFE,
  1,100,2,
  233,
  239,
  0xFE,
  1,140,2,
  62,
  65,
  1,147,2,
  62,
  65,
  0xFE,
  1,145,2,
  62,
  65,
  0xFE,
  1,144,1,
  65,
  0xFE,
  0xFE,
  1,144,2,
  62,
  65,
  0xFE,
  0xFF
};

static
const int vsop87_index_translation_table[8*6*6] = {
    0,
    1,
    2,
   -1,
   -1,
   -1,
    3,
    4,
    5,
    6,
   -1,
   -1,
    7,
    8,
    9,
   10,
   11,
   12,
   13,
   14,
   15,
   16,
   17,
   18,
   19,
   20,
   21,
   22,
   23,
   24,
   25,
   26,
   27,
   28,
   29,
   30,
   31,
   32,
   33,
   -1,
   -1,
   -1,
   34,
   35,
   36,
   37,
   -1,
   -1,
   38,
   39,
   40,
   41,
   42,
   43,
   44,
   45,
   46,
   47,
   48,
   49,
   50,
   51,
   52,
   53,
   54,
   55,
   56,
   57,
   58,
   59,
   60,
   61,
   62,
   63,
   64,
   -1,
   -1,
   -1,
   65,
   66,
   67,
   68,
   69,
   70,
   71,
   72,
   73,
   74,
   75,
   76,
   77,
   78,
   79,
   80,
   81,
   82,
   83,
   84,
   85,
   86,
   87,
   88,
   89,
   90,
   91,
   92,
   93,
   94,
   95,
   96,
   97,
   -1,
   -1,
   -1,
   98,
   99,
  100,
  101,
  102,
  103,
  104,
  105,
  106,
  107,
  108,
  109,
  110,
  111,
  112,
  113,
  114,
  115,
  116,
  117,
  118,
  119,
  120,
  121,
  122,
  123,
  124,
  125,
  126,
   -1,
  127,
  128,
  129,
  130,
  131,
  132,
  133,
  134,
  135,
  136,
  137,
  138,
  139,
  140,
  141,
  142,
  143,
   -1,
  144,
  145,
  146,
  147,
  148,
   -1,
  149,
  150,
  151,
  152,
   -1,
   -1,
  153,
  154,
  155,
  156,
   -1,
   -1,
  157,
  158,
  159,
  160,
  161,
  162,
  163,
  164,
  165,
  166,
  167,
  168,
  169,
  170,
  171,
  172,
  173,
  174,
  175,
  176,
  177,
  178,
  179,
  180,
  181,
  182,
  183,
  184,
  185,
   -1,
  186,
  187,
  188,
  189,
  190,
   -1,
  191,
  192,
  193,
  194,
  195,
  196,
  197,
  198,
  199,
  200,
  201,
  202,
  203,
  204,
  205,
  206,
  207,
   -1,
  208,
  209,
  210,
  211,
  212,
   -1,
  213,
  214,
  215,
  216,
   -1,
   -1,
  217,
  218,
  219,
  220,
   -1,
   -1,
  221,
  222,
  223,
  224,
  225,
  226,
  227,
  228,
  229,
  230,
  231,
  232,
  233,
  234,
  235,
  236,
  237,
  238,
  239,
  240,
  241,
  242,
  243,
  244,
  245,
  246,
  247,
  248,
   -1,
   -1,
  249,
  250,
  251,
  252,
   -1,
   -1,
};


/* Original VSOP87 values come from IAU 1976 constants.
   Thanks to Jean-Louis Simon for this information. */
#define GAUSS_GRAV_CONST (0.01720209895*0.01720209895)

static
const double vsop87_mu[8] = {
  (1.0+1.0/6023600)*GAUSS_GRAV_CONST,
  (1.0+1.0/408523.5)*GAUSS_GRAV_CONST,
  (1.0+1.0/328900.5)*GAUSS_GRAV_CONST,
  (1.0+1.0/3098710)*GAUSS_GRAV_CONST,
  (1.0+1.0/1047.355)*GAUSS_GRAV_CONST,
  (1.0+1.0/3498.5)*GAUSS_GRAV_CONST,
  (1.0+1.0/22869)*GAUSS_GRAV_CONST,
  (1.0+1.0/19314)*GAUSS_GRAV_CONST
};

static
void PrepareLambdaArray(int nr_of_lambdas,
                        const int max_lambda_factor[],
                        const double lambda[],
                        double cos_sin_lambda[]) {
    /* initialize cos_sin_lambda:
       (cos,sin)(1*lambda[0]),(cos,sin)(-1*lambda[0]),(cos,sin)(2*lambda[0]),...
       (cos,sin)(1*lambda[1]),(cos,sin)(-1*lambda[1]),(cos,sin)(2*lambda[1]),...
    */
  int i;
  for (i=0;i<nr_of_lambdas;i++) {
    const int max_factor = max_lambda_factor[i];
    double *cslp = cos_sin_lambda;
    cslp[0] = cos(lambda[i]);
    cslp[1] = sin(lambda[i]);
    cslp[2] =  cslp[0];
    cslp[3] = -cslp[1];
    int m;
    for (m=2;m<=max_factor;m++) {
      cslp += 4;
        /* addition theorem:
           cos(m*l) = cos(m0*l+m1*l) = cos(m0*l)*cos(m1*l)-sin(m0*l)*sin(m1*l)
           sin(m*l) = sin(m0*l+m1*l) = cos(m0*l)*sin(m1*l)+sin(m0*l)*cos(m1*l)
        */
      const double *m0 = cos_sin_lambda + ((((m+0)>>1)-1)<<2);
      const double *m1 = cos_sin_lambda + ((((m+1)>>1)-1)<<2);
      cslp[0] = m0[0] * m1[0] - m0[1] * m1[1];
      cslp[1] = m0[0] * m1[1] + m0[1] * m1[0];
      cslp[2] =  cslp[0];
      cslp[3] = -cslp[1];
    }
    cos_sin_lambda += (max_factor<<2);
  }
}

static
void AccumulateVsop87Terms(const unsigned char *instructions,
                           const double *coefficients,
                           const double cos_sin_lambda[],
                           double accu[],
                           double *sp) {
    /* Accumulates the series given in instructions/coefficients.
       The argument of the series is cos_sin_lambda which has
       been initialized using PrepareLambdaArray().
       accu is the output accumulator.
       sp must point to a memory area holding 2*nr_of_lambdas double.
       area must be supplied by the caller, it will be destroyed during the
       calculation.
    */
  sp[0] = 1.0;
  sp[1] = 0.0;
  for (;;) {
    int lambda_index = *instructions++;
    if (lambda_index < 0xFE) {
      lambda_index = (lambda_index<<8)|(*instructions++);
      int term_count = (*instructions++);
        /* calculate new argument and push it on the stack */
      const double *const cos_sin = cos_sin_lambda + (lambda_index<<1);
      sp[2] = cos_sin[0]*sp[0] - cos_sin[1]*sp[1];
      sp[3] = cos_sin[0]*sp[1] + cos_sin[1]*sp[0];
      sp += 2;
      while (--term_count >= 0) {
        accu[*instructions++] += (coefficients[0]*sp[0]
                                + coefficients[1]*sp[1]);
        coefficients += 2;
      }
    } else {
      if (lambda_index == 0xFF) break;
        /* pop argument from the stack */
      sp -= 2;
    }
  }
}

static
void CalcVsop87Elem(double t,double elem[8*6]) {
  double lambda[12];
  int i;
  for (i=0;i<12;i++) lambda[i] = lambda_0[i] + lambda_1[i] * t;
  double cos_sin_lambda[203*4];
  PrepareLambdaArray(12,vsop87_max_lambda_factor,lambda,cos_sin_lambda);
  double accu[sizeof(vsop87_constants)/sizeof(vsop87_constants[0])];
  for (i=0;i<(sizeof(vsop87_constants)/sizeof(vsop87_constants[0]));i++) {
    accu[i] = 0.0;
  }
  double stack[12*2];
  AccumulateVsop87Terms(vsop87_instructions,vsop87_coefficients,cos_sin_lambda,
                        accu,stack);

  for (i=0;i<8*6;i++) {
    elem[i] = 0.0;
  }
    /* terms of order t^alpha: */
  double use_polynomials = (6.1 - fabs(t)) / 0.1;
  if (use_polynomials > 0) {
    if (use_polynomials > 1.0) use_polynomials = 1.0;
    for (i=0;i<8*6;i++) {
      int alpha;
      for (alpha=5;alpha>0;alpha--) {
        const int j = vsop87_index_translation_table[i*6+alpha];
        if (j >= 0) {
          elem[i] += accu[j] + vsop87_constants[j];
          elem[i] *= t;
        }
      }
      elem[i] *= use_polynomials;
    }
  }
    /* terms of order t^0: */
  for (i=0;i<8*6;i++) {
    const int j = vsop87_index_translation_table[i*6];
    elem[i] += accu[j] + vsop87_constants[j];
  }
    /* longitudes: */
  for (i=0;i<8;i++) {
    elem[i*6+1] = fmod(elem[i*6+1]+t*vsop87_l[i],2*M_PI);
    if (elem[i*6+1] < 0.0) elem[i*6+1] += 2*M_PI;
  }

/*
  for (i=0;i<8;i++) {
    printf("%f %d %15.10f %15.10f %15.10f %15.10f %15.10f %15.10f\n",
           2451545.0+t*365250.0,i,
           elem[i*6+0],elem[i*6+1],elem[i*6+2],
           elem[i*6+3],elem[i*6+3],elem[i*6+4]);
  }
*/
}

  /* dirty caching in static variables */
static double vsop87_t0 = -1e99;
static double vsop87_elem[8*6];

/* 10 days: */
#define ELEM_VALIDITY 10

#define VAL_FACTOR (365250.0/ELEM_VALIDITY)

void GetVsop87Coor(double jd,int body,double *xyz) {
  const double t = (jd - 2451545.0) / 365250.0;
  if (fabs(t-vsop87_t0) > 0.5/VAL_FACTOR) {
    vsop87_t0 = t;
    CalcVsop87Elem(vsop87_t0,vsop87_elem);
  }
  EllipticToRectangularA(vsop87_mu[body],vsop87_elem+body*6,
                         365250.0*(t-vsop87_t0),xyz);
}

void GetVsop87OsculatingCoor(double jd0,double jd,int body,double *xyz) {
  const double t0 = (jd0 - 2451545.0) / 365250.0;
  if (fabs(t0-vsop87_t0) > 0.5/VAL_FACTOR) {
    vsop87_t0 = t0;
    CalcVsop87Elem(vsop87_t0,vsop87_elem);
  }
  const double t = (jd - 2451545.0) / 365250.0;
  EllipticToRectangularA(vsop87_mu[body],vsop87_elem+body*6,
                         365250.0*(t-vsop87_t0),xyz);
}
